Package: BeeBDC
Title: Occurrence Data Cleaning
Version: 1.3.4
Authors@R: c(
    person("James B.", "Dorey", , "jbdorey@me.com", role = c("aut", "cre", "cph"),
           comment = c(ORCID = "0000-0003-2721-3842")),
    person("Robert L.", "O'Reilly", , "robert.oreilly@flinders.edu.au", role = "aut",
           comment = c(ORCID = "0000-0001-5291-7396")),
    person("Silas", "Bossert", , "silas.bossert@wsu.edu", role = "aut",
           comment = c(ORCID = "0000-0002-3620-5468")),
    person("Erica E.", "Fischer", , "fischeer@mail.gvsu.edu", role = "aut",
           comment = c(ORCID = "0000-0002-8202-158X"))
  )
Description: Flags and checks occurrence data that are in Darwin Core
    format.  The package includes generic functions and data as well as
    some that are specific to bees. This package is meant to build upon
    and be complimentary to other excellent occurrence cleaning packages,
    including 'bdc' and 'CoordinateCleaner'.  This package uses datasets
    from several sources and particularly from the Discover Life Website,
    created by Ascher and Pickering (2020).  For further information,
    please see the original publication and package website.  Publication
    - Dorey et al. (2023) <doi:10.1101/2023.06.30.547152> and package
    website - Dorey et al. (2023) <https://github.com/jbdorey/BeeBDC>.
License: GPL (>= 3)
URL: https://jbdorey.github.io/BeeBDC/
        https://github.com/jbdorey/BeeBDC
BugReports: https://github.com/jbdorey/BeeBDC/issues
Depends: R (>= 4.1.0)
Imports: circlize, CoordinateCleaner, cowplot, dplyr, forcats, ggplot2,
        ggspatial, here, igraph, lubridate, mgsub, openxlsx, paletteer,
        readr, rnaturalearth, sf, stringr, tidyselect
Suggests: bdc, BiocManager, classInt, ComplexHeatmap, countrycode,
        curl, devtools, emld, formatR, galah, hexbin, htmltools,
        htmlwidgets, httr, iNEXT, janitor, kableExtra, knitr, leaflet,
        magrittr, mosaic, pkgdown, plotly, prettydoc, purrr, R.utils,
        renv, rgnparser, rlang, rmarkdown, rmdformats,
        rnaturalearthdata, rvest, SpadeR, taxadb, terra, testthat,
        tibble, tidyr, utils, xml2
Config/testthat/edition: 3
Encoding: UTF-8
Language: en-gb
LazyData: TRUE
LazyDataCompression: xz
Config/roxygen2/version: 8.0.0
NeedsCompilation: no
Packaged: 2026-05-22 05:05:20 UTC; jamesdorey
Author: James B. Dorey [aut, cre, cph] (ORCID:
    <https://orcid.org/0000-0003-2721-3842>),
  Robert L. O'Reilly [aut] (ORCID:
    <https://orcid.org/0000-0001-5291-7396>),
  Silas Bossert [aut] (ORCID: <https://orcid.org/0000-0002-3620-5468>),
  Erica E. Fischer [aut] (ORCID: <https://orcid.org/0000-0002-8202-158X>)
Maintainer: James B. Dorey <jbdorey@me.com>
Repository: CRAN
Date/Publication: 2026-05-22 06:52:05 UTC
Built: R 4.6.0; ; 2026-05-22 12:04:33 UTC; unix

# Generated by roxygen2: do not edit by hand

export("%>%")
export(BeeBDCQuery)
export(ChaoWrapper)
export(ColTypeR)
export(GBIFissues)
export(PaigeIntegrater)
export(USGS_formatter)
export(atlasDownloader)
export(beesChecklist)
export(beesTaxonomy)
export(chordDiagramR)
export(continentOutlieRs)
export(coordUncerFlagR)
export(countryHarmoniseR)
export(countryNameCleanR)
export(countryOutlieRs)
export(dataProvTables)
export(dataSaver)
export(dateFindR)
export(diagonAlley)
export(dirMaker)
export(dupePlotR)
export(dupeSummary)
export(fileFinder)
export(flagAbsent)
export(flagLicense)
export(flagRecorder)
export(flagSummaryTable)
export(formattedCombiner)
export(ggRichnessWrapper)
export(harmoniseR)
export(iNEXTwrapper)
export(idMatchR)
export(importOccurrences)
export(interactiveMapR)
export(jbd_CfC_chunker)
export(jbd_Ctrans_chunker)
export(jbd_coordCountryInconsistent)
export(jbd_coordinates_precision)
export(jbd_coordinates_transposed)
export(jbd_create_figures)
export(manualOutlierFindeR)
export(plotFlagSummary)
export(readr_BeeBDC)
export(repoFinder)
export(repoMerge)
export(richnessEstimateR)
export(richnessPrepR)
export(summaryFun)
export(summaryMaps)
export(taxadbToBeeBDC)
importFrom(CoordinateCleaner,cc_val)
importFrom(CoordinateCleaner,clean_coordinates)
importFrom(circlize,chordDiagram)
importFrom(circlize,circos.clear)
importFrom(circlize,circos.par)
importFrom(circlize,circos.trackPlotRegion)
importFrom(circlize,get.cell.meta.data)
importFrom(circlize,highlight.sector)
importFrom(circlize,mm_h)
importFrom(dplyr,"%>%")
importFrom(dplyr,across)
importFrom(dplyr,all_of)
importFrom(dplyr,arrange)
importFrom(dplyr,as_tibble)
importFrom(dplyr,bind_cols)
importFrom(dplyr,bind_rows)
importFrom(dplyr,contains)
importFrom(dplyr,desc)
importFrom(dplyr,distinct)
importFrom(dplyr,filter)
importFrom(dplyr,full_join)
importFrom(dplyr,group_by)
importFrom(dplyr,if_else)
importFrom(dplyr,left_join)
importFrom(dplyr,lst)
importFrom(dplyr,mutate)
importFrom(dplyr,n)
importFrom(dplyr,n_groups)
importFrom(dplyr,pull)
importFrom(dplyr,relocate)
importFrom(dplyr,rename)
importFrom(dplyr,row_number)
importFrom(dplyr,select)
importFrom(dplyr,tibble)
importFrom(dplyr,where)
importFrom(ggplot2,ggtitle)
importFrom(ggplot2,xlab)
importFrom(ggplot2,ylab)
importFrom(ggspatial,north_arrow_fancy_orienteering)
importFrom(grDevices,gray)
importFrom(grid,unit)
importFrom(here,here)
importFrom(here,i_am)
importFrom(lubridate,as_date)
importFrom(lubridate,month)
importFrom(lubridate,ymd)
importFrom(paletteer,paletteer_dynamic)
importFrom(readr,col_character)
importFrom(readr,cols)
importFrom(readr,read_csv)
importFrom(readr,write_excel_csv)
importFrom(sf,sf_use_s2)
importFrom(stats,complete.cases)
importFrom(stats,setNames)
importFrom(stringr,str_c)
importFrom(stringr,str_replace)
importFrom(stringr,str_to_sentence)
importFrom(utils,unzip)

<!-- README.md is generated from README.Rmd. Please edit that file -->

# <a href='https://github.com/jbdorey/BeeBDC'><img src="https://photos.smugmug.com/photos/i-MpLFKTT/0/741daa6d/X4/i-MpLFKTT-X4.png" alt="BeeBDC logo of a cuckoo bee sweeping up occurrence records in South America" align="right" width="155"/></a>

# **BeeBDC: an occurrence data cleaning package**

<!-- badges: start -->

[![CRANstatus](https://www.r-pkg.org/badges/version/BeeBDC)](https://CRAN.R-project.org/package=BeeBDC)
[![downloads](https://cranlogs.r-pkg.org/badges/grand-total/BeeBDC)](https://cranlogs.r-pkg.org:443/badges/grand-total/BeeBDC)
[![R-CMD-check](https://github.com/jbdorey/BeeBDC/actions/workflows/R-CMD-check.yaml/badge.svg)](https://github.com/jbdorey/BeeBDC/actions/workflows/R-CMD-check.yaml)
[![License](https://img.shields.io/badge/license-GPL%20(%3E=%203)-lightgrey.svg?style=flat)](http://www.gnu.org/licenses/gpl-3.0.html)

<!-- badges: end -->

## **Overview**

The reliable implementation of biodiversity data continues to be a
challenge for researchers. We present the **BeeBDC** package which
provides novel and updated functions for flagging, cleaning,
visualising, and analysing occurrence datasets. Our package is general
and can be applied to any taxon; however, we also provide some functions
and data that are specific for use with bee occurrence data;
specifically due to their input data. We add new functionality and keep
conventions in other fantastic R packages, especially
[bdc](https://brunobrr.github.io/bdc/) and
[CoordinateCleaner](https://ropensci.github.io/CoordinateCleaner/articles/Cleaning_GBIF_data_with_CoordinateCleaner.html),
while also removing many dependencies on sp-related packages. Hence, our
package name is **Bee** **B**iodiversity **D**ata **C**leaning
(**BeeBDC**).

We provide a full workflow that uses **BeeBDC**, **bdc**, and
**CoordinateCleaner** to clean occurrence data in our Articles page and
encourage users to read and also cite this primary
[publication](https://doi.org/10.1101/2023.06.30.547152). For our
parallelised implementation of iChao and iNEXT species richness
estimations, cite this [primary
publication](https://doi.org/10.21203/rs.3.rs-6372769/v1).

### **The *BeeBDC* vignettes**

The **BeeBDC** vignettes are split into several, depending on waht
you’re hoping to do with the package. So far, it is broadly split into
the (1) data cleaning workflow and (2) species richness estimation.

1.  Data cleaning is broken into
    - 1.1 The [full cleaning
      workflow](https://jbdorey.github.io/BeeBDC/articles/BeeBDC_main.html)
    - 1.2 A [more basic
      workflow](https://jbdorey.github.io/BeeBDC/articles/basic_workflow.html)
    - 1.3 A [bee-data specific
      workflow](https://jbdorey.github.io/BeeBDC/articles/BeeBDC_data_preparation.html)
      to prepare those bee datasets
2.  A short, but complete, vignette to [estimate species
    richness](https://jbdorey.github.io/BeeBDC/articles/Species_richness_estimation.html)

### ***BeeBDC*’s structure**

The **BeeBDC** toolkit is intentionally organized using conventions in
**bdc** and **CoordinateCleaner**.

> Like in the **bdc** package, we provide a suggested workflow here.
> While our functions can mostly be run out of order, there are a few
> exceptions mentioned throughout the documentation. Additionally, many
> functions require the database_id column that is generated early on in
> the **BeeBDC** or **bdc** workflows. When running very large datasets
> (e.g., the global bee occurrence dataset) you may require a machine
> that has a minimum amount of RAM (~32 GB). However, we do try to
> provide work-arounds, especially by alowing some functions to be
> broken into consumable chunks. Paper DOI -
> <https://doi.org/10.1101/2023.06.30.547152>; Package GitHub -
> <https://github.com/jbdorey/BeeBDC/>

#### ![Workflow figure from Dorey et al. 2023](https://photos.smugmug.com/photos/i-V37Vg2w/4/X4/i-V37Vg2w-X4.jpg)

## **Installation**

### **Install *BeeBDC***

You can install *BeeBDC* from CRAN or GitHub.

``` r
  # Install BeeBDC from CRAN
install.packages("BeeBDC")

  # Or using the development version from GitHub (keeping in mind this may not be as stable)
remotes::install_github("https://github.com/jbdorey/BeeBDC.git", 
                          # To use the development version use "devel"; otherwise choose "main"
                        ref = "devel", force = TRUE)
```

### **Install *sf* and *terra***

> ### First time using the *sf* or *terra* packages?
>
> The first time that you use **terra** or **sf** on a new computer you
> may need to install some dependencies. Try to install the **terra**
> and **sf** packages first but then come back here if that doesn’t
> work.
>
> ##### Windows:
>
> On **Windows**, you need to first install
> [Rtools](https://cran.r-project.org/bin/windows/Rtools/) to get a C++
> compiler that **R** can use. You need a recent version of **Rtools42**
> (rtools42-5355-5357).
>
> ##### MacOS:
>
> On **macOS**, you can use [MacPorts](https://www.macports.org/) or
> [Homebrew](https://brew.sh/).
>
> With **MacPorts** you can do
>
> `sudo port install R-terra`
>
> With **Homebrew**, you need to first install GDAL:
>
> `brew install pkg-config`
>
> `brew install gdal`
>
> Followed by (note the additional configuration argument needed for
> Homebrew)
>
> ``` r
>   # Install terra
> install.packages("terra", type = "source", configure.args = "--with-proj-lib=$(brew --prefix)/lib/")
>   # install sf
> install.packages("sf", type = "source", configure.args = "--with-proj-lib=$(brew --prefix)/lib/")
>
> library(terra)
> library(sf)
> ```

Load the package with:

``` r
library(BeeBDC)
```

### **Optional packages**

Because **BeeBDC** provides broad functionality that might not be
required by all users, some dependencies are optional (but required for
some functions). Optional packages can be downloaded prior to starting
your workflow, if desired. *However, you will be prompted to download
these packages if they aren’t already installed when you run those
functions*. The packages **BiocManager** and **devtools** may also be
required to download some extra packages.

1.  The package, **rnaturalearthhires**, is a data package that allows
    the usage of higher-resolution country maps and is very useful for
    multiple **BeeBDC** functions.

``` r
  # rnaturalearthhires can be installed using devtools from their github page
if (!require("devtools", quietly = TRUE))
    install.packages("devtools")
  # Install rnaturalearthhires
devtools::install_github("ropensci/rnaturalearthhires")
```

2.  The package, **ComplexHeatmap**, is only used for one **BeeBDC**
    function (`BeeBDC::chordDiagramR()`) and is less critical.

``` r
  # ComplexHeatmap can be installed using BiocManager
if (!require("BiocManager", quietly = TRUE))
    install.packages("BiocManager")
  # Install ComplexHeatmap
BiocManager::install("ComplexHeatmap")
```

3.  The package **taxadb** is used by `BeeBDC::taxadbToBeeBDC()` to
    download and transform taxonomy files for **any taxon** and from
    multiple providers (e.g., GBIF and ITIS) to work with **BeeBDC**.

``` r
  # taxadb can be installed using install.packages
    install.packages("taxadb")
```

4.  The packages **iNEXT** and **SpadeR** can be downloaded in order to
    estimate species richness using occurrence data (see the [species
    richness
    estimation](https://jbdorey.github.io/BeeBDC/articles/Species_richness_estimation.html)
    vignette). Implemented in the functions `BeeBDC::iNEXTwrapper()`,
    `BeeBDC::ChaoWrapper()`, `BeeBDC::ggRichnessWrapper()`, and
    `BeeBDC::richnessEstimateR()`.

``` r
  # iNEXT and SpadeR can be installed using install.packages
    install.packages("iNEXT")
    install.packages("SpadeR")
```

## **BeeBDC workflow components**

#### 1. [**Data merge**](https://jbdorey.github.io/BeeBDC/articles/BeeBDC_main.html#data-merge)

Integrate and merge different datasets from major the data
repositories - GBIF, SCAN, iDigBio, the USGS, and ALA.

- `atlasDownloader()` Downloads ALA data and creates a new file in the
  path to put those data. This function can also request downloads from
  other atlases ([see
  here](https://galah.ala.org.au/R/articles/choosing_an_atlas.html)).
  However, it will only send the download to your email and you must do
  the rest yourself at this point.
- `repoMerge()` Locates data from GBIF, ALA, iDigBio, and SCAN within a
  directory and reads it in along with its eml metadata.
- `repoFinder()` Find GBIF, ALA, iDigBio, and SCAN files in a directory.
- `importOccurrences()` Looks for and imports the most-recent version of
  the occurrence data created by the `BeeBDC::repoMerge()` function.
- `USGS_formatter()` The function finds, imports, formats, and creates
  metadata for the USGS dataset.
- `formattedCombiner()` Merges the Darwin Core version of the USGS
  dataset that was created using `BeeBDC::USGS_formatter()` with the
  main dataset.
- `dataSaver()` Used at the end of 1.x in the example workflow in order
  to save the occurrence dataset and its associated eml metadata.

#### 2. [**Data preperation**](https://jbdorey.github.io/BeeBDC/articles/BeeBDC_main.html#data-preparation)

The reading in and formatting of the major and minor \[bee\] occurrence
repositories as well as some data modifications. This section is mostly,
but not entirely, related to bee occurrence data.

- `fileFinder()` A function which can be used to find files within a
  user-defined directory based on a user-provided character string.
- `PaigeIntegrater()` Replaces publicly available data with data that
  has been manually cleaned and error-corrected for use in the paper
  Chesshire, P. R., Fischer, E. E., Dowdy, N. J., Griswold, T.,
  Hughes, A. C., Orr, M. J., . . . McCabe, L. M. (In Press).
  Completeness analysis for over 3000 United States bee species
  identifies persistent data gaps. Ecography.
- `readr_BeeBDC()` Read in a variety of data files that are specific to
  certain smaller data providers. There is an internal readr function
  for each dataset and each one of these functions is called by
  readr_BeeBDC. While these functions are internal, they are displayed
  in the documentation of readr_BeeBDC for clarity.
- `idMatchR()` This function attempts to match database_ids from a prior
  **bdc** or **BeeBDC** run in order to keep this column somewhat
  consistent between iterations. However, not all records contain
  sufficient information for this to work flawlessly.

#### 3. [**Initial flags**](https://jbdorey.github.io/BeeBDC/articles/BeeBDC_main.html#initial-flags)

Flagging and carpentry of several, mostly general, data issues. See
**bdc**’s
[pre-filter](https://brunobrr.github.io/bdc/articles/prefilter.html) for
more related functions.

- `countryNameCleanR()` This is a basic function for a user to manually
  fix some country name inconsistencies.
- `jbd_CfC_chunker()` Because the
  `BeeBDC::jbd_country_from_coordinates()` function is very
  RAM-intensive, this wrapper allows a user to specify chunk-sizes and
  only analyse a small portion of the occurrence data at a time. The
  prefix jbd\_ is used to highlight the difference between this function
  and the original `bdc::bdc_country_from_coordinates()`.
- `jbd_Ctrans_chunker()` Because the
  `BeeBDC::jbd_coordinates_transposed()` function is very RAM-intensive,
  this wrapper allows a user to specify chunk-sizes and only analyse a
  small portion of the occurrence data at a time. The prefix jbd\_ is
  used to highlight the difference between this function and the
  original `bdc::bdc_coordinates_transposed()`. These functions will
  preferably use the countryCode column generated by
  `bdc::bdc_country_standardized()`.
- `jbd_coordCountryInconsistent()` Compares stated country name in an
  occurrence record with record’s coordinates using rnaturalearth data.
  The prefix, jbd\_ is meant to distinguish this function from the
  original `bdc::bdc_coordinates_country_inconsistent()`. This functions
  will preferably use the countryCode and country_suggested columns
  generated by `bdc::bdc_country_standardized()`; please run it on your
  dataset prior to running this function.
- `flagAbsent()` Flags occurrences that are “ABSENT” for the
  occurrenceStatus (or some other user-specified) column.
- `GBIFissues()` This function will flag records which are subject to a
  user-specified vector of GBIF issues.
- `flagRecorder()` This function is used to save the flag data for your
  occurrence data as you run the **BeeBDC** script. It will read and
  append existing files, if asked to. Your flags should also be saved in
  the occurrence file itself automatically.

#### 4. [**Taxonomy**](https://jbdorey.github.io/BeeBDC/articles/BeeBDC_main.html#taxonomy)

Harmonisation of scientific names against a taxonomy downloaded from
**taxadb**, from the provided [Discover
Life](https://www.discoverlife.org) website’s taxonomic reference, or a
custom taxonomy.

- `taxadbToBeeBDC()` Uses the **taxadb** to download a species taxonomy
  from any of their sources and transforms it into the **BeeBDC** format
  that can then be exported as a .csv or into the **R** environment to
  be be fed directly into `BeeBDC::harmoniseR()`. This means that the
  taxonomy from ANY taxon can be used. See also `BeeBDC::beesTaxonomy()`
  for the best global bee taxonomy.
- `harmoniseR()` Uses the Discover Life taxonomy to harmonise bee
  occurrences and flag those that do not match the checklist. This
  function could be hijacked to service other taxa if a user matched the
  format of the beesTaxonomy file. `BeeBDC::harmoniseR()` prefers to use
  the names_clean columns that is generated by `bdc::bdc_clean_names()`.
  While this is not required, you may find better results by running
  that function on your dataset first.

#### 5. [**Space**](https://jbdorey.github.io/BeeBDC/articles/BeeBDC_main.html#space)

Flagging of erroneous, suspicious, and low-precision geographic
coordinates.

- `jbd_coordinates_precision()` This function flags occurrences where
  BOTH latitude and longitude values are rounded. This contrasts with
  the original function, `bdc::bdc_coordinates_precision()` that will
  flag occurrences where only one of latitude OR longitude are rounded.
  The **BeeBDC** approach saves occurrences that may have had terminal
  zeros rounded in one coordinate column.
- `diagonAlley()` A simple function that looks for potential latitude
  and longitude fill-down errors by identifying consecutive occurrences
  with coordinates at regular intervals. This is accomplished by using a
  sliding window with the length determined by minRepeats.
- `coordUncerFlagR()` To use this function, the user must choose a
  column, probably “coordinateUncertaintyInMeters” and a threshold above
  which occurrences will be flagged for geographic uncertainty.
- `countryOutlieRs()` This function flags country-level outliers using
  the checklist provided with this package. For additional context and
  column names, see `beesChecklist`.
- `continentOutlieRs()` This function flags continent-level outliers
  using the checklist provided with this package. This function works
  much the same as `countryOutlieRs()`, but at a lower resolution. For
  additional context and column names, see `beesChecklist`.
- `jbd_create_figures()` Creates figures (i.e., bar plots, maps, and
  histograms) reporting the results of data quality tests implemented
  the bdc and **BeeBDC** packages. Works like
  `bdc::bdc_create_figures()`, but it allows the user to specify a save
  path.

#### 6. [**Time**](https://jbdorey.github.io/BeeBDC/articles/BeeBDC_main.html#time)

Flagging and, whenever possible, correction of inconsistent collection
date.

- `dateFindR()` A function made to search other columns for dates and
  add them to the eventDate column. The function searches the columns
  locality, fieldNotes, locationRemarks, and verbatimEventDate for the
  relevant information.

#### 7. [**De-duplication**](https://jbdorey.github.io/BeeBDC/articles/BeeBDC_main.html#de-duplication)

- `dupeSummary()` This function uses user-specified inputs and columns
  to identify duplicate occurrence records. Duplicates are identified
  iteratively and will be tallied up, duplicate pairs clustered, and
  sorted at the end of the function. The function is designed to work
  with Darwin Core data with a database_id column, but it is also
  modifiable to work with other columns.

#### 8. [**Filtering**](https://jbdorey.github.io/BeeBDC/articles/BeeBDC_main.html#data-filtering)

- `manualOutlierFindeR()` Uses expert-identified outliers with source
  spreadsheets that may be edited by users. The function will also use
  the duplicates file made using `BeeBDC::dupeSummary()` to identify
  duplicates of the expert-identified outliers and flag those as well.
  The function will add a flagging column called .expertOutlier where
  records that are FALSE are the expert outliers.
- `summaryFun()` Using all flag columns (column names starting with
  “.”), this function either creates or updates the .summary flag column
  which is FALSE when ANY of the flag columns are FALSE. Columns can be
  excluded and removed after creating the .summary column. Additionally,
  the occurrence dataset can be filtered to only those where .summary =
  TRUE at the end of the function.

#### 9. [**Figures and tables**](https://jbdorey.github.io/BeeBDC/articles/BeeBDC_main.html#figures-and-tables)

- `chordDiagramR()` This function outputs a figure which shows the
  relative size and direction of occurrence points duplicated between
  data providers, such as, SCAN, GBIF, ALA, etc. This function requires
  the outputs generated by `BeeBDC::dupeSummary()`.
- `dupePlotR()` Creates a plot with two bar graphs. One shows the
  absolute number of duplicate records for each data source while the
  other shows the proportion of records that are duplicated within each
  data source. This function requires a dataset that has been run
  through `BeeBDC::dupeSummary()`.
- `plotFlagSummary()` Creates a compound bar plot that shows the
  proportion of records that pass or fail each flag (rows) and for each
  data source (columns). The function can also optionally return a point
  map for a user-specified species when plotMap = TRUE. This function
  requires that your dataset has been run through some filtering
  functions - so that is can display logical columns starting with “.”.
- `summaryMaps()` Builds an output figure that shows the number of
  species and the number of occurrences per country. Breaks the data
  into classes for visualisation. Users may filter data to their taxa of
  interest to produce figures of interest.
- `interactiveMapR()` Uses the occurrence data (preferably uncleaned)
  and outputs interactive .html maps that can be opened in your browser
  to a specific directory. The maps can highlight if an occurrence has
  passed all filtering (.summary == TRUE) or failed at least one filter
  (.summary == FALSE). This can be modified by first running
  `BeeBDC::summaryFun()` to set the columns that you want to be
  highlighted. It can also highlight occurrences flagged as
  expert-identified or country outliers.
- `dataProvTables()` This function will attempt to find and build a
  table of data providers that have contributed to the input data,
  especially using the ‘institutionCode’ column. It will also look for a
  variety of other columns to find data providers using a an internally
  set sequence of if-else statements. Hence, this function is quite
  specific for bee data, but should work for other taxa in similar
  institutions.
- `flagSummaryTable()` Takes a flagged dataset and returns the total
  number of fails (FALSE) per flag (columns starting with “.”) and per
  species. Users may define the column to group the summary by. While it
  is intended to work with the scientificName column, users may select
  any grouping column (e.g., country).

#### 10. [Species richness estimation](https://jbdorey.github.io/BeeBDC/articles/speciesRichness_example.html)

- `diversityPrepR()` Takes your occurrence dataset along with a taxonomy
  and checklist in order to produce a file that’s ready to be passed
  into the `BeeBDC::richnessEstimateR()` function in order to estimate
  species richness using iChao (non-parametric species richness;
  `BeeBDC::ChaoWrapper()`) and iNEXT (hill numbers;
  `BeeBDC::iNEXTwrapper()`) for countries, continents, or the entire
  globe.
- `richnessEstimateR()` Takes an output dataset from
  `BeeBDC::diversityPrepR()` to estimate species richness using iChao
  (non-parametric species richness; `BeeBDC::ChaoWrapper()`) and iNEXT
  (hill numbers; `BeeBDC::iNEXTwrapper()`) for countries, continents,
  and/or the entire globe. Has parallel functionality.
- `iNEXTwrapper()` A wrapper for `iNEXT::iNEXT()` to interpolate and
  extrapolate Hill numbers with order q (rarify species richness). The
  wrapper has the ability to estimate species richness for multiple
  sites (or countries) at once and to do this using multiple cores.
- `ChaoWrapper()` A wrapper for `SpadeR::ChaoSpecies()` to
  non-parametrically estimate species richness. The wrapper has the
  ability to estimate species richness for multiple sites (or countries)
  at once and to do this using multiple cores at once.

#### 11. **Datasets**

We provide two full datasets that are downloadable using the below two
functions

- `beesTaxonomy()` Downloads the taxonomic information for the bees of
  the world. Source of taxonomy is listed under “source” but are mostly
  derived from the Discover Life website. The data will be sourced from
  the **BeeBDC** article’s Figshare. Please see also
  `BeeBDC::taxadbToBeeBDC()` for the download of any other taxonomy (for
  any taxa or for bees).
- `beesChecklist()` Download the table contains taxonomic and country
  information for the bees of the world based on data collated on
  Discover Life. The data will be sourced from the **BeeBDC** article’s
  Figshare.

We further provide five test datasets that are available with
**BeeBDC**.

- `BeeBDC::bees3sp` This test dataset includes 105 random occurrence
  records from three bee species. The included species are:
  “*Agapostemon tyleri* Cockerell, 1917”, “*Centris rhodopus* Cockerell,
  1897”, and “*Perdita octomaculata* (Say, 1824)”.
- `BeeBDC::beesRaw` A small bee occurrence dataset with flags generated
  by **BeeBDC** used to run example script and test functions. For data
  types, see `ColTypeR()`.
- `BeeBDC::beesFlagged` A small bee occurrence dataset with flags
  generated by BeeBDC used to run example script and test functions. For
  data types, see `ColTypeR()`.
- `BeeBDC::beesCountrySubset` A very small bee occurrence dataset with
  the columns “scientificName” and “country_suggested” and data for four
  countries, Fiji, Uganda, Vietnam, and Zambia. This is the test dataset
  for the species richness functions.
- There are also two small test datasets of the beesTaxonomy and
  beesChecklist in the system files of the package that are filtered to
  include *only* those species that occur in bees3sp, beesRaw, and
  beesFlagged. These are accessible as follows but are only used
  internally for tests.

``` r
  # Access the test taxonomy file
system.file("extdata", "testTaxonomy.rda", package="BeeBDC") |> load()
  # View the file
View(testTaxonomy)
  # Access the test checklist file
system.file("extdata", "testChecklist.rda", package="BeeBDC") |> load()
  # View the file
View(testChecklist)
```

## Package info

#### **Package website**

See **BeeBDC** package website
(<https://jbdorey.github.io/BeeBDC/reference/index.html>) for detailed
explanation on each module.

#### **Getting help**

> This package is maintained by Dr James B Dorey, Lecturer in Biological
> Sciences at the University of Wollongong, Australia.

> If you encounter a clear bug, please file an issue
> [**here**](https://github.com/jbdorey/BeeBDC/issues). For questions or
> suggestion, flick us an email (jdorey@uow.edu.au).

#### **Citations**

- **Original paper, dataset, and package citation:** Dorey, J.B.,
  Fischer, E.E., Chesshire, P.R., Nava-Bolaños, A., O’Reilly, R.L.,
  Bossert, S., Collins, S.M., Lichtenberg, E.M., Tucker, E.,
  Smith-Pardo, A., Falcon-Brindis, A., Guevara, D.A., Ribeiro, B.R., de
  Pedro, D., Hung, J.K.-L., Parys, K.A., McCabe, L.M., Rogan, M.S.,
  Minckley, R.L., Velzco, S.J.E., Griswold, T., Zarrillo, T.A., Jetz,
  W., Sica, Y.V., Orr, M.C., Guzman, L.M., Ascher, J., Hughes, A.C. &
  Cobb, N.S. (2023) A globally synthesised and flagged bee occurrence
  dataset and cleaning workflow. *Scientific Data*, 10, 1–17.
  <https://www.doi.org/10.1038/S41597-023-02626-W>
  - **Figshare live data link:**
    <https://doi.org/10.25451/flinders.21709757>
- **Species richness estimation citation:** Dorey J. B., Gilpin, A.-M.,
  Johnson, N., Esquerre, D., Hughes, A. C., Ascher, J. S., & Orr, M. C.
  (2026). Estimating global bee species richness and taxonomic gaps.
  *Nature Communications*. <https://doi.org/10.1038/s41467-026-69029-4>
- **BeeBDC package citation:** Dorey, J. B., O’Reilly, R. L.,
  Bossert, S. & Fischer, E. (2023). **BeeBDC**: an occurrence data
  cleaning package. R package version 1.3.4. url:
  <https://github.com/jbdorey/BeeBDC>
- **Discover Life citation (for use of bee taxonomy and checklist):**
  Ascher, J.S. & Pickering, J. (2026) Discover Life bee species guide
  and world checklist (Hymenoptera: Apoidea: Anthophila).
  <https://www.discoverlife.org/mp/20q?guide=Apoidea_species>

This package and its data sets were created with the support, and as a
part, of the iDigBees project

<a href='https://idigbees.org'><img src="https://photos.smugmug.com/photos/i-rcd8CCs/0/b291edd1/O/i-rcd8CCs.png" alt="The iDigBees logo with a colourful bee and the iDigBees text on the right" align="left" width="300"/></a>

# This R script was written by James Dorey, starting on the 2nd of May 2022. The script intends
# to clean occurrence data from several sources using a combination of custom functions and 
# functions from the "bdc" package.
# For queries, please feel free to contact James Dorey at jbdorey@me.com

# WHAT TO ADD? You can list things here if you see the need.


#### 0.0 Script preparation ####
##### 0.1 Working directory ####
# Choose the path to the root folder in which all other folders can be found (or made by dirMaker)
RootPath <- "/Users/jamesdorey/Desktop/Uni/My_papers/Bee_SDM_paper"
# Read in this initial function
source(paste(RootPath, "/BDC_repo/NewFunctions/dirMaker.R", sep = ""))
library(magrittr)
# Create file paths and prepare for what's to come
dirMaker(
  RootPath = RootPath,
  # Input the location of the workflow script RELATIVE to the RootPath
  RDoc = "BDC_repo/BeeCleaning_SciData.R") %>%
  # Add paths created by this function to the environment()
  list2env(envir = parent.env(environment())) 
# Set the working directory
setwd(DataPath)
# Install reenv, IF NEEDED
#install.packages("renv")
renv::init() 


##### 0.2 Install packages (if needed) #####
# Install only those packages that are not already present in your system
# Choose packages that need to be installed 
# You may need to install gdal on your computer. This can be done on mac by using
# Homebrew in the terminal and the command "brew install gdal"
list.of.packages <- c("R.utils",           # To use gunzip
                      "magrittr",          # to use pipes
                      "dplyr",             #  Part of the tidyverse
                      "forcats",           # tidyverse for working with factors
                      "tidyr",             #  Part of the tidyverse
                      "rlist",             # Package to save lists
                      "galah",             #  To download ALA data
                      "praise",            #  To whispers sweet nothings 
                      "EML",               #  To work with .eml files
                      "emld",              #  To work with .eml files
                      "rlang",             #  Part of the tidyverse - core functions
                      "xml2",              #  Part of the tidyverse - reads .xml files
                      "stringr",           #  Part of the tidyverse - works with text strings
                      "lubridate",         #  Part of the tidyverse - works with dates
                      "tidyselect",        #  Part of the tidyverse
                      "mgsub",             #  To perform multiple text substitutions
                      "bdc",               # data cleaning package
                      "rvest",             # Package for interfacing with and downloading from the internet
                      "rnaturalearth",     #  Global vector map data 
                      "rnaturalearthdata", #  To access the above global map data
                      "countrycode",       # Package to deal with country codes
                      "rworldmap",
                      "readxl",
                      "cowplot",           # ggplot2 helper package
                      "ggspatial")         #  Makes ggplot2 create north arrows or scale bars
# Install sf and terra seperately
# renv::install(c("sf","terra"), type = "binary")
renv::install(c("sf","terra"), type = "binary")
# List the new (not installed) packages and then if there are any, install them.
renv::install(packages = c(list.of.packages), 
              rebuild = FALSE) # try changing to TRUE if you're having package troubles

##### 0.3 Load packages ####
# Load all packages from the list specified above, 
lapply(c(list.of.packages, "sf","terra"), 
       library, character.only = TRUE)
### Load in R scripts and character strings in our package
sapply(list.files(file.path(ScriptPath), pattern = ".R$", full.names = TRUE), source)  # loads in all functions
# Save a snapshot of the environment
renv::snapshot()


#### 1.0 Data read ####
  ##### 1.1 Read data ####
# Read in the uncleaned, but flagged, dataset
beeData <- readr::read_csv("/Users/jamesdorey/Desktop/Uni/My_papers/Bee_SDM_paper/Data_acquisition_workflow/Output/Intermediate/05_unCleaned_database.csv",
                           col_types = ColTypeR())

  ##### 1.2 Bee taxonomy ####
# Read in the custom taxonomy file
  # This can be used to filter to a particular taxon
BeeTaxonomy <- BeeBDC::beesTaxonomy()

#### 2.0 Taxon example ####
  # Users could filter to a particular taxonomic group by editing the below
  # Select only unique selected bee genera
selectedGenera <- BeeTaxonomy %>%
    # Select only tribe anthophorini (for example)
  dplyr::filter(tolower(tribe) == tolower("anthophorini")) %>%
  distinct(genus)
  
  # Filter the data
taxonData <- beeData %>%
  dplyr::filter(genus %in% selectedGenera$genus)



#### 3.0 Country example ####
# Users could filter to a particular group of countries by editing the below
studyArea <- c("Canada", "United states", "Mexico", "Guatemala")
# Filter the data 
countryData <- beeData %>%
  dplyr::filter(country %in% studyArea)



#### 4.0 Filtering example ####
  ##### 4.1 Simple filter ####
  # Users can filter the flagged data by whichever columns they wish using the summaryFun...
source(paste(ScriptPath, "summaryFun.R", sep = "/"))
filteredData <- 
    # The input dataset
  beeData %>%
    # Run the summary function
  summaryFun(
  # Use the above input dataset to filter
  data = .,
  # Choose the columns to NOT filter (or NULL to filter all columns)
  dontFilterThese = c(".gridSummary", ".lonFlag", ".latFlag", ".uncer_terms",
                      ".uncertaintyThreshold"),
  # In the output, do you want to REMOVE all filtering columns (TRUE), or keep them (FALSE)
  removeFilterColumns = TRUE,
  # In the output, do you want to only keep clean data according to your filtering (TRUE),
    # Or keep all data and simply update the .summary column (FALSE)
  filterClean = TRUE) # END summary function

##### 4.2 Uncertainty threshold ####
# Users may also want to filter by .uncertaintyThreshold, but want to specify their own 
  # Uncertainty threshold...
# Users can filter the flagged data by whichever columns they wish using the summaryFun...
source(paste(ScriptPath, "summaryFun.R", sep = "/"))
filteredData <- 
    # The input dataset
  beeData %>%
  # Remove any exiting .uncertaintyThreshold column
  dplyr::select(!tidyselect::any_of(".uncertaintyThreshold")) %>%
    # Chose the coordinate uncertainty to filter to...
  coordUncerFlagR(data = .,
                  uncerColumn = "coordinateUncertaintyInMeters",
                    # 10 km here
                  threshold = 10000) %>%
    # Now re-do the .summary column and filter the data using this new value
  summaryFun(
  # Select the input dataset to filter
  data = .,
  # Choose the columns to NOT filter (or NULL to filter all columns)
    # NOTE: the .uncertaintyThreshold is now removed and WILL be filtered
  dontFilterThese = c(".gridSummary", ".lonFlag", ".latFlag", ".uncer_terms"),
  # In the output, do you want to REMOVE all filtering columns (TRUE), or keep them (FALSE)
  removeFilterColumns = TRUE,
  # In the output, do you want to only keep clean data according to your filtering (TRUE),
  # Or keep all data and simply update the .summary column (FALSE)
  filterClean = TRUE)


##### 4.2 Date filter ####
  ###### a. bdc_year_outOfRange ####
# Users can filter the flagged data by whichever columns they wish using the summaryFun...
source(paste(ScriptPath, "summaryFun.R", sep = "/"))
filteredData <- 
  # The input dataset
  beeData %>%
    # Remove any exisitng .year_outOfRange column
  dplyr::select(!".year_outOfRange") %>%
  # Chose the minimum year to filter to...
  bdc::bdc_year_outOfRange(data = .,
                           eventDate = "year",
                           year_threshold = 1970) %>%
  # Now re-do the .summary column and filter the data using this new value
  summaryFun(
    # Select the input dataset to filter
    data = .,
    # Choose the columns to NOT filter (or NULL to filter all columns)
    dontFilterThese = c(".gridSummary", ".lonFlag", ".latFlag", ".uncer_terms",
                        ".uncertaintyThreshold"),
    # In the output, do you want to REMOVE all filtering columns (TRUE), or keep them (FALSE)
    removeFilterColumns = TRUE,
    # In the output, do you want to only keep clean data according to your filtering (TRUE),
    # Or keep all data and simply update the .summary column (FALSE)
    filterClean = TRUE)

###### b. year range ####
# Users can filter the flagged data by whichever columns they wish using the summaryFun...
source(paste(ScriptPath, "summaryFun.R", sep = "/"))
filteredData <- 
  # The input dataset
  beeData %>%
  # Chose the year range...
  dplyr::filter(year > 1950 & year < 1970) %>%
  # Now re-do the .summary column and filter the data using this new value
  summaryFun(
    # Select the input dataset to filter
    data = .,
    # Choose the columns to NOT filter (or NULL to filter all columns)
    dontFilterThese = c(".gridSummary", ".lonFlag", ".latFlag", ".uncer_terms",
                        ".uncertaintyThreshold"),
    # In the output, do you want to REMOVE all filtering columns (TRUE), or keep them (FALSE)
    removeFilterColumns = TRUE,
    # In the output, do you want to only keep clean data according to your filtering (TRUE),
    # Or keep all data and simply update the .summary column (FALSE)
    filterClean = TRUE)


    # Users may choose any number of filtering steps form the main workflow to include above 
    # summaryFun(), just use pipes '%>%' between the function and use '.' as the data input
    # because this wil lfeed in the data aoutput from the above function into the proceeding one.



#### 5. Summary figures ####
##### 5.1 Duplicate chordDiagrams ####
utils::install.packages("circlize")
if(!require("BiocManager", quietly = TRUE)){
  utils::install.packages("BiocManager")}
BiocManager::install("ComplexHeatmap", force = TRUE)
utils::install.packages("paletteer")
library(paletteer)# Find palettes here
renv::snapshot()

# Read in the most-RECENT file
duplicates <- fileFinder(path = "PATH TO A FOLDER CONTAINING THE duplicateRun_ - could be supp. materials folder",
                          fileName = "duplicateRun_") %>%
  readr::read_csv() %>%
  # Select only the stingless bee data
  dplyr::filter(database_id %in% stinglessData$database_id |
                  database_id_match %in% stinglessData$database_id)

# Choose the global figure parameters
par(mar = c(2, 2, 2, 2)/2, mfrow = c(1,1))

# Create the chorDiagram. You can leave many of the below values out but we show here
# the defaults
source(paste(ScriptPath, "chordDiagramR.R", sep = "/"))
chordDiagramR(
  # The duplicate data from the dupeSummary function output  
  dupeData = duplicates,
  outPath = OutPath_Figures,
  fileName = "ChordDiagram.pdf",
  # These can be modified to help fit the final pdf that's exported.
  width = 9,
  height = 7.5,
  bg = "white",
  # How few distinct dataSources should a group have to be listed as "other"
  smallGrpThreshold = 3,
  title = "Duplicated record sources",
  # The default list of colour palettes to choose from usign the paleteer package
  palettes = c("cartography::blue.pal", "cartography::green.pal", 
               "cartography::sand.pal", "cartography::orange.pal", "cartography::red.pal",
               "cartography::purple.pal", "cartography::brown.pal"),
  canvas.ylim = c(-1.0,1.0), 
  canvas.xlim = c(-0.6, 0.25),
  text.col = "black",
  legendX = grid::unit(6, "mm"),
  legendY = grid::unit(18, "mm"),
  legendJustify = c("left", "bottom"),
  niceFacing = TRUE)
# Save 7*6


##### 5.2 Duplicate histogram ####
# Find the existing flag file and read it in
flagColumns <- fileFinder(path = "PATH TO A FOLDER CONTAINING THE flagsRecorded_ - could be supp. materials folder",
                           fileName = "flagsRecorded_") %>%
  readr::read_csv() %>%
    # WARNING: alternate path if wanting to produce figures for the selected taxonData (2.0 above)
  # Select only the taxonData data
  dplyr::filter(database_id %in% taxonData$database_id)

# Read in the function
source(paste(ScriptPath, "dupePlotR.R", sep = "/"))
# Create a figure shoring the total number of duplicates, kept duplicates, and unique
# records for each datasource (simplified to the text before the first underscore) and
# the proportion of the above for each data source
dupePlotR(
  flagColumns = flagColumns,
  # The outPath to save the plot as
  outPath = paste0(DataPath, "/Output", "/Figures"),
  fileName = "duplicatePlot.pdf",
  # Colours in order: duplicate, kept duplicate, unique
  dupeColours = c("#F2D2A2","#B9D6BC", "#349B90"),
  # Plot size and height
  base_height = 7, base_width = 7,
  legend.position = c(0.85, 0.8),
  # Extra variables can be fed into forcats::fct_recode() to change names on plot
  GBIF = "GBIF", SCAN = "SCAN", iDigBio = "iDigBio", USGS = "USGS", ALA = "ALA", 
  ASP = "ASP"
)


##### 5.3 Flags by source ####
# Read in the function
source(paste(ScriptPath, "plotFlagSummary.R", sep = "/"))
  ###### a. All taxa in dataset ####
# Visualise all flags for each dataSource (simplified to the text before the first underscore)
plotFlagSummary(
    # WARNING: alternate path if wanting to produce figures for the selected taxonData (2.0 above)
      # Select only the taxonData data
  plotData = taxonData,
  # Colours in order of pass (TRUE), fail (FALSE), and NA
  flagColours = c("#127852", "#A7002D", "#BDBABB"),
  fileName = paste0("FlagsPlot_", Sys.Date(),".pdf"),
  outPath = paste0(DataPath, "/Output/Figures"),
  width = 15, height = 9,
  # Extra variables can be fed into forcats::fct_recode() to change names on plot
  GBIF = "GBIF", SCAN = "SCAN", iDigBio = "iDigBio", USGS = "USGS", ALA = "ALA", 
  ASP = "ASP", CAES = "CAES", 'B. Mont.' = "BMont", 'B. Minkley' = "BMin", Ecd = "Ecd",
  Gaiarsa = "Gai", EPEL = "EPEL"
)

  ###### b. Single sp. summary ####
# Visualise all flags for each dataSource (simplified to the text before the first underscore)
  # A clever user might also realise the potential to summarise and produce outputs in other columns
plotFlagSummary(
  # WARNING: alternate path if wanting to produce figures for the selected taxonData (2.0 above)
  # Select only the taxonData data
  plotData = beeData,
  # Colours in order of pass (TRUE), fail (FALSE), and NA
  flagColours = c("#127852", "#A7002D", "#BDBABB"),
  fileName = paste0("FlagsPlot_Lfijiense", Sys.Date(),".pdf"),
  outPath = paste0(DataPath, "/Output/Figures"),
  width = 15, height = 9,
  # OPTIONAL:
         #  # Filter to species
           speciesName = "Lasioglossum fijiense",
             # column to look in
           nameColumn = "species",
           # Save the filtered data
           saveFiltered = TRUE,
     # Filter column to display on map
           filterColumn = ".summary",
           plotMap = TRUE,
       # amount to jitter points if desired, e.g. 0.25 or NULL
     jitterValue = NULL,
       # Map opacity value for points between 0 and 1
     mapAlpha = 1,
  # Extra variables can be fed into forcats::fct_recode() to change names on plot
  GBIF = "GBIF", SCAN = "SCAN", iDigBio = "iDigBio", USGS = "USGS", ALA = "ALA", 
  ASP = "ASP", CAES = "CAES", 'B. Mont.' = "BMont", 'B. Minkley' = "BMin", Ecd = "Ecd",
  Gaiarsa = "Gai", EPEL = "EPEL"
)



##### 5.4 Maps ####
# Import CLEANED dataset (you can change this option)
  # WARNING: alternate path if wanting to produce figures for the selected taxonData (2.0 above)
    # Select only the taxonData data
mapData <- taxonData %>% 
  dplyr::filter(.summary == TRUE)


# Read in the function
# TO ADD: Change legend title and text to be nicer
# Add A and B
source(paste(ScriptPath, "summaryMaps.R", sep = "/"))
summaryMaps(
  mapData = mapData,
  width = 10, height = 10,
  class_n = 15,
  class_Style = "jenks",
  fileName = paste0(DataPath, "/Output/Figures/", "CountryMaps_jenks.pdf")
)



#### 6.0 Save data ####
mapData %>%
  readr::write_excel_csv(paste0(DataPath, "/Output/Intermediate/", "cleanTaxon_",
                          Sys.Date(), ".csv"))

# This R script was written by James Dorey, starting on the 2nd of May 2022. The script serves
# as a workflow for the BeeBDC package to clean and flag bee, and other, occurrence data.
# It also uses functions from several sources and particularly from the "bdc" package.
# For queries, please feel free to contact James Dorey at jbdorey@me.com


#### 0.0 Script preparation ####
##### 0.1 Working directory ####
# Choose the path to the root folder in which all other folders can be found (or made by dirMaker)
RootPath <- "/Users/jamesdorey/Desktop/Uni/My_papers/Bee_SDM_paper"

# Set the working directory
setwd(paste0(RootPath,"/Data_acquisition_workflow"))

# Initialise renv the project if needed
  # renv::init(project = paste0(RootPath,"/Data_acquisition_workflow")) 
renv::activate(project = paste0(RootPath,"/Data_acquisition_workflow"))


# Install BeeBDC from CRAN
utils::install.packages("BeeBDC")
# You could also install BeeBDC's development version using the below: 
# WARNING the development version may not pass all CRAN or GitHub tests.
     remotes::install_github("https://github.com/jbdorey/BeeBDC.git", user="jbdorey", 
                               # To use the development version, do below, otherwise choose "main"
                             ref = "devel", 
                             force = TRUE)

##### 0.2 Load packages ####
# Save a snapshot of the environment
renv::snapshot(project = paste0(RootPath,"/Data_acquisition_workflow"))
# Load all packages from the list specified above,
lapply(c("ComplexHeatmap", "BeeBDC", "magrittr"), 
       library, character.only = TRUE)


# Create file paths and prepare for what's to come
BeeBDC::dirMaker(
  RootPath = RootPath,
  # Input the location of the workflow script RELATIVE to the RootPath
  RDoc = "Packages/BeeBDC_development/Workflows/BeeBDC_fullWorkflow.R") %>%
  # Add paths created by this function to the environment()
  list2env(envir = parent.env(environment()))  

#### 1.0 Data merge ####
##### 1.1 Download ALA data ####
# Downloads ALA data and creates a new file in the HomePath to put those data
BeeBDC::atlasDownloader(path = DataPath,
               userEmail = "jbdorey@me.com",
               atlas = "ALA",
               ALA_taxon = "Apiformes")

##### 1.2 Import and merge ALA, SCAN, iDigBio and GBIF data ####
# Supply the path to where the data are
# save_type is either "csv_files" or "R_file"
DataImp <- BeeBDC::repoMerge(path = DataPath, 
                        # Find data - Many problems can be solved by running repoFinder(path = DataPath)
                        # And looking for problems
                      occ_paths = BeeBDC::repoFinder(path = DataPath),
                      save_type = "R_file")

# Load in the most-recent version of these data if needed 
# This will return a list with 
# 1. the occurrence dataset with attributes and 
# 2. the appended eml file
DataImp <- BeeBDC::importOccurrences(path = DataPath,
                             fileName = "BeeData_")

##### 1.3 Import USGS Data ####
# The USGS_formatter will find, import, format, and create metadata for the USGS dataset
# pubDate must be in day-month-year format and must be supplied by the user here.
USGS_data <- BeeBDC::USGS_formatter(path = DataPath, pubDate = "19-11-2022")

##### 1.4 Formatted Source Importer ####
# Formatted source importer. Use this importer to find files that have been formatted and need to 
# be added to the larger data file (e.g., made by repoMerge and USGS_formatter)
# Combine the USGS data and the existing big dataset
Complete_data <- BeeBDC::formattedCombiner(path = DataPath, 
                                    strings = c("USGS_[a-zA-Z_]+[0-9]{4}-[0-9]{2}-[0-9]{2}"), 
                                    # This should be the list-format with eml attached
                                    existingOccurrences = DataImp$Data_WebDL,
                                    existingEMLs = DataImp$eml_files) 
# In the catalogNumber, remove ".*specimennumber:" as what comes after should be the USGS
# number to match for duplicates
Complete_data$Data_WebDL <- Complete_data$Data_WebDL %>%
  dplyr::mutate(catalogNumber = stringr::str_replace(catalogNumber,
                                                     pattern = ".*\\| specimennumber:",
                                                     replacement = ""))

##### 1.5 Save data ####
# Choose the type of data format you want
BeeBDC::dataSaver(path = DataPath,# The main path to look for data in
           save_type = "CSV_file", # "R_file" OR "CSV_file"
           occurrences = Complete_data$Data_WebDL, # The existing datasheet
           eml_files = Complete_data$eml_files, # The existing EML files
           file_prefix = "Fin_") # The prefix for the fileNames
rm(Complete_data, DataImp, USGS_data)


#### 2.0 Data preparation ####
##### 2.1 Standardise datasets ####
  # You may either use 
    # (a) the bdc import method (works well with general datasets) or 
    # (b) the BeeBDC import method (works well with above data merge)
  # The bdc import is NOT truly supported here, but provided as an example. Please go to section
    # 2.1b below.
###### a. bdc import ####
warning(paste0("The bdc method here is not truly implemented and supported. If you use it you must do so alone.",
               " This is just a place-holder for people using the bdc package more heavily.",
               "\nPreferably, go directly to 2.1b - BeeBDC import."))
# Read in the bdc metadata
bdc_metadata <- readr::read_csv(paste(DataPath, "Output", "bdc_integration.csv", sep = "/"))
# Standardise the dataset to bdc
db_standardized <- bdc::bdc_standardize_datasets(
  metadata = bdc_metadata,
  format = "csv",
  overwrite = TRUE,
  save_database = TRUE)
# read in configuration description file of the column header info
config_description <- readr::read_csv(paste(DataPath, "Output", "bdc_configDesc.csv",
                                            sep = "/"), 
                                      show_col_types = FALSE, trim_ws = TRUE)

###### b. BeeBDC import ####
# You can also just read the data in using the below script. This will 
# likely be quicker and more-reliable. Find the path
occPath <- BeeBDC::fileFinder(path = DataPath, fileName = "Fin_BeeData_combined_")
# read in the file
db_standardized <- readr::read_csv(occPath, 
                                   # Use the basic ColTypeR function to determine types
                                   col_types = BeeBDC::ColTypeR(), trim_ws = TRUE) %>%
  # add the database_id columns
  dplyr::mutate(database_id = paste("Dorey_data_", 1:nrow(.), sep = ""),
                .before = family)

###### c. optional thin ####
    # You can thin the dataset for TESTING ONLY!
    # check_pf <- check_pf %>%
    #   # take every 100th record
    #   filter(row_number() %% 100 == 1)


##### 2.2 Paige dataset ####
# Integrate Paige Chesshire's cleaned dataset.
# Import Paige's cleaned N. American data
# IF you haven't figured out by now, dont worry about the column name warning - not all columns occur here.
PaigeNAm <- readr::read_csv(paste(DataPath, "Paige_data", "NorAmer_highQual_only_ALLfamilies.csv",
                                  sep = "/"), col_types = BeeBDC::ColTypeR()) %>%
  # Change the column name from Source to dataSource to match the rest of the data.
  dplyr::rename(dataSource = Source) %>%
  # add a NEW database_id column
  dplyr::mutate(
    database_id = paste0("Paige_data_", 1:nrow(.)),
    .before = scientificName)

# Merge Paige's data with downloaded data
db_standardized <- BeeBDC::PaigeIntegrater(
  db_standardized = db_standardized,
  PaigeNAm = PaigeNAm,
  # This is a list of columns by which to match Paige's data to the most-recent download with. 
  # Each vector will be matched individually
  columnStrings = list(
    c("decimalLatitude", "decimalLongitude", 
      "recordNumber", "recordedBy", "individualCount", "samplingProtocol",
      "associatedTaxa", "sex", "catalogNumber", "institutionCode", "otherCatalogNumbers",
      "recordId", "occurrenceID", "collectionID"),         # Iteration 1
    c("catalogNumber", "institutionCode", "otherCatalogNumbers",
      "recordId", "occurrenceID", "collectionID"), # Iteration 2
    c("decimalLatitude", "decimalLongitude", 
      "recordedBy", "genus", "specificEpithet"),# Iteration 3
    c("id", "decimalLatitude", "decimalLongitude"),# Iteration 4
    c("recordedBy", "genus", "specificEpithet", "locality"), # Iteration 5
    c("recordedBy", "institutionCode", "genus", 
      "specificEpithet","locality"),# Iteration 6
    c("occurrenceID","decimalLatitude", "decimalLongitude"),# Iteration 7
    c("catalogNumber","decimalLatitude", "decimalLongitude"),# Iteration 8
    c("catalogNumber", "locality") # Iteration 9
  ) )

# Remove spent data
rm(PaigeNAm)

##### 2.3 USGS ####
# The USGS dataset also partially occurs on GBIF from BISON, however, the occurrence codes are in
# a silly place... We will correct these here to help identify duplicates later
db_standardized <- db_standardized %>%
    # Remove the discoverlife html if it is from USGS
  dplyr::mutate(occurrenceID = dplyr::if_else(
    stringr::str_detect(occurrenceID, "USGS_DRO"),
    stringr::str_remove(occurrenceID, "http://www\\.discoverlife\\.org/mp/20l\\?id="),
    occurrenceID)) %>%
    # Use otherCatalogNumbers when occurrenceID is empty AND when USGS_DRO is detected there
  dplyr::mutate(
    occurrenceID = dplyr::if_else(
      stringr::str_detect(otherCatalogNumbers, "USGS_DRO") & is.na(occurrenceID),
      otherCatalogNumbers, occurrenceID)) %>%
    # Make sure that no eventIDs have snuck into the occurrenceID columns 
      # For USGS_DRO, codes with <6 digits are event ids
  dplyr::mutate(
    occurrenceID = dplyr::if_else(stringr::str_detect(occurrenceID, "USGS_DRO", negate = TRUE),
        # Keep occurrenceID if it's NOT USGS_DRO
       occurrenceID, 
        # If it IS USGS_DRO and it has => 6 numbers, keep it, else, NA
      dplyr::if_else(stringr::str_detect(occurrenceID, "USGS_DRO[0-9]{6,10}"),
                     occurrenceID, NA_character_)),
    catalogNumber = dplyr::if_else(stringr::str_detect(catalogNumber, "USGS_DRO", negate = TRUE),
        # Keep catalogNumber if it's NOT USGS_DRO
      catalogNumber, 
        # If it IS USGS_DRO and it has => 6 numbers, keep it, else, NA
      dplyr::if_else(stringr::str_detect(catalogNumber, "USGS_DRO[0-9]{6,10}"),
                     catalogNumber, NA_character_)))


##### 2.4 Additional datasets ####
  # Import additional and potentially private datasets.
    # Private dataset functions are provided but the data not integrated here until those datasets 
    # become freely available post-publication.
  # There will be some warnings were a few rows may not be formatted correctly or where dates fail
    # to parse. This is normal.
###### a. EPEL ####
  # Guzman, L. M., Kelly, T. & Elle, E. A data set for pollinator diversity and their interactions 
  # with plants in the Pacific Northwest. Ecology n/a, e3927 (2022). 
  # <https://doi.org/10.1002/ecy.3927>
EPEL_Data <- BeeBDC::readr_BeeBDC(dataset = "EPEL",
                                path = paste0(DataPath, "/Additional_Datasets"),
                      inFile = "/InputDatasets/bee_data_canada.csv",
                      outFile = "jbd_EPEL_data.csv",
                      dataLicense = "https://creativecommons.org/licenses/by-nc-sa/4.0/")
###### b. Allan Smith-Pardo ####
ASP_Data <- BeeBDC::readr_BeeBDC(dataset = "ASP",
                               path = paste0(DataPath, "/Additional_Datasets"),
                      inFile = "/InputDatasets/Allan_Smith-Pardo_Dorey_ready2.csv",
                      outFile = "jbd_ASP_data.csv",
                      dataLicense = "https://creativecommons.org/licenses/by-nc-sa/4.0/")
###### c. Minckley ####
BMin_Data <- BeeBDC::readr_BeeBDC(dataset = "BMin",
                                path = paste0(DataPath, "/Additional_Datasets"),
                        inFile = "/InputDatasets/Bob_Minckley_6_1_22_ScanRecent-mod_Dorey.csv",
                        outFile = "jbd_BMin_data.csv",
                        dataLicense = "https://creativecommons.org/licenses/by-nc-sa/4.0/")
###### d. BMont ####
  # Delphia, C. M. Bumble bees of Montana. 
  # <https://www.mtent.org/projects/Bumble_Bees/bombus_species.html>. (2022)
BMont_Data <- BeeBDC::readr_BeeBDC(dataset = "BMont",
                                 path = paste0(DataPath, "/Additional_Datasets"),
                          inFile = "/InputDatasets/Bombus_Montana_dorey.csv",
                          outFile = "jbd_BMont_data.csv",
                          dataLicense = "https://creativecommons.org/licenses/by-sa/4.0/")
###### e. Ecd ####
  # Ecdysis. Ecdysis: a portal for live-data arthropod collections, 
  # <https://serv.biokic.asu.edu/ecdysis/index.php> (2022).
Ecd_Data <- BeeBDC::readr_BeeBDC(dataset = "Ecd",
                               path = paste0(DataPath, "/Additional_Datasets"),
                      inFile = "/InputDatasets/Ecdysis_occs.csv",
                      outFile = "jbd_Ecd_data.csv",
                      dataLicense = "https://creativecommons.org/licenses/by-nc-sa/4.0/")
###### f. Gai ####
  # Gaiarsa, M. P., Kremen, C. & Ponisio, L. C. Pollinator interaction flexibility across scales 
  # affects patch colonization and occupancy. Nature Ecology & Evolution 5, 787-793 (2021). 
  # <https://doi.org/10.1038/s41559-021-01434-y>
Gai_Data <- BeeBDC::readr_BeeBDC(dataset = "Gai",
                               path = paste0(DataPath, "/Additional_Datasets"),
                      inFile = "/InputDatasets/upload_to_scan_Gaiarsa et al_Dorey.csv",
                      outFile = "jbd_Gai_data.csv",
                      dataLicense = "https://creativecommons.org/licenses/by-nc-sa/4.0/")
###### g. CAES ####
# This is a little slower and will have a few date warnings - 215 failed to parse. 
CAES_Data <- BeeBDC::readr_BeeBDC(dataset = "CAES",
                                path = paste0(DataPath, "/Additional_Datasets"),
                        inFile = "/InputDatasets/CT_BEE_DATA_FROM_PBI.xlsx",
                        outFile = "jbd_CT_Data.csv",
                        sheet = "Sheet1",
                        dataLicense = "https://creativecommons.org/licenses/by-nc-sa/4.0/")

###### h. GeoL ####
GeoL_Data <- BeeBDC::readr_BeeBDC(dataset = "GeoL",
                                path = paste0(DataPath, "/Additional_Datasets"),
                        inFile = "/InputDatasets/Geolocate and BELS_certain and accurate.xlsx",
                        outFile = "jbd_GeoL_Data.csv",
                        dataLicense = "https://creativecommons.org/licenses/by-nc-sa/4.0/")


###### i. EaCO ####
EaCO_Data <- BeeBDC::readr_BeeBDC(dataset = "EaCO",
                                path = paste0(DataPath, "/Additional_Datasets"),
                        inFile = "/InputDatasets/Eastern Colorado bee 2017 sampling.xlsx",
                        outFile = "jbd_EaCo_Data.csv",
                        dataLicense = "https://creativecommons.org/licenses/by-nc-sa/4.0/")


###### j. FSCA ####
  # Florida State Collection of Arthropods
FSCA_Data <- BeeBDC::readr_BeeBDC(dataset = "FSCA",
                                path = paste0(DataPath, "/Additional_Datasets"),
                        inFile = "InputDatasets/fsca_9_15_22_occurrences.csv",
                        outFile = "jbd_FSCA_Data.csv",
                        dataLicense = "https://creativecommons.org/licenses/by-nc-sa/4.0/")

###### k. Texas SMC ####
# # Published or unpublished data from Texas literature not in an online database, usually copied 
# into spreadsheet from document format, or otherwise copied from a very differently-formatted spreadsheet
# # Unpublished or partially published data were obtained with express permission from the lead author
SMC_Data <- BeeBDC::readr_BeeBDC(dataset = "SMC",
                               path = paste0(DataPath, "/Additional_Datasets"),
                      inFile = "/InputDatasets/TXbeeLitOccs_31Oct22.csv", 
                      outFile = "jbd_SMC_Data.csv",
                      dataLicense = "https://creativecommons.org/licenses/by-nc-sa/4.0/")

###### l. Texas Bal ####
# # Data with GPS coordinates from Ballare et al. 2019, https://doi.org/10.1002/eap.1869. 
# The version on Dryad is missing site GPS coordinates (by accident). Kim is okay with these data 
# being made public as long as her paper is referenced.
# - Elinor Lichtenberg
Bal_Data <- BeeBDC::readr_BeeBDC(dataset = "Bal",
                               path = paste0(DataPath, "/Additional_Datasets"),
                      inFile = "/InputDatasets/Beedata_ballare.xlsx", 
                      outFile = "jbd_Bal_Data.csv",
                      sheet = "animal_data",
                      dataLicense = "https://creativecommons.org/licenses/by-nc-sa/4.0/")

###### m. Palouse Lic ####
# # *Attached: My canola data. I tried to get this in DarwinCore format. 
# These data go with the manuscript published here: 
# https://doi.org/10.1111/jen.13188 
# These are the data I will be putting on SCAN. 
# - Elinor Lichtenberg
Lic_Data <- BeeBDC::readr_BeeBDC(dataset = "Lic",
                               path = paste0(DataPath, "/Additional_Datasets"),
                      inFile = "/InputDatasets/Lichtenberg_canola_records.csv", 
                      outFile = "jbd_Lic_Data.csv",
                      dataLicense = "https://creativecommons.org/licenses/by-nc-sa/4.0/")

###### n. Arm ####
Arm_Data <- BeeBDC::readr_BeeBDC(dataset = "Arm",
                               path = paste0(DataPath, "/Additional_Datasets"),
                      inFile = "/InputDatasets/Bee database Armando_Final.xlsx",
                      outFile = "jbd_Arm_Data.csv",
                      sheet = "Sheet1",
                      dataLicense = "https://creativecommons.org/licenses/by-nc-sa/4.0/")

###### o. Dor #####
Dor_Data <- BeeBDC::readr_BeeBDC(dataset = "Dor",
                               path = paste0(DataPath, "/Additional_Datasets"),
                      inFile = "/InputDatasets/DoreyData.csv",
                      outFile = "jbd_Dor_Data.csv",
                      dataLicense = "https://creativecommons.org/licenses/by-nc-sa/4.0/")

###### p. VicWam #####
VicWam_Data <- BeeBDC::readr_BeeBDC(dataset = "VicWam",
                                 path = paste0(DataPath, "/Additional_Datasets"),
                                 inFile = "/InputDatasets/Combined_Vic_WAM_databases.xlsx",
                                 outFile = "jbd_VicWam_Data.csv",
                                 dataLicense = "https://creativecommons.org/licenses/by-nc-sa/4.0/",
                                 sheet = "Combined")


##### 2.5 Merge all ####
# Remove these spent datasets
rm(EPEL_Data, ASP_Data, BMin_Data, BMont_Data, Ecd_Data, Gai_Data, CAES_Data, 
   # INHS_Data, MABC_Data, EcoS_Data, KP_Data,
   GeoL_Data, EaCO_Data, FSCA_Data, SMC_Data, Bal_Data, Lic_Data, Arm_Data, Dor_Data,
   VicWam_Data)
# Read in and merge all
db_standardized <- db_standardized %>%
  dplyr::bind_rows(
    readr::read_csv(paste0(DataPath, "/Additional_Datasets", 
                           "/jbd_ASP_data.csv"), col_types = BeeBDC::ColTypeR()),
    readr::read_csv(paste0(DataPath, "/Additional_Datasets", 
                           "/jbd_EPEL_data.csv"), col_types = BeeBDC::ColTypeR()),
    readr::read_csv(paste0(DataPath, "/Additional_Datasets", 
                           "/jbd_BMin_data.csv"), col_types = BeeBDC::ColTypeR()),
    readr::read_csv(paste0(DataPath, "/Additional_Datasets", 
                           "/jbd_BMont_data.csv"), col_types = BeeBDC::ColTypeR()),
    readr::read_csv(paste0(DataPath, "/Additional_Datasets", 
                           "/jbd_Ecd_data.csv"), col_types = BeeBDC::ColTypeR()),
    readr::read_csv(paste0(DataPath, "/Additional_Datasets", 
                           "/jbd_Gai_data.csv"), col_types = BeeBDC::ColTypeR()),
    readr::read_csv(paste0(DataPath, "/Additional_Datasets", 
                           "/jbd_CT_Data.csv"), col_types = BeeBDC::ColTypeR()),
     readr::read_csv(paste0(DataPath, "/Additional_Datasets", 
                           "/jbd_GeoL_Data.csv"), col_types = BeeBDC::ColTypeR()),
    readr::read_csv(paste0(DataPath, "/Additional_Datasets", 
                           "/jbd_EaCo_Data.csv"), col_types = BeeBDC::ColTypeR()), 
    readr::read_csv(paste0(DataPath, "/Additional_Datasets", 
                           "/jbd_SMC_Data.csv"), col_types = BeeBDC::ColTypeR()),
    readr::read_csv(paste0(DataPath, "/Additional_Datasets", 
                           "/jbd_Bal_Data.csv"), col_types = BeeBDC::ColTypeR()),
    readr::read_csv(paste0(DataPath, "/Additional_Datasets", 
                           "/jbd_Lic_Data.csv"), col_types = BeeBDC::ColTypeR()),
    readr::read_csv(paste0(DataPath, "/Additional_Datasets", 
                           "/jbd_Arm_Data.csv"), col_types = BeeBDC::ColTypeR()),
    readr::read_csv(paste0(DataPath, "/Additional_Datasets", 
                           "/jbd_Dor_Data.csv"), col_types = BeeBDC::ColTypeR()),
    readr::read_csv(paste0(DataPath, "/Additional_Datasets", 
                           "/jbd_VicWam_Data.csv"), col_types = BeeBDC::ColTypeR())) %>% 
  # END bind_rows
  suppressWarnings(classes = "warning") # End suppressWarnings - due to col_types

##### 2.6 Match database_id ####
# Try to match database IDs with prior runs.
# read in a prior run of choice
priorRun <- fileFinder(path = DataPath,
                       file = "01_prefilter_database_9Aug22.csv") %>%
  readr::read_csv(file = ., col_types = ColTypeR())

# attempt to match records using the below function
# This function will attempt to find the database_ids from prior runs
# source(paste(ScriptPath, "idMatchR.R", sep = "/"))
db_standardized <- idMatchR(
  currentData = db_standardized,
  priorData = priorRun,
  # First matches will be given preference over later ones
  matchBy = tibble::lst(c("gbifID", "dataSource"),
                        c("catalogNumber", "institutionCode", "dataSource", "decimalLatitude",
                          "decimalLongitude"),
                        c("occurrenceID", "dataSource","decimalLatitude","decimalLongitude"),
                        c("recordId", "dataSource","decimalLatitude","decimalLongitude"),
                        c("id", "dataSource","decimalLatitude","decimalLongitude"),
                        # Because INHS was entered as it's own dataset but is now included in the GBIF download...
                        c("catalogNumber", "institutionCode", "dataSource",
                          "decimalLatitude","decimalLongitude")),
  # You can exclude datasets from prior by matching their prefixs - before first underscore:
  excludeDataset = c("ASP", "BMin", "BMont", "CAES", "EaCO", "Ecd", "EcoS",
                     "Gai", "KP", "EPEL", "CAES", "EaCO", "FSCA", "SMC", "Lic", "Arm",
                     "VicWam"))

# Remove redundant files
rm(priorRun)


#   # Save the dataset
db_standardized %>%
  readr::write_excel_csv(.,
                   paste(OutPath_Intermediate, "00_prefilter_database.csv",
                         sep = "/"))

#### 3.0 Initial flags ####
  # Read this back in if needed.
if(!exists("db_standardized")){
  db_standardized <- readr::read_csv(paste(OutPath_Intermediate, "00_prefilter_database.csv",
                                    sep = "/"), col_types = BeeBDC::ColTypeR())}

# See here for bdc prefilter tutorial - https://brunobrr.github.io/bdc/articles/prefilter.html
##### 3.1 SciName ####
  # Flag occurrences without scientificName provided
check_pf <- bdc::bdc_scientificName_empty(
  data = db_standardized,
  sci_name = "scientificName")

# now that this is saved, remove it to save space in memory
rm(db_standardized) 


##### 3.2 MissCoords ####
# Flag occurrences with missing lat and lon
check_pf <- bdc::bdc_coordinates_empty(
  data = check_pf,
  lat = "decimalLatitude",
  lon = "decimalLongitude")

##### 3.3 OutOfRange ####
# Flag occurrences that are not on earth (outside of -180 to 180 or -90 to 90 degrees)
check_pf <- bdc::bdc_coordinates_outOfRange(
  data = check_pf,
  lat = "decimalLatitude",
  lon = "decimalLongitude")

##### 3.4 ?Source ####
# Flag occurrences that don't match the basisOfRecord types below
check_pf <- bdc::bdc_basisOfRecords_notStandard(
  data = check_pf,
  basisOfRecord = "basisOfRecord",
  names_to_keep = c(
    # Keep all plus some at the bottom.
    "Event",
    "HUMAN_OBSERVATION",
    "HumanObservation",
    "LIVING_SPECIMEN",
    "LivingSpecimen",
    "MACHINE_OBSERVATION",
    "MachineObservation",
    "MATERIAL_SAMPLE",
    "O",
    "Occurrence",
    "MaterialSample",
    "OBSERVATION",
    "Preserved Specimen",
    "PRESERVED_SPECIMEN",
    "preservedspecimen Specimen",
    "Preservedspecimen",
    "PreservedSpecimen",
    "preservedspecimen",
    "pinned",
    "carded/pointed",
    "S",
    "Specimen",
    "Taxon",
    "UNKNOWN",
    "",
    NA,
    "NA",
    "LITERATURE", 
    "None", "Pinned Specimen", "Voucher reared", "Emerged specimen"
  ))


##### 3.5 CountryName ####
  # Try to harmonise country names
###### a. prepare dataset ####
  # Fix up country names based on common problems above and extract ISO2 codes for occurrences
check_pf_noNa <- BeeBDC::countryNameCleanR(
  data = check_pf,
  # Create a Tibble of common issues in country names and their replacements
  commonProblems = dplyr::tibble(problem = c('U.S.A.', 'US','USA','usa','UNITED STATES',
                                              'United States','U.S.A','MX','CA','Bras.','Braz.',
                                              'Brasil','CNMI','USA TERRITORY: PUERTO RICO'),
                                  fix = c('United States of America','United States of America',
                                          'United States of America','United States of America',
                                          'United States of America','United States of America',
                                          'United States of America','Mexico','Canada','Brazil',
                                          'Brazil','Brazil','Northern Mariana Islands','PUERTO.RICO'))
  )


###### b. run function ####
  # Get country name from coordinates using a wrapper around the bdc_country_from_coordinates function
  # Because our dataset is much larger than those used to design bdc, we have made it so that you
  # can analyse data in smaller pieces.
suppressWarnings(
  countryOutput <- BeeBDC::jbd_CfC_chunker(data = check_pf_noNa,
                                   lat = "decimalLatitude",
                                   lon = "decimalLongitude",
                                   country = "country",
                                   # How many rows to process at a time
                                   stepSize = 1000000,
                                   # Start row
                                   chunkStart = 1,
                                   path = OutPath_Intermediate,
                                   scale = "medium",
                                   mc.cores = 2),
  classes = "warning")

###### c. re-merge ####
# Left join these datasets
check_pf <- dplyr::left_join(check_pf, 
                             countryOutput, 
                             by = "database_id",
                             suffix = c("", "CO"))  %>% 
  # Take the new country name if the original is NA
  dplyr::mutate(country = dplyr::if_else(is.na(country),
                                         countryCO,
                                         country)) %>%
  # Remove duplicates if they arose from left_join!
  dplyr::distinct()

# Save the dataset
check_pf %>%
  readr::write_excel_csv(.,
                   paste(OutPath_Intermediate, "01_prefilter_database.csv",
                         sep = "/"))
# Save the countryOutput dataset
countryOutput %>%
  readr::write_excel_csv(.,
                   paste(OutPath_Intermediate, "countryOutput.csv",
                         sep = "/"))
# Read in IF needed
if(!exists("check_pf")){
check_pf <- readr::read_csv(paste(DataPath, 
             "Output", "Intermediate", "01_prefilter_database.csv", sep = "/"),
             col_types = BeeBDC::ColTypeR())}
# remove the interim datasets
rm(check_pf_noNa, countryOutput)

##### 3.6 StandardCoNames ####
# Run the function
  # Standardise country names and add ISO2 codes if needed
check_pf <- bdc::bdc_country_standardized(
  # Remove the countryCode and country_suggested columns to avoid an error with 
    # where two "countryCode" and "country_suggested" columns exist (i.e. if the dataset has been  
    # run before)
  data = check_pf %>% dplyr::select(!tidyselect::any_of(c("countryCode", "country_suggested"))),
  country = "country"
) %>%
    # Sometimes when "MX" is the country, "Mexico" is not correctly assigned
  dplyr::mutate(country_suggested = dplyr::if_else(country == "MX",
                                                   "Mexico",
                                                   country_suggested))

##### 3.7 TranspCoords ####
  # Flag and correct records when lat and long appear to be transposed. We have chunked 
    # this because it is too RAM-heavy to run on our large dataset
check_pf <- BeeBDC::jbd_Ctrans_chunker(
  # bdc_coordinates_transposed inputs
  data = check_pf,
  id = "database_id",
  lat = "decimalLatitude",
  lon = "decimalLongitude",
  country = "country",
  countryCode = "countryCode",
  border_buffer = 0.2, # in decimal degrees (~22 km at the equator)
  save_outputs = TRUE,
  sci_names = "scientificName",
  # chunker inputs
  stepSize = 1000000,  # How many rows to process at a time
  chunkStart = 1,  # Start row
  append = FALSE,  # If FALSE it may overwrite existing dataset
  path = OutPath_Check,
  mc.cores = 4
) 

# Get a summary of the number of transposed records
table(check_pf$coordinates_transposed, useNA = "always")

# Save the dataset
check_pf %>%
  readr::write_excel_csv(.,
                   paste(OutPath_Intermediate, "01_prefilter_database.csv",
                         sep = "/"))
gc()

##### 3.8 Coord-country ####
# Read data in again if needed
if(!exists("check_pf")){
  check_pf <- readr::read_csv(paste(OutPath_Intermediate, "01_prefilter_database.csv",
                                    sep = "/"), col_types = BeeBDC::ColTypeR())}
# Collect all country names in the country column
# rebuilt a bdc function to flag occurrences where the coordinates are inconsistent with the provided
  # country name
check_pf <- BeeBDC::jbd_coordCountryInconsistent(
  data = check_pf,
  lon = "decimalLongitude",
  lat = "decimalLatitude",
  scale = 50,
  pointBuffer = 0.01) 

# Save the dataset
check_pf %>%
  readr::write_excel_csv(.,
                   paste(OutPath_Intermediate, "01_prefilter_database.csv",
                         sep = "/"))

##### 3.9 GeoRefIssue ####
# This function Identifies records whose coordinates can potentially be extracted from locality 
  # information must be manually checked later
xyFromLocality <- bdc::bdc_coordinates_from_locality(
  data = check_pf,
  locality = "locality",
  lon = "decimalLongitude",
  lat = "decimalLatitude",
  save_outputs = TRUE
) %>%
# Save data if needed.
  readr::write_excel_csv(paste(OutPath_Check, "01_coordinates_from_locality.csv",
                         sep = "/"))
# Remove spent data
rm(xyFromLocality)

##### 3.10 flag Absent ####
# Flag the records marked as "absent"
check_pf <- BeeBDC::flagAbsent(data = check_pf,
                   PresAbs = "occurrenceStatus")

##### 3.11 flag License ####
# Flag the records that may not be used according to their license information
check_pf <- BeeBDC::flagLicense(data = check_pf,
                    strings_to_restrict = "all",
                    # DON'T flag if in the following dataSource(s)
                    excludeDataSource = NULL)

##### 3.12 GBIF issue ####
# Flag select issues that are flagged by GBIF
check_pf <- BeeBDC::GBIFissues(data = check_pf, 
                   issueColumn = "issue", 
                   GBIFflags = c("COORDINATE_INVALID", "ZERO_COORDINATE")) 


##### 3.13 Flag Reports ####
###### a. save flags ####
# SAVE the flags so far
# This function will make sure that you keep a copy of everything that has been flagged up until now.
# This will be updated throughout the script and accessed at the end, so be wary of moving files around manually.
flagFile <- BeeBDC::flagRecorder(
  data = check_pf,
  outPath = paste(OutPath_Report, sep =""),
  fileName = paste0("flagsRecorded_", Sys.Date(),  ".csv"),
    # These are the columns that will be kept along with the flags
  idColumns = c("database_id", "id", "catalogNumber", "occurrenceID", "dataSource"),
    # TRUE if you want to find a file from a previous part of the script to append to
  append = FALSE)

# produce the .summary column in main dataset - will be FALSE if ANY .filtering column is FALSE
check_pf <- BeeBDC::summaryFun(
  data = check_pf,
    # Don't filter these columns (or NULL)
  dontFilterThese = NULL,
    # Remove the filtering columns?
  removeFilterColumns = FALSE,
    # Filter to ONLY cleaned data?
  filterClean = FALSE)

###### b. reporting ####
(report <- bdc::bdc_create_report(data = check_pf,
                                  database_id = "database_id",
                                  workflow_step = "prefilter",
                                  save_report = TRUE)
)

###### c. figures ####
figures <-
  bdc::bdc_create_figures(data = check_pf,
                          database_id = "database_id",
                          workflow_step = "prefilter",
                          save_figures = TRUE)
# You can check figures using
figures$.coordinates_country_inconsistent

##### 3.14 save ####
# Save the intermediate dataset
check_pf %>%
  readr::write_excel_csv(., paste(OutPath_Intermediate, "01_prefilter_output.csv",
                            sep = "/"))

#### 4.0 Taxonomy ####
# See bdc tutorial here - https://brunobrr.github.io/bdc/articles/taxonomy.html
if(!exists("check_pf")){
# Read in the filtered dataset
database <-
  readr::read_csv( paste(OutPath_Intermediate, "01_prefilter_output.csv",
                         sep = "/"), col_types = BeeBDC::ColTypeR())
}else{
    # OR rename and remove
  database <- check_pf
  # Remove spent dataset
  rm(check_pf)}

# Remove names_clean if it already exists (i.e. you have run this before on this dataset)
database <- database %>%
  dplyr::select(!tidyselect::any_of("names_clean"))

##### 4.1 Prep data names ####
# This next step cleans the database's scientificName column. The cleaning tool will:
# Flag and remove family names pre-pended to species names;
# Flag and remove qualifiers denoting uncertain or provisional status of taxonomic identification (e.g., confer, species, affinis, and among others);
# Flag and remove infraspecific terms (e.g., variety [var.], subspecies [subsp], forma [f.], and their spelling variations);
# Standardize names, i.e., capitalize only the first letter of the genus name and remove extra whitespaces);
# Parse names, i.e., separate author, date, annotations from taxon name.
# ! MAC: You need to install gnparser through terminal - brew
  # brew tap gnames/gn
  # brew install gnparser
parse_names <-
  bdc::bdc_clean_names(sci_names = database$scientificName, save_outputs = TRUE)

# Keep only the .uncer_terms and names_clean columns
parse_names <-
  parse_names %>%
  dplyr::select(.uncer_terms, names_clean)
# Merge names with the complete dataset
database <- dplyr::bind_cols(database, parse_names)
rm(parse_names)


##### 4.2 Harmonise taxonomy ####
# Download the custom taxonomy file
taxonomyFile <- BeeBDC::beesTaxonomy(URL = "https://figshare.com/ndownloader/files/42402264?private_link=bce1f92848c2ced313ee")


# Harmonise the names in the occurrence tibble
#   # This flags the occurrences without a matched name and matches names to their correct name 
  # according to DiscoverLife
database <- BeeBDC::harmoniseR(path = DataPath, #The path to a folder that the output can be saved
                       taxonomy = taxonomyFile, # The formatted taxonomy file
                       data = database,
                       speciesColumn = "scientificName",
                       stepSize = 1000000,
                       mc.cores = 1)

# You don't need this file anymore...
rm(taxonomyFile)

# Save the harmonised file.
database %>%
  readr::write_excel_csv(.,
                   paste(OutPath_Intermediate, "02_taxonomy_database.csv",
                         sep = "/"))

##### 4.3 Save flags ####
# SAVE the flags so far
# This will find the most-recent flag file and append your new data to it.
# You can double-check the data and number of columns if you'd like to be thorough and sure that 
# all data are intact. <3 
flagFile <- BeeBDC::flagRecorder(
  data = database,
  outPath = paste(OutPath_Report, sep =""),
  fileName = paste0("flagsRecorded_", Sys.Date(),  ".csv"),
  idColumns = c("database_id", "id", "catalogNumber", "occurrenceID", "dataSource"),
  append = TRUE,
  printSummary = TRUE)


#### 5.0 Space #### 
# the final frontier or whatever.
# Read in the last database
if(!exists("database")){
database <-
  readr::read_csv(paste(OutPath_Intermediate, "02_taxonomy_database.csv", sep = "/"),
                  col_types = BeeBDC::ColTypeR())}

##### 5.1 Coord precision ####
# This function identifies records with a coordinate precision below a specified number of decimal 
# places. For example, the precision of a coordinate with 1 decimal place is 11.132 km at the 
# equator, i.e., the scale of a large city.
# "Coordinates with one, two, or three decimal places present a precision of
# ~11.1 km, ~1.1 km, and ~111 m at the equator, respectively."
# This function differs from the bdc function by ONLY flagging occurrences where BOTH lat and lon
    # are rounded (having only one or the other rounded could be due to rounding in excel).
check_space <-
  BeeBDC::jbd_coordinates_precision(
    data = database,
    lon = "decimalLongitude",
    lat = "decimalLatitude",
    ndec = 2 # number of decimals to be tested
  )
# Remove the spent dataset
rm(database)

# Save the harmonised file.
check_space %>%
  readr::write_excel_csv(.,
                   paste(OutPath_Intermediate, "03_space_inter_database.csv",
                         sep = "/"))

if(!exists("check_space")){
  # Read in the filtered dataset
  check_space <-
    readr::read_csv( paste(OutPath_Intermediate, "03_space_inter_database.csv",
                           sep = "/"), col_types = BeeBDC::ColTypeR())
}

##### 5.2 Common spatial issues ####
# Only run for occurrences through clean_coordinates that are spatially 'valid'.
# Not doing this might crash R.
tempSpace <- check_space %>% 
  dplyr::filter(!.coordinates_empty == FALSE) %>%
  dplyr::filter(!.coordinates_outOfRange == FALSE) %>%
# Next, we will flag common spatial issues using functions of the package CoordinateCleaner.
# Addresses some common issues in biodiversity datasets
  CoordinateCleaner::clean_coordinates(
    x =  .,
    lon = "decimalLongitude",
    lat = "decimalLatitude",
    species = "scientificName",
    countries = NULL, # Tests if coords are from x country. This is not needed.
    tests = c(
      "capitals",     # records within 0.5 km of capitals centroids
      "centroids",    # records within 1 km around country and province centroids
      "equal",      # records with equal coordinates
      "gbif",         # records within 1 km of GBIF headquarters. (says 1 degree in package, but code says 1000 m)
      "institutions", # records within 100m of zoo and herbaria
      "zeros"       # records with coordinates 0,0
      # "seas"        # Not flagged as this should be flagged by coordinate country inconsistent
    ),
    capitals_rad = 1000,
    centroids_rad = 500,
    centroids_detail = "both", # test both country and province centroids
    inst_rad = 100, # remove zoo and herbaria within 100m
    range_rad = 0,
    zeros_rad = 0.5,
    capitals_ref = NULL,
    centroids_ref = NULL,
    country_ref = NULL,
    country_refcol = "countryCode",
    inst_ref = NULL,
    range_ref = NULL,
    # seas_scale = 50,
    value = "spatialvalid" # result of tests are appended in separate columns
  ) %>%
    # Remove duplicate .summary column that can be replaced later and turn into a tibble
  dplyr::select(!tidyselect::starts_with(".summary")) %>%
  dplyr::tibble() 
# re-merge the datasets
check_space <- tempSpace %>%
  # Re-bind with the records that were excluded earlier
  dplyr::bind_rows(check_space %>% 
                     dplyr::filter(.coordinates_empty == FALSE | 
                                     .coordinates_outOfRange == FALSE) )
# Remove the temporary dataset
rm(tempSpace)

  # Save the intermediate dataset
check_space %>%
  readr::write_excel_csv(paste(OutPath_Intermediate, "03_space_inter_database.csv",
                         sep = "/"))

if(!exists("check_space")){
  check_space <- readr::read_csv(paste(OutPath_Intermediate, "03_space_inter_database.csv",
                                       sep = "/"),
                                 col_types = BeeBDC::ColTypeR())
}


##### 5.3 Diagonal + grid ####
# Finds sequential numbers that could be fill-down errors in lat and long. 
# groups by eventDate, recordedBy
# This is accomplished by using a sliding window with the length determined by minRepeats.
check_space <- BeeBDC::diagonAlley(
  data = check_space,
  # The minimum number of repeats needed to find a sequence in for flagging
  minRepeats = 6,
  mc.cores = 4)

# SPATIAL gridding from rasterisation:
# Select only the records with more than X occurrences
griddingDF <- check_space %>%
  # Exclude NA lat and lon values
  tidyr::drop_na(c("decimalLatitude", "decimalLongitude")) %>%
  # Group by the dataset name
  dplyr::group_by(datasetName) %>%
  # Remove rows that aren't unique for lat and long
  dplyr::distinct(decimalLongitude, decimalLatitude,
                  .keep_all = TRUE) %>%
  # Find the groups with 4 or more occurrence records 
  dplyr::filter(dplyr::n() >= 4) %>%
  dplyr::ungroup()

# Run the gridding analysis to find datasets that might be gridded
gridded_datasets <- CoordinateCleaner::cd_round(
  x = griddingDF,
  lon = "decimalLongitude",
  lat = "decimalLatitude",
  ds = "datasetName",
  T1 = 7,
  min_unique_ds_size = 4,
  test = "both",
  value = "dataset",
  graphs = FALSE,
  verbose = TRUE,
  reg_out_thresh = 2,
  reg_dist_min = 0.1,
  reg_dist_max = 2
) %>% 
  dplyr::tibble()
# The griddingDF is no longer needed. remove it.
rm(griddingDF)

# Integrate these results with the main dataset
check_space <- check_space %>%
  # Join the datasets
  dplyr::left_join(
    # Select the columns of interest
    dplyr::select(gridded_datasets, dataset, lon.flag, lat.flag, summary),
    by = c("datasetName" = "dataset")) %>%
  # Make new columns with more-consistent naming and change the NA vlaues to = TRUE (not flagged)
  dplyr::mutate(.lonFlag = tidyr::replace_na(lon.flag, TRUE),
                .latFlag = tidyr::replace_na(lat.flag, TRUE),
                .gridSummary = tidyr::replace_na(summary, TRUE)) %>%
  # Remove old columns
  dplyr::select(!c(lon.flag, lat.flag, summary))

# Save the gridded_datasets file for later examination
gridded_datasets %>%
  readr::write_excel_csv(paste(OutPath_Intermediate, "03_space_griddedDatasets.csv",
                         sep = "/"))
# Now remove this file
rm(gridded_datasets)

##### 5.4 Uncertainty ####
# Flag records that exceed a coordinateUncertaintyInMeters threshold
check_space <- BeeBDC::coordUncerFlagR(data = check_space,
                               uncerColumn = "coordinateUncertaintyInMeters",
                               threshold = 1000)

##### 5.5 Country + continent checklist ####
# Download the country-level checklist
beesChecklist <- BeeBDC::beesChecklist()

  # Flag records for country outliers
check_space <- countryOutlieRs(checklist = beesChecklist,
                        data = check_space,
                        keepAdjacentCountry = TRUE,
                        pointBuffer = 0.05,
                          # Scale of map to return, one of 110, 50, 10 OR 'small', 'medium', 'large'
                          # Smaller numbers will result in much longer calculation times. 
                          # We have not attempted a scale of 10.
                        scale = 50,
                        mc.cores = 1)

  # Flag records for continent outliers
check_space <- continentOutlieRs(checklist = beesChecklist,
                               data = check_space,
                               keepAdjacentContinent = FALSE,
                               pointBuffer = 0.05,
                               # Scale of map to return, one of 110, 50, 10 OR 'small', 'medium', 'large'
                               # Smaller numbers will result in much longer calculation times. 
                               # We have not attempted a scale of 10.
                               scale = 50,
                               mc.cores = 1)

  # A list of failed species-country combinations and their numbers can be output here
check_space %>%
  dplyr::filter(.countryOutlier == FALSE) %>%
  dplyr::select(database_id, scientificName, country) %>%
  dplyr::group_by(scientificName) %>% 
  dplyr::mutate(count_scientificName = dplyr::n()) %>%
  dplyr::distinct(scientificName, country, .keep_all = TRUE) %>% 
  readr::write_excel_csv(paste(OutPath_Intermediate, "03_space_failedCountryChecklist.csv",
                         sep = "/"))


##### 5.6 Map spatial errors ####
#rebuild the .summary column
check_space <- BeeBDC::summaryFun(
  data = check_space,
  dontFilterThese = NULL,
  removeFilterColumns = FALSE,
  filterClean = FALSE)
# Map ONE spatial flag at a time or map the .SUMMARY of all
# Make this selection in the col_to_map = section
check_space %>%
  dplyr::filter(.summary == FALSE) %>% # map only records flagged as FALSE
  bdc::bdc_quickmap(
    data = .,
    lon = "decimalLongitude",
    lat = "decimalLatitude",
    col_to_map = ".summary",
    size = 0.9
  )

##### 5.7 Space report ####
# Create the report
(report <-
   bdc::bdc_create_report(
     data = dplyr::tibble(check_space %>% dplyr::select(!.uncer_terms)),
     database_id = "database_id",
     workflow_step = "space",
     save_report = TRUE)
)

##### 5.8 Space figures ####
# Sadly not a figure of outer space :( 
# Create figures of spacial data filtering
(figures <-
    BeeBDC::jbd_create_figures(
      data = dplyr::tibble(check_space %>% dplyr::select(!.uncer_terms)),
      path = DataPath,
      database_id = "database_id",
      workflow_step = "space",
      save_figures = TRUE)
)

# Check figures using
# options are:
# .cap == Records around country capital centroid	
# .cen == Records around country or province centroids
# .dbl == Duplicated coordinates per species
# .equ == Identical coordinates
# .otl == Geographical outliers
# .gbf == Records around the GBIF headquarters
# .inst == Records around biodiversity institutions
# .rou == Rounded (probably imprecise) coordinates
# .urb == Records within urban areas - Not relevant for bees, I think.
# Example:
figures$.rou

# Save interim dataset
check_space %>%
  readr::write_excel_csv(paste(OutPath_Intermediate, "03_space_inter_database.csv",
                         sep = "/"))

##### 5.9 Save flags ####
# SAVE the flags so far
BeeBDC::flagRecorder(
  data = check_space,
  outPath = paste(OutPath_Report, sep =""),
  fileName = paste0("flagsRecorded_", Sys.Date(),  ".csv"),
  idColumns = c("database_id", "id", "catalogNumber", "occurrenceID", "dataSource"),
  append = TRUE,
  printSummary = TRUE)

##### 5.10 Save ####
  # Save the intermediate dataset
check_space %>%
  readr::write_excel_csv(.,
                   paste(OutPath_Intermediate, "03_space_database.csv",
                         sep = "/"))



#### 6.0 Time ####
# Read in the last database
if(!exists("check_space")){
  check_time <-
    readr::read_csv(paste(OutPath_Intermediate, "03_space_database.csv", sep = "/"),
                    col_types = BeeBDC::ColTypeR())
  }else{
  check_time <- check_space
      # Remove the spent file
  rm(check_space)}

  # You can plot a histogram of dates here, pre-cleaning to examine potential issues
hist(lubridate::ymd_hms(check_time$eventDate, truncated = 5), breaks = 20)
# Filter some silly dates that don't make sense...
check_time$year <- ifelse(check_time$year > lubridate::year(Sys.Date()) | check_time$year < 1600,
                        NA, check_time$year)
check_time$month <- ifelse(check_time$month > 12 | check_time$month < 1,
                         NA, check_time$month)
check_time$day <- ifelse(check_time$day > 31 | check_time$day < 1,
                       NA, check_time$day)
##### 6.1 Recover dates ####
# RESCUE some records with poor date data if possible - e.g., from other columns
check_time <- BeeBDC::dateFindR(data = check_time,
                        # Years above this are removed (from the recovered dates only)
                        maxYear = lubridate::year(Sys.Date()),
                        # Years below this are removed (from the recovered dates only)
                        minYear = 1700)

##### 6.2 No eventDate ####
# Flag records that simply lack collection date :(
check_time <-
  bdc::bdc_eventDate_empty(data = check_time, eventDate = "eventDate")

##### 6.3 Old records ####
# This will flag records prior to the date selected. 1950 is frequently chosen for SDM work. You may
  # not need to filter old records at all. Please just think critically about your use
check_time <-
  bdc::bdc_year_outOfRange(data = check_time,
                           eventDate = "year",
                           year_threshold = 1950)

##### 6.4 Time report ####
# Not all of it, just the time pertaining to our precise occurrence records. Obviously...
# Create a .summary column with all of the time flags where TRUE == records that passed 
# all filtering.
check_time <- BeeBDC::summaryFun(
  data = check_time,
  # Don't filter these columns (or NULL)
  dontFilterThese = c(".gridSummary", ".lonFlag", ".latFlag", ".uncer_terms"),
  # Remove the filtering columns?
  removeFilterColumns = FALSE,
  # Filter to ONLY cleaned data?
  filterClean = FALSE)

( report <-
    bdc::bdc_create_report(data = check_time,
                           database_id = "database_id",
                           workflow_step = "time",
                           save_report = FALSE)
)  

##### 6.5 Time figures ####
# Create figures
figures <-
  BeeBDC::jbd_create_figures(data = check_time,
                     path = DataPath,
                     database_id = "database_id",
                     workflow_step = "time",
                     save_figures = TRUE)

# Check figures using
figures$year

# Save the ~raw time dataset into the intermediate folder
check_time %>%
  readr::write_excel_csv(.,
                   paste(OutPath_Intermediate, "04_time_database.csv",
                         sep = "/"))

##### 6.6 Save flags ####
# SAVE the flags so far
BeeBDC::flagRecorder(
  data = check_time,
  outPath = paste(OutPath_Report, sep =""),
  fileName = paste0("flagsRecorded_", Sys.Date(),  ".csv"),
  idColumns = c("database_id", "id", "catalogNumber", "occurrenceID", "dataSource"),
  append = TRUE,
  printSummary = TRUE)

#### 7.0 De-duplication ####
# Raw dataset can be re-read here if it does not already exist
if(!exists("check_time")){
  check_time <-
    readr::read_csv(paste(OutPath_Intermediate, "04_time_database.csv",
                          sep = "/"),
                    col_types = BeeBDC::ColTypeR())}

##### 7.1 deDuplicate ####
# We will FLAG duplicates here. 
# These input columns can be hacked to de-duplicate as you wish.
check_time <- BeeBDC::dupeSummary(
  data = check_time,
  path = OutPath_Report,
  # options are "ID","collectionInfo", or "both"
  duplicatedBy = "collectionInfo", 
  # The columns to generate completeness info from (and to sort by completness)
  completeness_cols = c("decimalLatitude",  "decimalLongitude",
                        "scientificName", "eventDate"),
  # idColumns = c("gbifID", "occurrenceID", "recordId","id"),
  # The columns to ADDITIONALLY consider when finding duplicates in collectionInfo
  collectionCols = c("decimalLatitude", "decimalLongitude", "scientificName", "eventDate", 
                     "recordedBy"),
  # The columns to combine, one-by-one with the collectionCols
  collectInfoColumns = c("catalogNumber", "otherCatalogNumbers"),
    # Custom comparisons - as a list of columns to compare
     # RAW custom comparisons do not use the character and number thresholds
  CustomComparisonsRAW = dplyr::lst(c("catalogNumber", "institutionCode", "scientificName")),
     # Other custom comparisons use the character and number thresholds
  CustomComparisons = dplyr::lst(c("gbifID", "scientificName"),
                                  c("occurrenceID", "scientificName"),
                                  c("recordId", "scientificName"),
                                  c("id", "scientificName")),
  # The order in which you want to KEEP duplicated based on data source
  # try unique(check_time$dataSource)
  sourceOrder = c("CAES", "Gai", "Ecd","BMont", "BMin", "EPEL", "ASP", "KP", "EcoS", "EaCO",
                  "FSCA", "Bal", "SMC", "Lic", "Arm",
                  "USGS", "ALA", "VicWam", "GBIF","SCAN","iDigBio"),
    # Paige ordering is done using the database_id prefix, not the dataSource prefix.
  prefixOrder = c("Paige", "Dorey"),
    # Set the complexity threshold for id letter and number length
     # minimum number of characters when WITH the numberThreshold
  characterThreshold = 2,
     # minimum number of numbers when WITH the characterThreshold
  numberThreshold = 3,
     # Minimum number of numbers WITHOUT any characters
  numberOnlyThreshold = 5
) %>% # END dupeSummary
  dplyr::as_tibble(col_types = BeeBDC::ColTypeR())

# Save the dataset into the intermediate folder
check_time %>%
  readr::write_excel_csv(.,
                   paste(OutPath_Intermediate, "04_2_dup_database.csv",
                         sep = "/"))

##### 7.2 Save flags ####
# SAVE the flags so far
BeeBDC::flagRecorder(
  data = check_time,
  outPath = paste(OutPath_Report, sep =""),
  fileName = paste0("flagsRecorded_", Sys.Date(),  ".csv"),
  idColumns = c("database_id", "id", "catalogNumber", "occurrenceID", "dataSource"),
  append = TRUE,
  printSummary = TRUE)


#### 8.0 Data filtering ####
# Raw dataset can be re-read here if it does not already exist
if(!exists("check_time")){
 check_time <-
   readr::read_csv(paste(OutPath_Intermediate, "04_2_dup_database.csv",
                          sep = "/"), col_types = ColTypeR())}
##### 8.1 rm Outliers ####
# Read in the most-recent duplicates file as well
if(!exists("duplicates")){
  duplicates <- fileFinder(path = DataPath,
                            fileName = "duplicateRun_") %>%
    readr::read_csv()}
# identify the outliers and get a list of their database_ids
check_time <- manualOutlierFindeR(
  data = check_time,
  DataPath = DataPath,
  PaigeOutliersName = "removedBecauseDeterminedOutlier.csv",
  newOutliersName = "^All_outliers_ANB_23Jan24.xlsx",
  ColombiaOutliers_all = "All_Colombian_OutlierIDs.csv",
  # A .csv with manual outlier records that are too close to otherwise TRUE records
  NearTRUE = "nearTRUE.csv",
  duplicates = duplicates)

##### 8.2 Save uncleaned ####
  # Make sure that the .summary column is updated
check_time <- summaryFun(
  data = check_time,
  dontFilterThese = c(".gridSummary", ".lonFlag", ".latFlag", ".uncer_terms",
                      ".uncertaintyThreshold", ".sequential", ".year_outOfRange"),
  removeFilterColumns = FALSE,
  filterClean = FALSE)
# Save the uncleaned dataset
check_time %>% readr::write_excel_csv(.,
                                paste(OutPath_Intermediate, "/05_unCleaned_database_",
                                      Sys.Date(), ".csv",
                                      sep = ""))

##### 8.3 Save cleaned ####
# Now clean the dataset of extra columns and failed rows and save it...
cleanData <- BeeBDC::summaryFun(
  data = check_time,
  dontFilterThese = c(".gridSummary", ".lonFlag", ".latFlag", ".uncer_terms",
                      ".uncertaintyThreshold", ".sequential", ".year_outOfRange"),
  # Remove the filtering columns?
  removeFilterColumns = TRUE,
  # Filter to ONLY cleaned data?
  filterClean = TRUE) 

# Save this CLEANED dataset
cleanData %>%
  readr::write_excel_csv(.,
                         paste(OutPath_Intermediate, "/05_cleaned_database_",
                               Sys.Date(), ".csv",
                               sep = ""))
# Dataset can be re-read here
    #  cleanData <-
    #    readr::read_csv(paste(OutPath_Intermediate, "05_cleaned_database.csv",
    #                           sep = "/"))


#### 9.0 Summary figures and tables ####
##### 9.1 Duplicate chordDiagrams ####
# install ComplexHeatmap if needed
if (!require("BiocManager", quietly = TRUE))
BiocManager::install("ComplexHeatmap")

# Read in the most-RECENT duplicates file.
if(!exists("duplicates")){
  duplicates <- fileFinder(path = DataPath,
                            fileName = "duplicateRun_") %>%
    readr::read_csv()}

# Choose the global figure parameters
par(mar = c(2, 2, 2, 2)/2, mfrow = c(1,1))
# Create the chorDiagram. You can leave many of the below values out but we show here
# the defaults
BeeBDC::chordDiagramR(
  # The duplicate data from the dupeSummary function output  
  dupeData = duplicates,
  outPath = OutPath_Figures,
  fileName = "Fig2_ChordDiagram.pdf",
  # These can be modified to help fit the final pdf that's exported.
  width = 9,
  height = 7.5,
  bg = "white",
  # How few distinct dataSources should a group have to be listed as "other"
  smallGrpThreshold = 3,
  title = "Duplicated record sources",
  # The default list of colour palettes to choose from usign the paleteer package
  palettes = c("cartography::blue.pal", "cartography::green.pal", 
               "cartography::sand.pal", "cartography::orange.pal", "cartography::red.pal",
               "cartography::purple.pal", "cartography::brown.pal"),
  canvas.ylim = c(-1.0,1.0), 
  canvas.xlim = c(-0.6, 0.25),
  text.col = "black",
  legendX = grid::unit(6, "mm"),
  legendY = grid::unit(18, "mm"),
  legendJustify = c("left", "bottom"),
  niceFacing = TRUE)


##### 9.2 Duplicate histogram ####
# Use the uncleaned dataset (read it in, or change its name to beeData)
if(!exists("check_time")){
beeData <- readr::read_csv(paste(OutPath_Intermediate, "05_unCleaned_database.csv",
                                 sep = "/"),
                           col_types = BeeBDC::ColTypeR())
}else{
  beeData <- check_time
  rm(check_time)
}
# Create a figure showing the total number of duplicates, kept duplicates, and unique
# records for each data source (simplified to the text before the first underscore) and
# the proportion of the above for each data source
BeeBDC::dupePlotR(
  data = beeData,
  # The outPath to save the plot as
  outPath = OutPath_Figures,
  fileName = "Fig3_duplicatePlot.pdf",
  # Colours in order: duplicate, kept duplicate, unique
  dupeColours = c("#F2D2A2","#B9D6BC", "#349B90"),
  # Plot size and height
  base_height = 7, base_width = 7,
  legend.position = c(0.85, 0.8),
  # Extra variables can be fed into forcats::fct_recode() to change names on plot
  GBIF = "GBIF", SCAN = "SCAN", iDigBio = "iDigBio", USGS = "USGS", ALA = "ALA", 
  ASP = "ASP", CAES = "CAES", 'BMont' = "BMont", 'BMin' = "BMin", Ecd = "Ecd",
  Gaiarsa = "Gai", EPEL = "EPEL", VicWam = "VicWam"
)


##### 9.3 Flags by source ####
# Visualise all flags for each dataSource (simplified to the text before the first underscore)
BeeBDC::plotFlagSummary(
  data = beeData,
  # Colours in order of pass (TRUE), fail (FALSE), and NA
  flagColours = c("#127852", "#A7002D", "#BDBABB"),
  fileName = paste0("Fig4_FlagsPlot_", Sys.Date(),".pdf"),
  outPath = paste0(OutPath_Figures),
  width = 15, height = 9,
    # OPTIONAL:
      #   # Filter to species
      #       speciesName = "Holcopasites heliopsis",
      #         # column to look in
      #       nameColumn = "species",
      #        # Save the filtered data
      #       saveFiltered = TRUE,
      #   # Filter column to display on map
      #       filterColumn = ".summary",
      #       plotMap = TRUE,
      #   # amount to jitter points if desired, e.g. 0.25 or NULL
      #       jitterValue = NULL,
      #        # Map opacity value for points between 0 and 1
      #   mapAlpha = 1,
      #        # If a user wants to output the table used to make the figure, change this to TRUE
      #   saveTable = FALSE,
  # Extra variables can be fed into forcats::fct_recode() to change names on plot
  GBIF = "GBIF", SCAN = "SCAN", iDigBio = "iDigBio", USGS = "USGS", ALA = "ALA", 
  ASP = "ASP", CAES = "CAES", 'BMont' = "BMont", 'BMin' = "BMin", Ecd = "Ecd",
  Gaiarsa = "Gai", EPEL = "EPEL", VicWam = "VicWam"
)


##### 9.4 Maps ####
# Import CLEANED dataset (you can change this option)
if(!exists("cleanData")){
cleanData <- readr::read_csv(paste(OutPath_Intermediate, "05_cleaned_database.csv",
                                 sep = "/"),
                           col_types = BeeBDC::ColTypeR())}

  ###### a. Summary maps ####
  # Draw a global summary map for occurrence and species number by country
BeeBDC::summaryMaps(
  data = cleanData,
  width = 10, height = 10,
  class_n = 15,
  class_Style = "fisher",
  fileName = "Fig5_CountryMaps_fisher.pdf",
  outPath = OutPath_Figures,
  scale = 110
)

  ###### b. Interactive maps ####
  # Generate a list of random species names to map and check
beeData_interactive <- beeData %>%
    # Select only valid species
  dplyr::filter(.invalidName == TRUE) %>%
    # Get a distinct list of valid species names
  dplyr::distinct(scientificName, .keep_all = FALSE) %>% 
    # Select a random subset of species to map
  slice_sample(n = 100)

# Make the interactive maps
BeeBDC::interactiveMapR(
  # occurrence data
  data = beeData %>%
      # Select only those species
    dplyr::filter(scientificName %in% beeData_interactive),
  # Directory where to save files
  outPath = paste0(OutPath_Figures, "interactiveMaps", sep = "/"),
  # lat long columns
  lon = "decimalLongitude",
  lat = "decimalLatitude",
  # Occurrence dataset column with species names
  speciesColumn = "scientificName",
  # Which species to map - a character vector of names or "ALL"
  # Note: "ALL" is defined AFTER filtering for country
  speciesList = "ALL",
  countryList = NULL, # studyArea
  # Point jitter to see stacked points - jitters an amount in decimal degrees
  jitterValue = 0.01
)

  ##### 9.5 Data providers ####
# Read in the clean data if it's not already in the environment
if(!exists("cleanData")){
  cleanData <- readr::read_csv(paste(OutPath_Intermediate, "05_cleaned_database.csv",
                                sep = "/"),
                          col_types = BeeBDC::ColTypeR(),
  locale = readr::locale(encoding = "UTF-8"))}
institutionList_DL <- readxl::read_excel(paste(DiscLifePath, "Apoidea Bee Collections Master List jan 2023.xlsx",
                                               sep = "/"))
  # Create the table
dataProvTable <- BeeBDC::dataProvTables(data = cleanData,
                                        runBeeDataChecks = TRUE,
                                        outPath = OutPath_Report,
                                        fileName = "dataProvTable.csv")

  ##### 9.6 Flag summary ####
  # Produce a summary table of flags per species
summaryTable <- BeeBDC::flagSummaryTable(data = beeData, 
                                         column = "scientificName", 
                                         outPath = OutPath_Report,
                                         fileName = "flagTable.csv",
                                         percentThreshold = 0)
                                         





 

# Description of the BeeBDC country checklist dataset
# 16th of March 2023

#' Download a country-level checklist of bees from Discover Life
#'
#'  
#' Download the table contains taxonomic and country information for the bees of the world based 
#' on data collated on Discover Life. The data will be sourced from the BeeBDC article's 
#' Figshare.
#'  
#' Note that sometimes the download might not work without restarting R. In this case, you could
#' alternatively download the dataset from the URL below and then read it in using 
#' `base::readRDS("filePath.Rda")`. Note that as of version 1.3.2, this function internally uses 
#'  a modified version of the "download" function from the 
#' `downloader` package on CRAN. Additionally, BeeBDC will try a different 
#' download method with each failed attempt (for Windows: "auto" > "internal" > "libcurl" > "wget" > "curl"; 
#` for Mac/Unix: "libcurl" > "wget" > "curl" > "auto";
#' or "auto" if any particular method is not available).
#'  
#'  See [BeeBDC::beesTaxonomy()] for further context. 
#'
#' @param URL A character vector to the FigShare location of the dataset. The default will be to
#' the most-recent version.
#' @param mode A character passed on to `utils::download.file()`. Default = "wb" for Windows or "w" for Mac/Linux.
#' @param headers Character. Passed on to  `utils::download.file()`. Default = NULL.
#' @param token Character. Optional personal access token from FigShare account for authorisation. 
#' Default = NULL.
#' @param alternateURL Logical. If TRUE then the function will use the an alernate version of the 
#' download URL. Might be worth trying if it is failing. Default = FALSE.
#' @param ... Extra variables that can be passed to `downloader::download()`.
#' 
#' @return A downloaded beesChecklist.Rda file in the outPath and the same tibble returned to
#' the environment.
#' 
#'
#'  **Column details **
#'  
#'  **validName** The valid scientificName as it should occur in the scientificName column.
#'  
#'  **DiscoverLife_name** The full country name as it occurs on Discover Life.
#'  
#'  **rNaturalEarth_name** Country name from rnaturalearth's name_long.
#'  
#'  **shortName** A short version of the country name.
#'  
#'  **DiscoverLife_ISO** The ISO country name as it occurs on Discover Life.
#'  
#'  **Alpha-2** Alpha-2 from rnaturalearth.
#'  
#'  **Alpha-3** Alpha-3 from rnaturalearth.
#'  
#'  **official** Official country name = "yes" or only a Discover Life name = "no".
#'  
#'  **Source** A text strign denoting the source or author of the name-country pair.
#'  
#'  **matchCertainty** Quality of the name's match to the Discover Life checklist.
#'  
#'  **canonical** The valid species name without scientificNameAuthority.
#'  
#'  **canonical_withFlags** The validName without the scientificNameAuthority but with Discover Life flags.
#'  
#'  **family** Bee family.
#'  
#'  **subfamily** Bee subfamily.
#'  
#'  **genus** Bee genus.
#'  
#'  **subgenus** Bee subgenus.
#'  
#'  **infraspecies** Bee infraSpecificEpithet.
#'  
#'  **species** Bee specificEpithet.
#'  
#'  **scientificNameAuthorship** Bee scientificNameAuthorship.
#'  
#'  **taxon_rank** Rank of the taxon name.
#'  
#'  **Notes** Discover Life country name notes.
#'  
#'  **Previous checklists:**
#'  
#'   - 2026-01-12 **current**: https://open.flinders.edu.au/ndownloader/files/60945823
#'  
#'   - 2024-06-17: https://open.flinders.edu.au/ndownloader/files/47092720
#'   
#'   - Original: https://open.flinders.edu.au/ndownloader/files/42320598
#'  
#' 
#' @references This dataset was created using the Discover Life checklist and taxonomy. 
#' Dataset is from the publication: 
#' DOREY, J. B., CHESSHIRE, P. R., BOLAÑOS, A. N., O’REILLY, R. L., BOSSERT, S., COLLINS, S. M., LICHTENBERG, E. M., TUCKER, E., SMITH-PARDO, A., FALCON-BRINDIS, A., GUEVARA, D. A., RIBEIRO, B. R., DE PEDRO, D., FISCHER, E., HUNG, J. K.-L., PARYS, K. A., ROGAN, M. S., MINCKLEY, R. L., VELZCO, S. J. E., GRISWOLD, T., ZARRILLO, T. A., SICA, Y., ORR, M. C., GUZMAN, L. M., ASCHER, J., HUGHES, A. C. & COBB, N. S. In review. A globally synthesised and flagged bee occurrence dataset and cleaning workflow. Scientific Data.
#' The checklist data are mostly compiled from Discover Life data, www.discoverlife.org:
#' ASCHER, J. S. & PICKERING, J. 2020. Discover Life bee species guide and world checklist (Hymenoptera: Apoidea: Anthophila). http://www.discoverlife.org/mp/20q?guide=Apoidea_species.
#' 
#' @export
#' 
#' @examples
#'\dontrun{
#' beesChecklist <- BeeBDC::beesChecklist()
#'}
beesChecklist <- function(URL = "https://api.figshare.com/v2/file/download/60945823",
                          mode = NULL,
                          headers = NULL,
                          token = NULL,
                          alternateURL = FALSE,
                          ...){
  destfile <- checklist <- attempt <- nAttempts <- error_funcFile <- error_func <- NULL
  downloadReturn <- . <- NULL
  
    # If user wants to try the Flinders URL, which seems to not be working as well
  if(alternateURL == TRUE){
    URL <- URL %>%
      stringr::str_remove_all(".*/") %>%
      stringr::str_c("https://open.flinders.edu.au/ndownloader/files/", .)
  }
  
  #### 0.0 Prep ####
  # Set the number of attempts
  nAttempts = 6
    
    ##### 0.1 Errors ####
      ###### a. messages ####
    # Set up the error message function
  error_func <- function(e){
    message(paste("Checklist download attempt failed..."))
  }
  error_funcFile <- function(e){
    message(paste("Could not read checklist download..."))
  }
  
      ###### b. error catcher ####
  # Function to capture error outputs from here
        # Source - https://stackoverflow.com/a
        # Posted by Martin Morgan, modified by community. See post 'Timeline' for 
        # change history
        # Retrieved 2026-01-12, License - CC BY-SA 2.5
  errorCatcher <- function(fun) {
      warn <- err <- NULL
      res <- withCallingHandlers(
        tryCatch(fun(...), error=function(e) {
          err <<- conditionMessage(e)
          NULL
        }), warning=function(w) {
          warn <<- append(warn, conditionMessage(w))
          invokeRestart("muffleWarning")
        })
      list(res, warn=warn, err=err)
    }
  
    ##### 0.2 Check OS ####
    # Check operating system
  OS <- dplyr::if_else(.Platform$OS.type == "unix",
                                             "MacLinux",
                                             "Windows")
  writeLines(paste0("The operating system detected is ", OS, "."))
  
    ##### 0.3 Downloader function ####
  # Please note that this function is taken directly from the "downloader" package version 0.4.1
  # This is the purpose of the package, but I have taken their excellent function to avoid
  # Another dependency for BeeBDC. MY apologies and thanks to the authors.
  download <- function(URL, destfile = NULL, methodNum = NULL, headers = headers, ...) {
    # First, check protocol. If http or https, check platform:
    if (grepl('^https?://', URL)) {
      #### Windows ####
      if (tolower(.Platform$OS.type) == "windows") {
        # Set method to NULL to then be over-written
        method <- NULL
        # Try httr first
        if(methodNum == 1){
          message(paste0("Trying first download method using httr::GET..."))
          httr::GET(
            URL,
            httr::add_headers(
              Authorization = paste("token", token)
            ),
            httr::write_disk(destfile, overwrite = TRUE)
          )} # END methodNum == 1
        
        # Try different methods if failed
        if(methodNum == 2){method <- "auto"}
        if(methodNum == 3){method <- "wininet"}
          # Check also if libcurl is an option
        if(methodNum == 4 && capabilities("libcurl")){
          method <- "libcurl"}
          # Check also if wget is an option
        if(methodNum == 5 && nzchar(Sys.which("wget")[1])){
          method <- "wget"}
          # Check also if curl is an option
        if(methodNum == 6 && nzchar(Sys.which("curl")[1])){
          method <- "curl"}
          # If one of the above fails, use "internal"
        if(is.null(method)){
          method <- "internal"
        }
        
        if(is.null(mode)){
          mode <- "wb"}
        
        message(paste0("Trying download method ", method, " and mode ", mode, "..."))
        # download.file will complain about file size with something like:
        #       Warning message:
        #         In download.file(urURLl, ...) : downloaded length 19457 != reported length 200
        # because apparently it compares the length with the status code returned (?)
        # so we supress that
       if(methodNum > 1){
        message(paste0("Trying download method ", method, " and mode ", mode, "..."))
        downloadReturn <- utils::download.file(URL, 
                                               method = method, 
                                               destfile = destfile, 
                                               mode = mode,
                                               headers = headers) %>%
          errorCatcher()}
        
      } else {
        #### Mac/Linux ####
        method <- NULL
        # Try httr first
        if(methodNum == 1){
          message(paste0("Trying first download method using httr::GET..."))
          httr::GET(
            URL,
            httr::add_headers(
              Authorization = paste("token", token)
            ),
            httr::write_disk(destfile, overwrite = TRUE)
          )} # END methodNum == 1
        # If non-Windows, check for libcurl/curl/wget/lynx, then call download.file with
        # appropriate method.
        if (capabilities("libcurl") && methodNum == 2) {
          method <- "libcurl"
        } else if (nzchar(Sys.which("wget")[1]) && methodNum == 3) {
          method <- "wget"
        } else if (nzchar(Sys.which("curl")[1]) && methodNum == 4) {
          method <- "curl"
          # curl needs to add a -L option to follow redirects.
          # Save the original options and restore when we exit.
          orig_extra_options <- getOption("download.file.extra")
          on.exit(options(download.file.extra = orig_extra_options))
          options(download.file.extra = paste("-L", orig_extra_options))
        } else if (nzchar(Sys.which("lynx")[1]) && methodNum == 5) {
          method <- "lynx"
        } else if(methodNum == 6){
          method <- "auto"
        }
        if(is.null(method)){
          method <- "auto"
        }
        if(is.null(mode)){
          mode <- "wb"  
        }
        if(methodNum > 1){
        message(paste0("Trying download method ", method, " and mode ", mode, "..."))
        downloadReturn <- utils::download.file(URL, 
                                               method = method, 
                                               destfile = destfile, 
                                               mode = mode,
                                               headers = headers,
                                               ...) %>%
          errorCatcher()}
      }
    } else {
      downloadReturn <- utils::download.file(URL, destfile = destfile, 
                                             mode = "wb", headers = headers, ...) %>%
        errorCatcher()
    }
    return(downloadReturn)
  } # END download function
  
  
  #### 1.0 Download ####
  # Run a code to download the data and deal with potential internet issues
  checklist <- NULL                                 
  attempt <- 1 
  savePath <- file.path(tempdir(), "beesChecklist.Rda") %>% 
      # Change all backlashes to forward slashes -- sometimes they mix on Windows...
    stringr::str_replace_all("\\\\","/")
  writeLines(paste0("Saving file temporarily to ", savePath))
  suppressWarnings(
    while( is.null(checklist) && attempt <= nAttempts) {    
      # Initial message
      print( paste("Attempt: ", attempt, " of ", nAttempts))
        # Don't attempt for the last attempt
      if(attempt <= nAttempts){
# Windows or Mac/Linux
      # Download the file
      tryCatch(downloadReturn <- download(URL, destfile = savePath, 
                                          # Change the method based on attempt number
                                          methodNum = attempt,
                                          headers = headers),
          error = error_func, warning = error_func)
      # Load the file 
        tryCatch(
        checklist <- base::readRDS(savePath),
        error = error_funcFile, warning = error_funcFile)
      } # END if attempt < nAttempts

      
      if(attempt <= nAttempts){
      # Wait five seconds before the next request 
      if(is.null(checklist)){Sys.sleep(5)}
      } # END IF #2
      # Count the next attempt
      attempt <- attempt + 1   
      
      if(attempt > 1 && !is.null(downloadReturn)){
        # Output errors per run
        # Check download errors
        if(!stringr::str_detect(paste0(downloadReturn, collapse = ""), 
        "could not find function")){
        # Remove NULL elements
        downloadReturn <- downloadReturn[-which(sapply(downloadReturn, is.null))]
        # Paste message
        message(paste0("\n - Possible *download* error(s) returned:\n", paste0(
        names(downloadReturn), ": ", downloadReturn, collapse = "\n")))}
        # Check file errors
        fileError <- base::readRDS(savePath) %>% 
        errorCatcher()
        if(!stringr::str_detect(paste0(fileError, collapse = ""), 
        "could not find function")){
        # Remove NULL elements
        fileError <- fileError[-which(sapply(fileError, is.null))]
        message(paste0("\n - Possible *file* error(s) returned:\n", paste0(
        names(fileError), ": ", fileError, collapse = "\n")))}
      } # END if( attempt > 1){
    } # END while 
  )
  
    #### 2.0 Error list ####
  if(is.null(checklist)){
    # Check system capacities
    message(paste0(
      "System capabilities are:\n",
      " * Has libcurl? ", capabilities("libcurl"),
      "\n * Has wget? ", nzchar(Sys.which("wget")[1]),
      "\n * Has curl? ", nzchar(Sys.which("curl")[1])
    ))
    
    message(paste0(" - Checklist download failed. Please check your internet connection.\n",
                "Alternatively, feel free to paste the download url into your browser (",
                URL, ")",
                " and download the file directly. \n",
                "This file can then be read into R using:\n",
                "beesChecklist <- readRDS('path/to/downloaded/file/beesChecklist.Rda')"))
    
    
    stop("Errors finished.")
  }
  
  #### 3.0 Return ####
    # Return the data to the user
  return(checklist)
} # END function

# load in the test dataset
system.file("extdata", "testChecklist.rda", package="BeeBDC") |>
  load()


# Test the expected results
testthat::test_that("testChecklist expected number of columns", {
  testthat::expect_equal(ncol(testChecklist), 24)
})

# Test classes
testthat::test_that("testChecklist expected class", {
  testthat::expect_type(testChecklist, "list")
})
testthat::test_that("testChecklist expected class", {
  testthat::expect_equal(attributes(testChecklist)$class, c("tbl_df","tbl","data.frame" ))
})

# Define download headers
# headers <- c(Authorization = paste("token", "20271361303ce042ff9cce49ecd9c8f23594ed4032e83f150e535dbd2b67297ea88448e3ca29260f6c416f581834094a898089d90f2229111845f01ef8b75f46"))
# # Set some options for accessing the data
testthat::test_that("DownloadTest_skipOnline",{
  # Skip this test on MAC on GITHUB due to 403 error in tests
  if(stringr::str_detect(Sys.info() %>% as.character(), "Darwin") %>% any()){
    skip_on_ci()}
#  skip_on_ci()
Sys.getenv("BEEBDC_SECRET_GITHUB")
Authorization <- Sys.getenv("BEEBDC_SECRET_GITHUB")
userAgent <- paste0(sprintf("BeeBDC-tests/%s (R/%s.%s)",
                            as.character(utils::packageVersion("BeeBDC")),
                            R.version$major, R.version$minor),
                    "; R (",R.version$major,".", R.version$minor, " ",
                    R.version$platform, " ", R.version$arch," ",
                    R.version$os,")") 
options(timeout=400,
        HTTPUserAgent =  userAgent)


OS <- dplyr::if_else(.Platform$OS.type == "unix",
                     "MacLinux",
                     "Windows")
  # select mode — cna be varied by OS if needed
if(OS == "Windows"){
  mode <- "wb"
}else{
  mode <- "wb"
}

beesChecklist <- NULL
# load in the full dataset
beesChecklist <- BeeBDC::beesChecklist(mode = mode,
                                       headers = NULL,
                                       token = Authorization)

if(!is.null(beesChecklist)){
# Test the expected results
testthat::test_that("beesChecklist expected number of columns", {
  testthat::expect_equal(ncol(beesChecklist), 22)
})

# Test classes
testthat::test_that("beesChecklist expected class", {
  testthat::expect_type(beesChecklist, "list")
})
testthat::test_that("beesChecklist expected class", {
  testthat::expect_equal(attributes(beesChecklist)$class, c("tbl_df","tbl","data.frame" ))
})

} #END !is.null

}) # END testthat::test_that

# This function was written by James B Dorey on the 7th of September 2022.
  # It aims to match up database_id numbers applied between different runs by matching the
  # current and prior runs and a list of columns to find matches by. It finally matches the 
  # remaining occurrences by mergine by all columns.
  # For help, please contact jbdorey[at]me.com


#' Attempt to match database_ids from a prior run
#' 
#' This function attempts to match database_ids from a prior bdc or BeeBDC run in order to keep 
#' this column somewhat consistent between iterations. However, not all records contain sufficient
#' information for this to work flawlessly.
#'
#' @param currentData A data frame or tibble. The NEW occurrence records as input.
#' @param priorData A data frame or tibble. The PRIOR occurrence records as input.
#' @param matchBy A list of character vectors Should contain the columns to iteratively compare.
#' @param completeness_cols A character vector. The columns to check for completeness, arrange, 
#' and assign the relevant prior database_id.
#' @param excludeDataset A character vector. The dataSources that are to be excluded from data 
#' matching. These should be static dataSources from minor providers.
#'
#' @return The input data frame returned with an updated database_id column that shows the 
#' database_ids as in priorData where they could be matched. Additionally, a columnd called 
#' idContinuity is returned where TRUE indicates a match to a prior database_id and FALSE 
#' indicates that a new database_id was assigned.
#' 
#' @export
#' 
#' @importFrom dplyr %>%
#' @importFrom stats complete.cases
#' @importFrom dplyr desc across
#'
#' @examples
#' # Get the example data
#' data("beesRaw", package = "BeeBDC")
#' # Which datasets are static and should be excluded from matching?
#' excludeDataset <- c("BMin", "BMont", "CAES", "EaCO", "Ecd", "EcoS",
#'                     "Gai", "KP", "EPEL", "USGS", "FSCA", "SMC", "Bal", "Lic", "Arm", "BBD", 
#'                     "MEPB")
#'   # Match the data to itself just as an example of running the code.
#' beesRaw_out <- idMatchR(
#'   currentData = beesRaw,
#'   priorData = beesRaw,
#'   # First matches will be given preference over later ones
#'   matchBy = dplyr::lst(c("gbifID"),
#'                         c("catalogNumber", "institutionCode", "dataSource"),
#'                         c("occurrenceID", "dataSource"),
#'                         c("recordId", "dataSource"),
#'                         c("id"),
#'                         c("catalogNumber", "institutionCode")),
#'   # You can exclude datasets from prior by matching their prefixs - before first underscore:
#'   excludeDataset = excludeDataset)

idMatchR <- function(
    currentData = NULL,
    priorData = NULL,
    matchBy = NULL,
    completeness_cols = NULL,
    excludeDataset = NULL){
  # locally bind variables to the function
  dataSource <- completeness <- database_id <- . <- currentConcat <- dataSourceShort <- 
    database_id_matched <- idContinuity <- databaseName <- database_id_current <-
    databaseNum <- missingNum <- database_id_new <- databaseNum_current <- NULL
  
  requireNamespace("dplyr")

    #### 0.0 Prep ####
    ##### 0.1 Errors ####
      ###### a. fatal ####
  if(is.null(currentData)){
    stop("Please proivde a 'currentData' input.")
  }
  if(is.null(priorData)){
    stop("Please proivde a 'priorData' input.")
  }
  if(is.null(matchBy)){
    stop("Please proivde a 'matchBy' input or inputs.")
  }
  
  ###### b. Warnings ####
  if(is.null(completeness_cols)){
    message(paste("Warning message: \n",
                  " - No completeness_cols provided. Using default of: ",
                  "c('decimalLatitude',  'decimalLongitude', 'scientificName', and 'eventDate')",
                  sep=""))
    completeness_cols = c("decimalLatitude",  "decimalLongitude",
                          "scientificName", "eventDate")
  }
  
    ##### 0.2 Rm datasets ####  
    # Change the dataSource for matching
  if(!is.null(excludeDataset)){
    priorData <- priorData %>%
        # Remove all text after the first "_"
      dplyr::mutate(dataSource = stringr::str_remove(dataSource, "_.*")) %>%
        # Remove the datasets
      dplyr::filter(!dataSource %in% excludeDataset)
  }

  
##### 0.3 Format data ####
  # First save a version of the currentData to keep and return
  returnData <- currentData
  # Save a count of priorData rows
  priorRowCount <- nrow(priorData)
  
    ###### a. dataSource ####
    # If the user is matching by DataSource, then simplify that column to only the over-arching 
    # source.
  if(any(stringr::str_detect(string = unlist(matchBy), pattern = "dataSource"))){
      # PRIOR dataset - Only run if Rm datasets isn't running already
    if(is.null(excludeDataset)){
    priorData <- priorData %>%
      # Remove all text after the first "_"
      dplyr::mutate(dataSource = stringr::str_remove(dataSource, "_.*"))
    }
      # CURRENT dataset
    currentData <- currentData %>%
      # Remove all text after the first "_"
      dplyr::mutate(dataSource = stringr::str_remove(dataSource, "_.*"))
  }
  
  ###### b. simplify by select ####
    # Only select the columns that are called by the function
  priorData <- priorData %>%
      # Keep only the columns called for and the database_id
    dplyr::select(tidyselect::all_of(c("database_id", unique(unlist(matchBy), completeness_cols))))
  currentData <- currentData %>%
    # Keep only the columns called for and the database_id
    dplyr::select(tidyselect::all_of(c("database_id", unique(unlist(matchBy), completeness_cols))))
  
  ##### 0.4 Completeness and arrange ####
  ###### a. completeness ####
  # Get the sum of the complete.cases of four important fields. Preference will be given to keeping 
  # the most-complete records
  writeLines(paste(
    " - Generating a basic completeness summary from the ", 
    paste(completeness_cols, collapse = ", "), " columns.","\n",
    "This summary is simply the sum of complete.cases in each column. It ranges from zero to the N",
    " of columns. This will be used to sort duplicate rows and select the most-complete rows.",
    sep = ""
  ))
  priorData <- priorData %>% 
    dplyr::rowwise() %>%
    # Create a new column called "completeness" where higher values are more-complete 
    dplyr::mutate(completeness = sum(complete.cases(completeness_cols))) %>%     
    dplyr::ungroup() %>%
      # Arrange so that the most-complete are on top. This might be overkill.
    dplyr::arrange( desc(completeness))
  gc()
  
  
    #### 1.0 loop ####
  
  writeLines(" - Starting core loop...")
    # Set up a loop dataframe to enter into
  loopDF <- dplyr::tibble()
  # Create a dataset to put unique values into
  for(i in 1:length(matchBy)){
    # Select the ith CustomComparisons to match with
    currentMatch <- matchBy[[i]]
    
    ##### 1.1 single input ####
    if(length(currentMatch) == 1){
      matched <- priorData %>%
        # Remove NA values and get distinct
        tidyr::drop_na(tidyselect::all_of(currentMatch)) %>%
        dplyr::distinct(dplyr::across(tidyselect::all_of(currentMatch)),
                        .keep_all = TRUE) %>%
        # Add a new column with these values concatenated,
        # dplyr::mutate(currentConcat = tidyselect::all_of(currentMatch))
        # JOIN datasets
        dplyr::left_join(., 
                         # FORMAT MATCH
                         currentData %>% dplyr::select(
                           tidyselect::all_of(c("database_id", currentMatch))) %>%
                           # Remove NA values and get distinct
                           tidyr::drop_na(tidyselect::all_of(currentMatch)) %>%
                           dplyr::distinct(dplyr::across(tidyselect::all_of(currentMatch)),
                                           .keep_all = TRUE),
                           by = currentMatch,
                         suffix = c("", "_current")) %>%
        # Extract only the matched ids
        dplyr::select(tidyselect::any_of(c("database_id", "database_id_current"))) %>%
        # Remove empty matches
        tidyr::drop_na()
      
      # User output
      writeLines(paste0(" - we matched ", 
                        format(nrow(matched), big.mark = ","),
                        " records using ", 
                        paste0(currentMatch, collapse = ", "), "."))
      
      # Merge with loopDF
      loopDF <- matched %>% 
        dplyr::bind_rows(loopDF)
    } # End single IF statement
    
    #### 1.2 multiple inputs ####
    if(length(currentMatch) > 1){
      matched <- priorData %>%
      # Remove NA values and get distinct
      tidyr::drop_na(tidyselect::all_of(currentMatch)) %>%
      dplyr::distinct(dplyr::across(tidyselect::all_of(currentMatch)),
                      .keep_all = TRUE) %>%
      # Add a new column with these values concatenated,
      tidyr::unite(., tidyselect::all_of(currentMatch), col = currentConcat) %>% 
        # JOIN datasets
      dplyr::left_join(., 
                       # FORMAT MATCH
                       currentData %>% dplyr::select(
                         tidyselect::all_of(c("database_id", currentMatch))) %>%
                         # Remove NA values and get distinct
                         tidyr::drop_na(tidyselect::all_of(currentMatch)) %>%
                         dplyr::distinct(dplyr::across(tidyselect::all_of(currentMatch)), 
                                         .keep_all = TRUE) %>%
                         # Add a new column with these values concatenated,
                         tidyr::unite(., tidyselect::all_of(currentMatch), col = currentConcat),
                       by = "currentConcat",
                       suffix = c("", "_current")) %>%
        # Extract only the matched ids
      dplyr::select(tidyselect::any_of(c("database_id", "database_id_current"))) %>%
        # Remove empty matches
        tidyr::drop_na()
      
      # User output
      writeLines(paste0(" - we matched ", 
                        format(nrow(matched), big.mark = ","),
                        " records using ", paste0(currentMatch, collapse = ", "), "."))
      # Merge with loopDF
      loopDF <- matched %>% 
        dplyr::bind_rows(loopDF)
    } # End multiple IF statement

    #### 1.3 Rm matches ####
      # Remove the occurrences that were just matched before the next iteration
    priorData <- priorData %>%
      dplyr::filter(!database_id %in% loopDF$database_id)
    currentData <- currentData %>%
      dplyr::filter(!database_id %in% loopDF$database_id_current)
    
      # User output
    writeLines(paste0("This leaves ",
                      format(nrow(priorData), big.mark = ","),
                      " unmatched data in the priorData file"))
    }# END LOOP

    # Remove data no longer needed
  rm(priorData, currentData)
  
  
  #### 2.0 Data return ####
  writeLines(" - Combining ids and assigning new ones where needed...")
    # Add a column to that matched data:
      # idContinuity, that shows that these ids are continuous with prior versions
  loopDF <- loopDF %>%
    dplyr::mutate(idContinuity = TRUE)
  
    # Change the database_id column for return
  checkedData <- returnData %>%
    select(tidyselect::all_of(c("database_id", "dataSource"))) %>%
      # Filter to only the examined dataSources
    # Remove all text after the first "_"
    dplyr::mutate(dataSourceShort = stringr::str_remove(dataSource, "_.*")) %>%
      # Add the new database_id column, while removing the old one (database_id_current)
    dplyr::left_join(loopDF, by = c("database_id" = "database_id_current"),
                     suffix = c("", "_matched"), keep = FALSE) %>%
      # Remove existing [current] database_id columns
    dplyr::rename(database_id_current = database_id) %>%
      # Move this column to the start
    dplyr::relocate(database_id_matched) %>%
      # Rename the column
    dplyr::rename(database_id = database_id_matched) %>%
      # Highlight those records that might not be id-continuous
    dplyr::mutate(idContinuity = dplyr::if_else(is.na(idContinuity),
                                                FALSE, TRUE)) %>%
    # Add a new column with the database_id NAME
    dplyr::mutate(databaseName = stringr::str_extract(
      string = database_id, pattern = "[a-zA-Z_]+")) %>%
    # Fix those that failed to match
    dplyr::mutate(databaseName = dplyr::if_else(
      is.na(databaseName),
      stringr::str_extract(
        string = database_id_current, pattern = "[a-zA-Z_]+"),
      databaseName)) %>%
    # Add a new column with the database_id NUMBER
    dplyr::mutate(databaseNum = stringr::str_extract(
      string = database_id, "[0-9]+") %>% as.numeric(),
        # Get a column with the current numbers to start the MAX count from
      databaseNum_current = database_id_current %>% stringr::str_extract("[0-9]+") %>% 
        as.numeric()) %>%
    # Group by databaseName
    dplyr::group_by(databaseName) %>%
    # Sort
    dplyr::arrange(databaseNum, .by_group = TRUE)
  
    # Get new numbers for the new data
  newData <- checkedData  %>%
      # Apply by group
    dplyr::group_by(databaseName) %>%
      # Add a new column with the databaseNum numbers
    dplyr::mutate(missingNum = databaseNum) %>%
      # If the group is entirely unmatched, assign the first row in that group to equal 1
    dplyr::mutate(missingNum = dplyr::if_else(dplyr::row_number() == 1 & is.na(missingNum[[1]]),
                  1,
                  missingNum)) %>%
      # Fill down the missing numbers starting from 1+ the maximum within databaseName group.
    dplyr::mutate(missingNum = dplyr::if_else(is.na(missingNum), 
                                        (max(databaseNum_current, na.rm = TRUE)+
                                           dplyr::row_number()-sum(complete.cases(missingNum))  
                                         ),
                  missingNum)) %>%
      # Update the database_id column
    dplyr::mutate(database_id = stringr::str_c(databaseName, missingNum)) %>% 
      # Filter for only NA values on the databaseNum column
    dplyr::filter(is.na(databaseNum)) 
    
  
    # Now combine
  checkedData <- checkedData %>%
      # First, remove those newData from the checkedData
    dplyr::filter(!database_id_current %in% newData$database_id_current) %>%
      # now re-combine
    dplyr::bind_rows(newData) %>%
      # Remove groupings
    dplyr::ungroup() %>%
    dplyr::distinct(database_id, .keep_all = TRUE) %>%
    # Remove the excludeDataset
    dplyr::filter(!dataSourceShort %in% excludeDataset)

  # User output
  writeLines(paste0(" - We matched a total of ",
                    format(sum(complete.cases(checkedData$databaseNum)), big.mark = ","),
                    " database_id numbers. We then assigned new database_id numbers to ",
                    format(sum(complete.cases(checkedData$missingNum)), big.mark = ","),
                    " unmatched occurrences."
                    ))

  
    # Merge the new databse IDs with the returnData
  returnData <- returnData %>%
      # Join the checkedData dataset
    dplyr::left_join(checkedData %>% dplyr::select(
      tidyselect::any_of(c("database_id", "database_id_current", "idContinuity"))),
                     by = c("database_id" = "database_id_current"),
                     suffix = c("", "_new")) %>%
    # Update the database_id column to include the new database_ids, or the old ones where
      # new ones aren't available.
    dplyr::mutate(database_id = dplyr::if_else(is.na(database_id_new),
                                                # If from an excluded dataset, 
                                                # keep existing database_id
                                               database_id,
                                                # Otherwise Assign the newly matched id
                                               database_id_new)) %>%
    dplyr::select(!database_id_new)
  

    # Return the data
  return(returnData )
  
}# END function

requireNamespace("dplyr")
requireNamespace("stringr")

  # Read in a flagged test dataset
beesFlagged <- BeeBDC::beesFlagged

  # Create a dummy "priorData" dataset using the first fifty rows
priorRun <- beesFlagged %>%
  dplyr::slice_head(n = 50)


  #### 1.0 Exclude ASP ####
  # Run the function using the first fifty to be matched to their original database_id numbers
testOut <- BeeBDC::idMatchR(
  currentData = beesFlagged %>% dplyr::mutate(database_id = database_id %>%
                                                stringr::str_replace("[0-9]+","") %>%
                                                paste0(., dplyr::row_number())),
  priorData = priorRun,
    # default completeness_cols
  completeness_cols = c("decimalLatitude",  "decimalLongitude",
                        "scientificName", "eventDate"),
  # First matches will be given preference over later ones
  matchBy = dplyr::lst(c("gbifID"),
                        c("catalogNumber", "institutionCode", "dataSource"),
                        c("occurrenceID", "dataSource"),
                        c("recordId", "dataSource"),
                        c("id"),
                        # Because INHS was entered as it's own dataset but is now included in the GBIF download...
                        c("catalogNumber", "institutionCode")),
  # You can exclude datasets from prior by matching their prefixs — before first underscore:
  # Which datasets are static and should be excluded from matching?
  excludeDataset = c("ASP", "BMin", "BMont", "CAES", "EaCO", "Ecd", "EcoS",
                     "Gai", "KP", "EPEL", "CAES", "EaCO", "FSCA", "SMC", "Lic", "Arm"))


# Get a count of TRUE and FALSE column name matches
resultsMatched <- sum(testOut$database_id %in% beesFlagged$database_id)
resultsExcluded <- sum(testOut$database_id %in% (beesFlagged %>% dplyr::mutate(database_id = database_id %>%
                                                                                   stringr::str_replace("[0-9]+","") %>%
                                                                                   paste0(., dplyr::row_number())) %>%
                                                     dplyr::pull(database_id)))
resultsNotMatched <- sum(testOut$database_id %in% beesFlagged$database_id)

# Test the number of expected TRUE and FALSE columns and then test the output format (data frames and
# tibbles are a special case of lists)
testthat::test_that("idMatchR results successfuly matched", {
  testthat::expect_equal(resultsMatched, 50)
})
testthat::test_that("idMatchR results not matched", {
  testthat::expect_equal(resultsNotMatched, 50)
})
testthat::test_that("idMatchR results excluded because in excludeDatasets", {
  testthat::expect_equal(resultsExcluded, 1)
})

testthat::test_that("idMatchR expected class", {
  testthat::expect_type(testOut, "list")
})


  #### 2.0 Don't exclude ASP ####
# Run the function using the first fifty to be matched to their original database_id numbers
testOut2 <- BeeBDC::idMatchR(
  currentData = beesFlagged %>% dplyr::mutate(database_id = database_id %>%
                                                stringr::str_replace("[0-9]+","") %>%
                                                paste0(., dplyr::row_number())),
  priorData = priorRun,
  # First matches will be given preference over later ones
  matchBy = dplyr::lst(c("gbifID"),
                        c("catalogNumber", "institutionCode", "dataSource"),
                        c("occurrenceID", "dataSource"),
                        c("recordId", "dataSource"),
                        c("id"),
                        # Because INHS was entered as it's own dataset but is now included in the GBIF download...
                        c("catalogNumber", "institutionCode")),
  # You can exclude datasets from prior by matching their prefixs — before first underscore:
  # Which datasets are static and should be excluded from matching?
    # This time don't exclude the ASP data
  excludeDataset = NULL)

# Get a count of TRUE and FALSE column name matches
resultsMatched <- sum(testOut2$database_id %in% beesFlagged$database_id)
resultsExcluded <- sum(testOut2$database_id %in% (beesFlagged %>% dplyr::mutate(database_id = database_id %>%
                                                                                 stringr::str_replace("[0-9]+","") %>%
                                                                                 paste0(., dplyr::row_number())) %>%
                                                   dplyr::pull(database_id)))
resultsNotMatched <- sum(testOut2$database_id %in% beesFlagged$database_id)

# Test the number of expected TRUE and FALSE columns and then test the output format (data frames and
# tibbles are a special case of lists)
testthat::test_that("idMatchR results successfuly matched", {
  testthat::expect_equal(resultsMatched, 50)
})
testthat::test_that("idMatchR results not matched", {
  testthat::expect_equal(resultsNotMatched, 50)
})
testthat::test_that("idMatchR results excluded because in excludeDatasets", {
  testthat::expect_equal(resultsExcluded, 0)
})

# This function, written by James Dorey builds flags from the Ascher notes column
# For queries, please contact James Dorey at jbdorey[at]me.com
# This function was started on 13th May 2022 and last updated 17th May 2022

#' @importFrom dplyr %>%
#' 
#' 
# Build a function that finds flags in the notes column and standardizes them into the flags column
flag_converter <- function(SynFile = DLdf){
  # locally bind variables to the function
  DLdf <- . <- Combined <- NULL
  
requireNamespace("dplyr")
  
  #### 1.1 doubtful ####
  # Doubtful species strings to find and matches
  doubt_sp_str <-c("nomen dubium","doubtful synonymy","doubtful synonym")
  doubt_sp_replacements <- SynFile$notes %>% 
    # Extract those strings
    stringr::str_extract(paste(doubt_sp_str,
                               collapse = "|")) %>%
    # Repalce those strings
    stringr::str_replace(pattern = paste(doubt_sp_str,
                                         collapse = "|"),
                         replacement = "doubtful species")  %>%
    as.character() %>% as.data.frame()
  
  #### 1.2 Syn Issue ####
  # Syn Issue strings to find and matches
  synIss_str <-c("partim")
  synIss_replacements <- SynFile$notes %>% 
    # Extract those strings
    stringr::str_extract(paste(synIss_str,
                               collapse = "|")) %>%
    # Repalce those strings
    stringr::str_replace(pattern = paste(synIss_str,
                                         collapse = "|"),
                         replacement = "synonym issue")  %>%
    as.character() %>% as.data.frame()
  
  #### 1.3 Auth Issue ####
  authIss_str <-c("auct", "auct , not Fabricius", "Auctorum", "Auct", "_auct",
                  "auct.", "miscitation")
  authIss_replacements <- SynFile$notes %>% 
    # Extract those strings
    stringr::str_extract(paste(authIss_str,
                               collapse = "|")) %>%
    # Repalce those strings
    stringr::str_replace(pattern = paste(authIss_str,
                                         collapse = "|"),
                         replacement = "authorship issue")  %>%
    as.character() %>% as.data.frame()
  
  #### 1.4 Auth not provided ####
  # If the authorship column is empty, flag...
  noAuth <- ifelse(is.na(SynFile$authorship), "author not provided", NA)  %>%
    as.character() %>% as.data.frame()

  
  #### 2.0 Merge flags ####
  MergedFlags <- dplyr::bind_cols(doubt_sp_replacements, synIss_replacements, authIss_replacements,
                                  noAuth) %>%
    # Concatenate the columns into one
    tidyr::unite(., col = Combined, 
                 na.rm = TRUE, sep = ", ")
    
    
  return(MergedFlags)
} # END function

# This function was written by James B Dorey on the 5th of August 2022 to fix up some
  # Country names in occurrence records and extract country names from ISO2 codes


#' Fix country name issues using a user-input list
#' 
#' This function is basic for a user to manually fix some country name inconsistencies.
#'
#' @param data A data frame or tibble. Occurrence records as input.
#' @param ISO2_table A data frame or tibble with the columns ISO2 and long names for country names. Default 
#' is a static version from Wikipedia.
#' @param commonProblems A data frame or tibble. It must have two columns: 
#' one containing the user-identified problem and one with a user-defined fix
#'
#' @return Returns the input data, but with countries occurring in the user-supplied problem 
#' column ("commonProblems") replaced with those in the user-supplied fix column
#' @export
#' 
#' @importFrom dplyr %>%
#' @importFrom stats complete.cases
#'
#' @examples
#' beesFlagged_out <- countryNameCleanR(
#' data = BeeBDC::beesFlagged,
#' commonProblems = dplyr::tibble(problem = c('U.S.A.', 'US','USA','usa','UNITED STATES',
#'                         'United States','U.S.A','MX','CA','Bras.','Braz.',
#'                         'Brasil','CNMI','USA TERRITORY: PUERTO RICO'),
#'                         fix = c('United States of America','United States of America',
#'                                 'United States of America','United States of America',
#'                                 'United States of America','United States of America',
#'                                 'United States of America','Mexico','Canada','Brazil',
#'                                 'Brazil','Brazil','Northern Mariana Islands','PUERTO.RICO')))

countryNameCleanR <- function(
    data = NULL,
    ISO2_table = NULL,
    commonProblems = NULL){
  # locally bind variables to the function
  database_id <- decimalLatitude <- decimalLongitude <- country <- countryCode <- scientificName <-
    dataSource <- FullName <- . <- fix <- NULL
  
  #### 0.0 Prep ####
  
  ##### 0.1 warnings ####
  if(is.null(data)){
    stop(" - Please provide input data.")
  }
  if(!"countryCode" %in% colnames(data)){
    stop(" - No countryCode column in data.")
  }
  
  ##### 0.2 Data defaults ####
  if(is.null(ISO2_table)){
    writeLines(" - Using default country names and codes from https:en.wikipedia.org/wiki/ISO_3166-1_alpha-2 - static version from July 2022.")
    # Use countryCode to extract country name using the Wikipedia table - https://en.wikipedia.org/wiki/ISO_3166-1_alpha-2
    # Make a tibble with 
    ISO2_table <- dplyr::tibble(
      ISO2 = c("AD","AE","AF","AG","AI","AL","AM","AO","AQ","AR","AS","AT","AU","AW","AX","AZ","BA","BB","BD","BE","BF","BG","BH","BI","BJ","BL","BM","BN","BO","BQ","BR","BS","BT","BV","BW","BY","BZ","CA","CC","CD","CF","CG","CH","CI","CK","CL","CM","CN","CO","CR","CU","CV","CW","CX","CY","CZ","DE","DJ","DK","DM","DO","DZ","EC","EE","EG","EH","ER","ES","ET","FI","FJ","FK","FM","FO","FR","GA","GB","GD","GE","GF","GG","GH","GI","GL","GM","GN","GP","GQ","GR","GS","GT","GU","GW","GY","HK","HM","HN","HR","HT","HU","ID","IE","IL","IM","IN","IO","IQ","IR","IS","IT","JE","JM","JO","JP","KE","KG","KH","KI","KM","KN","KP","KR","KW","KY","KZ","LA","LB","LC","LI","LK","LR","LS","LT","LU","LV","LY","MA","MC","MD","ME","MF","MG","MH","MK","ML","MM","MN","MO","MP","MQ","MR","MS","MT","MU","MV","MW","MX","MY","MZ","NA","NC","NE","NF","NG","NI","NL","NO","NP","NR","NU","NZ","OM","PA","PE","PF","PG","PH","PK","PL","PM","PN","PR","PS","PT","PW","PY","QA","RE","RO","RS","RU","RW","SA","SB","SC","SD","SE","SG","SH","SI","SJ","SK","SL","SM","SN","SO","SR","SS","ST","SV","SX","SY","SZ","TC","TD","TF","TG","TH","TJ","TK","TL","TM","TN","TO","TR","TT","TV","TW","TZ","UA","UG","UM","US","UY","UZ","VA","VC","VE","VG","VI","VN","VU","WF","WS","YE","YT","ZA","ZM","ZW"),
      FullName = c("Andorra","United Arab Emirates","Afghanistan","Antigua and Barbuda","Anguilla","Albania","Armenia","Angola","Antarctica","Argentina","American Samoa","Austria","Australia","Aruba","\\u00c5land Islands","Azerbaijan","Bosnia and Herzegovina","Barbados","Bangladesh","Belgium","Burkina Faso","Bulgaria","Bahrain","Burundi","Benin","Saint Barth\\u00e9lemy","Bermuda","Brunei Darussalam","Bolivia (Plurinational State of)","Bonaire, Saint Eustatius and Saba","Brazil","Bahamas","Bhutan","Bouvet Island","Botswana","Belarus","Belize","Canada","Cocos (Keeling) Islands","Congo, Democratic Republic of the","Central African Republic","Congo","Switzerland","C\\u00f4te d'Ivoire","Cook Islands","Chile","Cameroon","China","Colombia","Costa Rica","Cuba","Cabo Verde","Cura\\u00e7ao","Christmas Island","Cyprus","Czechia","Germany","Djibouti","Denmark","Dominica","Dominican Republic","Algeria","Ecuador","Estonia","Egypt","Western Sahara","Eritrea","Spain","Ethiopia","Finland","Fiji","Falkland Islands (Malvinas)","Micronesia (Federated States of)","Faroe Islands","France","Gabon","United Kingdom of Great Britain and Northern Ireland","Grenada","Georgia","French Guiana","Guernsey","Ghana","Gibraltar","Greenland","Gambia","Guinea","Guadeloupe","Equatorial Guinea","Greece","South Georgia and the South Sandwich Islands","Guatemala","Guam","Guinea-Bissau","Guyana","Hong Kong","Heard Island and McDonald Islands","Honduras","Croatia","Haiti","Hungary","Indonesia","Ireland","Israel","Isle of Man","India","British Indian Ocean Territory","Iraq","Iran (Islamic Republic of)","Iceland","Italy","Jersey","Jamaica","Jordan","Japan","Kenya","Kyrgyzstan","Cambodia","Kiribati","Comoros","Saint Kitts and Nevis","Korea (Democratic People's Republic of)","Korea, Republic of","Kuwait","Cayman Islands","Kazakhstan","Lao People's Democratic Republic","Lebanon","Saint Lucia","Liechtenstein","Sri Lanka","Liberia","Lesotho","Lithuania","Luxembourg","Latvia","Libya","Morocco","Monaco","Moldova, Republic of","Montenegro","Saint Martin (French part)","Madagascar","Marshall Islands","North Macedonia","Mali","Myanmar","Mongolia","Macao","Northern Mariana Islands","Martinique","Mauritania","Montserrat","Malta","Mauritius","Maldives","Malawi","Mexico","Malaysia","Mozambique","Namibia","New Caledonia","Niger","Norfolk Island","Nigeria","Nicaragua","Netherlands","Norway","Nepal","Nauru","Niue","New Zealand","Oman","Panama","Peru","French Polynesia","Papua New Guinea","Philippines","Pakistan","Poland","Saint Pierre and Miquelon","Pitcairn","Puerto Rico","Palestine, State of","Portugal","Palau","Paraguay","Qatar","R\\u00e9union","Romania","Serbia","Russian Federation","Rwanda","Saudi Arabia","Solomon Islands","Seychelles","Sudan","Sweden","Singapore","Saint Helena, Ascension and Tristan da Cunha","Slovenia","Svalbard and Jan Mayen","Slovakia","Sierra Leone","San Marino","Senegal","Somalia","Suriname","South Sudan","Sao Tome and Principe","El Salvador","Sint Maarten (Dutch part)","Syrian Arab Republic","Eswatini","Turks and Caicos Islands","Chad","French Southern Territories","Togo","Thailand","Tajikistan","Tokelau","Timor-Leste","Turkmenistan","Tunisia","Tonga","Turkey","Trinidad and Tobago","Tuvalu","Taiwan, Province of China","Tanzania, United Republic of","Ukraine","Uganda","United States Minor Outlying Islands","United States of America","Uruguay","Uzbekistan","Holy See","Saint Vincent and the Grenadines","Venezuela (Bolivarian Republic of)","Virgin Islands (British)","Virgin Islands (U.S.)","Viet Nam","Vanuatu","Wallis and Futuna","Samoa","Yemen","Mayotte","South Africa","Zambia","Zimbabwe"))
  }
  
###### a. prep data ####
# Remove NA for lat and long
data_noNa <- data %>% 
  # Drop na in lat and long
  tidyr::drop_na(c("decimalLatitude", "decimalLongitude")) %>%
  # select a subset of columns to save RAM
  dplyr::select(c(database_id, decimalLatitude, decimalLongitude, country, countryCode,
                  scientificName, dataSource))

  #### 1.0 GBIF ####
# Now, because GBIF doesn't have a country column (weird, right?!), 
# Create country from countryCode
GBIF_occs <- data_noNa %>%
  dplyr::filter(stringr::str_detect(dataSource, pattern = "GBIF"))
  # # Remove NA values from data_noNa
  # GBIF_occs$countryCode <- GBIF_occs$countryCode %>% as.character() %>%
  #   stringr::str_replace(pattern = "NA", replacement = "")

# Join the ISO2_table names with the countryCode df
GBIF_occs <- GBIF_occs %>%
  # If there is NO countryCode, but there IS a country, add this to countryCode in case
  dplyr::mutate(countryCode = dplyr::if_else(countryCode == "" & stats::complete.cases(as.character(country)),
                                             country, as.character(countryCode))) %>%
  dplyr::left_join(ISO2_table, by = c("countryCode" = "ISO2")) %>%
  # Insert the country information to the correct spot
  dplyr::mutate(country = FullName) %>%
  dplyr::select(!FullName)
# Re-join datasets
data_noNa <- data_noNa %>%
  # Remove GBIF databse_ids
  dplyr::filter(!database_id %in% GBIF_occs$database_id) %>%
  # Re-bind those occs
  dplyr::bind_rows(GBIF_occs)
rm(GBIF_occs)


#### 2.0 All ####
# Join the ISO2_table names with the countryCode df
data_all <- data_noNa %>%
  # If there is NO countryCode, but there IS a country, add this to countryCode in case
  dplyr::mutate(countryCode = dplyr::if_else(countryCode == "" | is.na(countryCode)
                                               & stats::complete.cases(country) & 
                                               tolower(country) != "namibia",
                                             country, countryCode)) %>%
  dplyr::left_join(ISO2_table, by = c("countryCode" = "ISO2")) %>%
  # Insert the country information to the correct spot
  dplyr::mutate(country = FullName) %>%
  dplyr::select(!FullName)
# Re-join datasets
data_all <- data_all %>%
  # Remove GBIF databse_ids
  dplyr::filter(!database_id %in% data_noNa$database_id) %>%
  # Re-bind those occs
  dplyr::bind_rows(data_noNa)
rm(data_noNa)

# Remove illegal characters
data_all$country <- data_all$country %>%
  stringr::str_replace(., pattern = paste("\\[", "\\]", "\\?",
                                          sep=  "|"), replacement = "")
# Replace the problems as they occur
data_all <- data_all %>%
  dplyr::left_join(commonProblems, by = c("country" = "problem")) %>%
  dplyr::mutate(country = 
                  dplyr::if_else(country %in% as.character(commonProblems$problem),
                                 fix, country))
return(data_all)

}

requireNamespace("readr")
requireNamespace("dplyr")
library(dplyr) ## could not use %>% without loading as library


# make yourself some test data which has common problem names for the USA and one row that is already correct
testData <- dplyr::tribble(
              ~database_id, ~decimalLatitude, ~decimalLongitude,                  ~country, ~countryCode,                 ~scientificName, ~dataSource, ~recordedBy,        ~order,  ~family, ~genus, ~specificEpithet, ~scientificNameAuthorship,
               "fake SCAN",            38.54,        -121.75667,           "United States",           NA, "Apis mellifera Linneaus, 1758",      "fake",          NA, "Hymenoptera", "apidae", "Apis",      "mellifera",          "Linnaeus, 1758",
               "fake SCAN",            38.54,        -121.75667,                  "U.S.A.",           NA, "Apis mellifera Linneaus, 1759",      "fake",          NA, "Hymenoptera", "apidae", "Apis",      "mellifera",          "Linnaeus, 1758",
               "fake SCAN",            38.54,        -121.75667,                  "U.S.A.",           NA, "Apis mellifera Linneaus, 1760",      "fake",          NA, "Hymenoptera", "apidae", "Apis",      "mellifera",          "Linnaeus, 1758",
               "fake SCAN",            38.54,        -121.75667,                      "US",           NA, "Apis mellifera Linneaus, 1761",      "fake",          NA, "Hymenoptera", "apidae", "Apis",      "mellifera",          "Linnaeus, 1758",
               "fake SCAN",            38.54,        -121.75667,                     "USA",           NA, "Apis mellifera Linneaus, 1762",      "fake",          NA, "Hymenoptera", "apidae", "Apis",      "mellifera",          "Linnaeus, 1758",
               "fake SCAN",            38.54,        -121.75667,                     "usa",           NA, "Apis mellifera Linneaus, 1763",      "fake",          NA, "Hymenoptera", "apidae", "Apis",      "mellifera",          "Linnaeus, 1758",
               "fake SCAN",            38.54,        -121.75667,           "UNITED STATES",           NA, "Apis mellifera Linneaus, 1764",      "fake",          NA, "Hymenoptera", "apidae", "Apis",      "mellifera",          "Linnaeus, 1758",
               "fake SCAN",            38.54,        -121.75667,"United States of America",           NA, "Apis mellifera Linneaus, 1758",      "fake",          NA, "Hymenoptera", "apidae", "Apis",      "mellifera",          "Linnaeus, 1758",
               "fake SCAN",            38.54,        -121.75667,                     "usa",           NA, "Apis mellifera Linneaus, 1765",      "fake",          NA, "Hymenoptera", "apidae", "Apis",      "mellifera",          "Linnaeus, 1758"
              )

# common problem table from the function documentation
commonProblems <- dplyr::tibble(problem = c('U.S.A.', 'US','USA','usa','UNITED STATES',
                                            'United States','U.S.A','MX','CA','Bras.','Braz.',
                                            'Brasil','CNMI','USA TERRITORY: PUERTO RICO'),
                                 fix = c('United States of America','United States of America',
                                        'United States of America','United States of America',
                                        'United States of America','United States of America',
                                        'United States of America','Mexico','Canada','Brazil',
                                        'Brazil','Brazil','Northern Mariana Islands','PUERTO.RICO'))


# run the function!
testOut <- BeeBDC::countryNameCleanR(data = testData, commonProblems = commonProblems)


# check that country column has only one value (correct value)
numberOfCountries <- testOut %>% group_by(country) %>% summarise()


# number of NA values in fix column
numberNAs <- sum(is.na(testOut$fix))


# Test the number of expected fixed and NA values in the fix column and
# verify addition of new column
testthat::test_that("number of country names in the fixed country column" , {
  testthat::expect_equal(ncol(numberOfCountries), 1)
})

testthat::test_that("number of unchanged rows in data", {
  testthat::expect_equal(numberNAs, 1)
})

#' Internal function. Get a world map with country names and iso code
#'
#' This is a helper function used to obtain names, iso code, and the limits
#' (polygon) of world countries. Data from the package 'rnaturalearth'. The prefix, jbd_ is meant
#' to distinguish this function from the original bdc::bdc_get_world_map().
#' 
#' @param scale Passed to rnaturalearth's ne_countries().
#' Scale of map to return, one of 110, 50, 10 or 'small', 'medium', 'large'. Default = "large".
#'
#' @noRd
#' @importFrom dplyr %>%
#'
#' @examples
#' \donttest{
#'   # This is an internal function.
#' worldmap <- jbd_get_world_map(scale = "large")
#' }
jbd_get_world_map <- function(scale = "large") {
  name_en <- iso_n3 <- iso2c <- iso3c <- NULL

  check_require_cran("rnaturalearth")

  suppressWarnings({
    worldmap <- rnaturalearth::ne_countries(scale = scale, returnclass = "sf") 
    
      # For large scales
    if(scale %in% c("large", "10")){
      worldmap <- worldmap %>%
        dplyr::mutate(iso2c = countrycode::countrycode(worldmap$name_en,
                                                       origin = "country.name.en",
                                                       destination = "iso2c"),
                      iso3c = countrycode::countrycode(worldmap$name_en,
                                                       origin = "country.name.en",
                                                       destination = "iso3c")
        ) %>% # END mutate
        dplyr::select(name_en, tidyselect::starts_with("iso")) 
    }else{ # For other scales...
      # Add some iso code to some countries polygons
        # Also, add the name column to be used
      worldmap <- worldmap %>%
        dplyr::mutate(name_en = countrycode::countrycode(iso_n3 %>% as.numeric(),
                                                         origin = "iso3n",
                                                         destination = "country.name.en"),
                      iso2c = countrycode::countrycode(worldmap$iso_n3 %>% as.numeric(),
                                                       origin = "iso3n",
                                                       destination = "iso2c"),
                      iso3c = countrycode::countrycode(worldmap$iso_n3 %>% as.numeric(),
                                                       origin = "iso3n",
                                                       destination = "iso3c")
                      ) %>%# END mutate
        dplyr::select(name_en, tidyselect::starts_with("iso")) 
    }# END else

    is.na(worldmap) %>% colSums() # number of polygons without isocode

    worldmap <- worldmap %>%
      dplyr::select(iso2c, iso3c) %>%
      sf::st_make_valid() 
  })
  return(worldmap)
}

# This function was written by James Dorey to build in the Bee biodiversity process from the [currently] 
# upcoming paper to estimate the number of bee species in the world, entitled 
    # "How many bee species are there? A quantitative global estimate"
# This function comes out of the article titled "How many bee species are there? A quantitative global estimate"
# This function was written from the 13th of February 2025. For questions, please email James
# at jbdorey [at] me.com


#' Prepare occurrence, taxonomy, and checklist data for richness estimation
#' 
#' Takes your occurrence dataset along with a taxonomy and checklist in order to produce 
#' a file that's ready to be passed into the [BeeBDC::richnessEstimateR()] function in order to 
#' estimate species richness using
#' iChao and iNEXT (hill numbers) for countries, continents, or the entire globe. 
#'
#' @param data A data frame or tibble. Occurrence records as input. Needs to include the 
#' `scientificName` column and a "country" column (see `countryColumn`)
#' @param taxonomyFile A data frame or tibble. The taxonomy file to use. 
#' Default = [BeeBDC::beesTaxonomy()] but see [BeeBDC::taxadbToBeeBDC()] for other taxa.
#' @param checklistFile A data frame or tibble. The taxonomy to use. 
#' Default = [BeeBDC::beesChecklist()]; use this as a template to convert other taxa checklists.
#' @param curveFunction A function. The mathematical function that describes the curve used to 
#' randomly sample from in order to fill empty sample sizes for species present in the checklist
#' but not present in the occurrence dataset. Default is `function(x) (228.7531 * x * x^-log(12.1593))`,
#' which is taken from the paper "How many bee species are there? A quantitative global estimate" by 
#' Dorey et al. Models can be fit using [mosaic::fitModel()].
#' @param sampleSize Numeric. The size of the sample randomly drawn from the provided curve.
#' See `curveFunction`. Default = 10000.
#' @param countryColumn Character. The column from which country names should be sought.
#' @param limitGlobal Character vector. A character vector of the countries to filter the data to
#' in order limit the extent of the global-level analysis. Defualt = NULL.
#' @param outPath Character. The output path where the curve plot (curvePlot.pdf) and the output
#' Rdata (richnessInputs.Rda) file will be saved. Default = `tempdir()`.
#' 
#' @return Saves an Rdata file with the data needed to feed into the [BeeBDC::richnessEstimateR()]
#'  function. The [BeeBDC::richnessEstimateR()] function will then 
#' use iChao and/or iNEXT to estimate species richness for countries, continents, and the globe. 
#' Also returns the RData file to the environment.
#' 
#' @importFrom dplyr %>%
#' 
#' @seealso [BeeBDC::countryHarmoniseR()] to harmonise country names based on a short list.
#' 
#' @export
#'
#' @examples
#' 
#' \dontrun{
#' data(beesCountrySubset)
#' if(require("htmlwidgets"){
#' 
#' estimateDataExample <- richnessPrepR(
#'   data = beesCountrySubset,
#'   # Download the taxonomy
#'   taxonomyFile = BeeBDC::beesTaxonomy(),
#'   # Download the checklist
#'   checklistFile = BeeBDC::beesChecklist(),
#'   curveFunction = function(x) (228.7531 * x * x^-log(12.1593)),
#'   sampleSize = 10000,
#'   countryColumn = "country_suggested",
#'   limitGlobal = NULL,
#'   outPath = tempdir()
#' )
#' } # End if require
#' } # End dontrun

richnessPrepR <- function(
    data = NULL,
    # Download the taxonomy
    taxonomyFile = BeeBDC::beesTaxonomy(),
    # Download the checklist
    checklistFile = BeeBDC::beesChecklist(),
    curveFunction = function(x) (228.7531 * x * x^-log(12.1593)),
    sampleSize = 10000,
    countryColumn = "country_suggested",
    limitGlobal = NULL,
    outPath = tempdir()
) {  
  # locally bind variables to the function
  . <- taxonomyAcceptedNames <- taxonomy_synonyms <- data_counts <- taxonomic_status <- NULL
  taxonomic_status<- rNaturalEarth_name<- scientificName<- count_n<- validName<- name<- 
    continent<- name_long<- continent.x<- continent.y <- country_suggested <- n<- NULL
  
  
  # Load required packages 
  requireNamespace("rlang")
  requireNamespace("dplyr")
  
  # Record start time
  startTime <- Sys.time()
  
  #### 0.0 Prep ####
  ##### 0.1 Errors ####
  ###### a. FATAL errors ####
  if(is.null(data)){
    stop(" - Please provide an argument for data. I'm a program not a magician.")
  }
  
    ##### 0.2 Prep input data ####
  # Designing data
    # data <- readr::read_csv("/Users/jamesdorey/Desktop/Uni/Packages/BeeBDC_development/beesCountry.csv") %>%
    #   dplyr::select(c("scientificName", "country_suggested"))
  data <- data %>%
    dplyr::rename(countryColumn = tidyselect::any_of(countryColumn)) 
  
  
  #### 1.0 Prep taxonomy and checklist ####
    ##### 1.1 Taxonomy manipulation ####
  # Get this just for valid species 
  taxonomyAcceptedNames <- taxonomyFile %>%
    dplyr::filter(taxonomic_status == "accepted")
  # Get this just for synonyms species 
  taxonomy_synonyms <- taxonomyFile %>%
    dplyr::filter(taxonomic_status == "synonym")
  
  
    ##### 1.2 Checklist manipulation ####
  # Now, for each country pass on the newly-extracted info to the country checklist
  checklistFile <- checklistFile %>% 
    dplyr::left_join(taxonomyFile %>%
                       dplyr::select(tidyselect::any_of(c("validName", "year", 
                                                          "author", "authorCount"))),
                     by = "validName")
  
    # If chosen, limit the countries in the checklist in order to limit the global analysis
  if(!is.null(limitGlobal)){
    checklistFile <- checklistFile %>%
      dplyr::filter(rNaturalEarth_name %in% limitGlobal)
  }
  
  #### 2. Prepare occ data ####
  ##### 2.1 Prepare all counts ####
  ###### a. total counts ####
  # Get counts of species in the dataset
  data_counts <- data %>% 
    dplyr::group_by(scientificName) %>%
    dplyr::count() %>%
    dplyr::mutate(dataFrom = "points")
  
  ##### 2.2 Country counts ####
  # Get the counts of species per country based on point occurrences
  siteSpeciesCounts <- data %>%
    countryHarmoniseR(data = .,
                      countryColumn = "country_suggested") %>%
    dplyr::group_by(scientificName, country_suggested) %>%
    dplyr::count() %>%
    dplyr::mutate(dataFrom = "points") %>%
    dplyr::filter(!is.na(country_suggested)) %>%
    dplyr::group_by(n) %>% 
    dplyr::mutate(count_n = dplyr::n())
  
  
  
  #### 3.0 Sample the curve ####
    ##### 3.1 Plot curve ####
  # Build a plot of the points and the model
  suppressWarnings({
  (curvePlot <- ggplot2::ggplot(data = siteSpeciesCounts, 
                               ggplot2::aes(x = n, y = count_n)) +
     # Literature curve data 
     ggplot2::geom_function(fun = curveFunction, 
                            ggplot2::aes(colour = "#FFC125"), linetype = 1, linewidth = 1) +
     ggplot2::geom_point(data = siteSpeciesCounts,
                         ggplot2::aes(x = n, y = count_n, colour = "#FFC125"),
                         alpha = 0.65) +
     ggplot2::scale_colour_manual(values = c("#FFC125"),
                                  label = c("Literature"),
                                  name = "Data source") +
     ggplot2::labs(x = "Number of occurrences", y = "Count")+ 
     
     # Limit the x-axis to see the curve
     ggplot2::xlim(1,100) + 
     ggplot2::theme(legend.position.inside = c(0.76, 0.86),
                    legend.position = "inside",
                    panel.background = ggplot2::element_rect(fill = NA,
                                                             colour = "black",
                                                             linetype = "solid"),
                    panel.border = ggplot2::element_rect(fill = NA,
                                                         colour = "black",
                                                         linetype = "solid")) )
  
  # Save the plot
  ggplot2::ggsave(paste0(outPath, "/curvePlot.pdf"), 
                  plot = curvePlot, 
                  width = 6, height = 6, units = "in", dpi = 300)
  }) # END suppressWarnings
  
    ##### 3.2 Extract and sample curve ####
  # Get the coordinates from the gam model in ggplot2 and put them in a tibble
  curveExtraction <- dplyr::tibble(
    x_coords = ggplot2::ggplot_build(curvePlot)$data[[1]]$x,
    y_coords = ggplot2::ggplot_build(curvePlot)$data[[1]]$y)
  
  # Set the seed for a consistent result
  set.seed(1234)
  # Get a random sample based on the empirical dataset
  litSample <- sample(x = curveExtraction$x_coords, prob = curveExtraction$y_coords, 
                      replace = TRUE,
                      size = sampleSize) %>%
    round(0)
  
  
  
  
  ##### 3.3 Apply to missing species #####
  # Find the species that are not represented in the occurrence dataset
  noPointSpecies <- taxonomyAcceptedNames %>% 
    dplyr::pull(validName) %>%
    setdiff(data$scientificName)  %>%
    # Add a sample size that is drawn from the same distriubtion of the literature and no-occurrences
    # dataset and generated in litSample
    dplyr::tibble(scientificName = .,
                  n = litSample[1:length(.)],
                  dataFrom = "randomisedSample")
  
  # 
  data_totalCounts <- data_counts %>% 
    dplyr::bind_rows(noPointSpecies)
  
  
  #### 4.0 Manipulate data ####
  
  # Make a function to sample the country sample sizes
  countrySampler <- function(df = NULL){ 
    dataFrom <- NULL
    # Get the maximum value for curveExtraction$x_coords
    max_x <- max(curveExtraction$x_coords)
    # Test that there is sufficient length (i.e., at least one records with n > NA)
    lengthTest <- df$n[complete.cases(df$n)] %>% 
      length()
    # Get the maximum value for the country's sample size or set the minium to n = 1
    if(lengthTest > 0){
      max_n = max(df$n[complete.cases(df$n)])
    }else{
      max_n = 1
    }
    # Find the length of the indput data
    inputLength <- length(df$n) + 100
    # Take the smaller of these two values
    maxSample <- dplyr::if_else(max_x > max_n, max_n, max_x)
    # Now apply the maxxed distribution to the country's checklist data
    df <- df %>%
      dplyr::mutate(n = dplyr::if_else(is.na(n), 
                                       # Draw from the literature distribution
                                       sample(x = curveExtraction$x_coords[1:maxSample],
                                              prob = curveExtraction$y_coords[1:maxSample], 
                                              replace = TRUE,
                                              size = inputLength)[dplyr::row_number()] %>%
                                         round(0),
                                       # Or keep n
                                       n),
                    dataFrom = dplyr::if_else(is.na(dataFrom), "checklist", dataFrom))
    return(df)
  }# END countrySampler
  
  # Now, for those species-country combinations without ANY points, but that should have points
  # from the checklist, make n = a number drawn from the literature records distribution
  # and state that they came from the checklist
  country_speciesChecklistCounts <- checklistFile %>%
    dplyr::select(tidyselect::any_of(c("validName", "rNaturalEarth_name", "year",
                                       "author", "authorCount"))) %>%
    dplyr::full_join(siteSpeciesCounts, by = c("validName" = "scientificName",
                                                   "rNaturalEarth_name" = "country_suggested")) %>%
    # Group by country and then split into a list per group
    dplyr::group_by(rNaturalEarth_name) %>%
    dplyr::group_split() %>%
    # For each country, apply the literature distribution but with an n max of the empirical maximum
    # for that country
    lapply(., countrySampler) %>%
    dplyr::bind_rows() %>%
    # Fix up the year, author, and authorCount columns by removing and re-adding them
    dplyr::select(!tidyselect::any_of(c("year", "author", "authorCount"))) %>% 
    dplyr::left_join(taxonomyFile %>%
                       dplyr::select(tidyselect::any_of(c("validName", "year", 
                                                          "author", "authorCount"))),
                     by = "validName") %>%
    dplyr::distinct(validName, rNaturalEarth_name, .keep_all = TRUE)
  
  # Create the Chao input data including checklist data
  countryChaoData_checklist <- country_speciesChecklistCounts %>%
    dplyr::select(validName, rNaturalEarth_name, n) %>%
    tidyr::pivot_wider(names_from = rNaturalEarth_name,
                       values_from = n,
                       values_fill = 0) %>%
    # Create the rownames
    column_to_rownames_internal("validName") %>%
    dplyr::tibble()
  
  # Create the Chao input data including ONLY occurrence data
  countryChaoData_occs <- siteSpeciesCounts %>%
    dplyr::select(scientificName, country_suggested, n) %>%
    tidyr::pivot_wider(names_from = country_suggested,
                       values_from = n,
                       values_fill = 0) %>%
    # Create the rownames
    column_to_rownames_internal("scientificName") %>%
    dplyr::tibble()
  
  
  ##### 4.1 Continent ####
  # Download a world map to convert countries to continents
  worldMap <- rnaturalearth::ne_countries(returnclass = "sf",
                                          scale = 50, type = "countries") 
  
  # Turn the country occ data into a continent one
  continentOccs <- siteSpeciesCounts %>%
    dplyr::ungroup() %>%
    countryHarmoniseR(countryColumn = "country_suggested") %>%
    # Join first by name
    dplyr::left_join(worldMap %>% dplyr::select(name, continent) %>%
                       sf::st_drop_geometry(), by = c("country_suggested" = "name")) %>%
    # Then join by name_long
    dplyr::left_join(worldMap %>% dplyr::select(name_long, continent) %>%
                       sf::st_drop_geometry(), by = c("country_suggested" = "name_long")) %>%
    # Now merge these continent columns
    dplyr::mutate(continent = dplyr::if_else(is.na(continent.x),
                                             continent.y, continent.x)) %>%
    # drop interim and old count columns
    dplyr::select(!c("continent.x", "continent.y")) %>%
    dplyr::group_by(scientificName, continent) %>%
    dplyr::mutate(sum = sum(n)) %>%
    dplyr::mutate(name_continent = stringr::str_c(scientificName, continent, sep = "__"))
  
  
  # Turn the country checklist data into a continent one
  continentChecklist <- checklistFile %>% 
    dplyr::select(validName, rNaturalEarth_name) %>%
    dplyr::ungroup() %>%
    # Change some country names to better match the continent data
    countryHarmoniseR(countryColumn = "rNaturalEarth_name",
                      continentAnalysis = TRUE) %>%
    # Join first by name
    dplyr::left_join(worldMap %>% dplyr::select(name, continent) %>%
                       sf::st_drop_geometry(), by = c("rNaturalEarth_name" = "name")) %>%
    # Then join by name_long
    dplyr::left_join(worldMap %>% dplyr::select(name_long, continent) %>%
                       sf::st_drop_geometry(), by = c("rNaturalEarth_name" = "name_long")) %>%
    # Now merge these continent columns
    dplyr::mutate(continent = dplyr::if_else(is.na(continent.x),
                                             continent.y, continent.x)) %>%
    # drop interim and old count columns
    dplyr::select(!c("continent.x", "continent.y")) %>%
    dplyr::mutate(name_continent = stringr::str_c(validName, continent, sep = "__"))
  
  # Find the names in the checklist that are missing from the occurrence records
  continentNoPoints <- dplyr::symdiff(continentOccs$name_continent, continentChecklist$name_continent) %>%
    # Extract continent and species names into tibble columns and remove the . column
    dplyr::tibble(scientificName = stringr::str_extract(., ".*__") %>% stringr::str_remove("__"),
                  continent = stringr::str_extract(., "__.*") %>% stringr::str_remove("__"),
                  n = NA,
                  dataFrom = "litEstimate") %>%
    dplyr::select(!.) 
  
  # Combine the occurrence and checklist data 
  continentCounts <- continentOccs %>%
    dplyr::select(!c("country_suggested", "n", "name_continent")) %>%
    dplyr::rename(n = sum) %>%
    dplyr::bind_rows(continentNoPoints) %>%
    # Group by country and then split into a list per group
    dplyr::group_by(continent) %>%
    dplyr::group_split() %>%
    # For each country, apply the literature distribution but with an n max of the empirical maximum
    # for that country
    lapply(., countrySampler) %>%
    dplyr::bind_rows() %>%
    dplyr::distinct(scientificName, continent, .keep_all = TRUE) 
  
  # Pivot wider 
  continentWider <- continentCounts %>%
    dplyr::select(scientificName, continent, n) %>%
    tidyr::drop_na() %>% 
    tidyr::pivot_wider(names_from = continent,
                       values_from = n,
                       values_fill = 0) %>%
    # Create the rownames
    column_to_rownames_internal("scientificName") %>%
    dplyr::tibble()
  
  
  #### 5. Save data ####
  
  ##### 5.1 data prep ####
  # Get the difference between the taxonomy and occurrence records
  taxonomyNoPoints <- taxonomyAcceptedNames %>% 
    dplyr::pull(validName) %>%
    setdiff(data$scientificName)
  
  # Re-apply the original column name
  names(continentChecklist)[names(continentChecklist) == "countryColumn"] <- countryColumn
  names(siteSpeciesCounts)[names(siteSpeciesCounts) == "countryColumn"] <- siteSpeciesCounts
  
  
  # Make a list of the data inputs so that this can be run easily on another computer
  richnessInputs <- dplyr::lst(
    # The curve created by implementing mosaic::fitModel over the literature samples  
    curveExtraction,
    checklistFile,
    taxonomyFile,
    taxonomyAcceptedNames,
    taxonomyNoPoints,
    data_counts,
    continentChecklist,
    continentNoPoints,
    worldMap,
    # Get the counts of species per country based on point occurrences
    siteSpeciesCounts) 
  
    #### 5.2 Save data ####
  # Save the file for use on another computer
  richnessInputs %>%
    base::saveRDS(., 
                  file = paste0(outPath, "/richnessInputs.Rda"))
  # Return the data
return(richnessInputs)
  
}

requireNamespace("dplyr")
skip_if_not_installed("rnaturalearthdata")

# Load a test dataset
beesRaw <- BeeBDC::beesRaw
# Load the small testTaxonomy file
system.file("extdata", "testTaxonomy.rda", package="BeeBDC") |>
  load()
# Load the small checklist file
system.file("extdata", "testChecklist.rda", package="BeeBDC") |>
  load()

testOut <- richnessPrepR(
  data = beesRaw,
  # Download the taxonomy
  taxonomyFile = testTaxonomy,
  # Download the checklist
  checklistFile = testChecklist,
  curveFunction = function(x) (228.7531 * x * x^-log(12.1593)),
  sampleSize = 10000,
  countryColumn = "country_suggested",
  limitGlobal = NULL,
  outPath = tempdir()
)


# Test the expected results
testthat::test_that("richnessPrepR results successfuly matched", {
  testthat::expect_equal(length(testOut) , 10)
})
testthat::test_that("richnessPrepR names match", {
  testthat::expect_equal(names(testOut) ,
                         c("curveExtraction","checklistFile","taxonomyFile","taxonomyAcceptedNames",
                           "taxonomyNoPoints","data_counts","continentChecklist","continentNoPoints",
                           "worldMap","siteSpeciesCounts"    
                           ) )
})

# This function was written by James Dorey to chunk the bdc_country_from_coordinates function
  # to allow bigger datasets to be analysed without consuming too much RAM.
# This function was written on the 12th of May 2022. For questions, please email jbdorey[at]me.com
#' Get country names from coordinates
#' 
#' Because the [bdc::bdc_country_from_coordinates()] function is very RAM-intensive, this wrapper 
#' allows a user to specify chunk-sizes and only analyse a small portion of the occurrence data at a 
#' time. The prefix jbd_ is used to highlight the difference between this function and the original
#' [bdc::bdc_country_from_coordinates()].
#'
#' @param data A data frame or tibble. Occurrence records to use as input.
#' @param lat Character. The name of the column to use as latitude. Default = "decimalLatitude".
#' @param lon Character. The name of the column to use as longitude. Default = "decimalLongitude".
#' @param country Character. The name of the column containing country names. Default = "country.
#' @param stepSize Numeric. The number of occurrences to process in each chunk. Default = 1000000.
#' @param chunkStart Numeric. The chunk number to start from. This can be > 1 when you need to 
#' restart the function from a certain chunk. For example, can be used if R failed unexpectedly.
#' @param path Character. The directory path to a folder in which to save the running countrylist
#' csv file.
#' @param scale Passed to rnaturalearth's ne_countries().
#' Scale of map to return, one of 110, 50, 10 or 'small', 'medium', 'large'. Default = "large".
#' @param mc.cores Numeric. If > 1, the function will run in parallel
#' using mclapply using the number of cores specified. If = 1 then it will be run using a serial
#' loop. NOTE: Windows machines must use a value of 1 (see ?parallel::mclapply). Additionally,
#' be aware that each thread can use large chunks of memory.
#'  Default = 1.
#'
#' @return A data frame containing database_ids and a country column 
#' that needs to be re-merged with the data input.
#' @export
#' 
#' @importFrom dplyr %>%
#'
#' @examples
#' if(requireNamespace("rnaturalearthdata")){
#' library("dplyr")
#' data(beesFlagged)
#' HomePath = tempdir()
#' # Tibble of common issues in country names and their replacements
#' commonProblems <- dplyr::tibble(problem = c('U.S.A.', 'US','USA','usa','UNITED STATES',
#' 'United States','U.S.A','MX','CA','Bras.','Braz.','Brasil','CNMI','USA TERRITORY: PUERTO RICO'),
#'                                  fix = c('United States of America','United States of America',
#'                                  'United States of America','United States of America',
#'                                  'United States of America','United States of America',
#'                                  'United States of America','Mexico','Canada','Brazil','Brazil',
#'                                  'Brazil','Northern Mariana Islands','Puerto Rico'))
#'                                  
#' beesFlagged <- beesFlagged %>%
#'       # Replace a name to test
#'    dplyr::mutate(country = stringr::str_replace_all(country, "Brazil", "Brasil"))
#' 
#' beesFlagged_out <- countryNameCleanR(
#'   data = beesFlagged,
#'   commonProblems = commonProblems)
#' 
#' suppressWarnings(
#'   countryOutput <- jbd_CfC_chunker(data = beesFlagged_out,
#'                                    lat = "decimalLatitude",
#'                                    lon = "decimalLongitude",
#'                                    country = "country",
#'                                    # How many rows to process at a time
#'                                    stepSize = 1000000,
#'                                    # Start row
#'                                    chunkStart = 1,
#'                                    path = HomePath,
#'                                    scale = "medium"),
#'   classes = "warning")
#' 
#' 
#' # Left join these datasets
#' beesFlagged_out <- left_join(beesFlagged_out, countryOutput, by = "database_id")  %>% 
#'   # merge the two country name columns into the "country" column
#'   dplyr::mutate(country = dplyr::coalesce(country.x, country.y)) %>%
#'   # remove the now redundant country columns 
#'   dplyr::select(!c(country.x, country.y)) %>%
#'   # put the column back 
#'   dplyr::relocate(country) %>% 
#'   # Remove duplicates if they arose!
#'   dplyr::distinct()
#' 
#' # Remove illegal characters
#' beesFlagged_out$country <- beesFlagged_out$country %>%
#'   stringr::str_replace(., pattern = paste("\\[", "\\]", "\\?",
#'                                           sep=  "|"), replacement = "")
#' } # END if require


jbd_CfC_chunker <- function(data = NULL,
                            lat = "decimalLatitude",
                            lon = "decimalLongitude",
                            country = "country",
                            # How many rows to process at a time
                            stepSize = 1000000,
                            # Start row
                            chunkStart = 1,
                            scale = "medium",
                            path = tempdir(),
                            mc.cores = 1){
  BeeBDC_order <- . <- .data <- id_temp <- name_long <- geometry <- inData <- country_OG <- NULL
  #### 0.0 Prep ####
  startTime <- Sys.time()
    ##### 0.1 nChunks ####
  # Find the number of chunks needed to complete the run
  nChunks = ceiling(nrow(data)/stepSize)
    # Find the max nrow
  nrowMax <- nrow(data)
  # The chunkEnd is the same as the stepSize initially, but the chunkEnd will change with each 
    # iteration
    ##### 0.3 chunkEnd ####
  chunkEnd = (chunkStart + stepSize) - 1

  
  ##### 0.4 Text out ####
  # Write user output
  message(paste(" - Running chunker with:", "\n",
                   "stepSize = ", 
                   format(stepSize, big.mark=",",scientific=FALSE), "\n",
                   "chunkStart = ", 
                   format(chunkStart, big.mark=",",scientific=FALSE), "\n",
                   "chunkEnd = ", 
                   format(chunkEnd, big.mark=",",scientific=FALSE), 
                   sep = ""))
 
  
  #### 1.0 Parallel ####
      ##### 1.1 Input function for parallel ####
    funCoordCountry <-
      function(inData) {
        suppressWarnings({
          check_require_cran("rnaturalearth")
          # check_require_github("ropensci/rnaturalearthdata")
        })

        # create an id_temp
        inData$id_temp <- 1:nrow(inData)
        minimum_colnames <- c(lat, lon)
        if(!all(minimum_colnames %in% colnames(inData))) {
          stop(
            "These columns names were not found in your database: ",
            paste(minimum_colnames[!minimum_colnames %in% colnames(inData)],
                  collapse = ", "),
            call. = FALSE
          )}
        # check if inData has a country column
        has_country <- any(colnames(inData) == country)
        
        if(!has_country) {
          inData$country <- NA}
        
        # converts coordinates columns to numeric
        inData <- inData %>%
          dplyr::mutate(decimalLatitude = as.numeric(.data[[lat]]),
                        decimalLongitude = as.numeric(.data[[lon]]))
        
        worldmap <- rnaturalearth::ne_countries(scale = scale, returnclass = "sf") %>%
          sf::st_make_valid()
        
        data_no_country <- inData %>%
          dplyr::filter(is.na(country) | country == "")
        
        if(nrow(data_no_country) == 0) {
          inData <- inData %>% dplyr::select(-id_temp)
        }else{
          # converts coordinates columns to spatial points
          suppressWarnings({
            data_no_country <-
              CoordinateCleaner::cc_val(
                x = data_no_country,
                lon = lon,
                lat = lat,
                verbose = FALSE
              ) %>%
              sf::st_as_sf(
                .,
                coords = c("decimalLongitude", "decimalLatitude"),
                remove = FALSE
              ) %>%
              sf::st_set_crs(., sf::st_crs(worldmap))
          })
          
          worldmap <-
            sf::st_as_sf(worldmap) %>% dplyr::select(name_long)
          
          # Extract country names from coordinates
          suppressWarnings({
            suppressMessages({
              ext_country <-
                data_no_country %>%
                dplyr::select(id_temp, geometry) %>%
                sf::st_intersection(., worldmap)
            })
          })
          
          ext_country$geometry <- NULL
          
          res <- dplyr::left_join(data_no_country, ext_country, by = "id_temp") %>%
            dplyr::distinct(id_temp, .keep_all = TRUE)
          
          id_replace <- res$id_temp 
          inData[id_replace, "country"] <- res$name_long
          inData <- inData %>% dplyr::select(-id_temp)
        }
        return(dplyr::as_tibble(inData))
      }
    
    ##### 1.2 Run mclapply ####
    # User output
    writeLines(paste(" - Starting parallel operation. Unlike the serial operation (mc.cores = 1)",
                     ", a parallel operation will not provide running feedback. Please be patient",
                     " as this function may take some time to complete. Each chunk will be run on",
                     " a seperate thread so also be aware of RAM usage."))
    loop_check_pf = data %>%
          # Make a new column with the ordering of rows
      dplyr::mutate(BeeBDC_order = dplyr::row_number()) %>%
          # Group by the row number and step size
      dplyr::group_by(BeeBDC_group = ceiling(BeeBDC_order/stepSize)) %>%
          # Split the dataset up into a list by group
      dplyr::group_split(.keep = TRUE) %>%
    # Run the actual function
      parallel::mclapply(., funCoordCountry,
                         mc.cores = mc.cores
    ) %>%
      # Combine the lists of tibbles
      dplyr::bind_rows()
    CountryList = dplyr::tibble(database_id = loop_check_pf$database_id, 
                                country = loop_check_pf$country, 
                                BeeBDC_order = loop_check_pf$BeeBDC_order) %>%
      # Arrange these
      dplyr::arrange(BeeBDC_order) %>%
        # Remove extra columns
      dplyr::select(!tidyselect::any_of("BeeBDC_order"))
  
  
    #### 2.0 Return ####
  colnames(CountryList) <- c("database_id", "country")
  
  endTime <- Sys.time()
  message(paste(
    " - Completed in ", 
    round(difftime(endTime, startTime), digits = 2 ),
    " ",
    units(round(endTime - startTime, digits = 2)),
    sep = ""))
  
    # Get a subset of the input data
  data <- data %>% 
    dplyr::select(tidyselect::any_of(c("database_id", "country")))
  
  # Clean a little
  CountryList <-  CountryList %>%
    # Drop na rows
    tidyr::drop_na(country)
  
    # Get a summary of the output
  summaryTable <- CountryList %>%
    dplyr::left_join(data, by = "database_id",
                     suffix = c("", "_OG")) %>%
      # Assign changed == 1 if the country name has changed from the original
    dplyr::mutate(changed = dplyr::if_else(is.na(country_OG), 1, 0))
  
  writeLines(paste0(" - We have updated the country names of ", 
                    format(sum(summaryTable$changed), big.mark = ","), 
                    " occurrences that previously had no country name assigned."))
  
  return(CountryList)
} # END function

requireNamespace("rnaturalearth")
requireNamespace("BeeBDC")
requireNamespace("dplyr")

# If rnaturalearthdata is present, run tests
if(requireNamespace("rnaturalearthdata")){

bees3sp <- BeeBDC::bees3sp

# Set up some fake data
fakeData <- dplyr::tibble(
  database_id = c("a1","a2","a3","a4", "a5","b1", "b2") ,
  decimalLatitude = c(-17.556329, 17.556329, 17.556329, 63.43333, -17.90000,-28.3497, 30.556553),
  decimalLongitude = c(178.007245, 178.007245, 178.007245, -17.90000, 63.43333, 114.6328, 108.832062),
  country = c("", "", "", "", "","",""),
  countryCode = c("FJ", "FJ", "US", "BO", "BO","AU","CH"),
  scientificName = c("sp1","sp1","sp1","sp1", "sp2","sp3","sp4")
)


countryOutput <- BeeBDC::jbd_CfC_chunker(data = bees3sp %>%
                                     tidyr::drop_na(decimalLatitude),
                                     lat = "decimalLatitude",
                                     lon = "decimalLongitude",
                                     country = "country",
                                     # How many rows to process at a time
                                     stepSize = 50,
                                     # Start row
                                     chunkStart = 1,
                                     path = tempdir(),
                                     scale = "medium")


  # Test expected number of rows
testthat::test_that("jbd_CfC_chunker row count expected", {
  testthat::expect_equal(nrow(countryOutput), 86)
})
  # Test output class
testthat::test_that("jbd_CfC_chunker expected class", {
  testthat::expect_type(countryOutput, "list")
})

# Test output class
testthat::test_that("jbd_CfC_chunker check that the input and output contries are the same — because nothing was changed for this test dataset.", {
  testthat::expect_equal(countryOutput$country, bees3sp %>%
                          tidyr::drop_na(decimalLatitude) %>%
                          dplyr::pull(country))
})


# Do a second round of test with fail data
countryOutput <- BeeBDC::jbd_CfC_chunker(data = bees3sp %>%
                                           tidyr::drop_na(decimalLatitude) %>%
                                           dplyr::bind_rows(fakeData),
                                         lat = "decimalLatitude",
                                         lon = "decimalLongitude",
                                         country = "country",
                                         # How many rows to process at a time
                                         stepSize = 50,
                                         # Start row
                                         chunkStart = 1,
                                         path = tempdir(),
                                         scale = "medium")


# Test expected number of rows
testthat::test_that("jbd_CfC_chunker row count expected", {
  testthat::expect_equal(nrow(countryOutput), 89) # Three extra rows from the fake data
})
# Test output class
testthat::test_that("jbd_CfC_chunker expected class", {
  testthat::expect_type(countryOutput, "list")
})

# Test output class
testthat::test_that("jbd_CfC_chunker check that the input and output contries are the same — because nothing was changed for this test dataset.", {
  testthat::expect_equal(countryOutput$country, c(bees3sp %>%
                           tidyr::drop_na(decimalLatitude) %>%
                           dplyr::pull(country), "Fiji", "Australia", "China"))
})



#     # Do a third round of test with fail data and multiple cores
#     countryOutput <- BeeBDC::jbd_CfC_chunker(data = bees3sp %>%
#                                                tidyr::drop_na(decimalLatitude) %>%
#                                                dplyr::bind_rows(fakeData),
#                                              lat = "decimalLatitude",
#                                              lon = "decimalLongitude",
#                                              country = "country",
#                                              # How many rows to process at a time
#                                              stepSize = 50,
#                                              # Start row
#                                              chunkStart = 1,
#                                              path = tempdir(),
#                                              scale = "medium", # Test at medium scale in this instance.
#                                              mc.cores = 2)
#     
#     
#     # Test expected number of rows
#     testthat::test_that("jbd_CfC_chunker row count expected", {
#       testthat::expect_equal(nrow(countryOutput), 89) # Three extra rows from the fake data
#     })
#     # Test output class
#     testthat::test_that("jbd_CfC_chunker expected class", {
#       testthat::expect_type(countryOutput, "list")
#     })
#     
#     # Test output class
#     testthat::test_that("jbd_CfC_chunker check that the input and output contries are the same — because nothing was changed for this test dataset.", {
#       testthat::expect_equal(countryOutput$country, c(bees3sp %>%
#                                                         tidyr::drop_na(decimalLatitude) %>%
#                                                         dplyr::pull(country), "Fiji", "Australia", "China"))
#     })
#     
#     


} # END if require

# Description of the BeeBDC test taxonomic dataset
# 17th June 2024

#' An example of the beesTaxonomy file
#'
#'
#' A small test taxonomy file for package tests. This dataset was built by filtering the 
#' taxonomy data from the three test datasets, beesFlagged, beesRaw, bees3sp.
#'
#' @docType data
#'
#' @usage data("testTaxonomy", package = "BeeBDC")
#'
#' @format An object of class \code{"tibble"}
#' \describe{
#'  \item{taxonomic_status}{Taxonomic status. Values are "accepted" or "synonym"}
#'  \item{source}{Source of the name.}
#'  \item{accid}{The id of the accepted taxon name or "0" if taxonomic_status == accepted.}
#'  \item{id}{The id number for the taxon name.}
#'  \item{kingdom}{The biological kingdom the taxon belongs to. For bees, kingdom == Animalia.}
#'  \item{phylum}{The biological phylum the taxon belongs to. For bees, phylum == Arthropoda.}
#'  \item{class}{The biological class the taxon belongs to. For bees, class == Insecta.}
#'  \item{order}{The biological order the taxon belongs to. For bees, order == Hymenoptera.}
#'  \item{family}{The family of bee which the species belongs to.}
#'  \item{subfamily}{The subfamily of bee which the species belongs to.}
#'  \item{tribe}{The tribe of bee which the species belongs to.}
#'  \item{subtribe}{The subtribe of bee which the species belongs to.}
#'  \item{validName}{The valid scientific name as it should occur in the 'scientificName" column in a Darwin Core file.}
#'  \item{canonical}{The scientificName without the scientificNameAuthority.}
#'  \item{canonical_withFlags}{The scientificName without the scientificNameAuthority and with Discover Life taxonomy flags.}
#'  \item{genus}{The genus the bee species belongs to.}
#'  \item{subgenus}{The subgenus the bee species belongs to.}
#'  \item{species}{The specific epithet for the bee species.}
#'  \item{infraspecies}{The infraspecific epithet for the bee addressed.}
#'  \item{authorship}{The author who described the bee species.}
#'  \item{taxon_rank}{Rank for the bee taxon addressed in the entry.}
#'  \item{notes}{Additional notes about the name/taxon.}
#' }
#' @references This dataset is a subset of the beesTaxonomy file described in:
#' Dorey, J.B., Fischer, E.E., Chesshire, P.R., Nava-Bolaños, A., O’Reilly, R.L., Bossert, S., Collins, S.M., Lichtenberg, E.M., Tucker, E., Smith-Pardo, A., Falcon-Brindis, A., Guevara, D.A., Ribeiro, B.R., de Pedro, D., Hung, J.K.-L., Parys, K.A., McCabe, L.M., Rogan, M.S., Minckley, R.L., Velzco, S.J.E., Griswold, T., Zarrillo, T.A., Jetz, W., Sica, Y.V., Orr, M.C., Guzman, L.M., Ascher, J., Hughes, A.C. & Cobb, N.S. (2023) A globally synthesised and flagged bee occurrence dataset and cleaning workflow. Scientific Data, 10, 1–17. https://www.doi.org/10.1038/S41597-023-02626-W
#' @keywords datasets
#' @examples
#' 
#' beesRaw <- BeeBDC::testTaxonomy
#' head(testTaxonomy)
#' 
"testTaxonomy"

#' Identify transposed geographic coordinates
#'
#' This function flags and corrects records when latitude and longitude appear
#' to be transposed. 
#' This function will preferably use the countryCode column generated by 
#' [bdc::bdc_country_standardized()].
#'
#' @family prefilter
#' @param data A data frame or tibble. Containing a unique identifier for each record,
#' geographical coordinates, and country names. Coordinates must be expressed
#' in decimal degrees and WGS84.
#' @param idcol A character string. The column name with a unique record identifier.
#' Default = "database_id".
#' @param sci_names A character string. The column name with species' scientific
#' names. Default = "scientificName".
#' @param lat A character string. The column name with latitudes. Coordinates must
#' be expressed in decimal degrees and WGS84. Default = "decimalLatitude".
#' @param lon A character string. The column name with longitudes. Coordinates must be
#' expressed in decimal degrees and WGS84. Default = "decimalLongitude".
#' @param country A character string. The column name with the country
#' assignment of each occurrence record. Default = "country".
#' @param countryCode A character string. The column name containing an ISO-2 country code for 
#' each record.
#' @param border_buffer Numeric. Must have value greater than or equal to 0. 
#' A distance in decimal degrees used to
#' created a buffer around each country. Records within a given country and at
#' a specified distance from the border will be not be corrected.
#' Default = 0.2 (~22 km at the equator).
#' @param save_outputs Logical. Indicates if a table containing transposed coordinates should be 
#' saved for further inspection. Default = FALSE.
#' @param fileName A character string. The out file's name.
#' @param path A character string. A path as a character vector for where to create the directories
#' and save the figures. If 
#' no path is provided (the default), the directories will be created using [here::here()].
#' @param scale Passed to rnaturalearth's ne_countries().
#' Scale of map to return, one of 110, 50, 10 or 'small', 'medium', 'large'. Default = "large".
#' @param mc.cores Numeric. If > 1, the jbd_correct_coordinates function will run in parallel
#' using mclapply using the number of cores specified. If = 1 then it will be run using a serial
#' loop. NOTE: Windows machines must use a value of 1 (see ?parallel::mclapply). Additionally,
#' be aware that each thread can use large chunks of memory.
#'  Default = 1.#'
#' @details This test identifies transposed coordinates based on mismatches between the 
#' country provided for a record and the record's latitude and longitude coordinates. Transposed
#' coordinates often fall outside of the indicated country (i.e., in other
#' countries or in the sea). Different coordinate transformations are
#' performed to correct country/coordinates mismatches. Importantly, verbatim
#' coordinates are replaced by the corrected ones in the returned database. A
#' database containing verbatim and corrected coordinates is created in
#' "Output/Check/01_coordinates_transposed.csv" if save_outputs == TRUE. The
#' columns "country" and "countryCode" can be retrieved by using the function
#' [bdc::bdc_country_standardized].
#'
#' @return A tibble containing the column "coordinates_transposed" which indicates if 
#' verbatim coordinates were not transposed (TRUE). Otherwise
#' records are flagged as (FALSE) and, in this case, verbatim coordinates are
#' replaced by corrected coordinates.
#'
#' @importFrom readr write_excel_csv
#' @importFrom dplyr tibble rename mutate select contains pull
#' @importFrom here here
#'
#' @export
#' 
#' @importFrom dplyr %>%
#'
#' @examples
#' \donttest{
#' if(require("rnaturalearthdata") && require("bdc")){
#' database_id <- c(1, 2, 3, 4)
#' scientificName <- c(
#'   "Rhinella major", "Scinax ruber",
#'   "Siparuna guianensis", "Psychotria vellosiana"
#' )
#' decimalLatitude <- c(63.43333, -14.43333, -41.90000, -46.69778)
#' decimalLongitude <- c(-17.90000, -67.91667, -13.25000, -13.82444)
#' country <- c("BOLIVIA", "bolivia", "Brasil", "Brazil")
#'
#' x <- data.frame(
#'   database_id, scientificName, decimalLatitude,
#'   decimalLongitude, country
#' )
#'
#' # Get country codes
#' x <- bdc::bdc_country_standardized(data = x, country = "country")
#'
#' jbd_coordinates_transposed(
#'   data = x,
#'   idcol = "database_id",
#'   sci_names = "scientificName",
#'   lat = "decimalLatitude",
#'   lon = "decimalLongitude",
#'   country = "country_suggested",
#'   countryCode = "countryCode",
#'   border_buffer = 0.2,
#'   save_outputs = FALSE,
#'   scale = "medium"
#' ) 
#' } # END if require
#' } # RND donttest
#'
jbd_coordinates_transposed <- function(data,
           idcol = "database_id",
           sci_names = "scientificName",
           lat = "decimalLatitude",
           lon = "decimalLongitude",
           country = "country",
           countryCode = "countryCode",
           border_buffer = 0.2,
           save_outputs = FALSE,
           fileName = NULL,
           scale = "large",
           path = NULL,
           mc.cores = 1) {
    decimalLatitude <- decimalLongitude <- database_id <- scientificName <- NULL
    
    requireNamespace("dplyr")

    suppressWarnings({
      check_require_cran("rnaturalearth")
      check_require_cran("readr")
    })
    
    # Ensure that working directories are maintain on exit from function
    oldwd <- getwd()           # code line i 
    on.exit(setwd(oldwd))        # code line i+1 
    
    # Copy original wd
    OGwd <- getwd()
    
    sf::sf_use_s2(TRUE)
    
    data <- dplyr::tibble(data)
    minimum_colnames <-
      c(idcol, sci_names, lat, lon, country, countryCode)
    
    if (length(minimum_colnames) < 6) {
      stop("Fill all function arguments: idcol, sci_names, lon, lat, and
         country")
    }
    
    if (!all(minimum_colnames %in% colnames(data))) {
      stop(
        "These columns names were not found in your database: ",
        paste(minimum_colnames[!minimum_colnames %in% colnames(data)],
              collapse = ", "),
        call. = FALSE
      )
    }

    # Temporarily change names of the collumn .summary to avoid error of duplicated coordinates
    if(".summary" %in% names(data)){
      w <- which(names(data) == ".summary")
      names(data)[w] <- "temp_summary"
    }
    
    # Standardizing columns names
    data <-
      data %>%
      dplyr::rename(
        database_id = {{ idcol }},
        decimalLatitude = {{ lat }},
        decimalLongitude = {{ lon }},
        scientificName = {{ sci_names }},
        countryCode = {{ countryCode }}
      )

    # converts coordinates columns to numeric
    data <-
      data %>%
      dplyr::mutate(
        decimalLatitude = as.numeric(decimalLatitude),
        decimalLongitude = as.numeric(decimalLongitude)
      )

    worldmap <- jbd_get_world_map(scale = scale)  # get world map and country iso

    # Correct latitude and longitude transposed
    message("Correcting latitude and longitude transposed\n")
    corrected_coordinates <-
      jbd_correct_coordinates(
        data = data,
        x = "decimalLongitude",
        y = "decimalLatitude",
        sp = "scientificName",
        idcol = idcol,
        cntr_iso2 = "countryCode",
        world_poly = worldmap,
        world_poly_iso = "iso2c",
        border_buffer = border_buffer,
        mc.cores = mc.cores
      )
    
    if (!is.null(corrected_coordinates)) {
      
      # Exports a table with verbatim and transposed xy
      corrected_coordinates <-
        corrected_coordinates %>%
        dplyr::select(database_id, scientificName, dplyr::contains("decimal"))
      
      if (save_outputs) {
        jbd_create_dir(path = path)
        setwd(path)
        corrected_coordinates %>%
          readr::write_excel_csv(
            paste(path, fileName, sep = "/"),
                           append = TRUE)
        message(
          paste(
            "\nCheck database containing coordinates corrected in:\nOutput/Check/01_coordinates_transposed.csv",
            "\n", "This will be appended to any existing rows from THIS run."
          )
        )
      }
      
      # finding the position of records with lon/lat modified
      w <-
        which(data %>% dplyr::pull(database_id) %in% (corrected_coordinates %>% dplyr::pull(database_id)))
      
      data[w, "decimalLatitude"] <-
        corrected_coordinates[, "decimalLatitude_modified"]
      
      data[w, "decimalLongitude"] <-
        corrected_coordinates[, "decimalLongitude_modified"]
      
      # Flags transposed coordinates
      data$coordinates_transposed <- TRUE
      data[w, "coordinates_transposed"] <- FALSE
      
      # Return collumn .summary
      if("temp_summary" %in% names(data)){
        w <- which(names(data) == "temp_summary")
        names(data)[w] <- ".summary"
      }
      
      message(
        paste(
          "\njbd_coordinates_transposed:\nCorrected",
          format(sum(data$coordinates_transposed == FALSE, na.rm = TRUE), big.mark = ","),
          "records.\nOne columns were added to the database.\n"
        )
      )
      
      return(data)
    } else{
      
      # Return collumn .summary
      if("temp_summary" %in% names(data)){
        w <- which(names(data) == "temp_summary")
        names(data)[w] <- ".summary"
      }
      
      message("No latitude and longitude were transposed\n")
      return(data)
    }
    setwd(OGwd)
  }

  #' Internal function. Detects and corrects transposed geographic coordinates
  #'
  #' This functions detects mismatches between country names informed coordinates.
  #' Once detects, transposed coordinates are corrected by the used of different
  #' coordinates transformations by using the 'jbd_coord_trans' function.
  #'
  #' @param data data.frame. Containing an unique identifier for each records,
  #' geographical coordinates, and country names. Coordinates must be expressed in
  #' decimal degree and in WGS84.
  #' @param x character string. The column name with longitude. Default =
  #' "decimalLongitude".
#' @param y character string. The column name with latitude Default =
#' "decimalLatitude".
#' @param sp character string. The column name with species scientific name.
#' Default = "scientificName".
#' @param idcol idcol character string. The column name with an unique record
#' identifier. #' Default = "idcol".
#' @param cntr_iso2 character string. The column name with the country code
#' assignment of each record. Default = "country_code".
#' @param world_poly polygon. Borders of the world.
#' @param world_poly_iso charterer sting. Iso2 code column of country polygon
#' database
#' @param border_buffer numeric. A distance in decimal degrees used to created a
#' buffer around the country. Records within a given country and at a specified
#' distance from its coast will be not be corrected. Default = 0.2 (~20 km at
#' the equator).
#' @param mc.cores Numeric. If > 1, the jbd_correct_coordinates function will run in parallel
#' using mclapply using the number of cores specified. If = 1 then it will be run using a serial
#' loop. NOTE: Windows machines must use a value of 1 (see ?parallel::mclapply). Additionally,
#' be aware that each thread can use large chunks of memory.
#'  Default = 1.
#'
#' @return Internal function
#'
#' @importFrom CoordinateCleaner cc_val clean_coordinates
#' @importFrom dplyr filter mutate as_tibble select all_of pull bind_rows distinct relocate left_join %>%
#'
#' @noRd
#'
#' @examples
#' \donttest{
#' # This is an internal function, that is exported for clarity and for users to potentially trace
#' errors.
#' }
jbd_correct_coordinates <-
  function(data,
           x,
           y,
           sp,
           idcol,
           cntr_iso2,
           world_poly,
           world_poly_iso,
           border_buffer,
           mc.cores = 1) {
    . <- decimalLatitude <- decimalLongitude <- .summary <- iso2c <- jbd_coord_trans <- ':=' <- NULL
    indexMatch <- countryCode <- countryMatch <- database_id <- NULL
    #### 0.0 Perperation ####
    ##### 0.1 Simplify map ####
      # Simplify the world map ONCE to be used later
    simplePoly <- world_poly %>% sf::st_drop_geometry() %>%
      dplyr::mutate(indexMatch = dplyr::row_number())
    
    ##### 0.2 Create function ####
      # Create the coord_trans function for internal use
    jbd_coord_trans <-
      function(data) {
        . <- ':=' <- indexMatch <- NULL
        data <-
          data %>% dplyr::select(
            dplyr::all_of(x),
            dplyr::all_of(y),
            dplyr::all_of(cntr_iso2),
            dplyr::all_of(idcol)
          ) 
        
        names(data)[names(data) == idcol] <- "idcol"
        
        d1 <- data.frame(x = data[, x], y = -data[, y], idcol = data[, "idcol"])
        d2 <- data.frame(x = -data[, x], y = data[, y], idcol = data[, "idcol"])
        d3 <- data.frame(x = -data[, x], y = -data[, y], idcol = data[, "idcol"])
        d4 <- data.frame(x = data[, y], y = data[, x], idcol = data[, "idcol"])
        d5 <- data.frame(x = data[, y], y = -data[, x], idcol = data[, "idcol"])
        d6 <- data.frame(x = -data[, y], y = data[, x], idcol = data[, "idcol"])
        d7 <- data.frame(x = -data[, y], y = -data[, x], idcol = data[, "idcol"])
        
        d.list <- list(d1, d2, d3, d4, d5, d6, d7)
        rm(list = paste0("d", 1:7))
        d.list <- lapply(d.list, function(x) {
          colnames(x) <- c("x", "y", "idcol")
          return(x)
        })
        
        over_list <- list()
        
        for (d in 1:length(d.list)) {
            # Check for coordinate validity first
          caluse <- d.list[[d]] %>% 
              # Remove coordinates that dont land on [our] earth
            dplyr::filter(!y > 90) %>% dplyr::filter(!y < -90) %>%
            dplyr::filter(!x > 180) %>% dplyr::filter(!x < -180) 

            # IF The coordinates do land on [our] earth, then turn them into sf objects and 
          # check if they overlap with a country
          if(nrow(caluse) > 0){
            caluse <- caluse %>%
            sf::st_as_sf(., coords = c("x", "y"), crs = sf::st_crs("WGS84")) %>%
            sf::st_make_valid()
          suppressWarnings({
            overresult <- sf::st_intersects(caluse, world_poly) %>%
              # return a tibble with the index of each match or NA where there was no match
              dplyr::tibble(indexMatch = . ) %>%
              # Convert to numeric
              dplyr::mutate(indexMatch = indexMatch %>% as.numeric()) %>%
              dplyr::left_join(simplePoly,
                               by = "indexMatch") %>%
                # Add in the database_id
              dplyr::bind_cols(caluse %>% sf::st_drop_geometry())
          })}else{
            overresult = tibble()
          }
          
          if(nrow(overresult) > 0){
            colnames(d.list[[d]]) <-
              c(paste0(x, "_modified"), paste0(y, "_modified"), "idcol")
            over_list[[d]] <- dplyr::left_join(d.list[[d]], data, by = "idcol") %>%
              dplyr::left_join(overresult, by = "idcol")
            rm(caluse)
            filt <-
              which(over_list[[d]][cntr_iso2] == over_list[[d]][world_poly_iso]) 
          }else{
            filt = dplyr::tibble()
          }
          if (length(filt) > 0) {
            over_list[[d]] <- over_list[[d]][filt,]
          } else {
            over_list[[d]] <- NULL
          }
          rm(list = c("overresult", "filt"))
        }
        
        rm(d.list)
        
        non_empty_list_test <- !sapply(over_list <- over_list, is.null)
        
        if (any(non_empty_list_test)) {
          over_list <- over_list[non_empty_list_test]
          over_list <- dplyr::bind_rows(over_list) 
        } else{
          over_list <- dplyr::tibble(
            decimalLongitude = double(),
            decimalLatitude = double(),
            countryCode = character(),
            database_id = character()
          )
        }
        # Return the database_id column to its correct name
        colnames(over_list)[colnames(over_list) == "idcol"] = idcol
  
        return(over_list)
      }
    
    
    
      #### 1.0 data prep ####
    x_mod <- paste0(x, "_modified")
    y_mod <- paste0(y, "_modified")
    
    occ_country <- data %>% dplyr::filter(!is.na(data[[cntr_iso2]]))
    
      #### 2.0 CoordinateCleaner ####
    # Filter occurrences database to avoid error in clean_coordinates errors
    suppressWarnings({
      suppressMessages({
        occ_country <-
          occ_country %>%
          CoordinateCleaner::cc_val(., lon = x, lat = y) %>%
          dplyr::mutate(
            decimalLatitude = as.numeric(decimalLatitude),
            decimalLongitude = as.numeric(decimalLongitude)
          )
      })
    })
    
    # Detect records outside a country
      # Convert to sf object
    suppressWarnings({
    countryTest <- occ_country %>% 
      sf::st_as_sf(coords = c(x, y), crs = sf::st_crs(world_poly)) %>%
        # Perform intersect operation with world_poly
      sf::st_intersects(., world_poly) %>% 
        # return a tibble with the index of each match or NA where there was no match
      dplyr::tibble(indexMatch = . ) %>%
        # Convert to numeric
      dplyr::mutate(indexMatch = indexMatch %>% as.character() %>% as.numeric()) %>%
      dplyr::left_join(simplePoly,
                       by = "indexMatch")
    })# END suppressWarnings
      # Join with the original dataset to find the database_ids of those occurrences that 1. do not
        # match with their supplied country code and/or 2. fall in the ocean (are NA)
    countryTest <- occ_country %>%
        # Get a subset of columns
      dplyr::select(tidyselect::all_of(c("database_id", cntr_iso2))) %>%
        # Bind columns with the original data
      dplyr::bind_cols(countryTest) %>%
        # Make a column to test country matches by
      dplyr::mutate(countryMatch = dplyr::if_else(countryCode == iso2c,
                                                  TRUE, FALSE)) %>%
        # Filter to failed and ocean occurrences
      dplyr::filter(countryMatch == FALSE | is.na(countryMatch))
    
    # Separate those records outside their countries
    occ_country <-
      occ_country %>%
      dplyr::filter(database_id %in% countryTest$database_id)
    
    # now this database have all those records with potential error that be
    # corrected
    message(occ_country %>% nrow(), " occurrences will be tested")
    
    # If occ_country have no data 
    if(nrow(occ_country)==0){
      return(NULL)
    }
    
    # Split database by country code - cntr_iso2
    occ_country <-
      split(occ_country, occ_country[cntr_iso2])
    
    # JBD edit - Remove empty elements from list before testing.
    occ_country <- occ_country[sapply(occ_country, function(x) dim(x)[1]) > 0]
    
    
    # jbd_coord_trans() function will try different coordinate transformations
    # to correct georeferenced occurrences
    coord_test <- list()
    
    
    #### 3.0 Run function ####
    ##### 3.1 Run mclapply ####
      # Run the actual function
coord_test <- parallel::mclapply(occ_country, jbd_coord_trans,
                     mc.cores = mc.cores
)


    # elimination from the list those countries without correction
    filt <- sapply(coord_test, function(x) nrow(x) > 0)
    
    if(any(filt)){
      coord_test <- coord_test[filt]
      
      # Elimination of those records near to country border (to avoid flip
      # coordinates or sign that fall too close to country border)
      
      for (i in 1:length(coord_test)) {
        n <-
          coord_test[[i]] %>%
          dplyr::select(dplyr::all_of(cntr_iso2)) %>%
          unique() %>%
          dplyr::pull()
        
          # Select only the relevant polygon to buffer
        my_country2 <- world_poly %>%
          dplyr::filter(iso2c %in% n)
        
        # Here filter polygon based on your country iso2c code
        my_country2 <-
          my_country2 %>%
          dplyr::filter(iso2c %in% n) %>%
          # JBD - France was failing to buffer using raster due to TopologyException. Use sf instead.
          sf::st_as_sf() %>% sf::st_buffer(border_buffer) 
        # JBD - turned off for above reason
        # 0.5 degree ~50km near to equator
        # my_country2 <- raster::buffer(my_country, width = border_buffer)
        #  > my_country2
        #  class       : SpatialPolygons 
        #  features    : 1 
        #  extent      : -180, 180, -90, -60.51621  (xmin, xmax, ymin, ymax)
        #  crs         : NA 
        
        coord_sp <- sf::st_as_sf(coord_test[[i]] %>% dplyr::select({{ x }}, {{ y }}),
                                 coords = c(x, y))
        
        sf::st_crs(coord_sp) <- sf::st_crs(my_country2)
        over_occ <- sf::st_join(coord_sp, my_country2) %>%
          dplyr::pull(iso2c) 
        
        # Eliminate as corrected those records too close to country border
        coord_test[[i]] <-
          coord_test[[i]] %>% dplyr::filter(is.na(over_occ))
      }
      
      # Elimination of those records with more than two possible corrections
      coord_test <-
        dplyr::bind_rows(coord_test) %>%
        dplyr::as_tibble() # binding dataframes allocated in the list in a single one
      
      coord_test <-
        coord_test[!duplicated(coord_test[idcol]), ] %>%
        dplyr::relocate(dplyr::all_of(idcol), dplyr::all_of(x), dplyr::all_of(y))
      
      # Merge coord_test with other columns of occurrence database
      coord_test <-
        dplyr::left_join(coord_test,
                         data %>% dplyr::select(-c({{ x }}, {{ y }}, {{ cntr_iso2 }})),
                         by = idcol
        )
      
      return(coord_test)
    }else{
      return(NULL)
    }
  }

#### 5. formatted combiner ####
# Currently works to combine the USGS data with the rest of the data when the string is as below.
# But, it should also work for any properly-formatted files
# strings = c("USGS_[a-zA-Z_]+[0-9]{4}-[0-9]{2}-[0-9]{2}")

#' Combine the formatted USGS data with the main dataset
#' 
#' Merges the Darwin Core version of the USGS dataset that was created using [BeeBDC::USGS_formatter()]
#' with the main dataset.
#'
#' @param path A directory as character. The directory to look in for the formatted USGS data. 
#' @param strings A regex string. The string to find the most-recent formatted USGS dataset.
#' @param existingOccurrences A data frame. The existing occurrence dataset.
#' @param existingEMLs An EML file. The existing EML data file to be appended.
#'
#' @return A list with the combined occurrence dataset and the updated EML file.
#' @export
#' 
#' @importFrom dplyr %>%
#' @importFrom stats setNames
#'
#' @examples
#' \dontrun{
#' DataPath <- tempdir()
#' strings = c("USGS_DRO_flat_27-Apr-2022")
#'     # Combine the USGS data and the existing big dataset
#' Complete_data <- formattedCombiner(path = DataPath, 
#'                                     strings = strings, 
#'                                     # This should be the list-format with eml attached
#'                                     existingOccurrences = DataImp$Data_WebDL,
#'                                     existingEMLs = DataImp$eml_files) 
#'                                     }

formattedCombiner <- function(path, 
                               strings, 
                               existingOccurrences,
                               existingEMLs){
  # locally bind variables to the function
  . <- NULL
  
  requireNamespace("dplyr")
  requireNamespace("xml2")
  # Find all of the previously-produced data files
  BeeData_Locs <- file.info(list.files(path, full.names = T, pattern = strings,
                                       recursive = TRUE))
  # Check if the data are present
  if(nrow(BeeData_Locs) == 0){ # If there are no data matching the name...
    stop(" - Bugger it, R can't find any files produced by our package in the path provided :(")
  }
  # Find the most-recent files based on date in file name
  file_dates <- stringr::str_extract(rownames(BeeData_Locs), 
                                     pattern = "[0-9]{4}-[0-9]{2}-[0-9]{2}") %>%
    # Sort from most- to least-recent files
    sort(decreasing = TRUE) 
  # Return the strings containing this date
  most_recent <- stringr::str_subset(rownames(BeeData_Locs), 
                                     pattern = file_dates[1]) 
  # Return information to user
  writeLines(paste(" - Great, R has detected some files. These files include: ", "\n",
                   paste(most_recent, collapse = "\n") ), sep = "")
  
  #### CSV import ####
  # IF there is ONLY a complete .csv file among the most-recent files...
  if(any(stringr::str_detect(most_recent, paste(strings,".csv", sep = ""))) == TRUE &&
     all(stringr::str_detect(most_recent, paste(strings,".rds", sep = ""))) == FALSE){
    writeLines(paste(" - .csv export version found. Loading this file..."))
    ColTypes <- ColTypeR()
    # Find the most-recent .csv occurrence file
    # Find the file that deos NOT include "attribute" or "problems" in the string
    occ_file <- most_recent[stringr::str_which(most_recent, negate = TRUE,
                                               paste("attribute|problems", sep = ""))] %>%
      # Read in .csv file and supress warnings
      readr::read_csv(col_types = ColTypes) %>%
      suppressWarnings(., classes = "warning")
    # Check if attributes file is in .rds format or not and preferentially read this if present
    Rdata_test <- most_recent[stringr::str_which(most_recent, paste("attribute", sep = ""))] %>%
      stringr::str_detect(pattern = ".rds") 
    # Find the most-recent .rds attributes file
    if(Rdata_test == TRUE){
      # Find the file that does include "attribute in the name
      attr_file <- most_recent[stringr::str_which(most_recent, paste("attribute", sep = ""))] %>%
        base::readRDS()
    } # END .rds portion of csv
    # Find the most-recent .csv attributes file
    if(Rdata_test == FALSE){
      # Find the file that does include "attribute in the name
      attr_file <- most_recent[stringr::str_which(most_recent, paste("attribute", sep = ""))] %>%
        readr::read_csv(col_types = readr::cols_only(
          dataSource = "c", alternateIdentifier = "c", title = "c", pubDate = "D", dateStamp = "c",
          doi = "c", downloadLink = "c", abstract = "c", citations = "c", downloadCitation = "c",
          rights = "c"))
      # Turn the abstract, citations, and rights into lists
      attr_file$abstract <- dplyr::lst(attr_file$abstract)
      attr_file$citations <- dplyr::lst(attr_file$citations)
      attr_file$rights <- dplyr::lst(attr_file$rights)
    } # END .rds portion of csv
    # Look for a .xml file
    if(any(stringr::str_detect(most_recent,pattern = ".xml")) == TRUE){
      # Find and read in the .xml file
      xml_file <- most_recent[stringr::str_which(most_recent, paste(".xml", sep = ""))] %>%
        file() %>%
        xml2::read_xml()
    } # END xml look
  } #END IF .csv
  
  #### Merge ####
  # print user information
  writeLines( paste(" - Merging occurrence and attribute files.", "\n",
                    "Depending on file size, this could take some time...","\n",
                    sep = ""))
  ##### dataset and attrs ####
  # Merge the datasets
    # If the dataset appears as a list, make it not a list.
  if(any(class(existingOccurrences) %in% "list")){
    existingOccurrences <- existingOccurrences[[1]]
  }
  bound_data <- dplyr::bind_rows(existingOccurrences %>%
                                   # Convert column types
                                   readr::type_convert(col_types = ColTypeR()), 
                                 occ_file)
  # Extract the existing attributes file
  extAttr_file <-  attributes(existingOccurrences)
  # Bind the dataSource tables together
  extAttr_file$dataSource <- dplyr::bind_rows(extAttr_file$dataSource, attr_file)
  # Update this in the bound_data tibble 
  attr(existingOccurrences, "dataSource") <- extAttr_file$dataSource
  # Replace the existing data with the newly-bound data
  existingOccurrences <- bound_data
  

  ##### xmls ####
  # IF the xml file exists, append it to the eml_files 
  if(exists("xml_file") == TRUE){
    # append the new xml_file to the existing ones
    existingEMLs <- append(existingEMLs, xml_file)
  }
  
  # Combine these files back into a list
  existing_data <- list(existingOccurrences, existingEMLs) %>%
    setNames(c("Data_WebDL", "eml_files"))

  
  # Return end product and print completion note
  writeLines(paste(" - Fin.", sep = "\n"))
  # Return the outfile
  return(existing_data)
} # COMPLETE formattedCombiner

requireNamespace("readr")
requireNamespace("emld")
requireNamespace("dplyr")

# If rnaturalearthdata is present, run tests
if(requireNamespace("rnaturalearthdata")){

# Be sure that the testData is not already in tempdir
tempdirFiles <- list.files(tempdir(), include.dirs = F, full.names = T, recursive = FALSE)
# Remove tempdir files from the prior run
unlink(tempdirFiles[stringr::str_detect(tempdirFiles, "files")] %>% normalizePath(),
       recursive = TRUE)

  # Create a test USGS dataset and save it to a temporary location
USGS_testData <- dplyr::tribble(
                                            ~scientificName,           ~identifiedBy, ~dateIdentified, ~decimalLatitude, ~decimalLongitude,   ~stateProvince, ~country,            ~county,                       ~municipality,             ~eventDate,  ~eventTime,       ~recordedBy,       ~eventID,       ~Location,                                                                                                                              ~samplingProtocol, ~samplingEffort, ~coordinateUncertaintyInMeters, ~datasetID, ~institutionCode,        ~datasetName,    ~occurrenceID,        ~recordId,                                                                                                                                                                                                                                                                                                                                                                      ~fieldNotes, ~sex,              ~id, ~dataSource,
                                    "Lasioglossum imitatum",            "Sam Droege",      "1/1/2005",          35.5917,          -83.0602, "North Carolina",    "USA",            "Swain",             "Great Smoky Mountains", "2002-04-02T10:30:00Z", 2.00811e+13, "Harold W. Ikerd",  "USGS_DRO337", "Purchase Knob",     "Technique used: pan trap|Bowls full upon collection: 5|Sampling Bowl/trap size: bowl 6.0oz|Trap colour: yellow-uv|Trap liquid: soap dawn",              NA,                             3L, "USGS_DRO",           "USGS", "USGS_DRO database", "USGS_DRO000001", "USGS_DRO000001",                                                                                                                                                                                                  "field_note: Field no blooms|DateEntered: 1/27/2017 9:35:36|DateScanned: 1/1/1900|ipAddress: 134.192.250.22|position: 23|time1: 2002-04-02 10:30:00|time2: 2002-04-02 16:30:00",  "f", "USGS_DRO000001", "USGS_data",
                                                         NA,                      NA,              NA,          35.5917,          -83.0602, "North Carolina",    "USA",            "Swain",             "Great Smoky Mountains", "2002-04-02T12:00:00Z", 2.00811e+13, "Harold W. Ikerd",  "USGS_DRO339", "Purchase Knob",    "Technique used: pan trap|Bowls full upon collection: 5|Sampling Bowl/trap size: bowl 12.0oz|Trap colour: pale blue|Trap liquid: soap dawn",              NA,                             3L, "USGS_DRO",           "USGS", "USGS_DRO database", "USGS_DRO000013", "USGS_DRO000013",                                                                                                                                                                                                                                                      "field_note: edge FX ;on road|ipAddress: 134.192.250.22|position: 23|time1: 2002-04-02 12:00:00|time2: 2002-04-02 17:30:00",   NA, "USGS_DRO000013", "USGS_data",
                                           "Andrena rugosa",           "John Ascher",      "10/1/200",          35.6114,          -83.5429,      "Tennessee",    "USA",           "Blount",             "Great Smoky Mountains", "2002-04-01T09:00:00Z", 2.00811e+13, "Harold W. Ikerd",  "USGS_DRO283",   "Goshen Prog",     "Technique used: pan trap|Bowls full upon collection: 5|Sampling Bowl/trap size: bowl 6.0oz|Trap colour: yellow-uv|Trap liquid: soap dawn",              NA,                             4L, "USGS_DRO",           "USGS", "USGS_DRO database", "USGS_DRO000376", "USGS_DRO000376",                                                                           "field_note: Phacelia fimbriata;Claytonia caroliniana;Erythronium americanum;Violoa rotundifolia; V. hastata; V.sororia; V.pedata;Violoa rotundifolia|DateEntered: 1/27/2017 9:35:36|DateScanned: 12/27/20|ipAddress: 172.130.3.67|position: 00|time1: 2002-04-01 09:00:00|time2: 2002-04-01 15:30:00",  "f", "USGS_DRO000376", "USGS_data",
                             "Plenoculus davisi_atlanticus",         "Matthias Buck",      "1/1/2009",       39.0507167,       -76.7778167,       "Maryland",    "USA",     "Anne Arundel", "Patuxent Wildlife Research Refuge", "2002-05-24T12:00:00Z", 2.00811e+13, "Harold W. Ikerd",  "USGS_DRO309",              NA, "Technique used: pan trap|Bowls full upon collection: 5|Sampling Bowl/trap size: bowl 12.0oz|Trap colour: pale blue-uv|Trap liquid: soap dawn",              NA,                             4L, "USGS_DRO",           "USGS", "USGS_DRO database", "USGS_DRO001232", "USGS_DRO001232",                                                                                                                                                                          "field_note: sandy area under powerline on north track|DateEntered: 1/27/2017 9:35:36|DateScanned: 4/21/200|ipAddress: 172.130.3.67|position: 04|time1: 2002-05-24 12:00:00|time2: 2002-05-25 11:00:00",  "m", "USGS_DRO001232", "USGS_data",
                                  "Chrysura kyrae/pacifica",          "Erika Tucker",      "12/6/200",          39.0457,          -76.7896,       "Maryland",    "USA",  "Prince George's", "Patuxent Wildlife Research Refuge", "2002-04-30T09:20:00Z",  2.0081e+13,      "Sam Droege",   "USGS_DRO70",              NA,   "Technique used: pan trap|Bowls full upon collection: 20|Sampling Bowl/trap size: bowl 12.0oz|Trap colour: white|Trap liquid: in field note",              NA,                             4L, "USGS_DRO",           "USGS", "USGS_DRO database", "USGS_DRO002016", "USGS_DRO002016",  "field_note: Blooming-a few sping beauties, mertensia, garlic mustard, not much else;habitat-bottomland;other-detergent test;other-sunny/wind 4mph/60's;test failed 4x5=20 bowls put out but sharpie permanent marker failed.|DateEntered: 1/27/2017 9:35:36|DateScanned: 12/9/200|ipAddress: 159.189.28.40|position: 04|time1: 2002-04-30 09:20:00|time2: 2002-04-30 00:00:00",  "u", "USGS_DRO002016", "USGS_data",
                                    "Melissodes subillatus",          "Karen Wright",      "7/1/2016",           38.909,           -76.683,       "Maryland",    "USA",  "Prince George's",                                  NA, "2002-06-18T00:00:00Z", 2.00811e+13,                NA,  "USGS_DRO376",              NA,         "Technique used: pan trap|Bowls full upon collection: 5|Sampling Bowl/trap size: bowl 12.0oz|Trap colour: blue|Trap liquid: soap dawn",              NA,                             3L, "USGS_DRO",           "USGS", "USGS_DRO database", "USGS_DRO002363", "USGS_DRO002363",                                                                                                                                                                                       "field_note: student conservation folks|DateEntered: 1/27/2017 9:35:36|DateScanned: 7/12/201|ipAddress: 134.192.250.22|position: 04|time1: 2002-06-18 00:00:00|time2: 2002-06-18 00:00:00",  "f", "USGS_DRO002363", "USGS_data",
                       "Augochloropsis metallica_metallica",           "Michael Orr",      "3/8/2011",           38.909,           -76.683,       "Maryland",    "USA",  "Prince George's",                                  NA, "2002-06-18T00:00:00Z", 2.00811e+13,                NA,  "USGS_DRO377",              NA,        "Technique used: pan trap|Bowls full upon collection: 5|Sampling Bowl/trap size: bowl 12.0oz|Trap colour: white|Trap liquid: soap dawn",              NA,                             3L, "USGS_DRO",           "USGS", "USGS_DRO database", "USGS_DRO002395", "USGS_DRO002395",                                                                                                                                                                                       "field_note: student conservation folks|DateEntered: 1/27/2017 9:35:36|DateScanned: 3/8/2011|ipAddress: 134.192.250.22|position: 04|time1: 2002-06-18 00:00:00|time2: 2002-06-18 00:00:00",  "f", "USGS_DRO002395", "USGS_data",
                                       "Osmia marilaunidii",          "Harold Ikerd",      "5/31/201",          33.3404,          -97.1366,          "Texas",    "USA",           "Denton",                              "RVAC", "2002-04-27T00:00:00Z", 2.01605e+13, "Harold W. Ikerd",  "USGS_DRO480",              NA,                           "Technique used: pan trap|Bowls full upon collection: 5|Sampling Bowl/trap size: bowl 12.0oz|Trap colour: pale blue",              NA,                             4L, "USGS_DRO",           "USGS", "USGS_DRO database", "USGS_DRO004865", "USGS_DRO004865",                                                                                                                                                                                                                              "DateEntered: 1/27/2017 9:35:36|DateScanned: 5/31/201|ipAddress: 159.189.28.152|position: 08|time1: 2002-04-27 00:00:00|time2: 2002-04-27 00:00:00",  "f", "USGS_DRO004865", "USGS_data",
                                     "Lasioglossum droegei",           "Jason Gibbs",      "12/9/201",          29.1361,         -103.1786,          "Texas",    "USA",         "Brewster",            "Big Bend National Park", "2002-05-01T06:00:00Z", 2.00811e+13, "Harold W. Ikerd",  "USGS_DRO856",              NA,         "Technique used: pan trap|Bowls full upon collection: 5|Sampling Bowl/trap size: bowl 12.0oz|Trap colour: blue|Trap liquid: soap dawn",              NA,                             4L, "USGS_DRO",           "USGS", "USGS_DRO database", "USGS_DRO010505", "USGS_DRO010505",                                                                                                                                                                                                                               "DateEntered: 1/27/2017 9:35:36|DateScanned: 12/13/20|ipAddress: 134.192.250.8|position: 56|time1: 2002-05-01 06:00:00|time2: 2002-05-02 12:00:00",  "f", "USGS_DRO010505", "USGS_data",
                                        "Osmia atriventris",       "Molly Rightmyer",      "10/19/20",          38.6916,          -77.0573,       "Maryland",    "USA",  "Prince George's",                        "Piscataway", "2003-04-02T16:40:00Z", 2.00811e+13,      "Sam Droege",  "USGS_DRO954",            "25",                          "Technique used: pan trap|Bowls full upon collection: 30|Sampling Bowl/trap size: bowl 3.25oz|Trap liquid: soap dawn",              NA,                             4L, "USGS_DRO",           "USGS", "USGS_DRO database", "USGS_DRO016225", "USGS_DRO016225",                                                                                                                                                    "field_note: 15 w 15 y, 0 missing, irregularly mown field|note: Transect 25|DateEntered: 1/27/2017 9:35:36|DateScanned: 10/20/20|ipAddress: 134.192.250.8|position: 04|time1: 2003-04-02 16:40:00|time2: 2003-04-03 17:40:00",  "m", "USGS_DRO016225", "USGS_data",
                                      "Epeolus scutellaris", "Rebekah Andrus Nelson",      "9/1/2007",          41.0704,          -71.8582,       "New York",    "USA",          "Suffolk",          "Montauk Point State Park", "2005-09-06T00:00:00Z", 2.00812e+13,      "Sam Droege", "USGS_DRO2169",              NA,                                                                                                                     "Technique used: hand net",              NA,                             4L, "USGS_DRO",           "USGS", "USGS_DRO database", "USGS_DRO039487", "USGS_DRO039487",                "field_note: sunny in the low 80's fair amount of wind;collected around the lighthouse on the point as well as on the bluff immediately adjacent on the Camp Hero Site.  Collected primarily off of goldenrod.|DateEntered: 1/27/2017 9:35:36|DateScanned: 10/9/200|ipAddress: 172.162.21.134|position: 09|time1: 2005-09-06 00:00:00|time2: 2005-09-06 00:00:00",  "m", "USGS_DRO039487", "USGS_data"
                       ) %>%
  # Save a temporary version of these data
  readr::write_excel_csv(., paste0(tempdir(), "/USGS_formatted_2023-01-27.csv"))

USGS_attrFile <- dplyr::tribble(
                   ~dataSource, ~alternateIdentifier,              ~title,     ~pubDate,   ~dateStamp,           ~doi,                                          ~downloadLink, ~abstract,               ~citations,                                                     ~downloadCitation,                                                                          ~rights,
                   "USGS_data",       "Not provided", "USGS_DRO database", "2022/11/19", "2022/11/19", "Not provided", "Not provided, contact Sam Droege at sdroege@usgs.gov",        NA, "Citations not provided", "Sam Droege. (2022-11-19). United States Geological Survey bee data.", "Rights are not provided. Please seek permission for data use from Same Droege."
                   ) %>%
  readr::write_excel_csv(., paste0(tempdir(), "/USGS_attribute_files2023-01-27.csv"))

  # Create a test database of occurrences
existingTestdb <- dplyr::tribble(
                                                               ~scientificName,        ~family, ~subfamily,         ~genus,       ~subgenus, ~subspecies,                 ~species, ~specificEpithet, ~infraspecificEpithet,                      ~acceptedNameUsage,   ~taxonRank, ~scientificNameAuthorship, ~identificationQualifier, ~higherClassification, ~identificationReferences, ~typeStatus, ~previousIdentifications, ~verbatimIdentification,        ~identifiedBy,       ~dateIdentified, ~decimalLatitude, ~decimalLongitude,                  ~stateProvince,           ~country,      ~continent,                                              ~locality, ~island,        ~county,            ~municipality, ~countryCode, ~level0Gid,     ~level0Name, ~level1Gid,          ~level1Name,         ~license,                                                                                                              ~issue,              ~eventDate,  ~eventTime, ~day, ~month, ~year,       ~basisOfRecord, ~type, ~occurrenceStatus, ~recordNumber,                                   ~recordedBy,                                     ~eventID,                                                                                        ~samplingProtocol, ~samplingEffort, ~individualCount, ~organismQuantity, ~coordinatePrecision, ~coordinateUncertaintyInMeters, ~spatiallyValid,                                                     ~catalogNumber,     ~gbifID,                             ~datasetID, ~institutionCode,                                                 ~datasetName,                                                                                                           ~otherCatalogNumbers,                                                        ~occurrenceID, ~taxonKey,                                ~coreid,                                       ~recordId,                                          ~collectionID,          ~verbatimScientificName,                                         ~verbatimEventDate, ~associatedTaxa, ~associatedOrganisms,                                                                               ~fieldNotes,     ~sex,                                             ~rights,                                            ~rightsHolder,                                                                                                 ~accessRights,                                                                                                              ~associatedReferences, ~bibliographicCitation,                                                                        ~references,                                                                                                                               ~informationWithheld, ~isDuplicateOf, ~hasCoordinate, ~hasGeospatialIssues,                                                                                                                                                                                                                                                     ~assertions,        ~occurrenceYear,                                    ~id, ~duplicateStatus, ~associatedOccurrences,                                                                                                                                          ~locationRemarks,          ~dataSource,
                                                                      "APIDAE",       "Apidae",         NA,             NA,              NA,          NA,                       NA,     "planifrons",                    NA, "Platytrigona planifrons (Smith, 1865)",     "family",                        NA,                       NA,                    NA,                        NA,          NA,                       NA,                      NA,      "Rasmussen, C.",                "2007",             -7.2,            146.65,               "Morobe Province", "Papua New Guinea",              NA,                                                     NA,      NA,             NA,                       NA,         "PG",         NA,              NA,         NA,                   NA, "CC-BY 3.0 (Au)",                                                                                                                  NA,                      NA,          NA,   NA,     6L, 1950L, "PRESERVED_SPECIMEN",    NA,         "PRESENT",            NA,                                   "Faddy, H.", "urn:australianmuseum.net.au:Events:1093889",                                                                                                       NA,              NA,               NA,                NA,                0.001,                         10000L,            TRUE,                                                         "K.240452",          NA,                                     NA,             "AM",                                                           NA,                                                                  "ecatalogue.irn:1212006 | urn:catalog:AM:Entomology:K.240452",              "urn:lsid:ozcam.taxonomy.org.au:AM:Entomology:K.240452",        NA,                                     NA,                                              NA,                                                     NA,                               NA,                                                         NA,              NA,                   NA,                                                                                        NA,       NA,                                                  NA,                                                       NA,                                                                                                            NA,                                                                                                                                 NA,                     NA,                                                                                 NA,                                                                                                                                                 NA,             NA,             NA,                   NA, "GEODETIC_DATUM_ASSUMED_WGS84 | MISSING_GEODETICDATUM | MISSING_GEOREFERENCEDBY | MISSING_GEOREFERENCEPROTOCOL | MISSING_GEOREFERENCESOURCES | MISSING_GEOREFERENCEVERIFICATIONSTATUS | MISSING_GEOREFERENCE_DATE | MISSING_TAXONRANK | TAXON_MATCH_HIGHERRANK", "1949-12-31T14:00:00Z", "0a3c9ba7-c666-4a42-8e93-f60789977722",               NA,                     NA,   "ecatalogue.LocCollectionEventLocal: \"Papua New Guinea, Morobe Province, Bulolo (7¬∞ 12' S, 146¬∞ 39' E) /06/1950 - /06/1950, Faddy, H.(Collector)\";",      "ALA_Apiformes",
                                                      "Bombus Latreille, 1802",       "Apidae",         NA,       "Bombus",              NA,          NA,                       NA,               NA,                    NA,                                      NA,      "GENUS",                        NA,                       NA,                    NA,                        NA,          NA,                       NA,                      NA,                   NA,                    NA,         44.99712,           1.00448,                              NA,                 NA,              NA,                                                     NA,      NA,     "Dordogne",                "Fleurac",         "FR",      "FRA",        "France", "FRA.10_1", "Nouvelle-Aquitaine",      "CC_BY_4_0",                                                                                             "RECORDED_DATE_INVALID",                      NA,          NA,   NA,     NA,    NA,  "HUMAN_OBSERVATION",    NA,         "PRESENT",            NA,                                   "(SPIPOLL)",       "fcb8e5ba-2a2b-11e9-a1a8-005056010096",                                                                                                       NA,              NA,               NA,                NA,                   NA,                          3777L,              NA,                                                                 NA,         NA, "5900C02E-AD2B-555E-E053-2614A8C02A2B",               NA,                                                           NA,                                                                                                                             NA,                               "fcb8e5ba-2a2b-11e9-a1a8-005056010096",  1340278L,                                     NA,                                              NA,                                                     NA,         "Bombus Latreille, 1802",                                                         NA,              NA,                   NA,                                                                                        NA,       NA,                                                  NA,                                                       NA,                                                                                                            NA,                                                                                                                                 NA,                     NA,                                                                                 NA, "G√©ographie soumise √† floutage √† la demande du producteur|Geographic information generalized during aggregation at the request of the producer",             NA,           TRUE,                FALSE,                                                                                                                                                                                                                                                              NA,                     NA,                                     NA,               NA,                     NA,                                                                                                                                                        NA,        "GBIF_Apidae",
                                        "Bombus pensylvanicus (De Geer, 1773)",       "Apidae",         NA,       "Bombus",              NA,          NA,   "Bombus pensylvanicus",  "pensylvanicus",                    NA,                                      NA,    "SPECIES",                        NA,                       NA,                    NA,                        NA,          NA,                       NA,                      NA,                   NA,                    NA,          33.9315,           -93.592,                      "Arkansas",                 NA,              NA,                                          "3 mi N Hope",      NA,    "Hempstead",                       NA,         "US",      "USA", "United States",  "USA.4_1",           "Arkansas",      "CC_BY_4_0",                                                               "GEODETIC_DATUM_ASSUMED_WGS84;INSTITUTION_MATCH_NONE",   "1954-08-19T00:00:00",          NA,  19L,     8L, 1954L, "PRESERVED_SPECIMEN",    NA,         "PRESENT",            NA,                                            NA,                                           NA,                                                                                                       NA,              NA,               NA,                NA,                   NA,                             NA,              NA,                                                      "BOMBUS32831",  767135245L,                                     NA,       "USDA-ARS",                                                           NA,                                                                                                                             NA,                                                          "767135245",  1340416L,                                     NA,                                              NA,                                                     NA,           "Bombus pensylvanicus",                                                         NA,              NA,                   NA,                                                                                        NA,       NA,                                                  NA,                                                       NA,                                                                                                            NA,                                                                                                                                 NA,                     NA,                                                                                 NA,                                                                                                                                                 NA,             NA,           TRUE,                FALSE,                                                                                                                                                                                                                                                              NA,                     NA,                                     NA,               NA,                     NA,                                                                                                                                                        NA,        "GBIF_Apidae",
                                             "Nomada subcornuta (Kirby, 1802)",       "Apidae",         NA,       "Nomada",              NA,          NA,      "Nomada subcornuta",     "subcornuta",                    NA,                                      NA,    "SPECIES",                        NA,                       NA,                    NA,                        NA,          NA,                       NA,                      NA,    "Paukkunen, Juho",                    NA,         63.05649,          27.68866,                            "Sb",                 NA,              NA,                                                     NA,      NA, "Siilinj√§rvi",                       NA,         "FI",      "FIN",       "Finland",  "FIN.1_1",    "Eastern Finland",      "CC_BY_4_0", "GEODETIC_DATUM_ASSUMED_WGS84;COORDINATE_PRECISION_INVALID;INSTITUTION_MATCH_FUZZY;INSTITUTION_COLLECTION_MISMATCH",   "1954-05-30T00:00:00",          NA,  30L,     5L, 1954L, "PRESERVED_SPECIMEN",    NA,         "PRESENT",            NA,                                            NA,                                           NA,                                                                                                       NA,              NA,               NA,                NA,                   NA,                             NA,              NA,                                                        "JP08-1503",  335348142L,                                     NA,            "MZH",                                                           NA,                                                                                                                             NA,                                                                   NA,  7442089L,                                     NA,                                              NA,                                                     NA,              "Nomada subcornuta",                                                         NA,              NA,                   NA,                                                                                        NA,       NA,                                                  NA,                                                       NA,                                                                                                            NA,                                                                                                                                 NA,                     NA,                                                                                 NA,                                                                                                                                                 NA,             NA,           TRUE,                FALSE,                                                                                                                                                                                                                                                              NA,                     NA,                                     NA,               NA,                     NA,                                                                                                                                                        NA,        "GBIF_Apidae",
                                               "Apis mellifera Linnaeus, 1758",       "Apidae",         NA,         "Apis",              NA,          NA,         "Apis mellifera",      "mellifera",                    NA,                                      NA,    "SPECIES",                        NA,                       NA,                    NA,                        NA,          NA,                       NA,                      NA,         "nina fogel", "2021-09-04T15:41:50",        34.011922,       -118.461783,                    "California",                 NA,              NA,                                                     NA,      NA,             NA,                       NA,         "US",      "USA", "United States",  "USA.5_1",         "California",   "CC_BY_NC_4_0",                                                                                                "COORDINATE_ROUNDED",   "2021-09-03T10:41:34", "17:41:34Z",   3L,     9L, 2021L,  "HUMAN_OBSERVATION",    NA,         "PRESENT",            NA,                                    "yarrow71",                                           NA,                                                                                                       NA,              NA,               NA,                NA,                   NA,                             6L,              NA,                                                         "93593164",         NA,                                     NA,    "iNaturalist",                    "iNaturalist research-grade observations",                                                                                                                             NA,                  "https://www.inaturalist.org/observations/93593164",  1341976L,                                     NA,                                              NA,                                                     NA,                 "Apis mellifera",                  "Fri Sep 03 2021 10:41:34 GMT-0700 (PDT)",              NA,                   NA,                                                                                        NA,       NA,                                                  NA,                                               "yarrow71",                                                                                                            NA,                                                                                                                                 NA,                     NA,                                "https://www.inaturalist.org/observations/93593164",                                                                                                                                                 NA,             NA,           TRUE,                FALSE,                                                                                                                                                                                                                                                              NA,                     NA,                                     NA,               NA,                     NA,                                                                                                                                                        NA,        "GBIF_Apidae",
                                                        "Nomada Scopoli, 1770",       "Apidae",         NA,       "Nomada",              NA,          NA,                       NA,               NA,                    NA,                                      NA,      "GENUS",                        NA,                       NA,                    NA,                        NA,          NA,                       NA,                      NA,                   NA,                    NA,               NA,                NA,                        "Kansas",                 NA, "NORTH_AMERICA",                                          "unspecified",      NA,        "Miami",                       NA,         "US",         NA,              NA,         NA,                   NA,      "CC_BY_4_0",                                                                                            "COLLECTION_MATCH_FUZZY",   "1951-05-27T00:00:00",          NA,  27L,     5L, 1951L, "PRESERVED_SPECIMEN",    NA,         "PRESENT",            NA,                                "Beer, Robert",                                           NA,                                                                                                       NA,              NA,               NA,                NA,                   NA,                             NA,              NA,                                                          "1253091",  894574519L,                                     NA,             "KU",                       "Snow Entomological Museum Collection",                                                                                                                             NA,                               "28de669c-8e2d-46ce-bc25-36c0db48aa1b",  1343230L,                                     NA,                                              NA,                                                     NA,                         "Nomada",                                                         NA,              NA,                   NA,                                                                                        NA,       NA,                                                  NA,                                                       NA, "http://biodiversity.ku.edu/research/university-kansas-biodiversity-institute-data-publication-and-use-norms",                                                                                                                                 NA,                     NA,                                                                                 NA,                                                                                                                                                 NA,             NA,          FALSE,                FALSE,                                                                                                                                                                                                                                                              NA,                     NA,                                     NA,               NA,                     NA,                                                                                                                                                        NA,        "GBIF_Apidae",
                                       "Hoplitis tuberculata (Nylander, 1848)", "Megachilidae",         NA,     "Hoplitis",              NA,          NA,   "Hoplitis tuberculata",    "tuberculata",                    NA,                                      NA,    "SPECIES",                        NA,                       NA,                    NA,                        NA,          NA,                       NA,                      NA,    "Paukkunen, Juho",                    NA,         61.06093,          24.49662,                            "Ta",                 NA,              NA,                                                     NA,      NA, "H√§meenlinna",                       NA,         "FI",      "FIN",       "Finland",  "FIN.5_1",    "Western Finland",      "CC_BY_4_0", "GEODETIC_DATUM_ASSUMED_WGS84;COORDINATE_PRECISION_INVALID;INSTITUTION_MATCH_FUZZY;INSTITUTION_COLLECTION_MISMATCH",   "1968-06-12T00:00:00",          NA,  12L,     6L, 1968L, "PRESERVED_SPECIMEN",    NA,         "PRESENT",            NA,                                            NA,                                           NA,                                                                                                       NA,              NA,               NA,                NA,                   NA,                             NA,              NA,                                                       "JP09-54463",  335339819L,                                     NA,            "MZH",                                                           NA,                                                                                                                             NA,                                                                   NA,  1337165L,                                     NA,                                              NA,                                                     NA,           "Hoplitis tuberculata",                                                         NA,              NA,                   NA,                                                                                        NA,       NA,                                                  NA,                                                       NA,                                                                                                            NA,                                                                                                                                 NA,                     NA,                                                                                 NA,                                                                                                                                                 NA,             NA,           TRUE,                FALSE,                                                                                                                                                                                                                                                              NA,                     NA,                                     NA,               NA,                     NA,                                                                                                                                                        NA,  "GBIF_Megachilidae",
                                                        "Lasioglossum packeri",   "Halictidae",         NA, "Lasioglossum",              NA,          NA,   "Lasioglossum packeri",        "packeri",                    NA,                                      NA,    "SPECIES",                        NA,                       NA,                    NA,                        NA,          NA,                       NA,                      NA,         "Sam Droege",                    NA,          43.7172,         -102.1011,                  "South Dakota",                 NA,              NA,                                                     NA,      NA,   "Pennington", "Badlands National Park",         "US",      "USA", "United States", "USA.42_1",       "South Dakota",        "CC0_1_0",                                                               "GEODETIC_DATUM_ASSUMED_WGS84;INSTITUTION_MATCH_NONE",   "2011-08-19T00:00:00",          NA,  19L,     8L, 2011L, "PRESERVED_SPECIMEN",    NA,         "PRESENT",            NA,                               "Badlands Crew",                               "USGS_DRO9257", "trapType:pan trap; trapCount:29; trapVolume:bowl 3.25oz; trapColor:in field note; trapLiquid:soap dawn",              NA,               NA,                NA,                   NA,                             NA,              NA, "SamplingEventNumber:USGS_DRO9257 | SpecimenNumber:USGS_DRO270402", 1456689405L,                                     NA,          "BISON", "USGS PWRC - Native Bee Inventory and Monitoring Lab (BIML)",                                                                                                                             NA,               "http://www.discoverlife.org/mp/20l?id=USGS_DRO270402", 10599483L,                                     NA,                                              NA, "https://bison.usgs.gov/ipt/resource?r=usgs-pwrc-biml",           "Lasioglossum packeri", "StartDateTime:201108180830xx; EndDateTime:201108190830xx",              NA,                   NA,  "\"Sunny, upper 80's, afternoon thunderstorm, moderate wind, 10 y 10 w 10 b, 1 tipped\"", "FEMALE",                                                  NA,                                                       NA,                                                                                                            NA,                                                                                                                                 NA,                     NA,                                                                                 NA,                                                                                                                                                 NA,             NA,           TRUE,                FALSE,                                                                                                                                                                                                                                                              NA,                     NA,                                     NA,               NA,                     NA,                                                                                                                                                        NA,    "GBIF_Halictidae",
                                       "Lasioglossum hoffmanni (Strand, 1915)",   "Halictidae",         NA, "Lasioglossum",              NA,          NA, "Lasioglossum hoffmanni",      "hoffmanni",                    NA,                                      NA,    "SPECIES",                        NA,                       NA,                    NA,                        NA,          NA,                       NA,                      NA,                   NA,                    NA,               NA,                NA,                              NA,                 NA,              NA,                          "Kyushu, Fukuoka-ken, Kashii",      NA,             NA,                       NA,         "JP",         NA,              NA,         NA,                   NA,   "CC_BY_NC_4_0",                                                                                            "INSTITUTION_MATCH_NONE",   "1960-03-03T00:00:00",          NA,   3L,     3L, 1960L, "PRESERVED_SPECIMEN",    NA,         "PRESENT",            NA,                                            NA,                                           NA,                                                                                                       NA,              NA,               NA,                NA,                   NA,                             NA,              NA,                                                            "21083",  729476432L,                                     NA,           "ELKU",                                                           NA,                                                                                                                             NA,                                         "urn:catalog:ELKU:HYM:21083",  1353546L,                                     NA,                                              NA,                                                     NA,         "Lasioglossum hoffmanni",                                                         NA,              NA,                   NA,                                                                                        NA, "FEMALE",                                                  NA,                                                       NA,                                                                                                            NA,                                                                                                                                 NA,                     NA,                                                                                 NA,                                                                                                                                                 NA,             NA,          FALSE,                FALSE,                                                                                                                                                                                                                                                              NA,                     NA,                                     NA,               NA,                     NA,                                                                                                                                                        NA,    "GBIF_Halictidae",
                                              "Andrena nitida (M√ºller, 1776)",   "Andrenidae",         NA,      "Andrena",              NA,          NA,         "Andrena nitida",         "nitida",                    NA,                                      NA,    "SPECIES",                        NA,                       NA,                    NA,                        NA,          NA,                       NA,                      NA,       "Straka Jakub",                    NA,         47.28543,           8.26479,                            "Ag",                 NA,              NA,                                                     NA,      NA,             NA,                       NA,         "CH",      "CHE",   "Switzerland", "CHE.12_1",            "Lucerne",      "CC_BY_4_0",                                                                                                                  NA,   "2004-04-26T00:00:00",          NA,  26L,     4L, 2004L,  "HUMAN_OBSERVATION",    NA,         "PRESENT",            NA,                            "Wermelinger Beat",                                           NA,                                                                                          "Combined trap",              NA,               NA,                NA,                   NA,                          3535L,              NA,                                                  "CSCF-API-661285",         NA,                             "CSCF-API",            "WSL",                            "Swiss National Apoidea Databank",                                                                                                                             NA,                                                    "CSCF-API-661285",  1357545L,                                     NA,                                              NA,                                                     NA, "Andrena nitida (M√ºller, 1776)",                                                         NA,              NA,                   NA,                                                                                        NA,       NA,                                                  NA, "Info Fauna - Centre Suisse de Cartographie de la Faune",                                                                                                            NA, "Obrist M.K., Wermelinger B., Moretti M., Gossner M.M., Duelli P. 2021. Hymenoptera. EnviDat. http://doi.org/10.16904/envidat.200",                     NA,                                                                                 NA,                                                                                                                                                 NA,             NA,           TRUE,                FALSE,                                                                                                                                                                                                                                                              NA,                     NA,                                     NA,               NA,                     NA,                                                                                                                                                        NA,    "GBIF_Andrenidae",
                                                        "Lasioglossum lusoria",   "Halictidae",         NA, "Lasioglossum",              NA,          NA,                       NA,        "lusoria",                    NA,                                      NA,           NA,          "(Cresson 1872)",                       NA,                    NA,                        NA,          NA,                       NA,                      NA,                   NA,                    NA,               NA,                NA,                          "Utah",              "USA",              NA, "Dugway Proving Grounds; Wig Mt., 7.5 km N (site 35B)",      NA,       "Tooele",                       NA,           NA,         NA,              NA,         NA,                   NA,               NA,                                                                                                                  NA,                      NA,          NA,  20L,     5L, 1998L,  "PreservedSpecimen",    NA,                NA,            NA,                                    "T. Toler",                                           NA,                                                                                                       NA,              NA,               NA,                NA,                   NA,                             NA,              NA,                                                       "BBSL648975",          NA,                                     NA,       "USDA-ARS",                                                           NA,                                                                                                                             NA,                                                          "658910273",        NA, "d02c2ac6-808a-4bc3-b8bb-3983c25cb172",                                              NA,                                                     NA,                               NA,                                                         NA,              NA,                   NA,                                                                                        NA,       NA,                                                  NA,                                                       NA,                                                                                                            NA,                                                                                                                                 NA,                     NA,                                                                                 NA,                                                                                                                                                 NA,             NA,             NA,                   NA,                                                                                                                                                                                                                                                              NA,                     NA,                                     NA,               NA,                     NA,                                                                                                                                                        NA, "iDigBio_halictidae",
                                                             "Bombus fervidus",       "Apidae",         NA,       "Bombus",              NA,          NA,                       NA,       "fervidus",                    NA,                                      NA,           NA,       "(Fabricius, 1798)",                       NA,                    NA,                        NA,          NA,                       NA,                      NA,    "R.I. Velez-Ruiz",                "2012",        44.311357,        -96.798388,                  "South Dakota",    "United States",              NA,                                            "Brookings",      NA,    "Brookings",                       NA,           NA,         NA,              NA,         NA,                   NA,               NA,                                                                                                                  NA,              "9/9/1961",          NA,   9L,     9L, 1961L,  "PreservedSpecimen",    NA,                NA,            NA,                                "H.C. Severin",                                           NA,                                                                                                       NA,              NA,               NA,                NA,                   NA,                          5005L,              NA,                                                             "3572",          NA,                                     NA,           "SDSU",                                                           NA,                                                                                                                             NA,                               "4ccd163a-bea6-46a4-89eb-e1e842532bd9",        NA, "513a265c-8d05-44b9-975f-7e0093599fd2", "urn:uuid:4ccd163a-bea6-46a4-89eb-e1e842532bd9",                 "79a96c8c-5719-4e0e-8a24-d597c48c1d62",                               NA,                                                 "9/9/1961",              NA,                   NA,                                                                                        NA,       NA,                                                  NA,                                                       NA,                                                                                                            NA,                                                                                                                                 NA,                     NA,                                                                                 NA,                                                                                                                                                 NA,             NA,             NA,                   NA,                                                                                                                                                                                                                                                              NA,                     NA,                                     NA,               NA,                     NA,                                                                                                                                                        NA,     "iDigBio_apidae",
                    "Bombus (Thoracobombus) pascuorum floralis (Gmelin, 1790)",       "Apidae",         NA,       "Bombus", "Thoracobombus",          NA,                       NA,      "pascuorum",            "floralis",                                      NA, "subspecies",          "(Gmelin, 1790)",                       NA,  "Hymenoptera|Apidae",                        NA,          NA,                       NA,                      NA,                   NA,                    NA,               NA,                NA, "Mecklenburg-Vorpommern, R?gen",          "Germany",              NA,                                             "Sassnitz",      NA,             NA,                       NA,           NA,         NA,              NA,         NA,                   NA,               NA,                                                                                                                  NA, "1910-08-01/1910-08-31",          NA,   NA,     NA,    NA,  "PreservedSpecimen",    NA,                NA,            NA,                                            NA,                                           NA,                                                                                                       NA,              NA,               1L,                NA,                   NA,                             NA,              NA,                                                   "ZMA.INS.699461",          NA,                                     NA,               NA,                                                           NA,                                                                                                                             NA, "https://data.biodiversitydata.nl/naturalis/specimen/ZMA.INS.699461",        NA, "ec08390b-e77a-461f-9ece-a9e33c63b506",                                              NA,                                                     NA,                               NA,                                                   "Aug-10",              NA,                   NA,                                                                                        NA,   "male",                                                  NA,                                                       NA,                                                                                                            NA,                                                                                                                                 NA,                     NA,                                                                                 NA,                                                                                                                                                 NA,             NA,             NA,                   NA,                                                                                                                                                                                                                                                              NA,                     NA,                                     NA,               NA,                     NA,                                                                                                                                                        NA,     "iDigBio_apidae",
                                                               "Bombus mixtus",       "Apidae",         NA,             NA,              NA,          NA,                       NA,               NA,                    NA,                                      NA,           NA,           "Cresson, 1878",                      "A",                    NA,                        NA,          NA,                       NA,                      NA,  "Jessica J. Rykken",           "20/7/2016",         63.74163,        -149.35381,                        "Alaska",    "United States",              NA,                          "Denali NPP., Primrose Ridge",      NA,             NA,                       NA,           NA,         NA,              NA,         NA,                   NA,               NA,                                                                                                                  NA,             "20/7/2016",          NA,  20L,     7L, 2016L,  "PreservedSpecimen",    NA,                NA,            NA,             "Collector(s): Jessica J. Rykken",                                           NA,                                                                                              "vane trap",              NA,               1L,                NA,                   NA,                            28L,              NA,                                                  "UAM:Ento:396686",          NA,                                     NA,            "UAM",                                                           NA, "field number=2-T-V1_160802, U. S. National Park Service accession=DENA-00705, U. S. National Park Service catalog=DENA 49179",    "http://arctos.database.museum/guid/UAM:Ento:396686?seid=4282896",        NA,                                     NA, "urn:uuid:cc7a8624-31b1-42be-8e19-27e0059e9d2a",                 "8cafaa53-4c65-4155-ae9c-13e51dabe170",                               NA,                                     "7/20/2016 - 8/2/2016",              NA,                   NA,                                                                                        NA, "female", "http://creativecommons.org/publicdomain/zero/1.0/",                            "University of Alaska Museum",                                                                                                            NA,                                                                                                                                 NA,                     NA, "https://scan-bugs.org:443/portal/collections/individual/index.php?occid=39814377",                                                                                                                                                 NA,             NA,             NA,                   NA,                                                                                                                                                                                                                                                              NA,                     NA,                             "39814377",               NA,                     NA,                                                                                                                                                        NA,        "SCAN_Apidae",
                                                            "Andrena crataegi",   "Andrenidae",         NA,      "Andrena",              NA,          NA,                       NA,       "crataegi",                    NA,                                      NA,           NA,         "Robertson, 1893",                       NA,                    NA,                        NA,          NA,                       NA,                      NA,                   NA,                "2005",         48.41222,         -97.41028,                              NA,                 NA,              NA,                                              "Grafton",      NA,             NA,                       NA,           NA,         NA,              NA,         NA,                   NA,               NA,                                                                                                                  NA,              "7/7/1968",          NA,   7L,     7L, 1968L,  "PreservedSpecimen",    NA,                NA,            NA,                                "W.E. LaBerge",                                           NA,                                                                                                       NA,              NA,               1L,                NA,                   NA,                         10000L,              NA,                                     "INHS Insect Collection 99514",          NA,                                     NA,           "INHS",                                                           NA,                                                                                                                             NA,                               "aefb4875-994f-4466-b0a3-405606d996e0",        NA,                                     NA, "urn:uuid:9024adfa-a4dd-4cc3-8c25-3bdd0ca5486e",                 "d93d943b-2390-4e2c-9050-11a9ec9a2a96",                               NA,                                                 "7/7/1968",              NA,                   NA,                                                                                        NA,       NA,    "http://creativecommons.org/licenses/by-nc/4.0/",                                                       NA,                                                                                                            NA,                                                                                                                                 NA,                     NA, "https://scan-bugs.org:443/portal/collections/individual/index.php?occid=60961242",                                                                                                                                                 NA,             NA,             NA,                   NA,                                                                                                                                                                                                                                                              NA,                     NA,                             "60961242",               NA,                     NA,                                                                                                                                                        NA,    "SCAN_Andrenidae",
                                          "Notanthidium adornatum Urban, 1997", "Megachilidae",         NA,             NA,              NA,          NA,                       NA,               NA,                    NA,                                      NA,           NA,                        NA,                       NA,                    NA,                        NA,          NA,                       NA,                      NA,                   NA,                    NA,               NA,                NA,                       "Mendoza",        "ARGENTINA",              NA,                        "El Sosneado to Bardas Blancas",      NA,             NA,                       NA,           NA,         NA,              NA,         NA,                   NA,               NA,                                                                                                                  NA,             "6/12/1983",          NA,   6L,    12L, 1983L,  "PreservedSpecimen",    NA,                NA,            NA,                                  "L. E. Pena",                                           NA,                                                                                                "Netting",              NA,               1L,                NA,                   NA,                             NA,              NA,                                                "AMNH_BEE 00047627",          NA,                                     NA,           "AMNH",                                                           NA,                                                                                                                             NA,                               "83f79136-d8e1-11e2-99a2-0026552be7ea",        NA,                                     NA, "urn:uuid:b1b0bb31-846a-4575-a6fe-5440697d70e9",                 "66acc5ec-da39-42c7-811a-824c25f96161",                               NA,                                                "12/6/1983",              NA,                   NA,                                                                                        NA, "Female",    "http://creativecommons.org/licenses/by-nc/4.0/",                                                   "AMNH",                                                                                     "Not-for-profit use only",                                                                                                                                 NA,                     NA, "https://scan-bugs.org:443/portal/collections/individual/index.php?occid=49444799",                                                                                                                                                 NA,             NA,             NA,                   NA,                                                                                                                                                                                                                                                              NA,                     NA,                             "49444799",               NA,                     NA,                                                                                                                                                        NA,  "SCAN_Megachilidae",
                                                  "Lasioglossum microlepoides",   "Halictidae",         NA, "Lasioglossum",              NA,          NA,                       NA,  "microlepoides",                    NA,                                      NA,           NA,           "(Ellis, 1914)",                       NA,                    NA,                        NA,          NA,                       NA,                      NA, "T.L. Griswold 1999",                    NA,               36,              -115,                        "Nevada",    "United States",              NA,                                             "withheld",      NA,        "Clark",                       NA,           NA,         NA,              NA,         NA,                   NA,               NA,                                                                                                                  NA,                      NA,          NA,   NA,     NA,    NA, "PRESERVED_SPECIMEN",    NA,                NA,            NA, "M. Andres, K. Keen, K. Receveur, C. Schultz",                                           NA,                                                                                                       NA,              NA,               NA,                NA,                   NA,                             NA,              NA,                                                       "BBSL266213",          NA,                                     NA,       "USDA-ARS",                                                           NA,                                                                                                                             NA,                                                                   NA,        NA,                                     NA, "urn:uuid:da7d3cc7-381c-4a38-be46-d3a7deb4e791",                 "b870b19f-b0ea-4d3c-b3f0-53fbaf062e0c",                               NA,                                                         NA,              NA,                   NA,                                                                                        NA,       NA, "http://creativecommons.org/publicdomain/zero/1.0/",                                                       NA,                                                                                                            NA,                                                                                                                                 NA,                     NA, "https://scan-bugs.org:443/portal/collections/individual/index.php?occid=25412520",                                                                                                                                                 NA,             NA,             NA,                   NA,                                                                                                                                                                                                                                                              NA,                     NA,                             "25412520",               NA,                     NA,                                                                                                                                                        NA,    "SCAN_Halictidae"
                    )

  # create an empty eml file
test_eml <- emld::template("creator")

  

  # Run the function
Complete_data <- BeeBDC::formattedCombiner(path = paste0(tempdir()), 
                                           strings = c("USGS_[a-zA-Z_]+[0-9]{4}-[0-9]{2}-[0-9]{2}"), 
                                           # This should be the list-format with eml attached
                                           existingOccurrences = existingTestdb,
                                           existingEMLs = test_eml)


  # Test class expectations
testthat::test_that("formattedCombiner expected class", {
  testthat::expect_type(Complete_data, "list")
})
testthat::test_that("formattedCombiner expected class", {
  testthat::expect_type(Complete_data$Data_WebDL, "list")
})
testthat::test_that("formattedCombiner expected class", {
  testthat::expect_type(Complete_data$eml_files, "list")
})


# Test length of total
testthat::test_that("formattedCombiner item length", {
  testthat::expect_equal(length(Complete_data), 2)
})
  # Test length of rows
testthat::test_that("formattedCombiner item length", {
  testthat::expect_equal(nrow(Complete_data$Data_WebDL), 28)
})
    # Test against the sub-nrows from input datasets
testthat::test_that("formattedCombiner item length", {
  testthat::expect_equal(nrow(Complete_data$Data_WebDL), 
                         sum(nrow(USGS_testData), nrow(existingTestdb)))
})


} # END if require namespace

# This function is a modification of a bdc function and flags columns as FALSE when they are  
  # present in the strings_to_restrict. 
# This function was written on the 22nd of May 2022 by James Dorey. Email James at
  # jbdorey[at]me.com for help.#


#' Flag license protected records
#' 
#' This function will search for strings that indicate a record is restricted in its use and will 
#' flag the restricted records.
#'
#' @param data A data frame or tibble. Occurrence records as input.
#' @param strings_to_restrict A character vector. Should contain the strings used to detect protected records.
#' Default =  c("All Rights Reserved", "All rights reserved", "All rights reserved.", "ND", "Not for public")
#' @param excludeDataSource Optional. A character vector. A vector of the data sources (dataSource) 
#' that will not be flagged as protected, even if they are. This is useful if you have a private 
#' dataset that should be listed  as "All rights reserved" which you want to be ignored by this flag.
#'
#' @return Returns the data with a new column, .unLicensed, where FALSE = records that are protected by 
#' a license.
#' @export
#' 
#' @importFrom dplyr %>%
#'
#' @examples
#'   # Read in the example data
#' data("beesRaw")
#'   # Run the function
#' beesRaw_out <- flagLicense(data = beesRaw,
#'                         strings_to_restrict = "all",
#'                         # DON'T flag if in the following data# source(s)
#'                         excludeDataSource = NULL)

flagLicense <- function(data = NULL,
           strings_to_restrict = "all",
           excludeDataSource = NULL) {
    .data <- .unLicensed <- dataSource <-  NULL
    requireNamespace("dplyr")
    requireNamespace("rlang") 

    #### 1.0 Preperation ####
      ##### 1.1 strings_to_restrict ####
      # Flag if these strigns are present for NO USE!
    if (strings_to_restrict[1] == "all") {
      strings_to_restrict <-
        c("All Rights Reserved", 
          "All rights reserved",
          "All rights reserved.",
          "ND", # noDerivatives to be distributed
          "Not for public") # I fear this has been uploaded in error?
    }
      ##### 1.2 Missing columns ####
        ###### a. dataSource ####
      # If the dataSource column is not in the dataset, fill it in with "NA"s
    if(!any(colnames(data) %in% "dataSource")){
    data <- data %>%
      dplyr::mutate(dataSource = NA_character_)
    message("No dataSource provided. Filling this column with NAs...")
    }
    ###### b. rights ####
    # If the rights column is not in the dataset, fill it in with "NA"s
    if(!any(colnames(data) %in% "rights")){
      data <- data %>%
        dplyr::mutate(rights = NA_character_)
      message("No rights provided. Filling this column with NAs...")
    }
    
    ###### c. license ####
    # If the license column is not in the dataset, fill it in with "NA"s
    if(!any(colnames(data) %in% "license")){
      data <- data %>%
        dplyr::mutate(license = NA_character_)
      message("No license provided. Filling this column with NAs...")
    }
    
    ###### a. accessRights ####
    # If the accessRights column is not in the dataset, fill it in with "NA"s
    if(!any(colnames(data) %in% "accessRights")){
      data <- data %>%
        dplyr::mutate(accessRights = NA_character_)
      message("No accessRights provided. Filling this column with NAs...")
    }
    
    
    
    # Make a new column called .unLicensed to be TRUE when ... the restricted flags are present in 
      # any of the rights, license, or accessRights columns.
    data <-
      data %>%
      dplyr::mutate(
        .unLicensed = 
          dplyr::if_else(grepl(paste(strings_to_restrict, collapse = "|"), data$rights) |
                           grepl(paste(strings_to_restrict, collapse = "|"), data$license) |
                           grepl(paste(strings_to_restrict, collapse = "|"), data$accessRights) == TRUE,
                           dplyr::if_else(dataSource %in% excludeDataSource,
                                            # Flagged, but EXCLUDED - keep these data
                                         TRUE, 
                                            # Flagged AND included - DON'T use these data
                                         FALSE), 
                          # Not flagged - keep these data
                         TRUE))
    
    # Return user output
    message(
      paste(
        "\\.unLicensed:\n",
        "Flagged",
        format(sum(data$.unLicensed == FALSE, na.rm = TRUE), big.mark = ","),
        "records that may NOT be used.\n",
        "One column was added to the database.\n"
      )
    )
      # Return the data to the user
    return(data)
  }

requireNamespace("dplyr")
library(dplyr) # couldn't use %>% without this


# make some test data - this is derived from GBIF data but it is HEAVILY EDITED and USELESS
testData <- dplyr::tribble( # even step coordinates (different for lat/long)
  ~database_id, ~datasetName,    ~id, ~institutionCode, ~collectionCode,         ~ownerInstitutionCode,                    ~basisOfRecord,                  ~occurrenceID,      ~catalogNumber, ~otherCatalogNumbers,   ~kingdom,      ~phylum,    ~class,        ~order,  ~family,  ~scientificName, ~taxonID, ~scientificNameAuthorship, ~genus, ~specificEpithet,        ~recordedBy, ~eventDate, ~year, ~month, ~day,        ~verbatimEventDate,        ~country, ~stateProvince,  ~locality,                  ~locationRemarks, ~decimalLatitude, ~decimalLongitude, ~minimumElevationInMeters,                                             ~rights,                                     ~rightsHolder, ~accessRights,                                       ~recordId,   ~occurrenceStatus,
  "fake SCAN1", "fakeDataset",   13775122L,                      "CAS",        "ANTWEB",        "UCDC, Davis, CA, USA", "PreservedSpecimen",     "CAS:ANTWEB:casent0106100",     "casent0106100",                   NA, "Animalia", "Arthropoda", "Insecta", "Hymenoptera", "apidae", "apis mellifera",  235783L,          "Linnaeus, 1758", "Apis",      "mellifera",        "P.S. Ward",  "6/28/05", 2005L,     6L,  28L, "28 Jun 2005/29 Jun 2005", "United States",   "California",    "Davis", "coordinates obtained from Label",           38.541,         -121.7567,                       15L, "http://creativecommons.org/publicdomain/zero/1.0/", "The California Academy of Sciences - AntWeb.org",            NA, "urn:uuid:46a46727-6535-4e70-88e7-a42c98f806ed",           "PRESENT",
  "fake SCAN2", "fakeDataset",   13775123L,                      "CAS",        "ANTWEB",        "UCDC, Davis, CA, USA", "PreservedSpecimen", "CAS:ANTWEB:casent0106100-d01", "casent0106100-d01",                   NA, "Animalia", "Arthropoda", "Insecta", "Hymenoptera", "apidae", "apis mellifera",  235783L,          "Linnaeus, 1758", "Apis",      "mellifera",        "P.S. Ward",  "6/28/05", 2005L,     6L,  28L, "28 Jun 2005/29 Jun 2005", "United States",   "California",    "Davis", "coordinates obtained from Label",           38.542,         -121.7568,                       15L,                               "All rights reserved", "The California Academy of Sciences - AntWeb.org",            NA, "urn:uuid:d6ff3ddb-4695-4aaf-ab89-c251ab2fc7e6",           "PRESENT",
  "fake SCAN3", "fakeDataset",   13775124L,                      "CAS",        "ANTWEB",        "UCDC, Davis, CA, USA", "PreservedSpecimen", "CAS:ANTWEB:casent0106100-d11", "casent0106100-d11",                   NA, "Animalia", "Arthropoda", "Insecta", "Hymenoptera", "apidae", "apis mellifera",  235783L,          "Linnaeus, 1758", "Apis",      "mellifera",        "P.S. Ward",  "6/28/05", 2005L,     6L,  28L, "28 Jun 2005/29 Jun 2005", "United States",   "California",    "Davis", "coordinates obtained from Label",           38.543,         -121.7569,                       15L,                              "All rights reserved.", "The California Academy of Sciences - AntWeb.org",            NA, "urn:uuid:4504a6e6-ab1c-4de7-ac43-a3588a63ae90",           "PRESENT",
  "fake SCAN4", "fakeDataset",          NA,                      "CAS",        "ANTWEB",        "UCDC, Davis, CA, USA", "PreservedSpecimen",                             NA,                  NA,                   NA, "Animalia", "Arthropoda", "Insecta", "Hymenoptera", "apidae", "apis mellifera",  235783L,          "Linnaeus, 1758", "Apis",      "mellifera",        "P.S. Ward",  "6/28/05", 2005L,     6L,  28L, "28 Jun 2005/29 Jun 2005", "United States",   "California",    "Davis", "coordinates obtained from Label",           38.544,         -121.7570,                       15L,                               "All rights reserved", "The California Academy of Sciences - AntWeb.org",            NA,                                              NA,            "ABSENT",
  "fake SCAN5", "fakeDataset",          NA,                      "CAS",        "ANTWEB",        "UCDC, Davis, CA, USA", "PreservedSpecimen",                             NA,                  NA,                   NA, "Animalia", "Arthropoda", "Insecta", "Hymenoptera", "apidae", "apis mellifera",  235783L,          "Linnaeus, 1758", "Apis",      "mellifera",        "P.S. Ward",  "6/28/05", 2005L,     6L,  28L, "28 Jun 2005/29 Jun 2005", "United States",   "California",    "Davis", "coordinates obtained from Label",           38.545,         -121.7571,                       15L, "http://creativecommons.org/publicdomain/zero/1.0/", "The California Academy of Sciences - AntWeb.org",            NA,                                              NA,            "ABSENT",
  "fake SCAN6", "fakeDataset",          NA,                      "CAS",        "ANTWEB",        "UCDC, Davis, CA, USA", "PreservedSpecimen",                             NA,                  NA,                   NA, "Animalia", "Arthropoda", "Insecta", "Hymenoptera", "apidae", "apis mellifera",  235783L,          "Linnaeus, 1758", "Apis",      "mellifera",        "P.S. Ward",  "6/28/05", 2005L,     6L,  28L, "28 Jun 2005/29 Jun 2005", "United States",   "California",    "Davis", "coordinates obtained from Label",           38.546,         -121.7572,                       15L, "http://creativecommons.org/publicdomain/zero/1.0/", "The California Academy of Sciences - AntWeb.org",            NA,                                              NA,            "ABSENT",
  "fake SCAN7", "fakeDataset",         NA,                      "CAS",        "ANTWEB",        "UCDC, Davis, CA, USA", "PreservedSpecimen",                             NA,                  NA,                   NA, "Animalia", "Arthropoda", "Insecta", "Hymenoptera", "apidae", "apis mellifera",  235783L,          "Linnaeus, 1758", "Apis",      "mellifera",        "P.S. Ward",  "6/28/05", 2005L,     6L,  28L, "28 Jun 2005/29 Jun 2005", "United States",   "California",    "Davis", "coordinates obtained from Label",            38.54,         -121.7567,                       15L, "http://creativecommons.org/publicdomain/zero/1.0/", "The California Academy of Sciences - AntWeb.org",            NA,                                              NA,           "PRESENT",
  "fake SCAN8", "fakeDataset",         NA,                      "CAS",        "ANTWEB",        "UCDC, Davis, CA, USA", "PreservedSpecimen",                             NA,                  NA,                   NA, "Animalia", "Arthropoda", "Insecta", "Hymenoptera", "apidae", "apis mellifera",  235783L,          "Linnaeus, 1758", "Apis",      "mellifera",        "P.S. Ward",  "6/28/05", 2005L,     6L,  28L, "28 Jun 2005/29 Jun 2005", "United States",   "California",    "Davis", "coordinates obtained from Label",            38.54,         -121.7567,                       15L, "http://creativecommons.org/publicdomain/zero/1.0/", "The California Academy of Sciences - AntWeb.org",            NA,                                              NA,            "ABSENT",
  "fake SCAN9", "fakeDataset",  13852774L,                      "CAS",        "ANTWEB", "CAS, San Francisco, CA, USA", "PreservedSpecimen",     "CAS:ANTWEB:casent0249086",     "casent0249086",                   NA, "Animalia", "Arthropoda", "Insecta", "Hymenoptera", "apidae", "apis mellifera",  235783L,          "Linnaeus, 1758", "Apis",      "mellifera", "C. Sch‚àö‚àÇning",   "9/1/09", 2009L,     9L,   1L,                "1-Sep-09",         "Kenya",             NA, "Chogoria",                                NA,         -0.20222,           37.5083,                     2277L,                               "All Rights Reserved", "The California Academy of Sciences - AntWeb.org",            NA, "urn:uuid:964139d3-fb07-4a57-9b7c-e9a5830f4c82",            "ABSENT",
  "fake SCAN10", "fakeDataset",         NA,                      "CAS",        "ANTWEB",        "UCDC, Davis, CA, USA", "PreservedSpecimen",                             NA,                  NA,                   NA, "Animalia", "Arthropoda", "Insecta", "Hymenoptera", "apidae", "apis mellifera",  235783L,          "Linnaeus, 1758", "Apis",     "mellifera",         "P.S. Ward",  "6/29/05", 2005L,     6L,  28L, "28 Jun 2005/29 Jun 2005", "United States",   "California",    "Davis", "coordinates obtained from Label",           38.546,         -121.7572,                       15L,                               "All Rights Reserved", "The California Academy of Sciences - AntWeb.org",            NA,                                              NA,            "ABSENT",
  "fake SCAN11", "newDataset",         NA,                      "CAS",        "ANTWEB",        "UCDC, Davis, CA, USA", "PreservedSpecimen",                             NA,                  NA,                   NA, "Animalia", "Arthropoda", "Insecta", "Hymenoptera", "apidae", "apis mellifera",  235783L,          "Linnaeus, 1758", "Apis",      "mellifera",        "P.S. Ward",  "6/29/05", 2005L,     6L,  28L, "28 Jun 2005/29 Jun 2005", "United States",   "California",    "Davis", "coordinates obtained from Label",            38.54,         -121.7567,                       15L,                               "All Rights Reserved", "The California Academy of Sciences - AntWeb.org",            NA,                                              NA,           "PRESENT",
  "fake SCAN12", "newDataset",         NA,                      "CAS",        "ANTWEB",        "UCDC, Davis, CA, USA", "PreservedSpecimen",                             NA,                  NA,                   NA, "Animalia", "Arthropoda", "Insecta", "Hymenoptera", "apidae", "apis mellifera",  235783L,          "Linnaeus, 1758", "Apis",      "mellifera",        "P.S. Ward",  "6/29/05", 2005L,     6L,  28L, "28 Jun 2005/29 Jun 2005", "United States",   "California",    "Davis", "coordinates obtained from Label",            38.54,         -121.7567,                       15L, "http://creativecommons.org/publicdomain/zero/1.0/", "The California Academy of Sciences - AntWeb.org",            NA,                                              NA,           "PRESENT",
  "fake SCAN13", "newDataset",  13852774L,                      "CAS",        "ANTWEB", "CAS, San Francisco, CA, USA", "PreservedSpecimen",     "CAS:ANTWEB:casent0249086",     "casent0249086",                   NA, "Animalia", "Arthropoda", "Insecta", "Hymenoptera", "apidae", "apis mellifera",  235783L,          "Linnaeus, 1758", "Apis",      "mellifera", "C. Sch‚àö‚àÇning",   "9/1/09", 2009L,     9L,   1L,                "1-Sep-09",         "Kenya",             NA, "Chogoria",                                NA,         -0.20222,           37.5083,                     2277L, "http://creativecommons.org/publicdomain/zero/1.0/", "The California Academy of Sciences - AntWeb.org",            NA, "urn:uuid:964139d3-fb07-4a57-9b7c-e9a5830f4c82",            "ABSENT",

  "fake SCAN14", "fakeDataset",          NA,                      "CAS",        "ANTWEB",        "UCDC, Davis, CA, USA", "PreservedSpecimen",                             NA,                  NA,                   NA, "Animalia", "Arthropoda", "Insecta", "Hymenoptera", "apidae", "apis mellifera",  235783L,          "Linnaeus, 1758", "Apis",      "mellifera",            "Ward",  "4/29/05", 2005L,     6L,  28L, "28 Jun 2005/29 Jun 2005", "United States",   "California",    "Davis", "coordinates obtained from Label",            38.54,          -121.756,                       15L,                              "All rights reserved.", "The California Academy of Sciences - AntWeb.org",            NA,                                              NA,           "PRESENT",
  "fake SCAN15", "fakeDataset",          NA,                      "CAS",        "ANTWEB",        "UCDC, Davis, CA, USA", "PreservedSpecimen",                             NA,                  NA,                   NA, "Animalia", "Arthropoda", "Insecta", "Hymenoptera", "apidae", "apis mellifera",  235783L,          "Linnaeus, 1758", "Apis",      "mellifera",            "Ward",  "4/29/05", 2005L,     6L,  28L, "28 Jun 2005/29 Jun 2005", "United States",   "California",    "Davis", "coordinates obtained from Label",            38.54,          -121.756,                       15L, "http://creativecommons.org/publicdomain/zero/1.0/", "The California Academy of Sciences - AntWeb.org",            NA,                                              NA,           "PRESENT",
  "fake SCAN16", "fakeDataset",          NA,                      "CAS",        "ANTWEB",        "UCDC, Davis, CA, USA", "PreservedSpecimen",                             NA,                  NA,                   NA, "Animalia", "Arthropoda", "Insecta", "Hymenoptera", "apidae", "apis mellifera",  235783L,          "Linnaeus, 1758", "Apis",      "mellifera",            "Ward",  "4/29/05", 2005L,     6L,  28L, "28 Jun 2005/29 Jun 2005", "United States",   "California",    "Davis", "coordinates obtained from Label",            38.54,          -121.756,                       15L,                              "All rights reserved.", "The California Academy of Sciences - AntWeb.org",            NA,                                              NA,           "PRESENT",
  "fake SCAN17", "fakeDataset",          NA,                      "CAS",        "ANTWEB",        "UCDC, Davis, CA, USA", "PreservedSpecimen",                             NA,                  NA,                   NA, "Animalia", "Arthropoda", "Insecta", "Hymenoptera", "apidae", "apis mellifera",  235783L,          "Linnaeus, 1758", "Apis",      "mellifera",            "Ward",  "4/29/05", 2005L,     6L,  28L, "28 Jun 2005/29 Jun 2005", "United States",   "California",    "Davis", "coordinates obtained from Label",            38.54,          -121.756,                       15L, "http://creativecommons.org/publicdomain/zero/1.0/", "The California Academy of Sciences - AntWeb.org",            NA,                                              NA,           "PRESENT",

  "fake SCAN18", "fakeDataset",   13775123L,                      "CAS",        "ANTWEB",        "UCDC, Davis, CA, USA", "PreservedSpecimen", "CAS:ANTWEB:casent0106100-d01", "casent0106100-d01",                   NA, "Animalia", "Arthropoda", "Insecta", "Hymenoptera", "apidae", "apis mellifera",  235783L,          "Linnaeus, 1758", "Apis",      "mellifera",         "BeeNerd",  "6/30/05", 2005L,     6L,  28L, "28 Jun 2005/29 Jun 2005", "United States",   "California",    "Davis", "coordinates obtained from Label",           28.742,           -121.75,                       15L,                                    "Not for public", "The California Academy of Sciences - AntWeb.org",            NA, "urn:uuid:d6ff3ddb-4695-4aaf-ab89-c251ab2fc7e6",            "ABSENT",
  "fake SCAN19", "fakeDataset",   13775124L,                      "CAS",        "ANTWEB",        "UCDC, Davis, CA, USA", "PreservedSpecimen", "CAS:ANTWEB:casent0106100-d11", "casent0106100-d11",                   NA, "Animalia", "Arthropoda", "Insecta", "Hymenoptera", "apidae", "apis mellifera",  235783L,          "Linnaeus, 1758", "Apis",      "mellifera",         "BeeNerd",  "6/30/05", 2005L,     6L,  28L, "28 Jun 2005/29 Jun 2005", "United States",   "California",    "Davis", "coordinates obtained from Label",           28.743,           -121.75,                       15L,                                    "Not for public", "The California Academy of Sciences - AntWeb.org",            NA, "urn:uuid:4504a6e6-ab1c-4de7-ac43-a3588a63ae90",            "ABSENT",
  "fake SCAN20", "fakeDataset",          NA,                      "CAS",        "ANTWEB",        "UCDC, Davis, CA, USA", "PreservedSpecimen",                             NA,                  NA,                   NA, "Animalia", "Arthropoda", "Insecta", "Hymenoptera", "apidae", "apis mellifera",  235783L,          "Linnaeus, 1758", "Apis",      "mellifera",         "BeeNerd",  "6/30/05", 2005L,     6L,  28L, "28 Jun 2005/29 Jun 2005", "United States",   "California",    "Davis", "coordinates obtained from Label",           28.744,           -121.75,                       15L, "http://creativecommons.org/publicdomain/zero/1.0/", "The California Academy of Sciences - AntWeb.org",            NA,                                              NA,            "ABSENT",  
  "fake SCAN21", "fakeDataset",          NA,                      "CAS",        "ANTWEB",        "UCDC, Davis, CA, USA", "PreservedSpecimen",                             NA,                  NA,                   NA, "Animalia", "Arthropoda", "Insecta", "Hymenoptera", "apidae", "apis mellifera",  235783L,          "Linnaeus, 1758", "Apis",      "mellifera",         "BeeNerd",  "6/30/05", 2005L,     6L,  28L, "28 Jun 2005/29 Jun 2005", "United States",   "California",    "Davis", "coordinates obtained from Label",           28.745,           -121.75,                       15L,                                                "ND", "The California Academy of Sciences - AntWeb.org",            NA,                                              NA,            "ABSENT",                     
)



# Run the function
testOut <- BeeBDC::flagLicense(data = testData, 
                              strings_to_restrict = "all", 
                              excludeDataSource = NULL)


# test number of TRUE and FALSE values in the flag column, .occurrenceAbsent
resultsT <- length(testOut$.unLicensed[testOut$.unLicensed == TRUE])
resultsF <- length(testOut$.unLicensed[testOut$.unLicensed == FALSE])

testthat::test_that("flagLicense column .unLicensed results TRUE", {
  testthat::expect_equal(resultsT, 10)
})

testthat::test_that("flagLicense column .unLicensed results FALSE", {
  testthat::expect_equal(resultsF, 11)
})


# test the order of the TRUE and FALSE values in the flag column, .unLicensed
correct <- c(TRUE, FALSE, FALSE, FALSE, TRUE, TRUE, TRUE, TRUE, FALSE, FALSE, FALSE,
             TRUE, TRUE, FALSE, TRUE, FALSE, TRUE, FALSE, FALSE, TRUE, FALSE)

testthat::test_that("flag column .unLicensed results correct series", {
  testthat::expect_equal(correct, testOut$.unLicensed)
})

# This function was written by James B Dorey and Angela Nava Bolańos from the 12 th of October 2022
# Its purpose is to create series of output figures or csv files for specified species
# Please contact jbdorey[at]me.com for help
#' Creates interactive html maps for species
#' 
#' Uses the occurrence data (preferably uncleaned) and outputs interactive .html maps that can be opened
#' in your browser to a specific directory. The maps can highlight if an occurrence has passed all filtering
#' (.summary == TRUE) or failed at least one filter (.summary == FALSE). This can be modified by first running
#' [BeeBDC::summaryFun()] to set the columns that you want to be highlighted. It can also highlight occurrences
#' flagged as expert-identified or country outliers.
#' 
#' @param data A data frame or tibble. Occurrence records to use as input.
#' @param outPath A directory as character. Directory where to save output maps.
#' @param lon Character. The name of the longitude column. Default = "decimalLongitude".
#' @param lat Character. The name of the latitude column. Default = "decimalLatitude".
#' @param speciesColumn Character. The name of the column containing species names (or another factor)
#' to build individual maps from. Default = "scientificName".
#' @param speciesList A character vector. Should contain species names as they appear in the 
#' speciesColumn to make maps of. User can also specify "ALL" in order to make maps of all 
#' species present in the data. Hence, a user may first filter their data and then use "ALL".
#' Default = "ALL".
#' @param countryList A character vector. Country names to map, or NULL for to map ALL countries.
#' @param jitterValue Numeric. The amount, in decimal degrees, to jitter the map points by - this 
#' is important for separating stacked points with the same coordinates.
#' @param onlySummary Logical. If TRUE, the function will not look to plot country or 
#' expert-identified outliers in different colours.
#' @param overWrite Logical. If TRUE, the function will overwrite existing files in the provided
#' directory that have the same name.
#' Default = TRUE.
#' @param customColumn1 Character. Allows the user to report on a column of their choosing in the output.
#' @param customColumn2 Character. Allows the user to report on a column of their choosing in the output.
#' @param TrueAlwaysTop If TRUE, the quality (TRUE) points will always be displayed on top of other points. 
#' If FALSE, then whichever layer was turned on most-recently will be displayed on top.
#' @param excludeSpecies Character. A character vector of species names to exclude, especially
#' if they are so large as to become a problem when making maps. For bee data, for example, 
#' "Apis mellifera Linnaeus, 1758" has too many data points and is default excluded. 
#' @param pointColours A character vector of colours. In order provide colour for TRUE, FALSE, countryOutlier, and customOutlier.
#' Default = c("blue", "darkred","#ff7f00", "black").
#' @param returnPlot Logical. If TRUE, return the plot to the environment. Default = FALSE.
#' 
#'
#' @return Exports .html interactive maps of bee occurrences to the specified directory.
#' @export
#' 
#' @importFrom dplyr %>%
#' @importFrom dplyr across where
#'
#' @examples
#' if(require("htmlwidgets") && require("leaflet")){
#' 
#' OutPath_Figures <- tempdir()
#' 
#' interactiveMapR(
#' # occurrence data - start with entire dataset, filter down to these species
#' data = BeeBDC::bees3sp, # %>%
#'   # Select only those species in the 100 randomly chosen
#'   # dplyr::filter(scientificName %in% beeData_interactive$scientificName),
#'   # Select only one species to map
#'   # dplyr::filter(scientificName %in% "Agapostemon sericeus (Forster, 1771)"),
#' # Directory where to save files
#' outPath = paste0(OutPath_Figures, "/interactiveMaps_TEST"),
#' # lat long columns
#' lon = "decimalLongitude",
#' lat = "decimalLatitude",
#' # Occurrence dataset column with species names
#' speciesColumn = "scientificName",
#' # Which species to map - a character vector of names or "ALL"
#' # Note: "ALL" is defined AFTER filtering for country
#' speciesList = "ALL",
#' # studyArea
#' countryList = NULL, 
#' # Point jitter to see stacked points - jitters an amount in decimal degrees
#' jitterValue = 0.01,
#' # If TRUE, it will only map the .summary column. Otherwise, it will map .summary
#' # which will be over-written by countryOutliers and manualOutliers
#' onlySummary = TRUE,
#' excludeSpecies = c("Apis mellifera Linnaeus, 1758"),
#' overWrite = TRUE,
#'   # Colours for points which are flagged as TRUE, FALSE, countryOutlier, and customOutlier
#' pointColours = c("blue", "darkred","#ff7f00", "black")
#' )
#' 
#' } # END if require

interactiveMapR <- function(
      # occurrence data
    data = NULL,
      # Directory where to save files
    outPath = NULL,
      # lat long columns
    lon = "decimalLongitude",
    lat = "decimalLatitude",
      # Occurrence dataset column with species names
    speciesColumn = "scientificName",
      # Which species to map - a character vector of names or "ALL"
    speciesList = "ALL",
    countryList = NULL,
    jitterValue = NULL,
    onlySummary = TRUE,
    overWrite = TRUE,
    customColumn1 = NULL,
    customColumn2 = NULL,
    TrueAlwaysTop = FALSE,
    excludeSpecies = c("Apis mellifera Linnaeus, 1758"),
    pointColours = c("blue", "darkred","#ff7f00", "black"),
    returnPlot = FALSE
    ){
  # locally bind variables to the function
  country <- .data <- scientificName <- expertOutlier <- .countryOutlier <- .summary <-
    providers <- databaseSpp <- .expertOutlier <- NULL
    
  
  requireNamespace("htmlwidgets")
  requireNamespace("leaflet")
  requireNamespace("dplyr")
  
  # Ensure that working directories are maintain on exit from function
  oldwd <- getwd()           # code line i 
  on.exit(setwd(oldwd))        # code line i+1 
  
  
#### 0.0 Prep ####
  ##### 0.1 Errors ####
  ###### a. FATAL errors ####
  if(is.null(data)){
    stop(paste0(" - No data was given. Please specify the data that you want to map ",
                "for your data-cleaning adventures. I'll do the rest."))
  }
  if(is.null(outPath)){
    stop(paste0(" - No outPath was given. Please specify the directory to save the maps to."))
  }
    #customColumn1
  if(!is.null(customColumn1)){
    if(!is.character(customColumn1)){
      stop(paste0(" - customColumn1 must be character."))
    }
    if(!customColumn1 %in% colnames(data)){
      stop(paste0(" - customColumn1 must be a column in the input data"))
    }
    # Temporarily rename the customColumn1 to "customColumn1" within the function
    data <- data %>%
      dplyr::rename("customColumn1" = tidyselect::any_of(customColumn1))
    
  } # END customColumn1
  #customColumn2
  if(!is.null(customColumn2)){
    if(!is.character(customColumn2)){
      stop(paste0(" - customColumn2 must be character."))
    }
    if(!customColumn2 %in% colnames(data)){
      stop(paste0(" - customColumn2 must be a column in the input data"))
    }
    # Temporarily rename the customColumn2 to "customColumn2" within the function
    data <- data %>%
      dplyr::rename("customColumn2" = tidyselect::any_of(customColumn2))
  } # END customColumn2
  
  
  ##### 0.2 Packages ####
  # Save the original directory 
  olddir <- getwd()
  

  ##### 0.3 Directories ####
  # Create directory if it does not exist
if (!dir.exists(outPath)) {
  dir.create(outPath, recursive = TRUE)}
  # Set directory
setwd(outPath) #directory of work

  # data$IDall <- paste0(1:nrow(data)) #to add an ID by row

  #### 1.0 Data prep ####
    ##### 1.1 Remove na+ ####
data <- data %>%
  tidyr::drop_na(tidyselect::any_of(c(lon, lat)))  %>%
    # Rename the lat and lon to darwincore
  dplyr::rename("decimalLongitude" = tidyselect::all_of(lon),
                "decimalLatitude" = tidyselect::all_of(lat) )

# Stop if no lat/lon
if(nrow(data) == 0){
  stop("It looks like there may be no lat and lon data. Check that it exists and is not NA.")
}

# If there is no .expertOutlier then add one as all NA
if(!".expertOutlier" %in% colnames(data)){
  message("The column .expertOutlier was not found. One will be created with all values = TRUE.")
  data <- data %>% 
    dplyr::mutate(.expertOutlier = TRUE)
}
# If there is no .countryOutlier then add one as all NA
if(!".countryOutlier" %in% colnames(data)){
  message("The column .countryOutlier was not found. One will be created with all values = TRUE.")
  data <- data %>% 
    dplyr::mutate(.countryOutlier = TRUE)
}
# If there is no .summary then add one as all NA
if(!".summary" %in% colnames(data)){
  message("The column .summary was not found. One will be created with all values = TRUE.")
  data <- data %>% 
    dplyr::mutate(.summary = TRUE)
}



##### 1.2 Country list ####
# Select only the countries user provides
if(!is.null(countryList)){
  data <- data %>%
    dplyr::filter(country %in% countryList)
}

    ##### 1.3 Species list ####
if(any(stringr::str_detect(speciesList, "ALL")) == FALSE){
# Prepare the data for the loop
  data <- data %>% 
  # Select ONLY the species requested
  dplyr::filter(.data[[speciesColumn]] %in% speciesList)
}else{
  speciesList <- unique(data[[speciesColumn]])
} # END if else statement


##### 1.4 excludeApis_mellifera ####
if(!is.null(excludeSpecies)){
  data <- data %>%
    dplyr::filter(!scientificName %in% excludeSpecies )
  speciesList <- setdiff(speciesList, excludeSpecies)
}


##### 1.5 Overwrite ####
if(overWrite == FALSE){
    # Find completed species
  existingFiles <- list.files(path = outPath) %>%
    stringr::str_remove("\\.html")
    # remove them from the to-do list
  speciesList <- setdiff(speciesList, existingFiles)
    # STOP if no maps will be produced
  if(length(speciesList) == 0){
    stop("With overWrite = FALSE, there are no new maps to produce.")
  }
    # Re-filter the data to use only wanted species
  data <- data %>% 
    # Select ONLY the species requested
    dplyr::filter(.data[[speciesColumn]] %in% speciesList)
}

    ##### 1.6 Jitter ####
  # If the user specifies a jitter value, add that calue
if(!is.null(jitterValue)){
  data <- data %>%
    dplyr::mutate(
      decimalLongitude = base::jitter(data[[lon]], amount = jitterValue),
      decimalLatitude = base::jitter(data[[lat]], amount = jitterValue)
    )
}else{
    # If no jitter, ensure that the lat lon columns are the same
  data <- data %>%
    dplyr::mutate(
      decimalLongitude = data[[lon]],
      decimalLatitude = data[[lat]])
} # END Jitter


  # Make a new column to colour by if onlySummary == FALSE
if(onlySummary == FALSE){
  data <- data %>% 
    dplyr::mutate(mapLevels = dplyr::if_else(.expertOutlier == FALSE,
                                             "expertOutlier",
                    dplyr::if_else(.countryOutlier == FALSE | is.na(.countryOutlier),
      "countryOutlier", dplyr::if_else(.summary == FALSE,
                                       "FALSE", "TRUE"))) %>%
        factor(c("TRUE", "FALSE", "countryOutlier", "expertOutlier"), 
               levels = c("TRUE", "FALSE", "countryOutlier", "expertOutlier"), 
               ordered = TRUE)
      )
  colPal = leaflet::colorFactor(pointColours, 
                             levels = c("TRUE", "FALSE",
                                        "countryOutlier", "expertOutlier"))
  
}else{ # Make colour palette for == TRUE
    # Only take the first two, if more are provided
  pointColours <- pointColours[1:2]
  colPal = leaflet::colorFactor(pointColours, 
                             levels = c("TRUE", "FALSE"))
}
  


# ensure UTF-8 encoding
old <- options()         # code line i 
on.exit(options(old))      # code line i+1 

# Ensure that there are no NA entries in the species list
speciesList <- speciesList[complete.cases(speciesList)]

options(encoding = "UTF-8")

data <- data %>% dplyr::mutate(dplyr::across(dplyr::where(is.character), 
                              function(x){iconv(x, 
                                                to = "UTF-8",
                                                sub = "")}))

#### 2.0 produce maps ####
#function for leaflet maps
for (x in 1:length(speciesList)){
    # Filter to the xth species
  databaseLoop <- data %>% 
    dplyr::filter(.data[[speciesColumn]] == speciesList[[x]] %>% iconv(x,
                                                                       from = "UTF-8",
                                                                       to = "UTF-8",
                                                                       sub = ""))
    # Split data into classes
  if(onlySummary == FALSE){
    databaseLoop <- split(databaseLoop, f= databaseLoop$mapLevels, drop = TRUE)
  }else{
  databaseLoop <- split(databaseLoop, databaseLoop$.summary)}
                                    #here you can change the number of spp
  
  # Make the base map
  mdatabaseSpp <- leaflet::leaflet(data = databaseLoop ) %>% 
      # Add map panes
    leaflet::addMapPane(name = "maplabels_FALSE", zIndex = 410) %>% 
    leaflet::addMapPane(name = "maplabels_TRUE", zIndex = 420) %>% # higher zIndex rendered on top
    leaflet::addMapPane(name = "maplabels_default", zIndex = 600) %>%
    # Base groups
    leaflet::addTiles(group = "OSM (default)") %>%
    leaflet::addProviderTiles("Stadia.StamenTonerLite", group = "Toner Lite",
                              layerId = 300,
                              options = leaflet::providerTileOptions(zIndex = 500))
    # For the names in the list, apply the points function
      # Apply each walkName in a for loop to add to the map.
for(i in 1:length(names(databaseLoop))){
  walkName <- names(databaseLoop)[[i]]
  databaseSpp <- databaseLoop[[walkName]]
  mdatabaseSpp <- databaseLoop[[walkName]] %>%
    leaflet::addCircleMarkers(map = mdatabaseSpp,
                              data = databaseSpp,
                              lng = ~decimalLongitude, lat = ~decimalLatitude, ###then you can specify what do you want in the popup window from your data
                              group = walkName,
                             if(TrueAlwaysTop == TRUE){
                             options = leaflet::leafletOptions(
                               pane = if(walkName == "TRUE"){"maplabels_TRUE"
                                 }else{"maplabels_FALSE"})}else{
                                   options = leaflet::leafletOptions(
                                     pane = "maplabels_default")
                                 },
                              popup = stringr::str_c(
                                sep = "",
                                ###### a. basic data ####
                                "<b>Basic data </b> - ",
                                "ID: ", databaseSpp$database_id, " ", #databaseSpp is the name of data and ID the name of the column
                                if("family" %in% colnames(databaseSpp)){
                                  paste0("Family: ", databaseSpp$family, 
                                         ";   ")},
                                if("scientificName" %in% colnames(databaseSpp)){
                                  paste0("Species: ", databaseSpp$scientificName, 
                                         ";   ")},
                                            if("institutionCode" %in% colnames(databaseSpp)){
                                                           paste0(" institutionCode: ", databaseSpp$institutionCode, 
                                                          ";   ")},
                                            if("catalogNumber" %in% colnames(databaseSpp)){
                                                           paste0("catalogNumber: ", databaseSpp$catalogNumber, 
                                                                  ";   ")},
                                            if("verbatimScientificName" %in% colnames(databaseSpp)){
                                                           paste0("Original name: ", databaseSpp$verbatimScientificName, 
                                                                  ";   ")},
                                            if("scientificNameAuthorship" %in% colnames(databaseSpp)){
                                                           paste0("Authority: ", databaseSpp$scientificNameAuthorship, 
                                                                  ";   ")},
                                ###### b. summary data ####
                                if(".summary" %in% colnames(databaseSpp)){
                                  paste0("<p></p> <b>Summary flag</b> - ", databaseSpp$.summary)},
                                ###### c. initial data ####
                                "<p></p><b>Initial flags</b> - ",
                                            if(".coordinates_empty" %in% colnames(databaseSpp)){
                                                           paste0("No coordinates: ", databaseSpp$.coordinates_empty, 
                                                                  ";   ")},
                                            if(".coordinates_outOfRange" %in% colnames(databaseSpp)){
                                                           paste0("Point off map: ", databaseSpp$.coordinates_outOfRange, 
                                                                  ";   ")},
                                            if(".basisOfRecords_notStandard" %in% colnames(databaseSpp)){
                                                           paste0("Excluded basis of record: ", databaseSpp$.basisOfRecords_notStandard, 
                                                                  ";   ")},
                                            if(".coordinates_country_inconsistent" %in% colnames(databaseSpp)){
                                                           paste0("Coords. & country inconsistent: ", databaseSpp$.coordinates_country_inconsistent, 
                                                                  ";   ")},
                                            if(".occurrenceAbsent" %in% colnames(databaseSpp)){
                                                           paste0("Absent record: ", databaseSpp$.occurrenceAbsent, 
                                                                  ";   ")},
                                            if(".unLicensed" %in% colnames(databaseSpp)){
                                                           paste0("Protected by license: ", databaseSpp$.unLicensed, 
                                                                  ";   ")},
                                ###### d. taxonomy data ####
                                              # Taxonomy
                                "<p></p><b>Taxonomy flags</b> - ",
                                            if(".scientificName_empty" %in% colnames(databaseSpp)){
                                                           paste0("No scientific name: ", databaseSpp$.scientificName_empty, 
                                                                  ";   ")},
                                            if(".invalidName" %in% colnames(databaseSpp)){
                                                           paste0("Name didn't match: ", databaseSpp$.invalidName, 
                                                                  ";   ")},
                                            if(".uncer_terms" %in% colnames(databaseSpp)){
                                                           paste0("Taxonomy qualifier: ", databaseSpp$.uncer_terms, 
                                                                  ";   ")},
                                ###### e. space data ####
                                              # space
                                "<p></p><b>Space flags</b> - ",
                                            if(".rou" %in% colnames(databaseSpp)){
                                                           paste0("Coordinates rounded: ", databaseSpp$.rou, 
                                                                  ";   ")},
                                            if(".uncertaintyThreshold" %in% colnames(databaseSpp)){
                                              paste0("High coordinate uncertainty: ", databaseSpp$.uncertaintyThreshold, 
                                                     ";   ")},
                                            if(".cap" %in% colnames(databaseSpp)){
                                             paste0("Capital centroid: ", databaseSpp$.cap, 
                                                    ";   ")},
                                            if(".cen" %in% colnames(databaseSpp)){
                                                           paste0("Country centroid: ", databaseSpp$.cen, 
                                                                  ";   ")},
                                            if(".gbf" %in% colnames(databaseSpp)){
                                                           paste0("Point on GBIF HQ: ", databaseSpp$.gbf, 
                                                                  ";   ")},
                                            if(".equ" %in% colnames(databaseSpp)){
                                                           paste0("Coordinates equal: ", databaseSpp$.equ, 
                                                                  ";   ")},
                                            if(".inst" %in% colnames(databaseSpp)){
                                                           paste0("Point on institution: ", databaseSpp$.inst, 
                                                                  ";   ")},
                                            if(".zer" %in% colnames(databaseSpp)){
                                                           paste0("Coordinates zero: ", databaseSpp$.zer, 
                                                                  ";   ")},
                                            if(".val" %in% colnames(databaseSpp)){
                                                           paste0("Coordinates zero: ", databaseSpp$.val, 
                                                                  ";   ")},
                                           if(".sea" %in% colnames(databaseSpp)){
                                             paste0("In sea: ", databaseSpp$.sea, 
                                                    ";   ")},
                                           
                                            if(".countryOutlier" %in% colnames(databaseSpp)){
                                                           paste0("Country outliers: ", databaseSpp$.countryOutlier, 
                                                                  ";   ")},
                                            if(".stateOutlier" %in% colnames(databaseSpp)){
                                              paste0("State outliers: ", databaseSpp$.stateOutlier, 
                                                     ";   ")},
                                           if(".expertOutlier" %in% colnames(databaseSpp)){
                                             paste0("Expert-identified outliers: ", databaseSpp$.expertOutlier, 
                                                    ";   ")},
                                            if(".sequential" %in% colnames(databaseSpp)){
                                                           paste0("Coordinate fill-down: ", databaseSpp$.sequential, 
                                                                  ";   ")},
                                            if(".latFlag" %in% colnames(databaseSpp)){
                                                           paste0("Gridded latitudes: ", databaseSpp$.latFlag, 
                                                                  ";   ")},
                                            if(".lonFlag" %in% colnames(databaseSpp)){
                                                           paste0("Gridded latitudes: ", databaseSpp$.lonFlag, 
                                                                  ";   ")},
                                            if(".gridSummary" %in% colnames(databaseSpp)){
                                                           paste0("Gridded lat & lon: ", databaseSpp$.gridSummary, 
                                                                  ";   ")},
                                ###### f. time data ####
                                            # Time
                                "<p></p><b>Time flags</b> - ",
                                            if(".eventDate_empty" %in% colnames(databaseSpp)){
                                                           paste0("No event date: ", databaseSpp$.eventDate_empty, 
                                                                  ";   ")},
                                            if(".year_outOfRange" %in% colnames(databaseSpp)){
                                                           paste0("Year out of range: ", databaseSpp$.year_outOfRange, 
                                                                  ";   ")},
                                ###### g. duplicate data ####
                                            # Duplicate
                                if(".duplicates" %in% colnames(databaseSpp)){
                                               paste0("<p></p><b>Duplicate flag</b> - ", databaseSpp$.duplicates,
                                                      ";   ")},
                                ###### h. collection data ####
                                # Time
                                "<p></p><b>Collection data</b> - ",
                                if("recordedBy" %in% colnames(databaseSpp)){
                                  paste0("Collector(s): ", databaseSpp$recordedBy, 
                                         ";   ")},
                                if("year" %in% colnames(databaseSpp)){
                                  paste0("Year: ", databaseSpp$year, 
                                         ";   ")},
                                if("identifiedBy" %in% colnames(databaseSpp)){
                                  paste0("Identified by: ", databaseSpp$identifiedBy, 
                                         ";   ")},
                                if("country" %in% colnames(databaseSpp)){
                                  paste0("Country: ", databaseSpp$country, 
                                         ";   ")},
                                if("references" %in% colnames(databaseSpp)){
                                  paste0("References: ", databaseSpp$references, 
                                         "   ")},
                                
                                # Custom column reporting
                                if(!is.null(customColumn1)|!is.null(customColumn2)){
                                  "<p></p><b>Custom column(s)</b> - "
                                },
                              
                                if(!is.null(customColumn1)){
                                  paste0(customColumn1, ": ", databaseSpp$customColumn1, 
                                         "   ")},
                                if(!is.null(customColumn2)){
                                  paste0(customColumn2, ": ", databaseSpp$customColumn2, 
                                         "   ")}
   
                                
                                            ), #you can add what do you want from columns of your data
                    
                  ###### i. colour ####
                              fillOpacity = if(walkName %in% c("TRUE", "FALSE")){0.4}else{0.7},
                              opacity = if(walkName %in% c("TRUE", "FALSE")){0.65}else{1},
                              #opacity =  if(walkName %in% c("TRUE", "FALSE")){0.25}else{1},
                              #stroke =  if(walkName %in% c("TRUE", "FALSE")){TRUE}else{FALSE},

                  # colour determined by if else
                              fillColor = 
                                if(onlySummary == FALSE){
                                  colPal(databaseSpp$mapLevels)
                                }else{colPal(databaseSpp$.summary)},
                              # Stroke color
                              color = if(walkName %in% c("TRUE", "FALSE")){
                                colPal(databaseSpp$.summary)}else{
                                  # colour TRUE border
                                if(all(databaseSpp$.summary) == TRUE){
                                  pointColours[1]
                              }else{pointColours[2]}},
                                # Internal size
                             radius = if(walkName %in% c("TRUE", "FALSE")){
                               5}else{6},
                                # border size
                             weight = if(walkName %in% c("TRUE", "FALSE")){
                               1.5}else{2.5}) #to change the size of points  
    } # END for loop

    
  ###### j. controller ####
    # Add the layers control
    mdatabaseSpp <- mdatabaseSpp %>%
      leaflet::addLegend(color = pointColours[length(names(databaseLoop)):1],
                labels = names(databaseLoop), 
                group = names(databaseLoop)) %>%
    leaflet::addLayersControl(
      baseGroups = c("OSM (default)", "Toner Lite"),
      overlayGroups = names(databaseLoop),
      options = leaflet::layersControlOptions(collapsed = FALSE, autoZIndex = FALSE,
                                     sortLayers = FALSE))

  ###### k. save ####
  #then, it is to save in html format
  htmlwidgets::saveWidget(plotly::as_widget(mdatabaseSpp), 
                          file.path(outPath, #directory to save files
                                    paste0(speciesList[[x]],".html")),
                          selfcontained = TRUE,
                          title = paste0(speciesList[[x]]))
} # END for

# reset to original directory
setwd(olddir)

if(returnPlot == TRUE){
return(mdatabaseSpp)
}

} # END function


#NOTE: to can view spp by spp in the viewer of R, you can run only from line 13 to line 31 changing 
  # numbers of spp in the line 13

requireNamespace("xml2")
requireNamespace("dplyr")

# If rnaturalearthdata is present, run tests
if(requireNamespace("leaflet")){


# Load in a test dataset
bees3sp <- BeeBDC::bees3sp


interactiveMapR(
  # occurrence data — start with entire dataset, filter down to these species
  data = bees3sp,
  # Directory where to save files
  outPath = paste0(tempdir(), "/interactiveMaps"),
  # lat long columns
  lon = "decimalLongitude",
  lat = "decimalLatitude",
  # Occurrence dataset column with species names
  speciesColumn = "scientificName",
  # Which species to map — a character vector of names or "ALL"
  # Note: "ALL" is defined AFTER filtering for country
  speciesList = "ALL",
  countryList = NULL, # studyArea
  # Point jitter to see stacked points — jitters an amount in decimal degrees
  jitterValue = 0.01,
  # If TRUE, it will only map the .summary column. Otherwise, it will map .summary
  # which will be over-written by countryOutliers and manualOutliers
  onlySummary = FALSE,
  TrueAlwaysTop = TRUE,
  # TRUE, FALSE, countryOutlier, customOutlier
  pointColours = c("blue", "darkred","#ff7f00", "black")
)


# Check directory that the plot was saved
testthat::test_that("interactiveMapR plot saved?", {
  testthat::expect_true(any(stringr::str_detect(list.files(tempdir()), "interactiveMaps")))
})
# Check directory that the plot was saved
testthat::test_that("interactiveMapR plot saved?", {
  testthat::expect_equal(length(list.files(paste0(tempdir(), "/interactiveMaps"))), 3)
})

  # Find the .html files that were made
directories <- list.files(paste0(tempdir(), "/interactiveMaps"), 
                         full.names = TRUE)
  # Read in the html files for each species
A_tyleri <- xml2::read_html(directories[stringr::str_detect(directories, "tyleri")])
C_rhodopus <- xml2::read_html(directories[stringr::str_detect(directories, "rhodopus")])
P_octomaculata <- xml2::read_html(directories[stringr::str_detect(directories, "octomaculata")])

# Check directory that the plot was saved
testthat::test_that("interactiveMapR detect Agapostemon tyleri Cockerell, 1917", {
  testthat::expect_true(stringr::str_detect(A_tyleri %>% as.character(), 
                                            "Agapostemon tyleri Cockerell, 1917"))
})
testthat::test_that("interactiveMapR detect Centris rhodopus Cockerell, 1897", {
  testthat::expect_true(stringr::str_detect(C_rhodopus %>% as.character(), 
                                            "Centris rhodopus Cockerell, 1897"))
})
testthat::test_that("interactiveMapR detect Perdita octomaculata (Say, 1824)", {
  testthat::expect_true(stringr::str_detect(P_octomaculata %>% as.character(), 
                                            "Perdita octomaculata \\(Say, 1824\\)"))
})


# Test class
testthat::test_that("interactiveMapR expected class", {
  testthat::expect_equal(attributes(A_tyleri)$class, c("xml_document", "xml_node" ))
})


} # END if require namespace

# This function was written by James B Dorey on the 29th of September 2022
# Its purpose is to visualise duplicate occurrence data by using a compound bargraph
# Please contact jbdorey[at]me.com for help

#' Create a compound bar graph of duplicate data sources
#' 
#' Creates a plot with two bar graphs. One shows the absolute number of duplicate records for each 
#' data source
#' while the other shows the proportion of records that are duplicated within each data source. 
#' This function requires a dataset that has been run through [BeeBDC::dupeSummary()].
#'
#' @param data A data frame or tibble. Occurrence records as input.
#' @param outPath Character. The path to a directory (folder) in which the output should be saved.
#' @param fileName Character. The name of the output file, ending in '.pdf'.
#' @param legend.position.inside The position of the legend as coordinates. Default = c(0.85, 0.8).
#' @param base_height Numeric. The height of the plot in inches. Default = 7.
#' @param base_width Numeric. The width of the plot in inches. Default = 7.
#' @param ... Other arguments to be used to change factor levels of data sources.
#' @param dupeColours A vector of colours for the levels duplicate, kept duplicate, and unique.
#' Default = c("#F2D2A2","#B9D6BC", "#349B90").
#' @param returnPlot Logical. If TRUE, return the plot to the environment. Default = FALSE.
#'
#' @return Outputs a .pdf figure.
#' @export
#' 
#' @importFrom dplyr %>%
#'
#' @examples
#' 
#' # This example will show a warning for the factor levels taht are not present in the specific 
#' # test dataset
#' dupePlotR(
#'   data = beesFlagged,
#'   # The outPath to save the plot as
#'     # Should be something like: #paste0(OutPath_Figures, "/duplicatePlot_TEST.pdf"),
#'   outPath = tempdir(), 
#'   fileName = "duplicatePlot_TEST.pdf",
#'   # Colours in order: duplicate, kept duplicate, unique
#'   dupeColours = c("#F2D2A2","#B9D6BC", "#349B90"),
#'   # Plot size and height
#'   base_height = 7, base_width = 7,
#'   legend.position.inside = c(0.85, 0.8),
#'   # Extra variables can be fed into forcats::fct_recode() to change names on plot
#'   GBIF = "GBIF", SCAN = "SCAN", iDigBio = "iDigBio", USGS = "USGS", ALA = "ALA", 
#'   ASP = "ASP", CAES = "CAES", 'B. Mont.' = "BMont", 'B. Minckley' = "BMin", Ecd = "Ecd",
#'   Gaiarsa = "Gai", EPEL = "EPEL", Lic = "Lic", Bal = "Bal", Arm = "Arm"
#'   )
dupePlotR <- function(
    data = NULL,
    outPath = NULL,
    fileName = NULL,
    legend.position.inside = c(0.85, 0.8),
    base_height = 7,
    base_width = 7,
    # Factor levels to be changed
    ...,
      # Colours in order: duplicate, kept duplicate, unique
    dupeColours = c("#F2D2A2","#B9D6BC", "#349B90"),
    returnPlot = FALSE
){
  # locally bind variables to the function
  database_id <- duplicateStatus <- dataSource <- simpleSource <- NULL 
  
    # Load dependencies
  requireNamespace("ggspatial")
  requireNamespace("forcats")
  requireNamespace("dplyr")
  requireNamespace("cowplot")
  
  #### 0.0 Prep ####
  ##### 0.1 errors ####
  ###### a. FATAL errors ####
  if(is.null(data)){
    stop(" - Please provide an argument for data I'm a program not a magician.")
  }
  if(is.null(outPath)){
    stop(" - Please provide an argument for outPath Seems reckless to let me just guess.")
  }
  if(is.null(fileName)){
    stop(" - Please provide an argument for fileName Seems reckless to let me just guess.")
  }
  if(sum(data$.duplicates) == 0){
    stop(" - No duplicates flagged in the dataset. Stopping process.")
  }
  
  
  #### 1.0 data prep. ####
  # Create the formatted file to create the figure
  # Add duplicates
  dupeTibble <- data %>%
    # Select relevant columns
    dplyr::select(database_id, duplicateStatus, dataSource) %>%
    # Simplify the dataSources
    dplyr::mutate(simpleSource = stringr::str_replace(
      string = dataSource, pattern = "_.*", replacement = "")) 
  
  dupeTibble$duplicateStatus <- dupeTibble$duplicateStatus %>%
    forcats::fct_relevel("Duplicate","Kept duplicate","Unique")
  
  rm(data)
  
  # Recode the simpleSource to be consistent with elsewhere and then order it
  factorised <- dupeTibble$simpleSource %>% 
      # Recode the factors as the user requests
    forcats::fct_recode(...) %>%
      # Turn into a tibble
    dplyr::tibble() %>%
      # Name the single column
    stats::setNames(c("simpleSource")) %>%
      # Group by this column and then count the number of occurrences for each level
    dplyr::group_by(simpleSource) %>%
    dplyr::mutate(count = dplyr::n()) %>%
      # Name these columns
    stats::setNames(c("simpleSource", "count")) 
    # Now re-order the factor by this count and then feed back into dupeTibble
  dupeTibble$simpleSource <-  forcats::fct_reorder(factorised$simpleSource, 
                                                    factorised$count, .desc = TRUE)

    #### 2.0 Total duplicates ####
  dupeBar <- ggplot2::ggplot(dupeTibble, 
                              ggplot2::aes(simpleSource, fill = duplicateStatus)) +
      ggplot2::geom_bar(position = "fill") +
      ggplot2::scale_fill_manual("legend", values = c("Duplicate" = dupeColours[1],
                                                      "Kept duplicate" = dupeColours[2],
                                                      "Unique" = dupeColours[3])) +
      ggplot2::theme(axis.text.x = ggplot2::element_text(angle = 45, hjust = 1),
                     panel.background = ggplot2::element_rect(fill = "white", colour = NA),
                     axis.line = ggplot2::element_line(colour = "black"),
                     legend.position="none") +
      ggplot2::ylab("Proportion of records") + ggplot2::xlab("data source")
  
  #### 3.0 Proportion of duplicates ####
  dupHist <- ggplot2::ggplot(dupeTibble, ggplot2::aes(simpleSource, fill = duplicateStatus)) +
      ggplot2::geom_bar() +
      ggplot2::scale_fill_manual("legend", values = c("Duplicate" = dupeColours[1],
                                             "Kept duplicate" = dupeColours[2],
                                             "Unique" = dupeColours[3])) +
      ggplot2::theme(axis.text.x = ggplot2::element_blank(),
            axis.title.x= ggplot2::element_blank(),
            panel.background = ggplot2::element_rect(fill = "white", colour = NA),
            axis.line = ggplot2::element_line(colour = "black")) +
      ggplot2::ylab("Number of records") + ggplot2::xlab("data source")
  
  #### 4.0 combine + save ####
  # plot the figures together
  (dupPlot <- cowplot::plot_grid(dupHist + 
                                   ggplot2::theme(legend.position.inside = legend.position.inside,
                                         legend.title = ggplot2::element_blank()),
                                 dupeBar, 
                                 labels = c("(a)","(b)"),
                                 ncol = 1, align = 'v', axis = 'l'))
  # Save the plot
  cowplot::save_plot(filename = paste(outPath, fileName, sep = "/"),
                     plot = dupPlot,
                     base_width = base_width,
                     base_height = base_height)
  
  if(returnPlot == TRUE){
    return(dupPlot)}
  
  
} # END function

requireNamespace("dplyr")
requireNamespace("ggplot2")
requireNamespace("BeeBDC")


beesFlagged <- BeeBDC::beesFlagged


# Create a figure showing the total number of duplicates, kept duplicates, and unique
# records for each datasource (simplified to the text before the first underscore) and
# the proportion of the above for each data source
testOut <- BeeBDC::dupePlotR(
  data = beesFlagged,
  # The outPath to save the plot as
  outPath = tempdir(),
  fileName = "duplicatePlot.pdf",
  # Colours in order: duplicate, kept duplicate, unique
  dupeColours = c("#F2D2A2","#B9D6BC", "#349B90"),
  # Plot size and height
  base_height = 7, base_width = 7,
  legend.position = c(0.85, 0.8),
  # Extra variables can be fed into forcats::fct_recode() to change names on plot
  GBIF = "GBIF", SCAN = "SCAN", iDigBio = "iDigBio", USGS = "USGS", ALA = "ALA", 
  ASP = "ASP", 
  returnPlot = TRUE
)



# Test classes
   testthat::test_that("dupePlotR expected class", {
     testthat::expect_type(testOut, "object")
   })

   #  # Check directory that the plot was saved
   testthat::test_that("dupePlotR plot saved?", {
     testthat::expect_true(inherits(testOut, c("ggplot", "ggplot2::ggplot")))
   })

# Check directory that the plot was saved
testthat::test_that("dupePlotR plot saved?", {
  testthat::expect_true(any(stringr::str_detect(list.files(tempdir()), "duplicatePlot.pdf")))
})

  # This function was written by James Dorey on the 16th of June 2022 to join
    # Paige's cleaned dataset as best as possible.

#' Integrate manually-cleaned data from Paige Chesshire
#' 
#' Replaces publicly available data with data that has been manually cleaned and error-corrected for use in
#' the paper Chesshire, P. R., Fischer, E. E., Dowdy, N. J., Griswold, T., Hughes, A. C., Orr, M. J., . . . McCabe, L. M. (In Press). Completeness analysis for over 3000 United States bee species identifies persistent data gaps. Ecography. 
#'
#' @param db_standardized A data frame or tibble. Occurrence records as input.
#' @param PaigeNAm A data frame or tibble. The Paige Chesshire dataset.
#' @param columnStrings A list of character vectors. Each vector is a set of columns that will be 
#' used to iteratively match the public dataset against the Paige dataset.
#'
#' @importFrom dplyr %>%
#' @importFrom stats complete.cases
#'
#' @return Returns db_standardized (input occurrence records) with the Paige Chesshire data integrated.
#' @export
#'
#'
#' @examples
#' \dontrun{
#' library(dplyr)
#' # set the DataPath to tempdir for this example
#' DataPath <- tempdir()
#' # Integrate Paige Chesshire's cleaned dataset.
# Import Paige's cleaned N. American data
# IF you haven't figured out by now, dont worry about the column name warning - not all columns occur here.
#'PaigeNAm <- readr::read_csv(paste(DataPath, "Paige_data", "NorAmer_highQual_only_ALLfamilies.csv",
#'                                  sep = "/"), col_types = ColTypeR()) %>%
#'  # Change the column name from Source to dataSource to match the rest of the data.
#'  dplyr::rename(dataSource = Source) %>%
#'  # add a NEW database_id column
#'  dplyr::mutate(
#'    database_id = paste0("Paige_data_", 1:nrow(.)),
#'    .before = scientificName)
#'
#'  # Set up the list of character vectors to iteratively check for matches with public data.
#'columnList <- list(
#'  c("decimalLatitude", "decimalLongitude", 
#'    "recordNumber", "recordedBy", "individualCount", "samplingProtocol",
#'    "associatedTaxa", "sex", "catalogNumber", "institutionCode", "otherCatalogNumbers",
#'    "recordId", "occurrenceID", "collectionID"), # Iteration 1
#'  c("catalogNumber", "institutionCode", "otherCatalogNumbers",
#'    "recordId", "occurrenceID", "collectionID"), # Iteration 2
#'  c("decimalLatitude", "decimalLongitude", 
#'    "recordedBy", "genus", "specificEpithet"), # Iteration 3
#'  c("id", "decimalLatitude", "decimalLongitude"), # Iteration 4
#'  c("recordedBy", "genus", "specificEpithet", "locality"), # Iteration 5
#'  c("recordedBy", "institutionCode", "genus", 
#'    "specificEpithet","locality"),# Iteration 6
#'  c("occurrenceID","decimalLatitude", "decimalLongitude"), # Iteration 7
#'  c("catalogNumber","decimalLatitude", "decimalLongitude"), # Iteration 8
#'  c("catalogNumber", "locality") # Iteration 9
#') 
#'
#'# Merge Paige's data with downloaded data
#'db_standardized <- BeeBDC::PaigeIntegrater(
#'  db_standardized = db_standardized,
#'  PaigeNAm = PaigeNAm,
#'  columnStrings = columnList)
#' }
#' 
#' 
PaigeIntegrater <- function(
    db_standardized = NULL,
    PaigeNAm = NULL,
    columnStrings = NULL){
  
  # locally bind variables to the function
  occurrenceID<-database_id<-database_id_p<-database_id_d<-finalLatitude<-finalLongitude<-
    Dorey_match<-decimalLatitude<-decimalLongitude<-scientificName<-genus<-specificEpithet<-
    infraspecificEpithet<-country<-coordinateUncertaintyInMeters<-decimalLatitude_m<-
    database_id_m<-decimalLongitude_m<-scientificName_m<-genus_m<-specificEpithet_m<-
    infraspecificEpithet_m<-country_m<-coordinateUncertaintyInMeters_m <- . <- NULL
  
  requireNamespace("dplyr")
  

#### 1.0 occurrenceID ####
# Make a temporary dataset
tempData <- db_standardized %>%
  dplyr::filter(stats::complete.cases(occurrenceID))
# Find the matches for occurrenceID
occMatched <- dplyr::tibble(
  Dorey_match = tempData$database_id[cbind(
    match(PaigeNAm$occurrenceID, tempData$occurrenceID )
  )], # Match by occurrenceID
  Paige_match = PaigeNAm$database_id)
# User output
writeLines(paste0(
  " - INITIAL match with occurrenceID only ", 
  format(sum(stats::complete.cases(occMatched$Dorey_match)), big.mark = ","), " of ",
  format(nrow(occMatched), big.mark = ","), " Paige occurrences.\n",
  "There are ", 
  format(nrow(occMatched) - sum(stats::complete.cases(occMatched$Dorey_match)), big.mark = ","),
  " occurrences remaining to match."))
# Save the number remaining
numMatched <- (nrow(occMatched) - sum(stats::complete.cases(occMatched$Dorey_match)))
  # Set matchedPaige to feed into the loop
matchedPaige <- occMatched

#### 2.0 Loop ####
  # loop through the number of column strings
for(i in 1:length(columnStrings)){
  message(paste0(" - Starting iteration ", i))
  # Get the Paige occurrence records that are not matched above
  matchedPaige <- matchedPaige %>%
    dplyr::filter(stats::complete.cases(matchedPaige$Dorey_match))
  unMatchedPaige <- PaigeNAm %>%
    # Remove the already-matched records
    dplyr::filter(!database_id %in% matchedPaige$Paige_match)
  # Select the columns to match by
  colOverlap <- unlist(columnStrings[i])
  # Get a subset of the db_standardized to feed in below
  temp_db <- db_standardized %>%
    # Get distinct data for theabove columns
    dplyr::distinct(dplyr::across(tidyselect::all_of(colOverlap)), 
                    .keep_all = TRUE) %>%
    dplyr::select(c(database_id, tidyselect::all_of(colOverlap))) %>%
      # Remove already-matched occurrences
    dplyr::filter(!database_id %in% matchedPaige$Dorey_match)
  
  # GET THE MATCHED occurrences
  matchedPaige <- unMatchedPaige %>%
    # Merge datasets
    dplyr::left_join(temp_db, by = c(colOverlap),
                     suffix = c("_p", "_d")) %>%
    # Select the id columns
    dplyr::select(database_id_p, database_id_d) %>%
    # Keep ONLY the matched columns
    dplyr::filter(stats::complete.cases(database_id_p)) %>%
    # Rename those columns
    dplyr::rename(Dorey_match = database_id_d, Paige_match = database_id_p) %>%
    # bind with the last lot of matched names
    dplyr::bind_rows(matchedPaige)
  
  # User output
  writeLines(paste0(
    "Matched ",
    format(sum(stats::complete.cases(matchedPaige$Dorey_match)), big.mark = ","), " of ",
    format(nrow(matchedPaige), big.mark = ","), " Paige occurrences.\n",
    "There are ", 
    format(nrow(matchedPaige) - sum(stats::complete.cases(matchedPaige$Dorey_match)), big.mark = ","),
    " occurrences remaining to match.\n",
    "This step has found ", 
    format(
      numMatched - (nrow(matchedPaige) - sum(stats::complete.cases(matchedPaige$Dorey_match))),
    big.mark = ","),
    " extra occurrences from the last iteration."
    ))
      # Update numMatched for next iteration
  numMatched = (nrow(matchedPaige) - sum(stats::complete.cases(matchedPaige$Dorey_match)))
} # END loop

#### 3.0 Append #### 
  # Update the data from Paige
  writeLines(" - Updating Paige datasheet to merge...")
matchedPaige <- PaigeNAm %>%
    # Select the matched records.
  dplyr::filter(database_id %in% matchedPaige$Paige_match) %>%
    # Replace the lat/lon columns
  dplyr::mutate(
    decimalLatitude = finalLatitude,
    decimalLongitude = finalLongitude) %>%
  dplyr::select(!c(finalLatitude, finalLongitude)) %>%
    # Add on the associated Dorey database_id
  dplyr::left_join(matchedPaige, by = c("database_id" = "Paige_match") ) %>%
    # Make sure that all Dorey_match's are unique
  dplyr::distinct(Dorey_match, .keep_all = TRUE)

writeLines(" - Updating the final datasheet with new information from Paige...")
  # Merge the new information
db_standardized <- db_standardized %>%
        # Join select fields of the Paige data
      dplyr::left_join(matchedPaige %>%
                         dplyr::select(., c(Dorey_match, decimalLatitude, decimalLongitude,
                                            scientificName, genus, specificEpithet,
                                            infraspecificEpithet, database_id, country,
                                            coordinateUncertaintyInMeters)),
      by = c("database_id" = "Dorey_match"), suffix = c("", "_m")) %>%
        # Rename those fields to replace existing fields 
  dplyr::mutate(
    decimalLatitude = dplyr::if_else(stats::complete.cases(decimalLatitude_m), decimalLatitude_m,
                                     decimalLatitude),
    database_id = dplyr::if_else(stats::complete.cases(database_id_m), database_id_m,
                                 database_id),
    decimalLongitude = dplyr::if_else(stats::complete.cases(decimalLongitude_m), decimalLongitude_m,
                                      decimalLongitude),
    scientificName = dplyr::if_else(stats::complete.cases(scientificName_m), scientificName_m,
                                    scientificName),
    genus = dplyr::if_else(stats::complete.cases(genus_m), genus_m,
                           genus),
    specificEpithet = dplyr::if_else(stats::complete.cases(specificEpithet_m), specificEpithet_m,
                                     specificEpithet),
    infraspecificEpithet = dplyr::if_else(stats::complete.cases(infraspecificEpithet_m), 
                                          infraspecificEpithet_m,
                                          infraspecificEpithet),
    country = dplyr::if_else(stats::complete.cases(country_m), country_m,
                             country),
    coordinateUncertaintyInMeters = dplyr::if_else(stats::complete.cases(coordinateUncertaintyInMeters_m),
                                                   coordinateUncertaintyInMeters_m,
                                                   coordinateUncertaintyInMeters)) %>%
    # Remove the additional columns
  dplyr::select(!c(decimalLatitude_m, decimalLongitude_m,
                   scientificName_m, genus_m, specificEpithet_m,
                   infraspecificEpithet_m, database_id_m, country_m,
                   coordinateUncertaintyInMeters_m))
    
  # Return the object
return(db_standardized)
} # END function

requireNamespace("readr")
requireNamespace("BeeBDC")
requireNamespace("dplyr")


## existing data
db_standardized <- dplyr::tribble(
              ~decimalLatitude, ~decimalLongitude, ~recordNumber,                                  ~recordedBy, ~individualCount, ~samplingProtocol, ~associatedTaxa,     ~sex,      ~catalogNumber, ~institutionCode, ~otherCatalogNumbers, ~recordId,                          ~occurrenceID, ~collectionID,            ~genus, ~specificEpithet, ~id,                                                       ~locality, ~database_id,     ~scientificName, ~infraspecificEpithet, ~country,                             ~coordinateUncertaintyInMeters,
                         2.233,              -90L,            NA,                              "James Ogilvie",               1L,                NA,              NA, "FEMALE", "AMNH_IZC 00323864",           "AMNH",                   NA,        NA, "85fb767f-1f99-482f-896c-69dce8b59097",            NA, "Pseudoanthidium",      "truncatum",  NA,                                                              NA,  "db_fake_1",      "Fake species",  "species",         "The incorporated states of nowhere",    100L,
                         2.233,              -90L,            NA,                         "T. D. A. Cockerell",               1L,                NA,              NA, "FEMALE", "AMNH_IZC 00323581",           "AMNH",                   NA,        NA, "63f56ed1-7391-4814-a298-c1b67ca986be",            NA,        "Halictus",        "simplex",  NA,                                                 "near Weing Sa",  "db_fake_2",      "Fake species",  "species",         "The incorporated states of nowhere",    100L,
                         2.233,              -90L,            NA, "T. D. A. Cockerell & Wilmatte P. Cockerell",               1L,                NA,              NA,   "MALE", "AMNH_IZC 00323583",           "AMNH",                   NA,        NA, "cde3718b-8621-4b3f-b203-4436961a4b73",            NA,        "Halictus",        "simplex",  NA, "Lake Waskesiu, at the beginning of the portage to Heart Lakes",  "db_fake_3",      "Fake species",  "species",         "The incorporated states of nowhere",    100L,
                         2.233,              -90L,            NA,                              "C. S. Brimley",               1L,                NA,              NA, "FEMALE", "AMNH_IZC 00324010",           "AMNH",                   NA,        NA, "82e1eab7-5d92-4358-aede-1fab15ba4eda",            NA,        "Halictus",        "simplex",  NA,                                                              NA,  "db_fake_4",      "Fake species",  "species",         "The incorporated states of nowhere",    100L,
                         2.233,              -90L,            NA,                         "T. D. A. Cockerell",               1L,                NA,              NA, "FEMALE", "AMNH_IZC 00323630",           "AMNH",                   NA,        NA, "c161e78e-c091-4483-adf4-6b54045283d7",            NA,        "Halictus",        "simplex",  NA,                     "Prince Albert National Park, Lake Ajawaan",  "db_fake_5",      "Fake species",  "species",         "The incorporated states of nowhere",    100L,
                         2.233,              -90L,            NA,                         "T. D. A. Cockerell",               1L,                NA,              NA, "FEMALE", "AMNH_IZC 00323584",           "AMNH",                   NA,        NA, "002a4861-5374-46ac-84d4-087762ba0c88",            NA,        "Halictus",        "simplex",  NA,                                                              NA,  "db_fake_6",      "Fake species",  "species",         "The incorporated states of nowhere",    100L,
                         2.233,              -90L,            NA,       "T. D. A. Cockerell & Alice M. Mackie",               1L,                NA,              NA, "FEMALE", "AMNH_IZC 00323616",           "AMNH",                   NA,        NA, "5e848030-8e9d-4f64-873d-08fee631e7dd",            NA,        "Halictus",        "simplex",  NA,                                                              NA,  "db_fake_7",      "Fake species",  "species",         "The incorporated states of nowhere",    100L,
                         2.233,              -90L,            NA,                            "Alice M. Mackie",               1L,                NA,              NA,   "MALE", "AMNH_IZC 00323619",           "AMNH",                   NA,        NA, "250ba041-2727-43d0-a39e-8d3470f5a8bf",            NA,        "Halictus",        "simplex",  NA,                                                              NA,  "db_fake_8",      "Fake species",  "species",         "The incorporated states of nowhere",    100L,
                         2.233,              -90L,            NA,                         "T. D. A. Cockerell",               1L,                NA,              NA, "FEMALE", "AMNH_IZC 00323591",           "AMNH",                   NA,        NA, "4cad3ea9-11b3-4589-b1d1-0035a84ea861",            NA,        "Halictus",        "simplex",  NA,                                                              NA,  "db_fake_9",      "Fake species",  "species",         "The incorporated states of nowhere",    100L,
                         2.233,              -90L,            NA,                         "T. D. A. Cockerell",               1L,                NA,              NA,   "MALE", "AMNH_IZC 00323601",           "AMNH",                   NA,        NA, "dc3ebe14-f11f-45e9-b121-d34a8693284e",            NA,        "Halictus",        "simplex",  NA,                                                              NA, "db_fake_10",      "Fake species",  "species",         "The incorporated states of nowhere",    100L,
                         2.233,              -90L,            NA,                                "A. H. Manes",               1L,                NA,              NA, "FEMALE", "AMNH_IZC 00324011",           "AMNH",                   NA,        NA, "6f1d8a81-3a4a-4650-be8c-7102f073e90e",            NA,        "Halictus",        "simplex",  NA,                                                              NA, "db_fake_11",      "Fake species",  "species",         "The incorporated states of nowhere",    100L,
                         2.233,              -90L,            NA,                         "T. D. A. Cockerell",               1L,                NA,              NA, "FEMALE", "AMNH_IZC 00323627",           "AMNH",                   NA,        NA, "406433bc-9c08-4e94-9768-d8df0bdd92cb",            NA,        "Halictus",        "simplex",  NA,                                                              NA, "db_fake_12",      "Fake species",  "species",         "The incorporated states of nowhere",    100L,
                         2.233,              -90L,            NA,                            "Alice M. Mackie",               1L,                NA,              NA, "FEMALE", "AMNH_IZC 00323611",           "AMNH",                   NA,        NA, "bf53740e-e96c-4b5a-a0e9-6b4bea3b2448",            NA,        "Halictus",        "simplex",  NA,                                        "Kaieteur National Park", "db_fake_13",      "Fake species",  "species",         "The incorporated states of nowhere",    100L,
                         2.233,              -90L,            NA,                         "T. D. A. Cockerell",               1L,                NA,              NA, "FEMALE", "AMNH_IZC 00323613",           "AMNH",                   NA,        NA, "2e6d4df0-0a46-4837-b340-ed6f8166a831",            NA,        "Halictus",        "simplex",  NA,                                               "Kum Puang Creek", "db_fake_14",      "Fake species",  "species",         "The incorporated states of nowhere",    100L,
                         2.233,              -90L,            NA,                            "Alice M. Mackie",               1L,                NA,              NA,       NA, "AMNH_IZC 00323617",           "AMNH",                   NA,        NA, "c0642780-37dc-4625-8a30-eacd5dbd1b70",            NA,        "Halictus",        "simplex",  NA,                                                              NA, "db_fake_15",      "Fake species",  "species",         "The incorporated states of nowhere",    100L,
                         2.233,              -90L,            NA,                         "T. D. A. Cockerell",               1L,                NA,              NA, "FEMALE",       "Fake_data_1",           "AMNH",                   NA,        NA, "259d5501-c49c-4a0b-99fc-1595ffe403b1",            NA,        "Halictus",        "simplex",  NA,                                                              NA, "db_fake_16",      "Fake species",  "species",         "The incorporated states of nowhere",    100L,
                         2.233,              -90L,            NA,                                 "Easter Egg",               NA,                NA,              NA,       NA,       "Fake_data_2",      "Bee Place",                   NA,        NA,                        "not_a_number1",            NA,  "Augochloropsis",        "fulgida",  NA,                                                  "non-existent", "db_fake_17",      "Fake species",  "species",         "The incorporated states of nowhere",    100L,
                         2.233,              -90L,            NA,                                 "Easter Egg",               NA,                NA,              NA,       NA,       "Fake_data_3",      "Bee Place",                   NA,        NA,                        "not_a_number2",            NA,  "Augochloropsis",        "fulgida",  NA,                                                  "non-existent", "db_fake_18",      "Fake species",  "species",         "The incorporated states of nowhere",    100L,
                         2.233,              -90L,            NA,                                 "Easter Egg",               NA,                NA,              NA,       NA,       "Fake_data_4",      "Bee Place",                   NA,        NA,                        "not_a_number3",            NA,  "Augochloropsis",        "fulgida",  NA,                                                  "non-existent", "db_fake_19",      "Fake species",  "species",         "The incorporated states of nowhere",    100L,
                         2.233,              -90L,            NA,                                 "Easter Egg",               NA,                NA,              NA,       NA,       "Fake_data_5",      "Bee Place",                   NA,        NA,                        "not_a_number4",            NA,  "Augochloropsis",        "fulgida",  NA,                                                  "non-existent", "db_fake_20",      "Fake species",  "species",         "The incorporated states of nowhere",    100L,
                         2.233,              -90L,            NA,                                 "Easter Egg",               NA,                NA,              NA,       NA,       "Fake_data_6",      "Bee Place",                   NA,        NA,                        "not_a_number5",            NA,  "Augochloropsis",        "fulgida",  NA,                                                  "non-existent", "db_fake_21",      "Fake species",  "species",         "The incorporated states of nowhere",    100L,
                         2.233,              -90L,            NA,                                 "Easter Egg",               NA,                NA,              NA,       NA,       "Fake_data_7",      "Bee Place",                   NA,        NA,                        "not_a_number6",            NA,  "Augochloropsis",        "fulgida",  NA,                                                  "non-existent", "db_fake_22",      "Fake species",  "species",         "The incorporated states of nowhere",    100L,
                         2.233,              -90L,            NA,                                 "Easter Egg",               NA,                NA,              NA,       NA,       "Fake_data_8",      "Bee Place",                   NA,        NA,                        "not_a_number7",            NA,  "Augochloropsis",        "fulgida",  NA,                                                  "non-existent", "db_fake_23",      "Fake species",  "species",         "The incorporated states of nowhere",    100L,
                         2.233,              -90L,            NA,                                 "Easter Egg",               NA,                NA,              NA,       NA,       "Fake_data_9",      "Bee Place",                   NA,        NA,                        "not_a_number8",            NA,  "Augochloropsis",        "fulgida",  NA,                                                  "non-existent", "db_fake_24",      "Fake species",  "species",         "The incorporated states of nowhere",    100L,
                         2.233,              -90L,            NA,                                 "Easter Egg",               NA,                NA,              NA,       NA,      "Fake_data_10",      "Bee Place",                   NA,        NA,                        "not_a_number9",            NA,  "Augochloropsis",        "fulgida",  NA,                                                  "non-existent", "db_fake_25",      "Fake species",  "species",         "The incorporated states of nowhere",    100L,
                         2.233,              -90L,            NA,                                       "Nope",               NA,                NA,              NA,       NA,      "Fake_data_11",      "Bee Place",                   NA,        NA,                       "not_a_number10",            NA,  "Augochloropsis",        "fulgida",  NA,                                                  "non-existent", "db_fake_26",      "Fake species",  "species",         "The incorporated states of nowhere",    100L,
                         2.233,              -90L,            NA,                                       "Nope",               NA,                NA,              NA,       NA,      "Fake_data_12",      "Bee Place",                   NA,        NA,                       "not_a_number11",            NA,  "Augochloropsis",        "fulgida",  NA,                                                  "non-existent", "db_fake_27",      "Fake species",  "species",         "The incorporated states of nowhere",    100L,
                         2.233,              -90L,            NA,                                       "Nope",               NA,                NA,              NA,       NA,      "Fake_data_13",      "Bee Place",                   NA,        NA,                       "not_a_number12",            NA,    "Lasioglossum",  "semicaeruleum",  NA,                                                  "non-existent", "db_fake_28",      "Fake species",  "species",         "The incorporated states of nowhere",    100L,
                         2.233,              -90L,            NA,                                       "Nope",               NA,                NA,              NA,       NA,      "Fake_data_14",      "Bee Place",                   NA,        NA,                       "not_a_number13",            NA,    "Lasioglossum",  "semicaeruleum",  NA,                                                  "non-existent", "db_fake_29",      "Fake species",  "species",         "The incorporated states of nowhere",    100L,
                         2.233,              -90L,            NA,                                       "Nope",               NA,                NA,              NA,       NA,      "Fake_data_15",      "Bee Place",                   NA,        NA,                       "not_a_number14",            NA,    "Lasioglossum",  "semicaeruleum",  NA,                                                  "non-existent", "db_fake_30",      "Fake species",  "species",         "The incorporated states of nowhere",    100L
              )




## Paige corrections - this is FAKED DATA AND SHOULD NOT BE USED FOR ANYTHING EVER OTHER THAN THIS TEST
PaigeNAm <- dplyr::tribble(
              ~X.1,     ~X,            ~scientificName,         ~id, ~institutionCode, ~collectionCode, ~ownerInstitutionCode, ~collectionID,       ~basisOfRecord,                                                     ~occurrenceID, ~catalogNumber, ~otherCatalogNumbers,   ~kingdom,      ~phylum,    ~class,        ~order,      ~family, ~taxonID, ~scientificNameAuthorship,        ~genus, ~specificEpithet, ~taxonRank, ~infraspecificEpithet,      ~identifiedBy,        ~dateIdentified, ~identificationReferences, ~identificationRemarks, ~taxonRemarks, ~identificationQualifier, ~typeStatus,                              ~recordedBy, ~recordNumber,             ~eventDate, ~year,    ~month, ~day, ~startDayOfYear, ~endDayOfYear, ~verbatimEventDate, ~occurrenceRemarks, ~habitat, ~fieldNumber, ~informationWithheld, ~dataGeneralizations, ~dynamicProperties, ~associatedTaxa, ~associatedOccurrences, ~reproductiveCondition, ~establishmentMeans, ~lifeStage,     ~sex, ~individualCount, ~samplingProtocol, ~samplingEffort, ~preparations, ~country,     ~stateProvince,              ~county, ~municipality,                         ~locality, ~locationRemarks, ~decimalLatitude, ~decimalLongitude, ~geodeticDatum, ~coordinateUncertaintyInMeters, ~verbatimCoordinates, ~georeferencedBy, ~georeferenceProtocol, ~georeferenceSources, ~georeferenceVerificationStatus, ~georeferenceRemarks, ~minimumElevationInMeters, ~maximumElevationInMeters, ~minimumDepthInMeters, ~maximumDepthInMeters, ~verbatimDepth, ~verbatimElevation, ~disposition, ~language, ~recordEnteredBy,              ~modified, ~rights, ~rightsHolder, ~accessRights, ~recordId,                                                          ~references, ~finalLatitude, ~finalLongitude, ~Identifier_Level, ~PrecisionLevel, ~Source, ~geolocate_LocalityID, ~geolocate_ResultID, ~geolocate_Latitude, ~geolocate_Longitude, ~geolocate_UncertaintyRadiusMeters, ~geolocate_UncertaintyPolygon, ~geolocate_Score, ~geolocate_Precision, ~geolocate_ParsePattern, ~geolocate_locFieldUsed, ~geolocate_NumResults, ~wordCount,              ~finalName,    ~CountryName,
                1L, 20102L,     "Agapostemon borealis", 1317390166L,           "USNM",    "Entomology",                    NA,            NA, "PRESERVED_SPECIMEN", "http://n2t.net/ark:/65665/3084be498-793e-4b29-bb1d-0ad85253ef61",             NA,                   NA, "Animalia", "Arthropoda", "Insecta", "Hymenoptera", "Halictidae",       NA,                        NA, "Agapostemon",       "borealis",  "SPECIES",                    NA,                 NA,                     NA,                        NA,                     NA,            NA,                       NA,          NA,                             "Livingston",            NA, "1896-04-04T00:00:00Z", 1896L,       "4",   4L,             95L,            NA,         "4/4/1996",                 NA,       NA,           NA,                   NA,                   NA,                 NA,              NA,                     NA,                     NA,                  NA,         NA, "FEMALE",               1L,                NA,              NA,      "Pinned",     "CA", "British Columbia",                   NA,            NA,                       "Vancouver",               NA,            2.233,          -90,             NA,                         10206L,                   NA,               NA,           "GEOLocate",                   NA,                              NA,                   NA,                        NA,                        NA,                    NA,                    NA,             NA,                 NA,           NA,        NA,               NA,                     NA,      NA,            NA,            NA,        NA,                                                                   NA,          49.25,        -123.133,        "Specimen",          "High",  "GBIF",                    NA,                  NA,                  NA,                   NA,                                 NA,                            NA,               NA,                   NA,                      NA,                      NA,                    NA,         2L,   "Agapostemon texanus",        "Canada",
                2L, 20103L, "Agapostemon californicus", 1101319745L,            "UCM",          "UCMC",                    NA,            NA, "PRESERVED_SPECIMEN",                            "7f6a4b80-642d-40f2-8dc7-eb45235f157c",             NA,                   NA, "Animalia", "Arthropoda", "Insecta", "Hymenoptera", "Halictidae",       NA,                        NA, "Agapostemon",   "californicus",  "SPECIES",                    NA, "T.D.A. Cockerell", "1939-01-01T00:00:00Z",                        NA,                     NA,            NA,                       NA,          NA,                       "T.D.A. Cockerell",            NA,                     NA,    NA, "Jun. 18",   NA,              NA,            NA,          "Jun. 18",                 NA,       NA,           NA,                   NA,                   NA,                 NA,              NA,                     NA,                     NA,                  NA,    "Adult",       NA,               1L,                NA,              NA,      "pinned",     "US",       "California", "Los Angeles County",            NA,             "San Clemente Island",               NA,             2.233,         -90,             NA,                         21236L,                   NA,               NA,                    NA,                   NA,                              NA,                   NA,                        NA,                        NA,                    NA,                    NA,             NA,                 NA,           NA,        NA,               NA, "2017-06-29T00:00:00Z",      NA,            NA,            NA,        NA,                                                                   NA,           32.9,       -118.5009,        "Specimen",          "High",  "GBIF",                    NA,                  NA,                  NA,                   NA,                                 NA,                            NA,               NA,                   NA,                      NA,                      NA,                    NA,         2L,   "Agapostemon texanus", "United States",
                5L, 23836L,    "Agapostemon leunculus",  658242510L,             "KU",          "SEMC",                    NA,            NA, "PRESERVED_SPECIMEN",                            "3aa3dd4b-1ed6-11e3-bfac-90b11c41863e",       "495676",                   NA, "Animalia", "Arthropoda", "Insecta", "Hymenoptera", "Halictidae",       NA,                        NA, "Agapostemon",      "leunculus",  "SPECIES",                    NA,       "Roberts, R",                     NA,                        NA,                     NA,            NA,                       NA,          NA,             "Univ. of Kans. Mex. Exped.",            NA, "1961-06-11T00:00:00Z", 1961L,       "6",  11L,              NA,            NA,                 NA,                 NA,       NA,           NA,                   NA,                   NA,                 NA,              NA,                     NA,                     NA,                  NA,    "Adult",       NA,               NA,                NA,              NA,  "Pinned - 1",     "MX",         "Veracruz",                   NA,            NA,       "Alazan, 7 mi NW nr Tuxpan",               NA,          20.9915,         -97.54551,             NA,                             NA,                   NA,               NA,                    NA,                   NA,                              NA,                   NA,                        NA,                        NA,                    NA,                    NA,             NA,                 NA,           NA,        NA,               NA,                     NA,      NA,            NA,            NA,        NA,                                                                   NA,        20.9915,       -97.54551,        "Specimen",          "High",  "GBIF",                    NA,                  NA,                  NA,                   NA,                                 NA,                            NA,               NA,                   NA,                      NA,                      NA,                    NA,         2L, "Agapostemon leunculus",        "Mexico",
                6L, 23837L,    "Agapostemon leunculus", 1060982746L,           "EMEC",          "EMEC",                    NA,            NA, "PRESERVED_SPECIMEN",                                    "urn:catalog:EMEC:EMEC:511822",   "EMEC511822",                   NA, "Animalia", "Arthropoda", "Insecta", "Hymenoptera", "Halictidae",       NA,                        NA, "Agapostemon",      "leunculus",  "SPECIES",                    NA,                 NA,                     NA,                        NA,                     NA,            NA,                       NA,          NA, "A. E. Michelbacher, M. M. Michelbacher",            NA, "1962-06-29T00:00:00Z", 1962L,       "6",  29L,              NA,            NA,                 NA,                 NA,       NA,           NA,                   NA,                   NA,                 NA,              NA,                     NA,                     NA,                  NA,    "Adult",       NA,               1L,                NA,              NA,         "pin",     "MX",          "Sinaloa",                   NA,            NA,                         "Rosario",               NA,         22.99325,         -105.8586,             NA,                          1405L,                   NA,               NA,                    NA,                   NA,                              NA,                   NA,                        NA,                        NA,                    NA,                    NA,             NA,                 NA,           NA,        NA,               NA, "2017-06-24T00:00:00Z",      NA,            NA,            NA,        NA, "http://essigdb.berkeley.edu/cgi/eme_query?bnhm_id=EMEC511822&one=T",       22.99325,       -105.8586,        "Specimen",          "High",  "GBIF",                    NA,                  NA,                  NA,                   NA,                                 NA,                            NA,               NA,                   NA,                      NA,                      NA,                    NA,         2L, "Agapostemon leunculus",        "Mexico",
                7L, 23838L,    "Agapostemon leunculus",  219014691L,         "IBUNAM",          "EBCC",                    NA,            NA,            "UNKNOWN",                                                   "not_a_number1",       "300182",                   NA, "Animalia", "Arthropoda", "Insecta", "Hymenoptera", "Halictidae",       NA,                        NA, "Agapostemon",      "leunculus",  "SPECIES",                    NA,   "Douglas Yanega",                     NA,                        NA,                     NA,            NA,                       NA,          NA,            "Alicia Rodr√É¬≠guez Palafox",            NA, "1993-04-23T00:00:00Z", 1993L,       "4",  23L,              NA,            NA,                 NA,                 NA,       NA,           NA,                   NA,                   NA,                 NA,              NA,                     NA,                     NA,                  NA,         NA,       NA,               NA,                NA,              NA,            NA,     "MX",          "Chiapas",                   NA,            NA, "Ocosingo, Ej. Shanil, 52.9 km N",               NA,         17.21667,             -92.1,             NA,                             NA,                   NA,               NA,                    NA,                   NA,                              NA,                   NA,                        NA,                        NA,                    NA,                    NA,             NA,                 NA,           NA,        NA,               NA,                     NA,      NA,            NA,            NA,        NA,                                                                   NA,       17.21667,           -92.1,        "Specimen",          "High",  "GBIF",                    NA,                  NA,                  NA,                   NA,                                 NA,                            NA,               NA,                   NA,                      NA,                      NA,                    NA,         2L, "Agapostemon leunculus",        "Mexico",
                8L, 23839L,    "Agapostemon leunculus", 1060924636L,           "EMEC",          "EMEC",                    NA,            NA, "PRESERVED_SPECIMEN",                                    "urn:catalog:EMEC:EMEC:511782",   "EMEC511782",                   NA, "Animalia", "Arthropoda", "Insecta", "Hymenoptera", "Halictidae",       NA,                        NA, "Agapostemon",      "leunculus",  "SPECIES",                    NA,                 NA,                     NA,                        NA,                     NA,            NA,                       NA,          NA, "A. E. Michelbacher, M. M. Michelbacher",            NA, "1962-07-27T00:00:00Z", 1962L,       "7",  27L,              NA,            NA,                 NA,                 NA,       NA,           NA,                   NA,                   NA,                 NA,              NA,                     NA,                     NA,                  NA,    "Adult",       NA,               1L,                NA,              NA,         "pin",     "MX",           "Oaxaca",                   NA,            NA,                          "Oaxaca",               NA,         17.06371,         -96.73192,             NA,                          8339L,                   NA,               NA,                    NA,                   NA,                              NA,                   NA,                        NA,                        NA,                    NA,                    NA,             NA,                 NA,           NA,        NA,               NA, "2017-05-04T00:00:00Z",      NA,            NA,            NA,        NA, "http://essigdb.berkeley.edu/cgi/eme_query?bnhm_id=EMEC511782&one=T",       17.06371,       -96.73192,        "Specimen",          "High",  "GBIF",                    NA,                  NA,                  NA,                   NA,                                 NA,                            NA,               NA,                   NA,                      NA,                      NA,                    NA,         2L, "Agapostemon leunculus",        "Mexico",
                9L, 23840L,    "Agapostemon leunculus", 1060924589L,           "EMEC",          "EMEC",                    NA,            NA, "PRESERVED_SPECIMEN",                                    "urn:catalog:EMEC:EMEC:511766",   "EMEC511766",                   NA, "Animalia", "Arthropoda", "Insecta", "Hymenoptera", "Halictidae",       NA,                        NA, "Agapostemon",      "leunculus",  "SPECIES",                    NA,                 NA,                     NA,                        NA,                     NA,            NA,                       NA,          NA, "A. E. Michelbacher, M. M. Michelbacher",            NA, "1962-07-27T00:00:00Z", 1962L,       "7",  27L,              NA,            NA,                 NA,                 NA,       NA,           NA,                   NA,                   NA,                 NA,              NA,                     NA,                     NA,                  NA,    "Adult",       NA,               1L,                NA,              NA,         "pin",     "MX",           "Oaxaca",                   NA,            NA,                          "Oaxaca",               NA,         2.233,         -96.73192,             NA,                          8339L,                   NA,               NA,                    NA,                   NA,                              NA,                   NA,                        NA,                        NA,                    NA,                    NA,             NA,                 NA,           NA,        NA,               NA, "2017-05-04T00:00:00Z",      NA,            NA,            NA,        NA, "http://essigdb.berkeley.edu/cgi/eme_query?bnhm_id=EMEC511766&one=T",       17.06371,       -96.73192,        "Specimen",          "High",  "GBIF",                    NA,                  NA,                  NA,                   NA,                                 NA,                            NA,               NA,                   NA,                      NA,                      NA,                    NA,         2L, "Agapostemon leunculus",        "Mexico",
               10L, 23841L,    "Agapostemon leunculus",  219014682L,         "IBUNAM",          "EBCC",                    NA,            NA,            "UNKNOWN",                                                 "not_a_number11",       "300173",                   NA, "Animalia", "Arthropoda", "Insecta", "Hymenoptera", "Halictidae",       NA,                        NA, "Agapostemon",      "leunculus",  "SPECIES",                    NA,   "Douglas Yanega",                     NA,                        NA,                     NA,            NA,                       NA,          NA,                             "J. L. Neff",            NA, "1993-04-12T00:00:00Z", 1993L,       "4",  12L,              NA,            NA,                 NA,                 NA,       NA,           NA,                   NA,                   NA,                 NA,              NA,                     NA,                     NA,                  NA,         NA,       NA,               NA,                NA,              NA,            NA,     "MX",          "Chiapas",                   NA,            NA,       "Tuxtla Gutierrez, 37 km E",               NA,            2.233,         90,             NA,                             NA,                   NA,               NA,                    NA,                   NA,                              NA,                   NA,                        NA,                        NA,                    NA,                    NA,             NA,                 NA,           NA,        NA,               NA,                     NA,      NA,            NA,            NA,        NA,                                                                   NA,          16.75,       -92.91667,        "Specimen",          "High",  "GBIF",                    NA,                  NA,                  NA,                   NA,                                 NA,                            NA,               NA,                   NA,                      NA,                      NA,                    NA,         2L, "Agapostemon leunculus",        "Mexico",
               11L, 23842L,    "Agapostemon leunculus",  219014681L,         "IBUNAM",          "EBCC",                    NA,            NA,            "UNKNOWN",                                                                NA,       "300172",                   NA, "Animalia", "Arthropoda", "Insecta", "Hymenoptera", "Halictidae",       NA,                        NA, "Agapostemon",      "leunculus",  "SPECIES",                    NA,   "Douglas Yanega",                     NA,                        NA,                     NA,            NA,                       NA,          NA,                     "Wallace E. LaBerge",            NA, "1993-04-10T00:00:00Z", 1993L,       "4",  10L,              NA,            NA,                 NA,                 NA,       NA,           NA,                   NA,                   NA,                 NA,              NA,                     NA,                     NA,                  NA,         NA,       NA,               NA,                NA,              NA,            NA,     "MX",          "Chiapas",                   NA,            NA,           "Ocozocoautla, 17 km N",               NA,            16.89,            -93.46,             NA,                             NA,                   NA,               NA,                    NA,                   NA,                              NA,                   NA,                        NA,                        NA,                    NA,                    NA,             NA,                 NA,           NA,        NA,               NA,                     NA,      NA,            NA,            NA,        NA,                                                                   NA,          16.89,          -93.46,        "Specimen",          "High",  "GBIF",                    NA,                  NA,                  NA,                   NA,                                 NA,                            NA,               NA,                   NA,                      NA,                      NA,                    NA,         2L, "Agapostemon leunculus",        "Mexico",
               12L, 23843L,    "Agapostemon leunculus",  657759380L,             "KU",          "SEMC",                    NA,            NA, "PRESERVED_SPECIMEN",                            "1fc6766a-1ed6-11e3-bfac-90b11c41863e",       "496619",                   NA, "Animalia", "Arthropoda", "Insecta", "Hymenoptera", "Halictidae",       NA,                        NA, "Agapostemon",      "leunculus",  "SPECIES",                    NA,       "Roberts, R",                     NA,                        NA,                     NA,            NA,                       NA,          NA,             "Univ. of Kans. Mex. Exped.",            NA, "1954-08-22T00:00:00Z", 1954L,       "8",  22L,              NA,            NA,                 NA,                 NA,       NA,           NA,                   NA,                   NA,                 NA,              NA,                     NA,                     NA,                  NA,    "Adult",       NA,               NA,                NA,              NA,  "Pinned - 1",     "MX",  "San Luis Potosi",                   NA,            NA,             "Ciudad Maiz, 5 mi E",               NA,         22.44681,         -99.56705,             NA,                             NA,                   NA,               NA,                    NA,                   NA,                              NA,                   NA,                        NA,                        NA,                    NA,                    NA,             NA,                 NA,           NA,        NA,               NA,                     NA,      NA,            NA,            NA,        NA,                                                                   NA,       22.44681,       -99.56705,        "Specimen",          "High",  "GBIF",                    NA,                  NA,                  NA,                   NA,                                 NA,                            NA,               NA,                   NA,                      NA,                      NA,                    NA,         2L, "Agapostemon leunculus",        "Mexico",
               13L, 23844L,    "Agapostemon leunculus", 1060924663L,           "EMEC",          "EMEC",                    NA,            NA, "PRESERVED_SPECIMEN",                                    "urn:catalog:EMEC:EMEC:511764",   "EMEC511764",                   NA, "Animalia", "Arthropoda", "Insecta", "Hymenoptera", "Halictidae",       NA,                        NA, "Agapostemon",      "leunculus",  "SPECIES",                    NA,                 NA,                     NA,                        NA,                     NA,            NA,                       NA,          NA, "A. E. Michelbacher, M. M. Michelbacher",            NA, "1962-07-27T00:00:00Z", 1962L,       "7",  27L,              NA,            NA,                 NA,                 NA,       NA,           NA,                   NA,                   NA,                 NA,              NA,                     NA,                     NA,                  NA,    "Adult",       NA,               1L,                NA,              NA,         "pin",     "MX",           "Oaxaca",                   NA,            NA,                          "Oaxaca",               NA,         17.06371,         -96.73192,             NA,                          8339L,                   NA,               NA,                    NA,                   NA,                              NA,                   NA,                        NA,                        NA,                    NA,                    NA,             NA,                 NA,           NA,        NA,               NA, "2017-05-04T00:00:00Z",      NA,            NA,            NA,        NA, "http://essigdb.berkeley.edu/cgi/eme_query?bnhm_id=EMEC511764&one=T",       17.06371,       -96.73192,        "Specimen",          "High",  "GBIF",                    NA,                  NA,                  NA,                   NA,                                 NA,                            NA,               NA,                   NA,                      NA,                      NA,                    NA,         2L, "Agapostemon leunculus",        "Mexico",
               14L, 23845L,    "Agapostemon leunculus",  657733166L,             "KU",          "SEMC",                    NA,            NA, "PRESERVED_SPECIMEN",                            "1e5ca92c-1ed6-11e3-bfac-90b11c41863e",       "495671",                   NA, "Animalia", "Arthropoda", "Insecta", "Hymenoptera", "Halictidae",       NA,                        NA, "Agapostemon",      "leunculus",  "SPECIES",                    NA,       "Roberts, R",                     NA,                        NA,                     NA,            NA,                       NA,          NA,             "Univ. of Kans. Mex. Exped.",            NA, "1961-06-11T00:00:00Z", 1961L,       "6",  11L,              NA,            NA,                 NA,                 NA,       NA,           NA,                   NA,                   NA,                 NA,              NA,                     NA,                     NA,                  NA,    "Adult",       NA,               NA,                NA,              NA,  "Pinned - 1",     "MX",         "Veracruz",                   NA,            NA,       "Alazan, 7 mi NW nr Tuxpan",               NA,          20.9915,         -97.54551,             NA,                             NA,                   NA,               NA,                    NA,                   NA,                              NA,                   NA,                        NA,                        NA,                    NA,                    NA,             NA,                 NA,           NA,        NA,               NA,                     NA,      NA,            NA,            NA,        NA,                                                                   NA,        20.9915,       -97.54551,        "Specimen",          "High",  "GBIF",                    NA,                  NA,                  NA,                   NA,                                 NA,                            NA,               NA,                   NA,                      NA,                      NA,                    NA,         2L, "Agapostemon leunculus",        "Mexico",
               15L, 23846L,    "Agapostemon leunculus", 1060924629L,           "EMEC",          "EMEC",                    NA,            NA, "PRESERVED_SPECIMEN",                                    "urn:catalog:EMEC:EMEC:511794",   "EMEC511794",                   NA, "Animalia", "Arthropoda", "Insecta", "Hymenoptera", "Halictidae",       NA,                        NA, "Agapostemon",      "leunculus",  "SPECIES",                    NA,                 NA,                     NA,                        NA,                     NA,            NA,                       NA,          NA, "A. E. Michelbacher, M. M. Michelbacher",            NA, "1962-07-26T00:00:00Z", 1962L,       "7",  26L,              NA,            NA,                 NA,                 NA,       NA,           NA,                   NA,                   NA,                 NA,              NA,                     NA,                     NA,                  NA,    "Adult",       NA,               1L,                NA,              NA,         "pin",     "MX",           "Oaxaca",                   NA,            NA,                          "Oaxaca",               NA,         17.06371,         -96.73192,             NA,                          8339L,                   NA,               NA,                    NA,                   NA,                              NA,                   NA,                        NA,                        NA,                    NA,                    NA,             NA,                 NA,           NA,        NA,               NA, "2017-05-04T00:00:00Z",      NA,            NA,            NA,        NA, "http://essigdb.berkeley.edu/cgi/eme_query?bnhm_id=EMEC511794&one=T",       17.06371,       -96.73192,        "Specimen",          "High",  "GBIF",                    NA,                  NA,                  NA,                   NA,                                 NA,                            NA,               NA,                   NA,                      NA,                      NA,                    NA,         2L, "Agapostemon leunculus",        "Mexico",
               16L, 23847L,    "Agapostemon leunculus",  219014679L,         "IBUNAM",          "EBCC",                    NA,            NA,            "UNKNOWN",                                                                NA,       "300170",                   NA, "Animalia", "Arthropoda", "Insecta", "Hymenoptera", "Halictidae",       NA,                        NA, "Agapostemon",      "leunculus",  "SPECIES",                    NA,   "Douglas Yanega",                     NA,                        NA,                     NA,            NA,                       NA,          NA,             "Felipe Noguera Mart√É¬≠nez",            NA, "1993-04-10T00:00:00Z", 1993L,       "4",  10L,              NA,            NA,                 NA,                 NA,       NA,           NA,                   NA,                   NA,                 NA,              NA,                     NA,                     NA,                  NA,         NA,       NA,               NA,                NA,              NA,            NA,     "MX",          "Chiapas",                   NA,            NA,           "Ocozocoautla, 17 km N",               NA,            16.89,            -93.46,             NA,                             NA,                   NA,               NA,                    NA,                   NA,                              NA,                   NA,                        NA,                        NA,                    NA,                    NA,             NA,                 NA,           NA,        NA,               NA,                     NA,      NA,            NA,            NA,        NA,                                                                   NA,          16.89,          -93.46,        "Specimen",          "High",  "GBIF",                    NA,                  NA,                  NA,                   NA,                                 NA,                            NA,               NA,                   NA,                      NA,                      NA,                    NA,         2L, "Agapostemon leunculus",        "Mexico",
               17L, 23848L,    "Agapostemon leunculus",  219014687L,         "IBUNAM",          "EBCC",                    NA,            NA,            "UNKNOWN",                                                                NA,       "300178",                   NA, "Animalia", "Arthropoda", "Insecta", "Hymenoptera", "Halictidae",       NA,                        NA, "Agapostemon",      "leunculus",  "SPECIES",                    NA,   "Douglas Yanega",                     NA,                        NA,                     NA,            NA,                       NA,          NA,                        "Claudia Mendoza",            NA, "1993-04-13T00:00:00Z", 1993L,       "4",  13L,              NA,            NA,                 NA,                 NA,       NA,           NA,                   NA,                   NA,                 NA,              NA,                     NA,                     NA,                  NA,         NA,       NA,               NA,                NA,              NA,            NA,     "MX",          "Chiapas",                   NA,            NA,                        "Teopisca",               NA,         16.53333,         -92.46667,             NA,                             NA,                   NA,               NA,                    NA,                   NA,                              NA,                   NA,                        NA,                        NA,                    NA,                    NA,             NA,                 NA,           NA,        NA,               NA,                     NA,      NA,            NA,            NA,        NA,                                                                   NA,       16.53333,       -92.46667,        "Specimen",          "High",  "GBIF",                    NA,                  NA,                  NA,                   NA,                                 NA,                            NA,               NA,                   NA,                      NA,                      NA,                    NA,         2L, "Agapostemon leunculus",        "Mexico"
              ) %>%
  # Change the column name from Source to dataSource to match the rest of the data.
  dplyr::rename(dataSource = Source) %>%
  # EXTRACT WAS HERE
  # add a NEW database_id column
  dplyr::mutate(
    database_id = paste0("Paige_data_", 1:nrow(.)),
    .before = scientificName)

  
  


columnList <-  list(c("decimalLatitude", "decimalLongitude", 
                   "recordNumber", "recordedBy", "individualCount", "samplingProtocol",
                   "associatedTaxa", "sex", "catalogNumber", "institutionCode", "otherCatalogNumbers",
                   "recordId", "occurrenceID", "collectionID"), # Iteration 1
c("catalogNumber", "institutionCode", "otherCatalogNumbers",
  "recordId", "occurrenceID", "collectionID"), # Iteration 2
c("decimalLatitude", "decimalLongitude", 
  "recordedBy", "genus", "specificEpithet"), # Iteration 3
c("id", "decimalLatitude", "decimalLongitude"), # Iteration 4
c("recordedBy", "genus", "specificEpithet", "locality"), # Iteration 5
c("recordedBy", "institutionCode", "genus", 
  "specificEpithet","locality"),# Iteration 6
c("occurrenceID","decimalLatitude", "decimalLongitude"), # Iteration 7
c("catalogNumber","decimalLatitude", "decimalLongitude"), # Iteration 8
c("catalogNumber", "locality") # Iteration 9
)

## generate new data using function
testOut <- BeeBDC::PaigeIntegrater(db_standardized = db_standardized, 
                                  PaigeNAm = PaigeNAm, 
                                  columnStrings = columnList)
                                   

## test the function
# the correct answer 
correct <- dplyr::tribble(
             ~decimalLatitude, ~decimalLongitude, ~recordNumber,                                  ~recordedBy, ~individualCount, ~samplingProtocol, ~associatedTaxa,     ~sex,      ~catalogNumber, ~institutionCode, ~otherCatalogNumbers, ~recordId,                          ~occurrenceID, ~collectionID,            ~genus, ~specificEpithet, ~id,                                                       ~locality,   ~database_id,         ~scientificName, ~infraspecificEpithet,                             ~country, ~coordinateUncertaintyInMeters,
                        2.233,               -90,            NA,                              "James Ogilvie",               1L,                NA,              NA, "FEMALE", "AMNH_IZC 00323864",           "AMNH",                   NA,        NA, "85fb767f-1f99-482f-896c-69dce8b59097",            NA, "Pseudoanthidium",      "truncatum",  NA,                                                              NA,    "db_fake_1",          "Fake species",             "species", "The incorporated states of nowhere",                           100L,
                        2.233,               -90,            NA,                         "T. D. A. Cockerell",               1L,                NA,              NA, "FEMALE", "AMNH_IZC 00323581",           "AMNH",                   NA,        NA, "63f56ed1-7391-4814-a298-c1b67ca986be",            NA,        "Halictus",        "simplex",  NA,                                                 "near Weing Sa",    "db_fake_2",          "Fake species",             "species", "The incorporated states of nowhere",                           100L,
                        2.233,               -90,            NA, "T. D. A. Cockerell & Wilmatte P. Cockerell",               1L,                NA,              NA,   "MALE", "AMNH_IZC 00323583",           "AMNH",                   NA,        NA, "cde3718b-8621-4b3f-b203-4436961a4b73",            NA,        "Halictus",        "simplex",  NA, "Lake Waskesiu, at the beginning of the portage to Heart Lakes",    "db_fake_3",          "Fake species",             "species", "The incorporated states of nowhere",                           100L,
                        2.233,               -90,            NA,                              "C. S. Brimley",               1L,                NA,              NA, "FEMALE", "AMNH_IZC 00324010",           "AMNH",                   NA,        NA, "82e1eab7-5d92-4358-aede-1fab15ba4eda",            NA,        "Halictus",        "simplex",  NA,                                                              NA,    "db_fake_4",          "Fake species",             "species", "The incorporated states of nowhere",                           100L,
                        2.233,               -90,            NA,                         "T. D. A. Cockerell",               1L,                NA,              NA, "FEMALE", "AMNH_IZC 00323630",           "AMNH",                   NA,        NA, "c161e78e-c091-4483-adf4-6b54045283d7",            NA,        "Halictus",        "simplex",  NA,                     "Prince Albert National Park, Lake Ajawaan",    "db_fake_5",          "Fake species",             "species", "The incorporated states of nowhere",                           100L,
                        2.233,               -90,            NA,                         "T. D. A. Cockerell",               1L,                NA,              NA, "FEMALE", "AMNH_IZC 00323584",           "AMNH",                   NA,        NA, "002a4861-5374-46ac-84d4-087762ba0c88",            NA,        "Halictus",        "simplex",  NA,                                                              NA,    "db_fake_6",          "Fake species",             "species", "The incorporated states of nowhere",                           100L,
                        2.233,               -90,            NA,       "T. D. A. Cockerell & Alice M. Mackie",               1L,                NA,              NA, "FEMALE", "AMNH_IZC 00323616",           "AMNH",                   NA,        NA, "5e848030-8e9d-4f64-873d-08fee631e7dd",            NA,        "Halictus",        "simplex",  NA,                                                              NA,    "db_fake_7",          "Fake species",             "species", "The incorporated states of nowhere",                           100L,
                        2.233,               -90,            NA,                            "Alice M. Mackie",               1L,                NA,              NA,   "MALE", "AMNH_IZC 00323619",           "AMNH",                   NA,        NA, "250ba041-2727-43d0-a39e-8d3470f5a8bf",            NA,        "Halictus",        "simplex",  NA,                                                              NA,    "db_fake_8",          "Fake species",             "species", "The incorporated states of nowhere",                           100L,
                        2.233,               -90,            NA,                         "T. D. A. Cockerell",               1L,                NA,              NA, "FEMALE", "AMNH_IZC 00323591",           "AMNH",                   NA,        NA, "4cad3ea9-11b3-4589-b1d1-0035a84ea861",            NA,        "Halictus",        "simplex",  NA,                                                              NA,    "db_fake_9",          "Fake species",             "species", "The incorporated states of nowhere",                           100L,
                        2.233,               -90,            NA,                         "T. D. A. Cockerell",               1L,                NA,              NA,   "MALE", "AMNH_IZC 00323601",           "AMNH",                   NA,        NA, "dc3ebe14-f11f-45e9-b121-d34a8693284e",            NA,        "Halictus",        "simplex",  NA,                                                              NA,   "db_fake_10",          "Fake species",             "species", "The incorporated states of nowhere",                           100L,
                        2.233,               -90,            NA,                                "A. H. Manes",               1L,                NA,              NA, "FEMALE", "AMNH_IZC 00324011",           "AMNH",                   NA,        NA, "6f1d8a81-3a4a-4650-be8c-7102f073e90e",            NA,        "Halictus",        "simplex",  NA,                                                              NA,   "db_fake_11",          "Fake species",             "species", "The incorporated states of nowhere",                           100L,
                        2.233,               -90,            NA,                         "T. D. A. Cockerell",               1L,                NA,              NA, "FEMALE", "AMNH_IZC 00323627",           "AMNH",                   NA,        NA, "406433bc-9c08-4e94-9768-d8df0bdd92cb",            NA,        "Halictus",        "simplex",  NA,                                                              NA,   "db_fake_12",          "Fake species",             "species", "The incorporated states of nowhere",                           100L,
                        2.233,               -90,            NA,                            "Alice M. Mackie",               1L,                NA,              NA, "FEMALE", "AMNH_IZC 00323611",           "AMNH",                   NA,        NA, "bf53740e-e96c-4b5a-a0e9-6b4bea3b2448",            NA,        "Halictus",        "simplex",  NA,                                        "Kaieteur National Park",   "db_fake_13",          "Fake species",             "species", "The incorporated states of nowhere",                           100L,
                        2.233,               -90,            NA,                         "T. D. A. Cockerell",               1L,                NA,              NA, "FEMALE", "AMNH_IZC 00323613",           "AMNH",                   NA,        NA, "2e6d4df0-0a46-4837-b340-ed6f8166a831",            NA,        "Halictus",        "simplex",  NA,                                               "Kum Puang Creek",   "db_fake_14",          "Fake species",             "species", "The incorporated states of nowhere",                           100L,
                        2.233,               -90,            NA,                            "Alice M. Mackie",               1L,                NA,              NA,       NA, "AMNH_IZC 00323617",           "AMNH",                   NA,        NA, "c0642780-37dc-4625-8a30-eacd5dbd1b70",            NA,        "Halictus",        "simplex",  NA,                                                              NA,   "db_fake_15",          "Fake species",             "species", "The incorporated states of nowhere",                           100L,
                        2.233,               -90,            NA,                         "T. D. A. Cockerell",               1L,                NA,              NA, "FEMALE",       "Fake_data_1",           "AMNH",                   NA,        NA, "259d5501-c49c-4a0b-99fc-1595ffe403b1",            NA,        "Halictus",        "simplex",  NA,                                                              NA,   "db_fake_16",          "Fake species",             "species", "The incorporated states of nowhere",                           100L,
                     17.21667,             -92.1,            NA,                                 "Easter Egg",               NA,                NA,              NA,       NA,       "Fake_data_2",      "Bee Place",                   NA,        NA,                        "not_a_number1",            NA,     "Agapostemon",      "leunculus",  NA,                                                  "non-existent", "Paige_data_5", "Agapostemon leunculus",             "species",                                 "MX",                           100L,
                        2.233,               -90,            NA,                                 "Easter Egg",               NA,                NA,              NA,       NA,       "Fake_data_3",      "Bee Place",                   NA,        NA,                        "not_a_number2",            NA,  "Augochloropsis",        "fulgida",  NA,                                                  "non-existent",   "db_fake_18",          "Fake species",             "species", "The incorporated states of nowhere",                           100L,
                        2.233,               -90,            NA,                                 "Easter Egg",               NA,                NA,              NA,       NA,       "Fake_data_4",      "Bee Place",                   NA,        NA,                        "not_a_number3",            NA,  "Augochloropsis",        "fulgida",  NA,                                                  "non-existent",   "db_fake_19",          "Fake species",             "species", "The incorporated states of nowhere",                           100L,
                        2.233,               -90,            NA,                                 "Easter Egg",               NA,                NA,              NA,       NA,       "Fake_data_5",      "Bee Place",                   NA,        NA,                        "not_a_number4",            NA,  "Augochloropsis",        "fulgida",  NA,                                                  "non-existent",   "db_fake_20",          "Fake species",             "species", "The incorporated states of nowhere",                           100L,
                        2.233,               -90,            NA,                                 "Easter Egg",               NA,                NA,              NA,       NA,       "Fake_data_6",      "Bee Place",                   NA,        NA,                        "not_a_number5",            NA,  "Augochloropsis",        "fulgida",  NA,                                                  "non-existent",   "db_fake_21",          "Fake species",             "species", "The incorporated states of nowhere",                           100L,
                        2.233,               -90,            NA,                                 "Easter Egg",               NA,                NA,              NA,       NA,       "Fake_data_7",      "Bee Place",                   NA,        NA,                        "not_a_number6",            NA,  "Augochloropsis",        "fulgida",  NA,                                                  "non-existent",   "db_fake_22",          "Fake species",             "species", "The incorporated states of nowhere",                           100L,
                        2.233,               -90,            NA,                                 "Easter Egg",               NA,                NA,              NA,       NA,       "Fake_data_8",      "Bee Place",                   NA,        NA,                        "not_a_number7",            NA,  "Augochloropsis",        "fulgida",  NA,                                                  "non-existent",   "db_fake_23",          "Fake species",             "species", "The incorporated states of nowhere",                           100L,
                        2.233,               -90,            NA,                                 "Easter Egg",               NA,                NA,              NA,       NA,       "Fake_data_9",      "Bee Place",                   NA,        NA,                        "not_a_number8",            NA,  "Augochloropsis",        "fulgida",  NA,                                                  "non-existent",   "db_fake_24",          "Fake species",             "species", "The incorporated states of nowhere",                           100L,
                        2.233,               -90,            NA,                                 "Easter Egg",               NA,                NA,              NA,       NA,      "Fake_data_10",      "Bee Place",                   NA,        NA,                        "not_a_number9",            NA,  "Augochloropsis",        "fulgida",  NA,                                                  "non-existent",   "db_fake_25",          "Fake species",             "species", "The incorporated states of nowhere",                           100L,
                        2.233,               -90,            NA,                                       "Nope",               NA,                NA,              NA,       NA,      "Fake_data_11",      "Bee Place",                   NA,        NA,                       "not_a_number10",            NA,  "Augochloropsis",        "fulgida",  NA,                                                  "non-existent",   "db_fake_26",          "Fake species",             "species", "The incorporated states of nowhere",                           100L,
                        16.75,         -92.91667,            NA,                                       "Nope",               NA,                NA,              NA,       NA,      "Fake_data_12",      "Bee Place",                   NA,        NA,                       "not_a_number11",            NA,     "Agapostemon",      "leunculus",  NA,                                                  "non-existent", "Paige_data_8", "Agapostemon leunculus",             "species",                                 "MX",                           100L,
                        2.233,               -90,            NA,                                       "Nope",               NA,                NA,              NA,       NA,      "Fake_data_13",      "Bee Place",                   NA,        NA,                       "not_a_number12",            NA,    "Lasioglossum",  "semicaeruleum",  NA,                                                  "non-existent",   "db_fake_28",          "Fake species",             "species", "The incorporated states of nowhere",                           100L,
                        2.233,               -90,            NA,                                       "Nope",               NA,                NA,              NA,       NA,      "Fake_data_14",      "Bee Place",                   NA,        NA,                       "not_a_number13",            NA,    "Lasioglossum",  "semicaeruleum",  NA,                                                  "non-existent",   "db_fake_29",          "Fake species",             "species", "The incorporated states of nowhere",                           100L,
                        2.233,               -90,            NA,                                       "Nope",               NA,                NA,              NA,       NA,      "Fake_data_15",      "Bee Place",                   NA,        NA,                       "not_a_number14",            NA,    "Lasioglossum",  "semicaeruleum",  NA,                                                  "non-existent",   "db_fake_30",          "Fake species",             "species", "The incorporated states of nowhere",                           100L
             )

  
testthat::test_that("PaigeIntegrater revised datatable with correct values", {
    testthat::expect_equal(testOut, correct)
})

  # These functions were written by James B Dorey beginning on the 17th of June 2022 to read in,
    # Format and save various datasets.
  # For questions, please email jbdorey[at]me.com


#### x.0 readr_BeeBDC function ####
#' A wrapper for all of the data readr_functions
#' 
#' Read in a variety of data files that are specific to certain smaller data providers. 
#' There is an internal readr function for each dataset and each one of these functions is called
#' by readr_BeeBDC. While these functions are internal, they are displayed in the documentation of 
#' readr_BeeBDC for clarity.
#' 
#' This function wraps several internal readr functions. Users may call
#' readr_BeeBDC and select the dataset name to import a certain dataset. These datasets include:
#' 
#' Excel (.xlsx) formatted datasets: CAES, MABC, Col, Bal, MEPB, MUPJ, Arm, JoLa, and VicWam.
#' 
#' CSV (.csv) formatted datasets: EPEL, ASP, BMin, BMont, Ecd, Gai, KP, EcoS, GeoL, EaCo, FSCA, SMC,
#' Lic, Dor, BBD, STRI, and PALA
#' 
#' See Dorey et al. 2023 BeeBDC... for further details.
#'
#' @param dataset Character. The name of the dataset to be read in. For example readr_CAES can
#' be called using "readr_CAES" or "CAES". This is not caps sensitive.
#' @param path A character path. The path to the directory containing the data.
#' @param inFile Character or character path. The name of the file itself (can also be the 
#' remainder of a path including the file name).
#' @param outFile Character or character path. The name of the Darwin Core format file to be saved.
#' @param dataLicense Character. The license to accompany each record in the Darwin Core 'license' 
#' column.
#' @param sheet A character String. For those datasets read from an .xlsx format, provide the 
#' sheet name. 
#' NOTE: This will be ignored for .csv readr_ functions and required for .xlsx readr_ functions.
#'
#' @return A data frame that is in Darwin Core format.
#' @export
#' 
#' @importFrom readr read_csv write_excel_csv
#' @importFrom dplyr rename mutate select if_else  %>%
#' @importFrom lubridate ymd month
#' @importFrom stringr str_c
#' 
#'
#' @examples
#' \dontrun{
#' # An example using a .xlsx file
#'Arm_Data <- readr_BeeBDC(
#'     dataset = "Arm",
#'     path = paste0(tempdir(), "/Additional_Datasets"),
#'     inFile = "/InputDatasets/Bee database Armando_Final.xlsx",
#'     outFile = "jbd_Arm_Data.csv",
#'     sheet = "Sheet1",
#'     dataLicense = "https://creativecommons.org/licenses/by-nc-sa/4.0/")
#'     
#'     
#'     # An example using a .csv file
#'EPEL_Data <- readr_BeeBDC(
#'   dataset = "readr_EPEL",
#'   path = paste0(tempdir(), "/Additional_Datasets"),
#'   inFile = "/InputDatasets/bee_data_canada.csv",
#'   outFile = "jbd_EPEL_data.csv",
#'   dataLicense = "https://creativecommons.org/licenses/by-nc-sa/4.0/")
#' }
readr_BeeBDC <- function(
    dataset = NULL,
    path = NULL,
    inFile = NULL,
    outFile = NULL,
    dataLicense = NULL,
    sheet = NULL){
  
  #### x.0 Prep ####
  ##### x.1 Errors ####
  ###### a. FATAL errors ####
  if(is.null(dataset)){
    stop(" - Please provide an argument for dataset. This can be from any of the readr_ functions")
  }
  if(is.null(path)){
    stop(" - Please provide an argument for the path to a folder containing your file.")
  }
  if(is.null(inFile)){
    stop(" - Please provide an argument for the inFile name to find.")
  }
  if(is.null(outFile)){
    stop(" - Please provide an argument for outFile to save as.")
  }
  if(is.null(dataLicense)){
    stop(" - Please provide an argument for dataLicense.")
  }
  
  ##### x.2 Data types ####
  # Create the lists of potential data types for .xlsx or .csv inputs
  excelTypes <- c("CAES", "MABC", "Col", "Bal", "MEPB", "MPUJ", "Arm", "JoLa", "VicWam")
  csvTypes <- c("EPEL", "ASP", "BMin", "BMont", "Ecd", "Gai", "KP", "EcoS", "GeoL",
                "EaCo", "FSCA", "SMC", "Lic", "Dor", "BBD", "PALA", "STRI")
  
  # Wrong entry test
  if(!tolower(dataset) %in% tolower(c(paste0("readr_",csvTypes), csvTypes,
                              paste0("readr_",excelTypes), excelTypes)) ){
    stop("The dataset does not match a readr_BeeBDC function")
  }
  
  
  #### x.0 Choose function ####
  ##### x.1 Excel functions ####
  # EXCEL test
  if(tolower(dataset) %in% tolower(c(paste0("readr_",excelTypes), excelTypes)) ){
    message("A .xlsx data type was chosen...")
    # If no sheet is provided
    if(is.null(sheet)){
      stop(" - No sheet argument was provided. Please check for the sheet name to import.")
    }
    ###### a. CAES ####
    if(tolower(dataset) %in% tolower(c("readr_CAES", "CAES")) ){
      dataOut <- readr_CAES(path = path,
                        inFile = inFile,
                        outFile = outFile,
                        dataLicense = dataLicense,
                        sheet = sheet)}
    ##### b. MABC ####
    if(tolower(dataset) %in% tolower(c("readr_MABC", "MABC")) ){
      dataOut <- readr_MABC(path = path,
                        inFile = inFile,
                        outFile = outFile,
                        dataLicense = dataLicense,
                        sheet = sheet)}
    ##### c. Col ####
    if(tolower(dataset) %in% tolower(c("readr_Col", "Col")) ){
      dataOut <- readr_Col(path = path,
                       inFile = inFile,
                       outFile = outFile,
                       dataLicense = dataLicense,
                       sheet = sheet)}
    ##### d. Bal ####
    if(tolower(dataset) %in% tolower(c("readr_Bal", "Bal")) ){
      dataOut <- readr_Bal(path = path,
                       inFile = inFile,
                       outFile = outFile,
                       dataLicense = dataLicense,
                       sheet = sheet)}
    ##### e. MEPB ####
    if(tolower(dataset) %in% tolower(c("readr_MEPB", "MEPB")) ){
      dataOut <- readr_MEPB(path = path,
                        inFile = inFile,
                        outFile = outFile,
                        dataLicense = dataLicense,
                        sheet = sheet)}
    ##### f. MPUJ ####
    if(tolower(dataset) %in% tolower(c("readr_MPUJ", "MPUJ")) ){
      dataOut <- readr_MPUJ(path = path,
                        inFile = inFile,
                        outFile = outFile,
                        dataLicense = dataLicense,
                        sheet = sheet)}
    ##### f. Arm ####
    if(tolower(dataset) %in% tolower(c("readr_Arm", "Arm")) ){
      dataOut <- readr_Arm(path = path,
                       inFile = inFile,
                       outFile = outFile,
                       dataLicense = dataLicense,
                       sheet = sheet)}
    ##### g. JoLa ####
    if(tolower(dataset) %in% tolower(c("readr_JoLa", "JoLa")) ){
      dataOut <- readr_JoLa(path = path,
                        inFile = inFile,
                        outFile = outFile,
                        dataLicense = dataLicense,
                        sheet = sheet)}
    ##### h. VicWam ####
    if(tolower(dataset) %in% tolower(c("readr_VicWam", "VicWam")) ){
      dataOut <- readr_VicWam(path = path,
                            inFile = inFile,
                            outFile = outFile,
                            dataLicense = dataLicense,
                            sheet = sheet)}
  }# END Excel functions
  
  ##### x.2 CSV functions ####
  # CSV test
  if(tolower(dataset) %in% tolower(c(paste0("readr_",csvTypes), csvTypes)) ){
    message("A .csv data type was chosen...")
    
    ###### a. EPEL ####
    if(tolower(dataset) %in% tolower(c("readr_EPEL", "EPEL")) ){
      dataOut <- readr_EPEL(path = path,
                        inFile = inFile,
                        outFile = outFile,
                        dataLicense = dataLicense)}
    ##### b. ASP ####
    if(tolower(dataset) %in% tolower(c("readr_ASP", "ASP")) ){
      dataOut <- readr_ASP(path = path,
                       inFile = inFile,
                       outFile = outFile,
                       dataLicense = dataLicense)}
    ##### c. BMin ####
    if(tolower(dataset) %in% tolower(c("readr_BMin", "BMin")) ){
      dataOut <- readr_BMin(path = path,
                        inFile = inFile,
                        outFile = outFile,
                        dataLicense = dataLicense)}
    ##### d. BMont ####
    if(tolower(dataset) %in% tolower(c("readr_BMont", "BMont")) ){
      dataOut <- readr_BMont(path = path,
                         inFile = inFile,
                         outFile = outFile,
                         dataLicense = dataLicense)}
    ##### e. Ecd ####
    if(tolower(dataset) %in% tolower(c("readr_Ecd", "Ecd")) ){
      dataOut <- readr_Ecd(path = path,
                       inFile = inFile,
                       outFile = outFile,
                       dataLicense = dataLicense)}
    ##### f. Gai ####
    if(tolower(dataset) %in% tolower(c("readr_Gai", "Gai")) ){
      dataOut <- readr_Gai(path = path,
                       inFile = inFile,
                       outFile = outFile,
                       dataLicense = dataLicense)}
    ##### g. KP ####
    if(tolower(dataset) %in% tolower(c("readr_KP", "KP")) ){
      dataOut <- readr_KP(path = path,
                      inFile = inFile,
                      outFile = outFile,
                      dataLicense = dataLicense)}
    ##### h. EcoS ####
    if(tolower(dataset) %in% tolower(c("readr_EcoS", "EcoS")) ){
      dataOut <- readr_EcoS(path = path,
                        inFile = inFile,
                        outFile = outFile,
                        dataLicense = dataLicense)}
    ##### i. GeoL ####
    if(tolower(dataset) %in% tolower(c("readr_GeoL", "GeoL")) ){
      dataOut <- readr_GeoL(path = path,
                        inFile = inFile,
                        outFile = outFile,
                        dataLicense = dataLicense)}
    ##### j. EaCO ####
    if(tolower(dataset) %in% tolower(c("readr_EaCO", "EaCo")) ){
      dataOut <- readr_EaCO(path = path,
                        inFile = inFile,
                        outFile = outFile,
                        dataLicense = dataLicense)}
    ##### k. FSCA ####
    if(tolower(dataset) %in% tolower(c("readr_FSCA", "FSCA")) ){
      dataOut <- readr_FSCA(path = path,
                        inFile = inFile,
                        outFile = outFile,
                        dataLicense = dataLicense)}
    ##### l. SMC ####
    if(tolower(dataset) %in% tolower(c("readr_SMC", "SMC")) ){
      dataOut <- readr_SMC(path = path,
                       inFile = inFile,
                       outFile = outFile,
                       dataLicense = dataLicense)}
    ##### m. Lic ####
    if(tolower(dataset) %in% tolower(c("readr_Lic", "Lic")) ){
      dataOut <- readr_Lic(path = path,
                       inFile = inFile,
                       outFile = outFile,
                       dataLicense = dataLicense)}
    ##### n. Dor ####
    if(tolower(dataset) %in% tolower(c("readr_Dor", "Dor")) ){
      dataOut <- readr_Dor(path = path,
                       inFile = inFile,
                       outFile = outFile,
                       dataLicense = dataLicense)}
    ##### o. BBD ####
    if(tolower(dataset) %in% tolower(c("readr_BBD", "BBD")) ){
      dataOut <- readr_BBD(path = path,
                       inFile = inFile,
                       outFile = outFile,
                       dataLicense = dataLicense)}
    ##### p. STRI ####
    if(tolower(dataset) %in% tolower(c("readr_STRI", "STRI")) ){
      dataOut <- readr_STRI(path = path,
                        inFile = inFile,
                        outFile = outFile,
                        dataLicense = dataLicense)}
    ##### g. PALA ####
    if(tolower(dataset) %in% tolower(c("readr_PALA", "PALA")) ){
      dataOut <- readr_PALA(path = path,
                        inFile = inFile,
                        outFile = outFile,
                        dataLicense = dataLicense)}
    
    
    
  } #END csv
  
  return(dataOut)
} # END readr_BeeBDC





#### 1.0 EPEL ####
#' @rdname readr_BeeBDC
#' 
#' 
#'
readr_EPEL <- function(path = NULL,
                      inFile = NULL,
                      outFile = NULL,
                      dataLicense = NULL){
  # locally bind variables to the function
  catalog_number<-pollinator_family<-pollinator_genus<-pollinator_species<-collection_method<-
    day_collected<-month_collected<-year_collected<-location_description<-latitude<-longitude<-
    basis_of_record<-genus<-specificEpithet<-year<-day<-eventDate<-collector_number<-
    location_name<-habitat<-.<-catalogNumber <- month<-NULL
  
  #### 1.1 Prep ####
  requireNamespace("dplyr")
  requireNamespace("lubridate")

  #### 1.2 Read+ ####
  EPEL_Data <- readr::read_csv(paste(path, inFile, sep = "/"),
                              trim_ws = TRUE) %>%
  # Rename some columns to make the consistent with DarwinCore
  dplyr::rename(
    catalogNumber =	catalog_number,
    family =	pollinator_family,
    genus =	pollinator_genus,
    specificEpithet =	pollinator_species,
    samplingProtocol =	collection_method,
    day =	day_collected,
    month =	month_collected,
    year =	year_collected,
    locality =	location_description,
    decimalLatitude =	latitude,
    decimalLongitude =	longitude,
    basisOfRecord =	basis_of_record) %>%
  # Make new columns
  dplyr::mutate(
    # Merge to scientific name
    scientificName = paste(genus, specificEpithet),
    # Add data source
    dataSource = "EPEL_Anthophila",
    # Add eventDate
    eventDate = lubridate::ymd(paste(year, month, day, sep = "-"),
                               truncated = 2),
    month = lubridate::month(eventDate)) %>%
      # Add fieldNotes
    dplyr::mutate(
      fieldNotes = stringr::str_c(
        dplyr::if_else(!is.na(collector_number),
                       paste0("Collector_number: ", collector_number), ""),
        dplyr::if_else(!is.na(location_name),
                       paste0("location_name: ", location_name), ""),
        dplyr::if_else(!is.na(habitat),
                       paste0("habitat: ", habitat), ""),
        sep = "|") %>%
        # Remove extra bars "|".
        stringr::str_replace_all(pattern = "(\\|){2,9}",
                                 replacement = "\\|") %>%
        stringr::str_replace_all(pattern = "(\\|$)+|(^\\|)+",
                                 replacement = "")) %>%
  # Remove these input columns
  dplyr::select(!c(collector_number, location_name, habitat)) %>%
  # add the database_id column
  dplyr::mutate(
    database_id = paste("EPEL_data_", 1:nrow(.), sep = ""),
    .before = catalogNumber) %>%
    dplyr::mutate(license = dataLicense) %>%
  # add the database_id column
    dplyr::mutate(
      datasetName = "Elle Pollination Ecology Lab",
      datasetID = "EPEL",
      institutionCode = "SFU"
    )
  
  #### 1.3 Out ####
# Save the dataset
readr::write_excel_csv(EPEL_Data, file = paste(path, outFile, sep = "/"))
# Return data
return(EPEL_Data)
} # END readr_EPEL


#### 2.0 ASP ####
#' @rdname readr_BeeBDC
#' 
#' 
#'
readr_ASP <- function(path = NULL,
                         inFile = NULL,
                         outFile = NULL,
                      dataLicense = NULL){
  # locally bind variables to the function
  Tribe <- Morphospecies <- Successional_Stage <- genus <- specificEpithet <- NULL
  eventDate <- catalogNumber <- . <- elevation <- NULL
  #### 2.1 Prep ####
  requireNamespace("dplyr")
  
  requireNamespace("lubridate")


  
  #### 2.2 Read+ ####
ASP_data <- readr::read_csv(paste(path, inFile, sep = "/"),
                            trim_ws = TRUE) %>%
    dplyr::mutate(verbatimElevation = elevation) %>%
  dplyr::mutate(
    # Add previousIdentifications
    previousIdentifications = stringr::str_c(
      dplyr::if_else(!is.na(Tribe),
                     paste0("Tribe: ", Tribe), ""),
      dplyr::if_else(!is.na(Morphospecies),
                     paste0("Morphospecies: ", Morphospecies), ""),
      sep = "|") %>%
      # Remove extra bars "|".
      stringr::str_replace_all(pattern = "(\\|){2,9}",
                               replacement = "\\|") %>%
      stringr::str_replace_all(pattern = "(\\|$)+|(^\\|)+",
                               replacement = "")) %>%
  dplyr::mutate(
    # Add locationRemarks
    locationRemarks = paste0("Successional Stage:", Successional_Stage),
    # Add scientificName
    scientificName = paste(genus, specificEpithet),
    # Add dataSource
    dataSource = "ASP_Anthophila",
    # Add eventDate
    eventDate = lubridate::dmy(eventDate,
                               truncated = 2),
    # I'm sorry but this catalogNumber is useless. I'm going to edit it more-unique
    catalogNumber = dplyr::if_else(!is.na(catalogNumber),
                                   paste0("ASP_", catalogNumber), "")) %>%
  # add the database_id column
  dplyr::mutate(
    database_id = paste("ASP_data_", 1:nrow(.), sep = ""),
    .before = catalogNumber) %>%
  dplyr::mutate(license = dataLicense) %>%
  # add the database_id column
  dplyr::mutate(
    datasetName = "Allan Smith-Pardo",
    datasetID = "ASP"
  )

  #### 2.3 Out ####
# Save the dataset
readr::write_excel_csv(ASP_data, file = paste(path, outFile, sep = "/"))
# Return data
return(ASP_data)
} # END readr_ASP


#### 3.0 BMin ####
 
 #' @rdname readr_BeeBDC
#' 
#' 
#'

readr_BMin <- function(path = NULL,
                      inFile = NULL,
                      outFile = NULL,
                      dataLicense = NULL){
  # locally bind variables to the function
  eventDate <- . <- catalogNumber <- NULL
  catalog_number <- pollinator_family <- pollinator_genus <- pollinator_species <- NULL
  collection_method <- day_collected <- month_collected <- year_collected <- NULL
  location_description <- latitude <- longitude <- basis_of_record <- genus <- NULL
  specificEpithet <- year <- day <- eventDate <- collector_number <- location_name <- NULL
  habitat <- . <- catalogNumber <- NULL
  
  
  #### 3.1 Prep ####
  requireNamespace("dplyr")
  
  requireNamespace("lubridate")


  #### 3.2 Read+ ####
BMin_data <- readr::read_csv(paste(path, inFile, sep = "/"),
                               trim_ws = TRUE) %>%
  dplyr::mutate(
    # Format eventDate
    eventDate = lubridate::dmy(eventDate,
                               truncated = 2, quiet = TRUE),
    dataSource = "BMin_Anthophila") %>%
  # add the database_id column
  dplyr::mutate(
    database_id = paste("BMin_data_", 1:nrow(.), sep = ""),
    .before = catalogNumber)  %>%
  dplyr::mutate(license = dataLicense) %>%
  # add the database_id column
  dplyr::mutate(
    datasetName = "Robert Minckley",
    datasetID = "BMin",
    institutionCode = "University of Rochester"
  )
  #### 3.3 Out ####
# Save the dataset
readr::write_excel_csv(BMin_data, file = paste(path, outFile, sep = "/"))
# Return data
return(BMin_data)
} # END readr_BMin



#### 4.0 BMont ####
#' @rdname readr_BeeBDC
#' 
#' 
#'
readr_BMont <- function(path = NULL,
                      inFile = NULL,
                      outFile = NULL,
                      dataLicense = NULL){
  # locally bind variables to the function
  occurence_lsid <- fieldNotes <- GPS_device <- . <- catalogNumber <- eventDate <- dateTest <- NULL
  tempDate <- mdy <- VerbatimEventDate <- ymd<-NULL
  
  
  #### 4.1 Prep ####
  requireNamespace("dplyr")
  
  requireNamespace("lubridate")


  
  #### 4.2 Read+ ####
BMont_data <- readr::read_csv(paste(path, inFile, sep = "/"),
                              trim_ws = TRUE) %>%
  dplyr::rename(
    occurrenceID = occurence_lsid,
    verbatimEventDate = VerbatimEventDate) %>%
  dplyr::mutate(
    # Add fieldNotes 
     fieldNotes = stringr::str_c(
      dplyr::if_else(!is.na(fieldNotes),
                     paste0("fieldNotes: ", fieldNotes), ""),
      dplyr::if_else(!is.na(GPS_device),
                     paste0("GPS_device: ", GPS_device), ""),
      sep = "|") %>%
       # Remove extra bars "|".
       stringr::str_replace_all(pattern = "(\\|){2,9}",
                                replacement = "\\|") %>%
       stringr::str_replace_all(pattern = "(\\|$)+|(^\\|)+",
                                replacement = "")) %>%
    dplyr::mutate(dataSource = "BMont_Anthophila") %>%
  # add the database_id column
  dplyr::mutate(
    database_id = paste("BMont_data_", 1:nrow(.), sep = ""),
    .before = catalogNumber) %>%
  # Find the date ranges
  dplyr::mutate(dateTest = dplyr::if_else(stringr::str_count(eventDate) > 11,
                                          stringr::str_replace(string = eventDate,
                                                               pattern = "/", replacement = ","), 
                                          "FALSE")) %>%
  # Append these to fieldNotes
  dplyr::mutate(fieldNotes = dplyr::if_else(dateTest != FALSE & !is.na(dateTest),
                                            paste0(fieldNotes,
                                                   "|dateRange: ", eventDate), fieldNotes)) %>%
  # Now remove those eventDates with a range and instead insert the start date
  dplyr::mutate(tempDate = dplyr::if_else(dateTest != FALSE & !is.na(dateTest),
                                          paste0(stringr::str_replace(eventDate,
                                                                      pattern = ".*/",
                                                                      replacement = "")), 
                                          eventDate)) %>%
  # Pick up dates of different formats and format together.
  dplyr::mutate(mdy = lubridate::mdy(tempDate, quiet = TRUE)) %>%
  dplyr::mutate(ymd = lubridate::ymd(tempDate, quiet = TRUE)) %>%
  tidyr::unite(col = eventDate,
               mdy, ymd, sep = "", na.rm = TRUE,
               remove = FALSE) %>% 
  # Remove working columns
  dplyr::select(!c(dateTest, tempDate, mdy, ymd)) %>%
  dplyr::mutate(license = dataLicense) %>%
  # add the database_id column
  dplyr::mutate(
    datasetName = "Bombus Montana",
    datasetID = "BMon"
  )
  #### 4.3 Out ####
# Save the dataset
readr::write_excel_csv(BMont_data, file = paste(path, outFile, sep = "/"))
  # Return data
return(BMont_data)
} # END readr_BMont


#### 5.0 Ecd ####
#' @rdname readr_BeeBDC
#' 
#' 
#' 
#'
readr_Ecd <- function(path = NULL,
                       inFile = NULL,
                       outFile = NULL,
                      dataLicense = NULL){
  # locally bind variables to the function
  year <- day <- institutionCode <- id <- . <- catalogNumber <- recordID <-month<- NULL
  
  #### 5.1 Prep ####
  requireNamespace("dplyr")
  
  requireNamespace("lubridate")


  
  #### 5.2 Read+ ####
Ecd_data <- readr::read_csv(paste(path, inFile, sep = "/"),
                            trim_ws = TRUE) %>%
    dplyr::rename(recordId = recordID) %>%
  dplyr::mutate(
    # Format eventDate
    eventDate = lubridate::ymd(paste(year, month, day, sep = "-"),
                               truncated = 2, quiet = TRUE),
    dataSource = "Ecd_Anthophila",
    # I want to make sure id is unique...
    id = paste(institutionCode, id, sep = "_")) %>%
  # add the database_id column
  dplyr::mutate(
    database_id = paste("Ecd_data_", 1:nrow(.), sep = ""),
    .before = catalogNumber)  %>%
  dplyr::mutate(license = dataLicense) %>%
  # add the database_id column
  dplyr::mutate(
    datasetName = "Ecdysis",
    datasetID = "Ecd"
  )

  #### 5.3 Out ####
# Save the dataset
readr::write_excel_csv(Ecd_data, file = paste(path, outFile, sep = "/"))
# Return data
return(Ecd_data)
} # END readr_Ecd

###### 6.0 Gai ####
#' @rdname readr_BeeBDC
#' 
#' 
#' 
#'
readr_Gai <- function(path = NULL,
                       inFile = NULL,
                       outFile = NULL,
                      dataLicense = NULL){
  # locally bind variables to the function
  species<-subspecies<-SpecimenLocation<-eventTime<-EndTime<-TempStart<-TempEnd<-WindStart<-
    WindEnd<-SkyStart<-SkyEnd<-Site<-siteLocality<-syd<-eventDate<-.<-institutionCode <- NULL
  
  #### 6.1 Prep ####
  requireNamespace("dplyr")
  
  requireNamespace("lubridate")


  #### 6.2 Read+ ####
Gai_data <- readr::read_csv(paste(path, inFile, sep = "/"),
                            trim_ws = TRUE) %>%
  # Make columns DarwinCore-compatible
  dplyr::rename(
    collectionCode = 'Collection Code',
    otherCatalogNumbers = 'Other Catalog Number',
    specificEpithet = species,
    infraspecificEpithet = subspecies)  %>%
  # Add locationRemarks from a bunch of other columns
  dplyr::mutate(
    locationRemarks = stringr::str_c(
    dplyr::if_else(!is.na(SpecimenLocation), paste0("SpecimenLocation: ", SpecimenLocation), ""),
    dplyr::if_else(!is.na(eventTime),
                   paste0("StartTime: ", eventTime), ""),
    dplyr::if_else(!is.na(EndTime),
                   paste0("EndTime: ", EndTime), ""),
    dplyr::if_else(!is.na(TempStart),
                   paste0("TempStart: ", TempStart), ""),
    dplyr::if_else(!is.na(TempEnd),
                   paste0("TempEnd: ", TempEnd), ""),
    dplyr::if_else(!is.na(WindStart),
                   paste0("WindStart: ", WindStart), ""),
    dplyr::if_else(!is.na(WindEnd),
                   paste0("WindEnd: ", WindEnd), ""),
    dplyr::if_else(!is.na(SkyStart),
                   paste0("SkyStart: ", SkyStart), ""),
    dplyr::if_else(!is.na(SkyEnd),
                   paste0("SkyEnd: ", SkyEnd), ""),
    dplyr::if_else(!is.na(Site),
                   paste0("Site: ", Site), ""),
    dplyr::if_else(!is.na(siteLocality),
                   paste0("siteLocality: ", siteLocality), ""),
    dplyr::if_else(!is.na(syd),
                   paste0("syd: ", syd), ""),
    sep = "|") %>%
      # Remove extra bars "|".
      stringr::str_replace_all(pattern = "(\\|){2,9}",
                               replacement = "\\|") %>%
      stringr::str_replace_all(pattern = "(\\|$)+|(^\\|)+",
                               replacement = "")) %>%
  dplyr::mutate(
    # Format eventDate
    eventDate = lubridate::mdy(eventDate,
                               truncated = 2, quiet = FALSE),
    dataSource = "Gai_Anthophila") %>%
  # add the database_id column
  dplyr::mutate(
    database_id = paste("Gai_data_", 1:nrow(.), sep = ""),
    .before = institutionCode)  %>%
  dplyr::mutate(license = dataLicense) %>%
  # add the database_id column
  dplyr::mutate(
    datasetName = "Gaiarsa et al. 2021",
    datasetID = "Gai"
  )
  #### 6.3 Out ####
# Save the dataset
readr::write_excel_csv(Gai_data, file =  paste(path, outFile, sep = "/"))
# Return data
return(Gai_data)
} # END readr_Gai

#### 7.0 CAES ####
#' @rdname readr_BeeBDC
#' 
#' 
#' 
#'
readr_CAES <- function(path = NULL,
                       inFile = NULL,
                       outFile = NULL,
                       dataLicense = NULL,
                       sheet = "Sheet1"){
  # locally bind variables to the function
  Tribe <- Morphospecies <- Successional_Stage <- genus <- specificEpithet <- NULL
  eventDate <- catalogNumber <- . <- NULL
  .<-PBIUSI<-Family<-Subfamily<-Genus<-species<-Country<-State_Prov<-Sec_Subdiv<-Locality<-Lat<-
    Lon<-Start_Date<-Collector<-Sex<-Inst_Code<-Det_By<-Det_Date<-Coll_Method<-Spec_Count<-
    Elev_m<-Trip_Code<-Project<-Det_History<-Tribe<-Host_Genus<-Host_Common_Name<-
    Host_Relation<-Host_Location<-Loc_Notes<-Lat_Lon_Method<-End_Date<-eventDate<-
    Elev_Det<-Macro_Habitat<-Micro_Habitat<-Pres_Method<-Spec_Notes<-genus<-specificEpithet<-
    Lat_Lon_Accuracy<-Host_species<-Host_Family<-catalogNumber <- NULL
  
    #### 7.1 Prep ####
    # This will load the requireNamespaced packages. These packages may still need to be installed to 
      # R using install.packages("dplyr")... etc.
  requireNamespace("dplyr")
  
  requireNamespace("lubridate")


  #### 7.2 Read+ ####
    # Reads in the .csv file, trims the white spaces, and formats the columns to the correct type
CAES_data <- openxlsx::read.xlsx(paste(path, inFile, sep = "/"), sheet = sheet) %>%
  # Turn spaces into "_" in column names
  dplyr::rename_with(., ~ gsub(" ", "_", .x, fixed = TRUE)) %>%
  # Make columns DarwinCore-compatible
  dplyr::rename(
    catalogNumber = PBIUSI,
    family = Family,
    subfamily = Subfamily,
    genus = Genus,
    specificEpithet = species,
    country = Country,
    stateProvince = State_Prov,
    county = Sec_Subdiv,
    locality = Locality,
    decimalLatitude = Lat,
    decimalLongitude = Lon,
    eventDate = Start_Date,
    recordedBy = Collector,
    sex = Sex,
    institutionCode = Inst_Code,
    identifiedBy = Det_By,
    dateIdentified = Det_Date,
    samplingProtocol = Coll_Method,
    individualCount = Spec_Count,
    elevation = Elev_m,
    otherCatalogNumbers = Trip_Code,
    bibliographicCitation = Project,
    identificationRemarks = Det_History)  %>%
      # Add a bunch of columns from other columns
    dplyr::mutate(
    # Add previousIdentifications
    previousIdentifications = paste0(
        # ONLY do this IF there is something in the cell - is.na() finds "NA" values. the "!" reverses this to find stats::complete.cases only.
      dplyr::if_else(!is.na(Tribe), paste0("Tribe: ", Tribe),"")),
    # Add associatedTaxa
      # This will ONLY concatenate columns where they have a value.
    associatedTaxa = stringr::str_c(
      dplyr::if_else(!is.na(Host_Genus), paste0("Host_id: ", Host_Genus), ""),
      dplyr::if_else(!is.na(Host_Common_Name),
                     paste0("Host_commonName: ", Host_Common_Name), ""),
      dplyr::if_else(!is.na(Host_Relation),
                     paste0("Host_relation: ", Host_Relation), ""),
      dplyr::if_else(!is.na(Host_Location),
                     paste0("Host_location: ", Host_Location), ""),
      sep = "|") %>%
      # Remove extra bars "|".
      stringr::str_replace_all(pattern = "(\\|){2,9}",
                               replacement = "\\|") %>%
      stringr::str_replace_all(pattern = "(\\|$)+|(^\\|)+",
                               replacement = "")) %>%
    # Do the same as the last mutate, but for fieldNotes
  dplyr::mutate(
    fieldNotes = stringr::str_c(
      dplyr::if_else(!is.na(Loc_Notes),
                     paste0("fieldNotes: ", Loc_Notes), ""),
      dplyr::if_else(!is.na(Lat_Lon_Method),
                     paste0("GPS_device: ", Lat_Lon_Method), ""),
      dplyr::if_else(!is.na(End_Date),
                     paste0("Sampling period: ", eventDate, " to ", End_Date), ""),
      dplyr::if_else(!is.na(Elev_Det),
                     paste0("Elevation_by: ", Elev_Det), ""),
      dplyr::if_else(!is.na(Macro_Habitat),
                     paste0("Macro_Habitat: ", Macro_Habitat), ""),
      dplyr::if_else(!is.na(Micro_Habitat),
                     paste0("Micro_Habitat: ", Micro_Habitat), ""),
      dplyr::if_else(!is.na(Pres_Method),
                     paste0("Preservation_method: ", Pres_Method), ""),
      dplyr::if_else(!is.na(Spec_Notes),
                     paste0("Specimen_notes: ", Spec_Notes), ""),
      sep = "|") %>%
      # Remove extra bars "|".
      stringr::str_replace_all(pattern = "(\\|){2,9}",
                               replacement = "\\|") %>%
      stringr::str_replace_all(pattern = "(\\|$)+|(^\\|)+",
                               replacement = "")) %>%
        # Add scientificName
      dplyr::mutate(
        scientificName = stringr::str_c(
          dplyr::if_else(!is.na(genus),
                         genus, ""),
          dplyr::if_else(!is.na(specificEpithet),
                         specificEpithet, ""),
          sep = " ")) %>%
  # Get the coordinateUncertaintyInMeters by taking the UPPER uncertainty limit from the provided 
    # ranges. I.e., "100-1000" becomes "1000"
  dplyr::mutate(
    coordinateUncertaintyInMeters =  stringr::str_extract(
      Lat_Lon_Accuracy, pattern = "-.*|[0-9]+" ),
    # Format eventDate
    eventDate = lubridate::ymd(
        # First convert from silly excel numeric format to real dates...
      eventDate %>%
        as.numeric() %>%
        as.Date(., origin = "1899-12-30"),
      truncated = 2, quiet = FALSE), # 215 failed to parse. 
    dataSource = "CAES_Anthophila") %>%
  # Add the scientificNameAuthorship column using the specificEpithet column
      # dplyr::mutate(scientificNameAuthorship = stringr::str_replace(
      #   string = verbatimScientificName,
      #   pattern = paste0(".*", specificEpithet, " "),
      #   replacement = ""
      # )) %>%
  # Remove those now redundant columns
  dplyr::select(!c(Tribe, Host_Genus, Host_species, Host_Family, Host_Common_Name,
                   Host_Relation, Host_Location, Loc_Notes, Lat_Lon_Method,
                   End_Date, Elev_Det, Macro_Habitat, Micro_Habitat, Pres_Method, Spec_Notes,
                   Lat_Lon_Accuracy)) %>%
  # Remove double white-spaces
  apply(., 2, stringr::str_squish) %>% dplyr::as_tibble() %>% 
  # add the database_id column
  dplyr::mutate(
    database_id = paste("CAES_data_", 1:nrow(.), sep = ""),
    .before = catalogNumber)  %>%
  dplyr::mutate(license = dataLicense) %>%
  # add the database_id column
  dplyr::mutate(
    datasetName = "Connecticut Agricultural Experiment Station",
    datasetID = "CAES"
  )

  #### 7.3 Out ####
# Save the dataset
readr::write_excel_csv(CAES_data, file = paste(path, outFile, sep = "/"))
  # Return the data from the function to the user
return(CAES_data)
} # END readr_CAES






#### 9.0 KP ####
#' @rdname readr_BeeBDC
#' 
#' 
#' 
#'
readr_KP <- function(path = NULL,
                     inFile = NULL,
                     outFile = NULL,
                     dataLicense = NULL){
  # locally bind variables to the function
  .<-ID<-Family<-Subfamily<-Genus<-Det<-Number<-Collection_method<-Collector<-Order<-Suborder<-
    subgenus<-species<-subspecies<-author<-whole_sci_name<-Country<-State<-County_Parish<-
    Location<-Lat<-Long<-decimalLatitude<-decimalLongitude<-Tribe<-sp_group<-Male<-Female<-
    genus<-specificEpithet<-infraspecificEpithet<-Collection_date<-catalogNumber <- NULL
  
  #### 9.1 Prep ####
  # This will load the requireNamespaced packages. These packages may still need to be installed to 
  # R using install.packages("dplyr")... etc.
  requireNamespace("dplyr")
  
  requireNamespace("lubridate")


  
  #### 9.2 Read+ ####
  # Reads in the .xlsx file, trims the white spaces, and formats the columns to the correct type
KP_data <- openxlsx::read.xlsx(paste(path, inFile, sep = "/")) %>%
    # Turn spaces into "_" in column names
    dplyr::rename_with(., ~ gsub(" ", "_", .x, fixed = TRUE)) %>%
    # Make columns DarwinCore-compatible
    dplyr::rename(
      catalogNumber = ID,
      family = Family,
      subfamily = Subfamily,
      genus = Genus,
      identifiedBy = Det,
      individualCount = Number,
      samplingProtocol = Collection_method,
      recordedBy = Collector,
      order = Order,
      suborder = Suborder,
      subgenus = subgenus,
      specificEpithet = species,
      infraspecificEpithet = subspecies,
      scientificNameAuthorship = author,
      verbatimScientificName = whole_sci_name,
      country = Country,
      stateProvince = State,
      county = County_Parish,
      locality = Location,
      decimalLatitude = Lat,
      decimalLongitude = Long)  %>%
    # round the coordinates to six decimal places
  dplyr::mutate(
    decimalLatitude = decimalLatitude %>% as.numeric() %>% base::round(digits = 6) %>%
      suppressWarnings(classes = "warning"),
    decimalLongitude = decimalLongitude %>% as.numeric() %>% base::round(digits = 6) %>%
      suppressWarnings(classes = "warning")) %>%
    # Add a bunch of columns from other columns
  dplyr::mutate(
    previousIdentifications = stringr::str_c(
      dplyr::if_else(!is.na(Tribe),
                     paste0("Tribe: ", Tribe), ""),
      dplyr::if_else(!is.na(sp_group),
                     paste0("sp_group: ", sp_group), ""),
      sep = "|") %>%
      # Remove extra bars "|".
      stringr::str_replace_all(pattern = "(\\|){2,9}",
                               replacement = "\\|") %>%
      stringr::str_replace_all(pattern = "(\\|$)+|(^\\|)+",
                               replacement = "")) %>%
    # Do the same as the last mutate, but for fieldNotes
    dplyr::mutate(
      sex = stringr::str_c(
        dplyr::if_else(!is.na(Male) & Male != 0,
                       paste0(Male, " M"), ""),
        dplyr::if_else(!is.na(Female) & Female != 0,
                       paste0(Female, " F"), ""),
        sep = "|") %>%
        # Remove extra bars "|".
        # Remove extra bars "|".
        stringr::str_replace_all(pattern = "(\\|){2,9}",
                                 replacement = "\\|") %>%
        stringr::str_replace_all(pattern = "(\\|$)+|(^\\|)+",
                                 replacement = "")) %>%
    # Format eventDate and add dataSource
    dplyr::mutate(
      # Create scientificName
      scientificName = stringr::str_c(
        dplyr::if_else(!is.na(genus),
                       paste0(genus), ""),
        dplyr::if_else(!is.na(specificEpithet),
                       paste0(specificEpithet), ""),
        dplyr::if_else(!is.na(infraspecificEpithet),
                       paste0(infraspecificEpithet), ""),
        sep = " ") %>% stringr::str_squish() %>% stringr::str_trim(side = "both"),
      # Format eventDate
      eventDate = lubridate::ymd_hms(Collection_date,
                                 truncated = 5, quiet = FALSE, tz = "UTC"), # 215 failed to parse. 
      dataSource = "KP_Anthophila") %>%
    # Remove those now redundant columns
    dplyr::select(!c(Male, Female, sp_group, Tribe, Collection_date)) %>%
    # Remove double white-spaces
    apply(., 2, stringr::str_squish) %>% dplyr::as_tibble() %>% 
    # add the database_id column
    dplyr::mutate(
      database_id = paste("KP_data_", 1:nrow(.), sep = ""),
      .before = catalogNumber)  %>%
    dplyr::mutate(license = dataLicense) %>%
  # add the database_id column
  dplyr::mutate(
    datasetName = "USDA ARS Southeastern USA",
    datasetID = "KP"
  )
  
  #### 9.3 Out ####
  # Save the dataset
  readr::write_excel_csv(KP_data, file = paste(path, outFile, sep = "/"))
  # Return the data from the function to the user
  return(KP_data)
} # END readr_KP










#### 11.0 EcoS ####
#' @rdname readr_BeeBDC
#' 
#' 
#' 
#'
readr_EcoS <- function(path = NULL,
                     inFile = NULL,
                     outFile = NULL,
                     dataLicense = NULL){
  # locally bind variables to the function
  Species <- . <- scientificName <- Latitude <- Longitude <- Year <- catalogNumber <- NULL
  Collection <- ID_project <- NULL
  
  #### 11.1 Prep ####
  # This will load the requireNamespaced packages. These packages may still need to be installed to 
  # R using install.packages("dplyr")... etc.
  requireNamespace("dplyr")
  
  requireNamespace("lubridate")


  
  #### 11.2 Read+ ####
  # Reads in the .csv file, trims the white spaces, and formats the columns to the correct type
  EcoS_data <- readr::read_csv(paste(path, inFile, sep = "/"),
                                trim_ws = TRUE, guess_max = 33000) %>%
      # Add institution information
    dplyr::mutate(
      institutionCode = Collection,
      datasetName = Collection,
      catalogNumber = ID_project,
      otherCatalogNumbers = stringr::str_c(Collection, ID_project, sep = "_")) %>%
  # Add taxonomic information
    dplyr::mutate(
      scientificName = Species) %>%
      # Split genus and species names
    tidyr::separate(
      data = ., col = scientificName,
      into = c("genus", "specificEpithet"),
               sep = "_", remove = FALSE) %>%
      # Lat Lon
    dplyr::mutate(
      decimalLatitude = Latitude,
      decimalLongitude = Longitude) %>%
      # Year
    dplyr::mutate(
      year = Year) %>%
      # Add dataset information
    dplyr::mutate(dataSource = "EcoS_Anthophila") %>%
    # Remove any double white-spaces
    apply(., 2, stringr::str_squish) %>% dplyr::as_tibble() %>% 
    # add the database_id column
    dplyr::mutate(
      database_id = paste("EcoS_data_", 1:nrow(.), sep = ""),
      .before = catalogNumber)  %>%
    dplyr::mutate(license = dataLicense) %>%
    # keep only valid columns
    dplyr::select( tidyselect::any_of(names(ColTypeR()[[1]])))  %>%
    # add the database_id column
    dplyr::mutate(
      datasetName = "EcoSur",
      datasetID = "EcoS"
    )
    
  
  #### 11.3 Out ####
  # Save the dataset
  readr::write_excel_csv(EcoS_data, file = paste(path, outFile, sep = "/"))
  # Return the data from the function to the user
  return(EcoS_data)
} # END readr_EcoS








#### 12.0 GeoL ####
#' @rdname readr_BeeBDC
#' 
#' 
#' 
#'
readr_GeoL <- function(path = NULL,
                       inFile = NULL,
                       outFile = NULL,
                       dataLicense = NULL){
  # locally bind variables to the function
  geolocate_Latitude<-geolocate_Longitude<-geolocate_UncertaintyRadiusMeters<-database_id<-
    bels_decimallatitude<-bels_decimallongitude<-bels_coordinateuncertaintyinmeters<-datasource<-
    scientificname<-infraspecificepithet<-specificepithet<-acceptednameusage<-taxonrank<-
    scientificnameauthorship<-countrycode<-stateprovince<-eventdate<-basisofrecord<-
    occurrencestatus<-recordnumber<-recordedby<-eventid<-samplingprotocol<-samplingeffort<-
    individualcount<-catalognumber<-rightsholder<-institutioncode<-datasetname<-
    othercatalognumbers<-occurrenceid<-coreid<-recordid<-collectionid<-
    verbatimscientificname<-verbatimeventdate<-id <- . <- NULL
  rightsHolder <- continent <- type <- samplingProtocol <- NULL
  island <- municipality <- verbatimEventDate <- catalogNumber <- NULL
  
  #### 12.1 Prep ####
  # This will load the requireNamespaced packages. These packages may still need to be installed to 
  # R using install.packages("dplyr")... etc.
  requireNamespace("dplyr")
  
  requireNamespace("lubridate")


  
  #### 12.2 Read+ ####
    ###### a. GeoL_high ####
  # Reads in the .xlsx file, trims the white spaces, and formats the columns to the correct type
  GeoL_data <- openxlsx::read.xlsx(paste(path, inFile, sep = "/"),
                               sheet = "GEOLOCATE HIGH") %>%
    # Return spaces in column names to keep the consistent with file before renaming
    stats::setNames(., stringr::str_replace_all(colnames(.), "\\.", " ")) %>%
      # Convert GeoLocate columns into dwc columns
    dplyr::rename(
      decimalLatitude = geolocate_Latitude, decimalLongitude = geolocate_Longitude,
      coordinateUncertaintyInMeters = geolocate_UncertaintyRadiusMeters) %>%
    # keep only valid columns
    dplyr::select( tidyselect::any_of(names(ColTypeR()[[1]]))) %>%
      # Remove blanks
    tidyr::drop_na(database_id) %>%
      # Temporarily add an identifier column
    dplyr::mutate(
      tempSource = "GeoL",
      rightsHolder = rightsHolder  %>% as.character(),
      island = island %>% as.character(),
      municipality = municipality %>% as.character(),
      verbatimEventDate = verbatimEventDate %>% as.character()
    )
  # User output
  writeLines(paste0(
    " - We have read in ", 
    format(nrow(GeoL_data), big.mark = ","), " occurrence records from the 'GEOLOCATE HIGH' sheet." 
  ))
  
  ###### b. BELS_high ####
  # Reads in the .xlsx file, trims the white spaces, and formats the columns to the correct type
  BELS_data <- openxlsx::read.xlsx(paste(path, inFile, sep = "/"),
                                  sheet = "BELS High") %>%
    # Convert GeoLocate columns into dwc columns
    dplyr::rename(
      decimalLatitude = bels_decimallatitude, 
      decimalLongitude = bels_decimallongitude,
      coordinateUncertaintyInMeters = bels_coordinateuncertaintyinmeters,
      dataSource = datasource,
      scientificName = scientificname,
      infraspecificEpithet = infraspecificepithet,
      specificEpithet = "_specificepithet",
      species = specificepithet,
      acceptedNameUsage = acceptednameusage,
      taxonRank = taxonrank,
      scientificNameAuthorship = scientificnameauthorship,
      countryCode = countrycode,
      stateProvince = stateprovince,
      eventDate = eventdate,
      basisOfRecord = basisofrecord,
      occurrenceStatus = occurrencestatus,
      recordNumber = recordnumber,
      recordedBy = recordedby,
      eventID = eventid,
      samplingProtocol = samplingprotocol,
      samplingEffort = samplingeffort,
      individualCount = individualcount,
      catalogNumber = catalognumber,
      rightsHolder = rightsholder,
      institutionCode = institutioncode,
      datasetName = datasetname,
      otherCatalogNumbers = othercatalognumbers,
      occurrenceID = occurrenceid,
      coreid = coreid,
      recordId = recordid,
      collectionID = collectionid,
      verbatimScientificName = verbatimscientificname,
      verbatimEventDate = verbatimeventdate,
      id = id) %>%
      # Correct some formatting
    dplyr::mutate(continent = continent %>% as.character(),
                  type = type %>% as.character(),
                  id = id %>% as.character(),
                  samplingProtocol = samplingProtocol %>% as.character(),
                  island = island %>% as.character(),
                  municipality = municipality %>% as.character(),
                  verbatimEventDate = verbatimEventDate %>% as.character()) %>%
    # keep only valid columns
    dplyr::select( tidyselect::any_of(names(ColTypeR()[[1]]))) %>%
    # Remove blanks
    tidyr::drop_na(database_id) %>%
    # Temporarily add an identifier column
    dplyr::mutate(
      tempSource = "Bels"
    )
  # User output
  writeLines(paste0(
    " - We have read in ", 
    format(nrow(BELS_data), big.mark = ","), " occurrence records from the 'BELS High' sheet." 
  ))
  
  ###### c. merge ####
  GeoL_data <- GeoL_data %>%
      # Remove data that occurs in BELS_data
    dplyr::filter(!database_id %in% BELS_data$database_id) %>%
      # Combine datasets 
    dplyr::bind_rows(BELS_data) %>%
    dplyr::select(!database_id) %>%
    # add the database_id column
    dplyr::mutate(
      database_id = paste("GEOL_data_", 1:nrow(.), sep = ""),
      .before = catalogNumber)
    
    # User output
  writeLines(paste0(
    " - We have kept ", 
    format(sum(GeoL_data$tempSource == "GeoL", na.rm = FALSE), big.mark = ","), 
    " occurrences from GeoLocate, and ",
    format(sum(GeoL_data$tempSource == "Bels", na.rm = FALSE), big.mark = ","),
    " records from BELS (",
    format(nrow(GeoL_data), big.mark = ","),
    " in total). BELS was given preference over GeoLocate"
  ))
    
  #### 12.3 Out ####
  # Save the dataset
  readr::write_excel_csv(GeoL_data, file = paste(path, outFile, sep = "/"))
  # Return the data from the function to the user
  return(GeoL_data)
} # END readr_GeoL









#### 13.0 EaCO ####
#' @rdname readr_BeeBDC
#' 
#' 
#' 
#'
readr_EaCO <- function(path = NULL,
                       inFile = NULL,
                       outFile = NULL,
                       dataLicense = NULL){
  # locally bind variables to the function
  County<-State<-Genus<-genus<-species<-dateRange<-dateSet<-dateCollected<-treatment<-
    trapNumber<-samplingRound<-coordinates<-decimalLatitude<-decimalLongitude<-recordNumber<-
    .<-catalogNumber <- NULL
  
  #### 13.1 Prep ####
  # This will load the requireNamespaced packages. These packages may still need to be installed to 
  # R using install.packages("dplyr")... etc.
  requireNamespace("dplyr")
  
  requireNamespace("lubridate")


  
  #### 13.2 Read+ ####
  # Reads in the .csv file, trims the white spaces, and formats the columns to the correct type
  EaCO_data <- openxlsx::read.xlsx(paste(path, inFile, sep = "/")) %>%
    # Return spaces in column names to keep the consistent with file before renaming
    stats::setNames(., stringr::str_replace_all(colnames(.), "\\.", " ")) %>%
    # Rename columns
    dplyr::rename(
      recordNumber = 'Specimen Number',
      county = County,
      stateProvince = State,
      genus = Genus,
      species = 'Species (if available)',
      dateRange = 'Date Range for collection',
      treatment = 'Treatment type',
      trapNumber = 'Trap number',
      samplingRound = 'Sampling Round',
      coordinates = 'GPS Coordinates of Traps') %>%
    # Drop rows without species names or that aren't bees or aren't useable names
    dplyr::filter(!is.na(genus), !genus == "",
                  !is.na(species), !species == "",
                  !stringr::str_detect(species, pattern = "[0-9]"),
                  !stringr::str_detect(species, pattern = "/"),
                  !stringr::str_detect(tolower(species), pattern = "sp.")) %>%
      # Split dates
    tidyr::separate(
      col = dateRange, sep = "-",
      into = c("dateSet", "dateCollected")) %>%
      # Add year and convert to ymd format
    dplyr::mutate(
      dateSet = stringr::str_c(dateSet, "/2017") %>% lubridate::mdy(truncated = 2),
      dateCollected = stringr::str_c(dateCollected, "/2017") %>% lubridate::mdy(truncated = 2),
      year = 2017) %>%
      # Add scientificName
    dplyr::mutate(scientificName = stringr::str_c(genus, species, sep = " ")) %>%
      # Add field notes
    dplyr::mutate(
      fieldNotes = stringr::str_c(
        dplyr::if_else(!is.na(dateSet), paste0("dateSet: ", dateSet), ""),
        dplyr::if_else(!is.na(dateCollected),
                       paste0("dateCollected: ", dateCollected), ""),
        dplyr::if_else(!is.na(treatment),
                       paste0("treatmentType: ", treatment), ""),
        dplyr::if_else(!is.na(trapNumber),
                       paste0("trapNumber: ", trapNumber), ""),
        dplyr::if_else(!is.na(samplingRound),
                       paste0("samplingRound: ", samplingRound), ""),
        sep = "|") %>%
        # Remove extra bars "|".
        stringr::str_replace_all(pattern = "(\\|){2,9}",
                                 replacement = "\\|") %>%
        stringr::str_replace_all(pattern = "(\\|$)+|(^\\|)+",
                                 replacement = "")) %>%
      # Add samplingEffort
    dplyr::mutate(
      samplingEffort = dateCollected - dateSet
    ) %>%
      # rescue coordinates
    tidyr::separate(
      col = coordinates,
      into = c("decimalLatitude", "decimalLongitude"),
      sep = ",") %>%
    dplyr::mutate(
      decimalLatitude = decimalLatitude %>% stringr::str_remove(pattern = "[a-zA-Z]"),
      decimalLongitude = decimalLongitude %>% stringr::str_remove(pattern = "[a-zA-Z]")) %>%
    dplyr::mutate(
      catalogNumber = stringr::str_c("EastColarado_",recordNumber, sep = "")) %>%
    # Add dataset information
    dplyr::mutate(dataSource = "EaCO_Anthophila") %>%
    # Remove any double white-spaces
    apply(., 2, stringr::str_squish) %>% dplyr::as_tibble() %>% 
    # add the database_id column
    dplyr::mutate(
      database_id = paste("EaCO_data_", 1:nrow(.), sep = ""),
      .before = catalogNumber)  %>%
    dplyr::mutate(license = dataLicense) %>%
    # keep only valid columns
    dplyr::select( tidyselect::any_of(names(ColTypeR()[[1]])))  %>%
    # add the database_id column
    dplyr::mutate(
      datasetName = "Eastern Colorado (Arathi Seshadri)",
      datasetID = "EaCo",
      institutionCode = "USDA ARS"
    )
  
  
  #### 13.3 Out ####
  # Save the dataset
  readr::write_excel_csv(EaCO_data, file = paste(path, outFile, sep = "/"))
  # Return the data from the function to the user
  return(EaCO_data)
} # END readr_EaCO








#### 14.0 MABC ####
#' @rdname readr_BeeBDC
#' 
#' 
#' 
#'
readr_MABC <- function(path = NULL,
                       inFile = NULL,
                       outFile = NULL,
                       dataLicense = NULL,
                       sheet = "Hoja1"){
  # locally bind variables to the function
  genus <- specificEpithet <- collectionSite <- siteCode <- hour <- tribe <- eventDate <- NULL
  . <- catalogNumber <- NULL
  
  #### 14.1 Prep ####
  # This will load the requireNamespaced packages. These packages may still need to be installed to 
  # R using install.packages("dplyr")... etc.
  requireNamespace("dplyr")
  
  requireNamespace("lubridate")


  
  #### 14.2 Read+ ####
  # Reads in the .csv file, trims the white spaces, and formats the columns to the correct type
  MABC_data <- openxlsx::read.xlsx(paste(path, inFile, sep = "/"),
                                  sheet = sheet) %>%
    # Return spaces in column names to keep the consistent with file before renaming
    stats::setNames(., stringr::str_replace_all(colnames(.), "\\.", " ")) %>%
    # Rename columns
    dplyr::rename(
      catalogNumber = 'Ejemplar',
      eventDate = 'Fecha colecta',
      country = paste0('Pa\u00eds'),
      stateProvince = 'Estado/Provincia',
      municipality = 'Municipio',
      locality = 'Localidad',
      samplingProtocol = 'Metodo colecta',
      decimalLatitude = 'Coordenadas Lat',
      decimalLongitude = 'Coordenadas Long',
      verbatimElevation = 'Altitud',
      georeferenceVerificationStatus = paste0('Datos georeferenciaci\u00f3n'),
      recordedBy = 'Colector',
      identifiedBy = 'Identificador',
      family = 'Familia',
      subfamily = 'Subfamilia',
      genus = 'Genero',
      subgenus = 'Subgenero',
      specificEpithet = 'Especie',
      infraspecificEpithet = 'Subespecie',
      species = 'Nombre especie',
      taxonID = paste0('C\u00f3digo especie'),
      sex = 'Sexo',
      # Non-standard fields
      tribe = 'Tribu',
      collectionSite = 'Sitio Colecta',
      siteCode = paste0('C\u00f3digo sitio'),
      hour = 'Hora') %>%
    # Add scientificName
    dplyr::mutate(scientificName = stringr::str_c(genus, specificEpithet, sep = " ")) %>%
    # Add field notes
    dplyr::mutate(
      fieldNotes = stringr::str_c(
        dplyr::if_else(!is.na(collectionSite), paste0("collectionSite: ", collectionSite), ""),
        dplyr::if_else(!is.na(siteCode),
                       paste0("siteCode: ", siteCode), ""),
        dplyr::if_else(!is.na(hour),
                       paste0("time: ", hour), ""),
        sep = "|") %>%
        # Remove extra bars "|".
        stringr::str_replace_all(pattern = "(\\|){2,9}",
                                 replacement = "\\|") %>%
        stringr::str_replace_all(pattern = "(\\|$)+|(^\\|)+",
                                 replacement = "")) %>%
    # Add identificationRemarks
    dplyr::mutate(
      identificationRemarks = stringr::str_c(
        dplyr::if_else(!is.na(tribe), paste0("tribe: ", tribe), ""))) %>%
  # Add year and ensure ymd format
  dplyr::mutate(eventDate = lubridate::dmy(eventDate))  %>%
    # Add dataset information
    dplyr::mutate(dataSource = "MABC_Anthophila") %>%
    # Remove any double white-spaces
    apply(., 2, stringr::str_squish) %>% dplyr::as_tibble() %>% 
    # add the database_id column
    dplyr::mutate(
      database_id = paste("MABC_data_", 1:nrow(.), sep = ""),
      .before = catalogNumber)  %>%
    dplyr::mutate(license = dataLicense) %>%
    # keep only valid columns
    dplyr::select( tidyselect::any_of(names(ColTypeR()[[1]])))  %>%
    # add the database_id column
    dplyr::mutate(
      datasetName = "MABC",
      datasetID = "MABC"
    )
  
  
  #### 14.3 Out ####
  # Save the dataset
  readr::write_excel_csv(MABC_data, file = paste(path, outFile, sep = "/"))
  # Return the data from the function to the user
  return(MABC_data)
} # END readr_MABC






#### 15.0 Col ####
#' @rdname readr_BeeBDC
#' 
#' 
#' 
#'
readr_Col <- function(path = NULL,
                       inFile = NULL,
                       outFile = NULL,
                       dataLicense = NULL,
                      sheet = sheet){
  # locally bind variables to the function
  day <- year <- eventDateInitial <- eventDate <- month2 <- day2 <- eventDate2 <- NULL
  scientificName <- . <- catalogNumber <-month<- NULL
  
  #### 15.1 Prep ####
  # This will load the requireNamespaced packages. These packages may still need to be installed to 
  # R using install.packages("dplyr")... etc.
  requireNamespace("dplyr")
  
  requireNamespace("lubridate")


  
  #### 15.2 Read+ ####
    ###### a. Col_data ####
  # Reads in the .csv file, trims the white spaces, and formats the columns to the correct type
  Col_data <- openxlsx::read.xlsx(paste(path, inFile, sep = "/"),
                                 sheet = sheet) %>%
    # Return spaces in column names to keep the consistent with file before renaming
    stats::setNames(., stringr::str_replace_all(colnames(.), "\\.", " ")) %>%
      # Fix some special cases that already involve "." in the column names
    stats::setNames(., stringr::str_replace_all(colnames(.), "  ", ". ")) %>%
    stats::setNames(., stringr::str_replace_all(colnames(.), " $", ".")) %>%
    # Rename columns
    dplyr::rename(
      catalogNumber = paste0('C\u00f3digo de Barras'),
      recordedBy = 'Colectores [Aggregated]',
      recordedByID = 'Colectores asociados',
      eventDateInitial = paste0('Fecha colecci\u00f3n inicial'),
      order = 'Orden',
      family = 'Familia',
      genus = paste0('G\u00e9nero'),
      specificEpithet = 'Especie',
      scientificNameAuthorship = 'Especie Author',
      typeStatus = 'Tipo',
      identifiedBy = 'Determinador [Formatted]',
      dateIdentified = paste0('Fecha determinaci\u00f3n'),
      country = paste0('Pa\u00eds'),
      stateProvince = 'Departamento',
      municipality = 'Municipio',
      locationRemarks = 'Corregimiento Departamental',
      locality = 'Localidad',
      decimalLatitude = 'Latitud georref. dec.',
      decimalLongitude = 'Longitud georref. dec.',
      scientificName = 'Nombre Completo',
      day = "dia",
      month = "mes",
      year = "Ano"
      # Previous column names:
        # collectionID = 'Colecci\\u00f3n/Guid',
        # collectionCode = 'C\\u00f3digo',
        # occurrenceID = 'Collection Object/GUID',
        # identificationRemarks = 'Observaciones generales',
        # eventID = 'Evento de Recolecci\\u00f3n/Guid',
        # recordedByID = 'Numero de colector',
        # eventDateFinal = 'Fecha colecci\\u00f3n final',
        # class = 'Clase',
        # infraspecificEpithet = 'Subespecie',
        # namePublishedInID = 'Referencia original',
        # identificationID = 'Determinaciones/Guid',
      # identificationQualifier = 'Obs. Determinaci\\u00f3n',
        # locationID = 'LocalityID',
        # geodeticDatum = 'Datum geod\\u00e9sico',
        # coordinateUncertaintyInMeters = 'Precisi\\u00f3n coord. georref.',
        # minimumElevationInMeters = 'Elevaci\\u00f3n m\\u00ednima',
        # maximumElevationInMeters = 'Elevaci\\u00f3n m\\u00e1xima',
        # georeferenceProtocol = 'Protocolo de georreferenciaci\\u00f3n',
        # verbatimLatitude = 'Informaci\\u00f3n geogr\\u00e1fica/Latitude1 literal',
        # verbatimLongitude = 'Longitude1 literal',
        # locationRemarks = 'Observaciones ninf Geogr\\u00e1fica',
        # lifeStage = 'Estado de desarrollo',
        # habitat = 'H\\u00e1bitat',
        # continent = 'Continente',
        # tribe = 'Tribu',
        # subSpeciesAuthor = 'Subespecie Author'
    ) %>%
    # Fix some date issues:
      # a. replace 00 dates with "" to be picked u pby truncated
    dplyr::mutate(
      day = dplyr::if_else(day == "0" | day == "00",
                           "", day) %>%
        stringr::str_remove("to.*|^0") %>%
        as.numeric(na.rm = TRUE),
      month = dplyr::if_else(month == "0" | month == "00",
                           "", month) %>%
        stringr::str_remove("^0") %>%
        as.numeric(na.rm = TRUE),
      year = year %>% as.numeric()
    ) %>%
    dplyr::mutate(
      eventDate = lubridate::ymd(paste(year, month, day, sep = "/"), 
                                 truncated = 2),
      .after = eventDateInitial) %>%
    # b. Fix inverted day-month
  dplyr::mutate(
    day2 = dplyr::if_else(is.na(eventDate),
                         month,
                         1),
    month2 = dplyr::if_else(is.na(eventDate),
                          day,
                          1),
    eventDate2 = dplyr::if_else(is.na(eventDate),
      lubridate::ymd(paste(year, month2, day2, sep = "/"), 
                                truncated = 2),
      eventDate),
    .after = year
  ) %>%
    # Combine into final date values
    dplyr::mutate(
        # Take eventDate if it's not empty, and eventDate2 if it is empty
      eventDate = dplyr::if_else(is.na(eventDate),
                                 eventDate2,
                                 eventDate),
      day = dplyr::if_else(is.na(day),
                           day2,
                            day),
      month = dplyr::if_else(is.na(month),
                             month2,
                             month)) %>%
    dplyr::select(!c(eventDate2, day2, month2)) %>%
    # Remvoe extra spaces in subgenus
    dplyr::mutate(scientificName = scientificName %>%
                    stringr::str_replace(pattern = "\\( ",
                                         replacement = "\\(")) %>%
    # Remove any double white-spaces
    apply(., 2, stringr::str_squish) %>% dplyr::as_tibble() %>% 
    # add the database_id column
    dplyr::mutate(
      database_id = paste("Col_data_", 1:nrow(.), sep = ""),
      .before = catalogNumber)  %>%
    # Add dataset information
    dplyr::mutate(dataSource = "Col_Anthophila") %>%
    dplyr::mutate(license = dataLicense)  %>%
    # add the database_id column
    dplyr::mutate(
      datasetName = "Colombia - Diego Alexander Guevara Farias",
      datasetID = "Col"
    )
  
  
  #### 15.3 Out ####
  # Save the dataset
  readr::write_excel_csv(Col_data, file = paste(path, outFile, sep = "/"))
  # Return the data from the function to the user
  return(Col_data)
} # END readr_Col






#### 16.0 FSCA ####
#' @rdname readr_BeeBDC
#' 
#' 
#' 
#'
readr_FSCA <- function(path = NULL,
                       inFile = NULL,
                       outFile = NULL,
                       dataLicense = NULL){
  # locally bind variables to the function
  . <- catalogNumber <- recordID <- NULL
  
  #### 16.1 Prep ####
  # This will load the requireNamespaced packages. These packages may still need to be installed to 
  # R using install.packages("dplyr")... etc.
  requireNamespace("dplyr")
  
  requireNamespace("lubridate")


  
  #### 16.2 Read+ ####
  # Reads in the .csv file, trims the white spaces, and formats the columns to the correct type
  FSCA_data <- readr::read_csv(paste(path, inFile, sep = "/"),
                                  trim_ws = TRUE) %>%
    # Add dataset information
    dplyr::mutate(dataSource = "FSCA_Anthophila") %>%
    # Remove any double white-spaces
    apply(., 2, stringr::str_squish) %>% dplyr::as_tibble() %>% 
    # add the database_id column
    dplyr::mutate(
      database_id = paste("FSCA_data_", 1:nrow(.), sep = ""),
      .before = 1)  %>%
    dplyr::mutate(license = dataLicense) %>%
    # keep only valid columns
    dplyr::select( tidyselect::any_of(names(ColTypeR()[[1]]))) %>%
    # add the database_id column
    dplyr::mutate(
      datasetName = "Florida State Collection of Arthropods",
      datasetID = "FSCA",
      institutionCode = "FSCA"
    )
  
  
  #### 16.3 Out ####
  # Save the dataset
  readr::write_excel_csv(FSCA_data, file = paste(path, outFile, sep = "/"))
  # Return the data from the function to the user
  return(FSCA_data)
} # END readr_FSCA




#### 17.0 SMC ####
#' @rdname readr_BeeBDC
#' 
#' 
#' 
#'
readr_SMC <- function(path = NULL,
                      inFile = NULL,
                      outFile = NULL,
                      dataLicense = NULL){
  # locally bind variables to the function
  collectionMethod <- locale <- latitude <- longitude <- organismName <- scientificName <- NULL
  observationDate <- eventDate <- . <- NULL
  
  #### 17.1 Prep ####
  requireNamespace("dplyr")
  
  requireNamespace("lubridate")


  #### 17.2 Read+ ####
  SMC_Data <- readr::read_csv(paste(path, inFile, sep = "/"),
                              trim_ws = TRUE, guess_max = 33000) %>%
    # Rename some columns to make format consistent with DarwinCore
    dplyr::rename(
      samplingProtocol = collectionMethod,
      locality = locale,
      decimalLatitude =	latitude,
      decimalLongitude = longitude,
      scientificName = organismName) %>%
    # Make new columns
    dplyr::mutate(
      # Remove underscore from scientificName strings
      scientificName = gsub("_", " ", scientificName),
      # Add basis of record
      basisOfRecord =	"specimen",
      # Add eventDate
      eventDate = lubridate::mdy(observationDate),
      # Parse eventDate into day, month, and year
      month = lubridate::month(eventDate),
      day = lubridate::day(eventDate),
      year = lubridate::year(eventDate)) %>%
    # Add dataSource information
    dplyr::mutate(dataSource = "SMC_Anthophila") %>%
    # add the database_id column
    dplyr::mutate(
      database_id = paste("SMC_data_", 1:nrow(.), sep = "")) %>%
    dplyr::mutate(license = dataLicense) %>%
    # add the database_id column
    dplyr::mutate(
      datasetName = "Texas SMC literature data",
      datasetID = "SMC"
    )
  # Save the dataset
  readr::write_excel_csv(SMC_Data, file = paste(path, outFile, sep = "/"))
  # Return data
  return(SMC_Data)
} # END readr_SMC




#### 18.0 Bal ####
#' @rdname readr_BeeBDC
#' 
#' 
#' 
#'
readr_Bal <- function(path = NULL,
                      inFile = NULL,
                      outFile = NULL,
                      dataLicense = NULL,
                      sheet = "animal_data"){
  # locally bind variables to the function
  siteID<-year<-animalID<-abundance<-samplingMethod<-censusType<-decimalLatitude<-
    decimalLongitude<-studyLocation<-siteDescription<-studyID<-locationID<-.<-
    samplingIntensity<-eventDate<-catalogNumber <- NULL
  
  #### 18.1 Prep ####
  # This will load the requireNamespaced packages. These packages may still need to be installed to 
  # R using install.packages("dplyr")... etc.
  requireNamespace("dplyr")
  
  requireNamespace("lubridate")


  
  #### 18.2 Read+ ####
  # Reads in the .xlsx file, trims the white spaces, and formats the columns to the correct type
  Bal_data <- openxlsx::read.xlsx(paste(path, inFile, sep = "/"),
                                 sheet = sheet, startRow = 2) %>%
    # Return spaces in column names to keep the consistent with file before renaming
    stats::setNames(., stringr::str_replace_all(colnames(.), "\\.", " ")) %>%
    # Make columns DarwinCore-compatible
    dplyr::rename(
      locationID = siteID,
      year = year,
      eventDate = date,
      scientificName = animalID,
      individualCount = abundance,
      samplingProtocol = samplingMethod,
      fieldNotes = censusType,
      decimalLatitude = decimalLatitude,
      decimalLongitude = decimalLongitude,
      locality = studyLocation,
      locationRemarks = siteDescription) %>%
    # Add some columns
    dplyr::mutate(
      catalogNumber = stringr::str_c(studyID, "_", locationID,"_", 1:nrow(.), sep = ""),
      samplingEffort = stringr::str_c(samplingIntensity, " hours")
    ) %>%
    # round the coordinates to six decimal places
    dplyr::mutate(
      decimalLatitude = decimalLatitude %>% as.numeric() %>% base::round(digits = 6) %>%
        suppressWarnings(classes = "warning"),
      decimalLongitude = decimalLongitude %>% as.numeric() %>% base::round(digits = 6) %>%
        suppressWarnings(classes = "warning")) %>%
    # Format eventDate and add dataSource
    # Format eventDate
    dplyr::mutate(
      eventDate = lubridate::dmy(eventDate,
                                 truncated = 2, quiet = FALSE),
      # Parse eventDate into day, month, and year
      month = lubridate::month(eventDate),
      day = lubridate::day(eventDate),
      year = lubridate::year(eventDate),
      dataSource = "Bal_Anthophila") %>%
    # add the database_id column
    dplyr::mutate(
      database_id = paste("Bal_data_", 1:nrow(.), sep = ""),
      .before = catalogNumber)  %>%
    dplyr::mutate(license = dataLicense) %>%
    # add the database_id column
    dplyr::mutate(
      datasetName = "Ballare et al. 2019",
      datasetID = "Bal"
    )
  
  #### 18.3 Out ####
  # Save the dataset
  readr::write_excel_csv(Bal_data, file = paste(path, outFile, sep = "/"))
  # Return the data from the function to the user
  return(Bal_data)
} # END readr_Bal







#### 19.0 Lic ####
#' @rdname readr_BeeBDC
#' 
#' 
#' 
#'
readr_Lic <- function(path = NULL,
                      inFile = NULL,
                      outFile = NULL,
                      dataLicense = NULL){
  # locally bind variables to the function
  Kingdom<-Order<-Family_or_grp<-Genus<-Species<-sex<-Collector<-Determiner<-genus<-species<-
    occurrenceID<-eventID<-eventDate<-.<-catalogNumber<-family <- NULL
  
  #### 19.1 Prep ####
  # This will load the requireNamespaced packages. These packages may still need to be installed to 
  # R using install.packages("dplyr")... etc.
  requireNamespace("dplyr")
  
  requireNamespace("lubridate")


  
  #### 19.2 Read+ ####
  # Reads in the .csv file, trims the white spaces, and formats the columns to the correct type
  Lic_data <- readr::read_csv(paste(path, inFile, sep = "/"),
                              trim_ws = TRUE, guess_max = 33000) %>%
    dplyr::rename(
      kingdom = Kingdom,
      order = Order,
      family = Family_or_grp,
      genus = Genus,
      species = Species,
      sex = sex,
      recordedBy = Collector,
      identifiedBy = Determiner) %>%
    # Add taxonomic information
    dplyr::mutate(
      scientificName = stringr::str_c(
        dplyr::if_else(!is.na(genus),
                       paste0(genus), ""),
        dplyr::if_else(!is.na(species),
                       paste0(species), ""),
        sep = " ")) %>%
    # Make a catalogue number
    dplyr::mutate(
      catalogNumber = stringr::str_c(
        dplyr::if_else(!is.na(occurrenceID),
                       paste0(occurrenceID), ""),
        dplyr::if_else(!is.na(eventID),
                       paste0(eventID), ""),
        sep = "_")) %>%
    # Format eventDate
    dplyr::mutate(
      eventDate = lubridate::dmy(eventDate,
                                 truncated = 2, quiet = FALSE),
      # Parse eventDate into day, month, and year
      month = lubridate::month(eventDate),
      day = lubridate::day(eventDate),
      year = lubridate::year(eventDate)) %>% 
    # Add dataset information
    dplyr::mutate(dataSource = "Lic_Anthophila") %>%
    # Remove any double white-spaces
    apply(., 2, stringr::str_squish) %>% dplyr::as_tibble() %>% 
    # add the database_id column
    dplyr::mutate(
      database_id = paste("Lic_data_", 1:nrow(.), sep = ""),
      .before = catalogNumber)  %>%
    dplyr::mutate(license = dataLicense) %>%
    # add the database_id column
    dplyr::mutate(
      datasetName = "Elinor Lichtenberg Canola Data",
      datasetID = "Lic"
    )
  
  # filter to bee families only
    Lic_data <- Lic_data %>%
      dplyr::filter(tolower(family) %in% 
                      tolower(c("Andrenidae","Apidae","Colletidae","Halictidae","Megachilidae",
                                "Melittidae","Stenotritidae")))
  
  
  #### 19.3 Out ####
  # Save the dataset
  readr::write_excel_csv(Lic_data, file = paste(path, outFile, sep = "/"))
  # Return the data from the function to the user
  return(Lic_data)
} # END readr_Lic






#### 20.0 Arm ####
#' @rdname readr_BeeBDC
#' 
#' 
#' 
#'
readr_Arm <- function(path = NULL,
                       inFile = NULL,
                       outFile = NULL,
                       dataLicense = NULL,
                       sheet = "Sheet1"){
  # locally bind variables to the function
  fam<-genus<-sp<-species<-sex<-locality<-munic<-state<-y<-x<-elev<-specificEpithet<-ecoregion<-
    veget<-g<-m<-s<-G<-M<-S<-day<-year<-.<-family <- month<-NULL
  
  #### 20.1 Prep ####
  # This will load the requireNamespaced packages. These packages may still need to be installed to 
  # R using install.packages("dplyr")... etc.
  requireNamespace("dplyr")
  
  requireNamespace("lubridate")


  
  #### 20.2 Read+ ####
  # Reads in the .csv file, trims the white spaces, and formats the columns to the correct type
  Arm_data <- openxlsx::read.xlsx(paste(path, inFile, sep = "/"), sheet = sheet) %>%
    # Make columns DarwinCore-compatible
    dplyr::rename(
      family = fam,
      genus = genus,
      specificEpithet = sp,
      species = species,
      sex = sex,
      locality = locality,
      municipality = munic,
      stateProvince = state,
      decimalLatitude = y,
      decimalLongitude = x,
      verbatimElevation = elev)  %>%
    # Add a bunch of columns from other columns
    dplyr::mutate(
      # Add scientificName
      # This will ONLY concatenate columns where they have a value.
      scientificName = stringr::str_c(
        dplyr::if_else(!is.na(genus), genus, ""),
        dplyr::if_else(!is.na(specificEpithet),
                       specificEpithet, ""),
        sep = " ") ) %>%
    # Do the same as the last mutate, but for fieldNotes
    dplyr::mutate(
      fieldNotes = stringr::str_c(
        dplyr::if_else(!is.na(ecoregion),
                       paste0("ecoregion: ", ecoregion), ""),
        dplyr::if_else(!is.na(veget),
                       paste0("vegetation: ", veget), ""),
        sep = "|") %>%
        # Remove extra bars "|".
        stringr::str_replace_all(pattern = "(\\|){2,9}",
                                 replacement = "\\|") %>%
        stringr::str_replace_all(pattern = "(\\|$)+|(^\\|)+",
                                 replacement = "")) %>%
    # Add scientificName
    dplyr::mutate(
      verbatimLatitude = stringr::str_c(
        dplyr::if_else(!is.na(g),
                       as.character(g), ""),
        dplyr::if_else(!is.na(m),
                       as.character(m), ""),
        
        dplyr::if_else(!is.na(s),
                       as.character(s), ""),
        sep = " "),
      verbatimLongitude = stringr::str_c(
        dplyr::if_else(!is.na(G),
                       as.character(G), ""),
        dplyr::if_else(!is.na(M),
                       as.character(M), ""),
        dplyr::if_else(!is.na(S),
                       as.character(S), ""),
        sep = " ")) %>%
    # Get the coordinateUncertaintyInMeters by taking the UPPER uncertainty limit from the provided 
    # ranges. I.e., "100-1000" becomes "1000"
    dplyr::mutate(
      # Format eventDate
      eventDate = lubridate::dmy(paste(
        day, month, year, sep = "-"
      )), # 215 failed to parse. 
      dataSource = "Arm_Anthophila") %>%
    # Remove double white-spaces
    apply(., 2, stringr::str_squish) %>% dplyr::as_tibble() %>% 
    # add the database_id column
    dplyr::mutate(
      database_id = paste("Arm_data_", 1:nrow(.), sep = ""),
      .before = family)  %>%
    dplyr::mutate(license = dataLicense) %>%
      # Remove spent columns
    dplyr::select(!tidyselect::any_of(c("veget", "ecoregion", "g", "m", "s", "G", "M", "S"))) %>%
    # add the database_id column
    dplyr::mutate(
      datasetName = "Armando Falcon-Brindis",
      datasetID = "Arm"
    )
  
  #### 20.3 Out ####
  # Save the dataset
  readr::write_excel_csv(Arm_data, file = paste(path, outFile, sep = "/"))
  # Return the data from the function to the user
  return(Arm_data)
} # END readr_Arm








#### 21.0 Dorey ####
#' @rdname readr_BeeBDC
#' 
#' 
#' 
#'
readr_Dor <- function(path = NULL,
                      inFile = NULL,
                      outFile = NULL,
                      dataLicense = NULL){
  # locally bind variables to the function
  . <- catalogNumber <- eventDate <- stateOrProvince <- NULL
  
  #### 21.1 Prep ####
  requireNamespace("dplyr")
  
  requireNamespace("lubridate")


  
  #### 21.2 Read+ ####
  Dor_data <- readr::read_csv(paste(path, inFile, sep = "/"),
                              trim_ws = TRUE) %>%
    dplyr::rename(stateProvince = stateOrProvince) %>% 
    # add the database_id column
    dplyr::mutate(
      database_id = paste("Dor_data_", 1:nrow(.), sep = ""),
      .before = catalogNumber) %>%
    # Add dataset information
    dplyr::mutate(dataSource = "Dor_Anthophila") %>%
    # Format date
    dplyr::mutate(eventDate = eventDate %>% lubridate::dmy(., truncated = 2),
                  year = lubridate::year(eventDate),
                  month = lubridate::month(eventDate),
                  day = lubridate::day(eventDate)) %>%
    # Pick up dates of different formats and format together.
    dplyr::mutate(license = dataLicense) %>%
    # add the database_id column
    dplyr::mutate(
      datasetName = "James B Dorey Bee Data",
      datasetID = "Dorey"
    )
  #### 21.3 Out ####
  # Save the dataset
  readr::write_excel_csv(Dor_data, file = paste(path, outFile, sep = "/"))
  # Return data
  return(Dor_data)
} # END readr_Dor



#### 22.0 MEPB ####
#' @rdname readr_BeeBDC
#' 
#' 
#' 
#'
readr_MEPB <- function(path = NULL,
                       inFile = NULL,
                       outFile = NULL,
                       dataLicense = NULL,
                       sheet = NULL){
  # locally bind variables to the function
  catalog_number<-pollinator_family<-pollinator_genus<-pollinator_species<-collection_method<-
    day_collected<-month_collected<-year_collected<-location_description<-latitude<-longitude<-
    basis_of_record<-genus<-specificEpithet<-year<-day<-eventDate<-collector_number<-
    location_name<-habitat<-.<-catalogNumber <- NULL
  
  #### 22.1 Prep ####
  requireNamespace("dplyr")
  
  requireNamespace("lubridate")


  
  #### 22.2 Read+ ####
  MEPB_data <- openxlsx::read.xlsx(paste(path, inFile, sep = "/"), sheet = sheet)  %>%
    # Return spaces in column names to keep the consistent with file before renaming
    stats::setNames(., stringr::str_replace_all(colnames(.), "\\.", " ")) %>%
    # Fix broken encodings
    dplyr:: mutate(
      dplyr::across(
        .cols = dplyr::everything(),
        .fns = ~ stringr::str_replace_all(.,
                                 pattern = c("\\u00c3\\u00b3"="\\u00f3",  
                                             "\\u00c3\\u00a9"="\\u00e9", 
                                             "\\u00c3\\u00b1"="\\u00f1",
                                             "\\u00c2\\u00b0"="\\u00b0",
                                             "\\u00c3"="\\u00e1") 
        ))) %>%
    # add the database_id column
    dplyr::mutate(
      database_id = paste("MEPB_data_", 1:nrow(.), sep = ""),
      .before = catalogNumber) %>%
    # Add dataset information
    dplyr::mutate(dataSource = "MEPB_Anthophila") %>%
    # Format date
    dplyr::mutate(eventDate = lubridate::ymd(stringr::str_c(year %>% as.numeric(), 
                                                            month %>% as.numeric(), 
                                                            day %>% as.numeric(), 
                                                            sep = "/"),
                                             truncated = 2)) %>%
    # Pick up dates of different formats and format together.
    dplyr::mutate(license = dataLicense) %>%
    # add the database_id column
    dplyr::mutate(
      datasetID = "MEPB"
    )
  #### 22.3 Out ####
  # Save the dataset
  readr::write_excel_csv(MEPB_data, file = paste(path, outFile, sep = "/"))
  # Return data
  return(MEPB_data)
} # END readr_MEPB




#### 23.0 Brazil ####
#' @rdname readr_BeeBDC
#' 
#' 
#' 
#'
readr_BBD <- function(path = NULL,
                      inFile = NULL,
                      outFile = NULL,
                      dataLicense = NULL){
  # locally bind variables to the function
  . <- catalogNumber <- year <- day <- dateLastModified <- dateLastModified2 <- NULL
  identifiedBy <- Spcslink.identifiedby <-month <- NULL
  
  #### 23.1 Prep ####
  requireNamespace("dplyr")
  
  requireNamespace("lubridate")


  
  #### 23.2 Read+ ####
  lookupCols <- c(verbatimScientificName = "Scientificname_ORIGINAL",
                  verbatimLatitude = "Lat_original",
                  verbatimLongitude = "Long_original")
  BBD_data <- readr::read_csv(paste(path, inFile, sep = "/"),
                              trim_ws = TRUE) %>%
    # Rename columns
    dplyr::rename(
      tidyselect::any_of(lookupCols),
      id = "CodeBBdatabase_curated",
      scientificName = "Scientific name corrected",
      family = "Family",
      institutionCode = "institutioncode",
      # infraspecificEpithet = "Spcslink.subspecies",
      # scientificNameAuthorship = "Spcslink.scientificnameauthor",
      day = "Day",
      month = "Month",
      year = "Year",
      country = "Country",
      stateProvince = "State",
      decimalLatitude = "Latitude_dec.degrees",
      decimalLongitude = "Longitude_dec.degrees",
      coordinateUncertaintyInMeters = "Precision.of.coord.meters",
      verbatimLocality = "Locality.original",
      georeferenceRemarks = "NotasLatLong",
      county = "Spcslink.county", 
      recordedBy = "Collector",
      collectionCode = "Collection",
      references = "Source",
      sex = "Sex",
      identifiedBy = "Det_By",
      catalogNumber = "Codigo",
      otherCatalogNumbers = "Spcslink.collectioncode",
      dateIdentified = "Spcslink.yearidentified",
      dateLastModified = "Spcslink.datelastmodified",
      basisOfRecord = "Spcslink.basisofrecord",
      # occurrenceID = "Spcslink.collectioncode",
      eventRemarks = "NotesOnLocality") %>%
    #dplyr::mutate(
    #  decimalLatitude = decimalLatitude %>% stringr::str_replace("\\,", "\\."),
    #  decimalLongitude = decimalLongitude %>% stringr::str_replace("\\,", "\\."))
    # add the database_id column
    dplyr::mutate(
      database_id = paste("BBD_data_", 1:nrow(.), sep = ""),
      .before = catalogNumber) %>%
    # Edit catalogNumber for AMNH
    dplyr::mutate(catalogNumber = stringr::str_replace(
      catalogNumber, pattern = "AMNHBEE", replacement = "AMNH_BEE"),
      catalogNumber = stringr::str_replace(
        catalogNumber, pattern = "^0\\.", replacement = ".")) %>%
    # Add dataset information
    dplyr::mutate(dataSource = "BBD_Anthophila") %>%
    # Format date
    dplyr::mutate(eventDate = lubridate::ymd(paste0(
      year, month, day, sep = "/"), truncated = 2, quiet = TRUE)) %>%
    dplyr::mutate(dateLastModified2 = 
                    lubridate::dmy_hms(dateLastModified, truncated = 5, quiet = TRUE),
      .after = dateLastModified) %>%
    dplyr::mutate(dateLastModified2 = dplyr::if_else(is.na(dateLastModified2),
                                                         lubridate::ymd_hms(dateLastModified, truncated = 5, quiet = TRUE),
                                                     dateLastModified2)) %>%
    dplyr::mutate(dateLastModified2 = dplyr::if_else(is.na(dateLastModified2),
                                                   lubridate::mdy_hms(dateLastModified, truncated = 5, quiet = TRUE),
                                                   dateLastModified2)) %>%
    dplyr::mutate(dateLastModified2 = dplyr::if_else(is.na(dateLastModified2),
                                                   lubridate::ymd(dateLastModified, truncated = 2, quiet = TRUE),
                                                   dateLastModified2)) %>%
    dplyr::mutate(dateLastModified2 = dplyr::if_else(is.na(dateLastModified2),
                                                   as.Date(dateLastModified %>% as.numeric(na.rm = TRUE),
                                                           origin = "1899-12-30") %>%
                                                     lubridate::ymd_hms(., truncated = 5, quiet = TRUE),
                                                   dateLastModified2),
                                                   .after = dateLastModified) %>%
    dplyr::mutate(dateLastModified = dateLastModified2) %>%
    dplyr::select(!dateLastModified2) %>%
    # Pick up dates of different formats and format together.
    dplyr::mutate(license = dataLicense) %>%
    # add the database_id column
    dplyr::mutate(
      datasetName = "BBD_Brazil",
      datasetID = "BBD"
    ) %>%
      # If identifiedBy is not filled where it should be, try the other column rpovided
    dplyr::mutate(identifiedBy = dplyr::if_else(is.na(identifiedBy),
                                                Spcslink.identifiedby,
                                                identifiedBy))
  
  
  #### 23.3 Out ####
  # Save the dataset
  readr::write_excel_csv(BBD_data, file = paste(path, outFile, sep = "/"))
  # Return data
  return(BBD_data)
} # END readr_BBD




#### 24.0 MPUJ ####
#' @rdname readr_BeeBDC
#' 
#' 
#' 
#'
readr_MPUJ <- function(path = NULL,
                      inFile = NULL,
                      outFile = NULL,
                      dataLicense = NULL,
                      sheet = sheet){
  # locally bind variables to the function
  collector1stName <- collectorsLastName <- determined1stName <- determinedLastName <- NULL
  day <- year <- eventDate <- endDate <- . <- catalogNumber <- fieldNotes <- NULL
  `Start Date (Year)` <- `Start Date (Month)` <- `Start Date (Day)` <- month<-NULL
  
  #### 24.1 Prep ####
  # This will load the requireNamespaced packages. These packages may still need to be installed to 
  # R using install.packages("dplyr")... etc.
  requireNamespace("dplyr")
  requireNamespace("lubridate")


  
  #### 24.2 Read+ ####
  ###### a. MPUJ_data ####
  # Reads in the .csv file, trims the white spaces, and formats the columns to the correct type
  MPUJ_data <- openxlsx::read.xlsx(paste(path, inFile, sep = "/"),
                                 sheet = sheet) %>%
      # Return spaces in column names to keep the consistent with file before renaming
    stats::setNames(., stringr::str_replace_all(colnames(.), "\\.", " ")) %>%
    # Rename columns
    dplyr::rename(
      catalogNumber = 'Catalog Number',
      individualCount = "Count",
      otherCatalogNumbers = "Alt Cat Number",
      typeStatus = "Type Status",
      identificationQualifier = "Qualifier",
      sex = "Sex",
      samplingProtocol = "Method",
      habitat = "Habitat",
      continent = "Continent",
      country = "Country",
      stateProvince = "State",
      county = "County",
      locality = "Locality Name",
      minimumElevationInMeters = "Min Elevation",
      maximumElevationInMeters = "Max Elevation",
      fieldNotes = "Locality and Habitat Notes",
      decimalLongitude = "Latitude1",
      decimalLatitude = "Longitude1",
      verbatimLatitude = "Lat1text",
      verbatimLongitude = "Long1text",
      scientificName = "Full Name",
      kingdom = "Kingdom",
      order = "Order",
      family = "Family",
      subfamily = "Subfamily",
      genus = "Genus",
      specificEpithet = "Species",
      infraspecificEpithet = "Subspecies",
      scientificNameAuthorship = "Species Author",
      day = `Start Date (Day)`,
      month = `Start Date (Month)`,
      year = `Start Date (Year)`,
      associatedTaxa = "Associated Taxa",
      associatedOccurrences = "Associated Ocurrence",
      lifeStage = "Stage",
      collector1stName = 'Collectors/First Name',
      collectorsLastName = 'Collectors/Last Name',
      determined1stName = 'Determiner/First Name',
      determinedLastName = 'Determiner/Last Name',
      endDate = "End Date",
      verbatimEventDate = "Verbatim Date"
      ) %>%
    dplyr::mutate(year = year %>% as.numeric(),
                  month = month %>% as.numeric(),
                  day = day %>% as.numeric()) %>%
    dplyr::mutate(recordedBy = stringr::str_c(collector1stName, collectorsLastName,
                                              sep = " "),
                  identifiedBy = stringr::str_c(determined1stName, determinedLastName,
                                              sep = " ")) %>%
    dplyr::mutate(eventDate = lubridate::dmy(stringr::str_c(day, month, year, sep = "/"), 
                                             truncated = 2)) %>%
    dplyr::mutate(
      fieldNotes = 
        stringr::str_c(
          dplyr::if_else(!is.na(fieldNotes),
                         paste0("fieldNotes: ", fieldNotes), ""),
        dplyr::if_else(!is.na(eventDate),
                       paste0("startDate: ", eventDate), ""),
        dplyr::if_else(!is.na(endDate),
                       paste0("endDate: ", endDate), ""),
        sep = "|")
    ) %>%
    dplyr::mutate(basisOfRecord = "Preserved specimen") %>%
    # Remove any double white-spaces
    apply(., 2, stringr::str_squish) %>% dplyr::as_tibble() %>% 
    # add the database_id column
    dplyr::mutate(
      database_id = paste("MPUJ_data_", 1:nrow(.), sep = ""),
      .before = catalogNumber)  %>%
    # Add dataset information
    dplyr::mutate(dataSource = "MPUJ_Anthophila") %>%
    dplyr::mutate(license = dataLicense)  %>%
    # add the database_id column
    dplyr::mutate(
      datasetName = "Colombia MPUJ - Diego Alexander Guevara Farias",
      datasetID = "MPUJ"
    )
  
  #### 24.3 Out ####
  # Save the dataset
  readr::write_excel_csv(MPUJ_data, file = paste(path, outFile, sep = "/"))
  # Return the data from the function to the user
  return(MPUJ_data)
} # END readr_MPUJ



#### 25.0 STRI ####
#' @rdname readr_BeeBDC
#' 
#' 
#' 
#'
#' 
readr_STRI <- function(path = NULL,
                       inFile = NULL,
                       outFile = NULL,
                       dataLicense = NULL){
  # locally bind variables to the function
  fieldNotes <- Catalognumber <- . <- day <- year <- catalogNumber <-  recordId <-month<- NULL
  
  #### 25.1 Prep ####
  # This will load the requireNamespaced packages. These packages may still need to be installed to 
  # R using install.packages("dplyr")... etc.
  requireNamespace("dplyr")
  
  requireNamespace("lubridate")

  
  
  #### 25.2 Read+ ####
  ###### a. STRI_data ####
  # Reads in the .csv file, trims the white spaces, and formats the columns to the correct type
  STRI_data <- readr::read_csv(paste(path, inFile, sep = "/"),
                                  trim_ws = TRUE) %>% 
    dplyr::rename(recordId = "recordID") %>%
    # Rename columns
    dplyr::mutate(
      fieldNotes = stringr::str_c(
        dplyr::if_else(!is.na(Catalognumber),
                       paste0("secondary catalog #: ", Catalognumber), ""),
                       sep = "|"),
      species = "scientificName") %>%
      # Format dates
    dplyr::mutate(
        # If day is not recorded, set to first day of month
      day = dplyr::if_else(day == 0, 1, day),
      eventDate = lubridate::dmy(stringr::str_c(day, month, year, sep = "/"), 
                                             truncated = 2)) %>%
    # Remove any double white-spaces
    apply(., 2, stringr::str_squish) %>% dplyr::as_tibble() %>% 
    # add the database_id column
    dplyr::mutate(
      database_id = paste("STRI_data_", 1:nrow(.), sep = ""),
      .before = catalogNumber)  %>%
    # Add dataset information
    dplyr::mutate(dataSource = "STRI_Anthophila") %>%
    dplyr::mutate(license = dataLicense)  %>%
    # add the database_id column
    dplyr::mutate(
      datasetName = "STRI",
      datasetID = "STRI"
    )
  
  #### 25.3 Out ####
  # Save the dataset
  readr::write_excel_csv(STRI_data, file = paste(path, outFile, sep = "/"))
  # Return the data from the function to the user
  return(STRI_data)
} # END readr_STRI




#### 26.0 PALA ####
#' @rdname readr_BeeBDC
#' 
#' 
#' 
#'
#' 
readr_PALA <- function(path = NULL,
                       inFile = NULL,
                       outFile = NULL,
                       dataLicense = NULL){
  # locally bind variables to the function
  romanNumerals <- numeralConversion <- catalogNumber <- type <- country <- municipality <- NULL
  locality <- decimalLatitude <- decimalLongitude <- verbatimElevation <- verbatimEventDate <- NULL
  recordedBy <- collectionCode <- otherCatalogNumbers <- associatedTaxa <- taxonRemarks <- NULL
  family <- genus <- references <- specificEpithet <- scientificName <- . <- eventDate <- NULL
  
  #### 26.1 Prep ####
  # This will load the requireNamespaced packages. These packages may still need to be installed to 
  # R using install.packages("dplyr")... etc.
  requireNamespace("dplyr")
  requireNamespace("lubridate")

  requireNamespace("mgsub")
  
  
  #### 26.2 Read+ ####
  ###### a. month strings ####
    # Prepare month strings to convert from roman numerals
  romanNumerals <- c("i","ii","iii","iv","v","vi","vii","viii","ix","x","xi","xii",
                     "I","II","III","IV","V","VI","VII","VIII","IX","X","XI","XII")
  numeralConversion <- c("Jan", "Feb", "Mar", "Apr", "May", "Jun",
                         "Jul", "Aug", "Sep", "Oct","Nov", "Dec",
                         "Jan", "Feb", "Mar", "Apr", "May", "Jun",
                         "Jul", "Aug", "Sep", "Oct","Nov", "Dec")
  
  ###### b. PALA_data ####
  # Reads in the .csv file, trims the white spaces, and formats the columns to the correct type
  PALA_data <- readr::read_csv(paste(path, inFile, sep = "/"),
                               trim_ws = TRUE) %>%
  # Rename columns
    dplyr::rename(
      catalogNumber = "catalogNumber",
      type = "Type",
      country = "Country",
      municipality = "Muninciplaity",
      locality = "Site",
      decimalLatitude = "Latitud",
      decimalLongitude = "Longitude",
      verbatimElevation = "elevation",
      verbatimEventDate = "date",
      recordedBy = "recordedby",
      collectionCode = "Collection",
      otherCatalogNumbers = "othercatalognumber",
      associatedTaxa = "AssociatedTaxa",
      taxonRemarks = "taxonremarks",
      family = "Family",
      genus = "Genus",
      specificEpithet = "species",
      references = "Citation"
    ) %>%
      # Add in sciName
    dplyr::mutate(scientificName = stringr::str_c(genus, specificEpithet, sep = " ")) %>%
      # Format date
    dplyr::mutate(eventDate = verbatimEventDate %>%
                    mgsub::mgsub(
                      pattern = paste("[-/ \\.]", romanNumerals, "[ -/\\.]", sep = ""),
                      replacement = numeralConversion) %>%
                    lubridate::dmy(truncated = 2, quiet = TRUE),
                  .after = verbatimEventDate) %>% 
      # Add day, month, year
    dplyr::mutate(
      day = lubridate::day(eventDate),
      month = lubridate::month(eventDate),
      year = lubridate::year(eventDate),
      .after = eventDate
    ) %>%
    # Remove any double white-spaces
    apply(., 2, stringr::str_squish) %>% dplyr::as_tibble() %>% 
    # add the database_id column
    dplyr::mutate(
      database_id = paste("PALA_data_", 1:nrow(.), sep = ""),
      .before = catalogNumber)  %>%
    # Add dataset information
    dplyr::mutate(dataSource = "PALA_Anthophila") %>%
    dplyr::mutate(license = dataLicense)  %>%
    # add the database_id column
    dplyr::mutate(
      datasetName = "PALA",
      datasetID = "PALA"
    )
  
  #### 26.3 Out ####
  # Save the dataset
  readr::write_excel_csv(PALA_data, file = paste(path, outFile, sep = "/"))
  # Return the data from the function to the user
  return(PALA_data)
} # END readr_PALA




#### 27.0 JoLa ####
#' @rdname readr_BeeBDC
#' 
#' 
#' 
#'
#' 
readr_JoLa <- function(path = NULL,
                       inFile = NULL,
                       outFile = NULL,
                       dataLicense = NULL,
                       sheet = c("pre-1950", "post-1950")){
  # locally bind variables to the function
  fieldNotes <- Catalognumber <-`Start Date (Year)` <- genus <- specificEpithet <-  NULL
  year <- . <- NULL
  
  #### 27.1 Prep ####
  # This will load the requireNamespaced packages. These packages may still need to be installed to 
  # R using install.packages("dplyr")... etc.
  requireNamespace("dplyr")
  requireNamespace("lubridate")

  
  
  
  #### 27.2 Read+ ####
  ###### a. JoLa_data ####
  # Reads in the .csv file, trims the white spaces, and formats the columns to the correct type
    # Read in both sheets and bind them together
  JoLa_data <-  openxlsx::read.xlsx(paste(path, inFile, sep = "/"),
                                   sheet = sheet[1]) %>% 
    dplyr::bind_rows(openxlsx::read.xlsx(paste(path, inFile, sep = "/"),
                                        sheet = sheet[2])) %>%
    # Return spaces in column names to keep the consistent with file before renaming
    stats::setNames(., stringr::str_replace_all(colnames(.), "\\.", " ")) %>%
      # Rename the columns
    dplyr::rename(
      specificEpithet = "Species",
      decimalLatitude = "Latitude1",
      decimalLongitude = "Longitude1",
      year = `Start Date (Year)`
    ) %>%
    # Add in higher taxonomic information
    dplyr::mutate(
      genus = "Lasioglossum",
      family = "Halictidae",
      order = "Hymenoptera",
      scientificName = stringr::str_c(genus, specificEpithet, sep = " ")
    ) %>%
    dplyr::mutate(
      eventDate = lubridate::ymd(year, truncated = 2)
    ) %>%
    # Remove any double white-spaces
    apply(., 2, stringr::str_squish) %>% dplyr::as_tibble() %>% 
    # add the database_id column
    dplyr::mutate(
      database_id = paste("JoLa_data_", 1:nrow(.), sep = ""),
      .before = 1)  %>%
    # Add dataset information
    dplyr::mutate(dataSource = "JoLa_Anthophila") %>%
    dplyr::mutate(license = dataLicense)  %>%
    # add the database_id column
    dplyr::mutate(
      datasetName = "JoLa",
      datasetID = "JoLa"
    )
  
  #### 27.3 Out ####
  # Save the dataset
  readr::write_excel_csv(JoLa_data, file = paste(path, outFile, sep = "/"))
  # Return the data from the function to the user
  return(JoLa_data)
} # END readr_JoLa



#### 28.0 VicWam ####
#' @rdname readr_BeeBDC
#' 
#' 
#' 
#'
#' 
readr_VicWam <- function(path = NULL,
                       inFile = NULL,
                       outFile = NULL,
                       dataLicense = NULL,
                       sheet = "Combined"){
  # locally bind variables to the function
  fieldNotes <- Catalognumber <- recordNumber <- genus <- specificEpithet <-  NULL
  year <- . <- VicWam_data <- otherCatalogNumbers <- institutionCode <- class <- order <- NULL
  infraspecificEpithet <- NULL
  NEAREST <- DISTANCE <- DIST_UNIT <- DIRECTION <- locality <- locality2 <- PLACEACCURACY <- 
    coordinateUncertaintyInMeters <- DTFR <- day <- month <- year <- eventDate <- eventDate2 <- 
    DTTO <- dayTO <- monthTO <- yearTO <- eventDateTO <- eventDateTO2 <- LABELFAMILY <- 
    LABELGENUS <- LABELSPECIES <- associatedTaxa <- stateProvince <- country <- dataSource <- 
    license <- decimalLatitude <- decimalLongitude <- lengthTest <- decimalLatitude2 <- 
    decimalLongitude2 <- NULL
  
  #### 28.1 Prep ####
  # This will load the requireNamespaced packages. These packages may still need to be installed to 
  # R using install.packages("dplyr")... etc.
  requireNamespace("dplyr")
  requireNamespace("lubridate")
  
  
  #### 28.2 Read+ ####
  ###### a. JoLa_data ####
  # Reads in the .csv file, trims the white spaces, and formats the columns to the correct type
  # Read in both sheets and bind them together
  VicWam_data <-  openxlsx::read.xlsx(paste(path, inFile, sep = "/"),
                                    sheet = sheet[1]) %>%
      # Start by renaming columns
    dplyr::rename(
      recordNumber = "Reg_DoreyExtension",
      otherCatalogNumbers = "COLLNUM",
      institutionCode = "INSTITUTE",
      class = "CLASS",
      order = "ORDER",
      "superfamily" = "SUPERFAMILY",
      "family" = "FAMILY",
      "subfamily" = "SUBFAMILY",
      "tribe" = "TRIBE",
      "genus" = "GENUS",
      "subgenus" = "SUBGENUS",
      "specificEpithet" = "SPECIES",
      "infraspecificEpithet" = "SUBSPECIES",
      "identifiedBy" = "DTMNDBY",
      "dateIdentified" = "DTMNDDT",
      "lifeStage" = "LIFEHISTORY",
      "sex" = "SEX",
      "identificationQualifier" = "NAMEQUALIFIER",
      "typeStatus" = "SPCMTYPE",
      "individualCount" = "SPECNUM",
      "stateProvince" = "STATE",
      "locality" = "SITE",
      "decimalLatitude" = "LATDEC",
      "decimalLongitude" = "LONGDEC",
      "recordedBy" = "COLLTOR",
      "basisOfRecord" = "STORAGE",
      "verbatimLatitude" = "LATITUDE",
      "verbatimLongitude" = "LONGITUDE",
      "occurrenceRemarks" = "NOTES"
    ) %>%
      # Create the scientificName
    dplyr::mutate(scientificName = stringr::str_c(
      dplyr::if_else(!is.na(genus),
                     paste0(genus), ""),
      dplyr::if_else(!is.na(specificEpithet),
                     paste0(specificEpithet), ""),
      dplyr::if_else(!is.na(infraspecificEpithet),
                     paste0(infraspecificEpithet), ""),
      sep = " "
    ) %>% stringr::str_squish()) %>%
      # modify locality
    dplyr::mutate(locality2 = stringr::str_c(dplyr::if_else(stats::complete.cases(NEAREST),
                                                             NEAREST, ""),
                                              dplyr::if_else(stats::complete.cases(DISTANCE),
                                                             as.character(DISTANCE), ""),
                                              dplyr::if_else(stats::complete.cases(DIST_UNIT),
                                                             DIST_UNIT, ""),
                                              dplyr::if_else(stats::complete.cases(DIRECTION),
                                                             paste0(DIRECTION, "\u00b0"), 
                                                             ""),
                                              sep = " ") %>% as.character() %>%
                    stringr::str_squish(),
                  .after = locality) %>%
    dplyr::mutate(locality = dplyr::if_else(is.na(locality),
                                            locality2, locality)) %>%
    dplyr::select(!locality2) %>%
    # Extract coordinateUncertaintyInMeters
    dplyr::mutate(
      coordinateUncertaintyInMeters = PLACEACCURACY %>% stringr::str_extract(
        "[0-9]+\\s[m(km)(mi)]+") %>% stringr::str_replace_all(c("km" = "000",
                                                                "m" = "")) %>%
        stringr::str_replace(" ","") %>% as.numeric(),
      .after = PLACEACCURACY
    ) %>%
      # Create eventDate
    dplyr::mutate(
      day = DTFR %>% stringr::str_extract("^[0-9]+/") %>% stringr::str_remove_all("/") %>%
        dplyr::if_else(. == "00", NA_character_, .),
      month = DTFR %>% stringr::str_extract("/[0-9]+/") %>% stringr::str_remove_all("/") %>%
        dplyr::if_else(. == "00", NA_character_, .),
      year = DTFR %>% stringr::str_extract("/[0-9]+$") %>% stringr::str_remove_all("/") %>%
        dplyr::if_else(. == "0000", NA_character_, .),
      eventDate = lubridate::dmy(stringr::str_c(day, month, year, sep = "/"), truncated = 2),
      eventDate2 = dplyr::if_else(!stringr::str_detect(DTFR, "[a-zA-Z]"),  # convert from silly excel numeric format to real dates...
        as.Date(as.numeric(DTFR), origin = "1899-12-30") %>% as.character(),
        DTFR) %>% lubridate::ymd(truncated = 2),
        # if eventDate is empty, use eventDate2
      eventDate = dplyr::if_else(is.na(eventDate),
                                 eventDate2, eventDate),
      .after = DTFR
    ) %>% dplyr::select(!eventDate2) %>%
      # Create the date to
    dplyr::mutate(
      dayTO = DTTO %>% stringr::str_extract("^[0-9]+/") %>% stringr::str_remove_all("/") %>%
        dplyr::if_else(. == "00", NA_character_, .),
      monthTO = DTTO %>% stringr::str_extract("/[0-9]+/") %>% stringr::str_remove_all("/") %>%
        dplyr::if_else(. == "00", NA_character_, .),
      yearTO = DTTO %>% stringr::str_extract("/[0-9]+$") %>% stringr::str_remove_all("/") %>%
        dplyr::if_else(. == "0000", NA_character_, .),
      eventDateTO = lubridate::dmy(stringr::str_c(dayTO, monthTO, yearTO, sep = "/"), truncated = 2),
      eventDateTO2 = dplyr::if_else(!stringr::str_detect(DTTO, "[a-zA-Z]"),  # convert from silly excel numeric format to real dates...
                                  as.Date(as.numeric(DTTO), origin = "1899-12-30") %>% as.character(),
                                  DTTO %>% as.character()) %>% lubridate::ymd(truncated = 2),
      # if eventDateTO is empty, use eventDateTO2
      eventDateTO = dplyr::if_else(is.na(eventDateTO),
                                   eventDateTO2, eventDateTO) %>% as.character(),
      eventDateTO = dplyr::if_else(is.na(eventDateTO),
                                   stringr::str_c(
                                     dplyr::if_else(stats::complete.cases(yearTO),
                                                    yearTO,""),
                                     dplyr::if_else(stats::complete.cases(monthTO),
                                                    monthTO,""),
                                     dplyr::if_else(stats::complete.cases(dayTO),
                                                    dayTO,""),
                                     sep = " ") %>% stringr::str_squish() %>% 
                                     stringr::str_replace(" ", "-"),
                                   eventDateTO),
      .after = DTTO
    ) %>% 
    dplyr::select(!tidyselect::any_of(c("eventDateTO2", "dayTO", "monthTO", "yearTO"))) %>%
      # Add field notes
    dplyr::mutate(
    fieldNotes = stringr::str_c(
      dplyr::if_else(!is.na(eventDate),
                     paste0("startDate: ", eventDate), ""),
      dplyr::if_else(!is.na(eventDateTO),
                     paste0("endDate: ", eventDateTO), ""),
      dplyr::if_else(!is.na(LABELFAMILY),
                     paste0("associatedFamily: ", LABELFAMILY), ""),
      dplyr::if_else(!is.na(LABELGENUS),
                     paste0("associatedGenus: ", LABELGENUS), ""),
      dplyr::if_else(!is.na(LABELSPECIES),
                     paste0("associatedSpecies: ", LABELSPECIES), ""),
      sep = "|")) %>% 
        # Add associatedTaxa
    dplyr::mutate(associatedTaxa = stringr::str_c(
      dplyr::if_else(!is.na(LABELGENUS),
                     paste0(LABELGENUS), ""),
      dplyr::if_else(!is.na(LABELSPECIES),
                     paste0(LABELSPECIES), ""),
      sep = " ") %>% stringr::str_squish()) %>%
        # Worst case, use family
      dplyr::mutate(associatedTaxa = stringr::str_c(
        dplyr::if_else(is.na(associatedTaxa),
                       paste0(LABELFAMILY), associatedTaxa)
      )) %>% 
      # Format country name
    dplyr::mutate(country = stringr::str_to_sentence(country)) %>%
    dplyr::mutate(stateProvince = stringr::str_replace_all(stateProvince,
                                   c("^ACT$" = "Australian Capital Territory",
                                     "New South wales" = "New South Wales",
                                     "^NSW" = "New South Wales",
                                     "^NT$" = "Northern Territory",
                                     "^Qld$" = "Queensland",
                                     "^SA$" = "South Australia",
                                     "^Tas$" = "Tasmania",
                                     "^Vic$" = "Victoria",
                                     "^WA$" = "Western Australia",
                                     "Western australia" = "Western Australia"))) %>% 
    # Remove any double white-spaces
    apply(., 2, stringr::str_squish) %>% dplyr::as_tibble() %>% 
    # add the database_id column
    dplyr::mutate(
      database_id = paste("VicWam_data_", 1:nrow(.), sep = ""),
      .before = 1)  %>%
    # Add dataset information
    dplyr::mutate(dataSource = "VicWam_Anthophila") %>%
    dplyr::mutate(license = dataLicense)  %>%
    # add the database_id column
    dplyr::mutate(
      datasetName = "VicWam",
      datasetID = "VicWam"
    ) %>%
      # Format some lats and lons to decimal degrees
    # first, squish out extra spaces
    dplyr::mutate(
      decimalLatitude = decimalLatitude %>% stringr::str_squish(),
      decimalLongitude = decimalLongitude %>% stringr::str_squish(),
    ) %>%
      # make a column to test for the lat/lon length
    dplyr::mutate(lengthTest = dplyr::if_else(decimalLatitude %>% stringr::str_detect("S"),
                                              dplyr::if_else(stringr::str_count(decimalLatitude,
                                                                                "\\s") > 2,
                                                             "Long", "Short"),
                                              "NA"),
                  .before = decimalLatitude) %>%
    # Convert to DD
    # Long coordinates
    dplyr::mutate(decimalLatitude2 = dplyr::if_else(decimalLatitude %>% stringr::str_detect("S") &
                                                      lengthTest == "Long",
                                                    stringr::str_c(stringr::str_extract(decimalLatitude, "^[0-9]{2}") %>%
                                                                     as.numeric() +
                                                                     (stringr::str_extract(decimalLatitude,
                                                                                           "\\s[0-9]{2}\\s") %>% 
                                                                        as.numeric() / 60) +
                                                                     (stringr::str_extract(decimalLatitude, 
                                                                                           "[0-9]{2}\\sS") %>% 
                                                                        stringr::str_remove("S") %>% 
                                                                        as.numeric() / 3600)) %>%
                                                      as.character(),
                                                    decimalLatitude),
                  decimalLongitude2 = dplyr::if_else(decimalLongitude %>% stringr::str_detect("E") &
                                                       lengthTest == "Long",
                                                     stringr::str_c(stringr::str_extract(decimalLongitude, "^[0-9]+") %>%
                                                                      as.numeric() +
                                                                      (stringr::str_extract(decimalLongitude, 
                                                                                            "\\s[0-9]{2}\\s") %>%
                                                                         as.numeric() / 60) +
                                                                      (stringr::str_extract(decimalLongitude, 
                                                                                            "\\s[0-9]+\\sE") %>% 
                                                                         stringr::str_remove("E") %>% 
                                                                         as.numeric() / 3600)) %>%
                                                       as.character(),
                                                     decimalLongitude),
                  .after = decimalLatitude) %>%
    # Short coordinates
    dplyr::mutate(decimalLatitude2 = dplyr::if_else(decimalLatitude %>% stringr::str_detect("S") &
                                                      lengthTest == "Short",
                                                    stringr::str_c(stringr::str_extract(decimalLatitude, "^[0-9]{2}") %>%
                                                                     as.numeric() +
                                                                     (stringr::str_extract(decimalLatitude,
                                                                                           "\\s[0-9]{2}\\sS") %>% 
                                                                        stringr::str_remove("S") %>%
                                                                        as.numeric() / 60) ) %>%
                                                      as.character(),
                                                    decimalLatitude2) %>% as.numeric(),
                  decimalLongitude2 = dplyr::if_else(decimalLongitude %>% stringr::str_detect("E") &
                                                       lengthTest == "Short",
                                                     stringr::str_c(stringr::str_extract(decimalLongitude, "^[0-9]+") %>%
                                                                      as.numeric() +
                                                                      (stringr::str_extract(decimalLongitude, 
                                                                                            "\\s[0-9]{2}+\\sE") %>%
                                                                         stringr::str_remove("E") %>%
                                                                         as.numeric() / 60) ) %>%
                                                       as.character(),
                                                     decimalLongitude2) %>% as.numeric(),
                  .after = decimalLatitude) %>%
      # Change lat/lon to correct accuracy
    dplyr::mutate(
      decimalLatitude2 = dplyr::if_else(lengthTest == "Long",
                                        round(decimalLatitude2, digits = 2),
                                        dplyr::if_else(lengthTest == "Short",
                                                       round(decimalLatitude2, digits = 1),
                                                       decimalLatitude2)),
      decimalLongitude2 = dplyr::if_else(lengthTest == "Long",
                                         round(decimalLongitude2, digits = 2),
                                         dplyr::if_else(lengthTest == "Short",
                                                        round(decimalLongitude2, digits = 1),
                                                        decimalLongitude2))
    ) %>%
      # Add coordinateUncertaintyInMeters for these as well based on the length of coordinates
    dplyr::mutate(
      coordinateUncertaintyInMeters = coordinateUncertaintyInMeters %>% as.numeric(),
      coordinateUncertaintyInMeters = dplyr::if_else(
      is.na(coordinateUncertaintyInMeters),
      dplyr::if_else(lengthTest == "Long", 1000, 10000),
      coordinateUncertaintyInMeters
    )) %>%
      # Replace the lat/lon columns with the working columns
    dplyr::select(!c("decimalLongitude", "decimalLatitude")) %>%
    dplyr::rename(decimalLongitude = "decimalLongitude2",
                  decimalLatitude = "decimalLatitude2") %>%
    # All of Australia is in the southern hemisphere, correct these latitudes to negative
    dplyr::mutate(decimalLatitude = dplyr::if_else(
      stateProvince %in% c("Australian Capital Territory", "New South Wales",
                           "Northern Territory", "Queensland", "South Australia",
                           "Tasmania", "Victoria", "Western Australia", ""),
      abs(decimalLatitude) * - 1,
      decimalLatitude)) %>%
      # Do the same for longitude but hte entire dataset is in the eastern hemisphere
    dplyr::mutate(decimalLongitude = abs(decimalLongitude))
  
  #### 28.3 Out ####
  # Save the dataset
  readr::write_excel_csv(VicWam_data, file = paste(path, outFile, sep = "/"))
  # Return the data from the function to the user
  return(VicWam_data)
} # END readr_VicWam

  # Test each of the readr_functions

  # Load required packages
requireNamespace("readr")
requireNamespace("BeeBDC")
requireNamespace("dplyr")
requireNamespace("openxlsx")

library(dplyr) ## could not use %>% without loading as library

#### 1.0 EPEL ####

testData <- dplyr::tribble(
  ~catalog_number, ~pollinator_family, ~pollinator_genus, ~pollinator_species, ~collection_method, ~collector_number, ~day_collected, ~month_collected, ~year_collected,                   ~location_description, ~location_name,    ~habitat, ~latitude,  ~longitude, ~basis_of_record,
  "SFU7052435916",       "Halictidae",    "Secret",             "sp. 1",    "Pantrap, blue",               20L,            30L,            "Apr",           2007L, "Bear Hill Regional Park, Victoria CRD",           "BH", "Garry Oak", 48.545605, -123.406769,       "specimen",
  "SFU705917",       "Andrenidae",         "Secret",     "sp.",    "Pantrap, blue",               20L,            30L,            "Apr",           2007L, "Bear Hill Regional Park, Victoria CRD",           "BH", "Garry Oak", 48.545605, -123.406769,       "specimen",
  "SFU7052346919",       "Halictidae",    "Secret",       "sp.",    "Pantrap, blue",               20L,            30L,            "Apr",           2007L, "Bear Hill Regional Park, Victoria CRD",           "BH", "Garry Oak", 48.545605, -123.406769,       "specimen",
  "SFU7052645920",       "Andrenidae",         "Secret",     "sp.",    "Pantrap, blue",               20L,            30L,            "Apr",           2007L, "Bear Hill Regional Park, Victoria CRD",           "BH", "Garry Oak", 48.545605, -123.406769,       "specimen",
  "SFU705645921",       "Andrenidae",         "Secret",     "sp.",    "Pantrap, blue",               20L,            30L,            "Apr",           2007L, "Bear Hill Regional Park, Victoria CRD",           "BH", "Garry Oak", 48.545605, -123.406769,       "specimen"
)

# Be sure that the testData is not already in tempdir
testDataPath <- file.info(list.files(tempdir(), full.names = T, 
                                     pattern = "testData.csv", recursive = TRUE))
unlink(rownames(testDataPath))

# Save a temporary version of these data
readr::write_excel_csv(testData, paste0(tempdir(), "/testData.csv"))

testOut1 <- BeeBDC::readr_BeeBDC(dataset = "EPEL",
                               path = tempdir(),
                              inFile = "testData.csv",
                              outFile = "testDataOut.csv",
                              dataLicense = "https://creativecommons.org/licenses/by-nc-sa/4.0/")


# Get a count of TRUE and FALSE column name matches
resultsT <- sum(colnames(testOut1) %in% (ColTypeR()[[1]] %>% names()) == TRUE)
resultsF <- sum(colnames(testOut1) %in% (ColTypeR()[[1]] %>% names()) == FALSE)

# Test the number of expected TRUE and FALSE columns and then test the output format (data frames and
# tibbles are a special case of lists)
testthat::test_that("readr_EPEL results columns TRUE", {
  testthat::expect_equal(resultsT, 21)
})
testthat::test_that("readr_EPEL results columns FALSE", {
  testthat::expect_equal(resultsF, 0)
})

testthat::test_that("readr_EPEL expected class", {
  testthat::expect_type(testOut1, "list")
})




#### 2.0 ASP ####
    testData <- dplyr::tribble(
      ~catalogNumber,        ~family,     ~subfamily,       ~Tribe,         ~genus,      ~subgenus, ~Morphospecies, ~specificEpithet,   ~locality, ~Successional_Stage, ~decimalLatitude, ~decimalLongitude, ~coordinateUncertaintyInMeters, ~elevation, ~eventTime, ~samplingProtocol,   ~eventDate, ~sex, ~associatedTaxa,      ~continent,         ~recordedBy,
      2L, "Megachilidae", "Megachilinae", "Anthidiini", "Anthodioctes", "Nananthidium",        "Msp 1",            "m16", "Calandria",                "RA",            6.773,          -75.1036,                           100L,      1019L,     "p.m.",               "J",     "3/6/97",  "H",              NA, "South America", "Allan Smith-Pardo",
      8L, "Megachilidae", "Megachilinae", "Anthidiini", "Anthodioctes", "Anthodioctes",        "Msp 2",      "mapirense", "Calandria",                "RA",            6.773,          -75.1036,                           100L,      1019L,       "m.",               "J",  "23/9/1997",  "H",              NA, "South America", "Allan Smith-Pardo",
      9L, "Megachilidae", "Megachilinae", "Anthidiini", "Anthodioctes", "Anthodioctes",        "Msp 2",      "mapirense", "Calandria",                "RA",            6.773,          -75.1036,                           100L,      1019L,       "m.",               "J", "31/10/1997",  "H",              NA, "South America", "Allan Smith-Pardo",
      10L, "Megachilidae", "Megachilinae", "Anthidiini", "Anthodioctes", "Anthodioctes",        "Msp 2",      "mapirense", "Calandria",                "RA",            6.773,          -75.1036,                           100L,      1019L,     "p.m.",               "J",  "22/8/1997",  "H",              NA, "South America", "Allan Smith-Pardo"
    )
    
# Be sure that the testData is not already in tempdir
testDataPath <- file.info(list.files(tempdir(), full.names = T, 
                                     pattern = "testData.csv", recursive = TRUE))
unlink(rownames(testDataPath))

    # Save a temporary version of these data
    readr::write_excel_csv(testData, paste0(tempdir(), "/testData.csv"))
    
    
    testOut1 <- BeeBDC::readr_BeeBDC(dataset = "ASP",
                                   path = tempdir(),
                                   inFile = "testData.csv",
                                   outFile = "testDataOut.csv",
                                   dataLicense = "https://creativecommons.org/licenses/by-nc-sa/4.0/")
    
    
    # Get a count of TRUE and FALSE column name matches
    resultsT <- sum(colnames(testOut1) %in% (ColTypeR()[[1]] %>% names()) == TRUE)
    resultsF <- sum(colnames(testOut1) %in% (ColTypeR()[[1]] %>% names()) == FALSE)
    
    
    # Test the number of expected TRUE and FALSE columns and then test the output format (data frames and
    # tibbles are a special case of lists)
    testthat::test_that("readr_ASP results columns TRUE", {
      testthat::expect_equal(resultsT, 26)
    })
    
    testthat::test_that("readr_ASP results columns FALSE", {
      testthat::expect_equal(resultsF, 4)
    })
    
    testthat::test_that("readr_ASP expected class", {
      testthat::expect_type(testOut1, "list")
    })



#### 3.0 BMin ####
    
    library(dplyr) ## could not use %>% without loading as library
    
    
    testData <- dplyr::tribble(
      ~catalogNumber,      ~Phylum, ~higherClassification,        ~Order,      ~family,        ~genus, ~specificEpithet,        ~scientificName, ~scientificNameAuthorship,     ~sex,         ~associatedTaxa,    ~identifiedBy, ~dateIdentified, ~recordedBy,   ~eventDate,      ~continent,    ~country,           ~stateProvince, ~county,                                              ~locality, ~decimalLatitude, ~decimalLongitude,             ~collectionID,       ~basisOfRecord,
      "RLMC00005185", "Arthropoda",             "Insecta", "Hymenoptera", "Colletidae",   "Cadeguala",     "albopilosa", "Cadeguala albopilosa",         "(Spinola, 1851)",   "Male",                      NA,       "L Packer",              NA,  "L Packer", "31/10/2001", "South America",     "Chile", "Region de la Araucania",      NA, "Parque Nacional Nahuelbuta, Piedra del Aguilla trail",         -37.8251,           -73.037, "University of Rochester", "preserved specimen",
      "RLMC00005186", "Arthropoda",             "Insecta", "Hymenoptera", "Colletidae",   "Cadeguala",     "albopilosa", "Cadeguala albopilosa",         "(Spinola, 1851)", "Female", "Aristotelia chilensis",       "L Packer",              NA, "A-I Gavel",     "2/7/05", "South America", "Argentina",   "Provincia del Chubut",      NA,                                "INTA Trevelin, site 2",     -43.09925833,         -71.54235, "University of Rochester", "preserved specimen",
      "RLMC00005187", "Arthropoda",             "Insecta", "Hymenoptera", "Andrenidae", "Spinoliella",         "rozeni",   "Spinoliella rozeni",        "Toro & Ruz, 1972",   "Male",                      NA, "AH Smith-Pardo",              NA,  "L Packer", "13/10/2001", "South America",     "Chile",                "Atacama",      NA,             "Tres Playitas, 10km north of Huasco Bajo",         -28.4066,          -71.1891, "University of Rochester", "preserved specimen",
      "RLMC00005188", "Arthropoda",             "Insecta", "Hymenoptera", "Andrenidae", "Spinoliella",         "rozeni",   "Spinoliella rozeni",        "Toro & Ruz, 1972", "Female",                      NA, "AH Smith-Pardo",              NA,  "L Packer", "13/11/1997", "South America",     "Chile",                "Atacama",      NA,                                             "Chanaral",         -26.3328,          -70.6347, "University of Rochester", "preserved specimen"
    )
    
    # Be sure that the testData is not already in tempdir
    testDataPath <- file.info(list.files(tempdir(), full.names = T, 
                                         pattern = "testData.csv", recursive = TRUE))
    unlink(rownames(testDataPath))
    
    # Save a temporary version of these data
    readr::write_excel_csv(testData, paste0(tempdir(), "/testData.csv"))
    
    
    testOut1 <- BeeBDC::readr_BeeBDC(dataset = "BMin",
                                   path = tempdir(),
                                  inFile = "testData.csv",
                                  outFile = "testDataOut.csv",
                                  dataLicense = "https://creativecommons.org/licenses/by-nc-sa/4.0/")
    
    
    # Get a count of TRUE and FALSE column name matches
    resultsT <- sum(colnames(testOut1) %in% (ColTypeR()[[1]] %>% names()) == TRUE)
    resultsF <- sum(colnames(testOut1) %in% (ColTypeR()[[1]] %>% names()) == FALSE)
    
    
    # Test the number of expected TRUE and FALSE columns and then test the output format (data frames and
    # tibbles are a special case of lists)
    testthat::test_that("readr_BMin results columns TRUE", {
      testthat::expect_equal(resultsT, 28)
    })
    
    testthat::test_that("readr_BMin results columns FALSE", {
      testthat::expect_equal(resultsF, 2)
    })
    
    testthat::test_that("readr_BMin expected class", {
      testthat::expect_type(testOut1, "list")
    })
    
    
#### 4.0 BMont ####
    
    testData <- dplyr::tribble(
      ~occurence_lsid, ~language,      ~basisOfRecord, ~catalogNumber,                                  ~scientificNameID,     ~scientificName,     ~sex, ~lifeStage, ~organismQuantity,                                                       ~institutionCode,                                                                                 ~fieldNotes,                                ~locality, ~decimalLatitude, ~decimalLongitude,    ~GPS_device,        ~country, ~stateProvince,   ~county, ~municipality,                                                          ~georeferenceRemarks, ~VerbatimEventDate, ~eventDate, ~eventTime, ~recordedBy, ~samplingProtocol, ~fieldNumber,              ~identifiedBy, ~dateIdentified, ~typeStatus,   ~kingdom,      ~phylum,     ~class,        ~order,  ~family,   ~genus, ~specificEpithet, ~infraspecificEpithet, ~habitat,       ~associatedTaxa,                                            ~source,              ~modified,
      "urn:lsid:biosci.ohio-state.edu:osuc_occurrences:MTEC__014085",      "en", "PreservedSpecimen",  "MTEC 014085", "urn:lsid:biosci.ohio-state.edu:osuc_names:169432",  "Bombus sylvicola", "female",    "queen",                1L, "University of Colorado Museum of Natural History, Boulder, CO (UCMC)",                    "[MONT. Powell Co. Flint Crk. Mts. 8500ft-9500ft, 28 July 1961 B Vogel]", "Flint Creek Mountains, Granite Co., MT",          46.3624,         -113.1077, "Google Earth", "United States",      "Montana", "Granite",            NA, "Label says Powell Co., but the Flint Creek Range is actually in Granite Co.",        "28-Jul-61",  "7/28/61",         NA, "Vogel, B.",  "none specified",           NA, "Thorp, R. W. (Robbin W.)",           1966L,          NA, "Animalia", "Arthropoda", "Hexapoda", "Hymenoptera", "Apidae", "Bombus",      "sylvicola",                    NA,       NA,                    NA, "http://hol.osu.edu/spmInfo.html?id=MTEC%20014085", "2015-11-06T12:23:32Z",
      "urn:lsid:biosci.ohio-state.edu:osuc_occurrences:MTEC__014086",      "en", "PreservedSpecimen",  "MTEC 014086", "urn:lsid:biosci.ohio-state.edu:osuc_names:169432",  "Bombus sylvicola", "female",    "queen",                1L, "University of Colorado Museum of Natural History, Boulder, CO (UCMC)",                    "[MONT. Powell Co. Flint Crk. Mts. 8500ft-9500ft, 28 July 1961 B Vogel]", "Flint Creek Mountains, Granite Co., MT",          46.3624,         -113.1077, "Google Earth", "United States",      "Montana", "Granite",            NA, "Label says Powell Co., but the Flint Creek Range is actually in Granite Co.",        "28-Jul-61",  "7/28/61",         NA, "Vogel, B.",  "none specified",           NA, "Thorp, R. W. (Robbin W.)",           1966L,          NA, "Animalia", "Arthropoda", "Hexapoda", "Hymenoptera", "Apidae", "Bombus",      "sylvicola",                    NA,       NA,                    NA, "http://hol.osu.edu/spmInfo.html?id=MTEC%20014086", "2015-11-06T12:23:32Z",
      "urn:lsid:biosci.ohio-state.edu:osuc_occurrences:MTEC__014087",      "en", "PreservedSpecimen",  "MTEC 014087", "urn:lsid:biosci.ohio-state.edu:osuc_names:169432",  "Bombus sylvicola", "female",    "queen",                1L, "University of Colorado Museum of Natural History, Boulder, CO (UCMC)",                                  "[MONT. Carbon Co. Beartooth Plateau, 9 Jly 1963 B Vogel]",      "Beartooth Plateau, Carbon Co., MT",          45.0214,         -109.5779, "Google Earth", "United States",      "Montana",  "Carbon",            NA,                                                                            NA,         "9-Jul-63",   "7/9/63",         NA, "Vogel, B.",  "none specified",           NA, "Thorp, R. W. (Robbin W.)",           1966L,          NA, "Animalia", "Arthropoda", "Hexapoda", "Hymenoptera", "Apidae", "Bombus",      "sylvicola",                    NA,       NA,                    NA, "http://hol.osu.edu/spmInfo.html?id=MTEC%20014087", "2015-09-28T11:15:16Z",
      "urn:lsid:biosci.ohio-state.edu:osuc_occurrences:MTEC__014088",      "en", "PreservedSpecimen",  "MTEC 014088", "urn:lsid:biosci.ohio-state.edu:osuc_names:128375", "Bombus flavifrons", "female",    "queen",                1L, "University of Colorado Museum of Natural History, Boulder, CO (UCMC)",                                    "[MONT: Carbon Co. E. Rosebud L. 20 June 1962 B. Vogel]",        "E. Rosebud Lake, Carbon Co., MT",          45.2002,         -109.6412, "Google Earth", "United States",      "Montana",  "Carbon",            NA,                                                                            NA,        "20-Jun-62",  "6/20/62",         NA, "Vogel, B.",  "none specified",           NA, "Thorp, R. W. (Robbin W.)",           1966L,          NA, "Animalia", "Arthropoda", "Hexapoda", "Hymenoptera", "Apidae", "Bombus",     "flavifrons",                    NA,       NA,                    NA, "http://hol.osu.edu/spmInfo.html?id=MTEC%20014088", "2015-09-28T11:15:17Z",
      "urn:lsid:biosci.ohio-state.edu:osuc_occurrences:MTEC__014089",      "en", "PreservedSpecimen",  "MTEC 014089", "urn:lsid:biosci.ohio-state.edu:osuc_names:128375", "Bombus flavifrons", "female",    "queen",                1L, "University of Colorado Museum of Natural History, Boulder, CO (UCMC)", "[MONT: Carbon Co. E. Rosebud L. 20 June 1962 B. Vogel] [on Penstemon confertas] confertus",        "E. Rosebud Lake, Carbon Co., MT",          45.2002,         -109.6412, "Google Earth", "United States",      "Montana",  "Carbon",            NA,                                                                            NA,        "20-Jun-62",  "6/20/62",         NA, "Vogel, B.",  "none specified",           NA, "Thorp, R. W. (Robbin W.)",           1966L,          NA, "Animalia", "Arthropoda", "Hexapoda", "Hymenoptera", "Apidae", "Bombus",     "flavifrons",                    NA,       NA, "Penstemon confertus", "http://hol.osu.edu/spmInfo.html?id=MTEC%20014089", "2015-09-28T11:15:17Z",
      "urn:lsid:biosci.ohio-state.edu:osuc_occurrences:MTEC__014090",      "en", "PreservedSpecimen",  "MTEC 014090", "urn:lsid:biosci.ohio-state.edu:osuc_names:128375", "Bombus flavifrons", "female",   "worker",                1L, "University of Colorado Museum of Natural History, Boulder, CO (UCMC)", "[MONT: Carbon Co. E. Rosebud L. 20 June 1962 B. Vogel] [on Penstemon confertas] confertus",        "E. Rosebud Lake, Carbon Co., MT",          45.2002,         -109.6412, "Google Earth", "United States",      "Montana",  "Carbon",            NA,                                                                            NA,        "20-Jun-62",  "6/20/62",         NA, "Vogel, B.",  "none specified",           NA, "Thorp, R. W. (Robbin W.)",           1966L,          NA, "Animalia", "Arthropoda", "Hexapoda", "Hymenoptera", "Apidae", "Bombus",     "flavifrons",                    NA,       NA, "Penstemon confertus", "http://hol.osu.edu/spmInfo.html?id=MTEC%20014090", "2015-09-28T11:15:18Z"
    )
    # Be sure that the testData is not already in tempdir
    testDataPath <- file.info(list.files(tempdir(), full.names = T, 
                                         pattern = "testData.csv", recursive = TRUE))
    unlink(rownames(testDataPath))
    
    # Save a temporary version of these data
    readr::write_excel_csv(testData, paste0(tempdir(), "/testData.csv"))
    
    testOut1 <- BeeBDC::readr_BeeBDC(dataset = "BMont",
                                   path = tempdir(),
                                   inFile = "testData.csv",
                                   outFile = "testDataOut.csv",
                                   dataLicense = "https://creativecommons.org/licenses/by-nc-sa/4.0/")
    
    
    
    # Get a count of TRUE and FALSE column name matches
    resultsT <- sum(colnames(testOut1) %in% (ColTypeR()[[1]] %>% names()) == TRUE)
    resultsF <- sum(colnames(testOut1) %in% (ColTypeR()[[1]] %>% names()) == FALSE)
    
    
    # Test the number of expected TRUE and FALSE columns and then test the output format (data frames and
    # tibbles are a special case of lists)
    testthat::test_that("readr_BMont results columns TRUE", {
      testthat::expect_equal(resultsT, 38)
    })
    testthat::test_that("readr_BMont results columns FALSE", {
      testthat::expect_equal(resultsF, 8)
    })
    
    testthat::test_that("readr_BMont expected class", {
      testthat::expect_type(testOut1, "list")
    })
    
    
    
#### 5.0 Ecd  ####
    
    
    testData <- dplyr::tribble(
      ~id, ~institutionCode, ~collectionCode, ~ownerInstitutionCode,     ~basisOfRecord,                          ~occurrenceID, ~catalogNumber, ~otherCatalogNumbers,                                                                                                         ~higherClassification,   ~kingdom,      ~phylum,    ~class,        ~order,  ~family,        ~scientificName, ~taxonID, ~scientificNameAuthorship,     ~genus, ~subgenus, ~specificEpithet, ~verbatimTaxonRank, ~infraspecificEpithet, ~taxonRank, ~identifiedBy, ~dateIdentified, ~identificationReferences, ~identificationRemarks, ~taxonRemarks, ~identificationQualifier, ~typeStatus,          ~recordedBy, ~associatedCollectors, ~recordNumber, ~eventDate, ~year, ~month, ~day, ~startDayOfYear, ~endDayOfYear, ~verbatimEventDate, ~occurrenceRemarks,                           ~habitat, ~substrate, ~verbatimAttributes, ~fieldNumber, ~eventID, ~informationWithheld, ~dataGeneralizations, ~dynamicProperties, ~associatedOccurrences, ~associatedSequences, ~associatedTaxa, ~reproductiveCondition, ~establishmentMeans, ~cultivationStatus, ~lifeStage, ~sex, ~individualCount, ~preparations,        ~country, ~stateProvince,    ~county, ~municipality,                 ~locality, ~locationRemarks, ~localitySecurity, ~localitySecurityReason, ~decimalLatitude, ~decimalLongitude, ~geodeticDatum, ~coordinateUncertaintyInMeters, ~verbatimCoordinates, ~georeferencedBy, ~georeferenceProtocol, ~georeferenceSources, ~georeferenceVerificationStatus, ~georeferenceRemarks, ~minimumElevationInMeters, ~maximumElevationInMeters, ~minimumDepthInMeters, ~maximumDepthInMeters, ~verbatimDepth, ~verbatimElevation, ~disposition, ~language, ~recordEnteredBy,       ~modified, ~`sourcePrimaryKey-dbpk`, ~collID,                                       ~recordID,                                                                         ~references,
      637436L,            "ASU",   "BIO386-2020",                    NA, "HumanObservation", "bd5eea21-0995-45e6-b626-7a3cb361d878",             NA,                   NA,                   "Animalia|Arthropoda|Hexapoda|Insecta|Pterygota|Neoptera|Hymenoptera|Apocrita|Aculeata|Apoidea|Apidae|Apis", "Animalia", "Arthropoda", "Insecta", "Hymenoptera", "Apidae",       "Apis mellifera",  235783L,          "Linnaeus, 1758",     "Apis",        NA,      "mellifera",                 NA,                    NA,  "Species",            NA,              NA,                        NA,                     NA,            NA,                       NA,          NA, "Melkonoff, Natalie",                    NA,            NA,  "9/26/20", 2020L,     9L,  26L,            270L,            NA,                 NA,                 NA, "On Asclepias angustifolia flower",         NA,                  NA,           NA,       NA,                   NA,                   NA,                 NA,                     NA,                   NA,              NA,                     NA,                  NA,                 0L,         NA,   NA,               NA,            NA, "United States",      "Arizona", "Maricopa",            NA, "Desert Botanical Garden",               NA,                0L,                      NA,        33.460551,       -111.942522,        "WGS84",                            50L,                   NA,               NA,                    NA,                   NA,                              NA,                   NA,                        NA,                        NA,                    NA,                    NA,             NA,                 NA,           NA,        NA,               NA, "9/26/20 16:14",                       NA,     23L, "urn:uuid:bd5eea21-0995-45e6-b626-7a3cb361d878", "https://serv.biokic.asu.edu/ecdysis/collections/individual/index.php?occid=637436",
      637531L,            "ASU",   "BIO386-2020",                    NA, "HumanObservation", "a19fa123-6f6f-4c2a-92c7-f2d8409873b1",             NA,                   NA,  "Animalia|Arthropoda|Hexapoda|Insecta|Pterygota|Neoptera|Hymenoptera|Apocrita|Aculeata|Apoidea|Apidae|Apinae|Bombini|Bombus", "Animalia", "Arthropoda", "Insecta", "Hymenoptera", "Apidae", "Bombus vosneskenskii",  603453L,       "Radoszkowski 1862",   "Bombus",        NA,  "vosneskenskii",                 NA,                    NA,  "Species",            NA,              NA,                        NA,                     NA,            NA,                       NA,          NA,          "Ma, Janey",                    NA,            NA,  "9/21/20", 2020L,     9L,  21L,            265L,            NA,                 NA,      "Hymenoptera",                                 NA,         NA,                  NA,           NA,       NA,                   NA,                   NA,                 NA,                     NA,                   NA,              NA,                     NA,                  NA,                 0L,         NA,   NA,               NA,            NA, "United States",           "AZ", "Maricopa",            NA,                "Glendale",               NA,                0L,                      NA,          33.5771,          112.1906,             NA,                            50L,                   NA,               NA,                    NA,                   NA,                              NA,                   NA,                        NA,                        NA,                    NA,                    NA,             NA,                 NA,           NA,        NA,               NA, "9/26/20 22:36",                       NA,     23L, "urn:uuid:a19fa123-6f6f-4c2a-92c7-f2d8409873b1", "https://serv.biokic.asu.edu/ecdysis/collections/individual/index.php?occid=637531",
      637539L,            "ASU",   "BIO386-2020",                    NA, "HumanObservation", "1d420150-ce0f-4510-bdf9-e65125621e43",             NA,                   NA, "Animalia|Arthropoda|Hexapoda|Insecta|Pterygota|Neoptera|Hymenoptera|Apocrita|Aculeata|Apoidea|Apidae|Xylocopinae|Xylocopini", "Animalia", "Arthropoda", "Insecta", "Hymenoptera", "Apidae",             "Xylocopa",  204218L,         "Latreille, 1802", "Xylocopa",        NA,               NA,                 NA,                    NA,    "Genus",            NA,              NA,                        NA,                     NA,            NA,                       NA,          NA,          "Ma, Janey",                    NA,            NA,  "9/21/20", 2020L,     9L,  21L,            265L,            NA,                 NA,    "Hymenopterans",                                 NA,         NA,                  NA,           NA,       NA,                   NA,                   NA,                 NA,                     NA,                   NA,              NA,                     NA,                  NA,                 0L,         NA,   NA,               NA,            NA, "United States",      "Arizona", "Maricopa",            NA,                "Glendale",               NA,                0L,                      NA,          33.5771,           112.186,             NA,                            50L,                   NA,               NA,                    NA,                   NA,                              NA,                   NA,                        NA,                        NA,                    NA,                    NA,             NA,                 NA,           NA,        NA,               NA, "9/26/20 22:51",                       NA,     23L, "urn:uuid:1d420150-ce0f-4510-bdf9-e65125621e43", "https://serv.biokic.asu.edu/ecdysis/collections/individual/index.php?occid=637539",
      637542L,            "ASU",   "BIO386-2020",                    NA, "HumanObservation", "c56b011d-70f7-4ee8-b0b9-f37a32fb5cd4",             NA,                   NA,                               "Animalia|Arthropoda|Hexapoda|Insecta|Pterygota|Neoptera|Hymenoptera|Apocrita|Aculeata|Apoidea", "Animalia", "Arthropoda", "Insecta", "Hymenoptera",       NA,         "Megachilidae",   91998L,                        NA,         NA,        NA,               NA,                 NA,                    NA,   "Family",            NA,              NA,                        NA,                     NA,            NA,                       NA,          NA,          "Ma, Janey",                    NA,            NA,  "9/24/20", 2020L,     9L,  24L,            268L,            NA,                 NA,      "Hymenoptera",                                 NA,         NA,                  NA,           NA,       NA,                   NA,                   NA,                 NA,                     NA,                   NA,              NA,                     NA,                  NA,                 0L,         NA,   NA,               NA,            NA, "United States",           "AZ", "Maricopa",            NA,                "Glendale",               NA,                0L,                      NA,          33.5771,           112.186,             NA,                            50L,                   NA,               NA,                    NA,                   NA,                              NA,                   NA,                        NA,                        NA,                    NA,                    NA,             NA,                 NA,           NA,        NA,               NA, "9/26/20 22:53",                       NA,     23L, "urn:uuid:c56b011d-70f7-4ee8-b0b9-f37a32fb5cd4", "https://serv.biokic.asu.edu/ecdysis/collections/individual/index.php?occid=637542"
    )
    
    # Be sure that the testData is not already in tempdir
    testDataPath <- file.info(list.files(tempdir(), full.names = T, 
                                         pattern = "testData.csv", recursive = TRUE))
    unlink(rownames(testDataPath))
    
    # Save a temporary version of these data
    readr::write_excel_csv(testData, paste0(tempdir(), "/testData.csv"))
    
    
    testOut1 <- BeeBDC::readr_BeeBDC(dataset = "Ecd",
                                   path = tempdir(),
                                 inFile = "testData.csv",
                                 outFile = "testDataOut.csv",
                                 dataLicense = "https://creativecommons.org/licenses/by-nc-sa/4.0/")
    
    
    # Get a count of TRUE and FALSE column name matches
    resultsT <- sum(colnames(testOut1) %in% (ColTypeR()[[1]] %>% names()) == TRUE)
    resultsF <- sum(colnames(testOut1) %in% (ColTypeR()[[1]] %>% names()) == FALSE)
    
    
    # Test the number of expected TRUE and FALSE columns and then test the output format (data frames and
    # tibbles are a special case of lists)
    testthat::test_that("readr_Ecd results columns TRUE", {
      testthat::expect_equal(resultsT, 58)
    })
    
    testthat::test_that("readr_Ecd results columns FALSE", {
      testthat::expect_equal(resultsF, 38)
    })
    
    testthat::test_that("readr_Ecd expected class", {
      testthat::expect_type(testOut1, "list")
    })
    
    
    
#### 6.0 Gai ####
    
    testData <- dplyr::tribble(
      ~institutionCode,	~Collection.Code,	~collection.var,	     ~catalogNumber,	   ~Other.Catalog.Number,	~sex,	~identifiedBy,	~dateIdentified,	~recordedBy,	~SpecimenLocation,	~BeeNonbee,	~order,	       ~family,	   ~genus,      	~subgenus,	          ~species,	     ~subspecies,	     ~scientificName,            	    ~associatedTaxa,	~locality,	          ~country,	~eventDate,	~SampleRound,	~SiteStatus,	~samplingProtocol,	~eventTime,	~EndTime,	~TempStart,	~TempEnd,	~WindStart,	~WindEnd,	~SkyStart,	~SkyEnd,	~Site,	~siteLocality,	~countryCode,	~stateProvince,	~county,	~decimalLatitude,	~decimalLongitude,	~year,	       ~syd,	       ~verbatimIdentification,	                   ~SiteStatusBACI,	~ypr,	~syd.veg,        	~associatedTaxa2,	~basisOfRecord,
      "EMEC-UTB",       "ENT",            "kremen-baci",   "M2006SR3DQU_30015", 30015L,                 NA, "C. Kremen",      "7/6/06",        "C. Kremen",    "UBC",             "bee",  "Hymenoptera",  "Apidae",   "Bombus",          NA,              "californicus",      NA,          "Bombus californicus",            "Rubus discolor",  "North America",  "United States", "7/6/06",   3L,         "control",       "net",        "10:41:00", "12:36:00",    29.2,    31.6,      0,           1,      "clear",  "clear",     "DQU",  "C5a",            "US",  "California",  "Yolo",              38.33687, -121.53191,            2006L,   "DQU;2006;187",    "Bombus californicus",                        "control",      0L,   "DQU;2006;187",  "Rubus discolor",  "preserved specimen",
      "EMEC-UTB",      "ENT",           "kremen-baci",  "M2006SR3DQU_30016",   30016L,                 NA, "C. Kremen",     "7/6/06",         "C. Kremen",   "UBC",              "bee", "Hymenoptera",     "Apidae",   "Bombus",         NA,          "californicus",        NA,          "Bombus californicus",            "Rubus discolor", "North America",    "United States",  "7/6/06",   3L,       "control",        "net",        "10:41:00",  "12:36:00",   29.2,    31.6,       0,          1,      "clear", "clear",    "DQU", "C5a",               "US", "California", "Yolo",                38.33687,    -121.53191,         2006L,    "DQU;2006;187",    "Bombus californicus",                     "control",        0L,    "DQU;2006;187", "Rubus discolor", "preserved specimen",
      "EMEC-UTB",     "ENT",             "kremen-baci", "M2007SR2Barger_007",   20398L,                "f", "J. Gibbs",     "7/2/08",          "K. Ullmann", "ESSIG",              "bee", "Hymenoptera", "Halictidae", "Lasioglossum",  "(Dialictus)", "diversopunctatum",    NA,          "Lasioglossum diversopunctatum",   "Brassica sp.", "North America",      "United States", "6/20/07",   2L,      "control",        "net",         "9:49:00",  "11:45:00",    21.4,    24.4,     1.2,        2.6,     "clear", "clear", "Barger",  "H1",               "US", "California", "Yolo",                38.34575,    -121.49834,         2007L, "Barger;2007;171", "Lasioglossum (Dialictus) diversopunctatum", "hedgerow",        0L, "Barger;2007;171",   "Brassica sp.", "preserved specimen",
      "EMEC-UTB",    "ENT",             "kremen-baci", "M2007SR2Barger_008",   20399L,                 "f",  "J. Gibbs",     "7/2/08",         "K. Ullmann", "ESSIG",             "bee", "Hymenoptera", "Halictidae", "Lasioglossum", "(Dialictus)", "diversopunctatum",     NA,        "Lasioglossum diversopunctatum",      "Brassica sp.", "North America",    "United States", "6/20/07",   2L,       "control",         "net",       "9:49:00",  "11:45:00",    21.4,    24.4,     1.2,          2.6,    "clear", "clear", "Barger",  "H1",               "US", "California", "Yolo",                38.34575,    -121.49834,         2007L, "Barger;2007;171", "Lasioglossum (Dialictus) diversopunctatum", "hedgerow",       0L, "Barger;2007;171",   "Brassica sp.", "preserved specimen"   
    )
    
    # need to change column names, which need spaces for function to work
    colnames(testData) <- c("institutionCode", "Collection Code", "collection.var", "catalogNumber", "Other Catalog Number", "sex",	"identifiedBy",	"dateIdentified",	"recordedBy",	"SpecimenLocation",	"BeeNonbee", "order", "family", "genus", "subgenus", "species", "subspecies", "scientificName", "associatedTaxa", "locality", "country", "eventDate", "SampleRound", "SiteStatus", "samplingProtocol", "eventTime",	"EndTime", "TempStart", "TempEnd", "WindStart", "WindEnd", "SkyStart", "SkyEnd", "Site", "siteLocality", "countryCode", "stateProvince", "county", "decimalLatitude", "decimalLongitude", "year", "syd", "verbatimIdentification", "SiteStatusBACI", "ypr", "syd.veg", "associatedTaxa2",	"basisOfRecord")
    
    # Be sure that the testData is not already in tempdir
    testDataPath <- file.info(list.files(tempdir(), full.names = T, 
                                         pattern = "testData.csv", recursive = TRUE))
    unlink(rownames(testDataPath))
    
    # Save a temporary version of these data
    readr::write_excel_csv(testData, paste0(tempdir(), "/testData.csv"))
    
    testOut1 <- BeeBDC::readr_BeeBDC(dataset = "Gai",
                                   path = tempdir(),
                                 inFile = "testData.csv",
                                 outFile = "testDataOut.csv",
                                 dataLicense = "https://creativecommons.org/licenses/by-nc-sa/4.0/")
    
    
    # Get a count of TRUE and FALSE column name matches
    resultsT <- sum(colnames(testOut1) %in% (ColTypeR()[[1]] %>% names()) == TRUE)
    resultsF <- sum(colnames(testOut1) %in% (ColTypeR()[[1]] %>% names()) == FALSE)
    
    
    # Test the number of expected TRUE and FALSE columns and then test the output format (data frames and
    # tibbles are a special case of lists)
    testthat::test_that("readr_Gai results columns TRUE", {
      testthat::expect_equal(resultsT, 34)
    })
    
    testthat::test_that("readr_Gai results columns FALSE", {
      testthat::expect_equal(resultsF, 20)
    })
    
    testthat::test_that("readr_Gai expected class", {
      testthat::expect_type(testOut1, "list")
    })
    



#### 7.0 CAES  ####
    
    testData <- dplyr::tribble(
      ~PBIUSI,    ~Genus, ~species, ~Country,   ~State_Prov, ~Sec_Subdiv,                            ~Locality,     ~Lat,      ~Lon, ~Start_Date, ~End_Date,        ~Collector,                ~Sex, ~Inst_Code,                                                                                                                                                                                    ~Project,  ~Type,         ~Det_By, ~Det_Date, ~Det_History,    ~Loc_Notes, ~Accuracy, ~Coll_Method, ~Spec_Notes, ~Host_Family, ~Host_Genus, ~Host_species, ~Pres_Method, ~Spec_Count, ~Lat_Lon_Method, ~Lat_Lon_Accuracy, ~Elev_m, ~Elev_f, ~Elev_Det, ~Trip_Code, ~Dissections, ~Illustrations, ~Measurements, ~Photos, ~SEM, ~DNA, ~OrigUSI,      ~Family,   ~Subfamily,      ~Tribe, ~Macro_Habitat, ~Micro_Habitat, ~Host_subspecies, ~Host_Author, ~Host_Common_Name,    ~Host_Relation, ~Host_Location, ~Host_Emergence_Date, ~Host_with_Specimen,
      "AMNH_BEE 000266235443", "Secret", "species",    "USA", "Connecticut", "Fairfield",                         "New Canaan", 41.14666, -73.49472, 22786,        NA,      "M. Statham",      "Adult Female",     "AMNH", "Data citation: Digital Bee Collections Network, 2014 (and updates). Version: <download date>. National Science Foundation grant DBI 0956388; Contact: Ascher, John S. (dbsajs@nus.edu.sg)", "None",  "J. S. Ascher",     2007L,           NA, "GNIS coord.",        NA,    "Netting",          NA,           NA,          NA,            NA,     "Pinned",          1L,     "Gazetteer",                NA,     97L,    318L, "Unknown",         NA,         "NO",           "NO",          "NO",    "NO", "NO",   NA,       NA, "Andrenidae", "Andreninae", "Andrenini",             NA,             NA,               NA,           NA,                NA,                NA,             NA,                   NA,                "NO",
      "AMNH_BEE 00026254675", "Secret", "species",    "USA", "Connecticut", "Fairfield",            "Danbury, I-84 at exit 2", 41.39247, -73.52731,  38451,        NA,    "J. S. Ascher",        "Adult Male",     "AMNH", "Data citation: Digital Bee Collections Network, 2014 (and updates). Version: <download date>. National Science Foundation grant DBI 0956388; Contact: Ascher, John S. (dbsajs@nus.edu.sg)", "None",  "J. S. Ascher",     2007L,           NA,            NA,        NA,    "Netting",          NA, "Salicaceae",     "Salix",        "spp.",     "Pinned",          1L,              NA,                NA,      NA,      NA,        NA,         NA,         "NO",           "NO",          "NO",    "NO", "NO",   NA,       NA, "Andrenidae", "Andreninae", "Andrenini",             NA,             NA,               NA,           NA,                NA, "associated with",             NA,                   NA,                "NO",
      "UCMS_ENT 00022347759", "Secret", "species",    "USA", "Connecticut",   "Tolland", "Coventry Twp., near Eagleville Dam", 41.76916, -72.30444, 26414,        NA, "L. R. Schechter", "Adult sex unknown",      "GSC", "Data citation: Digital Bee Collections Network, 2014 (and updates). Version: <download date>. National Science Foundation grant DBI 0956388; Contact: Ascher, John S. (dbsajs@nus.edu.sg)", "None", "W. E. LaBerge",        NA,           NA,            NA,        NA,    "Unknown",          NA,           NA,          NA,            NA,     "Pinned",          1L,              NA,                NA,      NA,      NA,        NA,         NA,         "NO",           "NO",          "NO",    "NO", "NO",   NA,       NA, "Andrenidae", "Andreninae", "Andrenini",             NA,             NA,               NA,           NA,                NA,                NA,             NA,                   NA,                "NO",
      "UCMS_ENT 00023457760", "Secret", "species",    "USA", "Connecticut",   "Tolland",        "Mansfield Twp., Gurleyville", 41.81416,   -72.255, 26044,        NA,     "G. I. Stage", "Adult sex unknown",      "GSC", "Data citation: Digital Bee Collections Network, 2014 (and updates). Version: <download date>. National Science Foundation grant DBI 0956388; Contact: Ascher, John S. (dbsajs@nus.edu.sg)", "None", "W. E. LaBerge",        NA,           NA,            NA,        NA,    "Unknown",          NA,           NA,          NA,            NA,     "Pinned",          1L,              NA,                NA,      NA,      NA,        NA,         NA,         "NO",           "NO",          "NO",    "NO", "NO",   NA,       NA, "Andrenidae", "Andreninae", "Andrenini",             NA,             NA,               NA,           NA,                NA,                NA,             NA,                   NA,                "NO",
      "UCMS_ENT 002345027762", "Secret", "species",    "USA", "Connecticut",   "Tolland", "Coventry Twp., near Eagleville Dam", 41.76916, -72.30444, 26417,        NA,    "J. R. Gordon", "Adult sex unknown",      "GSC", "Data citation: Digital Bee Collections Network, 2014 (and updates). Version: <download date>. National Science Foundation grant DBI 0956388; Contact: Ascher, John S. (dbsajs@nus.edu.sg)", "None", "W. E. LaBerge",        NA,           NA,            NA,        NA,    "Unknown",          NA,           NA,          NA,            NA,     "Pinned",          1L,              NA,                NA,      NA,      NA,        NA,         NA,         "NO",           "NO",          "NO",    "NO", "NO",   NA,       NA, "Andrenidae", "Andreninae", "Andrenini",             NA,             NA,               NA,           NA,                NA,                NA,             NA,                   NA,                "NO"
    )
    
    
    # Be sure that the testData is not already in tempdir
    testDataPath <- file.info(list.files(tempdir(), full.names = T, 
                                         pattern = "testData.xlsx", recursive = TRUE))
    unlink(rownames(testDataPath))
    
    # Save a temporary version of these data
    openxlsx::write.xlsx(testData, paste0(tempdir(), "/testData.xlsx"),
                         sheetName = "Sheet1")
    
    testOut1 <- BeeBDC::readr_BeeBDC(dataset = "CAES",
                                     path = tempdir(),
                                  inFile = "testData.xlsx",
                                  outFile = "testDataOut.csv",
                                  sheet = "Sheet1",
                                  dataLicense = "https://creativecommons.org/licenses/by-nc-sa/4.0/")
    
    
    
    # Get a count of TRUE and FALSE column name matches
    resultsT <- sum(colnames(testOut1) %in% (ColTypeR()[[1]] %>% names()) == TRUE)
    resultsF <- sum(colnames(testOut1) %in% (ColTypeR()[[1]] %>% names()) == FALSE)
    
    # Test the number of expected TRUE and FALSE columns and then test the output format (data frames and
    # tibbles are a special case of lists)
    testthat::test_that("readr_BBD results columns TRUE", {
      testthat::expect_equal(resultsT, 31)
    })
    testthat::test_that("readr_BBD results columns FALSE", {
      testthat::expect_equal(resultsF, 16)
    })
    
    testthat::test_that("readr_BBD expected class", {
      testthat::expect_type(testOut1, "list")
    })
    
    
    
    

#### 9.0 KP  ####
    
    # data below is FAKE
    testData <- dplyr::tribble(
      ~ID, ~institutionCode,             ~Det, ~Number, ~Male, ~Female, ~Collection_method,          ~Collection_date,  ~Collector,        ~Order,    ~Suborder,        ~Family,     ~Subfamily,        ~Tribe,         ~Genus,         ~subgenus, ~sp_group,      ~species, ~subspecies,            ~author,                             ~whole_sci_name, ~Country,     ~State,         ~County_Parish,                        ~Location,                 ~Lat,               ~Long,
      "discoverlife fake1 USGS-DRO. fake1",           "USGS",      "S. Droege",      1L,    0L,      1L,              "Net", "2019-08-07 00:00:00 UTC", "J. Fowler", "Hymenoptera", "Anthophila",       "Apidae",             NA,            NA,         "Apis",                NA,        NA,   "mellifera",          NA,   "Linnaeus, 1758",             "Apis mellifera Linnaeus, 1758",    "USA",       "NY",            "Kings Co.",           "Brooklyn Bridge Park",              "40.69",                 -73,
      "discoverlife fake2 USGS-DRO. fake2",           "USGS",      "S. Droege",      1L,    1L,      0L,              "Net", "2019-09-06 00:00:00 UTC", "J. Fowler", "Hymenoptera", "Anthophila",       "Apidae",             NA,            NA,         "Apis",                NA,        NA,   "mellifera",          NA,   "Linnaeus, 1758",             "Apis mellifera Linnaeus, 1758",    "USA",       "NY",            "Kings Co.",           "Brooklyn Bridge Park",              "40.69",                 -73,
      "discoverlife fake3 USGS-DRO. fake3",           "USGS", "S. Droege 2019",      1L,    1L,      0L,              "Net", "2019-04-18 00:00:00 UTC", "S. Droege", "Hymenoptera", "Anthophila",       "Apidae",             NA,            NA,         "Apis",                NA,        NA,   "mellifera",          NA,   "Linnaeus, 1758",             "Apis mellifera Linnaeus, 1758",    "USA",       "NY",            "Kings Co.",           "Brooklyn Bridge Park",              "40.69",                 -73,
      "discoverlife fake4 USGS-DRO. fake4",           "USGS",      "S. Droege",      1L,    1L,      0L,              "Net", "2016-03-25 00:00:00 UTC", "S. Droege", "Hymenoptera", "Anthophila",       "Apidae",             NA,            NA,         "Apis",                NA,        NA,   "mellifera",          NA,   "Linnaeus, 1758",             "Apis mellifera Linnaeus, 1758",    "USA",       "NY",            "Kings Co.",           "Brooklyn Bridge Park",              "40.69",                 -73
    )
    
    # Be sure that the testData is not already in tempdir
    testDataPath <- file.info(list.files(tempdir(), full.names = T, 
                                         pattern = "testData.xlsx", recursive = TRUE))
    unlink(rownames(testDataPath))
    
    # Save a temporary version of these data
    openxlsx::write.xlsx(testData, paste0(tempdir(), "/testData.xlsx"))
    
    
    testOut1 <- BeeBDC::readr_BeeBDC(dataset = "KP",
                                   path = tempdir(),
                                inFile = "testData.xlsx",
                                outFile = "testDataOut.csv",
                                dataLicense = "All rights reserved")
    
    
    # Get a count of TRUE and FALSE column name matches
    resultsT <- sum(colnames(testOut1) %in% (ColTypeR()[[1]] %>% names()) == TRUE)
    resultsF <- sum(colnames(testOut1) %in% (ColTypeR()[[1]] %>% names()) == FALSE)
    
    
    # Test the number of expected TRUE and FALSE columns and then test the output format (data frames and
    # tibbles are a special case of lists)
    testthat::test_that("readr_KP results columns TRUE", {
      testthat::expect_equal(resultsT, 30)
    })
    
    testthat::test_that("readr_KP results columns FALSE", {
      testthat::expect_equal(resultsF, 1)
    })
    
    testthat::test_that("readr_KP expected class", {
      testthat::expect_type(testOut1, "list")
    })
    
    
    
#### 11.0 EcoS  ####
    
    testData <- dplyr::tribble(
      ~Collection, ~ID_project,                    ~Species, ~Longitude, ~Latitude, ~Year,
      "CECON",          1L, "Agapostemon_atrocaeruleus",    -89.439,    16.357, 2006L,
      "CECON",          2L,     "Agapostemon_leunculus",    -89.377,     14.62, 2003L,
      "CECON",          3L,     "Agapostemon_leunculus",    -90.211,    15.266, 2011L,
      "CECON",          4L,  "Agapostemon_melliventris",     -90.06,    14.926, 2004L,
      "CECON",          5L,     "Anthidium_maculifrons",    -91.492,    15.332, 2009L,
      "CECON",          6L,     "Anthidium_maculifrons",    -89.363,    14.607, 2003L
    )
    
    # Be sure that the testData is not already in tempdir
    testDataPath <- file.info(list.files(tempdir(), full.names = T, 
                                         pattern = "testData.csv", recursive = TRUE))
    unlink(rownames(testDataPath))
    
    # Save a temporary version of these data
    readr::write_excel_csv(testData, paste0(tempdir(), "/testData.csv"))
    
    testOut1 <- BeeBDC::readr_BeeBDC(dataset = "EcoS",
                                  path = tempdir(),
                                  inFile = "testData.csv",
                                  outFile = "testDataOut.csv",
                                  dataLicense = "https://creativecommons.org/licenses/by-nc-sa/4.0/")
    
    
    # Get a count of TRUE and FALSE column name matches
    resultsT <- sum(colnames(testOut1) %in% (ColTypeR()[[1]] %>% names()) == TRUE)
    resultsF <- sum(colnames(testOut1) %in% (ColTypeR()[[1]] %>% names()) == FALSE)
    
    
    # Test the number of expected TRUE and FALSE columns and then test the output format (data frames and
    # tibbles are a special case of lists)
    testthat::test_that("readr_EcoS results columns TRUE", {
      testthat::expect_equal(resultsT, 14)
    })
    
    testthat::test_that("readr_EcoS results columns FALSE", {
      testthat::expect_equal(resultsF, 0)
    })
    
    testthat::test_that("readr_EcoS expected class", {
      testthat::expect_type(testOut1, "list")
    })
    
    
    
#### 12.0 GeoL  ####
    
    testData <- dplyr::tribble(
      ~database_id,                              ~scientificName,      ~family,   ~subfamily,          ~genus, ~subgenus, ~subspecies,                     ~species, ~specificEpithet, ~infraspecificEpithet, ~acceptedNameUsage, ~taxonRank, ~scientificNameAuthorship,      ~continent,        ~country, ~country_suggested, ~countryCode, ~stateProvince,         ~county, ~municipality,                                                                   ~locality, ~island,                                             ~license,                                                                                                     ~issue,             ~eventDate, ~day, ~month, ~year,       ~basisOfRecord,            ~type, ~occurrenceStatus, ~recordNumber,                      ~recordedBy, ~eventID, ~samplingProtocol, ~samplingEffort, ~individualCount,                  ~catalogNumber, ~rightsHolder, ~institutionCode, ~decimallatitude, ~decimallongitude, ~datasetID, ~datasetName, ~otherCatalogNumbers,                                                          ~occurrenceID, ~coreid,                                       ~recordId,                          ~collectionID,                             ~verbatimScientificName,      ~verbatimEventDate,           ~sex,                                          ~rights, ~occurrenceYear,           ~id,       ~dataSource,                 ~names_clean,                     ~verbatim_scientificName, ~geolocate_Latitude, ~geolocate_Longitude, ~geolocate_UncertaintyRadiusMeters, ~geolocate_Score, ~geolocate_Precision,                  ~geolocate_ParsePattern, ~geolocate_locFieldUsed, ~geolocate_NumResults,
      "Ecd_data_20840",          "Agapostemon texanus Cresson, 1872", "Halictidae", "Halictinae",   "Agapostemon",        NA,          NA,        "Agapostemon texanus",        "texanus",                    NA,                 NA,  "Species",           "Cresson, 1872",              NA, "United states",    "United states",         "US",   "California", "Santa Barbara",            NA,                                   "University of California, Santa Barbara",      NA, "https://creativecommons.org/licenses/by-nc-sa/4.0/",                                                                                                         NA, "1984-05-17T00:00:00Z",  17L,     5L, 1984L,  "PreservedSpecimen",               NA,                NA,            NA,                               NA,       NA,                NA,              NA,               1L,              "UCSB-IZC00005603",            NA,           "UCSB",               NA,                NA,         NA,           NA,                   NA,                                 "3de7811e-1169-4fe9-a451-59ba9c653449",      NA,                                              NA,                                     NA,                                                  NA,             "17/5/1984",         "Male",                                               NA,              NA, "UCSB_828348",  "Ecd_Anthophila",                           NA,          "Agapostemon texanus Cresson, 1872",           34.416323,          -119.846392,                                84L,             100L,               "High", "UNIVERSITY OF CALIFORNIA SANTA BARBARA",              "locality",                   15L,
      "Dorey_data_761973",       "Perdita trifasciata Timberlake, 1953", "Andrenidae", "Panurginae",       "Perdita",        NA,          NA,        "Perdita trifasciata",    "trifasciata",                    NA,                 NA,  "SPECIES",        "Timberlake, 1953",              NA, "United states",    "United states",         "US",   "New mexico",              NA,            NA,                                                         "Cottonwood Spring",      NA,                                            "CC0_1_0",                                                         "OCCURRENCE_STATUS_INFERRED_FROM_INDIVIDUAL_COUNT", "1990-09-03T00:00:00Z",   3L,     9L, 1990L,  "MATERIAL_CITATION",               NA,         "PRESENT",            NA, "T. L. Griswold & Walnut Canyon",       NA,                NA,              NA,               1L,                              NA,            NA,               NA,               NA,                NA,         NA,           NA,                   NA, "03DA51566E5A3810FF43FF7FFDD46C19.mc.3B1BEA1D6E583810FC56FF08FC936808",      NA,                                              NA,                                     NA,                    "Perdita trifasciata Timberlake",                      NA,       "FEMALE",                                               NA,              NA,            NA, "GBIF_Andrenidae",        "Perdita trifasciata",       "Perdita trifasciata Timberlake, 1953",           36.876388,           -108.39286,                                90L,              92L,               "High",                      "COTTONWOOD SPRING",              "locality",                   33L,
      "Dorey_data_13971366", "Augochlorella pomoniella (Cockerell, 1915)", "Halictidae", "Halictinae", "Augochlorella",        NA,          NA,   "Augochlorella pomoniella",     "pomoniella",                    NA,                 NA,         NA,       "(Cockerell, 1915)",              NA, "United States",    "United States",         "US",   "California",          "Inyo",            NA,                   "17 mi E Big Pine, Death Valley Road, pinon juniper zone",      NA,                                                   NA,                                                                                                         NA, "1992-06-21T00:00:00Z",  21L,     6L, 1992L,  "PreservedSpecimen",               NA,                NA,            NA,        "M.E. Irwin, D.K. Yeates",       NA,                NA,              NA,               1L, "INHS Insect Collection 352770",            NA,           "INHS",               NA,                NA,         NA,           NA,                   NA,                                                              "3247430",      NA, "urn:uuid:ab6083b7-ad01-4537-a7c4-e1cff7f2172c", "d93d943b-2390-4e2c-9050-11a9ec9a2a96",                                                  NA, "1992-06-21/1992-06-24",       "Female", "http://creativecommons.org/licenses/by-nc/3.0/",              NA,    "58444644", "SCAN_Halictidae",                           NA, "Augochlorella pomoniella (Cockerell, 1915)",           37.142698,          -118.123715,                                90L,             100L,               "High",     "DEATH VALLEY RD AT DEATH VALLEY RD",              "locality",                    2L,
      "CAES_data_4284",       "Andrena wellesleyana Robertson, 1897", "Andrenidae", "Andreninae",       "Andrena",        NA,          NA,       "Andrena wellesleyana",   "wellesleyana",                    NA,                 NA,         NA,         "Robertson, 1897",              NA, "United states",    "United states",         "US",  "Connecticut",     "New Haven",            NA, "North Haven, 0.5 km NNW of interchange 12 (US Highway 5) on Interstate 91",      NA, "https://creativecommons.org/licenses/by-nc-sa/4.0/",                                                                                                         NA, "1998-03-31T00:00:00Z",  31L,     3L, 1998L,                   NA,               NA,                NA,            NA,                 "Chris T. Maier",       NA,         "Netting",              NA,               1L,             "UCMS_ENT 00076127",            NA,           "CAES",               NA,                NA,         NA,           NA,                   NA,                                                                     NA,      NA,                                              NA,                                     NA, "Andrena (Parandrena) wellesleyana Robertson, 1910",                      NA, "Adult Female",                                               NA,              NA,            NA, "CAES_Anthophila",       "Andrena wellesleyana",       "Andrena wellesleyana Robertson, 1897",           41.403421,           -72.853785,                               170L,              89L,               "High",         "Distance NNW of INTERCHANGE 12",              "locality",                    4L,
      "CAES_data_4285",       "Andrena wellesleyana Robertson, 1897", "Andrenidae", "Andreninae",       "Andrena",        NA,          NA,       "Andrena wellesleyana",   "wellesleyana",                    NA,                 NA,         NA,         "Robertson, 1897",              NA, "United states",    "United states",         "US",  "Connecticut",     "New Haven",            NA, "North Haven, 0.5 km NNW of interchange 12 (US Highway 5) on Interstate 91",      NA, "https://creativecommons.org/licenses/by-nc-sa/4.0/",                                                                                                         NA, "1997-04-21T00:00:00Z",  21L,     4L, 1997L,                   NA,               NA,                NA,            NA,                 "Chris T. Maier",       NA,         "Netting",              NA,               1L,             "UCMS_ENT 00076128",            NA,           "CAES",               NA,                NA,         NA,           NA,                   NA,                                                                     NA,      NA,                                              NA,                                     NA, "Andrena (Parandrena) wellesleyana Robertson, 1911",                      NA, "Adult Female",                                               NA,              NA,            NA, "CAES_Anthophila",       "Andrena wellesleyana",       "Andrena wellesleyana Robertson, 1897",           41.403421,           -72.853785,                               170L,              89L,               "High",         "Distance NNW of INTERCHANGE 12",              "locality",                    4L,
      "Dorey_data_1010149", "Acamptopoeum submetallicum (Spinola, 1851)", "Andrenidae", "Panurginae",  "Acamptopoeum",        NA,          NA, "Acamptopoeum submetallicum",  "submetallicum",                    NA,                 NA,  "SPECIES",         "(Spinola, 1851)", "SOUTH_AMERICA",         "Chile",            "Chile",         "CL",        "Maule",         "Talca",            NA,                                                                 "Rio Claro",      NA,                                       "CC_BY_NC_4_0", "OCCURRENCE_STATUS_INFERRED_FROM_INDIVIDUAL_COUNT;INSTITUTION_MATCH_FUZZY;INSTITUTION_COLLECTION_MISMATCH", "1980-11-28T00:00:00Z",  28L,    11L, 1980L, "PRESERVED_SPECIMEN", "PhysicalObject",         "PRESENT",            NA,                               NA,       NA,                NA,              NA,               1L,                          "1291",            NA,           "PUCV",               NA,                NA,         NA,           NA,                   NA,                                                       "PUCV:ENTO:1291",      NA,                                              NA,                                     NA,                        "Acamptopoeum submetallicum",                      NA,       "FEMALE",                                               NA,              NA,            NA, "GBIF_Andrenidae", "Acamptopoeum submetallicum", "Acamptopoeum submetallicum (Spinola, 1851)",          -34.997203,           -70.818563,                               210L,              84L,               "High",                              "RIO CLARO",              "locality",                    4L
    )
    
    testData2 <- dplyr::tribble(
      ~database_id,                            ~scientificname,      ~family,   ~subfamily,         ~genus, ~subgenus, ~infraspecificepithet,         ~specificepithet, ~`_specificepithet`, ~`_infraspecificepithet`, ~acceptednameusage, ~taxonrank, ~scientificnameauthorship, ~unnamed_column_1, ~continent,        ~country, ~country_suggested, ~countrycode,  ~stateprovince,  ~county, ~municipality,                    ~locality, ~island,    ~license,                                                                  ~issue,             ~eventdate, ~day, ~month, ~year,       ~basisofrecord, ~type, ~occurrencestatus, ~recordnumber,   ~recordedby, ~eventid, ~samplingprotocol, ~samplingeffort, ~individualcount,             ~catalognumber, ~rightsholder, ~institutioncode, ~unnamed_column_2, ~decimallatitude, ~decimallongitude, ~unnamed_column_3, ~datasetid, ~datasetname, ~othercatalognumbers,                          ~occurrenceid, ~coreid,                                       ~recordid,                          ~collectionid,                        ~verbatimscientificname, ~verbatimeventdate,     ~sex,                                          ~rights, ~unnamed_column_4, ~occurrenceyear,       ~id,       ~datasource, ~unnamed_column_5,      ~names_clean,                   ~verbatim_scientificname, ~bels_match_country, ~bels_interpreted_countrycode,                  ~bels_matchwithcoords,              ~bels_matchverbatimcoords,                  ~bels_matchsanscoords, ~bels_decimallatitude, ~bels_decimallongitude, ~bels_geodeticdatum, ~bels_coordinateuncertaintyinmeters,   ~bels_georeferencedby, ~bels_georeferenceddate,                                                                        ~bels_georeferenceprotocol,       ~bels_georeferencesources, ~bels_georeferenceremarks, ~bels_georeference_score, ~bels_georeference_source, ~bels_best_of_n_georeferences,              ~bels_match_type,
      "Dorey_data_7221776",              "Augochlora pura (Say, 1837)", "Halictidae", "Halictinae",   "Augochlora",        NA,                    NA,        "Augochlora pura",              "pura",                       NA,                 NA,  "SPECIES",             "(Say, 1837)",                NA,         NA, "United states",    "United states",         "US", "West Virginia",       NA,            NA, "Coopers Rock State Forest,",      NA, "CC_BY_4_0", "TAXON_MATCH_HIGHERRANK;INSTITUTION_MATCH_FUZZY;COLLECTION_MATCH_FUZZY", "1987-08-17T00:00:00Z",  17L,     8L, 1987L, "PRESERVED_SPECIMEN",    NA,         "PRESENT",            NA,            NA,       NA,                NA,              NA,               NA, "Insect Collection 358298",            NA,           "INHS",                NA,               NA,                NA,        436516747L,         NA,           NA,                   NA,                                     NA,      NA,                                              NA,                                     NA, "Augochlora (Augochlora) pura pura (Say 1837)",                 NA,       NA,                                               NA,                NA,              NA,        NA, "GBIF_Halictidae",                NA, "Augochlora pura",              "Augochlora pura (Say, 1837)",                "US",                          "US", "uswestvirginiacoopersrockstateforest", "uswestvirginiacoopersrockstateforest", "uswestvirginiacoopersrockstateforest",             39.664554,             -79.773401,         "epsg:4326",                                  6L,                      NA,                      NA,                                                                                                NA,                              NA,                        NA,                       0L,                    "GBIF",                            2L, "match using verbatim coords",
      "Dorey_data_7242038",              "Augochlora pura (Say, 1837)", "Halictidae", "Halictinae",   "Augochlora",        NA,                    NA,        "Augochlora pura",              "pura",                       NA,                 NA,  "SPECIES",             "(Say, 1837)",                NA,         NA, "United states",    "United states",         "US", "West Virginia",       NA,            NA, "Coopers Rock State Forest,",      NA, "CC_BY_4_0", "TAXON_MATCH_HIGHERRANK;INSTITUTION_MATCH_FUZZY;COLLECTION_MATCH_FUZZY", "1987-08-17T00:00:00Z",  17L,     8L, 1987L, "PRESERVED_SPECIMEN",    NA,         "PRESENT",            NA,            NA,       NA,                NA,              NA,               NA, "Insect Collection 358299",            NA,           "INHS",                NA,               NA,                NA,        436516760L,         NA,           NA,                   NA,                                     NA,      NA,                                              NA,                                     NA, "Augochlora (Augochlora) pura pura (Say 1837)",                 NA,       NA,                                               NA,                NA,              NA,        NA, "GBIF_Halictidae",                NA, "Augochlora pura",              "Augochlora pura (Say, 1837)",                "US",                          "US", "uswestvirginiacoopersrockstateforest", "uswestvirginiacoopersrockstateforest", "uswestvirginiacoopersrockstateforest",             39.664554,             -79.773401,         "epsg:4326",                                  6L,                      NA,                      NA,                                                                                                NA,                              NA,                        NA,                       0L,                    "GBIF",                            2L, "match using verbatim coords",
      "Dorey_data_7220088",              "Augochlora pura (Say, 1837)", "Halictidae", "Halictinae",   "Augochlora",        NA,                    NA,        "Augochlora pura",              "pura",                       NA,                 NA,  "SPECIES",             "(Say, 1837)",                NA,         NA, "United states",    "United states",         "US", "West Virginia",       NA,            NA, "Coopers Rock State Forest,",      NA, "CC_BY_4_0", "TAXON_MATCH_HIGHERRANK;INSTITUTION_MATCH_FUZZY;COLLECTION_MATCH_FUZZY", "1987-08-17T00:00:00Z",  17L,     8L, 1987L, "PRESERVED_SPECIMEN",    NA,         "PRESENT",            NA,            NA,       NA,                NA,              NA,               NA, "Insect Collection 358300",            NA,           "INHS",                NA,               NA,                NA,        436516763L,         NA,           NA,                   NA,                                     NA,      NA,                                              NA,                                     NA, "Augochlora (Augochlora) pura pura (Say 1837)",                 NA,       NA,                                               NA,                NA,              NA,        NA, "GBIF_Halictidae",                NA, "Augochlora pura",              "Augochlora pura (Say, 1837)",                "US",                          "US", "uswestvirginiacoopersrockstateforest", "uswestvirginiacoopersrockstateforest", "uswestvirginiacoopersrockstateforest",             39.664554,             -79.773401,         "epsg:4326",                                  6L,                      NA,                      NA,                                                                                                NA,                              NA,                        NA,                       0L,                    "GBIF",                            2L, "match using verbatim coords",
      "Dorey_data_7236978",              "Augochlora pura (Say, 1837)", "Halictidae", "Halictinae",   "Augochlora",        NA,                    NA,        "Augochlora pura",              "pura",                       NA,                 NA,  "SPECIES",             "(Say, 1837)",                NA,         NA, "United states",    "United states",         "US", "West Virginia",       NA,            NA, "Coopers Rock State Forest,",      NA, "CC_BY_4_0", "TAXON_MATCH_HIGHERRANK;INSTITUTION_MATCH_FUZZY;COLLECTION_MATCH_FUZZY", "1987-08-17T00:00:00Z",  17L,     8L, 1987L, "PRESERVED_SPECIMEN",    NA,         "PRESENT",            NA,            NA,       NA,                NA,              NA,               NA, "Insect Collection 358301",            NA,           "INHS",                NA,               NA,                NA,        436516766L,         NA,           NA,                   NA,                                     NA,      NA,                                              NA,                                     NA, "Augochlora (Augochlora) pura pura (Say 1837)",                 NA,       NA,                                               NA,                NA,              NA,        NA, "GBIF_Halictidae",                NA, "Augochlora pura",              "Augochlora pura (Say, 1837)",                "US",                          "US", "uswestvirginiacoopersrockstateforest", "uswestvirginiacoopersrockstateforest", "uswestvirginiacoopersrockstateforest",             39.664554,             -79.773401,         "epsg:4326",                                  6L,                      NA,                      NA,                                                                                                NA,                              NA,                        NA,                       0L,                    "GBIF",                            2L, "match using verbatim coords",
      "Dorey_data_15031296", "Lasioglossum sisymbrii (Cockerell, 1895)", "Halictidae", "Halictinae", "Lasioglossum",        NA,                    NA, "Lasioglossum sisymbrii",         "sisymbrii",                       NA,                 NA,         NA,       "(Cockerell, 1895)",                NA,         NA, "United states",    "United states",         "US",       "Arizona", "Mohave",            NA,                     "Topock",      NA,          NA,                                                                      NA, "1972-04-08T00:00:00Z",   8L,     4L, 1972L,  "PreservedSpecimen",    NA,                NA,            NA, "B. Apperson",       NA,                NA,              NA,               1L,                "HYM 95971",         "MNA",            "MNA",                NA,               NA,                NA,                NA,         NA,           NA,                   NA, "1154c20d-211a-47ba-8320-784b8e320e6f",      NA, "urn:uuid:ada6a97d-7d8e-4765-a607-289b4abcaa39", "48b700e0-fcdb-4196-b642-70fd7b8a71ef",                                             NA,         "8/4/1972", "Female", "http://creativecommons.org/licenses/by-nc/4.0/",                NA,              NA, 57139867L, "SCAN_Halictidae",                NA,                NA, "Lasioglossum sisymbrii (Cockerell, 1895)",                "US",                          "US",                "usarizonamohavetopock",                "usarizonamohavetopock",                "usarizonamohavetopock",            34.7183075,           -114.4871902,         "epsg:4326",                                 14L, "Gary W. Shugart (PSM)",              "3/8/2003", "MaNIS/HerpNET/ORNIS Georeferencing Guidelines, Guide to Best Practices for Georeferencing 2006.", "AZ GNIS downloaded March 2003",                        NA,                      30L,                   "ORNIS",                            1L, "match using verbatim coords"
    )
    
    
    # Be sure that the testData is not already in tempdir
    testDataPath <- file.info(list.files(tempdir(), full.names = T, 
                                         pattern = "testData.xlsx", recursive = TRUE))
    unlink(rownames(testDataPath))
    
    # Save a temporary version of these data
    # First create a workbook
    wb <- openxlsx::createWorkbook()
    # Add sheets to it
    openxlsx::addWorksheet(wb, "GEOLOCATE HIGH")
    openxlsx::addWorksheet(wb, "BELS High")
    # Fill those sheets with data
    openxlsx::writeData(wb, "GEOLOCATE HIGH", testData)
    openxlsx::writeData(wb, "BELS High", testData2)
    # Save to file
    openxlsx::saveWorkbook(wb, file = paste0(tempdir(), "/testData.xlsx"), overwrite = TRUE)
    
    
    testOut1 <- BeeBDC::readr_BeeBDC(dataset = "GeoL",
                                   path = tempdir(),
                                  inFile = "testData.xlsx",
                                  outFile = "testDataOut.csv",
                                  dataLicense = "https://creativecommons.org/licenses/by-nc-sa/4.0/")
    
    
    
    # Get a count of TRUE and FALSE column name matches
    resultsT <- sum(colnames(testOut1) %in% (ColTypeR()[[1]] %>% names()) == TRUE)
    resultsF <- sum(colnames(testOut1) %in% (ColTypeR()[[1]] %>% names()) == FALSE)
    
    # Test the number of expected TRUE and FALSE columns and then test the output format (data frames and
    # tibbles are a special case of lists)
    testthat::test_that("readr_GeoL results columns TRUE", {
      testthat::expect_equal(resultsT, 59)
    })
    testthat::test_that("readr_GeoL results columns FALSE", {
      testthat::expect_equal(resultsF, 1)
    })
    
    testthat::test_that("readr_GeoL expected class", {
      testthat::expect_type(testOut1, "list")
    })
    
    
    
    
    
    
#### 13.0 EaCO  ####
    
    testData <- dplyr::tribble(
      ~'Specimen Number', ~`Farmers (names redacted)`, ~'Date Blue vane traps were active', ~'Date Range for collection', ~'Sampling Round', ~'Treatment type', ~'Trap number',          ~County, ~State, ~'GPS Coordinates of Traps',         ~Genus, ~`Species (if available)`, ~`other (example coleoptera)`, ~`notes (example hymenoptera-ant)`, ~comment,
      6376L,                          3L,                                NA,              "05/24-06/01",              1L,        "Grazed",         "S7", "Kit Carson Co.",   "CO",      "39.210N, -102.884W", "Species",                     "135",                            NA,                                 NA,       NA,
      6051L,                          3L,                                NA,              "05/24-06/01",              1L,      "Ungrazed",         "S6", "Kit Carson Co.",   "CO",      "39.210N, -102.883W", "Species",                     "165",                            NA,                                 NA,       NA,
      3291L,                          2L,                                NA,              "06/21-06/28",              2L,        "Grazed",         "T4",    "Perkins Co.",   "NE",      "40.799N, -101.944W",  "Species",             "secret",                            NA,                                 NA,       NA,
      3457L,                          2L,                                NA,              "06/21-06/28",              2L,      "Ungrazed",         "T8",    "Perkins Co.",   "NE",      "40.799N, -101.936W",  "Species",             "secret",                            NA,                                 NA,       NA,
      103L,                          1L,                                NA,              "05/24-06/01",              1L,        "Fallow",         "F4",       "Weld Co.",   "CO",      "40.503N, -103.901W",   "Species",               "secret",                            NA,                                 NA,       NA
    )
    
    
    # Be sure that the testData is not already in tempdir
    testDataPath <- file.info(list.files(tempdir(), full.names = T, 
                                         pattern = "testData.xslx", recursive = TRUE))
    unlink(rownames(testDataPath))
    
    # Save a temporary version of these data
    openxlsx::write.xlsx(testData, paste0(tempdir(), "/testData.xlsx"), sheetName="Sheet1")
    
    testOut1 <- BeeBDC::readr_BeeBDC(dataset = "EaCo",
                                   path = tempdir(),
                                  inFile = "testData.xlsx",
                                  outFile = "testDataOut.csv",
                                  dataLicense = "https://creativecommons.org/licenses/by-nc-sa/4.0/")
    
    
    
    # Get a count of TRUE and FALSE column name matches
    resultsT <- sum(colnames(testOut1) %in% (ColTypeR()[[1]] %>% names()) == TRUE)
    resultsF <- sum(colnames(testOut1) %in% (ColTypeR()[[1]] %>% names()) == FALSE)
    
    # Test the number of expected TRUE and FALSE columns and then test the output format (data frames and
    # tibbles are a special case of lists)
    testthat::test_that("readr_EaCO results columns TRUE", {
      testthat::expect_equal(resultsT, 18)
    })
    testthat::test_that("readr_EaCO results columns FALSE", {
      testthat::expect_equal(resultsF, 0)
    })
    
    testthat::test_that("readr_EaCO expected class", {
      testthat::expect_type(testOut1, "list")
    })
    
    
    
    
    
#### 14.0 MABC  ####
    
    testData <- dplyr::tribble(
      ~Ejemplar, ~'Fecha colecta',    ~País, ~`Estado/Provincia`, ~Municipio,       ~Localidad,            ~'Sitio Colecta', ~'Código sitio', ~'Metodo colecta', ~Hora, ~'Coordenadas Lat', ~'Coordenadas Long', ~Altitud, ~'Datos georeferenciación',         ~Colector,   ~Identificador,     ~Familia,      ~Subfamilia,          ~Tribu,      ~Genero, ~Subgenero,      ~Especie, ~Subespecie,       ~'Nombre especie', ~'Código especie', ~Sexo,
      "MABC-04-0000001",   "10/10/1982", "México",           "Jalisco",         NA,        "Chamela", "Estacion Biológica UNAM",            NA,              NA,    2,               NA,                NA,       NA,                       NA, "Stephen Bullock", "R. R. Snelling", "Colletidae", "Diphaglossinae", "Dissoglossini", "Secret",         NA,     "species",          NA,  "Secret species",              NA,   "M",
      "MABC-04-0000002",   "23/10/1982", "México",           "Jalisco",         NA,        "Chamela", "Estacion Biológica UNAM",            NA,              NA,    7,               NA,                NA,       NA,                       NA, "Stephen Bullock", "R. R. Snelling", "Colletidae", "Diphaglossinae", "Dissoglossini", "Secret",         NA,     "species",          NA,  "Secret species",              NA,   "H",
      "MABC-04-0000003",    "8/11/2003", "México",   "Baja California", "Ensenada", "Nuevo Rosarito",          "Nuevo Rosarito",       "BCENR",              NA,    NA,           28.634,          -114.017,       NA,             "Originales", "Stephen Bullock", "Terry Griswold", "Colletidae",     "Colletinae",     "Colletini",   "Secret",         NA, "species",          NA, "Secret species",        "COLALB",   "M",
      "MABC-04-0000004",    "8/11/2003", "México",   "Baja California", "Ensenada", "Nuevo Rosarito",          "Nuevo Rosarito",       "BCENR",              NA,    NA,           28.634,          -114.017,       NA,             "Originales", "Stephen Bullock", "Terry Griswold", "Colletidae",     "Colletinae",     "Colletini",   "Secret",         NA, "species",          NA, "Secret species",        "COLALB",   "M",
      "MABC-04-0000005",    "8/11/2003", "México",   "Baja California", "Ensenada", "Nuevo Rosarito",          "Nuevo Rosarito",       "BCENR",              NA,    NA,           28.634,          -114.017,       NA,             "Originales", "Stephen Bullock", "Terry Griswold", "Colletidae",     "Colletinae",     "Colletini",   "Secret",         NA, "species",          NA, "Secret species",        "COLALB",   "M"
    )
    
    # Be sure that the testData is not already in tempdir
    testDataPath <- file.info(list.files(tempdir(), full.names = T, 
                                         pattern = "testData.xlsx", recursive = TRUE))
    unlink(rownames(testDataPath))
    
    # Save a temporary version of these data
    openxlsx::write.xlsx(testData, paste0(tempdir(), "/testData.xlsx"), sheetName="Hoja1")
    
    testOut1 <- BeeBDC::readr_BeeBDC(dataset = "MABC",
                                   path = tempdir(),
                                  inFile = "testData.xlsx",
                                  outFile = "testDataOut.csv",
                                  sheet = "Hoja1",
                                  dataLicense = "All rights reserved")
    
    
    # Get a count of TRUE and FALSE column name matches
    resultsT <- sum(colnames(testOut1) %in% (ColTypeR()[[1]] %>% names()) == TRUE)
    resultsF <- sum(colnames(testOut1) %in% (ColTypeR()[[1]] %>% names()) == FALSE)
    
    
    testthat::test_that("readr_MABC results columns TRUE", {
      testthat::expect_equal(resultsT, 28)
    })
    testthat::test_that("readr_MABC results columns FALSE", {
      testthat::expect_equal(resultsF, 0)
    })
    
    testthat::test_that("readr_MABC expected class", {
      testthat::expect_type(testOut1, "list")
    })
    
    
    
    
    
#### 15.0 Col  ####
    
    # this is FAKE DATA BECAUSE THE REAL DATA IS UNDER RESTRCTIVE LICENSE
    testData <- dplyr::tribble(
      ~Código.de.Barras, ~institutionCode, ~`Colectores.[Aggregated]`, ~Colectores.asociados, ~Fecha.colección.inicial,  ~Ano, ~mes, ~dia,        ~Orden, ~Familia,         ~Género,    ~Especie,            ~Especie.Author, ~Tipo, ~`Determinador.[Formatted]`, ~Fecha.determinación,            ~País,        ~Departamento, ~Municipio, ~Corregimiento.Departamental,                                              ~Localidad, ~Latitud.georref..dec., ~Longitud.georref..dec.,          ~Nombre.Completo,
      "ICN_butfake1",            "ICN",              "anonymous a",                    NA,             "01/01/2010", 2010L, "01", "01", "Hymenoptera", "Apidae",          "Apis", "mellifera",           "Linneaus, 1758",    NA,                          NA,                   NA, "ESTADOS UNIDOS",            "ARIZONA",   "COCHISE", "1 milla above nion saddle",      "USA. Arizona. Cochise. 1 milla above nion saddle",             32.1139688,            -109.9211756,          "Apis mellifera",
      "ICN_butfake2",            "ICN",              "anonymous b",                    NA,             "01/01/2010", 2010L, "01", "01", "Hymenoptera", "Apidae",          "Apis", "mellifera",           "Linneaus, 1758",    NA,               "anonymous e",                   NA, "ESTADOS UNIDOS",            "ARIZONA",   "COCHISE", "1 milla above nion saddle",      "USA. Arizona. Cochise. 1 milla above nion saddle",             32.1139688,            -109.9211756,          "Apis mellifera",
      "ICN_butfake3",            "ICN",              "anonymous c",                    NA,             "01/01/2010", 2010L, "01", "01", "Hymenoptera", "Apidae",          "Apis", "mellifera",           "Linneaus, 1758",    NA,                          NA,                   NA, "ESTADOS UNIDOS",            "ARIZONA",  "COCHISE",  "1 milla above nion saddle",      "USA. Arizona. Cochise. 1 milla above nion saddle",             32.1139688,            -109.9211756,          "Apis mellifera",
      "ICN_butfake4",            "ICN",              "anonymous d",                    NA,            "01/01/20101", 2010L, "01", "01", "Hymenoptera", "Apidae",          "Apis", "mellifera",           "Linneaus, 1758",    NA,               "anonymous f",                   NA, "ESTADOS UNIDOS",            "ARIZONA",  "COCHISE",  "1 milla above nion saddle",      "USA. Arizona. Cochise. 1 milla above nion saddle",             32.1139688,            -109.9211756,          "Apis mellifera"
    )
    
    # need to change column names, function requires spaces to work 
    colnames(testData) <- c("Código de Barras", "institutionCode", "Colectores [Aggregated]", "Colectores asociados", "Fecha colección inicial", "Ano", "mes", "dia", "Orden", "Familia", "Género", "Especie", "Especie Author", "Tipo", "Determinador [Formatted]", "Fecha determinación", "País", "Departamento", "Municipio", "Corregimiento Departamental", "Localidad", "Latitud georref. dec.", "Longitud georref. dec.", "Nombre Completo")
    
    # Be sure that the testData is not already in tempdir
    testDataPath <- file.info(list.files(tempdir(), full.names = T, 
                                         pattern = "testData.xlsx", recursive = TRUE))
    unlink(rownames(testDataPath))
    
    # Save a temporary version of these data
    openxlsx::write.xlsx(testData, paste0(tempdir(), "/testData.xlsx"), sheetName = "Spanish headers")
    
    
    testOut1 <- BeeBDC::readr_BeeBDC(dataset = "Col",
                                   path = tempdir(),
                                 inFile = "testData.xlsx",
                                 outFile = "testDataOut.csv",
                                 sheet = "Spanish headers",
                                 dataLicense = "All rights reserved")
    
    
    # Get a count of TRUE and FALSE column name matches
    resultsT <- sum(colnames(testOut1) %in% (ColTypeR()[[1]] %>% names()) == TRUE)
    resultsF <- sum(colnames(testOut1) %in% (ColTypeR()[[1]] %>% names()) == FALSE)
    
    
    # Test the number of expected TRUE and FALSE columns and then test the output format (data frames and
    # tibbles are a special case of lists)
    testthat::test_that("readr_Col results columns TRUE", {
      testthat::expect_equal(resultsT, 28)
    })
    
    testthat::test_that("readr_Col results columns FALSE", {
      testthat::expect_equal(resultsF, 2)
    })
    
    testthat::test_that("readr_Col expected class", {
      testthat::expect_type(testOut1, "list")
    })
    
    
    
#### 16.0 FSCA  ####
    
    testData <- dplyr::tribble(
      ~id, ~institutionCode, ~collectionCode, ~ownerInstitutionCode,      ~basisOfRecord, ~occurrenceID,  ~catalogNumber, ~otherCatalogNumbers,                                                                                      ~higherClassification,   ~kingdom,      ~phylum,    ~class,        ~order,  ~family,       ~scientificName, ~taxonID, ~scientificNameAuthorship,      ~genus, ~subgenus, ~specificEpithet, ~verbatimTaxonRank, ~infraspecificEpithet, ~taxonRank,             ~identifiedBy, ~dateIdentified, ~identificationReferences, ~identificationRemarks, ~taxonRemarks, ~identificationQualifier, ~typeStatus,                    ~recordedBy, ~associatedCollectors, ~recordNumber, ~eventDate, ~year, ~month, ~day, ~startDayOfYear, ~endDayOfYear, ~verbatimEventDate,                                                                                                                             ~occurrenceRemarks, ~habitat, ~substrate, ~verbatimAttributes, ~fieldNumber, ~eventID, ~informationWithheld, ~dataGeneralizations, ~dynamicProperties, ~associatedOccurrences, ~associatedSequences, ~associatedTaxa, ~reproductiveCondition, ~establishmentMeans, ~cultivationStatus, ~lifeStage,     ~sex, ~individualCount, ~preparations,        ~country, ~stateProvince,   ~county, ~municipality,                                                                             ~locality, ~locationRemarks, ~localitySecurity, ~localitySecurityReason, ~decimalLatitude, ~decimalLongitude, ~geodeticDatum, ~coordinateUncertaintyInMeters,                  ~verbatimCoordinates, ~georeferencedBy, ~georeferenceProtocol, ~georeferenceSources, ~georeferenceVerificationStatus, ~georeferenceRemarks, ~minimumElevationInMeters, ~maximumElevationInMeters, ~minimumDepthInMeters, ~maximumDepthInMeters, ~verbatimDepth, ~verbatimElevation, ~disposition, ~language, ~recordEnteredBy,       ~modified, ~`sourcePrimaryKey-dbpk`, ~collID,                                       ~recordID,                                                                         ~references,
      2109207L,           "FSCA",          "FSCA",                    NA, "PreservedSpecimen",            NA, "FSCA 00008737",                   NA, "Organism|Animalia|Arthropoda|Insecta|Hymenoptera|Apoidea|Anthophila|Apidae|Apinae|Anthophorini|Habropoda", "Animalia", "Arthropoda", "Insecta", "Hymenoptera", "Apidae", "Habropoda laboriosa",   11043L,       "(Fabricius, 1804)", "Habropoda",        NA,      "laboriosa",                 NA,                    NA,  "Species", "Wiley, J. R. (James R.)",        "3/1/99",                        NA,  "Habropoda laboriosa",            NA,                       NA,          NA,    "Stange, L. A. (Lionel A.)",                    NA,            NA,  "3/25/91", 1991L,     3L,  25L,             84L,            NA,                 NA, "[Florida Alachua Co: Gainesvile Doyle Conner Bld, March 25, 1991 L.A. Stange col.] [Habropoda laboriosa (Fabricius) 1804 det. J. Wiley 1999]",       NA,         NA,                  NA,           NA,       NA,                   NA,                   NA,                 NA,                     NA,                   NA,              NA,                     NA,                  NA,                 NA,    "adult", "Female",               NA,            NA, "United States",      "Florida", "Alachua",            NA, "Doyle Conner Bldg. (32601), Dept. of Agriculture, Gainesville, Alachua Co., Florida",               NA,                0L,                      NA,         29.63438,         -82.37143,             NA,                             NA,                 "29.63438, -82.37143",               NA,                    NA,                   NA,                              NA,                   NA,                        NA,                        NA,                    NA,                    NA,             NA,                 NA,           NA,        NA,               NA, "9/12/22 14:58",          "FSCA 00008737",     14L, "urn:uuid:b08fcc6e-ba5d-437a-ba80-0ab32bfd0269", "https://library.big-bee.net/portal/collections/individual/index.php?occid=2109207",
      2109208L,           "FSCA",          "FSCA",                    NA, "PreservedSpecimen",            NA, "FSCA 00008738",                   NA, "Organism|Animalia|Arthropoda|Insecta|Hymenoptera|Apoidea|Anthophila|Apidae|Apinae|Anthophorini|Habropoda", "Animalia", "Arthropoda", "Insecta", "Hymenoptera", "Apidae", "Habropoda laboriosa",   11043L,       "(Fabricius, 1804)", "Habropoda",        NA,      "laboriosa",                 NA,                    NA,  "Species",               "Wiley, J.",        "3/1/99",                        NA,  "Habropoda laboriosa",            NA,                       NA,          NA, "Weems Jr., H. V. (Howard V.)",                    NA,            NA,  "3/26/84", 1984L,     3L,  26L,             86L,            NA,                 NA, "[Florida Alachua Co., 8 miles W of Gainesville of FL 26 -111-1984 H.V. Weems, Jr.] [Habropoda laboriosa (Fabricius) 1804 det. J. Wiley 1999]",       NA,         NA,                  NA,           NA,       NA,                   NA,                   NA,                 NA,                     NA,                   NA,              NA,                     NA,                  NA,                 NA,    "adult",   "Male",               NA,            NA, "United States",      "Florida", "Alachua",            NA,                                             "8 mi. W of Gainesville, Alachua Co., FL",               NA,                0L,                      NA,      29.66733333,      -82.53688889,             NA,                             NA, "29.6673333333333, -82.5368888888889",               NA,                    NA,                   NA,                              NA,                   NA,                        NA,                        NA,                    NA,                    NA,             NA,                 NA,           NA,        NA,               NA, "9/12/22 14:58",          "FSCA 00008738",     14L, "urn:uuid:342bc796-8184-4f59-a8b2-1c68a9f67456", "https://library.big-bee.net/portal/collections/individual/index.php?occid=2109208",
      2109209L,           "FSCA",          "FSCA",                    NA, "PreservedSpecimen",            NA, "FSCA 00008739",                   NA, "Organism|Animalia|Arthropoda|Insecta|Hymenoptera|Apoidea|Anthophila|Apidae|Apinae|Anthophorini|Habropoda", "Animalia", "Arthropoda", "Insecta", "Hymenoptera", "Apidae", "Habropoda laboriosa",   11043L,       "(Fabricius, 1804)", "Habropoda",        NA,      "laboriosa",                 NA,                    NA,  "Species",               "Wiley, J.",        "3/1/99",                        NA,  "Habropoda laboriosa",            NA,                       NA,          NA,                             NA,                    NA,            NA,  "3/17/85", 1985L,     3L,  17L,             76L,            NA,                 NA,                          "[Florida: Gainesville UF Campus Mtrap Bamila 17-III-1985] [Habropoda laboriosa (Fabricius) 1804 det. J. Wiley 1999]",       NA,         NA,                  NA,           NA,       NA,                   NA,                   NA,                 NA,                     NA,                   NA,              NA,                     NA,                  NA,                 NA,    "adult",   "Male",               NA,            NA, "United States",      "Florida", "Alachua",            NA,                                      "Gainesville, University of FL, Alachua Co., FL",               NA,                0L,                      NA,         29.65163,         -82.32483,             NA,                             NA,                 "29.65163, -82.32483",               NA,                    NA,                   NA,                              NA,                   NA,                        NA,                        NA,                    NA,                    NA,             NA,                 NA,           NA,        NA,               NA, "9/12/22 14:58",          "FSCA 00008739",     14L, "urn:uuid:af32f90c-2763-4a21-8f90-e93588a80c0f", "https://library.big-bee.net/portal/collections/individual/index.php?occid=2109209",
      2109210L,           "FSCA",          "FSCA",                    NA, "PreservedSpecimen",            NA, "FSCA 00008740",                   NA, "Organism|Animalia|Arthropoda|Insecta|Hymenoptera|Apoidea|Anthophila|Apidae|Apinae|Anthophorini|Habropoda", "Animalia", "Arthropoda", "Insecta", "Hymenoptera", "Apidae", "Habropoda laboriosa",   11043L,       "(Fabricius, 1804)", "Habropoda",        NA,      "laboriosa",                 NA,                    NA,  "Species",               "Wiley, J.",        "3/1/99",                        NA,  "Habropoda laboriosa",            NA,                       NA,          NA,                             NA,                    NA,            NA,  "3/17/85", 1985L,     3L,  17L,             76L,            NA,                 NA,                          "[Florida: Gainesville UF Campus Mtrap Bamila 17-III-1985] [Habropoda laboriosa (Fabricius) 1804 det. J. Wiley 1999]",       NA,         NA,                  NA,           NA,       NA,                   NA,                   NA,                 NA,                     NA,                   NA,              NA,                     NA,                  NA,                 NA,    "adult",   "Male",               NA,            NA, "United States",      "Florida", "Alachua",            NA,                                      "Gainesville, University of FL, Alachua Co., FL",               NA,                0L,                      NA,         29.65163,         -82.32483,             NA,                             NA,                 "29.65163, -82.32483",               NA,                    NA,                   NA,                              NA,                   NA,                        NA,                        NA,                    NA,                    NA,             NA,                 NA,           NA,        NA,               NA, "9/12/22 14:58",          "FSCA 00008740",     14L, "urn:uuid:ca52acc1-7026-46cd-8089-add7f4388734", "https://library.big-bee.net/portal/collections/individual/index.php?occid=2109210"
    )
    
    
    # Be sure that the testData is not already in tempdir
    testDataPath <- file.info(list.files(tempdir(), full.names = T, 
                                         pattern = "testData.csv", recursive = TRUE))
    unlink(rownames(testDataPath))
    
    # Save a temporary version of these data
    readr::write_excel_csv(testData, paste0(tempdir(), "/testData.csv"))
    
    
    testOut1 <- BeeBDC::readr_BeeBDC(dataset = "FSCA",
                                   path = tempdir(),
                                  inFile = "testData.csv",
                                  outFile = "testDataOut.csv",
                                  dataLicense = "https://creativecommons.org/licenses/by-nc-sa/4.0/")
    
    
    # Get a count of TRUE and FALSE column name matches
    resultsT <- sum(colnames(testOut1) %in% (ColTypeR()[[1]] %>% names()) == TRUE)
    resultsF <- sum(colnames(testOut1) %in% (ColTypeR()[[1]] %>% names()) == FALSE)
    
    
    # Test the number of expected TRUE and FALSE columns and then test the output format (data frames and
    # tibbles are a special case of lists)
    testthat::test_that("readr_FSCA results columns TRUE", {
      testthat::expect_equal(resultsT, 57)
    })
    
    testthat::test_that("readr_FSCA results columns FALSE", {
      testthat::expect_equal(resultsF, 0)
    })
    
    testthat::test_that("readr_FSCA expected class", {
      testthat::expect_type(testOut1, "list")
    })
    
    
#### 17.0 SMC  ####
    
    testData <- dplyr::tribble(
      ~organismName, ~individualCount,  ~county, ~stateProvince,                             ~locale, ~observationDate, ~collectionMethod,            ~references, ~institutionCode, ~latitude, ~longitude, ~georeferenceMethod, ~georeferenceMethodNotes,  ~georeferencedBy, ~determiner, ~lifeStage,                                                 ~Data.Source.Link, ~Notes,
      "Agapostemon_angelicus/texanus",               2L, "Bailey",        "Texas", "Muleshoe National Wildlife Refuge",      "3/17/2013",       "pan_traps", "Auerbach et al, 2019",            "TTU",  33.96377,  -102.7508,       "Digital Map",           "Google Earth", "Shannon Collins",          NA,    "Adult", "https://academic.oup.com/ee/article/48/4/968/5494819?login=true",     NA,
      "Agapostemon_melliventris",               2L, "Bailey",        "Texas", "Muleshoe National Wildlife Refuge",      "3/17/2013",       "pan_traps", "Auerbach et al, 2019",            "TTU",  33.96377,  -102.7508,       "Digital Map",           "Google Earth", "Shannon Collins",          NA,    "Adult", "https://academic.oup.com/ee/article/48/4/968/5494819?login=true",     NA,
      "Andrena_alamonis",               2L, "Bailey",        "Texas", "Muleshoe National Wildlife Refuge",      "3/17/2013",       "pan_traps", "Auerbach et al, 2019",            "TTU",  33.96377,  -102.7508,       "Digital Map",           "Google Earth", "Shannon Collins",          NA,    "Adult", "https://academic.oup.com/ee/article/48/4/968/5494819?login=true",     NA,
      "Agapostemon_angelicus",               6L, "Bailey",        "Texas", "Muleshoe National Wildlife Refuge",      "4/12/2013",       "pan_traps", "Auerbach et al, 2019",            "TTU",  33.96377,  -102.7508,       "Digital Map",           "Google Earth", "Shannon Collins",          NA,    "Adult", "https://academic.oup.com/ee/article/48/4/968/5494819?login=true",     NA,
      "Agapostemon_coloradinus",               1L, "Bailey",        "Texas", "Muleshoe National Wildlife Refuge",      "4/12/2013",       "pan_traps", "Auerbach et al, 2019",            "TTU",  33.96377,  -102.7508,       "Digital Map",           "Google Earth", "Shannon Collins",          NA,    "Adult", "https://academic.oup.com/ee/article/48/4/968/5494819?login=true",     NA
    )
    
    # Be sure that the testData is not already in tempdir
    testDataPath <- file.info(list.files(tempdir(), full.names = T, 
                                         pattern = "testData.csv", recursive = TRUE))
    unlink(rownames(testDataPath))
    
    # Save a temporary version of these data
    readr::write_excel_csv(testData, paste0(tempdir(), "/testData.csv"))
    
    testOut1 <- BeeBDC::readr_BeeBDC(dataset = "SMC",
                                   path = tempdir(),
                                 inFile = "testData.csv",
                                 outFile = "testDataOut.csv",
                                 dataLicense = "https://creativecommons.org/licenses/by-nc-sa/4.0/")
    
    
    
    # Get a count of TRUE and FALSE column name matches
    resultsT <- sum(colnames(testOut1) %in% (ColTypeR()[[1]] %>% names()) == TRUE)
    resultsF <- sum(colnames(testOut1) %in% (ColTypeR()[[1]] %>% names()) == FALSE)
    
    # Test the number of expected TRUE and FALSE columns and then test the output format (data frames and
    # tibbles are a special case of lists)
    testthat::test_that("readr_SMC results columns TRUE", {
      testthat::expect_equal(resultsT, 20)
    })
    testthat::test_that("readr_SMC results columns FALSE", {
      testthat::expect_equal(resultsF, 8)
    })
    
    testthat::test_that("readr_SMC expected class", {
      testthat::expect_type(testOut1, "list")
    })
    
    
    
    
#### 18.0 Bal  ####
    
    testData <- dplyr::tribble(
      ~studyID,           ~siteID, ~year,     ~date,                           ~animalID, ~abundance, ~abundanceMethod, ~samplingMethod, ~numCensus, ~samplingIntensity,                                                                                                                                                                                   ~censusType,     ~fieldDist,  ~flowering, ~decimalLatitude, ~decimalLongitude,                ~studyLocation, ~habitatType,                ~siteDescription,
      "Ball01", "Bastrop Gardens", 2013L, "22/5/2013",             "Agapostemon angelicus",         1L,      "Abundance",      "Pan Trap",         1L,                24L, "50 pan traps (painted blue, yellow or left white as in LeBuhn et al) placed in x formation 1 m apart from center of the 50 m2 plot, vegetation measured as described for blue vane trapped.", "not measured", "flowering",        30.157397,        -97.491703, "Bastrop, Bastrop County, TX",           NA, "Agriculture. Community garden",
      "Ball01", "Bastrop Gardens", 2013L, "22/5/2013",               "Agapostemon texanus",         1L,      "Abundance",      "Pan Trap",         1L,                24L, "50 pan traps (painted blue, yellow or left white as in LeBuhn et al) placed in x formation 1 m apart from center of the 50 m2 plot, vegetation measured as described for blue vane trapped.", "not measured", "flowering",        30.157397,        -97.491703, "Bastrop, Bastrop County, TX",           NA, "Agriculture. Community garden",
      "Ball01", "Bastrop Gardens", 2013L, "22/5/2013",            "Anthophora californica",         1L,      "Abundance",      "Pan Trap",         1L,                24L, "50 pan traps (painted blue, yellow or left white as in LeBuhn et al) placed in x formation 1 m apart from center of the 50 m2 plot, vegetation measured as described for blue vane trapped.", "not measured", "flowering",        30.157397,        -97.491703, "Bastrop, Bastrop County, TX",           NA, "Agriculture. Community garden",
      "Ball01", "Bastrop Gardens", 2013L, "22/5/2013",                "Ceratina shinnersi",         1L,      "Abundance",      "Pan Trap",         1L,                24L, "50 pan traps (painted blue, yellow or left white as in LeBuhn et al) placed in x formation 1 m apart from center of the 50 m2 plot, vegetation measured as described for blue vane trapped.", "not measured", "flowering",        30.157397,        -97.491703, "Bastrop, Bastrop County, TX",           NA, "Agriculture. Community garden",
      "Ball01", "Bastrop Gardens", 2013L, "22/5/2013",                  "Ceratina strenua",         1L,      "Abundance",      "Pan Trap",         1L,                24L, "50 pan traps (painted blue, yellow or left white as in LeBuhn et al) placed in x formation 1 m apart from center of the 50 m2 plot, vegetation measured as described for blue vane trapped.", "not measured", "flowering",        30.157397,        -97.491703, "Bastrop, Bastrop County, TX",           NA, "Agriculture. Community garden",
      "Ball01", "Bastrop Gardens", 2013L, "22/5/2013", "Lasioglossum (Dialictus) sp.TX-14",        12L,      "Abundance",      "Pan Trap",         1L,                24L, "50 pan traps (painted blue, yellow or left white as in LeBuhn et al) placed in x formation 1 m apart from center of the 50 m2 plot, vegetation measured as described for blue vane trapped.", "not measured", "flowering",        30.157397,        -97.491703, "Bastrop, Bastrop County, TX",           NA, "Agriculture. Community garden"
    ) 
    # Because readr_Bal ignores the first row of data, get the columns names and add them again 
    # to the data to be saved (and eventually ignored).
    testDataCols <-colnames(testData)
    # Convert into a data.frame instead of a tibble to relax the requirements on colmn classes matching
    testData <- testData %>% 
      as.data.frame(stringsAsFactors = FALSE) 
    # Re-bind the column names to the first row of data.
    testData <- testDataCols %>%
      rbind(testData)
    
    # Be sure that the testData is not already in tempdir
    testDataPath <- file.info(list.files(tempdir(), full.names = T, 
                                         pattern = "testData.xlsx", recursive = TRUE))
    unlink(rownames(testDataPath))
    
    # Save a temporary version of these data
    openxlsx::write.xlsx(testData, paste0(tempdir(), "/testData.xlsx"), sheetName = "animal_data")
    
    testOut1 <- BeeBDC::readr_BeeBDC(dataset = "Bal",
                                   path = tempdir(),
                                 inFile = "testData.xlsx",
                                 outFile = "testDataOut.csv",
                                 dataLicense = "https://creativecommons.org/licenses/by-nc-sa/4.0/",
                                 sheet = "animal_data")
    
    
    # Get a count of TRUE and FALSE column name matches
    resultsT <- sum(colnames(testOut1) %in% (ColTypeR()[[1]] %>% names()) == TRUE)
    resultsF <- sum(colnames(testOut1) %in% (ColTypeR()[[1]] %>% names()) == FALSE)
    
    # Test the number of expected TRUE and FALSE columns and then test the output format (data frames and
    # tibbles are a special case of lists)
    testthat::test_that("readr_Bal results columns TRUE", {
      testthat::expect_equal(resultsT, 19)
    })
    
    testthat::test_that("readr_Bal results columns FALSE", {
      testthat::expect_equal(resultsF, 8)
    })
    
    testthat::test_that("readr_Bal expected class", {
      testthat::expect_type(testOut1, "list")
    })
    
    
    
    
#### 19.0 Lic  ####
    
    testData <- dplyr::tribble(
      ~eventID, ~occurrenceID,      ~basisOfRecord,  ~eventDate,   ~Kingdom,        ~Order, ~Family_or_grp,      ~Tribe,         ~Genus, ~Species, ~Morphospecies, ~adult, ~sex,    ~Collector,          ~Determiner, ~individualCount, ~samplingProtocol, ~samplingEffort, ~sampleSizeValue, ~sampleSizeUnit, ~decimalLatitude, ~decimalLongitude, ~geodeticDatum, ~countryCode,                   ~country,
      "Be2:b",        "EML1", "PreservedSpecimen", "24/6/2014", "Animalia", "Hymenoptera",   "Andrenidae",          NA,             NA,       NA,            38L,   TRUE,   NA, "L. Rafferty", "Ivan Milosavljevic",               1L,   "blue pan trap",      "24 hours",               2L,         "traps",            46.52,           -116.87,        "WGS84",         "US", "United States of America",
      "Be2:b",        "EML2", "PreservedSpecimen", "24/6/2014", "Animalia", "Hymenoptera",   "Andrenidae",          NA,             NA,       NA,            77L,   TRUE,   NA, "L. Rafferty", "Ivan Milosavljevic",               2L,   "blue pan trap",      "24 hours",               2L,         "traps",            46.52,           -116.87,        "WGS84",         "US", "United States of America",
      "Be2:b",        "EML3", "PreservedSpecimen", "24/6/2014", "Animalia", "Hymenoptera",       "Apidae",          NA,             NA,       NA,            15L,   TRUE,   NA, "L. Rafferty", "Ivan Milosavljevic",               2L,   "blue pan trap",      "24 hours",               2L,         "traps",            46.52,           -116.87,        "WGS84",         "US", "United States of America",
      "Be2:b",        "EML4", "PreservedSpecimen", "24/6/2014", "Animalia", "Hymenoptera",   "Halictidae", "Halictini",             NA,       NA,            39L,   TRUE,   NA, "L. Rafferty", "Ivan Milosavljevic",               1L,   "blue pan trap",      "24 hours",               2L,         "traps",            46.52,           -116.87,        "WGS84",         "US", "United States of America",
      "Be2:b",        "EML5", "PreservedSpecimen", "24/6/2014", "Animalia", "Hymenoptera",   "Halictidae",          NA,  "Agapostemon",       NA,            13L,   TRUE,   NA, "L. Rafferty", "Ivan Milosavljevic",               1L,   "blue pan trap",      "24 hours",               2L,         "traps",            46.52,           -116.87,        "WGS84",         "US", "United States of America",
      "Be2:b",        "EML6", "PreservedSpecimen", "24/6/2014", "Animalia", "Hymenoptera",   "Halictidae",          NA, "Lasioglossum",       NA,            37L,   TRUE,   NA, "L. Rafferty", "Ivan Milosavljevic",               4L,   "blue pan trap",      "24 hours",               2L,         "traps",            46.52,           -116.87,        "WGS84",         "US", "United States of America"
    )
    
    
    # Be sure that the testData is not already in tempdir
    testDataPath <- file.info(list.files(tempdir(), full.names = T, 
                                         pattern = "testData.csv", recursive = TRUE))
    unlink(rownames(testDataPath))
    # Be sure that the testData is not already in tempdir
    testDataPath <- file.info(list.files(tempdir(), full.names = T, 
                                         pattern = "testData.csv", recursive = TRUE))
    unlink(rownames(testDataPath))
    # Save a temporary version of these data
    readr::write_excel_csv(testData, paste0(tempdir(), "/testData.csv"))
    
    testOut1 <- BeeBDC::readr_BeeBDC(dataset = "Lic",
                                   path = tempdir(),
                                 inFile = "testData.csv",
                                 outFile = "testDataOut.csv",
                                 dataLicense = "https://creativecommons.org/licenses/by-nc-sa/4.0/")
    
    
    # Get a count of TRUE and FALSE column name matches
    resultsT <- sum(colnames(testOut1) %in% (ColTypeR()[[1]] %>% names()) == TRUE)
    resultsF <- sum(colnames(testOut1) %in% (ColTypeR()[[1]] %>% names()) == FALSE)
    
    
    # Test the number of expected TRUE and FALSE columns and then test the output format (data frames and
    # tibbles are a special case of lists)
    testthat::test_that("readr_Lic results columns TRUE", {
      testthat::expect_equal(resultsT, 29)
    })
    
    testthat::test_that("readr_Lic results columns FALSE", {
      testthat::expect_equal(resultsF, 6)
    })
    
    testthat::test_that("readr_Lic expected class", {
      testthat::expect_type(testOut1, "list")
    })
    
    
#### 20.0 Arm  ####
    
    testData <- dplyr::tribble(
      ~fam,        ~genus,            ~sp,                   ~species, ~sex,         ~locality,    ~munic, ~state,  ~g,  ~m,    ~s,          ~y,   ~G,  ~M,    ~S,           ~x, ~elev,         ~ecoregion,        ~veget, ~day, ~month, ~year, ~institutionCode,
      "Halictidae", "Agapostemon", "melliventris", "Agapostemon melliventris",  "M",       "El Sauzal",  "Mulegé",  "BCS", 27L, 10L,  22.3, 27.17286111, 112L, 52L,   5.7,   -112.86825,  132L,  "Vizcaíno Desert",       "oasis",  13L,  "jul", 2017L,         "CIBNOR",
      "Halictidae", "Agapostemon",    "mexicanus",    "Agapostemon mexicanus",  "M", "Cd. Insurgentes", "Comondú",  "BCS", 25L,  8L, 28.97, 25.14138056, 111L, 48L, 40.03, -111.8111194,   60L, "Magdalena plains", "xeric scrub",   3L,  "jun", 2017L,         "CIBNOR",
      "Halictidae", "Agapostemon", "melliventris", "Agapostemon melliventris",  "M",       "El Sauzal",  "Mulegé",  "BCS", 27L, 10L,  22.3, 27.17286111, 112L, 52L,   5.7,   -112.86825,  132L,  "Vizcaíno Desert",       "oasis",  13L,  "jul", 2017L,         "CIBNOR"
    )
    
    
    # Be sure that the testData is not already in tempdir
    testDataPath <- file.info(list.files(tempdir(), full.names = T, 
                                         pattern = "testData.xlsx", recursive = TRUE))
    unlink(rownames(testDataPath))
    
    # Save a temporary version of these data
    openxlsx::write.xlsx(testData, paste0(tempdir(), "/testData.xlsx"),
                         sheetName = "Sheet1")
    
    testOut1 <- BeeBDC::readr_BeeBDC(dataset = "Arm",
                                   path = tempdir(),
                                   inFile = "testData.xlsx",
                                   outFile = "testDataOut.csv",
                                   sheet = "Sheet1",
                                   dataLicense = "https://creativecommons.org/licenses/by-nc-sa/4.0/")
    
    
    # Get a count of TRUE and FALSE column name matches
    resultsT <- sum(colnames(testOut1) %in% (BeeBDC::ColTypeR()[[1]] %>% names()) == TRUE)
    resultsF <- sum(colnames(testOut1) %in% (BeeBDC::ColTypeR()[[1]] %>% names()) == FALSE)
    
    # Test the number of expected TRUE and FALSE columns and then test the output format (data frames and
    # tibbles are a special case of lists)
    testthat::test_that("readr_Arm results columns TRUE", {
      testthat::expect_equal(resultsT, 25)
    })
    testthat::test_that("readr_Arm results columns FALSE", {
      testthat::expect_equal(resultsF, 0)
    })
    
    testthat::test_that("readr_Arm expected class", {
      testthat::expect_type(testOut1, "list")
    })
    
    
    
    
#### 21.0 Dorey  ####
    
    testData <- dplyr::tribble(
      ~basisOfRecord, ~catalogNumber, ~recordNumber,   ~locationID,      ~family,  ~subfamily,         ~genus, ~specificEpithet,           ~scientificName, ~identifiedBy,    ~sex, ~individualCount, ~decimalLatitude, ~decimalLongitude, ~elevationInMeters, ~stateOrProvince,        ~locality,   ~eventDate, ~eventTime, ~recordedBy,    ~country,                                                                                                                                                                                                                                                                                                               ~fieldNotes, ~samplingEffort,            ~associatedTaxa, ~Plant.ID.by, ~Plant.ID.using,                                                                                                                                                                                ~references, ~coordinateUncertaintyInMeters, ~institutionCode,                                                                                                                                                                     ~bibliographicCitation,
      "preservedSpecimen",             NA, "18JDEC450-A", "JBD_Site365", "Colletidae", "Hylaeinae", "Pharohylaeus",     "lactiferus", "Pharohylaeus lactiferus",    "JB Dorey",     "F",               1L,        -17.26785,         145.49384,               868L,            "QLD", "Hallorans Hill",     "3/2/19",     "8:55",  "JB Dorey", "Australia",                                                                                                                                                                                                                                                    "Sight-swept off vegetation on rainforests edge. Mostly overcast ~24C",              NA, "Brachychiton acerifolius",   "JB Dorey",       "Keybase", "DOREY, J. B. 2021. Missing for almost 100 years: the rare and potentially threatened bee Pharohylaeus lactiferus (Hymenoptera, Colltidae). Journal of Hymenoptera Research, 81, 165-180.",                            10L,          "Dorey", "DOREY, J. B. 2021. Missing for almost 100 years: the rare and potentially threatened bee Pharohylaeus lactiferus (Hymenoptera, Colltidae). Journal of Hymenoptera Research, 81, 165-180.",
      "preservedSpecimen",             NA, "18JDEC459-A", "JBD_Site365", "Colletidae", "Hylaeinae", "Pharohylaeus",     "lactiferus", "Pharohylaeus lactiferus",    "JB Dorey",    "2M",               2L,        -17.26761,         145.49474,               864L,            "QLD", "Hallorans Hill",     "5/2/19",    "12:12",  "JB Dorey", "Australia",                                                                                                                             "General sweep off Stenocarpus sinuatus 6-7 m up in the canopy of rainforest (probably regenerated around a remnant core). Partly sunny to raining. Other floral resources presented ~26C +i",    "15 minutes",     "Stenocarpus sinuatus",   "JB Dorey",       "Keybase", "DOREY, J. B. 2021. Missing for almost 100 years: the rare and potentially threatened bee Pharohylaeus lactiferus (Hymenoptera, Colltidae). Journal of Hymenoptera Research, 81, 165-180.",                            10L,          "Dorey", "DOREY, J. B. 2021. Missing for almost 100 years: the rare and potentially threatened bee Pharohylaeus lactiferus (Hymenoptera, Colltidae). Journal of Hymenoptera Research, 81, 165-180.",
      "preservedSpecimen",             NA,   "18JDEC460", "JBD_Site365", "Colletidae", "Hylaeinae", "Pharohylaeus",     "lactiferus", "Pharohylaeus lactiferus",    "JB Dorey",    "2M",               2L,        -17.26761,         145.49474,               864L,            "QLD", "Hallorans Hill",     "5/2/19",    "14:30",  "JB Dorey", "Australia",                                                                                                                                           "General sweep off Stenocarpus sinuatus 6-7 m up in the canopy of rainforest (probably regenerated around a remnant core). Overcast. Other floral resources presented  ~26C +i",    "17 minutes",     "Stenocarpus sinuatus",   "JB Dorey",       "Keybase", "DOREY, J. B. 2021. Missing for almost 100 years: the rare and potentially threatened bee Pharohylaeus lactiferus (Hymenoptera, Colltidae). Journal of Hymenoptera Research, 81, 165-180.",                            10L,          "Dorey", "DOREY, J. B. 2021. Missing for almost 100 years: the rare and potentially threatened bee Pharohylaeus lactiferus (Hymenoptera, Colltidae). Journal of Hymenoptera Research, 81, 165-180.",
      "preservedSpecimen",             NA, "18JDEC497-A", "JBD_Site365", "Colletidae", "Hylaeinae", "Pharohylaeus",     "lactiferus", "Pharohylaeus lactiferus",    "JB Dorey",     "M",               1L,        -17.26761,         145.49474,               864L,            "QLD", "Hallorans Hill",     "3/2/19",    "11:01",  "JB Dorey", "Australia",              "General/sight swept off Stenocarpus sinuatus. Males of Pharohylaeus frequently seen patrolling flowers ~slowly flying around each flower in what appeared to be a circuit. They would sometimes alight on leaves. No females were seen (or were at least caught) despite male activity. Hot and sunny ~30C",    "52 minutes",     "Stenocarpus sinuatus",   "JB Dorey",       "Keybase", "DOREY, J. B. 2021. Missing for almost 100 years: the rare and potentially threatened bee Pharohylaeus lactiferus (Hymenoptera, Colltidae). Journal of Hymenoptera Research, 81, 165-180.",                            10L,          "Dorey", "DOREY, J. B. 2021. Missing for almost 100 years: the rare and potentially threatened bee Pharohylaeus lactiferus (Hymenoptera, Colltidae). Journal of Hymenoptera Research, 81, 165-180.",
      "preservedSpecimen",             NA,    "19JDEC60", "JBD_Site380", "Colletidae", "Hylaeinae", "Pharohylaeus",     "lactiferus", "Pharohylaeus lactiferus",    "JB Dorey", "4M 1F",               5L,        -16.81119,         145.63464,               351L,            "QLD",        "Kuranda", "20/11/2019",    "11:53",  "JB Dorey", "Australia",                                                                                                                                                  "Tall flame tree (~13 m high) on street surrounded by rainforest. Several females and males observed. Bees seen almost immediately upon arrival. Sunny and warm ~30C +i",    "14 minutes", "Brachychiton acerifolius",   "JB Dorey",       "Keybase", "DOREY, J. B. 2021. Missing for almost 100 years: the rare and potentially threatened bee Pharohylaeus lactiferus (Hymenoptera, Colltidae). Journal of Hymenoptera Research, 81, 165-180.",                            10L,          "Dorey", "DOREY, J. B. 2021. Missing for almost 100 years: the rare and potentially threatened bee Pharohylaeus lactiferus (Hymenoptera, Colltidae). Journal of Hymenoptera Research, 81, 165-180.",
      "preservedSpecimen",             NA,    "19JDEC61", "JBD_Site365", "Colletidae", "Hylaeinae", "Pharohylaeus",     "lactiferus", "Pharohylaeus lactiferus",    "JB Dorey", "1M 2F",               3L,        -17.26804,         145.49175,               853L,            "QLD",       "Atherton", "21/11/2019",    "10:30",  "JB Dorey", "Australia", "General sweep of flame tree near rainforest, but on a street. Males and females seen in flowers, but Hylaeinae have pollen covering their backs. Sunny and hot ~27C +i. Additionally: Lasioglossum sp. female, Tetragonula female, Meroglossa itamuca males and females, 3 Hylaeus (Gnathoprosopis) albonitens females.",    "14 minutes", "Brachychiton acerifolius",   "JB Dorey",       "Keybase", "DOREY, J. B. 2021. Missing for almost 100 years: the rare and potentially threatened bee Pharohylaeus lactiferus (Hymenoptera, Colltidae). Journal of Hymenoptera Research, 81, 165-180.",                            10L,          "Dorey", "DOREY, J. B. 2021. Missing for almost 100 years: the rare and potentially threatened bee Pharohylaeus lactiferus (Hymenoptera, Colltidae). Journal of Hymenoptera Research, 81, 165-180."
    )
    # Be sure that the testData is not already in tempdir
    testDataPath <- file.info(list.files(tempdir(), full.names = T, 
                                         pattern = "testData.csv", recursive = TRUE))
    unlink(rownames(testDataPath))
    
    # Save a temporary version of these data
    readr::write_excel_csv(testData, paste0(tempdir(), "/testData.csv"))
    
    testOut1 <- BeeBDC::readr_BeeBDC(dataset = "Dor",
                                 path = tempdir(),
                                 inFile = "/testData.csv",
                                 outFile = "testDataOut.csv",
                                 dataLicense = "https://creativecommons.org/licenses/by-nc-sa/4.0/")
    
    
    # Get a count of TRUE and FALSE column name matches
    resultsT <- sum(colnames(testOut1) %in% (BeeBDC::ColTypeR()[[1]] %>% names()) == TRUE)
    resultsF <- sum(colnames(testOut1) %in% (BeeBDC::ColTypeR()[[1]] %>% names()) == FALSE)
    
    
    # Test the number of expected TRUE and FALSE columns and then test the output format (data frames and
    # tibbles are a special case of lists)
    testthat::test_that("readr_Dor results columns TRUE", {
      testthat::expect_equal(resultsT, 34)
    })
    
    testthat::test_that("readr_Dor results columns FALSE", {
      testthat::expect_equal(resultsF, 4)
    })
    
    testthat::test_that("readr_Dor expected class", {
      testthat::expect_type(testOut1, "list")
    })
    
    
    
#### 22.0 MEPB ####
    
    testData <- dplyr::tribble(
      ~id,            ~type, ~institutionID, ~collectionID, ~institutionCode,                              ~collectionCode,          ~datasetName,      ~basisOfRecord,          ~occurrenceID, ~catalogNumber, ~occurrenceRemarks, ~recordedBy, ~organismQuantity, ~organismQuantityType, ~sex, ~lifeStage, ~preparations, ~associatedTaxa, ~previousIdentifications, ~samplingProtocol, ~eventDate, ~year, ~month, ~day, ~verbatimEventDate,            ~habitat, ~continent,   ~country, ~countryCode, ~stateProvince,          ~county,                                        ~locality,                    ~verbatimLocality, ~verbatimElevation, ~minimumElevationInMeters, ~maximumElevationInMeters, ~verbatimCoordinates, ~verbatimLatitude, ~verbatimLongitude, ~decimalLatitude, ~decimalLongitude, ~geodeticDatum, ~coordinateUncertaintyInMeters,               ~georeferencedBy, ~georeferencedDate,                                                                                                                                                                                                                                                                                                                                                                              ~georeferenceProtocol,                                                                                                                                                                                                                                                                                                                                    ~georeferenceSources,                                                                                                                                                                                 ~georeferenceRemarks, ~identifiedBy, ~dateIdentified, ~identificationRemarks, ~identificationQualifier, ~scientificName,   ~kingdom,      ~phylum,    ~class,        ~order,  ~family,     ~genus, ~subgenus, ~specificEpithet, ~infraspecificEpithet, ~taxonRank, ~verbatimTaxonRank, ~scientificNameAuthorship, ~taxonomicStatus,
      "COMFENALCO:MEPB:955", "Objeto fÃ.sico",  "890900842-6",     "RNC:147",     "COMFENALCO", "Museo EntomolÃ³gico Piedras Blancas (MEPB)", "ColecciÃ³n INDERENA", "PreservedSpecimen",  "COMFENALCO:MEPB:955",           955L,                 NA, "P Cardona",                1L,          "Individuos",   NA,   "Adulto",        "Seco",              NA,                 "Apidae",                NA,     "2002", 2002L,     NA,   NA,             "2002",          "Hotel***",       "SA", "Colombia",         "CO",     "QuindÃ.o",       "Quimbaya",           "Vereda Kerman, Km 7 al Parque Panaca", "Vereda Kerman, Km 7, Parque Panaca",      "1339 - 1339",                     1182L,                     1182L,                   NA,                NA,                 NA,          4.60781,         -75.82079,        "WGS84",                            671, "Daniel Suarez (SiB Colombia)",           "4/1/17", "Escobar D, Jojoa LM, DÃ.az SR, Rudas E, AlbarracÃ.n RD, RamÃ.rez C, GÃ³mez JY, LÃ³pez CR, Saavedra J, Ortiz R, (2016). GeorreferenciaciÃ³n de localidades: Una guÃ.a de referencia para colecciones biolÃ³gicas. Instituto de InvestigaciÃ³n de Recursos BiolÃ³gicos Alexander von Humboldt â€“ Instituto de Ciencias Naturales, Universidad Nacional de Colombia. BogotÃ¡ D.C., Colombia. 144 p", "Plancha 224 del Instituto GeogrÃ¡fico AgustÃ.n Codazzi - IGAC. Base cartogrÃ¡fica oficial integrada.Escala 1:100.000. CodificaciÃ³n de la divisiÃ³n polÃ.tico administrativa de Colombia(Divipola) - DANE. NASA Land Processes Distributed Active Archive Center (LP DAAC). ASTER Global DEM. EOSDIS/Reverb ECHO. https://reverb.echo.nasa.gov. 2011.",                                         "Nivel 2. Se valida la localidad. Se valida la geografÃ.a superior. No reporta coordenadas. Incertidumbre por extensiÃ³n de la localidad y escala del mapa.",            NA,              NA,                     NA,                       NA,        "Apidae", "Animalia", "Arthropoda", "Insecta", "Hymenoptera", "Apidae",         NA,        NA,               NA,                    NA,  "Familia",              "sp.",                        NA,               NA,
      "COMFENALCO:MEPB:7534", "Objeto fÃ.sico",  "890900842-6",     "RNC:147",     "COMFENALCO", "Museo EntomolÃ³gico Piedras Blancas (MEPB)",           "DonaciÃ³n", "PreservedSpecimen", "COMFENALCO:MEPB:7534",          7534L,                 NA, "P Cardona",                1L,          "Individuos",   NA,   "Adulto",        "Seco",              NA,                 "Apidae",  "Captura manual",     "2002", 2002L,     NA,   NA,             "2002", "Intradomicilio***",       "SA", "Colombia",         "CO",     "QuindÃ.o",       "Quimbaya", "Parque Panaca, Km 7 de la vereda Kerman, hotel",  "Panaca, Km 7 Vereda Kerman, Hotel",                 NA,                     1182L,                     1182L,                   NA,                NA,                 NA,          4.60948,          -75.8252,        "WGS84",                            296, "Daniel Suarez (SiB Colombia)",           "4/1/17", "Escobar D, Jojoa LM, DÃ.az SR, Rudas E, AlbarracÃ.n RD, RamÃ.rez C, GÃ³mez JY, LÃ³pez CR, Saavedra J, Ortiz R, (2016). GeorreferenciaciÃ³n de localidades: Una guÃ.a de referencia para colecciones biolÃ³gicas. Instituto de InvestigaciÃ³n de Recursos BiolÃ³gicos Alexander von Humboldt â€“ Instituto de Ciencias Naturales, Universidad Nacional de Colombia. BogotÃ¡ D.C., Colombia. 144 p", "Plancha 224 del Instituto GeogrÃ¡fico AgustÃ.n Codazzi - IGAC. Base cartogrÃ¡fica oficial integrada.Escala 1:100.000. CodificaciÃ³n de la divisiÃ³n polÃ.tico administrativa de Colombia(Divipola) - DANE. NASA Land Processes Distributed Active Archive Center (LP DAAC). ASTER Global DEM. EOSDIS/Reverb ECHO. https://reverb.echo.nasa.gov. 2011.",                                         "Nivel 2. Se valida la localidad. Se valida la geografÃ.a superior. No reporta coordenadas. Incertidumbre por extensiÃ³n de la localidad y escala del mapa.",            NA,              NA,                     NA,                       NA,        "Apidae", "Animalia", "Arthropoda", "Insecta", "Hymenoptera", "Apidae",         NA,        NA,               NA,                    NA,  "Familia",              "sp.",                        NA,               NA,
      "COMFENALCO:MEPB:1175", "Objeto fÃ.sico",  "890900842-6",     "RNC:147",     "COMFENALCO", "Museo EntomolÃ³gico Piedras Blancas (MEPB)", "ColecciÃ³n INDERENA", "PreservedSpecimen", "COMFENALCO:MEPB:1175",          1175L,                 NA,          NA,                1L,          "Individuos",   NA,   "Adulto",        "Seco",              NA,               "Xylocopa",                NA,   "7/1/83", 1983L,     7L,   1L,           "7/1/83",                  NA,       "SA", "Colombia",         "CO",    "Antioquia",          "Urrao",                                      "Sin Datos",                                   NA,                 NA,                     2250L,                     2250L,                   NA,                NA,                 NA,      6.337984698,       -76.2756657,        "WGS84",                      426237168,             "Ricardo Ortiz G.",          "6/12/16", "Escobar D, Jojoa LM, DÃ.az SR, Rudas E, AlbarracÃ.n RD, RamÃ.rez C, GÃ³mez JY, LÃ³pez CR, Saavedra J, Ortiz R, (2016). GeorreferenciaciÃ³n de localidades: Una guÃ.a de referencia para colecciones biolÃ³gicas. Instituto de InvestigaciÃ³n de Recursos BiolÃ³gicos Alexander von Humboldt â€“ Instituto de Ciencias Naturales, Universidad Nacional de Colombia. BogotÃ¡ D.C., Colombia. 144 p",   "Instituto GeogrÃ¡fico AgustÃ.n Codazzi - IGAC. Base cartogrÃ¡fica oficial integrada. Plancha 145. Escala 1:100.000. CodificaciÃ³n de la divisiÃ³n polÃ.tico administrativa de Colombia(Divipola) - DANE. NASA Land Processes Distributed Active Archive Center (LP DAAC). ASTER Global DEM. EOSDIS/Reverb ECHO. https://reverb.echo.nasa.gov. 2011.", "Nivel 4. Georreferenciado en el centroide del municipio. La incertidumbre corresponde a la distancia desde el centroide hasta el lÃ.mite mas alejado del municipio mas la incertidumbre por escala",            NA,              NA,                     NA,                       NA,      "Xylocopa", "Animalia", "Arthropoda", "Insecta", "Hymenoptera", "Apidae", "Xylocopa",        NA,               NA,                    NA,  "GÃ©nero",              "sp.",         "Latreille, 1802",       "Aceptado",
      "COMFENALCO:MEPB:7531", "Objeto fÃ.sico",  "890900842-6",     "RNC:147",     "COMFENALCO", "Museo EntomolÃ³gico Piedras Blancas (MEPB)",           "DonaciÃ³n", "PreservedSpecimen", "COMFENALCO:MEPB:7531",          7531L,                 NA, "A Acevedo",                1L,          "Individuos",   NA,   "Adulto",        "Seco",              NA,                 "Apidae",            "Jama",   "3/1/84", 1984L,     3L,   1L,           "3/1/84",                  NA,       "SA", "Colombia",         "CO",    "Antioquia", "Puerto Triunfo",                                      "Sin Datos",                                   NA,                 NA,                      167L,                      167L,                   NA,                NA,                 NA,      5.953695157,      -74.68633006,        "WGS84",                      180120452,             "Ricardo Ortiz G.",          "6/12/16", "Escobar D, Jojoa LM, DÃ.az SR, Rudas E, AlbarracÃ.n RD, RamÃ.rez C, GÃ³mez JY, LÃ³pez CR, Saavedra J, Ortiz R, (2016). GeorreferenciaciÃ³n de localidades: Una guÃ.a de referencia para colecciones biolÃ³gicas. Instituto de InvestigaciÃ³n de Recursos BiolÃ³gicos Alexander von Humboldt â€“ Instituto de Ciencias Naturales, Universidad Nacional de Colombia. BogotÃ¡ D.C., Colombia. 144 p",   "Instituto GeogrÃ¡fico AgustÃ.n Codazzi - IGAC. Base cartogrÃ¡fica oficial integrada. Plancha 168. Escala 1:100.000. CodificaciÃ³n de la divisiÃ³n polÃ.tico administrativa de Colombia(Divipola) - DANE. NASA Land Processes Distributed Active Archive Center (LP DAAC). ASTER Global DEM. EOSDIS/Reverb ECHO. https://reverb.echo.nasa.gov. 2011.", "Nivel 4. Georreferenciado en el centroide del municipio. La incertidumbre corresponde a la distancia desde el centroide hasta el lÃ.mite mas alejado del municipio mas la incertidumbre por escala",            NA,              NA,                     NA,                       NA,        "Apidae", "Animalia", "Arthropoda", "Insecta", "Hymenoptera", "Apidae",         NA,        NA,               NA,                    NA,  "Familia",              "sp.",                        NA,               NA
    ) %>%
      dplyr::mutate(year = year %>% as.double(),
                    month = month %>% as.double(),
                    day = day %>% as.double())
    
    # Be sure that the testData is not already in tempdir
    testDataPath <- file.info(list.files(tempdir(), full.names = T, 
                                         pattern = "testData.xlsx", recursive = TRUE))
    unlink(rownames(testDataPath))
    
    # Save a temporary version of these data
    openxlsx::write.xlsx(testData, paste0(tempdir(), "/testData.xlsx"),
                         sheetName = "PiedrasBalncas Bees Data")
    
    
    testOut1 <- BeeBDC::readr_BeeBDC(dataset = "MEPB",
                                   path = tempdir(),
                                  inFile = "testData.xlsx",
                                  outFile = "testDataOut.csv",
                                  sheet = "PiedrasBalncas Bees Data",
                                  dataLicense = "https://creativecommons.org/licenses/by-nc-sa/4.0/")
    
    
    # Get a count of TRUE and FALSE column name matches
    resultsT <- sum(colnames(testOut1) %in% (ColTypeR()[[1]] %>% names()) == TRUE)
    resultsF <- sum(colnames(testOut1) %in% (ColTypeR()[[1]] %>% names()) == FALSE)
    
    
    # Test the number of expected TRUE and FALSE columns and then test the output format (data frames and
    # tibbles are a special case of lists)
    testthat::test_that("readr_MEPB results columns TRUE", {
      testthat::expect_equal(resultsT, 51)
    })
    
    testthat::test_that("readr_MEPB results columns FALSE", {
      testthat::expect_equal(resultsF, 19)
    })
    
    testthat::test_that("readr_MEPB expected class", {
      testthat::expect_type(testOut1, "list")
    })
    




#### 23.0 Brazil ####
    
    testData <- dplyr::tribble(
      ~CodeBBdatabase_curated, ~'Scientific name corrected', ~Native.to.Brazil, ~Family_Proposed.in.Moure.Catalogue,      ~Family,   ~SubFamily,        ~Tribe,            ~Codigo, ~institutioncode, ~Day, ~Month, ~Year, ~Date_precision, ~Country,           ~State, ~Latitude_dec.degrees, ~Longitude_dec.degrees, ~Precision.of.coord.meters, ~NotasLatLong, ~NotesOnLocality,                     ~Locality.original, ~Spcslink.county, ~Spcslink.continentocean,               ~Collector, ~Collection,                                                                                                                                                                                     ~Source,         ~Sex,                                                                                                                                                                                    ~Project,        ~Det_By, ~Spcslink.linha_numero, ~Spcslink.datelastmodified, ~Spcslink.collectioncode, ~Spcslink.basisofrecord, ~Spcslink.identifiedby, ~Spcslink.yearidentified, ~Spcslink.barcode,
      "BBD_498851",      "Secret species",             "yes",                            "Apidae", "Andrenidae", "Andreninae", "Calliopsini",                 NA,               NA,   NA,     NA,    NA,              NA, "Brazil", "Santa Catarina",          -27.14783261,           -52.34069428,    "municipality centroid",            NA,               NA,                                     NA,               NA,                       NA,  "Moure's Bee Catalogue",          NA,                                                                                                                                                                                          NA,           NA,                                                                                                                                                                                          NA,             NA,                     NA,                         NA,                       NA,                      NA,                     NA,                       NA,                NA,
      "BBD_00082426",      "Secret species",             "yes",                            "Apidae", "Andrenidae", "Andreninae", "Calliopsini",                 NA,             "KU",   NA,    11L, 1939L,              NA, "Brazil",      "Sao Paulo",             -23.53377,              -46.69189,                         NA,            NA,               NA,                            "Sao Paulo",               NA,            "Neotropical",                "unknown",   "KU-SEMC",                                                                                                                                                             "SpeciesLink 22/12/2020 17:23h",           NA,                                                                                                                                                                                          NA,             NA,                270288L,                40933.39514,                 100623465863.0,     "PreservedSpecimen",           "Ruz, Luisa",                       NA,                NA,
      "BBD_00082419",      "Secret species",             "yes",                            "Apidae", "Andrenidae", "Andreninae", "Calliopsini",                 NA,             "KU",  26L,     2L, 1954L,              NA, "Brazil", "Santa Catarina",           -27.1833333,            -52.3833333,                         NA,            NA,               NA,                        "Nova Teutonia",               NA,            "Neotropical",        "Plaumann, Fritz",   "KU-SEMC",                                                                                                                                                             "SpeciesLink 22/12/2020 17:23h",           NA,                                                                                                                                                                                          NA,             NA,                270340L,                40933.39375,                 1006265758.0,     "PreservedSpecimen",                     NA,                       NA,                NA,
      "BBD_00043536",      "Secret species",             "yes",                            "Apidae", "Andrenidae", "Andreninae", "Calliopsini", "AMNHBEE 00028690",               NA,   1L,     1L, 1970L,              NA, "BRAZIL",   "Minas Gerais",             -20.13333,                  -43.5,                         NA,            NA,               NA, "Serra do Caraca (Saof Santa Barbara)",               NA,                       NA,         "F. M. Oliveira",      "AMNH", "Data citation: Digital Bee Collections Network, 2014 (and updates). Version: <download date>. National Science Foundation grant DBI 0956388; Contact: Ascher, John S. (dbsajs@nus.edu.sg)", "Adult Male", "Data citation: Digital Bee Collections Network, 2014 (and updates). Version: <download date>. National Science Foundation grant DBI 0956388; Contact: Ascher, John S. (dbsajs@nus.edu.sg)", "J. S. Ascher",                     NA,                         NA,                       NA,                      NA,                     NA,                       NA,                NA,
      "BBD_00082342",      "Secret species",             "yes",                            "Apidae", "Andrenidae", "Andreninae", "Calliopsini",                 NA,             "KU",   NA,    11L, 1951L,              NA, "Brazil", "Santa Catarina",           -27.1833333,            -52.3833333,                         NA,            NA,               NA,                        "Nova Teutonia",               NA,            "Neotropical",            "Plaumann, L",   "KU-SEMC",                                                                                                                                                             "SpeciesLink 22/12/2020 17:23h",           NA,                                                                                                                                                                                          NA,             NA,                270388L,                40933.39375,                 100245624566694.0,     "PreservedSpecimen",                     NA,                       NA,                NA
    )
    
    
    # Be sure that the testData is not already in tempdir
    testDataPath <- file.info(list.files(tempdir(), full.names = T, 
                                         pattern = "testData.csv", recursive = TRUE))
    unlink(rownames(testDataPath))
    
    # Save a temporary version of these data
    readr::write_excel_csv(testData, paste0(tempdir(), "/testData.csv"))
    
    testOut1 <- BeeBDC::readr_BeeBDC(dataset = "BBD",
                                   path = tempdir(),
                                 inFile = "testData.csv",
                                 outFile = "testDataOut.csv",
                                 dataLicense = "All rights reserved")
    
    
    # Get a count of TRUE and FALSE column name matches
    resultsT <- sum(colnames(testOut1) %in% (ColTypeR()[[1]] %>% names()) == TRUE)
    resultsF <- sum(colnames(testOut1) %in% (ColTypeR()[[1]] %>% names()) == FALSE)
    
    # Test the number of expected TRUE and FALSE columns and then test the output format (data frames and
    # tibbles are a special case of lists)
    testthat::test_that("readr_BBD results columns TRUE", {
      testthat::expect_equal(resultsT, 27)
    })
    testthat::test_that("readr_BBD results columns FALSE", {
      testthat::expect_equal(resultsF, 15)
    })
    
    testthat::test_that("readr_BBD expected class", {
      testthat::expect_type(testOut1, "list")
    })
    
    




#### 24.0 MPUJ ####
    
    testData <- dplyr::tribble(
      ~Catalog.Number, ~`Collectors/First.Name`, ~`Collectors/Last.Name`, ~Count, ~Sex, ~Stage, ~Reproductive.Condition, ~Behavior,             ~Name, ~Alt.Cat.Number, ~Associated.Taxa, ~Associated.Ocurrence, ~Method,  ~Start.Date,    ~End.Date, ~`Start.Date.(Year)`, ~`Start.Date.(Month)`, ~`Start.Date.(Day)`, ~Start.Time, ~End.Time, ~Verbatim.Date, ~Habitat,      ~Continent,   ~Country,         ~State,                      ~County,                                               ~Locality.Name, ~Min.Elevation, ~Max.Elevation, ~Locality.and.Habitat.Notes,   ~Latitude1, ~Latitude2,     ~Lat1text,    ~Longitude1, ~Longitude2,     ~Long1text,               ~Full.Name,   ~Kingdom,        ~Order,  ~Family,    ~Subfamily,         ~Genus,      ~Species, ~Subspecies,     ~Species.Author, ~Type.Status, ~Qualifier, ~`Determiner/Last.Name`, ~`Determiner/First.Name`,
      "MPUJ_ENT0046822",                       NA,                      NA,     1L,   NA,     NA,                      NA,        NA, "Montado en seco",              NA,               NA,                    NA,      NA, "29-08-1987",           NA,                1987L,                    8L,                 29L,          NA,        NA,             NA,       NA, "South America", "Colombia",       "Tolima", "San Sebastián de Mariquita",                                                  "Mariquita",             NA,             NA,                          NA,     5.198894,         NA, "5.198894° N",      -74.89295,          NA,  "74.89295° W",   "Xylocopa aeneipennis", "Animalia", "Hymenoptera", "Apidae", "Xylocopinae",     "Xylocopa", "aeneipennis",          NA,    "(DeGeer, 1773)",           NA,         NA,               "Zanella",               "Fernando",
      "MPUJ_ENT0046821",                      "N",                 "Novoa",     1L,   NA,     NA,                      NA,        NA, "Montado en seco",              NA,               NA,                    NA,      NA, "27-03-2009",           NA,                2009L,                    3L,                 27L,          NA,        NA,             NA,       NA, "South America", "Colombia",        "Chocó",                     "Acandi",                      "Capurganá, Jardín Botánico del Darién",             90,             90,                          NA,     8.639755,         NA, "8.639755° N",     -77.350424,          NA, "77.350424° W",   "Xylocopa aeneipennis", "Animalia", "Hymenoptera", "Apidae", "Xylocopinae",     "Xylocopa", "aeneipennis",          NA,    "(DeGeer, 1773)",           NA,         NA,        "Guevara Farias",        "Diego Alexander",
      "MPUJ_ENT0089543",                      "R",                "Ovalle",     1L,   NA,     NA,                      NA,        NA, "Montado en seco",              NA,               NA,                    NA,      NA, "15-02-1986",           NA,                1986L,                    2L,                 15L,          NA,        NA,             NA,       NA, "South America", "Colombia", "Bogotá, D.C.",               "Bogotá, D.C.",                    "Las Villas, alrededores de la plazoleta",           2680,             NA,                          NA,           NA,         NA,            NA,             NA,          NA,             NA,      "Thygater aethiops", "Animalia", "Hymenoptera", "Apidae",      "Apinae",     "Thygater",    "aethiops",          NA,     "(Smith, 1854)",           NA,         NA,        "Guevara Farias",        "Diego Alexander",
      "MPUJ_ENT0089540",                      "R",                "Ovalle",     1L,   NA,     NA,                      NA,        NA, "Montado en seco",              NA,               NA,                    NA,      NA, "15-02-1986",           NA,                1986L,                    2L,                 15L,          NA,        NA,             NA,       NA, "South America", "Colombia", "Bogotá, D.C.",               "Bogotá, D.C.",                    "Las Villas, alrededores de la plazoleta",           2680,             NA,                          NA,           NA,         NA,            NA,             NA,          NA,             NA,      "Thygater aethiops", "Animalia", "Hymenoptera", "Apidae",      "Apinae",     "Thygater",    "aethiops",          NA,     "(Smith, 1854)",           NA,         NA,        "Guevara Farias",        "Diego Alexander",
      "MPUJ_ENT0036978",                      "M",                "Nariño",     1L,   NA,     NA,                      NA,        NA, "Montado en seco",              NA,               NA,                    NA,      NA, "07-03-2016", "11-03-2016",                2016L,                    3L,                  7L,          NA,        NA,             NA,       NA, "South America", "Colombia",       "Boyacá",                "Santa María", "sendero Hyca Quye, aproximadamente 5.5km NW de Santa María",            900,            900,                          NA,      4.89811,         NA,  "4.89811° N",      -73.29344,          NA,  "73.29344° W",    "Paratrigona anduzei", "Animalia", "Hymenoptera", "Apidae",      "Apinae",  "Paratrigona",     "anduzei",          NA,   "(Schwarz, 1943)",           NA,         NA,        "Guevara Farias",        "Diego Alexander",
      "MPUJ_ENT0072731",                "Rodrigo",                "García",     1L,   NA,     NA,                      NA,        NA, "Montado en seco",              NA,               NA,                    NA,      NA, "01-12-2021", "03-12-2021",                2021L,                   12L,                  1L,          NA,        NA,             NA,       NA, "South America", "Colombia",       "Caldas",                     "Samana",                                        "Vereda Monte Cristo",            850,             NA,                          NA,     5.503694,         NA, "5.503694° N",      -75.00061,          NA,  "75.00061° W", "Tetragonisca angustula", "Animalia", "Hymenoptera", "Apidae",      "Apinae", "Tetragonisca",   "angustula",          NA, "(Latreille, 1811)",           NA,         NA,        "Guevara Farias",        "Diego Alexander"
    )
    
    # need to change column names, which need spaces for function to work
    colnames(testData) <- c("Catalog Number", "Collectors/First Name", "Collectors/Last Name", "Count", "Sex", "Stage", "Reproductive Condition", "Behavior", "Name", "Alt Cat Number", "Associated Taxa", "Associated Ocurrence", "Method", "Start Date", "End Date", "Start Date (Year)", "Start Date (Month)", "Start Date (Day)", "Start Time", "End Time", "Verbatim Date", "Habitat", "Continent", "Country", "State", "County", "Locality Name", "Min Elevation", "Max Elevation", "Locality and Habitat Notes", "Latitude1", "Latitude2", "Lat1text", "Longitude1", "Longitude2", "Long1text", "Full Name", "Kingdom", "Order", "Family", "Subfamily", "Genus", "Species", "Subspecies", "Species Author", "Type Status", "Qualifier", "Determiner/Last Name", "Determiner/First Name")
    
    # Be sure that the testData is not already in tempdir
    testDataPath <- file.info(list.files(tempdir(), full.names = T, 
                                         pattern = "testData.xlsx", recursive = TRUE))
    unlink(rownames(testDataPath))
    
    # Save a temporary version of these data
    openxlsx::write.xlsx(testData, paste0(tempdir(), "/testData.xlsx"), sheetName = "Sheet1")
    
    testOut1 <- BeeBDC::readr_BeeBDC(dataset = "MPUJ",
                                   path = tempdir(),
                                  inFile = "testData.xlsx",
                                  outFile = "testDataOut.csv",
                                  sheet = "Sheet1", 
                                  dataLicense = "https://creativecommons.org/licenses/by-nc-sa/4.0/")
    
    
    # Get a count of TRUE and FALSE column name matches
    resultsT <- sum(colnames(testOut1) %in% (ColTypeR()[[1]] %>% names()) == TRUE)
    resultsF <- sum(colnames(testOut1) %in% (ColTypeR()[[1]] %>% names()) == FALSE)
    
    
    # Test the number of expected TRUE and FALSE columns and then test the output format (data frames and
    # tibbles are a special case of lists)
    testthat::test_that("readr_MPUJ results columns TRUE", {
      testthat::expect_equal(resultsT, 41)
    })
    
    testthat::test_that("readr_MPUJ results columns FALSE", {
      testthat::expect_equal(resultsF, 17)
    })
    
    testthat::test_that("readr_MPUJ expected class", {
      testthat::expect_type(testOut1, "list")
    })
    



#### 25.0 STRI ####
    
    testData <- dplyr::tribble(
      ~Catalognumber, ~institutionCode, ~collectionCode, ~ownerInstitutionCode,                          ~collectionID,      ~basisOfRecord,                          ~occurrenceID, ~catalogNumber, ~otherCatalogNumbers,                                                                                           ~higherClassification,   ~kingdom,      ~phylum,    ~class,        ~order,  ~family,       ~scientificName, ~taxonID, ~scientificNameAuthorship,     ~genus, ~subgenus, ~specificEpithet, ~verbatimTaxonRank, ~infraspecificEpithet, ~taxonRank,  ~identifiedBy, ~dateIdentified, ~identificationReferences, ~identificationRemarks, ~taxonRemarks, ~identificationQualifier, ~typeStatus,      ~recordedBy, ~recordNumber, ~eventDate, ~year, ~month, ~day, ~startDayOfYear, ~endDayOfYear, ~verbatimEventDate,                                                                                                ~occurrenceRemarks, ~habitat, ~fieldNumber, ~eventID, ~informationWithheld, ~dataGeneralizations, ~dynamicProperties, ~associatedOccurrences, ~associatedSequences, ~associatedTaxa, ~reproductiveCondition, ~establishmentMeans, ~lifeStage, ~sex, ~individualCount, ~preparations, ~locationID, ~continent, ~waterBody, ~islandGroup, ~island,        ~country,     ~stateProvince, ~county, ~municipality,        ~locality, ~locationRemarks, ~decimalLatitude, ~decimalLongitude, ~geodeticDatum, ~coordinateUncertaintyInMeters,  ~verbatimCoordinates, ~georeferencedBy, ~georeferenceProtocol, ~georeferenceSources, ~georeferenceVerificationStatus, ~georeferenceRemarks, ~minimumElevationInMeters, ~maximumElevationInMeters, ~minimumDepthInMeters, ~maximumDepthInMeters, ~verbatimDepth, ~verbatimElevation, ~disposition, ~language, ~recordEnteredBy,      ~modified,                                             ~rights, ~rightsHolder, ~accessRights,                              ~recordID,                                                                  ~references,
      911208L,      "Field-Obs",    "Entomology",                    NA, "6e16d724-0be0-4101-a230-b841ebd1f0a5", "PreservedSpecimen", "48739ca8-e401-43c2-b52a-137ad191d726",             NA,                   NA, "Organism|Animalia|Arthropoda|Insecta|Pterygota|Hymenoptera|Apocrita|Apoidea|Apidae|Apinae|Euglossini|Euglossa", "Animalia", "Arthropoda", "Insecta", "Hymenoptera", "Apidae", "Euglossa allosticta",   48972L,             "Moure, 1969", "Euglossa",        NA,     "allosticta",                 NA,                    NA,  "Species", "David Roubik",              NA,                        NA,                     NA,            NA,                       NA,          NA,   "David Roubik",            NA,   "2/2/06", 2006L,     2L,   2L,             33L,            NA,                 NA,                                   "Euglossa allosticta; Euglossa allosticta, Face. Taken with Automontage system",       NA,           NA,       NA,                   NA,                   NA,                 NA,                     NA,                   NA,              NA,                     NA,                  NA,         NA,   NA,               NA,            NA,          NA,         NA,         NA,           NA,      NA, "French Guiana",                 NA,      NA,            NA,               NA,               NA,               NA,                NA,             NA,                             NA,                    NA,               NA,                    NA,                   NA,                              NA,                   NA,                        NA,                        NA,                    NA,                    NA,             NA,                 NA,           NA,        NA,               NA, "1/8/15 11:31", "http://creativecommons.org/publicdomain/zero/1.0/",            NA,            NA, "62e76208-6975-43ae-b805-5f02766fc9ce", "https://panamabiota.org/stri/collections/individual/index.php?occid=911208",
      911091L,      "Field-Obs",    "Entomology",                    NA, "6e16d724-0be0-4101-a230-b841ebd1f0a5", "PreservedSpecimen", "a4e323b1-f79d-4826-86ba-26c00c7ba7ea",             NA,                   NA,  "Organism|Animalia|Arthropoda|Insecta|Pterygota|Hymenoptera|Apocrita|Apoidea|Apidae|Apinae|Meliponini|Trigona", "Animalia", "Arthropoda", "Insecta", "Hymenoptera", "Apidae",    "Trigona amalthea",   55214L,         "(Olivier, 1789)",  "Trigona",        NA,       "amalthea",                 NA,                    NA,  "Species",             NA,              NA,                        NA,                     NA,            NA,                       NA,          NA, "Miriam Arrueta",            NA,         NA,    NA,     NA,   NA,              NA,            NA,                 NA,  "Orchid Bee: Trigona amalthea; David Roubik Cerro Brewster Orchid Bee collection: Trigona amalthea, Dorsal view",       NA,           NA,       NA,                   NA,                   NA,                 NA,                     NA,                   NA,              NA,                     NA,                  NA,         NA,   NA,               NA,            NA,          NA,         NA,         NA,           NA,      NA,        "PANAMA",         "SAN BLAS",      NA,            NA, "CERRO BREWSTER",               NA,         9.370558,        -79.247071,             NA,                             NA, "17 692500E 1036300N",               NA,                    NA,                   NA,                              NA,                   NA,                        NA,                        NA,                    NA,                    NA,             NA,                 NA,           NA,        NA,               NA, "1/8/15 11:31", "http://creativecommons.org/publicdomain/zero/1.0/",            NA,            NA, "2c2a5b65-de7d-4ceb-9445-ada9fcfbb7ad", "https://panamabiota.org/stri/collections/individual/index.php?occid=911091",
      911092L,      "Field-Obs",    "Entomology",                    NA, "6e16d724-0be0-4101-a230-b841ebd1f0a5", "PreservedSpecimen", "9e7a88c2-09be-4b9c-b59b-f458ea424800",             NA,                   NA,  "Organism|Animalia|Arthropoda|Insecta|Pterygota|Hymenoptera|Apocrita|Apoidea|Apidae|Apinae|Meliponini|Trigona", "Animalia", "Arthropoda", "Insecta", "Hymenoptera", "Apidae",    "Trigona amalthea",   55214L,         "(Olivier, 1789)",  "Trigona",        NA,       "amalthea",                 NA,                    NA,  "Species",             NA,              NA,                        NA,                     NA,            NA,                       NA,          NA, "Miriam Arrueta",            NA,         NA,    NA,     NA,   NA,              NA,            NA,                 NA, "Orchid Bee: Trigona amalthea; David Roubik Cerro Brewster Orchid Bee collection: Trigona amalthea, Oblique view",       NA,           NA,       NA,                   NA,                   NA,                 NA,                     NA,                   NA,              NA,                     NA,                  NA,         NA,   NA,               NA,            NA,          NA,         NA,         NA,           NA,      NA,        "PANAMA", "SAN BLAS (KUNA Y",      NA,            NA, "CERRO BREWSTER",               NA,         9.370558,        -79.247071,             NA,                             NA, "17 692500E 1036300N",               NA,                    NA,                   NA,                              NA,                   NA,                        NA,                        NA,                    NA,                    NA,             NA,                 NA,           NA,        NA,               NA, "1/8/15 11:31", "http://creativecommons.org/publicdomain/zero/1.0/",            NA,            NA, "29b950c2-4d7b-423b-aba1-ea96ccba523b", "https://panamabiota.org/stri/collections/individual/index.php?occid=911092"
    )
    
    
    # Be sure that the testData is not already in tempdir
    testDataPath <- file.info(list.files(tempdir(), full.names = T, 
                                         pattern = "testData.csv", recursive = TRUE))
    unlink(rownames(testDataPath))
    
    # Save a temporary version of these data
    readr::write_excel_csv(testData, paste0(tempdir(), "/testData.csv"))
    
    testOut1 <- BeeBDC::readr_BeeBDC(dataset = "STRI",
                                   path = tempdir(),
                                   inFile = "testData.csv",
                                   outFile = "testDataOut.csv",
                                   dataLicense = "https://creativecommons.org/licenses/by-nc-sa/4.0/")
    
    
    # Get a count of TRUE and FALSE column name matches
    resultsT <- sum(colnames(testOut1) %in% (BeeBDC::ColTypeR()[[1]] %>% names()) == TRUE)
    resultsF <- sum(colnames(testOut1) %in% (BeeBDC::ColTypeR()[[1]] %>% names()) == FALSE)
    
    # Test the number of expected TRUE and FALSE columns and then test the output format (data frames and
    # tibbles are a special case of lists)
    testthat::test_that("readr_STRI results columns TRUE", {
      testthat::expect_equal(resultsT, 65)
    })
    testthat::test_that("readr_STRI results columns FALSE", {
      testthat::expect_equal(resultsF, 34)
    })
    
    testthat::test_that("readr_STRI expected class", {
      testthat::expect_type(testOut1, "list")
    })
    



#### 26.0 PALA ####
    
    testData <- dplyr::tribble(
      ~Type,          ~Country, ~Muninciplaity, ~Gender,                                    ~Site,   ~Latitud, ~Longitude,     ~elevation,          ~date,               ~recordedby, ~catalogNumber, ~Collection, ~othercatalognumber, ~AssociatedTaxa,  ~taxonremarks,      ~Family,         ~Genus,      ~species, ~Citation,
      NA,                NA,   "Tamaulipas",    "1?", "10 km W of Antiguo Morelos, Highway 80",   "22.555",  "-99.167", "elev. 470 m.",  "6 Jul. 1990",         "Ilan Yarom leg.", "PALASEMC0026",      "SEMC",                  NA,              NA, "Lyucatanense", "Halictidae", "Lasioglossum", "yucatanense",        NA,
      NA,        "Campeche",             NA,    "1?",              "10 mi. north of HopelchÈn",    "19.89",   "-89.84",             NA, "17 Apr. 1962",       "L. A. Stange leg.",  "PALAUCD0001",      "UCDC",                  NA,              NA,     "Lpaxtoni", "Halictidae", "Lasioglossum",     "paxtoni",        NA,
      NA, "San Luis Potosi",             NA,    "1?",                 "12 km NW of El Naranjo", "22.59∞ N", "99.38∞ W", "elev. 400 m.",  "5 Jul. 1990", "Robert L. Minckley leg.", "PALASEMC0009",      "SEMC",                  NA,              NA, "Lyucatanense", "Halictidae", "Lasioglossum", "yucatanense",        NA
    )
    
    
    # Be sure that the testData is not already in tempdir
    testDataPath <- file.info(list.files(tempdir(), full.names = T, 
                                         pattern = "testData.csv", recursive = TRUE))
    unlink(rownames(testDataPath))
    
    # Save a temporary version of these data
    readr::write_excel_csv(testData, paste0(tempdir(), "/testData.csv"))
    
    testOut1 <- BeeBDC::readr_BeeBDC(dataset = "PALA",
                                   path = tempdir(),
                                   inFile = "testData.csv",
                                   outFile = "testDataOut.csv",
                                   dataLicense = "https://creativecommons.org/licenses/by-nc-sa/4.0/")
    
    
    # Get a count of TRUE and FALSE column name matches
    resultsT <- sum(colnames(testOut1) %in% (BeeBDC::ColTypeR()[[1]] %>% names()) == TRUE)
    resultsF <- sum(colnames(testOut1) %in% (BeeBDC::ColTypeR()[[1]] %>% names()) == FALSE)
    
    # Test the number of expected TRUE and FALSE columns and then test the output format (data frames and
    # tibbles are a special case of lists)
    testthat::test_that("readr_PALA results columns TRUE", {
      testthat::expect_equal(resultsT, 26)
    })
    testthat::test_that("readr_PALA results columns FALSE", {
      testthat::expect_equal(resultsF, 3)
    })
    
    testthat::test_that("readr_PALA expected class", {
      testthat::expect_type(testOut1, "list")
    })
    
    



#### 27.0 JoLa ####
    
    testData1 <- dplyr::tribble(
      ~Species,    ~Latitude1,     ~Longitude1, ~`Start.Date.(Year)`,
      "angelicum",         33.62,         -117.93,                1920L,
      "angelicum",         33.97,         -117.32,                1924L,
      "angelicum",         33.97,         -117.32,                1925L
    )
    
    testData2 <- dplyr::tribble(
      ~Species,    ~Latitude1,     ~Longitude1, ~`Start.Date.(Year)`,
      "angelicum",         33.97,         -118.44,                1981L,
      "angelicum",         34.15,         -118.14,                2000L,
      "angelicum",         34.15,         -118.14,                2000L
    )
    
    
    testData <- list(testData1, testData2)
    
    # Be sure that the testData is not already in tempdir
    testDataPath <- file.info(list.files(tempdir(), full.names = T, 
                                         pattern = "testData.xlsx", recursive = TRUE))
    unlink(rownames(testDataPath))
    
    # Save a temporary version of these data
    openxlsx::write.xlsx(testData, paste0(tempdir(), "/testData.xlsx"),
                         sheetName = c("pre-1950", "post-1950"))
    
    testOut1 <- BeeBDC::readr_BeeBDC(dataset = "JoLa",
                                   path = tempdir(),
                                   inFile = "testData.xlsx",
                                   outFile = "testDataOut.csv",
                                   sheet = c("pre-1950", "post-1950"),
                                   dataLicense = "https://creativecommons.org/licenses/by-nc-sa/4.0/")
    
    
    
    # Get a count of TRUE and FALSE column name matches
    resultsT <- sum(colnames(testOut1) %in% (BeeBDC::ColTypeR()[[1]] %>% names()) == TRUE)
    resultsF <- sum(colnames(testOut1) %in% (BeeBDC::ColTypeR()[[1]] %>% names()) == FALSE)
    rows <- nrow(testData1) + nrow(testData2)
    
    # Test the number of expected TRUE and FALSE columns and then test the output format (data frames and
    # tibbles are a special case of lists)
    testthat::test_that("readr_JoLa results columns TRUE", {
      testthat::expect_equal(resultsT, 14)
    })
    testthat::test_that("readr_JoLa results columns FALSE", {
      testthat::expect_equal(resultsF, 0)
    })
    testthat::test_that("readr_JoLa correct number of rows", {
      testthat::expect_equal(nrow(testOut1), rows)
    })
    
    testthat::test_that("readr_JoLa expected class", {
      testthat::expect_type(testOut1, "list")
    })
    


    
    #### 28.0 VicWam ####
      # Make the test data
    testData <- dplyr::tribble(
                   ~REGNO, ~Reg_DoreyExtension, ~OLDNO, ~COLLNUM, ~INSTITUTE, ~CATBY,      ~DATENT, ~CHECKEDBY, ~DATMOD, ~on_loan, ~LOAN_NO, ~cabinet, ~drawer,    ~CLASS,        ~ORDER, ~SUPERFAMILY,      ~FAMILY, ~SUBFAMILY, ~TRIBE, ~GENUS, ~SUBGENUS, ~SPECIES, ~SUBSPECIES, ~DTMNDBY, ~DTMNDDT, ~IDMETHOD, ~LIFEHISTORY,              ~SEX, ~NAMEQUALIFIER, ~SPCMTYPE, ~designator, ~DESIGNATORYEAR, ~SPECNUM, ~MULTIPLEMOUNTS, ~PREVIOUSID, ~verbiage,    ~country,            ~STATE, ~ISLANDGR, ~ISLAND,                 ~SITE,       ~NEAREST, ~DISTANCE, ~DIST_UNIT, ~DIRECTION_TX, ~DIRECTION, ~LATITUDE, ~LONGITUDE,    ~LATDEC,    ~LONGDEC,  ~PLACEDERIVATION,                        ~PLACEACCURACY, ~ALTITUDEM, ~ALTITUDEF, ~ALTITUDEDERIVATION, ~ALTITUDEACCURACY,        ~COLLTOR, ~COLLMETH,        ~DTFR, ~DTTO, ~STORAGE, ~quarantine, ~QUARANTINEINFO, ~reared, ~REARINGINFO, ~MICROHABITAT, ~MACROHABITAT, ~HOSTCOMMON,     ~LABELFAMILY,  ~LABELGENUS, ~LABELSPECIES, ~CENSUSGENUS,         ~CENSUSSPECIES,          ~REMARKS, ~voucher, ~VOUCHERINFO,                                                                                         ~NOTES, ~JD_Genus,            ~JD_Species,      ~JD_Genus_Species,             ~JD_PlantGS,     ~JD_Family, ~Bee_GS_asALA, ~JD_PlantGenus_APNI,
                   23798L,         "WAM 23798",     NA,       NA,      "WAM",  "BPH", "07/07/2010",         NA,      NA,       NA,       NA,       NA,      NA, "Insecta", "Hymenoptera",    "Apoidea", "Colletidae",         NA,     NA,     NA,        NA,       NA,          NA,       NA,       NA, "curated",      "adult",                NA,             NA,        NA,          NA,              NA,       1L,              NA,          NA,        NA, "AUSTRALIA",              "SA",        NA,      NA,                    NA,      "Hartley",        4L,       "mi",        "East",         90, "35°12`S", "139°05`E",    "-35.2",  "139.0833", "AUSLIG Database",                  "within 2 km radius",         NA,         NA,                  NA,                NA, "Houston, T.F.",        NA, "10/01/1971",    NA, "pinned",          NA,              NA,      NA,           NA,            NA,            NA,          NA,               NA, "Eucalyptus",            NA,           NA,                     NA,         "on host",       NA,           NA,                                              "donation from Elizabeth Exley (UQ) 21 June 2004",        NA,                     NA,                     NA,            "Eucalyptus",    "Myrtaceae",            NA,                  NA,
                   24021L,         "WAM 24021",     NA,       NA,      "WAM",  "BPH", "08/07/2010",         NA,      NA,       NA,       NA,       NA,      NA, "Insecta", "Hymenoptera",    "Apoidea", "Colletidae",         NA,     NA,     NA,        NA,       NA,          NA,       NA,       NA, "curated",      "adult", "unsexed/unknown",             NA,        NA,          NA,              NA,       1L,              NA,          NA,        NA, "AUSTRALIA",              "SA",        NA,      NA,                    NA,      "Minnipa",       48L,       "km",         "NNE",       22.5, "32°25`S", "135°09`E", "-32.4166",    "135.15", "AUSLIG Database", "within 1 km radius (nearest minute)",         NA,         NA,                  NA,                NA, "Houston, T.F.",        NA, "18/10/1973",    NA, "pinned",          NA,              NA,      NA,           NA,            NA,            NA,          NA,               NA, "Eucalyptus",            NA,           NA,                     NA,         "on host",       NA,           NA,                                              "donation from Elizabeth Exley (UQ) 21 June 2004",        NA,                     NA,                     NA,            "Eucalyptus",    "Myrtaceae",            NA,                  NA,
                   24034L,         "WAM 24034",     NA,       NA,      "WAM",  "BPH", "08/07/2010",         NA,      NA,       NA,       NA,       NA,      NA, "Insecta", "Hymenoptera",    "Apoidea", "Colletidae",         NA,     NA,     NA,        NA,       NA,          NA,       NA,       NA, "curated",      "adult",                NA,             NA,        NA,          NA,              NA,       1L,              NA,          NA,        NA, "AUSTRALIA",              "SA",        NA,      NA,                    NA,      "Hartley",        4L,       "mi",        "East",         90, "35°12`S", "139°05`E",    "-35.2",  "139.0833", "AUSLIG Database",                  "within 2 km radius",         NA,         NA,                  NA,                NA, "Houston, T.F.",        NA, "10/01/1971",    NA, "pinned",          NA,              NA,      NA,           NA,            NA,            NA,          NA,               NA, "Eucalyptus",            NA,           NA,                     NA,         "on host",       NA,           NA,                                              "donation from Elizabeth Exley (UQ) 21 June 2004",        NA,                     NA,                     NA,            "Eucalyptus",    "Myrtaceae",            NA,                  NA,
                   82259L,         "WAM 82259",     NA,       NA,      "WAM",  "BPH", "08/07/2010",         NA,      NA,       NA,       NA,       NA,      NA, "Insecta", "Hymenoptera",    "Apoidea", "Colletidae",         NA,     NA,     NA,        NA,       NA,          NA,       NA,       NA, "curated",      "adult",                NA,             NA,        NA,          NA,              NA,       1L,              NA,          NA,        NA, "AUSTRALIA",              "SA",        NA,      NA,                    NA,  "Oodla Wirra",        6L,       "mi",          "NE",         45, "32°49`S", "139°08`E", "-32.8166",  "139.1333", "AUSLIG Database",                  "within 2 km radius",         NA,         NA,                  NA,                NA, "Houston, T.F.",        NA, "17/12/1969",    NA, "pinned",          NA,              NA,      NA,           NA,            NA,            NA,          NA,               NA, "Eucalyptus",            NA,           NA,                     NA,         "on host",       NA,           NA, "donation from Elizabeth Exley (UQ) 21 June 2004; 82259 to 82260 double-mounted onto same pin",        NA,                     NA,                     NA,            "Eucalyptus",    "Myrtaceae",            NA,                  NA,
                   82260L,         "WAM 82260",     NA,       NA,      "WAM",  "BPH", "08/07/2010",         NA,      NA,       NA,       NA,       NA,      NA, "Insecta", "Hymenoptera",    "Apoidea", "Colletidae",         NA,     NA,     NA,        NA,       NA,          NA,       NA,       NA, "curated",      "adult",                NA,             NA,        NA,          NA,              NA,       1L,              NA,          NA,        NA, "AUSTRALIA",              "SA",        NA,      NA,                    NA,  "Oodla Wirra",        6L,       "mi",          "NE",         45, "32°49`S", "139°08`E", "-32.8166",  "139.1333", "AUSLIG Database",                  "within 2 km radius",         NA,         NA,                  NA,                NA, "Houston, T.F.",        NA, "17/12/1969",    NA, "pinned",          NA,              NA,      NA,           NA,            NA,            NA,          NA,               NA, "Eucalyptus",            NA,           NA,                     NA,         "on host",       NA,           NA, "donation from Elizabeth Exley (UQ) 21 June 2004; 82259 to 82260 double-mounted onto same pin",        NA,                     NA,                     NA,            "Eucalyptus",    "Myrtaceae",            NA,                  NA,
                   82261L,         "WAM 82261",     NA,       NA,      "WAM",  "BPH", "08/07/2010",         NA,      NA,       NA,       NA,       NA,      NA, "Insecta", "Hymenoptera",    "Apoidea", "Colletidae",         NA,     NA,     NA,        NA,       NA,          NA,       NA,       NA, "curated",      "adult",                NA,             NA,        NA,          NA,              NA,       1L,              NA,          NA,        NA, "AUSTRALIA",              "SA",        NA,      NA,                    NA,  "Oodla Wirra",        6L,       "mi",          "NE",         45, "32°49`S", "139°08`E", "-32.8166",  "139.1333", "AUSLIG Database",                  "within 2 km radius",         NA,         NA,                  NA,                NA, "Houston, T.F.",        NA, "17/12/1969",    NA, "pinned",          NA,              NA,      NA,           NA,            NA,            NA,          NA,               NA, "Eucalyptus",            NA,           NA,                     NA,         "on host",       NA,           NA,                                              "donation from Elizabeth Exley (UQ) 21 June 2004",        NA,                     NA,                     NA,            "Eucalyptus",    "Myrtaceae",            NA,                  NA,
                   41780L,         "HYM 41780",     NA,       NA,      "Vic",  "HYM",           NA,         NA,      NA,       NA,       NA,       NA,      NA,        NA, "Hymenoptera",           NA, "Colletidae",         NA,     NA,     NA,        NA,       NA,          NA,       NA,       NA,        NA,           NA,         "Unknown",             NA,        NA,          NA,              NA,       NA,              NA,          NA,        NA,          NA, "South Australia",        NA,      NA, "1 km S Port Germein", "Port Germein",        NA,         NA,            NA,         NA,        NA,         NA, "33 05  S", "138 00  E",                NA,                                    NA,         NA,         NA,                  NA,                NA,   "Walker, K L",        NA, "24/10/1990",    NA,       NA,          NA,              NA,      NA,           NA,            NA,            NA,          NA, "Zygophyllaceae",   "Nitraria", "billardieri",           NA, "Nitraria billardieri", "EntoSex = adult",       NA,           NA,                                                                                             NA,        NA, "Nitraria billardieri", "Nitraria billardieri", "Nitraria billardierei", "Nitrariaceae",            NA,          "Nitraria",
                   41774L,         "HYM 41774",     NA,       NA,      "Vic",  "HYM",           NA,         NA,      NA,       NA,       NA,       NA,      NA,        NA, "Hymenoptera",           NA, "Colletidae",         NA,     NA,     NA,        NA,       NA,          NA,       NA,       NA,        NA,           NA,         "Unknown",             NA,        NA,          NA,              NA,       NA,              NA,          NA,        NA,          NA, "South Australia",        NA,      NA, "1 km S Port Germein", "Port Germein",        NA,         NA,            NA,         NA,        NA,         NA, "33 05  S", "138 00  E",                NA,                                    NA,         NA,         NA,                  NA,                NA,   "Walker, K L",        NA, "24/10/1990",    NA,       NA,          NA,              NA,      NA,           NA,            NA,            NA,          NA, "Zygophyllaceae",   "Nitraria", "billardieri",           NA, "Nitraria billardieri", "EntoSex = adult",       NA,           NA,                                                                                             NA,        NA, "Nitraria billardieri", "Nitraria billardieri", "Nitraria billardierei", "Nitrariaceae",            NA,          "Nitraria"
                   )

    
    
    # Be sure that the testData is not already in tempdir
    testDataPath <- file.info(list.files(tempdir(), full.names = T, 
                                         pattern = "testData.xlsx", recursive = TRUE))
    unlink(rownames(testDataPath))
    
    # Save a temporary version of these data
    openxlsx::write.xlsx(testData, paste0(tempdir(), "/testData.xlsx"),
                         sheetName = "Combined")
    
    suppressWarnings(
    testOut1 <- readr_BeeBDC(dataset = "VicWam",
                                     path = tempdir(),
                                     inFile = "testData.xlsx",
                                     outFile = "testDataOut.csv",
                                     sheet = "Combined",
                                     dataLicense = "https://creativecommons.org/licenses/by-nc-sa/4.0/")
    )
    
    
    # Get a count of TRUE and FALSE column name matches
    resultsT <- sum(colnames(testOut1) %in% (BeeBDC::ColTypeR()[[1]] %>% names()) == TRUE)
    resultsF <- sum(colnames(testOut1) %in% (BeeBDC::ColTypeR()[[1]] %>% names()) == FALSE)

    
    # Test the number of expected TRUE and FALSE columns and then test the output format (data frames and
    # tibbles are a special case of lists)
    testthat::test_that("readr_VicWam results columns TRUE", {
      testthat::expect_equal(resultsT, 40)
    })
    testthat::test_that("readr_VicWam results columns FALSE", {
      testthat::expect_equal(resultsF, 59)
    })
    testthat::test_that("readr_VicWam correct number of rows", {
      testthat::expect_equal(nrow(testOut1), nrow(testData))
    })
    
    testthat::test_that("readr_VicWam expected class", {
      testthat::expect_type(testOut1, "list")
    })

###### 0.2.1 NS func. ####

#' @importFrom dplyr %>%
# This function splits up the input name into its component elements
nameSplitR <- function(NameInput, 
                            Authority_patterns = NULL){
  requireNamespace("mgsub")

  if(is.null(Authority_patterns)){
  # split up the authority from the species name based on regular expressions
      # Help can be found testing regex here: https://www.regextester.com/95629
    Authority_patterns <- paste0("([a-zA-Z.\\-\\_]+,)", "|", # Find words ending in comma
                           "[a-zA-Z0-9.\\-\\_]+ and ","|", # Or find words before " and "
                           "\\s[:upper:][a-z]+$", "|", # Find a space then an upper-case letter (i.e., an author name after the species name)
                                                  # and at the end of the string
                           "\\s[:upper:][a-z]+\\s\\&", "|", # Find uppercase followed by "&" like "Smith &..."
                           "\\s[:upper:][a-z]+[A-Z][a-z]+\\s\\&", "|", # As above, but including multiple capitals, like "LaBerge &..."
                           "\\s[:upper:][a-z]+\\s[0-9]{4}", "|", # Find Author followed by no comma, space, and a year like " Cresson 1863"
                           "\\s[:upper:]\\'[A-Z]", "|", # Deal with names like "O'Brien"
                           "(\\([A-Z][a-z]+)(?=\\s)", "|", #, "|", # Find a capitalized word (with or without a bracket) - \\([A-Z][a-z]+) - that is followed by a space (not a bracket) - (?=\\s)
                           "(\\([A-Z]{1,2}[\\.]?)(?=\\s)", "|", # Find those few authorities that start with multiple capital letters - (SS Saunders, 1850), or with initials - (W. F. Kirby, 1900)
                           "(\\([A-Z][a-z\\u016f\\u00c0-\\u017e]+\\-[A-Z]+)(?=[\\s\\,])","|", # Match as above, but with special characters
                           "(\\([A-Z][a-z\\u016f\\u00c0-\\u017e]+)(?=[\\s\\,])","|", # Match as above, but with special characters
                           " \\(?de | \\(?van der | \\?van ", "|", # Finds the prefixs for last names "de Villers", "van der Zanden" etc.
                           "(\\([A-Z][a-z\\u016f\\u00c0-\\u017e]+\\-[A-Z][a-z\\u016f\\u00c0-\\u017e]+)(?=[\\s\\,])", "|", # Matches European/African characters with a hyphen between
                           "( not_\\(Cockerell)|\\u016fozi\\u016f|auct \\,| sensu auct not| 1914| sensu auct") # Find other haphazard matches
  }
  
  # Split the name based on Auth. Patterns
  SpSplit <-  stringr::str_split(NameInput, Authority_patterns, simplify = TRUE) %>% 
    mgsub::mgsub( c(" \\($", " $"), c("","")) # Get only the species name in ELEMENT ONE
  AuthSplit <- substr( NameInput[1],  # using the raw NameInput
                       nchar(SpSplit[[1]])+2, nchar(NameInput[1])) # find authority by removing name
  ##### subgenus test #####
  # Test for presence of subgenus
  if( grepl( "\\)",SpSplit[1], fixed = FALSE) == TRUE &
      !grepl( "\\)\\s\\(", SpSplit[1], fixed = FALSE)|
      grepl( "\\s[A-Z][a-z]", SpSplit[1], fixed = FALSE) == TRUE &
      !grepl( "\\)\\s\\(", SpSplit[1], fixed = FALSE)){ # Check to see if there is a subgenus using the presence of a a bracket
    # if there IS a subgenus present
    # SPLIT into individual words (by any white space)
    SynCols_split <- stringr::str_split(SpSplit[1],"\\s", simplify = TRUE)  # Express as matrix
    Gen_Name <- SynCols_split[1] # Get genus
    SubGen_Name <- mgsub::mgsub(SynCols_split[2], c("\\)", "\\("), c("",""),)  # Get subgenus and remove brackets from the subgenus column
    SpName <- SynCols_split[3] # Get species
    SubSpName <- ifelse(length(SynCols_split)>3, paste(c(SynCols_split[4:length(SynCols_split)]), 
                                                       sep=" ",collapse=" "), "NA")  # Get subspecies, if present
  }else{ # if there is NOT a subgenus present
    # SPLIT into individual words (by any white space)
    SynCols_split <- stringr::str_split(SpSplit[1],"\\s", simplify = TRUE)  # Express as matrix
    Gen_Name <- SynCols_split[1] # Get genus
    SubGen_Name <- "NA" # NO subgenus
    SpName <- SynCols_split[2] # Get species
    SubSpName <- ifelse(length(SynCols_split)>2, paste(c(SynCols_split[3:length(SynCols_split)]), 
                                                       sep=" ",collapse=" "), "NA")  # Get subspecies, if present
  } # END else
  
  ##### flag test ####
  flagTest <- sub( "(.*?[A-Za-z\\s]?){0,1}([\\(0-9\\)]{4}\\,?)|^[\\w]+", "", AuthSplit) %>%  # Extract the text after the year in authorship
    trimws( which = "left" , whitespace = "[\\, \\] \\) ]")  # Trim some special characters from the FRONT of the string
  
  if(flagTest != ""){ # IF there are flags present
    AuthSplit2 <- substr(AuthSplit,  # using the raw AuthSplit
                         start = 1, stop = (nchar(AuthSplit) - nchar(flagTest))) %>% # extract just the authority by removing the flag
      trimws( which = "right" , whitespace = "[\\, \\] \\) ]")
    Flags <- substr(AuthSplit,  
                    start = (nchar(AuthSplit2)+1), stop = (nchar(AuthSplit))) %>% # extract just the FLAG by removing the author
      trimws( which = "left" , whitespace = "[\\, \\] \\) ]")
  }else{ # END flag test - IF
    AuthSplit2 <- AuthSplit
    Flags <- ""
  } # END flag test - ELSE
  
  # Return values of interest
  AuthSplitOut <- list("flags" = Flags, "validName" = NameInput, 
                       "canonical" = paste( Gen_Name, "(", SubGen_Name , ")", SpName, SubSpName, sep	= " ") %>% 
                         mgsub::mgsub(c("NA","\\( "," \\)"),c("","\\(","\\)") ) %>%
                         mgsub::mgsub(c("  ","\\(\\)"),c(" ","") ), # Formerly in full - SpSplit[1] 
                       "genus" = Gen_Name,
                       "subgenus" = SubGen_Name, "species" = SpName, "infraspecies" = SubSpName,
                       "authorship" = AuthSplit2)
  
} # END the NameAuth_Splitter FUNCTION

# This function was written by James Dorey to chunk the bdc_coordinates_transposed function
# to allow bigger datasets to be analysed without consuming too much RAM.
# This function was written on the 31st of May 2022. For questions, please email jbdorey[at]me.com
#' Wraps jbd_coordinates_transposed to identify  and fix transposed occurrences
#' 
#' Because the [BeeBDC::jbd_coordinates_transposed()] function is very RAM-intensive, this wrapper 
#' allows a user to specify chunk-sizes and only analyse a small portion of the occurrence data at a 
#' time. The prefix jbd_ is used to highlight the difference between this function and the original
#' [bdc::bdc_coordinates_transposed()].
#' This function will preferably use the countryCode column generated by 
#' [bdc::bdc_country_standardized()].
#'
#' @param data A data frame or tibble. Occurrence records as input.
#' @param lat Character. The column with latitude in decimal degrees. Default = "decimalLatitude".
#' @param lon Character. The column with longitude in decimal degrees. Default = "decimalLongitude".
#' @param country Character. The name of the column containing country names. Default = "country".
#' @param idcol Character. The column name with a unique record identifier. Default = "database_id".
#' @param countryCode Character. Identifies the column containing ISO-2 country codes 
#' Default = "countryCode".
#' @param sci_names Character. The column containing scientific names. Default = "scientificName".
#' @param border_buffer Numeric. The buffer, in decimal degrees, around points to help match them
#' to countries. Default = 0.2 (~22 km at equator).
#' @param save_outputs Logical. If TRUE, transposed occurrences will be saved to their own file.
#' @param stepSize Numeric. The number of occurrences to process in each chunk. Default = 1000000.
#' @param chunkStart Numeric. The chunk number to start from. This can be > 1 when you need to restart 
#' the function from a certain chunk; for example if R failed unexpectedly. 
#' @param progressiveSave Logical. If TRUE then the country output list will be saved between
#' each iteration so that `append` can be used if the function is stopped part way through.
#' @param path Character. The path to a file in which to save the 01_coordinates_transposed_
#' output.
#' @param append Logical. If TRUE, the function will look to append an existing file.
#' @param scale Passed to rnaturalearth's ne_countries().
#' Scale of map to return, one of 110, 50, 10 or 'small', 'medium', 'large'. Default = "large".
#' @param mc.cores Numeric. If > 1, the jbd_correct_coordinates function will run in parallel
#' using mclapply using the number of cores specified. If = 1 then it will be run using a serial
#' loop. NOTE: Windows machines must use a value of 1 (see ?parallel::mclapply). Additionally,
#' be aware that each thread can use large chunks of memory.
#'  Default = 1.#'
#' @return Returns the input data frame with a new column, coordinates_transposed, where FALSE = columns
#' that had coordinates transposed.
#' @export
#' 
#' @importFrom dplyr %>%
#'
#' @examples
#' if(requireNamespace("rnaturalearthdata")){
#' library(dplyr)
#'   # Import and prepare the data
#' data(beesFlagged)
#' beesFlagged <- beesFlagged %>% dplyr::select(!c(.val, .sea)) %>%
#'   # Cut down the dataset to un example quicker
#' dplyr::filter(dplyr::row_number() %in% 1:20)
#'   # Run the function
#' beesFlagged_out <- jbd_Ctrans_chunker(
#' # bdc_coordinates_transposed inputs
#' data = beesFlagged,
#' idcol = "database_id",
#' lat = "decimalLatitude",
#' lon = "decimalLongitude",
#' country = "country_suggested",
#' countryCode = "countryCode",
#' # in decimal degrees (~22 km at the equator)
#' border_buffer = 1, 
#' save_outputs = FALSE,
#' sci_names = "scientificName",
#' # chunker inputs
#' # How many rows to process at a time
#' stepSize = 1000000,  
#' # Start row
#' chunkStart = 1,  
#' # Progressively save the output between each iteration?
#' progressiveSave = FALSE,
#' path = tempdir(),
#' # If FALSE it may overwrite existing dataset
#' append = FALSE,
#'   # Users should select scale = "large" as it is more thoroughly tested
#' scale = "medium",
#' mc.cores = 1
#' ) 
#' table(beesFlagged_out$coordinates_transposed, useNA = "always")
#' } # END if require
#' 
jbd_Ctrans_chunker <- function(
    data = NULL,
    lat = "decimalLatitude",
    lon = "decimalLongitude",
    idcol = "databse_id",
    country = "country_suggested",
    countryCode = "countryCode",
    sci_names = "scientificName",
    border_buffer = 0.2, # in decimal degrees (~22 km at the equator)
    save_outputs = TRUE,
    
     # How many rows to process at a time
    stepSize = 1000000,
     # Start row
    chunkStart = 1,
    progressiveSave = TRUE,
    path = tempdir(),
     # If FALSE it may overwrite existing dataset
    append = TRUE,
    scale = "large",
    mc.cores = 1){
  database_id <- NULL
  
  requireNamespace("dplyr")
  
  #### 0.0 Prep ####
  startTime <- Sys.time()
    # Select the output file name to be saved as you run
  fileName <- paste("01_coordinates_transposed_", Sys.Date(), ".csv", sep = "")
    #### 0.1 nChuncks ####
   # Find the number of chunks needed to complete the run
  nChunks = ceiling(nrow(data)/stepSize)
  # Find the max nrow
  nrowMax <- nrow(data)
   # IF a run failed you can start again from the same spot using append = TRUE
  #### 0.2 Append ####
  if(append == TRUE){
    suppressWarnings({
      # Read in the Tranps_tibble csv
    Tranps_tibble = readr::read_csv(paste0(path, "/","Tranps_tibble.csv"), col_types = ColTypeR())
    })
     # set the chunkStart to the number of rows completed plus one
    chunkStart = nrow(Tranps_tibble) + 1
    nChunks = ceiling((nrow(data)-chunkStart)/stepSize)
  } # END append IF statement
  # The chunkEnd is the same as the stepSize initially, but the chunkEnd will change with each iteration
  # It will also differ if append == true based on where the run is at.
  chunkEnd = (chunkStart + stepSize) - 1
  
  #### 0.3 User text ####
  # Write user output
  writeLines(paste(" - Running chunker with:", "\n",
                   "stepSize = ", 
                   format(stepSize, big.mark=",",scientific=FALSE), "\n",
                   "chunkStart = ", 
                   format(chunkStart, big.mark=",",scientific=FALSE), "\n",
                   "chunkEnd = ", 
                   format(chunkEnd, big.mark=",",scientific=FALSE), "\n",
                   "append = ", append, 
                   sep = ""))
  #### 1.0 Run Loop ####
  # Loop - from chunkStart to the end, process rows in batches of chunkEnd
  for(i in 1:nChunks){
    # Select rows from chunkStart to chunkEnd
    loop_check_pf = data[chunkStart:chunkEnd,] %>%
      tidyr::drop_na(tidyselect::all_of(idcol)) %>%
        # Drop unused factors
      base::droplevels()

    # User output
    writeLines(paste(" - Starting chunk ", i, "...", "\n",
                     "From ",  
                     format(chunkStart, big.mark=",",scientific=FALSE), " to ", 
                     format(chunkEnd, big.mark=",",scientific=FALSE),
                     sep = ""))
      ##### 1.1 Function ####
    # Run the bdc_country_from_coordinates function from the BeeBDC package
    loop_check_pf <- jbd_coordinates_transposed(
      data = loop_check_pf,
      lat = lat,
      lon = lon,
      sci_names = sci_names,
      country = country,
      countryCode = countryCode,
      idcol = idcol,
      border_buffer = border_buffer,
      save_outputs = save_outputs,
      path = path,
      fileName = fileName,
      scale = scale,
      mc.cores = mc.cores)
    
    #### 1.2 Save + bind file ####
    # Save a smaller csv file with the database_id and country name to be matched later
    # For the first instance in the loop...
    if(i == 1 && append == FALSE){
      Tranps_tibble = dplyr::tibble(loop_check_pf)
    }else{
      Tranps_tibble = dplyr::bind_rows(Tranps_tibble, 
                                       loop_check_pf)
    }
    # Set chunkStart to be chunkEnd +1 for the next row
    chunkStart = chunkStart + stepSize
    chunkEnd = chunkEnd + stepSize
    # If chunkEnd surpasses nrowMax, then assign nrowMax.
    if(chunkEnd > nrowMax){
      chunkEnd = nrowMax
    }
    # Make room on the RAM by cleaning up the garbage
    # user output
      #### 1.3 User text ####
    # Print use output
    writeLines(paste(" - Finished chunk ", i, " of ", nChunks, ". ",
                     "Total records examined: ", 
                     format(nrow(Tranps_tibble), big.mark=",",scientific=FALSE),
                     sep = "") )
    # Save as a csv after each iteration
    if(progressiveSave == TRUE){
    readr::write_excel_csv(Tranps_tibble, file = paste0(path, "/", "Tranps_tibble.csv"))}
  } # END loop
  
  
  #### 2.0 Clean and return ####
  # Remove NA values
  Tranps_tibble <- Tranps_tibble %>%
    dplyr::filter(!is.na("database_id")) %>%
    # Remove any duplicates that have been introduced
    dplyr::distinct()
  
  
  endTime <- Sys.time()
  
  message(paste(
    " - Completed in ", 
    round(difftime(endTime, startTime), digits = 2 ),
    " ",
    units(round(endTime - startTime, digits = 2)),
    sep = ""))
  return(Tranps_tibble)
} # END function

requireNamespace("dplyr")

# If rnaturalearthdata is present, run tests
if(requireNamespace("rnaturalearthdata")){

bees3sp <- BeeBDC::bees3sp

fakeData <- dplyr::tibble(
  database_id = c("a1","a2","a3","a4", "a5") ,
  decimalLatitude = c(-17.556329, 17.556329, 17.556329, 63.43333, -17.90000),
  decimalLongitude = c(178.007245, 178.007245, 178.007245, -17.90000, 63.43333),
  country = c("Fiji", "Fiji", "United States", "Bolivia", "Bolivia"),
  countryCode = c("FJ", "FJ", "US", "BO", "BO"),
  scientificName = c("sp1","sp1","sp1","sp1", "sp2")
)

testOut <- BeeBDC::jbd_Ctrans_chunker(
  # bdc_coordinates_transposed inputs
  data = bees3sp %>%
    dplyr::filter(complete.cases(decimalLatitude)) %>%
    dplyr::select(c(database_id, decimalLatitude, decimalLongitude, country, countryCode,
                    scientificName)) %>%
    dplyr::filter(dplyr::row_number() %in% 1:20) %>%
    dplyr::bind_rows(fakeData),
  idcol = "database_id",
  lat = "decimalLatitude",
  lon = "decimalLongitude",
  country = "country",
  countryCode = "countryCode",
    # Larger buffer to speed up operation for test
  border_buffer = 0.9, # in decimal degrees (~22 km at the equator)
  save_outputs = FALSE,
  sci_names = "scientificName",
  # chunker inputs
  stepSize = 100,  # How many rows to process at a time
  chunkStart = 1,  # Start row
  path = tempdir(),
  append = FALSE,  # If FALSE it may overwrite existing dataset
  progressiveSave = FALSE,
  scale = "medium",
  mc.cores = 1
) 

# Get a count of TRUE and FALSE column name matches
resultsT <- sum(testOut$coordinates_transposed == TRUE)
resultsF <- sum(testOut$coordinates_transposed == FALSE)

# Test the number of expected TRUE and FALSE columns and then test the output format (data frames and
# tibbles are a special case of lists)
testthat::test_that("jbd_Ctrans_chunker results TRUE", {
  testthat::expect_equal(resultsT, 22)
})
testthat::test_that("jbd_Ctrans_chunker results FALSE", {
  testthat::expect_equal(resultsF, 3)
})


# Test expected number of rows
testthat::test_that("jbd_Ctrans_chunker row count expected", {
  testthat::expect_equal(nrow(testOut), 25)
})
# Test output class
testthat::test_that("jbd_Ctrans_chunker expected class", {
  testthat::expect_type(testOut, "list")
})

} # END if require

# This function was started on the 20th of January 2025 by James Dorey in order to make some 
# country countryColumns more consistent for his paper titled "How many bee species are there? A 
# quantitative global estimate".
  # For questions, email him on jbdorey@me.com



#' Match and harmonise country names
#' 
#' A small function that is useful for harmonising country names so that they mathc between datasets.
#' This relies on the country names matching one of the exceptions pre-coded in and is so far used 
#' internally.
#'
#' @param data A data frame or tibble. 
#' @param countryColumn Character. The column that contains country names to be harmonised.
#' @param continentAnalysis Logical. Set to TRUE in order to match small entities to continents;
#' limited use for the moment.
#' @param shorterNames Logical. If TRUE, some very long country names will be shortened. This can be
#' helpful for writing country names in plots where long names are annoying.
#' 
#'
#' @return Returns the original data frame or tibble but with harmonised country names
#' 
#' @importFrom dplyr %>%
#' 
#' 
#' @export
#'
#' @examples
#' # load in the test dataset
#' system.file("extdata", "testTaxonomy.rda", package="BeeBDC") |> load()
#' harmonisedCountries <- countryHarmoniseR(
#'   data = testTaxonomy,
#'   countryColumn = "country_suggested"
#' )


countryHarmoniseR <- function(
    data = NULL,
    countryColumn = NULL,
    shorterNames = TRUE,
      # set to TRUE in order to match small entities to continents
    continentAnalysis = FALSE){
  . <- NULL
    
  
  #### 1.0 Main changes ####
  # Change country names based on exceptions
  data <- data %>%
    dplyr::rename(countryColumn = tidyselect::any_of(countryColumn)) %>%
    # Shorten some country countryColumns
    dplyr::mutate(countryColumn = dplyr::if_else(countryColumn %in% c("United States of America","United States"),
                                        "United States",countryColumn),
                  countryColumn = dplyr::if_else(countryColumn == "Russia", "Russian Federation", countryColumn),
                  countryColumn = dplyr::if_else(countryColumn %in% c("Lao People's Democratic Republic", "Lao PDR"),
                                        "Lao",countryColumn),
                  countryColumn = dplyr::if_else(countryColumn %in% c("Republic of Cabo Verde","Cape Verde"),
                                        "Cabo Verde",countryColumn),
                  countryColumn = dplyr::if_else(countryColumn %in% c("S\u00e3o Tom\u00e9 and Principe",
                                                                      "Sao Tome and Principe",
                                                                      "S\u00e3o Tom\u00e9 and Pr\u00edncipe"),
                                        "S\u00e3o Tom\u00e9",countryColumn),
                  countryColumn = dplyr::if_else(countryColumn %in% c("Republic of the Congo","Republic of Congo"),
                                        "Republic of Congo",countryColumn),
                  countryColumn = dplyr::if_else(stringr::str_detect(countryColumn, "Northern") & stringr::str_detect(countryColumn, "Cyprus"),
                                        "Northern Cyprus", countryColumn),
                  countryColumn = dplyr::if_else(countryColumn %in% c("N. Cyprus"),
                                        "Northern Cyprus",countryColumn),
                  countryColumn = dplyr::if_else(countryColumn == "The Gambia", "Gambia", countryColumn),
                  countryColumn = dplyr::if_else(stringr::str_detect(countryColumn, "North") & stringr::str_detect(countryColumn, "Macedonia"),
                                        "Macedonia", countryColumn),
                  countryColumn = dplyr::if_else(countryColumn %in% c("Kingdom of eSwatini","Swaziland", "eSwatini"), 
                                        "Eswatini", countryColumn),
                  countryColumn = dplyr::if_else(countryColumn %in% c("Barbuda", "Antigua"),
                                        "Antigua and Barbuda", countryColumn),
                  countryColumn = dplyr::if_else(countryColumn %in% c("Brussels","Flemish Region","Walloon Region"),
                                        "Belgium", countryColumn),
                  countryColumn = dplyr::if_else(countryColumn %in% c("Caribbean Netherlands","Martinique"),
                                                      # RecountryColumnd Cura\u00e7ao to match to continent
                                                      "Cura\u00e7ao", countryColumn),
                  countryColumn = dplyr::if_else(countryColumn == "Cocos Islands","Indian Ocean Territories", countryColumn),
                  countryColumn = dplyr::if_else(countryColumn %in% c("Federation of Bosnia and Herzegovina",
                                                    "Republic Srpska","Vojvodina", "Bosnia"),
                                        "Bosnia and Herzegovina", countryColumn),
                  countryColumn = dplyr::if_else(countryColumn == "Gibraltar","Spain", countryColumn),
                  countryColumn = dplyr::if_else(countryColumn == "Guadeloupe","Barbados", countryColumn),
                  countryColumn = dplyr::if_else(countryColumn == "Mayotte","Madagascar", countryColumn),
                  countryColumn = dplyr::if_else(countryColumn %in% c("R\u00e9union", "Reunion"),"Mauritius", countryColumn),
                  countryColumn = dplyr::if_else(countryColumn %in% c("Svalbard Islands","Bouvet"),
                                        "Norway", countryColumn),
                  countryColumn = dplyr::if_else(countryColumn %in% c("West Bank","Gaza"),"Palestine", countryColumn),
                  countryColumn = dplyr::if_else(countryColumn == "French Southern Territories",
                                        "French Southern and Antarctic Lands", countryColumn),
                  countryColumn = dplyr::if_else(countryColumn == "South Georgia and South Sandwich Islands","South Georgia and the Islands", countryColumn),
                  countryColumn = dplyr::if_else(countryColumn %in% c("Aland Islands"),
                                                 "\u00c5land Islands", countryColumn),
                  countryColumn = dplyr::if_else(countryColumn == "Kosovo",
                                                      "Serbia", countryColumn),
                  countryColumn = dplyr::if_else(countryColumn == "Falkland Islands","Falkland Islands / Malvinas", countryColumn),
                  countryColumn = dplyr::if_else(stringr::str_detect(countryColumn, "Darussalam"),"Brunei Darussalam", countryColumn),
                  countryColumn = dplyr::if_else(countryColumn %in% c("Scotland","Wales","England"),
                                        "United Kingdom", countryColumn),
                  countryColumn = dplyr::if_else(countryColumn == "Somaliland",
                                                      "Somalia", countryColumn),
                  countryColumn = dplyr::if_else(countryColumn == "Tokelau","New Zealand", countryColumn),
                  countryColumn = dplyr::if_else(countryColumn == "Czech Republic", "Czechia", countryColumn),
                  countryColumn = dplyr::if_else(countryColumn == "South Korea", "Republic of Korea", countryColumn),
                  countryColumn = dplyr::if_else(countryColumn == "Wallis and Futuna Islands", "Wallis and Futuna", countryColumn),
                  countryColumn = dplyr::if_else(countryColumn == "Faeroe Islands","Faroe Islands", countryColumn),
                  countryColumn = dplyr::if_else(countryColumn == "Saint Martin","Saint-Martin", countryColumn),
                  )
  
  #### 2.0 Shorter names ####
  if(shorterNames == TRUE){
    data <- data %>%
      dplyr::rename(countryColumn = tidyselect::any_of(countryColumn)) %>%
      # Shorten some country countryColumns
      dplyr::mutate(countryColumn = dplyr::if_else(countryColumn %in% c("United States of America","United States"),
                                                   "USA",countryColumn),
                    countryColumn = dplyr::if_else(countryColumn == "Democratic Republic of the Congo","DRC",countryColumn),
                    countryColumn = dplyr::if_else(countryColumn %in% c("Lao People's Democratic Republic", "Lao PDR"),
                                                   "Lao",countryColumn),
                    countryColumn = dplyr::if_else(countryColumn == "United Arab Emirates","UAE",countryColumn),
                    countryColumn = dplyr::if_else(countryColumn == "Central African Republic","CAR",countryColumn),
                    countryColumn = dplyr::if_else(countryColumn == "Saint Vincent and the Grenadines","SV & Grenadines",countryColumn),
                    countryColumn = dplyr::if_else(countryColumn == "Northern Mariana Islands","N. Mariana",countryColumn),
                    countryColumn = dplyr::if_else(countryColumn %in% c("Republic of the Congo","Republic of Congo"),
                                                   "Rep. Congo",countryColumn)
      )
  }
  
  
  #### 3.0 Continent changes ####
  if(continentAnalysis == TRUE){
    data = data %>% 
      dplyr::mutate(
        countryColumn = dplyr::if_else(countryColumn == "French Guiana","Brazil", countryColumn),
      )
  }
  
    # Re-apply the original column name
  names(data)[names(data) == "countryColumn"] <- countryColumn
  
  # Return the data
  return(data)
  
}

#' Build a per-species summary for each and all flags
#' 
#' Takes a flagged dataset and returns the total number of fails (FALSE) per flag (columns starting
#' with ".") and per species. It will ignore the .scientificName_empty and .invalidName columns as 
#' species are not assigned.
#' Users may define the column to group the summary by. While it is intended to work with 
#' the scientificName column, users may select any grouping column (e.g., country).
#'
#' @param data A data frame or tibble. The flagged dataset.
#' @param column Character. The name of the column to group by and summarise the failed occurrences.
#' Default = "scientificName".
#' @param outPath A character path. The path to the directory in which the figure will be saved.
#' Default = OutPath_Report. If is NULL then no file will be saved to the disk.
#' @param fileName Character. The name of the file to be saved, ending in ".csv". 
#' Default = "flagTable.csv".
#' @param percentImpacted Logical. If TRUE (the default), the program will write the percentage of 
#' species impacted and over the percentThreshold for each flagging column.
#' @param percentThreshold Numeric. A number between 0 and 100 to indicate the percent of 
#' individuals (>; within each species) that is impacted by a flag, and to be included in the 
#' percentImpacted. Default = 0.
#' 
#' @importFrom dplyr %>%
#'
#' @return A tibble with a column for each flag column (starting with ".") showing the number of 
#' failed (FALSE) occurrences per group. Also shows the (i) total number of records, (ii) total
#' number of failed records, and (iii) the percentage of failed records.
#' @export
#'
#' @examples
#' # Load the toy flagged bee data
#' data("beesFlagged")
#'
#'   # Run the function and build the flag table
#' flagTibble <- flagSummaryTable(data = beesFlagged,
#'                               column = "scientificName",
#'                               outPath = paste0(tempdir()),
#'                               fileName = "flagTable.csv")
#'                               
#' 
flagSummaryTable <- function(
    data = NULL,
    column = "scientificName",
    outPath = OutPath_Report,
    fileName = "flagTable.csv",
    percentImpacted = TRUE,
    percentThreshold = 0){
  # locally bind variables to the function
  flagColumns <- dataFlags <- speciesColumn <- loopCol <- summaryColumn <- . <- NULL
  flagCol <- .summary <- totalFailed <- totalFailed <- total <- OutPath_Report <- NULL
  .scientificName_empty <- .invalidName <- NULL
  
  # Load required packages
  requireNamespace("dplyr")
  requireNamespace("tidyselect")
  requireNamespace("stats")
  requireNamespace("readr")
  
  #### 0.0 Prep ####
  ##### 0.1 Errors ####
  ###### a. FATAL errors ####
  if(is.null(data)){
    stop(" - Please provide an argument for data. I'm a program not a magician.")
  }
  if(percentThreshold > 100 | percentThreshold < 0){
    stop(" - percentThreshold must range from 0 to 100.")
  }
  
  
  #### 1.0 Data prep ####
    # Filter for bad names to not include in the table
  if(any(colnames(data) %in% ".scientificName_empty")){
    data <- data %>%
      dplyr::filter(!.scientificName_empty == FALSE)
  }
  if(any(colnames(data) %in% ".invalidName")){
    data <- data %>%
      dplyr::filter(!.invalidName == FALSE)
  }
  
  # Re-do the .summary column to be sure its up to date
data <- data %>%
      # Remove those without scientific name and that flag column
    dplyr::select(!tidyselect::any_of(".scientificName_empty")) %>%
      # Do the same for invalid names
    dplyr::select(!tidyselect::any_of(".invalidName")) %>%
      # Refresh the .summary column
    BeeBDC::summaryFun() 

  # Get a character vector of the flag columns 
  flagColumns <- data %>% dplyr::select(tidyselect::starts_with(".")) %>% colnames()

  # Select only the species name and flag columns
  dataFlags <- data %>%
  dplyr::select(tidyselect::all_of( c(column, flagColumns))) 

  # Create a column of only species names to add the summary of each column to below
speciesColumn <- dataFlags %>%
  dplyr::distinct(dplyr::across(tidyselect::all_of(column)))

  #### 2.0 Flag column loop ####
  # Loop through each column to get species level counts, starting from column two so as not to 
    # count the column
for(i in 1:(ncol(dataFlags)-1) ){
    # For the ith column, get a COUNT of the FALSE (failed) occurrences per species (or column) 
  loopCol <- dataFlags %>%
      # Select the relevant columns and group by them
    dplyr::select(tidyselect::all_of( c(column, flagColumns[[i]]))) %>%
    dplyr::group_by(dplyr::across( tidyselect::all_of(c(column, flagColumns[[i]])))) %>%
      # Count the occurrences of both TRUE and FALSE
    dplyr::count(name = "n") %>%
      # Set the column names temporarily and then REMOVE the TRUE counts
    stats::setNames(paste0( c(column, "flagCol", "n" ))) %>%
    dplyr::filter(!flagCol == TRUE) %>%
     # ungroup
    dplyr::ungroup() %>%
      # select the relevant columns
    dplyr::select(tidyselect::all_of( c(column, "n"))) %>%
      # Rename the columns
    stats::setNames(paste0( c(column, flagColumns[[i]]) )) 
  
    # Add the count to the speciesColumn tibble
  speciesColumn <- speciesColumn %>% 
    dplyr::left_join(loopCol, by = column) %>%
    replace(is.na(.), 0)
}

#### 3.0 Summary columns ####
  # Add in the totals and percentage to the last columns
summaryColumn <- speciesColumn %>%
    # Add a count of total records
  dplyr::left_join(dataFlags %>%
                     dplyr::group_by(dplyr::across(tidyselect::all_of(column))) %>%
                     dplyr::count(name = "total"),
                   by = column) %>%
    # Change the .summary column to a total failed column
  dplyr::rename(totalFailed = .summary) %>%
  dplyr::relocate(totalFailed, .after = total) %>%
    # Add percentage failed
  dplyr::mutate(percentFailed = (totalFailed/total)*100)

if(percentImpacted == TRUE){
  # Turn the percentThreshold into a proportion instead of a percentage
  percentThreshold <- percentThreshold/100
  # Get the percentages of species that are impacted by each flag
percentImpacted <- summaryColumn %>%
  dplyr::summarise(dplyr::across(tidyselect::starts_with("."), 
                                 function(x){
                                   sum(x/total > percentThreshold)/length(x)*100
                                 })) %>%
    # Transpose the tibble
  tidyr::pivot_longer(cols = tidyselect::starts_with("."))
    # Provide use output
writeLines(paste0("The percentages of species impacted by each flag in your analysis are as follows: \n",
       paste0("  ", percentImpacted$name, " = ", round(percentImpacted$value, 2), "%", 
              collapse = "\n"))
       ) # END  writeLines
  } # END percentImpacted == TRUE

  #### 4.0 Output ####
  # If user provided an outPath then save the file
if(!is.null(outPath)){
  summaryColumn %>%
    readr::write_excel_csv(file = paste(outPath, fileName, sep = "/"))
}

  # return the table as a tibble
return(summaryColumn)

} # END function

requireNamespace("dplyr")


  # Read in the test data
beesFlagged <- BeeBDC::beesFlagged

  # Run the function
flagTibble <- flagSummaryTable(data = beesFlagged,
                              column = "scientificName",
                              outPath = tempdir(),
                              fileName = "/filterTable.csv")

   # Get the number of 100% fails and 0% fails
fails100 <- sum(flagTibble$percentFailed == 100)
fails0 <- sum(flagTibble$percentFailed == 0)

# Test class
testthat::test_that("flagSummaryTable expected class", {
  testthat::expect_type(flagTibble, "list")
})
  # Test specific class
testthat::test_that("flagSummaryTable expected class", {
  testthat::expect_true(any(stringr::str_detect(attributes(flagTibble)$class, 
                                                "tbl_df")))
})

# Test 100% fails size
testthat::test_that("flagSummaryTable 100% failed size test", {
  testthat::expect_equal(fails100, 63)
})
testthat::test_that("flagSummaryTable 0% failed size test", {
  testthat::expect_equal(fails0, 15)
})

# Test number of filtering columns
testthat::test_that("flagSummaryTable test the number of flag columns in and out are equal, minus two. Excluding '.summary'", {
  testthat::expect_equal(colnames(beesFlagged %>% 
                                    dplyr::select(tidyselect::starts_with(".")) %>%
                                    dplyr::select(!tidyselect::any_of(".summary"))) %>%
                           length - 2, 
                         colnames(flagTibble %>% 
                                    dplyr::select(tidyselect::starts_with(".")) %>%
                                    dplyr::select(!tidyselect::any_of(".summary"))) %>%
                           length)
})

# This function was written by James B Dorey to chose and flag some issues as listed by GBIF.
  # This function was written on the 30th of May 2022. For questions, please contact James at
  # jbdorey[at]me.com

# Possible flags:
  # allDates: RECORDED_DATE_INVALID, RECORDED_DATE_UNLIKELY
  # allMetadata: AMBIGUOUS_COLLECTION, COLLECTION_MATCH_NONE, COUNTRY_DERIVED_FROM_COORDINATES,
    # INSTITUTION_MATCH_NONE, DIFFERENT_OWNER_INSTITUTION, COUNTRY_INVALID, AMBIGUOUS_INSTITUTION,
    # COLLECTION_MATCH_FUZZY, INSTITUTION_COLLECTION_MISMATCH, INSTITUTION_MATCH_FUZZY
  # allObservations: OCCURRENCE_STATUS_INFERRED_FROM_INDIVIDUAL_COUNT, BASIS_OF_RECORD_INVALID,
    # INDIVIDUAL_COUNT_INVALID
  # allSpatial: GEODETIC_DATUM_ASSUMED_WGS84, COORDINATE_ROUNDED, COORDINATE_PRECISION_INVALID
    # FOOTPRINT_WKT_INVALID, PRESUMED_NEGATED_LONGITUDE, CONTINENT_INVALID
    # PRESUMED_NEGATED_LATITUDE, COORDINATE_INVALID, COUNTRY_COORDINATE_MISMATCH
    # COORDINATE_UNCERTAINTY_METERS_INVALID, ZERO_COORDINATE, GEODETIC_DATUM_INVALID
  # allTaxo: TAXON_MATCH_HIGHERRANK, TYPE_STATUS_INVALID, TAXON_MATCH_FUZZY


#' Flags records with GBIF issues
#' 
#' This function will flag records which are subject to a user-specified vector of GBIF issues.
#'
#' @param data A data frame or tibble. Occurrence records as input.
#' @param issueColumn Character. The column in which to look for GBIF issues. Default = "issue".
#' @param GBIFflags Character vector. The GBIF issues to flag. Users may choose their own vector of issues to flag or
#' use a pre-set vector or vectors, including c("allDates", "allMetadata", "allObservations", 
#' "allSpatial", "allTaxo", or "all").
#' 
#' Default = c("COORDINATE_INVALID", "PRESUMED_NEGATED_LONGITUDE", "PRESUMED_NEGATED_LATITUDE", "COUNTRY_COORDINATE_MISMATCH", "ZERO_COORDINATE")
#'
#' @return Returns the data with a new column, ".GBIFflags", where FALSE = records with any of the provided 
#' GBIFflags.
#' @export
#' 
#' @importFrom dplyr %>%
#'
#' @examples
#' # Import the example data
#' data(beesRaw)
#' # Run the function
#' beesRaw_Out <- GBIFissues(data = beesRaw, 
#'    issueColumn = "issue", 
#'    GBIFflags = c("COORDINATE_INVALID", "ZERO_COORDINATE")) 
#' 
#' 
GBIFissues <- function (data = NULL, 
                            issueColumn = "issue", 
                            GBIFflags = NULL) 
{
  .data <- .GBIFflags <- NULL
  #### 0.0 Warnings ####
  if(is.null(data)){
    stop("\n - Please provide an argument for data. I'm a program, not a magician.")
  }
  if(is.null(GBIFflags)){
    warning("\n - GBIFflags not provided. Please provide an argument. I'm a program, not a magician.")
    writeLines(paste(
      " - Possible options are:\n",
      paste("TAXON_MATCH_HIGHERRANK", "TYPE_STATUS_INVALID", "TAXON_MATCH_FUZZY",
      "GEODETIC_DATUM_ASSUMED_WGS84", "COORDINATE_ROUNDED", "COORDINATE_PRECISION_INVALID",
      "FOOTPRINT_WKT_INVALID", "PRESUMED_NEGATED_LONGITUDE", "CONTINENT_INVALID",
      "PRESUMED_NEGATED_LATITUDE", "COORDINATE_INVALID", "COUNTRY_COORDINATE_MISMATCH",
      "COORDINATE_UNCERTAINTY_METERS_INVALID", "ZERO_COORDINATE", "GEODETIC_DATUM_INVALID",
      "OCCURRENCE_STATUS_INFERRED_FROM_INDIVIDUAL_COUNT", "BASIS_OF_RECORD_INVALID",
      "INDIVIDUAL_COUNT_INVALID", 
      "AMBIGUOUS_COLLECTION", "COLLECTION_MATCH_NONE", 
      "COUNTRY_DERIVED_FROM_COORDINATES",
      "INSTITUTION_MATCH_NONE", "DIFFERENT_OWNER_INSTITUTION", "COUNTRY_INVALID", 
      "AMBIGUOUS_INSTITUTION", "COLLECTION_MATCH_FUZZY", 
      "INSTITUTION_COLLECTION_MISMATCH", "INSTITUTION_MATCH_FUZZY",
      "RECORDED_DATE_INVALID", "RECORDED_DATE_UNLIKELY", collapse = "", sep = ", "), "\n",
    " - Or:\n",
    "allDates, allMetadata, allObservations, allTaxo, allSpatial, or all. ",
    "We recommend thinking about what is required.", sep = ""
    ))
    message(paste(
      "Using default of:\n",
      paste("COORDINATE_INVALID", "PRESUMED_NEGATED_LONGITUDE", "PRESUMED_NEGATED_LATITUDE", 
             "COUNTRY_COORDINATE_MISMATCH", "ZERO_COORDINATE", sep = ", "),
    sep = ""))
    GBIFflags <- c("COORDINATE_INVALID", "PRESUMED_NEGATED_LONGITUDE", "PRESUMED_NEGATED_LATITUDE", 
      "COUNTRY_COORDINATE_MISMATCH", "ZERO_COORDINATE")
  }
  
  #### 1.0 Flag options ####
    # Flag allDates
  if (any(GBIFflags == "allDates")) {
    GBIFflags <- c("RECORDED_DATE_INVALID", "RECORDED_DATE_UNLIKELY")} 
  # Flag allMetadata
  if (any(GBIFflags == "allMetadata")) {
    GBIFflags <- c("AMBIGUOUS_COLLECTION", "COLLECTION_MATCH_NONE", 
                       "COUNTRY_DERIVED_FROM_COORDINATES",
                       "INSTITUTION_MATCH_NONE", "DIFFERENT_OWNER_INSTITUTION", "COUNTRY_INVALID", 
                       "AMBIGUOUS_INSTITUTION", "COLLECTION_MATCH_FUZZY", 
                       "INSTITUTION_COLLECTION_MISMATCH", "INSTITUTION_MATCH_FUZZY")} 
  # Flag allObservations
  if (any(GBIFflags == "allObservations")) {
    GBIFflags <- c("OCCURRENCE_STATUS_INFERRED_FROM_INDIVIDUAL_COUNT", "BASIS_OF_RECORD_INVALID",
                       "INDIVIDUAL_COUNT_INVALID")} 
  # Flag allSpatial
  if (any(GBIFflags == "allSpatial")) {
    GBIFflags <- c("GEODETIC_DATUM_ASSUMED_WGS84", "COORDINATE_ROUNDED", "COORDINATE_PRECISION_INVALID",
                       "FOOTPRINT_WKT_INVALID", "PRESUMED_NEGATED_LONGITUDE", "CONTINENT_INVALID",
                       "PRESUMED_NEGATED_LATITUDE", "COORDINATE_INVALID", "COUNTRY_COORDINATE_MISMATCH",
                       "COORDINATE_UNCERTAINTY_METERS_INVALID", "ZERO_COORDINATE", "GEODETIC_DATUM_INVALID")} 
  # Flag allTaxo
  if (any(GBIFflags == "allTaxo")) {
    GBIFflags <- c("TAXON_MATCH_HIGHERRANK", "TYPE_STATUS_INVALID", "TAXON_MATCH_FUZZY")} 
  # Flag all
  if (any(GBIFflags == "all")) {
    GBIFflags <- c("TAXON_MATCH_HIGHERRANK", "TYPE_STATUS_INVALID", "TAXON_MATCH_FUZZY",
                   "GEODETIC_DATUM_ASSUMED_WGS84", "COORDINATE_ROUNDED", "COORDINATE_PRECISION_INVALID",
                   "FOOTPRINT_WKT_INVALID", "PRESUMED_NEGATED_LONGITUDE", "CONTINENT_INVALID",
                   "PRESUMED_NEGATED_LATITUDE", "COORDINATE_INVALID", "COUNTRY_COORDINATE_MISMATCH",
                   "COORDINATE_UNCERTAINTY_METERS_INVALID", "ZERO_COORDINATE", "GEODETIC_DATUM_INVALID",
                   "OCCURRENCE_STATUS_INFERRED_FROM_INDIVIDUAL_COUNT", "BASIS_OF_RECORD_INVALID",
                   "INDIVIDUAL_COUNT_INVALID", 
                   "AMBIGUOUS_COLLECTION", "COLLECTION_MATCH_NONE", 
                   "COUNTRY_DERIVED_FROM_COORDINATES",
                   "INSTITUTION_MATCH_NONE", "DIFFERENT_OWNER_INSTITUTION", "COUNTRY_INVALID", 
                   "AMBIGUOUS_INSTITUTION", "COLLECTION_MATCH_FUZZY", 
                   "INSTITUTION_COLLECTION_MISMATCH", "INSTITUTION_MATCH_FUZZY",
                   "RECORDED_DATE_INVALID", "RECORDED_DATE_UNLIKELY")} 
  
  
    # Add flag column
  data <- data %>% dplyr::mutate(.GBIFflags = !tolower(.data[[issueColumn]]) %in% 
                                   tolower(GBIFflags))
    # User output
  message(paste(" - jbd_GBIFissues:\nFlagged", 
                format(sum(data$.GBIFflags == FALSE, na.rm = TRUE), big.mark = ","),
                       "\n ",
                "The .GBIFflags column was added to the database.", "\n",
                sep = " "))
    # Return the dataset
  return(data)
}

#### 4. USGS formatter ####
#' Find, import, and format USGS data to Darwin Core
#' 
#' The function finds, imports, formats, and creates metadata for the USGS dataset.
#'
#' @param path A character path to a directory that contains the USGS data, which will be found using
#' [BeeBDC::fileFinder()]. The function will look for "USGS_DRO_flat".
#' @param pubDate Character. The publication date of the dataset to update the metadata and citation.
#'
#' @return Returns a list with the occurrence data, "USGS_data", and the EML data, "EML_attributes".
#' @export
#' 
#' @importFrom dplyr %>%
#'
#' @examples
#' \dontrun{
#' USGS_data <- USGS_formatter(path = DataPath, pubDate = "19-11-2022")
#' }
#' 
USGS_formatter <- function(
    path, 
    pubDate){
  # locally bind variables to the function
  . <-readr <- lubridate <- problems <- how0 <- how1 <- how2 <- how3 <- how4 <- days <- 
    field_note <- note <- SpeciesNotes <- SpeciesNotes<-DateEntered<-DateScanned<-ip<-position<-
    time1<-time2<- occurrenceID <- NULL
  
  #### require and checks ####
  # Load required packages
  requireNamespace("lubridate")
  requireNamespace("dplyr")
  # File name to search for
  USGS_fileName <- "USGS_DRO_flat"
  # Find the USGS data from the HomePath
  USGS_loc <- fileFinder(path, USGS_fileName)
  # Define ColsToKeep
  ColsToKeep <- ColTypeR()[[1]] %>% names()
  
  # Check if the data are present
  if(is.null(USGS_loc) == TRUE){ # If there are no data matching the name...
    stop(" - Oh dear, it looks like there are no USGS data in the HomePath. If you want to include such data, make sure that they exist.")
  }
  # Check if the ddate has been enetered
  if(exists("pubDate") == FALSE){ # If there are no data matching the name...
    stop(paste(" - Oh no! It looks like you have not provided pubDate = dd-mm-yyy. ",
               "Please include this as the date that Sam Droege passed on his data in the above format."))
  }
  
  #### Find and import ####
  if(grepl(pattern =  ".txt.gz", USGS_loc) == TRUE){ # If the zipped file is present
    writeLines("Only one zip file detected. Unzipping file to be read in.")
    # unzip the file
    R.utils::gunzip(
      # File to unzip
      fileName = USGS_loc, 
      # Where to put the extracted file
      destname = stringr::str_remove(USGS_loc, ".gz"),
      overwrite = FALSE,
      remove = FALSE) 
    USGS_loc <- stringr::str_remove(USGS_loc, ".gz")
    # User output
    writeLines(paste(" - Unzipped file to: ", USGS_loc))
  }
  # If already extracted, use the extracted file
  if(stringr::str_detect(pattern = USGS_fileName, 
                         # Select the string that matches only (avoids a warning but works without)
                         string = USGS_loc)){
    writeLines(paste(" - Reading in data file. This should not take too long.","\n",
                     "There may be some errors upon reading in depending on the state of the data.",
                     "\n", "One might consider reporting errors to Sam Droege to improve the dataset."))
    # Read in the data file using "$" as the delimiter 
    USGS_data <- readr::read_delim(USGS_loc[[1]],
                                   delim = "$")
    # Make a copy of the problems, if they exist
    USGS_problems <- readr::problems(USGS_data)
  }
  
  #### Metadata building ####
  Attributes_USGS <- dplyr::tibble(
    dataSource = "USGS_data",
    alternateIdentifier = "Not provided",
    title = "USGS_DRO database",
    pubDate = lubridate::dmy(pubDate),
    dateStamp = pubDate,
    doi = "Not provided",
    downloadLink = "Not provided, contact Sam Droege at sdroege@usgs.gov",
    abstract = dplyr::lst(
      "Note:  If you are getting data from Sam Droege, you are getting one of two file types. Either an Excel spreadsheet with data already extracted for you, or the entire data set which is compressed into a file called USGS_DRO_flat.gz.  The file USGS_DRO_flat.gz is how this file is usually shipped when we send the entire database.  If you get this file it is compressed and needs to be uncompressed with 7-zip, gzip, or something similar ...after unzipping it becomes a very large txt file which can be imported into a database program for manipulation and searching.  Data can also be extracted from BISON and GBIF and we assume the file structure below is reflected in their data also.
      Be sure to check for spelling errors...we do global checks only every year or two.
      Column Headings for USGS Bee Flat File (or an Excel Sheet if we send a subset)
      ID. - Unique 6 digit specimen number with the database identifier (USGS_DRO) in front
      name - Scientific name; ('Destroyed' = label destroyed without being used; 'NONBEE' = An insect that is not a bee, but was given a label; Blank = not yet identified, can indicate that specimen is waiting for identification, is lost, or was missed during data entry) 
      Sex - 'f' = female; 's' = male; 'u' = unknown
      DeterminedBy - Who did the identification of the specimen
      DeterminedWhen - Roughly when the identification was done
      WhoScanned - Who scanned or entered the data
      DateScanned - Date data were scanned
      SpeciesNotes - Any notes about the species such as changes in identification
      DateEntered - Date entered into the new database (done automatically)
      COLLECTION.db - Collection event number.  Up to 5 digit code preceded by database identifier (USGS_DRO) of the date, time, place of the collection event
      gmt - Greenwich Mean Time of I think the time in which the Collection Event Number was created
      Latitude - Latitude in decimal degrees
      Longitude - Longitude in decimal degrees
      accuracy - Degree of association of specimens to the Latitude/Longitude (1 = Specimens from the level of a country/state; 2 = specimens found in area of a county; 3 = specimens found in area of a park or refuge; 4 and 5 = Specimens located very close to coordinants
      elevation - Elevation of location (rarely used)
      country - Country where collection occurred
      state - State where collection occurred
      county - County where collection occurred
      city - Nearest city or the geographic unit (i.e. Park or refuge) where collected 
      site - A site name, number, or designation within a city/park
      position - similar to \"site\" but used less often
      time1 - Date/time collection started or traps put out (format is yyyymmddmmss)
      time2 - Date/time collection ended or traps picked up
      days - Number of trapping days (used inconsistently)
      who - The collector
      email - Almost always Sam Droege's OLD email address
      how0 - Technique used to capture bees
      how1 - Number of traps that remained full of trap fluid when traps picked up
      how2 - Bowl/ trap size
      how3 - Trap color (rarely used since mostly we use several colors)
      how4 - Trap liquid/soap
      habitat - Rarely used
      field_note - 'field_note' and 'note' used interchangeably to take notes about collection event
      note -  'field_note' and 'note' used interchangeably to take notes about collection event"
    ),
    citations = dplyr::lst("Citations not provided"),
    downloadCitation = paste("Sam Droege. (", 
                             lubridate::as_date(pubDate, format = "%m-%d-%y"),
                             "). United States Geological Survey bee data.",
                             sep = ""),
    rights = dplyr::lst("Rights are not provided. Please seek permission for data use from Same Droege.") 
  ) # END metadata
  # combine the input eml and the attributes tibble into a list for output from the function
  EML_attributes <- dplyr::lst("No_eml_from_USGS", Attributes_USGS)
  names(EML_attributes) <- c("source_eml","Source_tibble") 
  
  
  #### Format the data ####
  writeLines(" - Formatting the USGS dataset...")
  
  writeLines(" - Formatting the dateTime...")
  # Convert time1 and time2 to dateTime format
  # Convert time1
  USGS_data$time1 <- USGS_data$time1 %>% 
    lubridate::ymd_hms(., truncated = 5)   
  # Convert time2
  USGS_data$time2 <- USGS_data$time2 %>% 
    lubridate::ymd_hms(., truncated = 5)  
  
  writeLines(" - Creating samplingProtocol and samplingEffort columns...")
  # Create new columns with extra information that doesn't fit the established columns well
  # Merge all of the extra collection info
  USGS_data <- USGS_data %>% dplyr::mutate(
    samplingProtocol = stringr::str_c(
      dplyr::if_else(!is.na(how0),
                     paste0("Technique used: ", how0), ""),
      dplyr::if_else(!is.na(how1),
                     paste0("Bowls full upon collection: ", how1), ""),
      dplyr::if_else(!is.na(how2),
                     paste0("Sampling Bowl/trap size: ", how2), ""),
      dplyr::if_else(!is.na(how3),
                     paste0("Trap colour: ", how3), ""),
      dplyr::if_else(!is.na(how4),
                     paste0("Trap liquid: ", how4), ""),
      sep = "|") %>%
      # Remove extra bars.
      # Remove extra bars.
      stringr::str_replace_all(pattern = "(\\|){2,5}",
                               replacement = "\\|") %>%
      stringr::str_replace_all(pattern = "(\\|$)+|(^\\|)+",
                               replacement = "")) %>%
      # Add in samplingEffort
    dplyr::mutate(
      samplingEffort = stringr::str_c(
        dplyr::if_else(!is.na(days),
                       paste0("sampling days: ", days), ""),
        sep = "")) %>%
      # Add in dataset information
    dplyr::mutate(
      datasetName = "USGS_DRO database",
      datasetID = "USGS_DRO",
      institutionCode = "USGS"
    )

  
  writeLines(" - Creating the fieldNotes and dataSource columns...")
  # Enter these extra data into a new column.
  USGS_data <- USGS_data %>% dplyr::mutate(
    fieldNotes = stringr::str_c(
      dplyr::if_else(!is.na(field_note),
                     paste0("field_note: ", field_note), ""),
      dplyr::if_else(!is.na(note),
                     paste0("note: ", note), ""),
      dplyr::if_else(!is.na(SpeciesNotes),
                     paste0("SpeciesNotes: ", SpeciesNotes), ""),
      dplyr::if_else(!is.na(DateEntered),
                     paste0("DateEntered: ", DateEntered), ""),
      dplyr::if_else(!is.na(DateScanned),
                     paste0("DateScanned: ", DateScanned), ""),
      dplyr::if_else(!is.na(ip),
                     paste0("ipAddress: ", ip), ""),
      dplyr::if_else(!is.na(position),
                     paste0("position: ", position), ""),
      dplyr::if_else(!is.na(time1),
                     paste0("time1: ", time1), ""),
      dplyr::if_else(!is.na(time2),
                     paste0("time2: ", time2), ""),
      sep = "|") %>%
    # Remove extra bars.
    stringr::str_replace_all(pattern = "(\\|){2,9}",
                 replacement = "\\|") %>%
      stringr::str_replace_all(pattern = "(\\|$)+|(^\\|)+",
                 replacement = "")) 
  # Trim white spaces
    # stringr::str_trim(side = "both")
  
  # Set the dataSource
  USGS_data$dataSource <- "USGS_data"

  writeLines(" - Renaming and selecting columns...")
  # These data must be formatted to match the other data sets.
  # that we created at the top of the R-script
  USGS_data <- USGS_data %>%  # The data frame to match with
    dplyr::rename("occurrenceID" = "ID.",
                  "scientificName" = "name",
                  "sex" = "sex",
                  "identifiedBy" = "DeterminedBy", 
                  "dateIdentified" = "DeterminedWhen", 
                  "eventID" = "COLLECTION.db", 
                  "eventTime" = "gmt", 
                  "decimalLatitude" = "latitude", 
                  "decimalLongitude" = "longitude", 
                  "coordinateUncertaintyInMeters" = "accuracy", 
                  "stateProvince" = "state", 
                  "municipality" = "city", 
                  "Location" = "site", 
                  "recordedBy" = "who", 
                  "eventDate" = "time1") %>%
    dplyr::mutate(
      id = occurrenceID,
      recordId = occurrenceID,
    ) %>%
    # select columns that match the following string
    dplyr::select( dplyr::matches(      
      # Use the carrot "^" to signify start of string and dollar sign "$" to signify end of 
      # string. Effectively, this will only return an exact match.
      paste("^",ColsToKeep,"$",sep="") ))  
  #### Save data ####
  # Check for and create outpath if needed
  outPath <- outFile_maker(path = path)
  # Notfiy user that occurrence file is being written
  writeLines( paste(" - Writing occurrence data file...", "\n",
                    "Number of rows (records): ", format(nrow(USGS_data), big.mark=",",scientific=FALSE), "\n",
                    "Written to file called ", paste("USGS_formatted_", Sys.Date(), ".csv", sep = ""),
                    " at location ", outPath,
                    sep = "")) 
  # Write the occurence file
  readr::write_excel_csv(USGS_data, paste(outPath, "/USGS_formatted_", Sys.Date(), ".csv", sep = ""))
  # Notify user that the .eml file is being written
  writeLines( paste(" - Writing attributes file...", "\n",
                    "Written to file called ", paste("USGS_attribute_files", Sys.Date(),".xml", sep="" ),
                    " at location ", outPath,
                    sep = "")) 
  # Format the attributes for exporting lists
  EML_attributes$Source_tibble$abstract <- EML_attributes$Source_tibble$abstract %>% unlist()
  EML_attributes$Source_tibble$citations <- EML_attributes$Source_tibble$citations %>% unlist()
  EML_attributes$Source_tibble$rights <- EML_attributes$Source_tibble$rights %>% unlist()
  # Write the attribute file
  readr::write_excel_csv(EML_attributes$Source_tibble, file = paste(outPath, 
                                                       "/USGS_attribute_files", 
                                                       Sys.Date(),".csv", sep="" ))
  # IF there were problems detected, write these to a .csv file and notify the user
  if(nrow(USGS_problems) > 0){
    writeLines(" - Problems detected with the tibble. Saving to a .csv file...")
    readr::write_excel_csv(USGS_problems, file = paste(outPath, "/USGS_problems", 
                                          Sys.Date(),".csv", sep="" ))
  }
  
  # Return end product and print completion note
  writeLines(paste(" - Fin.", sep = "\n"))
  return( dplyr::lst(USGS_data, EML_attributes) )
} # END USGS_import

  # Copy in some of the USGS dataset
testData <- "ID.$name$sex$DeterminedBy$DeterminedWhen$WhoScanned$DateScanned$SpeciesNotes$DateEntered$COLLECTION.db$gmt$ip$latitude$longitude$accuracy$elevation$country$state$county$city$site$position$time1$time2$days$who$email$how0$how1$how2$how3$how4$habitat$field_note$note
USGS_DRO000001$Lasioglossum imitatum$f$Sam Droege$2005-01-01$Unknown$1900-01-01$$2017-01-27 09:35:35.770000000$USGS_DRO337$20081107112006$134.192.250.22$35.5917$-83.0602$3$1530$USA$North Carolina$Swain$Great Smoky Mountains$Purchase Knob$23$200204021030xx$200204021630xx$$Harold W. Ikerd$sdroege@usgs.gov$pan trap$5$bowl 6.0oz$yellow-uv$soap dawn$Field$Field no blooms$\n
USGS_DRO000002$Lasioglossum imitatum$f$Sam Droege$2005-01-01$Unknown$1900-01-01$$2017-01-27 09:35:35.770000000$USGS_DRO337$20081107112006$134.192.250.22$35.5917$-83.0602$3$1530$USA$North Carolina$Swain$Great Smoky Mountains$Purchase Knob$23$200204021030xx$200204021630xx$$Harold W. Ikerd$sdroege@usgs.gov$pan trap$5$bowl 6.0oz$yellow-uv$soap dawn$Field$Field no blooms$\n
USGS_DRO000003$Augochlorella aurata$f$Sam Droege$2005-01-01$Unknown$1900-01-01$$2017-01-27 09:35:35.770000000$USGS_DRO337$20081107112006$134.192.250.22$35.5917$-83.0602$3$1530$USA$North Carolina$Swain$Great Smoky Mountains$Purchase Knob$23$200204021030xx$200204021630xx$$Harold W. Ikerd$sdroege@usgs.gov$pan trap$5$bowl 6.0oz$yellow-uv$soap dawn$Field$Field no blooms$\n
USGS_DRO000004$Augochlorella aurata$f$Sam Droege$2005-01-01$Unknown$1900-01-01$$2017-01-27 09:35:35.770000000$USGS_DRO337$20081107112006$134.192.250.22$35.5917$-83.0602$3$1530$USA$North Carolina$Swain$Great Smoky Mountains$Purchase Knob$23$200204021030xx$200204021630xx$$Harold W. Ikerd$sdroege@usgs.gov$pan trap$5$bowl 6.0oz$yellow-uv$soap dawn$Field$Field no blooms$\n
USGS_DRO000005$Halictus confusus$f$Sam Droege$2005-01-01$Unknown$1900-01-01$$2017-01-27 09:35:35.770000000$USGS_DRO337$20081107112006$134.192.250.22$35.5917$-83.0602$3$1530$USA$North Carolina$Swain$Great Smoky Mountains$Purchase Knob$23$200204021030xx$200204021630xx$$Harold W. Ikerd$sdroege@usgs.gov$pan trap$5$bowl 6.0oz$yellow-uv$soap dawn$Field$Field no blooms$\n
USGS_DRO000006$Nomada maculata$m$Sam Droege$2005-01-01$Unknown$1900-01-01$$2017-01-27 09:35:35.770000000$USGS_DRO337$20081107112006$134.192.250.22$35.5917$-83.0602$3$1530$USA$North Carolina$Swain$Great Smoky Mountains$Purchase Knob$23$200204021030xx$200204021630xx$$Harold W. Ikerd$sdroege@usgs.gov$pan trap$5$bowl 6.0oz$yellow-uv$soap dawn$Field$Field no blooms$\n
USGS_DRO000007$Lasioglossum pilosum$f$Sam Droege$2005-01-01$Unknown$1900-01-01$$2017-01-27 09:35:35.770000000$USGS_DRO337$20081107112006$134.192.250.22$35.5917$-83.0602$3$1530$USA$North Carolina$Swain$Great Smoky Mountains$Purchase Knob$23$200204021030xx$200204021630xx$$Harold W. Ikerd$sdroege@usgs.gov$pan trap$5$bowl 6.0oz$yellow-uv$soap dawn$Field$Field no blooms$"


writeLines(testData, paste0(tempdir(), "/19-Nov-22_USGS_DRO_flat.txt"))

testOut <- BeeBDC::USGS_formatter(path = tempdir(), 
                                 pubDate = "19-11-2022")



# Test the expected results
testthat::test_that("USGS_formatter test length of occurrence data", {
  testthat::expect_equal(nrow(testOut$USGS_data), 7)
})
testthat::test_that("USGS_formatter test length of EML data", {
  testthat::expect_equal(nrow(testOut$EML_attributes$Source_tibble), 1)
})
testthat::test_that("USGS_formatter test length of list", {
  testthat::expect_equal(length(testOut), 2)
})


# Test classes
testthat::test_that("USGS_formatter expected class", {
  testthat::expect_type(testOut, "list")
})
testthat::test_that("USGS_formatter expected class", {
  testthat::expect_equal(attributes(testOut$USGS_data)$class, c("tbl_df","tbl" ,"data.frame"))
})
testthat::test_that("USGS_formatter expected attribute names", {
  testthat::expect_equal(attributes(testOut$EML_attributes)$names, c("source_eml","Source_tibble"))
})

# Description of the BeeBDC test checklist dataset
# 17th June 2024

#' An example of the beesChecklist file
#'
#'
#' A small test checklist file for package tests. This dataset was built by filtering the 
#' checklist data from the three test datasets, beesFlagged, beesRaw, bees3sp.
#'
#' @docType data
#'
#' @usage data("testChecklist", package = "BeeBDC")
#'
#' @format An object of class \code{"tibble"}
#' \describe{
#'  \item{validName}{The valid scientificName as it should occur in the scientificName column.}
#'  \item{DiscoverLife_name}{The full country name as it occurs on Discover Life.}
#'  \item{rNaturalEarth_name}{Country name from rnaturalearth's name_long and type = "map_units".}
#'  \item{shortName}{A short version of the country name.}
#'  \item{continent}{The continent where that country is found.}
#'  \item{DiscoverLife_ISO}{The ISO country name as it occurs on Discover Life.}
#'  \item{Alpha-2}{Alpha-2 from rnaturalearth.}
#'  \item{iso_a3_eh}{iso_a3_eh from rnaturalearth.}
#'  \item{official}{Official country name = "yes" or only a Discover Life name = "no".}
#'  \item{Source}{A text strign denoting the source or author of the name-country pair.}
#'  \item{matchCertainty}{Quality of the name's match to the Discover Life checklist.}
#'  \item{canonical}{The valid species name without scientificNameAuthority.}
#'  \item{canonical_withFlags}{The validName without the scientificNameAuthority but with Discover Life flags.}
#'  \item{family}{Bee family.}
#'  \item{subfamily}{Bee subfamily.}
#'  \item{genus}{Bee genus.}
#'  \item{subgenus}{Bee subgenus.}
#'  \item{infraspecies}{Bee infraSpecificEpithet.}
#'  \item{species}{Bee specificEpithet.}
#'  \item{scientificNameAuthorship}{Bee scientificNameAuthorship.}
#'  \item{taxon_rank}{Rank of the taxon name.}
#'  \item{Notes}{Discover Life country name notes.}
#'  }
#' @references This dataset is a subset of the beesChecklist file described in:
#' Dorey, J.B., Fischer, E.E., Chesshire, P.R., Nava-Bolaños, A., O’Reilly, R.L., Bossert, S., Collins, S.M., Lichtenberg, E.M., Tucker, E., Smith-Pardo, A., Falcon-Brindis, A., Guevara, D.A., Ribeiro, B.R., de Pedro, D., Hung, J.K.-L., Parys, K.A., McCabe, L.M., Rogan, M.S., Minckley, R.L., Velzco, S.J.E., Griswold, T., Zarrillo, T.A., Jetz, W., Sica, Y.V., Orr, M.C., Guzman, L.M., Ascher, J., Hughes, A.C. & Cobb, N.S. (2023) A globally synthesised and flagged bee occurrence dataset and cleaning workflow. Scientific Data, 10, 1–17. https://www.doi.org/10.1038/S41597-023-02626-W
#' @keywords datasets
#' @examples
#' beesRaw <- BeeBDC::testChecklist
#' head(testChecklist)
#' 
"testChecklist"

# This script was written by James Dorey to use Paige Chesshire's taxonomy updates to update the 
  # Orr and Ascher combined taxonomy
  # This function checks if there is a match in the current taxonomy or not
# For queries, please contact James Dorey at jbdorey[at]me.com
# This function was started on 15th May 2022 and last updated 17th May 2022
#' @importFrom dplyr %>%

PaigesSyns <- function(PaigeSheet = Paige_sheet_loc,
                       SynFile = SynL_AO){
  # locally bind variables to the function
  . <- Paige_sheet_loc <- SynL_AO <- canonical <- id <- NULL
  
  # Read in the relevant sheets from the excel file
  SynChanges <- openxlsx::read.xlsx(PaigeSheet, sheet = "Synonym_changes")
  
  
#### 2.0 Synonyms ####
# Set up an empty tibble...
SynAnnotate_df <- dplyr::tibble()
# Run a loop
for(i in 1:nrow(SynChanges)){
  # Find the ith name to change
  loopName_ori <- SynChanges$OriginalScientificName_DL_s1[i]
  loopName_DL <- SynChanges$DLifeName_s2[i]
  loopName_Final <- SynChanges$Final_Name[i]
  ##### 2.1 One name ####
  # If there is only one name...
  if(loopName_ori == loopName_DL || loopName_DL %in% c("Remove","remove (illogical)")){
    # Find that name in the original list
    SLAO_row <- SynFile %>% 
      dplyr::filter(canonical == loopName_ori)
    # Now find the accepted row based on the accid
    acc_SLAO_row <- SynFile %>% 
      dplyr::filter(id %in% SLAO_row$accid)
    # Check to see if any of the DiscoverLife accepted names match Paige's name
    taxaCurrentTest <- any(acc_SLAO_row$canonical %in% loopName_Final)
    # . note this gives a TRUE or FALSE to see if the names match the current taxonomy.
    if(taxaCurrentTest == TRUE){
      SynAnnotate_df <- dplyr::tibble(
        FinalName <- loopName_Final, originalName = loopName_ori, DisLifeName = loopName_DL,
        PotentialMatches = nrow(SLAO_row), iteration = i,
        Correct = TRUE, note = "One name provided, VALID") %>%
        dplyr::bind_rows(SynAnnotate_df, .)
    }else{
      SynAnnotate_df <- dplyr::tibble(
        FinalName <- loopName_Final, originalName = loopName_ori, DisLifeName = loopName_DL,
        PotentialMatches = nrow(SLAO_row), iteration = i,
        Correct = FALSE, note = "One name provided, INVALID")%>%
        dplyr::bind_rows(SynAnnotate_df, .)
    }
  }# End 1. valid name
  ##### 2.2 Two names ####
  # If there is only one name...
  if(loopName_ori != loopName_DL && !loopName_DL %in% c("Remove","remove (illogical)")){
    # Find that name in the original list
    SLAO_row <- SynFile %>% 
      dplyr::filter(canonical %in% c(loopName_ori, loopName_DL))
    # Now find the accepted row based on the accid
    acc_SLAO_row <- SynFile %>% 
      dplyr::filter(id %in% SLAO_row$accid)
    # Check to see if any of the DiscoverLife accepted names match Paige's name
    taxaCurrentTest <- any(acc_SLAO_row$canonical %in% loopName_Final)
    # . note this gives a TRUE or FALSE to see if the names match the current taxonomy.
    if(taxaCurrentTest == TRUE){
      SynAnnotate_df <- dplyr::tibble(
        FinalName <- loopName_Final, originalName = loopName_ori, DisLifeName = loopName_DL,
        PotentialMatches = nrow(SLAO_row), iteration = i,
        Correct = TRUE, note = "Two names provided, VALID") %>%
        dplyr::bind_rows(SynAnnotate_df, .)
    }else{
      SynAnnotate_df <- dplyr::tibble(
        FinalName <- loopName_Final, originalName = loopName_ori, DisLifeName = loopName_DL,
        PotentialMatches = nrow(SLAO_row), iteration = i,
        Correct = FALSE, note = "Two names provided, INVALID")%>%
        dplyr::bind_rows(SynAnnotate_df, .)
    } # END else
  } # END 2.2 Two names
}# End 2.0 Synonym loop

  # Use output
writeLines( paste("The output can be used to manually check names. ", 
                  "There are ", sum(SynAnnotate_df$Correct == FALSE), " names that require checking",
                  " and ", sum(SynAnnotate_df$Correct == TRUE), " that appear fine.",
                  sep = ""))

return(SynAnnotate_df)
} # END function

####  2. repoMerge ####
#' Import occurrences from GBIF, ALA, iDigBio, and SCAN downloads
#' 
#' Locates data from GBIF, ALA, iDigBio, and SCAN within a directory and reads it in along with its eml
#' metadata. Please keep the original download folder names and architecture unchanged.
#' NOTE: This function uses family-level data to identify taxon downloads. If this, or something new,
#' becomes an issue, please contact James Dorey (the developer) as there are likely to be exceptions
#' to how files are downloaded. current as of versions 1.0.4.
#'
#' @param path A directory as a character. The directory to recursively look in for the above data.
#' @param save_type Character. The data type to save the resulting file as. Options are: 
#' csv_files" or "R_file".
#' @param occ_paths A list of directories. Preferably produced using [BeeBDC::repoFinder()] the 
#' function asks for a list of paths to the relevant input datasets. You can fault-find errors 
#' in this function by checking the output of [BeeBDC::repoFinder()].
#'
#' @return A list with a data frame of merged occurrence records, "Data_WebDL", and a list of eml 
#' files contained in "eml_files". Also saves these files in the requested format.
#' @export
#' 
#' @importFrom dplyr %>%
#' @importFrom readr col_character cols
#' @importFrom stats setNames
#'
#' @examples
#' \dontrun{
#' if(require("emld")){
#' 
#' DataImp <- repoMerge(path = DataPath, 
#' # Find data - Many problems can be solved by running [BeeBDC::repoFinder(path = DataPath)]
#' # And looking for problems
#' occ_paths = BeeBDC::repoFinder(path = DataPath),
#' save_type = "R_file")
#' } # END if require
#' } # END donttest
repoMerge <- function(path, save_type, occ_paths){
  . <- family <- data_i <- NULL
  
    #### 0.0 Prep ####
  requireNamespace("dplyr")
    # Remove empty elements
  occ_paths <- occ_paths[lapply(occ_paths,length)>0]
  
  #### 1.0 Data read+merge ####
  ###### 1.1 Loop prep. ####
  {startTime <- Sys.time()
  # print user information
  writeLines( paste(" - Reading and joining ",length(unlist(occ_paths))," occurrence files.", "\n",
                    "Depending on file size and number, this could take some time.","\n",
                    sep = ""))
  # Make an internal copy of the template for use in the loop as the template tibble
  data_template <- matrix(ncol = length(BeeBDC::ColTypeR()[[1]] %>% names()), nrow = 0) %>% as.data.frame() %>% 
    stats::setNames(BeeBDC::ColTypeR()[[1]] %>% names()) %>% dplyr::tibble() %>%
    dplyr::mutate(dplyr::across(dplyr::everything(), as.character))
  # Copy the template
  Data_WebDL <- data_template
  # Make an empty eml file for the loop
  eml_files <- dplyr::tibble()
  # Set up a counter to keep track of the number of files processed
  counter = 1
  
  ##### 1.2 Data loop ####
  for(j in 1:length(occ_paths)){ # Start path loop
    for(i in 1:length(unlist(occ_paths[j]))){
      # Firstly, select the download path for the loop
      path_i <- occ_paths[j] %>% unlist() %>% .[i]
      
      ###### a. loop data ####
      # Use the custom function dataReader to read in the occurrence data in the correct format
      data_i <- dataReader(path_i = path_i, home_path = path) #%>%
        #readr::type_convert(col_types = BeeBDC::ColTypeR())
        # If there is a date range of days, take only the first day
      data_i$day <- data_i$day %>% 
        stringr::str_replace(pattern = "-.*", replacement = "") %>% 
        stringr::str_squish() #%>%
        #as.integer()
      
      
      ###### b. attr. mangement ####
      # Use the custom function attr_builder to extract attribute information from file
      data_Attributes <- data_i %>%
          # Group by family
        dplyr::group_by(family) %>%
          # Split the groups into their own lists
        dplyr::group_split(.keep = TRUE) %>%
        lapply(., attr_builder,
               path_i = path_i) 
      
        # Extract all of the Source_tibble tibbles and bind them together
      data_Source_tibbles <- data_Attributes %>%
        lapply(., function(x) dplyr::tibble(x$Source_tibble)) %>%
        # Combine the lists of tibbles
        dplyr::bind_rows()
        # Recombine the source eml dat aand the source tibble data
      data_Attributes <- list(data_Attributes[[1]][[1]], data_Source_tibbles) %>%
        stats::setNames(c("source_eml","Source_tibble"))
      
      
      
        # Add an additional test for iDigBio to extract families
      if(any(data_Attributes$Source_tibble$dataSource %>% stringr::str_detect("iDigBio"))){
        data_Attributes$Source_tibble <- data_Attributes$Source_tibble %>% 
          dplyr::mutate(
            dataSource = stringr::str_c("iDigBio_", unique(data_i$family) %>% 
                                          stringr::str_replace_na("NA")),
            taxon = stringr::str_c(unique(data_i$family) %>% stringr::str_replace_na("NA"))
          )} # END if iDigBio

      
      # Test if the attributes are sourced from a single family, or from more. Should == TRUE if it 
      # is from a single family and == FALSE if from a multiple family download.
      singleFamilyTest <- nrow(data_Attributes$Source_tibble) == 1
      
        ###### b1. single family ####
      if(singleFamilyTest == TRUE){
        # Annotate the dataSource column with information from the attributes (based on one input 
          # family)
        data_i$dataSource <- data_Attributes$Source_tibble$dataSource
      } # END singleFamilyTest == TRUE
      ###### b2. many families ####
      if(singleFamilyTest == FALSE){
        # If the family column is NA, allow linking of the dataSource by temporarily modifying that
          # column to include the source_NA
        data_i <- data_i %>%
          dplyr::mutate(family = dplyr::if_else(is.na(family),
                                                "NA",
                                                family))
        # Annotate the dataSource column with information from the attributes (based on many input 
          # families)
        data_i <- data_i %>%
            # Remove the empty dataSource column
          dplyr::select(!tidyselect::any_of("dataSource")) %>%
            # Match the family-level dataSources to the dataSource column using left_join
          dplyr::left_join(data_Attributes$Source_tibble %>%
                             dplyr::select(c("dataSource", "taxon")),
                           by = c("family" = "taxon"))
      } # END singleFamilyTest == FALSE
      
      
      ## Source table ##  
      # Bind with either the template or running GBIF_Data_WebDL tibble
      Data_WebDL <- dplyr::bind_rows(Data_WebDL, data_i) # END rbind.fill
      # Extract existing attributes
      ExistingAttrs <- attributes(Data_WebDL)
      # Combine with new attributes
      # combine tibbles
      SrcTbl_new <- dplyr::bind_rows(ExistingAttrs$dataSource, data_Attributes$Source_tibble)
      # Add the new attribute information into the file
      attr(Data_WebDL, which = "dataSource") <- SrcTbl_new
      ## eml files ##
      # Make the eml file into a list
      new_eml_file <- list(data_Attributes$source_eml)
      # Name that list with the appropriate source and name
      names(new_eml_file) <- if(singleFamilyTest == TRUE){
                  # If singleFamilyTest == TRUE, then use that single family dataSource name as the
                  # eml name
        data_Attributes$Source_tibble$dataSource
        }else{
                  # If singleFamilyTest == FALSE, then concatenate those names into a text string
                  # separated by "; " for each family present
        stringr::str_c(data_Attributes$Source_tibble$dataSource, collapse = "; ")}
        
      # Append this to the running eml list file
      eml_files <- append(eml_files, new_eml_file) %>%
        emld::as_emld( from = "list")
      
      
      ###### c. Progress print ####
      # Print the progress and number of records (rows) to user
      writeLines( paste(" - Completed", i,"of", length(unlist(occ_paths[j])), names(occ_paths[j]),
                        "files. ","\n",
                        counter, "of", length(unlist(occ_paths)), "total files processed","\n",
                        "Cumulative number of rows =", 
                        format(nrow(Data_WebDL), big.mark=",",scientific=FALSE), sep = " ") )
      # progress the counter by 1
      counter = counter + 1
    } # End GBIF loop
  } # End path loop
  
  ##### 3. Save data ####
  # user-input includes the path and save_type, which are supplied from repoMerge()
  dataSaver(path = path, # Path to HomeDirectory
             save_type = save_type, # save_type is either "csv_files" or "R_file"
             occurrences = Data_WebDL, # Input data file
             eml_files = eml_files, # Input eml_files
             file_prefix = "BeeData_") 

  # Completion message to user with endTime
  endTime <- Sys.time()
  writeLines( paste(" - Completed in ", round(endTime - startTime, digits = 2), " ",
                    units(round(endTime - startTime, digits = 2)), sep = ""))
  } # END Data Loop
  return( dplyr::lst(Data_WebDL, eml_files))
} # END repoMerge function

# This function was written by James Dorey to harmonise the names of bees using the Ascher-Orr-Chesshire
# bee taxonomies.
# The function first merges names based on scientificName, then merging the bdc cleaned_name and
# scientificNameAuthorship and matching those, followed by matching to canoncial with flags, then canonical. 
# In all of these cases, names that are ambiguous at that level are removed so that only confident
# matches are maintaned. 
# This function was written between the 18th and 20th of May 2022. For questions, please email James
# at jbdorey[at]me.com


#' Harmonise taxonomy of occurrence data
#' 
#' Uses BeeBDC-formatted taxonomy data to harmonise occurrence records and flag those that do not match
#' the taxonomy [BeeBDC::harmoniseR()] prefers to use the names_clean columns that is generated 
#' by [bdc::bdc_clean_names()]. While this is not required, you may find better results by running
#' that function on your dataset first.
#' It is possible to download taxonomy file for other taxa using [BeeBDC::taxadbToBeeBDC()]
#' which can download taxonomies from ITIS, GBIF, and more. You could also match the format of the
#' [BeeBDC::beesTaxonomy()] file.
#'
#' @param path A directory as character. The path to a folder that the output can be saved.
#' @param taxonomy A data frame or tibble. The taxonomy file to use. 
#' Default = [BeeBDC::beesTaxonomy()]; for other taxa see first [BeeBDC::taxadbToBeeBDC()].
#' @param data A data frame or tibble. Occurrence records as input.
#' @param speciesColumn Character. The name of the column containing species names. Default = "scientificName".
#' @param rm_names_clean Logical. If TRUE then the names_clean column will be removed at the end of
#' this function to help reduce confusion about this column later. Default = TRUE
#' @param checkVerbatim Logical. If TRUE then the verbatimScientificName will be checked as well 
#' for species matches. This matching will ONLY be done after harmoniseR has failed for the other 
#' name columns. NOTE: this column is *not* first run through `bdc::bdc_clean_names`. Default = FALSE
#' @param relaxAmbiguous Logical. If TRUE, then ambiguous names will be returned as "TRUE" for 
#' .invalidName. However, they will be listed as "ambiguouslyMatchedName" 
#' under the "scientificNameAuthorship" column and the speciesColumn will NOT include the authority
#' information. Hence, the names can be used but in should remain clear in the data that they must
#' be used with caution. Default = FALSE.
#' @param matchHigherTaxonomy Logical. If TRUE, then BeeBDC will try to match higher taxonomies
#' (e.g., family names) using genera and such. This will work better if you have run the data through
#' `bdc::bdc_clean_names` beforehand. The function will update the taxonRank column, if it is 
#' otherwise empty. Default = TRUE.
#' @param stepSize Numeric. The number of occurrences to process in each chunk. Default = 1000000.
#' @param mc.cores Numeric. If > 1, the function will run in parallel
#' using mclapply using the number of cores specified. If = 1 then it will be run using a serial
#' loop. NOTE: Windows machines must use a value of 1 (see ?parallel::mclapply). Additionally,
#' be aware that each thread can use large chunks of memory.
#'  Default = 1.
#'
#' @return The occurrences are returned with update taxonomy columns, including: scientificName, 
#' species, family, subfamily, genus, subgenus, specificEpithet, infraspecificEpithet, and 
#' scientificNameAuthorship. A new column, .invalidName, is also added and is FALSE when the occurrence's
#' name did not match the supplied taxonomy. 
#' 
#' @importFrom dplyr %>%
#' 
#' @seealso [BeeBDC::taxadbToBeeBDC()] to download any taxonomy (of any taxa or of bees) and
#' [BeeBDC::beesTaxonomy()] for the bee taxonomy download.
#' 
#' @export
#'
#' @examples
#' # load in the test dataset
#' system.file("extdata", "testTaxonomy.rda", package="BeeBDC") |> load()
#' 
#' # See also
#' ?BeeBDC::taxadbToBeeBDC()
#' 
#' beesRaw_out <- BeeBDC::harmoniseR(
#'   #The path to a folder that the output can be saved
#' path = tempdir(),
#' # The formatted taxonomy file
#' taxonomy = testTaxonomy, 
#' data = BeeBDC::beesFlagged,
#' speciesColumn = "scientificName")
#' table(beesRaw_out$.invalidName, useNA = "always")

harmoniseR <- function(
    data = NULL,
    path = NULL, #The path to a folder that the output can be saved
    taxonomy = BeeBDC::beesTaxonomy(), # The formatted taxonomy file
    speciesColumn = "scientificName",
    rm_names_clean = TRUE,
    checkVerbatim = FALSE,
    relaxAmbiguous = FALSE,
    matchHigherTaxonomy = TRUE,
    stepSize = 1000000,
    mc.cores = 1
) {  
  # locally bind variables to the function
  . <- id <- validName<-canonical<-canonical_withFlags<-family<-subfamily<-genus<-subgenus<-
    species<-infraspecies<-authorship<-taxonomic_status<-flags<-accid<-validName_valid<-
    family_valid<-subfamily_valid<-canonical_withFlags_valid<-genus_valid<-subgenus_valid<-
    species_valid<-infraspecies_valid<-authorship_valid<-database_id<-names_clean<-
    scientificNameAuthorship<-taxonRank<-authorFound<-SciNameAuthorSimple<-.invalidName<- .data <-
    kingdom <-
    authorSimple<-united_SciName<-verbatimScientificName <- scientificName <- BeeBDC_order <- NULL
  
  
  
  # Load required packages 
  requireNamespace("rlang")
  requireNamespace("dplyr")
  
  # Record start time
  startTime <- Sys.time()
  
  #### 0.0 Prep ####
  ##### 0.1 Errors ####
  ###### a. FATAL errors ####
  if(is.null(data)){
    stop(" - Please provide an argument for data. I'm a program not a magician.")
  }
  if(is.null(taxonomy)){
    stop(" - Please provide an argument for taxonomy I'm a program not a magician.")
  }
  if(is.null(path)){
    stop(" - Please provide an argument for path I'm a program not a magician.")
  }
  if(!"verbatimScientificName" %in% colnames(data) & checkVerbatim == TRUE){  
    stop(paste0(" - If 'checkVerbatim = TRUE', then the verbatimScientificName column must be 
  present in the data."))
  }
  
  
  
  #### 1.0  _match columns ####
  # Make a synonym index list
  writeLines(paste(" - Formatting taxonomy for matching..."))
  # save the original column names
  OG_colnames <- unique(c("database_id", colnames(data)))
  # Save the original number of rows
  OG_rowNum <- nrow(data)
  
  ##### 1.1 Prepare columns ####
  # To make the function more general, allow some column changing internally.
  ###### a. rename to scientificName ####
  # Temporarily rename the speciesColumn to "scientificName" within the function
  data <- data %>%
    dplyr::rename("scientificName" = tidyselect::any_of(speciesColumn))
  ###### b. temp names_clean ####
  # IF the names_clean column does not exist, temporarily add it to the dataset using the 
  # scientificName column's data.
  if(!"names_clean" %in% colnames(data)){
    data <- data %>%
      dplyr::mutate(names_clean = scientificName)
    message(paste0("The names_clean column was not found and will be temporarily copied from",
                   " scientificName"))
  }
  ###### c. database_id ####
  # If the database_id column isn't in the dataset, then add it for internal use
  if(!"database_id" %in% colnames(data)){
    data <- data %>%
      dplyr::mutate(database_id = paste0("BeeBDC_TempCode_", dplyr::row_number()), .before = 1)
    message("The database_idcolumn was not found, making this column with 'BeeBDC_TempCode_'...")
  }
  ###### d. scientificNameAuthorship ####
  # If there is no scientificNameAuthorship, make all NA
  if(!"scientificNameAuthorship" %in% colnames(data)){
    data <- data %>%
      dplyr::mutate(scientificNameAuthorship = NA_character_)
    message("The scientificNameAuthorship column was not found, making this column full of NAs.")
  }
  ###### e. taxonRank ####
  # If there is no taxonRank, make all NA
  if(!"taxonRank" %in% colnames(data)){
    data <- data %>%
      dplyr::mutate(taxonRank = NA_character_)
    message("The taxonRank column was not found, making this column full of NAs.")
  }
  ###### f. species ####
  # If there is no species, make all NA
  if(!"species" %in% colnames(data)){
    data <- data %>%
      dplyr::mutate(species = scientificName)
    message("The species column was not found, filling this column with scientificName.")
  }
  # Record the colnames of the data 
  originalColnames <- colnames(data)
  
  # Remove non-ambiguous tags 
  taxonomy <- taxonomy %>%
    dplyr::mutate(flags = flags %>%
                    stringr::str_remove_all("non-ambiguous canonical|	non-ambiguous can_wFlags"))
  
  # Add a new column which has the canonical names matched to the synonyms
  taxonomy <- taxonomy %>%
    dplyr::left_join(x = ., 
                     # left join ONLY the validName, canonical, and canonical_withFlags
                     y = dplyr::select(taxonomy, 
                                       tidyselect::any_of(
                                         c("id", "validName", "canonical", "canonical_withFlags", 
                                           "family", "subfamily", "genus", "subgenus", "species", 
                                           "infraspecies", "authorship"))), 
                     by = c("accid" = "id"), suffix = c("", "_valid"),
                     multiple = "all")
  # Now, also duplicate the accepted names into the ._matched columns  
  AccMatched <- taxonomy %>% 
    # select only the ACCEPTED NAMES
    dplyr::filter(taxonomic_status == "accepted") %>%
    # duplicate the valid columns into the matched column locations
    dplyr::mutate(validName_valid = validName,
                  canonical_valid = canonical,
                  canonical_withFlags_valid = canonical_withFlags,
                  family_valid = family,
                  subfamily_valid = subfamily,
                  genus_valid = genus, 
                  subgenus_valid = subgenus, 
                  species_valid = species, 
                  infraspecies_valid = infraspecies, 
                  authorship_valid = authorship)
  
  # Merge these datasets
  taxonomy <- taxonomy %>%
    # First filter for the reverse of above - SYNONYM NAMES
    dplyr::filter(taxonomic_status == "synonym") %>%
    # combine
    dplyr::bind_rows(AccMatched)
  
  
  rm(AccMatched)
  
  
  #### 2.0 Harmonise data ####
  writeLines(paste("\n",
                   " - Harmonise the occurrence data with unambiguous names...", sep = ""))
  # Create the parallel-able function
  unAmbiguousFunction <- function(data){
    ##### 2.1 Valid Name ####
    ###### a. prep synonyms ####
    # Filter out the AMBIGUOUS validNames prior to matching
    currenttaxonomy <- taxonomy %>%
      # REMOVE ambiguous validNames
      dplyr::filter(!stringr::str_detect(
        # Replace NA in flags with "" to allow matching
        tidyr::replace_na(flags, ""),
        "ambiguous validName")) 
    
    ###### b. assign names ####
    # Clean up some illegal characters
    data$scientificName <- data$scientificName %>%
      stringr::str_replace(pattern = "^\"", replacement = "") %>%
      stringr::str_replace(pattern = "\"$", replacement = "")
    
    # Match names first with the validName column
    occs_21 <- data %>%
      dplyr::left_join(currenttaxonomy %>%
                         dplyr::select(c(id, accid, validName, canonical_withFlags, canonical, validName_valid,
                                         family_valid, subfamily_valid,
                                         canonical_withFlags_valid, genus_valid, subgenus_valid, 
                                         species_valid, infraspecies_valid, authorship_valid)),
                       # Match scientific name with the valid synonym name
                       by = c("scientificName" = "validName"),
                       suffix = c("", "_harmon"),
                       multiple = "all", relationship = "many-to-many") 
    #   # Add a column to express the name-match quality - "high" IF there is a match at this point
    # dplyr::mutate(nameQuality = dplyr::if_else(stats::complete.cases(validName_valid),
    #   "high", "NA")) 3,703
    
    ###### c. return Occs ####
    # Return the matched data
    occs_21 <- occs_21 %>%
      dplyr::filter(stats::complete.cases(validName_valid)) # 1,927
    
    
    ##### 2.2 validName_comb ####
    # Now we will try and match the valid name by combining the names_clean and scientificNameAuthorship columns
    ###### a. prep synonyms ####
    # For those that did not match, attempt to match them with the Canonical with flags column...
    # Filter out the AMBIGUOUS validNames prior to matching
    ## SAME as 2.1 ##
    currenttaxonomy <- taxonomy %>%
      # REMOVE ambiguous validNames
      dplyr::filter(!stringr::str_detect(
        # Replace NA in flags with "" to allow matching
        tidyr::replace_na(flags, ""),
        "ambiguous validName")) 
    
    ###### b. assign names ####
    # Match names first with the validName column
    occs_22 <- data %>%
      # remove already-matched names
      dplyr::filter(!database_id %in% occs_21$database_id) %>%
      # Make a new column by combining names_clean and scientificNameAuthorship
      tidyr::unite(col = "united_SciName", names_clean, scientificNameAuthorship, sep = " ",
                   na.rm = TRUE) 
    
    # Match names first with the validName column
    occs_22 <- occs_22 %>%
      dplyr::left_join(currenttaxonomy %>%
                         dplyr::select(c(id, accid, validName, canonical_withFlags, canonical, validName_valid,
                                         family_valid, subfamily_valid,
                                         canonical_withFlags_valid, genus_valid, subgenus_valid, 
                                         species_valid, infraspecies_valid, authorship_valid)),
                       # Match scientific name with the valid synonym name
                       by = c("united_SciName" = "validName"),
                       suffix = c("", "_harmon"),
                       multiple = "all", relationship = "many-to-many") 
    
    ###### c. return Occs ####
    # Return the matched data
    runningOccurrences <- occs_22 %>%
      dplyr::filter(stats::complete.cases(validName_valid) & validName_valid != "NA") %>%
      # Bind the previous rows
      dplyr::bind_rows(occs_21) # 2,678
    # Remove this spent files
    rm(occs_21, occs_22)
    
    
    ##### 2.3 canonical_wFlags ####
    ###### a. prep synonyms ####
    # For those that did not match, attempt to match them with the Canonical with flags column...
    # Filter out the AMBIGUOUS validNames prior to matching
    currenttaxonomy <- taxonomy %>%
      # REMOVE ambiguous validNames and can_wFlags
      dplyr::filter(!stringr::str_detect(
        # Replace NA in flags with "" to allow matching
        tidyr::replace_na(flags, ""),
        paste("ambiguous validName", 
              "ambiguous can_wFlags",
              sep = "|"))) %>%
      # remove the rows where the canonical and canonical_withFlags match
      # ONLY matches those with added canonicals flags
      dplyr::filter(!canonical == canonical_withFlags)
    
    
    ###### b. assign names ####
    # Match names first with the validName column
    occs_23 <- data %>%
      # remove already-matched names
      dplyr::filter(!database_id %in% runningOccurrences$database_id) %>%
      dplyr::left_join(currenttaxonomy %>% 
                         dplyr::select(c(id, accid, validName, canonical_withFlags, canonical, validName_valid,
                                         family_valid, subfamily_valid,
                                         canonical_withFlags_valid, genus_valid, subgenus_valid, 
                                         species_valid, infraspecies_valid, authorship_valid)),
                       # Match scientific name with the valid synonym name
                       by = c("species" = "canonical_withFlags"),
                       suffix = c("", "_harmon"),
                       multiple = "all", relationship = "many-to-many") 
    
    ###### c. return Occs ####
    # Return the matched data
    runningOccurrences <- occs_23 %>%
      dplyr::filter(stats::complete.cases(validName_valid)) %>%
      # Bind the previous rows
      dplyr::bind_rows(runningOccurrences)
    # Remove this spent file 
    rm(occs_23)
    
    
    ##### 2.4 canonical ####
    ###### a. prep synonyms ####
    # For those that did not match, attempt to match them with the Canonical with flags column...
    # Filter out the AMBIGUOUS validNames prior to matching
    currenttaxonomy <- taxonomy %>%
      # REMOVE ambiguous names
      dplyr::filter(!stringr::str_detect(
        # Replace NA in flags with "" to allow matching
        tidyr::replace_na(flags, ""), 
        paste("ambiguous validName", 
              "ambiguous can_wFlags",
              "ambiguous canonical",
              sep = "|"))) 
    
    ###### b. assign names ####
    # Match names first with the validName column
    occs_24 <- data %>%
      # Keep the unmatched names
      dplyr::filter(!database_id %in% runningOccurrences$database_id) %>%
      dplyr::left_join(currenttaxonomy %>% 
                         dplyr::select(c(id, accid, validName, canonical_withFlags, canonical, 
                                         validName_valid,
                                         family_valid, subfamily_valid,
                                         canonical_withFlags_valid, genus_valid, subgenus_valid, 
                                         species_valid, infraspecies_valid, authorship_valid)),
                       # Match scientific name with the valid synonym name
                       by = c("names_clean" = "canonical"),
                       suffix = c("", "_harmon"),
                       multiple = "all", relationship = "many-to-many") 
    
    ###### c. return Occs ####
    # Return the matched data
    runningOccurrences <- occs_24 %>%
      dplyr::filter(stats::complete.cases(validName_valid)) %>%
      # Bind the previous rows
      dplyr::bind_rows(runningOccurrences) %>%
      # Make sure no duplicates have snuck in
      dplyr::distinct(database_id, .keep_all = TRUE)
    # Remove spent file
    rm(occs_24)
    
    
    ##### 2.5 sciName_comb ####
    # Now we will try and match the valid name by combining the scientificName and scientificNameAuthorship columns
    ###### a. prep synonyms ####
    # For those that did not match, attempt to match them with the Canonical with flags column...
    # Filter out the AMBIGUOUS validNames prior to matching
    ## SAME as 2.1 ##
    currenttaxonomy <- taxonomy %>%
      # REMOVE ambiguous validNames
      dplyr::filter(!stringr::str_detect(
        # Replace NA in flags with "" to allow matching
        tidyr::replace_na(flags, ""),
        "ambiguous validName")) 
    
    ###### b. assign names ####
    # Match names first with the validName column
    occs_25 <- data %>%
      # remove already-matched names
      dplyr::filter(!database_id %in% runningOccurrences$database_id) %>%
      # Make a new column by combining names_clean and scientificNameAuthorship
      tidyr::unite(col = "united_SciName", names_clean, scientificNameAuthorship, sep = " ",
                   na.rm = TRUE) 
    
    # Match names first with the validName column
    occs_25 <- occs_25 %>%
      dplyr::left_join(currenttaxonomy %>% 
                         dplyr::select(c(id, accid, validName, canonical_withFlags, canonical, validName_valid,
                                         family_valid, subfamily_valid,
                                         canonical_withFlags_valid, genus_valid, subgenus_valid, 
                                         species_valid, infraspecies_valid, authorship_valid)),
                       # Match scientific name with the valid synonym name
                       by = c("united_SciName" = "validName"),
                       suffix = c("", "_harmon"),
                       multiple = "all", relationship = "many-to-many") 
    
    ###### c. return Occs ####
    # Return the matched data
    runningOccurrences <- occs_25 %>%
      dplyr::filter(stats::complete.cases(validName_valid)) %>%
      # Bind the previous rows
      dplyr::bind_rows(runningOccurrences) %>%
      # Make sure no duplicates have snuck in
      dplyr::distinct(database_id, .keep_all = TRUE)
    # Remove spent file
    rm(occs_25)
    
    
    
    ##### 2.6 No subgenus validName ####
    # Match scientificName with validName; remove subgenus from both
    ###### a. prep synonyms ####
    # For those that did not match, attempt to match them with the Canonical with flags column...
    # Filter out the AMBIGUOUS validNames prior to matching
    # For those that did not match, attempt to match them with the Canonical with flags column...
    # Filter out the AMBIGUOUS validNames prior to matching
    currenttaxonomy <- taxonomy %>%
      # REMOVE ambiguous names
      dplyr::filter(!stringr::str_detect(
        # Replace NA in flags with "" to allow matching
        tidyr::replace_na(flags, ""), 
        paste("ambiguous validName", 
              "ambiguous canonical",
              sep = "|"))) 
    
    ###### b. assign names ####
    # Match names first with the validName column
    occs_26 <- data %>%
      # remove already-matched names
      dplyr::filter(!database_id %in% runningOccurrences$database_id) %>%
      dplyr::mutate(scientificNameMatch = scientificName %>% 
                      # Replace subgenus with nothing
                      stringr::str_replace("\\([A-Za-z]+\\)", "") %>%
                      stringr::str_squish())
    
    # Match names first with the validName column
    occs_26 <- occs_26 %>%
      dplyr::left_join(currenttaxonomy %>% 
                         dplyr::select(c(id, accid, validName, canonical_withFlags, canonical, validName_valid,
                                         family_valid, subfamily_valid,
                                         canonical_withFlags_valid, genus_valid, subgenus_valid, 
                                         species_valid, infraspecies_valid, authorship_valid)) %>%
                         dplyr::mutate(validNameMatch = validName %>% 
                                         # Replace subgenus with nothing
                                         stringr::str_replace("\\([A-Za-z]+\\)", "") %>%
                                         stringr::str_squish()),
                       # Match scientific name with the valid synonym name
                       by = c("scientificNameMatch" = "validNameMatch"),
                       suffix = c("", "_harmon"),
                       multiple = "all", relationship = "many-to-many") 
    
    ###### c. return Occs ####
    # Return the matched data
    runningOccurrences <- occs_26 %>%
      dplyr::filter(stats::complete.cases(validName_valid)) %>%
      # Bind the previous rows
      dplyr::bind_rows(runningOccurrences) %>%
      # Make sure no duplicates have snuck in
      dplyr::distinct(database_id, .keep_all = TRUE)
    # Remove spent file
    rm(occs_26, currenttaxonomy)
    
    ##### 2.7 No subgenus canonical ####
    # Match scientificName with canonical; remove subgenus from both
    ###### a. prep synonyms ####
    # For those that did not match, attempt to match them with the Canonical with flags column...
    # Filter out the AMBIGUOUS validNames prior to matching
    # For those that did not match, attempt to match them with the Canonical with flags column...
    # Filter out the AMBIGUOUS validNames prior to matching
    currenttaxonomy <- taxonomy %>%
      # REMOVE ambiguous names
      dplyr::filter(!stringr::str_detect(
        # Replace NA in flags with "" to allow matching
        tidyr::replace_na(flags, ""), 
        paste("ambiguous validName", 
              "ambiguous canonical",
              sep = "|"))) 
    
    ###### b. assign names ####
    # Match names first with the canonical column
    occs_27 <- data %>%
      # remove already-matched names
      dplyr::filter(!database_id %in% runningOccurrences$database_id) %>%
      dplyr::mutate(scientificNameMatch = scientificName %>% 
                      # Replace subgenus with nothing
                      stringr::str_replace("\\([A-Za-z]+\\)", "") %>%
                      stringr::str_squish())
    
    # Match names first with the canonical column
    occs_27 <- occs_27 %>%
      dplyr::left_join(currenttaxonomy %>% 
                         dplyr::select(c(id, accid, validName, canonical_withFlags, canonical, validName_valid,
                                         family_valid, subfamily_valid,
                                         canonical_withFlags_valid, genus_valid, subgenus_valid, 
                                         species_valid, infraspecies_valid, authorship_valid)) %>%
                         dplyr::mutate(canonicalMatch = canonical %>% 
                                         # Replace subgenus with nothing
                                         stringr::str_replace("\\([A-Za-z]+\\)", "") %>%
                                         stringr::str_squish()),
                       # Match scientific name with the canonical synonym name
                       by = c("scientificNameMatch" = "canonicalMatch"),
                       suffix = c("", "_harmon"),
                       multiple = "all", relationship = "many-to-many") 
    
    ###### c. return Occs ####
    # Return the matched data
    runningOccurrences <- occs_27 %>%
      dplyr::filter(stats::complete.cases(validName_valid)) %>%
      # Bind the previous rows
      dplyr::bind_rows(runningOccurrences) %>%
      # Make sure no duplicates have snuck in
      dplyr::distinct(database_id, .keep_all = TRUE)
    # Remove spent file
    rm(occs_27, currenttaxonomy)
    return(runningOccurrences)
  } # END unAmbiguousFunction
  
  # Run the function
  runningOccurrences <- data %>%
    # Make a new column with the ordering of rows
    dplyr::mutate(BeeBDC_order = dplyr::row_number()) %>%
    # Group by the row number and step size
    dplyr::group_by(BeeBDC_group = ceiling(BeeBDC_order/stepSize)) %>%
    # Split the dataset up into a list by group
    dplyr::group_split(.keep = TRUE) %>%
    # Run the actual function
    parallel::mclapply(., unAmbiguousFunction,
                       mc.cores = mc.cores
    ) %>%
    # Combine the lists of tibbles
    dplyr::bind_rows() 
  
  #### 3.0 Ambiguous names ####
  writeLines(paste("\n",
                   " - Attempting to harmonise the occurrence data with ambiguous names...", sep = ""))
  ambiguousFunction <- function(ambData){
    ##### 3.1 Prepare datasets ####
    ###### a. prep synonyms ####
    # Synonym list of ambiguous names
    # Filter TO the AMBIGUOUS validNames prior to matching
    ambiguousSynonyms <- taxonomy %>%
      # Keep only ambiguous validNames
      dplyr::filter(stringr::str_detect(
        # Replace NA in flags with "" to allow matching
        tidyr::replace_na(flags, ""), 
        "ambiguous")) %>%
      # Remove non-ambiguous matches
      dplyr::filter(!stringr::str_detect(
        # Replace NA in flags with "" to allow matching
        tidyr::replace_na(flags, ""),
        "non-")) %>%
      # Remove ambiguous validName matches because these are ambiguous even with authorities.
      # Perhaps in future these can be matched by geography.
      dplyr::filter(!stringr::str_detect(
        # Replace NA in flags with "" to allow matching
        tidyr::replace_na(flags, ""),
        "ambiguous validName")) %>%
      # Remove those without authorship
      dplyr::filter(!is.na(authorship)) %>% 
      # # Change the authorship to be easier to match by removing capitals and punctuation
      dplyr::mutate(authorSimple = stringr::str_remove_all(authorship, 
                                                           pattern = "[:punct:]") %>% tolower())
    
    
    ###### b. author in sciName  ####
    if(!all(is.na(ambData$scientificNameAuthorship))){
      # Create a list of scientificNameAuthorships that can be found in scientificName, where the former is empty
      ambiguousAuthorFound <- ambData %>%
        # check only the data without authorship
        dplyr::filter(is.na(scientificNameAuthorship)) %>%
        # Select only UNDER genus-level IDs
        dplyr::filter(taxonRank %in% c("Especie", "forma", "Infrasubspecies", "Race",
                                       "species", "Species", "SPECIES", "subsp.", "subspecies",
                                       "Subspecies", "SUBSPECIES", "syn", "var.", "variety",
                                       "Variety", "VARIETY", NA, "NA")) %>%
        # Make a new column with the authorship extracted from scientificName
        dplyr::mutate(authorFound = stringr::str_extract(
          # Use a simplified scientificName string
          string = stringr::str_remove_all(scientificName, 
                                           pattern = "[:punct:]") %>% tolower(),
          pattern = paste(ambiguousSynonyms$authorSimple, collapse = "|"))) %>%
        # Keep only matched names
        dplyr::filter(stats::complete.cases(authorFound)) %>%
        # Return only the database_id and authorFound for merging...
        dplyr::select(tidyselect::all_of(c("database_id", "authorFound")))
      
      # Add the author to those data that were lacking
      ambData <- ambData %>%
        # Add authorFound to original dataset
        dplyr::left_join(ambiguousAuthorFound, by = "database_id",multiple = "all") %>%
        # If scientificNameAuthorship is empty, use authorFound from ambiguousAuthorFound
        dplyr::mutate(scientificNameAuthorship = 
                        dplyr::if_else(is.na(scientificNameAuthorship),
                                       # If missing replace the na with the authorFound
                                       authorFound,
                                       # IF already complete, keep the original
                                       scientificNameAuthorship))
      
      # Remove used data
      rm(ambiguousAuthorFound)}
    
    ###### c. ambiguous data ####
    # Filter occurrence dataset to those with ambiguous names AND authorship values
    data_amb <- ambData %>%
      # Keep those with authorship recorded
      dplyr::filter(stats::complete.cases(scientificNameAuthorship)) %>%
      # Keep those that are in the ambiguous names list
      dplyr::filter(scientificName %in% ambiguousSynonyms$validName |
                      scientificName %in% ambiguousSynonyms$canonical_withFlags |
                      scientificName %in% ambiguousSynonyms$canonical) %>%
      # Simplify scientificNameAuthorship to make easier matches
      dplyr::mutate(SciNameAuthorSimple = stringr::str_remove_all(scientificNameAuthorship, 
                                                                  pattern = "[:punct:]") %>% tolower())
    
    
    ##### 3.2 Valid Name ####
    ###### a. assign names ####
    # Match names first with the validName column
    runningAmb_occs <- data_amb %>%
      # Select only rows with both scientificName and SciNameAuthorSimple
      dplyr::filter(stats::complete.cases(scientificName) & stats::complete.cases(SciNameAuthorSimple)) %>%
      # Add taxonomy information to the occurrence data.
      dplyr::left_join(ambiguousSynonyms %>%
                         dplyr::select(c(id, accid, validName, canonical_withFlags, canonical, validName_valid,
                                         family_valid, subfamily_valid,
                                         canonical_withFlags_valid, genus_valid, subgenus_valid, 
                                         species_valid, infraspecies_valid, authorship_valid, authorSimple)),
                       # Match scientific name with the valid synonym name
                       by = c("scientificName" = "validName",
                              "SciNameAuthorSimple" = "authorSimple"),
                       suffix = c("", "_harmon"),
                       multiple = "all",relationship = "many-to-many") 
    
    ###### b. return Occs ####
    # Return the matched data_amb
    runningAmb_occs <- runningAmb_occs %>%
      dplyr::filter(stats::complete.cases(validName_valid)) # 1,927
    
    
    ##### 3.3 validName_comb ####
    # Now we will try and match the valid name by combining the names_clean and scientificNameAuthorship columns
    
    
    ###### a. assign names ####
    # Match names first with the validName column
    occs_33 <- data_amb %>%
      # remove already-matched names
      dplyr::filter(!database_id %in% runningAmb_occs$database_id) %>%
      # Make a new column by combining names_clean and scientificNameAuthorship
      tidyr::unite(col = "united_SciName", names_clean, scientificNameAuthorship, sep = " ",
                   na.rm = TRUE) 
    
    # Match names first with the validName column
    occs_33 <- occs_33 %>%
      # Select only rows with both united_SciName and SciNameAuthorSimple
      dplyr::filter(stats::complete.cases(united_SciName) & stats::complete.cases(SciNameAuthorSimple)) %>%
      # Add taxonomy information to the occurrence data.
      dplyr::left_join(ambiguousSynonyms %>%
                         dplyr::select(c(id, accid, validName, canonical_withFlags, canonical, validName_valid,
                                         family_valid, subfamily_valid,
                                         canonical_withFlags_valid, genus_valid, subgenus_valid, 
                                         species_valid, infraspecies_valid, authorship_valid, authorSimple)),
                       # Match scientific name with the valid synonym name
                       by = c("united_SciName" = "validName",
                              "SciNameAuthorSimple" = "authorSimple"),
                       suffix = c("", "_harmon"),
                       multiple = "all", relationship = "many-to-many") 
    
    ###### b. return Occs ####
    # Return the matched data_amb
    runningAmb_occs <- occs_33 %>%
      dplyr::filter(stats::complete.cases(validName_valid)) %>%
      # Bind the previous rows
      dplyr::bind_rows(runningAmb_occs) # 2,678
    # Remove this spent files
    rm(occs_33)
    
    
    ##### 3.4 canonical_wFlags ####
    ###### a. prep datasets ####
    # Synonym list of ambiguous names
    # Filter TO the AMBIGUOUS validNames prior to matching
    syns_34 <- ambiguousSynonyms %>%
      # remove the rows where the canonical and canonical_withFlags match
      dplyr::filter(!canonical == canonical_withFlags)
    
    
    ###### b. assign names ####
    # Match names first with the validName column
    occs_34 <- data_amb %>%
      # remove already-matched names
      dplyr::filter(!database_id %in% runningAmb_occs$database_id) %>%
      # Select only rows with both species and SciNameAuthorSimple
      dplyr::filter(stats::complete.cases(species) & stats::complete.cases(SciNameAuthorSimple)) %>%
      # Add taxonomy information to the occurrence data.
      dplyr::left_join(syns_34 %>%
                         dplyr::select(c(id, accid, validName, canonical_withFlags, canonical, validName_valid,
                                         family_valid, subfamily_valid,
                                         canonical_withFlags_valid, genus_valid, subgenus_valid, 
                                         species_valid, infraspecies_valid, authorship_valid, authorSimple)),
                       # Match scientific name with the valid synonym name
                       by = c("species" = "canonical_withFlags",
                              "SciNameAuthorSimple" = "authorSimple"),
                       suffix = c("", "_harmon"),
                       multiple = "all", relationship = "many-to-many") 
    
    ###### c. return Occs ####
    # Return the matched data_amb
    runningAmb_occs <- occs_34 %>%
      dplyr::filter(stats::complete.cases(validName_valid)) %>%
      # Bind the previous rows
      dplyr::bind_rows(runningAmb_occs)
    # Remove this spent file 
    rm(occs_34, syns_34)
    
    
    ##### 3.5 canonical ####
    ###### b. assign names ####
    # Match names first with the validName column
    occs_35 <- data_amb %>%
      # Keep the unmatched names
      dplyr::filter(!database_id %in% runningAmb_occs$database_id) %>%
      # Select only rows with both names_clean and SciNameAuthorSimple
      dplyr::filter(stats::complete.cases(names_clean) & stats::complete.cases(SciNameAuthorSimple)) %>%
      # Add taxonomy information to the occurrence data.
      dplyr::left_join(ambiguousSynonyms %>%
                         dplyr::select(c(id, accid, validName, canonical_withFlags, canonical, validName_valid,
                                         family_valid, subfamily_valid,
                                         canonical_withFlags_valid, genus_valid, subgenus_valid, 
                                         species_valid, infraspecies_valid, authorship_valid, authorSimple)),
                       # Match scientific name with the valid synonym name
                       by = c("names_clean" = "canonical",
                              "SciNameAuthorSimple" = "authorSimple"),
                       suffix = c("", "_harmon"),
                       multiple = "all", relationship = "many-to-many") 
    
    ###### c. return Occs ####
    # Return the matched data_amb
    runningAmb_occs <- occs_35 %>%
      dplyr::filter(stats::complete.cases(validName_valid)) %>%
      # Bind the previous rows
      dplyr::bind_rows(runningAmb_occs)%>%
      # Make sure no duplicates have snuck in
      dplyr::distinct(database_id, .keep_all = TRUE)
    # Remove spent file
    rm(occs_35)
    
    
    
    ##### 3.6 validName_comb ####
    # Now we will try and match the valid name by combining the names_clean and scientificNameAuthorship columns
    ###### a. assign names ####
    # Match names first with the validName column
    occs_36 <- data_amb %>%
      # remove already-matched names
      dplyr::filter(!database_id %in% runningAmb_occs$database_id) %>%
      # Make a new column by combining names_clean and scientificNameAuthorship
      tidyr::unite(col = "united_SciName", names_clean, scientificNameAuthorship, sep = " ",
                   na.rm = TRUE) 
    
    # Match names first with the validName column
    occs_36 <- occs_36 %>%
      # Select only rows with both united_SciName and SciNameAuthorSimple
      dplyr::filter(stats::complete.cases(united_SciName) & stats::complete.cases(SciNameAuthorSimple)) %>%
      # Add taxonomy information to the occurrence data.
      dplyr::left_join(ambiguousSynonyms %>%
                         dplyr::select(c(id, accid, validName, canonical_withFlags, canonical, validName_valid,
                                         family_valid, subfamily_valid,
                                         canonical_withFlags_valid, genus_valid, subgenus_valid, 
                                         species_valid, infraspecies_valid, authorship_valid, authorSimple)),
                       # Match scientific name with the valid synonym name
                       by = c("united_SciName" = "validName",
                              "SciNameAuthorSimple" = "authorSimple"),
                       suffix = c("", "_harmon"),
                       multiple = "all", relationship = "many-to-many") 
    
    ###### b. return Occs ####
    # Return the matched data_amb
    runningAmb_occs <- occs_36 %>%
      dplyr::filter(stats::complete.cases(validName_valid)) %>%
      # Bind the previous rows
      dplyr::bind_rows(runningAmb_occs) # 2,678
    # Remove this spent files
    rm(occs_36)
    
    
    ##### 3.7 No subgenus validName ####
    # Match scientificName with validName; remove subgenus from both
    ###### a. assign names ####
    # Match names first with the validName column
    occs_37 <- data_amb %>%
      # remove already-matched names
      dplyr::filter(!database_id %in% runningAmb_occs$database_id) %>%
      dplyr::mutate(scientificNameMatch = scientificName %>% 
                      # Replace subgenus with nothing
                      stringr::str_replace("\\([A-Za-z]+\\)", "") %>%
                      stringr::str_squish())
    
    occs_37 <- occs_37 %>%
      dplyr::left_join(ambiguousSynonyms %>% 
                         dplyr::select(c(id, accid, validName, canonical_withFlags, canonical, 
                                         validName_valid, family_valid, subfamily_valid,
                                         canonical_withFlags_valid, genus_valid, subgenus_valid, 
                                         species_valid, infraspecies_valid, authorship_valid)) %>%
                         dplyr::mutate(validNameMatch = validName %>% 
                                         # Replace subgenus with nothing
                                         stringr::str_replace("\\([A-Za-z]+\\)", "") %>%
                                         stringr::str_squish()),
                       # Match scientific name with the valid synonym name
                       by = c("scientificNameMatch" = "validNameMatch"),
                       suffix = c("", "_harmon"),
                       multiple = "all", relationship = "many-to-many") 
    
    ###### b. return Occs ####
    # Return the matched data_amb
    runningAmb_occs <- occs_37 %>%
      dplyr::filter(stats::complete.cases(validName_valid)) %>%
      # Bind the previous rows
      dplyr::bind_rows(runningAmb_occs)
    # Remove this spent files
    rm(occs_37)
    
    
    ##### 3.8 No subgenus canonical ####
    # Match scientificName with canonical; remove subgenus from both
    ###### a. assign names ####
    # Match names first with the canonical column
    occs_38 <- data_amb %>%
      # remove already-matched names
      dplyr::filter(!database_id %in% runningAmb_occs$database_id) %>%
      dplyr::mutate(scientificNameMatch = scientificName %>% 
                      # Replace subgenus with nothing
                      stringr::str_replace("\\([A-Za-z]+\\)", "") %>%
                      stringr::str_squish())
    
    occs_38 <- occs_38 %>%
      dplyr::left_join(ambiguousSynonyms %>% 
                         dplyr::select(c(id, accid, validName, canonical_withFlags, canonical, 
                                         validName_valid, family_valid, subfamily_valid,
                                         canonical_withFlags_valid, genus_valid, subgenus_valid, 
                                         species_valid, infraspecies_valid, authorship_valid)) %>%
                         dplyr::mutate(canonicalMatch = canonical %>% 
                                         # Replace subgenus with nothing
                                         stringr::str_replace("\\([A-Za-z]+\\)", "") %>%
                                         stringr::str_squish()),
                       # Match scientific name with the valid synonym name
                       by = c("scientificNameMatch" = "canonicalMatch"),
                       suffix = c("", "_harmon"),
                       multiple = "all", relationship = "many-to-many") 
    
    ###### b. return Occs ####
    # Return the matched data_amb
    runningAmb_occs <- occs_38 %>%
      dplyr::filter(stats::complete.cases(validName_valid)) %>%
      # Bind the previous rows
      dplyr::bind_rows(runningAmb_occs)
    # Remove this spent files
    rm(occs_38)
    return(runningAmb_occs)
  } # END ambiguousFunction
  
  # Run the function
  runningAmb_occs <- data %>%
    # Make a new column with the ordering of rows
    dplyr::mutate(BeeBDC_order = dplyr::row_number()) %>%
    # Group by the row number and step size
    dplyr::group_by(BeeBDC_group = ceiling(BeeBDC_order/stepSize)) %>%
    # Split the dataset up into a list by group
    dplyr::group_split(.keep = TRUE) %>%
    # Run the actual function
    parallel::mclapply(., ambiguousFunction,
                       mc.cores = mc.cores
    ) %>%
    # Combine the lists of tibbles
    dplyr::bind_rows() 

    
  ##### 3.9 Combine 2.x & 3.x ####
  # Merge the results from 2.x and 3.x 
  runningOccurrences <- runningOccurrences %>%
    # Remove the ambiguous data from the prior-matched data
    dplyr::filter(!database_id %in% runningAmb_occs$database_id) %>%
    # Add in the ambiguous data again.
    dplyr::bind_rows(runningAmb_occs)
  
  # Remove the spent runningAmb_occs data
  rm(runningAmb_occs)
  
  
  #### 4.0 verbatimScientificName ####
  if(checkVerbatim == TRUE){
    writeLines(paste0("checkVerbatim = TRUE. Checking the verbatimScientificName column..."))
    
    ##### 4.1 failedMatches ####
    # Find the data that did not match
    failedMatches <- data %>%
      # Remove the matched names from the OG dataset
      dplyr::filter(!database_id %in% runningOccurrences$database_id) %>%
      # Move the verbatimScientificName to scientificName
      dplyr::mutate(scientificName = verbatimScientificName)
    
    #### 4.2 Run unAmbiguous names ####
    # Run the function
    runningOccurrences_verb <- failedMatches %>%
      # Make a new column with the ordering of rows
      dplyr::mutate(BeeBDC_order = dplyr::row_number()) %>%
      # Group by the row number and step size
      dplyr::group_by(BeeBDC_group = ceiling(BeeBDC_order/stepSize)) %>%
      # Split the dataset up into a list by group
      dplyr::group_split(.keep = TRUE) %>%
      # Run the actual function
      parallel::mclapply(., unAmbiguousFunction,
                         mc.cores = mc.cores
      ) %>%
      # Combine the lists of tibbles
      dplyr::bind_rows() 
    
    
    #### 4.3 Run ambiguous names ####
    # Run the function
    runningAmb_occs_verb <- failedMatches %>%
      # Make a new column with the ordering of rows
      dplyr::mutate(BeeBDC_order = dplyr::row_number()) %>%
      # Group by the row number and step size
      dplyr::group_by(BeeBDC_group = ceiling(BeeBDC_order/stepSize)) %>%
      # Split the dataset up into a list by group
      dplyr::group_split(.keep = TRUE) %>%
      # Run the actual function
      parallel::mclapply(., ambiguousFunction,
                         mc.cores = mc.cores
      ) %>%
      # Combine the lists of tibbles
      dplyr::bind_rows() 
    
    
    #### 4.4 Combine 4.2-.3 ####
    runningOccurrences_verb <- runningOccurrences_verb %>%
      # Remove the ambiguous data from the prior-matched data
      dplyr::filter(!database_id %in% runningAmb_occs_verb$database_id) %>%
      # Add in the ambiguous data again.
      dplyr::bind_rows(runningAmb_occs_verb)
    
    
    #### 4.5 Combine 3.6 & 4.4 ####
    # Merge the results from 2.x and 3.x 
    runningOccurrences <- runningOccurrences %>%
      # Remove the ambiguous data from the prior-matched data
      dplyr::filter(!database_id %in% runningOccurrences_verb$database_id) %>%
      # Add in the ambiguous data again.
      dplyr::bind_rows(runningOccurrences_verb)
    
    # Remove the spent runningAmb_occs data
    rm(runningOccurrences_verb, runningAmb_occs_verb)
    
  }
  
  
  #### 5.0 Merge ####
  writeLines(" - Formatting merged datasets...")
  # merge datasets
  runningOccurrences <- runningOccurrences %>%
    # Put the scientific name into a new column called verbatimScientificName
    dplyr::mutate(verbatimScientificName = scientificName) %>%
    # select the columns we want to keep
    dplyr::select( c(tidyselect::any_of(OG_colnames), validName_valid, 
                     verbatimScientificName,
                     family_valid, subfamily_valid,
                     canonical_withFlags_valid, genus_valid, subgenus_valid, 
                     species_valid, infraspecies_valid, authorship_valid)) %>%
    # rename validName_valid to scientificName and place it where it used to sit.
    dplyr::mutate(scientificName = validName_valid, .after = database_id) %>%
    # Add in the other taxonomic data
    dplyr::mutate(species = canonical_withFlags_valid,
                  family = family_valid, 
                  subfamily = subfamily_valid,
                  genus = genus_valid,
                  subgenus = subgenus_valid,
                  specificEpithet = species_valid,
                  infraspecificEpithet = infraspecies_valid,
                  scientificNameAuthorship = authorship_valid,
                  .after = scientificName) %>%
    # Remove extra columns
    dplyr::select(!c(canonical_withFlags_valid, family_valid, subfamily_valid, genus_valid,
                     subgenus_valid, species_valid, infraspecies_valid, authorship_valid,
                     validName_valid)) %>%
    # Add the .invalidName columns as TRUE (not flagged)
    dplyr::mutate(.invalidName = TRUE)
  
  
  ##### 5.1 User output ####
  nMatchedRows <- nrow(runningOccurrences)
  nUnmatchedRows <- nrow(data) -  nrow(runningOccurrences)
  
  ###### a. failedMatches ####
  # Find the data that did not match
  failedMatches <- data %>%
    # Remove the matched names from the OG dataset
    dplyr::filter(!database_id %in% runningOccurrences$database_id) %>%
    # Add the .invalidName columns as TRUE (not flagged)
    dplyr::mutate(.invalidName = FALSE) 
  # Remove this spent file
  rm(data)
  
  
  ###### b. Add column ####
  runningOccurrences <- runningOccurrences %>%
    # Bind the failed matches
    dplyr::bind_rows(failedMatches) %>%
    # Make sure no duplicates have snuck in
    dplyr::distinct(database_id, .keep_all = TRUE)
  
  
  #### 6.0 Less strict ####
    ##### 6.1 Loosen ambiguous names ####
  # If the user wants to loosen the requirement to exclude ambiguous names then provide that option
  if(relaxAmbiguous == TRUE){
    runningAmbiguous <- runningOccurrences %>%
      # First, select the invalid names
      dplyr::filter(.invalidName == FALSE)
    writeLines(paste0(" You have chose to relax ambiguous names. Checking ",
                      format(nrow(runningAmbiguous), big.mark = ","),
               " failed names now..."))
    
    # Run the unambiguous function, but FIRST remove all of the ambiguous warnings by clearing the 
    # "flags" column in the taxonomy file.
    taxonomy <- taxonomy %>%
      dplyr::mutate(flags = "")
    
    runningAmbiguous_matched <- runningAmbiguous %>%
      # Make a new column with the ordering of rows
      dplyr::mutate(BeeBDC_order = dplyr::row_number()) %>%
      # Group by the row number and step size
      dplyr::group_by(BeeBDC_group = ceiling(BeeBDC_order/stepSize)) %>%
      # Split the dataset up into a list by group
      dplyr::group_split(.keep = TRUE) %>%
      # Run the actual function
      parallel::mclapply(., unAmbiguousFunction,
                         mc.cores = mc.cores
      ) %>%
      # Combine the lists of tibbles
      dplyr::bind_rows() %>%
      # Put the scientific name into a new column called verbatimScientificName
      dplyr::mutate(verbatimScientificName = scientificName) %>%
      # select the columns we want to keep
      dplyr::select( c(tidyselect::any_of(OG_colnames), validName_valid, 
                       verbatimScientificName,
                       family_valid, subfamily_valid,
                       canonical_withFlags_valid, genus_valid, subgenus_valid, 
                       species_valid, infraspecies_valid, authorship_valid)) %>%
      # rename validName_valid to scientificName and place it where it used to sit.
      dplyr::mutate(scientificName = validName_valid, .after = database_id) %>%
      # Add in the other taxonomic data
      dplyr::mutate(species = canonical_withFlags_valid,
                    family = family_valid, 
                    subfamily = subfamily_valid,
                    genus = genus_valid,
                    subgenus = subgenus_valid,
                    specificEpithet = species_valid,
                    infraspecificEpithet = infraspecies_valid,
                    scientificNameAuthorship = authorship_valid,
                    .after = scientificName) %>%
      # Remove extra columns
      dplyr::select(!c(canonical_withFlags_valid, family_valid, subfamily_valid, genus_valid,
                       subgenus_valid, species_valid, infraspecies_valid, authorship_valid,
                       validName_valid)) %>%
      # Add the .invalidName columns as TRUE (not flagged)
      dplyr::mutate(.invalidName = TRUE) %>%
        # Remove the authorship from the scientificName column
      dplyr::mutate(scientificName = stringr::str_remove(scientificName, scientificNameAuthorship) %>%
                      stringr::str_remove_all("\\(\\)") %>% 
                      stringr::str_squish(),
                    scientificNameAuthorship = "ambiguouslyMatchedName")
    # Provide some user feedback
    writeLines(paste0(" BeeBDC has matched ",
                      format(nrow(runningAmbiguous_matched), big.mark = ","),
                      " rows that would otherwise be excluded as ambiguous names.\n",
                      "Remember that these names will be listed as 'ambiguouslyMatchedName' ", 
                      "under the 'scientificNameAuthorship' column and the speciesColumn will NOT ",
                      "include the authority information. Hence, the names can be used but in",
                      "should remain clear in the data that they must be used with caution. "))
    
    # Combine this file with all of the failed names
    runningAmbiguous_combined <- runningAmbiguous %>% 
        # Remove the newly matched ambiguous data
      dplyr::filter(!database_id %in% runningAmbiguous_matched$database_id) %>%
        # Re-add it in again
      dplyr::bind_rows(runningAmbiguous_matched) 
    rm(runningAmbiguous_matched, runningAmbiguous)
    
      # Combine with the overall file
    # Combine this file with all of the failed names
    runningOccurrences <- runningOccurrences %>% 
      # Remove the newly matched ambiguous data
      dplyr::filter(!database_id %in% runningAmbiguous_combined$database_id) %>%
      # Re-add it in again
      dplyr::bind_rows(runningAmbiguous_combined) 
  }

    ##### 6.2 Match higher taxonomy ####
  # Names are only matched at the species level. Here, BeeBDC will also match higher taxonomy 
  # names so that family and such can be added in
if(matchHigherTaxonomy == TRUE){
  # Create and order of operations. Genus goes before subgenus to avoid problems with exact matches
  columnOrder <- c("kingdom"  ,"phylum"    ,"class",
                   "order"    ,"family"    ,"subfamily","tribe",       
                   "subtribe","subgenus","genus") %>% rev()
    # Get a simplified taxonomy of subgenus and higher
  simpleTaxonomy <- taxonomy %>%
    # Take only valid names
    dplyr::filter(taxonomic_status == "accepted") %>%
    dplyr::select(tidyselect::any_of(columnOrder)) %>%
    dplyr::distinct()
    
  runningHigher <- runningOccurrences %>%
    # Get the failed names
    dplyr::filter(.invalidName == FALSE) %>% 
      # Remove the columns of interest
    dplyr::select(!tidyselect::any_of(columnOrder))
  

  # Create a function to check each column for higher taxonomy from bottom up
  higherTaxon_fun <- function(
    higherData = runningHigher,
    higherTaxonomy = simpleTaxonomy,
    levelOrder = columnOrder,
    #level = "family",
    level = NULL){
    # Conditionally remove subgenus from match if searching for genus
    if(level == "genus"){
      higherTaxonomy <- higherTaxonomy %>%
      dplyr::mutate(subgenus = NA_character_)
    }
    if(!level == "subgenus"){
    # For each level, remove the preceding levels' columns
    higherTaxonomy <- higherTaxonomy[,stringr::str_which(string = levelOrder,
                                               pattern = paste0("^",level,"$")):length(levelOrder)]
    }
    # Extract the level to work with 
    taxonLevel <- higherTaxonomy %>%
        # Get only complete cases for that level
      dplyr::filter(complete.cases(.data[[level]])) %>%
        # Get only distinct data
      dplyr::distinct()
    
    # look for matches
    higherData_matched <- higherData %>%
      dplyr::left_join(taxonLevel,
                       by = c("names_clean" = level),
                       keep = TRUE) %>%
      # Re-order columns by the beebdc colTypeR order
      dplyr::select(tidyselect::any_of(c(names(BeeBDC::ColTypeR()$cols),
                                          # Keep original column names also
                                         originalColnames,
                                          # Ensure that all names are included 
                                         c("kingdom"  ,"phylum"    ,"class",
                                           "order"    ,"family"    ,"subfamily","tribe",       
                                           "subtribe","subgenus","genus")))) %>% 
      # Keep only matches
      dplyr::filter(complete.cases(kingdom)) %>% 
        # Add in the taxonRank column info
      dplyr::mutate(taxonRank = dplyr::if_else(is.na(taxonRank),
                                               level, NA_character_))
    # Return matches
    return(higherData_matched)
  }
    # Run the function to find all of the higher taxonomies
  runningHigher_matched <- lapply(
    X = as.list(columnOrder),
    FUN = higherTaxon_fun,
    higherData = runningHigher,
    higherTaxonomy = simpleTaxonomy,
    levelOrder = columnOrder
  ) %>%
    dplyr::bind_rows()  %>%
    dplyr::distinct(database_id, .keep_all = TRUE)
  
  
  # Combine this file with all of the failed names
  runningHigher_combined <- runningHigher %>% 
    # Remove the newly matched ambiguous data
    dplyr::filter(!database_id %in% runningHigher_matched$database_id) %>%
    # Re-add it in again
    dplyr::bind_rows(runningHigher_matched) 
  rm(runningHigher_matched, runningHigher)
  
  # Combine with the overall file
  # Combine this file with all of the failed names
  runningOccurrences <- runningOccurrences %>% 
    # Remove the newly matched ambiguous data
    dplyr::filter(!database_id %in% runningHigher_combined$database_id) %>%
    # Re-add it in again
    dplyr::bind_rows(runningHigher_combined) 
  rm(runningHigher_combined)
  
}
  
  
  #### 7.0 Return and output ####
  ##### 7.1 Repair sp column ####
  # Return the speciesColumn name to it's original state
  names(runningOccurrences)[names(runningOccurrences) == "scientificName"] <- speciesColumn
  if(rm_names_clean == TRUE){
    message("Removing the names_clean column...")
    runningOccurrences <- runningOccurrences %>% 
      dplyr::select(!tidyselect::any_of("names_clean"))
  }
  
  
  # Cut down the failed list...
  # failedMatches <- failedMatches %>%
  #   dplyr::select(tidyselect::any_of("taxonRank")) %>%
  #   dplyr::filter(!taxonRank %in% c("Especie", "forma", "Infrasubspecies", "Race",
  #                                    "species", "Species", "SPECIES", "subsp.", "subspecies",
  #                                   "Subspecies", "SUBSPECIES", "syn", "var.", "variety",
  #                                   "Variety", "VARIETY")) 
  
  
  ##### 7.2 Output ####
  writeLines(paste(
    " - We matched valid names to ", 
    format(sum(runningOccurrences$.invalidName == TRUE), big.mark = ","), " of ",
    format(OG_rowNum, big.mark = ","), " occurrence records. This leaves a total of ",
    format(sum(runningOccurrences$.invalidName == FALSE), big.mark = ","), " unmatched occurrence records.",
    #  " Of the unmatched records, approximately ", 
    # format( nrow(failedMatches), big.mark = ","), 
    #  " are only identified to genus rank or higher.",
    sep = ""))
  
  writeLines(paste("\nharmoniseR:"))
  message(paste(format(sum(runningOccurrences$.invalidName == FALSE), big.mark = ","))) 
  writeLines(paste(
    "records were flagged.\nThe column, '.invalidName' was added to the database.\n"))
  
  message(paste0(
    " - We updated the following columns: ", speciesColumn,", species, family, subfamily, genus, subgenus, ",
    "specificEpithet, infraspecificEpithet, and scientificNameAuthorship. ",
    "The previous ",speciesColumn," column was converted to verbatimScientificName"
  ))
  
  # End time message
  endTime <- Sys.time()
  message(paste(
    " - Completed in ", 
    round(difftime(endTime, startTime), digits = 2 ),
    " ",
    units(round(endTime - startTime, digits = 2)),
    sep = ""))
  
  ##### 7.3 Return ####
  # Return this file
  return(runningOccurrences)
}

requireNamespace("dplyr")

  # Load a test dataset
beesRaw <- BeeBDC::beesRaw

  # Load the small testTaxonomy file
system.file("extdata", "testTaxonomy.rda", package="BeeBDC") |>
  load()


#### 1.0 Run normal function ####
  # Run the function
testOut <- harmoniseR(
  path = tempdir(), #The path to a folder that the output can be saved
  taxonomy = testTaxonomy, # The formatted taxonomy file
  data = beesRaw,
  speciesColumn = "scientificName",
  rm_names_clean = TRUE,
  stepSize = 50,
  mc.cores = 1)


# Test the expected results
testthat::test_that("harmoniseR results unsuccessfuly matched", {
  testthat::expect_equal(sum(testOut$.invalidName == FALSE), 9)
})
testthat::test_that("harmoniseR results successfuly matched", {
  testthat::expect_equal(sum(testOut$.invalidName == TRUE), 91)
})


# Test classes
testthat::test_that("harmoniseR expected class", {
  testthat::expect_type(testOut, "list")
})
testthat::test_that("harmoniseR expected class", {
  testthat::expect_equal(attributes(testOut)$class, c("tbl_df","tbl" ,"data.frame"))
})



#### 2.0 Run checkVerbatim function ####
# Run the function
testOut <- harmoniseR(
  path = tempdir(), #The path to a folder that the output can be saved
  taxonomy = testTaxonomy, # The formatted taxonomy file
  data = beesRaw %>% 
      # Add some extra data to test checkVerbatim = TRUE
    dplyr::bind_rows(
      dplyr::tibble(database_id = c("verb112312", "verb212312"),
                    scientificName = c("Perdita Sandhouse, 1937","Perdita Sandhouse, 1937"),
                    verbatimScientificName = c("Perdita ignota", "Perdita ignota"),
                    family = c("Andrenidae", "Andrenidae"),
                    genus = c("Perdita","Perdita")
                    )
    ),
  speciesColumn = "scientificName",
  rm_names_clean = TRUE,
  checkVerbatim = TRUE,
  stepSize = 50,
  mc.cores = 1)


# Test the expected results
testthat::test_that("harmoniseR results unsuccessfuly matched", {
  testthat::expect_equal(sum(testOut$.invalidName == FALSE), 9)
})
testthat::test_that("harmoniseR results successfuly matched", {
  testthat::expect_equal(sum(testOut$.invalidName == TRUE), 93)
})


# Test classes
testthat::test_that("harmoniseR expected class", {
  testthat::expect_type(testOut, "list")
})
testthat::test_that("harmoniseR expected class", {
  testthat::expect_equal(attributes(testOut)$class, c("tbl_df","tbl" ,"data.frame"))
})

# This function was written by James B Dorey on the 19th December 2022
# Its purpose is to flag rounded coordinates. This replaces bdc_coordinates_precision by flagging ONLY
# Occurrences where both lat AND lon are rounded; not just one.
# Please contact jbdorey[at]me.com for help

#' Flags coordinates for imprecision
#' 
#' This function flags occurrences where BOTH latitude and longitude values are rounded. This 
#' contrasts with the original function, bdc::bdc_coordinates_precision() that will flag 
#' occurrences where only one of latitude OR longitude are rounded. The BeeBDC approach saves 
#' occurrences that may have had terminal zeros rounded in one coordinate column.
#'
#' @param data A data frame or tibble. Occurrence records as input.
#' @param lat Character. The name of the column to use as latitude. Default = "decimalLatitude".
#' @param lon Character. The name of the column to use as longitude. Default = "decimalLongitude".
#' @param ndec Numeric. The number of decimal places to flag in decimal degrees. For example, 
#' argument value of 2 would flag occurrences with nothing in the hundredths place (0.0x).
#' @param quieter Logical. If TRUE, the functino will run a little quieter. Default = FALSE.
#'
#' @return Returns the input data frame with a new column, .rou, where FALSE indicates occurrences 
#' that failed the test.
#' @export
#' 
#' @importFrom dplyr %>%
#'
#' @examples
#' beesRaw_out <- jbd_coordinates_precision(
#'   data = BeeBDC::beesRaw,
#'   lon = "decimalLongitude",
#'   lat = "decimalLatitude",
#'     # number of decimals to be tested
#'   ndec = 2
#' )
#' table(beesRaw_out$.rou, useNA = "always")

jbd_coordinates_precision <-
  function(data,
           lat = "decimalLatitude",
           lon = "decimalLongitude",
           ndec = NULL,
           quieter = FALSE) {
    . <- .ndec_all <- NULL
    
    #### 0.0 Prep ####
    ##### 0.1 errors ####
    ###### a. FATAL errors ####
    if(is.null(ndec)){
      stop(paste0(" - No ndec was provided. This is the minimum number of decimal places",
                  " that the coordinates should have to be considered valid"))
    }
    
    #### 1.0 Run function ####
      ##### 1.1 Prepare data ####
        # Remove a .rou column if it already exists
    data <- data %>%
      dplyr::select(!tidyselect::any_of(".rou"))
    
    
      ##### 1.2 Tests ####
  # Select the columns that you want
    df <-
      data %>%
      dplyr::select({{ lon }}, {{ lat }}) %>%
      as.data.frame()
    
      # get a character vector of the length (number of decimal places) for each lat or lon
    ndec_lat <- (df[, lat] %>%
                   as.character() %>%
                   stringr::str_split_fixed(., pattern = "[.]", n = 2))[, 2] %>%
      stringr::str_length()
    ndec_lon <- (df[, lon] %>%
                   as.character() %>%
                   stringr::str_split_fixed(., pattern = "[.]", n = 2))[, 2] %>%
      stringr::str_length()
    
    rm(df)
    
    ndec_list <- as.list(ndec)
    names(ndec_list) <- paste0(".", "ndec", ndec)
    
    for (i in 1:length(ndec)) {
      ndec_list[[i]] <- (ndec_lat >= ndec[i] | ndec_lon >= ndec[i])
    }
    ndec_list <- dplyr::bind_cols(ndec_list)
    ndec_list$.ndec_all <- apply(ndec_list, 1, all) # all flagged as low decimal precision
    
    ndec_list <-
      ndec_list %>%
      dplyr::select(.ndec_all) %>%
      dplyr::rename(.rou = .ndec_all)
    
    
      #### 2.0 User output ####
    if(quieter == FALSE){
    message("jbd_coordinates_precision:\nFlagged ", 
            format(sum(!ndec_list[".rou"]), big.mark = ","), 
            " records\nThe '.rou' column was added to the database.\n")}else{
                # QUIETER message
              message("jbd_coordinates_precision:\nRemoved ", 
                      format(sum(!ndec_list[".rou"]), big.mark = ","), 
                      " records.")
            }
    
    res <- dplyr::bind_cols(data, ndec_list)
    
    return(res)
  }

# This function was written by James Dorey to build taxonomy files using taxadb and transform them
# into the BeeBDC input format
# This function was written from the 19th of March 2024. For questions, please email James
# at jbdorey[at]me.com

#' Import and convert taxadb taxonomies to BeeBDC format
#' 
#' Uses the taxadb R package to download a requested taxonomy and then transforms it into the input
#' BeeBDC format. This means that any taxonomy in their databases can be used with BeeBDC. You can
#' also save the output to your computer and to the R environment for immediate use. See 
#' details below for a list of providers or see `taxadb::td_create()`.
#' 
#'
#' @param name Character. Taxonomic scientific name (e.g. "Aves"). 
#' As defined by  `taxadb::filter_rank()`.
#' @param rank Character. Taxonomic rank name. (e.g. "class"). 
#' As defined by `taxadb::filter_rank()`.
#' @param provider Character. From which provider should the hierarchy be returned?
#' Default is 'gbif', which can also be configured using options(default_taxadb_provide = ...").
#'  See `taxadb::td_create()` for a list of recognized providers. NOTE: gbif seems to have the most-complete 
#'  columns, especially in terms of scientificNameAuthorship, which is important for matching
#'  ambiguous names.
#'  As defined by `taxadb::filter_rank()`.
#' @param version Character. Which version of the taxadb provider database should we use? defaults 
#' to latest. See tl_import for details. Default = 22.12. 
#' As defined by `taxadb::filter_rank()`.
#' @param collect Logical. Should we return an in-memory data.frame 
#' (default, usually the most convenient), or a reference to lazy-eval table on disk 
#' (useful for very large tables on which we may first perform subsequent filtering operations.).
#' Default = TRUE. 
#' As defined by `taxadb::filter_rank()`.
#' @param ignore_case Logical. should we ignore case (capitalization) in matching names? 
#' Can be significantly slower to run. Default = TRUE. 
#' As defined by `taxadb::filter_rank()`.
#' @param db a connection to the taxadb database. See details of `taxadb::filter_rank()`. Default 
#' = Null which should work. 
#' As defined by `taxadb::filter_rank()`.
#' @param ... Arguments passed to `taxadb::td_create()`.
#' 
#' @param removeEmptyNames Logical. If True (default), it will remove entries without an entry
#' for specificEpithet.
#' @param outPath Character. The path to a directory (folder) in which the output should be saved.
#' @param fileName Character. The name of the output file, ending in '.csv'.
#'
#'
#' @return Returns a taxonomy file (to the R environment and to the disk, if a fileName is 
#' provided) as a tibble that can be used with `BeeBDC::harmoniseR()`.
#' 
#' @importFrom dplyr %>%
#' 
#' @seealso [BeeBDC::beesTaxonomy()] for the bee taxonomy and [BeeBDC::harmoniseR()] for the 
#' taxon-cleaning function where these taxonomies are implemented.
#' 
#' @export
#'
#' @examples
#' \dontrun{
#' 
#' if(require("taxadb")){
#' 
#'   # Run the function using the bee genus Apis as an example...
#' ApisTaxonomy <- BeeBDC::taxadbToBeeBDC(
#'   name = "Apis",
#'   rank = "Genus",
#'   provider = "gbif",
#'   version = "22.12",
#'   removeEmptyNames = TRUE,
#'   outPath = getwd(),
#'   fileName = NULL,
#'   ...
#'   )
#'   
#'   } # END if require
#'   } # End dontrun
#' 

taxadbToBeeBDC <- function(
    name = NULL,
    rank = NULL,
    provider = "gbif",
    version = "22.12",
    collect = TRUE,
    ignore_case = TRUE,
    db = NULL,
    removeEmptyNames = TRUE,
    outPath = getwd(),
    fileName = NULL,
    ...
) {  
  # locally bind variables to the function
  . <- taxonomy_taxadb <- taxonomyOut <- canonical <- authorship <- taxonomic_status <- species <-
    taxonID <- id <- accid <- id_matched <- NULL

  
  # Load required packages 
  requireNamespace("stringr")
  requireNamespace("dplyr")
  requireNamespace("taxadb")
  
  
  #### 0.0 Prep ####
  ##### 0.1 Errors ####
  ###### a. FATAL errors ####
  if(is.null(name)){
    stop(" - Please provide an argument for name I'm a program not a magician.")
  }
  if(is.null(rank)){
    stop(" - Please provide an argument for rank I'm a program not a magician.")
  }
  if(provider == "ncbi"){
    stop(" - Sorry, ncbi doesn't include a taxonID against which to match the acceptedNameUsageID.")
  }
  if (!provider %in% c("itis",
                 "ncbi",
                 "col",
                 "tpl",
                 "gbif",
                 "fb",
                 "slb",
                 "wd",
                 "ott",
                 "iucn")) {
    stop(provider, " provided is not a valid name")
  }
  
  
  ##### 0.2 taxadb test ####
  ###### a. test ####
  # Check if taxadb is installed
  # TRUE if taxadb is found
  suppressWarnings(
    suggestedTest <- system.file(package='taxadb') %>% 
      stringr::str_count() > 0 
  )

  ###### b. taxadb ####
  if(suggestedTest == FALSE){
    # Set up instructions for download on fail
    instructions <- paste(" Please try installing the package for yourself", 
                          "using the following command: \n",
                          "install.packages(\"taxadb\")")
    # Set up fail function for tryCatch
    error_func <- function(e){
      stop(paste("Failed to install the taxadb package.\n", 
                 instructions))
    }
    # Begin interactive input
    input <- 1
    if (interactive()){
      input <- utils::menu(c("Yes", "No"), 
                           title = paste0("Install the taxadb package? \n"))
    }
    if(input == 1){
      # Start taxadb install
      message("Installing the taxadb package.")
      tryCatch(
        utils::install.packages("taxadb"), 
        error = error_func, warning = error_func)
    } # END input == 1
    
    else{
      stop(writeLines(paste("The taxadb package is necessary for BeeBDC::taxadbToBeeBDC.\n", 
                            instructions)))
    } # END else
  } # END suggestedTest == FALSE
  

#### 1.0 Download taxonomy ####
  ##### 1.1 Download ####
  writeLines(" - Downloading taxonomy...")
  taxadb::td_create(provider = provider,
                    schema = "dwc",
                    version = version,
                    # Only provide inputs here if user-inputs are provided.
                    if(is.null(db)){db = taxadb::td_connect()
                    }else{db = db},
                    ...)
    # User output
  writeLines(paste0(" - taxadb save the taxonomy to: ",
                    taxadb::taxadb_dir()))
  
  ##### 1.2 Turn into data table ####
  # Run the filter_rank function to output the data table taxonomy
taxonomy_taxadb <- taxadb::filter_rank(name,
                                       rank,
                                       provider = provider,
                                       collect = collect,
                                       ignore_case = ignore_case,
                                        # Only provide inputs here if user-inputs are provided.
                                       if(is.null(db)){db = taxadb::td_connect()
                                       }else{db = db},
                                       version = version
) 
  
  ##### 1.3 missing columns ####
  if(!"scientificNameAuthorship" %in% colnames(taxonomy_taxadb)){
    taxonomy_taxadb <- taxonomy_taxadb %>%
      dplyr::mutate(scientificNameAuthorship = NA_character_)
    warning(paste0(" - BeeBDC: no scientificNameAuthorship in downloaded data. BeeBDC really ", 
                   "likes this column as it helps identify ambiguities."))
  }
  if(!"infraspecificEpithet" %in% colnames(taxonomy_taxadb)){
    taxonomy_taxadb <- taxonomy_taxadb %>%
      dplyr::mutate(infraspecificEpithet = NA_character_)
    warning(paste0(" - BeeBDC: no infraspecificEpithet in downloaded data. This can be a really",
                   " helpful column for some taxa."))
  }
  
#### 2.0 Transform data ####
  ##### 2.1 basic rename and mutate ####
    # Begin transforming the taxonomy to BeeBDC format
taxonomyOut <- taxonomy_taxadb %>%
      # Rename columns... sadly away from DWC... for now.
    dplyr::rename(taxonomic_status = "taxonomicStatus",
                  authorship = "scientificNameAuthorship",
                  infraspecies = "infraspecificEpithet",
                  species = "specificEpithet",
                  taxon_rank = "taxonRank",
                  canonical = "scientificName") %>%
      # Build new columns
    dplyr::mutate(validName = stringr::str_c(canonical, 
                                             dplyr::if_else(!is.na(authorship),
                                                            paste0(authorship), ""),
                                             sep = " ") %>%
                                               stringr::str_squish(),
                  canonical_withFlags = canonical,
                  valid = dplyr::if_else(taxonomic_status == "accepted",
                                         TRUE, FALSE))
  ##### 2.2 Remove empty names ####
      # Remove empty names
  if(removeEmptyNames == TRUE){
    taxonomyOut <- taxonomyOut %>%
      dplyr::filter(stats::complete.cases(species))
  }
  
  ##### 2.3 Add id and accid ####
    # Add id and accepted accid
  taxonomyOut <- taxonomyOut %>%
      # Add id as a simple count, top to bottom
    dplyr::mutate(id = 1:nrow(.)) %>%
      # Add accid for the accepted names
    dplyr::mutate(accid = dplyr::if_else(taxonomic_status == "accepted",
                                 0, NA_integer_)) 
    # Match the synonyms to their accepted names
  taxonomyOut <- taxonomyOut %>%
    dplyr::left_join(taxonomyOut %>% dplyr::select(taxonID, id),
                     by = c("acceptedNameUsageID" = "taxonID"),
                     suffix = c("", "_matched")) %>%
      # Transfer this id to the accid column
    dplyr::mutate(accid = dplyr::if_else(is.na(accid),
                                         id_matched, accid)) %>%
      # Drop the temporary column
    dplyr::select(!id_matched) %>%
      # Add in source
    dplyr::mutate(source = stringr::str_c("taxadb_",provider, sep = "")) %>%
      # Add in empty columns
    dplyr::mutate(flags = NA_character_,
                  notes = NA_character_)
  
  ##### 2.4 Clean missing columns ####
    # Clean up potentially missing columns
  if(!"subfamily" %in% colnames(taxonomyOut)){
    taxonomyOut <- taxonomyOut %>%
      dplyr::mutate(subfamily = NA_character_)
  }
  if(!"tribe" %in% colnames(taxonomyOut)){
    taxonomyOut <- taxonomyOut %>%
      dplyr::mutate(tribe = NA_character_)
  }
  if(!"subtribe" %in% colnames(taxonomyOut)){
    taxonomyOut <- taxonomyOut %>%
      dplyr::mutate(subtribe = NA_character_)
  }
  if(!"subgenus" %in% colnames(taxonomyOut)){
    taxonomyOut <- taxonomyOut %>%
      dplyr::mutate(subgenus = NA_character_)
  }

    ##### 2.5 Re-order columns ####
  taxonomyOut <- taxonomyOut %>%
    dplyr::relocate(c("flags","taxonomic_status","source","accid","id",
                  "kingdom","phylum","class","order","family",
                  "subfamily","tribe","subtribe","validName","canonical",
                  "canonical_withFlags","genus","subgenus","species","infraspecies","authorship",
                  "taxon_rank","valid","notes"))
  
  
  #### 3.0 Identify ambiguities ####
    # Test for duplicate and ambiguous names
  taxonomyOut <- taxoDuplicator(
    SynList = taxonomyOut,
    source1 = provider,
      # This is not needed in this instance
    source2 = "")
  
  #### 4.0 Save ####
    # Save the output file if a fileName is provided
  if(!is.null(fileName)){
    readr::write_excel_csv(taxonomyOut,
                           file = paste0(outPath, "/", fileName, sep = ""))
  }
  
    # Return the dataset
  return(taxonomyOut)
  
  
  
} # END taxadbToBeeBDC
  

requireNamespace("dplyr")
requireNamespace("BeeBDC")

# If rnaturalearthdata is present, run tests
if(requireNamespace("taxadb")){

  requireNamespace("taxadb")

  # Run the taxadbToBeeBDC function using the example
 ApisTaxonomy <- BeeBDC::taxadbToBeeBDC(name = "Apis",
                                        rank = "Genus",
                                        provider = "gbif",
                                        version = "22.12",
                                        removeEmptyNames = TRUE,
                                        outPath = tempdir(),
                                        fileName = "TEST_out.csv",
                                          # Must be NULL to avoid deprecation warning
                                        overwrite = NULL, lines = NULL)
 


# Test the expected results
# Test classes
testthat::test_that("taxadbToBeeBDC expected class", {
  testthat::expect_type(ApisTaxonomy, "list")
})
testthat::test_that("taxadbToBeeBDC expected class", {
  testthat::expect_equal(attributes(ApisTaxonomy)$class, c("tbl_df","tbl","data.frame"))
})

# Check directory that the plot was saved
testthat::test_that("taxadbToBeeBDC plot saved?", {
  testthat::expect_true(any(stringr::str_detect(list.files(tempdir()), "TEST_out.csv")))
})

# Check the number of columns in the dataset
testthat::test_that("taxadbToBeeBDC expected number of columns", {
  testthat::expect_equal(length(colnames(ApisTaxonomy)), 29)
})


} # END if require

   # This function was written by James B Dorey starting from the 3rd-7th of June 2022.
# It is intended to replace the bdc function bdc_coordinates_country_inconsistent for larger datasets
# where it is simply not feasible. 
  # Initial function finished: _____
  # For questions, please ask James at jbdorey[at]me.com

#' Flags coordinates that are inconsistent with the stated country name
#' 
#' Compares stated country name in an occurrence record with record's coordinates using 
#' rnaturalearth data. The prefix, jbd_ is meant
#' to distinguish this function from the original [bdc::bdc_coordinates_country_inconsistent()].
#' This functions will preferably use the countryCode and country_suggested columns 
#' generated by `bdc::bdc_country_standardized()`; please run it on your dataset prior to running
#' this function.
#'
#' @param data A data frame or tibble. Occurrence records as input.
#' @param lon Character. The name of the column to use as longitude. Default = "decimalLongitude".
#' @param lat Character. The name of the column to use as latitude. Default = "decimalLatitude".
#' @param scale Numeric or character. To be passed to [rnaturalearth::ne_countries()]'s scale.
#' Scale of map to return, one of 110, 50, 10 or "small", "medium", "large". 
#' Smaller values return higher-resolution maps.
#' @param pointBuffer Numeric. Amount to buffer points, in decimal degrees. If the point is outside 
#' of a country, but within this point buffer, it will not be flagged. Default = 0.01.
#' @param mc.cores Numeric. If > 1, the st_intersects function will run in parallel
#' using mclapply using the number of cores specified. If = 1 then it will be run using a serial
#' loop. NOTE: Windows machines must use a value of 1 (see ?parallel::mclapply). Additionally,
#' be aware that each thread can use large chunks of memory.
#'  Default = 1.
#' @param stepSize Numeric. The number of occurrences to process in each chunk. Default = 1000000.
#'
#' @return The input occurrence data with a new column, .coordinates_country_inconsistent
#' @export
#' 
#' @importFrom dplyr %>%
#'
#' @examples
#' if(requireNamespace("rnaturalearthdata")){
#' beesRaw_out <- jbd_coordCountryInconsistent(
#'   data = BeeBDC::beesRaw,
#'   lon = "decimalLongitude",
#'   lat = "decimalLatitude",
#'   scale = 50,
#'   pointBuffer = 0.01)
#' } # END if require

jbd_coordCountryInconsistent <- function(
    data = NULL,
    lon = "decimalLongitude",
    lat = "decimalLatitude",
    scale = 50,
    pointBuffer = 0.01,
    stepSize = 1000000,
    mc.cores = 1){
  
  database_id <- decimalLatitude <- decimalLongitude <- country <- name_long <- iso_a2 <- 
    geometry <- admin <- sovereignt <- name <- . <- NULL
  .coordinates_outOfRange <- .coordinates_empty <- indexMatch <- BeeBDC_order <- NULL
  countryCode <- country_suggested <- NULL
  .coordinates_empt <- .data <- .coordinates_empty <- .coordinates_outOfRange <- isna_ <- NULL
  
startTime <- Sys.time()
requireNamespace("rnaturalearth")
requireNamespace("dplyr")
requireNamespace("ggspatial")
requireNamespace("mgsub")
requireNamespace("terra")

  #### 0.0 Prep ####
    ###### 0.1 fatal errors ####
if(!any(colnames(data) %in% "country")){
stop("There is no column called 'country' in the dataset. This is a minimum requirement.")
  }
  
    ###### 0.2 Coord quality ####
if(!any(colnames(data) %in% ".coordinates_outOfRange")){
    writeLines("No '.coordinates_outOfRange' column found, running bdc_coordinates_outOfRange...")
  
  bdc_coordinates_outOfRange_internal <- function(data, lat = "decimalLatitude", lon = "decimalLongitude") {
      .data <- .coordinates_outOfRange <- NULL
      data_filtered <- data %>%
        dplyr::select(dplyr::all_of(c(lon, lat))) %>%
        dplyr::rename(lon = dplyr::all_of(lon), lat = dplyr::all_of(lat)) %>%
        dplyr::mutate(dplyr::across(dplyr::everything(), ~ as.numeric(.x)))
      data_flag <- data_filtered %>% dplyr::mutate(.coordinates_outOfRange = dplyr::case_when(
            lat < -90 | lat > 90 ~ FALSE,
            lon < -180 | lon > 180 ~ FALSE,TRUE ~ TRUE)) %>%
        dplyr::select(.coordinates_outOfRange)
      df <- dplyr::bind_cols(data, data_flag)
      message(paste("\nbdc_coordinates_outOfRange:\nFlagged",
          sum(df$.coordinates_outOfRange == FALSE),
          "records.\nOne column was added to the database.\n"))
      return(df)
    }

  data <- bdc_coordinates_outOfRange_internal(
    data = data,
    lat = lat,
    lon = lon)
}
###### 0.3 columns present ####
if(!any(colnames(data) %in% ".coordinates_empty")){
  writeLines("No '.coordinates_empty' column found, running bdc_coordinates_empty")
  
  check_col_internal <- function(data, col) {
    for (i in seq_along(col)) {
      if (!col[i] %in% colnames(data)) {
        stop("Column `", col[i], "` is not present in the data", call. = FALSE)}}}
  
  bdc_coordinates_empty_internal <- function(data, lat = "decimalLatitude", lon = "decimalLongitude") {
      if (!is.data.frame(data)) {
        stop(deparse(substitute(data)), " is not a data.frame", call. = FALSE)}
      check_col_internal(data, c(lat, lon))
      df <- data %>%
        dplyr::select(dplyr::all_of(c(lat, lon))) %>%
        dplyr::mutate(dplyr::across(dplyr::all_of(c(lat, lon)), ~ as.numeric(.x))) %>%
        dplyr::mutate(isna_ = dplyr::if_any(dplyr::all_of(c(lat, lon)), ~ is.na(.x)),
          .coordinates_empty = ifelse(isna_, FALSE, TRUE)) %>%
        dplyr::select(.coordinates_empty)
      df <- dplyr::bind_cols(data, df)
      message(paste("\nbdc_coordinates_empty:\nFlagged",
          sum(df$.coordinates_empty == FALSE),
          "records.\nOne column was added to the database.\n"))
      return(df)
    }
  
  data <- bdc_coordinates_empty_internal(
    data = data,
    lat = lat,
    lon = lon)
}
if(!any(colnames(data) %in% "country_suggested")){
  writeLines(paste0("No 'country_suggested' column found, adding an empty (NA) placeholder. This",
                    " column can be added by running bdc::bdc_country_standardized() on the ",
                    "input data."))
  data <- data %>%
    dplyr::mutate(country_suggested = NA_character_)
}
if(!any(colnames(data) %in% "countryCode")){
  writeLines(paste0("No 'countryCode' column found, adding an empty (NA) placeholder. This",
                    " column can be added by running bdc::bdc_country_standardized() on the ",
                    "input data."))
  data <- data %>%
    dplyr::mutate(countryCode = NA_character_)
}


  # Remove poor-quality coordinates
dataR <- data %>%
  dplyr::filter(!.coordinates_outOfRange == FALSE) %>%
  dplyr::filter(!.coordinates_empty == FALSE) 


    # Reduce dataset 
  dataR <- dataR %>%
    dplyr::select(
      tidyselect::any_of(c("database_id", "decimalLatitude", "decimalLongitude",
                         "country", "countryCode","country_suggested"))) %>%
      # Remove lat/lon NAs
    dplyr::filter(!is.na(decimalLatitude)) %>% dplyr::filter(!is.na(decimalLongitude))
  
  #### 1.1 Terrestrial map ####
      ##### 1.1 rnaturalearth DL ####
  writeLines(" - Downloading naturalearth map...")
  suppressWarnings({
  # Download the rnaturalearth countries
vectEarth <- rnaturalearth::ne_countries(scale = scale, type = "countries", 
                                    returnclass = "sf" )%>%
  dplyr::select(name_long, iso_a2, geometry, admin, sovereignt, name) %>%
  sf::st_make_valid()
# Simplify the world map ONCE to be used later
simplePoly <- vectEarth %>% sf::st_drop_geometry() %>%
  dplyr::mutate(indexMatch = dplyr::row_number())

  # Repair gemoetries
sf::sf_use_s2(FALSE)
})

  #### 2.0 Extractions ####
    ##### 2.1 Create function 1 ####
intersectFun <- function(sp){
  suppressMessages({
  extracted <- sf::st_intersects(sp, vectEarth, sparse = TRUE) %>% 
    # return a tibble with the index of each match or NA where there was no match
    dplyr::tibble(indexMatch = .) 
    # If first element is full, unlist each one
      extracted <- extracted %>%
      dplyr::mutate(indexMatch = indexMatch %>% as.character() %>% as.numeric() )
    # rejoin
  extracted <- extracted %>%
    dplyr::left_join(simplePoly,
                     by = "indexMatch") %>%
    # Add in the database_id
    dplyr::bind_cols(sp)
  }) # END suppressWarnings
  return(extracted)
  }# END intersectFun function



    ##### 2.2 Country name ####
writeLines(" - Extracting initial country names without buffer...")
suppressWarnings({
      # Turn the points into an sf object
sp <- sf::st_as_sf(dataR, coords = c(lon, lat),
                   crs = sf::st_crs("WGS84"))
  # Extract the country for the points from the vectEarth map
  country_extracted = sp %>%
    # Make a new column with the ordering of rows
    dplyr::mutate(BeeBDC_order = dplyr::row_number()) %>%
    # Group by the row number and step size
    dplyr::group_by(BeeBDC_group = ceiling(BeeBDC_order/stepSize)) %>%
    # Split the dataset up into a list by group
    dplyr::group_split(.keep = TRUE) %>%
    # Run the actual function
    parallel::mclapply(., intersectFun,
                       mc.cores = mc.cores
    ) %>%
    # Combine the lists of tibbles
    dplyr::bind_rows()
})
  

    ##### 2.3 Failures ####
  # Find those records that don't match.
failed_extract <- country_extracted %>%
    # Find the mis-matched countries
  dplyr::filter(!tolower(country) %in% c(tolower(name_long), tolower(admin),
                                                   tolower(sovereignt), tolower(name))) %>%
  dplyr::filter(!tolower(country_suggested) %in% c(tolower(name_long), tolower(admin),
                                                   tolower(sovereignt), tolower(name))) %>%
  dplyr::filter(!tolower(iso_a2) %in% c(tolower(country), tolower(countryCode))) %>%
    # Remove NA countries
  dplyr::filter(!is.na(country)) %>%
  dplyr::tibble() %>%
  sf::st_as_sf(crs = sf::st_crs("WGS84"))
    # Replace some country names as needed and re-remove
failed_extract$country <-
  mgsub::mgsub(failed_extract$country, 
               pattern = c("Martinique", "French guiana", "Federated States of Micronesia"),
               replacement = c("France", "France", "Micronesia"))
  # Remove new matches
failed_extract <- failed_extract %>%
  # Find the mis-matched countries
  dplyr::filter(!tolower(country) %in% c(tolower(name_long), tolower(admin),
                                                   tolower(sovereignt), tolower(name))) %>%
  dplyr::filter(!tolower(country_suggested) %in% c(tolower(name_long), tolower(admin),
                                                   tolower(sovereignt), tolower(name))) %>%
  dplyr::filter(!tolower(iso_a2) %in% c(tolower(country), tolower(countryCode))) %>%
  # Remove NA countries
  dplyr::filter(!is.na(country)) %>%
  dplyr::tibble() %>%
  sf::st_as_sf(crs = sf::st_crs("WGS84"))
# remove spent dataset
rm(country_extracted)


    ##### 2.4 Buffer fails ####
writeLines(" - Buffering naturalearth map by pointBuffer...")
  # Buffer the natural earth map
suppressWarnings({
  vectEarth <- vectEarth %>% 
  sf::st_buffer(dist = pointBuffer)
})

writeLines(" - Extracting FAILED country names WITH buffer...")
# Extract the country for the points from the vectEarth map
suppressWarnings({
  failed_extract_2 = failed_extract %>%
    dplyr::select(database_id, country, geometry, country_suggested, countryCode) %>%
    # Make a new column with the ordering of rows
    dplyr::mutate(BeeBDC_order = dplyr::row_number()) %>%
    # Group by the row number and step size
    dplyr::group_by(BeeBDC_group = ceiling(BeeBDC_order/stepSize)) %>%
    # Split the dataset up into a list by group
    dplyr::group_split(.keep = TRUE) %>%
    # Run the actual function
    parallel::mclapply(., intersectFun,
                       mc.cores = mc.cores
    ) %>%
    # Combine the lists of tibbles
    dplyr::bind_rows()
})
    # Find MATCHES #
  # With country
fExtr_1 <- failed_extract_2 %>% 
  dplyr::filter(!tolower(country) %in% c(tolower(name_long), tolower(admin),
                                         tolower(sovereignt), tolower(name))) %>%
  dplyr::filter(!tolower(country_suggested) %in% c(tolower(name_long), tolower(admin),
                                                   tolower(sovereignt), tolower(name)))
  # With iso_a2
fExtr_2 <- failed_extract_2 %>% 
  dplyr::filter(tolower(iso_a2) %in% c(tolower(country), tolower(countryCode)))

ids2keep <- dplyr::bind_rows(fExtr_1, fExtr_2) %>%
  # Find the mis-matched countries
  dplyr::filter(!tolower(country) %in% c(tolower(name_long), tolower(admin),
                                         tolower(sovereignt), tolower(name))) %>%
  dplyr::filter(!tolower(country_suggested) %in% c(tolower(name_long), tolower(admin),
                                                   tolower(sovereignt), tolower(name))) %>%
  dplyr::filter(tolower(iso_a2) %in% c(tolower(country), tolower(countryCode))) %>%
    # Keep only the database id
  dplyr::select(database_id)

  #### 3.0 Final fails ####
# Get the final fails by removing those to keep from the failed list
ids2remove <- failed_extract_2 %>%
  dplyr::filter(!database_id %in% ids2keep$database_id) %>%
  dplyr::select(database_id)


  #### 4.0 Finals ####
    # Create new column
data <- data %>% 
  dplyr::mutate(.coordinates_country_inconsistent = !database_id %in% ids2remove$database_id)

    # return message
message(paste("\njbd_coordinates_country_inconsistent:\nFlagged", 
              format(sum(data$.coordinates_country_inconsistent == FALSE, na.rm = TRUE), big.mark = ","),
              "records.\nThe column, '.coordinates_country_inconsistent',",
              "was added to the database.\n"), sep = "")
 endTime <- Sys.time()
    # Time output
 message(paste(
   " - Completed in ", 
   round(difftime(endTime, startTime), digits = 2 ),
   " ",
   units(round(endTime - startTime, digits = 2)),
   sep = ""))

  # Return the data
return(data)
} # END function
  

check_require_cran <- function(pkg) {
  if (!requireNamespace(pkg, quietly = TRUE)) {
    stop(paste0("Package `", pkg, "` needed for this function to work! Please, install it with: `install.packages(\"", pkg, "\")`"), call. = FALSE)
  } else {
    require(pkg, character.only = TRUE)
  }
}

check_require_github <- function(pkg) {
  userpkg <- pkg
  pkg <- basename(pkg)
  if (!requireNamespace(pkg, quietly = TRUE)) {
    stop(paste0("Package `", pkg, "` needed for this function to work! Please, install it with: remotes::install_github(\"", userpkg, "\")"), call. = FALSE)
  } else {
    require(pkg, character.only = TRUE)
  }
}

# This Function is designed to check country-level outliers using the Discover Life checklist.
# It was written by James B Dorey from the 8th of November 2022.

#' Flag country-level outliers with a provided checklist.
#'
#'This function flags country-level outliers using the checklist provided with this package. 
#'For additional context and column names, see [BeeBDC::beesChecklist()].
#'
#' @param checklist A data frame or tibble. The formatted checklist which was built based on the Discover Life website.
#' @param data A data frame or tibble. The a Darwin Core occurrence dataset.
#' @param keepAdjacentCountry Logical. If TRUE, occurrences in countries that are adjacent to checklist countries will be 
#' kept. If FALSE, they will be flagged.
#' @param pointBuffer Numeric. A buffer around points to help them align with a country or coastline.
#' This provides a good way to retain points that occur right along the coast or borders of the 
#' maps in rnaturalearth
#' @param scale Numeric. The value fed into the map scale parameter for
#'  [rnaturalearth::ne_countries()]'s scale parameter:
#' 	Scale of map to return, one of 110, 50, 10 or 'small', 'medium', 'large', where smaller numbers 
#' 	are higher resolution. WARNING: This function is tested on 110 and 50.
#' @param stepSize Numeric. The number of occurrences to process in each chunk. Default = 1000000.
#' @param mc.cores Numeric. If > 1, the function will run in parallel
#' using mclapply using the number of cores specified. If = 1 then it will be run using a serial
#' loop. NOTE: Windows machines must use a value of 1 (see ?parallel::mclapply). Additionally,
#' be aware that each thread can use large chunks of memory. If the cores throw issues, consider
#' setting mc.cores to 1.
#'  Default = 1.
#'
#' @return The input data with two new columns, .countryOutlier or .sea. There are three possible 
#' values for 
#' the new column: TRUE == passed, FALSE == failed (not in country or in the ocean),
#'  NA == did not overlap with rnaturalearth map.
#' 
#' @seealso [BeeBDC::continentOutlieRs()] for implementation at the continent level. Implementation 
#' at the continent level may be lighter and more manageable on the data side of things where 
#' country-level checklists don't exist. Additionally, see [BeeBDC::beesChecklist()] for input data.
#' Note, not all columns are necessary if you are building your own dataset. 
#' At a minimum you will need *validName*, *country*, *iso_a3_eh* (to match rnaturalearth).
#' 
#' @export
#' @importFrom dplyr %>%
#'
#' @examples
#' if(requireNamespace("rnaturalearthdata")){
#' library(magrittr)
#'   # Load in the test dataset
#' beesRaw <- BeeBDC::beesRaw
#'   # For the sake of this example, use the testChecklist
#' system.file("extdata", "testChecklist.rda", package="BeeBDC") |> load()
#'   # For real examples, you might download the beesChecklist from FigShare using 
#'   #  [BeeBDC::beesChecklist()]
#' 
#' beesRaw_out <- countryOutlieRs(checklist = testChecklist,
#'                                data = beesRaw %>%
#'                                dplyr::filter(dplyr::row_number() %in% 1:50),
#'                                keepAdjacentCountry = TRUE,
#'                                pointBuffer = 1,
#'                                scale = 50,
#'                                stepSize = 1000000,
#'                                mc.cores = 1)
#' table(beesRaw_out$.countryOutlier, useNA = "always")
#' } # END if require

countryOutlieRs <- function(
    checklist = NULL,
    data = NULL,
    keepAdjacentCountry = TRUE,
    pointBuffer = NULL,
    scale = 50,
    stepSize = 1000000,
    mc.cores = 1
    ){
  # locally bind variables to the function
  iso_a2<-iso_a3_eh<-name<-name_long<-continent<-geometry<-countryOutlieRs<-decimalLongitude<-
    countryOutlieRs<-decimalLatitude<-database_id<-countryOutlieRs<-scientificName<-species<-
    family<-subfamily<-genus<-countryOutlieRs<-specificEpithet<-countryOutlieRs<-
    scientificNameAuthorship<-country<-stateProvince<-eventDate<-countryOutlieRs<-
    institutionCode<-recordNumber<-catalogNumber<-dataSource<-countryOutlieRs<-
    verbatim_scientificName<-.<-neighbours<-rowNum<-countryOutlieRs<-neighboursText<-
    SciCountry<-validName<-countryOutlieRs<-SciCountry_noYear<-countryOutlieRs<-
    neighbourMatch_noYear<-countryOutlieRs<-exactMatch_noYear<-matchType<-countryMatch<-
    countryOutlieRs<-countryOutlieRs<-.countryOutlier <- BeeBDC_order <- BeeBDC_group <- points <- 
    inData <- indexMatch <- NULL
  
  # REMOVE - TEST thinning
   #  data <- data %>%
   #    filter(row_number() %% 100 == 1)
  startTime <- Sys.time()
#### 0.0 Warnings ####
  if(is.null(checklist)){
    stop("You must provide a checklist of countries")
  }
  if(is.null(data)){
    stop("You must provide occurrence data (data). Honestly, what do you think I was gonna do without that?")
  }
  # Find if an older version of beesChecklist is in use (by finding the column "Alpha-3")
  if("Alpha-3" %in% colnames(checklist)){
    warning(paste0("Warning: an older version of beesChecklist has been detected based on the",
                   " presence of the 'Alpha-3' column. This checklist is dated as of BeeBDC",
                   " version 1.1.2. We will change the column name here to 'iso_a3_eh' and use it,",
                   " but you will",
                   "find a newer checklist with more-recent versions of BeeBDC."))
    writeLines(" - Re-naming 'Alpha-3' column to new 'iso_a3_eh' column to match new rnaturalearth.")
    checklist <- checklist %>%
      dplyr::rename(iso_a3_eh = "Alpha-3")
  }

#### 1.0 Data prep ####
  ##### 1.1 data ####
    # Drop .countryOutlier if its already present
  data <- data %>%
    dplyr::select(!tidyselect::any_of(".countryOutlier")) %>%
      # Remove other columns made by this function
    dplyr::select(!tidyselect::starts_with(c("iso_a3_eh", "countryMatch")))

  ##### 1.2 rNaturalEarth ####
# Download world map using rnaturalearth packages
  suppressWarnings({
countryMap <- rnaturalearth::ne_countries(returnclass = "sf", country = NULL,
                                        type = "countries", scale = scale)  %>%
      # buffer by zero and make geometry valid to avoid potential issues with polygon intersection
    sf::st_make_valid() %>%
  # Select only a subset of the naturalearthdata columns to extract
  dplyr::select(iso_a2, iso_a3_eh, name, name_long, continent, geometry) %>%
  # Replace some country codes to match the checklist
  dplyr::mutate(iso_a3_eh = iso_a3_eh %>% stringr::str_replace_all(
    c("ALA" = "FIN", # Aland Islands == Finland
    "ASM" = "WSM", # American Samoa == Samoa
    "HKG" = "CHN", # Hong Kong == China
    "MAF" = "SXM"))) # Saint-Martin == Sint Maarten
})
  
  # Simplify the world map ONCE to be used later
  simplePoly <- countryMap %>% sf::st_drop_geometry() %>%
    dplyr::mutate(indexMatch = dplyr::row_number())
  
  # Dont's use spherical geometry
sf::sf_use_s2(FALSE)


  #### 2.0 Use occ. data ####
    ##### 2.1 Create functions ####
      ###### a. simple intersect ####
# Hijack st_intersection to allow it to be run in parallel
jbd_intersection <- function(inData){
  inData <- inData %>% tidyr::drop_na(decimalLongitude, decimalLatitude)
  suppressWarnings({ suppressMessages({
    # Turn inData into a simple point feature
  points <- sf::st_as_sf(inData,
                         coords = c("decimalLongitude", "decimalLatitude"),
                         na.fail = TRUE,
                         # Assign the CRS from the rnaturalearth map to the point inData
                         crs = sf::st_crs(countryMap)) %>%
    # Use a subset of columns
    dplyr::select(database_id, scientificName, species, family, subfamily, genus, specificEpithet, 
                  scientificNameAuthorship,
                  country, stateProvince, eventDate, institutionCode, recordNumber, catalogNumber,
                  dataSource, verbatim_scientificName, geometry)
    #Extract polygon information to points
    points_extract <- sf::st_intersects(points, countryMap) %>%
      # return a tibble with the index of each match or NA where there was no match
      dplyr::tibble(indexMatch = .) %>%
      # Convert to numeric
      dplyr::mutate(indexMatch = indexMatch %>% as.character() %>%
                      # deal with problems - Take the first number where two are provided
                      stringr::str_extract("[0-9]+") %>% 
                      # Remove zero to NA
                      stringr::str_replace("^[0]$", NA_character_) %>% as.numeric()) %>%
      dplyr::left_join(simplePoly,
                       by = "indexMatch") %>%
      # Add in the database_id
      dplyr::bind_cols(inData %>% sf::st_drop_geometry() %>% dplyr::select(!continent))
  }) })
    # Return the points
  return(points_extract)
  } # END jbd_intersection

      ###### b. buffered intersect ####
# Hijack st_intersection to allow it to be run in parallel
jbd_bufferedIntersection <- function(inData){
  suppressWarnings({ suppressMessages({
    inData <- inData %>% 
        # Use only complete lat and lon data
      tidyr::drop_na(decimalLongitude, decimalLatitude) %>%
        # Remove the previous column names from jbd_intersection
      dplyr::select(!tidyselect::any_of(c("iso_a2","iso_a3_eh","name","name_long","continent",
                                          "indexMatch")))
  # Turn inData into a simple point feature
  points <- sf::st_as_sf(inData,
                         coords = c("decimalLongitude", "decimalLatitude"),
                         na.fail = TRUE,
                         # Assign the CRS from the rnaturalearth map to the point inData
                         crs = sf::st_crs(countryMap)) %>%
    # Use a subset of columns
    dplyr::select(database_id, scientificName, species, family, subfamily, genus, specificEpithet, 
                  scientificNameAuthorship,
                  country, stateProvince, eventDate, institutionCode, recordNumber, catalogNumber,
                  dataSource, verbatim_scientificName, geometry) %>%
      # Buffer the points by the pointBuffer
    sf::st_buffer(dist = pointBuffer)
  
  #Extract polygon information to points
  points_extract <- sf::st_intersects(points, countryMap) %>%
    # return a tibble with the index of each match or NA where there was no match
    dplyr::tibble(indexMatch = . ) %>%
    # Convert to numeric
    dplyr::mutate(indexMatch = indexMatch %>% as.character() %>%
                    # deal with problems - Take the first number where two are provided
                    stringr::str_extract("[0-9]+") %>% 
                    # Remove zero to NA
                    stringr::str_replace("^[0]$", NA_character_) %>% as.numeric()) %>%
    dplyr::left_join(simplePoly,
                     by = "indexMatch") %>%
    # Add in the database_id
    dplyr::bind_cols(inData %>% sf::st_drop_geometry() %>% 
                       dplyr::select(!tidyselect::any_of("continent")))
  })  })
  # Return the points
  return(points_extract)
} # END jbd_intersection



##### 2.2 Extraction ####
###### a. exactCountry ####
writeLines(" - Extracting country data from points...")
points_extract <- data %>%
  # remove the existing iso_a3_eh column 
  dplyr::select(!tidyselect::any_of("iso_a3_eh")) %>%
  # Make a new column with the ordering of rows
  dplyr::mutate(BeeBDC_order = dplyr::row_number()) %>%
  # Group by the row number and step size
  dplyr::group_by(BeeBDC_group = ceiling(BeeBDC_order/stepSize)) %>%
  # Split the dataset up into a list by group
  dplyr::group_split(.keep = TRUE) %>%
  # Run the actual function
  parallel::mclapply(., jbd_intersection,
                     mc.cores = mc.cores
  ) %>%
  # Combine the lists of tibbles
  dplyr::bind_rows() %>%
    # Drop those occurrences that did not intersect with a country
  tidyr::drop_na(iso_a3_eh)

  
  if(!is.null(pointBuffer)){
  # Failed extractions
  points_failed <- data %>%
    dplyr::filter(!database_id %in% points_extract$database_id)
  
  writeLines(" - Buffering failed points by pointBuffer...")
  
  points_failed <- points_failed %>%
    # Make a new column with the ordering of rows
    dplyr::mutate(BeeBDC_order = dplyr::row_number()) %>%
    # Group by the row number and step size
    dplyr::group_by(BeeBDC_group = ceiling(BeeBDC_order/stepSize)) %>%
    # Split the dataset up into a list by group
    dplyr::group_split(.keep = TRUE) %>%
    # Run the actual function
    parallel::mclapply(., jbd_bufferedIntersection,
                       mc.cores = mc.cores
    ) %>%
    # Combine the lists of tibbles
    dplyr::bind_rows() %>%
    # Drop those occurrences that did not intersect with a country
    tidyr::drop_na(iso_a3_eh)
  
  if(nrow(points_failed) > 0){
  # Re-merge good with failed
  points_extract <- points_extract %>%
    sf::st_drop_geometry() %>%
      # remove buffer-matched occurrences
    dplyr::filter(!database_id %in% points_failed$database_id) %>%
      # replace these, but now matched
    dplyr::bind_rows(points_failed %>% sf::st_drop_geometry())
  }
  } # End if pointBuffer
  
  writeLines(" - Prepare the neighbouring country dataset...")
  ###### b. neighbouringCountries ####
    # Get a list of countries that share borders
  countriesBordering <- sf::st_intersects(countryMap, countryMap) %>%
    paste(., sep = ";")
    # Make a new tibble with these information
  neighbouringCountries <- dplyr::tibble(
    rowNum = 1:nrow(countryMap),
    country = countryMap$iso_a3_eh,
    neighbours = countriesBordering,
      # Modify the text in column
    neighboursText = mgsub::mgsub(string = neighbours,
                                  pattern = rowNum,
                                  replacement = country) %>%
      stringr::str_replace(string = .,
                           pattern = "c\\(", replacement = "") %>%
      stringr::str_replace(string = .,
                           pattern = "\\)", replacement = "") %>%
      stringr::str_replace(string = .,
                           pattern = ":", replacement = ", ")) 
  
    # Make a long-format tibble with neighbouring countries
  neighbouringCountries <- neighbouringCountries %>%
    tidyr::separate_rows(data = ., neighboursText,
                              sep = ",")  %>%
      # Remove country matching themselves
    dplyr::filter(!(country == neighboursText))
    # Remove extra spaces
  neighbouringCountries$neighboursText <- stringr::str_squish(neighbouringCountries$neighboursText)
    # Join the datasets together so that we can make a list of adjacent countries to match also
  neighbouringCountries <- checklist %>%
    dplyr::left_join(neighbouringCountries %>% 
                       dplyr::select(tidyselect::any_of(c("country", "neighboursText"))),
                     by = c("iso_a3_eh" = "country"),
                     multiple = "all", relationship = "many-to-many")
    
###### c. Sea points ####
    # Find the points that did not overlap with countries but that had coordinates
  seaPoints <- sf::st_as_sf(data %>% tidyr::drop_na(decimalLongitude, decimalLatitude),
                            coords = c("decimalLongitude", "decimalLatitude"),
                            na.fail = TRUE,
                            # Assign the CRS from the rnaturalearth map to the point data
                            crs = sf::st_crs(countryMap)) %>%
    # Use a subset of columns
    dplyr::select(database_id, scientificName, species, family, subfamily, genus, specificEpithet, 
                  scientificNameAuthorship,
                  country, stateProvince, eventDate, institutionCode, recordNumber, catalogNumber,
                  dataSource, verbatim_scientificName, geometry) %>% 
    sf::st_drop_geometry() %>%
    dplyr::filter(!database_id %in% points_extract$database_id) %>%
    dplyr::select(database_id)
  
    ##### 2.3 Compare ####
  writeLines(" - Compare points with the checklist...")
    # Get a smaller subset of the columns AND make a new columns with scientific name and country
  points_simple <- points_extract %>% 
    dplyr::select(database_id, iso_a3_eh, scientificName, country) %>%
    dplyr::mutate(SciCountry = stringr::str_c(scientificName, 
                                              iso_a3_eh, sep = "_")) %>% 
    # Remove grammar and caps from SciCountry
        dplyr::mutate(SciCountry = tolower(SciCountry) %>%
                          # Replace punctuation
                        stringr::str_replace_all("[\\(\\)\\,\\.\\-]", "") %>%
                        # Replace white spaces with underscores
                        stringr::str_replace_all(" ", "_")) %>%
    # Make the new column to match with full species name, (NO AUTHORHSIP YEAR), and country
    # Remove grammar and caps from SciCountry
    dplyr::mutate(SciCountry_noYear = tolower(SciCountry) %>%
                    # Replace numbers
                    stringr::str_replace_all("[0-9]", "")%>%
                    # replace double __
                    stringr::str_replace_all("__", "_"))
    
  
    ###### a. exactCountry ####
    # Do the same for the ascher checklist
  checklist_simple <- checklist %>%
      # Select subset
    dplyr::select(validName, 'iso_a3_eh') %>%
      # Harmonise column names
         dplyr::rename(iso_a3_eh = 'iso_a3_eh') %>%
      # Make the new column to match with full species name, authorship, and country
    dplyr::mutate(SciCountry = stringr::str_c(validName, iso_a3_eh, sep = "_"))%>% 
    # Remove grammar and caps from SciCountry
    dplyr::mutate(SciCountry = tolower(SciCountry) %>%
                    # Replace punctuation
                    stringr::str_replace_all("[\\(\\)\\,\\.\\-]", "") %>%
                    # Replace white spaces with underscores
                    stringr::str_replace_all(" ", "_")) %>%
  # Make the new column to match with full species name, (NO AUTHORHSIP YEAR), and country
    # Remove grammar and caps from SciCountry
    dplyr::mutate(SciCountry_noYear = tolower(SciCountry) %>%
                    # Replace numbers
                    stringr::str_replace_all("[0-9]", "") %>%
                    # replace double __
                    stringr::str_replace_all("__", "_"))
  
  
  
    # Make a new columns showing if that species is expected in that country
  points_match <- points_simple %>%
    #dplyr::filter(country == "United states") %>%
    #dplyr::mutate(exactMatch = dplyr::if_else(SciCountry %in% checklist_simple$SciCountry,
    #                                    TRUE, FALSE)) %>%
    dplyr::mutate(exactMatch_noYear = dplyr::if_else(SciCountry_noYear %in% checklist_simple$SciCountry_noYear,
                                              TRUE, FALSE)) %>%
    dplyr::left_join(dplyr::select(checklist_simple, SciCountry_noYear), 
                     by = "SciCountry_noYear",
                     multiple = "all", relationship = "many-to-many")

  
    ###### b. neighbouringCountries ####
  # Get a smaller subset of the data AND make a new columns with scientific name and country
  nchecklist_simple <- neighbouringCountries %>%
    # Select subset
    dplyr::select(validName, neighboursText) %>%
    # Make the new column to match with
    dplyr::mutate(SciCountry = stringr::str_c(validName, neighboursText, sep = "_")) %>%
    # Remove grammar and caps from SciCountry
    dplyr::mutate(SciCountry = tolower(SciCountry) %>%
    # Replace punctuation
    stringr::str_replace_all("[\\(\\)\\,\\.\\-]", "") %>%
    # Replace white spaces with underscores
    stringr::str_replace_all(" ", "_")) %>%
  # Make the new column to match with full species name, (NO AUTHORHSIP YEAR), and country
  # Remove grammar and caps from SciCountry
  dplyr::mutate(SciCountry_noYear = tolower(SciCountry) %>%
                  # Replace numbers
                  stringr::str_replace_all("[0-9]", "") %>%
                  # replace double __
                  stringr::str_replace_all("__", "_")) %>%
      # Get a unique set
    dplyr::distinct(SciCountry, .keep_all = TRUE)
  # Make a new columns showing if that species is expected in that country
  npoints_match <- points_simple %>%
    # dplyr::filter(country == "United states") %>%
    #dplyr::mutate(neighbourMatch = dplyr::if_else(SciCountry %in% nchecklist_simple$SciCountry,
    #                                          TRUE, FALSE)) %>%
    dplyr::mutate(neighbourMatch_noYear = dplyr::if_else(SciCountry_noYear %in% nchecklist_simple$SciCountry_noYear,
                                                     TRUE, FALSE)) 
  
  
  #### 3.0 Merge ####
  writeLines(" - Combining data...")
    # Merge both points_match datasets
  bpoints_match <- dplyr::tibble(points_match) %>%
      # Join the two datasets togehter keeping only neighbourMatch and assignmentCertainty from the 
      # neighbour-joined dataset
    dplyr::left_join(dplyr::select(npoints_match, c(database_id, neighbourMatch_noYear)),
                     by = "database_id",
                     multiple = "all", relationship = "many-to-many") %>%
      # Remove geometry column
    dplyr::select(!tidyselect::starts_with("geometry")) %>%
      # Combine exactMatch_noYear and neighbourMatch_noYear
    dplyr::mutate(matchType = dplyr::if_else(exactMatch_noYear == TRUE,
                                             "exact", dplyr::if_else(neighbourMatch_noYear == TRUE,
                                                                     "neighbour", 
                                                                     "noMatch"))) 
  
  #### 4.0 Output ####
    ##### 4.1 User output ####
  bpoints_match <- bpoints_match %>%
    # Select the columns to keep
    dplyr::select(database_id, iso_a3_eh, matchType) %>%
    dplyr::rename(
      countryMatch = matchType) 
      # Set flag for those that don't pass countryMatch
  
    ###### a. keepAC == TRUE ####
  if(keepAdjacentCountry == TRUE){
    bpoints_match <- bpoints_match %>%
      dplyr::mutate(
        .countryOutlier = dplyr::if_else(countryMatch != "exact" & countryMatch != "neighbour",
                         FALSE, TRUE)
      ) # END mutate
  } # END TRUE
  
    ###### b. keepAC == FALSE ####
  if(keepAdjacentCountry == FALSE){
    bpoints_match <- bpoints_match %>%
      dplyr::mutate(
        .countryOutlier = dplyr::if_else(countryMatch != "exact",
                                         FALSE, TRUE)
      ) # END mutate
  }# END FALSE
  
    # Keep only entirely unique records
  bpoints_match <- bpoints_match %>%
    dplyr::distinct() 
  
  ###### c. distinct buffer ####
    # For those buffered records that might have overlapped with >1 country, select the unfiltered one, if it exists.
  if(!is.null(pointBuffer)){
    writeLines(" - Sorting and removing potentially duplicated buffered points...")
    bpoints_match <- bpoints_match %>%
      dplyr::group_by(database_id) %>%
      dplyr::arrange(desc(.countryOutlier), .by_group = TRUE) %>%
      dplyr::distinct(database_id, .keep_all = TRUE)
    }
  
  
      # Merge with original dataset
    output <- data %>%
      dplyr::left_join(bpoints_match, by = "database_id",
                       multiple = "all", relationship = "many-to-many") %>%
        # Add in .sea usign the seaPoints
      dplyr::mutate(.sea = dplyr::if_else(database_id %in% seaPoints$database_id,
                                          FALSE, TRUE))
    
    
    writeLines(paste0(
      " - Finished. \n",
      "We have matched ", 
      format(sum(bpoints_match$countryMatch == "exact", na.rm = TRUE), big.mark = ","),
      " records to their exact country and ", 
      format(sum(bpoints_match$countryMatch == "neighbour", na.rm = TRUE), big.mark = ","), 
      " to an adjacent country\n", 
      "We failed to match ",
      format(sum(bpoints_match$countryMatch == "noMatch", na.rm = TRUE), big.mark = ","), 
      " occurrences to any 'exact' or 'neighbouring' country.\n",
      "There are ",
      format(sum(is.na(output$.countryOutlier)), big.mark = ","), 
      " 'NA' occurrences for the .countryOutlier column.\n"
    ))
    

    
    # return message
    message(paste("countryOutlieRs:\nFlagged", 
                  format(sum(output$.countryOutlier == FALSE, na.rm = TRUE), big.mark = ","), 
                  " for country outlier and flagged ",
                  format(sum(output$.sea == FALSE, na.rm = TRUE), big.mark = ","), 
                  " for in the .sea",
                  "records.\nThree columns were added to the database:\n 1. ",
                  "The '.countryOutlier' column was added which is a filtering column. \n 2. ",
                  "The 'countryMatch' columns indicates exact, neighbour, or noMatch. \n",
                  "3. The '.sea' column was added as a filtering column for points in the ocean.",
                  " The '.sea' column includes the user input buffer in its calculation."),
            sep = "")
    
    # rm(countryMap, points, points_extract, countriesBordering, neighbouringCountries, points_simple,
    #    checklist_simple, points_match, nchecklist_simple, npoints_match, bpoints_match)
  # Return file
    endTime <- Sys.time()
    # Time output
    message(paste(
      " - Completed in ", 
      round(difftime(endTime, startTime), digits = 2 )," ",
      units(round(endTime - startTime, digits = 2)),
      sep = ""))
return(output)

} # END countryOutlieRs

requireNamespace("dplyr")
requireNamespace("readr")



# If rnaturalearthdata is present, run tests
if(requireNamespace("rnaturalearthdata")){

# Load in the test checklist data
system.file("extdata", "testChecklist.rda", package="BeeBDC") |>
  load()
# Load in the test dataset
beesFlagged <- BeeBDC::beesFlagged

  # Input a potentially difficult location on the map to test into an NA lat/lon slot
beesFlagged$decimalLatitude[[1]] <-  31.887646484374983
beesFlagged$decimalLongitude[[1]] <- 78.719726562500085 
beesFlagged$decimalLatitude[[2]] <-  78.719726562500085
beesFlagged$decimalLongitude[[2]] <- 31.887646484374983 

testOut <- BeeBDC::countryOutlieRs(
    # Speed up operation by providing only the relevant entries in the testChecklist
  checklist = testChecklist,
  data = beesFlagged %>% 
    dplyr::select(!tidyselect::any_of(c("countryMatch", ".countryOutlier","iso_a3"))),
  keepAdjacentCountry = FALSE,
  # running without a larger buffer to speed up tests
  pointBuffer = NULL,
  # Scale of map to return, one of 110, 50, 10 OR 'small', 'medium', 'large'
  # Smaller numbers will result in much longer calculation times. 
  # We have not attempted a scale of 10.
  scale = 50)


# Test the number of expected TRUE and FALSE columns and then test the output format (data frames and
# tibbles are a special case of lists)
testthat::test_that("countryOutlieRs results TRUE/passed", {
  testthat::expect_equal(sum(testOut$.countryOutlier == TRUE, na.rm = TRUE), 72)
})
testthat::test_that("countryOutlieRs results FALSE/failed", {
  testthat::expect_equal(sum(testOut$.countryOutlier == FALSE, na.rm = TRUE), 5)
})
testthat::test_that("countryOutlieRs results NA/could not assess", {
  testthat::expect_equal(sum(is.na(testOut$.countryOutlier)), 23)
})

  # Test format
testthat::test_that("countryOutlieRs expected class", {
  testthat::expect_type(testOut, "list")
})
testthat::test_that("countryOutlieRs results NA/could not assess", {
  testthat::expect_true( all(stringr::str_detect(attributes(testOut)$class, 
                                             c("tbl_df","tbl","data.frame"))))
})

# Expected number of columns matches input data
testthat::test_that("countryOutlieRs number of columns", {
    # Remove the three columns that will be updated by this function and then test that they are
      # added back in.
  testthat::expect_equal(ncol(beesFlagged %>% 
                                dplyr::select(!tidyselect::any_of(c("countryMatch", 
                                                                    ".countryOutlier",
                                                                    "iso_a3")))) + 3,
                         ncol(testOut))
})

} # END if require

# Sorry everyone, I just want to cut down on tibble as a dependency. These are simple enough
# internal functions though.
#### 1.0 column_to_rownames ####
  # replaces tibble's column_to_rownames
column_to_rownames_internal <- function(.data, var = "rowname") {
  stopifnot(is.data.frame(.data))
  
  .data <- as.data.frame(.data)
  rownames(.data) <- .data[[var]]
  .data[[var]] <- NULL
  .data
}



  #### 2.0 rownames to column ####
  # replaces tibble's rownames_to_column and remove_rownames
rownames_to_column_int <- function(.data, var = "rowname") {
  # rename, because .data has special semantics in tidy evaluation
  df <- .data
  
  stopifnot(is.data.frame(df))
  
  new_df <- df %>%
    dplyr::mutate(var_in = rownames(df))
  
  
  names(new_df)[names(new_df) == "var_in"] <- var
    
  remove_rownames_int(new_df)
}

remove_rownames_int <- function(.data) {
  stopifnot(is.data.frame(.data))
  rownames(.data) <- NULL
  .data
}

##### 2.1 repoFinder ####
#' Find GBIF, ALA, iDigBio, and SCAN files in a directory
#'
#' @param path A directory as character. The path within which to recursively look for GBIF, ALA, 
#' iDigBio, and SCAN files.
#'
#' @return Returns a list of directories to each of the above data downloads
#' @export
#' 
#' @importFrom dplyr %>%
#'
#' @examples
#' \dontrun{
#' # Where DataPath is made by [BeeBDC::dirMaker()]
#' BeeBDC::repoFinder(path = DataPath)
#' }
repoFinder <- function(path){
  . <- NULL
  requireNamespace("dplyr")
  # Find ALL occurrence file downloads contained within the HomePath and return their location and 
  # their information
  AllOccLocs <- file.info(list.files(path, full.names = T, pattern = "occurrence(s)?(_raw)?\\.|^data.csv",
                                     recursive = TRUE))
  # Split out each data source's occurrence files and file paths based on their slight differences
  occ_paths <- list(
    # ALA paths
    rownames(AllOccLocs) %>% grep("/data.csv", ., value = TRUE),
    # GBIF paths
    rownames(AllOccLocs) %>% grep("/occurrence.txt", ., value = TRUE),
    # iDigBio paths
    rownames(AllOccLocs) %>% grep("/occurrence_raw.csv", ., value = TRUE),
    # SCAN paths
    rownames(AllOccLocs) %>% grep("/occurrences.csv", ., value = TRUE))
  # Return those paths
  names(occ_paths) <- c("ALA_data", "GBIF_data", "iDigBio_data", "SCAN_data")
  return(occ_paths)
}

requireNamespace("readr")
requireNamespace("BeeBDC")


# make some data files for repoFinder to search and save the files 
## ALA data
ALA_data <- dplyr::tribble(
              ~scientificName,  ~family, ~subfamily, ~genus, ~subgenus, ~subspecies, ~species, ~specificEpithet, ~infraspecificEpithet, ~acceptedNameUsage, ~taxonRank, ~scientificNameAuthorship, ~identificationQualifier, ~higherClassification, ~identificationReferences, ~typeStatus, ~previousIdentifications, ~identifiedBy, ~dateIdentified, ~decimalLatitude, ~decimalLongitude, ~verbatimLatitude, ~verbatimLongitude,       ~stateProvince,    ~country, ~continent,                        ~locality, ~island, ~county, ~municipality, ~countryCode, ~`dcterms:license`,             ~eventDate, ~eventTime, ~day, ~month, ~year,       ~basisOfRecord, ~`dcterms:type`, ~occurrenceStatus, ~recordNumber, ~recordedBy, ~eventID, ~samplingProtocol, ~samplingEffort, ~individualCount, ~organismQuantity, ~coordinatePrecision, ~coordinateUncertaintyInMeters, ~spatiallyValid, ~catalogNumber, ~datasetID, ~institutionCode, ~datasetName, ~otherCatalogNumbers,                                          ~occurrenceID, ~collectionID, ~verbatimEventDate, ~associatedTaxa, ~associatedOrganisms, ~fieldNotes,   ~sex, ~rights, ~rightsHolder, ~`dcterms:accessRights`, ~associatedReferences, ~`dcterms:bibliographicCitation`, ~references, ~informationWithheld, ~isDuplicateOf,                                                                                                                                                                                                                                      ~assertions,        ~occurrenceYear, ~duplicateStatus, ~associatedOccurrences, ~locationRemarks,
                     "APIDAE", "Apidae",         NA,     NA,        NA,          NA,       NA,      "australis",                    NA,                 NA,   "family",                        NA,                       NA,                    NA,                        NA,          NA,                       NA,            NA,              NA,            -22.7,             133.4,                NA,                 NA, "Northern Territory", "Australia",         NA,                 "60 k S Ti Tree",      NA,      NA,            NA,         "AU",            "CC-BY", "1996-10-02T14:00:00Z",         NA,   3L,    10L, 1996L, "PRESERVED_SPECIMEN",              NA,         "PRESENT",            NA,          NA,       NA,                NA,              NA,               NA,                NA,                   NA,                             NA,            TRUE,      "T157061",         NA,             "QM",           NA,             2089363L, "urn:lsid:ozcam.taxonomy.org.au:QM.Entomology:T157061",            NA,                 NA,              NA,                   NA,          NA,     NA,      NA,            NA,                      NA,                    NA,                               NA,          NA,                   NA,             NA, "COORDINATE_UNCERTAINTY_METERS_INVALID | MISSING_GEOREFERENCEDBY | MISSING_GEOREFERENCEPROTOCOL | MISSING_GEOREFERENCESOURCES | MISSING_GEOREFERENCEVERIFICATIONSTATUS | MISSING_GEOREFERENCE_DATE | MISSING_TAXONRANK | TAXON_MATCH_HIGHERRANK", "1995-12-31T13:00:00Z",               NA,                     NA,               NA,
                     "APIDAE", "Apidae",         NA,     NA,        NA,          NA,       NA,     "essingtoni",                    NA,                 NA,   "family",                        NA,                       NA,                    NA,                        NA,          NA,                       NA,            NA,              NA,            -15.5,             128.7,                NA,                 NA,  "Western Australia", "Australia",         NA, "Middle Springs 15k N Kunanurra",      NA,      NA,            NA,         "AU",            "CC-BY", "1985-09-11T14:00:00Z",         NA,  12L,     9L, 1985L, "PRESERVED_SPECIMEN",              NA,         "PRESENT",            NA,          NA,       NA,                NA,              NA,               NA,                NA,                   NA,                             NA,            TRUE,      "T157059",         NA,             "QM",           NA,             2089361L, "urn:lsid:ozcam.taxonomy.org.au:QM.Entomology:T157059",            NA,                 NA,              NA,                   NA,          NA, "MALE",      NA,            NA,                      NA,                    NA,                               NA,          NA,                   NA,             NA, "COORDINATE_UNCERTAINTY_METERS_INVALID | MISSING_GEOREFERENCEDBY | MISSING_GEOREFERENCEPROTOCOL | MISSING_GEOREFERENCESOURCES | MISSING_GEOREFERENCEVERIFICATIONSTATUS | MISSING_GEOREFERENCE_DATE | MISSING_TAXONRANK | TAXON_MATCH_HIGHERRANK", "1984-12-31T13:00:00Z",               NA,                     NA,               NA,
                     "APIDAE", "Apidae",         NA,     NA,        NA,          NA,       NA,        "cassiae",                    NA,                 NA,   "family",                        NA,                       NA,                    NA,                        NA,          NA,                       NA,            NA,              NA,            -23.7,             149.9,                NA,                 NA,         "Queensland", "Australia",         NA,                       "Duaringa",      NA,      NA,            NA,         "AU",            "CC-BY", "1987-09-06T14:00:00Z",         NA,   7L,     9L, 1987L, "PRESERVED_SPECIMEN",              NA,         "PRESENT",            NA,          NA,       NA,                NA,              NA,               NA,                NA,                   NA,                             NA,            TRUE,      "T157064",         NA,             "QM",           NA,             2089366L, "urn:lsid:ozcam.taxonomy.org.au:QM.Entomology:T157064",            NA,                 NA,              NA,                   NA,          NA, "MALE",      NA,            NA,                      NA,                    NA,                               NA,          NA,                   NA,             NA, "COORDINATE_UNCERTAINTY_METERS_INVALID | MISSING_GEOREFERENCEDBY | MISSING_GEOREFERENCEPROTOCOL | MISSING_GEOREFERENCESOURCES | MISSING_GEOREFERENCEVERIFICATIONSTATUS | MISSING_GEOREFERENCE_DATE | MISSING_TAXONRANK | TAXON_MATCH_HIGHERRANK", "1986-12-31T13:00:00Z",               NA,                     NA,               NA,
                     "APIDAE", "Apidae",         NA,     NA,        NA,          NA,       NA,      "australis",                    NA,                 NA,   "family",                        NA,                       NA,                    NA,                        NA,          NA,                       NA,            NA,              NA,            -23.7,             149.9,                NA,                 NA,         "Queensland", "Australia",         NA,                       "Duaringa",      NA,      NA,            NA,         "AU",            "CC-BY", "1996-10-12T14:00:00Z",         NA,  13L,    10L, 1996L, "PRESERVED_SPECIMEN",              NA,         "PRESENT",            NA,          NA,       NA,                NA,              NA,               NA,                NA,                   NA,                             NA,            TRUE,      "T157060",         NA,             "QM",           NA,             2089362L, "urn:lsid:ozcam.taxonomy.org.au:QM.Entomology:T157060",            NA,                 NA,              NA,                   NA,          NA,     NA,      NA,            NA,                      NA,                    NA,                               NA,          NA,                   NA,             NA, "COORDINATE_UNCERTAINTY_METERS_INVALID | MISSING_GEOREFERENCEDBY | MISSING_GEOREFERENCEPROTOCOL | MISSING_GEOREFERENCESOURCES | MISSING_GEOREFERENCEVERIFICATIONSTATUS | MISSING_GEOREFERENCE_DATE | MISSING_TAXONRANK | TAXON_MATCH_HIGHERRANK", "1995-12-31T13:00:00Z",               NA,                     NA,               NA
              )

write.csv(ALA_data, paste0(tempdir(), "/data.csv"), row.names = FALSE)


## GBIF data - this is data from not GBIF because GBIF data refused to be copy-pasted
GBIF_data <- dplyr::tribble(
  ~studyID,           ~siteID, ~year,     ~date,                           ~animalID, ~abundance, ~abundanceMethod, ~samplingMethod, ~numCensus, ~samplingIntensity,                                                                                                                                                                                   ~censusType,     ~fieldDist,  ~flowering, ~decimalLatitude, ~decimalLongitude,                ~studyLocation, ~habitatType,                ~siteDescription,
  "Ball01", "Bastrop Gardens", 2013L, "2013/5/22",             "Agapostemon angelicus",         1L,      "Abundance",      "Pan Trap",         1L,                24L, "50 pan traps (painted blue, yellow or left white as in LeBuhn et al) placed in x formation 1 m apart from center of the 50 m2 plot, vegetation measured as described for blue vane trapped.", "not measured", "flowering",        30.157397,        -97.491703, "Bastrop, Bastrop County, TX",           NA, "Agriculture. Community garden",
  "Ball01", "Bastrop Gardens", 2013L, "2013/5/22",               "Agapostemon texanus",         1L,      "Abundance",      "Pan Trap",         1L,                24L, "50 pan traps (painted blue, yellow or left white as in LeBuhn et al) placed in x formation 1 m apart from center of the 50 m2 plot, vegetation measured as described for blue vane trapped.", "not measured", "flowering",        30.157397,        -97.491703, "Bastrop, Bastrop County, TX",           NA, "Agriculture. Community garden",
  "Ball01", "Bastrop Gardens", 2013L, "2013/5/22",            "Anthophora californica",         1L,      "Abundance",      "Pan Trap",         1L,                24L, "50 pan traps (painted blue, yellow or left white as in LeBuhn et al) placed in x formation 1 m apart from center of the 50 m2 plot, vegetation measured as described for blue vane trapped.", "not measured", "flowering",        30.157397,        -97.491703, "Bastrop, Bastrop County, TX",           NA, "Agriculture. Community garden",
  "Ball01", "Bastrop Gardens", 2013L, "2013/5/22",                "Ceratina shinnersi",         1L,      "Abundance",      "Pan Trap",         1L,                24L, "50 pan traps (painted blue, yellow or left white as in LeBuhn et al) placed in x formation 1 m apart from center of the 50 m2 plot, vegetation measured as described for blue vane trapped.", "not measured", "flowering",        30.157397,        -97.491703, "Bastrop, Bastrop County, TX",           NA, "Agriculture. Community garden",
  "Ball01", "Bastrop Gardens", 2013L, "2013/5/22",                  "Ceratina strenua",         1L,      "Abundance",      "Pan Trap",         1L,                24L, "50 pan traps (painted blue, yellow or left white as in LeBuhn et al) placed in x formation 1 m apart from center of the 50 m2 plot, vegetation measured as described for blue vane trapped.", "not measured", "flowering",        30.157397,        -97.491703, "Bastrop, Bastrop County, TX",           NA, "Agriculture. Community garden",
  "Ball01", "Bastrop Gardens", 2013L, "2013/5/22", "Lasioglossum (Dialictus) sp.TX-14",        12L,      "Abundance",      "Pan Trap",         1L,                24L, "50 pan traps (painted blue, yellow or left white as in LeBuhn et al) placed in x formation 1 m apart from center of the 50 m2 plot, vegetation measured as described for blue vane trapped.", "not measured", "flowering",        30.157397,        -97.491703, "Bastrop, Bastrop County, TX",           NA, "Agriculture. Community garden"
) 

write.table(GBIF_data, paste0(tempdir(), "/occurrence.txt"), row.names = FALSE)


## iDigBio data - this is data from not iDigBio because iDigBio data refused to be copy-pasted
iDigBio_data <- dplyr::tribble(
  ~studyID,           ~siteID, ~year,     ~date,                           ~animalID, ~abundance, ~abundanceMethod, ~samplingMethod, ~numCensus, ~samplingIntensity,                                                                                                                                                                                   ~censusType,     ~fieldDist,  ~flowering, ~decimalLatitude, ~decimalLongitude,                ~studyLocation, ~habitatType,                ~siteDescription,
  "Ball01", "Bastrop Gardens", 2013L, "2013/5/22",             "Agapostemon angelicus",         1L,      "Abundance",      "Pan Trap",         1L,                24L, "50 pan traps (painted blue, yellow or left white as in LeBuhn et al) placed in x formation 1 m apart from center of the 50 m2 plot, vegetation measured as described for blue vane trapped.", "not measured", "flowering",        30.157397,        -97.491703, "Bastrop, Bastrop County, TX",           NA, "Agriculture. Community garden",
  "Ball01", "Bastrop Gardens", 2013L, "2013/5/22",               "Agapostemon texanus",         1L,      "Abundance",      "Pan Trap",         1L,                24L, "50 pan traps (painted blue, yellow or left white as in LeBuhn et al) placed in x formation 1 m apart from center of the 50 m2 plot, vegetation measured as described for blue vane trapped.", "not measured", "flowering",        30.157397,        -97.491703, "Bastrop, Bastrop County, TX",           NA, "Agriculture. Community garden",
  "Ball01", "Bastrop Gardens", 2013L, "2013/5/22",            "Anthophora californica",         1L,      "Abundance",      "Pan Trap",         1L,                24L, "50 pan traps (painted blue, yellow or left white as in LeBuhn et al) placed in x formation 1 m apart from center of the 50 m2 plot, vegetation measured as described for blue vane trapped.", "not measured", "flowering",        30.157397,        -97.491703, "Bastrop, Bastrop County, TX",           NA, "Agriculture. Community garden",
  "Ball01", "Bastrop Gardens", 2013L, "2013/5/22",                "Ceratina shinnersi",         1L,      "Abundance",      "Pan Trap",         1L,                24L, "50 pan traps (painted blue, yellow or left white as in LeBuhn et al) placed in x formation 1 m apart from center of the 50 m2 plot, vegetation measured as described for blue vane trapped.", "not measured", "flowering",        30.157397,        -97.491703, "Bastrop, Bastrop County, TX",           NA, "Agriculture. Community garden",
  "Ball01", "Bastrop Gardens", 2013L, "2013/5/22",                  "Ceratina strenua",         1L,      "Abundance",      "Pan Trap",         1L,                24L, "50 pan traps (painted blue, yellow or left white as in LeBuhn et al) placed in x formation 1 m apart from center of the 50 m2 plot, vegetation measured as described for blue vane trapped.", "not measured", "flowering",        30.157397,        -97.491703, "Bastrop, Bastrop County, TX",           NA, "Agriculture. Community garden",
  "Ball01", "Bastrop Gardens", 2013L, "2013/5/22", "Lasioglossum (Dialictus) sp.TX-14",        12L,      "Abundance",      "Pan Trap",         1L,                24L, "50 pan traps (painted blue, yellow or left white as in LeBuhn et al) placed in x formation 1 m apart from center of the 50 m2 plot, vegetation measured as described for blue vane trapped.", "not measured", "flowering",        30.157397,        -97.491703, "Bastrop, Bastrop County, TX",           NA, "Agriculture. Community garden"
) 

write.csv(iDigBio_data, paste0(tempdir(), "/occurrence_raw.csv"), row.names = FALSE)


## SCAN data
SCAN_data <- dplyr::tribble(
                     ~id, ~institutionCode, ~collectionCode,         ~ownerInstitutionCode,                          ~collectionID,      ~basisOfRecord,                  ~occurrenceID,      ~catalogNumber, ~otherCatalogNumbers,   ~kingdom,      ~phylum,    ~class,        ~order,  ~family,  ~scientificName, ~taxonID, ~scientificNameAuthorship, ~genus, ~specificEpithet, ~taxonRank, ~infraspecificEpithet,  ~identifiedBy, ~dateIdentified, ~identificationReferences, ~identificationRemarks, ~taxonRemarks, ~identificationQualifier, ~typeStatus,    ~recordedBy, ~recordNumber, ~eventDate, ~year, ~month, ~day, ~startDayOfYear, ~endDayOfYear,        ~verbatimEventDate, ~occurrenceRemarks, ~habitat, ~fieldNumber, ~informationWithheld, ~dataGeneralizations, ~dynamicProperties, ~associatedTaxa, ~associatedOccurrences, ~reproductiveCondition, ~establishmentMeans, ~lifeStage,     ~sex, ~individualCount, ~samplingProtocol, ~samplingEffort, ~preparations,        ~country, ~stateProvince, ~county, ~municipality,  ~locality,                  ~locationRemarks, ~decimalLatitude, ~decimalLongitude, ~geodeticDatum, ~coordinateUncertaintyInMeters, ~verbatimCoordinates, ~georeferencedBy, ~georeferenceProtocol, ~georeferenceSources, ~georeferenceVerificationStatus, ~georeferenceRemarks, ~minimumElevationInMeters, ~maximumElevationInMeters, ~minimumDepthInMeters, ~maximumDepthInMeters, ~verbatimDepth, ~verbatimElevation, ~disposition, ~language, ~recordEnteredBy,       ~modified,                                             ~rights,                                     ~rightsHolder, ~accessRights,                                       ~recordId,                                                                        ~references,
               13775122L,            "CAS",        "ANTWEB",        "UCDC, Davis, CA, USA", "dc646193-9811-40ce-872a-f005f7dc059a", "PreservedSpecimen",     "CAS:ANTWEB:casent0106100",     "casent0106100",                   NA, "Animalia", "Arthropoda", "Insecta", "Hymenoptera", "apidae", "apis mellifera",  235783L,          "Linnaeus, 1758", "Apis",      "mellifera",         NA,                    NA,             NA,              NA,                        NA,                     NA,            NA,                       NA,          NA,    "P.S. Ward",            NA,  "6/28/05", 2005L,     6L,  28L,            179L,            NA, "28 Jun 2005/29 Jun 2005",                 NA,       NA,   "PSW15443",                   NA,                   NA,                 NA,              NA,                     NA,                     NA,                  NA,         NA, "worker",               NA,                NA,              NA,         "pin", "United States",   "California",      NA,            NA,    "Davis", "coordinates obtained from Label",            38.54,        -121.75667,             NA,                             NA,                   NA,               NA,                    NA,                   NA,                              NA,                   NA,                       15L,                        NA,                    NA,                    NA,             NA,                 NA,           NA,        NA,               NA, "4/30/21 14:36", "http://creativecommons.org/publicdomain/zero/1.0/", "The California Academy of Sciences - AntWeb.org",            NA, "urn:uuid:46a46727-6535-4e70-88e7-a42c98f806ed", "https://scan-bugs.org:443/portal/collections/individual/index.php?occid=13775122",
               13775123L,            "CAS",        "ANTWEB",        "UCDC, Davis, CA, USA", "dc646193-9811-40ce-872a-f005f7dc059a", "PreservedSpecimen", "CAS:ANTWEB:casent0106100-d01", "casent0106100-d01",                   NA, "Animalia", "Arthropoda", "Insecta", "Hymenoptera", "apidae", "apis mellifera",  235783L,          "Linnaeus, 1758", "Apis",      "mellifera",         NA,                    NA,             NA,              NA,                        NA,                     NA,            NA,                       NA,          NA,    "P.S. Ward",            NA,  "6/28/05", 2005L,     6L,  28L,            179L,            NA, "28 Jun 2005/29 Jun 2005",                 NA,       NA,   "PSW15443",                   NA,                   NA,                 NA,              NA,                     NA,                     NA,                  NA,         NA, "worker",               NA,                NA,              NA,   "100% EtOH", "United States",   "California",      NA,            NA,    "Davis", "coordinates obtained from Label",            38.54,        -121.75667,             NA,                             NA,                   NA,               NA,                    NA,                   NA,                              NA,                   NA,                       15L,                        NA,                    NA,                    NA,             NA,                 NA,           NA,        NA,               NA, "4/30/21 14:36", "http://creativecommons.org/publicdomain/zero/1.0/", "The California Academy of Sciences - AntWeb.org",            NA, "urn:uuid:d6ff3ddb-4695-4aaf-ab89-c251ab2fc7e6", "https://scan-bugs.org:443/portal/collections/individual/index.php?occid=13775123",
               13775124L,            "CAS",        "ANTWEB",        "UCDC, Davis, CA, USA", "dc646193-9811-40ce-872a-f005f7dc059a", "PreservedSpecimen", "CAS:ANTWEB:casent0106100-d11", "casent0106100-d11",                   NA, "Animalia", "Arthropoda", "Insecta", "Hymenoptera", "apidae", "apis mellifera",  235783L,          "Linnaeus, 1758", "Apis",      "mellifera",         NA,                    NA,             NA,              NA,                        NA,                     NA,            NA,                       NA,          NA,    "P.S. Ward",            NA,  "6/28/05", 2005L,     6L,  28L,            179L,            NA, "28 Jun 2005/29 Jun 2005",                 NA,       NA,   "PSW15443",                   NA,                   NA,                 NA,              NA,                     NA,                     NA,                  NA,         NA, "worker",               NA,                NA,              NA,   "100% EtOH", "United States",   "California",      NA,            NA,    "Davis", "coordinates obtained from Label",            38.54,        -121.75667,             NA,                             NA,                   NA,               NA,                    NA,                   NA,                              NA,                   NA,                       15L,                        NA,                    NA,                    NA,             NA,                 NA,           NA,        NA,               NA, "4/30/21 14:36", "http://creativecommons.org/publicdomain/zero/1.0/", "The California Academy of Sciences - AntWeb.org",            NA, "urn:uuid:4504a6e6-ab1c-4de7-ac43-a3588a63ae90", "https://scan-bugs.org:443/portal/collections/individual/index.php?occid=13775124",
               13852774L,            "CAS",        "ANTWEB", "CAS, San Francisco, CA, USA", "dc646193-9811-40ce-872a-f005f7dc059a", "PreservedSpecimen",     "CAS:ANTWEB:casent0249086",     "casent0249086",                   NA, "Animalia", "Arthropoda", "Insecta", "Hymenoptera", "apidae", "apis mellifera",  235783L,          "Linnaeus, 1758", "Apis",      "mellifera",         NA,                    NA, "C. Sch√∂ning",              NA,                        NA,                     NA,            NA,                       NA,          NA, "C. Sch√∂ning",            NA,   "9/1/09", 2009L,     9L,   1L,            244L,            NA,                "1-Sep-09",                 NA,       NA,     "N109-9",                   NA,                   NA,                 NA,              NA,                     NA,                     NA,                  NA,         NA, "worker",               NA,                NA,              NA,         "pin",         "Kenya",             NA,      NA,            NA, "Chogoria",                                NA,         -0.20222,          37.50083,             NA,                             NA,                   NA,               NA,                    NA,                   NA,                              NA,                   NA,                     2277L,                        NA,                    NA,                    NA,             NA,                 NA,           NA,        NA,               NA, "4/30/21 15:00", "http://creativecommons.org/publicdomain/zero/1.0/", "The California Academy of Sciences - AntWeb.org",            NA, "urn:uuid:964139d3-fb07-4a57-9b7c-e9a5830f4c82", "https://scan-bugs.org:443/portal/collections/individual/index.php?occid=13852774"
               )

write.csv(SCAN_data, paste0(tempdir(), "/occurrences.csv"), row.names = FALSE)



# path to the test files goes here - load the thing, run the function
testOut <- BeeBDC::repoFinder(path = paste0(tempdir()))


# get variable type for each item in list
ALAout <- testOut[[1]]
GBIFout <- testOut[[2]]
iDigBioout <- testOut[[3]]
SCANout <- testOut[[4]]


# expected results - should all be present and all character values
testthat::test_that("repoFinder ALA expected", {
  testthat::expect_type(ALAout, "character")
})

testthat::test_that("repoFinder GBIF expected", {
  testthat::expect_type(GBIFout, "character")
})

testthat::test_that("repoFinder iDigBio expected", {
  testthat::expect_type(iDigBioout, "character")
})

testthat::test_that("repoFinder SCAN expected", {
  testthat::expect_type(SCANout, "character")
})

testthat::test_that("repoFinder expected class", {
  testthat::expect_type(testOut, "list")
})

  # This function was written by James Dorey to remove duplicates from the combined 
  # source2 and source1 - user input, but originally Orr 2021 and Discover Life
    # synonym lists. For questions, please email jbdorey[at]me.com
# This function was started on 17th May 2022 and last updated 17th May 2022
#' @importFrom dplyr %>%
#' @importFrom dplyr row_number

taxoDuplicator <- function(
    SynList = NULL,
    source1 = "DiscoverLife",
    source2 = "Orr_et_al_2021_CurrBiol"){
  # locally bind variables to the function
  validName <- accid <- id <- flags <- taxonomic_status <- canonical_withFlags <- canonical <- NULL
  
    # Load required packages
  requireNamespace("dplyr")
  
  #### 0.0 Prep ####
  ##### 0.1 Remove existing flags ####
  writeLines("Removing previous flags generated with this function")
    # Remove the xisitng flags generated from this function
  SynList <- SynList %>%
    dplyr::mutate(flags = stringr::str_remove_all(flags, "non-ambiguous can_wFlags") %>%
                    stringr::str_remove_all("non-ambiguous canonical") %>%
                    stringr::str_remove_all("ambiguous canonical") %>%
                    stringr::str_remove_all("ambiguous validName") %>%
                    stringr::str_remove_all("ambiguous can_wFlags") %>%
                    stringr::str_remove_all("ambiguous can_wFlags") %>%
                    stringr::str_remove_all("^, $") %>%
                    stringr::str_replace(", , ", ", ") )

  ##### 0.2 Find duplciates ####
  # Look for duplicated names in the DiscoverLife subset of data
  duplicates <- SynList %>% 
    #dplyr::filter(source == "DiscoverLife") %>%
    dplyr::group_by(validName) %>%
    dplyr::filter(dplyr::n() > 1)
  # User output
  writeLines(paste(" - ", format(nrow(duplicates), big.mark = ","),
                                 " duplicates found in the data.", sep = ""))
  
    # Build subsetted datasets to examine 
  S1accepted <- duplicates %>% dplyr::filter(accid == 0 & source %in% source1)
  S2Accepted <- duplicates %>% dplyr::filter(accid == 0 & source %in% source2)
  S1synonyms <- duplicates %>% dplyr::filter(accid != 0 & source %in% source1)
  S2synonyms <- duplicates %>% dplyr::filter(accid != 0 & source %in% source2)
  
  #### 1.0 S1_S2 #### 
    ##### 1.1 acc. names ####
    # Find all duplicated valid names that occur in the source2 list and the source1 list. 
    # This looks  
      # like All of them! These will later be REMOVED.
  S2Acc2remove <- S2Accepted %>%
    dplyr::filter(validName %in% S1accepted$validName)
  # Do any of the accids in the full list match these names' ids?
  S1IDmatch <- S2Acc2remove %>%
    dplyr::filter(id %in% SynList$accid)
  # Stop here becuase I have no matches, but this might be important down the track if someone 
    # finds them!
  if(nrow(S1IDmatch) > 0 ){
    return(S1IDmatch)
    stop(paste(" - That's odd! There is an S2 accepted name that is referred to by another name.",
               "This hasn't happened before, but you'll need to sort it out, chump!", "\n",
               "I have returned the list of offending names."))
  }
  # For now, because these are all duplicates, I will not return these data.
  
    ##### 1.2 synonyms ####
    # Find all duplicated SYNONYMS names that occur in the source2 list and the source1 list. 
  S2DupeSyns <- S2synonyms %>%
    dplyr::filter(validName %in% S1synonyms$validName)
    # Do any of the accids in the full list match these names' ids?
  S2IDmatch <- S2DupeSyns %>%
    dplyr::filter(id %in% SynList$accid)
    # Stop here becuase I have no matches, but this might be important down the track if someone 
    # finds them!
  if(nrow(S2IDmatch) > 0 ){
    return(S2IDmatch)
    stop(paste(" - That's odd! There is an S2 synonym that is referred to as an accepted name.",
               "This hasn't happened before, but you'll need to sort it out, chump!", "\n",
               "I have returned the list of offending names."))
  } 
    # Which names in the source2 list can we keep as unique synonyms?
  S2Unique <- S2synonyms %>%
    dplyr::filter(!validName %in% S1synonyms$validName)
      # Pass these names onto 3.0

  
  #### 2.0 S1 duplicates ####
    ##### 2.1 acc. names ####
    # Look for internal source1 duplicated ACCEPTED names
  S1duplicates <- S1accepted %>%
    dplyr::group_by(validName) %>%
    dplyr::filter(dplyr::n() > 1)
  # Stop here becuase I have no matches, but this might be important down the track if someone 
    # finds them!
  if(nrow(S1duplicates) > 0 ){
    return(S1duplicates)
    stop(paste(" - That's odd! There is an internal S1 synonym.",
               "This hasn't happened before, but you'll need to sort it out, chump!", "\n",
               "I have returned the list of offending names."))
  }
  # Because none of these are duplicates, I will KEEP the original dataset S1accepted.
  
    ##### 2.2 valName synonyms ####
  # Look for internal source1 duplicated SYNONYMS
  S1duplicatesyns <- S1synonyms %>%
    dplyr::group_by(validName) %>%
    dplyr::filter(dplyr::n() > 1)
  
 S1dupes_nest <- S1duplicatesyns %>%
    # ungroup but nest the data by valid name instead
   dplyr::ungroup() %>%
   dplyr::nest_by(validName) 
 
 ###### a. source1 loop ####
 # Set up empty dataframes for loop
 ambiSyns <- dplyr::tibble()
 nonAmbiSyns <- dplyr::tibble()
    # Run a loop to examine each duplicate pair in the list
 if(nrow(S1dupes_nest) > 0){
 for(i in 1:nrow(S1dupes_nest)){
   # Get the first tibble
   LoopTibble <- S1dupes_nest$data[[i]] %>% 
     # add the validName column back in to each row
   dplyr::mutate(validName = S1dupes_nest$validName[[i]], .after = "subtribe")
   
       # FOR n == 2
       if(nrow(LoopTibble) == 2){
        # LOGICAL both duplicates match to the same accid
         logiTest <- all(duplicated(LoopTibble$accid) | duplicated(LoopTibble$accid, fromLast = TRUE))
          # IF the duplicates match the same accid (accepted name) - NON-ambiguous
         if(logiTest == TRUE){
           nonAmbiSyns <- nonAmbiSyns %>% dplyr::bind_rows(LoopTibble)
         } # END TRUE
           # IF the duplicates match different accid (accepted name) - AMBIGUOUS
         if(logiTest == FALSE){
           ambiSyns <- ambiSyns %>% dplyr::bind_rows(LoopTibble)
         } # END FALSE
       }# END n == 2
   
        # FOR n > 2
      if(nrow(LoopTibble) > 2){
          # Find non-ambiguous duplicates
        nrow_nonAmbi <- LoopTibble %>% dplyr::group_by(accid) %>% dplyr::filter(dplyr::n() > 1) %>%
          nrow()
          # Find ambiguous duplicates
        nrow_Ambi <- LoopTibble %>% dplyr::group_by(accid) %>% dplyr::filter(dplyr::n() == 1) %>%
          nrow()
          # IF ALL of these rows have the same accid, then they are just regular synonym duplicates
        if(nrow_nonAmbi == nrow(LoopTibble)){
          # Add the lowest id number to the nonAmbiSyns tibble
          LoopTibble <- LoopTibble %>% dplyr::arrange(id)
          nonAmbiSyns <- nonAmbiSyns %>% 
            dplyr::bind_rows(dplyr::filter(LoopTibble, dplyr::row_number() == 1))
        }else{ # ALL of the others have been ambiguous so far
            # Add these data to the ambiSyns dataframe
          ambiSyns <- ambiSyns %>% 
            dplyr::bind_rows(LoopTibble)
        } # END else
      } # END n > 2
 } # END Ambiguous loop
} # END length(S1dupes_nest) > 0
    ###### b. loop_clean ####
 if(nrow(nonAmbiSyns) > 0){
    # Take only one of each non-ambiguous synonyms
 nonAmbiSyns_deDuped <- nonAmbiSyns %>%
   dplyr::group_by(validName) %>%
   dplyr::filter(dplyr::row_number() == 1)
    # For ambiguous accids, add this to the flags
 ambiSyns$flags <- "ambiguous validName"
    ###### c. merge ####
  # Merge this back to the S1synonyms data. This will have duplicates removed and 
    # internally-ambiguous names flagged.
 S1synonyms <- S1synonyms %>%
    # REMOVE the duplicated valid names
   dplyr::filter(!validName %in% S1duplicatesyns$validName) %>%
    # ADD the cleaned rows back into the dataset
   dplyr::bind_rows(nonAmbiSyns_deDuped, ambiSyns) }else{
     
      # IF there are no non-ambiguous names then...
     # For ambiguous accids, add this to the flags
     ambiSyns$flags <- "ambiguous validName"
     
     # Merge this back to the S1synonyms data. This will have duplicates removed and 
     # internally-ambiguous names flagged.
     S1synonyms <- S1synonyms %>%
       # REMOVE the duplicated valid names
       dplyr::filter(!validName %in% S1duplicatesyns$validName) %>%
       # ADD the cleaned rows back into the dataset
       dplyr::bind_rows(ambiSyns)
   }
  # KEEP S1synonyms
 
 
  #### 3.0 source2 duplicates ####
    # Look for internal source1 duplicates
  S2Duplicates <- S2Unique %>%
    dplyr::group_by(validName) %>%
    dplyr::filter(dplyr::n() > 1)
    # Yep, there are source2 synonym duplicates to deal with!
  # Do any of the accids in the full list match these names' ids?
  S2IDmatches <- S2Duplicates %>%
    dplyr::filter(id %in% SynList$accid)
       # Stop here because I have no matches, but this might be important down the track if 
    # someone finds them!
       if(nrow(S2IDmatches) > 0 ){
         return(S2IDmatches)
         stop(paste(" - That's odd! There are accids matching to source2 synonym IDs.",
                    "This hasn't happened before, but you'll need to sort it out, chump!", "\n",
                    "I have returned the list of offending names. You're welcome."))
       }
      # Take only the lowest id number match
  S2Originals <- S2Duplicates %>% 
      # Sort by id number
    dplyr::arrange(id) %>%
      # Filter out ANY duplicated rows for validName
    dplyr::group_by(validName) %>%
      # take the first row
    dplyr::filter(dplyr::row_number() == 1)
    # KEEP S2Originals
  
  
  #### 4.0 Merge ####
  dupeMerge <- dplyr::bind_rows(S1accepted, S2Originals, S1synonyms) %>%
      # sort again by id
    dplyr::arrange(id)
  
    # Check to make sure that all ids are unique
  UniqueIDcheck <- dupeMerge %>%
    dplyr::arrange(id) %>%
    dplyr::filter(!duplicated(id))
  # FIRST, for now remove ambiguous names
  ambi_VNcheck <- UniqueIDcheck %>% dplyr::filter(flags %in% "ambiguous validName")
  NonAmbi_VNcheck <- UniqueIDcheck %>% dplyr::filter(!flags %in% "ambiguous validName") %>%
    dplyr::ungroup()
    # Check to make sure that all validNames are unique
  UniqueVNcheck <- NonAmbi_VNcheck %>%
    dplyr::group_by(validName) %>%
    dplyr::filter(dplyr::n() > 1)
  
     ##### 4.1 ValSyn_clean ####
      # look for matches between source1 accepted names and source2 synonyms
      # At present, all of these represent an accepted source1 name with a contradictory source2
    # name.
        # Keep the source1 name but warn the user if this changes...
     dupes2remove_UnVNcheck <- UniqueVNcheck %>%
      dplyr::group_by(validName) %>%
      dplyr::filter(accid != 0)
      # Stop if this is not half of the original (not all correspond to an source2 Syn)
    if(nrow(dupes2remove_UnVNcheck) != (nrow(UniqueVNcheck)/2)){
      stop(paste(" - This is new! There is a problem at 4.1 ValSyn_clean. Please go and have a look.",
                 "\n", "Good luck, LOL."))
    }
      # Remove dupes2remove_UnVNcheck (duplicates) from the list to return, and then add the new names.
  dupes2keep <- dupeMerge %>%
      # Remove duplicates...
    dplyr::filter(!id %in% dupes2remove_UnVNcheck$id) 
  
  # Merge these with the original dataset
  deDuplicated <- SynList %>% 
    # FIRST, remove all of the original duplicate rows
    dplyr::filter(!validName %in% duplicates$validName) %>% 
    # Add in the duplicates we want to keep
    dplyr::bind_rows(dupes2keep)

    ##### 4.2 Duplicate ids ####
    # There might be some rows with duplicate ids. These are now unique validNames. Assign these all NEW ids
  dupID <- deDuplicated %>%
    dplyr::filter(duplicated(id) | duplicated(id, fromLast = TRUE))
    # IF so, remove them
  if(nrow(dupID) > 0){
    # Remove these from the original dataset
  deDuplicated <- deDuplicated %>%
    dplyr::filter(!id %in% unique(dupID$id))
  # Replace the ids with new ones starting from +1 the max id number already existing
    # find the largest id...
      SeqStart <- max(deDuplicated$id)+1
      SeqEnd <- as.numeric(SeqStart+nrow(dupID))-1
    dupID$id <- seq(from = SeqStart, to = SeqEnd, by = 1)
  # re-merge
    deDuplicated <- deDuplicated %>%
      dplyr::bind_rows(dupID)
  } # END dupID
  
  ##### 4.3 Ambi accepted ####
    # Some ambiguous names are accepted names. Therefore, I will remove the associated ambiguous synonyms
    ambiAcc <- deDuplicated %>%
        # Find the duplicate names
      dplyr::filter(duplicated(validName)|duplicated(validName, fromLast = TRUE)) %>% 
        # Find the accepted names
      dplyr::filter(taxonomic_status == "accepted")
      # Get the number of accepted-assocaited ambiguous names that were removed.
    ambiAccCount <- nrow(dplyr::filter(deDuplicated, validName %in% ambiAcc$validName)) - nrow(ambiAcc)
        # REMOVE those names from the whole dataset
    deDuplicated <- deDuplicated %>% 
        # remove
      dplyr::filter(!validName %in% ambiAcc$validName) %>%
        # rejoin those accepted name rows
      dplyr::bind_rows(ambiAcc)
    
    #### 5.0 Final Ambi ####
      ##### 5.1 can_wFl synonyms ####
    # Look for internal source1 duplicated SYNONYMS
    S1duplicatesyns_51 <- deDuplicated %>%
      dplyr::group_by(canonical_withFlags) %>%
      dplyr::filter(canonical_withFlags %>% stringr::str_detect(
        "_"
      )) %>%
      dplyr::filter(dplyr::n() > 1) 

    S1dupes_nest <- S1duplicatesyns_51 %>%
      # ungroup but nest the data by valid name instead
      dplyr::ungroup() %>%
      dplyr::nest_by(canonical_withFlags) 
    
    ###### a. source1 loop ####
    # Set up empty dataframes for loop
    ambiSyns_51 <- dplyr::tibble()
    nonAmbiSyns_51 <- dplyr::tibble()
    # IF S1duplicatesyns_51 is EMPTy, do not run.
    if(nrow(S1duplicatesyns_51) > 0){
      # Run a loop to examine each duplicate pair in the list
      for(i in 1:nrow(S1dupes_nest)){
        # Get the first tibble
        LoopTibble <- S1dupes_nest$data[[i]] %>% 
          # add the canonical_withFlags column back in to each row
          dplyr::mutate(canonical_withFlags = S1dupes_nest$canonical_withFlags[[i]], .after = "canonical")
        
        # FOR n == 2
        if(nrow(LoopTibble) == 2){
          # LOGICAL both duplicates match to the same accid
          logiTest <- all(duplicated(LoopTibble$accid) | duplicated(LoopTibble$accid, fromLast = TRUE))
          # IF the duplicates match the same accid (accepted name) - NON-ambiguous
          if(logiTest == TRUE){
            nonAmbiSyns_51 <- nonAmbiSyns_51 %>% dplyr::bind_rows(LoopTibble)
          } # END TRUE
          # IF the duplicates match different accid (accepted name) - AMBIGUOUS
          if(logiTest == FALSE){
              # If one of these matches the other, they are NOT ambiguous.
            accTEST <- any(LoopTibble$id %in% LoopTibble$accid)
                 if(accTEST == FALSE){
               ambiSyns_51 <- ambiSyns_51 %>% dplyr::bind_rows(LoopTibble)
                 }
          } # END FALSE
        }# END n == 2
        
        # FOR n > 2
        if(nrow(LoopTibble) > 2){
          # Find non-ambiguous duplicates
          nrow_nonAmbi <- LoopTibble %>% dplyr::group_by(accid) %>% 
            dplyr::filter(dplyr::n() > 1) %>%
            nrow()
          # Find ambiguous duplicates
          nrow_Ambi <- LoopTibble %>% dplyr::group_by(accid) %>% 
            dplyr::filter(dplyr::n() == 1) %>%
            nrow()
          # IF ALL of these rows have the same accid, then they are just regular synonym duplicates
          if(nrow_nonAmbi == nrow(LoopTibble)){
            #  # Add the lowest id number to the nonAmbiSyns_51 tibble
            ambiSyns_51 <- ambiSyns_51 %>% 
              dplyr::bind_rows(LoopTibble)
          }else{ # ALL of the others have been ambiguous so far
            # Logical - if ALL but one accid matches an id, take the to mean they are all pointing at
            # the same record. None shold match for now.
            accTest <- sum(LoopTibble$id %in% LoopTibble$accid) == nrow(LoopTibble)-1
               # Add these data to the ambiSyns_51 dataframe
               if(accTest == FALSE){ # Ad all as synonyms
               ambiSyns_51 <- ambiSyns_51 %>% 
                 dplyr::bind_rows(LoopTibble)
               }else(
                 stop(" - unique problem at 5.1. :(")
               )
          } # END else
        } # END n > 2
      } # END Ambiguous loop
    }else{
      ambiSyns_51 = dplyr::tibble()
      nonAmbiSyns_51 = dplyr::tibble()
    } # END big IF
    
    ###### b. loop_clean ####
      # NON-AMBIGUOUS - because accid matches
    if(nrow(nonAmbiSyns_51) > 0){
      nonAmbiSyns_51_nAmb <- nonAmbiSyns_51  %>%
        # Filter for ONLY the names that AREN'T already flagged as ambiguous
        dplyr::filter(!flags %in% c("ambiguous validName")) %>%
        dplyr::filter(canonical_withFlags %>% stringr::str_detect("_homonym"))
      # For ambiguous accids, add this to the flags
      nonAmbiSyns_51_nAmb$flags <- paste(nonAmbiSyns_51_nAmb$flags, "non-ambiguous can_wFlags", 
                                         sep = ", ") %>%
        # REMOVE EMPTYS
        stringr::str_replace(pattern = "NA, ", "")
      
      # internally-ambiguous names flagged.
      deDuplicated_51 <- deDuplicated %>%
          # REMOVE the matching ids
        dplyr::filter(!id %in% nonAmbiSyns_51_nAmb$id) %>%
         # ADD the new rows
        dplyr::bind_rows(nonAmbiSyns_51_nAmb) 
    } else{
        # If not, pass this new name onto the next section
      deDuplicated_51 <- deDuplicated
    }
    
      # AMBIGUOUS 2
    if(nrow(ambiSyns_51) > 0){
      ambiSyns_51_NavN <- ambiSyns_51 %>%
        # Filter for ONLY the names that AREN'T already flagged as ambiguous
        dplyr::filter(!flags %in% c("ambiguous validName")) 
      # For ambiguous accids, add this to the flags
      ambiSyns_51_NavN$flags <- paste(ambiSyns_51_NavN$flags, "ambiguous can_wFlags", sep = ", ") %>%
          # REMOVE EMPTYS
        stringr::str_replace(pattern = "NA, ", "")
        # Filter the VALID ambiguities and ADD to the wflags
      ambiSyns_51_all <- ambiSyns_51 %>% 
          # filter
        dplyr::filter(!id %in% ambiSyns_51_NavN$id) %>%
          # add
        dplyr::bind_rows(ambiSyns_51_NavN)
      # Merge this back to the deDuplicated data. This will have duplicates removed and 
      # internally-ambiguous names flagged.
      deDuplicated_51 <- deDuplicated_51 %>%
        # REMOVE the duplicated valid names
        dplyr::filter(!canonical_withFlags %in% ambiSyns_51_all$canonical_withFlags) %>%
        # ADD the cleaned rows back into the dataset
        dplyr::bind_rows(ambiSyns_51_all) 
    }else{
      ambiSyns_51 = dplyr::tibble()
      # If not, pass this new name onto the next section
      deDuplicated_51 <- deDuplicated
    } # END ambiSyns_51 IF
    


    ##### 5.2 canon synonyms ####
    # Look for internal source1 duplicated SYNONYMS
    S1duplicatesyns_52 <- deDuplicated_51 %>%
      dplyr::group_by(canonical) %>%
      dplyr::filter(dplyr::n() > 1) 
    
    S1dupes_nest <- S1duplicatesyns_52 %>%
      # ungroup but nest the data by valid name instead
      dplyr::ungroup() %>%
      dplyr::nest_by(canonical) 
    
    ###### a. source1 loop ####
    # Set up empty dataframes for loop
    ambiSyns_52 <- dplyr::tibble()
    nonAmbiSyns_52 <- dplyr::tibble()
    # IF S1duplicatesyns_52 is EMPTy, do not run.
    if(nrow(S1duplicatesyns_52) > 0){
      # Run a loop to examine each duplicate pair in the list
      for(i in 1:nrow(S1dupes_nest)){
        # Get the first tibble
        LoopTibble <- S1dupes_nest$data[[i]] %>% 
          # add the canonical column back in to each row
          dplyr::mutate(canonical = S1dupes_nest$canonical[[i]], .after = "validName")
        
        # FOR n == 2
        if(nrow(LoopTibble) == 2){
          # LOGICAL both duplicates match to the same accid
          logiTest <- all(duplicated(LoopTibble$accid) | duplicated(LoopTibble$accid, fromLast = TRUE))
          # IF the duplicates match the same accid (accepted name) - NON-ambiguous
          if(logiTest == TRUE){
            nonAmbiSyns_52 <- nonAmbiSyns_52 %>% dplyr::bind_rows(LoopTibble)
          } # END TRUE
          # IF the duplicates match different accid (accepted name) - AMBIGUOUS
          if(logiTest == FALSE){
            # If one of these matches the other, they are NOT ambiguous.
            accTEST <- any(LoopTibble$id %in% LoopTibble$accid)
            if(accTEST == FALSE){
              ambiSyns_52 <- ambiSyns_52 %>% dplyr::bind_rows(LoopTibble)
            }
          } # END FALSE
        }# END n == 2
        
        # FOR n > 2
        if(nrow(LoopTibble) > 2){
          # Find non-ambiguous duplicates
          nrow_nonAmbi <- LoopTibble %>% dplyr::group_by(accid) %>% 
            dplyr::filter(dplyr::n() > 1) %>%
            nrow()
          # Find ambiguous duplicates
          nrow_Ambi <- LoopTibble %>% dplyr::group_by(accid) %>% 
            dplyr::filter(dplyr::n() == 1) %>%
            nrow()
          # IF ALL of these rows have the same accid, then they are just regular synonym duplicates
          if(nrow_nonAmbi == nrow(LoopTibble)){
            #  # Add the lowest id number to the nonAmbiSyns_52 tibble
            ambiSyns_52 <- ambiSyns_52 %>% 
              dplyr::bind_rows(LoopTibble)
          }else{ # ALL of the others have been ambiguous so far
            # Logical - if ALL but one accid matches an id, take the to mean they are all pointing at
            # the same record. None shold match for now.
            accTest <- sum(LoopTibble$id %in% LoopTibble$accid) == nrow(LoopTibble)-1
            # Add these data to the ambiSyns_52 dataframe
            if(accTest == FALSE){ # Ad all as synonyms
              ambiSyns_52 <- ambiSyns_52 %>% 
                dplyr::bind_rows(LoopTibble)
            }else(
              stop(" - unique problem at 5.2! :(")
            )
          } # END else
        } # END n > 2
      } # END Ambiguous loop
    }else{
      ambiSyns_52 = dplyr::tibble()
      nonAmbiSyns_52 = dplyr::tibble()
    } # END big IF
    ###### b. loop_clean ####
    # NON-AMBIGUOUS because accids match
    if(nrow(nonAmbiSyns_52) > 0){
      # Take only one of each non-ambiguous synonyms
      nonAmbiSyns_deDuped_52 <- nonAmbiSyns_52 %>%
        dplyr::group_by(validName) %>%
        dplyr::filter(dplyr::row_number() == 1)
    }
    
    # NON-AMBIGUOUS
    if(nrow(nonAmbiSyns_52) > 0){
      nonAmbiSyns_52_nAmb <- nonAmbiSyns_52  %>%
        # Filter for ONLY the names that AREN'T already flagged as ambiguous
        dplyr::filter(!flags %in% c("ambiguous validName", "ambiguous can_wFlags", 
                                    "ambiguous can_wFlags")) 
      # For ambiguous accids, add this to the flags
      nonAmbiSyns_52_nAmb$flags <- paste(nonAmbiSyns_52_nAmb$flags, "non-ambiguous canonical", sep = ", ") %>%
        # REMOVE EMPTYS
        stringr::str_replace(pattern = "NA, ", "")
      
      # internally-ambiguous names flagged.
      deDuplicated_52 <- deDuplicated_51 %>%
        # REMOVE the matching ids
        dplyr::filter(!id %in% nonAmbiSyns_52_nAmb$id) %>%
        # ADD the new rows
        dplyr::bind_rows(nonAmbiSyns_52_nAmb) 
    } else{
      # If not, pass this new name onto the next section
      deDuplicated_52 <- deDuplicated_51
    }
    
    # AMBIGUOUS
    if(nrow(ambiSyns_52) > 0){
      ambiSyns_52_NavN <- ambiSyns_52 %>%
        # Filter for ONLY the names that AREN'T already flagged as ambiguous
        dplyr::filter(!flags %in% c("ambiguous validName", "ambiguous can_wFlags")) 
      # For ambiguous accids, add this to the flags
      ambiSyns_52_NavN$flags <- paste(ambiSyns_52_NavN$flags, "ambiguous canonical", sep = ", ") %>%
        # REMOVE EMPTYS
        stringr::str_replace(pattern = "NA, ", "")
      # Filter the VALID ambiguities and ADD to the wflags
      ambiSyns_52_all <- ambiSyns_52 %>% 
        # filter
        dplyr::filter(!id %in% ambiSyns_52_NavN$id) %>%
        # add
        dplyr::bind_rows(ambiSyns_52_NavN)
      # Merge this back to the deDuplicated data. This will have duplicates removed and 
      # internally-ambiguous names flagged.
      deDuplicated_52 <- deDuplicated_52 %>%
        # REMOVE the duplicated valid names
        dplyr::filter(!canonical %in% ambiSyns_52_all$canonical) %>%
        # ADD the cleaned rows back into the dataset
        dplyr::bind_rows(ambiSyns_52_all) 
    }else{
      ambiSyns_52 = dplyr::tibble()
    } # END ambiSyns_52 IF
    # KEEP deDuplicated_52
    
    
    

    # What an adventure that was!
    # Now, lets try and return some user information 
writeLines(paste(    " - Cleaning complete! From an initial dataset of ", 
                 format(nrow(SynList), big.mark = ","), " names, there ",
                 "remain ", format(nrow(deDuplicated_52), big.mark = ",")," names.",  "\n",
                     " - We removed:", "\n"   ,
                 nrow(S1duplicates), " source1 accepted names,", "\n"   ,
                 nrow(S2Acc2remove), " source2 'accepted' names,", "\n"))
                  # 2.2 - synonyms removed
if(exists("nonAmbiSyns_deDuped")){
                   writeLines(paste(
                 format(nrow(nonAmbiSyns)-nrow(nonAmbiSyns_deDuped), big.mark = ","),
                        " source1 synonyms,", "\n"   ))}
writeLines(paste(
                 format(nrow(S2synonyms) - nrow(S2Unique), big.mark = ",")
                 , " source2 synonyms internally duplicated,", "\n"   ,
                 nrow(S2Duplicates)-nrow(S2Originals), " source2 synonyms duplicated with the source1 list,", "\n"   ,
                 nrow(dupes2remove_UnVNcheck), " subsequent duplicates after merging,", "\n",
                  # AMBIGUOUS flagged
                    " - We flagged:", "\n"   ,
                 sum(deDuplicated_52$flags %in% "ambiguous validName"), 
                  " ambiguous validName, ", "\n"   ,
                 sum(deDuplicated_52$flags %in% "ambiguous can_wFlags"), 
                  " ambiguous canonical_withFlags names, ", "\n"   ,
                 sum(deDuplicated_52$flags %in% "ambiguous canonical"), 
                 " ambiguous canonical names, ", "\n",
                 sum(deDuplicated_52$flags %in% "non-ambiguous can_wFlags"), 
                 " NON-ambiguous, but duplicated, canonical_withFlags names, ", "\n",
                 sum(deDuplicated_52$flags %in% "non-ambiguous canonical"), 
                 " NON-ambiguous, but duplicated, canonical names, ", "\n",
                     " - We removed: ", "\n",
                 ambiAccCount, " ambiguous synonyms associated with accepted names.", "\n",
                    " - We re-assigned:", "\n"   ,
                 nrow(dupID), " duplicated [non-duplicate] ids",
sep = "")) 

  #### 6.0 Clean flags ####
deDuplicated_52 <- deDuplicated_52 %>%
  dplyr::mutate(flags = 
                    # Fix non-ambiguous canonical repeat
                  dplyr::if_else(flags %>% stringr::str_count("non-ambiguous canonical") > 1,
                                 stringr::str_remove_all(flags, "non-ambiguous canonical") %>%
                                   stringr::str_c("non-ambiguous canonical"), flags),
                    # Fix contradictory non- and is-
                flags = dplyr::if_else(stringr::str_detect(
                  flags, "ambiguous canonical, non-ambiguous canonical") |
                    stringr::str_detect(flags, "non-ambiguous canonical, ambiguous canonical"),
                                   stringr::str_remove_all(flags, "ambiguous canonical") %>% 
                                     stringr::str_remove_all("non-ambiguous canonical") %>% 
                                     stringr::str_remove("^, |, $") %>%
                                     stringr::str_replace(", , ", ", ") %>%
                                     stringr::str_c(" ambiguous canonical "), flags),
                flags = dplyr::if_else(stringr::str_detect(
flags, 
"non-ambiguous can_wFlags, ambiguous canonical, non-ambiguous can_wFlags, ambiguous canonical"),
stringr::str_c(flags, " non-ambiguous can_wFlags, ambiguous canonical "), flags
                ),
### 3
                flags = dplyr::if_else(stringr::str_detect(flags,
"non-ambiguous can_wFlags, non-ambiguous can_wFlagsnon-ambiguous canonical"),
"non-ambiguous can_wFlags, non-ambiguous canonical", flags),
### 4
                flags = dplyr::if_else(stringr::str_detect(flags,
"non-ambiguous can_wFlags, ambiguous canonical, non-ambiguous can_wFlags, ambiguous canonical non-ambiguous can_wFlags, ambiguous canonical"),
"non-ambiguous can_wFlags, ambiguous canonical", flags),
### 5
                flags = dplyr::if_else(stringr::str_detect(flags,
"ambiguous can_wFlags, ambiguous can_wFlags, ambiguous canonical"),
"ambiguous can_wFlags, ambiguous canonical", flags),
### 6
flags = dplyr::if_else(stringr::str_detect(flags,
                                           "non- ambiguous canonical"),
                       "non-ambiguous canonical", flags),
### 7
flags = dplyr::if_else(stringr::str_detect(flags,
                                           "ambiguous canonical, ambiguous canonical"),
                       "ambiguous canonical", flags),
### cleanup
                flags = flags %>% stringr::str_squish() %>%
  stringr::str_remove("^, |, $") %>%
  stringr::str_replace(", , ", ", ") 
  )

  # Return the cleaned dataset
return(deDuplicated_52)
}
  #### END ####

# This function was made by James Dorey to manage flags in the synonym lists
# For queries, please contact James Dorey at jbdorey[at]me.com
# This function was started on 15th May 2022 and last updated 17th May 2022
#' @importFrom dplyr %>%


FlagManager <- function(InSynList = DLdf,
                        flagCol = flags){
  # locally bind variables to the function
  . <- DLdf <- flags <- genus <- validName <- canonical <- NULL
  
  #### Flags option ####
  if(flagCol == "flags"){
    base::writeLines(" - Using the flags column.")
  # If there is no canonical_withFlags, make one with the canonical column, assuming it has flags
  suppressWarnings( CWFtest <- InSynList$canonical_withFlags,
                    classes = "warning")
  if(is.null(CWFtest) == TRUE){
    canonical_withFlags <- InSynList$canonical
    InSynList <- dplyr::mutate(InSynList, canonical_withFlags, .after = "canonical")
  }
    # IF all of the canonical_withFlags column is empty, replace it with the canonical valuse
  if(all(CWFtest == "NA") == TRUE){
    InSynList$canonical_withFlags <- InSynList$canonical
  }
  
  base::writeLines(" - 1. Remove flag from validName column...")
  ##### START ProgBar 1 ####
  # Initializes the progress bar
  pb1 <- utils::txtProgressBar(min = 0,      # Minimum value of the progress bar
                       max = nrow(InSynList), # Maximum value of the progress bar
                       style = 3,    # Progress bar style (also available style = 1 and style = 2)
                       width = NA,   # Progress bar width. Defaults to getOption("width")
                       char = "=")   # Character used to create the bar
  
  # Remove flag from validName
  for(i in 2:nrow(InSynList)){ 
    if(is.na(InSynList$flags[i]) == FALSE){ # Enter statement only IF there is a flag present
      InSynList$validName[i] <- sub(InSynList$flags[i], "", InSynList$validName[i] , fixed = TRUE) %>%
        trimws( which = "right" , whitespace = "[\\, ]")  # Trim some special characters from the END of the string
    } # END IF statement
    
    # Sets the progress bar to the current state
    utils::setTxtProgressBar(pb1, i)
  } # END loop COLUMNS
  #### END progBar 1 ####
  close(pb1) # Close the connection
  
  # Clean up double spaces in data frame
  InSynList <- as.data.frame(apply(InSynList, 2, FUN = stringr::str_squish ), stringsAsFactors = FALSE)
  # Remove empty rows that contain the following strings
  InSynList <- subset(InSynList, genus!="NA" & genus!="F" & genus!="V" & genus!="Discover" & 
                        genus!="Updated:" & genus!="Other" & validName!="Kinds of" & 
                        validName!="Scientific name" & canonical != "Kim, NA")
  
  # Find annotations and add the instead to the flags column
  # Which columns to search through
  FlagAnnot_cols <- c("genus",	"subgenus",	"species",	"infraspecies", "authorship")
  # Which flags to find
  FlagsAnnot <- c(
    # Name flags
    "_homonym[1-4]?","_homonytm","_homony","_sic","_sensu_lato",
    "_unavailable","_invalid","_nomen_nudum","_emend",
    " var "," subvar "," forma "," form "," race "," auct ","^var ","^subvar ","^forma ",
    "^form ","^race ",
    # unsure flags
    " f ","^f "," m ","^m "," r ","^r "," ab ","^ab ",
    # Specific cases
    "_Urban_not_Pasteels","_Friese_not_Megerle","_Friese_not_H\\u00fcbner","_Friese_not_Stimpson",
    # Author flags
    "Auctorum, sensu","sensu auct not","Auct non","Auct, not","auct, not","_auct not_",
    "_auct","^auct ") %>% 
    paste(collapse="|")
  
  
  base::writeLines(" - 2. Find and add flags to the 'flags' column...")
  #### START ProgBar 2 ####
  # Initializes the progress bar
  pb2 <- utils::txtProgressBar(min = 0,      # Minimum value of the progress bar
                       max = length(FlagAnnot_cols), # Maximum value of the progress bar
                       style = 3,    # Progress bar style (also available style = 1 and style = 2)
                       width = NA,   # Progress bar width. Defaults to getOption("width")
                       char = "=")   # Character used to create the bar
  
  # Find and add flags to flag column
  for(j in 1:length(FlagAnnot_cols)){ # COLUMNS
    for(i in 1:nrow(InSynList)){ # ROWS
      if(grepl(FlagsAnnot,  InSynList[i,FlagAnnot_cols[j]], fixed = FALSE) == TRUE){
        # Extract the flag from the string and add the column name
        ExtractedFlag <- stringr::str_extract_all(InSynList[i,FlagAnnot_cols[j]], 
                                                  FlagsAnnot, simplify = TRUE) %>%
          paste(FlagAnnot_cols[j],.,sep=" ", collapse = "|") %>%
          gsub("  ", " ",. ) # Remove double spaces
        if( is.na(InSynList[i,"flags"]) == TRUE){ # If there is NO flag, insert this flag as is
          InSynList[i,"flags"] <- ExtractedFlag
        }else{ # If there IS a flag, ADD this to the existing flag
          InSynList[i,"flags"] <- paste(InSynList[i,"flags"], ExtractedFlag, sep =  " | ")
        } # END if else statement
      } # END find flag IF statement
    } # END for loop of flag annotations - ROWS
    
    # Sets the progress bar to the current state
    utils::setTxtProgressBar(pb2, j)
  } # END loop COLUMNS
  #### END progBar 2 ####
  close(pb2) # Close the connection
  
  # Trim internal white spaces in flag column
  InSynList$flags <- gsub(" \\| | \\||\\| ", "|", InSynList$flags)  
  # Remove these strings from the relevant columns now that they are saved as flags
  TempCols <- as.data.frame(apply(InSynList[c("canonical",FlagAnnot_cols) ], 2,
                                  function(y) gsub(FlagsAnnot, " ", y))) 
  # Replace existing columns with these new, trimmed, columns
  InSynList[c(FlagAnnot_cols)] <- TempCols[, FlagAnnot_cols]
  # Add the canonical column after the canonical column with flags - later convert to 
    # canonical and canonical_withFlags
  CanTest <- InSynList$canonical
    # IF there IS already a canonical column repalce that column with the new one...
  if(is.null(CanTest) == FALSE){
    InSynList$canonical <- TempCols$canonical
  }else{
    InSynList <- dplyr::mutate(InSynList, TempCols$canonical, .after = "canonical")
  }
  #  Clean up double spaces in data frame
  InSynList <- as.data.frame(apply(InSynList, 2, FUN = stringr::str_squish ), stringsAsFactors = FALSE)
  } # END IF flags col
  
#### Notes option ####
  if(flagCol == "notes"){
    base::writeLines(" - Using the notes column.")
    # If there is no canonical_withFlags, make one with the canonical column, assuming it has flags
    suppressWarnings( CWFtest <- InSynList$canonical_withFlags,
                      classes = "warning")
    if(is.null(CWFtest) == TRUE){
      canonical_withFlags <- InSynList$canonical
      InSynList <- dplyr::mutate(InSynList, canonical_withFlags, .after = "canonical")
    }
    # IF all of the canonical_withFlags column is empty, replace it with the canonical valuse
    if(all(CWFtest == "NA") == TRUE){
      InSynList$canonical_withFlags <- InSynList$canonical
    }
    
    base::writeLines(" - 1. Remove flag from validName column...")
    ##### START ProgBar 1 ####
    # Initializes the progress bar
    pb1 <- utils::txtProgressBar(min = 0,      # Minimum value of the progress bar
                          max = nrow(InSynList), # Maximum value of the progress bar
                          style = 3,    # Progress bar style (also available style = 1 and style = 2)
                          width = NA,   # Progress bar width. Defaults to getOption("width")
                          char = "=")   # Character used to create the bar
    
    # Remove flag from validName
    for(i in 2:nrow(InSynList)){ 
      if(is.na(InSynList$notes[i]) == FALSE){ # Enter statement only IF there is a flag present
        InSynList$validName[i] <- sub(InSynList$notes[i], "", InSynList$validName[i] , fixed = TRUE) %>%
          trimws( which = "right" , whitespace = "[\\, ]")  # Trim some special characters from the END of the string
      } # END IF statement
      
      # Sets the progress bar to the current state
      utils::setTxtProgressBar(pb1, i)
    } # END loop COLUMNS
    #### END progBar 1 ####
    close(pb1) # Close the connection
    
    # Clean up double spaces in data frame
    InSynList <- as.data.frame(apply(InSynList, 2, FUN = stringr::str_squish ), stringsAsFactors = FALSE)
    # Remove empty rows that contain the following strings
    InSynList <- subset(InSynList, genus!="NA" & genus!="F" & genus!="V" & genus!="Discover" & 
                          genus!="Updated:" & genus!="Other" )
    
    # Find annotations and add the instead to the flags column
    # Which columns to search through
    FlagAnnot_cols <- c("genus",	"subgenus",	"species",	"infraspecies", "authorship")
    # Which flags to find
    FlagsAnnot <- c(
      # Name flags
      "_homonym[1-4]?","_homonytm","_homony","_sic","_sensu_lato",
      "_unavailable","_invalid","_nomen_nudum","_emend",
      " var "," subvar "," forma "," form "," race "," auct ","^var ","^subvar ","^forma ",
      "^form ","^race ",
      # unsure flags
      " f ","^f "," m ","^m "," r ","^r "," ab ","^ab ",
      # Specific cases
      "_Urban_not_Pasteels","_Friese_not_Megerle","_Friese_not_H\\u00fcbner","_Friese_not_Stimpson",
      # Author flags
      "Auctorum, sensu","sensu auct not","Auct non","Auct, not","auct, not","_auct not_",
      "_auct","^auct ") %>%
      paste(collapse="|")
    
    
    base::writeLines(" - 2. Find and add flags to the 'flags' column...")
    #### START ProgBar 2 ####
    # Initializes the progress bar
    pb2 <- utils::txtProgressBar(min = 0,      # Minimum value of the progress bar
                          max = length(FlagAnnot_cols), # Maximum value of the progress bar
                          style = 3,    # Progress bar style (also available style = 1 and style = 2)
                          width = NA,   # Progress bar width. Defaults to getOption("width")
                          char = "=")   # Character used to create the bar
    
    # Find and add flags to notes column
    for(j in 1:length(FlagAnnot_cols)){ # COLUMNS
      for(i in 1:nrow(InSynList)){ # ROWS
        if(grepl(FlagsAnnot,  InSynList[i,FlagAnnot_cols[j]], fixed = FALSE) == TRUE){
          # Extract the flag from the string and add the column name
          ExtractedFlag <- stringr::str_extract_all(InSynList[i,FlagAnnot_cols[j]], 
                                                    FlagsAnnot, simplify = TRUE) %>%
            paste(FlagAnnot_cols[j],.,sep=" ", collapse = "|") %>%
            gsub("  ", " ",. ) # Remove double spaces
          if( is.na(InSynList[i,"notes"]) == TRUE){ # If there is NO flag, insert this flag as is
            InSynList[i,"notes"] <- ExtractedFlag
          }else{ # If there IS a flag, ADD this to the existing flag
            InSynList[i,"notes"] <- paste(InSynList[i,"notes"], ExtractedFlag, sep =  " | ")
          } # END if else statement
        } # END find flag IF statement
      } # END for loop of flag annotations - ROWS
      
      # Sets the progress bar to the current state
      utils::setTxtProgressBar(pb2, j)
    } # END loop COLUMNS
    #### END progBar 2 ####
    close(pb2) # Close the connection
    
    # Trim internal white spaces in flag column
    InSynList$notes <- gsub(" \\| | \\||\\| ", "|", InSynList$notes)  
    # Remove these strings from the relevant columns now that they are saved as flags
    TempCols <- as.data.frame(apply(InSynList[c("canonical",FlagAnnot_cols) ], 2,
                                    function(y) gsub(FlagsAnnot, " ", y))) 
    # Replace existing columns with these new, trimmed, columns
    InSynList[c(FlagAnnot_cols)] <- TempCols[, FlagAnnot_cols]
    # Add the canonical column after the canonical column with flags - later convert to 
    # canonical and canonical_withFlags
    CanTest <- InSynList$canonical
    # IF there IS already a canonical column repalce that column with the new one...
    if(is.null(CanTest) == FALSE){
      InSynList$canonical <- TempCols$canonical
    }else{
      InSynList <- dplyr::mutate(InSynList, TempCols$canonical, .after = "canonical")
    }
    #  Clean up double spaces in data frame
    InSynList <- as.data.frame(apply(InSynList, 2, FUN = stringr::str_squish ), stringsAsFactors = FALSE)
  } # END IF notes col
  
  #### neither ####
  if(flagCol != "flags" & flagCol != "notes"){
    base::writeLines("!! The flagCol option must equal 'flags' or 'notes' !!")
  }
  
  #backupdf <- InSynList
  #InSynList <- backupdf
  
  # change column names
  #colnames(InSynList) <- c("flags","taxonomic_status","source","accid","id","kingdom","phylum","class","order","family","subfamily","tribe","subtribe","validName","canonical_withFlags","canonical","genus","subgenus","species","infraspecies","authorship")
  return(InSynList)
} # END FlagManager

# This function was written on the 25th of February to find manually identified outliers in 
  # bee data
# For questions, contact James B Dorey at jbdorey[at]me.com


#' Finds outliers, and their duplicates, as determined by experts
#' 
#' Uses expert-identified outliers with source spreadsheets that may be edited by users. The function 
#' will also use the duplicates file made using [BeeBDC::dupeSummary()] to identify duplicates of the 
#' expert-identified outliers and flag those as well.
#' The function will add a flagging column called `.expertOutlier` where records that are FALSE are
#' the expert outliers.
#'
#' @param data A data frame or tibble. Occurrence records as input.
#' @param DataPath A character path to the directory that contains the outlier spreadsheets.
#' @param PaigeOutliersName A character patch. Should lead to outlier spreadsheet from Paige Chesshire (csv file).
#' @param newOutliersName A character path. Should lead to appropriate outlier spreadsheet (xlsx file).
#' @param ColombiaOutliers_all A character path. Should lead to spreadsheet of bee outliers from Colombia (csv file).
#' @param duplicates A data frame or tibble. The duplicate file produced by [BeeBDC::dupeSummary()].
#' @param NearTRUE Optional. A character file name to the csv file. If you want to remove expert
#' outliers that are too close to TRUE points, use the name of the NearTRUE.csv.
#' Note: This implementation is only basic for now unless there is a greater need in the future.
#' @param NearTRUE_threshold Numeric. The threshold (in km) for the distance to TRUE points to 
#' keep expert outliers.
#'
#' @return Returns the data with a new column, `.expertOutlier` where records that are FALSE are
#' the expert outliers.
#' @export
#' 
#' @importFrom dplyr %>%
#'
#' @examples
#' \dontrun{
#'   # Read example data
#'   data(beesFlagged)
#' # Read in the most-recent duplicates file as well
#' if(!exists("duplicates")){
#'   duplicates <- fileFinder(path = DataPath,
#'                             fileName = "duplicateRun_") %>%
#'     readr::read_csv()}
#' # identify the outliers and get a list of their database_ids
#' beesFlagged_out <- manualOutlierFindeR(
#'   data = beesFlagged,
#'   DataPath = DataPath,
#'   PaigeOutliersName = "removedBecauseDeterminedOutlier.csv",
#'   newOutliersName = "^All_outliers_ANB_14March.xlsx",
#'   ColombiaOutliers_all = "All_Colombian_OutlierIDs.csv",
#'   duplicates = duplicates)
#' }
#' 
manualOutlierFindeR <- function(
    data = NULL,
    DataPath = NULL,
    PaigeOutliersName = "removedBecauseDeterminedOutlier.csv",
    newOutliersName = "All_outliers_ANB.xlsx",
    ColombiaOutliers_all = "All_Colombian_OutlierIDs.csv",
    duplicates = NULL,
    NearTRUE = NULL,
    NearTRUE_threshold = 5
    ){
  # locally bind variables to the function
  OutPath_Report <- eventDate <- near_truepoints_KM <- database_id <- occurrenceID <- NULL
  institutionCode <- database_id_keep <- catalogNumber <- .expertOutlier <- . <- NULL
  
  #### 0.0 Prep ####
  ##### 0.1 Errors ####
  ###### a. FATAL errors ####
  if(is.null(data)){
    stop(paste0(" - No data was given. Please specify the occurrence data."))
  }
  if(is.null(DataPath)){
    stop(paste0(" - No DataPath was given. Please specify the directory that contains the outliers."))
  }
  if(is.null(PaigeOutliersName)){
    stop(paste0(" - No PaigeOutliersName was given. Please specify the outliers' file name."))
  }
  if(is.null(newOutliersName)){
    stop(paste0(" - No newOutliersName was given. Please specify the outliers' file name."))
  }
  if(is.null(duplicates)){
    stop(paste0(" - No duplicates was given. Please provide the duplicates dataset as generated by jbd_dupeSummary."))
  }
  ###
  
  #### 1.0 Data prep ####
    ##### 1.1 Find data ####
  writeLines(" - Looking for the datasets...")
      ###### a. Paige outliers ####
  # Find the outliers from chesshire et al. 2023
  PaigeOutliers <- fileFinder(path = DataPath,
                               fileName = PaigeOutliersName) %>%
    readr::read_csv( col_types = readr::cols(.default = "c")) %>%
    suppressWarnings()
  
      ###### b. new outliers ####
  # Find the new outliers from the three sheets concatenates by Angela
  outliersAll <- fileFinder(path = DataPath,
                             fileName = newOutliersName) %>%
    openxlsx::read.xlsx("Outliers_FromCanadaToPanama_ANB") %>%
    dplyr::mutate(dplyr::across(tidyselect::all_of(colnames(.)), as.character)) %>%
    dplyr::bind_rows(fileFinder(path = DataPath,
                                 fileName = newOutliersName) %>%
                       openxlsx::read.xlsx("Tracys_outliers") %>%
                       dplyr::mutate(dplyr::across(tidyselect::all_of(colnames(.)), as.character))
                     ) %>%
    dplyr::bind_rows(fileFinder(path = DataPath,
                                 fileName = newOutliersName) %>%
                       openxlsx::read.xlsx("Colombian_outliers") %>%
                       dplyr::mutate(dplyr::across(tidyselect::all_of(colnames(.)), as.character))
                     ) %>%
    dplyr::bind_rows(fileFinder(path = DataPath,
                                 fileName = newOutliersName) %>%
                       openxlsx::read.xlsx("Outliers_SppInStatus3") %>%
                       dplyr::mutate(dplyr::across(tidyselect::all_of(colnames(.)), as.character))
                     ) %>%
    readr::write_excel_csv(paste(tempdir(), "newOutliers.csv", sep = "/"))
  # Read back in with the correct column classes
  outliersAll <- fileFinder(path = tempdir(),
                             fileName = "newOutliers.csv") %>%
    readr::read_csv(col_types = readr::cols(.default = "c"), lazy = FALSE) %>%
    dplyr::mutate(eventDate = eventDate %>%
                    lubridate::ymd_hms(truncated = 5)) %>%
    suppressWarnings()
  
  ###### c. Colombia ####
  ColombiaOutliers <- fileFinder(path = DataPath,
                                  fileName = ColombiaOutliers_all) %>%
    readr::read_csv( col_types = readr::cols(.default = "c")) %>%
    suppressWarnings()
  
  ######  d. remove NearTRUE ####
  # If user provies a NearTRUE input
  if(!is.null(NearTRUE)){
    # Find and read the csv
    NearTRUE_data <- fileFinder(path = DataPath,
                                 fileName = NearTRUE) %>% 
      readr::read_csv() %>% 
      dplyr::filter(near_truepoints_KM >= NearTRUE_threshold)
    
    # Remove those below a threshold in from NearTRUE in outliersAll
    outliersAll <- outliersAll %>%
      dplyr::filter(!database_id %in% NearTRUE_data$database_id)
    # Remove those below a threshold in from NearTRUE in ColombiaOutliers
    
    ColombiaOutliers <- ColombiaOutliers %>%
      dplyr::filter(!database_id %in% NearTRUE_data$database_id)
  }
  
  ###### e. eventDate ####
    # format data eventDate
  data <- data %>%
    dplyr::mutate(eventDate = eventDate %>%
                    lubridate::ymd_hms(truncated = 5))
  
  
    ##### 1.2 Process Paige ####
  writeLines(" - Processing the Paige outliers...")
  
  # Find PaigeOutliers in the occurrence data by occurrenceID and institutionCode
  Outl_occID <- data %>%
    tidyr::drop_na(occurrenceID) %>%
    dplyr::filter(occurrenceID %in% PaigeOutliers$occurrenceID &
                    institutionCode %in% PaigeOutliers$institutionCode) 
  # Find PaigeOutliers by occurrenceID and institutionCode
  Outliers_matched <- data %>%
    # Remove matched IDs
    dplyr::filter(!occurrenceID %in% Outl_occID$occurrenceID) %>%
    tidyr::drop_na(catalogNumber, institutionCode) %>%
    dplyr::filter(catalogNumber %in% PaigeOutliers$catalogNumber &
                    institutionCode %in% PaigeOutliers$institutionCode) %>%
    # Re-bind the outlier matches 
    dplyr::bind_rows(Outl_occID)

  # Combine the Paige and new outliers 
  outliersAll <- outliersAll %>%
      # Convert to the correct column types
    readr::type_convert(col_types = ColTypeR()) %>%
    dplyr::bind_rows(Outliers_matched) 

  
  #### 2.0 Find outlier duplicates ####
    ##### 2.1 Find duplicates ####
  writeLines(" - Looking for duplicates of the outliers...")
    # Get a list of the outliers and their duplicates
  outlierDuplicates <- duplicates %>%
    dplyr::filter(database_id %in% outliersAll$database_id | 
                    database_id_keep %in% outliersAll$database_id)
  duplicateList <- c(outlierDuplicates %>% dplyr::pull(database_id), 
                     outlierDuplicates %>% dplyr::pull(database_id_keep)) %>% unique()
  
    ##### 2.2 Combine duplicates ####
  # Get a list of all outliers and their duplicates - database_id
  outList <- c(outliersAll %>% dplyr::pull(database_id), duplicateList,
               ColombiaOutliers %>% pull(database_id)) %>% unique()
  
  #### 3.0 Flag records ####
  # Find the occurrences that did not match
  data <- data %>%
    # Add the .expertOutlier columns as TRUE (not flagged)
    dplyr::mutate(.expertOutlier = dplyr::if_else(
      database_id %in% outList,
      FALSE, TRUE)) 
  # Return user output
  message(
    paste(
      "\\manualOutlierFindeR:\n",
      "Flagged",
      format(sum(data$.expertOutlier == FALSE, na.rm = TRUE), big.mark = ","),
      "expert-identified outliers:\n",
      "The column '.expertOutlier' was added to the database.\n"
    )
  )
  
  # Return data
return(data)
  
}

# This function was written by James B Dorey on the 29th of September 2022
# Its purpose is to visualise some data spatially by country
# Please contact jbdorey[at]me.com for help


#' Create country-level summary maps of species and occurrence numbers
#' 
#' Builds an output figure that shows the number of species and the number of occurrences per 
#' country. Breaks the data into classes for visualisation. Users may filter data to their taxa 
#' of interest to produce figures of interest.
#'
#' @param data A data frame or tibble. Occurrence records as input.
#' @param class_n Numeric. The number of categories to break the data into.
#' @param class_Style Character. The class style passed to [classInt::classIntervals()]. Options are chosen 
#' style: one of "fixed", "sd", "equal", "pretty", "quantile", "kmeans", "hclust", "bclust", 
#' "fisher", "jenks", "dpih", "headtails", or "maximum". Default = "fisher"
#' @param outPath A character vector the path to the save location for the output figure.
#' @param fileName A character vector with file name 
#' for the output figure, ending with '.pdf'.
#' @param width Numeric. The width, in inches, of the resulting figure. Default = 5.
#' @param height Numeric. The height, in inches, of the resulting figure. Default = 10.
#' @param dpi Numeric. The resolution of the resulting plot. Default = 300.
#' @param returnPlot Logical. If TRUE, return the plot to the environment. Default = FALSE.
#' @param scale Numeric or character. Passed to rnaturalearth's ne_countries().
#' Scale of map to return, one of 110, 50, 10 or 'small', 'medium', 'large'. Default = 110.
#' @param pointBuffer Numeric. Amount to buffer points, in decimal degrees. If the point is outside 
#' of a country, but within this point buffer, it will count towards that country. It's a good idea
#' to keep this value consistent with the prior flags applied. Default = 0.01.
#'
#' @return Saves a figure to the user-specified outpath and name with a global map of bee 
#' occurrence species and count data from the input dataset. 
#' @export
#' 
#' @importFrom dplyr %>%
#' @importFrom sf sf_use_s2
#' @importFrom ggspatial north_arrow_fancy_orienteering
#' @importFrom grDevices gray
#' @importFrom ggplot2 xlab ylab ggtitle
#'
#' @examples
#' if(requireNamespace("rnaturalearthdata")){
#' # Read in data
#' data(beesFlagged)
#' OutPath_Figures <- tempdir()
#' # This simple example using the test data has very few classes due to the small amount of input 
#' # data.
#' summaryMaps(
#' data = beesFlagged,
#' width = 10, height = 10,
#' class_n = 4,
#' class_Style = "fisher",
#' outPath = OutPath_Figures,
#' fileName = paste0("CountryMaps_fisher_TEST.pdf"),
#' )
#' } # END if require
#' 
summaryMaps <- function(
    data = NULL,
    class_n = 15,
    class_Style = "fisher",
    outPath = NULL,
    fileName = NULL,
    width = 5, 
    height = 10,
    dpi = 300,
    returnPlot = FALSE,
    scale = 110,
    pointBuffer = 0.01
){
  # locally bind variables to the function
  name_long<-iso_a3<-name<-geometry<-decimalLongitude<-decimalLatitude<-database_id<-
    scientificName<-species<-country<-stateProvince<-dataSource<-count<-class_count<-
    class_count2<-occCount <- indexMatch <- . <- iso_a2 <- n<-NULL
  
  requireNamespace("dplyr")
  requireNamespace("classInt")
  requireNamespace("rnaturalearth")
  requireNamespace("ggspatial")
  
  
  #### 0.0 Prep ####
  ##### 0.1 errors ####
  ###### a. FATAL errors ####
  if(is.null(data)){
    stop(" - Please provide an argument for data I'm a program not a magician.")
  }
  if(is.null(fileName)){
    stop(" - No argument provided for fileName. Please provide a fileName.")
  }
  if(is.null(outPath)){
    stop(" - No argument provided for outPath Please provide an outPath.")
  }
  
  

  #### 1.0 Download base map ####
    ###### 1.1 naturalEarth ####
  # Download world map using rnaturalearth packages
  worldMap <- rnaturalearth::ne_countries(returnclass = "sf", country = NULL,
                                          type = "map_units", scale = scale) %>%
    # Select only a subset of the naturalearthdata columns to extract
    dplyr::select(name_long, iso_a3, iso_a2, name, name_long, geometry) %>%
    sf::st_make_valid()

  # This stops a plotting error
  sf::sf_use_s2(FALSE)
  
  
    ###### 1.2 occurrences ####
    # Filter the data columns
  data <- data %>%
    # Use a subset of columns
    dplyr::select(tidyselect::any_of(c("database_id", "scientificName", "species", 
                  "country", "stateProvince", "dataSource", "geometry",
                  "decimalLongitude", "decimalLatitude"))) %>% 
      # Drop the points without coordinates
    tidyr::drop_na(tidyselect::all_of(c("decimalLongitude", "decimalLatitude")))
  
  # Make all of the US virgin islands species into US species
    #  data$countryCode <- stringr::str_replace(string = data$countryCode, 
    #                                              pattern = "VI", replacement = "US")
  # Turn occData into a simple point feature
  dataPoints <- sf::st_as_sf(data,
                         coords = c("decimalLongitude", "decimalLatitude"),
                         na.fail = TRUE,
                         # Assign the CRS from the rnaturalearth map to the point data
                         crs = sf::st_crs(worldMap)) 
  
    ##### 1.3 Extraction ####
  writeLines(" - Extracting country data from points...")
  suppressWarnings({
    # Set geometries to constant for the sake of the map
  sf::st_agr(worldMap) = "constant"
  sf::st_agr(dataPoints) = "constant"
  
  # Simplify the world map ONCE to be used later
  simplePoly <- worldMap %>% sf::st_drop_geometry() %>%
    dplyr::mutate(indexMatch = dplyr::row_number())
  
    #Extract polygon information to points
  extracted <- sf::st_intersects(dataPoints, worldMap, sparse = TRUE) %>% 
    # return a tibble with the index of each match or NA where there was no match
    dplyr::tibble(indexMatch = .) 
  # If first element is full, unlist each one
  extracted <- extracted %>%
    dplyr::mutate(indexMatch = indexMatch %>% as.character() %>%
                      # deal with problems - Take the first number where two are provided
                    stringr::str_extract("[0-9]+") %>% 
                      # Remove zero to NA
                    stringr::str_replace("^[0]$", NA_character_) %>%
                      # Make numeric
                    as.numeric()
                    ) %>%
    # drop geometry
    sf::st_drop_geometry() 
})
# rejoin
data <- extracted %>%
  dplyr::left_join(simplePoly,
                   by = "indexMatch") %>%
  # Add in the database_id
  dplyr::bind_cols(data)

rm(extracted)
  
  writeLines("Extraction complete.")
  
  ##### 1.4 Buffer fails ####
  writeLines(" - Buffering naturalearth map by pointBuffer...")
    ###### a. buffer map ####
  # Buffer the natural earth map
  suppressWarnings({
    worldMap <- worldMap %>% 
      sf::st_buffer(dist = pointBuffer)
  })
    ###### b. extract fails ####
  
  #Extract polygon information to points
  suppressWarnings({
  extracted2 <- sf::st_intersects(dataPoints %>% dplyr::filter(
    database_id %in% (data %>% dplyr::filter(is.na(name_long)) %>% dplyr::pull(database_id))),
                                 worldMap %>% sf::st_make_valid(), sparse = TRUE) %>% 
    # return a tibble with the index of each match or NA where there was no match
    dplyr::tibble(indexMatch = .) 
  # If first element is full, unlist each one
  extracted2 <- extracted2 %>%
    dplyr::mutate(indexMatch = indexMatch %>% as.character() %>%
                    # deal with problems - Take the first number where two are provided
                    stringr::str_extract("[0-9]+") %>% 
                    # Remove zero to NA
                    stringr::str_replace("^[0]$", NA_character_) %>%
                    # Make numeric
                    as.numeric()
    ) %>%
    # drop geometry
    sf::st_drop_geometry() 
})
# rejoin
extracted2 <- extracted2 %>%
  dplyr::left_join(simplePoly,
                   by = "indexMatch") %>%
  # Add in the database_id
  dplyr::bind_cols(data %>% dplyr::filter(is.na(name_long)) %>% 
                     dplyr::select(!tidyselect::any_of(c("name_long", "iso_a3", "indexMatch",
                                                         "iso_a2", "name"))))

  # Rejoin with data
data <- data %>%
  dplyr::filter(!database_id %in% extracted2$database_id) %>%
  dplyr::bind_rows(extracted2)

rm(extracted2)
  
  
  #### 2.0 Species map ####
    ##### 2.1 Data prep ####
  # Get the unique country-species pairs
  spMapData <- data %>%
    dplyr::distinct(scientificName, name_long, .keep_all = TRUE) %>%
    # Group by the country
    dplyr::group_by(name_long) %>%
    # Get a count of the records per country
    dplyr::mutate(count = n()) %>% dplyr::ungroup() %>%
    # Get unique
    dplyr::distinct(name_long, .keep_all = TRUE) 
  
  ##### 2.2 Breaks ####
  # make class intervals.
  # Class intervals from ?classIntervals: fixed", "sd", "equal", "pretty", "quantile", 
  # "kmeans", "hclust", "bclust", "fisher", "jenks", "dpih" or "headtails"
  classes <- classInt::classIntervals(spMapData$count, n = class_n, 
                                      style = class_Style, dig.lab=20,
                                      dataPrecision=0)
  # Next we'll create a new column in our sf object using the base R cut() function to cut up our 
  # percent variable into distinct groups:
  spMapData <- spMapData %>%
    dplyr::mutate(class_count = cut(count, 
                                    classes$brks %>% round(digits = 0),
                                    include.lowest = T, dig.lab = 10)) %>%
      # format the class_count column to remove spaces, add comma break, and join min and max
    dplyr::mutate(class_count2 = class_count %>%
                    stringr::str_remove("\\[|\\]|\\(|\\)") %>%
                    stringr::str_remove("\\]") %>%
                    stringr::str_replace(",", "-")) %>%
    tidyr::separate(col = class_count2, into = c("min", "max"), sep = "-") %>%
    dplyr::mutate(min = min %>% as.numeric() %>% format(big.mark = ",") %>% 
                    stringr::str_remove("\\s+"),
                  max = max %>% as.numeric() %>% format(big.mark = ",") %>% 
                    stringr::str_remove("\\s+"),
                  class_count2 = stringr::str_c(min, max, sep = "-") ) 
  
  # Join the map and occurrence data
  fullMap <-  dplyr::full_join(worldMap,  spMapData %>% sf::st_drop_geometry(),
                               by = c("name_long" = "name_long")) %>%
    # Remove na rows
    tidyr::drop_na(count)
  
  ##### 2.3 Draw map ####
  # Make the map
  (spCountryMap <- ggplot2::ggplot(data = fullMap, ) +
     # Add in a blank base-map to highlight countries with no data
     ggplot2::geom_sf(data = worldMap, size = 0.15, fill = "white")+ 
      # Plot and colour the terrestrial base map
      ggplot2::geom_sf(ggplot2::aes(fill = class_count), size = 0.15)+ 
      # Set map limits, if wanted
      ggplot2::coord_sf(expand = FALSE, ylim = c(-60,90), lims_method = "geometry_bbox") + 
      # Map formatting
      # Add in the map's north arrow
      ggspatial::annotation_north_arrow(location = "tl", which_north = "true", 
                             pad_x = ggplot2::unit(0.1, "cm"), pad_y = ggplot2::unit(0.1, "cm"), 
                             style = ggspatial::north_arrow_fancy_orienteering()) + # Add in NORTH ARROW
      ggplot2::theme(panel.grid.major = ggplot2::element_line(color = grDevices::gray(.1, alpha = 0.1), 
                                            linetype = "dashed", linewidth = 0.5), # Add grid lines
            panel.border = ggplot2::element_rect(color = grDevices::gray(.1, alpha = 1), 
                                        linetype = "solid", linewidth = 0.5,
                                        fill = NA), # add panel border
            # Add background - colour in the ocean
            panel.background = ggplot2::element_rect(fill = "aliceblue") )+ 
      # Change map colour scheme - CHOOSE YOUR OWN ADVENTURE
      # For Dorey colour scheme use the below
      ggplot2::scale_fill_viridis_d(option = "inferno",
                           na.value = "grey50",
                           name = "Class count",
                           labels = fullMap %>%
                             dplyr::arrange(count) %>% 
                             dplyr::distinct(class_count2) %>%
                             # options = "magma", "inferno", "plasma", "cividis"
                             dplyr::pull(class_count2)) + 
      # Add in X and Y labels
     ggplot2::xlab("Longitude") + ggplot2::ylab("Latitude") + 
      # Add in the title
     ggplot2::ggtitle( "Number of species per country")  )
  
  rm(spMapData)
  
  
  #### 3.0 occurrence map ####
  ##### 2.1 data prep ####
  # Get the unique county-database_id pairs
  mapTable <- data %>%
      # Group by the country
    dplyr::group_by(name_long) %>%
      # Get a count of the records per country
    dplyr::mutate(occCount = dplyr::n()) %>%
    # Get unique
    dplyr::distinct(name_long, .keep_all = TRUE) %>%
      # Select only these columns
    dplyr::select(name_long, occCount)
  
  # Join the map and occurrence data
  fullMap <-  dplyr::full_join(worldMap,  mapTable  %>% sf::st_drop_geometry(),
                               by = c("name_long" = "name_long")) %>%
    # Remove na rows
    tidyr::drop_na(occCount)
  
  ##### 2.2 Breaks ####
  # make class intervals.
  # Class intervals from ?classIntervals: fixed", "sd", "equal", "pretty", "quantile", 
  # "kmeans", "hclust", "bclust", "fisher", "jenks", "dpih" or "headtails"
  classes <- classInt::classIntervals(fullMap$occCount, n = class_n, 
                                      style = class_Style, dig.lab=20,
                                      dataPrecision=0)
  # Next we'll create a new column in our sf object using the base R cut() function to cut up our 
  # percent variable into distinct groups:
  fullMap <- fullMap %>%
    dplyr::mutate(class_count = cut(occCount, 
                                    classes$brks %>% round(digits = 0),
                                    include.lowest = T, dig.lab = 10)) %>%
  # format the class_count column to remove spaces, add comma break, and join min and max
  dplyr::mutate(class_count2 = class_count %>%
                  stringr::str_remove("\\[|\\]|\\(|\\)") %>%
                  stringr::str_remove("\\]") %>%
                  stringr::str_replace(",", "-")) %>%
    tidyr::separate(col = class_count2, into = c("min", "max"), sep = "-") %>%
    dplyr::mutate(min = min %>% as.numeric() %>% format(big.mark = ",") %>% 
                    stringr::str_remove("\\s+"),
                  max = max %>% as.numeric() %>% format(big.mark = ",") %>% 
                    stringr::str_remove("\\s+"),
                  class_count2 = stringr::str_c(min, max, sep = "-") ) 
  
  ##### 2.3 Draw map ####
  # Make the map
  (occCountryMap <- ggplot2::ggplot(data = fullMap) +
      # Add in a blank base-map to highlight countries with no data
     ggplot2::geom_sf(data = worldMap, size = 0.15, fill = "white")+ 
      # Plot and colour the terrestrial base map
      ggplot2::geom_sf(ggplot2::aes(fill = class_count), size = 0.15)+ 
      # Set map limits, if wanted
     ggplot2::coord_sf(expand = FALSE, ylim = c(-60,90), lims_method = "geometry_bbox") + 
     # Map formatting
      # Add in the map's north arrow
      ggspatial::annotation_north_arrow(location = "tl", which_north = "true", 
                                        pad_x = ggplot2::unit(0.1, "cm"), pad_y = ggplot2::unit(0.1, "cm"), 
                                        style = ggspatial::north_arrow_fancy_orienteering()) + # Add in NORTH ARROW
      ggplot2::theme(panel.grid.major = ggplot2::element_line(color = grDevices::gray(.1, alpha = 0.1), 
                                                              linetype = "dashed", 
                                                              linewidth = 0.5), # Add grid lines
                     panel.border = ggplot2::element_rect(color = grDevices::gray(.1, alpha = 1), 
                                                          linetype = "solid", linewidth = 0.5,
                                                          fill = NA), # add panel border
                     # Add background - colour in the ocean
                     panel.background = ggplot2::element_rect(fill = "aliceblue") )+ 
      # Change map colour scheme - CHOOSE YOUR OWN ADVENTURE
      # For Dorey colour scheme use the below
      ggplot2::scale_fill_viridis_d(option = "inferno",
                                    na.value = "grey50",
                                    name = "Class count",
                                    labels = fullMap %>%
                                      dplyr::arrange(occCount) %>% 
                                      dplyr::distinct(class_count2) %>%
                                      # options = "magma", "inferno", "plasma", "cividis"
                                      dplyr::pull(class_count2)) + 
      # Add in X and Y labels
      ggplot2::xlab("Longitude") + ggplot2::ylab("Latitude") + 
      # Add in the title
     ggplot2::ggtitle( "Number of occurrences per country")  )
  
  #### 4.0 combine + save ####
  # plot the figures together
  (combinedPlot <- cowplot::plot_grid(spCountryMap,
                                  #    + 
                                  # theme(legend.position = legend.position,
                                  #       legend.title = element_blank()),
                                  occCountryMap, 
                                  labels = c("(a)","(b)"),
                                 ncol = 1, align = 'v', axis = 'l'))
  # Save the plot
  cowplot::save_plot(filename = paste(outPath, "/", fileName, sep = ""),
                     plot = combinedPlot,
                     base_width = width,
                     base_height = height, dpi = dpi)
  if(returnPlot == TRUE){
  return(combinedPlot)}
  
  
} # END function

# If rnaturalearthdata is present, run tests
if(requireNamespace("rnaturalearthdata")){
  # Load in example data
beesFlagged <- BeeBDC::beesFlagged


# Draw a global summary map for occurrence and species number by country
testMap <- BeeBDC::summaryMaps(
  data = beesFlagged,
  width = 10, height = 10,
  class_n = 3,
  class_Style = "fisher",
  outPath = tempdir(),
  fileName = "CountryMaps_fisher.pdf",
  returnPlot = TRUE,
  scale = 110
)


# Test class
   testthat::test_that("summaryMaps expected class", {
     testthat::expect_type(testMap, "object")
   })

   testthat::test_that("dupePlotR plot saved?", {
     testthat::expect_true(inherits(testMap, c("ggplot", "ggplot2::ggplot")))
   })


  # Test some internals
testthat::test_that("summaryMaps internals - testMap$layers[[1]]$geom_params$grob$name", {
  testthat::expect_equal(testMap$layers[[1]]$geom_params$grob$name, "layout")
})


} # END if require

# This function was written by James B Dorey on 8th November 2022 to create a .summary column and
  # replace the bdc_summary_col function which fails with NA values

#' Create or update the .summary flag column
#' 
#' Using all flag columns (column names starting with "."), this function either creates or updates
#' the .summary flag column which is FALSE when ANY of the flag columns are FALSE. Columns can be excluded
#'  and removed after creating the .summary column. Additionally, the occurrence dataset
#' can be filtered to only those where .summary = TRUE at the end of the function.
#'
#' @param data A data frame or tibble. Occurrence records to use as input.
#' @param dontFilterThese A character vector of flag columns to be ignored in the creation or updating
#'  of the .summary column. Cannot be specified with onlyFilterThese.
#' @param onlyFilterThese A character vector. The inverse of dontFilterThese, where columns identified
#' here will be filtered and no others. Cannot be specified with dontFilterThese.
#' @param removeFilterColumns Logical. If TRUE all columns starting with "." will be removed in the 
#' output data. This only makes sense to use when filterClean = TRUE. Default = FALSE.
#' @param filterClean Logical. If TRUE, the data will be filtered to only those occurrence where .summary
#' = TRUE (i.e., completely clean according to the used flag columns). Default = FALSE.
#'
#' @return Returns a data frame or tibble of the input data but modified based on the above parameters.
#' @export
#' 
#' @importFrom dplyr %>%
#' 
#'
#' @examples
#' # Read in example data
#' data(beesFlagged)
#' 
#' # To only update the .summary column
#' beesFlagged_out <- summaryFun(
#'     data = beesFlagged,
#'     dontFilterThese = c(".gridSummary", ".lonFlag", ".latFlag", ".uncer_terms", ".unLicensed"),
#'     removeFilterColumns = FALSE,
#'     filterClean = FALSE)
#'   # View output
#' table(beesFlagged_out$.summary, useNA = "always")
#' 
#' # Now filter to only the clean data and remove the flag columns
#' beesFlagged_out <- summaryFun(
#'   data = BeeBDC::beesFlagged,
#'   dontFilterThese = c(".gridSummary", ".lonFlag", ".latFlag", ".uncer_terms", ".unLicensed"),
#'   removeFilterColumns = TRUE,
#'   filterClean = TRUE)
#' # View output
#' table(beesFlagged_out$.summary, useNA = "always")
#' 
#' 
#' 

summaryFun <- function(
    data = NULL,
    dontFilterThese = NULL,
    onlyFilterThese = NULL,
    removeFilterColumns = FALSE,
    filterClean = FALSE){
  # locally bind variables to the function
  . <- rowSum <- .summaryNew <- .summary <- NULL
  
  #### 0.0 Prep ####
  if(is.null(data)){
    stop("You must provide a dataset in the 'data' argument.")
  }
  if(!is.null(dontFilterThese) & !is.null(onlyFilterThese)){
    stop("Please only choose dontFilterThese OR onlyFilterThese.")
  }
  
  ##### 0.1 onlyFilter to dontFilter ####
    # In order to use onlyFilterThese, simply transform it to dontFilterThese
  if(!is.null(onlyFilterThese)){
    dontFilterThese <- data %>%
      # Select all columns starting with "."
      dplyr::select(tidyselect::starts_with(".")) %>%
      colnames(.) %>%
      setdiff(., onlyFilterThese)
  }
  
  
  #### 1.0 Generate .summary column ####
    ##### 1.1 dontFilterThese present ####
    # User output
  if(!is.null(dontFilterThese)){
    writeLines(paste0(" - We will NOT flag the following columns. However, they will remain",
                      " in the data file.\n",
                      paste(dontFilterThese, collapse = ", ") ))
      # Run function
    dataOut <-
      data %>%
      # Which columns NOT to filter
      dplyr::select(!tidyselect::any_of(dontFilterThese)) %>%
      # Update .summary column
      # Select all columns starting with "."
      dplyr::select(tidyselect::starts_with(".")) %>% 
      # Delete the summary column if it's there
      dplyr::select(!tidyselect::starts_with(".summary")) %>%
      # Make FALSE == 1 and TRUE == 0
      dplyr::mutate_if(is.logical, ~abs(as.numeric(.) - 1)) %>%
      # IF rowSum > 0 then there is at least one flag
      dplyr::mutate(rowSum = base::rowSums(., na.rm = TRUE)) %>%
      # Add the .summary column
      dplyr::mutate(.summaryNew = dplyr::if_else(rowSum > 0,
                                                 FALSE, TRUE)) %>%
      dplyr::select(.summaryNew) %>%
      dplyr::bind_cols(data, .) %>%
      dplyr::mutate(.summary = .summaryNew) %>% dplyr::select(!.summaryNew) 
  }
  
  ##### 1.2 dontFilterThese NULL ####
  if(is.null(dontFilterThese)){
    writeLines(paste0(" - We will flag all columns starting with '.'"))
      # Run function
    dataOut <-
      data %>%
      # Update .summary column
      # Select all columns starting with "."
      dplyr::select(tidyselect::starts_with(".")) %>% 
      # Delete the summary column if it's there
      dplyr::select(!tidyselect::starts_with(".summary")) %>%
      # Make FALSE == 1 and TRUE == 0
      dplyr::mutate_if(is.logical, ~abs(as.numeric(.) - 1)) %>%
      # IF rowSum > 0 then there is at least one flag
      dplyr::mutate(rowSum = rowSums(., na.rm = TRUE)) %>%
      # Add the .summary column
      dplyr::mutate(.summaryNew = dplyr::if_else(rowSum > 0,
                                                 FALSE, TRUE)) %>%
      dplyr::select(.summaryNew) %>%
      dplyr::bind_cols(data, .) %>%
      dplyr::mutate(.summary = .summaryNew) %>% dplyr::select(!.summaryNew) 
  }
  
  ##### 1.3 User message ####
  message(paste(" - summaryFun:\nFlagged", 
                format(sum(dataOut$.summary == FALSE, na.rm = TRUE), big.mark = ","),
                "\n ",
                "The .summary column was added to the database.",
                sep = " "))
  

  #### 2.0 Optional extras ####
  ##### 2.1 Filter for clean ####
  # RFilter for only clean records here if user specifies
  if(filterClean == TRUE){
    dataOut <- dataOut   %>%
      # FILTER HERE
      dplyr::filter(.summary == TRUE) 
    message(paste(" - REMOVED all occurrences that were FALSE for the 'summary' column.")) 
  }
  
  ##### 2.2 Remove filtering columns ####
  # Remove filtering columns if user specifies
  if(removeFilterColumns == TRUE){
    dataOut <- dataOut %>%
      dplyr::select(!tidyselect::starts_with("."))
  }


#### 3.0 Output ####
return(dataOut)
} # End function

requireNamespace("BeeBDC")
requireNamespace("dplyr")


## test data - these have been abbreviated for sanity and also edited. And duplicated. They're not totally real
testData <- dplyr::tribble(
                     ~family,     ~subfamily,            ~genus,                    ~species, ~specificEpithet, ~taxonRank, ~typeStatus, ~decimalLatitude, ~decimalLongitude,       ~stateProvince,                                                       ~locality, ~county, ~municipality,    ~license,                                                                                                                  ~issue, ~eventDate, ~day, ~month, ~year,       ~basisOfRecord, ~country, ~type, ~occurrenceStatus,                                  ~recordedBy, ~coordinatePrecision, ~coordinateUncertaintyInMeters, ~spatiallyValid,      ~catalogNumber, ~institutionCode,       ~datasetName,            ~verbatimScientificName, ~verbatimEventDate,     ~sex, ~rights,                        ~rightsHolder, ~hasCoordinate, ~hasGeospatialIssues, ~occurrenceYear, ~duplicateStatus, ~verbatim_scientificName, ~verbatimLatitude, ~verbatimLongitude, ~.scientificName_empty, ~.coordinates_empty, ~.coordinates_outOfRange, ~.basisOfRecords_notStandard, ~country_suggested, ~countryCode, ~coordinates_transposed, ~.coordinates_country_inconsistent, ~.occurrenceAbsent, ~.unLicensed, ~.GBIFflags, ~.uncer_terms,               ~names_clean, ~.invalidName, ~.rou, ~.val, ~.equ, ~.zer, ~.cap, ~.cen, ~.gbf, ~.inst, ~.sequential, ~.lonFlag, ~.latFlag, ~.gridSummary, ~.uncertaintyThreshold, ~countryMatch, ~.countryOutlier, ~.sea, ~.eventDate_empty, ~.year_outOfRange, ~.duplicates, ~.expertOutlier, ~postal, ~stateMatch, ~stateMatch_certainty, ~.stateOutlier,
              "Megachilidae", "Megachilinae", "Pseudoanthidium", "Pseudoanthidium truncatum",      "truncatum",  "SPECIES",  "HOLOTYPE",               NA,                NA,             "Kigoma",                                                              NA,      NA,       "Ujiji", "CC_BY_4_0", "OCCURRENCE_STATUS_INFERRED_FROM_INDIVIDUAL_COUNT;TAXON_MATCH_HIGHERRANK;INSTITUTION_MATCH_FUZZY;COLLECTION_MATCH_NONE",  "8/15/31",  15L,     8L, 1931L, "PRESERVED_SPECIMEN",       NA,    NA,         "PRESENT",                              "James Ogilvie",                   NA,                             NA,              NA, "AMNH_IZC 00323864",           "AMNH", "AMNH Hymenoptera", "Pachyanthidium truncatum (Smith)",          "8/15/31", "FEMALE",      NA, "American Museum of Natural History",          FALSE,                FALSE,              NA, "Kept duplicate",                       NA,                NA,                 NA,                   TRUE,               FALSE,                     TRUE,                         TRUE,                 NA,           NA,                    TRUE,                               TRUE,               TRUE,         TRUE,        TRUE,          TRUE, "Pachyanthidium truncatum",          TRUE,  TRUE,  TRUE,  TRUE,  TRUE,  TRUE,  TRUE,  TRUE,   TRUE,         TRUE,      TRUE,      TRUE,          TRUE,                   TRUE,          TRUE,             TRUE,  TRUE,              TRUE,              TRUE,         TRUE,            TRUE,      NA,        TRUE,                  TRUE,           TRUE,
                "Halictidae",   "Halictinae",        "Halictus",          "Halictus simplex",        "simplex",    "GENUS",  "HOLOTYPE",               NA,                NA,                   NA,                                                 "near Weing Sa",      NA,            NA, "CC_BY_4_0", "OCCURRENCE_STATUS_INFERRED_FROM_INDIVIDUAL_COUNT;TAXON_MATCH_HIGHERRANK;INSTITUTION_MATCH_FUZZY;COLLECTION_MATCH_NONE", "12/25/27",  25L,    12L, 1927L, "PRESERVED_SPECIMEN",       NA,    NA,         "PRESENT",                         "T. D. A. Cockerell",                   NA,                             NA,              NA, "AMNH_IZC 00323581",           "AMNH", "AMNH Hymenoptera",                         "Halictus",         "12/25/27", "FEMALE",      NA, "American Museum of Natural History",          FALSE,                FALSE,              NA, "Kept duplicate",                       NA,                NA,                 NA,                   TRUE,                TRUE,                    FALSE,                         TRUE,                 NA,           NA,                    TRUE,                               TRUE,               TRUE,         TRUE,        TRUE,          TRUE,                 "Halictus",          TRUE,  TRUE,  TRUE,  TRUE,  TRUE,  TRUE,  TRUE,  TRUE,   TRUE,         TRUE,      TRUE,      TRUE,          TRUE,                   TRUE,          TRUE,             TRUE,  TRUE,              TRUE,              TRUE,         TRUE,            TRUE,      NA,        TRUE,                  TRUE,           TRUE,
                "Halictidae",   "Halictinae",        "Halictus",          "Halictus simplex",        "simplex",    "GENUS",  "HOLOTYPE",               NA,                NA,       "Saskatchewan", "Lake Waskesiu, at the beginning of the portage to Heart Lakes",      NA,            NA, "CC_BY_4_0", "OCCURRENCE_STATUS_INFERRED_FROM_INDIVIDUAL_COUNT;TAXON_MATCH_HIGHERRANK;INSTITUTION_MATCH_FUZZY;COLLECTION_MATCH_NONE",  "8/31/36",  31L,     8L, 1936L, "PRESERVED_SPECIMEN",       NA,    NA,         "PRESENT", "T. D. A. Cockerell & Wilmatte P. Cockerell",                   NA,                             NA,              NA, "AMNH_IZC 00323583",           "AMNH", "AMNH Hymenoptera",                         "Halictus",          "8/31/36",   "MALE",      NA, "American Museum of Natural History",          FALSE,                FALSE,              NA, "Kept duplicate",                       NA,                NA,                 NA,                   TRUE,                TRUE,                     TRUE,                        FALSE,                 NA,           NA,                    TRUE,                               TRUE,               TRUE,         TRUE,        TRUE,          TRUE,                 "Halictus",          TRUE,  TRUE,  TRUE,  TRUE,  TRUE,  TRUE,  TRUE,  TRUE,   TRUE,         TRUE,      TRUE,      TRUE,          TRUE,                   TRUE,          TRUE,             TRUE,  TRUE,              TRUE,              TRUE,         TRUE,            TRUE,      NA,        TRUE,                  TRUE,           TRUE,
                "Halictidae",   "Halictinae",        "Halictus",          "Halictus simplex",        "simplex",    "GENUS",  "HOLOTYPE",               NA,                NA,     "North Carolina",                                                              NA,      NA,     "Raleigh", "CC_BY_4_0", "OCCURRENCE_STATUS_INFERRED_FROM_INDIVIDUAL_COUNT;TAXON_MATCH_HIGHERRANK;INSTITUTION_MATCH_FUZZY;COLLECTION_MATCH_NONE",  "5/29/24",  29L,     5L, 1924L, "PRESERVED_SPECIMEN",       NA,    NA,         "PRESENT",                              "C. S. Brimley",                   NA,                             NA,              NA, "AMNH_IZC 00324010",           "AMNH", "AMNH Hymenoptera",                         "Halictus",          "5/29/24", "FEMALE",      NA, "American Museum of Natural History",          FALSE,                FALSE,              NA, "Kept duplicate",                       NA,                NA,                 NA,                   TRUE,                TRUE,                     TRUE,                         TRUE,                 NA,           NA,                   FALSE,                               TRUE,               TRUE,         TRUE,        TRUE,          TRUE,                 "Halictus",          TRUE,  TRUE,  TRUE,  TRUE,  TRUE,  TRUE,  TRUE,  TRUE,   TRUE,         TRUE,      TRUE,      TRUE,          TRUE,                   TRUE,          TRUE,             TRUE,  TRUE,              TRUE,              TRUE,         TRUE,            TRUE,      NA,        TRUE,                  TRUE,           TRUE,
                "Halictidae",   "Halictinae",        "Halictus",          "Halictus simplex",        "simplex",    "GENUS",  "HOLOTYPE",               NA,                NA,       "Saskatchewan",                     "Prince Albert National Park, Lake Ajawaan",      NA,            NA, "CC_BY_4_0", "OCCURRENCE_STATUS_INFERRED_FROM_INDIVIDUAL_COUNT;TAXON_MATCH_HIGHERRANK;INSTITUTION_MATCH_FUZZY;COLLECTION_MATCH_NONE",  "8/26/36",  26L,     8L, 1936L, "PRESERVED_SPECIMEN",       NA,    NA,         "PRESENT",                         "T. D. A. Cockerell",                   NA,                             NA,              NA, "AMNH_IZC 00323630",           "AMNH", "AMNH Hymenoptera",                         "Halictus",          "8/26/36", "FEMALE",      NA, "American Museum of Natural History",          FALSE,                FALSE,              NA, "Kept duplicate",                       NA,                NA,                 NA,                   TRUE,                TRUE,                     TRUE,                         TRUE,                 NA,           NA,                    TRUE,                              FALSE,               TRUE,         TRUE,        TRUE,          TRUE,                 "Halictus",          TRUE,  TRUE,  TRUE,  TRUE,  TRUE,  TRUE,  TRUE,  TRUE,   TRUE,         TRUE,      TRUE,      TRUE,          TRUE,                   TRUE,          TRUE,             TRUE,  TRUE,              TRUE,              TRUE,         TRUE,            TRUE,      NA,        TRUE,                  TRUE,           TRUE,
                "Halictidae",   "Halictinae",        "Halictus",          "Halictus simplex",        "simplex",    "GENUS",  "HOLOTYPE",               NA,                NA,                "Nan",                                                              NA,      NA,         "Nan", "CC_BY_4_0", "OCCURRENCE_STATUS_INFERRED_FROM_INDIVIDUAL_COUNT;TAXON_MATCH_HIGHERRANK;INSTITUTION_MATCH_FUZZY;COLLECTION_MATCH_NONE", "12/31/27",  31L,    12L, 1927L, "PRESERVED_SPECIMEN",       NA,    NA,         "PRESENT",                         "T. D. A. Cockerell",                   NA,                             NA,              NA, "AMNH_IZC 00323584",           "AMNH", "AMNH Hymenoptera",                         "Halictus",         "12/31/27", "FEMALE",      NA, "American Museum of Natural History",          FALSE,                FALSE,              NA, "Kept duplicate",                       NA,                NA,                 NA,                   TRUE,                TRUE,                     TRUE,                         TRUE,                 NA,           NA,                    TRUE,                               TRUE,              FALSE,         TRUE,        TRUE,          TRUE,                 "Halictus",          TRUE,  TRUE,  TRUE,  TRUE,  TRUE,  TRUE,  TRUE,  TRUE,   TRUE,         TRUE,      TRUE,      TRUE,          TRUE,                   TRUE,          TRUE,             TRUE,  TRUE,              TRUE,              TRUE,         TRUE,            TRUE,      NA,        TRUE,                  TRUE,           TRUE,
                "Halictidae",   "Halictinae",        "Halictus",          "Halictus simplex",        "simplex",    "GENUS",  "HOLOTYPE",               NA,                NA, "MeknÔøΩs-Tafilalet",                                                              NA,      NA,      "Ifrane", "CC_BY_4_0", "OCCURRENCE_STATUS_INFERRED_FROM_INDIVIDUAL_COUNT;TAXON_MATCH_HIGHERRANK;INSTITUTION_MATCH_FUZZY;COLLECTION_MATCH_NONE",  "8/25/31",  25L,     8L, 1931L, "PRESERVED_SPECIMEN",       NA,    NA,         "PRESENT",       "T. D. A. Cockerell & Alice M. Mackie",                   NA,                             NA,              NA, "AMNH_IZC 00323616",           "AMNH", "AMNH Hymenoptera",                         "Halictus",          "8/25/31", "FEMALE",      NA, "American Museum of Natural History",          FALSE,                FALSE,              NA, "Kept duplicate",                       NA,                NA,                 NA,                   TRUE,                TRUE,                     TRUE,                         TRUE,                 NA,           NA,                    TRUE,                               TRUE,               TRUE,        FALSE,        TRUE,          TRUE,                 "Halictus",          TRUE,  TRUE,  TRUE,  TRUE,  TRUE,  TRUE,  TRUE,  TRUE,   TRUE,         TRUE,      TRUE,      TRUE,          TRUE,                   TRUE,          TRUE,             TRUE,  TRUE,              TRUE,              TRUE,         TRUE,            TRUE,      NA,        TRUE,                  TRUE,           TRUE,
                "Halictidae",   "Halictinae",        "Halictus",          "Halictus simplex",        "simplex",    "GENUS",  "HOLOTYPE",               NA,                NA, "MeknÔøΩs-Tafilalet",                                                              NA,      NA,      "Ifrane", "CC_BY_4_0", "OCCURRENCE_STATUS_INFERRED_FROM_INDIVIDUAL_COUNT;TAXON_MATCH_HIGHERRANK;INSTITUTION_MATCH_FUZZY;COLLECTION_MATCH_NONE",  "8/26/30",  26L,     8L, 1930L, "PRESERVED_SPECIMEN",       NA,    NA,         "PRESENT",                            "Alice M. Mackie",                   NA,                             NA,              NA, "AMNH_IZC 00323619",           "AMNH", "AMNH Hymenoptera",                         "Halictus",          "8/26/30",   "MALE",      NA, "American Museum of Natural History",          FALSE,                FALSE,              NA, "Kept duplicate",                       NA,                NA,                 NA,                   TRUE,                TRUE,                     TRUE,                         TRUE,                 NA,           NA,                    TRUE,                               TRUE,               TRUE,         TRUE,       FALSE,          TRUE,                 "Halictus",          TRUE,  TRUE,  TRUE,  TRUE,  TRUE,  TRUE,  TRUE,  TRUE,   TRUE,         TRUE,      TRUE,      TRUE,          TRUE,                   TRUE,          TRUE,             TRUE,  TRUE,              TRUE,              TRUE,         TRUE,            TRUE,      NA,        TRUE,                  TRUE,           TRUE,
              "Megachilidae", "Megachilinae", "Pseudoanthidium", "Pseudoanthidium truncatum",      "truncatum",  "SPECIES",  "HOLOTYPE",               NA,                NA,             "Kigoma",                                                              NA,      NA,       "Ujiji", "CC_BY_4_0", "OCCURRENCE_STATUS_INFERRED_FROM_INDIVIDUAL_COUNT;TAXON_MATCH_HIGHERRANK;INSTITUTION_MATCH_FUZZY;COLLECTION_MATCH_NONE",  "8/15/31",  15L,     8L, 1931L, "PRESERVED_SPECIMEN",       NA,    NA,         "PRESENT",                              "James Ogilvie",                   NA,                             NA,              NA, "AMNH_IZC 00323864",           "AMNH", "AMNH Hymenoptera", "Pachyanthidium truncatum (Smith)",          "8/15/31", "FEMALE",      NA, "American Museum of Natural History",          FALSE,                FALSE,              NA, "Kept duplicate",                       NA,                NA,                 NA,                   TRUE,                TRUE,                     TRUE,                         TRUE,                 NA,           NA,                    TRUE,                               TRUE,               TRUE,         TRUE,        TRUE,         FALSE, "Pachyanthidium truncatum",          TRUE,  TRUE,  TRUE,  TRUE,  TRUE,  TRUE,  TRUE,  TRUE,   TRUE,         TRUE,      TRUE,      TRUE,          TRUE,                   TRUE,          TRUE,             TRUE,  TRUE,              TRUE,              TRUE,         TRUE,            TRUE,      NA,        TRUE,                  TRUE,           TRUE,
                "Halictidae",   "Halictinae",        "Halictus",          "Halictus simplex",        "simplex",    "GENUS",  "HOLOTYPE",               NA,                NA,                   NA,                                                 "near Weing Sa",      NA,            NA, "CC_BY_4_0", "OCCURRENCE_STATUS_INFERRED_FROM_INDIVIDUAL_COUNT;TAXON_MATCH_HIGHERRANK;INSTITUTION_MATCH_FUZZY;COLLECTION_MATCH_NONE", "12/25/27",  25L,    12L, 1927L, "PRESERVED_SPECIMEN",       NA,    NA,         "PRESENT",                         "T. D. A. Cockerell",                   NA,                             NA,              NA, "AMNH_IZC 00323581",           "AMNH", "AMNH Hymenoptera",                         "Halictus",         "12/25/27", "FEMALE",      NA, "American Museum of Natural History",          FALSE,                FALSE,              NA, "Kept duplicate",                       NA,                NA,                 NA,                   TRUE,                TRUE,                     TRUE,                         TRUE,                 NA,           NA,                    TRUE,                               TRUE,               TRUE,         TRUE,        TRUE,          TRUE,                 "Halictus",         FALSE,  TRUE,  TRUE,  TRUE,  TRUE,  TRUE,  TRUE,  TRUE,   TRUE,         TRUE,      TRUE,      TRUE,          TRUE,                   TRUE,          TRUE,             TRUE,  TRUE,              TRUE,              TRUE,         TRUE,            TRUE,      NA,        TRUE,                  TRUE,           TRUE,
                "Halictidae",   "Halictinae",        "Halictus",          "Halictus simplex",        "simplex",    "GENUS",  "HOLOTYPE",               NA,                NA,       "Saskatchewan", "Lake Waskesiu, at the beginning of the portage to Heart Lakes",      NA,            NA, "CC_BY_4_0", "OCCURRENCE_STATUS_INFERRED_FROM_INDIVIDUAL_COUNT;TAXON_MATCH_HIGHERRANK;INSTITUTION_MATCH_FUZZY;COLLECTION_MATCH_NONE",  "8/31/36",  31L,     8L, 1936L, "PRESERVED_SPECIMEN",       NA,    NA,         "PRESENT", "T. D. A. Cockerell & Wilmatte P. Cockerell",                   NA,                             NA,              NA, "AMNH_IZC 00323583",           "AMNH", "AMNH Hymenoptera",                         "Halictus",          "8/31/36",   "MALE",      NA, "American Museum of Natural History",          FALSE,                FALSE,              NA, "Kept duplicate",                       NA,                NA,                 NA,                   TRUE,                TRUE,                     TRUE,                         TRUE,                 NA,           NA,                    TRUE,                               TRUE,               TRUE,         TRUE,        TRUE,          TRUE,                 "Halictus",          TRUE, FALSE,  TRUE,  TRUE,  TRUE,  TRUE,  TRUE,  TRUE,   TRUE,         TRUE,      TRUE,      TRUE,          TRUE,                   TRUE,          TRUE,             TRUE,  TRUE,              TRUE,              TRUE,         TRUE,            TRUE,      NA,        TRUE,                  TRUE,           TRUE,
                "Halictidae",   "Halictinae",        "Halictus",          "Halictus simplex",        "simplex",    "GENUS",  "HOLOTYPE",               NA,                NA,     "North Carolina",                                                              NA,      NA,     "Raleigh", "CC_BY_4_0", "OCCURRENCE_STATUS_INFERRED_FROM_INDIVIDUAL_COUNT;TAXON_MATCH_HIGHERRANK;INSTITUTION_MATCH_FUZZY;COLLECTION_MATCH_NONE",  "5/29/24",  29L,     5L, 1924L, "PRESERVED_SPECIMEN",       NA,    NA,         "PRESENT",                              "C. S. Brimley",                   NA,                             NA,              NA, "AMNH_IZC 00324010",           "AMNH", "AMNH Hymenoptera",                         "Halictus",          "5/29/24", "FEMALE",      NA, "American Museum of Natural History",          FALSE,                FALSE,              NA, "Kept duplicate",                       NA,                NA,                 NA,                   TRUE,                TRUE,                     TRUE,                         TRUE,                 NA,           NA,                    TRUE,                               TRUE,               TRUE,         TRUE,        TRUE,          TRUE,                 "Halictus",          TRUE,  TRUE, FALSE,  TRUE,  TRUE,  TRUE,  TRUE,  TRUE,   TRUE,         TRUE,      TRUE,      TRUE,          TRUE,                   TRUE,          TRUE,             TRUE,  TRUE,              TRUE,              TRUE,         TRUE,            TRUE,      NA,        TRUE,                  TRUE,           TRUE,
                "Halictidae",   "Halictinae",        "Halictus",          "Halictus simplex",        "simplex",    "GENUS",  "HOLOTYPE",               NA,                NA,       "Saskatchewan",                     "Prince Albert National Park, Lake Ajawaan",      NA,            NA, "CC_BY_4_0", "OCCURRENCE_STATUS_INFERRED_FROM_INDIVIDUAL_COUNT;TAXON_MATCH_HIGHERRANK;INSTITUTION_MATCH_FUZZY;COLLECTION_MATCH_NONE",  "8/26/36",  26L,     8L, 1936L, "PRESERVED_SPECIMEN",       NA,    NA,         "PRESENT",                         "T. D. A. Cockerell",                   NA,                             NA,              NA, "AMNH_IZC 00323630",           "AMNH", "AMNH Hymenoptera",                         "Halictus",          "8/26/36", "FEMALE",      NA, "American Museum of Natural History",          FALSE,                FALSE,              NA, "Kept duplicate",                       NA,                NA,                 NA,                   TRUE,                TRUE,                     TRUE,                         TRUE,                 NA,           NA,                    TRUE,                               TRUE,               TRUE,         TRUE,        TRUE,          TRUE,                 "Halictus",          TRUE,  TRUE,  TRUE, FALSE,  TRUE,  TRUE,  TRUE,  TRUE,   TRUE,         TRUE,      TRUE,      TRUE,          TRUE,                   TRUE,          TRUE,             TRUE,  TRUE,              TRUE,              TRUE,         TRUE,            TRUE,      NA,        TRUE,                  TRUE,           TRUE,
                "Halictidae",   "Halictinae",        "Halictus",          "Halictus simplex",        "simplex",    "GENUS",  "HOLOTYPE",               NA,                NA,                "Nan",                                                              NA,      NA,         "Nan", "CC_BY_4_0", "OCCURRENCE_STATUS_INFERRED_FROM_INDIVIDUAL_COUNT;TAXON_MATCH_HIGHERRANK;INSTITUTION_MATCH_FUZZY;COLLECTION_MATCH_NONE", "12/31/27",  31L,    12L, 1927L, "PRESERVED_SPECIMEN",       NA,    NA,         "PRESENT",                         "T. D. A. Cockerell",                   NA,                             NA,              NA, "AMNH_IZC 00323584",           "AMNH", "AMNH Hymenoptera",                         "Halictus",         "12/31/27", "FEMALE",      NA, "American Museum of Natural History",          FALSE,                FALSE,              NA, "Kept duplicate",                       NA,                NA,                 NA,                   TRUE,                TRUE,                     TRUE,                         TRUE,                 NA,           NA,                    TRUE,                               TRUE,               TRUE,         TRUE,        TRUE,          TRUE,                 "Halictus",          TRUE,  TRUE,  TRUE,  TRUE, FALSE,  TRUE,  TRUE,  TRUE,   TRUE,         TRUE,      TRUE,      TRUE,          TRUE,                   TRUE,          TRUE,             TRUE,  TRUE,              TRUE,              TRUE,         TRUE,            TRUE,      NA,        TRUE,                  TRUE,           TRUE,
                "Halictidae",   "Halictinae",        "Halictus",          "Halictus simplex",        "simplex",    "GENUS",  "HOLOTYPE",               NA,                NA, "MeknÔøΩs-Tafilalet",                                                              NA,      NA,      "Ifrane", "CC_BY_4_0", "OCCURRENCE_STATUS_INFERRED_FROM_INDIVIDUAL_COUNT;TAXON_MATCH_HIGHERRANK;INSTITUTION_MATCH_FUZZY;COLLECTION_MATCH_NONE",  "8/25/31",  25L,     8L, 1931L, "PRESERVED_SPECIMEN",       NA,    NA,         "PRESENT",       "T. D. A. Cockerell & Alice M. Mackie",                   NA,                             NA,              NA, "AMNH_IZC 00323616",           "AMNH", "AMNH Hymenoptera",                         "Halictus",          "8/25/31", "FEMALE",      NA, "American Museum of Natural History",          FALSE,                FALSE,              NA, "Kept duplicate",                       NA,                NA,                 NA,                   TRUE,                TRUE,                     TRUE,                         TRUE,                 NA,           NA,                    TRUE,                               TRUE,               TRUE,         TRUE,        TRUE,          TRUE,                 "Halictus",          TRUE,  TRUE,  TRUE,  TRUE,  TRUE, FALSE,  TRUE,  TRUE,   TRUE,         TRUE,      TRUE,      TRUE,          TRUE,                   TRUE,          TRUE,             TRUE,  TRUE,              TRUE,              TRUE,         TRUE,            TRUE,      NA,        TRUE,                  TRUE,           TRUE,
                "Halictidae",   "Halictinae",        "Halictus",          "Halictus simplex",        "simplex",    "GENUS",  "HOLOTYPE",               NA,                NA, "MeknÔøΩs-Tafilalet",                                                              NA,      NA,      "Ifrane", "CC_BY_4_0", "OCCURRENCE_STATUS_INFERRED_FROM_INDIVIDUAL_COUNT;TAXON_MATCH_HIGHERRANK;INSTITUTION_MATCH_FUZZY;COLLECTION_MATCH_NONE",  "8/26/30",  26L,     8L, 1930L, "PRESERVED_SPECIMEN",       NA,    NA,         "PRESENT",                            "Alice M. Mackie",                   NA,                             NA,              NA, "AMNH_IZC 00323619",           "AMNH", "AMNH Hymenoptera",                         "Halictus",          "8/26/30",   "MALE",      NA, "American Museum of Natural History",          FALSE,                FALSE,              NA, "Kept duplicate",                       NA,                NA,                 NA,                   TRUE,                TRUE,                     TRUE,                         TRUE,                 NA,           NA,                    TRUE,                               TRUE,               TRUE,         TRUE,        TRUE,          TRUE,                 "Halictus",          TRUE,  TRUE,  TRUE,  TRUE,  TRUE,  TRUE, FALSE,  TRUE,   TRUE,         TRUE,      TRUE,      TRUE,          TRUE,                   TRUE,          TRUE,             TRUE,  TRUE,              TRUE,              TRUE,         TRUE,            TRUE,      NA,        TRUE,                  TRUE,           TRUE,
              "Megachilidae", "Megachilinae", "Pseudoanthidium", "Pseudoanthidium truncatum",      "truncatum",  "SPECIES",  "HOLOTYPE",               NA,                NA,             "Kigoma",                                                              NA,      NA,       "Ujiji", "CC_BY_4_0", "OCCURRENCE_STATUS_INFERRED_FROM_INDIVIDUAL_COUNT;TAXON_MATCH_HIGHERRANK;INSTITUTION_MATCH_FUZZY;COLLECTION_MATCH_NONE",  "8/15/31",  15L,     8L, 1931L, "PRESERVED_SPECIMEN",       NA,    NA,         "PRESENT",                              "James Ogilvie",                   NA,                             NA,              NA, "AMNH_IZC 00323864",           "AMNH", "AMNH Hymenoptera", "Pachyanthidium truncatum (Smith)",          "8/15/31", "FEMALE",      NA, "American Museum of Natural History",          FALSE,                FALSE,              NA, "Kept duplicate",                       NA,                NA,                 NA,                   TRUE,               FALSE,                     TRUE,                         TRUE,                 NA,           NA,                    TRUE,                               TRUE,               TRUE,         TRUE,        TRUE,          TRUE, "Pachyanthidium truncatum",          TRUE,  TRUE,  TRUE,  TRUE,  TRUE,  TRUE,  TRUE, FALSE,   TRUE,         TRUE,      TRUE,      TRUE,          TRUE,                   TRUE,          TRUE,             TRUE,  TRUE,              TRUE,              TRUE,         TRUE,            TRUE,      NA,        TRUE,                  TRUE,           TRUE,
                "Halictidae",   "Halictinae",        "Halictus",          "Halictus simplex",        "simplex",    "GENUS",  "HOLOTYPE",               NA,                NA,                   NA,                                                 "near Weing Sa",      NA,            NA, "CC_BY_4_0", "OCCURRENCE_STATUS_INFERRED_FROM_INDIVIDUAL_COUNT;TAXON_MATCH_HIGHERRANK;INSTITUTION_MATCH_FUZZY;COLLECTION_MATCH_NONE", "12/25/27",  25L,    12L, 1927L, "PRESERVED_SPECIMEN",       NA,    NA,         "PRESENT",                         "T. D. A. Cockerell",                   NA,                             NA,              NA, "AMNH_IZC 00323581",           "AMNH", "AMNH Hymenoptera",                         "Halictus",         "12/25/27", "FEMALE",      NA, "American Museum of Natural History",          FALSE,                FALSE,              NA, "Kept duplicate",                       NA,                NA,                 NA,                   TRUE,                TRUE,                    FALSE,                         TRUE,                 NA,           NA,                    TRUE,                               TRUE,               TRUE,         TRUE,        TRUE,          TRUE,                 "Halictus",          TRUE,  TRUE,  TRUE,  TRUE,  TRUE,  TRUE,  TRUE,  TRUE,  FALSE,         TRUE,      TRUE,      TRUE,          TRUE,                   TRUE,          TRUE,             TRUE,  TRUE,              TRUE,              TRUE,         TRUE,            TRUE,      NA,        TRUE,                  TRUE,           TRUE,
                "Halictidae",   "Halictinae",        "Halictus",          "Halictus simplex",        "simplex",    "GENUS",  "HOLOTYPE",               NA,                NA,       "Saskatchewan", "Lake Waskesiu, at the beginning of the portage to Heart Lakes",      NA,            NA, "CC_BY_4_0", "OCCURRENCE_STATUS_INFERRED_FROM_INDIVIDUAL_COUNT;TAXON_MATCH_HIGHERRANK;INSTITUTION_MATCH_FUZZY;COLLECTION_MATCH_NONE",  "8/31/36",  31L,     8L, 1936L, "PRESERVED_SPECIMEN",       NA,    NA,         "PRESENT", "T. D. A. Cockerell & Wilmatte P. Cockerell",                   NA,                             NA,              NA, "AMNH_IZC 00323583",           "AMNH", "AMNH Hymenoptera",                         "Halictus",          "8/31/36",   "MALE",      NA, "American Museum of Natural History",          FALSE,                FALSE,              NA, "Kept duplicate",                       NA,                NA,                 NA,                   TRUE,                TRUE,                     TRUE,                        FALSE,                 NA,           NA,                    TRUE,                               TRUE,               TRUE,         TRUE,        TRUE,          TRUE,                 "Halictus",          TRUE,  TRUE,  TRUE,  TRUE,  TRUE,  TRUE,  TRUE,  TRUE,   TRUE,        FALSE,      TRUE,      TRUE,          TRUE,                   TRUE,          TRUE,             TRUE,  TRUE,              TRUE,              TRUE,         TRUE,            TRUE,      NA,        TRUE,                  TRUE,           TRUE,
                "Halictidae",   "Halictinae",        "Halictus",          "Halictus simplex",        "simplex",    "GENUS",  "HOLOTYPE",               NA,                NA,     "North Carolina",                                                              NA,      NA,     "Raleigh", "CC_BY_4_0", "OCCURRENCE_STATUS_INFERRED_FROM_INDIVIDUAL_COUNT;TAXON_MATCH_HIGHERRANK;INSTITUTION_MATCH_FUZZY;COLLECTION_MATCH_NONE",  "5/29/24",  29L,     5L, 1924L, "PRESERVED_SPECIMEN",       NA,    NA,         "PRESENT",                              "C. S. Brimley",                   NA,                             NA,              NA, "AMNH_IZC 00324010",           "AMNH", "AMNH Hymenoptera",                         "Halictus",          "5/29/24", "FEMALE",      NA, "American Museum of Natural History",          FALSE,                FALSE,              NA, "Kept duplicate",                       NA,                NA,                 NA,                   TRUE,                TRUE,                     TRUE,                         TRUE,                 NA,           NA,                   FALSE,                               TRUE,               TRUE,         TRUE,        TRUE,          TRUE,                 "Halictus",          TRUE,  TRUE,  TRUE,  TRUE,  TRUE,  TRUE,  TRUE,  TRUE,   TRUE,         TRUE,     FALSE,      TRUE,          TRUE,                   TRUE,          TRUE,             TRUE,  TRUE,              TRUE,              TRUE,         TRUE,            TRUE,      NA,        TRUE,                  TRUE,           TRUE,
                "Halictidae",   "Halictinae",        "Halictus",          "Halictus simplex",        "simplex",    "GENUS",  "HOLOTYPE",               NA,                NA,       "Saskatchewan",                     "Prince Albert National Park, Lake Ajawaan",      NA,            NA, "CC_BY_4_0", "OCCURRENCE_STATUS_INFERRED_FROM_INDIVIDUAL_COUNT;TAXON_MATCH_HIGHERRANK;INSTITUTION_MATCH_FUZZY;COLLECTION_MATCH_NONE",  "8/26/36",  26L,     8L, 1936L, "PRESERVED_SPECIMEN",       NA,    NA,         "PRESENT",                         "T. D. A. Cockerell",                   NA,                             NA,              NA, "AMNH_IZC 00323630",           "AMNH", "AMNH Hymenoptera",                         "Halictus",          "8/26/36", "FEMALE",      NA, "American Museum of Natural History",          FALSE,                FALSE,              NA, "Kept duplicate",                       NA,                NA,                 NA,                   TRUE,                TRUE,                     TRUE,                         TRUE,                 NA,           NA,                    TRUE,                              FALSE,               TRUE,         TRUE,        TRUE,          TRUE,                 "Halictus",          TRUE,  TRUE,  TRUE,  TRUE,  TRUE,  TRUE,  TRUE,  TRUE,   TRUE,         TRUE,      TRUE,     FALSE,          TRUE,                   TRUE,          TRUE,             TRUE,  TRUE,              TRUE,              TRUE,         TRUE,            TRUE,      NA,        TRUE,                  TRUE,           TRUE,
                "Halictidae",   "Halictinae",        "Halictus",          "Halictus simplex",        "simplex",    "GENUS",  "HOLOTYPE",               NA,                NA,                "Nan",                                                              NA,      NA,         "Nan", "CC_BY_4_0", "OCCURRENCE_STATUS_INFERRED_FROM_INDIVIDUAL_COUNT;TAXON_MATCH_HIGHERRANK;INSTITUTION_MATCH_FUZZY;COLLECTION_MATCH_NONE", "12/31/27",  31L,    12L, 1927L, "PRESERVED_SPECIMEN",       NA,    NA,         "PRESENT",                         "T. D. A. Cockerell",                   NA,                             NA,              NA, "AMNH_IZC 00323584",           "AMNH", "AMNH Hymenoptera",                         "Halictus",         "12/31/27", "FEMALE",      NA, "American Museum of Natural History",          FALSE,                FALSE,              NA, "Kept duplicate",                       NA,                NA,                 NA,                   TRUE,                TRUE,                     TRUE,                         TRUE,                 NA,           NA,                    TRUE,                               TRUE,              FALSE,         TRUE,        TRUE,          TRUE,                 "Halictus",          TRUE,  TRUE,  TRUE,  TRUE,  TRUE,  TRUE,  TRUE,  TRUE,   TRUE,         TRUE,      TRUE,      TRUE,         FALSE,                   TRUE,          TRUE,             TRUE,  TRUE,              TRUE,              TRUE,         TRUE,            TRUE,      NA,        TRUE,                  TRUE,           TRUE,
                "Halictidae",   "Halictinae",        "Halictus",          "Halictus simplex",        "simplex",    "GENUS",  "HOLOTYPE",               NA,                NA, "MeknÔøΩs-Tafilalet",                                                              NA,      NA,      "Ifrane", "CC_BY_4_0", "OCCURRENCE_STATUS_INFERRED_FROM_INDIVIDUAL_COUNT;TAXON_MATCH_HIGHERRANK;INSTITUTION_MATCH_FUZZY;COLLECTION_MATCH_NONE",  "8/25/31",  25L,     8L, 1931L, "PRESERVED_SPECIMEN",       NA,    NA,         "PRESENT",       "T. D. A. Cockerell & Alice M. Mackie",                   NA,                             NA,              NA, "AMNH_IZC 00323616",           "AMNH", "AMNH Hymenoptera",                         "Halictus",          "8/25/31", "FEMALE",      NA, "American Museum of Natural History",          FALSE,                FALSE,              NA, "Kept duplicate",                       NA,                NA,                 NA,                   TRUE,                TRUE,                     TRUE,                         TRUE,                 NA,           NA,                    TRUE,                               TRUE,               TRUE,        FALSE,        TRUE,          TRUE,                 "Halictus",          TRUE,  TRUE,  TRUE,  TRUE,  TRUE,  TRUE,  TRUE,  TRUE,   TRUE,         TRUE,      TRUE,      TRUE,          TRUE,                  FALSE,          TRUE,             TRUE,  TRUE,              TRUE,              TRUE,         TRUE,            TRUE,      NA,        TRUE,                  TRUE,           TRUE,
                "Halictidae",   "Halictinae",        "Halictus",          "Halictus simplex",        "simplex",    "GENUS",  "HOLOTYPE",               NA,                NA, "MeknÔøΩs-Tafilalet",                                                              NA,      NA,      "Ifrane", "CC_BY_4_0", "OCCURRENCE_STATUS_INFERRED_FROM_INDIVIDUAL_COUNT;TAXON_MATCH_HIGHERRANK;INSTITUTION_MATCH_FUZZY;COLLECTION_MATCH_NONE",  "8/26/30",  26L,     8L, 1930L, "PRESERVED_SPECIMEN",       NA,    NA,         "PRESENT",                            "Alice M. Mackie",                   NA,                             NA,              NA, "AMNH_IZC 00323619",           "AMNH", "AMNH Hymenoptera",                         "Halictus",          "8/26/30",   "MALE",      NA, "American Museum of Natural History",          FALSE,                FALSE,              NA, "Kept duplicate",                       NA,                NA,                 NA,                   TRUE,                TRUE,                     TRUE,                         TRUE,                 NA,           NA,                    TRUE,                               TRUE,               TRUE,         TRUE,       FALSE,          TRUE,                 "Halictus",          TRUE,  TRUE,  TRUE,  TRUE,  TRUE,  TRUE,  TRUE,  TRUE,   TRUE,         TRUE,      TRUE,      TRUE,          TRUE,                   TRUE,         FALSE,             TRUE,  TRUE,              TRUE,              TRUE,         TRUE,            TRUE,      NA,        TRUE,                  TRUE,           TRUE,
              "Megachilidae", "Megachilinae", "Pseudoanthidium", "Pseudoanthidium truncatum",      "truncatum",  "SPECIES",  "HOLOTYPE",               NA,                NA,             "Kigoma",                                                              NA,      NA,       "Ujiji", "CC_BY_4_0", "OCCURRENCE_STATUS_INFERRED_FROM_INDIVIDUAL_COUNT;TAXON_MATCH_HIGHERRANK;INSTITUTION_MATCH_FUZZY;COLLECTION_MATCH_NONE",  "8/15/31",  15L,     8L, 1931L, "PRESERVED_SPECIMEN",       NA,    NA,         "PRESENT",                              "James Ogilvie",                   NA,                             NA,              NA, "AMNH_IZC 00323864",           "AMNH", "AMNH Hymenoptera", "Pachyanthidium truncatum (Smith)",          "8/15/31", "FEMALE",      NA, "American Museum of Natural History",          FALSE,                FALSE,              NA, "Kept duplicate",                       NA,                NA,                 NA,                   TRUE,                TRUE,                     TRUE,                         TRUE,                 NA,           NA,                    TRUE,                               TRUE,               TRUE,         TRUE,        TRUE,         FALSE, "Pachyanthidium truncatum",          TRUE,  TRUE,  TRUE,  TRUE,  TRUE,  TRUE,  TRUE,  TRUE,   TRUE,         TRUE,      TRUE,      TRUE,          TRUE,                   TRUE,          TRUE,            FALSE,  TRUE,              TRUE,              TRUE,         TRUE,            TRUE,      NA,        TRUE,                  TRUE,           TRUE,
                "Halictidae",   "Halictinae",        "Halictus",          "Halictus simplex",        "simplex",    "GENUS",  "HOLOTYPE",               NA,                NA,                   NA,                                                 "near Weing Sa",      NA,            NA, "CC_BY_4_0", "OCCURRENCE_STATUS_INFERRED_FROM_INDIVIDUAL_COUNT;TAXON_MATCH_HIGHERRANK;INSTITUTION_MATCH_FUZZY;COLLECTION_MATCH_NONE", "12/25/27",  25L,    12L, 1927L, "PRESERVED_SPECIMEN",       NA,    NA,         "PRESENT",                         "T. D. A. Cockerell",                   NA,                             NA,              NA, "AMNH_IZC 00323581",           "AMNH", "AMNH Hymenoptera",                         "Halictus",         "12/25/27", "FEMALE",      NA, "American Museum of Natural History",          FALSE,                FALSE,              NA, "Kept duplicate",                       NA,                NA,                 NA,                   TRUE,                TRUE,                     TRUE,                         TRUE,                 NA,           NA,                    TRUE,                               TRUE,               TRUE,         TRUE,        TRUE,          TRUE,                 "Halictus",          TRUE,  TRUE,  TRUE,  TRUE,  TRUE,  TRUE,  TRUE,  TRUE,   TRUE,         TRUE,      TRUE,      TRUE,          TRUE,                   TRUE,          TRUE,             TRUE, FALSE,              TRUE,              TRUE,         TRUE,            TRUE,      NA,        TRUE,                  TRUE,           TRUE,
                "Halictidae",   "Halictinae",        "Halictus",          "Halictus simplex",        "simplex",    "GENUS",  "HOLOTYPE",               NA,                NA,       "Saskatchewan", "Lake Waskesiu, at the beginning of the portage to Heart Lakes",      NA,            NA, "CC_BY_4_0", "OCCURRENCE_STATUS_INFERRED_FROM_INDIVIDUAL_COUNT;TAXON_MATCH_HIGHERRANK;INSTITUTION_MATCH_FUZZY;COLLECTION_MATCH_NONE",  "8/31/36",  31L,     8L, 1936L, "PRESERVED_SPECIMEN",       NA,    NA,         "PRESENT", "T. D. A. Cockerell & Wilmatte P. Cockerell",                   NA,                             NA,              NA, "AMNH_IZC 00323583",           "AMNH", "AMNH Hymenoptera",                         "Halictus",          "8/31/36",   "MALE",      NA, "American Museum of Natural History",          FALSE,                FALSE,              NA, "Kept duplicate",                       NA,                NA,                 NA,                   TRUE,                TRUE,                     TRUE,                         TRUE,                 NA,           NA,                    TRUE,                               TRUE,               TRUE,         TRUE,        TRUE,          TRUE,                 "Halictus",          TRUE,  TRUE,  TRUE,  TRUE,  TRUE,  TRUE,  TRUE,  TRUE,   TRUE,         TRUE,      TRUE,      TRUE,          TRUE,                   TRUE,          TRUE,             TRUE,  TRUE,             FALSE,              TRUE,         TRUE,            TRUE,      NA,        TRUE,                  TRUE,           TRUE,
                "Halictidae",   "Halictinae",        "Halictus",          "Halictus simplex",        "simplex",    "GENUS",  "HOLOTYPE",               NA,                NA,     "North Carolina",                                                              NA,      NA,     "Raleigh", "CC_BY_4_0", "OCCURRENCE_STATUS_INFERRED_FROM_INDIVIDUAL_COUNT;TAXON_MATCH_HIGHERRANK;INSTITUTION_MATCH_FUZZY;COLLECTION_MATCH_NONE",  "5/29/24",  29L,     5L, 1924L, "PRESERVED_SPECIMEN",       NA,    NA,         "PRESENT",                              "C. S. Brimley",                   NA,                             NA,              NA, "AMNH_IZC 00324010",           "AMNH", "AMNH Hymenoptera",                         "Halictus",          "5/29/24", "FEMALE",      NA, "American Museum of Natural History",          FALSE,                FALSE,              NA, "Kept duplicate",                       NA,                NA,                 NA,                   TRUE,                TRUE,                     TRUE,                         TRUE,                 NA,           NA,                    TRUE,                               TRUE,               TRUE,         TRUE,        TRUE,          TRUE,                 "Halictus",          TRUE,  TRUE,  TRUE,  TRUE,  TRUE,  TRUE,  TRUE,  TRUE,   TRUE,         TRUE,      TRUE,      TRUE,          TRUE,                   TRUE,          TRUE,             TRUE,  TRUE,              TRUE,             FALSE,         TRUE,            TRUE,      NA,        TRUE,                  TRUE,           TRUE,
                "Halictidae",   "Halictinae",        "Halictus",          "Halictus simplex",        "simplex",    "GENUS",  "HOLOTYPE",               NA,                NA,       "Saskatchewan",                     "Prince Albert National Park, Lake Ajawaan",      NA,            NA, "CC_BY_4_0", "OCCURRENCE_STATUS_INFERRED_FROM_INDIVIDUAL_COUNT;TAXON_MATCH_HIGHERRANK;INSTITUTION_MATCH_FUZZY;COLLECTION_MATCH_NONE",  "8/26/36",  26L,     8L, 1936L, "PRESERVED_SPECIMEN",       NA,    NA,         "PRESENT",                         "T. D. A. Cockerell",                   NA,                             NA,              NA, "AMNH_IZC 00323630",           "AMNH", "AMNH Hymenoptera",                         "Halictus",          "8/26/36", "FEMALE",      NA, "American Museum of Natural History",          FALSE,                FALSE,              NA, "Kept duplicate",                       NA,                NA,                 NA,                   TRUE,                TRUE,                     TRUE,                         TRUE,                 NA,           NA,                    TRUE,                               TRUE,               TRUE,         TRUE,        TRUE,          TRUE,                 "Halictus",          TRUE,  TRUE,  TRUE,  TRUE,  TRUE,  TRUE,  TRUE,  TRUE,   TRUE,         TRUE,      TRUE,      TRUE,          TRUE,                   TRUE,          TRUE,             TRUE,  TRUE,              TRUE,              TRUE,        FALSE,            TRUE,      NA,        TRUE,                  TRUE,           TRUE,
                "Halictidae",   "Halictinae",        "Halictus",          "Halictus simplex",        "simplex",    "GENUS",  "HOLOTYPE",               NA,                NA,                "Nan",                                                              NA,      NA,         "Nan", "CC_BY_4_0", "OCCURRENCE_STATUS_INFERRED_FROM_INDIVIDUAL_COUNT;TAXON_MATCH_HIGHERRANK;INSTITUTION_MATCH_FUZZY;COLLECTION_MATCH_NONE", "12/31/27",  31L,    12L, 1927L, "PRESERVED_SPECIMEN",       NA,    NA,         "PRESENT",                         "T. D. A. Cockerell",                   NA,                             NA,              NA, "AMNH_IZC 00323584",           "AMNH", "AMNH Hymenoptera",                         "Halictus",         "12/31/27", "FEMALE",      NA, "American Museum of Natural History",          FALSE,                FALSE,              NA, "Kept duplicate",                       NA,                NA,                 NA,                   TRUE,                TRUE,                     TRUE,                         TRUE,                 NA,           NA,                    TRUE,                               TRUE,               TRUE,         TRUE,        TRUE,          TRUE,                 "Halictus",          TRUE,  TRUE,  TRUE,  TRUE,  TRUE,  TRUE,  TRUE,  TRUE,   TRUE,         TRUE,      TRUE,      TRUE,          TRUE,                   TRUE,          TRUE,             TRUE,  TRUE,              TRUE,              TRUE,         TRUE,           FALSE,      NA,        TRUE,                  TRUE,           TRUE,
                "Halictidae",   "Halictinae",        "Halictus",          "Halictus simplex",        "simplex",    "GENUS",  "HOLOTYPE",               NA,                NA,     "North Carolina",                                                              NA,      NA,     "Raleigh", "CC_BY_4_0", "OCCURRENCE_STATUS_INFERRED_FROM_INDIVIDUAL_COUNT;TAXON_MATCH_HIGHERRANK;INSTITUTION_MATCH_FUZZY;COLLECTION_MATCH_NONE",  "5/29/24",  29L,     5L, 1924L, "PRESERVED_SPECIMEN",       NA,    NA,         "PRESENT",                              "C. S. Brimley",                   NA,                             NA,              NA, "AMNH_IZC 00324010",           "AMNH", "AMNH Hymenoptera",                         "Halictus",          "5/29/24", "FEMALE",      NA, "American Museum of Natural History",          FALSE,                FALSE,              NA, "Kept duplicate",                       NA,                NA,                 NA,                   TRUE,                TRUE,                     TRUE,                         TRUE,                 NA,           NA,                    TRUE,                               TRUE,               TRUE,         TRUE,        TRUE,          TRUE,                 "Halictus",          TRUE,  TRUE,  TRUE,  TRUE,  TRUE,  TRUE,  TRUE,  TRUE,   TRUE,         TRUE,      TRUE,      TRUE,          TRUE,                   TRUE,          TRUE,             TRUE,  TRUE,              TRUE,              TRUE,         TRUE,            TRUE,      NA,        TRUE,                  TRUE,          FALSE,
                "Halictidae",   "Halictinae",        "Halictus",          "Halictus simplex",        "simplex",    "GENUS",  "HOLOTYPE",               NA,                NA,       "Saskatchewan",                     "Prince Albert National Park, Lake Ajawaan",      NA,            NA, "CC_BY_4_0", "OCCURRENCE_STATUS_INFERRED_FROM_INDIVIDUAL_COUNT;TAXON_MATCH_HIGHERRANK;INSTITUTION_MATCH_FUZZY;COLLECTION_MATCH_NONE",  "8/26/36",  26L,     8L, 1936L, "PRESERVED_SPECIMEN",       NA,    NA,         "PRESENT",                         "T. D. A. Cockerell",                   NA,                             NA,              NA, "AMNH_IZC 00323630",           "AMNH", "AMNH Hymenoptera",                         "Halictus",          "8/26/36", "FEMALE",      NA, "American Museum of Natural History",          FALSE,                FALSE,              NA, "Kept duplicate",                       NA,                NA,                 NA,                   TRUE,                TRUE,                     TRUE,                         TRUE,                 NA,           NA,                    TRUE,                               TRUE,               TRUE,         TRUE,        TRUE,          TRUE,                 "Halictus",          TRUE,  TRUE,  TRUE,  TRUE,  TRUE,  TRUE,  TRUE,  TRUE,   TRUE,         TRUE,      TRUE,      TRUE,          TRUE,                   TRUE,          TRUE,             TRUE,  TRUE,              TRUE,              TRUE,         TRUE,            TRUE,      NA,        TRUE,                  TRUE,           TRUE,
                "Halictidae",   "Halictinae",        "Halictus",          "Halictus simplex",        "simplex",    "GENUS",  "HOLOTYPE",               NA,                NA,                "Nan",                                                              NA,      NA,         "Nan", "CC_BY_4_0", "OCCURRENCE_STATUS_INFERRED_FROM_INDIVIDUAL_COUNT;TAXON_MATCH_HIGHERRANK;INSTITUTION_MATCH_FUZZY;COLLECTION_MATCH_NONE", "12/31/27",  31L,    12L, 1927L, "PRESERVED_SPECIMEN",       NA,    NA,         "PRESENT",                         "T. D. A. Cockerell",                   NA,                             NA,              NA, "AMNH_IZC 00323584",           "AMNH", "AMNH Hymenoptera",                         "Halictus",         "12/31/27", "FEMALE",      NA, "American Museum of Natural History",          FALSE,                FALSE,              NA, "Kept duplicate",                       NA,                NA,                 NA,                   TRUE,                TRUE,                     TRUE,                         TRUE,                 NA,           NA,                    TRUE,                               TRUE,               TRUE,         TRUE,        TRUE,          TRUE,                 "Halictus",          TRUE,  TRUE,  TRUE,  TRUE,  TRUE,  TRUE,  TRUE,  TRUE,   TRUE,         TRUE,      TRUE,      TRUE,          TRUE,                   TRUE,          TRUE,             TRUE,  TRUE,              TRUE,              TRUE,         TRUE,            TRUE,      NA,        TRUE,                  TRUE,           TRUE
              )




# run the function
testOut <- BeeBDC::summaryFun(data = testData, dontFilterThese = NULL, removeFilterColumns = FALSE, filterClean = FALSE)


# test that function added a column
testthat::test_that("summaryFun added column", {
  testthat::expect_equal((ncol(testOut) - ncol(testData)), 1)
})

# test that .summary column contains correct values for test data
correct <- c(FALSE, FALSE, FALSE, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, 
             FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, 
             FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, TRUE, TRUE)
             

testthat::test_that("summaryFun .summary column with correct values", {
  testthat::expect_equal(testOut$.summary, correct)
})

# Description of the BeeBDC taxonomic dataset
# 16th of March 2023

#' Download a nearly complete taxonomy of bees globally
#'
#'
#' Downloads the taxonomic information for the bees of the world. 
#' Source of taxonomy is listed under "source" but are mostly derived from the Discover Life 
#' website. The data will be sourced from the BeeBDC article's Figshare.
#'  
#' Note that sometimes the download might not work without restarting R. In this case, you could
#' alternatively download the dataset from the URL below and then read it in using 
#' `base::readRDS("filePath.Rda")`. Note that as of version 1.3.2, this function internally uses 
#'  a modified version of the "download" function from the 
#' `downloader` package on CRAN. Additionally, for Windows machines, BeeBDC will try a different 
#' download method with each failed attempt (for Windows: "auto" > "internal" > "libcurl" > "wget" > "curl"; 
#` for Mac/Unix: "libcurl" > "wget" > "curl" > "auto";
#' or "auto" if any particular method is not available).
#' 
#' 
#'  **Column details** 
#'  
#'  **flags** Flags or comments about the taxon name.
#'  
#'  **taxonomic_status** Taxonomic status. Values are "accepted" or "synonym"
#'  
#'  **source** Source of the name.
#'  
#'  **accid** The id of the accepted taxon name or "0" if taxonomic_status == accepted.
#'  
#'  **id** The id number for the taxon name.
#'  
#'  **kingdom** The biological kingdom the taxon belongs to. For bees, kingdom == Animalia.
#'  
#'  **phylum** The biological phylum the taxon belongs to. For bees, phylum == Arthropoda.
#'  
#'  **class** The biological class the taxon belongs to. For bees, class == Insecta.
#'  
#'  **order** The biological order the taxon belongs to. For bees, order == Hymenoptera.
#'
#'  
#'  **family** The family of bee which the species belongs to.
#'  
#'  **subfamily** The subfamily of bee which the species belongs to.
#'  
#'  **tribe** The tribe of bee which the species belongs to.
#'  
#'  **subtribe** The subtribe of bee which the species belongs to.
#'  
#'  **validName** The valid scientific name as it should occur in the “scientificName” column in a Darwin Core file.
#'  
#'  **canonical** The scientificName without the scientificNameAuthority.
#'  
#'  **canonical_withFlags** The scientificName without the scientificNameAuthority and with Discover Life taxonomy flags.
#'  
#'  **genus** The genus the bee species belongs to.
#'  
#'  **subgenus** The subgenus the bee species belongs to.
#'  
#'  **species** The specific epithet for the bee species.
#'  
#'  **infraspecies** The infraspecific epithet for the bee addressed.
#'  
#'  **authorship** The author who described the bee species.
#'  
#'  **taxon_rank** Rank for the bee taxon addressed in the entry.
#'  
#'  **notes** Additional notes about the name/taxon.
#'  
#'  **Previous taxonomies:**
#'   - 2026-05-04 **current**: https://api.figshare.com/v2/file/download/64193731
#'  
#'   - 2026-01-12: https://open.flinders.edu.au/ndownloader/files/60945820
#'  
#'   - 2024-06-17: https://open.flinders.edu.au/ndownloader/files/47089969
#'   
#'   - 2023-11-29: https://open.flinders.edu.au/ndownloader/files/43331472
#'   
#'   - 2023-10-10: https://open.flinders.edu.au/ndownloader/files/42613126
#'   
#'   - 2023-09-20: https://open.flinders.edu.au/ndownloader/files/42402264
#'   
#'   - Original: https://open.flinders.edu.au/ndownloader/files/42320595
#' 
#'
#' @param URL A character vector to the FigShare location of the dataset. The default will be to
#' the most-recent version.
#' @param mode A character passed on to `utils::download.file()`. Default = "wb" for Windows or "w" for Mac/Linux.
#' @param headers Character. Passed on to  `utils::download.file()`. Default = NULL.
#' @param token Character. Optional personal access token from FigShare account for authorisation. 
#' Default = NULL.
#' @param alternateURL Logical. If TRUE then the function will use the an alernate version of the 
#' download URL. Might be worth trying if it is failing. Default = FALSE.
#' @param ... Extra variables that can be passed to `downloader::download()`.
#'
#'
#'  
#' @return A downloaded beesTaxonomy.Rda file in the [tempdir()] and the same tibble returned to
#' the environment.
#'  
#' 
#' @references This dataset was created using the Discover Life taxonomy.
#' Dataset is from the publication: 
#' DOREY, J. B., CHESSHIRE, P. R., BOLAÑOS, A. N., O’REILLY, R. L., BOSSERT, S., COLLINS, S. M., LICHTENBERG, E. M., TUCKER, E., SMITH-PARDO, A., FALCON-BRINDIS, A., GUEVARA, D. A., RIBEIRO, B. R., DE PEDRO, D., FISCHER, E., HUNG, J. K.-L., PARYS, K. A., ROGAN, M. S., MINCKLEY, R. L., VELZCO, S. J. E., GRISWOLD, T., ZARRILLO, T. A., SICA, Y., ORR, M. C., GUZMAN, L. M., ASCHER, J., HUGHES, A. C. & COBB, N. S. In review. A globally synthesised and flagged bee occurrence dataset and cleaning workflow. Scientific Data.
#' The taxonomy data are mostly compiled from Discover Life data, www.discoverlife.org:
#' ASCHER, J. S. & PICKERING, J. 2020. Discover Life bee species guide and world checklist (Hymenoptera: Apoidea: Anthophila). http://www.discoverlife.org/mp/20q?guide=Apoidea_species.
#'
#' @seealso [BeeBDC::taxadbToBeeBDC()] to download any other taxonomy (of any taxa or of bees)
#' and [BeeBDC::harmoniseR()] for the 
#' taxon-cleaning function where these taxonomies are implemented.
#' 
#' @export
#' 
#' @examples
#'\dontrun{
#' beesTaxonomy <- BeeBDC::beesTaxonomy()
#'}
#' 
#'
beesTaxonomy <- function(URL = "https://api.figshare.com/v2/file/download/64193731",
                         mode = NULL,
                         headers = NULL,
                         token = NULL,
                         alternateURL = FALSE,
                         ...){
  destfile <- taxonomy <- attempt <- nAttempts <- error_funcFile <- error_func <-  NULL
  downloadReturn <- . <- NULL
  
  # If user wants to try the Flinders URL, which seems to not be working as well
  if(alternateURL == TRUE){
    URL <- URL %>%
      stringr::str_remove_all(".*/") %>%
      stringr::str_c("https://open.flinders.edu.au/ndownloader/files/", .)
  }
  
    #### 0.0 Prep ####
  # Set the number of attempts
  nAttempts = 6
    
      ##### 0.1 Errors ####
        ###### a. messages ####
  # Set up the error message function
  error_func <- function(e){
    message(paste("Taxonomy download attempt failed..."))
  }
  error_funcFile <- function(e){
    message(paste("Could not read taxonomy download..."))
  }
  
  
        ###### b. error catcher ####
  # Function to capture error outputs from here
  # Source - https://stackoverflow.com/a
  # Posted by Martin Morgan, modified by community. See post 'Timeline' for 
  # change history
  # Retrieved 2026-01-12, License - CC BY-SA 2.5
  errorCatcher <- function(fun){
      warn <- err <- NULL
      res <- withCallingHandlers(
        tryCatch(fun(...), error=function(e) {
          err <<- conditionMessage(e)
          NULL
        }), warning=function(w) {
          warn <<- append(warn, conditionMessage(w))
          invokeRestart("muffleWarning")
        })
      list(res, warn=warn, err=err)
    }
  
  
    ##### 0.2 Check OS ####
  # Check operating system
  OS <- dplyr::if_else(.Platform$OS.type == "unix",
                       "MacLinux",
                       "Windows")
  
  writeLines(paste0("The operating system detected is ", OS, "."))
  
    ##### 0.3 Downloader function ####
    # Please note that this function is taken directly from the "downloader" package version 0.4.1
    # This is the purpose of the package, but I have taken their excellent function to avoid
    # Another dependency for BeeBDC. MY apologies and thanks to the authors.
  download <- function(URL, destfile = NULL, methodNum = NULL, headers = headers, ...) {
    # First, check protocol. If http or https, check platform:
    if (grepl('^https?://', URL)) {
      #### Windows ####
      if (tolower(.Platform$OS.type) == "windows") {
        # Try httr first
        if(methodNum == 1){
          message(paste0("Trying first download method using httr::GET..."))
          httr::GET(
            URL,
            httr::add_headers(
              Authorization = paste("token", token)
            ),
            httr::write_disk(destfile, overwrite = TRUE)
          )} # END methodNum == 1
        # Set method to NULL to then be over-written
        method <- NULL
        # Try different methods if failed
        if(methodNum == 2){method <- "auto"}
        if(methodNum == 3){method <- "wininet"}
        # Check also if libcurl is an option
        if(methodNum == 4 && capabilities("libcurl")){
          method <- "libcurl"}
        # Check also if wget is an option
        if(methodNum == 5 && nzchar(Sys.which("wget")[1])){
          method <- "wget"}
        # Check also if curl is an option
        if(methodNum == 6 && nzchar(Sys.which("curl")[1])){
          method <- "curl"}
        # If one of the above fails, use "internal"
        if(is.null(method)){
          method <- "internal"
        }
        
        if(is.null(mode)){
          mode <- "wb"}
        
        # download.file will complain about file size with something like:
        #       Warning message:
        #         In download.file(urURLl, ...) : downloaded length 19457 != reported length 200
        # because apparently it compares the length with the status code returned (?)
        # so we supress that
        if(methodNum > 1){
          message(paste0("Trying download method ", method, " and mode ", mode, "..."))
          downloadReturn <- utils::download.file(URL, 
                                                 method = method, 
                                                 destfile = destfile, 
                                                 mode = mode,
                                                 headers = headers,
                                                 ...) %>%
            errorCatcher()}
        
      } else {
        #### Mac/Linux ####
        method <- NULL
        # Try httr first
        if(methodNum == 1){
          message(paste0("Trying first download method using httr::GET..."))
          httr::GET(
            URL,
            httr::add_headers(
              Authorization = paste("token", token)
            ),
            httr::write_disk(destfile, overwrite = TRUE)
          )} # END methodNum == 1
        # If non-Windows, check for libcurl/curl/wget/lynx, then call download.file with
        # appropriate method.
        if (capabilities("libcurl") && methodNum == 2) {
          method <- "libcurl"
        } else if (nzchar(Sys.which("wget")[1]) && methodNum == 3) {
          method <- "wget"
        } else if (nzchar(Sys.which("curl")[1]) && methodNum == 4) {
          method <- "curl"
          # curl needs to add a -L option to follow redirects.
          # Save the original options and restore when we exit.
          orig_extra_options <- getOption("download.file.extra")
          on.exit(options(download.file.extra = orig_extra_options))
          options(download.file.extra = paste("-L", orig_extra_options))
        } else if (nzchar(Sys.which("lynx")[1]) && methodNum == 5) {
          method <- "lynx"
        } else if(methodNum == 6){
          method <- "auto"
        }
        if(is.null(method)){
          method <- "auto"
        }
        if(is.null(mode)){
          mode <- "wb"  
        }
        if(methodNum > 1){
          message(paste0("Trying download method ", method, " and mode ", mode, "..."))
          downloadReturn <- utils::download.file(URL, 
                                                 method = method, 
                                                 destfile = destfile, 
                                                 mode = mode,
                                                 headers = headers,
                                                 ...) %>%
            errorCatcher()}
          
      }
      
    } else {
      downloadReturn <- utils::download.file(URL, destfile = destfile, 
                                             mode = "wb", headers = headers, ...) %>%
        errorCatcher()
    }
    return(downloadReturn)
  } # END download function
  
  
  #### 1.0 Download ####
    # Run a code to download the data and deal with potential internet issues
  taxonomy <- NULL                                 
  attempt <- 1
  savePath <- file.path(tempdir(), "beesTaxonomy.Rda") %>% 
      # Change all backlashes to forward slashes -- sometimes they mix on Windows...
    stringr::str_replace_all("\\\\","/") 
  writeLines(paste0("Saving file temporarily to ", savePath))
  suppressWarnings(
  while( is.null(taxonomy) && attempt <= nAttempts) {   
      # Initial message
    print( paste("Attempt: ", attempt, " of ", nAttempts))
    # Don't attempt for the last attempt
    if(attempt <= nAttempts){
# WINDOWS or MAC/Linux
        # Download the file to the outPath 
        tryCatch(downloadReturn <- download(URL, destfile = savePath, 
                                              # Change the method based on attempt number
                                            methodNum = attempt,
                                            headers = headers),
                 error = error_func, warning = error_func)
        # Load the file from the outPath
        tryCatch(
          taxonomy <- base::readRDS(savePath),
          error = error_funcFile, warning = error_funcFile)
    } # END if attempt < nAttempts
    
    if(attempt <= nAttempts){
      # Wait five seconds before the next request 
    if(is.null(taxonomy)){Sys.sleep(5)}            
    } # END IF #2
    # Count the next attempt
    attempt <- attempt + 1   
    if(attempt > 1 && !is.null(downloadReturn)){
      # Output errors per run
      # Check download errors
      if(!stringr::str_detect(paste0(downloadReturn, collapse = ""), 
      "could not find function") && attempt > 1){
      # Remove NULL elements
      downloadReturn <- downloadReturn[-which(sapply(downloadReturn, is.null))]
      # Paste message
      message(paste0("\n - Possible *download* error(s) returned:\n", paste0(
      names(downloadReturn), ": ", downloadReturn, collapse = "\n")))}
      # Check file errors
      fileError <- base::readRDS(savePath) %>% 
      errorCatcher()
      if(!stringr::str_detect(paste0(fileError, collapse = ""), 
      "could not find function") && attempt > 1){
      # Remove NULL elements
      fileError <- fileError[-which(sapply(fileError, is.null))]
      # Paste message
      message(paste0("\n - Possible *file* error(s) returned:\n", paste0(
      names(fileError), ": ", fileError, collapse = "\n")))}
      } # END if( attempt > 1){
  } # END while
  )
  
  #### 2.0 Error list ####
  if(is.null(taxonomy)){
      # Check system capacities
    message(paste0(
      "System capabilities are:\n",
        " * Has libcurl? ", capabilities("libcurl"),
      "\n * Has wget? ", nzchar(Sys.which("wget")[1]),
      "\n * Has curl? ", nzchar(Sys.which("curl")[1])
    ))
      
    warning(paste0(" - Taxonomy download failed. Please check your internet connection.\n",
    "Alternatively, feel free to paste the download url into your browser (",
    URL, ")",
                   " and download the file directly. \n",
    "This file can then be read into R using:\n",
                   "beesTaxonomy <- readRDS('path/to/downloaded/file/beesTaxonomy.Rda')"))
    
    stop("Errors finished.")
  }

  #### 3.0 Return ####
  # Return the data to the user
  return(taxonomy)
} # END function

# load in the test dataset
system.file("extdata", "testTaxonomy.rda", package="BeeBDC") |>
  load()


# Test the expected results
testthat::test_that("testTaxonomy expected number of columns", {
  testthat::expect_equal(ncol(testTaxonomy), 24)
})
testthat::test_that("testTaxonomy expect that number of unique ids == number of overall rows", {
  testthat::expect_equal(length(unique(testTaxonomy$id)), nrow(testTaxonomy))
})
testthat::test_that("testTaxonomy expect that number of 0[accid] == accepted[taxostatus]", {
  testthat::expect_equal(sum(testTaxonomy$accid == 0), 
                         sum(testTaxonomy$taxonomic_status == "accepted"))
})

  # Test classes
testthat::test_that("testTaxonomy expected class", {
  testthat::expect_type(testTaxonomy, "list")
})
testthat::test_that("testTaxonomy expected class", {
  testthat::expect_equal(attributes(testTaxonomy)$class, c("spec_tbl_df", "tbl_df","tbl","data.frame" ))
})

testthat::test_that("DownloadTest_skipOnline",{
# Skip this test on MAC on GITHUB due to 403 error in tests
  if(stringr::str_detect(Sys.info() %>% as.character(), "Darwin") %>% any()){
    skip_on_ci()}
#  skip_on_ci()
  
# Define download headers
#headers <- c(Authorization = paste("token", "20271361303ce042ff9cce49ecd9c8f23594ed4032e83f150e535dbd2b67297ea88448e3ca29260f6c416f581834094a898089d90f2229111845f01ef8b75f46"))
Sys.getenv("BEEBDC_SECRET_GITHUB")
Authorization <- Sys.getenv("BEEBDC_SECRET_GITHUB")
## Set some options for accessing the data
  userAgent <- paste0(sprintf("BeeBDC-tests/%s (R/%s.%s)",
                              as.character(utils::packageVersion("BeeBDC")),
                              R.version$major, R.version$minor),
                      "; R (",R.version$major,".", R.version$minor, " ",
                      R.version$platform, " ", R.version$arch," ",
                      R.version$os,")") 
  
  options(timeout=400,
          HTTPUserAgent = userAgent)
  OS <- dplyr::if_else(.Platform$OS.type == "unix",
                       "MacLinux",
                       "Windows")
if(OS == "Windows"){
  mode <- "wb"
}else{
  mode <- "wb"
}
taxonomyFile <- NULL
# load in the full dataset
taxonomyFile <- BeeBDC::beesTaxonomy(mode = mode,
                                     headers = NULL,
                                     token = Authorization)


if(!is.null(taxonomyFile)){
# Test the expected results
testthat::test_that("beesTaxonomy expected number of columns", {
  testthat::expect_equal(ncol(taxonomyFile), 24)
})
testthat::test_that("beesTaxonomy expect that number of unique ids == number of overall rows", {
  testthat::expect_equal(length(unique(taxonomyFile$id)), nrow(taxonomyFile))
})
testthat::test_that("beesTaxonomy expect that number of 0[accid] == accepted[taxostatus]", {
  testthat::expect_equal(sum(taxonomyFile$accid == 0), 
                         sum(taxonomyFile$taxonomic_status == "accepted"))
})

# Test classes
testthat::test_that("beesTaxonomy expected class", {
  testthat::expect_type(taxonomyFile, "list")
})
testthat::test_that("beesTaxonomy expected class", {
  testthat::expect_equal(attributes(taxonomyFile)$class, c("spec_tbl_df", "tbl_df","tbl","data.frame" ))
})

} #END !is.null
}) # END testthat::test_that

#' Create figures reporting the results of the bdc/BeeBDC packages
#'
#' Creates figures (i.e., bar plots, maps, and histograms) reporting the results
#' of data quality tests implemented the bdc and BeeBDC packages. Works like [bdc::bdc_create_figures()],
#' but it allows the user to specify a save path.
#'
#' @param data A data frame or tibble. Needs to contain the results of data quality tests; that
#' is, columns starting with ".".
#' @param path A character directory. The path to a directory in which to save the figures. 
#' Default = OutPath_Figures.
#' @param workflow_step A character string. Name of the workflow step. Options
#' available are "prefilter", "space", and "time".
#' @param database_id A character string. The column name with a unique record
#' identifier. Default = "database_id".
#' @param save_figures Logical. Indicates if the figures should be saved for further inspection or 
#' use. Default = FALSE.
#'
#' @details This function creates figures based on the results of data quality
#' tests. A pre-defined list of test names is used for creating
#' figures depending on the name of the workflow step informed. Figures are
#' saved in "Output/Figures" if save_figures = TRUE.
#'
#' @return List containing figures showing the results of data quality tests
#' implemented in one module of bdc/BeeBDC. When save_figures = TRUE, figures are
#' also saved locally in a .png format.
#'
#' @importFrom dplyr %>%
#' @export
#'
#' @examples
#' \donttest{
#' 
#' if(require("bdc")){
#' 
#' database_id <- c("GBIF_01", "GBIF_02", "GBIF_03", "FISH_04", "FISH_05")
#' lat <- c(-19.93580, -13.01667, -22.34161, -6.75000, -15.15806)
#' lon <- c(-40.60030, -39.60000, -49.61017, -35.63330, -39.52861)
#' .scientificName_emptys <- c(TRUE, TRUE, TRUE, FALSE, FALSE)
#' .coordinates_empty <- c(TRUE, TRUE, TRUE, TRUE, TRUE)
#' .invalid_basis_of_records <- c(TRUE, FALSE, TRUE, FALSE, TRUE)
#' .summary <- c(TRUE, FALSE, TRUE, FALSE, FALSE)
#'
#' x <- data.frame(
#'   database_id,
#'   lat,
#'   lon,
#'   .scientificName_emptys,
#'   .coordinates_empty,
#'   .invalid_basis_of_records,
#'   .summary
#' )
#'
#'figures <- 
#' jbd_create_figures(
#'   data = x, 
#'   database_id = "database_id",
#'   workflow_step = "prefilter",
#'   save_figures = FALSE
#' )
#' 
#' } # End if require
#' } # End dont test
jbd_create_figures <-
  function(data,
           path = OutPath_Figures,
           database_id = "database_id",
           workflow_step = NULL, 
           save_figures = FALSE) {
    . <- .data <- n_flagged <- n_total <- freq <- NULL
    . <- V1 <- Name <- freq <- year <- decimalLongitude <- n <- NULL
    decimalLatitude <- . <- long <- lat <- group <- `NA` <- .summary <- NULL
    OutPath_Figures <- OutPath_Report <- NULL
    
    if(is.null(path)){
      warning("Please provide a path!")
    }
    
    suppressWarnings({
      check_require_cran("cowplot")
      check_require_cran("readr")
      check_require_cran("ggspatial")
      requireNamespace("bdc")
    })
    
    match.arg(arg = workflow_step,
              choices = c("prefilter", "space", "time")
    )
    
    temp <- data %>% dplyr::select(tidyselect::starts_with("."))
    
    convertNames <- dplyr::tibble(
      # Space
      .cap = "Records around country capital centroid"	,
      .cen = "Records around country or province centroids",
      .dbl = "Duplicated coordinates per species",
      .equ = "Identical coordinates",
      .otl = "Geographical outliers",
      .gbf = "Records around the GBIF headquarters",
      .inst = "Records around biodiversity institutions",
      .rou = "Rounded (probably imprecise) coordinates",
      .urb = "Records within urban areas",
      .zer = "Plain zeros",
      # The rest
      .scientificName_empty = "Scientific name empty",
      .coordinates_empty = "Coordinates empty",
      .coordinates_outOfRange = "Coordinates out of range",
      .invalid_basis_of_records = "Invalid basis of record",
      .coordinates_country_inconsistent = "Coordinates inconsistent with country",
      .uncer_term = "Uncertainty term in name",
      .eventDate_empty = "Empty eventDate",
      .year_outOfRange = "Year out of range",
      .summary = "Summary",
      summary_all_tests = "Summary of all tests"
    )
    
    if (all((colSums(temp, na.rm = TRUE) - nrow(temp)) == 0)) {
      message("Figures were not created.\nNo records flagged as 'FALSE' in columns starting with '.'")
    }
    
    if (ncol(temp) == 0) {
      message(
        "Figures were not created.\nAt least one column 'starting with '.' containing results of data-quality tests must be provided"
      )
    }
    
    # Formatting y axis of ggplot bar
    fancy_scientific <- function(l) {
      format(l, big.mark = ",", digits = 2, nsmall = 1)
    }
    
    # Total number of records
    suppressMessages({
      n_records <-
        data %>%
        dplyr::summarise(n = dplyr::n()) %>%
        dplyr::pull(n)
      
      # Total number of records per database
      n_record_database <-
        data %>%
        dplyr::mutate(
          database_id = gsub("[[:digit:]]+", "", database_id),
          database_id = gsub("_", "", database_id)
        ) %>%
        dplyr::group_by(database_id) %>%
        dplyr::summarise(n_total = dplyr::n())
      
      our_theme <-
        ggplot2::theme_minimal() +
        ggplot2::theme(axis.title = ggplot2::element_text(size = 18),
                       axis.text = ggplot2::element_text(size = 12),
                       panel.grid.major.x = ggplot2::element_line(color = "#cbcbcb"),
                       panel.grid.major.y = ggplot2::element_blank(),
                       plot.margin = ggplot2::unit(c(0.5, 0.5, 0.5, 0.5), "cm")
        )
      
      # prefilter
      if (workflow_step == "prefilter") {
        tests <-
          c(
            ".scientificName_empty",
            ".coordinates_empty",
            ".coordinates_outOfRange",
            ".invalid_basis_of_records",
            ".coordinates_country_inconsistent",
            ".summary"
          )
        
        names_tab <- names(data)
        col_to_tests <- dplyr::intersect(tests, names_tab)
        
        if (file.exists("Output/Check/01_coordinates_transposed.csv")) {
          col_to_tests <- c(col_to_tests, "coordinates_transposed")
        }
      }
      
      # space
      if (workflow_step == "space") {
        tests <-
          c(
            ".equ",
            ".zer",
            ".cap",
            ".cen",
            ".otl",
            ".gbf",
            ".inst",
            ".dpl",
            ".rou",
            ".urb",
            ".summary"
          )
        
        names_tab <- names(data)
        col_to_tests <- dplyr::intersect(tests, names_tab)
      }
      
      # time
      if (workflow_step == "time") {
        tests <-
          c(
            ".eventDate_empty",
            ".year_outOfRange",
            ".summary",
            "summary_all_tests"
          )
        
        names_tab <- names(data)
        col_to_tests <- dplyr::intersect(tests, names_tab)
      }
      
      # function to create bar plots for each dataset separately
      create_barplot_database <-
        function(data, column_to_map, workflow_step = workflow_step) {
          temp <-
            data %>%
            dplyr::mutate(
              database_id = gsub("[[:digit:]]+", "", database_id),
              database_id = gsub("_", "", database_id)
            ) %>%
            dplyr::filter(., .data[[column_to_map]] == FALSE) %>%
            dplyr::group_by(database_id, .data[[column_to_map]]) %>%
            dplyr::summarise(n_flagged = dplyr::n()) %>%
            dplyr::full_join(., n_record_database, by = "database_id") %>%
            dplyr::mutate(freq = round(n_flagged / n_total, 5) * 100)
          
          temp[is.na(temp)] <- 0
          
          b <-
            ggplot2::ggplot(temp, ggplot2::aes(x = stats::reorder(database_id, -freq), y = freq)) +
            ggplot2::geom_col(colour = "white", fill = "royalblue") +
            ggplot2::coord_flip() +
            our_theme +
            ggplot2::labs(x = "Dataset", y = "% of records flagged",
                          title = convertNames[column_to_map == convertNames]) +
            ggplot2::geom_hline(
              yintercept = 0,
              linewidth = 1,
              colour = "#333333"
            ) +
            ggplot2::scale_y_continuous(expand = c(0, 0), 
                                        labels = fancy_scientific)
          
        }
      
      # function to create barplots considering all datasets together
      create_barplot_all_tests <-
        function(data,
                 column_to_map = "summary_all_tests",
                 workflow_step = workflow_step) {
          temp <-
            data %>%
            dplyr::select(tidyselect::starts_with(".")) %>%
            dplyr::mutate_if(is.character, ~ as.logical(as.character(.))) %>%
            dplyr::summarise_all(., .funs = sum) %>%
            t() %>%
            dplyr::as_tibble(rownames = "NA", .name_repair = "unique") %>%
            stats::setNames(c('NA', "V1")) %>%
            dplyr::mutate(V1 = nrow(data) - V1) %>%
            dplyr::mutate(
              freq =
                round((V1 / n_records * 100), 2)
            ) %>%
            dplyr::rename(
              Name = `NA`,
              n_flagged = V1
            )
          
          temp[is.na(temp)] <- 0
          
          if (all(temp$freq == 0)) {
            temp$freq <- 0.000001
          } else {
            gg <-
              ggplot2::ggplot(temp, ggplot2::aes(
                x = stats::reorder(Name, -freq),
                y = freq
              ))
          }
          
          b <-
            gg +
            ggplot2::geom_col(colour = "white", fill = "royalblue") +
            ggplot2::coord_flip() +
            our_theme +
            ggplot2::labs(x = "Tests", y = "% of records flagged",
                          title = column_to_map) +
            ggplot2::geom_hline(
              yintercept = 0,
              linewidth = 1,
              colour = "#333333"
            ) +
            ggplot2::scale_y_continuous(expand = c(0, 0), 
                                        labels = fancy_scientific)
          
        }
      
      # Names of columns available for creating barplot
      bar <- c(
        ".scientificName_empty",
        ".coordinates_empty",
        ".coordinates_outOfRange",
        ".invalid_basis_of_records",
        ".coordinates_country_inconsistent",
        ".uncer_term",
        ".rou",
        ".equ",
        ".zer",
        ".cap",
        ".cen",
        ".otl",
        ".gbf",
        ".inst",
        ".urb",
        ".dpl",
        ".eventDate_empty",
        ".year_outOfRange",
        ".summary",
        "summary_all_tests"
      )
      
      # Names of columns available for creating maps
      maps <-
        c(
          ".coordinates_country_inconsistent",
          ".equ",
          ".cap",
          ".cen",
          ".otl",
          ".inst",
          ".urb",
          ".dpl",
          ".rou"
        )
      
      # Names of column available for creating histogram
      hist <- c("year")
      
      # Find which names were provided
      w_bar <- dplyr::intersect(col_to_tests, bar)
      w_maps <- dplyr::intersect(col_to_tests, maps)
      w_tranposed <- dplyr::intersect(col_to_tests, "coordinates_transposed")
      w_hist <- hist
      
      # List for saving figures
      res <- list()
      
      # Create bar plots
      if (length(w_bar) == 0 & workflow_step %in% c("prefilter", "space")) {
        message("At least one column 'starting with '.' must be provided")
      }
      
      if (length(w_bar) != 0) {
        w <- which(colSums(!data[{{ w_bar }}], na.rm = TRUE) == 0)
        
        if (length(w) != 0) {
          message(
            "No records flagged for the following tests:\n",
            paste(w_bar[w], collapse = " ")
          )
          w_bar <- w_bar[-w]
        }
        
        if (nrow(n_record_database) != 1 & length(w_bar) != 0) {
          for (i in 1:length(w_bar)) {
            bp <- 
              create_barplot_database(
                data = data,
                column_to_map = w_bar[i],
                workflow_step = workflow_step
              )
            
            bp_list <- list(bp)
            names(bp_list) <-  w_bar[i]
            res <- c(res, bp_list)
          }
        }
        
        # summary of all tests
        if (nrow(n_record_database) != 1 & length(w_bar) != 0) {
          bp_all <-
            create_barplot_all_tests(
              data = data,
              column_to_map = "summary_all_tests",
              workflow_step = workflow_step)
          
          bp_all_list <- list(bp_all)
          names(bp_all_list) <- "summary_all_tests"
          res <- c(res, bp_all_list)
        }
        
      }
      
      # Create maps of invalid vs valid records
      if (length(w_maps) == 0 & workflow_step %in% c("prefilter", "space")) {
        message("At least one of the following columns must be provided for creating maps\n", paste0(maps, sep = " "))
      }
      
      if (length(w_maps) != 0) {
        w <- which(colSums(!data[{{ w_maps }}]) == 0)
        
        if (length(w) != 0) {
          message(
            "No records flagged for the following tests:\n",
            paste(w_maps[w], collapse = " ")
          )
          w_maps <- w_maps[-w]
        }
        
        # Worldmap
        m <- rnaturalearth::ne_countries(returnclass = "sf")
        
        # new theme
        our_theme2 <-
          ggplot2::theme_classic() +
          ggplot2::theme(panel.border = ggplot2::element_blank(),
                         panel.grid.major = ggplot2::element_blank(),
                         panel.grid.minor = ggplot2::element_blank(),
          )
        
        for (i in 1:length(w_maps)) {
          d <-
            CoordinateCleaner::cc_val(
              data,
              lon = "decimalLongitude",
              lat = "decimalLatitude",
              verbose = FALSE,
              value = "clean"
            )
          
          d <-
            d %>%
            dplyr::select(tidyselect::all_of(w_maps[i]), decimalLongitude, decimalLatitude) %>%
            dplyr::filter(.data[[w_maps[i]]] == FALSE)
          
          if (nrow(d) > 0){
            p <-
              ggplot2::ggplot() +
              ggplot2::geom_sf(
                data = m,
                ggplot2::aes(),
                fill = "gray75", colour = "gray88"
              ) +
              ggplot2::coord_sf() + 
              ggplot2::geom_hex(
                data = d,
                ggplot2::aes(x = decimalLongitude, y = decimalLatitude),
                binwidth = 3
                #bins = 75
              ) +
              #ggplot2::coord_quickmap() +
              ggplot2::theme_void() +
              ggplot2::labs(fill = "# of Records",
                            title = convertNames[w_maps[i]][[1]]) +
              ggplot2::scale_fill_viridis_c() +
              our_theme2
            
            p_list <- list(p)
            names(p_list) <-  w_maps[i]
            res <- c(res, p_list)
          }
        }
      }
      
      # Create maps of transposed and corrected coordinates
      if (length(w_tranposed) == 0 & workflow_step == "prefilter") {
        message("file 'Output/Check/01_coordinates_transposed.csv' not found. Maps showing the results of bdc_coordinates_transposed test will not be created")
      }
      
      if (length(w_tranposed) != 0) {
        temp <- readr::read_csv("Output/Check/01_coordinates_transposed.csv")
        
        p1 <-
          bdc::bdc_quickmap(
            data = temp,
            lon = "decimalLongitude",
            lat = "decimalLatitude",
            col_to_map = "#EC364F",
            size = 0.7
          )
        
        p2 <-
          bdc::bdc_quickmap(
            data = temp,
            lon = "decimalLongitude_modified",
            lat = "decimalLatitude_modified",
            col_to_map = "royalblue",
            size = 0.7
          )
        
        pt <- cowplot::plot_grid(p1, p2, nrow = 2)
        
        pt_list <- list(pt)
        names(pt_list) <-  "coordinates_transposed"
        res <- c(res, pt_list)
      }
      
      if (length(w_hist) == 0 & workflow_step == "time") {
        message("Column 'year' not found")
      }
      
      if (length(w_hist) != 0 & workflow_step == "time") {
        data <-
          data %>%
          dplyr::filter(.summary == TRUE)
        
        min_year <- min(data$year, na.rm = T)
        max_year <- max(data$year, na.rm = T)
        
        t <- data %>%
          ggplot2::ggplot(ggplot2::aes(x = year)) +
          ggplot2::geom_histogram(
            colour = "white",
            fill = "royalblue", position = "identity", bins = 80
          ) +
          our_theme +
          ggplot2::labs(x = "Year", y = "Number of records",
                        title = convertNames[w_bar[i]][[1]]) +
          ggplot2::geom_hline(
            yintercept = 0,
            linewidth = 1,
            colour = "#333333"
          ) +
          ggplot2::scale_y_continuous(labels = fancy_scientific)
        
        t_list <- list(t)
        names(t_list) <-  "year"
        res <- c(res, t_list)
        
      }
    }) # suppresswarning
    
      #### start IF ####
    if (save_figures == TRUE){
      
      for (i in seq_along(res)) {
        column_to_map <- names(res)[i]
        ggplot2::ggsave(
          paste(path,
            "/",
            workflow_step,
            "_",
            column_to_map,
            "_",
            "BAR",
            ".png",
            sep = ""
          ),
          res[[i]],
          dpi = 300,
          width = 6,
          height = 3,
          units = "cm",
          scale = 5
        )
      }
      
      message("Check figures in ", paste(path, sep = "/"))
    }
    
    return(res)
  }

requireNamespace("dplyr")
skip_if_not_installed("bdc")

beesFlagged <- BeeBDC::beesFlagged


#### 1.0 prefilter ####
figuresP <-
   BeeBDC::jbd_create_figures(
     data = dplyr::tibble(beesFlagged %>% dplyr::select(!.uncer_terms)),
     path = paste0(tempdir()),
     database_id = "database_id",
     workflow_step = "prefilter",
     save_figures = TRUE)


# Test the number of time figures
testthat::test_that("jbd_create_figures number of figures", {
  testthat::expect_equal(length(figuresP), 5)
})
   testthat::test_that("jbd_create_figures expected class", {
     testthat::expect_type(figuresP, "list")
   })
   testthat::test_that("jbd_create_figures expected class", {
     testthat::expect_type(figuresP[[1]], "object")
   })



#### 2.0 space ####
figuresS <-
  BeeBDC::jbd_create_figures(
    data = dplyr::tibble(beesFlagged %>% dplyr::select(!.uncer_terms)),
    path = paste0(tempdir()),
    database_id = "database_id",
    workflow_step = "space",
    save_figures = TRUE)

# Test the number of space figures
testthat::test_that("jbd_create_figures number of figures", {
  testthat::expect_equal(length(figuresS), 4)
})
   testthat::test_that("jbd_create_figures expected class", {
     testthat::expect_type(figuresS, "list")
   })
   testthat::test_that("jbd_create_figures expected class", {
     testthat::expect_type(figuresS[[1]], "object")
   })

#### 3.0 time ####
figuresT <-
    BeeBDC::jbd_create_figures(
      data = dplyr::tibble(beesFlagged %>% dplyr::select(!.uncer_terms)),
      path = paste0(tempdir()),
      database_id = "database_id",
      workflow_step = "time",
      save_figures = TRUE)


# Test the number of time figures
testthat::test_that("jbd_create_figures number of figures", {
  testthat::expect_equal(length(figuresT), 5)
})
 testthat::test_that("jbd_create_figures expected class", {
   testthat::expect_type(figuresT, "list")
 })
 testthat::test_that("jbd_create_figures expected class", {
   testthat::expect_type(figuresT[[1]], "object")
 })

#####  2.3 dataReader ####
#' @importFrom dplyr %>%
# Read in occurence data with the correct format to be merged
dataReader <-  function(path_i, home_path){
    # locally bind variables to the function
dplyr <- mgsub <- setNames <- . <- family <- day <- NULL
  
requireNamespace("dplyr", "mgsub")

  #Set up bee family list
  Bee_Families <- c("Andrenidae","Apidae", "Colletidae","Halictidae","Megachilidae","Melittidae",
                    "Stenotritidae","andrenidae","apidae", "colletidae","halictidae","megachilidae",
                    "melittidae","stenotritidae")  # Find the paths
  occ_paths <- repoFinder(path = path_i)
  # Get the column types
  ColTypes <- ColTypeR()
  # Get the columns to keep
  ColsToKeep <- names(ColTypes$cols)
  # Make an internal copy of the template for use in the loop as the template tibble
  data_template <- matrix(ncol = length(BeeBDC::ColTypeR()[[1]] %>% names()), nrow = 0) %>% as.data.frame() %>% 
    setNames(BeeBDC::ColTypeR()[[1]] %>% names()) %>% dplyr::tibble() %>%
    dplyr::mutate(dplyr::across(dplyr::everything(), as.character))
  
  #### ALA data ####
  if(grepl("ALA_data", names(path_i)) == "TRUE"){
    # Import these data
    data_i <- readr::read_csv(path_i, col_names = TRUE,
                              # read in all columns as character for now
                              col_types = readr::cols(.default = readr::col_character()),
                              name_repair = "minimal") %>% 
      # Suppress warnings from read_csv
      suppressWarnings(., classes = "warning") %>% 
      # Select only the unique columns
      dplyr::select(unique(colnames(.), fromLast = TRUE))
    # Change column names to match other datasets
    # Remove dcterms: prefixes from some column names
    colnames(data_i) <- mgsub::mgsub(colnames(data_i), 
                                     c("dcterms:"),
                                     c(""))
    # Filter the columns to only those that we want to select, based on the ColsToKeep vector 
    # that we created at the top of the R-script
    if("recordID" %in% colnames(data_i)){
      data_i <- data_i %>%
        # The dataframe to match with
        dplyr::rename("id" = "recordID") 
    }
      
    data_i <- data_i %>%  
      # select columns that match the following string
      dplyr::select( dplyr::matches(      
        # Use the carrot "^" to signify start of string and dollar sign "$" to signify end of 
        # string. Effectivly, this will only return an exact match.
        paste("^",ColsToKeep,"$",sep="") ))  
    # Define the column types to match our standard
    data_i <- data_i %>% # readr::type_convert(data_i, col_types = cols(.default = "c")) %>%
        # Keep only the columns defined in ColTypeR
      dplyr::select(tidyselect::any_of(c(ColsToKeep))) %>%
      # Select only Bee families (in case this is needed, e.g. if you downloaded "Apoidea")
      dplyr::filter(family %in% c(Bee_Families, "", "NA") )
  } # END ALA data IF statement
  
  #### GBIF data ####
  if(grepl("GBIF_data", names(path_i)) == "TRUE"){
    # Read in each file and then merge together
    data_i <- readr::read_tsv(path_i, 
                              quote = "", col_names = TRUE,
                              col_types = readr::cols(.default = readr::col_character())) %>% 
      # Supress warnings from read_tsv
      suppressWarnings(., classes = "warning") %>% 
      # Include all columns from original template file
      dplyr::bind_rows(., data_template) %>%
      # Keep only the columns defined in ColTypeR
      dplyr::select(tidyselect::any_of(c(ColsToKeep))) 
      
    
  } # END GBIF data IF statement

  ####  iDigBio data   #### 
  if(grepl("iDigBio_data", names(path_i)) == "TRUE"){ # Start iDigBio IF statement
    # Import these data
    data_i <- readr::read_csv(path_i, col_names = TRUE,
                              # read in all columns as character for now
                              col_types = readr::cols(.default = readr::col_character()),
                              # Do not keep the some columns
                                col_select = !c("abcd:typifiedName",  "aec:associatedTaxa",
                                                "ala:photographer","ala:species","ala:subfamily",
                                                "ala:subspecies","ala:superfamily","chrono:ChronometricAge",
                                                "dc:language","dc:rights","dc:type",
                                                "dcterms:accessRights","dcterms:bibliographicCitation","dcterms:language",                    
                                                "dcterms:license","dcterms:modified","dcterms:references",                 
                                                "dcterms:rights","dcterms:rightsHolder","dcterms:source",                   
                                                "dcterms:type","obis:ExtendedMeasurementOrFact",
                                                "symbiota:recordEnteredBy","symbiota:recordID",
                                                "symbiota:verbatimScientificName",   
                                                "taxonRankID","zan:ChronometricDate")) %>%
      # Supress warnings from read_csv
      suppressWarnings(., classes = "warning")
    # Change column names to match other datasets
    # Remove dwc. (darwin core), idigbio. and gbif. prefixes from column names
    colnames(data_i) <- mgsub::mgsub(colnames(data_i), 
                                     c("dwc.", "idigbio.", "gbif.", "aec:"),
                                     c("", "","",""))
    # Filter the columns to only those that we want to select, based on the ColsToKeep vector 
    # that we created at the top of the R-script
    data_i <- data_i %>%  # The dataframe to match with
      # select columns that match the following string
      dplyr::select( dplyr::matches(      
        # Use the carrot "^" to signify start of string and dollar sign "$" to signify end of 
        # string. Effectivly, this will only return an exact match.
        paste("^",ColsToKeep,"$",sep="") ))  %>%
          # Remove day ranges that will stop occurrences from being read in.
      dplyr::mutate(day = stringr::str_replace(day, pattern = " -.*|/.*|-.*", replacement = "" ))
    # Define the column types to match our standard
    data_i <- data_i %>%
      # Select only Bee families (in case this is needed, e.g. if you downloaded "Apoidea")
      dplyr::filter(family %in% c(Bee_Families, "", "NA", NA) ) %>%
      # Keep only the columns defined in ColTypeR
      dplyr::select(tidyselect::any_of(c(ColsToKeep))) # %>%
      # Define the column types to match our standard
      # readr::type_convert(col_types = ColTypes) 
  } # End iDigBio IF statement
      
  
  ####  SCAN data #### 
  if(grepl("SCAN_data", names(path_i)) == "TRUE"){ # Start SCAN IF statement
    data_i <- readr::read_csv(path_i, col_names = TRUE,
                              # read in all columns as character for now
                              col_types = readr::cols(.default = readr::col_character())) %>% 
      # Supress warnings from read_csv
      suppressWarnings(., classes = "warning") %>% 
      # Filter the columns to only those that we want to select, based on the ColsToKeep vector 
      # that we created at the top of the R-script
      dplyr::rename("fieldNotes" = "occurrenceRemarks") %>%
      # select columns that match the following string
      dplyr::select( tidyselect::any_of(ColsToKeep)) %>%
      # Select only Bee families (in case )
      dplyr::filter(family %in% c(Bee_Families, "", "NA") ) %>%
    # Keep only the columns defined in ColTypeR
    dplyr::select(tidyselect::any_of(c(ColsToKeep))) # %>%
      # Define the column types to match our standard
        # readr::type_convert(col_types = ColTypes, 
        #                     guess_integer = TRUE) 
  }  # END SCAN IF statement
  return(data_i)
}

# Additional inputs were made by mjwestgate to allow continued support with galah V2


##### 1. atlasDownloader ####
#' Download occurrence data from the Atlas of Living Australia (ALA)
#'
#'
#' Downloads ALA data and creates a new file in the path to put those data. This function can also
#' request downloads from other atlases (see: http://galah.ala.org.au/articles/choosing_an_atlas.html).
#' However, it will only send the download to your email and you must do the rest yourself at this point.
#' 
#' @param path A character directory. The path to a folder where the download will be stored.
#' @param userEmail A character string. The email used associated with the user's ALA account; 
#' user must make an ALA account to download data. 
#' @param ALA_taxon A character string. The taxon to download from ALA. Uses [galah::galah_identify()]
#' @param DL_reason Numeric. The reason for data download according to [galah::galah_config()]
#' @param atlas Character. The atlas to download occurrence data from - see here https://galah.ala.org.au/R/articles/choosing_an_atlas.html for details.
#' Note: the default is "ALA" and is probably the only atlas which will work seamlessly with the rest
#' of the workflow. However, different atlases can still be downloaded and a doi will be sent to 
#' your email.
#'
#' @return Completes an ALA data download and saves those data to the path provided.
#' 
#' @importFrom dplyr %>%
#' @importFrom utils unzip
#' @export
#'
#' @examples
#' \dontrun{
#' atlasDownloader(path = DataPath,
#'                userEmail = "InsertYourEmail",
#'                ALA_taxon = "Apiformes",
#'                DL_reason = 4)
#'                }
atlasDownloader <- function(path, userEmail = NULL, ALA_taxon, DL_reason = 4, atlas = "ALA"){
  # locally bind variables to the function
  . <- file_name <- NULL
  
  #### Intro checks ####
  writeLines(paste("1.","\n",
                   " - Note: galah has a 50 million record download limit.", "\n",
                   "You may call atlas_counts() to check.", "\n",
                   " - Additionally, you must register your email with your ", atlas, " otherwise you will get an ",
                   "error message.", "\n",
                   "See here - https://www.ala.org.au - or your relevant atlas","\n",
                   " - Valid donwload reasons include can be found by running show_all_reasons()",
                   sep = ""))
  # Check for a userEmail input present and halt if FALSE
  if(exists("userEmail") == FALSE){
    stop("You must provide a userEmail for the ",atlas," download.")
  }
  # Check for a userEmail format and halt if FALSE
  if(grepl( ".[^@]+@{1}.+\\..+", userEmail) == FALSE){
    stop("The email you entered might be incorrect, please double-check the format.")
  }
  requireNamespace("galah")
  # Define ColsToKeep
  ColsToKeep <- BeeBDC::ColTypeR()[[1]] %>% names()
  # Create a new working directory for ALA data in the path provided
  dir.create(paste0(path, "/", atlas, "_galah_path", sep = ""), showWarnings = FALSE)
  atlas_galah_path <- paste0(path, "/", atlas, "_galah_path")
  # Set up the ALA download configuration
  writeLines(" - Setting galah configuration.")
  galah::galah_config(directory = atlas_galah_path, 
                      download_reason_id = DL_reason,
                      verbose=TRUE, 
                      email = userEmail, 
                      send_email = TRUE, 
                      atlas = atlas) 

  #### ALA download ####
  # Choose ALA columns to download
  # Thankfully, ALA has a fantastic r package, galah, that allows easy download of occurrence data.
  # Thank you, ALA <3
  # DOWNLOAD ALA data here
  # Apiformes is an informal name that is helpful to select the bee families out of the superfamily Apoidea.
  writeLines(paste("2.","\n",
                   " - Beginning atlas download via galah.", "\n",
                   "A progress bar of your download should appear shortly. You will also receive an email ",
                   "when your download is complete.", sep = ""))
  # Use of `Sys.Date()` comes with the risk that consecutive downloads on the same day will 
    # overwrite each other, even if they are for different queries
  # Note: `file_name` given above is chosen for consistency with previous version of BeeBDC
  file_name <- paste0("galah_download_", Sys.Date(), ".zip")
  ALA_Occurence_download <- galah::galah_call() %>%
    galah::galah_identify(ALA_taxon) %>%
    galah::galah_select(tidyselect::any_of(ColsToKeep)) %>% 
    galah::atlas_occurrences(mint_doi = FALSE, file = file_name)
  # get download attributes from file and make it into a dataframe
  attrs_ALA_Occurence_download <- attributes(ALA_Occurence_download) 
  
  writeLines(paste("3.","\n"," - atlas download is complete.", "\n",
                   "The script will now unzip all of the data and metadata to ",
                   atlas_galah_path, ". This may take a short while.",
                   sep = ""))
  
  # unzip the file
  utils::unzip(
    # File to unzip
    zipfile = paste(atlas_galah_path,
                    "/galah_download_", Sys.Date(), ".zip", 
                    sep = ""), 
    # Where to put the extracted file
    exdir = paste(atlas_galah_path,
                  "/galah_download_folder", 
                  sep = ""),
    overwrite = TRUE) 

  
  #### Save data ####
  # Save some download information
  dplyr::tibble(
    downloaders_email = userEmail,
    taxon = ALA_taxon,
    doi = attr(ALA_Occurence_download, "doi"),
    search_url = attr(ALA_Occurence_download, "search_url"),
    # data_type = attrs_ALA_Occurence_download$data_type, # not supported post galah v.2
    # data_request = paste(dplyr::lst(attrs_ALA_Occurence_download$data_request)), # not supported post galah v.2
    ALA_download_reason = DL_reason,
    download_date = Sys.Date()) %>%
    write_excel_csv(file = paste(atlas_galah_path,
                           "/galah_download_folder/", 
                           "galah_DL_info.csv", 
                           sep = ""))
  # Write user instructions
  writeLines(paste("4.","\n"," - Fin.",
                   sep = ""))
  
  return(ALA_Occurence_download)
}
##### Current end ALA ####

# This function was written by James B. Dorey starting on the 19th of April 2023 to wrap iNEXT
# and output table for multiple species/countries/variables at once

#' Parallel estimation of species richness in a community using iNEXT
#' 
#' A wrapper for `iNEXT::iNEXT()` to interpolate and extrapolate Hill numbers with order q 
#' (rarify species richness). The wrapper has the ability to estimate species richness for 
#' multiple sites (or countries) at once and to do this using multiple cores.
#'
#' @param data A data frame or tibble. A data frame containing "abundance"-type data per variable
#' (population, country, species...) in columns.
#' @param variableColumn Character. The column to be used to group the data. Probably "country"
#' or "site". Default =  "groupVariable".
#' @param valueColumn Character. The column containing the count data. Defualt =  "n".
#' @param q a number or vector specifying the richness order(s) of Hill numbers.
#' @param datatype data type of input data: individual-based abundance data 
#' (datatype = "abundance"), sampling-unit-based incidence frequencies data 
#' (datatype = "incidence_freq") or species by sampling-units incidence matrix 
#' (datatype = "incidence_raw").
#' @param size an integer vector of sample sizes (number of individuals or sampling units) for 
#' which richness estimates will be computed. If NULL, then richness estimates will be computed 
#' for those sample sizes determined by the specified/default endpoint and knots.
#' @param endpoint an integer specifying the sample size that is the endpoint for 
#' rarefaction/extrapolation. If NULL, then endpoint = double the reference sample size.
#' @param knots an integer specifying the number of equally-spaced knots (say K, default is 40) 
#' between size 1 and the endpoint; each knot represents a particular sample size for which 
#' richness estimate will be calculated. If the endpoint is smaller than the reference sample 
#' size, then iNEXT() computes only the rarefaction esimates for approximately K evenly spaced
#' knots. If the endpoint is larger than the reference sample size, then iNEXT() computes 
#' rarefaction estimates for approximately K/2 evenly spaced knots between sample size 1 and 
#' the reference sample size, and computes extrapolation estimates for approximately K/2 evenly 
#' spaced knots between the reference sample size and the endpoint.
#' @param se a logical variable to calculate the bootstrap standard error and conf confidence 
#' interval.
#' @param conf 	a positive number < 1 specifying the level of confidence interval; default is 0.95.
#' @param nboot an integer specifying the number of replications; default is 50.
#' @param mc.cores Numeric. If > 1, the function will run in parallel
#' using mclapply using the number of cores specified. If = 1 then it will be run using a serial
#' loop. NOTE: Windows machines must use a value of 1 (see ?parallel::mclapply). Additionally,
#' be aware that each thread can use large chunks of memory.
#' Default = 1.
#'
#' @return Returns a list containing two tibbles. The first is a tibble that concatenates the outputs
#' from the basic data and rare species information in columns per input variable (column). The second is a tibble
#' that concatenates the various species richness estimates, with input variables in chunks of rows.
#' Additionally a console output will list the variables (columns) that lacked sufficient data
#' to be analysed.
#' 
#' 
#' @importFrom dplyr %>%
#' 
#' @seealso [iNEXT::iNEXT()] 
#' 
#' @export
#'
#' @examples
#' 
#' \dontrun{
#'   # Read in some example data and use [BeeBDC::richnessPrepR()] to create the example input data
#'  #' data(beesCountrySubset)
#' 
#' estimateDataExample <- BeeBDC::richnessPrepR(
#'   data = beesCountrySubset,
#'   # Download the taxonomy
#'   taxonomyFile = BeeBDC::beesTaxonomy(),
#'   # Download the checklist
#'   checklistFile = BeeBDC::beesChecklist(),
#'   curveFunction = function(x) (228.7531 * x * x^-log(12.1593)),
#'   sampleSize = 10000,
#'   countryColumn = "country_suggested",
#'   limitGlobal = NULL,
#'   outPath = tempdir()
#' )
#' 
#'   # In this function we can directly feed in estimateDataExample$site_speciesCounts 
#'  iNextOut <- iNEXTwrapper(data = estimateDataExample$site_speciesCounts,
#'                           variableColumn = "country_suggested",
#'                           valueColumn = "n",
#'                           q = 0,
#'                           datatype = "abundance",
#'                           conf = 0.95,
#'                           se = TRUE,
#'                           nboot = 50,
#'                           size = NULL,
#'                           endpoint = NULL,
#'                           knots = 40,
#'                           mc.cores = 1)
#' }

iNEXTwrapper <- function(data = NULL,
                         variableColumn = "groupVariable",
                         valueColumn = "n",
                         q = 0,
                         datatype = "abundance",
                         conf = 0.95,
                         se = TRUE,
                         nboot = 50,
                         size = NULL,
                         endpoint = NULL,
                         knots = 40,
                         mc.cores = 1){
  
  groupVariable <- groupCount <- . <- NULL
  
  # Load required packages 
  requireNamespace("stringr")
  requireNamespace("dplyr")
  requireNamespace("iNEXT")
  
  
  #### 0.0 Prep ####
  ##### 0.1 Errors ####
  ###### a. FATAL errors ####
  if(is.null(data)){
    stop(" - Please provide an argument for data I'm a program not a magician.")
  }
  
  
  ##### 0.2 iNEXT test ####
  ###### a. test ####
  # Check if iNEXT is installed
  # TRUE if iNEXT is found
  suppressWarnings(
    suggestedTest <- system.file(package='iNEXT') %>% 
      stringr::str_count() > 0 
  )
  
  ###### b. iNEXT ####
  if(suggestedTest == FALSE){
    # Set up instructions for download on fail
    instructions <- paste(" Please try installing the package for yourself", 
                          "using the following command: \n",
                          "install.packages(\"iNEXT\")")
    # Set up fail function for tryCatch
    error_func <- function(e){
      stop(paste("Failed to install the iNEXT package.\n", 
                 instructions))
    }
    # Begin interactive input
    input <- 1
    if (interactive()){
      input <- utils::menu(c("Yes", "No"), 
                           title = paste0("Install the iNEXT package? \n"))
    }
    if(input == 1){
      # Start iNEXT install
      message("Installing the iNEXT package.")
      tryCatch(
        utils::install.packages("iNEXT"), 
        error = error_func, warning = error_func)
    } # END input == 1
    
    else{
      stop(writeLines(paste("The iNEXT package is necessary for BeeBDC::iNEXTwrapper\n", 
                            instructions)))
    } # END else
  } # END suggestedTest == FALSE
  
  
  ##### 0.3 Functions ####
  # Duplicate a function within ChaoSpecies to not run data-poor species
  #f <- function(i, data) {
  #  length(data[which(data == i)])
  #}
  
  # wrap the iNEXT function
  wrapper <- function(inputData_i = df_list[[1]]){
      ###### a. data prep ####
      # Format the data coming in
    inputData_i <- inputData_i[[1]] %>%
      dplyr::select(tidyselect::any_of(c("groupVariable", "groupCount"))) %>%  # Create the rownames
      dplyr::tibble()
    
     ###### b. iNEXT ####
    # Don't run species where there are no counts that are less than k
    # i.e. ONLY run species that won't throw an error due to weird/poor cases of data
    # if(any(inputData_i[[1]] > 0 & inputData_i[[1]] < 5) & 
    #    !(f(1, inputData_i[[1]]) == sum(inputData_i[[1]], na.rm = TRUE))
    # ){
      # Run the ChaoSpecies function
      spOutput <- iNEXT::iNEXT(
        x = inputData_i$groupCount,
        datatype = datatype,
        q = q,
        conf = conf,
        se = se,
        nboot = nboot,
        size = size,
        endpoint = endpoint,
        knots = knots)

        ###### c. DataInfo ####
        # Extract the dataInfo and then replace the Assemblage with the groupVariable value
      dataInfo_out <- spOutput$DataInfo %>%
        dplyr::mutate(Assemblage = inputData_i$groupVariable %>% unique())
      
      ###### d. AsyEst ####
        # Save AsyEst and replace the rownames with a column 
      AsyEst_out <- spOutput$AsyEst %>%
        dplyr::mutate(statistic = rownames(.), .before = 1) %>% 
        dplyr::tibble() %>%
          # Add the grouping variable to the start
        dplyr::mutate(groupVariable = inputData_i$groupVariable %>% unique(),
                      .before = 1)
        
      
      ###### e. iNextEst ####
    # Save the iNextEst as a list that's named per groupVariable
      iNextEst <- spOutput #%>%
        #dplyr::lst() %>%
        #stats::setNames(inputData_i$groupVariable %>% unique())
      
      
      # Return the data as a list
      return(list( dataInfo_out, AsyEst_out, iNextEst) %>%
               stats::setNames(c( "DataInfo", "AsyEst", "iNextEst")))
    # }
  }
  
  #### 1.0 Data prep ####
    ##### 1.1 Prep tibble ####
    # Prep the correct column
  # Temporarily rename the variableColumn to "groupVariable" within the function
  data <- data %>%
    dplyr::rename("groupVariable" = tidyselect::any_of(variableColumn))
  # Temporarily rename the variableColumn to "groupVariable" within the function
  data <- data %>%
    dplyr::rename("groupCount" = tidyselect::any_of(valueColumn))
  
    # Turn the data into a list by the country
  data_list <- data %>%
    tidyr::drop_na(groupVariable) %>% 
    dplyr::group_by(groupVariable) %>%
    dplyr::group_split()
    # Create an empty list to add to
  df_list <- dplyr::lst()
  variableNames <- c()
  
    ##### 1.2 Loop prep ####
  # Loop the data to make it into a list (it's a two-level list)
  for(i in 1:length(data_list)){
    # Extract the list and give it the country name
    loopList <- data_list[[i]] %>%
      dplyr::lst() %>%
      stats::setNames(data_list[[i]]$groupVariable[[1]])
    # Add to the variable list to the greater list
    df_list <- df_list %>%
      append( dplyr::lst(loopList))
    # Add the variable name to a running list to ensure name order is maintained
    variableNames <- c(variableNames, data_list[[i]]$groupVariable[[1]])
  }
  # Set the names for the list
  df_list <- df_list %>%
    stats::setNames(variableNames)
  
 
  
  
  #### 2.0 Run functions ####    
  # Run the function per species or level
  iNEXTOutput <- parallel::mclapply(
    X = df_list,
    FUN = wrapper,
    mc.cores = mc.cores
  ) 
  
  
  #### 3.0 Process outputs ####
  ##### 3.1 seperate F + P ####
  # Find the non-null variables and extract those
  non_empty_list_test <- !sapply(iNEXTOutput <- iNEXTOutput, is.null)
  # Save a list of variables that could not be analysed
  failures <- iNEXTOutput[!non_empty_list_test] %>%
    # Get the failure names 
    names()
  # Remove the failed list
  iNEXTOutput <- iNEXTOutput[non_empty_list_test]
  
  
  ##### 3.2 dataInfo outputs ####
  # Now, combine each level of the list across variables
  # Extract the richness table stuff
  dataInfoOut <- lapply(iNEXTOutput, function(x) x[["DataInfo"]]) %>%
    # Bind together with the original two columns
    dplyr::bind_rows()
  
  
  ##### 3.3 AsyEst outputs ####
  # Row bind the richness statistics into a single table
  iNEXTOutputOut <- lapply(iNEXTOutput, function(x) x[["AsyEst"]]) %>%
    dplyr::bind_rows()
  # Return the variableColumn name to its original state
  names(iNEXTOutputOut)[names(iNEXTOutputOut) == "groupVariable"] <- variableColumn
  
  ##### 3.4 iNEXT outputs ####
    # Combine the iNextEsts ino a single list
  iNextEstOut <- iNEXTOutput %>%
    lapply(., function(x) x) %>%
    dplyr::lst() %>%
    stats::setNames("iNextEst")
  
  
  ##### 3.5 Combine ####
  output <- dplyr::lst(dataInfoOut, iNEXTOutputOut, iNextEstOut) %>% 
    stats::setNames(c("DataInfo", "AsyEst", "iNextEst"))
  
  #### 4.0 User output ####
  # provide some user output
  if(length(failures > 0)){
    writeLines(paste0(
      " - We could not examine the following variables (because of insufficent data or sample size): ",
      paste(failures, collapse = ", ")
    ))}
  
  message(paste0(" - Outputs can be found in a list with two tibbles called 'DataInfo' and",
                    " 'AsyEst' and a list of iNext outputs per groupVariable in iNextEst'."))

  
  # Return the output
  return(output)
  
} # END function

  

  
  

# This function was written by James Dorey to load and output updated summary information from flagging
  # This function was written from the 9th of June 2022. For questions, please email James
# at jbdorey[at]me.com


#' Loads, appends, and saves occurrence flag data 
#' 
#' This function is used to save the flag data for your occurrence data as you run the BeeBDC script. 
#' It will read and append existing files, if asked to. Your flags should also be saved in the occurrence 
#' file itself automatically.
#'
#' @param data A data frame or tibble. Occurrence records as input.
#' @param outPath A character path. Where the file should be saved.
#' @param fileName Character. The name of the file to be saved
#' @param idColumns A character vector. The names of the columns that are to be kept along with the
#' flag columns. These columns should be useful for identifying unique records with flags.
#' Default = c("database_id", "id", "catalogNumber", "occurrenceID", "dataSource").
#' @param append Logical. If TRUE, this will find and append an existing file generated by this function.
#' @param printSummary Logical. If TRUE, print a [summary()] of all filter columns - i.e. those which 
#' tidyselect::starts_with(".")
#'
#' @return Saves a file with id and flag columns and returns this as an object.
#' @export
#' 
#' @importFrom dplyr %>%
#'
#' @examples
#' # Load the example data
#' data("beesFlagged")
#' 
#'   # Run the function
#'   OutPath_Report <- tempdir()
#' flagFile <- flagRecorder(
#'   data = beesFlagged,
#'   outPath = paste(OutPath_Report, sep =""),
#'   fileName = paste0("flagsRecorded_", Sys.Date(), ".csv"),
#'   # These are the columns that will be kept along with the flags
#'   idColumns = c("database_id", "id", "catalogNumber", "occurrenceID", "dataSource"),
#'   # TRUE if you want to find a file from a previous part of the script to append to
#'   append = FALSE)

 
flagRecorder <- function(
    data = NULL,
    outPath = NULL,
    fileName = NULL,
    idColumns = c("database_id", "id", "catalogNumber", "occurrenceID", "dataSource"),
    append = NULL,
    printSummary = FALSE){
  # locally bind variables to the function
  database_id <- . <- rowSum <- .summary <- NULL
  
    #### 0.0 Prep ####
      ##### 0.1 Packages ####
  requireNamespace("dplyr")
  requireNamespace("tidyselect")
  requireNamespace("lubridate")
  requireNamespace("readr")

      ##### 0.2 Warnings ####
  if(is.null(data)){
    warning(" - Please provide a dataset.")
  }
  if(is.null(outPath)){
    warning(" - Please provide an outPath to where the bdc folders are dataset.")
  }
  if(is.null(idColumns)){
    warning(paste(" - No ID columns were selected! We will keep the following id columns intact:\n",
                  "database_id, id, catalogNumber, and occurrenceID.",
            sep = ""))
    idColumns = c("database_id", "id", "catalogNumber", "occurrenceID")
  }
  if(is.null(append)){
    warning(" - Please provide a append argument - TRUE (find existing file) or FALSE (start from scratch).")
  }
  
  
    ##### 0.3 Data in ####
      ###### a. Occurrence dataset ####
  # Select the columns from the input occurrence dataset
  data <- data %>%
    dplyr::select( dplyr::all_of(idColumns), tidyselect::starts_with(".")) 
  
      ###### b. Existing data ####
  if(append == TRUE){
    # Find an existing file
  flagPath <- fileFinder(path = outPath, fileName = "flagsRecorded_")
    # Read it in
  flagColumns <- readr::read_csv(flagPath)
    # Find the new columns that need adding in
  newColumns <- setdiff(colnames(data), colnames(flagColumns))
    # Merge the new columns to the data tibble
  data <- flagColumns %>%
    dplyr::left_join(
        # Select only the new columns to add, from the new tibble.
      dplyr::select(data, c(database_id, dplyr::all_of(newColumns))),
        # Merge by database_id
      by = "database_id", keep = FALSE)
    # Remove the spent dataframe
  rm(flagColumns)
  }
  
  
    #### 1.0 Yes .summary ####
  if(".summary" %in% colnames(data)){
    # Update .summary column
  summaryCol <- data  %>% 
    # Select all columns starting with "."
    dplyr::select(tidyselect::starts_with(".")) %>% 
    # Delete the summary column if it's there
    dplyr::select(!tidyselect::starts_with(".summary")) %>%
    # Make FALSE == 1 and TRUE == 0
    dplyr::mutate_if(is.logical, ~abs(as.numeric(.) - 1)) %>%
    # IF rowSum > 0 then there is at least one flag
    dplyr::mutate(rowSum = rowSums(., na.rm = TRUE)) %>%
    # Add the .summary column
    dplyr::mutate(.summary = dplyr::if_else(rowSum > 0,
                                            FALSE, TRUE)) %>%
    dplyr::select(.summary) 
    # Add this column in
  data <- data %>%
    dplyr::mutate(.summary = summaryCol$.summary)
  # User output
  message(" - .summary column detected. This will be over-written.")
  }
  
  #### 2.0 No .summary ####
  if(!".summary" %in% colnames(data)){
    # Update .summary column
    summaryCol <- data  %>% 
      # Select al lcolumns starting with "."
      dplyr::select(tidyselect::starts_with(".")) %>% 
      # Delete the summary column if it's there
      dplyr::select(!tidyselect::starts_with(".summary")) %>%
      # Make FALSE == 1 and TRUE == 0
      dplyr::mutate_if(is.logical, ~abs(as.numeric(.) - 1)) %>%
      # IF rowSum > 0 then there is at least one flag
      dplyr::mutate(rowSum = rowSums(., na.rm = TRUE)) %>%
      # Add the .summary column
      dplyr::mutate(.summary = dplyr::if_else(rowSum > 0,
                                              FALSE, TRUE)) %>%
      dplyr::select(.summary) 
    # Add this column in
    data <- data %>%
      dplyr::mutate(.summary = summaryCol$.summary)
      # User output
    message(" - NO .summary column detected. This will added to the data.")
  }
  
  #### 3.0 Save ####
    # Save this information as the csv flagsRecorded_DATE.csv
  readr::write_excel_csv(data,
                   paste0(outPath, "/", fileName, sep = ""))
    # User output
  message(paste(
    " - Data saved to ", paste0(outPath, "/", fileName, sep = ""), 
    sep = ""))
  
    # User output
  writeLines(paste(
    " - Selected ", ncol(data), " columns. These include:\n",
    paste(colnames(data)[1:ncol(data)-1], collapse = ", "),
    ", and ", paste(colnames(data)[ncol(data)]),
    sep = ""
  ))
  
  # Print summary if requested
  if(printSummary == TRUE){
      summary(dplyr::select(data, tidyselect::starts_with(".")))
  }
  
  # Return this file
  return(data)
} # END flagRecorder

# Description of the BeeBDC toy dataset beesCountrySubset.csv written by James B. Dorey on the 
# 28th of February 2025

#' A simple scientificName and country_suggested test dataset
#'
#' A small bee occurrence dataset with 1,488 scientificName-country_suggested combinations across 
#' four countries; Fiji, Uganda, Vietnam, and Zambia.
#'
#' @docType data
#'
#' @usage data("beesCountrySubset", package = "BeeBDC")
#'
#' @format An object of class \code{"tibble"}
#' \describe{
#'  \item{scientificName}{Full scientificName as shown on DiscoverLife}
#'  \item{country_suggested}{A country name suggested by the [bdc::bdc_country_standardized()] function.}
#' }
#' @references This data set was created by during the writing of the paper:
#' Dorey, J.B., (upcoming) How many bee species are there? A quantitative global estimate. Journal
#' @keywords datasets
#' @examples
#'
#' beesCountrySubset <- BeeBDC::beesCountrySubset
#' head(beesCountrySubset)
#'
"beesCountrySubset"

# This function was written by James B Dorey to identify occurrence records with potential fill-down
  # errors in the decimalLatitude and decimalLongitude columns. This function was written between
  # the 27th and 28th of May 2022. Please contact James at jbdorey[at]me.com with questions if needed.


#' Find fill-down errors
#' 
#' A simple function that looks for potential latitude and longitude fill-down errors by 
#' identifying consecutive occurrences with coordinates at regular intervals. This is accomplished
#' by using a sliding window with the length determined by minRepeats.
#' 
#' The sliding window (and hence fill-down errors) will only be examined 
#' within the user-defined groupingColumns; if any of those 
#' columns are empty, that record will be excluded.
#'
#' @param data A data frame or tibble. Occurrence records as input.
#' @param minRepeats Numeric. The minimum number of lat or lon repeats needed to flag a record
#' @param groupingColumns Character. The column(s) to group the analysis by and search for fill-down
#' errors within. Default = c("eventDate", "recordedBy", "datasetName").
#' @param ndec Numeric. The number of decimal places below which records will not be considered
#' in the diagonAlley function. This is fed into [BeeBDC::jbd_coordinates_precision()]. Default = 3.
#' @param stepSize Numeric. The number of occurrences to process in each chunk. Default = 1000000.
#' @param mc.cores Numeric. If > 1, the function will run in parallel
#' using mclapply using the number of cores specified. If = 1 then it will be run using a serial
#' loop. NOTE: Windows machines must use a value of 1 (see ?parallel::mclapply). Additionally,
#' be aware that each thread can use large chunks of memory.
#'  Default = 1.
#'
#' @return The function returns the input data with a new column, .sequential, where FALSE = 
#' records that have consecutive latitudes or longitudes greater than or equal to the user-defined 
#' threshold.
#' @export
#' 
#' @importFrom dplyr %>%
#' @importFrom stats complete.cases
#'
#' @examples
#' # Read in the example data
#'   data(beesRaw)
#'  # Run the function
#'   beesRaw_out <- diagonAlley(
#'     data = beesRaw,
#'     # The minimum number of repeats needed to find a sequence in for flagging
#'     minRepeats = 4,
#'     groupingColumns = c("eventDate", "recordedBy", "datasetName"),
#'     ndec = 3,
#'     stepSize = 1000000,
#'     mc.cores = 1)
#'   
#' 
diagonAlley <- function(
  data = NULL,
  minRepeats = NULL,
  groupingColumns = c("eventDate", "recordedBy", "datasetName"),
  ndec = 3,
  stepSize = 1000000,
  mc.cores = 1
  ){
  # locally bind variables to the function
  eventDate<-recordedBy<-decimalLatitude<-decimalLongitude<-database_id<-.data<-leadingLat<-
    laggingLat<-diffLead_Lat<-diffLag_Lat<-diffLat<- . <- NULL
  .rou <- leadingLon <- laggingLon <- diffLead_Lon <- diffLag_Lon <- diffLon <- NULL
  
  #### 0.0 Warnings ####
  if(is.null(data)){
    stop("\n - Please provide an argument for data. I'm a program, not a magician.")
  }
  if(is.null(minRepeats)){
    warning("\n - minRepeats not provided. Using default value of four")
    minRepeats = 4
  }
  if(is.null(groupingColumns)){
    warning("\n - groupingColumns not provided. Using the default of, eventDate, recordedBy, ",
            "and datasetName.")
    groupingColumns <- c("eventDate", "recordedBy", "datasetName")
  }

  #### 1.0 prepare data ####
  startTime <- Sys.time()
  
    ##### 1.1 ndec ####
    # If an ndec is provided, then filter to remove decimal places lower than ndec
  if(!is.null(ndec)){
    writeLines("Removing rounded coordinates with BeeBDC::jbd_coordinates_precision...")
    runningData <- data %>%
    BeeBDC::jbd_coordinates_precision(
      data = .,
      lon = "decimalLongitude",
      lat = "decimalLatitude",
      ndec = ndec,
      quieter = TRUE) %>%
      dplyr::filter(!.rou == FALSE) %>% 
      dplyr::select(!.rou)
  }else{
    runningData <- data
    }
  
    ##### 1.2 Initial filtering and prep ####
  runningData <- runningData %>%
    # Select fewer columns to make it easier on the old computer
    dplyr::select(database_id, decimalLatitude, decimalLongitude, 
                  tidyselect::all_of(groupingColumns)) %>%
      # Remove incomplete values 
    tidyr::drop_na( tidyselect::all_of(groupingColumns)) %>%
    tidyr::drop_na(decimalLatitude, decimalLongitude) %>%
      # Group the data by eventDate and recordedBy
    dplyr::group_by(  dplyr::across(tidyselect::all_of(groupingColumns))) %>%
      # Arrange from biggest to lowest decimalLatitude and then decimalLongitude by grouping
    dplyr::arrange(dplyr::desc(decimalLatitude), dplyr::desc(decimalLongitude), 
                   .by_group = TRUE) %>%
      # Remove duplicate lat and longs
    #dplyr::distinct(decimalLatitude, decimalLongitude, .keep_all = TRUE)  %>%
      # Select those groups with four or more occurrences
    dplyr::filter(dplyr::n() >= minRepeats) 
  
  #### 2.0 Identify sequences ####
  if(nrow(runningData) > 0){

      ##### 2.1 Create function ####
    # Set up the loop function
    LatLonFun <- function(funData){
      for(i in 1:length(funData)){
        # Run the sliding window
        for(j in 1:(nrow(funData) - minRepeats+1)){
          # select an amount of rows based on minRepeats
          windowj <- funData[j:(j+minRepeats-1),]
          # If all differences are equal, then add to a running list of database_ids
          if(all(windowj$diff == windowj$diff[1])){
            flaggedRecords <- flaggedRecords %>% 
              dplyr::bind_rows(windowj %>%  
                                 dplyr::select(database_id) )}  # END if statement
          } # END J loop
        # Keep distinct flaggedRecords
        # Run distinct every 1000th iteration, or at the end
        if(i %in% seq(0, length(funData), 1000) | 
           i == length(funData)){
          flaggedRecords <- flaggedRecords %>%
            dplyr::distinct(.keep_all = TRUE)}
        } # END i loop
      return(flaggedRecords)
      }# End LatLonFun
    
    ##### 2.2 Lat sequences ####
    writeLines(" - Starting the latitude sequence...")
  # Find the groups where ALL of the differences between values is the same (is.sequential)
    # Return their database_id
  runningData_Lat <- runningData %>% 
    # Sort 
    dplyr::arrange(dplyr::desc(.data$decimalLatitude), .by_group = TRUE) %>%
    dplyr::distinct(dplyr::across(c(decimalLongitude, decimalLatitude,
                                    tidyselect::all_of(groupingColumns))), .keep_all = TRUE) %>%
    # Add leading columns with the value of the next one
    dplyr::mutate(leadingLat = dplyr::lag(decimalLatitude)) %>%
    dplyr::mutate(laggingLat = dplyr::lead(decimalLatitude)) %>%
    # Add new new columns with the difference 
    dplyr::mutate(diffLead_Lat = (decimalLatitude - leadingLat)) %>%
    dplyr::mutate(diffLag_Lat = (decimalLatitude - laggingLat)) %>%
    # COMBINE these columns so that they are all complete from lead AND lag (no NAs)
    dplyr::mutate(diff = dplyr::if_else(is.na(diffLead_Lat),
                                           -diffLag_Lat,
                                           diffLead_Lat), 
                  diff = diff %>% round(digits = 9)) %>%
    # Remove extra columns
    dplyr::select(!c(leadingLat, laggingLat, diffLead_Lat, diffLag_Lat)) %>%
    # Remove groups below the threshold
    # Group by lat and lon and the groupingColumns
    dplyr::group_by(dplyr::across(tidyselect::all_of(groupingColumns))) %>%
    dplyr::filter(!dplyr::n() < minRepeats)
  
  
  # Re-join with the runningData and match up duplicate lat/lon within groups and assign the same 
  # diff values
  runningData_Lat <- runningData_Lat %>% 
    dplyr::bind_rows(runningData %>% 
                       dplyr::filter(!database_id %in% runningData_Lat$database_id)) %>%
    # Group by lat and lon and the groupingColumns
    dplyr::group_by(decimalLatitude, decimalLongitude, 
                    dplyr::across(tidyselect::all_of(groupingColumns))) %>%
    dplyr::arrange(decimalLatitude) %>% 
    # Assign matching occurrences to the same diff number so that they will also be flagged
    dplyr::mutate(diff = diff[[1]]) %>%
    tidyr::drop_na(diff) %>%
    dplyr::filter(!diff == 0) %>%
    dplyr::ungroup()
  
    # Turn each of the groups into its own tibble within a list
  runningData_LatGrp <- runningData_Lat %>% 
      # Re-group by the groupingColumns and then filter to those that pass the minimum repeats
    dplyr::group_by( dplyr::across(tidyselect::all_of(groupingColumns)), .add = TRUE) %>%
    dplyr::filter(dplyr::n() >= minRepeats) %>%
    dplyr::mutate(diff = diff %>% as.character()) %>%
      # Split groups into a list
    dplyr::group_split()
  
    # Remove the spent dataset
  rm(runningData_Lat)
  # Remove excess columns from list
  runningData_LatGrp <- lapply(runningData_LatGrp, function(x) x[(names(x) %in% c("database_id", "diff"))])
  
  # Set up a tibble for the flagged records
  flaggedRecords <- dplyr::tibble()
    # Run the loop function in parallel
  flagRecords_Lat <- runningData_LatGrp %>% 
    parallel::mclapply(LatLonFun, mc.cores = mc.cores) %>%
      # Re-bind the list elements 
    dplyr::bind_rows() 
  
    # Remove the spent dataset
  rm(runningData_LatGrp)
  
  
  ##### 2.3 Lon sequences ####
  writeLines(" - Starting the longitude sequence...")
  # Find the groups where ALL of the differences between values is the same (is.sequential)
    # Return their database_id
  runningData_Lon <- runningData %>% 
    # Sort 
    dplyr::arrange(dplyr::desc(.data$decimalLongitude), .by_group = TRUE) %>%
    dplyr::distinct(dplyr::across(c(decimalLongitude, decimalLatitude,
                                  tidyselect::all_of(groupingColumns))), .keep_all = TRUE) %>%
    # Add leading columns with the value of the next one
    dplyr::mutate(leadingLon = dplyr::lag(decimalLongitude)) %>%
    dplyr::mutate(laggingLon = dplyr::lead(decimalLongitude)) %>%
    # Add new new columns with the difference 
    dplyr::mutate(diffLead_Lon = (decimalLongitude - leadingLon)) %>%
    dplyr::mutate(diffLag_Lon = (decimalLongitude - laggingLon)) %>%
    # COMBINE these columns so that they are all complete from lead AND lag (no NAs)
    dplyr::mutate(diff = dplyr::if_else(is.na(diffLead_Lon),
                                           -diffLag_Lon,
                                           diffLead_Lon), 
                  diff = diff %>% round(digits = 9)) %>%
    # Remove extra columns
    dplyr::select(!c(leadingLon, laggingLon, diffLead_Lon, diffLag_Lon)) %>%
      # Remove groups below the threshold
    # Group by lat and lon and the groupingColumns
    dplyr::group_by(dplyr::across(tidyselect::all_of(groupingColumns))) %>%
    dplyr::filter(!dplyr::n() < minRepeats)
  
  
    # Re-join with the runningData and match up duplicate lat/lon within groups and assign the same 
      # diffLon values
  runningData_Lon <- runningData_Lon %>% 
    dplyr::bind_rows(runningData %>% 
                       dplyr::filter(!database_id %in% runningData_Lon$database_id)) %>%
        # Group by lat and lon and the groupingColumns
    dplyr::group_by(decimalLatitude, decimalLongitude, 
                    dplyr::across(tidyselect::all_of(groupingColumns))) %>%
    dplyr::arrange(decimalLongitude) %>% 
    # Assign matching occurrences to the same diffLon number so that they will also be flagged
    dplyr::mutate(diff = diff[[1]]) %>%
    tidyr::drop_na(diff) %>%
    dplyr::filter(!diff == 0) %>%
    dplyr::ungroup()
  
  # Turn each of the groups into its own tibble within a list
  runningData_LonGrp <- runningData_Lon %>% 
    # Re-group by the groupingColumns and then filter to those that pass the minimum repeats
    dplyr::group_by( dplyr::across(tidyselect::all_of(groupingColumns)), .add = TRUE) %>%
    dplyr::filter(dplyr::n() >= minRepeats) %>%
    dplyr::mutate(diff = diff %>% as.character()) %>%
    # Split groups into a list
    dplyr::group_split()
  
    # Remove the spent dataset
  rm(runningData_Lon)
  # Remove excess columns from list
  runningData_LonGrp <- lapply(runningData_LonGrp, function(x) x[(names(x) %in% c("database_id", "diff"))])

  # Run the loop function in parallel
  flagRecords_Lon <- runningData_LonGrp %>% 
    parallel::mclapply(LatLonFun, mc.cores = mc.cores) %>%
    # Re-bind the list elements 
    dplyr::bind_rows() 
  
    ##### 2.4 Merge lat lon ####
    # Merge the minor runs
  flagRecords <- dplyr::bind_rows(flagRecords_Lat, flagRecords_Lon) %>% 
    dplyr::distinct()
  
  }else{
    flagRecords = dplyr::tibble(database_id = NA_character_)
  } # END nrow(runningData) > 0
    # Remove the spent dataset
  rm(runningData_LonGrp)

  
  #### 3.0 Merge ####
  writeLines(" - Merging results and adding the .sequential column...")
    # Add a new column called .sequential to flag sequential lats and longs as FALSE
  data <- data %>%
    dplyr::mutate(.sequential = !database_id %in% flagRecords$database_id)

    # Use output
  message("\ndiagonAlley:\nFlagged ", 
          format(sum(data$.sequential == FALSE, na.rm = TRUE), big.mark = ","),
          " records\nThe .sequential column was added to the database.\n")
  
    # Return runtime
  endTime <- Sys.time()
  
  message(paste(
    " - Completed in ", 
    round(difftime(endTime, startTime), digits = 2 )," ",
    units(round(endTime - startTime, digits = 2)),
    sep = ""))
  return(data)
}# END function

requireNamespace("readr")
requireNamespace("dplyr")
library(dplyr) # couldn't use %>% without this


# this data is derived from SCAN data, but IT HAS BEEN EDITED FOR TESTING AND IS NOT USEFUL/REAL DATA, but by all means go ahead and mine it
data <- dplyr::tribble( # even step coordinates (different for lat/long)
  ~database_id, ~datasetName,       ~id, ~institutionCode, ~collectionCode,         ~ownerInstitutionCode,      ~basisOfRecord,                  ~occurrenceID,      ~catalogNumber, ~otherCatalogNumbers,   ~kingdom,      ~phylum,    ~class,        ~order,  ~family,  ~scientificName, ~taxonID, ~scientificNameAuthorship, ~genus, ~specificEpithet,        ~recordedBy, ~eventDate, ~year, ~month, ~day,        ~verbatimEventDate,        ~country, ~stateProvince,  ~locality,                  ~locationRemarks, ~decimalLatitude, ~decimalLongitude, ~minimumElevationInMeters,                                             ~rights,                                     ~rightsHolder, ~accessRights,                                       ~recordId,                                                                        ~references,
  "fake SCAN1", "fakeDataset",    13775122L,                      "CAS",        "ANTWEB",        "UCDC, Davis, CA, USA", "PreservedSpecimen",     "CAS:ANTWEB:casent0106100",     "casent0106100",                   NA, "Animalia", "Arthropoda", "Insecta", "Hymenoptera", "apidae", "apis mellifera",  235783L,          "Linnaeus, 1758", "Apis",      "mellifera",        "P.S. Ward",  "6/28/05", 2005L,     6L,  28L, "28 Jun 2005/29 Jun 2005", "United States",   "California",    "Davis", "coordinates obtained from Label",           38.541,        -121.7567,                       15L, "http://creativecommons.org/publicdomain/zero/1.0/", "The California Academy of Sciences - AntWeb.org",            NA, "urn:uuid:46a46727-6535-4e70-88e7-a42c98f806ed", "https://scan-bugs.org:443/portal/collections/individual/index.php?occid=13775122",
  "fake SCAN2", "fakeDataset",    13775123L,                      "CAS",        "ANTWEB",        "UCDC, Davis, CA, USA", "PreservedSpecimen", "CAS:ANTWEB:casent0106100-d01", "casent0106100-d01",                   NA, "Animalia", "Arthropoda", "Insecta", "Hymenoptera", "apidae", "apis mellifera",  235783L,          "Linnaeus, 1758", "Apis",      "mellifera",        "P.S. Ward",  "6/28/05", 2005L,     6L,  28L, "28 Jun 2005/29 Jun 2005", "United States",   "California",    "Davis", "coordinates obtained from Label",           38.542,        -121.7568,                       15L, "http://creativecommons.org/publicdomain/zero/1.0/", "The California Academy of Sciences - AntWeb.org",            NA, "urn:uuid:d6ff3ddb-4695-4aaf-ab89-c251ab2fc7e6", "https://scan-bugs.org:443/portal/collections/individual/index.php?occid=13775123",
  "fake SCAN3", "fakeDataset",    13775124L,                      "CAS",        "ANTWEB",        "UCDC, Davis, CA, USA", "PreservedSpecimen", "CAS:ANTWEB:casent0106100-d11", "casent0106100-d11",                   NA, "Animalia", "Arthropoda", "Insecta", "Hymenoptera", "apidae", "apis mellifera",  235783L,          "Linnaeus, 1758", "Apis",      "mellifera",        "P.S. Ward",  "6/28/05", 2005L,     6L,  28L, "28 Jun 2005/29 Jun 2005", "United States",   "California",    "Davis", "coordinates obtained from Label",           38.543,        -121.7569,                       15L, "http://creativecommons.org/publicdomain/zero/1.0/", "The California Academy of Sciences - AntWeb.org",            NA, "urn:uuid:4504a6e6-ab1c-4de7-ac43-a3588a63ae90", "https://scan-bugs.org:443/portal/collections/individual/index.php?occid=13775124",
  "fake SCAN4", "fakeDataset",           NA,                      "CAS",        "ANTWEB",        "UCDC, Davis, CA, USA", "PreservedSpecimen",                             NA,                  NA,                   NA, "Animalia", "Arthropoda", "Insecta", "Hymenoptera", "apidae", "apis mellifera",  235783L,          "Linnaeus, 1758", "Apis",      "mellifera",        "P.S. Ward",  "6/28/05", 2005L,     6L,  28L, "28 Jun 2005/29 Jun 2005", "United States",   "California",    "Davis", "coordinates obtained from Label",           38.544,        -121.7570,                       15L, "http://creativecommons.org/publicdomain/zero/1.0/", "The California Academy of Sciences - AntWeb.org",            NA,                                              NA,                                                                                 NA,
  "fake SCAN5", "fakeDataset",           NA,                      "CAS",        "ANTWEB",        "UCDC, Davis, CA, USA", "PreservedSpecimen",                             NA,                  NA,                   NA, "Animalia", "Arthropoda", "Insecta", "Hymenoptera", "apidae", "apis mellifera",  235783L,          "Linnaeus, 1758", "Apis",      "mellifera",        "P.S. Ward",  "6/28/05", 2005L,     6L,  28L, "28 Jun 2005/29 Jun 2005", "United States",   "California",    "Davis", "coordinates obtained from Label",           38.545,        -121.7571,                       15L, "http://creativecommons.org/publicdomain/zero/1.0/", "The California Academy of Sciences - AntWeb.org",            NA,                                              NA,                                                                                 NA,
  "fake SCAN6", "fakeDataset",           NA,                      "CAS",        "ANTWEB",        "UCDC, Davis, CA, USA", "PreservedSpecimen",                             NA,                  NA,                   NA, "Animalia", "Arthropoda", "Insecta", "Hymenoptera", "apidae", "apis mellifera",  235783L,          "Linnaeus, 1758", "Apis",      "mellifera",        "P.S. Ward",  "6/28/05", 2005L,     6L,  28L, "28 Jun 2005/29 Jun 2005", "United States",   "California",    "Davis", "coordinates obtained from Label",           38.546,        -121.7572,                       15L, "http://creativecommons.org/publicdomain/zero/1.0/", "The California Academy of Sciences - AntWeb.org",            NA,                                              NA,                                                                                 NA,
            
  "fake SCAN7", "fakeDataset",           NA,                      "CAS",        "ANTWEB",        "UCDC, Davis, CA, USA", "PreservedSpecimen",                             NA,                  NA,                   NA, "Animalia", "Arthropoda", "Insecta", "Hymenoptera", "apidae", "apis mellifera",  235783L,          "Linnaeus, 1758", "Apis",      "mellifera",        "P.S. Ward",  "6/28/05", 2005L,     6L,  28L, "28 Jun 2005/29 Jun 2005", "United States",   "California",    "Davis", "coordinates obtained from Label",            38.54,        -121.7567,                       15L, "http://creativecommons.org/publicdomain/zero/1.0/", "The California Academy of Sciences - AntWeb.org",            NA,                                              NA,                                                                                 NA,
  "fake SCAN8", "fakeDataset",           NA,                      "CAS",        "ANTWEB",        "UCDC, Davis, CA, USA", "PreservedSpecimen",                             NA,                  NA,                   NA, "Animalia", "Arthropoda", "Insecta", "Hymenoptera", "apidae", "apis mellifera",  235783L,          "Linnaeus, 1758", "Apis",      "mellifera",        "P.S. Ward",  "6/28/05", 2005L,     6L,  28L, "28 Jun 2005/29 Jun 2005", "United States",   "California",    "Davis", "coordinates obtained from Label",            38.54,        -121.7567,                       15L, "http://creativecommons.org/publicdomain/zero/1.0/", "The California Academy of Sciences - AntWeb.org",            NA,                                              NA,                                                                                 NA,
  "fake SCAN9", "fakeDataset",    13852774L,                      "CAS",        "ANTWEB", "CAS, San Francisco, CA, USA", "PreservedSpecimen",     "CAS:ANTWEB:casent0249086",     "casent0249086",                   NA, "Animalia", "Arthropoda", "Insecta", "Hymenoptera", "apidae", "apis mellifera",  235783L,          "Linnaeus, 1758", "Apis",      "mellifera", "C. Sch‚àö‚àÇning",   "9/1/09", 2009L,     9L,   1L,                "1-Sep-09",         "Kenya",             NA, "Chogoria",                                NA,         -0.20222,          37.5083,                     2277L, "http://creativecommons.org/publicdomain/zero/1.0/", "The California Academy of Sciences - AntWeb.org",            NA, "urn:uuid:964139d3-fb07-4a57-9b7c-e9a5830f4c82", "https://scan-bugs.org:443/portal/collections/individual/index.php?occid=13852774",
  "fake SCAN10", "fakeDataset",           NA,                      "CAS",        "ANTWEB",        "UCDC, Davis, CA, USA", "PreservedSpecimen",                             NA,                  NA,                   NA, "Animalia", "Arthropoda", "Insecta", "Hymenoptera", "apidae", "apis mellifera",  235783L,          "Linnaeus, 1758", "Apis",      "mellifera",        "P.S. Ward",  "6/29/05", 2005L,     6L,  28L, "28 Jun 2005/29 Jun 2005", "United States",   "California",    "Davis", "coordinates obtained from Label",           38.546,        -121.7572,                       15L, "http://creativecommons.org/publicdomain/zero/1.0/", "The California Academy of Sciences - AntWeb.org",            NA,                                              NA,                                                                                 NA,
  "fake SCAN11", "newDataset",           NA,                      "CAS",        "ANTWEB",        "UCDC, Davis, CA, USA", "PreservedSpecimen",                             NA,                  NA,                   NA, "Animalia", "Arthropoda", "Insecta", "Hymenoptera", "apidae", "apis mellifera",  235783L,          "Linnaeus, 1758", "Apis",      "mellifera",        "P.S. Ward",  "6/29/05", 2005L,     6L,  28L, "28 Jun 2005/29 Jun 2005", "United States",   "California",    "Davis", "coordinates obtained from Label",            38.54,        -121.7567,                       15L, "http://creativecommons.org/publicdomain/zero/1.0/", "The California Academy of Sciences - AntWeb.org",            NA,                                              NA,                                                                                 NA,
  "fake SCAN12", "newDataset",           NA,                      "CAS",        "ANTWEB",        "UCDC, Davis, CA, USA", "PreservedSpecimen",                             NA,                  NA,                   NA, "Animalia", "Arthropoda", "Insecta", "Hymenoptera", "apidae", "apis mellifera",  235783L,          "Linnaeus, 1758", "Apis",      "mellifera",        "P.S. Ward",  "6/29/05", 2005L,     6L,  28L, "28 Jun 2005/29 Jun 2005", "United States",   "California",    "Davis", "coordinates obtained from Label",            38.54,        -121.7567,                       15L, "http://creativecommons.org/publicdomain/zero/1.0/", "The California Academy of Sciences - AntWeb.org",            NA,                                              NA,                                                                                 NA,
  "fake SCAN13", "newDataset",    13852774L,                      "CAS",        "ANTWEB", "CAS, San Francisco, CA, USA", "PreservedSpecimen",     "CAS:ANTWEB:casent0249086",     "casent0249086",                   NA, "Animalia", "Arthropoda", "Insecta", "Hymenoptera", "apidae", "apis mellifera",  235783L,          "Linnaeus, 1758", "Apis",      "mellifera", "C. Sch‚àö‚àÇning",   "9/1/09", 2009L,     9L,   1L,                "1-Sep-09",         "Kenya",             NA, "Chogoria",                                NA,         -0.20222,          37.5083,                     2277L, "http://creativecommons.org/publicdomain/zero/1.0/", "The California Academy of Sciences - AntWeb.org",            NA, "urn:uuid:964139d3-fb07-4a57-9b7c-e9a5830f4c82", "https://scan-bugs.org:443/portal/collections/individual/index.php?occid=13852774",
            
  "fake SCAN14", "fakeDataset",           NA,                      "CAS",        "ANTWEB",        "UCDC, Davis, CA, USA", "PreservedSpecimen",                             NA,                  NA,                   NA, "Animalia", "Arthropoda", "Insecta", "Hymenoptera", "apidae", "apis mellifera",  235783L,          "Linnaeus, 1758", "Apis",      "mellifera",        "Ward",  "4/29/05", 2005L,     6L,  28L, "28 Jun 2005/29 Jun 2005", "United States",   "California",    "Davis", "coordinates obtained from Label",           38.54,        -121.756,                       15L, "http://creativecommons.org/publicdomain/zero/1.0/", "The California Academy of Sciences - AntWeb.org",            NA,                                              NA,                                                                                 NA,
  "fake SCAN15", "fakeDataset",           NA,                      "CAS",        "ANTWEB",        "UCDC, Davis, CA, USA", "PreservedSpecimen",                             NA,                  NA,                   NA, "Animalia", "Arthropoda", "Insecta", "Hymenoptera", "apidae", "apis mellifera",  235783L,          "Linnaeus, 1758", "Apis",      "mellifera",        "Ward",  "4/29/05", 2005L,     6L,  28L, "28 Jun 2005/29 Jun 2005", "United States",   "California",    "Davis", "coordinates obtained from Label",            38.54,        -121.756,                       15L, "http://creativecommons.org/publicdomain/zero/1.0/", "The California Academy of Sciences - AntWeb.org",            NA,                                              NA,                                                                                 NA,
  "fake SCAN16", "fakeDataset",           NA,                      "CAS",        "ANTWEB",        "UCDC, Davis, CA, USA", "PreservedSpecimen",                             NA,                  NA,                   NA, "Animalia", "Arthropoda", "Insecta", "Hymenoptera", "apidae", "apis mellifera",  235783L,          "Linnaeus, 1758", "Apis",      "mellifera",        "Ward",  "4/29/05", 2005L,     6L,  28L, "28 Jun 2005/29 Jun 2005", "United States",   "California",    "Davis", "coordinates obtained from Label",            38.54,        -121.756,                       15L, "http://creativecommons.org/publicdomain/zero/1.0/", "The California Academy of Sciences - AntWeb.org",            NA,                                              NA,                                                                                 NA,
  "fake SCAN17", "fakeDataset",           NA,                      "CAS",        "ANTWEB",        "UCDC, Davis, CA, USA", "PreservedSpecimen",                             NA,                  NA,                   NA, "Animalia", "Arthropoda", "Insecta", "Hymenoptera", "apidae", "apis mellifera",  235783L,          "Linnaeus, 1758", "Apis",      "mellifera",        "Ward",  "4/29/05", 2005L,     6L,  28L, "28 Jun 2005/29 Jun 2005", "United States",   "California",    "Davis", "coordinates obtained from Label",            38.54,        -121.756,                       15L, "http://creativecommons.org/publicdomain/zero/1.0/", "The California Academy of Sciences - AntWeb.org",            NA,                                              NA,                                                                                 NA,
            
  "fake SCAN18", "fakeDataset",    13775123L,                      "CAS",        "ANTWEB",        "UCDC, Davis, CA, USA", "PreservedSpecimen", "CAS:ANTWEB:casent0106100-d01", "casent0106100-d01",                   NA, "Animalia", "Arthropoda", "Insecta", "Hymenoptera", "apidae", "apis mellifera",  235783L,          "Linnaeus, 1758", "Apis",      "mellifera",        "BeeNerd",  "6/30/05", 2005L,     6L,  28L, "28 Jun 2005/29 Jun 2005", "United States",   "California",    "Davis", "coordinates obtained from Label",           28.742,        -121.75,                       15L, "http://creativecommons.org/publicdomain/zero/1.0/", "The California Academy of Sciences - AntWeb.org",            NA, "urn:uuid:d6ff3ddb-4695-4aaf-ab89-c251ab2fc7e6", "https://scan-bugs.org:443/portal/collections/individual/index.php?occid=13775123",
  "fake SCAN19", "fakeDataset",    13775124L,                      "CAS",        "ANTWEB",        "UCDC, Davis, CA, USA", "PreservedSpecimen", "CAS:ANTWEB:casent0106100-d11", "casent0106100-d11",                   NA, "Animalia", "Arthropoda", "Insecta", "Hymenoptera", "apidae", "apis mellifera",  235783L,          "Linnaeus, 1758", "Apis",      "mellifera",        "BeeNerd",  "6/30/05", 2005L,     6L,  28L, "28 Jun 2005/29 Jun 2005", "United States",   "California",    "Davis", "coordinates obtained from Label",           28.743,        -121.75,                       15L, "http://creativecommons.org/publicdomain/zero/1.0/", "The California Academy of Sciences - AntWeb.org",            NA, "urn:uuid:4504a6e6-ab1c-4de7-ac43-a3588a63ae90", "https://scan-bugs.org:443/portal/collections/individual/index.php?occid=13775124",
  "fake SCAN20", "fakeDataset",           NA,                      "CAS",        "ANTWEB",        "UCDC, Davis, CA, USA", "PreservedSpecimen",                             NA,                  NA,                   NA, "Animalia", "Arthropoda", "Insecta", "Hymenoptera", "apidae", "apis mellifera",  235783L,          "Linnaeus, 1758", "Apis",      "mellifera",        "BeeNerd",  "6/30/05", 2005L,     6L,  28L, "28 Jun 2005/29 Jun 2005", "United States",   "California",    "Davis", "coordinates obtained from Label",           28.744,        -121.75,                       15L, "http://creativecommons.org/publicdomain/zero/1.0/", "The California Academy of Sciences - AntWeb.org",            NA,                                              NA,                                                                                 NA,
  "fake SCAN21", "fakeDataset",           NA,                      "CAS",        "ANTWEB",        "UCDC, Davis, CA, USA", "PreservedSpecimen",                             NA,                  NA,                   NA, "Animalia", "Arthropoda", "Insecta", "Hymenoptera", "apidae", "apis mellifera",  235783L,          "Linnaeus, 1758", "Apis",      "mellifera",        "BeeNerd",  "6/30/05", 2005L,     6L,  28L, "28 Jun 2005/29 Jun 2005", "United States",   "California",    "Davis", "coordinates obtained from Label",           28.745,        -121.75,                       15L, "http://creativecommons.org/publicdomain/zero/1.0/", "The California Academy of Sciences - AntWeb.org",            NA,                                              NA,                                                                                 NA,
            
  "fake SCAN22", "fakeDataset",    13775123L,                      "CAS",        "ANTWEB",        "UCDC, Davis, CA, USA", "PreservedSpecimen", "CAS:ANTWEB:casent0106100-d01", "casent0106100-d01",                   NA, "Animalia", "Arthropoda", "Insecta", "Hymenoptera", "apidae", "apis mellifera",  235783L,          "Linnaeus, 1758", "Apis",      "mellifera",        "easterEgg",  "6/30/15", 2005L,     6L,  28L, "28 Jun 2005/29 Jun 2005", "United States",   "California",    "Davis", "coordinates obtained from Label",           28.7400,        -121.7512,                       15L, "http://creativecommons.org/publicdomain/zero/1.0/", "The California Academy of Sciences - AntWeb.org",            NA, "urn:uuid:d6ff3ddb-4695-4aaf-ab89-c251ab2fc7e6", "https://scan-bugs.org:443/portal/collections/individual/index.php?occid=13775123",
  "fake SCAN23", "fakeDataset",    13775124L,                      "CAS",        "ANTWEB",        "UCDC, Davis, CA, USA", "PreservedSpecimen", "CAS:ANTWEB:casent0106100-d11", "casent0106100-d11",                   NA, "Animalia", "Arthropoda", "Insecta", "Hymenoptera", "apidae", "apis mellifera",  235783L,          "Linnaeus, 1758", "Apis",      "mellifera",        "easterEgg",  "6/30/15", 2005L,     6L,  28L, "28 Jun 2005/29 Jun 2005", "United States",   "California",    "Davis", "coordinates obtained from Label",           28.7400,        -121.7513,                       15L, "http://creativecommons.org/publicdomain/zero/1.0/", "The California Academy of Sciences - AntWeb.org",            NA, "urn:uuid:4504a6e6-ab1c-4de7-ac43-a3588a63ae90", "https://scan-bugs.org:443/portal/collections/individual/index.php?occid=13775124",
  "fake SCAN24", "fakeDataset",           NA,                      "CAS",        "ANTWEB",        "UCDC, Davis, CA, USA", "PreservedSpecimen",                             NA,                  NA,                   NA, "Animalia", "Arthropoda", "Insecta", "Hymenoptera", "apidae", "apis mellifera",  235783L,          "Linnaeus, 1758", "Apis",      "mellifera",        "easterEgg",  "6/30/15", 2005L,     6L,  28L, "28 Jun 2005/29 Jun 2005", "United States",   "California",    "Davis", "coordinates obtained from Label",           28.7400,        -121.7514,                       15L, "http://creativecommons.org/publicdomain/zero/1.0/", "The California Academy of Sciences - AntWeb.org",            NA,                                              NA,                                                                                 NA,
  "fake SCAN25", "fakeDataset",           NA,                      "CAS",        "ANTWEB",        "UCDC, Davis, CA, USA", "PreservedSpecimen",                             NA,                  NA,                   NA, "Animalia", "Arthropoda", "Insecta", "Hymenoptera", "apidae", "apis mellifera",  235783L,          "Linnaeus, 1758", "Apis",      "mellifera",        "easterEgg",  "6/30/15", 2005L,     6L,  28L, "28 Jun 2005/29 Jun 2005", "United States",   "California",    "Davis", "coordinates obtained from Label",           28.7400,        -121.7515,                       15L, "http://creativecommons.org/publicdomain/zero/1.0/", "The California Academy of Sciences - AntWeb.org",            NA,                                              NA,                                                                                 NA,
  "fake SCAN26", "fakeDataset",    13775124L,                      "CAS",        "ANTWEB",        "UCDC, Davis, CA, USA", "PreservedSpecimen", "CAS:ANTWEB:casent0106100-d11", "casent0106100-d11",                   NA, "Animalia", "Arthropoda", "Insecta", "Hymenoptera", "apidae", "apis mellifera",  235783L,          "Linnaeus, 1758", "Apis",      "mellifera",        "P.S. Ward",  "6/28/05", 2005L,     6L,  28L, "28 Jun 2005/29 Jun 2005", "United States",   "California",    "Davis", "coordinates obtained from Label",           38.543,        -121.7569,                       15L, "http://creativecommons.org/publicdomain/zero/1.0/", "The California Academy of Sciences - AntWeb.org",            NA, "urn:uuid:4504a6e6-ab1c-4de7-ac43-a3588a63ae90", "https://scan-bugs.org:443/portal/collections/individual/index.php?occid=13775124",
            # no matching
  "fake SCAN27", "anotherDataset",  13775123L,                      "CAS",        "ANTWEB",        "UCDC, Davis, CA, USA", "PreservedSpecimen", "CAS:ANTWEB:casent0106100-d01", "casent0106100-d01",                   NA, "Animalia", "Arthropoda", "Insecta", "Hymenoptera", "apidae", "apis mellifera",  235783L,          "Linnaeus, 1758", "Apis",      "mellifera",        "Yadad",  "14/04/09", 2009L,     4L,  14L, "", "United States",   "California",    "Davis", "coordinates obtained from Label",           21.08,        -128.75,                       15L, "http://creativecommons.org/publicdomain/zero/1.0/", "The California Academy of Sciences - AntWeb.org",            NA, "urn:uuid:d6ff3ddb-4695-4aaf-ab89-c251ab2fc7e6", "https://scan-bugs.org:443/portal/collections/individual/index.php?occid=13775123",
  "fake SCAN28", "fakeDataset",    13775124L,                      "CAS",        "ANTWEB",        "UCDC, Davis, CA, USA", "PreservedSpecimen", "CAS:ANTWEB:casent0106100-d11", "casent0106100-d11",                   NA, "Animalia", "Arthropoda", "Insecta", "Hymenoptera", "apidae", "apis mellifera",  235783L,          "Linnaeus, 1758", "Apis",      "mellifera",        "Yadad",  "21/06/92", 1992L,     6L,  21L, "", "United States",   "California",    "Davis", "coordinates obtained from Label",           50.75400,        -12.73,                       15L, "http://creativecommons.org/publicdomain/zero/1.0/", "The California Academy of Sciences - AntWeb.org",            NA, "urn:uuid:4504a6e6-ab1c-4de7-ac43-a3588a63ae90", "https://scan-bugs.org:443/portal/collections/individual/index.php?occid=13775124",
  "fake SCAN29", "fakeDataset",           NA,                      "CAS",        "ANTWEB",        "UCDC, Davis, CA, USA", "PreservedSpecimen",                             NA,                  NA,                   NA, "Animalia", "Arthropoda", "Insecta", "Hymenoptera", "apidae", "apis mellifera",  235783L,          "Linnaeus, 1758", "Apis",      "mellifera",        "Yadad",  "2/06/25", 2025L,     6L,  2L, "", "United States",   "California",    "Davis", "coordinates obtained from Label",           13.2400,        -150.7514,                       15L, "http://creativecommons.org/publicdomain/zero/1.0/", "The California Academy of Sciences - AntWeb.org",            NA,                                              NA,                                                                                 NA,
  "fake SCAN30", "fakeDataset_butCooler",           NA,            "CAS",        "ANTWEB",        "UCDC, Davis, CA, USA", "PreservedSpecimen",                             NA,                  NA,                   NA, "Animalia", "Arthropoda", "Insecta", "Hymenoptera", "apidae", "apis mellifera",  235783L,          "Linnaeus, 1758", "Apis",      "mellifera",        "Yadad",  "8/01/00", 2000L,     1L,  8L, "", "United States",   "California",    "Davis", "coordinates obtained from Label",           09.712400,        -11.753215,                       15L, "http://creativecommons.org/publicdomain/zero/1.0/", "The California Academy of Sciences - AntWeb.org",            NA,                                              NA,                                                                                 NA,
  "fake SCAN31", "fakeDataset",    13775124L,                     "CAS",        "ANTWEB",        "UCDC, Davis, CA, USA", "PreservedSpecimen", "CAS:ANTWEB:casent0106100-d11", "casent0106100-d11",                   NA, "Animalia", "Arthropoda", "Insecta", "Hymenoptera", "apidae", "apis mellifera",  235783L,          "Linnaeus, 1758", "Apis",      "mellifera",        "P.S. Ward",  "31/12/76", 1876L,     12L,  31L, "", "United States",   "California",    "Davis", "coordinates obtained from Label",           38.5423,        -21.752369,                       15L, "http://creativecommons.org/publicdomain/zero/1.0/", "The California Academy of Sciences - AntWeb.org",            NA, "urn:uuid:4504a6e6-ab1c-4de7-ac43-a3588a63ae90", "https://scan-bugs.org:443/portal/collections/individual/index.php?occid=13775124",
    # Too few decimals
  "fake SCAN32", "newDataset",    13775122L,                      "CAS",        "ANTWEB",        "UCDC, Davis, CA, USA", "PreservedSpecimen",     "CAS:ANTWEB:casent0106100",     "casent0106100",                   NA, "Animalia", "Arthropoda", "Insecta", "Hymenoptera", "apidae", "apis mellifera",  235783L,          "Linnaeus, 1758", "Apis",      "mellifera",        "E Fischer",  "6/28/95", 2095L,     6L,  28L, "28 Jun 2095/29 Jun 2095", "United States",   "OuterSpace",    "Davis", "coordinates obtained from Label",           38.54,        -121.75,                       15L, "http://creativecommons.org/publicdomain/zero/1.0/", "The California Academy of Sciences - AntWeb.org",            NA, "urn:uuid:46a46727-6535-4e70-88e7-a42c98f806ed", "https://scan-bugs.org:443/portal/collections/individual/index.php?occid=13775122",
  "fake SCAN33", "newDataset",    13775123L,                      "CAS",        "ANTWEB",        "UCDC, Davis, CA, USA", "PreservedSpecimen", "CAS:ANTWEB:casent0106100-d01", "casent0106100-d01",                   NA, "Animalia", "Arthropoda", "Insecta", "Hymenoptera", "apidae", "apis mellifera",  235783L,          "Linnaeus, 1758", "Apis",      "mellifera",        "E Fischer",  "6/28/95", 2095L,     6L,  28L, "28 Jun 2095/29 Jun 2095", "United States",   "OuterSpace",    "Davis", "coordinates obtained from Label",           38.55,        -121.75,                       15L, "http://creativecommons.org/publicdomain/zero/1.0/", "The California Academy of Sciences - AntWeb.org",            NA, "urn:uuid:d6ff3ddb-4695-4aaf-ab89-c251ab2fc7e6", "https://scan-bugs.org:443/portal/collections/individual/index.php?occid=13775123",
  "fake SCAN34", "newDataset",    13775124L,                      "CAS",        "ANTWEB",        "UCDC, Davis, CA, USA", "PreservedSpecimen", "CAS:ANTWEB:casent0106100-d11", "casent0106100-d11",                   NA, "Animalia", "Arthropoda", "Insecta", "Hymenoptera", "apidae", "apis mellifera",  235783L,          "Linnaeus, 1758", "Apis",      "mellifera",        "E Fischer",  "6/28/95", 2095L,     6L,  28L, "28 Jun 2095/29 Jun 2095", "United States",   "OuterSpace",    "Davis", "coordinates obtained from Label",           38.56,        -121.75,                       15L, "http://creativecommons.org/publicdomain/zero/1.0/", "The California Academy of Sciences - AntWeb.org",            NA, "urn:uuid:4504a6e6-ab1c-4de7-ac43-a3588a63ae90", "https://scan-bugs.org:443/portal/collections/individual/index.php?occid=13775124",
  "fake SCAN35", "newDataset",           NA,                      "CAS",        "ANTWEB",        "UCDC, Davis, CA, USA", "PreservedSpecimen",                             NA,                  NA,                   NA, "Animalia", "Arthropoda", "Insecta", "Hymenoptera", "apidae", "apis mellifera",  235783L,          "Linnaeus, 1758", "Apis",      "mellifera",        "E Fischer",  "6/28/95", 2095L,     6L,  28L, "28 Jun 2095/29 Jun 2095", "United States",   "OuterSpace",    "Davis", "coordinates obtained from Label",           38.57,        -121.75,                       15L, "http://creativecommons.org/publicdomain/zero/1.0/", "The California Academy of Sciences - AntWeb.org",            NA,                                              NA,                                                                                 NA
  )


# run the function!
testOut <- BeeBDC::diagonAlley(data = data, 
                              minRepeats = 4,
                              groupingColumns = c("eventDate", "recordedBy", "datasetName"),
                              ndec = 3,
                              mc.cores = 1) %>%
  dplyr::select(c(database_id, "eventDate", "recordedBy", "datasetName", decimalLatitude, 
                  decimalLongitude, .sequential)) %>% invisible()


# Test the number of expected TRUE and FALSE values in the .sequential column
resultsT <- length(testOut$.sequential[testOut$.sequential == TRUE])
resultsF <- length(testOut$.sequential[testOut$.sequential == FALSE])

testthat::test_that("diagonAlley column .sequential results TRUE", {
  testthat::expect_equal(resultsT, 18)
})

testthat::test_that("diagonAlley column .sequential results FALSE", {
  testthat::expect_equal(resultsF, 17)
})


# Check sequence of TRUE and FALSE values in the .sequential column
correct <- c(FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, TRUE, TRUE, TRUE, TRUE, TRUE, TRUE, 
             TRUE, TRUE, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, TRUE, TRUE, 
             TRUE, TRUE, TRUE, TRUE, TRUE, TRUE, TRUE)

testthat::test_that("diagonAlley column .sequential results correct series", {
  testthat::expect_equal(correct, testOut$.sequential)
})

  # This function was written by James Dorey on the 20th of May 2022 in ored to define the columns
    # used during occurrence record cleaning
  # For help you may email James at jbdorey[at]me.com

#' Sets up column names and types
#' 
#' This function uses [readr::cols_only()] to assign a column name and the type of data 
#' (e.g., [readr::col_character()], 
#' and [readr::col_integer()]). To see the default columns simply run [BeeBDC::ColTypeR()]. 
#' This is intended for use with [readr::read_csv()]. Columns that are not present will NOT be included 
#' in the resulting tibble unless they are specified using [...].
#'
#' @param ... Additional arguments. These can be specified in addition to the ones default to the 
#' function.  For example:
#' * newCharacterColumn = [readr::col_character()],
#' * newNumericColumn = [readr::col_integer()],
#' * newLogicalColumn = [readr::col_logical()]
#' @param standardFormat Character. Some taxa may have a standard format for data. Presently,
#' Only bees have had this encoded here as "bee". Default = NULL.
#' 
#' @importFrom dplyr %>%
#'
#' @return Returns an object of class col_spec. 
#' See [readr::as.col_spec()] for additional context and explication.
#' @export
#' 
#' @references For using the bee standard —  Clos, B. D., Seltmann, K. C., 
#' Turley, N. E., Maffei, C., Tucker, E. M., Lane, I. G., Levenson, H. K., & Woodard, H. S. 
#' (2025). Improving the standardization of wild bee occurrence data: towards a formal 
#' wild bee data standard. Authorea. doi: https://doi.org/10.22541/au.173862402.22787949/v2 
#'
#' @examples 
#'   # You can simply return the below for default values
#'   library(dplyr)
#' BeeBDC::ColTypeR() 
#' 
#'   # To add new columns you can write
#' ColTypeR(newCharacterColumn = readr::col_character(), 
#'          newNumericColumn = readr::col_integer(), 
#'          newLogicalColumn = readr::col_logical()) 
#' 
#' # Try reading in one of the test datasets as an example:
#' beesFlagged %>% dplyr::as_tibble(col_types = BeeBDC::ColTypeR())
#'   # OR
#' beesRaw %>% dplyr::as_tibble(col_types = BeeBDC::ColTypeR())
#' 
#'   # OR, using the bee standard format from:
#'   
#' beesRaw %>% dplyr::as_tibble(col_types = BeeBDC::ColTypeR(standardFormat = "bee"))
#' 
#' 
ColTypeR <- function(...,
                     standardFormat = NULL){
    # If no standardFormat is assigned, assign it to "NULL"
  if(is.null(standardFormat)){
    standardFormat <- "NULL"
  }
  
  if(standardFormat == "NULL"){
  ColTypes <- readr::cols_only(
    # Character Strings
    # CHR - taxonomy
    database_id  = readr::col_character(), scientificName = readr::col_character(), 
    kingdom = readr::col_character(),phylum = readr::col_character(),class = readr::col_character(),
    order = readr::col_character(),
    family = readr::col_character(), subfamily = readr::col_character(), 
    tribe = readr::col_character(), subtribe = readr::col_character(),
    genus = readr::col_character(), 
    subgenus = readr::col_character(), subspecies = readr::col_character(), species = readr::col_character(), 
    specificEpithet = readr::col_character(), infraspecificEpithet = readr::col_character(), 
    acceptedNameUsage = readr::col_character(), taxonRank = readr::col_character(),
    scientificNameAuthorship = readr::col_character(), 
    identificationQualifier = readr::col_character(), higherClassification = readr::col_character(), 
    identificationReferences = readr::col_character(), typeStatus = readr::col_character(), 
    previousIdentifications = readr::col_character(), verbatimIdentification = readr::col_character(), 
    identifiedBy = readr::col_character(), dateIdentified = readr::col_character(),
    # DBL - Locality info
    decimalLatitude = readr::col_double(), decimalLongitude = readr::col_double(),
    verbatimLatitude = readr::col_character(), verbatimLongitude = readr::col_character(),
    verbatimElevation = readr::col_character(),
    # CHR/Factor - Locality info
    stateProvince = readr::col_character(), country = readr::col_character(), continent = readr::col_factor(), 
    locality = readr::col_character(), island = readr::col_character(),
    county = readr::col_character(), municipality = readr::col_character(),
    # CHR/Factor - Country codes
    countryCode = readr::col_factor(), level0Gid = readr::col_factor(), level0Name = readr::col_factor(), 
    level1Gid = readr::col_factor(), level1Name = readr::col_factor(), license = readr::col_factor(), 
    issue = readr::col_character(), 
    # Date/Time - Collection time
    eventDate = readr::col_character(), 
    eventTime = readr::col_character(), 
    startDayOfYear = readr::col_integer(),
    endDayOfYear = readr::col_integer(),
    # Int - Collection time
    day = readr::col_integer(), month = readr::col_integer(), year = readr::col_integer(),
    # Factor - Collection info
    basisOfRecord = readr::col_factor(), type = readr::col_factor(), occurrenceStatus = readr::col_factor(), 
    # CHR - Collection info
    recordNumber = readr::col_character(), recordedBy = readr::col_character(), eventID = readr::col_character(), 
    Location = readr::col_character(), samplingProtocol = readr::col_character(), 
    # Int — Collection info
    samplingEffort = readr::col_character(), fieldNumber = readr::col_character(), 
    individualCount = readr::col_double(), organismQuantity = readr::col_double(), 
    # mixed - Information uncertainty
    coordinatePrecision = readr::col_double(), coordinateUncertaintyInMeters = readr::col_double(), 
    spatiallyValid = readr::col_logical(),
    # CHR - Database information
    catalogNumber = readr::col_character(), gbifID = readr::col_character(), datasetID = readr::col_character(),
    institutionCode = readr::col_character(), datasetName = readr::col_character(), 
    otherCatalogNumbers = readr::col_character(), occurrenceID = readr::col_character(), 
    taxonKey = readr::col_character(), coreid = readr::col_character(), 
    recordId = readr::col_character(), collectionID = readr::col_character(),
    associatedSequences = readr::col_character(), institutionID = readr::col_character(),
    # CHR - Verbatim information
    verbatimScientificName = readr::col_character(), verbatimEventDate = readr::col_character(),
    # CHR/Factor - Aux info
    associatedTaxa = readr::col_character(), associatedOrganisms = readr::col_character(), 
    fieldNotes = readr::col_character(), sex = readr::col_character(),
    # CHR - Rights info
    rights = readr::col_character(), rightsHolder = readr::col_character(), accessRights = readr::col_character(), 
    dctermsLicense = readr::col_character(), dctermsType = readr::col_character(), 
    dctermsAccessRights = readr::col_character(), associatedReferences = readr::col_character(), 
    bibliographicCitation = readr::col_character(), dctermsBibliographicCitation = readr::col_character(), 
    references = readr::col_character(),
    # Record notes
    # CHR
    flags = readr::col_character(), informationWithheld = readr::col_character(), 
    isDuplicateOf = readr::col_character(),
    # Logical
    hasCoordinate = readr::col_logical(), hasGeospatialIssues = readr::col_logical(), 
    # Factor
    assertions = readr::col_factor(),
    # mix - ALA columns
    occurrenceYear = readr::col_datetime(), id = readr::col_character(), duplicateStatus = readr::col_factor(), 
    associatedOccurrences = readr::col_character(), 
    # CHR - SCAN column
    locationRemarks = readr::col_character(),
    # CHR - dataset origin column
    dataSource = readr::col_character(),
      # bdc columns
     dataBase_scientificName = readr::col_character(), .rou = readr::col_logical(),                         
    .val = readr::col_logical(), .equ = readr::col_logical(), .zer = readr::col_logical(), .cap = readr::col_logical(),                         
    .cen = readr::col_logical(), .sea = readr::col_logical(), .otl = readr::col_logical(), .gbf = readr::col_logical(),                         
    .inst = readr::col_logical(), .dpl = readr::col_logical(), .summary = readr::col_logical(),
    names_clean = readr::col_character(), verbatim_scientificName = readr::col_character(), 
    .uncer_terms = readr::col_logical(), .eventDate_empty = readr::col_logical(), 
    .year_outOfRange = readr::col_logical(),
    .duplicates = readr::col_logical(), .lonFlag = readr::col_logical(), .latFlag = readr::col_logical(), 
    .gridSummary = readr::col_logical(), .basisOfRecords_notStandard = readr::col_logical(),
    .scientificName_empty = readr::col_logical(), .coordinates_empty = readr::col_logical(),
    .coordinates_outOfRange = readr::col_logical(), coordinates_transposed = readr::col_logical(),
    country_suggested = readr::col_character(),  .countryOutlier = readr::col_logical(),
    countryMatch = readr::col_character(), .expertOutlier = readr::col_logical(),
      # jbd flags
    .occurrenceAbsent = readr::col_logical(), .coordinates_country_inconsistent = readr::col_logical(), 
    .unLicensed = readr::col_logical(), .invalidName = readr::col_logical(),
    .sequential = readr::col_logical(), idContinuity = readr::col_logical(), 
    .uncertaintyThreshold = readr::col_logical(),
    .GBIFflags = readr::col_logical(), 
      # Paige columns
    finalLatitude = readr::col_double(), finalLongitude = readr::col_double(), 
    
    Source = readr::col_character(),
    # Dynamic dots for extra columns specified by the user
    ...
  ) # END ColTypes
} # END standarFormat
  
  #### 2.0 Standard formats ####
    ##### 2.1 Bees ####
  if(tolower(standardFormat) == "bee"){
    ColTypes <- readr::cols_only(
    # Core terms
      # Record-level
    institutionCode = readr::col_character(), basisOfRecord = readr::col_factor(), informationWithheld = readr::col_character(), 
      # Occurrence
    occurrenceID = readr::col_character(), catalogNumber = readr::col_character(), recordedBy = readr::col_character(),
    individualCount = readr::col_double(), occurrenceStatus = readr::col_factor(), 
      # Event
    eventDate = readr::col_character(), 
    eventTime = readr::col_character(),  year = readr::col_integer(), month = readr::col_integer(),
    day = readr::col_integer(),samplingProtocol = readr::col_character(), sampleSizeValue = readr::col_double(),
    sampleSizeUnit = readr::col_character(), samplingEffort = readr::col_character(),
      # Location
    decimalLatitude = readr::col_double(), decimalLongitude = readr::col_double(),
    coordinateUncertaintyInMeters = readr::col_double(), 
      # Identification
    identifiedBy = readr::col_character(), 
      # Taxon
    scientificName = readr::col_character(), genus = readr::col_character(), specificEpithet = readr::col_character(),
    
    # Recommended terms
      # Record-level
    license = readr::col_factor(), institutionID = readr::col_character(), collectionID = readr::col_character(), 
    collectionCode = readr::col_character(), 
      # Occurrence
    recordedByID = readr::col_character(), sex = readr::col_character(),  associatedTaxa = readr::col_character(),
      # Event
    eventID = readr::col_character(), eventRemarks = readr::col_character(), 
      # MaterialEntity
    disposition = readr::col_factor(), 
      # Location
    locationID = readr::col_character(), stateProvince = readr::col_character(), 
    geodeticDatum = readr::col_character(), 
      # Identification 
    verbatimIdentification = readr::col_character(), identificationQualifier = readr::col_character(),
    identifiedByID = readr::col_character(), dateIdentified = readr::col_character(),
    identificationReferences = readr::col_character(), taxonID  = readr::col_character(),
    nameAccordingTo = readr::col_character(), family = readr::col_character(),
    infraspecificEpithet = readr::col_character(), scientificNameAuthorship = readr::col_character(), 
    
    # Optional terms
      # Record-level
    rightsHolder = readr::col_character(), dynamicProperties = readr::col_character(), 
      # Occurrence
    lifeStage = readr::col_factor(), caste = readr::col_factor(), behavior = readr::col_factor(),
    vitality = readr::col_factor(), associatedMedia = readr::col_character(), 
    associatedOccurrences = readr::col_character(), occurrenceRemarks = readr::col_character(),
      # MaterialEntity
    preparations = readr::col_character(), associatedSequences = readr::col_character(),
    materialEntityRemarks = readr::col_character(),
      # Event
    fieldNumber = readr::col_character(), habitat = readr::col_character(),
      # Location
    country = readr::col_character(), county = readr::col_character(),
    locality = readr::col_character(), verbatimElevation = readr::col_character(),
    coordinatePrecision = readr::col_double(), georeferencedBy = readr::col_character(),
    georeferenceRemarks = readr::col_character(),
      # Identification
    typeStatus = readr::col_character(), identificationRemarks = readr::col_character(),
    namePublishedIn = readr::col_character(), tribe = readr::col_character(),
    subgenus = readr::col_character(), taxonRank = readr::col_character(),
    vernacularName = readr::col_character(),
    # Dynamic dots for extra columns specified by the user
    ...
    ) # END ColTypes 
  } 
  
  return(ColTypes)
} # END ColTypeR

requireNamespace("testthat")
testthat::test_that("ColTypeR class is correct", {
  testthat::expect_type(BeeBDC::ColTypeR(), "list")
})

testthat::test_that("ColTypeR class is correct", {
  testthat::expect_equal(attributes(BeeBDC::ColTypeR())$class, "col_spec")
})

# This function was written by James B Dorey on the 29th of September 2022
# Its purpose is to visualise duplicate occurrence data by using a chord diagram
# Please contact jbdorey[at]me.com for help

#' Build a chord diagram of duplicate occurrence links
#' 
#' This function outputs a figure which shows the relative size and direction of occurrence points 
#' duplicated between data providers, such as, SCAN, GBIF, ALA, etc. This function requires the 
#' outputs generated by [BeeBDC::dupeSummary()].
#' 
#'
#' @param dupeData A tibble or data frame. The duplicate file produced by [BeeBDC::dupeSummary()].
#' @param outPath Character. The path to a directory (folder) in which the output should be saved.
#' @param fileName Character. The name of the output file, ending in '.pdf'.
#' @param width Numeric. The width of the figure to save (in inches). Default = 7.
#' @param height Numeric. The height of the figure to save (in inches). Default = 6.
#' @param bg The plot's background colour. Default = "white".
#' @param smallGrpThreshold Numeric. The upper threshold of sub-dataSources to be listed as "other".
#' Default = 3.
#' @param title A character string. The figure title. Default = "Duplicated record sources".
#' @param palettes A vector of the palettes to be used. One palette for each major dataSource and "other"
#' using the `paletteer` package. Default = c("cartography::blue.pal", "cartography::green.pal", 
#' "cartography::sand.pal", "cartography::orange.pal", "cartography::red.pal",
#' "cartography::purple.pal", "cartography::brown.pal")
#' @param canvas.ylim Canvas limits from [circlize::circos.par()]. Default = c(-1.0,1.0).
#' @param canvas.xlim Canvas limits from [circlize::circos.par()]. Default = c(-0.6, 0.25).
#' @param text.col A character string. Text colour
#' @param legendX The x position of the legends, as measured in current viewport.
#'  Passed to ComplexHeatmap::draw(). Default = grid::unit(6, "mm").
#' @param legendY The y position of the legends, as measured in current viewport.
#'  Passed to ComplexHeatmap::draw(). Default = grid::unit(18, "mm").
#' @param legendJustify A character vector declaring the justification of the legends. 
#' Passed to ComplexHeatmap::draw(). Default = c("left", "bottom").
#' @param niceFacing TRUE/FALSE. The niceFacing option automatically adjusts the text facing 
#' according to their positions in the circle. Passed to [circlize::highlight.sector()].
#' @param self.link 1 or 2 (numeric). Passed to [circlize::chordDiagram()]:	
#' if there is a self link in one sector, 1 means the link will be degenerated as a 'mountain' and the width corresponds to the value for this connection. 2 means the width of the starting root and the ending root all have the width that corresponds to the value for the connection.
#'
#'
#' @return Saves a figure to the provided file path.
#' 
#' @importFrom circlize circos.clear circos.par chordDiagram mm_h circos.trackPlotRegion get.cell.meta.data highlight.sector circos.clear
#' @importFrom stringr str_replace str_c
#' @importFrom dplyr bind_cols full_join mutate select group_by if_else arrange n filter %>%
#' @importFrom paletteer paletteer_dynamic
#' @importFrom grid unit
#' 
#' 
#' 
#' @export
#'
#' @examples
#' \dontrun{
#'   # Create a basic example dataset of duplicates to visualise
#' basicData <- dplyr::tribble(
#'                             ~dataSource,    ~dataSource_keep,
#'                       "GBIF_Halictidae",         "USGS_data",
#'                       "GBIF_Halictidae",         "USGS_data",
#'                       "GBIF_Halictidae",         "USGS_data",
#'                       "GBIF_Halictidae",         "USGS_data",
#'                       "GBIF_Halictidae",         "USGS_data",
#'                       "GBIF_Halictidae",         "USGS_data",
#'                       "SCAN_Halictidae",   "GBIF_Halictidae",
#'                    "iDigBio_halictidae",   "GBIF_Halictidae",
#'                    "iDigBio_halictidae",   "SCAN_Halictidae",
#'                    "iDigBio_halictidae",   "SCAN_Halictidae",
#'                       "SCAN_Halictidae",   "GBIF_Halictidae",
#'                        "iDigBio_apidae",       "SCAN_Apidae",
#'                           "SCAN_Apidae",    "Ecd_Anthophila",
#'                        "iDigBio_apidae",    "Ecd_Anthophila",
#'                           "SCAN_Apidae",    "Ecd_Anthophila",
#'                        "iDigBio_apidae",    "Ecd_Anthophila",
#'                     "SCAN_Megachilidae", "SCAN_Megachilidae",
#'                       "CAES_Anthophila",   "CAES_Anthophila",
#'                       "CAES_Anthophila",   "CAES_Anthophila"
#'  )
#'
#' 
#'  chordDiagramR(
# # The duplicate data from the dupeSummary function output  
#' dupeData = basicData,
#' outPath = tempdir(),
#' fileName = "ChordDiagram.pdf",
#' # These can be modified to help fit the final pdf that's exported.
#' width = 9,
#' height = 7.5,
#' bg = "white",
#' # How few distinct dataSources should a group have to be listed as "other"
#' smallGrpThreshold = 3,
#' title = "Duplicated record sources",
#' # The default list of colour palettes to choose from using the paleteer package
#' palettes = c("cartography::blue.pal", "cartography::green.pal", 
#'              "cartography::sand.pal", "cartography::orange.pal", "cartography::red.pal",
#'              "cartography::purple.pal", "cartography::brown.pal"),
#' canvas.ylim = c(-1.0,1.0), 
#' canvas.xlim = c(-0.6, 0.25),
#' text.col = "black",
#' legendX = grid::unit(6, "mm"),
#' legendY = grid::unit(18, "mm"),
#' legendJustify = c("left", "bottom"),
#' niceFacing = TRUE)}
chordDiagramR <- function(
    # The duplicate data from the dupeSummary function output  
  dupeData = NULL,
  outPath = NULL,
  fileName = NULL,
  width = 7,
  height = 6,
  bg = "white",
  # How few distinct dataSources should a group have to be listed as "other"
  smallGrpThreshold = 3,
  title = "Duplicated record sources",
  # The default list of colour palettes to choose from 
  palettes = c("cartography::blue.pal", "cartography::green.pal", 
               "cartography::sand.pal", "cartography::orange.pal", "cartography::red.pal",
               "cartography::purple.pal", "cartography::brown.pal"),
  canvas.ylim = c(-1.0,1.0), 
  canvas.xlim = c(-0.6, 0.25),
  text.col = "black",
  legendX = grid::unit(6, "mm"),
  legendY = grid::unit(18, "mm"),
  legendJustify = c("left", "bottom"),
  niceFacing = TRUE,
  self.link = 2){
  
  # locally bind variabls to the function
  Frequency <- Frequency_dupe <- sourceName <- . <- sourceCategories <- groupCount <- cur_group_id <-
    groupNumber <- groupPalette <- groupColours <- par <- NULL
  error_func_BCM <- CHtest <- error_func_CH <- input <- instructions <- NULL
  
    requireNamespace("circlize")
  requireNamespace("dplyr")
  requireNamespace("paletteer")
  requireNamespace("grid")
  
  #### 0.0 Prep ####
  ##### 0.1 errors ####
  ###### a. FATAL errors ####
  if(is.null(dupeData)){
    stop(" - Please provide an argument for dupeData. I'm a program not a magician.")
  }
  if(is.null(outPath)){
    stop(" - Please provide an argument for outPath Seems reckless to let me just guess.")
  }
  if(is.null(fileName)){
    stop(" - Please provide an argument for fileName Seems reckless to let me just guess.")
  }
  if(nrow(dupeData) == 0){
    stop(" - There are no duplicates in the dupeData object. Stopping process.")
  }
  
    ##### 0.2 maintain par ####
  # Make sure to maintain prior par on exit from the function
  oldpar <- par(no.readonly = TRUE) 
  on.exit(oldpar)
  
  ##### 0.3 BcM + ComplexHeatmap ####
    ###### a. test ####
  # Check if BiocManager is installed
  # TRUE if BiocManager is found
  suppressWarnings(
    BcMtest <- system.file(package='BiocManager') %>% 
      stringr::str_count() > 0 
  )
    # Check if ComplexHeatmap is installed
      # TRUE if ComplexHeatmap is found
  suppressWarnings(
  CHtest <- system.file(package='ComplexHeatmap') %>% 
    stringr::str_count() > 0 
  )

    ###### b. BiocManager ####
  if(CHtest == FALSE){
    if(BcMtest == FALSE){
    # Set up instructions for download on fail
    instructions <- paste(" Please try installing the package for yourself", 
                          "using the following command: \n", 
                          " install.packages(\"BiocManager\")")
    # Set up fail function for tryCatch
    error_func_BCM <- function(e){
      stop(paste("Failed to install the BiocManager package.\n", 
                 instructions))
    }
    # Begin interactive input
    input <- 1
    if (interactive()){
      input <- utils::menu(c("Yes", "No"), 
                           title = paste0("Install the BiocManager package? \n",
                            "NOTE: if you need to install BiocManager, you may need to restart R",
                            " before installing ComplexHeatmap."))
    }
    if(input == 1){
      # Check for BiocManager
      if( suppressWarnings(system.file(package='BiocManager')) %>% stringr::str_count() == 0){
        message("Installing the BiocManager package.")
        tryCatch(
          utils::install.packages("BiocManager"), 
          error = error_func_BCM, warning = error_func_BCM)
      }# END BiocManager check
    
    else{
      stop(writeLines(paste("The ComplexHeatmap package is necessary for BeeBDC::chordDiagramR.\n", 
                            instructions)))
    } # END else
    } # END input == 1
    }# END BcMtest == FALSE
    } # END CHtest == FALSE
  
  
  
    ###### c. ComplexHeatmap ####
  if(CHtest == FALSE){
      # Set up instructions for download on fail
    instructions <- paste(" Please try installing the package for yourself", 
                          "using the following command: \n",
                          "BiocManager::install(\"ComplexHeatmap\")")
      # Set up fail function for tryCatch
    error_func_CH <- function(e){
      stop(paste("Failed to install the ComplexHeatmap package.\n", 
                 instructions))
    }
      # Begin interactive input
    input <- 1
    if (interactive()){
      input <- utils::menu(c("Yes", "No"), 
                        title = paste0("Install the ComplexHeatmap package? \n"))
    }
    if(input == 1){
        # Start ComplexHeatmap install
      message("Installing the ComplexHeatmap package.")
      tryCatch(
        BiocManager::install("ComplexHeatmap"), 
        error = error_func_CH, warning = error_func_CH)
      } # END input == 1
      
    else{
      stop(writeLines(paste("The ComplexHeatmap package is necessary for BeeBDC::chordDiagramR.\n", 
                 instructions)))
    } # END else
  } # END CHtest == FALSE
  

    #### 1.0 Data prep ####
# Create a table to go into chord diagram
  suppressMessages(
chordData <- table(dplyr::bind_cols(dupeData$dataSource, dupeData$dataSource_keep)),
classes = "message")


  # Create tables of the counts of kept source and duplicate source
keptSource <- table(dupeData$dataSource) %>%
  as.data.frame() %>% dplyr::tibble() %>% 
  stats::setNames(c("sourceName", "Frequency")) 
dupeSource <- table(dupeData$dataSource_keep) %>%
  as.data.frame() %>% dplyr::tibble() %>% 
  stats::setNames(c("sourceName", "Frequency_dupe")) 
  # Merge the sources and get their sum (for a total frequency count to order by)
colourTable <- dplyr::full_join(keptSource, dupeSource, by = "sourceName") %>%
  dplyr::mutate(Frequency = (Frequency + Frequency_dupe)) %>%
    # Drop the Frequency_dupe column
  dplyr::select(!Frequency_dupe) %>%
  # Get broad source (before first underscore)
  dplyr::mutate( sourceCategories = (sourceName %>%
                                       stringr::str_replace(
                                         string = .,
                                         pattern = "_.*",
                                         replacement = ""
                                       ))) %>%
  dplyr::group_by(sourceCategories) %>%
  dplyr::mutate(  # Count group number
    groupCount = dplyr::n(),
    # Combine small groups (< smallGrpThreshold)
    sourceCategories = dplyr::if_else(
      groupCount < smallGrpThreshold, "Other", sourceCategories)) %>%
  dplyr::arrange(sourceName, .by_group = TRUE) %>%
  # Re-group
  dplyr::group_by(sourceCategories) %>%
  dplyr::mutate(groupNumber = dplyr::cur_group_id(),
                # Re-count
                groupCount = dplyr::n(),
                groupPalette = palettes[groupNumber]) %>%
  # assign colours
  dplyr::mutate(groupColours = 
                  paletteer::paletteer_dynamic(
                    palette = groupPalette[[1]],
                    n = groupCount[[1]]) %>% list(),
                colour = unlist(groupColours)[dplyr::row_number()])


  #### 2.0 Build plot ####
circlize::circos.clear()
  circlize::circos.par(canvas.ylim = canvas.ylim, canvas.xlim = canvas.xlim)
  
# Create the chord diagrame
circlize::chordDiagram(
  x = chordData,
  order = colourTable$sourceName,
  directional = 1,
  direction.type = c("arrows"),
  link.arr.type = "big.arrow",
  reduce = 0,
    # self links fold directly back onto themselves instead of going to far side
  self.link = self.link,
  grid.col = colourTable$colour,
  keep.diagonal = TRUE,
  # name, grid, axis
  annotationTrack = c("grid"),
  preAllocateTracks = list(
    track.height = circlize::mm_h(4),
    track.margin = c(circlize::mm_h(1), 0)),
  scale = FALSE
)
circlize::circos.trackPlotRegion(track.index = 1, panel.fun = function(x, y) {
  xlim = circlize::get.cell.meta.data("xlim")
  ylim = circlize::get.cell.meta.data("ylim")
  sector.name = circlize::get.cell.meta.data("sector.index")
}, bg.border = NA)
# Highlight inputs
for(i in 1:length(unique(colourTable$sourceCategories))){
  loopCat <- colourTable %>%
    dplyr::filter(sourceCategories == unique(colourTable$sourceCategories)[i])
  circlize::highlight.sector(stringr::str_c(loopCat$sourceName), 
                             track.index = 1, col = loopCat$colour[[1]], 
                             text = unique(loopCat$sourceCategories), cex = 0.8, 
                             text.col = text.col, niceFacing = niceFacing)
}
legendList <- c()
# Make legends by creating a list of legends for each sourceCategory
for(i in 1:length(unique(colourTable$sourceCategories))){
  loopCat <- colourTable %>%
    dplyr::filter(sourceCategories == unique(colourTable$sourceCategories)[i])
  legendList[[i]] <- ComplexHeatmap::Legend(labels = stringr::str_c(loopCat$sourceName), 
                                            title = unique(stringr::str_c(loopCat$sourceCategories)), 
                                            legend_gp = grid::gpar(fill = c(loopCat$colour)))
} # END legend loop

lgd_list <- ComplexHeatmap::packLegend(list = legendList)
ComplexHeatmap::draw(lgd_list, x = legendX, 
                     y = legendY, just = legendJustify)

circlize::circos.clear()

title(title)

grDevices::dev.copy2pdf(file = paste(outPath, "/", fileName, sep = ""),
                        height = height, width = width, bg = bg)

#dev.off()
} # END function

#### Aux. functions ####

##### a. dataSaver ####
#' Simple function to save occurrence AND EML data as a list
#' 
#' Used at the end of 1.x in the example workflow in order to save the occurrence dataset and its associated eml metadata.
#'
#' @param path Character. The main file path to look for data in.
#' @param save_type Character. The file format in which to save occurrence and EML data. 
#' Either "R_file" or "CSV_file"
#' @param occurrences The occurrences to save as a data frame or tibble.
#' @param eml_files A list of the EML files.
#' @param file_prefix Character. A prefix for the resulting output file.
#'
#' @return This function saves both occurrence and EML data as a list when save_type = "R_File" or
#' as individual csv files when save_type = "CSV_file".
#' @export
#' 
#' @importFrom dplyr %>%
#'
#' @examples
#' \dontrun{
#' dataSaver(path = tempdir(),# The main path to look for data in
#' save_type = "CSV_file", # "R_file" OR "CSV_file"
#' occurrences = Complete_data$Data_WebDL, # The existing datasheet
#' eml_files = Complete_data$eml_files, # The existing EML files
#' file_prefix = "Fin_") # The prefix for the file name
#' }
#' 
dataSaver <- function(path = NULL,
                       save_type = NULL,
                       occurrences = NULL,
                       eml_files = NULL,
                       file_prefix = NULL){
  # locally bind variables to the function
  . <- countComplete <- NULL
  
  # Check if the save type is present AND valid
  if(exists("save_type") == FALSE){ # If there are no data matching the name...
    stop(" - Heck! You must choose a save_type for the data output. Either .rds or .csv files")
  } # END Missing save_type type
  if(sum(stringr::str_count(pattern = c("R_file","CSV_file"), string = save_type)) == 0 ){
    stop(" - Heck! Please choose a valid save_type for the data output. Either .rds (R_file) or .csv files (csv_files)")
  } # END invalid save_type
  # If there is no file prefix... name it "BeeData_"
  if(is.null(file_prefix) == TRUE){
    file_prefix <- "BeeData_"
  }else{
    file_prefix <- paste(file_prefix,"BeeData_", sep = "")
  }
  # Notify user if an out_file has been created and make one, if required.
  outPath <- outFile_maker(path = path)
  # Extract the attribute data
  occurrences_attributes <- attributes(occurrences)
  
  # Find the empty columns
  colTest <- occurrences %>% 
    dplyr::summarise(dplyr::across(tidyselect::everything(), ~ sum(complete.cases(.)))) 
  colKeeps <- dplyr::tibble(column = colnames(colTest),
                 countComplete = t(colTest)[,1]) %>%
    dplyr::filter(countComplete > 0)
  # Cols to remove:
  colRemoves <- dplyr::tibble(column = colnames(colTest),
         countComplete = t(colTest)[,1]) %>%
    dplyr::filter(countComplete == 0)
    # Discard empty columns
  occurrences <- occurrences %>% 
    dplyr::select(tidyselect::all_of(colKeeps$column))
  message(paste0(
    " - We have removed empty columns. This is standard, but as an FYI, these columns are: ",
    paste(colRemoves$column, collapse = ", ")
  ))
  

  #### R save ####
  # Save R data
  if(save_type == "R_file"){
    writeLines( paste(" - Writing occurrence, attribute, and EML data file in .rds format...", "\n",
                      "Number of records: ", format(nrow(occurrences), big.mark=",",scientific=FALSE), "\n",
                      "Number of attribute sources: ", format(nrow(occurrences_attributes$dataSource), 
                                                              big.mark=",",scientific=FALSE), "\n",
                      "The ", length(names(eml_files)), " eml sources are ", 
                      paste(names(eml_files), collapse = ", "), "\n",
                      "Writing to file called ", paste(file_prefix, Sys.Date(), ".rds", sep = ""),
                      " at location ", outPath,"...",
                      sep = ""))
    # Save all of these data into a .rds format
    list(occurrences, eml_files) %>%
      saveRDS(., file = paste(outPath, "/", file_prefix, Sys.Date(), ".rds", sep = ""))
  }
  
  #### csv save ####
  # Save csv files
  if(save_type == "CSV_file"){
    ##### Occ. file ####
    writeLines( paste(" - Writing occurrence data file in csv format...", "\n",
                      "Number of rows (records): ", format(nrow(occurrences), big.mark=",",scientific=FALSE), "\n",
                      "Writing to file called ", paste(file_prefix, "combined_", Sys.Date(), ".csv", sep = ""),
                      " at location ", outPath,"...",
                      sep = "")) 
    # Write the occurence file
    readr::write_excel_csv(occurrences, paste(outPath, "/", file_prefix, "combined_", Sys.Date(), ".csv", sep = ""))
    
    #### Attr. file ####
    # Notfiy user that attribute data are being written
    occurrences_attributes <- attributes(occurrences)
    writeLines( paste(" - Writing attribute data file in csv format...", "\n",
                      "Number of rows (sources): ", format(nrow(occurrences_attributes$dataSource), 
                                                           big.mark=",",scientific=FALSE), "\n",
                      "Written to file called ", paste(file_prefix, "attributes_", Sys.Date(), ".csv", 
                                                       sep = ""),
                      " at location ", outPath,
                      sep = "")) 
    
    #### DataSource file #####
    # Write the citations file
    # Write the occurrence file
    readr::write_excel_csv(occurrences_attributes$dataSource, paste(outPath, "/" ,file_prefix, "attributes_", 
                                                      Sys.Date(), ".csv", sep = ""))
    
    #### All attributes ####
    # Update the names of each list element to reflect their source and taxonomic coverage
    names(occurrences_attributes$dataSource$citations) <- paste(occurrences_attributes$dataSource$dataSource, "_citations", sep = "")
    names(occurrences_attributes$dataSource$rights) <- paste(occurrences_attributes$dataSource$dataSource, "_rights", sep = "")
    names(occurrences_attributes$dataSource$abstract) <- paste(occurrences_attributes$dataSource$dataSource, "_abstract", sep = "")
    
    #### EML file ####
    if(!is.null(eml_files)){
    # Notify user that the .eml file is being written
    writeLines( paste(" - Writing eml file in xml format...", "\n",
                      "The ", length(names(eml_files)), " eml sources are ", 
                      paste(names(eml_files), collapse = ", "), "\n",
                      "Written to file called ", paste("eml_files", Sys.Date(),".xml", sep="" ),
                      " at location ", outPath,
                      sep = "")) 
    # Write the compounded .eml file as a .rds file
    saveRDS(eml_files, file = paste(outPath, "/eml_files", Sys.Date(),".rds", sep="" ))
    
    # Write the .rds file with all attribute information - this file can then be read into R again later 
    occurrences_attributes %>%
      saveRDS(., file = paste(outPath, "/", file_prefix, "completeAttributes_", Sys.Date(), ".rds", sep = ""))
  }}
  # Print completion note
  writeLines(paste(" - dataSaver. Fin.", sep = "\n"))
} # END dataSaver



#### +++++ ####



##### b. Set strings ####
#Set up bee family list
Bee_Families <- c("Andrenidae","Apidae", "Colletidae","Halictidae","Megachilidae","Melittidae",
                  "Stenotritidae","andrenidae","apidae", "colletidae","halictidae","megachilidae",
                  "melittidae","stenotritidae")


##### c. outFile_maker ####
outFile_maker <- function(path = path, file2make = "out_file"){
  # Write user output...
  writeLines(" - Checking for existing out_file directory...")
  # Look for outfile
  outFileLoc <- file.info(list.files(path, full.names = T, 
                                     pattern = file2make,
                                     recursive = TRUE,
                                     include.dirs = TRUE)
  )
  # IF there is not outfile, create one.
  if(nrow(outFileLoc) == 0){
    writeLines(paste(" - No existing,", file2make, " directory found. Creating directory...", sep = ""))
    dir.create(path = paste(path, file2make, sep = "/"))
  } # END create outfile
  # IF there IS an outfile, create one.
  if(nrow(outFileLoc) != 0){
    writeLines(paste(" - Existing ", file2make, "directory found. Data will be saved here.", sep = ""))
  } # END create outfile
  return(paste(path, file2make, sep = "/"))
} # END outFile_maker



#### ++++++ ####
#### d. fileFinder ####
#' Finds files within a directory
#' 
#' A function which can be used to find files within a user-defined directory based on a 
#' user-provided character string.
#'
#' @param path A directory as character. The directory to recursively search.
#' @param fileName A character/regex string. The file name to find.
#'
#' @return Returns a directory to the most-recent file that matches the provide file. Using regex
#' can greatly improve specificity.
#' The function will also write into the console the file that it has found - it is worthwhile to
#'  check that this is the correct file to avoid complications down the line
#' 
#' @importFrom stats complete.cases
#' @importFrom dplyr desc  %>%
#' 
#' @export
#'
#' @examples
#' \donttest{
#' # load dplyr
#' library(dplyr)
#' 
#'  # Make the RootPath to the tempdir for this example
#'   RootPath <- tempdir()
#'   
#'  # Load the example data
#'  data("beesRaw", package = "BeeBDC")
#' 
#' # Save and example dataset to the temp dir
#'   readr::write_csv(beesRaw, file = paste0(RootPath, "/beesRaw.csv"))
#' 
#'  # Now go find it!
#' fileFinder(path = RootPath, fileName = "beesRaw")
#' # more specifically the .csv version
#' fileFinder(path = RootPath, fileName = "beesRaw.csv")
#' }
fileFinder <- function(path, fileName){
  # locally bind variables to the function
  . <- dates <- NULL
  
  # Find all of the previously-produced data files
  locations <- file.info(list.files(path, full.names = T, pattern = fileName,
                                    recursive = TRUE))
  # Check if the data are present
  if(nrow(locations) == 0){ # If there are no data matching the name...
    stop(paste0(" - Bugger it, R can't find the file that you're looking for :(\n",
                "Please check that it exists in the provided directory."))
  }
  ##### Date from name ####
  # Extract only the file name to find the date from...
  FileName_dates <- stringr::str_replace(rownames(locations), ".+/", "") %>%
    # Remove additional text
    stringr::str_replace(., "[a-zA-Z\\/\\_]+[a-zA-Z\\/\\_]+", "") %>%
    # Find those files with dates in their name in d-m-y format
    dplyr::tibble(
      # Extract the dates from the file path rownames and supress the warning from non-matches
      suppressWarnings(lubridate::dmy(.), classes = "warning")) %>%
    # Set the column names of this new tibble
    stats::setNames(c("Locs","dates"))
       # IF there are no dates, look for ymd format
     if(sum(complete.cases(FileName_dates$dates)) == 0){
       FileName_dates <- stringr::str_replace(rownames(locations), ".+/", "") %>%
         # Remove additional text
         stringr::str_replace(., "[a-zA-Z\\/\\_]+[a-zA-Z\\/\\_]+", "") %>%
         # Find those files with dates in their name in d-m-y format
         dplyr::tibble(
           # Extract the dates from the file path rownames and supress the warning from non-matches
           suppressWarnings(lubridate::ymd(.), classes = "warning")) %>%
         # Set the column names of this new tibble
         stats::setNames(c("Locs","dates"))
     }
  # Insert the full locations in the Locs column
  FileName_dates$Locs <- rownames(locations)
  # Sort from most- to least-recent files
  FileName_dates <- dplyr::arrange(FileName_dates, dplyr::desc(dates))
  
  ##### Date from ctime ####
  # IF there are NO dates in the file names, use the file's ctime
  if(sum(complete.cases(FileName_dates$dates)) == 0){
    writeLines(" - No dates in file name(s). Finding most-recent from file save time...")
    # Find the most-recent file
    maxTime <- max(locations$ctime)
    # Extract the correct rowname (path)
    most_recent <- rownames(locations)[stringr::str_which(string = locations$ctime,
                                                 pattern = as.character(maxTime))]
    
  }else{
    writeLines(" - Dates found in file name(s). Finding most-recent file from file name...")
    # Return the strings containing this date
    most_recent <- FileName_dates[1,1] 
  } # END else
  # User output text
  writeLines(paste(
    " - Found the following file(s):", "\n",
    most_recent
  ))
  # Return this file location
  return(most_recent[[1]])
} # END fileFinder

#### ++++++ ####
#### e. re-export pipe operator ####
#' Pipe operator
#'
#' @name %>%
#' @rdname pipe
#' @keywords internal
#' @export
#' @importFrom dplyr %>%
#' @usage lhs \%>\% rhs
NULL

requireNamespace("readr")
requireNamespace("BeeBDC")
requireNamespace("dplyr")
requireNamespace("emld")

library(dplyr) ## could not use %>% without loading as library


## make some data
testData <- dplyr::tribble(
  ~catalog_number, ~pollinator_family, ~pollinator_genus, ~pollinator_species, ~collection_method, ~collector_number, ~day_collected, ~month_collected, ~year_collected,                   ~location_description, ~location_name,    ~habitat, ~latitude,  ~longitude, ~basis_of_record,
  "SFU7052435916",       "Halictidae",    "Secret",             "sp. 1",    "Pantrap, blue",               20L,            30L,            "Apr",           2007L, "Bear Hill Regional Park, Victoria CRD",           "BH", "Garry Oak", 48.545605, -123.406769,       "specimen",
  "SFU705917",       "Andrenidae",         "Secret",     "sp.",    "Pantrap, blue",               20L,            30L,            "Apr",           2007L, "Bear Hill Regional Park, Victoria CRD",           "BH", "Garry Oak", 48.545605, -123.406769,       "specimen",
  "SFU7052346919",       "Halictidae",    "Secret",       "sp.",    "Pantrap, blue",               20L,            30L,            "Apr",           2007L, "Bear Hill Regional Park, Victoria CRD",           "BH", "Garry Oak", 48.545605, -123.406769,       "specimen",
  "SFU7052645920",       "Andrenidae",         "Secret",     "sp.",    "Pantrap, blue",               20L,            30L,            "Apr",           2007L, "Bear Hill Regional Park, Victoria CRD",           "BH", "Garry Oak", 48.545605, -123.406769,       "specimen",
  "SFU705645921",       "Andrenidae",         "Secret",     "sp.",    "Pantrap, blue",               20L,            30L,            "Apr",           2007L, "Bear Hill Regional Park, Victoria CRD",           "BH", "Garry Oak", 48.545605, -123.406769,       "specimen"
)

  # Create a test attributes file
 testAttributes <- dplyr::tribble(
    ~dataSource, ~alternateIdentifier,              ~title,     ~pubDate,   ~dateStamp,           ~doi,                                          ~downloadLink, ~abstract,               ~citations,                                                     ~downloadCitation,                                                                          ~rights,
    "USGS_data",       "Not provided", "USGS_DRO database", "19/11/2022", "19/11/2022", "Not provided", "Not provided, contact Sam Droege at sdroege@usgs.gov",   "Empty", "Citations not provided", "Sam Droege. (2022-11-19). United States Geological Survey bee data.", "Rights are not provided. Please seek permission for data use from Same Droege."
    )
  # Add the attributes to the test data
 attributes(testData)$dataSource <- testAttributes

 # clear the temp directory - CAUTION THIS WILL COMPLETELY REMOVE ALL TEMP FILES FOR R SESSION
 file.remove(file.info(list.files(tempdir(), full.names = T, 
                                  recursive = TRUE)) %>% rownames())
## test the R file save method 
# write file
BeeBDC::dataSaver(path = tempdir(), 
                 save_type = "R_file", 
                 occurrences = testData)

# read files
list.files(tempdir(), full.names = T, pattern = "out_file",
           recursive = TRUE)

# Get a list of folders from the temp directory
folderList <- list.dirs(tempdir(), full.names = T, recursive = TRUE)

# Check that the "out_file" folder exists
testthat::test_that("dataSaver check that the out_file was created", {
  testthat::expect_true(any(stringr::str_detect(folderList, "out_file")))
})

# Get the paths for the files in that folder
testOutFiles <- file.info(list.files(paste(tempdir(), "out_file", sep = "/"), full.names = T, 
                             recursive = TRUE)) %>% rownames()

# Check the number of files == 1 (single .rds file)
testthat::test_that("dataSaver check that there is one file", {
  testthat::expect_true(length(testOutFiles) == 1)
})

testthat::test_that("dataSaver check that the one file is an .rds file", {
  testthat::expect_true(stringr::str_count(testOutFiles, "\\.rds") %>% sum() == 1)
})

# Test the output itself
testOut <- readRDS(testOutFiles[stringr::str_detect(testOutFiles,
                                                    "BeeData_[0-9]+-[0-9]+-[0-9]+")])

testthat::test_that("dataSaver RDS expected class", {
  testthat::expect_type(testOut, "list")
})

testOutData <- testOut[[1]]

testthat::test_that("read in occurrence data same as original data because no columns removed", {
  testthat::expect_equal(testData, testOutData)
})



## test the CSV save method
# clear the temp directory - CAUTION THIS WILL COMPLETELY REMOVE ALL TEMP FILES FOR R SESSION
file.remove(file.info(list.files(tempdir(), full.names = T, 
                                 recursive = TRUE)) %>% rownames())

# save the data
BeeBDC::dataSaver(path = tempdir(), 
                 save_type = "CSV_file", 
                 occurrences = testData)

# Get the paths for the files in that folder
testOutFiles <- file.info(list.files(paste(tempdir(), "out_file", sep = "/"), full.names = T, 
                                     recursive = TRUE)) %>% rownames()

# Check the file types
testthat::test_that("dataSaver check that there are two .csv files produced", {
  testthat::expect_true(stringr::str_count(testOutFiles, "\\.csv") %>% sum() == 2)
})

# Test the output itself
testOut2 <- readr::read_csv(file = testOutFiles[stringr::str_detect(testOutFiles, 
                                                                    "BeeData_combined")])

testthat::test_that("dataSaver csv expected class", {
  testthat::expect_type(testOut2, "list")
})

testthat::test_that("read in occurrence data same as original data because no columns removed", {
  testthat::expect_equal(testData, testOut2, ignore_attr = TRUE)
})

testOut3 <- readr::read_csv(file = testOutFiles[stringr::str_detect(testOutFiles, 
                                                                    "BeeData_attributes")])

testthat::test_that("dataSaver csv expected class", {
  testthat::expect_type(testOut3, "list")
})

testthat::test_that("read in occurrence data same as original data because no columns removed", {
  testthat::expect_equal(testAttributes, testOut3, ignore_attr = TRUE)
})

requireNamespace("readr")
requireNamespace("BeeBDC")
requireNamespace("dplyr")

library(dplyr) ## could not use %>% without loading as library


## fake data
testData <- dplyr::tribble(
  ~catalog_number, ~pollinator_family, ~pollinator_genus, ~pollinator_species, ~collection_method, ~collector_number, ~day_collected, ~month_collected, ~year_collected,                   ~location_description, ~location_name,    ~habitat, ~latitude,  ~longitude, ~basis_of_record,
  "SFU7052435916",       "Halictidae",    "Secret",             "sp. 1",    "Pantrap, blue",               20L,            30L,            "Apr",           2007L, "Bear Hill Regional Park, Victoria CRD",           "BH", "Garry Oak", 48.545605, -123.406769,       "specimen",
  "SFU705917",       "Andrenidae",         "Secret",     "sp.",    "Pantrap, blue",               20L,            30L,            "Apr",           2007L, "Bear Hill Regional Park, Victoria CRD",           "BH", "Garry Oak", 48.545605, -123.406769,       "specimen",
  "SFU7052346919",       "Halictidae",    "Secret",       "sp.",    "Pantrap, blue",               20L,            30L,            "Apr",           2007L, "Bear Hill Regional Park, Victoria CRD",           "BH", "Garry Oak", 48.545605, -123.406769,       "specimen",
  "SFU7052645920",       "Andrenidae",         "Secret",     "sp.",    "Pantrap, blue",               20L,            30L,            "Apr",           2007L, "Bear Hill Regional Park, Victoria CRD",           "BH", "Garry Oak", 48.545605, -123.406769,       "specimen",
  "SFU705645921",       "Andrenidae",         "Secret",     "sp.",    "Pantrap, blue",               20L,            30L,            "Apr",           2007L, "Bear Hill Regional Park, Victoria CRD",           "BH", "Garry Oak", 48.545605, -123.406769,       "specimen"
)


## write file
readr::write_excel_csv(testData, paste0(tempdir(), "/testData.csv"))


## find file 
testOut <- BeeBDC::fileFinder(path = paste0(tempdir()), fileName = "testData.csv")


## test
testthat::test_that("fileFinder expected class", {
  testthat::expect_type(testOut, "character")
})

# This function was written by James B. Dorey starting on the 14th of March 2023 to wrap ChaoSpecies
# and output table for multiple species/countries/variables at once


#' Parallel estimation of species richness in a community using iChao
#' 
#' ChaoSpecies: Estimation of species richness in a single community based on five types of data: 
#' Type (1) abundance data (datatype="abundance"), Type (1A) abundance-frequency counts 
#' (datatype="abundance_freq_count"), Type (2) incidence-frequency data 
#' (datatype = "incidence_freq"), Type (2A) incidence-frequency counts 
#' (datatype="incidence_freq_count"), and Type (2B) incidence-raw data (datatype="incidence_raw"); 
#' see SpadeR-package details for data input formats.
#'
#' @param data A data frame or tibble. A data frame containing "abundance"-type data per variable
#' (population, country, species...) in columns.
#' @param datatype Character. The type of input data, "abundance", "abundance_freq_count", 
#' "incidence_freq", "incidence_freq_count" or "incidence_raw". So far only tested with "abundance"
#' data. Default = "abundace".
#' @param k Numeric. the cut-off point (default = 10), which separates species into "abundant" and "rare" 
#' groups for abundance data for the estimator ACE; it separates species into "frequent" and 
#' "infrequent" groups for incidence data for the estimator ICE.
#' Default = 10.
#' @param conf Numeric. A positive number equal to or less than 1 specifying the level of 
#' confidence interval. Default = 0.95.
#' @param mc.cores Numeric. If > 1, the function will run in parallel
#' using mclapply using the number of cores specified. If = 1 then it will be run using a serial
#' loop. NOTE: Windows machines must use a value of 1 (see ?parallel::mclapply). Additionally,
#' be aware that each thread can use large chunks of memory.
#' Default = 1.
#'
#' @return Returns a list containing two tibbles. The first is a tibble that concatenates the outputs
#' from the basic data and rare species information in columns per input variable (column). The second is a tibble
#' that concatenates the various species richness estimates, with input variables in chunks of rows.
#' Additionally a console output will list the variables (columns) that lacked sufficient data
#' to be analysed.
#' 
#' 
#' @importFrom dplyr %>%
#' 
#' @seealso [SpadeR::ChaoSpecies()] 
#' 
#' @export
#'
#' @examples
#' \dontrun{
#'   # Read in some example data and use [BeeBDC::richnessPrepR()] to create the example input data
#'  #' data(beesCountrySubset)
#' 
#' estimateDataExample <- BeeBDC::richnessPrepR(
#'   data = beesCountrySubset,
#'   # Download the taxonomy
#'   taxonomyFile = BeeBDC::beesTaxonomy(),
#'   # Download the checklist
#'   checklistFile = BeeBDC::beesChecklist(),
#'   curveFunction = function(x) (228.7531 * x * x^-log(12.1593)),
#'   sampleSize = 10000,
#'   countryColumn = "country_suggested",
#'   limitGlobal = NULL,
#'   outPath = tempdir()
#' )
#' 
#'   # Transform the data for input
#' inputTestData <- estimateDataExample$site_speciesCounts %>%
#'   dplyr::select(scientificName, country_suggested, n) %>%
#'   tidyr::pivot_wider(names_from = country_suggested,
#'                      values_from = n,
#'                      values_fill = 0) %>%
#'   # Create the rownames
#' tibble::column_to_rownames("scientificName") %>%
#'   dplyr::tibble()
#' 
#' iChaoOut <- ChaoWrapper(
#' data = inputTestData,
#' datatype = "abundance",
#' k = 10,
#' conf = 0.95,
#' mc.cores = 1)
#' }


ChaoWrapper <- function(
    data = NULL,
    datatype = "abundance",
    k = 10,
    conf = 0.95,
    mc.cores = 1){
  # locally bind variables to the function
  i <- . <- inputData_i <- spOutput <- richnessOutput <- richnessTable <- basicOutput <-
    df_list <- loopVector <- wrapper <- basicCols <- variable <- rowname <- non_empty_list_test <-
    richnessOut <- basicOut <- failures <- output <- NULL
  
  # Load required packages 
  requireNamespace("stringr")
  requireNamespace("dplyr")
  requireNamespace("iNEXT")
  
  
  #### 0.0 Prep ####
  ##### 0.1 Errors ####
  ###### a. FATAL errors ####
  if(is.null(data)){
    stop(" - Please provide an argument for data I'm a program not a magician.")
  }
  
  
  ##### 0.2 SpadeR test ####
  ###### a. test ####
  # Check if SpadeR is installed
  # TRUE if SpadeR is found
  suppressWarnings(
    suggestedTest <- system.file(package='SpadeR') %>% 
      stringr::str_count() > 0 
  )
  
  ###### b. SpadeR ####
  if(suggestedTest == FALSE){
    # Set up instructions for download on fail
    instructions <- paste(" Please try installing the package for yourself", 
                          "using the following command: \n",
                          "install.packages(\"SpadeR\")")
    # Set up fail function for tryCatch
    error_func <- function(e){
      stop(paste("Failed to install the SpadeR package.\n", 
                 instructions))
    }
    # Begin interactive input
    input <- 1
    if (interactive()){
      input <- utils::menu(c("Yes", "No"), 
                           title = paste0("Install the SpadeR package? \n"))
    }
    if(input == 1){
      # Start SpadeR install
      message("Installing the SpadeR package.")
      tryCatch(
        utils::install.packages("SpadeR"), 
        error = error_func, warning = error_func)
    } # END input == 1
    
    else{
      stop(writeLines(paste("The SpadeR package is necessary for BeeBDC::ChaoWrapper\n", 
                            instructions)))
    } # END else
  } # END suggestedTest == FALSE
  
  
      ##### 0.2 Functions ####
  # Duplicate a function within ChaoSpecies to not run data-poor species
  f <- function(i, data) {
    length(data[which(data == i)])
  }
  
    # wrap the ChaoSpecies function in a wrapper to be passed to mclapply
  wrapper <- function(inputData_i = NULL){
    # Don't run species where there are no counts that are less than k
    # i.e. ONLY run species that won't throw an error due to weird/poor cases of data
    if(any(inputData_i[[1]] > 0 & inputData_i[[1]] < 5) & 
       !(f(1, inputData_i[[1]]) == sum(inputData_i[[1]], na.rm = TRUE))
    ){
      # Run the ChaoSpecies function
      spOutput <- SpadeR::ChaoSpecies(
        data = inputData_i[[1]][complete.cases(inputData_i[[1]])],
        datatype = datatype,
        k = k, 
        conf = conf)
        ###### a. richnessOutput ####
      # Get the richness measures and re-format
      richnessOutput <- spOutput$Species_table %>% as.data.frame() %>% 
        dplyr::mutate(rowname = rownames(.), .before = "Estimate") 
      # Add these data to one table
      richnessTable <- dplyr::tibble(
        variable = names(inputData_i),
        Name = c(richnessOutput$rowname),
        Estimate = c(richnessOutput$Estimate),
        's.e.' = c(richnessOutput$s.e.),
        '95%Lower' = c(richnessOutput$`95%Lower`),
        '95%Upper' = c(richnessOutput$`95%Upper`))
      
      
        ##### b. basicOutput ####
      # Get Basic information into one table
      basicOutput <- spOutput$Basic_data_information %>% as.data.frame() %>% 
          # Remove hanging white spaces at the start of the rowname
        dplyr::mutate(rowname = rownames(.) %>% 
                        gsub(pattern = "^\\s+", replacement = ""),
                      .before = "Variable") %>% 
        dplyr::tibble()
      
      # Set up the data
      basicTable <- basicOutput
      # rename Value to the species name
      names(basicTable) <- c("rowname", "variable", names(inputData_i))
      # Return the tables as a list
      return(list( basicTable, richnessTable) %>%
               stats::setNames(c( "basicTable", "richnessTable")))
    }
  }
  
#### 1.0 Data prep ####
  # Loop the data to make it into a list (it's a two-level list) because mclapply will strip one layer
df_list <- dplyr::lst()
for(i in 1:ncol(data)){
# Extract vector
      loopVector <- dplyr::pull(data, colnames(data)[i]) %>%
        dplyr::lst() %>%
        stats::setNames(., colnames(data)[i])
      # Add to the list
      df_list <- df_list %>%
      append( dplyr::lst(loopVector))
}
  # Set the names for the list
df_list <- df_list %>%
  stats::setNames(colnames(data))

  
#### 2.0 Run functions ####    
# Run the function per species or variable
richnessOutput <- parallel::mclapply(
  X = df_list,
  FUN = wrapper,
  mc.cores = mc.cores
) 


# Save the richness table's first two columns
basicCols <- richnessOutput[[1]]$basicTable %>%
  dplyr::select(rowname, variable)


#### 3.0 Process outputs ####
  ##### 3.1 Separate F + P ####
  # Find the non-null variables and extract those
non_empty_list_test <- !sapply(richnessOutput <- richnessOutput, is.null)
  # Save a list of variables that could not be analysed
failures <- richnessOutput[!non_empty_list_test] %>%
    # Get the failure names 
  names()
  # Remove the failed list
richnessOutput <- richnessOutput[non_empty_list_test]


  ##### 3.2 Basic outputs ####
  # Now, combine each level of the list across variables
  # Extract the richness table stuff
basicOut <- lapply(richnessOutput, function(x) x[["basicTable"]]) %>%
  # Remove the duplicate columns and only keept the output values
  lapply(., function(x) dplyr::select(x, !c("rowname", "variable"))) %>%
    # Bind together with the original two columns
  dplyr::bind_cols(basicCols, .)
  

  ##### 3.3 richness outputs ####
  # Row bind the richness statistics into a single table
richnessOut <- lapply(richnessOutput, function(x) x[["richnessTable"]]) %>%
  dplyr::bind_rows()


  ##### 3.4 Combine ####
output <- dplyr::lst(basicOut, richnessOut) %>% 
  stats::setNames(c("basicTable", "richnessTable"))
  
#### 4.0 User output ####
  # provide some user output on the failures
if(length(failures > 0)){
writeLines(paste0(
  " - We could not examine the following variables (because of insufficent data or sample size): ",
  paste(failures, collapse = ", ")
))}
  # provide user output about the file structure
message(paste0(" - Outputs can be found in a list with two tibbles called 'basicTable' and",
                  " 'richnessTable'."))

  # Return the output
return(output)

} # END ChaoWrapper JBD

#####  2.2 attr_builder ####
#' @importFrom lubridate as_date
#' @importFrom stringr str_to_sentence
#' @importFrom dplyr %>%

attr_builder <- function(path_i = path_i, occ_input = data_i){
  # locally bind variables to the function
  lubridate <- stringr <- family <- data_i <- NULL
  
  requireNamespace("lubridate")
  requireNamespace("dplyr")
  requireNamespace("xml2")
  # This function behaves differently depending on the data source, but returns common outputs.
  #### ALA START ####
  if(grepl("/data.csv", path_i) == "TRUE"){
    # Find and take the citations file 
    citations_i <- gsub("/data.csv", "/citation.csv", path_i) %>% 
      readr::read_csv(col_types = readr::cols(.default = readr::col_character())) 
    # Find and take the download information file 
    galahDL_i <- gsub("/data.csv", "/galah_DL_info.csv", path_i) %>% 
      readr::read_csv(n_max = 1, col_types = readr::cols(.default = readr::col_character())) 
    # Read in the closest thing ALA has to an abstract
    galahAbstract_i <- gsub("/data.csv", "/README.html", path_i) %>% 
      rvest::read_html() %>%
      rvest::html_text2()
    # Combine all of these attributes into a tibble
    Attributes_i <- dplyr::tibble(dataSource = paste("ALA_", 
                                                      galahDL_i$taxon, 
                                                      sep = ""),
                                   alternateIdentifier = if("search_url" %in% colnames(galahDL_i)){
                                     galahDL_i$search_url}else(NA_character_),
                                   title = "ALA Occurrence Download. ", 
                                   pubDate = galahDL_i$download_date %>%
                                                              lubridate::as_date(),
                                   dateStamp = galahDL_i$download_date,
                                   doi = galahDL_i$doi,
                                   downloadLink = galahDL_i$download_link,
                                   abstract = dplyr::lst(galahAbstract_i),
                                   citations = dplyr::lst(citations_i),
                                   downloadCitation = paste("ALA.org.au. (", 
                                                            lubridate::as_date(galahDL_i$download_date) %>% 
                                                              format("%d %B %Y"),
                                                            "). ALA Occurrence Download. ",
                                                            galahDL_i$doi,
                                                            sep = ""),
                                   rights = dplyr::lst("See occurrence records"),
                                  taxon = galahDL_i$taxon)
    
    # combine the input eml and the attributes tibble into a list for output from the function
    EML_attributes <- list("No_eml_from_ALA", Attributes_i)
    names(EML_attributes) <- c("source_eml","Source_tibble") 
    # output this list
    return(EML_attributes)
  } # ALA END 
  
  #### GBIF START ####
  if(grepl("occurrence.txt", path_i) == "TRUE"){
    # Find and take the metadata file
    sourceEML_i <- emld::as_emld(gsub("/occurrence.txt", "/metadata.xml", path_i), from = "xml" )
    # Find and take the citations file - convert into a list
    citations_i <- gsub("/occurrence.txt", "/citations.txt", path_i) %>% 
      readr::read_lines() %>%
      dplyr::lst()
    # Find and take the rights file - convert into a list
    rights_i <- gsub("/occurrence.txt", "/rights.txt", path_i) %>% 
      readr::read_lines() %>%
      dplyr::lst()
    # Find the download name
    fam_name <- tidyr::drop_na(
      occ_input, tidyselect::any_of("family")) %>%
      dplyr::pull(family) %>%
      unique()
    # Combine all of these attributes int o tibble
    Attributes_i <- dplyr::tibble(dataSource = paste("GBIF_", 
                                                      fam_name, 
                                                      sep = ""),
                                   alternateIdentifier =  sourceEML_i$dataset$alternateIdentifier,
                                   title = sourceEML_i$dataset$title, 
                                   pubDate = sourceEML_i$dataset$pubDate %>% 
                                     stringr::str_match("[0-9]{4}-[0-9]{2}-[0-9]{2}") %>%
                                     lubridate::as_date(),
 dateStamp = sourceEML_i$additionalMetadata$metadata$gbif$dateStamp,
 doi = paste("https://doi.org/",
             sourceEML_i$additionalMetadata$metadata$gbif$citation$identifier, sep = ""),
 downloadLink = sourceEML_i$additionalMetadata$metadata$gbif$physical$distribution$online$url$url,
 abstract = dplyr::lst(sourceEML_i$dataset$abstract),
 citations = dplyr::lst(citations_i),
 downloadCitation = paste("GBIF.org. (", 
                          lubridate::as_date(sourceEML_i$dataset$pubDate) %>% 
                            format("%d %B %Y"),
                          "). GBIF Occurrence Download. ",
                          paste("https://doi.org/",
                                sourceEML_i$additionalMetadata$metadata$gbif$citation$identifier, 
                                sep = ""),
                          sep = ""),
                                   rights = dplyr::lst(rights_i),
 taxon = fam_name)
    
    # combine the input eml and the attributes tibble into a list for output from the function
    EML_attributes <- list(sourceEML_i, Attributes_i)
    names(EML_attributes) <- c("source_eml","Source_tibble") 
    # output this list
    return(EML_attributes)
  } # GBIF END 
  
  
  #### iDigBio START ####
  if(grepl("occurrence_raw.csv", path_i) == "TRUE"){
    # Find and take the citations file - convert into a list
    citations_i <- gsub("/occurrence_raw.csv", "/records.citation.txt", path_i) %>% 
      readr::read_lines() %>%
      dplyr::lst()
    # Combine all of these attributes int o tibble
    Attributes_i <- dplyr::tibble(dataSource = paste("iDigBio_", citations_i$.[2] %>% 
                                                        stringr::str_match_all("[A-Za-z]+") %>%
                                                        unlist() %>% 
                                                        dplyr::last(), sep = ""),
                                  # Check for alt identifier and include if there
                                   alternateIdentifier =  
if(length(path_i %>% 
           stringr::str_match_all("[a-zA-Z0-9]+-[a-zA-Z0-9]+-[a-zA-Z0-9]+-[a-zA-Z0-9]+-[a-zA-Z0-9]+") %>%
           unlist()) > 0){
  # TRUE
               path_i %>% 
                 stringr::str_match_all("[a-zA-Z0-9]+-[a-zA-Z0-9]+-[a-zA-Z0-9]+-[a-zA-Z0-9]+-[a-zA-Z0-9]+") %>%
                 unlist()
      }else{
        # FALSE
        NA_character_},
                                   title = citations_i$.[1], 
                                   pubDate = citations_i$.[3] %>% 
  stringr::str_match("[0-9]{4}-[0-9]{2}-[0-9]{2}") %>%
                                     lubridate::as_date(),
                                   dateStamp = citations_i$.[3] %>% 
  stringr::str_match("[0-9]{4}-[0-9]{2}-[0-9]{2}T[0-9]{2}:[0-9]{2}:[0-9]{2}") %>%
                                     as.character(),
                                   doi = "Use downloadLink for iDigBio",
                                   downloadLink = paste("http://s.idigbio.org/idigbio-downloads/", 
                                                        path_i %>% 
stringr::str_match_all("[a-zA-Z0-9]+-[a-zA-Z0-9]+-[a-zA-Z0-9]+-[a-zA-Z0-9]+-[a-zA-Z0-9]+"),
                                                        ".zip", sep = ""),
                                   abstract = dplyr::lst(citations_i$.[1:4]),
                                   citations = dplyr::lst(citations_i$.[4:length(citations_i$.)]),
                                   downloadCitation = paste("iDigBio.org. (", 
                                                            citations_i$.[3] %>% 
stringr::str_match("[0-9]{4}-[0-9]{2}-[0-9]{2}") %>%
                                                              lubridate::as_date() %>% 
                                                              format("%d %B %Y"),
                                                            "). iDigBio Occurrence Download. ",
                                                  paste("http://s.idigbio.org/idigbio-downloads/", 
path_i %>% 
  stringr::str_match_all("[a-zA-Z0-9]+-[a-zA-Z0-9]+-[a-zA-Z0-9]+-[a-zA-Z0-9]+-[a-zA-Z0-9]+"),
                                                                  ".zip", sep = ""),
                                                            sep = ""),
                                   rights = dplyr::lst("See occurrence records"),
taxon = (citations_i$.[2] %>% 
  stringr::str_match_all("[A-Za-z]+") %>%
  unlist() %>% 
  dplyr::last())
) # END tibble()
    
    # combine the input eml and the attributes tibble into a list for output from the function
    EML_attributes <- list("No_eml_from_iDigBio", Attributes_i)
    names(EML_attributes) <- c("source_eml","Source_tibble") 
    # output this list
    return(EML_attributes)
  } # iDigBio END 
  
  #### SCAN START ####
  if(grepl("occurrences.csv", path_i) == "TRUE"){
    # Find and take the metadata file
    sourceEML_i <- xml2::read_xml(gsub("/occurrences.csv", "/eml.xml", path_i), from = "xml" ) %>% 
      emld::as_emld()
    # Combine all of these attributes int o tibble
    Attributes_i <- dplyr::tibble(dataSource = paste("SCAN_", 
                                                      unique(stringr::str_to_sentence(occ_input$family)), 
                                                      sep = ""),
alternateIdentifier =  sourceEML_i$additionalMetadata$metadata$symbiota$citation$identifier,
                                   title =  sourceEML_i$dataset$title$title, 
pubDate = sourceEML_i$additionalMetadata$metadata$symbiota$dateStamp %>% 
  stringr::str_match("[0-9]{4}-[0-9]{2}-[0-9]{2}") %>%
                                     lubridate::as_date(),
dateStamp = sourceEML_i$additionalMetadata$metadata$symbiota$dateStamp,
                                   doi = "SCAN does not provide a doi. See download link.",
downloadLink = "SCAN does not provide a download link.",
                                   abstract = dplyr::lst("SCAN does not provide a single abstract"),
                                   citations =     dplyr::lst(paste("SCAN. ", 
                          (sourceEML_i$additionalMetadata$metadata$symbiota$dateStamp %>% 
                             stringr::str_match("[0-9]{4}")),
                          ". http//:scan-bugs.org/portal/index.php. ",
                          "accessed on ", 
                          (sourceEML_i$additionalMetadata$metadata$symbiota$dateStamp %>% 
                             stringr::str_match("[0-9]{4}-[0-9]{2}-[0-9]{2}")),
                          ". ", sourceEML_i$additionalMetadata$metadata$symbiota$citation$citation, 
                          sep = "" ) ),
                                   downloadCitation = paste("SCAN. (", 
                         sourceEML_i$additionalMetadata$metadata$symbiota$dateStamp %>% 
                           stringr::str_match("[0-9]{4}-[0-9]{2}-[0-9]{2}") %>%
                           lubridate::as_date() %>% 
                           format("%d %B %Y"),
                         "). SCAN-Bugs Occurrence Download ",
                         "http//:scan-bugs.org/portal/index.php. uuid - ",
                         sourceEML_i$additionalMetadata$metadata$symbiota$citation$identifier,
                         sep = ""),
                                   rights = dplyr::lst("See .xml for rights"),
taxon = unique(stringr::str_to_sentence(occ_input$family)))
    # combine the input eml and the attributes tibble into a list for output from the function
    EML_attributes <- list(sourceEML_i, Attributes_i)
    names(EML_attributes) <- c("source_eml","Source_tibble") 
    # output this list
    return(EML_attributes)
  } # SCAN END
} # END attr_builder

# This Function is designed to check state-level outliers of the t2t project.
# It was written by James B Dorey from the 1st of August 2022.
#' @importFrom dplyr %>%

StateOutlieRs <- function(
    checklist = NULL,
    checklistColumns = NULL,
    countryList = NULL,
    occData = NULL,
    findHawaii = FALSE
    ){
    # locally bind variabls to the function
  . <- SciPost <- neighbourMatch <- nAssignmentCertainty <- exactMatch <- matchType <- 
    matchCertainty <- stateMatch <- jurisdiction_value<-stateValue<-adm1_code<-iso_3166_2<-name<-
    region<-postal<- geometry<-longitude<-latitude<-database_id<-scientificName<-species<-family<-
    subfamily<-genus<-specificEpithet<-country<-stateProvince<-eventDate<-institutionCode<-
    recordNumber<-catalogNumber<-dataSource<-verbatim_scientificName<-neighbours<-rowNum<-state<-
    neighboursText<-Genus<-Species<-assignmentCertainty <- NULL
  
#### 0.0 Warnings ####
  if(is.null(checklist)){
    stop("You must provide a checklist of states.")
  }
  if(is.null(checklistColumns)){
    stop("You must provide vector of checklistColumns.")
  }
  if(is.null(countryList)){
    stop("You must provide vector of countryList for rnaturalearth to download states from.")
  }
  if(is.null(occData)){
    stop("You must provide occurrence data (occData). Honestly, what do you think I was gonna do without that?")
  }

#### 1.0 Data prep ####
  ##### 1.1 Ascher ####

  # Find Hawaii if true
  if(findHawaii == TRUE){
    checklist <- checklist %>%
      dplyr::mutate(HI = 
                      # IF Hawaii is mentioned in the jurisdiction_value column, then the species is known from there (introduced or not)
                      dplyr::if_else(stringr::str_detect(
                        string = jurisdiction_value,
                        pattern = "Hawaii"),
                        "HI", ""))
    # Add Hawaii to checklistColumns
    checklistColumns <- c(checklistColumns, "HI")
  }
  # For the above columns, turn into long-format spreadhseet. Each species in EACH state will have a 
  # row of data. 
CL <- checklist %>%
  tidyr::pivot_longer(data = .,
                      cols = tidyselect::all_of(checklistColumns),
                      names_to = "state", values_to = "stateValue") %>%
    # remove empty cells for the values
  tidyr::drop_na(stateValue) %>%
    # add a certainty level == "Poor" for states with a "?"
  dplyr::mutate(assignmentCertainty = 
                  dplyr::if_else(stringr::str_detect(string = stateValue,
                                                     pattern = "\\?"),
                                 "Poor", "Good"))

  ##### 1.2 rNaturalEarth ####
# Download world map using rnaturalearth packages
stateMap <- rnaturalearth::ne_states(returnclass = "sf", country = countryList)  %>%
  # Select only a subset of the naturalearthdata columns to extract
  dplyr::select(adm1_code, iso_3166_2, name, region, postal, geometry)
  # Dont's use spherical geometry
sf::sf_use_s2(FALSE)

  # Can examine missing states between the two
if(length(setdiff(sort(unique(stateMap$postal)), sort(unique(CL$state))) > 0)){
  message(paste0("In the Ascher list, you are mising the following state(s) that occur in the rnaturalearth package:\n",
                 stringr::str_c(setdiff(sort(unique(stateMap$postal)), sort(unique(CL$state))), 
                                collapse = ", ")
                 ))}
if(length(setdiff( sort(unique(CL$state)), sort(unique(stateMap$postal))) > 0)){
  message(paste0("In the rnaturalearth list, you are mising the following state(s) that occur in the ascher list:\n",
                 stringr::str_c(setdiff( sort(unique(CL$state)), sort(unique(stateMap$postal))), 
                                collapse = ", ")
                 ))}


  #### 2.0 Use occ. data ####
  ##### 2.1 Angela data ####
  # Turn occData into a simple point feature
  points <- sf::st_as_sf(occData %>%
                           tidyr::drop_na(longitude, latitude),
                         coords = c("longitude", "latitude"),
                            # Assign the CRS from the rnaturalearth map to the point data
                            crs = sf::st_crs(stateMap)) %>%
      # Use a subset of columns
    dplyr::select(database_id, scientificName, species, family, subfamily, genus, specificEpithet, 
                  country, stateProvince, eventDate, institutionCode, recordNumber, catalogNumber,
                  dataSource, verbatim_scientificName, geometry)
  
    ##### 2.2 Extraction ####
    ###### a. exactState ####
  writeLines(" - Extracting state data from points...")
    #Extract polygon information to points
  points_extract <- sf::st_intersection(stateMap,
                              points)
  ###### a. neighbouringStates ####
    # Get a list of states that share borders
  statesBordering <- sf::st_intersects(stateMap, stateMap) %>%
    paste(., sep = ";")
    # Make a new tibble with these information
  neighbouringStates <- dplyr::tibble(
    rowNum = 1:nrow(stateMap),
    state = stateMap$postal,
    neighbours = statesBordering,
      # Modify the text in column
    neighboursText = mgsub::mgsub(string = neighbours,
                                  pattern = rowNum,
                                  replacement = state) %>%
      stringr::str_replace(string = .,
                           pattern = "c\\(", replacement = "") %>%
      stringr::str_replace(string = .,
                           pattern = "\\)", replacement = "") %>%
      stringr::str_replace(string = .,
                           pattern = ":", replacement = ", ")) 
  
    # Make a long-format tibble with neighbouring states
  neighbouringStates <- neighbouringStates %>%
    tidyr::separate_rows(data = ., neighboursText,
                              sep = ",")  %>%
      # Remove states matching themselves
    dplyr::filter(!(state == neighboursText))
    # Remove extra spaces
  neighbouringStates$neighboursText <- stringr::str_squish(neighbouringStates$neighboursText)
    # Join the datasets togehter so that we can make a list of adjacent states to match also
  neighbouringStates <- CL %>%
    dplyr::left_join(dplyr::select(neighbouringStates, c(state, neighboursText)),
                     by = "state", multiple = "all", relationship = "many-to-many")
    

  
    ##### 2.3 Compare ####
    # Get a smaller subset of the data AND make a new columns with scientific name and state
  points_simple <- points_extract %>% 
    dplyr::select(database_id, postal, genus, specificEpithet, country) %>%
    dplyr::mutate(SciPost = stringr::str_c(genus, specificEpithet, postal, sep = "_"))
  
    ###### a. exactState ####
    # Do the same for the ascher checklist
  CL_simple <- CL %>%
      # Select subset
    dplyr::select(Genus, Species, state, assignmentCertainty) %>%
      # Harmonise column names
    dplyr::rename(genus = Genus,
                  species = Species,
                  postal = state) %>%
      # Make the new column to match with
    dplyr::mutate(SciPost = stringr::str_c(genus, species, postal, sep = "_"))
    # Make a new columns showing if that species is expected in that state.
  points_match <- points_simple %>%
    dplyr::filter(stringr::str_detect(tolower(country), "united states")) %>%
    dplyr::mutate(exactMatch = dplyr::if_else(SciPost %in% CL_simple$SciPost,
                                        TRUE, FALSE)) %>%
      # join the assignmentCertainty column if there's a match
    dplyr::left_join(dplyr::select(CL_simple, SciPost, assignmentCertainty), 
                     by = "SciPost", multiple = "all", relationship = "many-to-many" )
    # Show a quick summary
    # table(points_match$exactMatch, useNA = "always")
    # table(points_match$assignmentCertainty, useNA = "always")
  
  
    ###### b. neighbouringStates ####
  # Get a smaller subset of the data AND make a new columns with scientific name and state
  nCL_simple <- neighbouringStates %>%
    # Select subset
    dplyr::select(Genus, Species, neighboursText, assignmentCertainty) %>%
    # Harmonise column names
    dplyr::rename(genus = Genus,
                  species = Species,
                  postal = neighboursText) %>%
    # Make the new column to match with
    dplyr::mutate(SciPost = stringr::str_c(genus, species, postal, sep = "_")) %>%
      # Get a unique set
    dplyr::distinct(SciPost, .keep_all = TRUE)
  # Make a new columns showing if that species is expected in that state.
  npoints_match <- points_simple %>%
    dplyr::filter(stringr::str_detect(tolower(country), "united states")) %>%
    dplyr::mutate(neighbourMatch = dplyr::if_else(SciPost %in% nCL_simple$SciPost,
                                              TRUE, FALSE)) %>%
      # Assign neighbourMatch for assignmentCertainty where occurrence was neighbour matched.
    dplyr::mutate(nAssignmentCertainty = dplyr::if_else(neighbourMatch == TRUE,
                                                       "neighbourMatch",""))
  
  # Show a quick summary
    # table(npoints_match$neighbourMatch, useNA = "always")
    # table(npoints_match$nAssignmentCertainty, useNA = "always")

  
  #### 3.0 Merge ####
  writeLines(" - Combining data...")
    # Merge both points_match datasets
  bpoints_match <- dplyr::tibble(points_match) %>%
      # Join the two datasets togehter keeping only neighbourMatch and assignmentCertainty from the 
      # neighbour-joined dataset
    dplyr::left_join(dplyr::select(npoints_match, c(database_id, neighbourMatch, 
                                                       nAssignmentCertainty)),
                     by = "database_id", multiple = "all", relationship = "many-to-many") %>%
      # Remove geometry column
    dplyr::select(!tidyselect::starts_with("geometry")) %>%
      # Combine exactMatch and neighbourMatch
    dplyr::mutate(matchType = dplyr::if_else(exactMatch == TRUE,
                                             "exact", dplyr::if_else(neighbourMatch == TRUE,
                                                                     "neighbour", 
                                                                     "noMatch"))) %>%
    # combine assignmentCertainty and nAssignmentCertainty
    dplyr::mutate(matchCertainty = dplyr::if_else(matchType == "exact",
                                                  assignmentCertainty, 
                                                  dplyr::if_else(matchType == "neighbour",
                                                                 nAssignmentCertainty, "NA")))
  
  #### 4.0 Output ####
    ##### 4.1 User output ####
  writeLines(paste0(
    " - Finished. \n",
    "We have matched ", 
    format(sum(bpoints_match$matchType == "exact", na.rm = TRUE), big.mark = ","),
    " records to their exact state and ", 
    format(sum(bpoints_match$matchType == "neighbour", na.rm = TRUE), big.mark = ","), 
    " to an adjacent state.\n", 
    "We failed to match ",
    format(sum(bpoints_match$matchType == "noMatch", na.rm = TRUE), big.mark = ","), 
    " occurrences to any 'exact' or 'neighbouring' state.\n",
    "Of the 'exact' matches, ",
    format(sum(bpoints_match$matchCertainty == "Good", na.rm = TRUE), big.mark = ","), 
    " are 'good matches' according to John Ascher and ",
    format(sum(bpoints_match$matchCertainty == "Poor", na.rm = TRUE), big.mark = ","), 
    " are 'poor matches' might benefit from checking"
  ))
  
  bpoints_match <- bpoints_match %>%
    # Select the columns to keep
    dplyr::select(database_id, postal, matchType, matchCertainty) %>%
    dplyr::rename(
      stateMatch = matchType,
      stateMatch_certainty = matchCertainty) %>%
      # Set flag for those that don't pass statematch
    dplyr::mutate(
      .stateOutlier = dplyr::if_else(stateMatch != "exact" & stateMatch != "neighbour",
                                     FALSE, TRUE))
      # Merge with orignal dataset
    output <- occData %>%
      dplyr::left_join(bpoints_match, by = "database_id", multiple = "all", relationship = "many-to-many") %>%
      dplyr::distinct(database_id, .keep_all = TRUE)

    # return message
    message(paste("\nchecklistOutlieR:\nFlagged", 
                  format(sum(output$.stateOutlier == FALSE, na.rm = TRUE), big.mark = ","), 
                  "records.\nThe column, '.stateOutlier',",
                  "was added to the database.\n"), sep = "")
  # Return file
return(output)

} # END checklistOutlieR

# This function was written by James B Dorey on the 20th of February 2023 to remove too-simple
  # id codes from occurrences. It is intended to be used internal in the jbd_dupeSummary function
#' @importFrom dplyr %>%
#' 
deSimplifieR <- function(inputData = NULL,
                         characterThreshold = 2,
                         numberThreshold = 3,
                         numberOnlyThreshold = 5)
{
  
  requireNamespace("dplyr")
  
  #### 1.0 Remove simple strings ####
    # Remove simple codes, only do for each column if that column exists
  inputData <- inputData %>%
  dplyr::mutate(
    if("occurrenceID" %in% colnames(inputData)){
    occurrenceID = dplyr::if_else( stringr::str_count(occurrenceID, "[A-Za-z]") >= characterThreshold &
                                     stringr::str_count(occurrenceID, "[0-9]") >= numberThreshold |
                                     stringr::str_count(occurrenceID, "[0-9]") >= numberOnlyThreshold, 
                                   occurrenceID, NA_character_)},
    if("recordId" %in% colnames(inputData)){
    recordId = dplyr::if_else( stringr::str_count(recordId, "[A-Za-z]") >= characterThreshold &
                                 stringr::str_count(recordId, "[0-9]") >= numberThreshold |
                                 stringr::str_count(recordId, "[0-9]") >= numberOnlyThreshold, 
                               recordId, NA_character_)},
    if("id" %in% colnames(inputData)){
    id = dplyr::if_else( stringr::str_count(id, "[A-Za-z]") >= characterThreshold &
                           stringr::str_count(id, "[0-9]") >= numberThreshold |
                           stringr::str_count(id, "[0-9]") >= numberOnlyThreshold, 
                         id, NA_character_)}, 
    if("catalogNumber" %in% colnames(inputData)){
    catalogNumber = dplyr::if_else( stringr::str_count(catalogNumber, "[A-Za-z]") >= characterThreshold &
                                      stringr::str_count(catalogNumber, "[0-9]") >= numberThreshold |
                                      stringr::str_count(catalogNumber, "[0-9]") >= numberOnlyThreshold, 
                                    catalogNumber, NA_character_)}, 
    if("otherCatalogNumbers" %in% colnames(inputData)){
    otherCatalogNumbers = dplyr::if_else( stringr::str_count(otherCatalogNumbers, "[A-Za-z]") >= characterThreshold &
                                            stringr::str_count(otherCatalogNumbers, "[0-9]") >= numberThreshold |
                                            stringr::str_count(otherCatalogNumbers, "[0-9]") >= numberOnlyThreshold, 
                                          otherCatalogNumbers, NA_character_)})


return(inputData)
}

# Description of the BeeBDC toy dataset beesRaw.csv written by James B. Dorey on the 
# 16th of March 2023

#' A dataset of 100 random bee occurrence records without flags or filters applied
#'
#' A small bee occurrence dataset with flags generated by BeeBDC used to run example script and test 
#' functions. For data types, see [BeeBDC::ColTypeR()].
#'
#' @docType data
#'
#' @usage data("beesRaw", package = "BeeBDC")
#'
#' @format An object of class \code{"tibble"}
#' \describe{
#'  \item{database_id}{Occurrence code generated in bdc or BeeBDC}
#'  \item{scientificName}{Full scientificName as shown on DiscoverLife}
#'  \item{family}{Family name}
#'  \item{subfamily}{Subfamily name}
#'  \item{genus}{Genus name}
#'  \item{subgenus}{Subgenus name}
#'  \item{subspecies}{Full name with subspecies name - ALA column}
#'  \item{specificEpithet}{The species name only}
#'  \item{infraspecificEpithet}{The subspecies name only}
#'  \item{acceptedNameUsage}{The full name, with authorship and date information if known, of the currently valid (zoological) or accepted (botanical) taxon.}
#'  \item{taxonRank}{The taxonomic rank of the most specific name in the scientificName.}
#'  \item{scientificNameAuthorship}{The authorship information for the scientificName formatted according to the conventions of the applicable nomenclaturalCode.}
#'  \item{identificationQualifier}{A brief phrase or a standard term ("cf.", "aff.") to express the determiner's doubts about the Identification.}
#'  \item{higherClassification}{A list (concatenated and separated) of taxa names terminating at the rank immediately superior to the taxon referenced in the taxon record.)}
#'  \item{identificationReferences}{A list (concatenated and separated) of references (publication, global unique identifier, URI) used in the Identification.}
#'  \item{typeStatus}{A list (concatenated and separated) of nomenclatural types (type status, typified scientific name, publication) applied to the subject.}
#'  \item{previousIdentifications}{A list (concatenated and separated) of previous assignments of names to the Organism.}
#'  \item{verbatimIdentification}{This term is meant to allow the capture of an unaltered original identification/determination, including identification qualifiers, hybrid formulas, uncertainties, etc. This term is meant to be used in addition to scientificName (and identificationQualifier etc.), not instead of it.}
#'  \item{identifiedBy}{A list (concatenated and separated) of names of people, groups, or organizations who assigned the Taxon to the subject.}
#'  \item{dateIdentified}{The date on which the subject was determined as representing the Taxon.}
#'  \item{decimalLatitude}{The geographic latitude (in decimal degrees, using the spatial reference system given in geodeticDatum) of the geographic center of a Location. Positive values are north of the Equator, negative values are south of it. Legal values lie between -90 and 90, inclusive.}
#'  \item{decimalLongitude}{The geographic longitude (in decimal degrees, using the spatial reference system given in geodeticDatum) of the geographic center of a Location. Positive values are east of the Greenwich Meridian, negative values are west of it. Legal values lie between -180 and 180, inclusive.}
#'  \item{stateProvince}{The name of the next smaller administrative region than country (state, province, canton, department, region, etc.) in which the Location occurs.}
#'  \item{continent}{The name of the continent in which the Location occurs.}
#'  \item{locality}{The specific description of the place.}
#'  \item{island}{The name of the island on or near which the Location occurs.}
#'  \item{county}{The full, unabbreviated name of the next smaller administrative region than stateProvince (county, shire, department, etc.) in which the Location occurs.}
#'  \item{municipality}{The full, unabbreviated name of the next smaller administrative region than county (city, municipality, etc.) in which the Location occurs. Do not use this term for a nearby named place that does not contain the actual location.}
#'  \item{license}{A legal document giving official permission to do something with the resource.}
#'  \item{issue}{A GBIF-defined issue.}
#'  \item{eventDate}{The date-time or interval during which an Event occurred. For occurrences, this is the date-time when the event was recorded. Not suitable for a time in a geological context.}
#'  \item{eventTime}{The time or interval during which an Event occurred.}
#'  \item{day}{The integer day of the month on which the Event occurred.}
#'  \item{month}{The integer month in which the Event occurred.}
#'  \item{year}{The four-digit year in which the Event occurred, according to the Common Era Calendar.}
#'  \item{basisOfRecord}{The specific nature of the data record. Recommended best practice is to use the standard label of one of the Darwin Core classes.PreservedSpecimen, FossilSpecimen, LivingSpecimen, MaterialSample, Event, HumanObservation, MachineObservation, Taxon, Occurrence, MaterialCitation}
#'  \item{country}{The name of the country or major administrative unit in which the Location occurs.}
#'  \item{type}{The nature or genre of the resource. StillImage, MovingImage, Sound, PhysicalObject, Event, Text.}
#'  \item{occurrenceStatus}{A statement about the presence or absence of a Taxon at a Location. present, absent.}
#'  \item{recordNumber}{An identifier given to the Occurrence at the time it was recorded. Often serves as a link between field notes and an Occurrence record, such as a specimen collector's number.}
#'  \item{recordedBy}{A list (concatenated and separated) of names of people, groups, or organizations responsible for recording the original Occurrence. The primary collector or observer, especially one who applies a personal identifier (recordNumber), should be listed first.}
#'  \item{eventID}{An identifier for the set of information associated with an Event (something that occurs at a place and time). May be a global unique identifier or an identifier specific to the data set.}
#'  \item{Location}{A spatial region or named place.}
#'  \item{samplingProtocol}{The names of, references to, or descriptions of the methods or protocols used during an Event. Examples	UV light trap, mist net, bottom trawl, ad hoc observation | point count, Penguins from space: faecal stains reveal the location of emperor penguin colonies, https://doi.org/10.1111/j.1466-8238.2009.00467.x, Takats et al. 2001.}
#'  \item{samplingEffort}{The amount of effort expended during an Event. Examples	40 trap-nights, 10 observer-hours, 10 km by foot, 30 km by car.}
#'  \item{individualCount}{The number of individuals present at the time of the Occurrence. Integer.}
#'  \item{organismQuantity}{A number or enumeration value for the quantity of organisms. Examples	27 (organismQuantity) with individuals (organismQuantityType). 12.5 (organismQuantity) with percentage biomass (organismQuantityType). r (organismQuantity) with Braun Blanquet Scale (organismQuantityType). many (organismQuantity) with individuals (organismQuantityType).}
#'  \item{coordinatePrecision}{A decimal representation of the precision of the coordinates given in the decimalLatitude and decimalLongitude.}
#'  \item{coordinateUncertaintyInMeters}{The horizontal distance (in meters) from the given decimalLatitude and decimalLongitude describing the smallest circle containing the whole of the Location. Leave the value empty if the uncertainty is unknown, cannot be estimated, or is not applicable (because there are no coordinates). Zero is not a valid value for this term.}
#'  \item{spatiallyValid}{Occurrence records in the ALA can be filtered by using the spatially valid flag. This flag combines a set of tests applied to the record to see how reliable are its spatial data components.}
#'  \item{catalogNumber}{An identifier (preferably unique) for the record within the data set or collection.}
#'  \item{gbifID}{The identifier assigned by GBIF for each record.}
#'  \item{datasetID}{An identifier for the set of data. May be a global unique identifier or an identifier specific to a collection or institution.}
#'  \item{institutionCode}{The name (or acronym) in use by the institution having custody of the object(s) or information referred to in the record. Examples	MVZ, FMNH, CLO, UCMP.}
#'  \item{datasetName}{The name identifying the data set from which the record was derived.}
#'  \item{otherCatalogNumbers}{A list (concatenated and separated) of previous or alternate fully qualified catalog numbers or other human-used identifiers for the same Occurrence, whether in the current or any other data set or collection.}
#'  \item{occurrenceID}{An identifier for the Occurrence (as opposed to a particular digital record of the occurrence). In the absence of a persistent global unique identifier, construct one from a combination of identifiers in the record that will most closely make the occurrenceID globally unique.}
#'  \item{taxonKey}{The GBIF-assigned taxon identifier number.}
#'  \item{collectionID}{An identifier for the collection or dataset from which the record was derived.}
#'  \item{verbatim_scientificName}{The verbatim (originally-provided) scientific name}
#'  \item{verbatimEventDate}{The verbatim original representation of the date and time information for an Event.}
#'  \item{associatedTaxa}{A list (concatenated and separated) of identifiers or names of taxa and the associations of this Occurrence to each of them.}
#'  \item{associatedOrganisms}{A list (concatenated and separated) of identifiers of other Organisms and the associations of this Organism to each of them.}
#'  \item{fieldNotes}{One of a) an indicator of the existence of, b) a reference to (publication, URI), or c) the text of notes taken in the field about the Event.}
#'  \item{sex}{The sex of the biological individual(s) represented in the Occurrence.}
#'  \item{rights}{A description of the usage rights applicable to the record.}
#'  \item{rightsHolder}{A person or organization owning or managing rights over the resource.}
#'  \item{accessRights}{Information about who can access the resource or an indication of its security status.}
#'  \item{associatedReferences}{A list (concatenated and separated) of identifiers (publication, bibliographic reference, global unique identifier, URI) of literature associated with the Occurrence.}
#'  \item{bibliographicCitation}{A bibliographic reference for the resource as a statement indicating how this record should be cited (attributed) when used.}
#'  \item{references}{A related resource that is referenced, cited, or otherwise pointed to by the described resource.}
#'  \item{informationWithheld}{Additional information that exists, but that has not been shared in the given record.}
#'  \item{isDuplicateOf}{Additional information that exists, but that has not been shared in the given record.}
#'  \item{hasCoordinate}{Variable indicating presence/absence of location coordinates.}
#'  \item{hasGeospatialIssues}{Variable indicating validity of geospatial data associated with record.}
#'  \item{occurrenceYear}{Year associated with Occurrence.}
#'  \item{id}{Variable with identifying value for the Occurrenc.}
#'  \item{duplicateStatus}{Variable indicating is Occurrence is duplicate or not.}
#'  \item{associatedOccurrences}{A list (concatenated and separated) of identifiers of other Occurrence records and their associations to this Occurrence.}
#'  \item{locationRemarks}{Comments or notes about the Location.}
#'  \item{dataSource}{BeeBDC assigned source of the data. Often written when the data is formatted by a BeeBDC::xxx_readr function or similar.}
#'  \item{verbatim_scientificName}{The verbatim (originally-provided) scientific name}
#' }
#' @references This data set was created by generating a random subset of 100 rows from the full, unfiltered and unflagged, BeeBDC dataset from the publication:
#' Dorey, J.B., Fischer, E.E., Chesshire, P.R., Nava-Bolaños, A., O’Reilly, R.L., Bossert, S., Collins, S.M., Lichtenberg, E.M., Tucker, E., Smith-Pardo, A., Falcon-Brindis, A., Guevara, D.A., Ribeiro, B.R., de Pedro, D., Hung, J.K.-L., Parys, K.A., McCabe, L.M., Rogan, M.S., Minckley, R.L., Velzco, S.J.E., Griswold, T., Zarrillo, T.A., Jetz, W., Sica, Y.V., Orr, M.C., Guzman, L.M., Ascher, J., Hughes, A.C. & Cobb, N.S. (2023) A globally synthesised and flagged bee occurrence dataset and cleaning workflow. Scientific Data, 10, 1–17. https://www.doi.org/10.1038/S41597-023-02626-W
#' @keywords datasets
#' @examples
#' 
#' beesRaw <- BeeBDC::beesRaw
#' head(beesRaw)
#'
"beesRaw"

# This function was written by James B Dorey on the 5th of October 2022
# Its purpose is to create a list of directories for beeData cleaning
# Please contact jbdorey[at]me.com for help


#' Set up global directory paths and create folders
#' 
#' This function sets up a directory for saving outputs (i.e. data, figures) generated through the 
#' use of the BeeBDC package, if the required folders do not already exist.
#'
#' @param RootPath A character String. The `RootPath` is the base path for your project, and all 
#' other paths should ideally be located within the `RootPath`. However, users may specify paths not 
#' contained in the RootPath
#' @param ScriptPath A character String. The `ScriptPath` is the path to any additional functions 
#' that you would like to read in for use with BeeBDC.
#' @param DataPath A character string. The path to the folder containing bee occurrence data 
#' to be flagged and/or cleaned
#' @param DataSubPath A character String. If a `DataPath` is not provided, this will be used as the `DataPath`
#' folder name within the `RootPath.` Default is "/Data_acquisition_workflow"
#' @param DiscLifePath A character String. The path to the folder which contains data from Ascher 
#' and Pcikering's Discover Life website.
#' @param OutPath A character String. The path to the folder where output data will be saved.
#' @param OutPathName A character String. The name of the `OutPath` subfolder located within the 
#' `RootPath.` Default is "Output".
#' @param RDoc A character String. The path to the current script or report, relative to the project 
#' root. Passing an absolute path raises an error. This argument is used by [here::i_am()] and incorrectly
#' setting this may result in `bdc` figures being saved to your computer's root directory
#' @param Report Logical. If TRUE, function creates a "Report" folder within the OutPath-defined 
#' folder. Default = TRUE.
#' @param Check Logical. If TRUE, function creates a "Check" folder within the OutPath-defined 
#' folder. Default = TRUE.
#' @param Figures Logical. If TRUE, function creates a "Figures" folder within the OutPath-defined 
#' folder. Default = TRUE.
#' @param Intermediate Logical. If TRUE, function creates a "Intermediate" folder within the 
#' OutPath-defined folder in which to save intermediate datasets. Default = TRUE.
#' @param useHere Logical. If TRUE, dirMaker will use [here::i_am()] to declare the relative path 
#' to 'RDoc'. This is aimed at preserving some functionality with where bdc saves summary figures 
#' and tables. Default = TRUE.
#' 
#'
#' @return Results in the generation of a list containing the BeeBDC-required directories in your global
#'  environment. This function should be run at the start of each session. Additionally, this 
#'  function will create the BeeBDC-required folders if they do not already exist in the supplied 
#'  directory
#'
#' @importFrom here i_am here
#' @importFrom dplyr %>%
#'
#' @export
#'
#' @examples
#'   # load dplyr
#'   library(dplyr)
#' # Standard/basic usage:
#' RootPath <- tempdir()
#' dirMaker(
#' RootPath = RootPath,
#' # Input the location of the workflow script RELATIVE to the RootPath
#' RDoc = NULL,
#' useHere = FALSE) %>%
#'   # Add paths created by this function to the environment()
#'   list2env(envir = environment())  
#'
#' # Custom OutPathName provided
#'   dirMaker(
#'  RootPath = RootPath,
#'  # Set some custom OutPath info
#'  OutPath = NULL,
#'  OutPathName = "T2T_Output",
#'  # Input the location of the workflow script RELATIVE to the RootPath
#'  RDoc = NULL,
#'  useHere = FALSE) %>%
#'    # Add paths created by this function to the environment()
#'    list2env(envir = environment())  
#'  # Set the working directory
#'
#' # Further customisations are also possible
#' dirMaker(
#'   RootPath = RootPath,
#'   ScriptPath = "...path/Bee_SDM_paper/BDC_repo/BeeBDC/R",
#'   DiscLifePath = "...path/BDC_repo/DiscoverLife_Data",
#'   OutPathName = "AsianPerspective_Output",
#'   # Input the location of the workflow script RELATIVE to the RootPath
#'   RDoc = NULL,
#'   useHere = FALSE) %>%
#'   # Add paths created by this function to the environment()
#'   list2env(envir = environment())  
#' 
#' 
#'


dirMaker <- function(
    RootPath = RootPath,
    ScriptPath = NULL,
    DataPath = NULL,
    DataSubPath = "/Data_acquisition_workflow",
    DiscLifePath = NULL,
    OutPath = NULL,
    OutPathName = "Output",
    Report = TRUE,
    Check = TRUE,
    Figures = TRUE,
    Intermediate = TRUE,
    RDoc = NULL,
    useHere = TRUE
){
  # Ensure that the working directory remains the same on exit
  oldwd <- getwd()           # code line i 
  on.exit(setwd(oldwd))        # code line i+1 
  
  #### 0.0 Prep ####
  ##### 0.1 errors ####
  ###### a. FATAL errors ####
  if(is.null(RootPath)){
    stop(paste0(" - No RootPath was given. Please specifcy the root drectory that you want to use ",
                "for your data-cleaning adventures. I'll do the rest."))
  }
  if(is.null(RDoc) & useHere == TRUE){
    stop(paste0(" - Please provide a path for RDoc with useHere = TRUE.",
                " This path MUST be relative to the RootPath. ",
                "Hence, if the RootPath is '/user/home/', and the path to the RDoc is ", 
                "'/user/home/beeData/cleaningWorkflow.R', ",
                "then RDoc == 'beeData/cleaningWorkflow.R"))
  }

    # Set the working directory
  setwd(RootPath)
  
  #### 1.0 Set paths ####
    ##### 1.1 ScriptPath ####
  # Create the ScriptPath if it does not already exist
  if(is.null(ScriptPath)){
  if (!dir.exists(paste0(RootPath, "/BDC_repo/BeeBDC/R"))) {
    dir.create(paste0(RootPath, "/BDC_repo/BeeBDC/R"), recursive = TRUE)
    message(paste0(" - We created the ", 
                   paste0(RootPath, "/BDC_repo/BeeBDC/R"),
                   "file. This file needs to have the NewFunctions added to it otherise things won't",
                   " work. These can be added from our GitHub"))
  }
  # Choose the location of the script
  ScriptPath <- paste(RootPath,
                      "/BDC_repo/BeeBDC/R", sep = "")
  }else{# If user provides an alternate ScriptRoot path
    if(ScriptPath != FALSE){
    if (!dir.exists(ScriptPath)) {
      dir.create(ScriptPath, recursive = TRUE)
      message(paste0(" - We created the ", 
                     ScriptPath,
                     "file. This file needs to have the NewFunctions added to it otherise things won't",
                     " work. These can be added from our GitHub"))
    }
  # Choose the location of the script
  ScriptPath <- ScriptPath}
    }
  
    ##### 1.2 DataPath ####
    # Create the DataPath if it does not already exist
  if(is.null(DataPath)){
  if (!dir.exists(paste0(RootPath, DataSubPath))) {
    dir.create(paste0(RootPath, DataSubPath), recursive = TRUE)
    # User message
    message(paste0(" - We created the ", 
                   paste0(RootPath, DataSubPath),
                   "file. This file needs to have the occurrence data that you want to use ",
                   "added to it otherise things won't",
                   " work. Please choose this data or download it from the supp. materials of our paper"))
  }
  # Choose the location of your data
  DataPath <- paste(RootPath, DataSubPath, sep = "")
  }else{
    if(DataPath != FALSE){
    if (!dir.exists(DataPath)) {
      dir.create(DataPath, recursive = TRUE)
      # User message
      message(paste0(" - We created the ", 
                     DataPath,
                     "file. This file needs to have the occurrence data that you want to use ",
                     "added to it otherise things won't",
                     " work. Please choose this data or download it from the supp. materials of our paper"))
    }
      # Choose the location of the DataPath
      DataPath <- DataPath}
  }
  
    ##### 1.3 DiscLifePath ####
  # Create the DiscLifePath if it does not already exist
  if(is.null(DiscLifePath) ){
  if (!dir.exists(paste0(RootPath, "/BDC_repo/DiscoverLife_Data"))) {
    dir.create(paste0(RootPath, "/BDC_repo/DiscoverLife_Data"), recursive = TRUE)
    {
      dir.create(paste0(RootPath, "/BDC_repo/DiscoverLife_Data"), recursive = TRUE)
        # User message
      message(paste0(" - We created the ", 
                     paste0(RootPath, "/BDC_repo/DiscoverLife_Data"),
                     "file. This file needs to have the DiscoverLife_Data added to it otherise things won't",
                     " work. These can be added from our GitHub"))
    }
  }
  # Choose the taxonomy path
  DiscLifePath <- paste(RootPath,
                        "/BDC_repo/DiscoverLife_Data", sep = "")
  }else{
    if(DiscLifePath != FALSE){
    if (!dir.exists(DiscLifePath)) {
      dir.create(DiscLifePath, recursive = TRUE)
      {
        dir.create(DiscLifePath, recursive = TRUE)
        # User message
        message(paste0(" - We created the ", 
                       DiscLifePath,
                       "file. This file needs to have the DiscoverLife_Data added to it otherise things won't",
                       " work. These can be added from our GitHub"))
      }
    }
      # Choose the location of the DiscLifePath
      DiscLifePath <- DiscLifePath}
  }
  
  
  ##### 1.4 OutPath ####
  # Create the OutPath if it does not already exist
  if(is.null(OutPath)){
    if (!dir.exists(paste0(DataPath, "/", OutPathName))) {
      dir.create(paste0(DataPath,  "/", OutPathName), recursive = TRUE)
      {
        dir.create(paste0(DataPath, "/", OutPathName), recursive = TRUE)
        # User message
        message(paste0(" - We created the ", 
                       paste0(DataPath,  "/", OutPathName),
                       "file."))
      }
    }
    # Choose the taxonomy path
    OutPath <- paste(DataPath,  "/", OutPathName, sep = "")
  }else{
    if(OutPath != FALSE){
    if (!dir.exists(OutPath)) {
      dir.create(OutPath, recursive = TRUE)
      {
        dir.create(OutPath, recursive = TRUE)
        # User message
        message(paste0(" - We created the ", 
                       OutPath,
                       "file."))
    }}
      # Choose the location of the OutPath
      OutPath <- OutPath}
  }
  
  #### 2.0 Create paths ####
      # If these bdc folders do not exist in the chosen directory, create them
  if (!dir.exists(paste0(OutPath, "/Check")) & Check != FALSE) {
    dir.create(paste0(OutPath, "/Check"), recursive = TRUE)
  }
  if (!dir.exists(paste0(OutPath, "/Figures")) & Figures != FALSE) {
    dir.create(paste0(OutPath, "/Figures"), recursive = TRUE)
  }
  if (!dir.exists(paste0(OutPath, "/Intermediate")) & Intermediate != FALSE) {
    dir.create(paste0(OutPath, "/Intermediate"), recursive = TRUE)
  }
  if (!dir.exists(paste0(OutPath, "/Report")) & Report != FALSE) {
    dir.create(paste0(OutPath, "/Report"), recursive = TRUE)
  }
  
    ##### 2.1 Make sub-paths ####
  if(Check != FALSE){
  OutPath_Check <- paste(OutPath, "/Check", sep = "")}
  if(Figures != FALSE){
  OutPath_Figures <- paste(OutPath, "/Figures", sep = "")}
  if(Intermediate != FALSE){
  OutPath_Intermediate <- paste(OutPath, "/Intermediate", sep = "")}
  if(Report != FALSE){
  OutPath_Report <- paste(OutPath, "/Report", sep = "")}
  
  #### 3.0 Set here::here ####
    # here::here needs to know where to find all output files. This is set here.
  if(useHere == TRUE){
  here::i_am(RDoc)}
  
  #### 4.0 Output ####
  return(list(ScriptPath, DataPath, DiscLifePath, OutPath,
              OutPath_Check, OutPath_Figures, OutPath_Intermediate, OutPath_Report) %>%
           stats::setNames(c("ScriptPath", "DataPath", "DiscLifePath", "OutPath",
                      "OutPath_Check", "OutPath_Figures", "OutPath_Intermediate",
                      "OutPath_Report")))
} # END function

# This function was written by James B Dorey on the 29th of September 2022
# Its purpose is to visualise all flags for each dataSource (simplified to the text before the 
  # first underscore)
# Please contact jbdorey[at]me.com for help
#' Generate a plot summarising flagged data
#' 
#' Creates a compound bar plot that shows the proportion of records that pass or fail each flag (rows)
#' and for each data source (columns). The function can also optionally return a point map for 
#' a user-specified species when plotMap = TRUE. This function requires that your dataset has been
#' run through some filtering functions - so that is can display logical columns starting with
#' ".".
#'
#' @param data A data frame or tibble. Occurrence records as input.
#' @param flagColours A character vector. Colours in order of pass (TRUE), fail (FALSE), and NA.
#' Default = c("#127852", "#A7002D", "#BDBABB").
#' @param fileName Character. The name of the file to be saved, ending in ".pdf".
#' If saving as a different file type, change file type suffix - See `device`.
#' @param outPath A character path. The path to the directory in which the figure will be saved.
#' Default = OutPath_Figures.
#' @param width Numeric. The width of the output figure in user-defined units Default = 15.
#' @param height Numeric. The height of the output figure in user-defined units Default = 9.
#' @param units Character. The units for the figure width and height passed to [ggplot2::ggsave()] 
#' ("in", "cm", "mm", or "px"). Default = "in".
#' @param dpi Numeric. Passed to [ggplot2::ggsave()]. Plot resolution. Also accepts a string input: "retina" (320), "print" (300), or 
#' "screen" (72). Applies only to raster output types. Default = 300.
#' @param bg Character. Passed to [ggplot2::ggsave()]. Background colour. If NULL, uses the plot.background fill value from the plot theme.
#' Default = "white."
#' @param device Character. Passed to [ggplot2::ggsave()]. Device to use. Can either be a device function (e.g. png), or one of "eps", "ps", "tex" (pictex), "pdf", "jpeg", "tiff", "png", "bmp", "svg" or "wmf" (windows only).
#' Default = "pdf". If not using default, change file name suffix in fileName argument.
#' @param speciesName Optional. Character. A species name, as it occurs in the user-input nameColumn.
#' If provided, the data will be filtered to this species for the plot.
#' @param nameColumn Optional. Character. If speciesName is not NULL, enter the column to look 
#' for the species in. A User might realise that, combined with speciesName, figures can be made for
#' a variety of factors.
#' @param saveFiltered Optional. Logical. If TRUE, the filtered data will be saved to the computer 
#' as a .csv file.
#' @param plotMap Logical. If TRUE, the function will produce a point map. Tested for use with one
#' species at a time; i.e., with speciesName is not NULL.
#' @param filterColumn Optional. The flag column to display on the map. Default = .summary.
#' @param mapAlpha Optional. Numeric. The opacity for the points on the map.
#' @param xbuffer Optional. Numeric vector. A buffer in degrees of the amount to increase the
#' min and max bounds along the
#' x-axis. This may require some experimentation, keeping in mind
#' the negative and positive directionality of hemispheres. Default = c(0,0).
#' @param ybuffer Optional. Numeric vector. A buffer in degrees of the amount to increase the
#' min and max bounds along the y-axis. This may require some experimentation, keeping in mind
#' the negative and positive directionality of hemispheres. Default = c(0,0).
#' @param ptSize Optional. Numeric. The size of the points as passed to ggplot2. Default = 1.
#' @param saveTable Optional. Logical. If TRUE, the function will save the data used to produce the 
#' compound bar plot.
#' @param jitterValue Optional. Numeric. The value to jitter points by in the map in decimal degrees.
#' @param returnPlot Logical. If TRUE, return the plot to the environment. Default = FALSE.
#' @param ... Optional. Extra variables to be fed into [forcats::fct_recode()] to change names on plot.
#' For example... 'B. Mont.' = "BMont", 'B. Minkley' = "BMin", Ecd = "Ecd", Gaiarsa = "Gai"
#'
#' @return Exports a compound bar plot that summarises all flag columns. Optionally can also return 
#' a point map for a particular species in tandem with the summary plot.
#' @export
#' 
#' @importFrom dplyr across desc %>%
#'
#' @examples
#' # import data
#' data(beesFlagged)
#' OutPath_Figures <- tempdir()
#'  # Visualise all flags for each dataSource (simplified to the text before the first underscore)
#' plotFlagSummary(
#'   data = beesFlagged,
#'   # Colours in order of pass (TRUE), fail (FALSE), and NA
#'   flagColours = c("#127852", "#A7002D", "#BDBABB"),
#'   fileName = paste0("FlagsPlot_TEST_", Sys.Date(),".pdf"),
#'   outPath = OutPath_Figures,
#'   width = 15, height = 9,
#'   # OPTIONAL:
#'   #\   #  # Filter to species
#'   #\   speciesName = "Holcopasites heliopsis",
#'   #\   # column to look in
#'   #\   nameColumn = "species",
#'   #\   # Save the filtered data
#'   #\   saveFiltered = TRUE,
#'   #\   # Filter column to display on map
#'   #\   filterColumn = ".summary",
#'   #\   plotMap = TRUE,
#'   #\   # amount to jitter points if desired, e.g. 0.25 or NULL
#'   #\   jitterValue = NULL,
#'   #\   # Map opacity value for points between 0 and 1
#'   #\   mapAlpha = 1,
#'   # Extra variables can be fed into forcats::fct_recode() to change names on plot
#'   GBIF = "GBIF", SCAN = "SCAN", iDigBio = "iDigBio", USGS = "USGS", ALA = "ALA", 
#'   ASP = "ASP", CAES = "CAES", 'B. Mont.' = "BMont", 'B. Minkley' = "BMin", Ecd = "Ecd",
#'   Gaiarsa = "Gai", EPEL = "EPEL"
#' )
#' 
#' 
#' 
plotFlagSummary <- function(
    data = NULL,
    flagColours = c("#127852", "#A7002D", "#BDBABB"),
    fileName = NULL,
    outPath = OutPath_Figures,
    width = 15, height = 9, units = "in",
    dpi = 300,
    bg = "white", device = "pdf",
      # OPTIONAL:
    speciesName = NULL,
    saveFiltered = FALSE,
    filterColumn = ".summary",
    nameColumn = NULL,
    plotMap = FALSE,
    mapAlpha = 0.5,
    xbuffer = c(0,0),
    ybuffer = c(0,0),
    ptSize = 1,
    saveTable = FALSE,
    # Jitter map? enter jitter amount
    jitterValue = NULL,
    returnPlot = FALSE,
    ...
){
  # locally bind variables to the function
  OutPath_Figures <- decimalLatitude <- decimalLongitude <- . <- dataSource <- NULL
  database <- flags <- value <- count <- .data <- NULL
  
  requireNamespace("ggspatial")
  requireNamespace("dplyr")
  requireNamespace("bdc")
  requireNamespace("forcats")
  
  
  #### 0.0 Prep ####
  ##### 0.1 errors ####
  ###### a. FATAL errors ####
  if(is.null(data)){
    stop(" - Please provide an argument for data I'm a program not a magician.")
  }
  if(is.null(outPath)){
    stop(" - Please provide an argument for outPath Seems reckless to let me just guess.")
  }
  if(is.null(speciesName) & saveFiltered == TRUE){
    stop(" - saveFiltered cannot be TRUE if no speciesName is provided to filter occurrences.\n",
         "This functionality is provided to save the filtered dataset for examination.")
  }
  ###### b. warnings ####
  if(is.null(speciesName) & plotMap == TRUE){
    warning(" - plotMap is not tested with no speciesName provided to filter occurrences.\n",
            "This functionality is provided to check the filtered dataset for examination and I fear ",
            "that this might result in an intense task to run... Maybe not... Enjoy!")
  }
  if(is.null(fileName)){
    writeLines(" - No argument provided for fileName. Using default of 'FlagsPlot_DATE.pdf'")
    fileName = paste0("FlagsPlot_", Sys.Date(),".pdf")
  }
  if(is.null(filterColumn)){
    writeLines(" - No argument provided for filterColumn Using default of '.summary'")
    filterColumn = ".summary"
  }
  if(!is.null(speciesName) & is.null(nameColumn)){
    writeLines(" - nameColumn is not provided. Defaulting to scientificName.\n")
    nameColumn = "scientificName"
  }

  
  
#### 1.0 Prepare data ####
  ##### 1.1 Optional species filter ####
    # If a species name is provided then filter to ONLY that/those species
  if(!is.null(speciesName)){
    writeLines(" - Filtering to selected species...")
      ###### a. filter ####
      # Filter data
    data <- data %>%
      dplyr::filter( data[[nameColumn]] %in% speciesName)
    writeLines(paste0(" - Selected species has ",
               format(nrow(data), big.mark = ","),
               " occurrences."))
    # OPTIONAL save filtered data
    ##### b. save ####
      # If a save location is provided, then save the filtered dataset
    if(saveFiltered == TRUE){
      data %>%
        readr::write_excel_csv(paste0(outPath, "/FlagsPlot_", speciesName,".csv"))
    }
    ##### c. map data ####
      # Save a version of the data for mapping, if asked by user
    if(plotMap == TRUE){
      mapData <- data %>%
          # Select the columns to use
        dplyr::select(c(decimalLatitude, decimalLongitude,
                        tidyselect::all_of(c(nameColumn, filterColumn)))) 
          # Sort the filterColumn to have TRUE on top of FALSE
      mapData <- mapData %>%
        dplyr::select(tidyselect::all_of(filterColumn)) %>%
        dplyr::mutate(dplyr::across(1, as.character)) %>%
        dplyr::bind_cols(mapData %>% dplyr::select(!tidyselect::all_of(filterColumn))) %>% 
        dplyr::arrange(dplyr::desc(.))
    }
  } # END !is.null(speciesName)
  

  ##### 1.3 Prepare for plot ####
  writeLines(" - Preparing data to plot...")
  # Make a column with the dataSource without numbers
  data <- data %>%
    # Make a new column with the dataSource names but not the specifics
    dplyr::mutate(database = stringr::str_replace(dataSource,
                                                  pattern = "_.*",
                                                  replacement = "")) %>%
    # Group by the new column (i.e. database)
    dplyr::group_by(database) %>%
    # Select only the filter columns (starting with ".") and the database column
    dplyr::select(database, tidyselect::starts_with(".")) %>%
    #mutate(across(is.logical, ~as.numeric(.x))) %>%
    #dplyr::filter(database %in% c("ASP", "Ecd")) %>%
    # Pivot the data to a longer format for plotting in ggplot2.
    tidyr::pivot_longer(cols = tidyselect::starts_with("."),
                        names_to = "flags",
                        values_to = "value") %>%
    dplyr::group_by(database, flags, value) %>%
    dplyr::summarise(count = dplyr::n())
  
  # Make flag type
  data$flagType <- data$flags %>%
    dplyr::recode(.coordinates_empty = "Initial", .coordinates_outOfRange = "Initial", 
                  .basisOfRecords_notStandard = "Initial", .coordinates_country_inconsistent = "Initial",
                  .occurrenceAbsent = "Initial", .unLicensed = "Initial", .GBIFflags = "Initial",
                  # Taxonomy
                  .scientificName_empty = "Taxonomy",.invalidName = "Taxonomy", .uncer_terms = "Taxonomy",
                  #Space
                  .rou = "Space", .uncertaintyThreshold = "Space", .cap = "Space", .cen = "Space",
                  .equ = "Space",.gbf = "Space", .inst = "Space", .zer = "Space",.sea = "Space",
                  .val = "Space",
                  .countryOutlier = "Space", .continentOutlier = "Space", .expertOutlier = "Space",
                  .sequential = "Space", .lonFlag = "Space", .latFlag = "Space", .gridSummary = "Space", 
                  # Time
                  .eventDate_empty = "Time", .year_outOfRange = "Time", 
                  # Summary
                  .duplicates = "Summary", .summary = "Summary") %>%
    factor(levels = c("Initial",
                         "Taxonomy","Space",
                         "Time","Summary")) %>% 
    forcats::fct_relevel("Initial",
                         "Taxonomy","Space",
                         "Time","Summary")
    
  
  # You can turn the flag columns into factors and order them here
  data$flags <- data$flags %>%
    dplyr::recode_factor(.coordinates_empty = "No coordinates", 
                         .coordinates_outOfRange = "Point off map", 
                         .basisOfRecords_notStandard = "Excluded basis of record", 
                         .coordinates_country_inconsistent = "Coords. & country inconsistent",
                         .occurrenceAbsent = "Absent record", 
                         .unLicensed = "Protected by license", 
                         .GBIFflags = "GBIF flags",
                         # Taxonomy
                         .scientificName_empty = "No scientific name",
                         .invalidName = "Name didn't match", 
                         .uncer_terms = "Taxonomy qualifier",
                         #Space
                         .rou = "Coordinates rounded",
                         .uncertaintyThreshold = "High coordinate uncertainty",
                         .cap = "Capital centroid", 
                         .cen = "Country centroid",
                         .equ = "Coordinates equal",
                         .gbf = "Point on GBIF HQ", 
                         .inst = "Point on institution", 
                         .zer = "Coordinates zero",
                         .sea = "Point in sea", 
                         .val = "Coordinates invalid",
                         .countryOutlier = "Country outliers",
                         .continentOutlier = "Continent outliers",
                         .expertOutlier = "Expert outliers",
                         .sequential = "Coordinate fill-down", 
                         .lonFlag = "Gridded longitudes", .latFlag = "Gridded latitudes",
                         .gridSummary = "Gridded lat & lon", 
                         # Time
                         .eventDate_empty = "No event date", .year_outOfRange = "Year out of range", 
                         # Summary
                         .duplicates = "Duplicate", .summary = "Summary") %>%
    forcats::fct_rev() 
  
  # Factorise and order by database
  data$database <- data$database %>%
    dplyr::recode_factor(...)
  
  ##### 1.4 Save table ####
    # If user choses to save the table that makes the plot, do so here.
  if(saveTable == TRUE){
    data %>%
      readr::write_excel_csv(paste0(outPath, "/",
                              stringr::str_replace(fileName,
                                                   pattern = "\\.pdf",
                                                   replacement = ".csv")))
  }

  # Make data$value into a character type
  data$value <- as.character(data$value)
  
  # NA levels may be removed from the plot here, but I prefer to keep them.
    # data <- data %>%
    #   tidyr::drop_na()
  
#### 2.0 Plot ####
    ##### 2.1 Build plot ####
  writeLines(" - Building plot...")
  plot <-  ggplot2::ggplot(data = data) +
    # Set up the plot facets
    ggplot2::facet_grid( flagType~database, scales = "free", space= "free_y") + 
    # Make the bar plots
    ggplot2::geom_bar(ggplot2::aes(y = flags, x = count, fill = as.factor(value)), 
                      position="fill", stat='identity') +
    # Colour and label the plots
    ggplot2::scale_fill_manual(labels = c("TRUE" = "Pass",
                                 "FALSE" = "Fail",
                                 "NA" = "NA"),
                      values = c("TRUE" = flagColours[1],
                                 "FALSE" = flagColours[2],
                                 "NA" = flagColours[3])
                      ) +
    # Set up the theme stuff
    ggplot2::theme(axis.text.x = ggplot2::element_blank(), # element_text(angle = 35, vjust = 0.5, hjust=0.5),
          #axis.text.y = element_text(colour = as.character(flagCols)),
          strip.placement = "outside",
          panel.background = ggplot2::element_rect(fill = "white"),
          plot.title = ggplot2::element_text(size = 20, face = "bold"))  + 
    ggplot2::labs(x ="Proportion flagged", y = "Flag columns", fill = "Filter") 
    
  ##### 2.2 Filtered title+map+save ####
  if(!is.null(speciesName)){
    ###### a. title ####
      # Add title if filtered by species name
    plot <- plot +
      ggplot2::ggtitle(paste0("Species: ", speciesName))
    
    ##### b. map ####
    if(plotMap == TRUE){
        # Make into an ordered factor
      mapData <- mapData %>%
        dplyr::mutate(dplyr::across(tidyselect::all_of(filterColumn),
                        ~factor(.x, levels = c(FALSE, TRUE), ordered = TRUE))) %>%
        tidyr::drop_na(decimalLatitude, decimalLongitude)
      
      # Download world map using rnaturalearth packages
      WorldMap_layer <- rnaturalearth::ne_countries(scale = "medium", returnclass = "sf", 
                                     country = NULL, type="map_units") 
      # Create the checklist map
      (PointMap <- ggplot2::ggplot(data = WorldMap_layer ) +
          # CORE plotting of map and data
          # Plot and colour the terrestrial base map
          ggplot2::geom_sf(ggplot2::aes(fill = NULL), size = 0.15)+ 
          # plot point data
            # POINTS IF IS NULL; i.e. DON'T jitter
          {if(is.null(jitterValue))
            ggplot2::geom_point(data = mapData %>% dplyr::filter(.[[1]] == "FALSE"),
                     mapping = ggplot2::aes(x = decimalLongitude, y = decimalLatitude,
                                   colour = .data[[filterColumn]]),
                                   size = ptSize,
                     alpha = mapAlpha)} +
          {if(is.null(jitterValue))
            ggplot2::geom_point(data = mapData %>% dplyr::filter(.[[1]] == "TRUE"),
                       mapping = ggplot2::aes(x = decimalLongitude, y = decimalLatitude,
                                     colour = .data[[filterColumn]]),
                       size = ptSize,
                       alpha = mapAlpha)} +
        # POINTS IF IS NOT NULL; i.e. jitter
          {if(!is.null(jitterValue))ggplot2::geom_jitter(mapData %>% 
                                                           dplyr::filter(.[[1]] == "FALSE"), 
                           mapping = ggplot2::aes(x = decimalLongitude, y = decimalLatitude,
                                         colour = .data[[filterColumn]]),
                           size = ptSize,
                           alpha = mapAlpha, width = jitterValue, height = jitterValue)}+ 
          {if(!is.null(jitterValue))ggplot2::geom_jitter(mapData %>% 
                                                           dplyr::filter(.[[1]] == "TRUE"), 
                                                mapping = ggplot2::aes(x = decimalLongitude, 
                                                                       y = decimalLatitude,
                                                              colour = .data[[filterColumn]]),
                                                size = ptSize,
                                                alpha = mapAlpha, width = jitterValue, 
                                                height = jitterValue)}+ 
          ggplot2::scale_color_manual(values = c(
                                        "TRUE" = flagColours[[1]], # "#013766",
                                        "FALSE" = flagColours[[2]]), #"#ac0e28"),
                             name = "Passed occ.") +
          # Set map limits, if wanted
          ggplot2::coord_sf(expand = TRUE, 
                   ylim = c(min(mapData$decimalLatitude, na.rm = TRUE)+ybuffer[[1]], 
                            max(mapData$decimalLatitude, na.rm = TRUE)+ybuffer[[2]]),
                   xlim = c(min(mapData$decimalLongitude, na.rm = TRUE)+xbuffer[[1]], 
                            max(mapData$decimalLongitude, na.rm = TRUE)+xbuffer[[2]]),
                   lims_method = "box") + 
          # Map formatting
          # Add in the map's north arrow
          ggspatial::annotation_north_arrow(location = "tl", which_north = "true", 
                                 pad_x = unit(0.1, "cm"), pad_y = unit(0.1, "cm"), 
                                 style = ggspatial::north_arrow_fancy_orienteering()) + # Add in NORTH ARROW
          ggplot2::theme(panel.grid.major = ggplot2::element_line(color = gray(.1, alpha = 0.1), 
                                                linetype = "dashed", linewidth = 0.5), # Add grid lines
                panel.border = ggplot2::element_rect(color = gray(.1, alpha = 1), 
                                            linetype = "solid", linewidth = 0.5,
                                            fill = NA), # add panel border
                panel.background = ggplot2::element_rect(fill = "aliceblue") ,
                plot.title = ggplot2::element_text(face = "italic"))+ # Add background - colour in the ocean
          # Change map colour scheme
          ggplot2::scale_fill_viridis_d(option = "magma") + # options = "magma", "inferno", "plasma", "cividis"
          # Add in X and Y labels
          ggplot2::xlab("Longitude") + ggplot2::ylab("Latitude") + 
          # Add in the title
          ggplot2::ggtitle( speciesName) )
      # save as the map as 10*6"
      ggplot2::ggsave(paste0("/Map_FlagsPlot_", speciesName, ".pdf"), plot = PointMap, device = "pdf", 
              width = 10, height = 5, dpi = 300, path = outPath)
    }
    
    ##### c. save ####
    # Filtered save plot
    if(!is.null(speciesName)){
      ggplot2::ggsave(filename = paste0("/FlagsPlot_", speciesName,".pdf"),
                      path = outPath,
                      plot = plot, dpi = dpi, bg = bg, device = device,
                      width = width, height = height, units = units)}
  }

  # Save the figure
  ##### 2.2 Save all option ####
  if(is.null(speciesName)){
  ggplot2::ggsave(filename = fileName,
                  path = outPath,
                  plot = plot, dpi = dpi, bg = bg, device = device,
                  width = width, height = height, units = units)}
 
  if(returnPlot == TRUE){
    return(plot)}
  
} # END function

requireNamespace("dplyr")
requireNamespace("ggplot2")
requireNamespace("dplyr")

# Load in the test data
beesFlagged <- BeeBDC::beesFlagged


# Visualise all flags for each dataSource (simplified to the text before the first underscore)
testOut <- plotFlagSummary(
  data = beesFlagged,
  # Colours in order of pass (TRUE), fail (FALSE), and NA
  flagColours = c("#127852", "#A7002D", "#BDBABB"),
  fileName = paste0("FlagsPlot_", Sys.Date(),".pdf"),
  outPath = tempdir(),
  width = 15, height = 9,
  GBIF = "GBIF", SCAN = "SCAN", iDigBio = "iDigBio", USGS = "USGS", ALA = "ALA", 
  ASP = "ASP", CAES = "CAES", 'B. Mont.' = "BMont", 'B. Minkley' = "BMin", Ecd = "Ecd",
  Gaiarsa = "Gai", EPEL = "EPEL",
  returnPlot = TRUE
)



# Test classes
 testthat::test_that("plotFlagSummary expected class", {
   testthat::expect_type(testOut, "object")
 })

   testthat::test_that("dupePlotR plot saved?", {
     testthat::expect_true(inherits(testOut, c("ggplot", "ggplot2::ggplot")))
   })


# Check directory that the plot was saved
testthat::test_that("plotFlagSummary plot saved?", {
  testthat::expect_true(any(stringr::str_detect(list.files(tempdir()), "FlagsPlot_")))
})

#' Query the bee taxonomy and country checklist
#'
#' A simple function to return information about a particular species, including name validity and
#' country occurrences.
#'
#' @param beeName Character or character vector. A single or several bee species names to search for
#' in the beesTaxonomy and beesChecklist tables.
#' @param searchChecklist Logical. If TRUE (default), search the country checklist for each species.
#' @param printAllSynonyms Logical. If TRUE, all synonyms will be printed out for each entered name.
#' default = FALSE.
#' @param beesChecklist A tibble. The bee checklist file for BeeBDC. If is NULL then 
#' [BeeBDC::beesChecklist()] will be called internally to download the file. Default = NULL.
#' @param beesTaxonomy A tibble. The bee taxonomy file for BeeBDC. If is NULL then 
#' [BeeBDC::beesTaxonomy()] will be called internally to download the file. Default = NULL.
#'
#' @return Returns a list with the elements 'taxonomyReport' and 'SynonymReport'. IF searchChecklist
#' is TRUE, then 'checklistReport' will also be returned.
#' @export
#'
#' @importFrom dplyr %>%
#'
#' @examples
#'   # For the sake of these examples, we will use the example taxonomy and checklist
#'   system.file("extdata", "testTaxonomy.rda", package="BeeBDC") |> load()
#'   system.file("extdata", "testChecklist.rda", package="BeeBDC") |> load()
#' 
#'   # Single entry example
#' testQuery <- BeeBDCQuery(
#'   beeName = "Lasioglossum bicingulatum",
#'   searchChecklist = TRUE,
#'   printAllSynonyms = TRUE,
#'   beesTaxonomy = testTaxonomy,
#'   beesChecklist = testChecklist)
#' 
#'   # Multiple entry example
#' testQuery <- BeeBDC::BeeBDCQuery(
#'   beeName = c("Lasioglossum bicingulatum", "Nomada flavopicta",
#'   "Lasioglossum fijiense (Perkins and Cheesman, 1928)"),
#'   searchChecklist = TRUE,
#'   printAllSynonyms = TRUE,
#'   beesTaxonomy = testTaxonomy,
#'   beesChecklist = testChecklist)
#'   
#'     # Example way to examine a report from the output list
#'   testQuery$checklistReport
#' 
#' 
#' 
BeeBDCQuery <- function(
    beeName = NULL,
    searchChecklist = TRUE,
    printAllSynonyms = FALSE,
    beesChecklist = NULL,
    beesTaxonomy = NULL){
  # locally bind variables to the function
  data <- validName <- accid <- inputName <- id <- 
    rowMatched <- . <- NULL
  
  #### 0.0 Prep ####
  ##### 0.1 Errors ####
  ###### a. FATAL errors ####
  if(is.null(beeName)){
    stop(" - Please provide an argument for beeName. I'm a program not a magician.")
  }

  
  #### 1.0 Data preperation ####
  # Download the datasets
  if(searchChecklist == TRUE & is.null(beesChecklist)){
    beesChecklist <- BeeBDC::beesChecklist()}
  if(is.null(beesTaxonomy)){
    beesTaxonomy <- BeeBDC::beesTaxonomy()}
  
    # Change beeName to be matched exactly
  beeNameExact <- paste0("^",  beeName, "$")
  
  #### 2.0 Taxonomy check ####
  message("Starting taxonomy report...")
  ##### 2.1 Taxonomy report ####
  # Get a report of the queried names and their matched rows in beesTaxonomy 
  report_beesTaxonomy <- 
    # Make a tibble with the input name(s)
    dplyr::tibble(inputName = beeName) %>%
    dplyr::left_join(beesTaxonomy, relationship = "many-to-many", keep = TRUE,
                     by = c("inputName" = "validName")) %>%
    # Do the same and bind using canonical
    dplyr::bind_rows(
      dplyr::tibble(inputName = beeName) %>%
        dplyr::left_join(beesTaxonomy, relationship = "many-to-many", keep = TRUE,
                         by = c("inputName" = "canonical"))
    ) %>%
    # Do the same and bind using canonical_withFlags
    dplyr::bind_rows(
      dplyr::tibble(inputName = beeName) %>%
        dplyr::left_join(beesTaxonomy, relationship = "many-to-many", keep = TRUE,
                         by = c("inputName" = "canonical_withFlags"))) %>%
    # Drop rows that did not match
    tidyr::drop_na(validName) %>%
    # Keep only distinct rows
    dplyr::distinct() 
      
    # Extract accepted names
  acceptedNames <- report_beesTaxonomy %>%
    dplyr::filter(accid == 0) %>%
    dplyr::mutate(inputID = 0, .after = inputName)
    # Extract accepted names for synonyms provided
  synonyms <- report_beesTaxonomy %>%
    dplyr::filter(accid != 0) %>% 
  # Filter to the inputName and inputID
    dplyr::select(tidyselect::all_of(c("inputName", "accid"))) %>% 
    dplyr::rename("inputID" = "accid") %>%
      # Rejoin with the beesTaxonomy, but ONLY have valid names
    dplyr::left_join(beesTaxonomy, by = c("inputID" = "id"), keep = TRUE) %>%
    dplyr::distinct()
    # recombine into report_beesTaxonomy
  report_beesTaxonomy <- dplyr::bind_rows(acceptedNames, synonyms)
    
  
    ##### 2.2 Taxonomy output ####
  for(i in 1:nrow(report_beesTaxonomy)){
    writeLines(paste0(
      report_beesTaxonomy$inputName[[i]], " is ", 
      # IF Synonym
      if(report_beesTaxonomy$inputID[[i]] > 0){
        paste0("a synonym of ",
               report_beesTaxonomy$validName[[i]], " with the taxon id number ", 
               report_beesTaxonomy$id[[i]],
               ".")
      }else{ # IF accepted
        paste0("an accepted name with the taxon id number ", report_beesTaxonomy$id[[i]], ".")
      } # END else
    ))
  } # End loop
  
    ##### 2.3 printAllSynonyms ####
  # Find the synonyms for the entered species
  synonymsMatched <- beesTaxonomy %>%
    dplyr::filter(accid %in% report_beesTaxonomy$id) %>%
    dplyr::left_join(report_beesTaxonomy %>% 
                       dplyr::select(tidyselect::all_of(c("inputName", "id"))),
                     by = c("accid" = "id"), relationship = "many-to-many") %>%
    dplyr::relocate(inputName, .before = "flags") %>%
    dplyr::arrange(accid, inputName)
  
    # If printAllSynonyms is TRUE then print the synonyms
  if(printAllSynonyms == TRUE){
    for(i in 1:length(report_beesTaxonomy$inputName)){
        # Get the synonyms for the ith entry
      synonymsLoop <- synonymsMatched %>%
        dplyr::filter(accid == report_beesTaxonomy$inputID[[i]])
      if(nrow(synonymsLoop) > 0){
        # Print user output
      writeLines(paste0(
        " - '", report_beesTaxonomy$inputName[[i]], "'",
        " has the synonyms: ", paste0(synonymsLoop$validName, collapse = ", ")
      ))} # END if > 0
      if(nrow(synonymsLoop) == 0){
        synonymsLoop <- synonymsMatched %>%
          dplyr::filter(accid == report_beesTaxonomy$id[[i]])
        # Print user output
        writeLines(paste0(
          " - '", report_beesTaxonomy$inputName[[i]], "'",
          " has the synonyms: ", unique(paste0(synonymsLoop$validName, collapse = ", ") %>%
                                          stringr::str_remove(report_beesTaxonomy$inputName[[i]]))
        ))} # END if > 0
      
      
    } # END loop
  } # END printAllSynonyms
  
  
  #### 3.0 Checklist ####
    ##### 3.1 Checklist user output ####
  if(searchChecklist == TRUE){
    message("Starting checklist report...")
      # Get the relevant rows of accepted names in the beesChecklist
    checklistMatched <- beesChecklist %>%
      dplyr::filter(validName %in% unique(report_beesTaxonomy$validName)) %>% 
      dplyr::arrange(validName)
 
      for(i in 1:length(unique(checklistMatched$validName))){
          # Select the ith species
        loopSpecies <- checklistMatched %>%
          dplyr::filter(validName == checklistMatched$validName[[i]])
          # User output
        writeLines(paste0(
          " - ", checklistMatched$validName[[i]],
          " is reportedly found in: \n",
          paste0(unique(loopSpecies$rNaturalEarth_name), collapse = ", ")
        ))
      }
    
  } # END searchChecklist
  
    #### 4.0 Return reports ####
    # Make a report of the species that did not match
  failedReport <- beeName %>%
    setdiff(., report_beesTaxonomy$inputName) %>%
    dplyr::tibble(unmatchedSpecies = .)
  
    # If searchChecklist is requested, then return the output as a list with it included
  if(searchChecklist == TRUE){
    output <- dplyr::lst(report_beesTaxonomy, synonymsMatched, checklistMatched, failedReport) %>%
      stats::setNames(c("taxonomyReport", "SynonymReport", "checklistReport", "failedReport"))
    writeLines(paste0(
      "The output will be returned as a list with the elements: ",
      "'taxonomyReport', 'SynonymReport', and 'checklistReport'. \n", "These can be accessed using",
      " 'output'$taxonomyReport, 'output'$SynonymReport, 'output'$checklistReport, or ",
      "'output'$failedReport."
    ))
  }else{
    output <- dplyr::lst(report_beesTaxonomy, synonymsMatched, failedReport) %>%
      stats::setNames(c("taxonomyReport", "SynonymReport", "failedReport"))
    writeLines(paste0(
      "The output will be returned as a list with the elements: ",
      "'taxonomyReport' and 'SynonymReport'. \n", "These can be accessed using",
      " 'output'$taxonomyReport, 'output'$SynonymReport, or ",
      "'output'$failedReport."
    ))
  }
 
   # Return the matched data
  return(output)
}  # END BeeBDCQuery

requireNamespace("dplyr")
requireNamespace("lubridate")
requireNamespace("mgsub")

# Download the datasets
# load in the small test dataset in the background
system.file("extdata", "testTaxonomy.rda", package="BeeBDC") |>
  load()
# Rename the file
beesTaxonomy <- testTaxonomy
rm(testTaxonomy)

# load in the small test dataset in the background
system.file("extdata", "testChecklist.rda", package="BeeBDC") |>
  load()
# Rename the file
beesChecklist <- testChecklist
rm(testChecklist)


if(!is.null(beesTaxonomy) & !is.null(beesChecklist)){
  # Run example queries
# Single entry example
 testQuerySing <- BeeBDC::BeeBDCQuery(
   beeName = "Nomia maneei",
   searchChecklist = TRUE,
   printAllSynonyms = TRUE,
   beesTaxonomy = beesTaxonomy,
   beesChecklist = beesChecklist)
 
   # Multiple entry example
 testQueryMult <- BeeBDC::BeeBDCQuery(
   beeName = c("Nomia maneei", "Bombus hypnorum",
   "Ceratina nanula Cockerell, 1897"),
   searchChecklist = TRUE,
   printAllSynonyms = TRUE,
   beesTaxonomy = beesTaxonomy,
   beesChecklist = beesChecklist)
 
  # Simple without checklist
 testQuerySimple <- BeeBDC::BeeBDCQuery(
   beeName = c("Nomia maneei", "Bombus hypnorum",
               "Ceratina nanula Cockerell, 1897"),
   searchChecklist = FALSE,
   printAllSynonyms = TRUE,
   beesTaxonomy = beesTaxonomy,
   beesChecklist = beesChecklist)

 
 # Test class
 testthat::test_that("BeeBDCQuery expected class", {
   testthat::expect_type(testQuerySing, "list")
 })
 testthat::test_that("BeeBDCQuery expected class", {
   testthat::expect_type(testQueryMult, "list")
 })
  # Test that elements are tibbles
 testthat::test_that("BeeBDCQuery expected class", {
   testthat::expect_true(any(stringr::str_detect(attributes(testQuerySing$taxonomyReport)$class, 
                                                 "tbl_df")))
 })
 testthat::test_that("BeeBDCQuery expected class", {
   testthat::expect_true(any(stringr::str_detect(attributes(testQueryMult$taxonomyReport)$class, 
                                                 "tbl_df")))
 })
 
 # Test output lengths
 testthat::test_that("BeeBDCQuery list size", {
   testthat::expect_equal(length(testQuerySing), 4)
 })
 testthat::test_that("BeeBDCQuery list size", {
   testthat::expect_equal(length(testQueryMult), 4)
 })
 testthat::test_that("BeeBDCQuery list size", {
   testthat::expect_equal(length(testQuerySimple), 3)
 })
 
 }

# This function was written by James Dorey to remove duplicates using between one and two methods
  # This was written between the 11th of June 2022. For help, please contact James at
  # jbdorey[at]me.com

#' Identifies duplicate occurrence records
#' 
#' This function uses user-specified inputs and columns to identify duplicate occurrence records. 
#' Duplicates are identified iteratively and will be tallied up, duplicate pairs clustered, and 
#' sorted at the end of the function.
#' The function is designed to work with Darwin Core data with a database_id column, 
#' but it is also modifiable to work with other columns.
#'
#' @param data A data frame or tibble. Occurrence records as input.
#' @param path A character path to the location where the duplicateRun_ file will be saved.
#' @param duplicatedBy A character vector. Options are c("ID", "collectionInfo", "both"). "ID" 
#' columns runs through a series of ID-only columns defined by idColumns. "collectionInfo" runs 
#' through a series of columns defined by collectInfoColumns, which are checked in combination 
#' with collectionCols. "both" runs both of the above.
#' @param idColumns A character vector. The columns to be checked individually for internal 
#' duplicates. Intended for use with ID columns only.
#' @param collectionCols A character vector. The columns to be checked in combination with each 
#' of the completeness_cols.
#' @param collectInfoColumns A character vector. The columns to be checked in combinatino with 
#' all of the collectionCols columns.
#' @param completeness_cols A character vector. A set of columns that are used to order and select 
#' duplicates by. For each occurrence, this function will calculate the sum of [complete.cases()].
#' Within duplicate clusters occurrences with a greater number of the completeness_cols filled 
#' in will be kept over those with fewer.
#' @param CustomComparisonsRAW A list of character vectors. Custom comparisons - as a list of 
#' columns to iteratively compare for duplicates. These differ from the CustomComparisons in 
#' that they ignore the minimum number and character thresholds for IDs.
#' @param CustomComparisons A list of character vectors. Custom comparisons - as a list of 
#' columns to iteratively compare for duplicates. These comparisons are made after character 
#' and number thresholds are accounted for in ID columns.
#' @param sourceOrder A character vector. The order in which you want to KEEP duplicated 
#' based on the dataSource column (i.e. what order to prioritize data sources). 
#' NOTE: These dataSources are simplified to the string prior 
#' to the first "_". Hence, "GBIF_Anthophyla" becomes "GBIF."
#' @param prefixOrder A character vector. Like sourceOrder, except based on the database_id prefix,
#' rather than the dataSource. Additionally, this is only examined if prefixOrder != NULL. 
#' Default = NULL.
#' @param dontFilterThese A character vector. This should contain the flag columns to be ignored 
#' in the creation or updating of the .summary column. Passed to  [BeeBDC::summaryFun()].
#' @param characterThreshold Numeric. The complexity threshold for ID letter length. This is the
#' minimum number of characters that need to be present in ADDITION TO the numberThreshold for an
#'  ID number to be tested for duplicates. Ignored by CustomComparisonsRAW. The columns that are 
#'  checked are occurrenceID, recordId, id, catalogNumber, and otherCatalogNumbers. Default = 2.
#' @param numberThreshold Numeric. The complexity threshold for ID number length. This is the
#' minimum number of numeric characters that need to be present in ADDITION TO the 
#' characterThreshold for an ID number to be tested for duplicates. Ignored by 
#' CustomComparisonsRAW. The columns that are checked are occurrenceID, recordId, id, 
#' catalogNumber, and otherCatalogNumbers. Default = 3.
#' @param numberOnlyThreshold Numeric. As numberThreshold except the characterThreshold is ignored.
#' Default = 5.
#' @param catalogSwitch Logical. If TRUE, and the catalogNumber is empty the function will copy over
#'  the otherCatalogNumbers into catalogNumber and visa versa. Hence, the function will attempt 
#'  to matchmore catalog numbers as both of these functions can be problematic. Default = TRUE.
#'
#' @return Returns data with an additional column called .duplicates where FALSE occurrences are 
#' duplicates and TRUE occurrences are either kept duplicates or unique. Also exports a .csv to 
#' the user-specified location with information about duplicate matching. This file is used by 
#' other functions including
#' [BeeBDC::manualOutlierFindeR()] and [BeeBDC::chordDiagramR()]
#'
#' @importFrom stats complete.cases setNames
#' @importFrom dplyr n_groups lst desc %>%
#' 
#' @seealso [BeeBDC::chordDiagramR()] for creating a chord diagram to visualise linkages between
#' dataSources and [BeeBDC::dupePlotR()] to visualise the numbers and proportions of duplicates in
#' each dataSource.
#' 
#' @export
#'
#' @examples
#' beesFlagged_out <- dupeSummary(
#' data = BeeBDC::beesFlagged,
#'   # Should start with paste0(DataPath, "/Output/Report/"), instead of tempdir():
#' path = paste0(tempdir(), "/"),
#' # options are "ID","collectionInfo", or "both"
#' duplicatedBy = "collectionInfo", # I'm only running ID for the first lot because we might 
#' # recover other info later
#' # The columns to generate completeness info from
#' completeness_cols = c("decimalLatitude",  "decimalLongitude",
#'                       "scientificName", "eventDate"),
#' # idColumns = c("gbifID", "occurrenceID", "recordId","id"),
#' # The columns to ADDITIONALLY consider when finding duplicates in collectionInfo
#' collectionCols = c("decimalLatitude", "decimalLongitude", "scientificName", "eventDate", 
#'                    "recordedBy"),
#' # The columns to combine, one-by-one with the collectionCols
#' collectInfoColumns = c("catalogNumber", "otherCatalogNumbers"),
#' # Custom comparisons - as a list of columns to compare
#' # RAW custom comparisons do not use the character and number thresholds
#' CustomComparisonsRAW = dplyr::lst(c("catalogNumber", "institutionCode", "scientificName")),
#' # Other custom comparisons use the character and number thresholds
#' CustomComparisons = dplyr::lst(c("gbifID", "scientificName"),
#'                                 c("occurrenceID", "scientificName"),
#'                                 c("recordId", "scientificName"),
#'                                 c("id", "scientificName")),
#' # The order in which you want to KEEP duplicated based on data source
#' # try unique(check_time$dataSource)
#' sourceOrder = c("CAES", "Gai", "Ecd","BMont", "BMin", "EPEL", "ASP", "KP", "EcoS", "EaCO",
#'                 "FSCA", "Bal", "SMC", "Lic", "Arm",
#'                 "USGS", "ALA", "GBIF","SCAN","iDigBio"),
#' # !!!!!! BELS > GeoLocate
#' # Set the complexity threshold for id letter and number length
#' # minimum number of characters when WITH the numberThreshold
#' characterThreshold = 2,
#' # minimum number of numbers when WITH the characterThreshold
#' numberThreshold = 3,
#' # Minimum number of numbers WITHOUT any characters
#' numberOnlyThreshold = 5)
#' 
#' 
dupeSummary <- function(
    data = NULL,
    path = NULL,
    duplicatedBy = NULL,
    # The columns to generate completeness info from
    completeness_cols = NULL,
    idColumns = NULL,
    # The columns to ADDITIONALLY consider when finding duplicates in collectionInfo
    collectionCols = NULL,
      # The columns to combine, one-by-one with the collectionCols
    collectInfoColumns = NULL,
    CustomComparisonsRAW = NULL,
    # Custom comparisons - as a list of 
    CustomComparisons = NULL,
      # The order in which you want to KEEP duplicated based on data source
    sourceOrder = NULL,
    prefixOrder = NULL,
      # Columns not to filter in .summary - default is below
    dontFilterThese = c(".gridSummary", ".lonFlag", ".latFlag", ".uncer_terms",
                        ".uncertaintyThreshold", ".unLicensed"),
    # Set the complexity threshold for id letter and number length
        # minimum number of characters when WITH the numberThreshold
    characterThreshold = 2,
        # minimum number of numbers when WITH the characterThreshold
    numberThreshold = 3,
        # Minimum number of numbers WITHOUT any characters
    numberOnlyThreshold = 5,
    catalogSwitch = TRUE
){
  # locally bind variables to the function
  database_id <- dataSource <- dupColumn_s <- completeness <- .summary <- database_id_match <-
    group <- database_id_Main <- dataSourceMain <- database_id_keep <- . <- NULL
  
  # Load required packages
  requireNamespace("dplyr")
  requireNamespace("lubridate")
  requireNamespace("igraph")
    # Record start time
  startTime <- Sys.time()
  
  #### 0.0 Prep ####
  ##### 0.1 Errors ####
  ###### a. FATAL errors ####
  if(is.null(data)){
    stop(" - Please provide an argument for data. I'm a program not a magician.")
  }
  if(is.null(sourceOrder)){
    stop(paste("Warning message: \n",
               " - No sourceOrder provided. This must be provided as the name of each dataSource ",
               "before the first '_' in the desired order.",
               sep=""))
  }
  ###### b. Warnings ####
  if(is.null(duplicatedBy)){
    message(paste("Warning message: \n",
          " - No duplicatedBy provided. Consider if you want to choose to find duplicates by (i)",
          " 'ID' columns only (for pre-cleaned data), by (ii) 'collectionInfo' columns only ",
          "(for cleaned data), or (ii) 'both'.\n",
          "NULL is acceptable, if quickDeDuplicate == TRUE",
          sep = ""))
  }
  if(is.null(idColumns) & stringr::str_detect(duplicatedBy, "ID|both")){
    message(paste("Warning message: \n",
                  " - No idColumns provided. Using default of: ",
                  "c('gbifID',  'occurrenceID', 'recordId', and 'id')",
                  sep=""))
    idColumns = c("gbifID", "occurrenceID", "recordId","id")
  }
  if(is.null(completeness_cols)){
    message(paste("Warning message: \n",
                  " - No completeness_cols provided. Using default of: ",
                  "c('decimalLatitude',  'decimalLongitude', 'scientificName', and 'eventDate')",
                  sep=""))
    completeness_cols = c("decimalLatitude",  "decimalLongitude",
                          "scientificName", "eventDate")
  }
  if(is.null(collectionCols)){
    message(paste("Warning message: \n",
                  " - No collectionCols provided. Using default of: ",
      "c('decimalLatitude',  'decimalLongitude', 'scientificName', 'eventDate', and 'recordedBy')",
                  sep=""))
    collectionCols = c("decimalLatitude", "decimalLongitude", "scientificName", "eventDate", 
                       "recordedBy")
  }
  if(is.null(collectInfoColumns)){
    message(paste("Warning message: \n",
                  " - No collectInfoColumns provided. Using default of: ",
      "c('recordNumber',  'eventID', 'catalogNumber', 'otherCatalogNumbers', and 'collectionID')",
                  sep=""))
    collectInfoColumns = c("recordNumber", "eventID", "catalogNumber", "otherCatalogNumbers",
                           "collectionID")
  }
  

  ##### 0.2 Data prep #####
    ###### a. completeness ####
  # Get the sum of the complete.cases of four important fields. Preference will be given to keeping 
  # the most-complete records
  writeLines(paste(
    " - Generating a basic completeness summary from the ", 
    paste(completeness_cols, collapse = ", "), " columns.","\n",
    "This summary is simply the sum of complete.cases in each column. It ranges from zero to the N",
    " of columns. This will be used to sort duplicate rows and select the most-complete rows.",
    sep = ""
  ))
  
  
  # Update the .summary column, ignoring the dontFilterThese columns.
  writeLines(" - Updating the .summary column to sort by...")
  data <- summaryFun(
    data = data,
    # Don't filter these columns (or NULL)
    dontFilterThese = dontFilterThese,
    # Remove the filtering columns?
    removeFilterColumns = FALSE,
    # Filter to ONLY cleaned data?
    filterClean = FALSE)
  
  # Create the completeness column based on the completeness_cols
  Loop_data <- data %>%
  dplyr::mutate(completeness = data %>% 
                     dplyr::select(tidyselect::all_of(completeness_cols)) %>% 
                     apply(., MARGIN = 1, function(x) sum(stats::complete.cases(x)))
  )
  
    ###### b. catalogSwitch ####
    # If the catalogNumber is empty, copy over the otherCatalogNumbers value and visa versa
  if(catalogSwitch == TRUE){
  Loop_data <- Loop_data %>%
    dplyr::mutate(
      otherCatalogNumbers = dplyr::if_else(is.na(otherCatalogNumbers),
                                           catalogNumber,
                                           otherCatalogNumbers),
      catalogNumber = dplyr::if_else(is.na(catalogNumber),
                                     otherCatalogNumbers,
                                     catalogNumber)
    )}
  
  ##### 0.3 format dateIdentified ####
    # Removed for now because dateIdentified is too poorly filled out.
  # Format the dateIdentified column
    #    writeLines(paste(
    #      " - Formatting the dateIdentified column to date format...", 
    #      sep = ""
    #    ))
    #    Loop_data$dateIdentified <- lubridate::ymd_hms(Loop_data$dateIdentified,
    #                                              truncated = 5, quiet = TRUE) %>%
    #      as.Date()
  
  # Add the dupColumn_s as NA for the first iteration
  Loop_data$dupColumn_s <- NA

    # Create a datset to put duplicates into
    runningDuplicates = dplyr::tibble()

    
    
    #### 1.0 CUSTOM_RAW ####
    if(!is.null(CustomComparisonsRAW)){
      message(" - Working on CustomComparisonsRAW duplicates...")
      # Get complete cases of CustomComparisonsRAW from each dataset
      
      ##### 1.1 Loop ####
      # Create a dataset to put unique vaules into
      for(i in 1:length(CustomComparisonsRAW)){
        # Select the ith CustomComparisonsRAW to match with
        currentColumn <- CustomComparisonsRAW[[i]]
        
        ###### a. Identify duplicates ####
        # Do the duplicate matching
        dupSummary <- Loop_data %>%
          # Select the columns to keep
          dplyr::select(tidyselect::all_of(c(currentColumn, "database_id",
                                             "dataSource", "dupColumn_s", "completeness",".summary"))
                        ) %>%
          # Drop any NA rows
          tidyr::drop_na( tidyselect::all_of(c(currentColumn))) %>%
          # Select the grouping (duplicate) columns
          dplyr::group_by( 
            dplyr::across(tidyselect::all_of(c(currentColumn)))) %>%
          # Select groups with more than one occurrence (duplicates)
          dplyr::filter(dplyr::n() > 1) %>%
          # Create a new column with the first occurrence in each group matching to the rest
          # Do the same for database_id
          dplyr::mutate(database_id_match = database_id[1], .after = database_id) %>%
          # Add the matching column as a column
          dplyr::mutate(
            dupColumn_s = stringr::str_c(
              dplyr::if_else(stats::complete.cases( dupColumn_s),
                             stringr::str_c(dupColumn_s, 
                                            paste0(currentColumn, collapse = ", "), sep = " & "), 
                             paste0(currentColumn, collapse = ", "))))
        
        
        # Get numbers just for an output text
        duplicates2record <- dupSummary %>%
          dplyr::filter( dplyr::row_number() > 1) %>%
          nrow()
        keptCount = dupSummary %>%
          dplyr::n_groups()
        
        ##### b. Running outputs ####
        # Bind the rows to a running file. Missing columns will be "NA"
        runningDuplicates = dplyr::bind_rows(runningDuplicates,
                                             dupSummary) %>% 
          dplyr::distinct(database_id, database_id_match, .keep_all = TRUE)
        
        ##### c. User output ####
        message(paste0(
          "\nCompleted iteration ", i, " of ", length(CustomComparisonsRAW), ":"
        ))
        writeLines(
          paste0(" - Identified ", 
                 format(duplicates2record, big.mark = ","), 
                 " duplicate records and kept ",
                 format(keptCount, big.mark = ","),
                 " unique records using the column(s): \n",
                 paste(currentColumn, collapse = ", ")))
        # Remove this temporary dataset
        rm(dupSummary)
      } # END for 1.1 CUSTOM_RAW LOOP
    } # END for 1.0 CUSTOM_RAW

    #### 2.0 Remove simple codes ####
      ##### 2.1 Code removal ####
      # Remove the too-simple codes after make the RAW comparisons
    Loop_data <- Loop_data %>%
      dplyr::mutate(
        if("occurrenceID" %in% colnames(Loop_data)){
          occurrenceID = dplyr::if_else( 
            stringr::str_count(occurrenceID, "[A-Za-z]") >= characterThreshold &
              stringr::str_count(occurrenceID, "[0-9]") >= numberThreshold |
              stringr::str_count(occurrenceID, "[0-9]") >= numberOnlyThreshold, 
                                         occurrenceID, NA_character_)},
        if("recordId" %in% colnames(Loop_data)){
          recordId = dplyr::if_else( 
            stringr::str_count(recordId, "[A-Za-z]") >= characterThreshold &
              stringr::str_count(recordId, "[0-9]") >= numberThreshold |
              stringr::str_count(recordId, "[0-9]") >= numberOnlyThreshold, 
                                     recordId, NA_character_)},
        if("id" %in% colnames(Loop_data)){
          id = dplyr::if_else( stringr::str_count(id, "[A-Za-z]") >= characterThreshold &
                                 stringr::str_count(id, "[0-9]") >= numberThreshold |
                                 stringr::str_count(id, "[0-9]") >= numberOnlyThreshold, 
                               id, NA_character_)}, 
        if("catalogNumber" %in% colnames(Loop_data)){
          catalogNumber = dplyr::if_else( 
            stringr::str_count(catalogNumber, "[A-Za-z]") >= characterThreshold &
              stringr::str_count(catalogNumber, "[0-9]") >= numberThreshold |
              stringr::str_count(catalogNumber, "[0-9]") >= numberOnlyThreshold, 
                                          catalogNumber, NA_character_)}, 
        if("otherCatalogNumbers" %in% colnames(Loop_data)){
          otherCatalogNumbers = dplyr::if_else( 
            stringr::str_count(otherCatalogNumbers, "[A-Za-z]") >= characterThreshold &
              stringr::str_count(otherCatalogNumbers, "[0-9]") >= numberThreshold |
              stringr::str_count(otherCatalogNumbers, "[0-9]") >= numberOnlyThreshold, 
                                                otherCatalogNumbers, NA_character_)})
    ##### 2.2. catalogSwitch ####
    # If the catalogNumber is empty, copy over the otherCatalogNumbers value and visa versa
    if(catalogSwitch == TRUE){
      Loop_data <- Loop_data %>%
        dplyr::mutate(
          otherCatalogNumbers = dplyr::if_else(is.na(otherCatalogNumbers),
                                               catalogNumber,
                                               otherCatalogNumbers),
          catalogNumber = dplyr::if_else(is.na(catalogNumber),
                                         otherCatalogNumbers,
                                         catalogNumber)
        )}
    
    
    #### 3.0 CUSTOM ####
    if(!is.null(CustomComparisons)){
      message(" - Working on CustomComparisons duplicates...")
      # Get complete cases of CustomComparisons from each dataset
      
      ##### 3.1 Loop ####
      # Create a dataset to put unique vaules into
      for(i in 1:length(CustomComparisons)){
        # Select the ith CustomComparisons to match with
        currentColumn <- CustomComparisons[[i]]
        
        ###### a. Identify duplicates ####
        # Do the duplicate matching
        dupSummary <- Loop_data %>%
          # Select the columns to keep
          dplyr::select(database_id, 
                        tidyselect::all_of(currentColumn),
                        dataSource, dupColumn_s, completeness,.summary) %>%
          # Drop any NA rows
          tidyr::drop_na( tidyselect::all_of(c(currentColumn))) %>%
          # Select the grouping (duplicate) columns
          dplyr::group_by( 
            dplyr::across(tidyselect::all_of(c(currentColumn)))) %>%
          # Select groups with more than one occurrence (duplicates)
          dplyr::filter(dplyr::n() > 1) %>%
          # Create a new column with the first occurrence in each group matching to the rest
          # Do the same for database_id
          dplyr::mutate(database_id_match = database_id[1], .after = database_id) %>%
          # Add the matching column as a column
          dplyr::mutate(
            dupColumn_s = stringr::str_c(
              dplyr::if_else(stats::complete.cases( dupColumn_s),
                             stringr::str_c(dupColumn_s, 
                                            paste0(currentColumn, collapse = ", "), sep = " & "), 
                             paste0(currentColumn, collapse = ", "))))
        
        
        # Get numbers just for an output text
        duplicates2record <- dupSummary %>%
          dplyr::filter( dplyr::row_number() > 1) %>%
          nrow()
        keptCount = dupSummary %>%
          dplyr::n_groups()
        
        ##### b. Running outputs ####
        # Bind the rows to a running file. Missing columns will be "NA"
        runningDuplicates = dplyr::bind_rows(runningDuplicates,
                                             dupSummary) %>% 
          dplyr::distinct(database_id, database_id_match, .keep_all = TRUE)
        
        ##### c. User output ####
        message(paste0(
          "\nCompleted iteration ", i, " of ", length(CustomComparisons), ":"
        ))
        writeLines(
          paste0(" - Identified ", 
                 format(duplicates2record, big.mark = ","), 
                 " duplicate records and kept ",
                 format(keptCount, big.mark = ","),
                 " unique records using the column(s): \n",
                 paste(currentColumn, collapse = ", ")))
        # Remove this temporary dataset
        rm(dupSummary)
      } # END for 3.1 CUSTOM LOOP
    } # END for 3.0 CUSTOM
    
    
  #### 4.0 ID ####
  if(duplicatedBy %in% c("ID","both")){
    message(" - Working on ID duplicates...")
    # Get complete cases of collectionInfo from each dataset
    
    ##### 4.1 Loop ####
    # Create a dataset to put unique values into
    for(i in 1:length(idColumns)){
      # Select the ith idColumns to match with
      currentColumn <- idColumns[i]
      
      ###### a. Identify duplicates ####
      # Do the duplicate matching
      dupSummary <- Loop_data %>%
        # Select the columns to keep
        dplyr::select(database_id, 
                      tidyselect::all_of(currentColumn),
                      dataSource, dupColumn_s, completeness,.summary) %>%
        # Drop any NA rows
        tidyr::drop_na(tidyselect::all_of(currentColumn)) %>%
        # Select the grouping (duplicate) columns
        dplyr::group_by( dplyr::across(dplyr::all_of(currentColumn))) %>%
        # Select groups with more than one occurrence (duplicates)
        dplyr::filter(dplyr::n() > 1) %>%
        # Create a new column with the first occurrence in each group matching to the rest to keep
        # Do the same for database_id
        dplyr::mutate(database_id_match = database_id[1], .after = database_id) %>%
        # Add the matching column as a column
        dplyr::mutate(
          dupColumn_s = stringr::str_c(
            dplyr::if_else(!is.na(dupColumn_s) ,
                           stringr::str_c(dupColumn_s, currentColumn, sep = " & "), 
                           currentColumn)))
      
      # Get numbers just for an output text
      duplicates2record <- dupSummary %>%
        dplyr::filter( dplyr::row_number() > 1) %>%
        nrow()
      keptCount = dupSummary %>%
        dplyr::n_groups()
      
      
      ##### b. Running outputs ####
      # Bind the rows to a running file. Missing columns will be "NA"
      runningDuplicates = dplyr::bind_rows(runningDuplicates,
                                           dupSummary) %>% 
        dplyr::distinct(database_id, database_id_match, .keep_all = TRUE)

      
      ##### c. User output ####
      message(paste0(
        "\nCompleted iteration ", i, " of ", length(idColumns), ":"
      ))
      writeLines(
        paste(" - Identified ", 
              format(duplicates2record, big.mark = ","), 
              " duplicate records and kept ",
              format(keptCount, big.mark = ","),
              " unique records using the column: \n",
              currentColumn), sep = "")
      # Remove this temporary dataset
      rm(dupSummary)
      
    } # END for LOOP
  } # END 4.0 ID
  
  
  
  #### 5.0 collectionInfo ####
  if(duplicatedBy %in% c("collectionInfo","both")){
    message(" - Working on collectionInfo duplicates...")
    # Get complete cases of collectionInfo from each dataset

          ##### 5.1 Loop ####
      # Create a dataset to put unique values into
    for(i in 1:length(collectInfoColumns)){
        # Select the ith collectInfoColumns to match with
      currentColumn <- collectInfoColumns[i]
      
      ###### a. Identify duplicates ####
        # Do the duplicate matching
    dupSummary <- Loop_data %>%
        # Select the columns to keep
      dplyr::select(database_id, 
                    tidyselect::all_of(collectionCols),
                    tidyselect::all_of(currentColumn),
                    dataSource, dupColumn_s, completeness, .summary) %>%
       # Drop any NA rows
      tidyr::drop_na( tidyselect::all_of(c(collectionCols, currentColumn))) %>%
        # Select the grouping (duplicate) columns
      dplyr::group_by( 
        dplyr::across(tidyselect::all_of(c(collectionCols, currentColumn)))) %>%
        # Select groups with more than one occurrence (duplicates)
      dplyr::filter(dplyr::n() > 1) %>%
        # Create a new column with the first occurrence in each group matching to the rest
        # Do the same for database_id
      dplyr::mutate(database_id_match = database_id[1], .after = database_id) %>%
      # Add the matching column as a column
      dplyr::mutate(
        dupColumn_s = stringr::str_c(
          dplyr::if_else(!is.na(dupColumn_s) ,
                         stringr::str_c(dupColumn_s, 
                                        paste0(currentColumn, collapse = ", "), sep = " & "), 
                         paste0(currentColumn, collapse = ", "))))

    
    # Get numbers just for an output text
    duplicates2record <- dupSummary %>%
      dplyr::filter( dplyr::row_number() > 1) %>%
      nrow()
    keptCount = dupSummary %>%
      dplyr::n_groups()
    
    
    ##### b. Running outputs ####
    # Bind the rows to a running file. Missing columns will be "NA"
    runningDuplicates = dplyr::bind_rows(runningDuplicates,
                                         dupSummary)%>% 
      dplyr::distinct(database_id, database_id_match, .keep_all = TRUE)


        ##### c. User output ####
    message(paste0(
      "\nCompleted iteration ", i, " of ", length(collectInfoColumns), ":"
    ))
    writeLines(
      paste0(" - Identified ", 
             format(duplicates2record, big.mark = ","), 
             " duplicate records and kept ",
             format(keptCount, big.mark = ","),
             " unique records using the columns: \n",
             paste(c(collectionCols), collapse = ", "), ", and ",
             currentColumn))
    # Remove this temporary dataset
    rm(dupSummary)
    
    } # END for 5.1 collectionInfo LOOP
    
  } # END for 5.0 collectionInfo
    

    #### 6.0 runningDuplicates File ####
    ##### 6.1 Clustering duplicates####
    writeLines(" - Clustering duplicate pairs...")
      # Cluster the id pairs into groups
    clusteredDuplicates <- runningDuplicates %>% 
      dplyr::select(database_id_match, database_id) %>% 
      igraph::graph_from_data_frame() %>%
      igraph::components()
      # Extract the id and the group only
    clusteredDuplicates <- clusteredDuplicates$membership %>% as.data.frame() %>%
      stats::setNames("group") %>%
      dplyr::mutate(database_id = rownames(.)) %>% dplyr::as_tibble()
      
      # Re-merge the relevant columns
    clusteredDuplicates <- clusteredDuplicates %>%
        # Re-merge the dupColumn_s column
      dplyr::left_join(runningDuplicates %>% 
                         dplyr::select(database_id, dupColumn_s) %>%
                         dplyr::distinct(database_id, .keep_all = TRUE),
                       by = "database_id") %>%
        # Re-merge the rest of the information
      dplyr::left_join(Loop_data %>%
                         dplyr::select( 
                           tidyselect::any_of(unique(c("database_id",
                                                       completeness_cols,
                                                       collectionCols,
                                                       collectInfoColumns,
                                                       dplyr::lst(CustomComparisons) %>% 
                                                         unlist() %>% as.character(),
                                                       "dataSource", "completeness",
                                                       ".summary"),
                                                     fromLast = TRUE,
                                                     na.rm = TRUE))),
                       by = "database_id") %>%
        # Group by the clustered group number
      dplyr::group_by(group) 
    
    # User output
    writeLines(paste0(
      "Duplicate pairs clustered. There are ", 
      format(nrow(clusteredDuplicates) - clusteredDuplicates %>% dplyr::n_groups(), 
             big.mark = ","), " duplicates across ", 
      format(clusteredDuplicates %>% dplyr::n_groups(), big.mark = ","),
      " kept duplicates."))
    
    ##### 6.2 Arrange data ####
    # Prepare data order
    if(!is.null(prefixOrder)){
      writeLines(" - Ordering prefixs...")
      prefixOrder = prefixOrder
      clusteredDuplicates <- clusteredDuplicates %>%
        # Make a new column with the database_id SOURCE, not the full database_id with numbers
        dplyr::mutate(database_id_Main = stringr::str_replace(database_id,
                                                              pattern = "_.*",
                                                              replacement = "") %>%
                        factor(levels = prefixOrder, ordered = TRUE) ) %>%
        # Sort so that certain datasets will be given preference over one another as user-defined.
        dplyr::arrange(database_id_Main) %>%
        # Remove this sorting column
        dplyr::select(!database_id_Main)
    }
    
    
    writeLines(paste0(" - Ordering data by 1. dataSource, 2. completeness",
               " and 3. .summary column..."))
    clusteredDuplicates <- clusteredDuplicates %>%
      # Extract only the actual source, not the taxonomic level
      dplyr::mutate(dataSourceMain = stringr::str_replace(dataSource,
                                                          pattern = "_.*",
                                                          replacement = "") %>%
                      factor(levels = sourceOrder, ordered = TRUE) ) %>%
      # Sort so that certain datasets will be given preference over one another as user-defined.
      dplyr::arrange(dataSourceMain) %>%
      # Sort so that higher completeness is given FIRST preference
      dplyr::arrange( dplyr::desc(completeness)) %>%
      # Sort by .summary so that TRUE is selected over FALSE
      dplyr::arrange( dplyr::desc(.summary)) %>%
      # Remove these sorting columns
      dplyr::select(!c(dataSourceMain)) 
    
    ##### 6.3 Keep first #####
    writeLines(paste0(" - Find and FIRST duplicate to keep and assign other associated", 
                      " duplicates to that one (i.e., across multiple tests a 'kept duplicate', ",
                      "could otherwise be removed)..."))
    # Find the first duplicate and assign the match to that one as the kept dupicate
    clusteredDuplicates <- clusteredDuplicates %>%
      dplyr::mutate(database_id_keep = database_id[1], .after = database_id) %>%
      dplyr::mutate(dataSource_keep = dataSource[1], .after = dataSource) %>%
      # Remove the row for the kept duplicate
      dplyr::filter(!database_id_keep == database_id)
    
       ##### 6.4 Save ####
      # Save the running 
    readr::write_excel_csv(clusteredDuplicates, 
                     file = paste0(path, 
                                  "/duplicateRun_", paste(duplicatedBy, collapse = "_"),
                                  "_", Sys.Date(),
                                  ".csv") %>% 
                       stringr::str_replace_all("//duplicateRun_", "/duplicateRun_"))
  writeLines(paste0(
    " - Duplicates have been saved in the file and location: ", 
    paste0(path, 
           "duplicateRun_", paste(duplicatedBy, collapse = "_"),
           "_", Sys.Date(),
           ".csv")
  ))
    
    #### 7.0 Flag .duplicates ####  
      # Add a flag to any database_id that occurs in the clusteredDuplicates file. The rest will be 
        # TRUE (not duplicates)
  Loop_data <- data %>%
      # Add .duplicates flag column
    dplyr::mutate(.duplicates = !database_id %in% clusteredDuplicates$database_id) %>%
      # Add in a column to show the duplicate status of each occurrence
    dplyr::mutate(duplicateStatus = dplyr::if_else(
      database_id %in% clusteredDuplicates$database_id,
      "Duplicate", 
      dplyr::if_else(
        database_id %in% clusteredDuplicates$database_id_keep,
        "Kept duplicate", "Unique"))
      )
  
  
    #### Final output ####
 
    writeLines(paste0(
      " - Across the entire dataset, there are now ",
      format(sum(Loop_data$.duplicates == FALSE), big.mark = ","), " duplicates from a total of ",
      format(nrow(Loop_data), big.mark = ","), " occurrences."
    ))

  endTime <- Sys.time()
  message(paste(
    " - Completed in ", 
    round(difftime(endTime, startTime), digits = 2 ),
    " ",
    units(round(endTime - startTime, digits = 2)),
    sep = ""))
    # Return data
  return(Loop_data)
} # END function

requireNamespace("dplyr")


testData <- dplyr::tibble(
  database_id = c("fakeID1","fakeID2", "fakeID2", "fakeID4","fakeID5", "fakeID6", "fakeID7",
                  "fakeID8","fakeID9", "fakeID10","fakeID11","fakeID12"),
  decimalLatitude = c(1.111, 2.111, 3.111, 4.111, 5.111, 6.111, 6.111, 6.111, 7.111, 7.111, 7.111, 7.111),
  decimalLongitude = c(10.111, 11.111, 12.111, 13.111, 14.111, 15.111, 15.111, 16.111, 16.111, 16.111, 17.111, 18.111),
  scientificName = c("a", "b", "c", "d", "e", "f", "f", "f", "g", "g", "h", "g"),      
  eventDate = c("1988-10-13 00:00:00 UTC", "1930-06-06 00:00:00 UTC", "1994-01-01 00:00:00 UTC",
                "1987-08-06 00:00:00 UTC", "1995-03-31 00:00:00 UTC", "2000-05-01 00:00:00 UTC",
                "1998-01-01 00:00:00 UTC", "1994-01-01 00:00:00 UTC","1994-01-01 00:00:00 UTC",
                 NA,"1994-01-01 00:00:00 UTC",
                "1972-04-28 00:00:00 UTC"),
  recordedBy = c("bee-er1, bee-er2", "bee-er3", "bee-er4", "bee-er5", "bee-er6","bee-er6",
                 "bee-er6", "bee-er6", "bee-er6", "bee-er7", "bee-er1", "bee-er2"),
  catalogNumber = c("beeCode0001","beeCode0002","beeCode0003","beeCode0004","beeCode0006","beeCode0006",
                    "beeCode0006","beeCode0006","beeCode0006","be01","boe6","beeCode0006"),
  otherCatalogNumbers = c("otherBeeCode0001","otherBeeCode0002","otherBeeCode0003","otherBeeCode0004",
                          "otherBeeCode0006","otherBeeCode0006",
                          "otherBeeCode0007","otherBeeCode0007","otherBeeCode0001","otherBeeCode0045",
                          "otherBeeCode0678","otherBeeCode194"),
  institutionCode = c("beeHouse","beeHouse","beeHouse","beeHouse","beeHouse","waspHouse","waspHouse",
                      "waspHouse","waspHouse","beeHouse","beeHouse","beeHouse"),     
  gbifID = c("asdflgjkh11","asdflgjkh11",NA_character_,NA_character_,NA_character_,NA_character_,
             NA_character_,NA_character_,NA_character_,NA_character_,NA_character_,NA_character_),
  occurrenceID = c(NA_character_,NA_character_,NA_character_,NA_character_,NA_character_,NA_character_,
                   NA_character_,NA_character_,"duplicatecode12345","duplicatecode12345",NA_character_,NA_character_),
  recordId = c(NA_character_,NA_character_,NA_character_,NA_character_,NA_character_,NA_character_,
               NA_character_,NA_character_,NA_character_,NA_character_,NA_character_,NA_character_),
  id = c("a","b","c","d","e","f","g","h","i","j","k","l"),
  dataSource = c("GBIF","GBIF","GBIF","GBIF","GBIF","GBIF","GBIF","GBIF","GBIF","GBIF","GBIF","GBIF")
)


check_time <- BeeBDC::dupeSummary(
  data = testData,
  path = tempdir(),
  # options are "ID","collectionInfo", or "both"
  duplicatedBy = "collectionInfo", 
  # The columns to generate completeness info from (and to sort by completness)
  completeness_cols = c("decimalLatitude",  "decimalLongitude",
                        "scientificName", "eventDate"),
  # idColumns = c("gbifID", "occurrenceID", "recordId","id"),
  # The columns to ADDITIONALLY consider when finding duplicates in collectionInfo
  collectionCols = c("decimalLatitude", "decimalLongitude", "scientificName", "eventDate", 
                     "recordedBy"),
  # The columns to combine, one-by-one with the collectionCols
  collectInfoColumns = c("catalogNumber", "otherCatalogNumbers"),
  # Custom comparisons — as a list of columns to compare
  # RAW custom comparisons do not use the character and number thresholds
  CustomComparisonsRAW = dplyr::lst(c("catalogNumber", "institutionCode", "scientificName")),
  # Other custom comparisons use the character and number thresholds
  CustomComparisons = dplyr::lst(c("gbifID", "scientificName"),
                                  c("occurrenceID", "scientificName"),
                                  c("recordId", "scientificName"),
                                  c("id", "scientificName")),
  # The order in which you want to KEEP duplicated based on data source
  # try unique(check_time$dataSource)
  sourceOrder = c("CAES", "Gai", "Ecd","BMont", "BMin", "EPEL", "ASP", "KP", "EcoS", "EaCO",
                  "FSCA", "Bal", "SMC", "Lic", "Arm",
                  "USGS", "ALA", "GBIF","SCAN","iDigBio"),
  # Prefix ordering is done using the database_id prefix, not the dataSource prefix.
  prefixOrder = c("Paige", "Dorey"),
  characterThreshold = 2,
  numberThreshold = 3,
  numberOnlyThreshold = 5
)

  # Get counts of the number of expected kept duplicates and duplicats
test_keptDuplicates <-  sum(check_time$duplicateStatus == "Kept duplicate")
test_duplicates <- sum(check_time$duplicateStatus == "Duplicate")

# Test duplicate numbers
testthat::test_that("dupeSummary kept duplicates", {
  testthat::expect_equal(test_keptDuplicates, 2)
})
testthat::test_that("dupeSummary duplicates", {
  testthat::expect_equal(test_duplicates, 3)
})
# Test concordance with TRUE/FALSE and number of duplicates
testthat::test_that("dupeSummary duplicates", {
  testthat::expect_equal(test_duplicates, sum(check_time$.duplicates == FALSE))
})

# Test class
testthat::test_that("dupeSummary expected class - simple", {
  testthat::expect_type(check_time, "list")
})
testthat::test_that("dupeSummary expected class - complex", {
  testthat::expect_true(any(stringr::str_detect(attributes(check_time)$class, "tbl_df|tbl")))
})

#### importOccurrences ####
#' Imports the most-recent repoMerge data
#' 
#' Looks for and imports the most-recent version of the occurrence data created by the [BeeBDC::repoMerge()] 
#' function.
#'
#' @param path A directory as a character. The directory to recursively look in for the above data.
#' @param fileName Character. A String of text to look for the most-recent dataset. 
#' Default = "^BeeData_". Find faults by modifying [BeeBDC::fileFinder()]
#' and logic-checking the file that's found.
#'
#' @return A list with a data frame of merged occurrence records, "Data_WebDL", and a list of EML 
#' files contained in "eml_files". 
#' @export
#' 
#' @importFrom dplyr %>%
#'
#' @examples
#' \dontrun{
#' DataImp <- importOccurrences(path = DataPath)
#' }
importOccurrences <- function(path = path,
           fileName = "^BeeData_" #occurrence file name. If not provided, R will search to match "BeeData_"
           ){ #spatial reference system as epsg code
  . <- NULL
  # Load required packages
  requireNamespace("dplyr")
  requireNamespace("lubridate")
  
  # if the fileName is not provided...
  if(!exists("fileName")){
    fileName = "^BeeData"
  }
  
  #### Find files ####
  # Find all of the previously-produced data files
  most_recent <- BeeBDC::fileFinder(path = path, fileName = fileName)
  
  # Return information to user
  writeLines(paste(" - Great, R has detected file(s), including... ", "\n",
                   paste(most_recent, collapse = "\n") ), sep = "")
  
  #### Detect format ####
    # Find the format of the most-recent files. This could potentially be .csv or .rds
    # TRUE IF .rds are present (that are not attribute files):
  (rdata_query <- any(most_recent[stringr::str_count(most_recent, pattern = "([aA]ttribute)|\\.rds") > 0] %>%
                        stringr::str_detect(., pattern = "([aA]ttribute)", negate = TRUE) == TRUE))
    # TRUE IF .csv data are present:
  (csv_query <- any(stringr::str_detect(most_recent, "([aA]ttributes)")) == TRUE &&
      any(stringr::str_detect(most_recent, ".*\\.csv{1}")) == TRUE)

  #### Both present ####
  # IF their is a complete .rds file among the most-recent files AND a .csv version...
  if(rdata_query == TRUE && csv_query == TRUE){
    writeLines(paste("\n", 
                     " - Oh boy, it looks like there are both .csv and .rds versions of your data!", 
                     "\n", "R will preferentially use the .rds file.", "\n",
                     "NOTE: the .rds file can be very slow to load"))
    # File to read:
    fileLoc <- most_recent[intersect(grep(".*\\.rds{1}", most_recent),
                                     grep("([aA]ttributes)", most_recent, invert = TRUE))]
    # Read in the .rds file
    writeLines(paste("Reading in ", fileLoc, "...", sep = ""))
   
    # Find the index of the string that matches and select that to read in
    occurDF <- fileLoc %>%
      readRDS()
  } #END IF both
  
  #### RData present ####
  # IF their is ONLY a complete .rds file among the most-recent files...
  if(rdata_query == TRUE && csv_query == FALSE){
    writeLines(paste(" - .rds export version found. Loading this file...", "\n",
                     "NOTE: the .rds file can be very slow to load"))
    # File to read:
    fileLoc <- most_recent[intersect(grep(".*\\.rds{1}", most_recent),
                                     grep("([aA]ttributes)", most_recent, invert = TRUE))]
    # Read in the .rds file
    writeLines(paste("Reading in ", 
                     fileLoc,
               "...", sep = ""))
    # Find the index of the string that matches and select that to read in
    occurDF <- as.character(fileLoc) %>%
      readRDS()
    writeLines("Completed reading in .rds file")
  } #END IF .rds

  #### CSV present ####
  # IF their is ONLY a complete .csv file among the most-recent files...
  if(csv_query == TRUE && rdata_query == FALSE){
    writeLines(paste(" - .csv exported version found. Loading this file..."))
    ColTypes <- ColTypeR()
    # Find the most-recent .csv occurrence file
    occurDF <- most_recent[intersect(grep(".*\\.csv", most_recent),
                                      grep("([aA]ttributes)", most_recent, invert = TRUE))] %>%
      read_csv(col_types = ColTypes)
    # Find the most-recent .rds attributes file
    attr_loc <- most_recent[stringr::str_which(most_recent, "(.*[aA]ttribute)(.*\\.rds)")] 
      # Check to see if the attributes file exists or not
    if(length(attr_loc) == 0){
      writeLines("No attribute file found... Please make sure that one exists to include the EML data")
    }else{
      writeLines(
        paste("Reading attribute file named ", attr_loc, "..." ))
      attr_file <- readRDS(attr_loc)
      }
    # Add the attributes file to the occurrence data file 
    attributes(occurDF) <- attr_file
    # Read in the EML file 
      # Find the folder that the attributes file is in.
    EML_home <- stringr::str_replace(attr_loc, pattern = "\\/[a-zA-Z0-9-_]+\\.rds$", "")
      # Find the .xml file in the same location as the attribute's folder
    EML_loc <- BeeBDC::fileFinder(path = EML_home, fileName = "eml.*\\.rds")
      # Read in the EML file
    EML_file <- readRDS(EML_loc)
  } #END IF .csv
  
  # Extract and save the data and the metadata based on their class
  for(i in 1:length(occurDF)){
      # If eml
    if(base::any(class(occurDF[[i]]) %in% c( "emld") )){
      eml_files <- occurDF[i]
    }
      # If tibble 
    if(base::any(class(occurDF[[i]]) %in% c("tbl_df", "data.frame", "tbl"))){
      Data_WebDL <- occurDF[i]
    }
  }

  
#### Return data ####
  # Re-combine the data and EML data
  Data_WebDL <- dplyr::lst(Data_WebDL, 
                          eml_files)
  
  # Return the Data_WebDL
  return(Data_WebDL)
  # Return end product and print completion note
  writeLines(paste(" - Fin.", sep = "\n"))
} # END data_importer

# This Function is designed to check continent-level outliers using the Discover Life checklist.
# It was written by James B Dorey from the 8th of November 2022.

#' Flag continent-level outliers with a provided checklist.
#'
#'This function flags continent-level outliers using the checklist provided with this package. 
#'For additional context and column names, see [BeeBDC::beesChecklist()].
#'
#' @param checklist A data frame or tibble. The formatted checklist which was built based on the Discover Life website.
#' @param data A data frame or tibble. The a Darwin Core occurrence dataset.
#' @param keepAdjacentContinent Logical. If TRUE, occurrences in continents that are adjacent to checklist continents will be 
#' kept. If FALSE, they will be flagged. Defualt = FALSE.
#' @param pointBuffer Numeric. A buffer around points to help them align with a continent or coastline.
#' This provides a good way to retain points that occur right along the coast or borders of the 
#' maps in rnaturalearth
#' @param scale Numeric. The value fed into the map scale parameter for
#'  [rnaturalearth::ne_countries()]'s scale parameter:
#' 	Scale of map to return, one of 110, 50, 10 or 'small', 'medium', 'large', where smaller numbers 
#' 	are higher resolution. WARNING: This function is tested on 110 and 50.
#' @param stepSize Numeric. The number of occurrences to process in each chunk. Default = 1000000.
#' @param mc.cores Numeric. If > 1, the function will run in parallel
#' using mclapply using the number of cores specified. If = 1 then it will be run using a serial
#' loop. NOTE: Windows machines must use a value of 1 (see ?parallel::mclapply). Additionally,
#' be aware that each thread can use large chunks of memory. If the cores throw issues, consider
#' setting mc.cores to 1.
#'  Default = 1.
#'
#' @return The input data with two new columns, .continentOutlier or .sea. There are three possible 
#' values for 
#' the new column: TRUE == passed, FALSE == failed (not in continent or in the ocean),
#'  NA == did not overlap with rnaturalearth map.
#' 
#' @seealso [BeeBDC::countryOutlieRs()] for implementation at the country level. Country-level 
#' implementation will be more data-hungry and, where data do not yet exist, difficult to implement.
#' Additionally, see [BeeBDC::beesChecklist()] for input data. Note, not all columns are 
#' necessary if you are building your own dataset. At a minimum you will need *validName* and
#' *continent*.
#' 
#' @export
#' @importFrom dplyr %>%
#'
#' @examples
#' if(requireNamespace("rnaturalearthdata")){
#' library(magrittr)
#'   # Load in the test dataset
#' beesRaw <- BeeBDC::beesRaw
#'   # For the sake of this example, use the testChecklist
#' system.file("extdata", "testChecklist.rda", package="BeeBDC") |> load()
#'   # For real examples, you might download the beesChecklist from FigShare using 
#'   #  [BeeBDC::beesChecklist()]
#' 
#' beesRaw_out <- continentOutlieRs(checklist = testChecklist,
#'                                data = beesRaw %>%
#'                                dplyr::filter(dplyr::row_number() %in% 1:50),
#'                                keepAdjacentContinent = FALSE,
#'                                pointBuffer = 1,
#'                                scale = 50,
#'                                stepSize = 1000000,
#'                                mc.cores = 1)
#' table(beesRaw_out$.continentOutlier, useNA = "always")
#' } # END if require

continentOutlieRs <- function(
    checklist = NULL,
    data = NULL,
    keepAdjacentContinent = FALSE,
    pointBuffer = NULL,
    scale = 50,
    stepSize = 1000000,
    mc.cores = 1
){
  # locally bind variables to the function
  name<-name_long<-continent<-geometry<-continentOutlieRs<-decimalLongitude<-
    decimalLatitude<-database_id<-countryOutlieRs<-scientificName<-species<-
    family<-subfamily<-genus<-specificEpithet<-
    scientificNameAuthorship<-country<-stateProvince<-eventDate<-
    institutionCode<-recordNumber<-catalogNumber<-dataSource<-countryOutlieRs<-
    verbatim_scientificName<-.<-neighbours<-rowNum<-neighboursText<-
    SciContinent<-validName<-SciContinent_noYear<-
    neighbourMatch_noYear<-exactMatch_noYear<-matchType<-continentMatch<-.continentOutlier <- 
    BeeBDC_order <- BeeBDC_group <- points <- 
    inData <- indexMatch <- NULL
  
  # REMOVE - TEST thinning
  #  data <- data %>%
  #    filter(row_number() %% 100 == 1)
  startTime <- Sys.time()
  #### 0.0 Warnings ####
  if(is.null(checklist)){
    stop("You must provide a checklist of countries")
  }
  if(is.null(data)){
    stop("You must provide occurrence data (data). Honestly, what do you think I was gonna do without that?")
  }
  
  #### 1.0 Data prep ####
  ##### 1.1 data ####
  # Drop .continentOutlier if its already present
  data <- data %>%
    dplyr::select(!tidyselect::any_of(".continentOutlier")) %>%
    # Remove other columns made by this function
    dplyr::select(!tidyselect::starts_with(c( "continentMatch", "continent")))
  
  ##### 1.2 rNaturalEarth ####
  # Download world map using rnaturalearth packages
  suppressWarnings({
    countryMap <- rnaturalearth::ne_countries(returnclass = "sf", country = NULL,
                                              type = "countries", scale = scale)  %>%
      # buffer by zero and make geometry valid to avoid potential issues with polygon intersection
      sf::st_make_valid() %>%
      # Select only a subset of the naturalearthdata columns to extract
      dplyr::select(name, name_long, continent, geometry) 
  })
  
  # Simplify the world map ONCE to be used later
  simplePoly <- countryMap %>% sf::st_drop_geometry() %>%
    dplyr::mutate(indexMatch = dplyr::row_number())
  
    # Make a simplified continent map
  suppressWarnings({
    suppressMessages(
  continentMap <- countryMap %>% 
    dplyr::group_by(continent) %>%
    dplyr::summarise(geometry = sf::st_union(geometry)) %>%
    dplyr::ungroup()
    )
  })
  # Simplify the continent map ONCE to be used later
  simpleConty <- continentMap %>% sf::st_drop_geometry() %>%
    dplyr::mutate(indexMatch = dplyr::row_number())
  
  # Dont's use spherical geometry
  sf::sf_use_s2(FALSE)
  
  
  #### 2.0 Use occ. data ####
  ##### 2.1 Create functions ####
  ###### a. simple intersect ####
  # Hijack st_intersection to allow it to be run in parallel
  jbd_intersection <- function(inData){
    inData <- inData %>% tidyr::drop_na(decimalLongitude, decimalLatitude)
    suppressWarnings({ suppressMessages({
      # Turn inData into a simple point feature
      points <- sf::st_as_sf(inData,
                             coords = c("decimalLongitude", "decimalLatitude"),
                             na.fail = TRUE,
                             # Assign the CRS from the rnaturalearth map to the point inData
                             crs = sf::st_crs(continentMap)) %>%
        # Use a subset of columns
        dplyr::select(database_id, scientificName, species, family, subfamily, genus, specificEpithet, 
                      scientificNameAuthorship,
                      stateProvince, eventDate, institutionCode, recordNumber, catalogNumber,
                      dataSource, verbatim_scientificName, geometry)
      #Extract polygon information to points
      points_extract <- sf::st_intersects(points, continentMap) %>%
        # return a tibble with the index of each match or NA where there was no match
        dplyr::tibble(indexMatch = .) %>%
        # Convert to numeric
        dplyr::mutate(indexMatch = indexMatch %>% as.character() %>%
                        # deal with problems - Take the first number where two are provided
                        stringr::str_extract("[0-9]+") %>% 
                        # Remove zero to NA
                        stringr::str_replace("^[0]$", NA_character_) %>% as.numeric()) %>%
        dplyr::left_join(simpleConty,
                         by = "indexMatch") %>%
        # Add in the database_id
        dplyr::bind_cols(inData %>% sf::st_drop_geometry())
    }) })
    # Return the points
    return(points_extract)
  } # END jbd_intersection
  
  ###### b. buffered intersect ####
  # Hijack st_intersection to allow it to be run in parallel
  jbd_bufferedIntersection <- function(inData){
    suppressWarnings({ suppressMessages({
      inData <- inData %>% 
        # Use only complete lat and lon data
        tidyr::drop_na(decimalLongitude, decimalLatitude) %>%
        # Remove the previous column names from jbd_intersection
        dplyr::select(!tidyselect::any_of(c("name","name_long","continent",
                                            "indexMatch")))
      # Turn inData into a simple point feature
      points <- sf::st_as_sf(inData,
                             coords = c("decimalLongitude", "decimalLatitude"),
                             na.fail = TRUE,
                             # Assign the CRS from the rnaturalearth map to the point inData
                             crs = sf::st_crs(continentMap)) %>%
        # Use a subset of columns
        dplyr::select(database_id, scientificName, species, family, subfamily, genus, specificEpithet, 
                      scientificNameAuthorship,
                      stateProvince, eventDate, institutionCode, recordNumber, catalogNumber,
                      dataSource, verbatim_scientificName, geometry) %>%
        # Buffer the points by the pointBuffer
        sf::st_buffer(dist = pointBuffer)
      
      #Extract polygon information to points
      points_extract <- sf::st_intersects(points, continentMap) %>%
        # return a tibble with the index of each match or NA where there was no match
        dplyr::tibble(indexMatch = . ) %>%
        # Convert to numeric
        dplyr::mutate(indexMatch = indexMatch %>% as.character() %>%
                        # deal with problems - Take the first number where two are provided
                        stringr::str_extract("[0-9]+") %>% 
                        # Remove zero to NA
                        stringr::str_replace("^[0]$", NA_character_) %>% as.numeric()) %>%
        dplyr::left_join(simpleConty,
                         by = "indexMatch") %>%
        # Add in the database_id
        dplyr::bind_cols(inData %>% sf::st_drop_geometry() %>% 
                           dplyr::select(!tidyselect::any_of("country")))
    })  })
    # Return the points
    return(points_extract)
  } # END jbd_intersection
  
  
  
  ##### 2.2 Extraction ####
  ###### a. exactContinent ####
  writeLines(" - Extracting continent data from points...")
  points_extract <- data %>%
    # Make a new column with the ordering of rows
    dplyr::mutate(BeeBDC_order = dplyr::row_number()) %>%
    # Group by the row number and step size
    dplyr::group_by(BeeBDC_group = ceiling(BeeBDC_order/stepSize)) %>%
    # Split the dataset up into a list by group
    dplyr::group_split(.keep = TRUE) %>%
    # Run the actual function
    parallel::mclapply(., jbd_intersection,
                       mc.cores = mc.cores
    ) %>%
    # Combine the lists of tibbles
    dplyr::bind_rows() %>%
    # Drop those occurrences that did not intersect with a continent
    tidyr::drop_na(continent)
  
  
  if(!is.null(pointBuffer)){
    # Failed extractions
    points_failed <- data %>%
      dplyr::filter(!database_id %in% points_extract$database_id)
    
    writeLines(" - Buffering failed points by pointBuffer...")
    
    points_failed <- points_failed %>%
      # Make a new column with the ordering of rows
      dplyr::mutate(BeeBDC_order = dplyr::row_number()) %>%
      # Group by the row number and step size
      dplyr::group_by(BeeBDC_group = ceiling(BeeBDC_order/stepSize)) %>%
      # Split the dataset up into a list by group
      dplyr::group_split(.keep = TRUE) %>%
      # Run the actual function
      parallel::mclapply(., jbd_bufferedIntersection,
                         mc.cores = mc.cores
      ) %>%
      # Combine the lists of tibbles
      dplyr::bind_rows() %>%
      # Drop those occurrences that did not intersect with a continent
      tidyr::drop_na(continent)
    
    if(nrow(points_failed) > 0){
      # Re-merge good with failed
      points_extract <- points_extract %>%
        sf::st_drop_geometry() %>%
        # remove buffer-matched occurrences
        dplyr::filter(!database_id %in% points_failed$database_id) %>%
        # replace these, but now matched
        dplyr::bind_rows(points_failed %>% sf::st_drop_geometry())
    }
  } # End if pointBuffer
  
  writeLines(" - Prepare the neighbouring continent dataset...")
  ###### b. neighbouringContinents ####
  # Get a list of continents that share borders
  continentsBordering <- sf::st_intersects(continentMap, continentMap) %>%
    paste(., sep = ";")
  # Make a new tibble with these information
  neighbouringContinents <- dplyr::tibble(
    rowNum = 1:nrow(continentMap),
    continent = continentMap$continent,
    neighbours = continentsBordering,
    # Modify the text in column
    neighboursText = mgsub::mgsub(string = neighbours,
                                  pattern = rowNum,
                                  replacement = continent) %>%
      stringr::str_replace(string = .,
                           pattern = "c\\(", replacement = "") %>%
      stringr::str_replace(string = .,
                           pattern = "\\)", replacement = "") %>%
      stringr::str_replace(string = .,
                           pattern = ":", replacement = ", ")) 
  
  # Make a long-format tibble with neighbouring countries
  neighbouringContinents <- neighbouringContinents %>%
    tidyr::separate_rows(data = ., neighboursText,
                         sep = ",")  %>%
    # Remove continent matching themselves
    dplyr::filter(!(continent == neighboursText))
  # Remove extra spaces
  neighbouringContinents <- neighbouringContinents %>%
    dplyr::mutate(neighboursText = stringr::str_squish(neighboursText),
                  neighboursText = stringr::str_remove_all(neighboursText, " \\(open ocean"))

  # Join the datasets together so that we can make a list of adjacent countries to match also
  neighbouringContinents <- checklist %>%
    dplyr::left_join(neighbouringContinents %>% 
                       dplyr::select(tidyselect::any_of(c("continent", "neighboursText"))),
                     by = c("continent" = "continent"),
                     multiple = "all", relationship = "many-to-many")
  
  ###### c. Sea points ####
  # Find the points that did not overlap with countries but that had coordinates
  seaPoints <- sf::st_as_sf(data %>% tidyr::drop_na(decimalLongitude, decimalLatitude),
                            coords = c("decimalLongitude", "decimalLatitude"),
                            na.fail = TRUE,
                            # Assign the CRS from the rnaturalearth map to the point data
                            crs = sf::st_crs(countryMap)) %>%
    # Use a subset of columns
    dplyr::select(database_id, scientificName, species, family, subfamily, genus, specificEpithet, 
                  scientificNameAuthorship,
                  stateProvince, eventDate, institutionCode, recordNumber, catalogNumber,
                  dataSource, verbatim_scientificName, geometry) %>% 
    sf::st_drop_geometry() %>%
    dplyr::filter(!database_id %in% points_extract$database_id) %>%
    dplyr::select(database_id)
  
  ##### 2.3 Compare ####
  writeLines(" - Compare points with the checklist...")
  # Get a smaller subset of the columns AND make a new columns with scientific name and country
  points_simple <- points_extract %>% 
    dplyr::select(database_id, continent, scientificName) %>%
    dplyr::mutate(SciContinent = stringr::str_c(scientificName, 
                                              continent, sep = "_")) %>% 
    # Remove grammar and caps from SciContinent
    dplyr::mutate(SciContinent = tolower(SciContinent) %>%
                    # Replace punctuation
                    stringr::str_replace_all("[\\(\\)\\,\\.\\-]", "") %>%
                    # Replace white spaces with underscores
                    stringr::str_replace_all(" ", "_")) %>%
    # Make the new column to match with full species name, (NO AUTHORHSIP YEAR), and country
    # Remove grammar and caps from SciContinent
    dplyr::mutate(SciContinent_noYear = tolower(SciContinent) %>%
                    # Replace numbers
                    stringr::str_replace_all("[0-9]", "")%>%
                    # replace double __
                    stringr::str_replace_all("__", "_"))
  
  
  ###### a. exactContinent ####
  # Do the same for the ascher checklist
  checklist_simple <- checklist %>%
    # Select subset
    dplyr::select(validName,continent) %>%
    # Make the new column to match with full species name, authorship, and continent
    dplyr::mutate(SciContinent = stringr::str_c(validName, continent, sep = "_")) %>% 
    # Remove grammar and caps from SciContinent
    dplyr::mutate(SciContinent = tolower(SciContinent) %>%
                    # Replace punctuation
                    stringr::str_replace_all("[\\(\\)\\,\\.\\-]", "") %>%
                    # Replace white spaces with underscores
                    stringr::str_replace_all(" ", "_")) %>%
    # Make the new column to match with full species name, (NO AUTHORHSIP YEAR), and continent
    # Remove grammar and caps from SciContinent
    dplyr::mutate(SciContinent_noYear = tolower(SciContinent) %>%
                    # Replace numbers
                    stringr::str_replace_all("[0-9]", "") %>%
                    # replace double __
                    stringr::str_replace_all("__", "_"))
  
  
  
  # Make a new columns showing if that species is expected in that continent
  points_match <- points_simple %>%
    #dplyr::filter(continent == "United states") %>%
    #dplyr::mutate(exactMatch = dplyr::if_else(SciContinent %in% checklist_simple$SciContinent,
    #                                    TRUE, FALSE)) %>%
    dplyr::mutate(exactMatch_noYear = dplyr::if_else(SciContinent_noYear %in% checklist_simple$SciContinent_noYear,
                                                     TRUE, FALSE)) %>%
    dplyr::left_join(dplyr::select(checklist_simple, SciContinent_noYear), 
                     by = "SciContinent_noYear",
                     multiple = "all", relationship = "many-to-many")
  
  
  ###### b. neighbouringContinents ####
  # Get a smaller subset of the data AND make a new columns with scientific name and continent
  nchecklist_simple <- neighbouringContinents %>%
    # Select subset
    dplyr::select(validName, neighboursText) %>%
    # Make the new column to match with
    dplyr::mutate(SciContinent = stringr::str_c(validName, neighboursText, sep = "_")) %>%
    # Remove grammar and caps from SciContinent
    dplyr::mutate(SciContinent = tolower(SciContinent) %>%
                    # Replace punctuation
                    stringr::str_replace_all("[\\(\\)\\,\\.\\-]", "") %>%
                    # Replace white spaces with underscores
                    stringr::str_replace_all(" ", "_")) %>%
    # Make the new column to match with full species name, (NO AUTHORHSIP YEAR), and continent
    # Remove grammar and caps from SciContinent
    dplyr::mutate(SciContinent_noYear = tolower(SciContinent) %>%
                    # Replace numbers
                    stringr::str_replace_all("[0-9]", "") %>%
                    # replace double __
                    stringr::str_replace_all("__", "_")) %>%
    # Get a unique set
    dplyr::distinct(SciContinent, .keep_all = TRUE)
  # Make a new columns showing if that species is expected in that continent
  npoints_match <- points_simple %>%
    dplyr::mutate(neighbourMatch_noYear = dplyr::if_else(SciContinent_noYear %in% nchecklist_simple$SciContinent_noYear,
                                                         TRUE, FALSE)) 
  
  
  #### 3.0 Merge ####
  writeLines(" - Combining data...")
  # Merge both points_match datasets
  bpoints_match <- dplyr::tibble(points_match) %>%
    # Join the two datasets togehter keeping only neighbourMatch and assignmentCertainty from the 
    # neighbour-joined dataset
    dplyr::left_join(dplyr::select(npoints_match, c(database_id, neighbourMatch_noYear)),
                     by = "database_id",
                     multiple = "all", relationship = "many-to-many") %>%
    # Remove geometry column
    dplyr::select(!tidyselect::starts_with("geometry")) %>%
    # Combine exactMatch_noYear and neighbourMatch_noYear
    dplyr::mutate(matchType = dplyr::if_else(exactMatch_noYear == TRUE,
                                             "exact", dplyr::if_else(neighbourMatch_noYear == TRUE,
                                                                     "neighbour", 
                                                                     "noMatch"))) 
  
  #### 4.0 Output ####
  ##### 4.1 User output ####
  bpoints_match <- bpoints_match %>%
    # Select the columns to keep
    dplyr::select(database_id, continent, matchType) %>%
    dplyr::rename(
      continentMatch = matchType) 
  # Set flag for those that don't pass continentMatch
  
  ###### a. keepAC == TRUE ####
  if(keepAdjacentContinent == TRUE){
    bpoints_match <- bpoints_match %>%
      dplyr::mutate(
        .continentOutlier = dplyr::if_else(continentMatch != "exact" & continentMatch != "neighbour",
                                         FALSE, TRUE)
      ) # END mutate
  } # END TRUE
  
  ###### b. keepAC == FALSE ####
  if(keepAdjacentContinent == FALSE){
    bpoints_match <- bpoints_match %>%
      dplyr::mutate(
        .continentOutlier = dplyr::if_else(continentMatch != "exact",
                                         FALSE, TRUE)
      ) # END mutate
  }# END FALSE
  
  # Keep only entirely unique records
  bpoints_match <- bpoints_match %>%
    dplyr::distinct() 
  
  ###### c. distinct buffer ####
  # For those buffered records that might have overlapped with >1 continent, select the unfiltered one, if it exists.
  if(!is.null(pointBuffer)){
    writeLines(" - Sorting and removing potentially duplicated buffered points...")
    bpoints_match <- bpoints_match %>%
      dplyr::group_by(database_id) %>%
      dplyr::arrange(desc(.continentOutlier), .by_group = TRUE) %>%
      dplyr::distinct(database_id, .keep_all = TRUE)
  }
  
  
  # Merge with original dataset
  output <- data %>%
    dplyr::left_join(bpoints_match, by = "database_id",
                     multiple = "all", relationship = "many-to-many") %>%
    # Add in .sea usign the seaPoints
    dplyr::mutate(.sea = dplyr::if_else(database_id %in% seaPoints$database_id,
                                        FALSE, TRUE))
  
  
  writeLines(paste0(
    " - Finished. \n",
    "We have matched ", 
    format(sum(bpoints_match$continentMatch == "exact", na.rm = TRUE), big.mark = ","),
    " records to their exact continent and ", 
    format(sum(bpoints_match$continentMatch == "neighbour", na.rm = TRUE), big.mark = ","), 
    " to an adjacent continent\n", 
    "We failed to match ",
    format(sum(bpoints_match$continentMatch == "noMatch", na.rm = TRUE), big.mark = ","), 
    " occurrences to any 'exact' or 'neighbouring' continent\n",
    "There are ",
    format(sum(is.na(output$.continentOutlier)), big.mark = ","), 
    " 'NA' occurrences for the .continentOutlier column.\n"
  ))
  
  
  
  # return message
  message(paste0("continentOutlieRs:\nFlagged ", 
                format(sum(output$.continentOutlier == FALSE, na.rm = TRUE), big.mark = ","), 
                " for continent outlier and flagged ",
                format(sum(output$.sea == FALSE, na.rm = TRUE), big.mark = ","), 
                " for in the .sea ",
                "records.\nThree columns were added to the database:\n1. ",
                "The '.continentOutlier' column was added which is a filtering column. \n2. ",
                "The 'continentMatch' columns indicates exact, neighbour, or noMatch. \n",
                "3. The '.sea' column was added as a filtering column for points in the ocean.",
                " The '.sea' column includes the user input buffer in its calculation."),
          sep = "")
  
  # rm(countryMap, points, points_extract, continentsBordering, neighbouringContinents, points_simple,
  #    checklist_simple, points_match, nchecklist_simple, npoints_match, bpoints_match)
  # Return file
  endTime <- Sys.time()
  # Time output
  message(paste(
    " - Completed in ", 
    round(difftime(endTime, startTime), digits = 2 )," ",
    units(round(endTime - startTime, digits = 2)),
    sep = ""))
  return(output)
  
} # END continentOutlieRs

requireNamespace("dplyr")
requireNamespace("readr")


  # If rnaturalearthdata is present, run tests
if(requireNamespace("rnaturalearthdata")){

# Load in the test checklist data
system.file("extdata", "testChecklist.rda", package="BeeBDC") |>
  load()
# Load in the test dataset
beesFlagged <- BeeBDC::beesFlagged

# Input a potentially difficult location on the map to test into an NA lat/lon slot
beesFlagged$decimalLatitude[[1]] <-  31.887646484374983
beesFlagged$decimalLongitude[[1]] <- 78.719726562500085 
beesFlagged$decimalLatitude[[2]] <-  78.719726562500085
beesFlagged$decimalLongitude[[2]] <- 31.887646484374983 
beesFlagged$decimalLatitude[[3]] <-  18.846970
beesFlagged$decimalLongitude[[3]] <- 18.846970 
beesFlagged$decimalLatitude[[4]] <-  30.328259
beesFlagged$decimalLongitude[[4]] <- 101.838181 
 
testOut <- continentOutlieRs(
  # Speed up operation by providing only the relevant entries in the testChecklist
  checklist = testChecklist,
  data = beesFlagged %>% 
    dplyr::select(!tidyselect::any_of(c("continentyMatch", ".continentOutlier"))),
  keepAdjacentContinent = FALSE,
  # running without a larger buffer to speed up tests
  pointBuffer = NULL,
  # Scale of map to return, one of 110, 50, 10 OR 'small', 'medium', 'large'
  # Smaller numbers will result in much longer calculation times. 
  # We have not attempted a scale of 10.
  scale = 50)


# Test the number of expected TRUE and FALSE columns and then test the output format (data frames and
# tibbles are a special case of lists)
testthat::test_that("continentOutlieRs results TRUE/passed", {
  testthat::expect_equal(sum(testOut$.continentOutlier == TRUE, na.rm = TRUE), 74)
})
testthat::test_that("continentOutlieRs results FALSE/failed", {
  testthat::expect_equal(sum(testOut$.continentOutlier == FALSE, na.rm = TRUE), 5)
})
testthat::test_that("continentOutlieRs results NA/could not assess", {
  testthat::expect_equal(sum(is.na(testOut$.continentOutlier)), 21)
})

# Test format
testthat::test_that("continentOutlieRs expected class", {
  testthat::expect_type(testOut, "list")
})
testthat::test_that("continentOutlieRs results NA/could not assess", {
  testthat::expect_true( all(stringr::str_detect(attributes(testOut)$class, 
                                                 c("tbl_df","tbl","data.frame"))))
})

# Expected number of columns matches input data
testthat::test_that("continentOutlieRs number of columns", {
  # Remove the three columns that will be updated by this function and then test that they are
  # added back in.
  testthat::expect_equal(ncol(beesFlagged %>% 
                                dplyr::select(!tidyselect::any_of(c("continentMatch", 
                                                                    ".continentOutlier")))) + 2,
                         ncol(testOut))
})

} # END if require

# This function, written by James Dorey merges the Ascher and other datasets with the columns:
  # "Original" and "Correct"
# For queries, please contact James Dorey at jbdorey[at]me.com
# This function was started on 13th May 2022 and last updated 17th May 2022
#' @importFrom dplyr %>%
#' @importFrom stats complete.cases 
#' @importFrom dplyr row_number
#' 
taxoMergeR <- function(currentNames = NULL,
                             newNames = NULL,
                             HigherNameList = NULL,
                       inKingdom = "Animalia",
                       inPhylum = "Arthropoda",
                       inClass = "Insecta",
                       inOrder = "Hymenoptera",
                       outPath = getwd(),
                       fileName = "taxonomy_taxoMergeR",
                       simpleNames = NULL,
                       problemStrings = NULL
                       ){
  # locally bind variables to the function
  Original <- Correct <- accid <- id <- validName <- canonical <- canonical_withFlags <- NULL
  genus <- species <- . <- tempIndex <- authorship <- taxonomic_status <- 
    authorship_nameSplit <- NULL
  
  requireNamespace("dplyr")

  
  #### 0.0 Prep ####
  ##### 0.1 Errors ####
  ###### a. FATAL errors ####
  if(is.null(simpleNames)){
    stop(" - Please provide an argument for simpleNames. This should be TRUE if using ",
         "'Genus species'",
         " format and FALSE if a more complex format like 'Genus (Subgenus) species Authority'")
  }
  
    #### 1.0 Data prep ####
    ##### 1.1 redundant names ####
  Original_newNames_Count <- nrow(newNames)
  # Remove redundant rows in NEW data (where Original == Correct or != unique())
  newNames <- newNames %>%
    # remove duplicate rows
    dplyr::distinct() %>%
    # Remove rows where Original == Correct
    dplyr::filter(Original != Correct) %>%
    # Remove empty spaces
    dplyr::mutate(Original = Original %>% stringr::str_squish(),
                  Correct = Correct %>% stringr::str_squish())
    # Simply count how many rows there are to output later
  Original_unNew_Count <- nrow(newNames)



  # Make a temporary index
  newNames$tempIndex <- 1:nrow(newNames)

  ##### 1.2 Unique synonyms [new] ####
    # In the newNames file, ONLY keep those synonyms that DO NOT occur in the list already
    ###### a. current names #####
    # Get all of the current names
  nameList <- currentNames %>%
    dplyr::select(accid, id, validName, canonical, canonical_withFlags, genus, species) %>%
    dplyr::mutate(Genus_species = stringr::str_c(dplyr::if_else(!is.na(genus), genus, ""),
                                                 dplyr::if_else(!is.na(species), species, ""),
                                                 sep = " "))
  
  ###### b. filter #### 
    # Filter out names that are already an Ascher synonym or accepted name
    # IF simple
  if(simpleNames == TRUE){
    newNames <- newNames %>%
    dplyr::filter(!Original %in% nameList$Genus_species)
  }
  
    # IF complex
  if(simpleNames == FALSE){
    newNames <- newNames %>%
    dplyr::filter(!Original %in% nameList$validName) %>%
    dplyr::filter(!Original %in% nameList$canonical) %>%
    dplyr::filter(!Original %in% nameList$canonical_withFlags) %>%
    dplyr::filter(!Original %in% nameList$Genus_species) 
  }
  
  ##### 1.3 Name splitting ####
  # Split the Original column into genus and species
  ###### a. simpleNames ####
  # Match simple Genus species format names
  # Do this in-line
    # if(simpleNames == TRUE){
    #   nameSplit <- newNames %>% 
    #     tidyr::separate(Original, into = c("genusNew", "speciesNew"), sep = " ",
    #                     remove = FALSE)}
  ###### b. complexNames ####
  # Remove problemStrings if they exist
  if(is.null(problemStrings) == FALSE){
    # Make a new loop string of names
    loopNames <- newNames$Original
    for(i in 1:length(problemStrings)){
      # Select the ith problemString
      str2remove = problemStrings[i]
      loopNames <- loopNames %>%
        stringr::str_replace_all(string = .,
                                 pattern = str2remove,
                                 replacement = "")
    }# END for problemStrings
    # Squish out extra white spaces
    loopNames <- loopNames %>% stringr::str_squish()
    # Add this as a new column
    newNames$Original_cleaned <- loopNames
  }else{# END if problemStrings
    newNames$Original_cleaned <- newNames$Original
  }
  
  # Match complex names
  if(simpleNames == FALSE){
    # create an empty tibble to populate
    loopTibble <- dplyr::tibble()
    # loop through each name to extract the relevant information
    for(i in 1:nrow(newNames)){
      # Extract the ith original name
      loopName <- newNames$Original_cleaned[i]
      loopSource <- newNames$Source[i]
      # Use nameSplitR to extract the information within
      loopData <- nameSplitR(NameInput = loopName, #NameInput
                             Authority_patterns = NULL) %>% 
        as.data.frame(stringsAsFactors = FALSE, row.names = FALSE) %>% dplyr::tibble() %>%        
        dplyr::mutate(Source = loopSource)
      # Add these data to loopTibble
      loopTibble <- loopTibble %>%
        dplyr::bind_rows(loopData)
    }
    # Sometimes a flag might be called due to a lack of author year. Make sure this does not happen.
    loopTibble$flags <- dplyr::if_else(stringr::str_detect(string = loopTibble$validName,
                                                           pattern = "[0-9]"),
                                       loopTibble$flags,
                                       "")
    # create a temporary tibble to populate
    matchTibble <- dplyr::tibble(
      tempIndex = newNames$tempIndex) %>%
      dplyr::bind_cols(loopTibble)
    ### END loop section
    # Add this information to newNames
    nameSplit <- newNames %>%
      # Remove the columns to be replace
      dplyr::select(!tidyselect::any_of(colnames(loopTibble))) %>%
      # add new info using the tempIndex column
      dplyr::left_join(matchTibble, by = "tempIndex")
  }# END simpleNames == FALSE
  
  
  #### 2.0 Simple Names ####
  if(simpleNames == TRUE){
  ##### 2.1 Acc. match ####
  # find all of the matching columns between New and Ascher using the final and flagged name columns
  Correct_matched <- newNames %>%
    dplyr::inner_join(currentNames, by = c("Correct" = "canonical_withFlags"), keep = TRUE,
                      relationship = "many-to-many")
  ##### 2.2 single match ####
  # subset those names that matched a single Ascher name
  SingleMatch <- Correct_matched %>%
    dplyr::group_by(tempIndex) %>%
    dplyr::filter(dplyr::n() == 1)
  # Split the Original column into genus and species
  suppressWarnings(
    SingleMatch_split <- SingleMatch %>% 
      tidyr::separate(Original, into = c("genusNew", "speciesNew"), sep = " ",
                      remove = FALSE),
    classes = "warning")
  
  
  # Find out if the Original New name (synonym) has a match and authority
  Syn_matched <- newNames %>%
    dplyr::inner_join(currentNames, by = c("Original" = "canonical_withFlags"), keep = TRUE,
                      relationship = "many-to-many") %>%
    # Keep only the current tempIndeces
    dplyr::filter(tempIndex %in% SingleMatch_split$tempIndex)%>%
    # Keep those with a single match
    dplyr::group_by(tempIndex) %>%
    dplyr::filter(dplyr::n() == 1)
  # Attach the authority to the SingleMatch_split tibble
  Single_newMatched <- SingleMatch_split %>%
    dplyr::left_join( dplyr::select(Syn_matched, c(tempIndex, authorship)), by = "tempIndex")
  
  ###### a. accepted ####
  # For those that match accepted names, use the id as the accid
  SOM_acc <- Single_newMatched %>%
    dplyr::filter(taxonomic_status == "accepted")
  
  # Merge these into a single tibble with the correct data 
  SOM_acc_final <- dplyr::bind_cols(
    flags = SOM_acc$flags,
    taxonomic_status = "synonym",
    source = SOM_acc$Source,
    accid = SOM_acc$id,  # Get the ACCEPTED name's id for accid
    id = NA,     # Enter later
    kingdom = inKingdom,
    phylum = inPhylum,
    class = inClass,
    order = inOrder,
    family = SOM_acc$family,
    subfamily = SOM_acc$subfamily,
    tribe = SOM_acc$tribe,
    subtribe = SOM_acc$subtribe,
    validName = stringr::str_c(
      dplyr::if_else(stats::complete.cases(SOM_acc$genusNew) & SOM_acc$genusNew != "NA",
                     SOM_acc$genusNew, ""),
      dplyr::if_else(stats::complete.cases(SOM_acc$authorship.y) & SOM_acc$authorship.y != "NA",
                     SOM_acc$authorship.y, ""),
      sep = " "),
    canonical_withFlags = "NA",
    canonical = stringr::str_c(
      dplyr::if_else(stats::complete.cases(SOM_acc$genusNew) & SOM_acc$genusNew != "NA",
                     SOM_acc$genusNew, ""),
      dplyr::if_else(stats::complete.cases(SOM_acc$genusNew) & SOM_acc$genusNew != "NA",
                     SOM_acc$genusNew, ""),
      dplyr::if_else(stats::complete.cases(SOM_acc$speciesNew) & SOM_acc$speciesNew != "NA",
                     SOM_acc$speciesNew, ""),
      sep = " "),
    genus = SOM_acc$genusNew,
    subgenus = "NA",
    species = SOM_acc$speciesNew,
    infraspecies = "NA",
    authorship = SOM_acc$authorship.y,
    taxon_rank = "species",
    valid = FALSE,
    tempIndex = SOM_acc$tempIndex,
    notes = SOM_acc$notes
  )
  
  ###### b. synonyms ####
  # For those that match SYNONYM names, use the accid as the accid - MATCH the actual accepted name.
  SOM_syn <- Single_newMatched %>% 
    dplyr::filter(taxonomic_status == "synonym")
  # Merge these into a single tibble with the correct data 
  SOM_syn_final <- dplyr::bind_cols(
    flags = SOM_syn$flags,
    taxonomic_status = "synonym",
    source = SOM_syn$Source,
    # Get the accepted id for accid. Hence, find the accepted name from the synonym this matched to
    accid = SOM_syn$accid,  
    id = NA,     # Assign later
    kingdom = inKingdom,
    phylum = inPhylum,
    class = inClass,
    order = inOrder,
    family = SOM_syn$family,
    subfamily = SOM_syn$subfamily,
    tribe = SOM_syn$tribe,
    subtribe = SOM_syn$subtribe,
    validName = stringr::str_c(
      dplyr::if_else(stats::complete.cases(SOM_syn$genusNew) & SOM_syn$genusNew != "NA",
                     SOM_syn$genusNew, ""),
      dplyr::if_else(stats::complete.cases(SOM_syn$speciesNew) & SOM_syn$speciesNew != "NA",
                     SOM_syn$speciesNew, ""),
      dplyr::if_else(stats::complete.cases(SOM_syn$authorship.y) & SOM_syn$authorship.y != "NA",
                     SOM_syn$authorship.y, ""),
      sep = " "),
    canonical_withFlags = "NA",
    canonical = stringr::str_c(
      dplyr::if_else(stats::complete.cases(SOM_syn$genusNew) & SOM_syn$genusNew != "NA",
                     SOM_syn$genusNew, ""),
      dplyr::if_else(stats::complete.cases(SOM_syn$speciesNew) & SOM_syn$speciesNew != "NA",
                     SOM_syn$speciesNew, ""),
      sep = " "),
    genus = SOM_syn$genusNew,
    subgenus = "NA",
    species = SOM_syn$speciesNew,
    infraspecies = "NA",
    authorship = SOM_syn$authorship.y,
    taxon_rank = "species",
    valid = FALSE,
    tempIndex = SOM_syn$tempIndex,
    notes = SOM_syn$notes
  )
  
  
  ##### 2.3 multiple matches ####
  ###### a. take accepted ####
  # subset those names that matched multiple Ascher names
  MultiMatch <- Correct_matched %>%
    # group by the temporary index number
    dplyr::group_by(tempIndex) %>%
    # Find all with more than one
    dplyr::filter(dplyr::n() > 1) %>% 
    # Sort in order of waht you want to keep
    dplyr::arrange(accid,
                   .by_group = TRUE) %>%
    # Filter for the first row only - this will direct to the accepted name hopefully
    dplyr::filter(dplyr::row_number() == 1)
  
  ###### a. find accepted (if needed) ####
  # If any rows do not == 0 (accepted names), then find those names.
  synMatched <- MultiMatch %>% 
    # Get those without an accepted match
    dplyr::filter(accid != 0) 
  # Remove the ascher columns
  synMatched_reduced <- synMatched %>%
    dplyr::select(!colnames(currentNames)) 
  # Re-add the accid
  synMatched_reduced$accid <- synMatched$accid
  synMatched_reduced$authorship_nameSplit <- synMatched$authorship
  # re-combine with the new accepted name's data
  synMatched <- synMatched_reduced %>% 
    dplyr::left_join(., currentNames, by = c("accid" = "id"))
  # These will need to have their accid become id - and changed back in a later step
  synMatched <- synMatched %>%
    dplyr::mutate(id = accid)
  
  ###### c. recombine ####
  MultiMatch <- MultiMatch %>%
    # Remove the old synonym records if they exist
    dplyr::filter(!tempIndex %in% synMatched$tempIndex) %>%
    dplyr::bind_rows(synMatched)
  
  ###### d. authorship ####
  # Give preference to the occurrence's authorship and then combine
  MultiMatch$authorship <- dplyr::if_else(MultiMatch$authorship_nameSplit == "" |
                                            is.na(MultiMatch$authorship_nameSplit),
                                          MultiMatch$authorship,
                                          MultiMatch$authorship_nameSplit )
  
  # Get the accepted names
  Mult_newAcc <- MultiMatch %>%
    dplyr::filter(taxonomic_status == "accepted")
  # Get the synonym names
  Mult_newSyn <- MultiMatch %>%
    dplyr::filter(taxonomic_status == "synonym")
  
  # IF there are synonyms that dont occur in the accepted names, STOP. Because this isn't a 
  # problem for me.
  MULTSynTest <- Mult_newSyn %>%
    dplyr::filter(!Original %in% Mult_newAcc$Original)
  if(nrow(MULTSynTest) > 0){
    stop(paste0("There are multiple-match synonyms that aren't in the mult.accepted list. ",
                "This is new. Look for 'MULTSynTest'"))}
  
  # Merge these into a single tibble with the correct data 
  MO_FINAL <- dplyr::bind_cols(
    flags = Mult_newAcc$flags,
    taxonomic_status = "synonym",
    source = Mult_newAcc$Source,
    accid = Mult_newAcc$id,  # Get the accepted id for accid
    id = NA,     # Get the new id from the number of rows plus i
    kingdom = inKingdom,
    phylum = inPhylum,
    class = inClass,
    order = inOrder,
    family = Mult_newAcc$family,
    subfamily = Mult_newAcc$subfamily,
    tribe = Mult_newAcc$tribe,
    subtribe = Mult_newAcc$subtribe,
    validName = stringr::str_c(
      dplyr::if_else(stats::complete.cases(Mult_newAcc$genus) & Mult_newAcc$genus != "NA",
                     Mult_newAcc$genus, ""),
      dplyr::if_else(stats::complete.cases(Mult_newAcc$species) & Mult_newAcc$species != "NA",
                     Mult_newAcc$species, ""),
      dplyr::if_else(stats::complete.cases(Mult_newAcc$authorship) & Mult_newAcc$authorship != "NA",
                     Mult_newAcc$authorship, ""),
      sep = " "),
    canonical_withFlags = "NA",
    canonical = stringr::str_c(
      dplyr::if_else(stats::complete.cases(Mult_newAcc$genus) & Mult_newAcc$genus != "NA",
                     Mult_newAcc$genus, ""),
      dplyr::if_else(stats::complete.cases(Mult_newAcc$species) & Mult_newAcc$species != "NA",
                     Mult_newAcc$species, ""),
      sep = " "),
    genus = Mult_newAcc$genus,
    subgenus = "NA",
    species = Mult_newAcc$species,
    infraspecies = "NA",
    authorship = Mult_newAcc$authorship,
    taxon_rank = "species",
    valid = FALSE,
    tempIndex = Mult_newAcc$tempIndex
    #notes = SM_GenSp$notes
  )
  } # END simpleNames == TRUE
  #
  
  
  
  #### 3.0 Complex names ####
  if(simpleNames == FALSE){
  ##### 3.1 Acc. match ####
    # find all of the matching columns between New and Ascher using the final and flagged 
      # name columns
  Correct_matched <- nameSplit %>%
      # Match with canonical_withFlags
    dplyr::inner_join(currentNames, by = c("Correct" = "canonical_withFlags"), keep = TRUE,
                      suffix = c("_nameSplit", ""), relationship = "many-to-many")
      # Match with validName for those that FAILED with canonical_withFlags
  Correct_matched2 <- nameSplit %>%
      # remove already matched
    dplyr::filter(!tempIndex %in% Correct_matched$tempIndex) %>%
    dplyr::inner_join(currentNames %>%
                          # Makes ure to sort by acceptednames first
                        dplyr::arrange(accid),
                        , by = c("Correct" = "validName"), keep = TRUE,
                      suffix = c("_nameSplit", ""), 
                        # Only keep the first match (will be the lowest accid)
                      multiple ="first")
  #merge
  Correct_matched <- Correct_matched %>%
    dplyr::bind_rows(Correct_matched2)
    rm(Correct_matched2)
    
    #### 3.2 single match ####
    # subset those names that matched a single Ascher name
  Single_newMatched <- Correct_matched %>%
    dplyr::group_by(tempIndex) %>%
    dplyr::filter(dplyr::n() == 1)
  
        # Give preference to the occurrence's authorship and then combine
  Single_newMatched$authorship <- dplyr::if_else(Single_newMatched$authorship_nameSplit == "" |
                                is.na(Single_newMatched$authorship_nameSplit),
                                Single_newMatched$authorship,
                                Single_newMatched$authorship_nameSplit ) 
       
        ###### a. accepted ####
       # For those that match accepted names, use the id as the accid
       SOM_acc <- Single_newMatched %>%
         dplyr::filter(taxonomic_status == "accepted")
       
    # Merge these into a single tibble with the correct data 
  SOM_acc_final <- dplyr::tibble(
    flags = SOM_acc$flags,
    taxonomic_status = "synonym",
    source = SOM_acc$Source,
    accid = SOM_acc$id,  # Get the ACCEPTED name's id for accid
    id = NA,     # Enter later
    kingdom = inKingdom,
    phylum = inPhylum,
    class = inClass,
    order = inOrder,
    family = SOM_acc$family,
    subfamily = SOM_acc$subfamily,
    tribe = SOM_acc$tribe,
    subtribe = SOM_acc$subtribe,
    validName = stringr::str_c(
      dplyr::if_else(stats::complete.cases(SOM_acc$genus_nameSplit) & SOM_acc$genus_nameSplit != "NA",
                     SOM_acc$genus_nameSplit, ""),
        # subgenus
      dplyr::if_else(stats::complete.cases(SOM_acc$subgenus_nameSplit) & 
                       SOM_acc$subgenus_nameSplit != "NA",
                     paste0("(",SOM_acc$subgenus_nameSplit ,")"), ""),
      dplyr::if_else(stats::complete.cases(SOM_acc$species_nameSplit) & 
                       SOM_acc$species_nameSplit != "NA",
                     SOM_acc$species_nameSplit, ""),
      dplyr::if_else(stats::complete.cases(SOM_acc$infraspecies_nameSplit) & 
                       SOM_acc$infraspecies_nameSplit != "NA",
                     SOM_acc$infraspecies_nameSplit, ""),
      dplyr::if_else(stats::complete.cases(SOM_acc$authorship) & SOM_acc$authorship != "NA",
                     SOM_acc$authorship, ""),
      sep = " "),
    canonical_withFlags = "NA",
    canonical = stringr::str_c(
      dplyr::if_else(stats::complete.cases(SOM_acc$genus_nameSplit) & SOM_acc$genus_nameSplit != "NA",
                     SOM_acc$genus_nameSplit, ""),
      # subgenus
      dplyr::if_else(stats::complete.cases(SOM_acc$subgenus_nameSplit) & 
                       SOM_acc$subgenus_nameSplit != "NA",
                     paste0("(",SOM_acc$subgenus_nameSplit ,")"), ""),
      dplyr::if_else(stats::complete.cases(SOM_acc$species_nameSplit) & SOM_acc$species_nameSplit != "NA",
                     SOM_acc$species_nameSplit, ""),
      dplyr::if_else(stats::complete.cases(SOM_acc$infraspecies_nameSplit) & 
                       SOM_acc$infraspecies_nameSplit != "NA",
                     SOM_acc$infraspecies_nameSplit, ""),
      sep = " "),
    genus = SOM_acc$genus_nameSplit,
    subgenus = dplyr::if_else(stats::complete.cases(SOM_acc$subgenus_nameSplit) & 
                                SOM_acc$subgenus_nameSplit != "NA",
                              SOM_acc$subgenus_nameSplit, ""),
    species = SOM_acc$species_nameSplit,
      # EDIT:
    infraspecies = SOM_acc$infraspecies_nameSplit,
    authorship = SOM_acc$authorship,
      # EDIT:
    taxon_rank = dplyr::if_else(stats::complete.cases(SOM_acc$infraspecies_nameSplit) & 
                                  SOM_acc$infraspecies_nameSplit != "NA",
                                "infraspecies", 
                                dplyr::if_else(stats::complete.cases(SOM_acc$species) & 
                                                 SOM_acc$species != "NA", 
                                               "species",
                                               dplyr::if_else(stats::complete.cases(SOM_acc$genus), 
                                                              "genus", "higher"))),
    valid = FALSE,
    tempIndex = SOM_acc$tempIndex,
    notes = SOM_acc$notes
    )

    ###### b. synonyms ####
  # For those that match SYNONYM names, use the accid as the accid - MATCH the actual accepted name.
  SOM_syn <- Single_newMatched %>% 
    dplyr::filter(taxonomic_status == "synonym")
  # Merge these into a single tibble with the correct data 
  SOM_syn_final <- dplyr::bind_cols(
    flags = SOM_syn$flags,
    taxonomic_status = "synonym",
    source = SOM_syn$Source,
    accid = SOM_syn$accid,  # Get the accepted id for accid. Hence, find the accepted name from 
    # the synonym this matched to
    id = NA,     # Assign later
    kingdom = inKingdom,
    phylum = inPhylum,
    class = inClass,
    order = inOrder,
    family = SOM_syn$family,
    subfamily = SOM_syn$subfamily,
    tribe = SOM_syn$tribe,
    subtribe = SOM_syn$subtribe,
    validName = stringr::str_c(
      dplyr::if_else(stats::complete.cases(SOM_syn$genus_nameSplit) & SOM_syn$genus_nameSplit != "NA",
                     SOM_syn$genus_nameSplit, ""),
      # subgenus
      dplyr::if_else(stats::complete.cases(SOM_syn$subgenus_nameSplit) & 
                       SOM_syn$subgenus_nameSplit != "NA",
                     paste0("(",SOM_syn$subgenus_nameSplit ,")"), ""),
      dplyr::if_else(stats::complete.cases(SOM_syn$species_nameSplit) & SOM_syn$species_nameSplit != "NA",
                     SOM_syn$species_nameSplit, ""),
      dplyr::if_else(stats::complete.cases(SOM_syn$infraspecies_nameSplit) & 
                       SOM_syn$infraspecies_nameSplit != "NA",
                     SOM_syn$infraspecies_nameSplit, ""),
      dplyr::if_else(stats::complete.cases(SOM_syn$authorship) & SOM_syn$authorship != "NA",
                     SOM_syn$authorship, ""),
      sep = " "),    
    canonical_withFlags = "NA",
      # EDIT:
    canonical = stringr::str_c(
      dplyr::if_else(stats::complete.cases(SOM_syn$genus_nameSplit) & SOM_syn$genus_nameSplit != "NA",
                     SOM_syn$genus_nameSplit, ""),
      # subgenus
      dplyr::if_else(stats::complete.cases(SOM_syn$subgenus_nameSplit) & 
                       SOM_syn$subgenus_nameSplit != "NA",
                     paste0("(",SOM_syn$subgenus_nameSplit ,")"), ""),
      dplyr::if_else(stats::complete.cases(SOM_syn$species_nameSplit) & SOM_syn$species_nameSplit != "NA",
                     SOM_syn$species_nameSplit, ""),
      dplyr::if_else(stats::complete.cases(SOM_syn$infraspecies_nameSplit) & 
                       SOM_syn$infraspecies_nameSplit != "NA",
                     SOM_syn$infraspecies_nameSplit, ""),
      sep = " "),
    genus = SOM_syn$genus_nameSplit,
    subgenus = dplyr::if_else(stats::complete.cases(SOM_syn$subgenus_nameSplit) & 
                                SOM_syn$subgenus_nameSplit != "NA",
                              paste0(SOM_syn$subgenus_nameSplit), ""),
    species = SOM_syn$species_nameSplit,
    # EDIT:
    infraspecies = SOM_syn$infraspecies_nameSplit,
    authorship = SOM_syn$authorship,
    # EDIT:
    taxon_rank = dplyr::if_else(stats::complete.cases(SOM_syn$infraspecies_nameSplit) & 
                                  SOM_syn$infraspecies_nameSplit != "NA",
                                "infraspecies", 
                                dplyr::if_else(stats::complete.cases(SOM_syn$species) & 
                                                 SOM_syn$species != "NA",
                                               "species",
                                               dplyr::if_else(stats::complete.cases(SOM_syn$genus) & 
                                                                SOM_syn$genus != "NA",
                                                              "genus", "higher"))),
    valid = FALSE,
    tempIndex = SOM_syn$tempIndex,
    notes = SOM_syn$notes
  )
  
  
    ##### 3.3 multiple matches ####
      ###### a. take accepted ####
    # subset those names that matched multiple Ascher names
  MultiMatch <- Correct_matched %>%
      # group by the temporary index number
    dplyr::group_by(tempIndex) %>%
      # Find all with more than one
    dplyr::filter(dplyr::n() > 1) %>% 
      # Sort in order of waht you want to keep
    dplyr::arrange(accid,
                   .by_group = TRUE) %>%
      # Filter for the first row only - this will direct to the accepted name hopefully
    dplyr::filter(dplyr::row_number() == 1)
  
  ###### a. find accepted (if needed) ####
    # If any rows do not == 0 (accepted names), then find those names.
  synMatched <- MultiMatch %>% 
      # Get those without an accepted match
    dplyr::filter(accid != 0) 
  # Remove the ascher columns
  synMatched_reduced <- synMatched %>%
    dplyr::select(!colnames(currentNames)) 
    # Re-add the accid
  synMatched_reduced$accid <- synMatched$accid
  # re-combine with the new accepted name's data
  synMatched <- synMatched_reduced %>% 
    dplyr::left_join(., currentNames, by = c("accid" = "id"))
    # These will need to have their accid become id - and changed back in a later step
  synMatched <- synMatched %>%
    dplyr::mutate(id = accid)
  
  ###### c. recombine ####
  MultiMatch <- MultiMatch %>%
      # Remove the old synonym records if they exist
    dplyr::filter(!tempIndex %in% synMatched$tempIndex) %>%
    dplyr::bind_rows(synMatched)
  
    ###### d. authorship ####
  # Give preference to the occurrence's authorship and then combine
  MultiMatch$authorship <- dplyr::if_else(MultiMatch$authorship_nameSplit == "" |
                                                   is.na(MultiMatch$authorship_nameSplit),
                                          MultiMatch$authorship,
                                          MultiMatch$authorship_nameSplit )
  # Remove redundant column
  MultiMatch <- MultiMatch %>% 
    dplyr::select(!c(authorship_nameSplit))
  
       
        # Get the accepted names
       Mult_newAcc <- MultiMatch %>%
         dplyr::filter(taxonomic_status == "accepted")
       # Get the synonym names
       Mult_newSyn <- MultiMatch %>%
         dplyr::filter(taxonomic_status == "synonym")
       
        # IF there are synonyms that dont occur in the accepted names, STOP. Because this isn't a 
        # problem for me.
       MULTSynTest <- Mult_newSyn %>%
         dplyr::filter(!Original %in% Mult_newAcc$Original)
       if(nrow(MULTSynTest) > 0){
         stop(paste0("There are multiple-match synonyms that aren't in the mult.accepted list.",
                     " This is new. Look for 'MULTSynTest'"))}
       
  # Merge these into a single tibble with the correct data 
  MO_FINAL <- dplyr::bind_cols(
    flags = Mult_newAcc$flags,
    taxonomic_status = "synonym",
    source = Mult_newAcc$Source,
    accid = Mult_newAcc$id,  # Get the accepted id for accid
    id = NA,     # Get the new id from the number of rows plus i
    kingdom = inKingdom,
    phylum = inPhylum,
    class = inClass,
    order = inOrder,
    family = Mult_newAcc$family,
    subfamily = Mult_newAcc$subfamily,
    tribe = Mult_newAcc$tribe,
    subtribe = Mult_newAcc$subtribe,
    validName = stringr::str_c(
      dplyr::if_else(stats::complete.cases(Mult_newAcc$genus_nameSplit) & 
                       Mult_newAcc$genus_nameSplit != "NA",
                     Mult_newAcc$genus_nameSplit, ""),
        #subgenus
      dplyr::if_else(stats::complete.cases(Mult_newAcc$subgenus_nameSplit) & 
                       Mult_newAcc$subgenus_nameSplit != "NA",
                     paste0("(",Mult_newAcc$subgenus_nameSplit ,")"), ""),
      dplyr::if_else(stats::complete.cases(Mult_newAcc$species_nameSplit) & 
                       Mult_newAcc$species_nameSplit != "NA",
                     Mult_newAcc$species_nameSplit, ""),
      dplyr::if_else(stats::complete.cases(Mult_newAcc$infraspecies_nameSplit) & 
                       Mult_newAcc$infraspecies_nameSplit != "NA",
                     Mult_newAcc$infraspecies_nameSplit, ""),
      dplyr::if_else(stats::complete.cases(Mult_newAcc$authorship) & Mult_newAcc$authorship != "NA",
                     Mult_newAcc$authorship, ""),
      sep = " "),
    canonical_withFlags = "NA",
    canonical = stringr::str_c(
      dplyr::if_else(stats::complete.cases(Mult_newAcc$genus_nameSplit) & 
                       Mult_newAcc$genus_nameSplit != "NA",
                     Mult_newAcc$genus_nameSplit, ""),
      #subgenus
      dplyr::if_else(stats::complete.cases(Mult_newAcc$subgenus_nameSplit) & 
                       Mult_newAcc$subgenus_nameSplit != "NA",
                     paste0("(",Mult_newAcc$subgenus_nameSplit ,")"), ""),
      dplyr::if_else(stats::complete.cases(Mult_newAcc$species_nameSplit) & 
                       Mult_newAcc$species_nameSplit != "NA",
                     Mult_newAcc$species_nameSplit, ""),
      dplyr::if_else(stats::complete.cases(Mult_newAcc$infraspecies_nameSplit) & 
                       Mult_newAcc$infraspecies_nameSplit != "NA",
                     Mult_newAcc$infraspecies_nameSplit, ""),
      sep = " "),
    genus = Mult_newAcc$genus_nameSplit,
    subgenus = dplyr::if_else(stats::complete.cases(Mult_newAcc$subgenus_nameSplit) & 
                                Mult_newAcc$subgenus_nameSplit != "NA",
                              paste0(Mult_newAcc$subgenus_nameSplit), ""),
    species = Mult_newAcc$species_nameSplit,
    # EDIT:
    infraspecies = Mult_newAcc$infraspecies_nameSplit,
    authorship = Mult_newAcc$authorship,
    # EDIT:
    taxon_rank = dplyr::if_else(
      stats::complete.cases(Mult_newAcc$infraspecies_nameSplit) & Mult_newAcc$infraspecies_nameSplit != "NA",
                                "infraspecies", 
                                dplyr::if_else(stats::complete.cases(Mult_newAcc$species) & 
                                                 Mult_newAcc$species != "NA", 
                                               "species",
                                               dplyr::if_else(stats::complete.cases(Mult_newAcc$genus) & 
                                                                Mult_newAcc$genus != "NA", 
                                                              "genus", "higher"))),
    valid = FALSE,
    tempIndex = Mult_newAcc$tempIndex
    #notes = SM_GenSp$notes
  )
  } # END simpleNames == FALSE
  
  #
  
  #### 4.0 Merge ####
    # Merge the output rows
  merged_names <- dplyr::bind_rows(SOM_acc_final, SOM_syn_final, MO_FINAL) 
  
  # Remove " NA" from end of validNames where there was no authority
  merged_names <- merged_names %>%
    dplyr::mutate(validName = stringr::str_replace(validName, 
                                                         pattern = " NA$", 
                                                         replacement = "") %>% 
                          stringr::str_squish())
    
    # Find those rows that did not match
  failed_names <- newNames %>%
    dplyr::filter(!tempIndex %in% merged_names$tempIndex)   
  
  # Select the desired columns
  merged_names <-  merged_names %>%
    dplyr::select(colnames(currentNames))
  # Add id numbers
  idMax <- max(currentNames$id) + 1
  idEnd <- idMax + nrow(merged_names) -1
  merged_names$id <- seq(from = idMax, to = idEnd, by = 1)
  
  # Remove silly names... for some reason I cant '!' inverse select these...
    # Not sure why this isn't filtering in the negative...
 #   toRemove <- newNames %>%
 #     dplyr::filter(!Correct %in% 
 #                     c("extinct", "genus", "Genus", "GENUS", "NIDB", "not_a_species",
 #                       "not a bee", "NOT A BEE", "not a species", "Not_a_bee", "NOT_A_BEE",
 #                       "untraced", "x", "Check"))
 #   
  
    # Write user output
  writeLines(paste(
   " - Names merged. ","\n",
   "We removed ", format(Original_newNames_Count - Original_unNew_Count, big.mark = ","), 
   " duplicate new synonyms ",
   "\n ", "We successfuly matched: ", "\n ",
   format(nrow(SOM_acc_final), big.mark = ","), " new names to the current accepted names;", "\n ",
   format(nrow(SOM_syn_final), big.mark = ","), 
   " new names to the current synonyms, and then their accepted name;",
   "\n ",format(nrow(MO_FINAL), big.mark = ","), 
   " new names that matched the current accepted and synonym names.",
   " These are matched to their accepted names;", "\n   ",
   "!! There were a total of ", format(nrow(failed_names), big.mark = ","), 
   " new names that failed",
   " to find a match in the the current list !!", "\n ",
   "We kept a total of ", format(nrow(merged_names), big.mark = ","), " new synonyms.",
  sep = ""))

  #### 3.0 Clean data ####  
  writeLines(paste(
    " - Cleaning new data...", sep = ""
  ))
  ## manage new data ##
  writeLines(paste(
    " - Adding higher names with the HigherNamer function...", sep = ""
  ))
    # Add higher order names
  merged_names_cl <- HigherNamer(HigherNameList = HigherNameList,
                      InSynList = merged_names)
    # manage flags 
  writeLines(paste(
    " - Managing flags with the FlagManager function...", sep = ""
  ))
  merged_names_cl <- FlagManager(InSynList = merged_names_cl,
                           flagCol = "notes")

  #### 4.0 Save ####
    writeLines(paste(
    " - Saving the matched new component of the data to the file ",
    outPath, "/", fileName, "_", Sys.Date(), ".csv",  " seperately...", sep = ""
  ))
  # Save the  current-matched new dataset
  readr::write_excel_csv(merged_names_cl, file = paste(outPath,"/",
    fileName, "_", Sys.Date(), ".csv", sep = ""))
  writeLines(paste(" - ", nrow(failed_names), 
    " names from the new list did not have an accepted or synonym match to the current list. They ",
    "will be removed. ", "\n",
    "Saving these no-match names to ",
    outPath, "/", fileName, "_", Sys.Date(), ".csv",  " seperately...", sep = ""
  ))
  # Save noMatch_df
  readr::write_excel_csv(failed_names, 
                   file = paste(outPath, "/",
                                fileName, "_failed_", Sys.Date(), ".csv",  sep = ""))

    # Add this dataset to the Ascher dataset
  merged_names_cl <- dplyr::tibble(merged_names_cl)
      # Convert some column types to merge with Ascher data
    merged_names_cl$id <- as.numeric(merged_names_cl$id)
    merged_names_cl$accid <- as.numeric(merged_names_cl$accid)
    merged_names_cl$valid <- as.logical(merged_names_cl$valid)
    
  namesFinal <- currentNames %>%
    dplyr::bind_rows(merged_names_cl)
  # Return this object
  return(namesFinal)
} # END Function

# This function is a modification of a bdc function and flags columns as FALSE when they are marked 
  # as "ABSENT". This function was written on the 22nd of May 2022 by James Dorey. Email James at
  # jbdorey[at]me.com for help.

#' Flags occurrences that are marked as absent
#' 
#' Flags occurrences that are "ABSENT" for the occurrenceStatus (or some other user-specified) column.
#'
#' @param data A data frame or tibble. Occurrence records as input.
#' @param PresAbs Character. The column in which the function will find "ABSENT" or "PRESENT" records.
#' Default = "occurrenceStatus"
#'
#' @return The input data with a new column called ".occurrenceAbsent" where FALSE == "ABSENT" records.
#' @export
#'
#'@importFrom dplyr %>%
#'
#' @examples
#'   # Bring in the data
#' data(beesRaw)
#'   # Run the function
#' beesRaw_out <- flagAbsent(data = beesRaw,
#' PresAbs = "occurrenceStatus")
#'   # See the result
#' table(beesRaw_out$.occurrenceAbsent, useNA = "always")
flagAbsent <-
  function(data = NULL,
           PresAbs = "occurrenceStatus") {
    .data <- .occurrenceAbsent <- individualCount <- NULL
    requireNamespace("dplyr")
    
      # Make a new column called .occurrenceAbsent to be TRUE when occurrenceStatus is "present" or NA
    data <-
      data %>%
      dplyr::mutate(
        .occurrenceAbsent =
          !.data[[PresAbs]] %in% c("ABSENT", "Absent", "absent")) %>%
        # Catch and flag individualCount == 0
      dplyr::mutate(.occurrenceAbsent = dplyr::if_else(complete.cases(individualCount),
                      dplyr::if_else(individualCount == 0,
                                                       FALSE, .occurrenceAbsent),
                      .occurrenceAbsent
                    ))
    

    # Return user output
    message(
      paste(
        "\\.occurrenceAbsent:\n",
        "Flagged",
        format(sum(data$.occurrenceAbsent == FALSE, na.rm = TRUE), big.mark = ","),
        "absent records:\n",
        "One column was added to the database.\n"
      )
    )
    return(data)
  }

requireNamespace("dplyr")
library(dplyr) # couldn't use %>% without this


# make some test data - this is derived from GBIF data but it is HEAVILY EDITED and USELESS
testData <- dplyr::tribble( # even step coordinates (different for lat/long)
  ~database_id, ~datasetName,  ~occurrenceStatus,          ~individualCount,
  "fake SCAN1", "fakeDataset",               "PRESENT",                1L,
  "fake SCAN2", "fakeDataset",               "PRESENT",                1L,
  "fake SCAN3", "fakeDataset",               "PRESENT",                1L,
  "fake SCAN4", "fakeDataset",                "ABSENT",                1L,
  "fake SCAN5", "fakeDataset",                "ABSENT",                1L,
  "fake SCAN6", "fakeDataset",                "ABSENT",                1L,
  "fake SCAN7", "fakeDataset",               "PRESENT",                1L,
  "fake SCAN8", "fakeDataset",                "ABSENT",                1L,
  "fake SCAN9", "fakeDataset",                  "ABSENT",                1L,
  "fake SCAN10", "fakeDataset",               "ABSENT",                1L,
  "fake SCAN11", "newDataset",               "PRESENT",                1L,
  "fake SCAN12", "newDataset",               "PRESENT",                1L,
  "fake SCAN13", "newDataset",                  "ABSENT",                1L,
  "fake SCAN14", "fakeDataset",            "PRESENT",                1L,
  "fake SCAN15", "fakeDataset",            "PRESENT",                1L,
  "fake SCAN16", "fakeDataset",            "PRESENT",                1L,
  "fake SCAN17", "fakeDataset",            "PRESENT",                1L,
  "fake SCAN18", "fakeDataset",             "ABSENT",                1L,
  "fake SCAN19", "fakeDataset",             "ABSENT",                1L,
  "fake SCAN20", "fakeDataset",             "ABSENT",                  1L,
  "fake SCAN21", "fakeDataset",             "ABSENT",                1L,
  "fake SCAN22", "fakeDataset",             "ABSENT",               1L,
  "fake SCAN23", "fakeDataset",            "PRESENT",               1L,
  "fake SCAN24", "fakeDataset",            "PRESENT",               0L,
  "fake SCAN25", "fakeDataset",            "PRESENT",               1L,
  "fake SCAN26", "fakeDataset",             "ABSENT",               1L,
  # no matching,
  "fake SCAN27", "fakeDataset",           "ABSENT",                 1L,
  "fake SCAN28", "fakeDataset",           "ABSENT",                 1L,
  "fake SCAN29", "fakeDataset",           "ABSENT",                 1L, 
  "fake SCAN30", "fakeDataset",           "ABSENT",                  1L,
  "fake SCAN31", "fakeDataset",           "ABSENT",                1L,
  # Too few decimals            
  "fake SCAN32", "newDataset",            "PRESENT",                1L,
  "fake SCAN33", "newDataset",            "PRESENT",                0L,
  "fake SCAN34", "newDataset",            "PRESENT",                1L,
  "fake SCAN35", "newDataset",            "PRESENT",                1L
)



# Run the function
testOut <- BeeBDC::flagAbsent(data = testData, PresAbs = "occurrenceStatus")


# test number of TRUE and FALSE values in the flag column, .occurrenceAbsent
resultsT <- length(testOut$.occurrenceAbsent[testOut$.occurrenceAbsent == TRUE])
resultsF <- length(testOut$.occurrenceAbsent[testOut$.occurrenceAbsent == FALSE])

testthat::test_that("flagAbsent column .occurrenceAbsent results TRUE", {
  testthat::expect_equal(resultsT, 15)
})

testthat::test_that("flagAbsent column .occurrenceAbsent results FALSE", {
  testthat::expect_equal(resultsF, 20)
})


# test the order of the TRUE and FALSE values in the flag column, .occurrenceAbsent
correct <- c(TRUE , TRUE,  TRUE, FALSE, FALSE, FALSE,  TRUE, FALSE, FALSE, FALSE,  TRUE,  TRUE, 
             FALSE,  TRUE,  TRUE,  TRUE,  TRUE, FALSE, FALSE, FALSE,
             FALSE, FALSE,  TRUE, FALSE,  TRUE, FALSE, FALSE, FALSE, 
             FALSE, FALSE, FALSE,  TRUE, FALSE,  TRUE,  TRUE)

testthat::test_that("flagAbsent column .occurrenceAbsent results correct series", {
  testthat::expect_equal(correct, testOut$.occurrenceAbsent)
})

#' Internal function. Create directories for saving the outputs of the bdc package
#'
#' Creates directories for saving the outputs of bdc package, namely
#' checks, figures, reports, and databases.
#'
#' @param path Character. A path as a character vector for where to create the directories. If 
#' no path is provided (the default), the directories will be created using [here::here()].
#'
#'
#' @importFrom here here
#'
#' @return None
#'
#' @details:
#' Function used to created four folder for saving results of some functions.
#' @noRd

#' @examples
#' \donttest{
#' jbd_create_dir(path = tempdir())
#' }
jbd_create_dir <- function(path = NULL) {
    # Option 1 if no path is provided, create directories using here::here
  if(is.null(path)){
    dir.create(here::here("Output/Check"), recursive = TRUE)
    dir.create(here::here("Output/Intermediate"), recursive = TRUE)
    dir.create(here::here("Output/Report"), recursive = TRUE)
    dir.create(here::here("Output/Figures"), recursive = TRUE)
  } # END option 1
  
  # Option 2 if a path is provided, create directories at that path
  if(!is.null(path) & !file.exists(path)){
    dir.create(paste0(path, "/Output/Check"), recursive = TRUE)
    dir.create(paste0(path, "/Output/Intermediate"), recursive = TRUE)
    dir.create(paste0(path, "/Output/Report"), recursive = TRUE)
    dir.create(paste0(path, "/Output/Figures"), recursive = TRUE)
  } # END option 2
  
} # END function

# This function was written by James Dorey estimate species richnesses using iChao and iNEXT
# This function was written starting on the 18th of February 2025 
# This function comes out of the article titled "How many bee species are there? A quantitative global estimate"
# at jbdorey[at]me.com

#' Estimate country, continental, and global species richnesses
#' 
#' Takes an output dataset from [BeeBDC::richnessPrepR()] to estimate species richness using
#' iChao and iNEXT (hill numbers) for countries, continents, and/or the entire globe. 
#'
#' @param data an RData file created using the [BeeBDC::richnessPrepR()] function.
#' @param sampleSize Numeric. The size of the sample randomly drawn from the provided curve.
#' See `curveFunction`. Default = 10000.
#' @param countrySamples Numeric. The number of times to sample the country species richness for both
#' iChao and iNEXT. If equal to zero (0), then this will not be analysed. Default = 5.
#' @param continentSamples Numeric. The number of times to sample the continent species richness for both
#' iChao and iNEXT. If equal to zero (0), then this will not be analysed. Default = 5.
#' @param globalSamples Numeric. The number of times to sample the global species richness for both
#' iChao and iNEXT. If equal to zero (0), then this will not be analysed. Default = 5.
#' @param filterToRecordedCountries Logical. If TRUE, the checklist will be filtered to the countries 
#' Where occurrence records were found. Default = TRUE. Change at your own peril.
#' @param countriesToExclude Character vector. You may decide to excluse some countries if they are 
#' being problematic or their sample sizes are too small. Default = NULL.
#' @param k Numeric. For iChao; the cut-off point (default = 10), which separates species into "abundant" and "rare" 
#' groups for abundance data for the estimator ACE; it separates species into "frequent" and 
#' "infrequent" groups for incidence data for the estimator ICE.
#' Default = 10.
#' @param outPath A directory as character. Directory where to save output figure. Default = tempdir().
#' @param fileName A character vector with file name 
#' for the output figure, ending with '.pdf'. Default = "continentSampled.pdf".
#' @param mc.cores Numeric. If > 1, the function will run in parallel
#' using mclapply using the number of cores specified. If = 1 then it will be run using a serial
#' loop. NOTE: Windows machines must use a value of 1 (see ?parallel::mclapply). Additionally,
#' be aware that each thread can use large chunks of memory.
#' Default = 1.
#' 
#' @return Outputs an R file with four tables ("Summary", "SiteOutput", "ContinentOutput", and 
#' "GlobalOutput"; depending on the number required). The summary table shows the Median overall 
#' estimates, while the remaining three shows the outputs from each iteration (useful for plotting, 
#' see relevant vignette). Some figures may also be saved to the selected outPath.
#' 
#' @importFrom dplyr %>%
#' 
#' @seealso [BeeBDC::countryHarmoniseR()] to harmonise country names based on a short list; 
#' [BeeBDC::richnessPrepR()] to produce the input data and for the required column names; as well 
#' as [BeeBDC::ChaoWrapper()] and [BeeBDC::iNEXTwrapper()] for the parallelised implementation of
#' `SpadeR` and `iNEXT` functions.
#' 
#' @export
#'
#' @examples
#' \dontrun{
#' 
#'   # Use the example data 
#' data(beesCountrySubset)
#' 
#'   # First, 
#' estimateDataExample <- BeeBDC::richnessPrepR(
#'   data = beesCountrySubset,
#'   # Download the taxonomy
#'   taxonomyFile = BeeBDC::beesTaxonomy(),
#'   # Download the checklist
#'   checklistFile = BeeBDC::beesChecklist(),
#'   curveFunction = function(x) (228.7531 * x * x^-log(12.1593)),
#'   sampleSize = 10000,
#'   countryColumn = "country_suggested",
#'   limitGlobal = NULL,
#'   outPath = tempdir()
#' )
#' 
#'  exampleEstimate <- richnessEstimateR(
#'    data = estimateDataExample,
#'    sampleSize = 10000,
#'    countrySamples = 1,
#'    continentSamples = 1,
#'    globalSamples = 1,
#'    filterToRecordedCountries = TRUE,
#'    mc.cores = 1,
#'    # Directory where to save files
#'    outPath = tempdir(),
#'    fileName = "Sampled.pdf"
#'  )
#'  
#' } # END dontrun
#' 


  # START THE REAL FUNCTION
richnessEstimateR <- function(
    data = NULL,
    sampleSize = 10000,
    countrySamples = 1,
    continentSamples = 1,
    globalSamples = 1,
    countriesToExclude = NULL,
    mc.cores = 1,
    k = 10,
    filterToRecordedCountries = TRUE,
    # Directory where to save files
    outPath = tempdir(),
    fileName = "continentSampled.pdf"
) {  
  # locally bind variables to the function
  . <- taxonomyAcceptedNames <- taxonomy_synonyms <- occData_counts <- NULL
  Name<- statistic<- Estimator<- `95% Lower`<- `95% Upper`<- `95%Upper`<- groupNo<- variable<-
    Observed<- Estimate<- `95%Lower`<- iNEXT_est<- observedRichness<- level<- niNEXT<- 
    iNEXT_lower<- iNEXT_upper<- iNEXT_increasePercent<- iNEXT_increase<- nChao<- iChao_est<- 
    iChao_lower<- iChao_upper<- Assemblage<- continent <- NULL
  rNaturalEarth_name <- combined_site_ChaoiNext <- combined_cont_ChaoiNext <- 
    combined_global_ChaoiNext <- output <- Est_s.e. <-n<- NULL
  
    #### 0.0 data prep ####
  ##### 0.1 Errors ####
  ###### a. FATAL errors ####
  if(is.null(data)){
    stop(" - Please provide an argument for data I'm a program not a magician.")
  }
  
  ##### 0.2 SpadeR test ####
  ###### a. test ####
  # Check if SpadeR is installed
  # TRUE if SpadeR is found
  suppressWarnings(
    suggestedTest <- system.file(package='SpadeR') %>% 
      stringr::str_count() > 0 
  )
  
  ###### b. SpadeR ####
  if(suggestedTest == FALSE){
    # Set up instructions for download on fail
    instructions <- paste(" Please try installing the package for yourself", 
                          "using the following command: \n",
                          "install.packages(\"SpadeR\")")
    # Set up fail function for tryCatch
    error_func <- function(e){
      stop(paste("Failed to install the SpadeR package.\n", 
                 instructions))
    }
    # Begin interactive input
    input <- 1
    if (interactive()){
      input <- utils::menu(c("Yes", "No"), 
                           title = paste0("Install the SpadeR package? \n"))
    }
    if(input == 1){
      # Start SpadeR install
      message("Installing the SpadeR package.")
      tryCatch(
        utils::install.packages("SpadeR"), 
        error = error_func, warning = error_func)
    } # END input == 1
    
    else{
      stop(writeLines(paste("The SpadeR package is necessary for BeeBDC::ChaoWrapper\n", 
                            instructions)))
    } # END else
  } # END suggestedTest == FALSE
  
  # Load required packages 
  requireNamespace("stringr")
  requireNamespace("dplyr")
  requireNamespace("iNEXT")
  
  

  
  
  ##### 0.3 iNEXT test ####
  ###### a. test ####
  # Check if iNEXT is installed
  # TRUE if iNEXT is found
  suppressWarnings(
    suggestedTest <- system.file(package='iNEXT') %>% 
      stringr::str_count() > 0 
  )
  
  ###### b. iNEXT ####
  if(suggestedTest == FALSE){
    # Set up instructions for download on fail
    instructions <- paste(" Please try installing the package for yourself", 
                          "using the following command: \n",
                          "install.packages(\"iNEXT\")")
    # Set up fail function for tryCatch
    error_func <- function(e){
      stop(paste("Failed to install the iNEXT package.\n", 
                 instructions))
    }
    # Begin interactive input
    input <- 1
    if (interactive()){
      input <- utils::menu(c("Yes", "No"), 
                           title = paste0("Install the iNEXT package? \n"))
    }
    if(input == 1){
      # Start iNEXT install
      message("Installing the iNEXT package.")
      tryCatch(
        utils::install.packages("iNEXT"), 
        error = error_func, warning = error_func)
    } # END input == 1
    
    else{
      stop(writeLines(paste("The iNEXT package is necessary for BeeBDC::iNEXTwrapper\n", 
                            instructions)))
    } # END else
  } # END suggestedTest == FALSE
  
  
  
  ##### 0.4 data prep ####
  # Building/testing data input
  richnessData <- data
  
  # Extract the datasets from data
  literatureCurve = richnessData$curveExtraction
  taxonomyAcceptedNames = richnessData$taxonomyAcceptedNames
  taxonomyNoPoints = richnessData$taxonomyNoPoints
  taxonomyFile = richnessData$taxonomyFile
  data_counts = richnessData$data_counts
  continentChecklist = richnessData$continentChecklist
  worldMap = richnessData$worldMap
  continentNoPoints = richnessData$continentNoPoints
  siteSpeciesCounts = richnessData$siteSpeciesCounts
  # combine some country names
  checklistFile = richnessData$checklistFile %>% 
    # Change country names
    countryHarmoniseR(countryColumn = "rNaturalEarth_name")
  
  # If no countriesToExclude is provided, make it an empty vector
  if(is.null(countriesToExclude)){
    countriesToExclude = c("")
  }
  
  ##### 0.5 filter to recorded countries ####
  if(filterToRecordedCountries == TRUE){
      # filter continent
    continentChecklist <- continentChecklist %>% 
      dplyr::filter(rNaturalEarth_name %in% siteSpeciesCounts$country_suggested)
      # Filter checklist 
    checklistFile <- checklistFile%>%
      dplyr::filter(rNaturalEarth_name %in% siteSpeciesCounts$country_suggested)

  }
  
  
  #### 1.0 functions ####
    ##### 1.2 richnessSampleR ####
  richnessSampleR <- function(
    # From X:Y iterations
    iterations = NULL,
    data = richnessData,
    # What scale?
    scale = NULL, # Country, Continent, or Global
    # Functions
    ChaoWrapper = ChaoWrapper,
    iNEXTwrapper = iNEXTwrapper
    
  ){
    rNaturalEarth_name <- validName <- country_suggested <- name <- name_long <- continent.x <- 
      continent.y <- scientificName <- dataFrom <- NULL
    #### 1.1.1 Prep ####

    # Extract the datasets from data
    literatureCurve = richnessData$curveExtraction
    taxonomyAcceptedNames = richnessData$taxonomyAcceptedNames
    taxonomyNoPoints = richnessData$taxonomyNoPoints
    taxonomyFile = richnessData$taxonomyFile
    data_counts = richnessData$data_counts
    continentChecklist = richnessData$continentChecklist
    worldMap = richnessData$worldMap
    continentNoPoints = richnessData$continentNoPoints
    siteSpeciesCounts = richnessData$siteSpeciesCounts
    
    # Fix some names in the siteSpeciesCounts
    siteSpeciesCounts <- siteSpeciesCounts %>% 
      countryHarmoniseR(countryColumn = "country_suggested")
    # combine some country names
    checklistFile = richnessData$checklistFile %>% 
      # Change country names
      countryHarmoniseR(countryColumn = "rNaturalEarth_name")
    if(filterToRecordedCountries == TRUE){
      # filter continent
      continentChecklist <- continentChecklist %>% 
        dplyr::filter(rNaturalEarth_name %in% siteSpeciesCounts$country_suggested)
      # Filter checklist 
      checklistFile <- checklistFile%>%
        dplyr::filter(rNaturalEarth_name %in% siteSpeciesCounts$country_suggested)
      
    }
    
    
    #### 1.1.2 Sample the curve ####
    # Sample the curve at the start of each iteration to generate new values
    # Get a random sample based on the empirical dataset
    litSample <- base::sample(x = literatureCurve$x_coords, prob = literatureCurve$y_coords, 
                        replace = TRUE,
                        size = sampleSize) %>%
      round(0) 
    
    

    # Find the species that are not represented in the occurrence dataset
    noPointSpecies <- taxonomyNoPoints  %>%
      # Add a sample size that is drawn from the same distriubtion of the literature and no-occurrences
      # dataset and generated in litSample
      dplyr::tibble(scientificName = .,
                    n = litSample[1:length(.)],
                    dataFrom = "randomisedSample")
    
    # 
    data_totalCounts <- data_counts %>% 
      dplyr::bind_rows(noPointSpecies) %>%
        # Drop NA points
      dplyr::filter(complete.cases(n))
    
    #### 1.1.3 Data prep ####
    ##### a. Country ####
    if(scale == "Country"){
      # Now, for those species-country combinations without ANY points, but that should have points
      # from the checklist, make n = a number drawn from the literature records distribution
      # and state that they came from the checklist
      country_speciesChecklistCounts <- checklistFile %>%
        dplyr::select(
          tidyselect::any_of(c("validName", "rNaturalEarth_name", "year", "author",
                              "authorCount"))) %>%
        dplyr::full_join(siteSpeciesCounts, by = c("validName" = "scientificName",
                                                       "rNaturalEarth_name" = "country_suggested")) %>%
        # Group by country and then split into a list per group
        dplyr::group_by(rNaturalEarth_name) %>%
        dplyr::group_split() %>%
        # For each country, apply the literature distribution but with an n max of the empirical maximum
        # for that country
        base::lapply(., countrySampler) %>%
        dplyr::bind_rows() %>%
        # Fix up the year, author, and authorCount columns by removing and re-adding them
        dplyr::select(!tidyselect::any_of(c("year", "author", "authorCount"))) %>% 
        dplyr::left_join(taxonomyFile %>%
                           dplyr::select(tidyselect::any_of(c("validName", "year", "author",
                                                            "authorCount"))),
                         by = "validName") %>%
        dplyr::distinct(validName, rNaturalEarth_name, .keep_all = TRUE)
      
      
      # Create the Chao input data including checklist data
      countryChaoData_checklist <- country_speciesChecklistCounts %>%
        dplyr::select(validName, rNaturalEarth_name, n) %>%
        tidyr::pivot_wider(names_from = rNaturalEarth_name,
                           values_from = n,
                           values_fill = 0) %>%
        ## Create the rownames
        column_to_rownames_internal("validName") %>%
        dplyr::tibble()
      
    } # END if country
    
    
    ##### b. Continent ####
    if(scale == "Continent"){
      # Turn the country occ data into a continent one
      continentOccs <- siteSpeciesCounts %>%
        dplyr::ungroup() %>%
        # Change some country names to better match the continent data
        countryHarmoniseR(countryColumn = "country_suggested") %>%
        # Join first by name
        dplyr::left_join(worldMap %>% dplyr::select(name, continent) %>%
                           countryHarmoniseR(countryColumn = "name") %>%
                           sf::st_drop_geometry(), by = c("country_suggested" = "name")) %>%
        # Then join by name_long
        dplyr::left_join(worldMap %>% dplyr::select(name_long, continent) %>%
                           countryHarmoniseR(countryColumn = "name_long") %>%
                           sf::st_drop_geometry(), by = c("country_suggested" = "name_long")) %>%
        # Now merge these continent columns
        dplyr::mutate(continent = dplyr::if_else(is.na(continent.x),
                                                 continent.y, continent.x)) %>%
        # drop interim and old count columns
        dplyr::select(!c("continent.x", "continent.y")) %>%
        dplyr::group_by(scientificName, continent) %>%
        dplyr::mutate(sum = sum(n)) %>%
        dplyr::mutate(name_continent = stringr::str_c(scientificName, continent, sep = "__"))
      
      
      # Combine the occurrence and checklist data 
      continentCounts <- continentOccs %>%
        dplyr::select(!base::c("country_suggested", "n", "name_continent")) %>%
        dplyr::rename(n = sum) %>%
        dplyr::bind_rows(continentNoPoints) %>%
        # Group by country and then split into a list per group
        dplyr::group_by(continent) %>%
        dplyr::group_split() %>%
        # For each country, apply the literature distribution but with an n max of the empirical maximum
        # for that country
        base::lapply(., countrySampler) %>%
        dplyr::bind_rows() %>%
        dplyr::distinct(scientificName, continent, .keep_all = TRUE) 
      
      # # Pivot wider 
      continentWider <- continentCounts %>%
        dplyr::select(scientificName, continent, n) %>%
        tidyr::drop_na() %>% 
        tidyr::pivot_wider(names_from = continent,
                           values_from = n,
                           values_fill = 0) %>%
        ## Create the rownames
        column_to_rownames_internal(var = "scientificName") %>%
        dplyr::tibble()
      
    } # END if continent
    
    
    
    #### 1.1.4 Analysis ####
    ##### a. Country ####
    if(tolower(scale) == "country"){
      ###### i. Chao ####
      suppressMessages(
      ChaoOut <- ChaoWrapper(data = countryChaoData_checklist,
                             k = 5,
                             datatype = "abundance",
                             conf = 0.95,
                             mc.cores = 1))
      
    } # END if country
    
    if(tolower(scale) == "country"){
      ###### ii. iNext ####
      # Run iNEXT for each country using the iNEXT wrapper
      suppressMessages(
      iNEXTout <- iNEXTwrapper(data = country_speciesChecklistCounts,
                               variableColumn = "rNaturalEarth_name",
                               valueColumn = "n",
                               datatype = "abundance",
                               mc.cores = 1))
    } # END if continent
    
    
    ##### b. Continent ####
    if(tolower(scale) == "continent"){
      ###### i. Chao ####
      suppressMessages(
      ChaoOut <- ChaoWrapper(data = continentWider,
                             k = 5,
                             datatype = "abundance",
                             conf = 0.95,
                             mc.cores = 1))
      
    } # END if country
    
    if(tolower(scale) == "continent"){
      ###### ii. iNext ####
      # Run iNEXT for each country using the iNEXT wrapper
      suppressMessages(
      iNEXTout <- iNEXTwrapper(data = continentCounts,
                               k = 5,
                               variableColumn = "continent",
                               datatype = "abundance",
                               conf = 0.95,
                               mc.cores = 1))
    } # END if continent
    
    
    ##### c. Global ####
    if(tolower(scale) == "global"){
      
      ###### i. iChao ####
      # Get the Chao values for global bee species diversity
      ChaoOut <- SpadeR::ChaoSpecies(data_totalCounts$n,
                                     datatype = "abundance", 
                                     k = 10, conf = 0.95)
      
      ##### ii. iNEXT ####
      global_iNEXT <- data_totalCounts %>%
        dplyr::select(scientificName, n) %>%   # Create the rownames
        column_to_rownames_internal(var = "scientificName") %>%
        dplyr::tibble()
      
      iNEXTout <- iNEXT::iNEXT(x = global_iNEXT$n, datatype = "abundance",
                               # 0 for species richness Shannon diversity (q = 1, the exponential of Shannon entropy) and Simpson diversity (q = 2, the inverse of Simpson concentration).
                               q = 0)
    } # END if global
    
    # Return a list of the objects produced
    return(dplyr::lst(ChaoOut, iNEXTout) %>%
             stats::setNames(c("ChaoOut", "iNEXTout")) %>%
             dplyr::lst() %>%
             # Set the name as the iteration number 
             stats::setNames(iterations))
  } # END richnessSampleR function
  
  
  ##### 1.3 countrySampler ####
  
  # Make a function to sample the country sample sizes
  countrySampler <- function(df = NULL,
                             iterations = NULL){ 
    dataFrom <- NULL
    # Set a random seed
    #set.seed(iterations*7)
    # Get the maximum value for literatureCurve$x_coords
    max_x <- base::max(literatureCurve$x_coords)
    # Test that there is sufficient length (i.e., at least one records with n > NA)
    lengthTest <- df$n[complete.cases(df$n)] %>% 
      length()
    # Get the maximum value for the country's sample size or set the minium to n = 1
    if(lengthTest > 0){
      max_n = base::max(df$n[complete.cases(df$n)])
    }else{
      max_n = 1
    }
    # Find the length of the indput data
    inputLength <- length(df$n) + 100
    # Take the smaller of these two values
    maxSample <- dplyr::if_else(max_x > max_n, max_n, max_x)
    # Now apply the maxxed distribution to the country's checklist data
    df <- df %>%
      dplyr::mutate(n = dplyr::if_else(is.na(n), 
                                       # Draw from the literature distribution
                                       sample(x = literatureCurve$x_coords[1:maxSample],
                                              prob = literatureCurve$y_coords[1:maxSample], 
                                              replace = TRUE,
                                              size = inputLength)[dplyr::row_number()] %>%
                                         round(0),
                                       # Or keep n
                                       n),
                    dataFrom = dplyr::if_else(is.na(dataFrom), "checklist", dataFrom))
    return(df)
  }# END countrySampler
  
  
  # Use the richnessSampleR function to iteratively generate richness estimates
#### 2.0 Global ####
  if(globalSamples > 0){
  globalSampled <- parallel::mclapply(
    X = 1:globalSamples,
    FUN = richnessSampleR,
    # FUNCTION INPUTS:
    # Input datasets
    data = richnessData,
    # What scale?
    scale = "Global", # Country, Continent, or Global
    # Functions
    ChaoWrapper = ChaoWrapper,
    iNEXTwrapper = iNEXTwrapper,
    mc.cores = mc.cores
  )
  
  counter <- 1
  # Extract all of the Chao data
  ChaoGlobalSampled <- globalSampled %>%
    lapply(X = .,
           function(x){
             rowname <- V1 <- NULL
             counter <<- counter + 1
             extraction <- x[[1]]$ChaoOut$Species_table %>%
               base::as.data.frame() %>%
               rownames_to_column_int(var = "Name") %>% 
               dplyr::mutate(Name = stringr::str_squish(Name))
             # Extract the sample size
             sampleSize <- x[[1]]$ChaoOut$Basic_data_information %>% 
               rownames_to_column_int(var = "Name") %>% 
               dplyr::mutate(Name = stringr::str_squish(Name)) %>%
               base::t() %>% base::as.data.frame() %>% rownames_to_column_int() %>% 
               dplyr::tibble() %>%
               dplyr::select(rowname, V1) %>% dplyr::filter(!V1 %in% c("Sample size", "n")) %>%
               dplyr::rename(n = V1)
             # Add the sample size 
             extraction <- extraction %>% 
               dplyr::mutate(n = sampleSize$n)
           }) 
  # Combine these together
  global_iChaoSamples <- ChaoGlobalSampled %>%
    dplyr::bind_rows() %>%
    # Select the desired statistic estimate
    dplyr::filter(stringr::str_detect(Name, "iChao1 \\(Chiu et al\\. 2014\\)")) %>%
    dplyr::mutate(Name = "iChao") %>% 
    dplyr::mutate(variable = "Global", .before = 1)
  
  # Extract all of the iNEXT data
  counter <- 1
  iNEXTGlobalSampled <- globalSampled %>%
    lapply(X = .,
           function(x){
             counter <<- counter + 1
             extraction <- x[[1]]$iNEXTout$AsyEst %>%
               dplyr::mutate(n = x[[1]]$iNEXTout$DataInfo$n %>%
                               as.character())
           }) 
  
  # Combine these together
  global_iNextSamples <- iNEXTGlobalSampled %>%
    dplyr::bind_rows() %>%
    dplyr::mutate(statistic = rownames(.)) %>% 
    # Select the desired statistic estimate
    dplyr::filter(stringr::str_detect(statistic, "Species Richness")) %>%
    dplyr::mutate(statistic = "iNEXT") %>% 
    # Rename to match with iChao
    dplyr::rename(Name = statistic,
                  Estimate = Estimator,
                  `95%Lower` = `95% Lower`,
                  `95%Upper` = `95% Upper`,
                  `s.e.` = `Est_s.e.`) %>%
    dplyr::mutate(variable = "Global", .before = 1)
  
  combined_global_ChaoiNext <- global_iNextSamples %>%
    dplyr::bind_rows(global_iChaoSamples) %>%
    dplyr::arrange( `95%Upper`) %>%
    # Add a column with group numbers, and then make them into groups of ten
    dplyr::mutate(groupNo = base::ceiling(dplyr::cur_group_id()/10)) %>%
    dplyr::group_by(groupNo)
  
  
  # Combine the data to calculate further statistics
  # Combine the iNEXT data
  globalcombined_iNEXT <- combined_global_ChaoiNext %>%
    dplyr::filter(Name == "iNEXT") %>%
    # Group by country
    dplyr::group_by(variable) %>%
    # Get the stats::medians of the sampled values
    dplyr::mutate(observedRichness = Observed,
                  niNEXT = stats::median(n %>% as.numeric(.)),
                  iNEXT_est = stats::median(Estimate),
                  iNEXT_lower = stats::median(`95%Lower`),
                  iNEXT_upper = stats::median(`95%Upper`),
                  iNEXT_increasePercent = ((iNEXT_est/observedRichness)-1)*100,
                  iNEXT_increase = iNEXT_est - observedRichness,
                  level = "Global") %>%
    dplyr::select(variable, observedRichness, level, niNEXT,
                  iNEXT_est, iNEXT_lower, iNEXT_upper, iNEXT_increasePercent, iNEXT_increase) %>%
    dplyr::distinct()
  
  # Combine the iChao data
  globalcombined_iChao <- combined_global_ChaoiNext %>%
    dplyr::filter(Name == "iChao") %>%
    # Group by country
    dplyr::group_by(variable) %>%
    # Get the stats::medians of the sampled values
    dplyr::mutate(nChao = stats::median(as.numeric(n)),
                  iChao_est = stats::median(Estimate),
                  iChao_lower = stats::median(`95%Lower`),
                  iChao_upper = stats::median(`95%Upper`)) %>%
    dplyr::select(variable, nChao, iChao_est, iChao_lower, iChao_upper) %>%
    dplyr::distinct() 
  
  # Combine them all
  combinedStatistics_sampled_global <- globalcombined_iChao %>%
    dplyr::left_join(globalcombined_iNEXT, by = "variable") %>% 
    # Calculate the iChao increases
    dplyr::mutate(iChao_increasePercent = ((iChao_est/observedRichness)-1)*100,
                  iChao_increase = iChao_est - observedRichness)
  
  } # END 2.0 Global
  
  
  
#### 3.0 Country ####
  if(countrySamples > 0){
  countrySampled <- parallel::mclapply(
    X = 1:countrySamples,
    FUN = richnessSampleR,
    # FUNCTION INPUTS:
    # Input datasets
    data = richnessData,
    # What scale?
    scale = "Country", # Country, Continent, or Global
    # Functions
    ChaoWrapper = ChaoWrapper,
    iNEXTwrapper = iNEXTwrapper,
    mc.cores = mc.cores
  )
  
  counter <- 1
  # Extract all of the Chao data
  ChaoCountrySampled <- countrySampled %>%
    lapply(X = .,
           function(x){
             V1 <- rowname<- NULL
             counter <<- counter + 1
             extraction <- x[[1]]$ChaoOut$richnessTable %>%
               dplyr::mutate(Name = stringr::str_squish(Name))
             # Extract the sample size
             sampleSize <- x[[1]]$ChaoOut$basicTable %>% 
               t() %>% base::as.data.frame() %>% 
               dplyr::mutate(rowname = rownames(.)) %>% 
               dplyr::tibble() %>%
               dplyr::select(rowname, V1) %>% dplyr::filter(!V1 %in% c("Sample size", "n")) %>%
               dplyr::rename(n = V1)
             # Add the sample size 
             extraction <- extraction %>% 
               dplyr::left_join(sampleSize, by = c("variable" = "rowname"))
           }) 
  # Combine these together
  country_iChaoSamples <- ChaoCountrySampled %>%
    dplyr::bind_rows() %>%
    # Select the desired statistic estimate
    dplyr::filter(stringr::str_detect(Name, "iChao1 \\(Chiu et al\\. 2014\\)")) %>%
    dplyr::mutate(Name = "iChao") %>% 
    # Group by continent
    dplyr::group_by(variable)
  
  # Extract all of the iNEXT data
  counter <- 1
  iNEXTCountrySampled <- countrySampled %>%
    lapply(X = .,
           function(x){
             counter <<- counter + 1
             extraction <- x[[1]]$iNEXTout$AsyEst 
             extraction <- extraction %>% 
               dplyr::left_join(x[[1]]$iNEXTout$DataInfo %>% 
                                  dplyr::select(Assemblage, n) %>%
                                  dplyr::mutate(n = as.character(n)),
                                by = c("rNaturalEarth_name" = "Assemblage"))
             
           }) 
  
  # Combine these together
  country_iNextSamples <- iNEXTCountrySampled %>%
    dplyr::bind_rows() %>%
    # Select the desired statistic estimate
    dplyr::filter(stringr::str_detect(statistic, "Species Richness")) %>%
    dplyr::mutate(statistic = "iNEXT") %>% 
    # Group by rNaturalEarth_name
    dplyr::group_by(rNaturalEarth_name) %>%
    # Rename to match with iChao
    dplyr::rename(variable = rNaturalEarth_name,
                  Name = statistic,
                  Estimate = Estimator,
                  `95%Lower` = `95% Lower`,
                  `95%Upper` = `95% Upper`,
                  `s.e.` = `Est_s.e.`)
  
  combined_site_ChaoiNext <- country_iNextSamples %>%
    dplyr::bind_rows(country_iChaoSamples) %>%
    # Remove low sample size or problematic countries
    dplyr::filter(!variable %in% countriesToExclude) %>% 
    dplyr::filter(!stringr::str_detect(variable, "Northern" # for Northern Cyprus
    )) %>%
    dplyr::arrange( `95%Upper`) %>%
    # Add a column with group numbers, and then make them into groups of ten
    dplyr::mutate(groupNo = base::ceiling(dplyr::cur_group_id()/10)) %>%
    dplyr::group_by(groupNo)
  
  suppressWarnings({
  # Make a plot of the iChao values and the confidence intervals
  (count_sampledPlot <- ggplot2::ggplot(data = combined_site_ChaoiNext) + 
      ggplot2::geom_violin(position="dodge", alpha=0.5,
                           ggplot2::aes(fill=Name, 
                                        y=`95%Lower`, x=variable), colour =  NA) +
      ggplot2::geom_violin(position="dodge", alpha=0.5,
                           ggplot2::aes(fill=Name,
                                        y=`95%Upper`, x=variable), colour =  NA) +
      ggplot2::geom_violin(position="dodge", alpha=1,
                           ggplot2::aes(fill=Name,
                                        y=Estimate, x=variable), colour =  "black") +
      ggplot2::scale_fill_manual(values=c("#55AD9B", "#FD9B63")) +
      ggplot2::theme_classic() + ggplot2::xlab("Country") + ggplot2::ylab("iChao") +
      ggplot2::facet_wrap( ggplot2::vars(variable), scales = "free", ncol = 5)+ 
      ggplot2::theme(
        strip.background = ggplot2::element_blank(),
        strip.text.x = ggplot2::element_blank()
      )
  )
  # Save the plot
  ggplot2::ggsave(file = paste0("country_", fileName), 
                  path = outPath,
                  plot = count_sampledPlot, device = "pdf",
                  width = 10, height = 50,
                  limitsize = FALSE)
  })
  
  # Combine the data to calculate further statistics
  # Combine the iNEXT data
  countrycombined_iNEXT <- combined_site_ChaoiNext %>%
    dplyr::filter(Name == "iNEXT") %>%
    # Group by country
    dplyr::group_by(variable) %>%
    # Get the stats::medians of the sampled values
    dplyr::mutate(observedRichness = Observed,
                  niNEXT = stats::median(n %>% as.numeric(.)),
                  iNEXT_est = stats::median(Estimate),
                  iNEXT_lower = stats::median(`95%Lower`),
                  iNEXT_upper = stats::median(`95%Upper`),
                  iNEXT_increasePercent = ((iNEXT_est/observedRichness)-1)*100,
                  iNEXT_increase = iNEXT_est - observedRichness,
                  level = "Country") %>%
    dplyr::select(variable, observedRichness, level, niNEXT,
                  iNEXT_est, iNEXT_lower, iNEXT_upper, iNEXT_increasePercent, iNEXT_increase) %>%
    dplyr::distinct()
  
  # Combine the iChao data
  countrycombined_iChao <- combined_site_ChaoiNext %>%
    dplyr::filter(Name == "iChao") %>%
    # Group by country
    dplyr::group_by(variable) %>%
    # Get the stats::medians of the sampled values
    dplyr::mutate(nChao = stats::median(as.numeric(n)),
                  iChao_est = stats::median(Estimate),
                  iChao_lower = stats::median(`95%Lower`),
                  iChao_upper = stats::median(`95%Upper`)) %>%
    dplyr::select(variable, nChao, iChao_est, iChao_lower, iChao_upper) %>%
    dplyr::distinct()
  
  # Combine them all
  combinedStatistics_sampled_country <- countrycombined_iChao %>%
    dplyr::left_join(countrycombined_iNEXT, by = "variable") %>% 
    # Calculate the iChao increases
    dplyr::mutate(iChao_increasePercent = ((iChao_est/observedRichness)-1)*100,
                  iChao_increase = iChao_est - observedRichness)
  
  
  } # END 3.0 Country
  
#### 4.0 Continent ####
  if(continentSamples > 0){
  # Iteratively sample the continent-level richness
  continentSampled <- parallel::mclapply(
    X = 1:continentSamples,
    FUN = richnessSampleR,
    # FUNCTION INPUTS:
    # Input datasets
    data = richnessData,
    # What scale?
    scale = "Continent", # Country, Continent, or Global
    # Functions
    ChaoWrapper = ChaoWrapper,
    iNEXTwrapper = iNEXTwrapper,
    mc.cores = mc.cores
  )
  
  counter <- 1
  # Extract all of the Chao data
  ChaoContinentSampled <- continentSampled %>%
    lapply(X = .,
           function(x){
             V1 <- rowname <- NULL
             counter <<- counter + 1
             extraction <- x[[1]]$ChaoOut$richnessTable
             # Extract the sample size
             sampleSize <- x[[1]]$ChaoOut$basicTable %>% 
               t() %>% base::as.data.frame() %>% 
               dplyr::mutate(rowname = rownames(.)) %>% 
               dplyr::tibble() %>%
               dplyr::select(rowname, V1) %>% dplyr::filter(!V1 %in% c("Sample size", "n")) %>%
               dplyr::rename(n = V1)
             # Add the sample size 
             extraction <- extraction %>% 
               dplyr::left_join(sampleSize, by = c("variable" = "rowname"))
           }) 
  # Combine these together
  continent_iChaoSamples <- ChaoContinentSampled %>%
    dplyr::bind_rows() %>%
    # Select the desired statistic estimate
    dplyr::filter(stringr::str_detect(Name, "iChao1 \\(Chiu et al\\. 2014\\)")) %>%
    dplyr::mutate(Name = "iChao") %>% 
    # Group by continent
    dplyr::group_by(variable)
  
  # Extract all of the iNEXT data
  counter <- 1
  iNEXTContinentSampled <- continentSampled %>%
    lapply(X = .,
           function(x){
             counter <<- counter + 1
             extraction <- x[[1]]$iNEXTout$AsyEst 
             extraction <- extraction %>% 
               dplyr::left_join(x[[1]]$iNEXTout$DataInfo %>% 
                                  dplyr::select(Assemblage, n) %>%
                                  dplyr::mutate(n = as.character(n)),
                                by = c("continent" = "Assemblage"))
             
           }) 
  
  # Combine these together
  continent_iNextSamples <- iNEXTContinentSampled %>%
    dplyr::bind_rows() %>%
    # Select the desired statistic estimate
    dplyr::filter(stringr::str_detect(statistic, "Species Richness")) %>%
    dplyr::mutate(statistic = "iNEXT") %>% 
    # Group by continent
    dplyr::group_by(continent) %>%
    # Rename to match with iChao
    dplyr::rename(variable = continent,
                  Name = statistic,
                  Estimate = Estimator,
                  `95%Lower` = `95% Lower`,
                  `95%Upper` = `95% Upper`,
                  `s.e.` = `Est_s.e.`)
  
  combined_cont_ChaoiNext <- continent_iChaoSamples %>%
    dplyr::bind_rows(continent_iNextSamples)
  
  
  # Combine the data to calculate further statistics
  # Combine the iNEXT data
  contCombined_iNEXT <- combined_cont_ChaoiNext %>%
    dplyr::filter(Name == "iNEXT") %>%
    # Group by country
    dplyr::group_by(variable) %>%
    # Get the stats::medians of the sampled values
    dplyr::mutate(observedRichness = Observed,
                  niNEXT = stats::median(n %>% as.numeric(.)),
                  iNEXT_est = stats::median(Estimate),
                  iNEXT_lower = stats::median(`95%Lower`),
                  iNEXT_upper = stats::median(`95%Upper`),
                  iNEXT_increasePercent = ((iNEXT_est/observedRichness)-1)*100,
                  iNEXT_increase = iNEXT_est - observedRichness,
                  level = "Continent") %>%
    dplyr::select(variable, observedRichness, level, niNEXT,
                  iNEXT_est, iNEXT_lower, iNEXT_upper, iNEXT_increasePercent, iNEXT_increase) %>%
    dplyr::distinct()
  
  # Combine the iChao data
  contCombined_iChao <- combined_cont_ChaoiNext %>%
    dplyr::filter(Name == "iChao") %>%
    # Group by country
    dplyr::group_by(variable) %>%
    # Get the stats::medians of the sampled values
    dplyr::mutate(nChao = stats::median(as.numeric(n)),
                  iChao_est = stats::median(Estimate),
                  iChao_lower = stats::median(`95%Lower`),
                  iChao_upper = stats::median(`95%Upper`)) %>%
    dplyr::select(variable, nChao, iChao_est, iChao_lower, iChao_upper) %>%
    dplyr::distinct()
  
  # Combine them all
  combinedStatistics_sampled_continent <- contCombined_iChao %>%
    dplyr::left_join(contCombined_iNEXT, by = "variable") %>% 
    # Calculate the iChao increases
    dplyr::mutate(iChao_increasePercent = ((iChao_est/observedRichness)-1)*100,
                  iChao_increase = iChao_est - observedRichness)
  
  suppressWarnings({
  # Make a plot of the iChao values and the confidence intervals
  (cont_sampledPlot <- ggplot2::ggplot(data = combined_cont_ChaoiNext) + 
      ggplot2::geom_violin(position="dodge", alpha=0.5,
                           ggplot2::aes(fill=Name, 
                                        y=`95%Lower`, x=variable)#,
                           #fill = "blue"
                           , colour =  NA) +
      ggplot2::geom_violin(position="dodge", alpha=0.5,
                           ggplot2::aes(fill=Name,
                                        y=`95%Upper`, x=variable)#,
                           #fill = "red"
                           , colour =  NA) +
      ggplot2::geom_violin(position="dodge", alpha=1,
                           ggplot2::aes(fill=Name,
                                        y=Estimate, x=variable),
                           #fill = "purple"
                           colour =  "black") +
      ggplot2::scale_fill_manual(values=c("#55AD9B", "#FD9B63")) +
      ggplot2::theme_classic() + ggplot2::xlab("Continent") + ggplot2::ylab("Species estimate") +
      ggplot2::guides(fill= ggplot2::guide_legend(title="Statistic"))) 
  # Save the plot
  ggplot2::ggsave(file = paste0( "continent_", fileName),
                  path = outPath,
                  plot = cont_sampledPlot, device = "pdf",
                  width = 7, height = 5)
})

} # END 4.0 Continent

  
#### 5.0 Combine data ####
    ###### 5.1 missing sheets ####
  if(!exists("combinedStatistics_sampled_country")){
    combinedStatistics_sampled_country <- dplyr::tibble(
      variable = NA_character_, nChao = NA_integer_, iChao_est = NA_integer_,
      iChao_lower = NA_integer_,iChao_upper = NA_integer_,          
      observedRichness = NA_integer_,level = NA_character_,niNEXT = NA_integer_,
      iNEXT_est = NA_integer_,iNEXT_lower = NA_integer_, iNEXT_upper = NA_integer_,
      iNEXT_increasePercent = NA_integer_,iNEXT_increase = NA_integer_,
      iChao_increasePercent = NA_integer_,iChao_increase  = NA_integer_) %>%
      tidyr::drop_na()
  }
  
  if(!exists("combinedStatistics_sampled_continent")){
    combinedStatistics_sampled_continent <- dplyr::tibble(
      variable = NA_character_, nChao = NA_integer_, iChao_est = NA_integer_,
      iChao_lower = NA_integer_,iChao_upper = NA_integer_,          
      observedRichness = NA_integer_,level = NA_character_,niNEXT = NA_integer_,
      iNEXT_est = NA_integer_,iNEXT_lower = NA_integer_, iNEXT_upper = NA_integer_,
      iNEXT_increasePercent = NA_integer_,iNEXT_increase = NA_integer_,
      iChao_increasePercent = NA_integer_,iChao_increase  = NA_integer_) %>%
      tidyr::drop_na()
  }
  if(!exists("combinedStatistics_sampled_global")){
    combinedStatistics_sampled_global <- dplyr::tibble(
      variable = NA_character_, nChao = NA_integer_, iChao_est = NA_integer_,
      iChao_lower = NA_integer_,iChao_upper = NA_integer_,          
      observedRichness = NA_integer_,level = NA_character_,niNEXT = NA_integer_,
      iNEXT_est = NA_integer_,iNEXT_lower = NA_integer_, iNEXT_upper = NA_integer_,
      iNEXT_increasePercent = NA_integer_,iNEXT_increase = NA_integer_,
      iChao_increasePercent = NA_integer_,iChao_increase  = NA_integer_) %>%
      tidyr::drop_na()
  }
  
  
    ##### 5.2 Combine all ####
  # Combine the iterative sampled data above
  combinedStatistics <- dplyr::bind_rows(combinedStatistics_sampled_country %>%
                                           dplyr::mutate(scale = "Country", .after = 1),
                                         combinedStatistics_sampled_continent %>%
                                           dplyr::mutate(scale = "Continent", .after = 1),
                                         combinedStatistics_sampled_global %>%
                                           dplyr::mutate(scale = "Global", .after = 1)
  ) %>%
    dplyr::rename(name = variable)
  
    # Combine all of the outputs into a list for return from the function
  output <- list(combinedStatistics, 
       if(exists("combined_site_ChaoiNext")){combined_site_ChaoiNext},
       if(exists("combined_cont_ChaoiNext")){combined_cont_ChaoiNext},
       if(exists("combined_global_ChaoiNext")){combined_global_ChaoiNext}) %>% 
    stats::setNames(c("Summary", "SiteOutput", "ContinentOutput", "GlobalOutput"))
  
  # Return the output 
  return(output)

} # End function

# This function was created by James Dorey on the 26th of May 2022. It will attempt to find dates
  # that dont occur in the EventDate column and restore them to avoid losing those occurrences in 
  # filtering. 
# For questions, ask James Dorey at jbdorey[at]me.com

#' Find dates in other columns
#' 
#' A function made to search other columns for dates and add them to the eventDate column. 
#' The function searches the columns locality, fieldNotes, locationRemarks, and verbatimEventDate 
#' for the relevant information. Additionally, for date ranges in the eventDate column, these will
#' Be translated into startDayOfYear and endDayOfYear and the eventDate will be kept as the last 
#' date, but formatted in the correct format. Ambiguous dates that can't be parsed reliably as 
#' day-month-year or month-day-year will be rounded to the nearest certainty; if month or day can
#' be identified with a spelled-out month or a day >12, these wil be correctly formatted. 
#' Year-month-day... is the most-reliable format
#'
#' @param data A data frame or tibble. Occurrence records as input.
#' @param maxYear Numeric. The maximum year considered reasonable to find. 
#' Default = lubridate::year(Sys.Date()).
#' @param minYear Numeric. The minimum year considered reasonable to find. Default = 1700.
#'
#' @importFrom stats complete.cases setNames
#'
#' @return The function results in the input occurrence data with but with updated eventDate, year, 
#' month, and day columns for occurrences where these data were a) missing and b) located in one of the 
#' searched columns.
#' 
#' @export
#' 
#' @importFrom dplyr %>%
#'
#' @examples
#' # Using the example dataset, you may not find any missing eventDates are rescued (dependent on 
#' # which version of the example dataset the user inputs.
#' beesRaw_out <- dateFindR(data = beesRaw,
#'                          # Years above this are removed (from the recovered dates only)
#'                          maxYear = lubridate::year(Sys.Date()),
#'                          # Years below this are removed (from the recovered dates only)
#'                          minYear = 1700)

dateFindR <- function(data = NULL,
           maxYear = lubridate::year(Sys.Date()),
           minYear = 1700) {
    # locally bind variables to the function
    eventDate<-database_id<-.<-verbatimEventDate<-fieldNotes<-locationRemarks<-ymd_vEV<-
      ymd_fieldNotes<-ymd_locationRemarks<-locality<-dmy_vEV<-dmy_locality<-dmy_fieldNotes<-
      dmy_locationRemarks<-mdy_vEV<-mdy_locality<-mdy_fieldNotes<-mdy_locationRemarks<-my_vEV<-
      my_locality<-my_fieldNotes<-my_locationRemarks<-amb_vEV<-amb_locality<-amb_fieldNotes<-
      amb_locationRemarks<-year <- endTime <- startTime <- originalDateCount <- 
      eventDate_in <- day <- startDayOfYear <- endDayOfYear <- ymd_eventDate_in <- 
      dmy_eventDate_in <- mdy_eventDate_in <- NULL
    
      # load required packages
    requireNamespace("dplyr")
    requireNamespace("lubridate")
    requireNamespace("mgsub")
    
    timeStart <- Sys.time()
    
    #### 0.0 prep ####
      ##### 0.1 Functions ####
      # Make a function to deal with ranges
    rangeFinder <- function(rangeData = NULL, rangeColumn = "eventDate_in",
                            timeStrings_in = timeStrings_robust){
      dateCol <- dateStart <- dateEnd <- startDayOfYear <- startDate <- endDayOfYear <- endDate <- NULL
      
      # Check and add startDayOfYear and endDayOfYear
      if(!"startDayOfYear" %in% colnames(rangeData)){
        rangeData <- rangeData %>%
          dplyr::mutate(startDayOfYear = NA_integer_)
      }
      if(!"endDayOfYear" %in% colnames(rangeData)){
        rangeData <- rangeData %>%
          dplyr::mutate(endDayOfYear = NA_integer_)
      }
      
      # Temporarily rename the rangeColumn to "dateCol" within the function
      rangeData <- rangeData %>%
        dplyr::rename("dateCol" = tidyselect::any_of(rangeColumn)) %>% 
          # Make sure that dateCol is a character column
        dplyr::mutate(dateCol = as.character(dateCol))
      
        # Sort out the ranges
      rangeData <- rangeData %>% 
      # DATE START
      dplyr::mutate(
        dateStart = dplyr::if_else(# Extract first for...
          # /
          stringr::str_detect(dateCol, dateRangeStrings_forwardSlash),
          stringr::str_extract(dateCol, ".*/") %>% stringr::str_remove("\\/"),
          # , 
          dplyr::if_else(stringr::str_detect(dateCol, dateRangeStrings_comma),
                         stringr::str_extract(dateCol, ".*,") %>% stringr::str_remove("\\,"),
                         # to
                         dplyr::if_else(stringr::str_detect(dateCol, dateRangeStrings_to),
                                        stringr::str_extract(dateCol, ".*to ") %>% stringr::str_remove("to"),
                                        # -
                                        dplyr::if_else(stringr::str_detect(dateCol, dateRangeStrings_dash),
                                                       stringr::str_extract(dateCol, ".*-") %>% stringr::str_remove("\\-"),
                                                       # startDateTime:
                                                       dplyr::if_else(stringr::str_detect(dateCol, dateRangeStrings_StEn),
                                                                      stringr::str_extract(dateCol, ".*EndDateTime:|.*endDateTime:") %>% 
                                                                        stringr::str_remove_all("[a-zA-Z]+|\\:|\\;|\\s"),
                                                                      dateCol))))),
        dateStart = stringr::str_squish(dateStart),
        .after = dateCol) %>%
        # DATE END
        dplyr::mutate(
          dateEnd = dplyr::if_else(
            # Extract first for...
            # /
            stringr::str_detect(dateCol, dateRangeStrings_forwardSlash),
            stringr::str_extract(dateCol, "/.*") %>% stringr::str_remove("\\/"),
            # , 
            dplyr::if_else(stringr::str_detect(dateCol, dateRangeStrings_comma),
                           stringr::str_extract(dateCol, ",.*") %>% stringr::str_remove("\\,"),
                           # to
                           dplyr::if_else(stringr::str_detect(dateCol, dateRangeStrings_to),
                                          stringr::str_extract(dateCol, " to.*") %>% stringr::str_remove("to"),
                                          # -
                                          dplyr::if_else(stringr::str_detect(dateCol, dateRangeStrings_dash),
                                                         stringr::str_extract(dateCol, "-.*") %>% stringr::str_remove("\\-"),
                                                         # EndDateTime:
                                                         dplyr::if_else(stringr::str_detect(dateCol, dateRangeStrings_StEn),
                                                                        stringr::str_extract(dateCol, "EndDateTime:.*|endDateTime:.*") %>%
                                                                          stringr::str_remove_all("[a-zA-Z]+|\\:|\\;|\\s"),
                                                                        dateCol))))),
          dateEnd = stringr::str_squish(dateEnd),
          .after = dateStart) %>%
        # IF the dateStart date has only day or only day month...
        dplyr::mutate(# DAY ONLY 
          dateStart = dplyr::if_else(stringr::str_detect(dateStart, "^[0-9]{1,2}$"),
                                     # Add in the end of dateEnd — TRUE1
                                     stringr::str_c(dateStart, 
                                                    stringr::str_remove(dateEnd,"^[0-9]{1,2}")),
                                     # DAY MONTH — FALSE
                                     dplyr::if_else(stringr::str_detect(dateStart, 
                                                                        stringr::str_c("^[0-9]{1,2}[\\s\\/\\-]", monthStrings,"$",
                                                                                       collapse = "|")),
                                                    # Add in the end of dateEnd — TRUE
                                                    stringr::str_c(dateStart, 
                                                                   stringr::str_remove(dateEnd,
                                                                                       stringr::str_c("^[0-9]{1,2}[\\s\\/\\-]", monthStrings,
                                                                                                      collapse = "|"))),
                                                    # FALSE
                                                    dateStart)
          ),
          .after = dateStart) %>% # END mutate
        # swap Roman numerals
        dplyr::mutate(
          dateEnd = mgsub::mgsub(dateEnd,
                                 pattern = paste0("[-/ \\.]", romanNumerals, "[ -/\\.]"),
                                 replacement = paste0(" ", numeralConversion, " "))  %>% 
            stringr::str_squish() %>% 
            # remove "Sept."
            stringr::str_replace(pattern = "[Ss]ept[\\s-/\\.]",replacement = "Sep "),
          dateStart = mgsub::mgsub(dateStart,
                                   pattern = paste0("[-/ \\.]", romanNumerals, "[ -/\\.]"),
                                   replacement = paste0(" ", numeralConversion, " "))  %>% 
            stringr::str_squish() %>% 
            # remove "Sept."
            stringr::str_replace(pattern = "[Ss]ept[\\s-/\\.]",replacement = "Sep "),
          eventDate = mgsub::mgsub(dateEnd,
                                   pattern = paste0("[-/ \\.]", romanNumerals, "[ -/\\.]"),
                                   replacement = paste0(" ", numeralConversion, " "))  %>% 
            stringr::str_squish() %>% 
            # remove "Sept."
            stringr::str_replace(pattern = "[Ss]ept[\\s-/\\.]", replacement = "Sep ")
        ) %>% 
        # Guess the formats
        dplyr::mutate(
          endDate = dateEnd %>% lubridate::parse_date_time(., timeStrings_in, quiet = TRUE) %>%
            lubridate::yday(),
          startDate = dateStart %>% lubridate::parse_date_time(., timeStrings_in, quiet = TRUE) %>%
            lubridate::yday(),
          eventDate = eventDate %>% lubridate::parse_date_time(., timeStrings_in, quiet = TRUE)
        ) %>%
        dplyr::mutate(
          startDayOfYear = dplyr::if_else(is.na(startDayOfYear), startDate, startDayOfYear),
          endDayOfYear = dplyr::if_else(is.na(endDayOfYear), endDate, startDayOfYear)
        ) %>%
        dplyr::select(!tidyselect::any_of(c("endDate", "startDate", "dateEnd", "dateStart")))
      # Return the rangeColumn name to it's original state
      names(rangeData)[names(rangeData) == "dateCol"] <- rangeColumn
      return(rangeData)
    }
    
      ##### 0.2 Strings ####
    ###### a. month strings ####
    monthStrings <- c("september","october","november","december",
                      "JAN", "FEB", "MAR", "APR", "MAY", "JUN",
                      "JUL", "AUG", "SEP", "OCT","NOV", "DEC",
                      "JANUARY", "FEBRUARY", "MARCH", "APRIL",
                      "MAY", "JUNE","JULY","AUGUST",
                      "SEPTEMBER","OCTOBER","NOVEMBER","DECEMBER",
                      "January", "February", "March", "April",
                      "May", "June","July","August",
                      "September","October","November","December",
                      "january", "february", "march", "april",
                      "may", "june","july","august",
                      "sept\\.","Sept\\.",
                      "sept",
                      "Jan", "Feb", "Mar", "Apr", "May", "Jun",
                      "Jul", "Aug", "Sep", "Oct","Nov", "Dec",
                      "jan", "feb", "mar", "apr", "may", "jun",
                      "jul", "aug", "sep", "oct","nov", "dec")
    romanNumerals <- c("i","ii","iii","iv","v","vi","vii","viii","ix","x","xi","xii",
                       "I","II","III","IV","V","VI","VII","VIII","IX","X","XI","XII")
    numeralConversion <- c("Jan", "Feb", "Mar", "Apr", "May", "Jun",
                           "Jul", "Aug", "Sep", "Oct","Nov", "Dec",
                           "Jan", "Feb", "Mar", "Apr", "May", "Jun",
                           "Jul", "Aug", "Sep", "Oct","Nov", "Dec")
    
          ###### b. date ranges ####
    writeLines(" - Preparing data...")
      # Get a count of how many eventDate rows are full
    originalDateCount <- sum(stats::complete.cases(data$eventDate))
    
      # Create sets of strings that can be used to match date ranges based on delimiter
    dateRangeStrings_forwardSlash <- stringr::str_c(
        # "2015-08-14T16:00:00/2015-08-15T16:00:00"
      "[0-9]{4}\\-[0-9]{2}\\-[0-9]{2}T[0-9]{2}\\:[0-9]{2}\\:[0-9]{2}\\/[0-9]{4}\\-[0-9]{2}\\-[0-9]{2}T[0-9]{2}\\:[0-9]{2}\\:[0-9]{2}",
        # "2019-07-29T14:00:00+02:00/2019-07-29T15:00:00+02:00"
      "[0-9]{4}\\-[0-9]{2}\\-[0-9]{2}T[0-9]{2}\\:[0-9]{2}\\:[0-9]{2}\\+[0-9]{2}:[0-9]{2}\\/[0-9]{4}\\-[0-9]{2}\\-[0-9]{2}T[0-9]{2}\\:[0-9]{2}\\:[0-9]{2}\\+[0-9]{2}:[0-9]{2}",
       # "2015-08-14/2015-08-15"
      "[0-9]{4}\\-[0-9]{2}\\-[0-9]{2}\\/[0-9]{4}\\-[0-9]{2}\\-[0-9]{2}",
      # 9.viii.2000/11.viii.2000
      stringr::str_c("[0-9]{1,2}[-\\.\\s]",
                     "(", stringr::str_c(romanNumerals, collapse = "|"), ")",
                     "[-\\.\\s][0-9]{4}\\/[0-9]{1,2}[-\\.\\s]",
                     "(", stringr::str_c(romanNumerals, collapse = "|"), ")",
                     "[-\\.\\s][0-9]{4}", collapse = "|"),
      # 28 Jun 2005/29 Jun 2005
      stringr::str_c("[0-9]{1,2}[\\s\\/-]",
                     "(", stringr::str_c(monthStrings, collapse = "|"), ")",
                     "[\\s\\/-][0-9]{4}\\/[0-9]{1,2}[\\s\\/-]",
                     "(", stringr::str_c(monthStrings, collapse = "|"), ")",
                     "[\\s\\/-][0-9]{4}", collapse = "|"),
      sep = "|"
    )
    dateRangeStrings_comma <- stringr::str_c(
      # "2015-03-02T17:37:47Z,2015-03-02 05:37 PM"  
      "[0-9]{4}\\-[0-9]{2}\\-[0-9]{2}T[0-9]{2}\\:[0-9]{2}\\:[0-9]{2}Z\\,[0-9]{4}\\-[0-9]{2}\\-[0-9]{2}",
      "[0-9]{4}\\-[0-9]{2}\\-[0-9]{2}T[0-9]{2}\\:[0-9]{2}\\:[0-9]{2}Z\\,[0-9]{4}\\-[0-9]{2}\\-[0-9]{2}T[0-9]{2}\\:[0-9]{2}\\:[0-9]{2}Z",
      # "2015-08-14,2015-08-15"
      "[0-9]{4}\\-[0-9]{2}\\-[0-9]{2}\\,[0-9]{4}\\-[0-9]{2}\\-[0-9]{2}",
      sep = "|"
    )
    dateRangeStrings_to <- stringr::str_c(
      # "2004-08-05 to 2004-08-06"
      "[0-9]{4}\\-[0-9]{2}\\-[0-9]{2} to [0-9]{4}\\-[0-9]{2}\\-[0-9]{2}",
      # 1940-06-26 to 1940-06-28
      "[0-9]{1,2}-[0-9]{1,2}-[0-9]{4}\\sto\\s[0-9]{1,2}-[0-9]{1,2}-[0-9]{4}",
      # "07 to 21 September 2009"
      stringr::str_c("[0-9]{1,2}\\sto\\s[0-9]{1,2}\\s",
                     monthStrings,"[\\s\\/-][0-9]{4}", collapse = "|"),
      # 20 April to 04 May 2009
      stringr::str_c("[0-9]{1,2}[\\s\\/-]",
                     "(", stringr::str_c(monthStrings, collapse = "|"),")",
                     "[\\s\\/-]to[\\s\\/-][0-9]{1,2}[\\s\\/-]",
                     "(", stringr::str_c(monthStrings, collapse = "|"),")",
                     "[\\s\\/-][0-9]{4}",  collapse = "|"),
      # 13-Jun-2018 to 13-Jul-2018
      stringr::str_c("[0-9]{1,2}[\\s\\/-]",
                     "(", stringr::str_c(monthStrings, collapse = "|"),")",
                     "[\\s\\/-][0-9]{4}",
                     "[\\s\\/-]to[\\s\\/-]","[0-9]{1,2}[\\s\\/-]",
                     "(", stringr::str_c(monthStrings, collapse = "|"),")",
                     "[\\s\\/-][0-9]{4}",  collapse = "|"),
      # Nov 15 to Jan 10.1918
      stringr::str_c("(", stringr::str_c(monthStrings, collapse = "|"),")",
                     "[\\s\\/-][0-9]{1,2}",
                     "[\\s\\/-]to[\\s\\/-]",
                     "(", stringr::str_c(monthStrings, collapse = "|"),")",
                     "[\\s\\/-][0-9]{1,2}",
                     "[\\s\\/-][0-9]{4}",  collapse = "|"),
      # Aug 30 1921 to Sept. 5 1921
      stringr::str_c( "(",stringr::str_c(monthStrings, collapse = "|"),")",
                     "[\\s\\/-][0-9]{1,2}[\\s\\/-][0-9]{4}",
                     "[\\s\\/-]to[\\s\\/-]",
                     "(", stringr::str_c(monthStrings, collapse = "|"),")",
                     "[\\s\\/-][0-9]{1,2}",
                     "[\\s\\/-][0-9]{4}",  collapse = "|"),
      sep = "|"
    )
    dateRangeStrings_dash <- stringr::str_c(
      # "28 September - 4 October 2005"
      stringr::str_c( "[0-9]{1,2}\\s",
                      "(",stringr::str_c(monthStrings, collapse = "|"),")",
                      "[\\s\\/-]-[\\s\\/-]", "[0-9]{1,2}\\s",
                      "(", stringr::str_c(monthStrings, collapse = "|"),")",
                      "[\\s\\/-][0-9]{4}",  collapse = "|"),
      # 2 October 2003 - 3 October 2003
      stringr::str_c( "[0-9]{1,2}\\s",
                      "(",stringr::str_c(monthStrings, collapse = "|"),")",
                      "[\\s\\/-]","[0-9]{4}",
                      "[\\s\\/-]-[\\s\\/-]", "[0-9]{1,2}\\s",
                      "(", stringr::str_c(monthStrings, collapse = "|"),")",
                      "[\\s\\/-][0-9]{4}",  collapse = "|"),
      # 15-17 OCTOBER 1976 Or 15-17 October 1976
      stringr::str_c( "[0-9]{1,2}-[0-9]{1,2}","[\\s\\/-]",
                      "(",stringr::str_c(monthStrings, collapse = "|"),")",
                      "[\\s\\/-][0-9]{4}",  collapse = "|"),
      sep = "|"
    )
    dateRangeStrings_StEn <- stringr::str_c(
      # "startDateTime:201908301000xx; EndDateTime:201908301000xx"
      "startDateTime:",
      "EndDateTime:",
      sep = "|"
    )
    
    
      # Testing for exceptions
    if(FALSE){
               TEST <- c("2015-08-14T16:00:00/2015-08-15T16:00:00",
                         "2015-08-14/2015-08-15",
                           # comma 
                         "2015-03-02T17:37:47Z,2015-03-02 05:37 PM",
                           # To
                         "2004-08-05 to 2004-08-06",
                         "07 to 21 September 2009",
                         "28 September - 4 October 2005",
                         "20 April to 04 May 2009",
                         "2 October 2003 - 3 October 2003",
                         "13-Jun-2018 to 13-Jul-2018",
                         "1940-06-26 to 1940-06-28",
                         "Nov 15 to Jan 10.1918",
                         "15-17 October 1976",
                         "15-17 OCTOBER 1976",
                         "Aug 30 1921 to Sept. 5 1921",
                         "startDateTime:201908301000xx; EndDateTime:201908301000xx",
                         "2019-07-29T14:00:00+02:00/2019-07-29T15:00:00+02:00",
                         "9.viii.2000/11.viii.2000",
                         "28 Jun 2005/29 Jun 2005")
               
               stringr::str_detect(TEST, stringr::str_c(
                 dateRangeStrings_forwardSlash, dateRangeStrings_comma,
                 dateRangeStrings_StEn, 
                 dateRangeStrings_to, 
                 dateRangeStrings_dash,
                 sep = "|"
               ))
               stringr::str_detect("15-17 OCTOBER 1976", stringr::str_c(
                 #dateRangeStrings_forwardSlash, dateRangeStrings_comma,
                 #dateRangeStrings_StEn, 
                 #dateRangeStrings_to, 
                 dateRangeStrings_dash,
                 sep = "|"
               ))
    } # END testing
    
    
      ###### c. time parse ####
      # The robust time strings
    timeStrings_robust <- c("Ymd", "YmdHM", "YmdH","YmdHMS","YmdHMSz",
                      # Explicit day MONTH year
                     "dBy","dBY HMS",
                      # Explicit MONTH day year
                     "Bdy","BdY HMS")
    
    
    ##### 0.3 Prep eventDate and find ranges ####
      # Create a new running dataset
    noDATEa <- data %>%
        # Save the original eventDate column
      dplyr::mutate(eventDate_in = eventDate,
                    .before = eventDate) %>%
      rangeFinder(rangeData = .,
                  rangeColumn = "eventDate_in")

    
      # Save this dataset to be merged at the end...
    ymd_hms_0 <- noDATEa %>%
      dplyr::filter(stats::complete.cases(eventDate)) %>%
      dplyr::select(database_id, eventDate, startDayOfYear, endDayOfYear) %>%
      stats::setNames(object = ., c("database_id", "date", "startDayOfYear", "endDayOfYear"))
    
    #### 1.0 easyDates ####
      # Retrieve dates that are much easier to recover...
    writeLines(" - Extracting dates from year, month, day columns...")
    ##### 1.1 year month day ####
    # Filter down to the records that again have no eventDate
    noDATEa <- noDATEa %>%
      dplyr::filter(is.na(eventDate))
      # Some records have date information in the dmy columns that can easily be retrieved
    noDATEa <- noDATEa %>%
      dplyr::mutate(eventDate = dplyr::if_else(is.na(as.character(eventDate)),
                                               lubridate::ymd(stringr::str_c(year, month, day,
                                                                             sep = "-"),
                                                              quiet = TRUE, truncated = 2),
                                               eventDate))

          # Save this dataset to be merged at the end...
    dmy_1 <- noDATEa %>%
      dplyr::filter(stats::complete.cases(eventDate)) %>%
      dplyr::select(database_id, eventDate) %>%
      stats::setNames(object = ., c("database_id", "date"))

      # Filter down to the records that again have no eventDate
    noDATEa <- noDATEa %>%
      dplyr::filter(is.na(eventDate))
    if("occurrenceYear" %in% colnames(noDATEa)){
      # Copy across the occurrenceYear column into the eventDate column
    noDATEa$eventDate <- noDATEa$occurrenceYear}
      # Save this dataset to be merged at the end...
    occYr_2 <- noDATEa %>%
      dplyr::filter(stats::complete.cases(eventDate))%>%
      dplyr::select(database_id, eventDate) %>%
      stats::setNames(object = ., c("database_id", "date"))
    
      ##### 1.1 Sept ####
        # Because some people write "Sept" which cannot be read by lubridate, it needs to be 
        # replaced in these columns
    noDATEa$locality <- noDATEa$locality %>%
      stringr::str_replace(pattern = "[Ss]ept[\\s-/]",
                           replacement = "Sep ")
    noDATEa$fieldNotes <- noDATEa$fieldNotes %>%
      stringr::str_replace(pattern = "[Ss]ept[\\s-/]",
                           replacement = "Sep ")
    noDATEa$locationRemarks <- noDATEa$locationRemarks %>%
      stringr::str_replace(pattern = "[Ss]ept[\\s-/]",
                           replacement = "Sep ")
    noDATEa$verbatimEventDate <- noDATEa$verbatimEventDate %>%
      stringr::str_replace(pattern = "[Ss]ept[\\s-/]",
                           replacement = "Sep ")
    
    #### 2.0 unAmb. str. dates ####
    writeLines(paste0(
      " - Extracting dates from fieldNotes, locationRemarks, and verbatimEventDate ",
      "columns in unambiguous ymd, dmy, mdy, and my formats..."))
      # Filter down to the records that again have no eventDate
    noDATEa <- noDATEa %>%
      dplyr::filter(is.na(eventDate))
    
      ##### 2.1 ymd ####
      # enter ymd strings...
    ymd_strings <- c(
      "[0-9]{4}[\\s- /]+[0-9]{2}[\\s- /]+[0-9]{2}",
      paste("[0-9]{4}", monthStrings,"[0-9]{2}", collapse = "|", sep = "[\\s-/]+"),
      paste("[0-9]{4}", romanNumerals,"[0-9]{2}", collapse = "|", sep = "[.\\s-/]+"),
      # "2015-08-14T16:00:00/2015-08-15T16:00:00"
      "[0-9]{4}\\-[0-9]{2}\\-[0-9]{2}T[0-9]{2}\\:[0-9]{2}\\:[0-9]{2}\\/[0-9]{4}\\-[0-9]{2}\\-[0-9]{2}T[0-9]{2}\\:[0-9]{2}\\:[0-9]{2}",
      # "2019-07-29T14:00:00+02:00/2019-07-29T15:00:00+02:00"
      "[0-9]{4}\\-[0-9]{2}\\-[0-9]{2}T[0-9]{2}\\:[0-9]{2}\\:[0-9]{2}\\+[0-9]{2}:[0-9]{2}\\/[0-9]{4}\\-[0-9]{2}\\-[0-9]{2}T[0-9]{2}\\:[0-9]{2}\\:[0-9]{2}\\+[0-9]{2}:[0-9]{2}",
      # "2015-08-14/2015-08-15"
      "[0-9]{4}\\-[0-9]{2}\\-[0-9]{2}\\/[0-9]{4}\\-[0-9]{2}\\-[0-9]{2}",
      # "2015-03-02T17:37:47Z,2015-03-02 05:37 PM"  
      "[0-9]{4}\\-[0-9]{2}\\-[0-9]{2}T[0-9]{2}\\:[0-9]{2}\\:[0-9]{2}Z\\,[0-9]{4}\\-[0-9]{2}\\-[0-9]{2}",
      "[0-9]{4}\\-[0-9]{2}\\-[0-9]{2}T[0-9]{2}\\:[0-9]{2}\\:[0-9]{2}Z\\,[0-9]{4}\\-[0-9]{2}\\-[0-9]{2}T[0-9]{2}\\:[0-9]{2}\\:[0-9]{2}Z",
      # "2015-08-14,2015-08-15"
      "[0-9]{4}\\-[0-9]{2}\\-[0-9]{2}\\,[0-9]{4}\\-[0-9]{2}\\-[0-9]{2}",
      # "2004-08-05 to 2004-08-06"
      "[0-9]{4}\\-[0-9]{2}\\-[0-9]{2} to [0-9]{4}\\-[0-9]{2}\\-[0-9]{2}",
      # 1940-06-26 to 1940-06-28
      "[0-9]{1,2}-[0-9]{1,2}-[0-9]{4}\\sto\\s[0-9]{1,2}-[0-9]{1,2}-[0-9]{4}")
    
    # Extract the matching strings
    ymd_unambiguous <- noDATEa %>%
      dplyr::mutate(
        ymd_vEV = stringr::str_extract(verbatimEventDate,
                                              pattern = paste0(ymd_strings, 
                                                              collapse = "|")),
        ymd_fieldNotes = stringr::str_extract(fieldNotes,
                                  pattern = paste0(ymd_strings, 
                                  collapse = "|")),
        ymd_locationRemarks = stringr::str_extract(locationRemarks,
                                  pattern = paste0(ymd_strings, 
                                  collapse = "|")),
        ymd_eventDate_in = stringr::str_extract(eventDate_in,
                                                   pattern = paste0(ymd_strings, 
                                                                    collapse = "|"))
        ) %>% # END mutate
      dplyr::select(database_id, 
                    ymd_vEV,
                    ymd_fieldNotes, 
                    ymd_locationRemarks,
                    ymd_eventDate_in)  
    
    # FORMAT the ymd_vEV column
    ymd_unambiguous$ymd_vEV <- ymd_unambiguous$ymd_vEV %>%
      # Remove 00 values to truncate the lubridate 
      stringr::str_replace(pattern = "/00$|/00/00$", replacement = "")
      #
        # FORMAT the ymd_fieldNotes column
      ymd_unambiguous$ymd_fieldNotes <- ymd_unambiguous$ymd_fieldNotes %>%
          # Remove 00 values to truncate the lubridate 
        stringr::str_replace(pattern = "/00$|/00/00$", replacement = "") 
      #
          # FORMAT the ymd_locationRemarks column
      ymd_unambiguous$ymd_locationRemarks <- ymd_unambiguous$ymd_locationRemarks %>%
        # Remove 00 values to truncate the lubridate 
        stringr::str_replace(pattern = "/00$|/00/00$", replacement = "") 
      #
      # FORMAT the ymd_eventDate column
      ymd_unambiguous$ymd_eventDate_in <- ymd_unambiguous$ymd_eventDate_in %>%
        # Remove 00 values to truncate the lubridate 
        stringr::str_replace(pattern = "/00$|/00/00$", replacement = "") 

        # Combine the columns
      ymd_keepers_21 <- ymd_unambiguous %>%
        dplyr::filter(stats::complete.cases(ymd_vEV)|
                        stats::complete.cases(ymd_locationRemarks) | stats::complete.cases(ymd_fieldNotes) |
                        stats::complete.cases(ymd_eventDate_in)) %>%
        tidyr::unite(col = date,
                     ymd_vEV, ymd_locationRemarks, ymd_fieldNotes, ymd_eventDate_in, na.rm = TRUE) %>% 
        rangeFinder(rangeData = .,
                    rangeColumn = "date")
        # add ymd_keepers_21 at the end
      

    ##### 2.2 dmy ####
    dmy_strings <- stringr::str_c(
      # 12-JUL-2002; 12 Jul 2002; 12/July/2002
    paste0("[0-9]{1,2}[\\s-/]+", monthStrings,"[\\s-/]+([0-9]{4}|[0-9]{2})", collapse = "|"),
      # 12-XII-2022; 12 XII 2022; 12 xii 2022;
    paste0("[0-9]{1,2}[.\\s-/]+", romanNumerals,"[.\\s-/]+([0-9]{4}|[0-9]{2})", collapse = "|"),
      # >12 <12 1992 - dmy
    "([1][3-9]|[2-3][0-9])[\\s-/ ]+(0[1-9]|[1-9]|1[0-2])[\\s-/]+([0-9]{4}|[0-9]{2})",
    stringr::str_c("[0-9]{1,2}[-\\.\\s]",romanNumerals,
                   "[-\\.\\s][0-9]{4}\\/[0-9]{1,2}[-\\.\\s]",
                   romanNumerals,"[-\\.\\s][0-9]{4}", collapse = "|"),
    # 28 Jun 2005/29 Jun 2005
    stringr::str_c("[0-9]{1,2}[\\s\\/-]",monthStrings,
                   "[\\s\\/-][0-9]{4}\\/[0-9]{1,2}[\\s\\/-]",
                   monthStrings,"[\\s\\/-][0-9]{4}", collapse = "|"),
    # "07 to 21 September 2009"
    stringr::str_c("[0-9]{1,2}\\sto\\s[0-9]{1,2}\\s",
                   monthStrings,"[\\s\\/-][0-9]{4}", collapse = "|"),
    # 20 April to 04 May 2009
    stringr::str_c("[0-9]{1,2}[\\s\\/-]",
                   "(", stringr::str_c(monthStrings, collapse = "|"),")",
                   "[\\s\\/-]to[\\s\\/-][0-9]{1,2}[\\s\\/-]",
                   "(", stringr::str_c(monthStrings, collapse = "|"),")",
                   "[\\s\\/-][0-9]{4}",  collapse = "|"),
    # 13-Jun-2018 to 13-Jul-2018
    stringr::str_c("[0-9]{1,2}[\\s\\/-]",
                   "(", stringr::str_c(monthStrings, collapse = "|"),")",
                   "[\\s\\/-][0-9]{4}",
                   "[\\s\\/-]to[\\s\\/-]","[0-9]{1,2}[\\s\\/-]",
                   "(", stringr::str_c(monthStrings, collapse = "|"),")",
                   "[\\s\\/-][0-9]{4}",  collapse = "|"),
    # Nov 15 to Jan 10.1918
    stringr::str_c("(", stringr::str_c(monthStrings, collapse = "|"),")",
                   "[\\s\\/-][0-9]{1,2}",
                   "[\\s\\/-]to[\\s\\/-]",
                   "(", stringr::str_c(monthStrings, collapse = "|"),")",
                   "[\\s\\/-][0-9]{1,2}",
                   "[\\s\\/-][0-9]{4}",  collapse = "|"),
    # Aug 30 1921 to Sept. 5 1921
    stringr::str_c( "(",stringr::str_c(monthStrings, collapse = "|"),")",
                    "[\\s\\/-][0-9]{1,2}[\\s\\/-][0-9]{4}",
                    "[\\s\\/-]to[\\s\\/-]",
                    "(", stringr::str_c(monthStrings, collapse = "|"),")",
                    "[\\s\\/-][0-9]{1,2}",
                    "[\\s\\/-][0-9]{4}",  collapse = "|"),
    # 9.viii.2000/11.viii.2000
    stringr::str_c("[0-9]{1,2}[-\\.\\s]",
                   "(", stringr::str_c(romanNumerals, collapse = "|"), ")",
                   "[-\\.\\s][0-9]{4}\\/[0-9]{1,2}[-\\.\\s]",
                   "(", stringr::str_c(romanNumerals, collapse = "|"), ")",
                   "[-\\.\\s][0-9]{4}", collapse = "|"),
    # 28 Jun 2005/29 Jun 2005
    stringr::str_c("[0-9]{1,2}[\\s\\/-]",
                   "(", stringr::str_c(monthStrings, collapse = "|"), ")",
                   "[\\s\\/-][0-9]{4}\\/[0-9]{1,2}[\\s\\/-]",
                   "(", stringr::str_c(monthStrings, collapse = "|"), ")",
                   "[\\s\\/-][0-9]{4}", collapse = "|"),
     sep = "|"
    )

        # Extract the matching strings
      dmy_unambiguous <- noDATEa %>%
          # First, remove the strings matched prior
        dplyr::filter(!database_id %in% ymd_keepers_21$database_id) %>%
        dplyr::mutate(
          dmy_vEV = stringr::str_extract(verbatimEventDate,
                                              pattern = dmy_strings),
          dmy_locality = stringr::str_extract(locality,
                                              pattern = dmy_strings),
          dmy_fieldNotes = stringr::str_extract(fieldNotes,
                                                pattern = dmy_strings),
          dmy_locationRemarks = stringr::str_extract(locationRemarks,
                                                     pattern = dmy_strings),
          dmy_eventDate_in = stringr::str_extract(eventDate_in,
                                                pattern = dmy_strings)
        ) %>% # END mutate
        dplyr::select(database_id, dmy_vEV, dmy_locality,
                      dmy_fieldNotes, dmy_locationRemarks, dmy_eventDate_in)  

      # FORMAT the dmy_vEV column
      dmy_unambiguous$dmy_vEV <- dmy_unambiguous$dmy_vEV %>%
        # Convert roman numerals to readable by lubridate
        mgsub::mgsub(
          pattern = paste0("[-/ \\.]", romanNumerals, "[ -/\\.]"),
          replacement = paste0(" ", numeralConversion, " ") ) %>% stringr::str_squish() %>%
        # Remove 00 values to truncate the lubridate 
        stringr::str_replace(pattern = "/00$|/00/00$|^00", replacement = "")
      #
      # FORMAT the dmy_locality column
      dmy_unambiguous$dmy_locality <- dmy_unambiguous$dmy_locality %>%
        # Convert roman numerals to readable by lubridate
        mgsub::mgsub(
          pattern = paste0("[-/ \\.]", romanNumerals, "[ -/\\.]"),
          replacement = paste0(" ", numeralConversion, " ") ) %>% stringr::str_squish() %>%
        # Remove 00 values to truncate the lubridate 
        stringr::str_replace(pattern = "/00$|/00/00$", replacement = "") 
      #
      # FORMAT the dmy_fieldNotes column
      dmy_unambiguous$dmy_fieldNotes <- dmy_unambiguous$dmy_fieldNotes %>%
        # Convert roman numerals to readable by lubridate
        mgsub::mgsub(
          pattern = paste0("[-/ \\.]", romanNumerals, "[ -/\\.]"),
          replacement = paste0(" ", numeralConversion, " ") ) %>% stringr::str_squish() %>%
        # Remove 00 values to truncate pthe lubridate 
        stringr::str_replace(pattern = "/00$|/00/00$", replacement = "") 
      #
      # FORMAT the dmy_locationRemarks column
      dmy_unambiguous$dmy_locationRemarks <- dmy_unambiguous$dmy_locationRemarks %>%
        # Convert roman numerals to readable by lubridate
        mgsub::mgsub(
          pattern = paste0("[-/ \\.]", romanNumerals, "[ -/\\.]"),
          replacement = paste0(" ", numeralConversion, " ") ) %>% stringr::str_squish() %>%
        # Remove 00 values to truncate the lubridate 
        stringr::str_replace(pattern = "/00$|/00/00$", replacement = "") 
      # FORMAT the dmy_eventDate_in column
      dmy_unambiguous$dmy_eventDate_in <- dmy_unambiguous$dmy_eventDate_in %>%
        # Convert roman numerals to readable by lubridate
        mgsub::mgsub(
          pattern = paste0("[-/ \\.]", romanNumerals, "[ -/\\.]"),
          replacement = paste0(" ", numeralConversion, " ") ) %>% stringr::str_squish() %>%
        # Remove 00 values to truncate the lubridate 
        stringr::str_replace(pattern = "/00$|/00/00$", replacement = "") 
      #
      # Combine the columns
      dmy_keepers_22 <- dmy_unambiguous %>%
        dplyr::filter(stats::complete.cases(dmy_vEV) |
                        stats::complete.cases(dmy_locality) | 
                        stats::complete.cases(dmy_locationRemarks) | 
                        stats::complete.cases(dmy_fieldNotes) |
                        stats::complete.cases(dmy_eventDate_in)) %>%
        tidyr::unite(col = date, 
                     dmy_vEV, dmy_locality, dmy_locationRemarks, dmy_fieldNotes, dmy_eventDate_in,
                     na.rm = TRUE) %>% 
        rangeFinder(rangeData = .,
                    rangeColumn = "date",
                    timeStrings_in = c(timeStrings_robust, "dmy"))
      # add dmy_keepers_22 at the end
    
    ##### 2.3 mdy ####
    mdy_strings <- c(
      # Aug 2, 2019
    paste0(monthStrings,"[\\s-/ ]+[0-9]{1,2}[\\s-/, ]+([0-9]{4}|[0-9]{2})", collapse = "|"),
      # Aug 1-10 2019
    paste0(monthStrings,"[0-9]+[-\\u2013][0-9]+[\\s-/ ]+([0-9]{4}|[0-9]{2})", collapse = "|"),
      # V. 17 1901
    paste0("(?:^|\\s|\\(|\"\')", romanNumerals,"[\\s-/\\.]+[0-9]{1,2}[\\s-/ ]+([0-9]{4}|[0-9]{2})", collapse = "|"),
    
    # <12 >12 1992 - mdy
    "(0?[1-9]|1[0-2])[\\s- /]+(1[3-9]|2[0-9]|3[01])[\\s- /]+([0-9]{4})",
    "(0?[1-9]|1[0-2])[\\s- /]+(1[3-9]|2[0-9]|3[01])[\\s- /]+([0-9]{2})",
    "(0?[1-9])[\\s- /]+(1[3-9]|2[0-9]|3[01])[\\s- /]+([0-9]{4})",
    "(0?[1-9])[\\s- /]+(1[3-9]|2[0-9]|3[01])[\\s- /]+([0-9]{2})",
    # Nov 15 to Jan 10.1918
    stringr::str_c("(", stringr::str_c(monthStrings, collapse = "|"),")",
                   "[\\s\\/-][0-9]{1,2}",
                   "[\\s\\/-]to[\\s\\/-]",
                   "(", stringr::str_c(monthStrings, collapse = "|"),")",
                   "[\\s\\/-][0-9]{1,2}",
                   "[\\s\\/-][0-9]{4}",  collapse = "|"),
    # Aug 30 1921 to Sept. 5 1921
    stringr::str_c( "(",stringr::str_c(monthStrings, collapse = "|"),")",
                    "[\\s\\/-][0-9]{1,2}[\\s\\/-][0-9]{4}",
                    "[\\s\\/-]to[\\s\\/-]",
                    "(", stringr::str_c(monthStrings, collapse = "|"),")",
                    "[\\s\\/-][0-9]{1,2}",
                    "[\\s\\/-][0-9]{4}",  collapse = "|")
    )
      # Get the IDs to remove...
      id2remove_23 <- c(ymd_keepers_21$database_id, dmy_keepers_22$database_id)
    
      # Extract the matching strings to three columns
      mdy_unambiguous <- noDATEa %>%
        # First, remove the strings matched prior
        dplyr::filter(!database_id %in% id2remove_23) %>%
        dplyr::mutate(
          mdy_vEV = stringr::str_extract(verbatimEventDate,
                                              pattern = paste0(mdy_strings, 
                                                               collapse = "|")),
          mdy_locality = stringr::str_extract(locality,
                                              pattern = paste0(mdy_strings, 
                                                               collapse = "|")),
          mdy_fieldNotes = stringr::str_extract(fieldNotes,
                                                pattern = paste0(mdy_strings, 
                                                                 collapse = "|")),
          mdy_locationRemarks = stringr::str_extract(locationRemarks,
                                                     pattern = paste0(mdy_strings, 
                                                                      collapse = "|")),
          mdy_eventDate_in = stringr::str_extract(eventDate_in,
                                                     pattern = paste0(mdy_strings, 
                                                                      collapse = "|"))
        ) %>% # END mutate
          # select a subset of columns
        dplyr::select(database_id, mdy_vEV, mdy_locality,
                      mdy_fieldNotes, mdy_locationRemarks, mdy_eventDate_in)  

      # FORMAT the mdy_vEV column
      mdy_unambiguous$mdy_vEV <- mdy_unambiguous$mdy_vEV %>%
        # Convert roman numerals to readable by lubridate
        mgsub::mgsub(
          pattern = paste0("^", romanNumerals, "[ -/\\.]"),
          replacement = paste0(numeralConversion, " ") ) %>% stringr::str_squish()

      #
      # FORMAT the mdy_locality column
      mdy_unambiguous$mdy_locality <- mdy_unambiguous$mdy_locality %>%
        # Convert roman numerals to readable by lubridate
        mgsub::mgsub(
          pattern = paste0("^", romanNumerals, "[ -/\\.]"),
          replacement = paste0(numeralConversion, " ") ) %>% stringr::str_squish() %>%
        stringr::str_replace( pattern = "^The ", replacement = "")  
      #
      # FORMAT the mdy_fieldNotes column
      mdy_unambiguous$mdy_fieldNotes <- mdy_unambiguous$mdy_fieldNotes %>%
        # Convert roman numerals to readable by lubridate
        mgsub::mgsub(
          pattern = paste0("^", romanNumerals, "[ -/\\.]"),
          replacement = paste0(numeralConversion, " ") ) %>% stringr::str_squish()

      #
      # FORMAT the mdy_locationRemarks column
      mdy_unambiguous$mdy_locationRemarks <- mdy_unambiguous$mdy_locationRemarks %>%
        # Convert roman numerals to readable by lubridate
        mgsub::mgsub(
          pattern = paste0("^", romanNumerals, "[-/ \\.]"),
          replacement = paste0(numeralConversion, " ") ) %>% stringr::str_squish() %>%
        lubridate::mdy(truncated = 2, quiet = TRUE) 
      # FORMAT the mdy_eventDate_in column
      mdy_unambiguous$mdy_eventDate_in <- mdy_unambiguous$mdy_eventDate_in %>%
        # Convert roman numerals to readable by lubridate
        mgsub::mgsub(
          pattern = paste0("^", romanNumerals, "[ -/\\.]"),
          replacement = paste0(numeralConversion, " ") ) %>% stringr::str_squish()
      #

      # Combine the columns
      mdy_keepers_23 <- mdy_unambiguous %>%
        dplyr::filter( stats::complete.cases(mdy_vEV) |
                         stats::complete.cases(mdy_locality) | 
                        stats::complete.cases(mdy_locationRemarks) | 
                         stats::complete.cases(mdy_fieldNotes) | 
                         stats::complete.cases(mdy_eventDate_in)) %>%
        tidyr::unite(col = date, 
                     mdy_vEV, mdy_locality, mdy_locationRemarks, mdy_fieldNotes, mdy_eventDate_in,
                     na.rm = TRUE, sep = "") %>% 
        rangeFinder(rangeData = .,
                    rangeColumn = "date",
                    timeStrings_in = c(timeStrings_robust, "mdy"))
        # KEEP mdy_keepers_23 at the end
    
    
    ##### 2.4 my ####
    my_strings <- c(
      # VIII-1946
    paste0("(?:^|\\s|\\(|\"|\')", romanNumerals, "[\\s-/ \\.]+([0-9]{4}|[0-9]{2})", collapse = "|"),
      # July 1995; July, 1995
    paste0(monthStrings, "[\\s-/ \\.]+([0-9]{4}|[0-9]{2})", collapse = "|"),
      # April 1899
    paste0(monthStrings, "[\\s-/ ]+([0-9]{4}|[0-9]{2})", collapse = "|"),
      # 1899 April 
    paste0("([0-9]{4}|[0-9]{2})[\\s-/ ]+", monthStrings, collapse = "|"),
      # 4/1957
    "(?:^|\\s|\\(|\"\')(0?[1-9]|1[0-2])[\\s- /]+[0-9]{4}"
    )
    
      # Get the IDs to remove...
      id2remove_24 <- c(mdy_keepers_23$database_id, id2remove_23)

      # Extract the matching strings to three columns
      my_unambiguous <- noDATEa %>%
        # First, remove the strings matched prior
        dplyr::filter(!database_id %in% id2remove_24) %>%
        dplyr::mutate(
          my_vEV = stringr::str_extract(verbatimEventDate,
                                             pattern = paste0(my_strings, 
                                                              collapse = "|")),
          my_locality = stringr::str_extract(locality,
                                              pattern = paste0(my_strings, 
                                                               collapse = "|")),
          my_fieldNotes = stringr::str_extract(fieldNotes,
                                                pattern = paste0(my_strings, 
                                                                 collapse = "|")),
          my_locationRemarks = stringr::str_extract(locationRemarks,
                                                     pattern = paste0(my_strings, 
                                                                      collapse = "|"))
        ) %>% # END mutate
        # select a subset of columns
        dplyr::select(database_id, my_vEV, my_locality,
                      my_fieldNotes, my_locationRemarks) 

      # FORMAT the my_vEV column
      my_unambiguous$my_vEV <- my_unambiguous$my_vEV %>%
        # Convert roman numerals to readable by lubridate
        mgsub::mgsub(
          pattern = paste0("(^|[-/ \\.])", romanNumerals, "[\\s -/\\.]"),
          replacement = paste0(" ", numeralConversion, " ") ) %>% stringr::str_squish()  %>%
        # format
        lubridate::my( quiet = TRUE) 
      #
      # FORMAT the my_locality column
      my_unambiguous$my_locality <- my_unambiguous$my_locality %>%
        # Convert roman numerals to readable by lubridate
        mgsub::mgsub(
          pattern = paste0("(^|[-/ \\.])", romanNumerals, "[\\s -/\\.]"),
          replacement = paste0(" ", numeralConversion, " ") ) %>% stringr::str_squish() %>%
          # format
        lubridate::my(quiet = TRUE) 
      #
      # FORMAT the my_fieldNotes column
      my_unambiguous$my_fieldNotes <- my_unambiguous$my_fieldNotes %>%
        # Convert roman numerals to readable by lubridate
        mgsub::mgsub(
          pattern = paste0("(^|[-/ \\.])", romanNumerals, "[\\s -/\\.]"),
          replacement = paste0(" ", numeralConversion, " ") ) %>% stringr::str_squish()  %>%
        # format
        lubridate::my(quiet = TRUE) 
      #
      # FORMAT the my_locationRemarks column
      my_unambiguous$my_locationRemarks <- my_unambiguous$my_locationRemarks %>%
        # Convert roman numerals to readable by lubridate
        mgsub::mgsub(
          pattern = paste0("(^|[-/ \\.])", romanNumerals, "[\\s -/\\.]"),
          replacement = paste0(" ", numeralConversion, " ") ) %>% stringr::str_squish()  %>%
        # format
        lubridate::my(quiet = TRUE) 
      #

      # Combine the columns
      my_keepers_24 <- my_unambiguous %>%
        dplyr::filter(stats::complete.cases(my_vEV) |
                        stats::complete.cases(my_locality) | 
                        stats::complete.cases(my_locationRemarks) | 
                        stats::complete.cases(my_fieldNotes)) %>%
        tidyr::unite(col = date, 
                     my_vEV, my_locality, my_locationRemarks, my_fieldNotes, 
                     na.rm = TRUE, sep = "-")  %>% 
        rangeFinder(rangeData = .,
                    rangeColumn = "date") 
          # Remove double-ups
      my_keepers_24$date <- stringr::str_replace(my_keepers_24$date,
                                                 pattern = "-[0-9]+-[0-9]+-[0-9]+", 
                                                 replacement = "")
    
      #### 3.0 Amb. str. dates ####
      writeLines(paste0(
        " - Extracting year from fieldNotes, locationRemarks, and verbatimEventDate ",
        "columns in ambiguous formats..."))
      ambiguousDateStrings <- c(
        # dmy or mdy; 10 02 1946
        "[0-9]{1,2}[\\s-/ ]+[0-9]{1,2}[\\s-/ ]+[0-9]{4}",
        "[0-9]{2}[\\s-/ ]+[0-9]{2}[\\s-/ ]+[0-9]{4}",
        "[0-9]{1,2}[\\s-/ ]+[0-9]{1,2}[\\s-/ ]+[0-9]{2}",
        "[0-9]{2}[\\s-/ ]+[0-9]{2}[\\s-/ ]+[0-9]{2}"
      )

      ambiguousDatePattern = paste0(ambiguousDateStrings, collapse="|")

      # Get the IDs to remove...
      id2remove_30 <- c(my_keepers_24$database_id, id2remove_24)
      
      # Extract the matching strings to three columns
      ambiguousNames <- noDATEa %>%
        # First, remove the strings matched prior
        dplyr::filter(!database_id %in% id2remove_30) %>%
        dplyr::mutate(
          amb_vEV = stringr::str_extract(verbatimEventDate, pattern=ambiguousDatePattern),
          amb_locality = stringr::str_extract(locality, pattern=ambiguousDatePattern),
          amb_fieldNotes = stringr::str_extract(fieldNotes, pattern=ambiguousDatePattern),
          amb_locationRemarks = stringr::str_extract(locationRemarks, pattern=ambiguousDatePattern)
        ) %>% # END mutate
        # select a subset of columns
        dplyr::select(database_id, amb_vEV, amb_locality,
                      amb_fieldNotes, amb_locationRemarks) 

      # FORMAT the amb_vEV column
      ambiguousNames$amb_vEV <- ambiguousNames$amb_vEV %>%
        # Remove 00 values to truncate the lubridate 
        stringr::str_replace(pattern = "00[-/ ]00[-/ ]", replacement = "01-01-") %>%
        stringr::str_replace(pattern = "00[-/ ]", replacement = "01-") %>%
        stringr::str_replace(pattern = "00-00-|[-/ ]00$|[-/ ]00[-/ ]00$|^00|^00[-/ ]00[-/ ]",
                             replacement = "") %>%
        # Attempt to build date from remaining string
        lubridate::dmy(truncated = 2, quiet = TRUE) %>%
        # Safely extract just the year
        lubridate::year() %>%
        # Build date from Jan 1st
        lubridate::make_date(1, 1)
      #
      # FORMAT the amb_locality column
      ambiguousNames$amb_locality <- ambiguousNames$amb_locality %>%
        # Remove 00 values to truncate the lubridate 
        stringr::str_replace(pattern = "00[-/ ]00[-/ ]", replacement = "01-01-") %>%
        stringr::str_replace(pattern = "00[-/ ]", replacement = "01-") %>%
        # Remove any remaining leading/trailing zeroes
        stringr::str_replace(pattern = "00-00-|[-/ ]00$|[-/ ]00[-/ ]00$|^00|^00[-/ ]00[-/ ]",
                             replacement = "") %>%
        # Attempt to build date from remaining string
        lubridate::dmy(truncated = 2, quiet = TRUE) %>%
        # Extract just the year
        lubridate::year() %>%
        # Build date from Jan 1st
        lubridate::make_date(1, 1)
      #
      # FORMAT the amb_fieldNotes column
      ambiguousNames$amb_fieldNotes <- ambiguousNames$amb_fieldNotes %>%
      # Remove 00 values to truncate the lubridate 
      stringr::str_replace(pattern = "00[-/ ]00[-/ ]", replacement = "01-01-") %>%
        stringr::str_replace(pattern = "00[-/ ]", replacement = "01-") %>%
        stringr::str_replace(pattern = "00-00-|[-/ ]00$|[-/ ]00[-/ ]00$|^00|^00[-/ ]00[-/ ]",
                             replacement = "") %>%
        # Attempt to build date from remaining string
        lubridate::dmy(truncated = 2, quiet = TRUE) %>%
        # Extract just the year
        lubridate::year() %>%
        # Build date from Jan 1st
        lubridate::make_date(1, 1)
      #
      # FORMAT the amb_locationRemarks column
      ambiguousNames$amb_locationRemarks <- ambiguousNames$amb_locationRemarks %>%
        # Remove 00 values to truncate the lubridate 
        stringr::str_replace(pattern = "00[-/ ]00[-/ ]", replacement = "01-01-") %>%
        stringr::str_replace(pattern = "00[-/ ]", replacement = "01-") %>%
        stringr::str_replace(pattern = "00-00-|[-/ ]00$|[-/ ]00[-/ ]00$|^00|^00[-/ ]00[-/ ]",
                            replacement = "") %>%
        # Attempt to build date from remaining string
        lubridate::dmy(truncated = 2, quiet = TRUE) %>%
        # Extract just the year
        lubridate::year() %>%
        # Build date from Jan 1st
        lubridate::make_date(1, 1)


      # Combine the columns
      amb_keepers_30 <- ambiguousNames %>%
        dplyr::filter(stats::complete.cases(amb_vEV)|
                        stats::complete.cases(amb_locality) | 
                        stats::complete.cases(amb_locationRemarks) | 
                        stats::complete.cases(amb_fieldNotes)) %>%
        tidyr::unite(col = date, amb_vEV,  
                     amb_locality, amb_locationRemarks, amb_fieldNotes, 
                     na.rm = TRUE) %>% 
        rangeFinder(rangeData = .,
                    rangeColumn = "date")
      # KEEP amb_keepers_30 at the end

      #### 4.0 Format+combine ####
      writeLines(paste0(
        " - Formating and combining the new data.."))
        ##### 4.1 formatting... ####
        # Extract only the date from occYr_2
      occYr_2$date <- as.character(occYr_2$date) %>% lubridate::ymd_hms() %>% lubridate::date()
        # Set as date format...
      ymd_keepers_21$date <- lubridate::ymd(ymd_keepers_21$eventDate)
      dmy_keepers_22$date <- lubridate::ymd(dmy_keepers_22$eventDate)
      mdy_keepers_23$date <- lubridate::ymd(mdy_keepers_23$eventDate)
      my_keepers_24$date <- lubridate::ymd(my_keepers_24$eventDate)
        # merge these data...

      saveTheDates <- dplyr::bind_rows(ymd_hms_0, dmy_1, occYr_2, 
        ymd_keepers_21, dmy_keepers_22, mdy_keepers_23) %>%
        dplyr::select(database_id, date, startDayOfYear, endDayOfYear) 


      ##### 4.2 Full dates ####
        # Join these dates to the original rows...
      datesOut_full <- data %>% 
        dplyr::right_join(saveTheDates, 
                          by = "database_id")
        # Fill the eventDate column
      datesOut_full$eventDate <- lubridate::ymd_hms(datesOut_full$date,
                                                    truncated = 5)
        # Fill the year, month, and day columns 
      datesOut_full$year <- lubridate::year(datesOut_full$date)
      datesOut_full$month <- lubridate::month(datesOut_full$date)
      datesOut_full$day <- lubridate::day(datesOut_full$date)
        # Remove records with non-sensical years
      datesOut_full <- datesOut_full %>%
          # remove FUTURE dates
        dplyr::filter(!year > maxYear) %>%
          # Remove PAST dates 
        dplyr::filter(!year < minYear) %>%
          # remove the date column
        dplyr::select(!date)

      ##### 4.3 No day ####
        # Join these dates to the original rows...
      datesOut_noDay <- data %>% 
        dplyr::right_join(my_keepers_24 %>%
                            dplyr::mutate(date = eventDate) %>%
                            dplyr::select(!eventDate), 
                          by = "database_id")
        # Fill the eventDate column
      datesOut_noDay$eventDate <- lubridate::ymd_hms(datesOut_noDay$date, quiet = TRUE,
                                                     truncated = 5)
        # Fill the year, month, and day columns 
      datesOut_noDay$year <- lubridate::year(datesOut_noDay$date)
      datesOut_noDay$month <- lubridate::month(datesOut_noDay$date)
        # Remove records with non-sensical years
      datesOut_noDay <- datesOut_noDay %>%
         # remove FUTURE dates
        dplyr::filter(!year > maxYear) %>%
         # Remove PAST dates 
        dplyr::filter(!year < minYear) %>%
         # remove the date column
        dplyr::select(!date)

      ##### 4.4 No month ####
        # Join these dates to the original rows...
      datesOut_noMonth <- data %>% 
        dplyr::right_join(amb_keepers_30 %>%
                            dplyr::mutate(date = eventDate) %>%
                            dplyr::select(!eventDate), 
                          by = "database_id")
        # Fill the eventDate column
      datesOut_noMonth$eventDate <- lubridate::ymd(datesOut_noMonth$date, quiet = TRUE)
        # Fill the year, month, and day columns 
      datesOut_noMonth$year <- lubridate::year(datesOut_noMonth$date)
        # Remove records with non-sensical years
      datesOut_noMonth <- datesOut_noMonth %>%
          # remove FUTURE dates
        dplyr::filter(!year > maxYear) %>%
         # Remove PAST dates 
        dplyr::filter(!year < minYear) %>%
          # remove the date column
        dplyr::select(!date)
    
    #### 5.0 Merge ####
      writeLines(paste0(
        " - Merging all data, nearly there..."))
        # Get all of the changed rows together
      datesMerged <- dplyr::bind_rows(
        datesOut_full, datesOut_noDay, datesOut_noMonth)

        # Format the original eventDate column into a new sheet - datesOut
      datesOut <- data
      datesOut$eventDate <- lubridate::ymd_hms(datesOut$eventDate,
                                 truncated = 5, quiet = TRUE)
      # Replace these in the original dataset
      datesOut <- datesOut %>% 
          # Remove the dates that are to be replaced
        dplyr::filter(!database_id %in% datesMerged$database_id) 

      # Extract year, month, and day where possible
        # year
      datesOut$year <- ifelse(is.na(datesOut$year),
                              lubridate::year(datesOut$eventDate),
                              datesOut$year)
        # month
      datesOut$month <- ifelse(is.na(datesOut$month),
                               lubridate::month(datesOut$eventDate),
                               datesOut$month)
        # day
      datesOut$day <- ifelse(is.na(datesOut$day),
                               lubridate::day(datesOut$eventDate),
                               datesOut$day)

      # Remove the months and days where the year is incorrect.
      datesOut$month <- ifelse(datesOut$year > maxYear | datesOut$year < minYear,
                               NA,
                               datesOut$month)
      datesOut$day <- ifelse(datesOut$year > maxYear | datesOut$year < minYear,
                             NA,
                             datesOut$day)
      # Remove non-sensical years now
      datesOut$year <- ifelse(datesOut$year > maxYear | datesOut$year < minYear,
                     NA,
                     datesOut$year)
      # Now check and replace the eventDate column if it's out of range
      datesOut$eventDate <- ifelse(lubridate::year(datesOut$eventDate) > maxYear | 
                                     lubridate::year(datesOut$eventDate) < minYear,
                                     NA,
                                     as.character(datesOut$eventDate))

    
        # For simplicity's sake, return the date columns as character...
      datesOut$eventDate  <- as.character(datesOut$eventDate)
      datesMerged$eventDate  <- as.character(datesMerged$eventDate)

      # MERGE all datsets
      dates_complete <- data %>%
        dplyr::mutate(eventDate = as.character(eventDate)) %>%
          # REMOVE the meddled-with rows
        dplyr::filter(!database_id %in% c( datesMerged$database_id, datesOut$database_id)) %>%
          # Merge the new rows back in
        dplyr::bind_rows(datesMerged, datesOut)

        # Return to date format <3 
      dates_complete$eventDate <- lubridate::ymd_hms(dates_complete$eventDate,
                         truncated = 5, quiet = TRUE)
    
      # Plot the dates
      # graphics::hist(dates_complete$eventDate, breaks = 100,
      #      main = "Histogram of eventDate output")
    
      timeEnd <- Sys.time()
    
    # Return user output
    writeLines(
        paste0(
        " - Finished. \n",
        "We now have ",
        format((sum(stats::complete.cases(dates_complete$eventDate)) - originalDateCount),
               big.mark = ","), 
        " more full eventDate cells than in the input data.\n",
        "We modified dates in \n",
        format(nrow(datesMerged), big.mark = ","), " occurrences.\n",
        " - As it stands, there are ", 
        format( sum(stats::complete.cases(dates_complete$eventDate)), big.mark = ","),
        " complete eventDates and ", 
        format( sum(is.na(dates_complete$eventDate)), big.mark = ","), 
        " missing dates.\n", 
        " - There are also ", 
        format( sum(stats::complete.cases(dates_complete$year)), big.mark = ","), 
        " complete year occurrences to filter from. This is up from an initial count of ",
        format( sum(stats::complete.cases(data$year)), big.mark = ","),
        " At this rate, you will stand to lose ",
        format( sum(is.na(dates_complete$year)), big.mark = ","), 
        " occurrences on the basis of missing",
        " year",
        " - Operation time: ", (timeEnd - timeStart)," ",
        units(round(timeEnd - timeStart, digits = 2)))
    )
    return(dates_complete)
  }

requireNamespace("dplyr")
library(dplyr)
library(lubridate)
library(stringr)

data = testData <- dplyr::tribble(
  ~database_id,               ~fieldNotes,                                                 ~eventDate, ~year, ~month, ~day,        ~verbatimEventDate, ~locality,                                ~locationRemarks, 
  "fake_SCAN01",                       "",                                                         "",    NA,     NA,   NA,                        "",   "Davis",      "coordinates obtained from Label; 6/28/05", 
  "fake_SCAN02",                       "",                                                         "",    NA,     NA,   NA,                        "",   "Davis", "coordinates obtained from Label: 3 march 2022", 
  "fake_SCAN03",            "28 Jun 2005",                                                         "",    NA,     NA,   NA,                        "",   "Davis",               "coordinates obtained from Label", 
  "fake_SCAN04",                       "",                                                         "",  2022,      2,    1,                        "",   "Davis",               "coordinates obtained from Label", 
  "fake_SCAN05",                       "",                                                         "",    NA,     NA,   NA, "28 Jun 2005/29 Jun 2005",   "Davis",               "coordinates obtained from Label", 
  "fake_SCAN06",                       "",                                                         "", 2005L,     6L,  28L, "28 Jun 2005/29 Jun 2005",   "Davis",               "coordinates obtained from Label", 
  "fake_SCAN07",                       "",                                                         "",    NA,     NA,   NA,              "28 IX 2005",   "Davis",               "coordinates obtained from Label", 
  "fake_SCAN08",                       "",                                                         "",    NA,     NA,   NA,                  "I 2022",   "Davis",               "coordinates obtained from Label", 
  "fake_SCAN09",                       "",                                                         "",    NA,     NA,   NA,                    "2022",   "Davis",               "coordinates obtained from Label", 
  "fake_SCAN10",                       "",                                               "2005-05-05",    NA,     NA,   NA,                    "2022",   "Davis",               "coordinates obtained from Label", 
  "fake_SCAN11",                       "",                                  "28 Jun 2005/29 Jun 2005",    NA,     NA,   NA,                        "",   "Davis",               "coordinates obtained from Label", 
  "fake_SCAN12",                       "",                                       "15-17 OCTOBER 1976",    NA,     NA,   NA,                    "2022",   "Davis",               "coordinates obtained from Label", 
  "fake_SCAN13",                       "",                            "28 September - 4 October 2005",    NA,     NA,   NA,                    "    ",   "Davis",               "coordinates obtained from Label", 
  "fake_SCAN14",                       "",      "2019-07-29T14:00:00+02:00/2019-07-29T15:00:00+02:00",    NA,     NA,   NA,                    "    ",   "Davis",               "coordinates obtained from Label", 
  "fake_SCAN15",                       "",                                  "07 to 21 September 2009",    NA,     NA,   NA,                    "    ",   "Davis",               "coordinates obtained from Label", 
  "fake_SCAN16",                       "",                              "Aug 30 1921 to Sept. 5 1921",    NA,     NA,   NA,                    "    ",   "Davis",               "coordinates obtained from Label", 
  "fake_SCAN17",                       "",                          "2 October 2003 - 3 October 2003",    NA,     NA,   NA,                    "    ",   "Davis",               "coordinates obtained from Label", 
  "fake_SCAN18",                       "",                                 "1940-06-26 to 1940-06-28",    NA,     NA,   NA,                    "    ",   "Davis",               "coordinates obtained from Label", 
  "fake_SCAN19",                       "", "startDateTime:201908301000xx; EndDateTime:201908301000xx",    NA,     NA,   NA,                    "    ",   "Davis",               "coordinates obtained from Label", 
  "fake_SCAN20",                       "",                                 "9.viii.2000/11.viii.2000",    NA,     NA,   NA,                    "    ",   "Davis",               "coordinates obtained from Label", 
  "fake_SCAN21",                       "",                                               "05-05-2005",    NA,     NA,   NA,                    "    ",   "Davis",               "coordinates obtained from Label", 
  "fake_SCAN22",                       "",                                               "05-28-2020",    NA,     NA,   NA,                    "    ",   "Davis",               "coordinates obtained from Label", 
  "fake_SCAN23",                       "",                                                 "06/28/05",    NA,     NA,   NA,                        "",   "Davis",               "coordinates obtained from Label", 
  "fake_SCAN24", "01.vii.2000/13.xi.2001",                                                         "",    NA,     NA,   NA,                        "",   "Davis",               "coordinates obtained from Label",
  "fake_SCAN25",            "III-05-2020",                                                         "",    NA,     NA,   NA,                        "",   "Davis",               "coordinates obtained from Label",
  "fake_SCAN26",           "XIII-10-2020",                                                         "",    NA,     NA,   NA,                        "",   "Davis",               "coordinates obtained from Label",
  "fake_SCAN27",               "5/6/2020",                                                         "",    NA,     NA,   NA,                        "",   "Davis",               "coordinates obtained from Label",
  "fake_SCAN28",  "15/6/2020 - 15/7/2020",                                                         "",    NA,     NA,   NA,                        "",   "Davis",               "coordinates obtained from Label",
  "fake_SCAN29",                 "Mar-22",                                                         "",    NA,     NA,   NA,                        "",   "Davis",               "coordinates obtained from Label"
  )


testOut <- dateFindR(data = testData,
                        # Years above this are removed (from the recovered dates only)
                        maxYear = lubridate::year(Sys.Date()),
                        # Years below this are removed (from the recovered dates only)
                        minYear = 1700) %>%
  dplyr::select(database_id, eventDate, fieldNotes, locationRemarks, verbatimEventDate, 
                day, month, year, startDayOfYear, endDayOfYear) %>%
  dplyr::arrange(desc(database_id))


# Test the expected results
testthat::test_that("dateFindR results successfully matched", {
  testthat::expect_equal(sum(complete.cases(testOut$eventDate) ), 27)
})
testthat::test_that("dateFindR results unsuccessfully matched", {
  testthat::expect_equal(sum(is.na(testOut$eventDate) ), 2)
})

testthat::test_that("dateFindR output dates match", {
  testthat::expect_equal(testOut %>% dplyr::arrange(database_id) %>% dplyr::pull(eventDate),
                         c("2005-06-28 00:00:00 UTC", "2022-03-03 00:00:00 UTC", "2005-06-28 00:00:00 UTC", "2022-02-01 00:00:00 UTC",
                           "2005-06-28 00:00:00 UTC", "2005-06-28 00:00:00 UTC", "2005-09-28 00:00:00 UTC", "2022-01-01 00:00:00 UTC",
                           NA                       , "2005-05-05 00:00:00 UTC", "2005-06-29 00:00:00 UTC", "1976-10-17 00:00:00 UTC",
                           "2005-10-04 00:00:00 UTC", "2019-07-29 13:00:00 UTC", "2009-09-21 00:00:00 UTC", "1921-09-05 00:00:00 UTC",
                           "2003-10-03 00:00:00 UTC", "1940-06-28 00:00:00 UTC", "2019-08-30 10:00:00 UTC", "2000-08-11 00:00:00 UTC",
                           "2005-05-20 05:00:00 UTC", "2020-05-28 00:00:00 UTC", "2005-06-28 00:00:00 UTC", "2000-07-01 00:00:00 UTC",
                           "2020-03-05 00:00:00 UTC",                        NA, "2020-01-01 00:00:00 UTC", "2020-06-15 00:00:00 UTC",
                           "2022-03-01 00:00:00 UTC") %>%
                           lubridate::ymd_hms(quiet = TRUE))
})

  # Test classes
testthat::test_that("dateFindR expected class", {
  testthat::expect_type(testOut, "list")
})
testthat::test_that("dateFindR expected class", {
  testthat::expect_equal(attributes(testOut)$class, c("tbl_df","tbl" ,"data.frame"))
})




troubleRows <- dplyr::tribble(
           ~database_id,                                                                                                                                          ~fieldNotes,                                                                                                     ~locality,                                ~eventDate,                     ~eventTime, ~startDayOfYear, ~endDayOfYear, ~day, ~month, ~year,   ~verbatimEventDate, ~locationRemarks,
   "NC_checklist_32750",                                                                                                                 "lot_id: 6994; preparation: pointed",        "\"GSMNP Oconaluftee residential area, 35°10'3\"\"N 83°18'16\"\"W, 2039ft, Old Field, Swain Co., NC\"",                     "2004-07-14T00:00:00",                             NA,              NA,            NA,  14L,     7L, 2004L,          "14-Jul-04",               NA,
    "NC_checklist_3666",   "[Savanna at Burgaw, N. C.] [T. B. Mitchell Oct. 23, 1957] [Hellanthus] [Andrena accepta Vier. ♀ det. Mitch. '62] [Andrena (Pterandrena) accepta]",                                      "Savanna, Burgaw, Pender Co., NC, Pender, North Carolina, United States",                                        NA,                             NA,              NA,            NA,   NA,     NA,    NA,                   NA,               NA,
   "NC_checklist_45856",                                              "C. Dasch Collection [Mt. Mitchell, N.C. 5500 ft.; added 12-JUL-2002] [Aug 17 1970; added 12-JUL-2002]",                                                                      "Mount Mitchell, 5500ft, Yancey Co., NC",                               "8/17/1970",                             NA,              NA,            NA,   NA,     NA,    NA,                   NA,               NA,
   "NC_checklist_45975",                                              "C. Dasch Collection [Coccino Cape Hatteras NC.; added 12-JUL-2002] [Aug. 24, 1971; added 12-JUL-2002]",                                                                        "Cape Hatteras, Coccino, Dare Co., NC",                               "8/24/1971",                             NA,              NA,            NA,   NA,     NA,    NA,                   NA,               NA,
   "NC_checklist_45007",                                                                                                 "[Mt. Mitchell N.C.5500 ft. Aug. 13, 1970 C. Dasch]",                                                                      "Mount Mitchell, 5500ft, Yancey Co., NC",                               "8/13/1970",                             NA,              NA,            NA,   NA,     NA,    NA,                   NA,               NA,
   "NC_checklist_87482",                                                                                                                                                   NA,                                                                                              "Southern Pines",                               "4/14/2010",                             NA,            104L,            NA,  14L,     4L, 1910L,          "4/14/2010",               NA,
   "NC_checklist_49805",                                                                                     "[Balsam, N.C.; added 12-JUL-2002] [VI-1-56; added 12-JUL-2002]",                                                                                     "Balsam, Jackson Co., NC",                                "6/1/1956",                             NA,              NA,            NA,   NA,     NA,    NA,                   NA,               NA,
   "NC_checklist_49033",                                              "C. Dasch Collection [Mt. Mitchell, N.C. 5500 ft.; added 12-JUL-2002] [Aug 13 1970; added 12-JUL-2002]",                                                                      "Mount Mitchell, 5500ft, Yancey Co., NC",                               "8/13/1970",                             NA,              NA,            NA,   NA,     NA,    NA,                   NA,               NA,
   "NC_checklist_46230",                                                                                                                                                   NA,                                                                                              "Southern Pines",                                        NA,                             NA,              NA,            NA,   NA,     NA,    NA,          "31-Jul-18",               NA,
   "NC_checklist_44059",                                                                                                                                                   NA,                                                              "Smokemont, Great Smoky Mountains National Park",                     "1835-08-01T00:00:00",                             NA,            243L,          243L,   NA,     8L, 1835L,           "Aug. '35",               NA,
  "NC_checklist_100075",                                                                                                                                                   NA,                                                                                                    "Savannah",                              "1954-00-00",                             NA,              NA,            NA,  10L,    13L, 1954L,         "1954-13-10",               NA,
   "NC_checklist_28295",                                                                                                                                                   NA,                                                                                   "Valley of Black Mountains",                                        NA,                             NA,              NA,            NA,   NA,     NA,    NA,          "6/20/1906",               NA,
   "NC_checklist_26803",                                                                                                                                                   NA,                                                                                              "Southern Pines",                                        NA,                             NA,              NA,            NA,   NA,     NA,    NA,          "6/13/1918",               NA,
   "NC_checklist_18034",                                                                                                                                                   NA,                                                                                                     "Raleigh",                                        NA,                             NA,              NA,            NA,   NA,     NA,    NA,     "mid April 1921",               NA,
   "NC_checklist_16325",                                                                                                                                                   NA,                                                                                                     "Raleigh",                                        NA,                             NA,              NA,            NA,   NA,     NA,    NA, "mid September 1921",               NA,
   "NC_checklist_36870",                                                                                                                 "lot_id: 6710; preparation: pointed", "\"GSMNP The Purchase, nr. house, 35°35'8\"\"N 83°4'23\"\"W, 1945ft, cut lawn/forest edge, Haywood Co., NC\"",                     "2003-06-27T00:00:00",                             NA,              NA,            NA,  27L,     6L, 2003L,          "27-Jun-03",               NA,
   "NC_checklist_27278",                                                                                                                                                   NA,                                                                              "Kitty Hawk, 703 Kitty Hawk Rd.", "2015-08-14T16:00:00/2015-08-15T16:00:00",                             NA,              NA,            NA,   NA,     NA,    NA,          "7/30/2011",               NA,
   "NC_checklist_58897",                                                                                                                                                   NA,                                                                                                    "Havelock",                   "2015-08-14/2015-08-15",                             NA,              NA,            NA,   NA,     4L,    9L,   "10/27/2015 11:56",               NA,
    "NC_checklist_3249",                                                        "[N CAROLINA: Carteret Co. Fort Macon State Park Sweep Netting 20-X-2005 Coll: Randy Newman]",                            "Fort Macon State Park, Carteret Co., NC, Carteret, North Carolina, United States",                                        NA,                             NA,              NA,            NA,   NA,     NA,    NA,                   NA,               NA
  ) %>%
  dplyr::select(database_id, eventDate, fieldNotes, locationRemarks, verbatimEventDate, locality,
                day, month, year, startDayOfYear, endDayOfYear)


testOut2 <- dateFindR(data = troubleRows,
                             # Years above this are removed (from the recovered dates only)
                             maxYear = lubridate::year(Sys.Date()),
                             # Years below this are removed (from the recovered dates only)
                             minYear = 1700) %>%
  dplyr::arrange(desc(database_id))



# Test the expected results
testthat::test_that("dateFindR results successfuly matched", {
  testthat::expect_equal(sum(complete.cases(testOut2$eventDate) ), 18)
})
testthat::test_that("dateFindR results unsuccessfuly matched", {
  testthat::expect_equal(sum(is.na(testOut2$eventDate) ), 1)
})

testthat::test_that("dateFindR output dates match", {
  testthat::expect_equal(testOut2$eventDate ,
                         c("1910-04-14 00:00:00 UTC", "2015-08-15 00:00:00 UTC", "2002-07-12 00:00:00 UTC", "2002-07-12 00:00:00 UTC",
                           "2018-07-31 00:00:00 UTC", "2002-07-12 00:00:00 UTC", "2002-07-12 00:00:00 UTC", "1970-08-13 00:00:00 UTC",
                           "1835-08-01 00:00:00 UTC", "2003-06-27 00:00:00 UTC", "2023-10-01 00:00:00 UTC", "2004-07-14 00:00:00 UTC",
                           "2005-10-20 00:00:00 UTC", "1906-06-20 00:00:00 UTC", "2015-08-15 16:00:00 UTC", "1918-06-13 00:00:00 UTC",
                           "1921-04-01 00:00:00 UTC", "1921-09-01 00:00:00 UTC", NA  ) %>%
                           lubridate::ymd_hms(quiet = TRUE))
})

# Test classes
testthat::test_that("dateFindR expected class", {
  testthat::expect_type(testOut2, "list")
})
testthat::test_that("dateFindR expected class", {
  testthat::expect_equal(attributes(testOut2)$class, c("tbl_df","tbl" ,"data.frame"))
})

# This function was written starting on the 24th of June 2024 by James B Dorey in order
# to wrap iNEXT::ggiNEXT to work with iNEXTwrapper and plot multiple accumulation curves
# at once across pages.

#' ggplot2 extension for a Chao- and iNEXT-wrapper outputs
#' 
#' `BeeBDC::ggRichnessWrapper()` takes data output from `BeeBDC::iNEXTwrapper()` and 
#' `BeeBDC::ChaoWrapper()` to produce plots for multiple groups in one go. The plots can be 
#' multiple per page and across multiple pages.
#' 
#' @param iNEXT_in a list of \code{iNEXT} objects computed by `BeeBDC::iNEXTwrapper()`.
#' @param iChao_in a list of R data created `BeeBDC::ChaoWrapper()`.
#' @param filterOut Character. A list of sites/countries to exclude from plotting; for example,
#' because sample size was inadequate. Default = NULL.
#' @param type three types of plots: sample-size-based rarefaction/extrapolation curve (\code{type = 1}); 
#' sample completeness curve (\code{type = 2}); coverage-based rarefaction/extrapolation curve (\code{type = 3}).
#' From `iNEXT::ggiNEXT()`
#' @param se a logical variable to display confidence interval around the estimated sampling curve.
#' From `iNEXT::ggiNEXT()`
##' @param facet.var create a separate plot for each value of a specified variable: 
#'  no separation \cr (\code{facet.var="None"}); 
#'  a separate plot for each diversity order (\code{facet.var="Order.q"}); 
#'  a separate plot for each assemblage (\code{facet.var="Assemblage"}); 
#'  a separate plot for each combination of order x assemblage (\code{facet.var="Both"}).              
#' From `iNEXT::ggiNEXT()`
#' @param color.var create curves in different colors for values of a specified variable:
#'  all curves are in the same color (\code{color.var="None"}); 
#'  use different colors for diversity orders (\code{color.var="Order.q"}); 
#'  use different colors for sites (\code{color.var="Assemblage"}); 
#'  use different colors for combinations of order x assemblage (\code{color.var="Both"}).  
#' From `iNEXT::ggiNEXT()`
#' @param grey a logical variable to display grey and white ggplot2 theme. 
#' From `iNEXT::ggiNEXT()`
#' @param legendPerPlot Logical. If TRUE, remove the legend from each plot. Default = FALSE
#' @param show_iNEXT Logical. If TRUE, show the estimate and 95% CIs if specified for iNEXT. Default = TRUE.
#' @param showPercent Logical. If TRUE, show the prrcentage increases. Default = TRUE.
#' @param ChaoColour Character. The to be used to graph Chao estimates (95% confidence intervals
#' are shown with reduced opacity). Default = "#55AD9B".
#' @param iNEXTcolour Character. The to be used to graph iNEXT estimates (95% confidence intervals
#' are shown with reduced opacity). Default = "#FD9B63".
#' @param Chao_estimate Character. The name of the Chao estimate to use from those calculated in 
#' `SpadeR::ChaoSpecies()`. The options are "Homogeneous Model","Homogeneous (MLE)",
#' "Chao1 (Chao, 1984)","Chao1-bc","iChao1 (Chiu et al. 2014)","ACE (Chao & Lee, 1992)",
#' "ACE-1 (Chao & Lee, 1992)","1st order jackknife","2nd order jackknife". 
#' Default = "iChao1 (Chiu et al. 2014)".
#' @param nrow Numeric. The number of rows per figure. Figures (that don't fit in the
#' nrow*ncol grid) will be saved into 
#' additional files. Default = 3.
#' @param ncol Numeric. The number of columns per figure. Figures (that don't fit in the
#' nrow*ncol grid) will be saved into 
#' additional files. Default = 4.
#' @param labels Character. The labels for each sub-plot (a, b, c, ...). The default is NULL,
#' which will provide labels a-z as required.
#' @param fileName Character. Prefix to the output files. Default = "richnessPlots.pdf". 
#' Changing the suffix is a convenient way to alter the file format.
#' @param outPath Character. The fodler in which to save the plots. Default = `tempdir()`
#' @param base_width Numeric. The width, in inches, to save the plot. Default = 8.3.
#' @param base_height Numeric. The height, in inches, to save the plot. Default = 11.7.
#' @param dpi Numeric. Plot resolution. Also accepts a string input: "retina" (320), 
#' "print" (300), or "screen" (72). Applies only to raster output types. Default = 300.
#' 
#' @param ... other arguments passed on to methods. Not currently used.
#' From `iNEXT::ggiNEXT()`
#' 
#' @return Saves pdf objects and returns a summary table for all levels
#' 
#' @export
#' 
#' @examples
#' \dontrun{
#' 
#     # Use the example data 
#' data(beesCountrySubset)
#' 
#'    # Transform data for iNEXT
#'  data_nextWrapper <- beesCountrySubset %>%
#'    dplyr::group_by(scientificName, country_suggested) %>%
#'    dplyr::count() 
#'    
#'  # Calculate iNEXT with the wrapper function
#'  output_iNEXTwrapper <- BeeBDC::iNEXTwrapper(data = data_nextWrapper,
#'                                              variableColumn = "country_suggested",
#'                                              valueColumn = "n",
#'                                              mc.cores = 1)
#'
#'  # Transform data for iChao
#' data_iChao <- beesCountrySubset %>%
#'   dplyr::group_by(scientificName, country_suggested) %>%
#'   dplyr::count() %>%
#'   dplyr::select(scientificName, country_suggested, n) %>%
#'   tidyr::pivot_wider(names_from = country_suggested,
#'                      values_from = n,
#'                      values_fill = 0) %>%
#'   ## Create the rownames
#'   tibble::column_to_rownames("scientificName") %>%
#'   dplyr::tibble()
#'   
#'   # Run the wrapper function
#' output_iChaowrapper <- BeeBDC::ChaoWrapper(data = data_iChao,
#'                                              datatype = "abundance",
#'                                              k = 10,
#'                                              conf = 0.95,
#'                                              mc.cores = 1)
#'                                              
#'    # Make the plots! 
#' plot_summary <- BeeBDC::ggRichnessWrapper(
#' iNEXT_in = output_iNEXTwrapper,
#' iChao_in = output_iChaowrapper,
#' nrow = 2,
#' ncol = 2,
#' labels = NULL,
#' fileName = "speciesRichnessPlots.pdf",
#' outPath = tempdir(),
#' base_width = 8.3,
#' base_height = 11.7, 
#' dpi = 300)
#'       
#' }                                       



ggRichnessWrapper <- function(
    iNEXT_in = country_iNEXT,
    iChao_in = NULL,
    filterOut = NULL,
    type = 1,
    se = TRUE,
    facet.var = "None",
    color.var = "Order.q",
    grey = FALSE,
    legendPerPlot = FALSE,
    show_iNEXT = TRUE,
    showPercent = TRUE,
    ChaoColour = "#55AD9B",
    iNEXTcolour = "#FD9B63",
    Chao_estimate = "iChao1 (Chiu et al. 2014)",
    nrow = 3,
    ncol = 4,
    labels = NULL,
    fileName = "richnessPlots.pdf",
    outPath = tempdir(),
    base_width = 8.3,
    base_height = 11.7, 
    dpi = 300,
    ...){
  
  decimalLatitude <- decimalLongitude <- database_id <- scientificName <- NULL
  country_iNEXT <- . <- Est_s.e. <- NULL
    
  requireNamespace("magrittr")
  requireNamespace("iNEXT")
  requireNamespace("cowplot")
  requireNamespace("ggplot2")
  
  
  #### 0.0 Prepare function ####
    ##### 0.1 Fatal warnings ####
  if (is.null(iNEXT_in)) {
    stop("No iNEXT_in provided. Please provide some data produced from BeeBDC::ggRichnessWrapper().")
  }
 
  ##### 0.2 iNEXT test ####
  ###### a. test ####
  # Check if iNEXT is installed
  # TRUE if iNEXT is found
  suppressWarnings(
    suggestedTest <- system.file(package='iNEXT') %>% 
      stringr::str_count() > 0 
  )
  
  ###### b. iNEXT ####
  if(suggestedTest == FALSE){
    # Set up instructions for download on fail
    instructions <- paste(" Please try installing the package for yourself", 
                          "using the following command: \n",
                          "install.packages(\"iNEXT\")")
    # Set up fail function for tryCatch
    error_func <- function(e){
      stop(paste("Failed to install the iNEXT package.\n", 
                 instructions))
    }
    # Begin interactive input
    input <- 1
    if (interactive()){
      input <- utils::menu(c("Yes", "No"), 
                           title = paste0("Install the iNEXT package? \n"))
    }
    if(input == 1){
      # Start iNEXT install
      message("Installing the iNEXT package.")
      tryCatch(
        utils::install.packages("iNEXT"), 
        error = error_func, warning = error_func)
    } # END input == 1
    
    else{
      stop(writeLines(paste("The iNEXT package is necessary for BeeBDC::iNEXTwrapper\n", 
                            instructions)))
    } # END else
  } # END suggestedTest == FALSE
  
  
    ##### 0.2 Prepare values ####
    # If no labels were provided, take them as the length of nrow and ncol a:z
  if(is.null(labels)){
    labels = LETTERS[1:(nrow*ncol)]
  }
  
  ##### 0.3 Functions ####
  getfun <- function(x) {
    if(length(grep("::", x)) > 0) {
      parts <- strsplit(x, "::")[[1]]
      getExportedValue(parts[1], parts[2])
    } else {
      x
    }
  }
  
  #### 1.0 Prepare data ####
    ##### 1.0 Extract listed countries ####
  dataExtracted <- iNEXT_in$iNextEst$iNextEst
  
    # Filter out levels in filterOut
  if(!is.null(filterOut)){
    dataExtracted <- dataExtracted[!names(dataExtracted) %in% filterOut]
  }
  
    # Create an empty data table to output
  if(!is.null(iChao_in)){
  statisticTable <- dplyr::tibble(
    level = NA_character_,
    n = NA_integer_, observedRichness = NA_integer_, 
    iNEXT_est = NA_integer_, iNEXT_lower = NA_integer_, iNEXT_upper = NA_integer_, 
    iNEXT_increasePercent = NA_integer_,
    iChao_est = NA_integer_, iChao_lower = NA_integer_, iChao_upper = NA_integer_, 
    iChao_increasePercent = NA_integer_
  ) %>%
    tidyr::drop_na()}else{
      statisticTable <- dplyr::tibble(
        level = NA_character_,
        n = NA_integer_, observedRichness = NA_integer_, 
        iNEXT_est = NA_integer_, iNEXT_lower = NA_integer_, iNEXT_upper = NA_integer_, 
        iNEXT_increasePercent = NA_integer_
      ) %>%
        tidyr::drop_na()
    }
  
  #### 1.0 Make plots ####
    ##### 1.1 Build plots ####
  ggiNEXT_fun <- function(dataIn = dataExtracted,
                      type = type,
                      se = se,
                      facet.var = facet.var,
                      color.var = color.var,
                      grey = grey){
    variable <- Name <- scale_x_continuous <- NULL
      # extract the iNEXT data itself
    dataWithin <- dataIn$iNextEst
    
    if(!is.null(iChao_in)){
      # Get the chao data for the current level
      levelChao <- iChao_in$richnessTable %>%
        dplyr::filter(variable == dataIn$DataInfo$Assemblage) %>%
        dplyr::filter(Name %>% stringr::str_squish() == Chao_estimate)
    }
    

        # Don't show choa in title if it's NULL
      if(is.null(iChao_in)){
        iNEXT_plot <- iNEXT::ggiNEXT(x = dataWithin,
                                     type = type,
                                     se = se,
                                     facet.var = facet.var,
                                     color.var = color.var,
                                     grey = grey) +
      ggplot2::ggtitle(paste0(dataIn$AsyEst$groupVariable %>% unique(),"\n",
                              "n = ", format(dataIn$DataInfo$n, big.mark = ","),"\n",
                              "; Obs. = ", format(dataIn$AsyEst$Observed[[1]] %>% round(0),
                                                    big.mark = ","),
                              "; Est. = ", format(dataIn$AsyEst$Estimator[[1]] %>% round(0),
                                                     big.mark = ","),
                                # Calculate percentage
                              if(showPercent == TRUE){
                                paste0(" (+",((1 - (dataIn$AsyEst$Observed[[1]] / dataIn$AsyEst$Estimator[[1]]))*100) %>% 
                                         round(0),"%)")
                              }, # END showPercent
                              "\n",
                              "Lower = ", format(dataIn$AsyEst$`95% Lower`[[1]] %>% round(0),
                                                 big.mark = ","),
                              "; Upper = ", format(dataIn$AsyEst$`95% Upper`[[1]] %>% round(0),
                                                 big.mark = ",")
                              )) + 
          ggplot2::theme_classic() } 
      # DO show choa in title if it's provided
      if(!is.null(iChao_in)){
        iNEXT_plot <- iNEXT::ggiNEXT(x = dataWithin,
                                     type = type,
                                     se = se,
                                     facet.var = facet.var,
                                     color.var = color.var,
                                     grey = grey) +
        ggplot2::ggtitle(paste0(dataIn$AsyEst$groupVariable %>% unique(),"\n",
                                "n = ", format(dataIn$DataInfo$n, big.mark = ","),
                                "; Obs. = ", format(dataIn$AsyEst$Observed[[1]] %>% round(0),
                                                  big.mark = ","),"\n",
                                  # iNEXT
                                "iNEXT = ", format(dataIn$AsyEst$Estimator[[1]] %>% round(0),
                                                    big.mark = ","),
                                # Calculate percentage
                                " (", format(dataIn$AsyEst$`95% Lower`[[1]] %>% round(0),
                                                   big.mark = ","),
                                "-", format(dataIn$AsyEst$`95% Upper`[[1]] %>% round(0),
                                                     big.mark = ","),
                                if(showPercent == TRUE){
                                  paste0("; +",((1 - (dataIn$AsyEst$Observed[[1]] / dataIn$AsyEst$Estimator[[1]]))*100) %>% 
                                           round(0),"%)")
                                }else{paste0(")")},
                                  # iCHAO
                                # Get the chao data for the current level
                                "\niChao = ", format(levelChao$Estimate %>% round(0),
                                                 big.mark = ","),
                                # Calculate percentage
                                " (", format(levelChao$`95%Lower` %>% round(0),
                                             big.mark = ","),
                                "-", format(levelChao$`95%Upper` %>% round(0),
                                            big.mark = ","),
                                if(showPercent == TRUE){
                                  paste0("; +",((1 - (dataIn$AsyEst$Observed[[1]] / levelChao$Estimate))*100) %>% 
                                           round(0),"%)")
                                }else{paste0(")")}
                                
        )) + 
          ggplot2::theme_classic() 
        }
    
    ###### a. legend ####
    
      # Remove the legend from each plot
    if(legendPerPlot == FALSE){
      iNEXT_plot <- iNEXT_plot + 
        ggplot2::theme(legend.position="none")
    } # END legendPerPlot
      
      ###### b. iNext 95% ####
      # Show the estimate and 95% CIs if specified for iNEXT
    if(show_iNEXT == TRUE){
      iNEXT_plot <- iNEXT_plot +
        ggplot2::geom_hline(yintercept = dataIn$AsyEst$Estimator[[1]], linetype="solid", color = iNEXTcolour) +
        ggplot2::geom_hline(yintercept = dataIn$AsyEst$`95% Lower`[[1]], linetype="dashed", color = iNEXTcolour) +
        ggplot2::geom_hline(yintercept = dataIn$AsyEst$`95% Upper`[[1]], linetype="dashed", color = iNEXTcolour) +
        ggplot2::geom_ribbon(ggplot2::aes(ymin = dataIn$AsyEst$`95% Upper`[[1]],  
                                          ymax = dataIn$AsyEst$`95% Lower`[[1]]),
                             fill = iNEXTcolour, alpha = 0.1, colour = NA) +
          # Have the plot start at zero on X and Y 
        ggplot2::scale_x_continuous(expand = c(0.001, 0.001)) + ggplot2::scale_y_continuous(expand = c(0.001, 0.001))
    } # END show_iNEXT
      
    ###### c. Chao 95% ####
      # Show the estimate and 95% CIs if specified for iChao
      if(!is.null(iChao_in)){
          # Get the chao data for the current level
        iNEXT_plot <- iNEXT_plot +
          ggplot2::geom_hline(yintercept = levelChao$Estimate, linetype="solid", color = ChaoColour) +
          ggplot2::geom_hline(yintercept = levelChao$`95%Lower`, linetype="dashed", color = ChaoColour) +
          ggplot2::geom_hline(yintercept = levelChao$`95%Upper`, linetype="dashed", color = ChaoColour) +
          ggplot2::geom_ribbon(ggplot2::aes(ymin = levelChao$`95%Lower`,  
                                            ymax = levelChao$`95%Upper`),
                               fill = ChaoColour, alpha = 0.1, colour = NA) 
      } # END iChao_in
    
     
    # Return the plot 
    return(iNEXT_plot)
  } # END ggiNEXT_fun 
  
  


  # Make the plots per country using lapply
  countryPlots <- dataExtracted %>%
    lapply(X = .,
           FUN = ggiNEXT_fun,
           type = type,
           se = se,
           facet.var = facet.var,
           color.var = color.var,
           grey = grey)
  

  ##### 1.2 Extract data table ####
  dataExtract_fun <- function(dataIn = dataExtracted,
                          type = type,
                          se = se,
                          facet.var = facet.var,
                          color.var = color.var,
                          grey = grey){
    variable <- Name <- NULL
    # extract the iNEXT data itself
    dataWithin <- dataIn$iNextEst
    
    if(!is.null(iChao_in)){
      # Get the chao data for the current level
      levelChao <- iChao_in$richnessTable %>%
        dplyr::filter(variable == dataIn$DataInfo$Assemblage) %>%
        dplyr::filter(Name %>% stringr::str_squish() == Chao_estimate)
    }
    
    # Simulataneously, create a row of data and add it to the output table
    
    if(!is.null(iChao_in)){
      statisticTable <-  dplyr::tibble(
        level = dataIn$AsyEst$groupVariable %>% unique(),
        n = dataIn$DataInfo$n, observedRichness = dataIn$AsyEst$Observed[[1]], 
        iNEXT_est = dataIn$AsyEst$Estimator[[1]] %>% round(0), 
        iNEXT_lower = dataIn$AsyEst$`95% Lower`[[1]] %>% round(0),
        iNEXT_upper = dataIn$AsyEst$`95% Upper`[[1]] %>% round(0), 
        iNEXT_increasePercent = ((1 - (dataIn$AsyEst$Observed[[1]] / dataIn$AsyEst$Estimator[[1]]))*100) %>%
          round(2),
        iChao_est = levelChao$Estimate %>% round(0), 
        iChao_lower = levelChao$`95%Lower` %>% round(0), iChao_upper = levelChao$`95%Upper` %>% round(0), 
        iChao_increasePercent = ((1 - (dataIn$AsyEst$Observed[[1]] / levelChao$Estimate))*100) %>% 
          round(2)
      ) }else{
        statisticTable <-  dplyr::tibble(
          level = dataIn$AsyEst$groupVariable %>% unique(),
          n = dataIn$DataInfo$n, observedRichness = dataIn$AsyEst$Observed[[1]], 
          iNEXT_est = dataIn$AsyEst$Estimator[[1]] %>% round(0), 
          iNEXT_lower = dataIn$AsyEst$`95% Lower`[[1]],  iNEXT_upper = dataIn$AsyEst$`95% Upper`[[1]], 
          iNEXT_increasePercent = ((1 - (dataIn$AsyEst$Observed[[1]] / dataIn$AsyEst$Estimator[[1]]))*100) %>%
            round(2)
        )
        }
    
    # Return the plot 
    return(statisticTable)
  } # END dataExtract_fun 

  
  # Make the plots per country using lapply
  statisticTable_out <- dataExtracted %>%
    lapply(X = .,
           FUN = dataExtract_fun,
           type = type,
           se = se,
           facet.var = facet.var,
           color.var = color.var,
           grey = grey) %>%
    dplyr::bind_rows()
  
  
  #### 2.0 Combine plots ####
    ##### 2.1 Chunk plot list ####
    # Get the number of chunks to be output
  numberOfChunks <- ceiling(length(countryPlots)/(nrow*ncol) )
  plotsPerPage <- (nrow*ncol)
  
  countryPlot_chunks <- dplyr::lst()
  j = 1
    # Put each chunk of plots into its own list
  for(i in 1:numberOfChunks){
    countryPlot_chunk_i <- countryPlots[j:(plotsPerPage + j)]
    countryPlot_chunks <- append(countryPlot_chunks, list(countryPlot_chunk_i))
    j = j + plotsPerPage 
  }
  
  
  ##### 2.2 Combine plots ####
    # A fucntion to do the cowplots
  multiPlotFun <- function(plotData = countryPlot_chunks,
                           nrow = nrow,
                           ncol = ncol,
                           labels = labels,
                           base_width = base_width,
                           base_height = base_height, 
                           dpi = dpi){
      # Make the plot grid
    plotOut <- do.call(getfun("cowplot::plot_grid"),
                       args = list(plotlist = plotData,
                                   nrow = nrow,
                                   ncol = ncol,
                                   labels = labels[1:length(plotData[lengths(plotData) > 0])]))
    
    # Save the plots
    cowplot::save_plot(
      plot = plotOut,
      filename = paste(outPath, "/", fileName, "_", names(plotData)[[1]], "_to_",
                       names(plotData)[[length(plotData)]], ".pdf", sep = ""),
      base_width = base_height,
      base_height = base_height, 
      dpi = dpi)
    
    # Return the plot grid
    return(plotOut)
  }
  
  
    # Combine the plots with cowplot
 combinedPlots <- lapply(X = countryPlot_chunks,
                         FUN = multiPlotFun,
                         nrow = nrow,
                         ncol = ncol,
                         labels = labels,
                         base_width = base_height,
                         base_height = base_height, 
                         dpi = dpi )
 
 return(statisticTable_out)

} # END ggRichnessWrapper

# This function is a modification of a bdc function and flags columns as FALSE when they have a  
# coordiante uncertainty over a user-set threshold. 
# This function was written on the 2nd of August 2022 by James Dorey. Email James at
# jbdorey[at]me.com for help.


#' Flag occurrences with an uncertainty threshold
#' 
#' To use this function, the user must choose a column, probably "coordinateUncertaintyInMeters" 
#' and a threshold above which occurrences will be flagged for geographic uncertainty.
#'
#' @param data A data frame or tibble. Occurrence records as input.
#' @param uncerColumn Character. The column to flag uncertainty in.
#' @param threshold Numeric. The uncertainty threshold. Values equal to, or greater than, this 
#' threshold will be flagged.
#'
#' @return The input data with a new column, .uncertaintyThreshold.
#' @export
#' @importFrom dplyr %>%
#'
#' @examples
#' # Run the function
#' beesRaw_out <- coordUncerFlagR(data = beesRaw,
#'                                uncerColumn = "coordinateUncertaintyInMeters",
#'                                threshold = 1000)
#' # View the output
#' table(beesRaw_out$.uncertaintyThreshold, useNA = "always")
coordUncerFlagR <-
  function(data = NULL,
           uncerColumn = "coordinateUncertaintyInMeters",
           threshold = NULL) {
    .data <- .occurrenceAbsent <- NULL
    requireNamespace("dplyr")
    
    # Make a new column called .occurrenceAbsent to be TRUE when occurrenceStatus is "present" or NA
    data <-
      data %>%
      dplyr::mutate(
        .uncertaintyThreshold =
          !.data[[uncerColumn]] > threshold) 

    # Return user output
    message(
      paste(
        "\\coordUncerFlagR:\n",
        "Flagged",
        format(sum(data$.uncertaintyThreshold == FALSE, na.rm = TRUE), big.mark = ","),
        "geographically uncertain records:\n",
        "The column '.uncertaintyThreshold' was added to the database.\n"
      )
    )
    return(data)
  }

requireNamespace("readr")
requireNamespace("BeeBDC")


# make us some data - this is FAKE DATA derived from records from SCAN
testData <- dplyr::tribble(
                   ~basisOfRecord,   ~kingdom,      ~phylum,    ~class,        ~order,  ~family,  ~scientificName, ~taxonID, ~scientificNameAuthorship, ~genus, ~specificEpithet, ~recordedBy, ~eventDate, ~year, ~month, ~day,        ~verbatimEventDate,        ~country, ~stateProvince, ~locality,                  ~locationRemarks, ~decimalLatitude, ~decimalLongitude, ~coordinateUncertaintyInMeters,                                             ~rights,                                     ~rightsHolder,
              "PreservedSpecimen", "Animalia", "Arthropoda", "Insecta", "Hymenoptera", "apidae", "apis mellifera",  235783L,          "Linnaeus, 1758", "Apis",      "mellifera", "P.S. Ward",  "6/28/05", 2005L,     6L,  28L, "28 Jun 2005/29 Jun 2005", "United States",   "California",   "Davis", "coordinates obtained from Label",            38.54,        -121.75667,                           200L, "http://creativecommons.org/publicdomain/zero/1.0/", "The California Academy of Sciences - AntWeb.org",
              "PreservedSpecimen", "Animalia", "Arthropoda", "Insecta", "Hymenoptera", "apidae", "apis mellifera",  235783L,          "Linnaeus, 1758", "Apis",      "mellifera", "P.S. Ward",  "6/28/05", 2005L,     6L,  28L, "28 Jun 2005/29 Jun 2005", "United States",   "California",   "Davis", "coordinates obtained from Label",            38.54,        -121.75667,                            10L, "http://creativecommons.org/publicdomain/zero/1.0/", "The California Academy of Sciences - AntWeb.org",
              "PreservedSpecimen", "Animalia", "Arthropoda", "Insecta", "Hymenoptera", "apidae", "apis mellifera",  235783L,          "Linnaeus, 1758", "Apis",      "mellifera", "P.S. Ward",  "6/28/05", 2005L,     6L,  28L, "28 Jun 2005/29 Jun 2005", "United States",   "California",   "Davis", "coordinates obtained from Label",            38.54,        -121.75667,                           500L, "http://creativecommons.org/publicdomain/zero/1.0/", "The California Academy of Sciences - AntWeb.org",
              "PreservedSpecimen", "Animalia", "Arthropoda", "Insecta", "Hymenoptera", "apidae", "apis mellifera",  235783L,          "Linnaeus, 1758", "Apis",      "mellifera", "P.S. Ward",  "6/28/05", 2005L,     6L,  28L, "28 Jun 2005/29 Jun 2005", "United States",   "California",   "Davis", "coordinates obtained from Label",            38.54,        -121.75667,                          1000L, "http://creativecommons.org/publicdomain/zero/1.0/", "The California Academy of Sciences - AntWeb.org",
              "PreservedSpecimen", "Animalia", "Arthropoda", "Insecta", "Hymenoptera", "apidae", "apis mellifera",  235783L,          "Linnaeus, 1758", "Apis",      "mellifera", "P.S. Ward",  "6/28/05", 2005L,     6L,  28L, "28 Jun 2005/29 Jun 2005", "United States",   "California",   "Davis", "coordinates obtained from Label",            38.54,        -121.75667,                           850L, "http://creativecommons.org/publicdomain/zero/1.0/", "The California Academy of Sciences - AntWeb.org",
              "PreservedSpecimen", "Animalia", "Arthropoda", "Insecta", "Hymenoptera", "apidae", "apis mellifera",  235783L,          "Linnaeus, 1758", "Apis",      "mellifera", "P.S. Ward",  "6/28/05", 2005L,     6L,  28L, "28 Jun 2005/29 Jun 2005", "United States",   "California",   "Davis", "coordinates obtained from Label",            38.54,        -121.75667,                           600L, "http://creativecommons.org/publicdomain/zero/1.0/", "The California Academy of Sciences - AntWeb.org",
              "PreservedSpecimen", "Animalia", "Arthropoda", "Insecta", "Hymenoptera", "apidae", "apis mellifera",  235783L,          "Linnaeus, 1758", "Apis",      "mellifera", "P.S. Ward",  "6/28/05", 2005L,     6L,  28L, "28 Jun 2005/29 Jun 2005", "United States",   "California",   "Davis", "coordinates obtained from Label",            38.54,        -121.75667,                           150L, "http://creativecommons.org/publicdomain/zero/1.0/", "The California Academy of Sciences - AntWeb.org",
              "PreservedSpecimen", "Animalia", "Arthropoda", "Insecta", "Hymenoptera", "apidae", "apis mellifera",  235783L,          "Linnaeus, 1758", "Apis",      "mellifera", "P.S. Ward",  "6/28/05", 2005L,     6L,  28L, "28 Jun 2005/29 Jun 2005", "United States",   "California",   "Davis", "coordinates obtained from Label",            38.54,        -121.75667,                            20L, "http://creativecommons.org/publicdomain/zero/1.0/", "The California Academy of Sciences - AntWeb.org"
              )


# run the function
testOut <- BeeBDC::coordUncerFlagR(data = testData, uncerColumn = "coordinateUncertaintyInMeters", threshold = 500)


# number of FALSE values in new column, .uncertaintyThreshold
resultsT <- length(testOut$.uncertaintyThreshold[testOut$.uncertaintyThreshold == TRUE])
resultsF <- length(testOut$.uncertaintyThreshold[testOut$.uncertaintyThreshold == FALSE])


# number of columns in old and new data
old <- ncol(testData)
new <- ncol(testOut)
testcols <- new - old


# Test the number of expected TRUE and FALSE values in the .uncertaintyThreshold column and
# verify addition of new column
testthat::test_that("coordUncerFlagR results columns TRUE", {
  testthat::expect_equal(resultsT, 5)
})

testthat::test_that("coordUncerFlagR results columns FALSE", {
  testthat::expect_equal(resultsF, 3)
})

testthat::test_that("verify addition of new column", {
  testthat::expect_equal(testcols, 1)
})

# This function was written by James Dorey to match higher order names to species.genus names
# For queries, please contact James Dorey at jbdorey[at]me.com
# This function was started on 15th May 2022 and last updated 17th May 2022
#' @importFrom dplyr %>%

HigherNamer <- function(HigherNameList = HigherOrders,
                        InSynList = DLdf){
  # locally bind variables to the function
  HigherOrders <- DLdf <- taxonomic_status <- validName <- family <- subfamily <- NULL
  subfamily <- tribe <- subtribe <- id <- NULL
  
  # Drop the completely NA row at the top
  InSynList <- InSynList %>%
    tidyr::drop_na(taxonomic_status) %>%
      # Drop some left-over Discover life stuff
    dplyr::filter(!validName %in% c("Kinds of Apoidea species", "Scientific name",
                                    "Other names")) %>%
    dplyr::filter(!stringr::str_detect(validName, "Updated: |Discover Life"))
  
  # Match and copy the Higher Order names across
  InSynList$family  <- HigherNameList$family[ cbind(match(InSynList$genus, HigherNameList$Genus ) )]
  InSynList$subfamily  <- HigherNameList$subfamily[ cbind(match(InSynList$genus, HigherNameList$Genus ) )]
  InSynList$tribe  <- HigherNameList$tribe[ cbind(match(InSynList$genus, HigherNameList$Genus ) )]
  InSynList$subtribe  <- HigherNameList$subtribe[ cbind(match(InSynList$genus, HigherNameList$Genus ) )]
  
  # Make sure these are nurmeric columns...
  InSynList$accid <- as.numeric(InSynList$accid)
  InSynList$id <- as.numeric(InSynList$id)
  
  # For those synonyms that do not have a higher taxonomy, take it from the accepted name
  emptyFamily <- InSynList %>%
    dplyr::filter(is.na(family))
    # Match to their accepted names
  emptyFamily <- emptyFamily %>%
    dplyr::select(!c(family, subfamily, tribe, subtribe)) %>%
    dplyr::left_join(InSynList %>% dplyr::select(c(family, subfamily, tribe, subtribe, id)),
                     by = c("accid" = "id"))
  
    # Re-join these data
  InSynList <- InSynList %>%
    dplyr::filter(!is.na(family)) %>%
    dplyr::bind_rows(emptyFamily)
  
  # Return the list
  return(InSynList)
} # END function

bees3sp <- BeeBDC::bees3sp


# Test the expected results
testthat::test_that("bees3sp expected number of columns", {
  testthat::expect_equal(ncol(bees3sp), 124)
})
testthat::test_that("bees3sp expected number of species", {
  testthat::expect_equal(length(unique(bees3sp$scientificName)), 3)
})


# Test classes
testthat::test_that("bees3sp expected class", {
  testthat::expect_type(bees3sp, "list")
})
testthat::test_that("bees3sp expected class", {
  testthat::expect_equal(attributes(bees3sp)$class, c("spec_tbl_df","tbl_df","tbl","data.frame" ))
})

beesRaw <- BeeBDC::beesRaw


# Test the expected results
testthat::test_that("beesRaw expected number of columns", {
  testthat::expect_equal(ncol(beesRaw), 90)
})
testthat::test_that("beesRaw expected number of rows", {
  testthat::expect_equal(nrow(beesRaw), 100)
})


# Test classes
testthat::test_that("beesRaw expected class", {
  testthat::expect_type(beesRaw, "list")
})
testthat::test_that("beesRaw expected class", {
  testthat::expect_equal(attributes(beesRaw)$class, c("spec_tbl_df","tbl_df","tbl","data.frame" ))
})

beesFlagged <- BeeBDC::beesFlagged


# Test the expected results
testthat::test_that("beesFlagged expected number of columns", {
  testthat::expect_equal(ncol(beesFlagged), 124)
})
testthat::test_that("beesFlagged expected number of rows", {
  testthat::expect_equal(nrow(beesFlagged), 100)
})


# Test classes
testthat::test_that("beesFlagged expected class", {
  testthat::expect_type(beesFlagged, "list")
})
testthat::test_that("beesFlagged expected class", {
  testthat::expect_equal(attributes(beesFlagged)$class, c("spec_tbl_df","tbl_df","tbl","data.frame" ))
})

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/dupePlotR.R
\name{dupePlotR}
\alias{dupePlotR}
\title{Create a compound bar graph of duplicate data sources}
\usage{
dupePlotR(
  data = NULL,
  outPath = NULL,
  fileName = NULL,
  legend.position.inside = c(0.85, 0.8),
  base_height = 7,
  base_width = 7,
  ...,
  dupeColours = c("#F2D2A2", "#B9D6BC", "#349B90"),
  returnPlot = FALSE
)
}
\arguments{
\item{data}{A data frame or tibble. Occurrence records as input.}

\item{outPath}{Character. The path to a directory (folder) in which the output should be saved.}

\item{fileName}{Character. The name of the output file, ending in '.pdf'.}

\item{legend.position.inside}{The position of the legend as coordinates. Default = c(0.85, 0.8).}

\item{base_height}{Numeric. The height of the plot in inches. Default = 7.}

\item{base_width}{Numeric. The width of the plot in inches. Default = 7.}

\item{...}{Other arguments to be used to change factor levels of data sources.}

\item{dupeColours}{A vector of colours for the levels duplicate, kept duplicate, and unique.
Default = c("#F2D2A2","#B9D6BC", "#349B90").}

\item{returnPlot}{Logical. If TRUE, return the plot to the environment. Default = FALSE.}
}
\value{
Outputs a .pdf figure.
}
\description{
Creates a plot with two bar graphs. One shows the absolute number of duplicate records for each
data source
while the other shows the proportion of records that are duplicated within each data source.
This function requires a dataset that has been run through \code{\link[=dupeSummary]{dupeSummary()}}.
}
\examples{

# This example will show a warning for the factor levels taht are not present in the specific 
# test dataset
dupePlotR(
  data = beesFlagged,
  # The outPath to save the plot as
    # Should be something like: #paste0(OutPath_Figures, "/duplicatePlot_TEST.pdf"),
  outPath = tempdir(), 
  fileName = "duplicatePlot_TEST.pdf",
  # Colours in order: duplicate, kept duplicate, unique
  dupeColours = c("#F2D2A2","#B9D6BC", "#349B90"),
  # Plot size and height
  base_height = 7, base_width = 7,
  legend.position.inside = c(0.85, 0.8),
  # Extra variables can be fed into forcats::fct_recode() to change names on plot
  GBIF = "GBIF", SCAN = "SCAN", iDigBio = "iDigBio", USGS = "USGS", ALA = "ALA", 
  ASP = "ASP", CAES = "CAES", 'B. Mont.' = "BMont", 'B. Minckley' = "BMin", Ecd = "Ecd",
  Gaiarsa = "Gai", EPEL = "EPEL", Lic = "Lic", Bal = "Bal", Arm = "Arm"
  )
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plotFlagSummary.R
\name{plotFlagSummary}
\alias{plotFlagSummary}
\title{Generate a plot summarising flagged data}
\usage{
plotFlagSummary(
  data = NULL,
  flagColours = c("#127852", "#A7002D", "#BDBABB"),
  fileName = NULL,
  outPath = OutPath_Figures,
  width = 15,
  height = 9,
  units = "in",
  dpi = 300,
  bg = "white",
  device = "pdf",
  speciesName = NULL,
  saveFiltered = FALSE,
  filterColumn = ".summary",
  nameColumn = NULL,
  plotMap = FALSE,
  mapAlpha = 0.5,
  xbuffer = c(0, 0),
  ybuffer = c(0, 0),
  ptSize = 1,
  saveTable = FALSE,
  jitterValue = NULL,
  returnPlot = FALSE,
  ...
)
}
\arguments{
\item{data}{A data frame or tibble. Occurrence records as input.}

\item{flagColours}{A character vector. Colours in order of pass (TRUE), fail (FALSE), and NA.
Default = c("#127852", "#A7002D", "#BDBABB").}

\item{fileName}{Character. The name of the file to be saved, ending in ".pdf".
If saving as a different file type, change file type suffix - See \code{device}.}

\item{outPath}{A character path. The path to the directory in which the figure will be saved.
Default = OutPath_Figures.}

\item{width}{Numeric. The width of the output figure in user-defined units Default = 15.}

\item{height}{Numeric. The height of the output figure in user-defined units Default = 9.}

\item{units}{Character. The units for the figure width and height passed to \code{\link[ggplot2:ggsave]{ggplot2::ggsave()}}
("in", "cm", "mm", or "px"). Default = "in".}

\item{dpi}{Numeric. Passed to \code{\link[ggplot2:ggsave]{ggplot2::ggsave()}}. Plot resolution. Also accepts a string input: "retina" (320), "print" (300), or
"screen" (72). Applies only to raster output types. Default = 300.}

\item{bg}{Character. Passed to \code{\link[ggplot2:ggsave]{ggplot2::ggsave()}}. Background colour. If NULL, uses the plot.background fill value from the plot theme.
Default = "white."}

\item{device}{Character. Passed to \code{\link[ggplot2:ggsave]{ggplot2::ggsave()}}. Device to use. Can either be a device function (e.g. png), or one of "eps", "ps", "tex" (pictex), "pdf", "jpeg", "tiff", "png", "bmp", "svg" or "wmf" (windows only).
Default = "pdf". If not using default, change file name suffix in fileName argument.}

\item{speciesName}{Optional. Character. A species name, as it occurs in the user-input nameColumn.
If provided, the data will be filtered to this species for the plot.}

\item{saveFiltered}{Optional. Logical. If TRUE, the filtered data will be saved to the computer
as a .csv file.}

\item{filterColumn}{Optional. The flag column to display on the map. Default = .summary.}

\item{nameColumn}{Optional. Character. If speciesName is not NULL, enter the column to look
for the species in. A User might realise that, combined with speciesName, figures can be made for
a variety of factors.}

\item{plotMap}{Logical. If TRUE, the function will produce a point map. Tested for use with one
species at a time; i.e., with speciesName is not NULL.}

\item{mapAlpha}{Optional. Numeric. The opacity for the points on the map.}

\item{xbuffer}{Optional. Numeric vector. A buffer in degrees of the amount to increase the
min and max bounds along the
x-axis. This may require some experimentation, keeping in mind
the negative and positive directionality of hemispheres. Default = c(0,0).}

\item{ybuffer}{Optional. Numeric vector. A buffer in degrees of the amount to increase the
min and max bounds along the y-axis. This may require some experimentation, keeping in mind
the negative and positive directionality of hemispheres. Default = c(0,0).}

\item{ptSize}{Optional. Numeric. The size of the points as passed to ggplot2. Default = 1.}

\item{saveTable}{Optional. Logical. If TRUE, the function will save the data used to produce the
compound bar plot.}

\item{jitterValue}{Optional. Numeric. The value to jitter points by in the map in decimal degrees.}

\item{returnPlot}{Logical. If TRUE, return the plot to the environment. Default = FALSE.}

\item{...}{Optional. Extra variables to be fed into \code{\link[forcats:fct_recode]{forcats::fct_recode()}} to change names on plot.
For example... 'B. Mont.' = "BMont", 'B. Minkley' = "BMin", Ecd = "Ecd", Gaiarsa = "Gai"}
}
\value{
Exports a compound bar plot that summarises all flag columns. Optionally can also return
a point map for a particular species in tandem with the summary plot.
}
\description{
Creates a compound bar plot that shows the proportion of records that pass or fail each flag (rows)
and for each data source (columns). The function can also optionally return a point map for
a user-specified species when plotMap = TRUE. This function requires that your dataset has been
run through some filtering functions - so that is can display logical columns starting with
".".
}
\examples{
# import data
data(beesFlagged)
OutPath_Figures <- tempdir()
 # Visualise all flags for each dataSource (simplified to the text before the first underscore)
plotFlagSummary(
  data = beesFlagged,
  # Colours in order of pass (TRUE), fail (FALSE), and NA
  flagColours = c("#127852", "#A7002D", "#BDBABB"),
  fileName = paste0("FlagsPlot_TEST_", Sys.Date(),".pdf"),
  outPath = OutPath_Figures,
  width = 15, height = 9,
  # OPTIONAL:
  #\   #  # Filter to species
  #\   speciesName = "Holcopasites heliopsis",
  #\   # column to look in
  #\   nameColumn = "species",
  #\   # Save the filtered data
  #\   saveFiltered = TRUE,
  #\   # Filter column to display on map
  #\   filterColumn = ".summary",
  #\   plotMap = TRUE,
  #\   # amount to jitter points if desired, e.g. 0.25 or NULL
  #\   jitterValue = NULL,
  #\   # Map opacity value for points between 0 and 1
  #\   mapAlpha = 1,
  # Extra variables can be fed into forcats::fct_recode() to change names on plot
  GBIF = "GBIF", SCAN = "SCAN", iDigBio = "iDigBio", USGS = "USGS", ALA = "ALA", 
  ASP = "ASP", CAES = "CAES", 'B. Mont.' = "BMont", 'B. Minkley' = "BMin", Ecd = "Ecd",
  Gaiarsa = "Gai", EPEL = "EPEL"
)



}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/flagLicense.R
\name{flagLicense}
\alias{flagLicense}
\title{Flag license protected records}
\usage{
flagLicense(data = NULL, strings_to_restrict = "all", excludeDataSource = NULL)
}
\arguments{
\item{data}{A data frame or tibble. Occurrence records as input.}

\item{strings_to_restrict}{A character vector. Should contain the strings used to detect protected records.
Default =  c("All Rights Reserved", "All rights reserved", "All rights reserved.", "ND", "Not for public")}

\item{excludeDataSource}{Optional. A character vector. A vector of the data sources (dataSource)
that will not be flagged as protected, even if they are. This is useful if you have a private
dataset that should be listed  as "All rights reserved" which you want to be ignored by this flag.}
}
\value{
Returns the data with a new column, .unLicensed, where FALSE = records that are protected by
a license.
}
\description{
This function will search for strings that indicate a record is restricted in its use and will
flag the restricted records.
}
\examples{
  # Read in the example data
data("beesRaw")
  # Run the function
beesRaw_out <- flagLicense(data = beesRaw,
                        strings_to_restrict = "all",
                        # DON'T flag if in the following data# source(s)
                        excludeDataSource = NULL)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/summaryMaps.R
\name{summaryMaps}
\alias{summaryMaps}
\title{Create country-level summary maps of species and occurrence numbers}
\usage{
summaryMaps(
  data = NULL,
  class_n = 15,
  class_Style = "fisher",
  outPath = NULL,
  fileName = NULL,
  width = 5,
  height = 10,
  dpi = 300,
  returnPlot = FALSE,
  scale = 110,
  pointBuffer = 0.01
)
}
\arguments{
\item{data}{A data frame or tibble. Occurrence records as input.}

\item{class_n}{Numeric. The number of categories to break the data into.}

\item{class_Style}{Character. The class style passed to \code{\link[classInt:classIntervals]{classInt::classIntervals()}}. Options are chosen
style: one of "fixed", "sd", "equal", "pretty", "quantile", "kmeans", "hclust", "bclust",
"fisher", "jenks", "dpih", "headtails", or "maximum". Default = "fisher"}

\item{outPath}{A character vector the path to the save location for the output figure.}

\item{fileName}{A character vector with file name
for the output figure, ending with '.pdf'.}

\item{width}{Numeric. The width, in inches, of the resulting figure. Default = 5.}

\item{height}{Numeric. The height, in inches, of the resulting figure. Default = 10.}

\item{dpi}{Numeric. The resolution of the resulting plot. Default = 300.}

\item{returnPlot}{Logical. If TRUE, return the plot to the environment. Default = FALSE.}

\item{scale}{Numeric or character. Passed to rnaturalearth's ne_countries().
Scale of map to return, one of 110, 50, 10 or 'small', 'medium', 'large'. Default = 110.}

\item{pointBuffer}{Numeric. Amount to buffer points, in decimal degrees. If the point is outside
of a country, but within this point buffer, it will count towards that country. It's a good idea
to keep this value consistent with the prior flags applied. Default = 0.01.}
}
\value{
Saves a figure to the user-specified outpath and name with a global map of bee
occurrence species and count data from the input dataset.
}
\description{
Builds an output figure that shows the number of species and the number of occurrences per
country. Breaks the data into classes for visualisation. Users may filter data to their taxa
of interest to produce figures of interest.
}
\examples{
if(requireNamespace("rnaturalearthdata")){
# Read in data
data(beesFlagged)
OutPath_Figures <- tempdir()
# This simple example using the test data has very few classes due to the small amount of input 
# data.
summaryMaps(
data = beesFlagged,
width = 10, height = 10,
class_n = 4,
class_Style = "fisher",
outPath = OutPath_Figures,
fileName = paste0("CountryMaps_fisher_TEST.pdf"),
)
} # END if require

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/idMatchR.R
\name{idMatchR}
\alias{idMatchR}
\title{Attempt to match database_ids from a prior run}
\usage{
idMatchR(
  currentData = NULL,
  priorData = NULL,
  matchBy = NULL,
  completeness_cols = NULL,
  excludeDataset = NULL
)
}
\arguments{
\item{currentData}{A data frame or tibble. The NEW occurrence records as input.}

\item{priorData}{A data frame or tibble. The PRIOR occurrence records as input.}

\item{matchBy}{A list of character vectors Should contain the columns to iteratively compare.}

\item{completeness_cols}{A character vector. The columns to check for completeness, arrange,
and assign the relevant prior database_id.}

\item{excludeDataset}{A character vector. The dataSources that are to be excluded from data
matching. These should be static dataSources from minor providers.}
}
\value{
The input data frame returned with an updated database_id column that shows the
database_ids as in priorData where they could be matched. Additionally, a columnd called
idContinuity is returned where TRUE indicates a match to a prior database_id and FALSE
indicates that a new database_id was assigned.
}
\description{
This function attempts to match database_ids from a prior bdc or BeeBDC run in order to keep
this column somewhat consistent between iterations. However, not all records contain sufficient
information for this to work flawlessly.
}
\examples{
# Get the example data
data("beesRaw", package = "BeeBDC")
# Which datasets are static and should be excluded from matching?
excludeDataset <- c("BMin", "BMont", "CAES", "EaCO", "Ecd", "EcoS",
                    "Gai", "KP", "EPEL", "USGS", "FSCA", "SMC", "Bal", "Lic", "Arm", "BBD", 
                    "MEPB")
  # Match the data to itself just as an example of running the code.
beesRaw_out <- idMatchR(
  currentData = beesRaw,
  priorData = beesRaw,
  # First matches will be given preference over later ones
  matchBy = dplyr::lst(c("gbifID"),
                        c("catalogNumber", "institutionCode", "dataSource"),
                        c("occurrenceID", "dataSource"),
                        c("recordId", "dataSource"),
                        c("id"),
                        c("catalogNumber", "institutionCode")),
  # You can exclude datasets from prior by matching their prefixs - before first underscore:
  excludeDataset = excludeDataset)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/testTaxonomy.R
\docType{data}
\name{testTaxonomy}
\alias{testTaxonomy}
\title{An example of the beesTaxonomy file}
\format{
An object of class \code{"tibble"}
\describe{
\item{taxonomic_status}{Taxonomic status. Values are "accepted" or "synonym"}
\item{source}{Source of the name.}
\item{accid}{The id of the accepted taxon name or "0" if taxonomic_status == accepted.}
\item{id}{The id number for the taxon name.}
\item{kingdom}{The biological kingdom the taxon belongs to. For bees, kingdom == Animalia.}
\item{phylum}{The biological phylum the taxon belongs to. For bees, phylum == Arthropoda.}
\item{class}{The biological class the taxon belongs to. For bees, class == Insecta.}
\item{order}{The biological order the taxon belongs to. For bees, order == Hymenoptera.}
\item{family}{The family of bee which the species belongs to.}
\item{subfamily}{The subfamily of bee which the species belongs to.}
\item{tribe}{The tribe of bee which the species belongs to.}
\item{subtribe}{The subtribe of bee which the species belongs to.}
\item{validName}{The valid scientific name as it should occur in the 'scientificName" column in a Darwin Core file.}
\item{canonical}{The scientificName without the scientificNameAuthority.}
\item{canonical_withFlags}{The scientificName without the scientificNameAuthority and with Discover Life taxonomy flags.}
\item{genus}{The genus the bee species belongs to.}
\item{subgenus}{The subgenus the bee species belongs to.}
\item{species}{The specific epithet for the bee species.}
\item{infraspecies}{The infraspecific epithet for the bee addressed.}
\item{authorship}{The author who described the bee species.}
\item{taxon_rank}{Rank for the bee taxon addressed in the entry.}
\item{notes}{Additional notes about the name/taxon.}
}
}
\usage{
data("testTaxonomy", package = "BeeBDC")
}
\description{
A small test taxonomy file for package tests. This dataset was built by filtering the
taxonomy data from the three test datasets, beesFlagged, beesRaw, bees3sp.
}
\examples{

beesRaw <- BeeBDC::testTaxonomy
head(testTaxonomy)

}
\references{
This dataset is a subset of the beesTaxonomy file described in:
Dorey, J.B., Fischer, E.E., Chesshire, P.R., Nava-Bolaños, A., O’Reilly, R.L., Bossert, S., Collins, S.M., Lichtenberg, E.M., Tucker, E., Smith-Pardo, A., Falcon-Brindis, A., Guevara, D.A., Ribeiro, B.R., de Pedro, D., Hung, J.K.-L., Parys, K.A., McCabe, L.M., Rogan, M.S., Minckley, R.L., Velzco, S.J.E., Griswold, T., Zarrillo, T.A., Jetz, W., Sica, Y.V., Orr, M.C., Guzman, L.M., Ascher, J., Hughes, A.C. & Cobb, N.S. (2023) A globally synthesised and flagged bee occurrence dataset and cleaning workflow. Scientific Data, 10, 1–17. https://www.doi.org/10.1038/S41597-023-02626-W
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/flagRecorder.R
\name{flagRecorder}
\alias{flagRecorder}
\title{Loads, appends, and saves occurrence flag data}
\usage{
flagRecorder(
  data = NULL,
  outPath = NULL,
  fileName = NULL,
  idColumns = c("database_id", "id", "catalogNumber", "occurrenceID", "dataSource"),
  append = NULL,
  printSummary = FALSE
)
}
\arguments{
\item{data}{A data frame or tibble. Occurrence records as input.}

\item{outPath}{A character path. Where the file should be saved.}

\item{fileName}{Character. The name of the file to be saved}

\item{idColumns}{A character vector. The names of the columns that are to be kept along with the
flag columns. These columns should be useful for identifying unique records with flags.
Default = c("database_id", "id", "catalogNumber", "occurrenceID", "dataSource").}

\item{append}{Logical. If TRUE, this will find and append an existing file generated by this function.}

\item{printSummary}{Logical. If TRUE, print a \code{\link[=summary]{summary()}} of all filter columns - i.e. those which
tidyselect::starts_with(".")}
}
\value{
Saves a file with id and flag columns and returns this as an object.
}
\description{
This function is used to save the flag data for your occurrence data as you run the BeeBDC script.
It will read and append existing files, if asked to. Your flags should also be saved in the occurrence
file itself automatically.
}
\examples{
# Load the example data
data("beesFlagged")

  # Run the function
  OutPath_Report <- tempdir()
flagFile <- flagRecorder(
  data = beesFlagged,
  outPath = paste(OutPath_Report, sep =""),
  fileName = paste0("flagsRecorded_", Sys.Date(), ".csv"),
  # These are the columns that will be kept along with the flags
  idColumns = c("database_id", "id", "catalogNumber", "occurrenceID", "dataSource"),
  # TRUE if you want to find a file from a previous part of the script to append to
  append = FALSE)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/jbd_CfC_chunker.R
\name{jbd_CfC_chunker}
\alias{jbd_CfC_chunker}
\title{Get country names from coordinates}
\usage{
jbd_CfC_chunker(
  data = NULL,
  lat = "decimalLatitude",
  lon = "decimalLongitude",
  country = "country",
  stepSize = 1e+06,
  chunkStart = 1,
  scale = "medium",
  path = tempdir(),
  mc.cores = 1
)
}
\arguments{
\item{data}{A data frame or tibble. Occurrence records to use as input.}

\item{lat}{Character. The name of the column to use as latitude. Default = "decimalLatitude".}

\item{lon}{Character. The name of the column to use as longitude. Default = "decimalLongitude".}

\item{country}{Character. The name of the column containing country names. Default = "country.}

\item{stepSize}{Numeric. The number of occurrences to process in each chunk. Default = 1000000.}

\item{chunkStart}{Numeric. The chunk number to start from. This can be > 1 when you need to
restart the function from a certain chunk. For example, can be used if R failed unexpectedly.}

\item{scale}{Passed to rnaturalearth's ne_countries().
Scale of map to return, one of 110, 50, 10 or 'small', 'medium', 'large'. Default = "large".}

\item{path}{Character. The directory path to a folder in which to save the running countrylist
csv file.}

\item{mc.cores}{Numeric. If > 1, the function will run in parallel
using mclapply using the number of cores specified. If = 1 then it will be run using a serial
loop. NOTE: Windows machines must use a value of 1 (see ?parallel::mclapply). Additionally,
be aware that each thread can use large chunks of memory.
Default = 1.}
}
\value{
A data frame containing database_ids and a country column
that needs to be re-merged with the data input.
}
\description{
Because the \code{\link[bdc:bdc_country_from_coordinates]{bdc::bdc_country_from_coordinates()}} function is very RAM-intensive, this wrapper
allows a user to specify chunk-sizes and only analyse a small portion of the occurrence data at a
time. The prefix jbd_ is used to highlight the difference between this function and the original
\code{\link[bdc:bdc_country_from_coordinates]{bdc::bdc_country_from_coordinates()}}.
}
\examples{
if(requireNamespace("rnaturalearthdata")){
library("dplyr")
data(beesFlagged)
HomePath = tempdir()
# Tibble of common issues in country names and their replacements
commonProblems <- dplyr::tibble(problem = c('U.S.A.', 'US','USA','usa','UNITED STATES',
'United States','U.S.A','MX','CA','Bras.','Braz.','Brasil','CNMI','USA TERRITORY: PUERTO RICO'),
                                 fix = c('United States of America','United States of America',
                                 'United States of America','United States of America',
                                 'United States of America','United States of America',
                                 'United States of America','Mexico','Canada','Brazil','Brazil',
                                 'Brazil','Northern Mariana Islands','Puerto Rico'))
                                 
beesFlagged <- beesFlagged \%>\%
      # Replace a name to test
   dplyr::mutate(country = stringr::str_replace_all(country, "Brazil", "Brasil"))

beesFlagged_out <- countryNameCleanR(
  data = beesFlagged,
  commonProblems = commonProblems)

suppressWarnings(
  countryOutput <- jbd_CfC_chunker(data = beesFlagged_out,
                                   lat = "decimalLatitude",
                                   lon = "decimalLongitude",
                                   country = "country",
                                   # How many rows to process at a time
                                   stepSize = 1000000,
                                   # Start row
                                   chunkStart = 1,
                                   path = HomePath,
                                   scale = "medium"),
  classes = "warning")


# Left join these datasets
beesFlagged_out <- left_join(beesFlagged_out, countryOutput, by = "database_id")  \%>\% 
  # merge the two country name columns into the "country" column
  dplyr::mutate(country = dplyr::coalesce(country.x, country.y)) \%>\%
  # remove the now redundant country columns 
  dplyr::select(!c(country.x, country.y)) \%>\%
  # put the column back 
  dplyr::relocate(country) \%>\% 
  # Remove duplicates if they arose!
  dplyr::distinct()

# Remove illegal characters
beesFlagged_out$country <- beesFlagged_out$country \%>\%
  stringr::str_replace(., pattern = paste("\\\\[", "\\\\]", "\\\\?",
                                          sep=  "|"), replacement = "")
} # END if require
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/repoMerge.R
\name{repoMerge}
\alias{repoMerge}
\title{Import occurrences from GBIF, ALA, iDigBio, and SCAN downloads}
\usage{
repoMerge(path, save_type, occ_paths)
}
\arguments{
\item{path}{A directory as a character. The directory to recursively look in for the above data.}

\item{save_type}{Character. The data type to save the resulting file as. Options are:
csv_files" or "R_file".}

\item{occ_paths}{A list of directories. Preferably produced using \code{\link[=repoFinder]{repoFinder()}} the
function asks for a list of paths to the relevant input datasets. You can fault-find errors
in this function by checking the output of \code{\link[=repoFinder]{repoFinder()}}.}
}
\value{
A list with a data frame of merged occurrence records, "Data_WebDL", and a list of eml
files contained in "eml_files". Also saves these files in the requested format.
}
\description{
Locates data from GBIF, ALA, iDigBio, and SCAN within a directory and reads it in along with its eml
metadata. Please keep the original download folder names and architecture unchanged.
NOTE: This function uses family-level data to identify taxon downloads. If this, or something new,
becomes an issue, please contact James Dorey (the developer) as there are likely to be exceptions
to how files are downloaded. current as of versions 1.0.4.
}
\examples{
\dontrun{
if(require("emld")){

DataImp <- repoMerge(path = DataPath, 
# Find data - Many problems can be solved by running [BeeBDC::repoFinder(path = DataPath)]
# And looking for problems
occ_paths = BeeBDC::repoFinder(path = DataPath),
save_type = "R_file")
} # END if require
} # END donttest
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/coordUncerFlagR.R
\name{coordUncerFlagR}
\alias{coordUncerFlagR}
\title{Flag occurrences with an uncertainty threshold}
\usage{
coordUncerFlagR(
  data = NULL,
  uncerColumn = "coordinateUncertaintyInMeters",
  threshold = NULL
)
}
\arguments{
\item{data}{A data frame or tibble. Occurrence records as input.}

\item{uncerColumn}{Character. The column to flag uncertainty in.}

\item{threshold}{Numeric. The uncertainty threshold. Values equal to, or greater than, this
threshold will be flagged.}
}
\value{
The input data with a new column, .uncertaintyThreshold.
}
\description{
To use this function, the user must choose a column, probably "coordinateUncertaintyInMeters"
and a threshold above which occurrences will be flagged for geographic uncertainty.
}
\examples{
# Run the function
beesRaw_out <- coordUncerFlagR(data = beesRaw,
                               uncerColumn = "coordinateUncertaintyInMeters",
                               threshold = 1000)
# View the output
table(beesRaw_out$.uncertaintyThreshold, useNA = "always")
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/manualOutlierFindeR.R
\name{manualOutlierFindeR}
\alias{manualOutlierFindeR}
\title{Finds outliers, and their duplicates, as determined by experts}
\usage{
manualOutlierFindeR(
  data = NULL,
  DataPath = NULL,
  PaigeOutliersName = "removedBecauseDeterminedOutlier.csv",
  newOutliersName = "All_outliers_ANB.xlsx",
  ColombiaOutliers_all = "All_Colombian_OutlierIDs.csv",
  duplicates = NULL,
  NearTRUE = NULL,
  NearTRUE_threshold = 5
)
}
\arguments{
\item{data}{A data frame or tibble. Occurrence records as input.}

\item{DataPath}{A character path to the directory that contains the outlier spreadsheets.}

\item{PaigeOutliersName}{A character patch. Should lead to outlier spreadsheet from Paige Chesshire (csv file).}

\item{newOutliersName}{A character path. Should lead to appropriate outlier spreadsheet (xlsx file).}

\item{ColombiaOutliers_all}{A character path. Should lead to spreadsheet of bee outliers from Colombia (csv file).}

\item{duplicates}{A data frame or tibble. The duplicate file produced by \code{\link[=dupeSummary]{dupeSummary()}}.}

\item{NearTRUE}{Optional. A character file name to the csv file. If you want to remove expert
outliers that are too close to TRUE points, use the name of the NearTRUE.csv.
Note: This implementation is only basic for now unless there is a greater need in the future.}

\item{NearTRUE_threshold}{Numeric. The threshold (in km) for the distance to TRUE points to
keep expert outliers.}
}
\value{
Returns the data with a new column, \code{.expertOutlier} where records that are FALSE are
the expert outliers.
}
\description{
Uses expert-identified outliers with source spreadsheets that may be edited by users. The function
will also use the duplicates file made using \code{\link[=dupeSummary]{dupeSummary()}} to identify duplicates of the
expert-identified outliers and flag those as well.
The function will add a flagging column called \code{.expertOutlier} where records that are FALSE are
the expert outliers.
}
\examples{
\dontrun{
  # Read example data
  data(beesFlagged)
# Read in the most-recent duplicates file as well
if(!exists("duplicates")){
  duplicates <- fileFinder(path = DataPath,
                            fileName = "duplicateRun_") \%>\%
    readr::read_csv()}
# identify the outliers and get a list of their database_ids
beesFlagged_out <- manualOutlierFindeR(
  data = beesFlagged,
  DataPath = DataPath,
  PaigeOutliersName = "removedBecauseDeterminedOutlier.csv",
  newOutliersName = "^All_outliers_ANB_14March.xlsx",
  ColombiaOutliers_all = "All_Colombian_OutlierIDs.csv",
  duplicates = duplicates)
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/jbd_coordCountryInconsistent.R
\name{jbd_coordCountryInconsistent}
\alias{jbd_coordCountryInconsistent}
\title{Flags coordinates that are inconsistent with the stated country name}
\usage{
jbd_coordCountryInconsistent(
  data = NULL,
  lon = "decimalLongitude",
  lat = "decimalLatitude",
  scale = 50,
  pointBuffer = 0.01,
  stepSize = 1e+06,
  mc.cores = 1
)
}
\arguments{
\item{data}{A data frame or tibble. Occurrence records as input.}

\item{lon}{Character. The name of the column to use as longitude. Default = "decimalLongitude".}

\item{lat}{Character. The name of the column to use as latitude. Default = "decimalLatitude".}

\item{scale}{Numeric or character. To be passed to \code{\link[rnaturalearth:ne_countries]{rnaturalearth::ne_countries()}}'s scale.
Scale of map to return, one of 110, 50, 10 or "small", "medium", "large".
Smaller values return higher-resolution maps.}

\item{pointBuffer}{Numeric. Amount to buffer points, in decimal degrees. If the point is outside
of a country, but within this point buffer, it will not be flagged. Default = 0.01.}

\item{stepSize}{Numeric. The number of occurrences to process in each chunk. Default = 1000000.}

\item{mc.cores}{Numeric. If > 1, the st_intersects function will run in parallel
using mclapply using the number of cores specified. If = 1 then it will be run using a serial
loop. NOTE: Windows machines must use a value of 1 (see ?parallel::mclapply). Additionally,
be aware that each thread can use large chunks of memory.
Default = 1.}
}
\value{
The input occurrence data with a new column, .coordinates_country_inconsistent
}
\description{
Compares stated country name in an occurrence record with record's coordinates using
rnaturalearth data. The prefix, jbd_ is meant
to distinguish this function from the original \code{\link[bdc:bdc_coordinates_country_inconsistent]{bdc::bdc_coordinates_country_inconsistent()}}.
This functions will preferably use the countryCode and country_suggested columns
generated by \code{bdc::bdc_country_standardized()}; please run it on your dataset prior to running
this function.
}
\examples{
if(requireNamespace("rnaturalearthdata")){
beesRaw_out <- jbd_coordCountryInconsistent(
  data = BeeBDC::beesRaw,
  lon = "decimalLongitude",
  lat = "decimalLatitude",
  scale = 50,
  pointBuffer = 0.01)
} # END if require
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/flagSummaryTable.R
\name{flagSummaryTable}
\alias{flagSummaryTable}
\title{Build a per-species summary for each and all flags}
\usage{
flagSummaryTable(
  data = NULL,
  column = "scientificName",
  outPath = OutPath_Report,
  fileName = "flagTable.csv",
  percentImpacted = TRUE,
  percentThreshold = 0
)
}
\arguments{
\item{data}{A data frame or tibble. The flagged dataset.}

\item{column}{Character. The name of the column to group by and summarise the failed occurrences.
Default = "scientificName".}

\item{outPath}{A character path. The path to the directory in which the figure will be saved.
Default = OutPath_Report. If is NULL then no file will be saved to the disk.}

\item{fileName}{Character. The name of the file to be saved, ending in ".csv".
Default = "flagTable.csv".}

\item{percentImpacted}{Logical. If TRUE (the default), the program will write the percentage of
species impacted and over the percentThreshold for each flagging column.}

\item{percentThreshold}{Numeric. A number between 0 and 100 to indicate the percent of
individuals (>; within each species) that is impacted by a flag, and to be included in the
percentImpacted. Default = 0.}
}
\value{
A tibble with a column for each flag column (starting with ".") showing the number of
failed (FALSE) occurrences per group. Also shows the (i) total number of records, (ii) total
number of failed records, and (iii) the percentage of failed records.
}
\description{
Takes a flagged dataset and returns the total number of fails (FALSE) per flag (columns starting
with ".") and per species. It will ignore the .scientificName_empty and .invalidName columns as
species are not assigned.
Users may define the column to group the summary by. While it is intended to work with
the scientificName column, users may select any grouping column (e.g., country).
}
\examples{
# Load the toy flagged bee data
data("beesFlagged")

  # Run the function and build the flag table
flagTibble <- flagSummaryTable(data = beesFlagged,
                              column = "scientificName",
                              outPath = paste0(tempdir()),
                              fileName = "flagTable.csv")
                              

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/jbd_create_figures.R
\name{jbd_create_figures}
\alias{jbd_create_figures}
\title{Create figures reporting the results of the bdc/BeeBDC packages}
\usage{
jbd_create_figures(
  data,
  path = OutPath_Figures,
  database_id = "database_id",
  workflow_step = NULL,
  save_figures = FALSE
)
}
\arguments{
\item{data}{A data frame or tibble. Needs to contain the results of data quality tests; that
is, columns starting with ".".}

\item{path}{A character directory. The path to a directory in which to save the figures.
Default = OutPath_Figures.}

\item{database_id}{A character string. The column name with a unique record
identifier. Default = "database_id".}

\item{workflow_step}{A character string. Name of the workflow step. Options
available are "prefilter", "space", and "time".}

\item{save_figures}{Logical. Indicates if the figures should be saved for further inspection or
use. Default = FALSE.}
}
\value{
List containing figures showing the results of data quality tests
implemented in one module of bdc/BeeBDC. When save_figures = TRUE, figures are
also saved locally in a .png format.
}
\description{
Creates figures (i.e., bar plots, maps, and histograms) reporting the results
of data quality tests implemented the bdc and BeeBDC packages. Works like \code{\link[bdc:bdc_create_figures]{bdc::bdc_create_figures()}},
but it allows the user to specify a save path.
}
\details{
This function creates figures based on the results of data quality
tests. A pre-defined list of test names is used for creating
figures depending on the name of the workflow step informed. Figures are
saved in "Output/Figures" if save_figures = TRUE.
}
\examples{
\donttest{

if(require("bdc")){

database_id <- c("GBIF_01", "GBIF_02", "GBIF_03", "FISH_04", "FISH_05")
lat <- c(-19.93580, -13.01667, -22.34161, -6.75000, -15.15806)
lon <- c(-40.60030, -39.60000, -49.61017, -35.63330, -39.52861)
.scientificName_emptys <- c(TRUE, TRUE, TRUE, FALSE, FALSE)
.coordinates_empty <- c(TRUE, TRUE, TRUE, TRUE, TRUE)
.invalid_basis_of_records <- c(TRUE, FALSE, TRUE, FALSE, TRUE)
.summary <- c(TRUE, FALSE, TRUE, FALSE, FALSE)

x <- data.frame(
  database_id,
  lat,
  lon,
  .scientificName_emptys,
  .coordinates_empty,
  .invalid_basis_of_records,
  .summary
)

figures <- 
jbd_create_figures(
  data = x, 
  database_id = "database_id",
  workflow_step = "prefilter",
  save_figures = FALSE
)

} # End if require
} # End dont test
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ggRichnessWrapper.R
\name{ggRichnessWrapper}
\alias{ggRichnessWrapper}
\title{ggplot2 extension for a Chao- and iNEXT-wrapper outputs}
\usage{
ggRichnessWrapper(
  iNEXT_in = country_iNEXT,
  iChao_in = NULL,
  filterOut = NULL,
  type = 1,
  se = TRUE,
  facet.var = "None",
  color.var = "Order.q",
  grey = FALSE,
  legendPerPlot = FALSE,
  show_iNEXT = TRUE,
  showPercent = TRUE,
  ChaoColour = "#55AD9B",
  iNEXTcolour = "#FD9B63",
  Chao_estimate = "iChao1 (Chiu et al. 2014)",
  nrow = 3,
  ncol = 4,
  labels = NULL,
  fileName = "richnessPlots.pdf",
  outPath = tempdir(),
  base_width = 8.3,
  base_height = 11.7,
  dpi = 300,
  ...
)
}
\arguments{
\item{iNEXT_in}{a list of \code{iNEXT} objects computed by \code{BeeBDC::iNEXTwrapper()}.}

\item{iChao_in}{a list of R data created \code{BeeBDC::ChaoWrapper()}.}

\item{filterOut}{Character. A list of sites/countries to exclude from plotting; for example,
because sample size was inadequate. Default = NULL.}

\item{type}{three types of plots: sample-size-based rarefaction/extrapolation curve (\code{type = 1});
sample completeness curve (\code{type = 2}); coverage-based rarefaction/extrapolation curve (\code{type = 3}).
From \code{iNEXT::ggiNEXT()}}

\item{se}{a logical variable to display confidence interval around the estimated sampling curve.
From \code{iNEXT::ggiNEXT()}}

\item{facet.var}{create a separate plot for each value of a specified variable:
no separation \cr (\code{facet.var="None"});
a separate plot for each diversity order (\code{facet.var="Order.q"});
a separate plot for each assemblage (\code{facet.var="Assemblage"});
a separate plot for each combination of order x assemblage (\code{facet.var="Both"}).
From \code{iNEXT::ggiNEXT()}}

\item{color.var}{create curves in different colors for values of a specified variable:
all curves are in the same color (\code{color.var="None"});
use different colors for diversity orders (\code{color.var="Order.q"});
use different colors for sites (\code{color.var="Assemblage"});
use different colors for combinations of order x assemblage (\code{color.var="Both"}).
From \code{iNEXT::ggiNEXT()}}

\item{grey}{a logical variable to display grey and white ggplot2 theme.
From \code{iNEXT::ggiNEXT()}}

\item{legendPerPlot}{Logical. If TRUE, remove the legend from each plot. Default = FALSE}

\item{show_iNEXT}{Logical. If TRUE, show the estimate and 95\% CIs if specified for iNEXT. Default = TRUE.}

\item{showPercent}{Logical. If TRUE, show the prrcentage increases. Default = TRUE.}

\item{ChaoColour}{Character. The to be used to graph Chao estimates (95\% confidence intervals
are shown with reduced opacity). Default = "#55AD9B".}

\item{iNEXTcolour}{Character. The to be used to graph iNEXT estimates (95\% confidence intervals
are shown with reduced opacity). Default = "#FD9B63".}

\item{Chao_estimate}{Character. The name of the Chao estimate to use from those calculated in
\code{SpadeR::ChaoSpecies()}. The options are "Homogeneous Model","Homogeneous (MLE)",
"Chao1 (Chao, 1984)","Chao1-bc","iChao1 (Chiu et al. 2014)","ACE (Chao & Lee, 1992)",
"ACE-1 (Chao & Lee, 1992)","1st order jackknife","2nd order jackknife".
Default = "iChao1 (Chiu et al. 2014)".}

\item{nrow}{Numeric. The number of rows per figure. Figures (that don't fit in the
nrow*ncol grid) will be saved into
additional files. Default = 3.}

\item{ncol}{Numeric. The number of columns per figure. Figures (that don't fit in the
nrow*ncol grid) will be saved into
additional files. Default = 4.}

\item{labels}{Character. The labels for each sub-plot (a, b, c, ...). The default is NULL,
which will provide labels a-z as required.}

\item{fileName}{Character. Prefix to the output files. Default = "richnessPlots.pdf".
Changing the suffix is a convenient way to alter the file format.}

\item{outPath}{Character. The fodler in which to save the plots. Default = \code{tempdir()}}

\item{base_width}{Numeric. The width, in inches, to save the plot. Default = 8.3.}

\item{base_height}{Numeric. The height, in inches, to save the plot. Default = 11.7.}

\item{dpi}{Numeric. Plot resolution. Also accepts a string input: "retina" (320),
"print" (300), or "screen" (72). Applies only to raster output types. Default = 300.}

\item{...}{other arguments passed on to methods. Not currently used.
From \code{iNEXT::ggiNEXT()}}
}
\value{
Saves pdf objects and returns a summary table for all levels
}
\description{
\code{BeeBDC::ggRichnessWrapper()} takes data output from \code{BeeBDC::iNEXTwrapper()} and
\code{BeeBDC::ChaoWrapper()} to produce plots for multiple groups in one go. The plots can be
multiple per page and across multiple pages.
}
\examples{
\dontrun{

data(beesCountrySubset)

   # Transform data for iNEXT
 data_nextWrapper <- beesCountrySubset \%>\%
   dplyr::group_by(scientificName, country_suggested) \%>\%
   dplyr::count() 
   
 # Calculate iNEXT with the wrapper function
 output_iNEXTwrapper <- BeeBDC::iNEXTwrapper(data = data_nextWrapper,
                                             variableColumn = "country_suggested",
                                             valueColumn = "n",
                                             mc.cores = 1)

 # Transform data for iChao
data_iChao <- beesCountrySubset \%>\%
  dplyr::group_by(scientificName, country_suggested) \%>\%
  dplyr::count() \%>\%
  dplyr::select(scientificName, country_suggested, n) \%>\%
  tidyr::pivot_wider(names_from = country_suggested,
                     values_from = n,
                     values_fill = 0) \%>\%
  ## Create the rownames
  tibble::column_to_rownames("scientificName") \%>\%
  dplyr::tibble()
  
  # Run the wrapper function
output_iChaowrapper <- BeeBDC::ChaoWrapper(data = data_iChao,
                                             datatype = "abundance",
                                             k = 10,
                                             conf = 0.95,
                                             mc.cores = 1)
                                             
   # Make the plots! 
plot_summary <- BeeBDC::ggRichnessWrapper(
iNEXT_in = output_iNEXTwrapper,
iChao_in = output_iChaowrapper,
nrow = 2,
ncol = 2,
labels = NULL,
fileName = "speciesRichnessPlots.pdf",
outPath = tempdir(),
base_width = 8.3,
base_height = 11.7, 
dpi = 300)
      
}                                       
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/atlasDownloader.R
\name{atlasDownloader}
\alias{atlasDownloader}
\title{Download occurrence data from the Atlas of Living Australia (ALA)}
\usage{
atlasDownloader(
  path,
  userEmail = NULL,
  ALA_taxon,
  DL_reason = 4,
  atlas = "ALA"
)
}
\arguments{
\item{path}{A character directory. The path to a folder where the download will be stored.}

\item{userEmail}{A character string. The email used associated with the user's ALA account;
user must make an ALA account to download data.}

\item{ALA_taxon}{A character string. The taxon to download from ALA. Uses \code{\link[galah:galah_identify]{galah::galah_identify()}}}

\item{DL_reason}{Numeric. The reason for data download according to \code{\link[galah:galah_config]{galah::galah_config()}}}

\item{atlas}{Character. The atlas to download occurrence data from - see here https://galah.ala.org.au/R/articles/choosing_an_atlas.html for details.
Note: the default is "ALA" and is probably the only atlas which will work seamlessly with the rest
of the workflow. However, different atlases can still be downloaded and a doi will be sent to
your email.}
}
\value{
Completes an ALA data download and saves those data to the path provided.
}
\description{
Downloads ALA data and creates a new file in the path to put those data. This function can also
request downloads from other atlases (see: http://galah.ala.org.au/articles/choosing_an_atlas.html).
However, it will only send the download to your email and you must do the rest yourself at this point.
}
\examples{
\dontrun{
atlasDownloader(path = DataPath,
               userEmail = "InsertYourEmail",
               ALA_taxon = "Apiformes",
               DL_reason = 4)
               }
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/richnessPrepR.R
\name{richnessPrepR}
\alias{richnessPrepR}
\title{Prepare occurrence, taxonomy, and checklist data for richness estimation}
\usage{
richnessPrepR(
  data = NULL,
  taxonomyFile = BeeBDC::beesTaxonomy(),
  checklistFile = BeeBDC::beesChecklist(),
  curveFunction = function(x) (228.7531 * x * x^-log(12.1593)),
  sampleSize = 10000,
  countryColumn = "country_suggested",
  limitGlobal = NULL,
  outPath = tempdir()
)
}
\arguments{
\item{data}{A data frame or tibble. Occurrence records as input. Needs to include the
\code{scientificName} column and a "country" column (see \code{countryColumn})}

\item{taxonomyFile}{A data frame or tibble. The taxonomy file to use.
Default = \code{\link[=beesTaxonomy]{beesTaxonomy()}} but see \code{\link[=taxadbToBeeBDC]{taxadbToBeeBDC()}} for other taxa.}

\item{checklistFile}{A data frame or tibble. The taxonomy to use.
Default = \code{\link[=beesChecklist]{beesChecklist()}}; use this as a template to convert other taxa checklists.}

\item{curveFunction}{A function. The mathematical function that describes the curve used to
randomly sample from in order to fill empty sample sizes for species present in the checklist
but not present in the occurrence dataset. Default is \code{function(x) (228.7531 * x * x^-log(12.1593))},
which is taken from the paper "How many bee species are there? A quantitative global estimate" by
Dorey et al. Models can be fit using \code{\link[mosaic:fitModel]{mosaic::fitModel()}}.}

\item{sampleSize}{Numeric. The size of the sample randomly drawn from the provided curve.
See \code{curveFunction}. Default = 10000.}

\item{countryColumn}{Character. The column from which country names should be sought.}

\item{limitGlobal}{Character vector. A character vector of the countries to filter the data to
in order limit the extent of the global-level analysis. Defualt = NULL.}

\item{outPath}{Character. The output path where the curve plot (curvePlot.pdf) and the output
Rdata (richnessInputs.Rda) file will be saved. Default = \code{tempdir()}.}
}
\value{
Saves an Rdata file with the data needed to feed into the \code{\link[=richnessEstimateR]{richnessEstimateR()}}
function. The \code{\link[=richnessEstimateR]{richnessEstimateR()}} function will then
use iChao and/or iNEXT to estimate species richness for countries, continents, and the globe.
Also returns the RData file to the environment.
}
\description{
Takes your occurrence dataset along with a taxonomy and checklist in order to produce
a file that's ready to be passed into the \code{\link[=richnessEstimateR]{richnessEstimateR()}} function in order to
estimate species richness using
iChao and iNEXT (hill numbers) for countries, continents, or the entire globe.
}
\examples{

\dontrun{
data(beesCountrySubset)
if(require("htmlwidgets"){

estimateDataExample <- richnessPrepR(
  data = beesCountrySubset,
  # Download the taxonomy
  taxonomyFile = BeeBDC::beesTaxonomy(),
  # Download the checklist
  checklistFile = BeeBDC::beesChecklist(),
  curveFunction = function(x) (228.7531 * x * x^-log(12.1593)),
  sampleSize = 10000,
  countryColumn = "country_suggested",
  limitGlobal = NULL,
  outPath = tempdir()
)
} # End if require
} # End dontrun
}
\seealso{
\code{\link[=countryHarmoniseR]{countryHarmoniseR()}} to harmonise country names based on a short list.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/aux_functions.R
\name{dataSaver}
\alias{dataSaver}
\title{Simple function to save occurrence AND EML data as a list}
\usage{
dataSaver(
  path = NULL,
  save_type = NULL,
  occurrences = NULL,
  eml_files = NULL,
  file_prefix = NULL
)
}
\arguments{
\item{path}{Character. The main file path to look for data in.}

\item{save_type}{Character. The file format in which to save occurrence and EML data.
Either "R_file" or "CSV_file"}

\item{occurrences}{The occurrences to save as a data frame or tibble.}

\item{eml_files}{A list of the EML files.}

\item{file_prefix}{Character. A prefix for the resulting output file.}
}
\value{
This function saves both occurrence and EML data as a list when save_type = "R_File" or
as individual csv files when save_type = "CSV_file".
}
\description{
Used at the end of 1.x in the example workflow in order to save the occurrence dataset and its associated eml metadata.
}
\examples{
\dontrun{
dataSaver(path = tempdir(),# The main path to look for data in
save_type = "CSV_file", # "R_file" OR "CSV_file"
occurrences = Complete_data$Data_WebDL, # The existing datasheet
eml_files = Complete_data$eml_files, # The existing EML files
file_prefix = "Fin_") # The prefix for the file name
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/taxadbToBeeBDC.R
\name{taxadbToBeeBDC}
\alias{taxadbToBeeBDC}
\title{Import and convert taxadb taxonomies to BeeBDC format}
\usage{
taxadbToBeeBDC(
  name = NULL,
  rank = NULL,
  provider = "gbif",
  version = "22.12",
  collect = TRUE,
  ignore_case = TRUE,
  db = NULL,
  removeEmptyNames = TRUE,
  outPath = getwd(),
  fileName = NULL,
  ...
)
}
\arguments{
\item{name}{Character. Taxonomic scientific name (e.g. "Aves").
As defined by  \code{taxadb::filter_rank()}.}

\item{rank}{Character. Taxonomic rank name. (e.g. "class").
As defined by \code{taxadb::filter_rank()}.}

\item{provider}{Character. From which provider should the hierarchy be returned?
Default is 'gbif', which can also be configured using options(default_taxadb_provide = ...").
See \code{taxadb::td_create()} for a list of recognized providers. NOTE: gbif seems to have the most-complete
columns, especially in terms of scientificNameAuthorship, which is important for matching
ambiguous names.
As defined by \code{taxadb::filter_rank()}.}

\item{version}{Character. Which version of the taxadb provider database should we use? defaults
to latest. See tl_import for details. Default = 22.12.
As defined by \code{taxadb::filter_rank()}.}

\item{collect}{Logical. Should we return an in-memory data.frame
(default, usually the most convenient), or a reference to lazy-eval table on disk
(useful for very large tables on which we may first perform subsequent filtering operations.).
Default = TRUE.
As defined by \code{taxadb::filter_rank()}.}

\item{ignore_case}{Logical. should we ignore case (capitalization) in matching names?
Can be significantly slower to run. Default = TRUE.
As defined by \code{taxadb::filter_rank()}.}

\item{db}{a connection to the taxadb database. See details of \code{taxadb::filter_rank()}. Default
= Null which should work.
As defined by \code{taxadb::filter_rank()}.}

\item{removeEmptyNames}{Logical. If True (default), it will remove entries without an entry
for specificEpithet.}

\item{outPath}{Character. The path to a directory (folder) in which the output should be saved.}

\item{fileName}{Character. The name of the output file, ending in '.csv'.}

\item{...}{Arguments passed to \code{taxadb::td_create()}.}
}
\value{
Returns a taxonomy file (to the R environment and to the disk, if a fileName is
provided) as a tibble that can be used with \code{BeeBDC::harmoniseR()}.
}
\description{
Uses the taxadb R package to download a requested taxonomy and then transforms it into the input
BeeBDC format. This means that any taxonomy in their databases can be used with BeeBDC. You can
also save the output to your computer and to the R environment for immediate use. See
details below for a list of providers or see \code{taxadb::td_create()}.
}
\examples{
\dontrun{

if(require("taxadb")){

  # Run the function using the bee genus Apis as an example...
ApisTaxonomy <- BeeBDC::taxadbToBeeBDC(
  name = "Apis",
  rank = "Genus",
  provider = "gbif",
  version = "22.12",
  removeEmptyNames = TRUE,
  outPath = getwd(),
  fileName = NULL,
  ...
  )
  
  } # END if require
  } # End dontrun

}
\seealso{
\code{\link[=beesTaxonomy]{beesTaxonomy()}} for the bee taxonomy and \code{\link[=harmoniseR]{harmoniseR()}} for the
taxon-cleaning function where these taxonomies are implemented.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/testChecklist.R
\docType{data}
\name{testChecklist}
\alias{testChecklist}
\title{An example of the beesChecklist file}
\format{
An object of class \code{"tibble"}
\describe{
\item{validName}{The valid scientificName as it should occur in the scientificName column.}
\item{DiscoverLife_name}{The full country name as it occurs on Discover Life.}
\item{rNaturalEarth_name}{Country name from rnaturalearth's name_long and type = "map_units".}
\item{shortName}{A short version of the country name.}
\item{continent}{The continent where that country is found.}
\item{DiscoverLife_ISO}{The ISO country name as it occurs on Discover Life.}
\item{Alpha-2}{Alpha-2 from rnaturalearth.}
\item{iso_a3_eh}{iso_a3_eh from rnaturalearth.}
\item{official}{Official country name = "yes" or only a Discover Life name = "no".}
\item{Source}{A text strign denoting the source or author of the name-country pair.}
\item{matchCertainty}{Quality of the name's match to the Discover Life checklist.}
\item{canonical}{The valid species name without scientificNameAuthority.}
\item{canonical_withFlags}{The validName without the scientificNameAuthority but with Discover Life flags.}
\item{family}{Bee family.}
\item{subfamily}{Bee subfamily.}
\item{genus}{Bee genus.}
\item{subgenus}{Bee subgenus.}
\item{infraspecies}{Bee infraSpecificEpithet.}
\item{species}{Bee specificEpithet.}
\item{scientificNameAuthorship}{Bee scientificNameAuthorship.}
\item{taxon_rank}{Rank of the taxon name.}
\item{Notes}{Discover Life country name notes.}
}
}
\usage{
data("testChecklist", package = "BeeBDC")
}
\description{
A small test checklist file for package tests. This dataset was built by filtering the
checklist data from the three test datasets, beesFlagged, beesRaw, bees3sp.
}
\examples{
beesRaw <- BeeBDC::testChecklist
head(testChecklist)

}
\references{
This dataset is a subset of the beesChecklist file described in:
Dorey, J.B., Fischer, E.E., Chesshire, P.R., Nava-Bolaños, A., O’Reilly, R.L., Bossert, S., Collins, S.M., Lichtenberg, E.M., Tucker, E., Smith-Pardo, A., Falcon-Brindis, A., Guevara, D.A., Ribeiro, B.R., de Pedro, D., Hung, J.K.-L., Parys, K.A., McCabe, L.M., Rogan, M.S., Minckley, R.L., Velzco, S.J.E., Griswold, T., Zarrillo, T.A., Jetz, W., Sica, Y.V., Orr, M.C., Guzman, L.M., Ascher, J., Hughes, A.C. & Cobb, N.S. (2023) A globally synthesised and flagged bee occurrence dataset and cleaning workflow. Scientific Data, 10, 1–17. https://www.doi.org/10.1038/S41597-023-02626-W
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/PaigeIntegrateR.R
\name{PaigeIntegrater}
\alias{PaigeIntegrater}
\title{Integrate manually-cleaned data from Paige Chesshire}
\usage{
PaigeIntegrater(db_standardized = NULL, PaigeNAm = NULL, columnStrings = NULL)
}
\arguments{
\item{db_standardized}{A data frame or tibble. Occurrence records as input.}

\item{PaigeNAm}{A data frame or tibble. The Paige Chesshire dataset.}

\item{columnStrings}{A list of character vectors. Each vector is a set of columns that will be
used to iteratively match the public dataset against the Paige dataset.}
}
\value{
Returns db_standardized (input occurrence records) with the Paige Chesshire data integrated.
}
\description{
Replaces publicly available data with data that has been manually cleaned and error-corrected for use in
the paper Chesshire, P. R., Fischer, E. E., Dowdy, N. J., Griswold, T., Hughes, A. C., Orr, M. J., . . . McCabe, L. M. (In Press). Completeness analysis for over 3000 United States bee species identifies persistent data gaps. Ecography.
}
\examples{
\dontrun{
library(dplyr)
# set the DataPath to tempdir for this example
DataPath <- tempdir()
# Integrate Paige Chesshire's cleaned dataset.
PaigeNAm <- readr::read_csv(paste(DataPath, "Paige_data", "NorAmer_highQual_only_ALLfamilies.csv",
                                 sep = "/"), col_types = ColTypeR()) \%>\%
 # Change the column name from Source to dataSource to match the rest of the data.
 dplyr::rename(dataSource = Source) \%>\%
 # add a NEW database_id column
 dplyr::mutate(
   database_id = paste0("Paige_data_", 1:nrow(.)),
   .before = scientificName)

 # Set up the list of character vectors to iteratively check for matches with public data.
columnList <- list(
 c("decimalLatitude", "decimalLongitude", 
   "recordNumber", "recordedBy", "individualCount", "samplingProtocol",
   "associatedTaxa", "sex", "catalogNumber", "institutionCode", "otherCatalogNumbers",
   "recordId", "occurrenceID", "collectionID"), # Iteration 1
 c("catalogNumber", "institutionCode", "otherCatalogNumbers",
   "recordId", "occurrenceID", "collectionID"), # Iteration 2
 c("decimalLatitude", "decimalLongitude", 
   "recordedBy", "genus", "specificEpithet"), # Iteration 3
 c("id", "decimalLatitude", "decimalLongitude"), # Iteration 4
 c("recordedBy", "genus", "specificEpithet", "locality"), # Iteration 5
 c("recordedBy", "institutionCode", "genus", 
   "specificEpithet","locality"),# Iteration 6
 c("occurrenceID","decimalLatitude", "decimalLongitude"), # Iteration 7
 c("catalogNumber","decimalLatitude", "decimalLongitude"), # Iteration 8
 c("catalogNumber", "locality") # Iteration 9
) 

# Merge Paige's data with downloaded data
db_standardized <- BeeBDC::PaigeIntegrater(
 db_standardized = db_standardized,
 PaigeNAm = PaigeNAm,
 columnStrings = columnList)
}


}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/dupeSummary.R
\name{dupeSummary}
\alias{dupeSummary}
\title{Identifies duplicate occurrence records}
\usage{
dupeSummary(
  data = NULL,
  path = NULL,
  duplicatedBy = NULL,
  completeness_cols = NULL,
  idColumns = NULL,
  collectionCols = NULL,
  collectInfoColumns = NULL,
  CustomComparisonsRAW = NULL,
  CustomComparisons = NULL,
  sourceOrder = NULL,
  prefixOrder = NULL,
  dontFilterThese = c(".gridSummary", ".lonFlag", ".latFlag", ".uncer_terms",
    ".uncertaintyThreshold", ".unLicensed"),
  characterThreshold = 2,
  numberThreshold = 3,
  numberOnlyThreshold = 5,
  catalogSwitch = TRUE
)
}
\arguments{
\item{data}{A data frame or tibble. Occurrence records as input.}

\item{path}{A character path to the location where the duplicateRun_ file will be saved.}

\item{duplicatedBy}{A character vector. Options are c("ID", "collectionInfo", "both"). "ID"
columns runs through a series of ID-only columns defined by idColumns. "collectionInfo" runs
through a series of columns defined by collectInfoColumns, which are checked in combination
with collectionCols. "both" runs both of the above.}

\item{completeness_cols}{A character vector. A set of columns that are used to order and select
duplicates by. For each occurrence, this function will calculate the sum of \code{\link[=complete.cases]{complete.cases()}}.
Within duplicate clusters occurrences with a greater number of the completeness_cols filled
in will be kept over those with fewer.}

\item{idColumns}{A character vector. The columns to be checked individually for internal
duplicates. Intended for use with ID columns only.}

\item{collectionCols}{A character vector. The columns to be checked in combination with each
of the completeness_cols.}

\item{collectInfoColumns}{A character vector. The columns to be checked in combinatino with
all of the collectionCols columns.}

\item{CustomComparisonsRAW}{A list of character vectors. Custom comparisons - as a list of
columns to iteratively compare for duplicates. These differ from the CustomComparisons in
that they ignore the minimum number and character thresholds for IDs.}

\item{CustomComparisons}{A list of character vectors. Custom comparisons - as a list of
columns to iteratively compare for duplicates. These comparisons are made after character
and number thresholds are accounted for in ID columns.}

\item{sourceOrder}{A character vector. The order in which you want to KEEP duplicated
based on the dataSource column (i.e. what order to prioritize data sources).
NOTE: These dataSources are simplified to the string prior
to the first "_". Hence, "GBIF_Anthophyla" becomes "GBIF."}

\item{prefixOrder}{A character vector. Like sourceOrder, except based on the database_id prefix,
rather than the dataSource. Additionally, this is only examined if prefixOrder != NULL.
Default = NULL.}

\item{dontFilterThese}{A character vector. This should contain the flag columns to be ignored
in the creation or updating of the .summary column. Passed to  \code{\link[=summaryFun]{summaryFun()}}.}

\item{characterThreshold}{Numeric. The complexity threshold for ID letter length. This is the
minimum number of characters that need to be present in ADDITION TO the numberThreshold for an
ID number to be tested for duplicates. Ignored by CustomComparisonsRAW. The columns that are
checked are occurrenceID, recordId, id, catalogNumber, and otherCatalogNumbers. Default = 2.}

\item{numberThreshold}{Numeric. The complexity threshold for ID number length. This is the
minimum number of numeric characters that need to be present in ADDITION TO the
characterThreshold for an ID number to be tested for duplicates. Ignored by
CustomComparisonsRAW. The columns that are checked are occurrenceID, recordId, id,
catalogNumber, and otherCatalogNumbers. Default = 3.}

\item{numberOnlyThreshold}{Numeric. As numberThreshold except the characterThreshold is ignored.
Default = 5.}

\item{catalogSwitch}{Logical. If TRUE, and the catalogNumber is empty the function will copy over
the otherCatalogNumbers into catalogNumber and visa versa. Hence, the function will attempt
to matchmore catalog numbers as both of these functions can be problematic. Default = TRUE.}
}
\value{
Returns data with an additional column called .duplicates where FALSE occurrences are
duplicates and TRUE occurrences are either kept duplicates or unique. Also exports a .csv to
the user-specified location with information about duplicate matching. This file is used by
other functions including
\code{\link[=manualOutlierFindeR]{manualOutlierFindeR()}} and \code{\link[=chordDiagramR]{chordDiagramR()}}
}
\description{
This function uses user-specified inputs and columns to identify duplicate occurrence records.
Duplicates are identified iteratively and will be tallied up, duplicate pairs clustered, and
sorted at the end of the function.
The function is designed to work with Darwin Core data with a database_id column,
but it is also modifiable to work with other columns.
}
\examples{
beesFlagged_out <- dupeSummary(
data = BeeBDC::beesFlagged,
  # Should start with paste0(DataPath, "/Output/Report/"), instead of tempdir():
path = paste0(tempdir(), "/"),
# options are "ID","collectionInfo", or "both"
duplicatedBy = "collectionInfo", # I'm only running ID for the first lot because we might 
# recover other info later
# The columns to generate completeness info from
completeness_cols = c("decimalLatitude",  "decimalLongitude",
                      "scientificName", "eventDate"),
# idColumns = c("gbifID", "occurrenceID", "recordId","id"),
# The columns to ADDITIONALLY consider when finding duplicates in collectionInfo
collectionCols = c("decimalLatitude", "decimalLongitude", "scientificName", "eventDate", 
                   "recordedBy"),
# The columns to combine, one-by-one with the collectionCols
collectInfoColumns = c("catalogNumber", "otherCatalogNumbers"),
# Custom comparisons - as a list of columns to compare
# RAW custom comparisons do not use the character and number thresholds
CustomComparisonsRAW = dplyr::lst(c("catalogNumber", "institutionCode", "scientificName")),
# Other custom comparisons use the character and number thresholds
CustomComparisons = dplyr::lst(c("gbifID", "scientificName"),
                                c("occurrenceID", "scientificName"),
                                c("recordId", "scientificName"),
                                c("id", "scientificName")),
# The order in which you want to KEEP duplicated based on data source
# try unique(check_time$dataSource)
sourceOrder = c("CAES", "Gai", "Ecd","BMont", "BMin", "EPEL", "ASP", "KP", "EcoS", "EaCO",
                "FSCA", "Bal", "SMC", "Lic", "Arm",
                "USGS", "ALA", "GBIF","SCAN","iDigBio"),
# !!!!!! BELS > GeoLocate
# Set the complexity threshold for id letter and number length
# minimum number of characters when WITH the numberThreshold
characterThreshold = 2,
# minimum number of numbers when WITH the characterThreshold
numberThreshold = 3,
# Minimum number of numbers WITHOUT any characters
numberOnlyThreshold = 5)


}
\seealso{
\code{\link[=chordDiagramR]{chordDiagramR()}} for creating a chord diagram to visualise linkages between
dataSources and \code{\link[=dupePlotR]{dupePlotR()}} to visualise the numbers and proportions of duplicates in
each dataSource.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/chordDiagramR.R
\name{chordDiagramR}
\alias{chordDiagramR}
\title{Build a chord diagram of duplicate occurrence links}
\usage{
chordDiagramR(
  dupeData = NULL,
  outPath = NULL,
  fileName = NULL,
  width = 7,
  height = 6,
  bg = "white",
  smallGrpThreshold = 3,
  title = "Duplicated record sources",
  palettes = c("cartography::blue.pal", "cartography::green.pal",
    "cartography::sand.pal", "cartography::orange.pal", "cartography::red.pal",
    "cartography::purple.pal", "cartography::brown.pal"),
  canvas.ylim = c(-1, 1),
  canvas.xlim = c(-0.6, 0.25),
  text.col = "black",
  legendX = grid::unit(6, "mm"),
  legendY = grid::unit(18, "mm"),
  legendJustify = c("left", "bottom"),
  niceFacing = TRUE,
  self.link = 2
)
}
\arguments{
\item{dupeData}{A tibble or data frame. The duplicate file produced by \code{\link[=dupeSummary]{dupeSummary()}}.}

\item{outPath}{Character. The path to a directory (folder) in which the output should be saved.}

\item{fileName}{Character. The name of the output file, ending in '.pdf'.}

\item{width}{Numeric. The width of the figure to save (in inches). Default = 7.}

\item{height}{Numeric. The height of the figure to save (in inches). Default = 6.}

\item{bg}{The plot's background colour. Default = "white".}

\item{smallGrpThreshold}{Numeric. The upper threshold of sub-dataSources to be listed as "other".
Default = 3.}

\item{title}{A character string. The figure title. Default = "Duplicated record sources".}

\item{palettes}{A vector of the palettes to be used. One palette for each major dataSource and "other"
using the \code{paletteer} package. Default = c("cartography::blue.pal", "cartography::green.pal",
"cartography::sand.pal", "cartography::orange.pal", "cartography::red.pal",
"cartography::purple.pal", "cartography::brown.pal")}

\item{canvas.ylim}{Canvas limits from \code{\link[circlize:circos.par]{circlize::circos.par()}}. Default = c(-1.0,1.0).}

\item{canvas.xlim}{Canvas limits from \code{\link[circlize:circos.par]{circlize::circos.par()}}. Default = c(-0.6, 0.25).}

\item{text.col}{A character string. Text colour}

\item{legendX}{The x position of the legends, as measured in current viewport.
Passed to ComplexHeatmap::draw(). Default = grid::unit(6, "mm").}

\item{legendY}{The y position of the legends, as measured in current viewport.
Passed to ComplexHeatmap::draw(). Default = grid::unit(18, "mm").}

\item{legendJustify}{A character vector declaring the justification of the legends.
Passed to ComplexHeatmap::draw(). Default = c("left", "bottom").}

\item{niceFacing}{TRUE/FALSE. The niceFacing option automatically adjusts the text facing
according to their positions in the circle. Passed to \code{\link[circlize:highlight.sector]{circlize::highlight.sector()}}.}

\item{self.link}{1 or 2 (numeric). Passed to \code{\link[circlize:chordDiagram]{circlize::chordDiagram()}}:
if there is a self link in one sector, 1 means the link will be degenerated as a 'mountain' and the width corresponds to the value for this connection. 2 means the width of the starting root and the ending root all have the width that corresponds to the value for the connection.}
}
\value{
Saves a figure to the provided file path.
}
\description{
This function outputs a figure which shows the relative size and direction of occurrence points
duplicated between data providers, such as, SCAN, GBIF, ALA, etc. This function requires the
outputs generated by \code{\link[=dupeSummary]{dupeSummary()}}.
}
\examples{
\dontrun{
  # Create a basic example dataset of duplicates to visualise
basicData <- dplyr::tribble(
                            ~dataSource,    ~dataSource_keep,
                      "GBIF_Halictidae",         "USGS_data",
                      "GBIF_Halictidae",         "USGS_data",
                      "GBIF_Halictidae",         "USGS_data",
                      "GBIF_Halictidae",         "USGS_data",
                      "GBIF_Halictidae",         "USGS_data",
                      "GBIF_Halictidae",         "USGS_data",
                      "SCAN_Halictidae",   "GBIF_Halictidae",
                   "iDigBio_halictidae",   "GBIF_Halictidae",
                   "iDigBio_halictidae",   "SCAN_Halictidae",
                   "iDigBio_halictidae",   "SCAN_Halictidae",
                      "SCAN_Halictidae",   "GBIF_Halictidae",
                       "iDigBio_apidae",       "SCAN_Apidae",
                          "SCAN_Apidae",    "Ecd_Anthophila",
                       "iDigBio_apidae",    "Ecd_Anthophila",
                          "SCAN_Apidae",    "Ecd_Anthophila",
                       "iDigBio_apidae",    "Ecd_Anthophila",
                    "SCAN_Megachilidae", "SCAN_Megachilidae",
                      "CAES_Anthophila",   "CAES_Anthophila",
                      "CAES_Anthophila",   "CAES_Anthophila"
 )


 chordDiagramR(
dupeData = basicData,
outPath = tempdir(),
fileName = "ChordDiagram.pdf",
# These can be modified to help fit the final pdf that's exported.
width = 9,
height = 7.5,
bg = "white",
# How few distinct dataSources should a group have to be listed as "other"
smallGrpThreshold = 3,
title = "Duplicated record sources",
# The default list of colour palettes to choose from using the paleteer package
palettes = c("cartography::blue.pal", "cartography::green.pal", 
             "cartography::sand.pal", "cartography::orange.pal", "cartography::red.pal",
             "cartography::purple.pal", "cartography::brown.pal"),
canvas.ylim = c(-1.0,1.0), 
canvas.xlim = c(-0.6, 0.25),
text.col = "black",
legendX = grid::unit(6, "mm"),
legendY = grid::unit(18, "mm"),
legendJustify = c("left", "bottom"),
niceFacing = TRUE)}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/continentOutliers.R
\name{continentOutlieRs}
\alias{continentOutlieRs}
\title{Flag continent-level outliers with a provided checklist.}
\usage{
continentOutlieRs(
  checklist = NULL,
  data = NULL,
  keepAdjacentContinent = FALSE,
  pointBuffer = NULL,
  scale = 50,
  stepSize = 1e+06,
  mc.cores = 1
)
}
\arguments{
\item{checklist}{A data frame or tibble. The formatted checklist which was built based on the Discover Life website.}

\item{data}{A data frame or tibble. The a Darwin Core occurrence dataset.}

\item{keepAdjacentContinent}{Logical. If TRUE, occurrences in continents that are adjacent to checklist continents will be
kept. If FALSE, they will be flagged. Defualt = FALSE.}

\item{pointBuffer}{Numeric. A buffer around points to help them align with a continent or coastline.
This provides a good way to retain points that occur right along the coast or borders of the
maps in rnaturalearth}

\item{scale}{Numeric. The value fed into the map scale parameter for
\code{\link[rnaturalearth:ne_countries]{rnaturalearth::ne_countries()}}'s scale parameter:
Scale of map to return, one of 110, 50, 10 or 'small', 'medium', 'large', where smaller numbers
are higher resolution. WARNING: This function is tested on 110 and 50.}

\item{stepSize}{Numeric. The number of occurrences to process in each chunk. Default = 1000000.}

\item{mc.cores}{Numeric. If > 1, the function will run in parallel
using mclapply using the number of cores specified. If = 1 then it will be run using a serial
loop. NOTE: Windows machines must use a value of 1 (see ?parallel::mclapply). Additionally,
be aware that each thread can use large chunks of memory. If the cores throw issues, consider
setting mc.cores to 1.
Default = 1.}
}
\value{
The input data with two new columns, .continentOutlier or .sea. There are three possible
values for
the new column: TRUE == passed, FALSE == failed (not in continent or in the ocean),
NA == did not overlap with rnaturalearth map.
}
\description{
This function flags continent-level outliers using the checklist provided with this package.
For additional context and column names, see \code{\link[=beesChecklist]{beesChecklist()}}.
}
\examples{
if(requireNamespace("rnaturalearthdata")){
library(magrittr)
  # Load in the test dataset
beesRaw <- BeeBDC::beesRaw
  # For the sake of this example, use the testChecklist
system.file("extdata", "testChecklist.rda", package="BeeBDC") |> load()
  # For real examples, you might download the beesChecklist from FigShare using 
  #  [BeeBDC::beesChecklist()]

beesRaw_out <- continentOutlieRs(checklist = testChecklist,
                               data = beesRaw \%>\%
                               dplyr::filter(dplyr::row_number() \%in\% 1:50),
                               keepAdjacentContinent = FALSE,
                               pointBuffer = 1,
                               scale = 50,
                               stepSize = 1000000,
                               mc.cores = 1)
table(beesRaw_out$.continentOutlier, useNA = "always")
} # END if require
}
\seealso{
\code{\link[=countryOutlieRs]{countryOutlieRs()}} for implementation at the country level. Country-level
implementation will be more data-hungry and, where data do not yet exist, difficult to implement.
Additionally, see \code{\link[=beesChecklist]{beesChecklist()}} for input data. Note, not all columns are
necessary if you are building your own dataset. At a minimum you will need \emph{validName} and
\emph{continent}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/iNEXTwrapper.R
\name{iNEXTwrapper}
\alias{iNEXTwrapper}
\title{Parallel estimation of species richness in a community using iNEXT}
\usage{
iNEXTwrapper(
  data = NULL,
  variableColumn = "groupVariable",
  valueColumn = "n",
  q = 0,
  datatype = "abundance",
  conf = 0.95,
  se = TRUE,
  nboot = 50,
  size = NULL,
  endpoint = NULL,
  knots = 40,
  mc.cores = 1
)
}
\arguments{
\item{data}{A data frame or tibble. A data frame containing "abundance"-type data per variable
(population, country, species...) in columns.}

\item{variableColumn}{Character. The column to be used to group the data. Probably "country"
or "site". Default =  "groupVariable".}

\item{valueColumn}{Character. The column containing the count data. Defualt =  "n".}

\item{q}{a number or vector specifying the richness order(s) of Hill numbers.}

\item{datatype}{data type of input data: individual-based abundance data
(datatype = "abundance"), sampling-unit-based incidence frequencies data
(datatype = "incidence_freq") or species by sampling-units incidence matrix
(datatype = "incidence_raw").}

\item{conf}{a positive number < 1 specifying the level of confidence interval; default is 0.95.}

\item{se}{a logical variable to calculate the bootstrap standard error and conf confidence
interval.}

\item{nboot}{an integer specifying the number of replications; default is 50.}

\item{size}{an integer vector of sample sizes (number of individuals or sampling units) for
which richness estimates will be computed. If NULL, then richness estimates will be computed
for those sample sizes determined by the specified/default endpoint and knots.}

\item{endpoint}{an integer specifying the sample size that is the endpoint for
rarefaction/extrapolation. If NULL, then endpoint = double the reference sample size.}

\item{knots}{an integer specifying the number of equally-spaced knots (say K, default is 40)
between size 1 and the endpoint; each knot represents a particular sample size for which
richness estimate will be calculated. If the endpoint is smaller than the reference sample
size, then iNEXT() computes only the rarefaction esimates for approximately K evenly spaced
knots. If the endpoint is larger than the reference sample size, then iNEXT() computes
rarefaction estimates for approximately K/2 evenly spaced knots between sample size 1 and
the reference sample size, and computes extrapolation estimates for approximately K/2 evenly
spaced knots between the reference sample size and the endpoint.}

\item{mc.cores}{Numeric. If > 1, the function will run in parallel
using mclapply using the number of cores specified. If = 1 then it will be run using a serial
loop. NOTE: Windows machines must use a value of 1 (see ?parallel::mclapply). Additionally,
be aware that each thread can use large chunks of memory.
Default = 1.}
}
\value{
Returns a list containing two tibbles. The first is a tibble that concatenates the outputs
from the basic data and rare species information in columns per input variable (column). The second is a tibble
that concatenates the various species richness estimates, with input variables in chunks of rows.
Additionally a console output will list the variables (columns) that lacked sufficient data
to be analysed.
}
\description{
A wrapper for \code{iNEXT::iNEXT()} to interpolate and extrapolate Hill numbers with order q
(rarify species richness). The wrapper has the ability to estimate species richness for
multiple sites (or countries) at once and to do this using multiple cores.
}
\examples{

\dontrun{
  # Read in some example data and use [BeeBDC::richnessPrepR()] to create the example input data
 #' data(beesCountrySubset)

estimateDataExample <- BeeBDC::richnessPrepR(
  data = beesCountrySubset,
  # Download the taxonomy
  taxonomyFile = BeeBDC::beesTaxonomy(),
  # Download the checklist
  checklistFile = BeeBDC::beesChecklist(),
  curveFunction = function(x) (228.7531 * x * x^-log(12.1593)),
  sampleSize = 10000,
  countryColumn = "country_suggested",
  limitGlobal = NULL,
  outPath = tempdir()
)

  # In this function we can directly feed in estimateDataExample$site_speciesCounts 
 iNextOut <- iNEXTwrapper(data = estimateDataExample$site_speciesCounts,
                          variableColumn = "country_suggested",
                          valueColumn = "n",
                          q = 0,
                          datatype = "abundance",
                          conf = 0.95,
                          se = TRUE,
                          nboot = 50,
                          size = NULL,
                          endpoint = NULL,
                          knots = 40,
                          mc.cores = 1)
}
}
\seealso{
\code{\link[iNEXT:iNEXT]{iNEXT::iNEXT()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/BeeBDCQuery.R
\name{BeeBDCQuery}
\alias{BeeBDCQuery}
\title{Query the bee taxonomy and country checklist}
\usage{
BeeBDCQuery(
  beeName = NULL,
  searchChecklist = TRUE,
  printAllSynonyms = FALSE,
  beesChecklist = NULL,
  beesTaxonomy = NULL
)
}
\arguments{
\item{beeName}{Character or character vector. A single or several bee species names to search for
in the beesTaxonomy and beesChecklist tables.}

\item{searchChecklist}{Logical. If TRUE (default), search the country checklist for each species.}

\item{printAllSynonyms}{Logical. If TRUE, all synonyms will be printed out for each entered name.
default = FALSE.}

\item{beesChecklist}{A tibble. The bee checklist file for BeeBDC. If is NULL then
\code{\link[=beesChecklist]{beesChecklist()}} will be called internally to download the file. Default = NULL.}

\item{beesTaxonomy}{A tibble. The bee taxonomy file for BeeBDC. If is NULL then
\code{\link[=beesTaxonomy]{beesTaxonomy()}} will be called internally to download the file. Default = NULL.}
}
\value{
Returns a list with the elements 'taxonomyReport' and 'SynonymReport'. IF searchChecklist
is TRUE, then 'checklistReport' will also be returned.
}
\description{
A simple function to return information about a particular species, including name validity and
country occurrences.
}
\examples{
  # For the sake of these examples, we will use the example taxonomy and checklist
  system.file("extdata", "testTaxonomy.rda", package="BeeBDC") |> load()
  system.file("extdata", "testChecklist.rda", package="BeeBDC") |> load()

  # Single entry example
testQuery <- BeeBDCQuery(
  beeName = "Lasioglossum bicingulatum",
  searchChecklist = TRUE,
  printAllSynonyms = TRUE,
  beesTaxonomy = testTaxonomy,
  beesChecklist = testChecklist)

  # Multiple entry example
testQuery <- BeeBDC::BeeBDCQuery(
  beeName = c("Lasioglossum bicingulatum", "Nomada flavopicta",
  "Lasioglossum fijiense (Perkins and Cheesman, 1928)"),
  searchChecklist = TRUE,
  printAllSynonyms = TRUE,
  beesTaxonomy = testTaxonomy,
  beesChecklist = testChecklist)
  
    # Example way to examine a report from the output list
  testQuery$checklistReport



}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/toyData_beesFlagged.R
\docType{data}
\name{beesFlagged}
\alias{beesFlagged}
\title{A flagged dataset of 100 random bee occurrence records}
\format{
An object of class \code{"tibble"}
\describe{
\item{database_id}{Occurrence code generated in bdc or BeeBDC}
\item{scientificName}{Full scientificName as shown on DiscoverLife}
\item{family}{Family name}
\item{subfamily}{Subfamily name}
\item{genus}{Genus name}
\item{subgenus}{Subgenus name}
\item{subspecies}{Full name with subspecies name - ALA column}
\item{specificEpithet}{The species name only}
\item{infraspecificEpithet}{The subspecies name only}
\item{acceptedNameUsage}{The full name, with authorship and date information if known, of the currently valid (zoological) or accepted (botanical) taxon.}
\item{taxonRank}{The taxonomic rank of the most specific name in the scientificName.}
\item{scientificNameAuthorship}{The authorship information for the scientificName formatted according to the conventions of the applicable nomenclaturalCode.}
\item{identificationQualifier}{A brief phrase or a standard term ("cf.", "aff.") to express the determiner's doubts about the Identification.}
\item{higherClassification}{A list (concatenated and separated) of taxa names terminating at the rank immediately superior to the taxon referenced in the taxon record.)}
\item{identificationReferences}{A list (concatenated and separated) of references (publication, global unique identifier, URI) used in the Identification.}
\item{typeStatus}{A list (concatenated and separated) of nomenclatural types (type status, typified scientific name, publication) applied to the subject.}
\item{previousIdentifications}{A list (concatenated and separated) of previous assignments of names to the Organism.}
\item{verbatimIdentification}{This term is meant to allow the capture of an unaltered original identification/determination, including identification qualifiers, hybrid formulas, uncertainties, etc. This term is meant to be used in addition to scientificName (and identificationQualifier etc.), not instead of it.}
\item{identifiedBy}{A list (concatenated and separated) of names of people, groups, or organizations who assigned the Taxon to the subject.}
\item{dateIdentified}{The date on which the subject was determined as representing the Taxon.}
\item{decimalLatitude}{The geographic latitude (in decimal degrees, using the spatial reference system given in geodeticDatum) of the geographic center of a Location. Positive values are north of the Equator, negative values are south of it. Legal values lie between -90 and 90, inclusive.}
\item{decimalLongitude}{The geographic longitude (in decimal degrees, using the spatial reference system given in geodeticDatum) of the geographic center of a Location. Positive values are east of the Greenwich Meridian, negative values are west of it. Legal values lie between -180 and 180, inclusive.}
\item{stateProvince}{The name of the next smaller administrative region than country (state, province, canton, department, region, etc.) in which the Location occurs.}
\item{continent}{The name of the continent in which the Location occurs.}
\item{locality}{The specific description of the place.}
\item{island}{The name of the island on or near which the Location occurs.}
\item{county}{The full, unabbreviated name of the next smaller administrative region than stateProvince (county, shire, department, etc.) in which the Location occurs.}
\item{municipality}{The full, unabbreviated name of the next smaller administrative region than county (city, municipality, etc.) in which the Location occurs. Do not use this term for a nearby named place that does not contain the actual location.}
\item{license}{A legal document giving official permission to do something with the resource.}
\item{issue}{A GBIF-defined issue.}
\item{eventDate}{The date-time or interval during which an Event occurred. For occurrences, this is the date-time when the event was recorded. Not suitable for a time in a geological context.}
\item{eventTime}{The time or interval during which an Event occurred.}
\item{day}{The integer day of the month on which the Event occurred.}
\item{month}{The integer month in which the Event occurred.}
\item{year}{The four-digit year in which the Event occurred, according to the Common Era Calendar.}
\item{basisOfRecord}{The specific nature of the data record. Recommended best practice is to use the standard label of one of the Darwin Core classes.PreservedSpecimen, FossilSpecimen, LivingSpecimen, MaterialSample, Event, HumanObservation, MachineObservation, Taxon, Occurrence, MaterialCitation}
\item{country}{The name of the country or major administrative unit in which the Location occurs.}
\item{type}{The nature or genre of the resource. StillImage, MovingImage, Sound, PhysicalObject, Event, Text.}
\item{occurrenceStatus}{A statement about the presence or absence of a Taxon at a Location. present, absent.}
\item{recordNumber}{An identifier given to the Occurrence at the time it was recorded. Often serves as a link between field notes and an Occurrence record, such as a specimen collector's number.}
\item{recordedBy}{A list (concatenated and separated) of names of people, groups, or organizations responsible for recording the original Occurrence. The primary collector or observer, especially one who applies a personal identifier (recordNumber), should be listed first.}
\item{eventID}{An identifier for the set of information associated with an Event (something that occurs at a place and time). May be a global unique identifier or an identifier specific to the data set.}
\item{Location}{A spatial region or named place.}
\item{samplingProtocol}{The names of, references to, or descriptions of the methods or protocols used during an Event. Examples	UV light trap, mist net, bottom trawl, ad hoc observation | point count, Penguins from space: faecal stains reveal the location of emperor penguin colonies, https://doi.org/10.1111/j.1466-8238.2009.00467.x, Takats et al. 2001.}
\item{samplingEffort}{The amount of effort expended during an Event. Examples	40 trap-nights, 10 observer-hours, 10 km by foot, 30 km by car.}
\item{individualCount}{The number of individuals present at the time of the Occurrence. Integer.}
\item{organismQuantity}{A number or enumeration value for the quantity of organisms. Examples	27 (organismQuantity) with individuals (organismQuantityType). 12.5 (organismQuantity) with percentage biomass (organismQuantityType). r (organismQuantity) with Braun Blanquet Scale (organismQuantityType). many (organismQuantity) with individuals (organismQuantityType).}
\item{coordinatePrecision}{A decimal representation of the precision of the coordinates given in the decimalLatitude and decimalLongitude.}
\item{coordinateUncertaintyInMeters}{The horizontal distance (in meters) from the given decimalLatitude and decimalLongitude describing the smallest circle containing the whole of the Location. Leave the value empty if the uncertainty is unknown, cannot be estimated, or is not applicable (because there are no coordinates). Zero is not a valid value for this term.}
\item{spatiallyValid}{Occurrence records in the ALA can be filtered by using the spatially valid flag. This flag combines a set of tests applied to the record to see how reliable are its spatial data components.}
\item{catalogNumber}{An identifier (preferably unique) for the record within the data set or collection.}
\item{gbifID}{The identifier assigned by GBIF for each record.}
\item{datasetID}{An identifier for the set of data. May be a global unique identifier or an identifier specific to a collection or institution.}
\item{institutionCode}{The name (or acronym) in use by the institution having custody of the object(s) or information referred to in the record. Examples	MVZ, FMNH, CLO, UCMP.}
\item{datasetName}{The name identifying the data set from which the record was derived.}
\item{otherCatalogNumbers}{A list (concatenated and separated) of previous or alternate fully qualified catalog numbers or other human-used identifiers for the same Occurrence, whether in the current or any other data set or collection.}
\item{occurrenceID}{An identifier for the Occurrence (as opposed to a particular digital record of the occurrence). In the absence of a persistent global unique identifier, construct one from a combination of identifiers in the record that will most closely make the occurrenceID globally unique.}
\item{taxonKey}{The GBIF-assigned taxon identifier number.}
\item{collectionID}{An identifier for the collection or dataset from which the record was derived.}
\item{verbatim_scientificName}{The verbatim (originally-provided) scientific name}
\item{verbatimEventDate}{The verbatim original representation of the date and time information for an Event.}
\item{associatedTaxa}{A list (concatenated and separated) of identifiers or names of taxa and the associations of this Occurrence to each of them.}
\item{associatedOrganisms}{A list (concatenated and separated) of identifiers of other Organisms and the associations of this Organism to each of them.}
\item{fieldNotes}{One of a) an indicator of the existence of, b) a reference to (publication, URI), or c) the text of notes taken in the field about the Event.}
\item{sex}{The sex of the biological individual(s) represented in the Occurrence.}
\item{rights}{A description of the usage rights applicable to the record.}
\item{rightsHolder}{A person or organization owning or managing rights over the resource.}
\item{accessRights}{Information about who can access the resource or an indication of its security status.}
\item{associatedReferences}{A list (concatenated and separated) of identifiers (publication, bibliographic reference, global unique identifier, URI) of literature associated with the Occurrence.}
\item{bibliographicCitation}{A bibliographic reference for the resource as a statement indicating how this record should be cited (attributed) when used.}
\item{references}{A related resource that is referenced, cited, or otherwise pointed to by the described resource.}
\item{informationWithheld}{Additional information that exists, but that has not been shared in the given record.}
\item{isDuplicateOf}{Additional information that exists, but that has not been shared in the given record.}
\item{hasCoordinate}{Variable indicating presence/absence of location coordinates.}
\item{hasGeospatialIssues}{Variable indicating validity of geospatial data associated with record.}
\item{occurrenceYear}{Year associated with Occurrence.}
\item{id}{Variable with identifying value for the Occurrenc.}
\item{duplicateStatus}{Variable indicating is Occurrence is duplicate or not.}
\item{associatedOccurrences}{A list (concatenated and separated) of identifiers of other Occurrence records and their associations to this Occurrence.}
\item{locationRemarks}{Comments or notes about the Location.}
\item{dataSource}{BeeBDC assigned source of the data. Often written when the data is formatted by a BeeBDC::xxx_readr function or similar.}
\item{verbatim_scientificName}{The verbatim (originally-provided) scientific name}
\item{.scientificName_empty}{Flag produced by \code{\link[bdc:bdc_scientificName_empty]{bdc::bdc_scientificName_empty()}} where FALSE == no scientific name provided and TRUE means that there is text in that column.}
\item{.coordinates_empty}{Flag produced by \code{\link[bdc:bdc_coordinates_empty]{bdc::bdc_coordinates_empty()}} where FALSE == no coordinates provided.}
\item{.coordinates_outOfRange}{Flag produced by \code{\link[bdc:bdc_coordinates_outOfRange]{bdc::bdc_coordinates_outOfRange()}} where FALSE == point off the earth. This function identifies records with out-of-range coordinates (not between -90 and 90 for latitude; between -180 and 180 for longitude).}
\item{.basisOfRecords_notStandard}{Flag produced by \code{\link[bdc:bdc_basisOfRecords_notStandard]{bdc::bdc_basisOfRecords_notStandard()}} where FALSE == an occurrence with a basisOfRecord not defined as acceptable by the user.}
\item{country_suggested}{A country name suggested by the \code{\link[bdc:bdc_country_standardized]{bdc::bdc_country_standardized()}} function.}
\item{countryCode}{A country code suggested by the \code{\link[bdc:bdc_country_standardized]{bdc::bdc_country_standardized()}} function.}
\item{coordinates_transposed}{A column indicating if coordinates were tansposed by \code{\link[=jbd_Ctrans_chunker]{jbd_Ctrans_chunker()}} where FALSE == transposed.}
\item{.coordinates_country_inconsistent}{A flag generated by \code{\link[=jbd_coordCountryInconsistent]{jbd_coordCountryInconsistent()}} where FALSE == an occurrence where the country name and coordinates did not match.}
\item{.occurrenceAbsent}{A flag generated by \code{\link[=flagAbsent]{flagAbsent()}} where FALSE == occurrences marked as "ABSENT" in the "occurrenceStatus" column}
\item{.unLicensed}{A flag generated by \code{\link[=flagLicense]{flagLicense()}} where FALSE == those occurrences protected by a restrictive license.}
\item{.GBIFflags}{A flag generated by \code{\link[=GBIFissues]{GBIFissues()}} where FALSE == an occurrence with user-specified GBIF issues to flag.}
\item{.uncer_terms}{A flag generated by \code{\link[bdc:bdc_clean_names]{bdc::bdc_clean_names()}} where FALSE == the presence of taxonomic uncertainty terms.}
\item{names_clean}{A column made by \code{\link[bdc:bdc_clean_names]{bdc::bdc_clean_names()}} indicating the cleaned scientificName}
\item{.invalidName}{A flag generated by \code{\link[=harmoniseR]{harmoniseR()}} where FALSE == occurrences whose scientificName did not match the Discover Life taxonomy.}
\item{.rou}{A flag generated by \code{\link[CoordinateCleaner:clean_coordinates]{CoordinateCleaner::clean_coordinates()}} where FALSE == rounded (probably imprecise) coordinates.}
\item{.val}{A flag generated by \code{\link[CoordinateCleaner:clean_coordinates]{CoordinateCleaner::clean_coordinates()}} where FALSE == invalid coordinates.}
\item{.equ}{A flag generated by \code{\link[CoordinateCleaner:clean_coordinates]{CoordinateCleaner::clean_coordinates()}} where FALSE == equal coordinates (e.g., 0.1, 0.1).}
\item{.zer}{A flag generated by \code{\link[CoordinateCleaner:clean_coordinates]{CoordinateCleaner::clean_coordinates()}} where FALSE == zeros as coordinates}
\item{.cap}{A flag generated by \code{\link[CoordinateCleaner:clean_coordinates]{CoordinateCleaner::clean_coordinates()}} where FALSE == records around country capital centroid.}
\item{.cen}{A flag generated by \code{\link[CoordinateCleaner:clean_coordinates]{CoordinateCleaner::clean_coordinates()}} where FALSE == records around country or province centroids.}
\item{.gbf}{A flag generated by \code{\link[CoordinateCleaner:clean_coordinates]{CoordinateCleaner::clean_coordinates()}} where FALSE == records around the GBIF headquarters.}
\item{.inst}{A flag generated by \code{\link[CoordinateCleaner:clean_coordinates]{CoordinateCleaner::clean_coordinates()}} where FALSE == records around biodiversity institutions.}
\item{.sequential}{A flag generated by \code{\link[=diagonAlley]{diagonAlley()}} where FALSE == records that are possibly the result of fill-down errors in sequence.}
\item{.lonFlag}{A flag generated by \code{\link[CoordinateCleaner:cd_round]{CoordinateCleaner::cd_round()}} where FALSE == potential gridding in the longitude column within dataset.}
\item{.latFlag}{A flag generated by \code{\link[CoordinateCleaner:cd_round]{CoordinateCleaner::cd_round()}} where FALSE == potential gridding in the latitude column within dataset.}
\item{.gridSummary}{A flag generated by \code{\link[CoordinateCleaner:cd_round]{CoordinateCleaner::cd_round()}} where FALSE == potential gridding in either the longitude or latitude columns within dataset.}
\item{.uncertaintyThreshold}{A flag generated by \code{\link[=coordUncerFlagR]{coordUncerFlagR()}} where FALSE == occurrences that did not pass a user-specified threshold in the "coordinateUncertaintyInMeters" column.}
\item{countryMatch}{A column made by \code{\link[=countryOutlieRs]{countryOutlieRs()}}. Summarises the occurrence-level result: where the species is not known to occur in that country (noMatch), it is known from a bordering country (neighbour), or it is known to occur in that country (exact).}
\item{.countryOutlier}{A flag generated by \code{\link[=countryOutlieRs]{countryOutlieRs()}} where FALSE == occurrences the do not occur in a country that concurs with the Discover Life country checklist OR an adjacent country.}
\item{.sea}{A flag generated by \code{\link[=countryOutlieRs]{countryOutlieRs()}} where FALSE == occurrences that are in the ocean.}
\item{.summary}{A flag generated by \code{\link[=summaryFun]{summaryFun()}} where FALSE == occurrences flagged as FALSE in any of the .flag columns. In this example it excludes flags in the ".gridSummary", ".lonFlag", ".latFlag", and ".uncer_terms" columns.}
\item{.eventDate_empty}{A flag generated by \code{\link[bdc:bdc_eventDate_empty]{bdc::bdc_eventDate_empty()}} where FALSE == occurrences with no eventDate provided.}
\item{.year_outOfRange}{A flag generated by \code{\link[bdc:bdc_year_outOfRange]{bdc::bdc_year_outOfRange()}} where FALSE == occurrences older than a threshold date. In this case 1950.}
\item{.duplicates}{A flag generated by \code{\link[=dupeSummary]{dupeSummary()}} where FALSE == occurrences identified as duplicates. There will be an associated kept duplicate (.duplictes == TRUE) for all duplicate clusters.}
}
}
\usage{
data("beesFlagged", package = "BeeBDC")
}
\description{
A small bee occurrence dataset with flags generated by BeeBDC used to run example script and test
functions. For data types, see \code{\link[=ColTypeR]{ColTypeR()}}.
}
\examples{

beesFlagged <- BeeBDC::beesFlagged
head(beesFlagged)

}
\references{
This data set was created by generating a random subset of 100 rows from the full BeeBDC dataset from the publication:
Dorey, J.B., Fischer, E.E., Chesshire, P.R., Nava-Bolaños, A., O’Reilly, R.L., Bossert, S., Collins, S.M., Lichtenberg, E.M., Tucker, E., Smith-Pardo, A., Falcon-Brindis, A., Guevara, D.A., Ribeiro, B.R., de Pedro, D., Hung, J.K.-L., Parys, K.A., McCabe, L.M., Rogan, M.S., Minckley, R.L., Velzco, S.J.E., Griswold, T., Zarrillo, T.A., Jetz, W., Sica, Y.V., Orr, M.C., Guzman, L.M., Ascher, J., Hughes, A.C. & Cobb, N.S. (2023) A globally synthesised and flagged bee occurrence dataset and cleaning workflow. Scientific Data, 10, 1–17. https://www.doi.org/10.1038/S41597-023-02626-W
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/summaryFun.R
\name{summaryFun}
\alias{summaryFun}
\title{Create or update the .summary flag column}
\usage{
summaryFun(
  data = NULL,
  dontFilterThese = NULL,
  onlyFilterThese = NULL,
  removeFilterColumns = FALSE,
  filterClean = FALSE
)
}
\arguments{
\item{data}{A data frame or tibble. Occurrence records to use as input.}

\item{dontFilterThese}{A character vector of flag columns to be ignored in the creation or updating
of the .summary column. Cannot be specified with onlyFilterThese.}

\item{onlyFilterThese}{A character vector. The inverse of dontFilterThese, where columns identified
here will be filtered and no others. Cannot be specified with dontFilterThese.}

\item{removeFilterColumns}{Logical. If TRUE all columns starting with "." will be removed in the
output data. This only makes sense to use when filterClean = TRUE. Default = FALSE.}

\item{filterClean}{Logical. If TRUE, the data will be filtered to only those occurrence where .summary
= TRUE (i.e., completely clean according to the used flag columns). Default = FALSE.}
}
\value{
Returns a data frame or tibble of the input data but modified based on the above parameters.
}
\description{
Using all flag columns (column names starting with "."), this function either creates or updates
the .summary flag column which is FALSE when ANY of the flag columns are FALSE. Columns can be excluded
and removed after creating the .summary column. Additionally, the occurrence dataset
can be filtered to only those where .summary = TRUE at the end of the function.
}
\examples{
# Read in example data
data(beesFlagged)

# To only update the .summary column
beesFlagged_out <- summaryFun(
    data = beesFlagged,
    dontFilterThese = c(".gridSummary", ".lonFlag", ".latFlag", ".uncer_terms", ".unLicensed"),
    removeFilterColumns = FALSE,
    filterClean = FALSE)
  # View output
table(beesFlagged_out$.summary, useNA = "always")

# Now filter to only the clean data and remove the flag columns
beesFlagged_out <- summaryFun(
  data = BeeBDC::beesFlagged,
  dontFilterThese = c(".gridSummary", ".lonFlag", ".latFlag", ".uncer_terms", ".unLicensed"),
  removeFilterColumns = TRUE,
  filterClean = TRUE)
# View output
table(beesFlagged_out$.summary, useNA = "always")



}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/jbd_coordinates_transposed.R
\name{jbd_coordinates_transposed}
\alias{jbd_coordinates_transposed}
\title{Identify transposed geographic coordinates}
\usage{
jbd_coordinates_transposed(
  data,
  idcol = "database_id",
  sci_names = "scientificName",
  lat = "decimalLatitude",
  lon = "decimalLongitude",
  country = "country",
  countryCode = "countryCode",
  border_buffer = 0.2,
  save_outputs = FALSE,
  fileName = NULL,
  scale = "large",
  path = NULL,
  mc.cores = 1
)
}
\arguments{
\item{data}{A data frame or tibble. Containing a unique identifier for each record,
geographical coordinates, and country names. Coordinates must be expressed
in decimal degrees and WGS84.}

\item{idcol}{A character string. The column name with a unique record identifier.
Default = "database_id".}

\item{sci_names}{A character string. The column name with species' scientific
names. Default = "scientificName".}

\item{lat}{A character string. The column name with latitudes. Coordinates must
be expressed in decimal degrees and WGS84. Default = "decimalLatitude".}

\item{lon}{A character string. The column name with longitudes. Coordinates must be
expressed in decimal degrees and WGS84. Default = "decimalLongitude".}

\item{country}{A character string. The column name with the country
assignment of each occurrence record. Default = "country".}

\item{countryCode}{A character string. The column name containing an ISO-2 country code for
each record.}

\item{border_buffer}{Numeric. Must have value greater than or equal to 0.
A distance in decimal degrees used to
created a buffer around each country. Records within a given country and at
a specified distance from the border will be not be corrected.
Default = 0.2 (~22 km at the equator).}

\item{save_outputs}{Logical. Indicates if a table containing transposed coordinates should be
saved for further inspection. Default = FALSE.}

\item{fileName}{A character string. The out file's name.}

\item{scale}{Passed to rnaturalearth's ne_countries().
Scale of map to return, one of 110, 50, 10 or 'small', 'medium', 'large'. Default = "large".}

\item{path}{A character string. A path as a character vector for where to create the directories
and save the figures. If
no path is provided (the default), the directories will be created using \code{\link[here:here]{here::here()}}.}

\item{mc.cores}{Numeric. If > 1, the jbd_correct_coordinates function will run in parallel
using mclapply using the number of cores specified. If = 1 then it will be run using a serial
loop. NOTE: Windows machines must use a value of 1 (see ?parallel::mclapply). Additionally,
be aware that each thread can use large chunks of memory.
Default = 1.#'}
}
\value{
A tibble containing the column "coordinates_transposed" which indicates if
verbatim coordinates were not transposed (TRUE). Otherwise
records are flagged as (FALSE) and, in this case, verbatim coordinates are
replaced by corrected coordinates.
}
\description{
This function flags and corrects records when latitude and longitude appear
to be transposed.
This function will preferably use the countryCode column generated by
\code{\link[bdc:bdc_country_standardized]{bdc::bdc_country_standardized()}}.
}
\details{
This test identifies transposed coordinates based on mismatches between the
country provided for a record and the record's latitude and longitude coordinates. Transposed
coordinates often fall outside of the indicated country (i.e., in other
countries or in the sea). Different coordinate transformations are
performed to correct country/coordinates mismatches. Importantly, verbatim
coordinates are replaced by the corrected ones in the returned database. A
database containing verbatim and corrected coordinates is created in
"Output/Check/01_coordinates_transposed.csv" if save_outputs == TRUE. The
columns "country" and "countryCode" can be retrieved by using the function
\link[bdc:bdc_country_standardized]{bdc::bdc_country_standardized}.
}
\examples{
\donttest{
if(require("rnaturalearthdata") && require("bdc")){
database_id <- c(1, 2, 3, 4)
scientificName <- c(
  "Rhinella major", "Scinax ruber",
  "Siparuna guianensis", "Psychotria vellosiana"
)
decimalLatitude <- c(63.43333, -14.43333, -41.90000, -46.69778)
decimalLongitude <- c(-17.90000, -67.91667, -13.25000, -13.82444)
country <- c("BOLIVIA", "bolivia", "Brasil", "Brazil")

x <- data.frame(
  database_id, scientificName, decimalLatitude,
  decimalLongitude, country
)

# Get country codes
x <- bdc::bdc_country_standardized(data = x, country = "country")

jbd_coordinates_transposed(
  data = x,
  idcol = "database_id",
  sci_names = "scientificName",
  lat = "decimalLatitude",
  lon = "decimalLongitude",
  country = "country_suggested",
  countryCode = "countryCode",
  border_buffer = 0.2,
  save_outputs = FALSE,
  scale = "medium"
) 
} # END if require
} # RND donttest

}
\concept{prefilter}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/richnessEstimateR.R
\name{richnessEstimateR}
\alias{richnessEstimateR}
\title{Estimate country, continental, and global species richnesses}
\usage{
richnessEstimateR(
  data = NULL,
  sampleSize = 10000,
  countrySamples = 1,
  continentSamples = 1,
  globalSamples = 1,
  countriesToExclude = NULL,
  mc.cores = 1,
  k = 10,
  filterToRecordedCountries = TRUE,
  outPath = tempdir(),
  fileName = "continentSampled.pdf"
)
}
\arguments{
\item{data}{an RData file created using the \code{\link[=richnessPrepR]{richnessPrepR()}} function.}

\item{sampleSize}{Numeric. The size of the sample randomly drawn from the provided curve.
See \code{curveFunction}. Default = 10000.}

\item{countrySamples}{Numeric. The number of times to sample the country species richness for both
iChao and iNEXT. If equal to zero (0), then this will not be analysed. Default = 5.}

\item{continentSamples}{Numeric. The number of times to sample the continent species richness for both
iChao and iNEXT. If equal to zero (0), then this will not be analysed. Default = 5.}

\item{globalSamples}{Numeric. The number of times to sample the global species richness for both
iChao and iNEXT. If equal to zero (0), then this will not be analysed. Default = 5.}

\item{countriesToExclude}{Character vector. You may decide to excluse some countries if they are
being problematic or their sample sizes are too small. Default = NULL.}

\item{mc.cores}{Numeric. If > 1, the function will run in parallel
using mclapply using the number of cores specified. If = 1 then it will be run using a serial
loop. NOTE: Windows machines must use a value of 1 (see ?parallel::mclapply). Additionally,
be aware that each thread can use large chunks of memory.
Default = 1.}

\item{k}{Numeric. For iChao; the cut-off point (default = 10), which separates species into "abundant" and "rare"
groups for abundance data for the estimator ACE; it separates species into "frequent" and
"infrequent" groups for incidence data for the estimator ICE.
Default = 10.}

\item{filterToRecordedCountries}{Logical. If TRUE, the checklist will be filtered to the countries
Where occurrence records were found. Default = TRUE. Change at your own peril.}

\item{outPath}{A directory as character. Directory where to save output figure. Default = tempdir().}

\item{fileName}{A character vector with file name
for the output figure, ending with '.pdf'. Default = "continentSampled.pdf".}
}
\value{
Outputs an R file with four tables ("Summary", "SiteOutput", "ContinentOutput", and
"GlobalOutput"; depending on the number required). The summary table shows the Median overall
estimates, while the remaining three shows the outputs from each iteration (useful for plotting,
see relevant vignette). Some figures may also be saved to the selected outPath.
}
\description{
Takes an output dataset from \code{\link[=richnessPrepR]{richnessPrepR()}} to estimate species richness using
iChao and iNEXT (hill numbers) for countries, continents, and/or the entire globe.
}
\examples{
\dontrun{

  # Use the example data 
data(beesCountrySubset)

  # First, 
estimateDataExample <- BeeBDC::richnessPrepR(
  data = beesCountrySubset,
  # Download the taxonomy
  taxonomyFile = BeeBDC::beesTaxonomy(),
  # Download the checklist
  checklistFile = BeeBDC::beesChecklist(),
  curveFunction = function(x) (228.7531 * x * x^-log(12.1593)),
  sampleSize = 10000,
  countryColumn = "country_suggested",
  limitGlobal = NULL,
  outPath = tempdir()
)

 exampleEstimate <- richnessEstimateR(
   data = estimateDataExample,
   sampleSize = 10000,
   countrySamples = 1,
   continentSamples = 1,
   globalSamples = 1,
   filterToRecordedCountries = TRUE,
   mc.cores = 1,
   # Directory where to save files
   outPath = tempdir(),
   fileName = "Sampled.pdf"
 )
 
} # END dontrun

}
\seealso{
\code{\link[=countryHarmoniseR]{countryHarmoniseR()}} to harmonise country names based on a short list;
\code{\link[=richnessPrepR]{richnessPrepR()}} to produce the input data and for the required column names; as well
as \code{\link[=ChaoWrapper]{ChaoWrapper()}} and \code{\link[=iNEXTwrapper]{iNEXTwrapper()}} for the parallelised implementation of
\code{SpadeR} and \code{iNEXT} functions.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/diagonAlley.R
\name{diagonAlley}
\alias{diagonAlley}
\title{Find fill-down errors}
\usage{
diagonAlley(
  data = NULL,
  minRepeats = NULL,
  groupingColumns = c("eventDate", "recordedBy", "datasetName"),
  ndec = 3,
  stepSize = 1e+06,
  mc.cores = 1
)
}
\arguments{
\item{data}{A data frame or tibble. Occurrence records as input.}

\item{minRepeats}{Numeric. The minimum number of lat or lon repeats needed to flag a record}

\item{groupingColumns}{Character. The column(s) to group the analysis by and search for fill-down
errors within. Default = c("eventDate", "recordedBy", "datasetName").}

\item{ndec}{Numeric. The number of decimal places below which records will not be considered
in the diagonAlley function. This is fed into \code{\link[=jbd_coordinates_precision]{jbd_coordinates_precision()}}. Default = 3.}

\item{stepSize}{Numeric. The number of occurrences to process in each chunk. Default = 1000000.}

\item{mc.cores}{Numeric. If > 1, the function will run in parallel
using mclapply using the number of cores specified. If = 1 then it will be run using a serial
loop. NOTE: Windows machines must use a value of 1 (see ?parallel::mclapply). Additionally,
be aware that each thread can use large chunks of memory.
Default = 1.}
}
\value{
The function returns the input data with a new column, .sequential, where FALSE =
records that have consecutive latitudes or longitudes greater than or equal to the user-defined
threshold.
}
\description{
A simple function that looks for potential latitude and longitude fill-down errors by
identifying consecutive occurrences with coordinates at regular intervals. This is accomplished
by using a sliding window with the length determined by minRepeats.
}
\details{
The sliding window (and hence fill-down errors) will only be examined
within the user-defined groupingColumns; if any of those
columns are empty, that record will be excluded.
}
\examples{
# Read in the example data
  data(beesRaw)
 # Run the function
  beesRaw_out <- diagonAlley(
    data = beesRaw,
    # The minimum number of repeats needed to find a sequence in for flagging
    minRepeats = 4,
    groupingColumns = c("eventDate", "recordedBy", "datasetName"),
    ndec = 3,
    stepSize = 1000000,
    mc.cores = 1)
  

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/dirMaker.R
\name{dirMaker}
\alias{dirMaker}
\title{Set up global directory paths and create folders}
\usage{
dirMaker(
  RootPath = RootPath,
  ScriptPath = NULL,
  DataPath = NULL,
  DataSubPath = "/Data_acquisition_workflow",
  DiscLifePath = NULL,
  OutPath = NULL,
  OutPathName = "Output",
  Report = TRUE,
  Check = TRUE,
  Figures = TRUE,
  Intermediate = TRUE,
  RDoc = NULL,
  useHere = TRUE
)
}
\arguments{
\item{RootPath}{A character String. The \code{RootPath} is the base path for your project, and all
other paths should ideally be located within the \code{RootPath}. However, users may specify paths not
contained in the RootPath}

\item{ScriptPath}{A character String. The \code{ScriptPath} is the path to any additional functions
that you would like to read in for use with BeeBDC.}

\item{DataPath}{A character string. The path to the folder containing bee occurrence data
to be flagged and/or cleaned}

\item{DataSubPath}{A character String. If a \code{DataPath} is not provided, this will be used as the \code{DataPath}
folder name within the \code{RootPath.} Default is "/Data_acquisition_workflow"}

\item{DiscLifePath}{A character String. The path to the folder which contains data from Ascher
and Pcikering's Discover Life website.}

\item{OutPath}{A character String. The path to the folder where output data will be saved.}

\item{OutPathName}{A character String. The name of the \code{OutPath} subfolder located within the
\code{RootPath.} Default is "Output".}

\item{Report}{Logical. If TRUE, function creates a "Report" folder within the OutPath-defined
folder. Default = TRUE.}

\item{Check}{Logical. If TRUE, function creates a "Check" folder within the OutPath-defined
folder. Default = TRUE.}

\item{Figures}{Logical. If TRUE, function creates a "Figures" folder within the OutPath-defined
folder. Default = TRUE.}

\item{Intermediate}{Logical. If TRUE, function creates a "Intermediate" folder within the
OutPath-defined folder in which to save intermediate datasets. Default = TRUE.}

\item{RDoc}{A character String. The path to the current script or report, relative to the project
root. Passing an absolute path raises an error. This argument is used by \code{\link[here:i_am]{here::i_am()}} and incorrectly
setting this may result in \code{bdc} figures being saved to your computer's root directory}

\item{useHere}{Logical. If TRUE, dirMaker will use \code{\link[here:i_am]{here::i_am()}} to declare the relative path
to 'RDoc'. This is aimed at preserving some functionality with where bdc saves summary figures
and tables. Default = TRUE.}
}
\value{
Results in the generation of a list containing the BeeBDC-required directories in your global
environment. This function should be run at the start of each session. Additionally, this
function will create the BeeBDC-required folders if they do not already exist in the supplied
directory
}
\description{
This function sets up a directory for saving outputs (i.e. data, figures) generated through the
use of the BeeBDC package, if the required folders do not already exist.
}
\examples{
  # load dplyr
  library(dplyr)
# Standard/basic usage:
RootPath <- tempdir()
dirMaker(
RootPath = RootPath,
# Input the location of the workflow script RELATIVE to the RootPath
RDoc = NULL,
useHere = FALSE) \%>\%
  # Add paths created by this function to the environment()
  list2env(envir = environment())  

# Custom OutPathName provided
  dirMaker(
 RootPath = RootPath,
 # Set some custom OutPath info
 OutPath = NULL,
 OutPathName = "T2T_Output",
 # Input the location of the workflow script RELATIVE to the RootPath
 RDoc = NULL,
 useHere = FALSE) \%>\%
   # Add paths created by this function to the environment()
   list2env(envir = environment())  
 # Set the working directory

# Further customisations are also possible
dirMaker(
  RootPath = RootPath,
  ScriptPath = "...path/Bee_SDM_paper/BDC_repo/BeeBDC/R",
  DiscLifePath = "...path/BDC_repo/DiscoverLife_Data",
  OutPathName = "AsianPerspective_Output",
  # Input the location of the workflow script RELATIVE to the RootPath
  RDoc = NULL,
  useHere = FALSE) \%>\%
  # Add paths created by this function to the environment()
  list2env(envir = environment())  



}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/USGS_formatter.R
\name{USGS_formatter}
\alias{USGS_formatter}
\title{Find, import, and format USGS data to Darwin Core}
\usage{
USGS_formatter(path, pubDate)
}
\arguments{
\item{path}{A character path to a directory that contains the USGS data, which will be found using
\code{\link[=fileFinder]{fileFinder()}}. The function will look for "USGS_DRO_flat".}

\item{pubDate}{Character. The publication date of the dataset to update the metadata and citation.}
}
\value{
Returns a list with the occurrence data, "USGS_data", and the EML data, "EML_attributes".
}
\description{
The function finds, imports, formats, and creates metadata for the USGS dataset.
}
\examples{
\dontrun{
USGS_data <- USGS_formatter(path = DataPath, pubDate = "19-11-2022")
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/countryNameCleanR.R
\name{countryNameCleanR}
\alias{countryNameCleanR}
\title{Fix country name issues using a user-input list}
\usage{
countryNameCleanR(data = NULL, ISO2_table = NULL, commonProblems = NULL)
}
\arguments{
\item{data}{A data frame or tibble. Occurrence records as input.}

\item{ISO2_table}{A data frame or tibble with the columns ISO2 and long names for country names. Default
is a static version from Wikipedia.}

\item{commonProblems}{A data frame or tibble. It must have two columns:
one containing the user-identified problem and one with a user-defined fix}
}
\value{
Returns the input data, but with countries occurring in the user-supplied problem
column ("commonProblems") replaced with those in the user-supplied fix column
}
\description{
This function is basic for a user to manually fix some country name inconsistencies.
}
\examples{
beesFlagged_out <- countryNameCleanR(
data = BeeBDC::beesFlagged,
commonProblems = dplyr::tibble(problem = c('U.S.A.', 'US','USA','usa','UNITED STATES',
                        'United States','U.S.A','MX','CA','Bras.','Braz.',
                        'Brasil','CNMI','USA TERRITORY: PUERTO RICO'),
                        fix = c('United States of America','United States of America',
                                'United States of America','United States of America',
                                'United States of America','United States of America',
                                'United States of America','Mexico','Canada','Brazil',
                                'Brazil','Brazil','Northern Mariana Islands','PUERTO.RICO')))
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/beesChecklist.R
\name{beesChecklist}
\alias{beesChecklist}
\title{Download a country-level checklist of bees from Discover Life}
\usage{
beesChecklist(
  URL = "https://api.figshare.com/v2/file/download/60945823",
  mode = NULL,
  headers = NULL,
  token = NULL,
  alternateURL = FALSE,
  ...
)
}
\arguments{
\item{URL}{A character vector to the FigShare location of the dataset. The default will be to
the most-recent version.}

\item{mode}{A character passed on to \code{utils::download.file()}. Default = "wb" for Windows or "w" for Mac/Linux.}

\item{headers}{Character. Passed on to  \code{utils::download.file()}. Default = NULL.}

\item{token}{Character. Optional personal access token from FigShare account for authorisation.
Default = NULL.}

\item{alternateURL}{Logical. If TRUE then the function will use the an alernate version of the
download URL. Might be worth trying if it is failing. Default = FALSE.}

\item{...}{Extra variables that can be passed to \code{downloader::download()}.}
}
\value{
A downloaded beesChecklist.Rda file in the outPath and the same tibble returned to
the environment.

**Column details **

\strong{validName} The valid scientificName as it should occur in the scientificName column.

\strong{DiscoverLife_name} The full country name as it occurs on Discover Life.

\strong{rNaturalEarth_name} Country name from rnaturalearth's name_long.

\strong{shortName} A short version of the country name.

\strong{DiscoverLife_ISO} The ISO country name as it occurs on Discover Life.

\strong{Alpha-2} Alpha-2 from rnaturalearth.

\strong{Alpha-3} Alpha-3 from rnaturalearth.

\strong{official} Official country name = "yes" or only a Discover Life name = "no".

\strong{Source} A text strign denoting the source or author of the name-country pair.

\strong{matchCertainty} Quality of the name's match to the Discover Life checklist.

\strong{canonical} The valid species name without scientificNameAuthority.

\strong{canonical_withFlags} The validName without the scientificNameAuthority but with Discover Life flags.

\strong{family} Bee family.

\strong{subfamily} Bee subfamily.

\strong{genus} Bee genus.

\strong{subgenus} Bee subgenus.

\strong{infraspecies} Bee infraSpecificEpithet.

\strong{species} Bee specificEpithet.

\strong{scientificNameAuthorship} Bee scientificNameAuthorship.

\strong{taxon_rank} Rank of the taxon name.

\strong{Notes} Discover Life country name notes.

\strong{Previous checklists:}
\itemize{
\item 2026-01-12 \strong{current}: https://open.flinders.edu.au/ndownloader/files/60945823
\item 2024-06-17: https://open.flinders.edu.au/ndownloader/files/47092720
\item Original: https://open.flinders.edu.au/ndownloader/files/42320598
}
}
\description{
Download the table contains taxonomic and country information for the bees of the world based
on data collated on Discover Life. The data will be sourced from the BeeBDC article's
Figshare.

Note that sometimes the download might not work without restarting R. In this case, you could
alternatively download the dataset from the URL below and then read it in using
\code{base::readRDS("filePath.Rda")}. Note that as of version 1.3.2, this function internally uses
a modified version of the "download" function from the
\code{downloader} package on CRAN. Additionally, BeeBDC will try a different
download method with each failed attempt (for Windows: "auto" > "internal" > "libcurl" > "wget" > "curl";
or "auto" if any particular method is not available).

See \code{\link[=beesTaxonomy]{beesTaxonomy()}} for further context.
}
\examples{
\dontrun{
beesChecklist <- BeeBDC::beesChecklist()
}
}
\references{
This dataset was created using the Discover Life checklist and taxonomy.
Dataset is from the publication:
DOREY, J. B., CHESSHIRE, P. R., BOLAÑOS, A. N., O’REILLY, R. L., BOSSERT, S., COLLINS, S. M., LICHTENBERG, E. M., TUCKER, E., SMITH-PARDO, A., FALCON-BRINDIS, A., GUEVARA, D. A., RIBEIRO, B. R., DE PEDRO, D., FISCHER, E., HUNG, J. K.-L., PARYS, K. A., ROGAN, M. S., MINCKLEY, R. L., VELZCO, S. J. E., GRISWOLD, T., ZARRILLO, T. A., SICA, Y., ORR, M. C., GUZMAN, L. M., ASCHER, J., HUGHES, A. C. & COBB, N. S. In review. A globally synthesised and flagged bee occurrence dataset and cleaning workflow. Scientific Data.
The checklist data are mostly compiled from Discover Life data, www.discoverlife.org:
ASCHER, J. S. & PICKERING, J. 2020. Discover Life bee species guide and world checklist (Hymenoptera: Apoidea: Anthophila). http://www.discoverlife.org/mp/20q?guide=Apoidea_species.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/aux_functions.R
\name{fileFinder}
\alias{fileFinder}
\title{Finds files within a directory}
\usage{
fileFinder(path, fileName)
}
\arguments{
\item{path}{A directory as character. The directory to recursively search.}

\item{fileName}{A character/regex string. The file name to find.}
}
\value{
Returns a directory to the most-recent file that matches the provide file. Using regex
can greatly improve specificity.
The function will also write into the console the file that it has found - it is worthwhile to
check that this is the correct file to avoid complications down the line
}
\description{
A function which can be used to find files within a user-defined directory based on a
user-provided character string.
}
\examples{
\donttest{
# load dplyr
library(dplyr)

 # Make the RootPath to the tempdir for this example
  RootPath <- tempdir()
  
 # Load the example data
 data("beesRaw", package = "BeeBDC")

# Save and example dataset to the temp dir
  readr::write_csv(beesRaw, file = paste0(RootPath, "/beesRaw.csv"))

 # Now go find it!
fileFinder(path = RootPath, fileName = "beesRaw")
# more specifically the .csv version
fileFinder(path = RootPath, fileName = "beesRaw.csv")
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ColTypeR.R
\name{ColTypeR}
\alias{ColTypeR}
\title{Sets up column names and types}
\usage{
ColTypeR(..., standardFormat = NULL)
}
\arguments{
\item{...}{Additional arguments. These can be specified in addition to the ones default to the
function.  For example:
\itemize{
\item newCharacterColumn = \code{\link[readr:col_character]{readr::col_character()}},
\item newNumericColumn = \code{\link[readr:col_integer]{readr::col_integer()}},
\item newLogicalColumn = \code{\link[readr:col_logical]{readr::col_logical()}}
}}

\item{standardFormat}{Character. Some taxa may have a standard format for data. Presently,
Only bees have had this encoded here as "bee". Default = NULL.}
}
\value{
Returns an object of class col_spec.
See \code{\link[readr:as.col_spec]{readr::as.col_spec()}} for additional context and explication.
}
\description{
This function uses \code{\link[readr:cols_only]{readr::cols_only()}} to assign a column name and the type of data
(e.g., \code{\link[readr:col_character]{readr::col_character()}},
and \code{\link[readr:col_integer]{readr::col_integer()}}). To see the default columns simply run \code{\link[=ColTypeR]{ColTypeR()}}.
This is intended for use with \code{\link[readr:read_csv]{readr::read_csv()}}. Columns that are not present will NOT be included
in the resulting tibble unless they are specified using \link{...}.
}
\examples{
  # You can simply return the below for default values
  library(dplyr)
BeeBDC::ColTypeR() 

  # To add new columns you can write
ColTypeR(newCharacterColumn = readr::col_character(), 
         newNumericColumn = readr::col_integer(), 
         newLogicalColumn = readr::col_logical()) 

# Try reading in one of the test datasets as an example:
beesFlagged \%>\% dplyr::as_tibble(col_types = BeeBDC::ColTypeR())
  # OR
beesRaw \%>\% dplyr::as_tibble(col_types = BeeBDC::ColTypeR())

  # OR, using the bee standard format from:
  
beesRaw \%>\% dplyr::as_tibble(col_types = BeeBDC::ColTypeR(standardFormat = "bee"))


}
\references{
For using the bee standard —  Clos, B. D., Seltmann, K. C.,
Turley, N. E., Maffei, C., Tucker, E. M., Lane, I. G., Levenson, H. K., & Woodard, H. S.
(2025). Improving the standardization of wild bee occurrence data: towards a formal
wild bee data standard. Authorea. doi: https://doi.org/10.22541/au.173862402.22787949/v2
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/countryHarmoniseR.R
\name{countryHarmoniseR}
\alias{countryHarmoniseR}
\title{Match and harmonise country names}
\usage{
countryHarmoniseR(
  data = NULL,
  countryColumn = NULL,
  shorterNames = TRUE,
  continentAnalysis = FALSE
)
}
\arguments{
\item{data}{A data frame or tibble.}

\item{countryColumn}{Character. The column that contains country names to be harmonised.}

\item{shorterNames}{Logical. If TRUE, some very long country names will be shortened. This can be
helpful for writing country names in plots where long names are annoying.}

\item{continentAnalysis}{Logical. Set to TRUE in order to match small entities to continents;
limited use for the moment.}
}
\value{
Returns the original data frame or tibble but with harmonised country names
}
\description{
A small function that is useful for harmonising country names so that they mathc between datasets.
This relies on the country names matching one of the exceptions pre-coded in and is so far used
internally.
}
\examples{
# load in the test dataset
system.file("extdata", "testTaxonomy.rda", package="BeeBDC") |> load()
harmonisedCountries <- countryHarmoniseR(
  data = testTaxonomy,
  countryColumn = "country_suggested"
)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/flagAbsent.R
\name{flagAbsent}
\alias{flagAbsent}
\title{Flags occurrences that are marked as absent}
\usage{
flagAbsent(data = NULL, PresAbs = "occurrenceStatus")
}
\arguments{
\item{data}{A data frame or tibble. Occurrence records as input.}

\item{PresAbs}{Character. The column in which the function will find "ABSENT" or "PRESENT" records.
Default = "occurrenceStatus"}
}
\value{
The input data with a new column called ".occurrenceAbsent" where FALSE == "ABSENT" records.
}
\description{
Flags occurrences that are "ABSENT" for the occurrenceStatus (or some other user-specified) column.
}
\examples{
  # Bring in the data
data(beesRaw)
  # Run the function
beesRaw_out <- flagAbsent(data = beesRaw,
PresAbs = "occurrenceStatus")
  # See the result
table(beesRaw_out$.occurrenceAbsent, useNA = "always")
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/beesCountrySubset.R
\docType{data}
\name{beesCountrySubset}
\alias{beesCountrySubset}
\title{A simple scientificName and country_suggested test dataset}
\format{
An object of class \code{"tibble"}
\describe{
\item{scientificName}{Full scientificName as shown on DiscoverLife}
\item{country_suggested}{A country name suggested by the \code{\link[bdc:bdc_country_standardized]{bdc::bdc_country_standardized()}} function.}
}
}
\usage{
data("beesCountrySubset", package = "BeeBDC")
}
\description{
A small bee occurrence dataset with 1,488 scientificName-country_suggested combinations across
four countries; Fiji, Uganda, Vietnam, and Zambia.
}
\examples{

beesCountrySubset <- BeeBDC::beesCountrySubset
head(beesCountrySubset)

}
\references{
This data set was created by during the writing of the paper:
Dorey, J.B., (upcoming) How many bee species are there? A quantitative global estimate. Journal
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/jbd_coordinates_precision.R
\name{jbd_coordinates_precision}
\alias{jbd_coordinates_precision}
\title{Flags coordinates for imprecision}
\usage{
jbd_coordinates_precision(
  data,
  lat = "decimalLatitude",
  lon = "decimalLongitude",
  ndec = NULL,
  quieter = FALSE
)
}
\arguments{
\item{data}{A data frame or tibble. Occurrence records as input.}

\item{lat}{Character. The name of the column to use as latitude. Default = "decimalLatitude".}

\item{lon}{Character. The name of the column to use as longitude. Default = "decimalLongitude".}

\item{ndec}{Numeric. The number of decimal places to flag in decimal degrees. For example,
argument value of 2 would flag occurrences with nothing in the hundredths place (0.0x).}

\item{quieter}{Logical. If TRUE, the functino will run a little quieter. Default = FALSE.}
}
\value{
Returns the input data frame with a new column, .rou, where FALSE indicates occurrences
that failed the test.
}
\description{
This function flags occurrences where BOTH latitude and longitude values are rounded. This
contrasts with the original function, bdc::bdc_coordinates_precision() that will flag
occurrences where only one of latitude OR longitude are rounded. The BeeBDC approach saves
occurrences that may have had terminal zeros rounded in one coordinate column.
}
\examples{
beesRaw_out <- jbd_coordinates_precision(
  data = BeeBDC::beesRaw,
  lon = "decimalLongitude",
  lat = "decimalLatitude",
    # number of decimals to be tested
  ndec = 2
)
table(beesRaw_out$.rou, useNA = "always")
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/jbd_Ctrans_chunker.R
\name{jbd_Ctrans_chunker}
\alias{jbd_Ctrans_chunker}
\title{Wraps jbd_coordinates_transposed to identify  and fix transposed occurrences}
\usage{
jbd_Ctrans_chunker(
  data = NULL,
  lat = "decimalLatitude",
  lon = "decimalLongitude",
  idcol = "databse_id",
  country = "country_suggested",
  countryCode = "countryCode",
  sci_names = "scientificName",
  border_buffer = 0.2,
  save_outputs = TRUE,
  stepSize = 1e+06,
  chunkStart = 1,
  progressiveSave = TRUE,
  path = tempdir(),
  append = TRUE,
  scale = "large",
  mc.cores = 1
)
}
\arguments{
\item{data}{A data frame or tibble. Occurrence records as input.}

\item{lat}{Character. The column with latitude in decimal degrees. Default = "decimalLatitude".}

\item{lon}{Character. The column with longitude in decimal degrees. Default = "decimalLongitude".}

\item{idcol}{Character. The column name with a unique record identifier. Default = "database_id".}

\item{country}{Character. The name of the column containing country names. Default = "country".}

\item{countryCode}{Character. Identifies the column containing ISO-2 country codes
Default = "countryCode".}

\item{sci_names}{Character. The column containing scientific names. Default = "scientificName".}

\item{border_buffer}{Numeric. The buffer, in decimal degrees, around points to help match them
to countries. Default = 0.2 (~22 km at equator).}

\item{save_outputs}{Logical. If TRUE, transposed occurrences will be saved to their own file.}

\item{stepSize}{Numeric. The number of occurrences to process in each chunk. Default = 1000000.}

\item{chunkStart}{Numeric. The chunk number to start from. This can be > 1 when you need to restart
the function from a certain chunk; for example if R failed unexpectedly.}

\item{progressiveSave}{Logical. If TRUE then the country output list will be saved between
each iteration so that \code{append} can be used if the function is stopped part way through.}

\item{path}{Character. The path to a file in which to save the 01_coordinates_transposed_
output.}

\item{append}{Logical. If TRUE, the function will look to append an existing file.}

\item{scale}{Passed to rnaturalearth's ne_countries().
Scale of map to return, one of 110, 50, 10 or 'small', 'medium', 'large'. Default = "large".}

\item{mc.cores}{Numeric. If > 1, the jbd_correct_coordinates function will run in parallel
using mclapply using the number of cores specified. If = 1 then it will be run using a serial
loop. NOTE: Windows machines must use a value of 1 (see ?parallel::mclapply). Additionally,
be aware that each thread can use large chunks of memory.
Default = 1.#'}
}
\value{
Returns the input data frame with a new column, coordinates_transposed, where FALSE = columns
that had coordinates transposed.
}
\description{
Because the \code{\link[=jbd_coordinates_transposed]{jbd_coordinates_transposed()}} function is very RAM-intensive, this wrapper
allows a user to specify chunk-sizes and only analyse a small portion of the occurrence data at a
time. The prefix jbd_ is used to highlight the difference between this function and the original
\code{\link[bdc:bdc_coordinates_transposed]{bdc::bdc_coordinates_transposed()}}.
This function will preferably use the countryCode column generated by
\code{\link[bdc:bdc_country_standardized]{bdc::bdc_country_standardized()}}.
}
\examples{
if(requireNamespace("rnaturalearthdata")){
library(dplyr)
  # Import and prepare the data
data(beesFlagged)
beesFlagged <- beesFlagged \%>\% dplyr::select(!c(.val, .sea)) \%>\%
  # Cut down the dataset to un example quicker
dplyr::filter(dplyr::row_number() \%in\% 1:20)
  # Run the function
beesFlagged_out <- jbd_Ctrans_chunker(
# bdc_coordinates_transposed inputs
data = beesFlagged,
idcol = "database_id",
lat = "decimalLatitude",
lon = "decimalLongitude",
country = "country_suggested",
countryCode = "countryCode",
# in decimal degrees (~22 km at the equator)
border_buffer = 1, 
save_outputs = FALSE,
sci_names = "scientificName",
# chunker inputs
# How many rows to process at a time
stepSize = 1000000,  
# Start row
chunkStart = 1,  
# Progressively save the output between each iteration?
progressiveSave = FALSE,
path = tempdir(),
# If FALSE it may overwrite existing dataset
append = FALSE,
  # Users should select scale = "large" as it is more thoroughly tested
scale = "medium",
mc.cores = 1
) 
table(beesFlagged_out$coordinates_transposed, useNA = "always")
} # END if require

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/importOccurrences.R
\name{importOccurrences}
\alias{importOccurrences}
\title{Imports the most-recent repoMerge data}
\usage{
importOccurrences(path = path, fileName = "^BeeData_")
}
\arguments{
\item{path}{A directory as a character. The directory to recursively look in for the above data.}

\item{fileName}{Character. A String of text to look for the most-recent dataset.
Default = "^BeeData_". Find faults by modifying \code{\link[=fileFinder]{fileFinder()}}
and logic-checking the file that's found.}
}
\value{
A list with a data frame of merged occurrence records, "Data_WebDL", and a list of EML
files contained in "eml_files".
}
\description{
Looks for and imports the most-recent version of the occurrence data created by the \code{\link[=repoMerge]{repoMerge()}}
function.
}
\examples{
\dontrun{
DataImp <- importOccurrences(path = DataPath)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/additionalData_readRs.R
\name{readr_BeeBDC}
\alias{readr_BeeBDC}
\alias{readr_EPEL}
\alias{readr_ASP}
\alias{readr_BMin}
\alias{readr_BMont}
\alias{readr_Ecd}
\alias{readr_Gai}
\alias{readr_CAES}
\alias{readr_KP}
\alias{readr_EcoS}
\alias{readr_GeoL}
\alias{readr_EaCO}
\alias{readr_MABC}
\alias{readr_Col}
\alias{readr_FSCA}
\alias{readr_SMC}
\alias{readr_Bal}
\alias{readr_Lic}
\alias{readr_Arm}
\alias{readr_Dor}
\alias{readr_MEPB}
\alias{readr_BBD}
\alias{readr_MPUJ}
\alias{readr_STRI}
\alias{readr_PALA}
\alias{readr_JoLa}
\alias{readr_VicWam}
\title{A wrapper for all of the data readr_functions}
\usage{
readr_BeeBDC(
  dataset = NULL,
  path = NULL,
  inFile = NULL,
  outFile = NULL,
  dataLicense = NULL,
  sheet = NULL
)

readr_EPEL(path = NULL, inFile = NULL, outFile = NULL, dataLicense = NULL)

readr_ASP(path = NULL, inFile = NULL, outFile = NULL, dataLicense = NULL)

readr_BMin(path = NULL, inFile = NULL, outFile = NULL, dataLicense = NULL)

readr_BMont(path = NULL, inFile = NULL, outFile = NULL, dataLicense = NULL)

readr_Ecd(path = NULL, inFile = NULL, outFile = NULL, dataLicense = NULL)

readr_Gai(path = NULL, inFile = NULL, outFile = NULL, dataLicense = NULL)

readr_CAES(
  path = NULL,
  inFile = NULL,
  outFile = NULL,
  dataLicense = NULL,
  sheet = "Sheet1"
)

readr_KP(path = NULL, inFile = NULL, outFile = NULL, dataLicense = NULL)

readr_EcoS(path = NULL, inFile = NULL, outFile = NULL, dataLicense = NULL)

readr_GeoL(path = NULL, inFile = NULL, outFile = NULL, dataLicense = NULL)

readr_EaCO(path = NULL, inFile = NULL, outFile = NULL, dataLicense = NULL)

readr_MABC(
  path = NULL,
  inFile = NULL,
  outFile = NULL,
  dataLicense = NULL,
  sheet = "Hoja1"
)

readr_Col(
  path = NULL,
  inFile = NULL,
  outFile = NULL,
  dataLicense = NULL,
  sheet = sheet
)

readr_FSCA(path = NULL, inFile = NULL, outFile = NULL, dataLicense = NULL)

readr_SMC(path = NULL, inFile = NULL, outFile = NULL, dataLicense = NULL)

readr_Bal(
  path = NULL,
  inFile = NULL,
  outFile = NULL,
  dataLicense = NULL,
  sheet = "animal_data"
)

readr_Lic(path = NULL, inFile = NULL, outFile = NULL, dataLicense = NULL)

readr_Arm(
  path = NULL,
  inFile = NULL,
  outFile = NULL,
  dataLicense = NULL,
  sheet = "Sheet1"
)

readr_Dor(path = NULL, inFile = NULL, outFile = NULL, dataLicense = NULL)

readr_MEPB(
  path = NULL,
  inFile = NULL,
  outFile = NULL,
  dataLicense = NULL,
  sheet = NULL
)

readr_BBD(path = NULL, inFile = NULL, outFile = NULL, dataLicense = NULL)

readr_MPUJ(
  path = NULL,
  inFile = NULL,
  outFile = NULL,
  dataLicense = NULL,
  sheet = sheet
)

readr_STRI(path = NULL, inFile = NULL, outFile = NULL, dataLicense = NULL)

readr_PALA(path = NULL, inFile = NULL, outFile = NULL, dataLicense = NULL)

readr_JoLa(
  path = NULL,
  inFile = NULL,
  outFile = NULL,
  dataLicense = NULL,
  sheet = c("pre-1950", "post-1950")
)

readr_VicWam(
  path = NULL,
  inFile = NULL,
  outFile = NULL,
  dataLicense = NULL,
  sheet = "Combined"
)
}
\arguments{
\item{dataset}{Character. The name of the dataset to be read in. For example readr_CAES can
be called using "readr_CAES" or "CAES". This is not caps sensitive.}

\item{path}{A character path. The path to the directory containing the data.}

\item{inFile}{Character or character path. The name of the file itself (can also be the
remainder of a path including the file name).}

\item{outFile}{Character or character path. The name of the Darwin Core format file to be saved.}

\item{dataLicense}{Character. The license to accompany each record in the Darwin Core 'license'
column.}

\item{sheet}{A character String. For those datasets read from an .xlsx format, provide the
sheet name.
NOTE: This will be ignored for .csv readr_ functions and required for .xlsx readr_ functions.}
}
\value{
A data frame that is in Darwin Core format.
}
\description{
Read in a variety of data files that are specific to certain smaller data providers.
There is an internal readr function for each dataset and each one of these functions is called
by readr_BeeBDC. While these functions are internal, they are displayed in the documentation of
readr_BeeBDC for clarity.
}
\details{
This function wraps several internal readr functions. Users may call
readr_BeeBDC and select the dataset name to import a certain dataset. These datasets include:

Excel (.xlsx) formatted datasets: CAES, MABC, Col, Bal, MEPB, MUPJ, Arm, JoLa, and VicWam.

CSV (.csv) formatted datasets: EPEL, ASP, BMin, BMont, Ecd, Gai, KP, EcoS, GeoL, EaCo, FSCA, SMC,
Lic, Dor, BBD, STRI, and PALA

See Dorey et al. 2023 BeeBDC... for further details.
}
\examples{
\dontrun{
# An example using a .xlsx file
Arm_Data <- readr_BeeBDC(
    dataset = "Arm",
    path = paste0(tempdir(), "/Additional_Datasets"),
    inFile = "/InputDatasets/Bee database Armando_Final.xlsx",
    outFile = "jbd_Arm_Data.csv",
    sheet = "Sheet1",
    dataLicense = "https://creativecommons.org/licenses/by-nc-sa/4.0/")
    
    
    # An example using a .csv file
EPEL_Data <- readr_BeeBDC(
  dataset = "readr_EPEL",
  path = paste0(tempdir(), "/Additional_Datasets"),
  inFile = "/InputDatasets/bee_data_canada.csv",
  outFile = "jbd_EPEL_data.csv",
  dataLicense = "https://creativecommons.org/licenses/by-nc-sa/4.0/")
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/dataProvTables.R
\name{dataProvTables}
\alias{dataProvTables}
\title{Build a table of data providers for bee occurrence records}
\usage{
dataProvTables(
  data = NULL,
  runBeeDataChecks = FALSE,
  outPath = OutPath_Report,
  fileName = NULL
)
}
\arguments{
\item{data}{A data frame or tibble. Occurrence records as input.}

\item{runBeeDataChecks}{Logical. If TRUE, will search in other columns for specific clues to
determine the institution.}

\item{outPath}{A character path. The path to the directory in which the figure will be saved.
Default = OutPath_Report.}

\item{fileName}{Character. The name of the file to be saved, ending in ".csv".}
}
\value{
Returns a table with the data providers, an specimen count, and a species count.
}
\description{
This function will attempt to find and build a table of data providers that have contributed
to the input data, especially using the 'institutionCode' column. It will also look for a
variety of other columns to find data providers using an internally set sequence of if-else
statements. Hence, this function is quite specific for bee data, but should work for other
taxa in similar institutions.
}
\examples{

data(beesFlagged)

testOut <- dataProvTables(
data = beesFlagged,
runBeeDataChecks = TRUE,
outPath = tempdir(),
fileName = "testFile.csv")

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ChaoWrapper.R
\name{ChaoWrapper}
\alias{ChaoWrapper}
\title{Parallel estimation of species richness in a community using iChao}
\usage{
ChaoWrapper(
  data = NULL,
  datatype = "abundance",
  k = 10,
  conf = 0.95,
  mc.cores = 1
)
}
\arguments{
\item{data}{A data frame or tibble. A data frame containing "abundance"-type data per variable
(population, country, species...) in columns.}

\item{datatype}{Character. The type of input data, "abundance", "abundance_freq_count",
"incidence_freq", "incidence_freq_count" or "incidence_raw". So far only tested with "abundance"
data. Default = "abundace".}

\item{k}{Numeric. the cut-off point (default = 10), which separates species into "abundant" and "rare"
groups for abundance data for the estimator ACE; it separates species into "frequent" and
"infrequent" groups for incidence data for the estimator ICE.
Default = 10.}

\item{conf}{Numeric. A positive number equal to or less than 1 specifying the level of
confidence interval. Default = 0.95.}

\item{mc.cores}{Numeric. If > 1, the function will run in parallel
using mclapply using the number of cores specified. If = 1 then it will be run using a serial
loop. NOTE: Windows machines must use a value of 1 (see ?parallel::mclapply). Additionally,
be aware that each thread can use large chunks of memory.
Default = 1.}
}
\value{
Returns a list containing two tibbles. The first is a tibble that concatenates the outputs
from the basic data and rare species information in columns per input variable (column). The second is a tibble
that concatenates the various species richness estimates, with input variables in chunks of rows.
Additionally a console output will list the variables (columns) that lacked sufficient data
to be analysed.
}
\description{
ChaoSpecies: Estimation of species richness in a single community based on five types of data:
Type (1) abundance data (datatype="abundance"), Type (1A) abundance-frequency counts
(datatype="abundance_freq_count"), Type (2) incidence-frequency data
(datatype = "incidence_freq"), Type (2A) incidence-frequency counts
(datatype="incidence_freq_count"), and Type (2B) incidence-raw data (datatype="incidence_raw");
see SpadeR-package details for data input formats.
}
\examples{
\dontrun{
  # Read in some example data and use [BeeBDC::richnessPrepR()] to create the example input data
 #' data(beesCountrySubset)

estimateDataExample <- BeeBDC::richnessPrepR(
  data = beesCountrySubset,
  # Download the taxonomy
  taxonomyFile = BeeBDC::beesTaxonomy(),
  # Download the checklist
  checklistFile = BeeBDC::beesChecklist(),
  curveFunction = function(x) (228.7531 * x * x^-log(12.1593)),
  sampleSize = 10000,
  countryColumn = "country_suggested",
  limitGlobal = NULL,
  outPath = tempdir()
)

  # Transform the data for input
inputTestData <- estimateDataExample$site_speciesCounts \%>\%
  dplyr::select(scientificName, country_suggested, n) \%>\%
  tidyr::pivot_wider(names_from = country_suggested,
                     values_from = n,
                     values_fill = 0) \%>\%
  # Create the rownames
tibble::column_to_rownames("scientificName") \%>\%
  dplyr::tibble()

iChaoOut <- ChaoWrapper(
data = inputTestData,
datatype = "abundance",
k = 10,
conf = 0.95,
mc.cores = 1)
}
}
\seealso{
\code{\link[SpadeR:ChaoSpecies]{SpadeR::ChaoSpecies()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/GBIFissues.R
\name{GBIFissues}
\alias{GBIFissues}
\title{Flags records with GBIF issues}
\usage{
GBIFissues(data = NULL, issueColumn = "issue", GBIFflags = NULL)
}
\arguments{
\item{data}{A data frame or tibble. Occurrence records as input.}

\item{issueColumn}{Character. The column in which to look for GBIF issues. Default = "issue".}

\item{GBIFflags}{Character vector. The GBIF issues to flag. Users may choose their own vector of issues to flag or
use a pre-set vector or vectors, including c("allDates", "allMetadata", "allObservations",
"allSpatial", "allTaxo", or "all").

Default = c("COORDINATE_INVALID", "PRESUMED_NEGATED_LONGITUDE", "PRESUMED_NEGATED_LATITUDE", "COUNTRY_COORDINATE_MISMATCH", "ZERO_COORDINATE")}
}
\value{
Returns the data with a new column, ".GBIFflags", where FALSE = records with any of the provided
GBIFflags.
}
\description{
This function will flag records which are subject to a user-specified vector of GBIF issues.
}
\examples{
# Import the example data
data(beesRaw)
# Run the function
beesRaw_Out <- GBIFissues(data = beesRaw, 
   issueColumn = "issue", 
   GBIFflags = c("COORDINATE_INVALID", "ZERO_COORDINATE")) 


}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/HarmoniseR.R
\name{harmoniseR}
\alias{harmoniseR}
\title{Harmonise taxonomy of occurrence data}
\usage{
harmoniseR(
  data = NULL,
  path = NULL,
  taxonomy = BeeBDC::beesTaxonomy(),
  speciesColumn = "scientificName",
  rm_names_clean = TRUE,
  checkVerbatim = FALSE,
  relaxAmbiguous = FALSE,
  matchHigherTaxonomy = TRUE,
  stepSize = 1e+06,
  mc.cores = 1
)
}
\arguments{
\item{data}{A data frame or tibble. Occurrence records as input.}

\item{path}{A directory as character. The path to a folder that the output can be saved.}

\item{taxonomy}{A data frame or tibble. The taxonomy file to use.
Default = \code{\link[=beesTaxonomy]{beesTaxonomy()}}; for other taxa see first \code{\link[=taxadbToBeeBDC]{taxadbToBeeBDC()}}.}

\item{speciesColumn}{Character. The name of the column containing species names. Default = "scientificName".}

\item{rm_names_clean}{Logical. If TRUE then the names_clean column will be removed at the end of
this function to help reduce confusion about this column later. Default = TRUE}

\item{checkVerbatim}{Logical. If TRUE then the verbatimScientificName will be checked as well
for species matches. This matching will ONLY be done after harmoniseR has failed for the other
name columns. NOTE: this column is \emph{not} first run through \code{bdc::bdc_clean_names}. Default = FALSE}

\item{relaxAmbiguous}{Logical. If TRUE, then ambiguous names will be returned as "TRUE" for
.invalidName. However, they will be listed as "ambiguouslyMatchedName"
under the "scientificNameAuthorship" column and the speciesColumn will NOT include the authority
information. Hence, the names can be used but in should remain clear in the data that they must
be used with caution. Default = FALSE.}

\item{matchHigherTaxonomy}{Logical. If TRUE, then BeeBDC will try to match higher taxonomies
(e.g., family names) using genera and such. This will work better if you have run the data through
\code{bdc::bdc_clean_names} beforehand. The function will update the taxonRank column, if it is
otherwise empty. Default = TRUE.}

\item{stepSize}{Numeric. The number of occurrences to process in each chunk. Default = 1000000.}

\item{mc.cores}{Numeric. If > 1, the function will run in parallel
using mclapply using the number of cores specified. If = 1 then it will be run using a serial
loop. NOTE: Windows machines must use a value of 1 (see ?parallel::mclapply). Additionally,
be aware that each thread can use large chunks of memory.
Default = 1.}
}
\value{
The occurrences are returned with update taxonomy columns, including: scientificName,
species, family, subfamily, genus, subgenus, specificEpithet, infraspecificEpithet, and
scientificNameAuthorship. A new column, .invalidName, is also added and is FALSE when the occurrence's
name did not match the supplied taxonomy.
}
\description{
Uses BeeBDC-formatted taxonomy data to harmonise occurrence records and flag those that do not match
the taxonomy \code{\link[=harmoniseR]{harmoniseR()}} prefers to use the names_clean columns that is generated
by \code{\link[bdc:bdc_clean_names]{bdc::bdc_clean_names()}}. While this is not required, you may find better results by running
that function on your dataset first.
It is possible to download taxonomy file for other taxa using \code{\link[=taxadbToBeeBDC]{taxadbToBeeBDC()}}
which can download taxonomies from ITIS, GBIF, and more. You could also match the format of the
\code{\link[=beesTaxonomy]{beesTaxonomy()}} file.
}
\examples{
# load in the test dataset
system.file("extdata", "testTaxonomy.rda", package="BeeBDC") |> load()

# See also
?BeeBDC::taxadbToBeeBDC()

beesRaw_out <- BeeBDC::harmoniseR(
  #The path to a folder that the output can be saved
path = tempdir(),
# The formatted taxonomy file
taxonomy = testTaxonomy, 
data = BeeBDC::beesFlagged,
speciesColumn = "scientificName")
table(beesRaw_out$.invalidName, useNA = "always")
}
\seealso{
\code{\link[=taxadbToBeeBDC]{taxadbToBeeBDC()}} to download any taxonomy (of any taxa or of bees) and
\code{\link[=beesTaxonomy]{beesTaxonomy()}} for the bee taxonomy download.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/dateFindR.R
\name{dateFindR}
\alias{dateFindR}
\title{Find dates in other columns}
\usage{
dateFindR(data = NULL, maxYear = lubridate::year(Sys.Date()), minYear = 1700)
}
\arguments{
\item{data}{A data frame or tibble. Occurrence records as input.}

\item{maxYear}{Numeric. The maximum year considered reasonable to find.
Default = lubridate::year(Sys.Date()).}

\item{minYear}{Numeric. The minimum year considered reasonable to find. Default = 1700.}
}
\value{
The function results in the input occurrence data with but with updated eventDate, year,
month, and day columns for occurrences where these data were a) missing and b) located in one of the
searched columns.
}
\description{
A function made to search other columns for dates and add them to the eventDate column.
The function searches the columns locality, fieldNotes, locationRemarks, and verbatimEventDate
for the relevant information. Additionally, for date ranges in the eventDate column, these will
Be translated into startDayOfYear and endDayOfYear and the eventDate will be kept as the last
date, but formatted in the correct format. Ambiguous dates that can't be parsed reliably as
day-month-year or month-day-year will be rounded to the nearest certainty; if month or day can
be identified with a spelled-out month or a day >12, these wil be correctly formatted.
Year-month-day... is the most-reliable format
}
\examples{
# Using the example dataset, you may not find any missing eventDates are rescued (dependent on 
# which version of the example dataset the user inputs.
beesRaw_out <- dateFindR(data = beesRaw,
                         # Years above this are removed (from the recovered dates only)
                         maxYear = lubridate::year(Sys.Date()),
                         # Years below this are removed (from the recovered dates only)
                         minYear = 1700)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/interactiveMapR.R
\name{interactiveMapR}
\alias{interactiveMapR}
\title{Creates interactive html maps for species}
\usage{
interactiveMapR(
  data = NULL,
  outPath = NULL,
  lon = "decimalLongitude",
  lat = "decimalLatitude",
  speciesColumn = "scientificName",
  speciesList = "ALL",
  countryList = NULL,
  jitterValue = NULL,
  onlySummary = TRUE,
  overWrite = TRUE,
  customColumn1 = NULL,
  customColumn2 = NULL,
  TrueAlwaysTop = FALSE,
  excludeSpecies = c("Apis mellifera Linnaeus, 1758"),
  pointColours = c("blue", "darkred", "#ff7f00", "black"),
  returnPlot = FALSE
)
}
\arguments{
\item{data}{A data frame or tibble. Occurrence records to use as input.}

\item{outPath}{A directory as character. Directory where to save output maps.}

\item{lon}{Character. The name of the longitude column. Default = "decimalLongitude".}

\item{lat}{Character. The name of the latitude column. Default = "decimalLatitude".}

\item{speciesColumn}{Character. The name of the column containing species names (or another factor)
to build individual maps from. Default = "scientificName".}

\item{speciesList}{A character vector. Should contain species names as they appear in the
speciesColumn to make maps of. User can also specify "ALL" in order to make maps of all
species present in the data. Hence, a user may first filter their data and then use "ALL".
Default = "ALL".}

\item{countryList}{A character vector. Country names to map, or NULL for to map ALL countries.}

\item{jitterValue}{Numeric. The amount, in decimal degrees, to jitter the map points by - this
is important for separating stacked points with the same coordinates.}

\item{onlySummary}{Logical. If TRUE, the function will not look to plot country or
expert-identified outliers in different colours.}

\item{overWrite}{Logical. If TRUE, the function will overwrite existing files in the provided
directory that have the same name.
Default = TRUE.}

\item{customColumn1}{Character. Allows the user to report on a column of their choosing in the output.}

\item{customColumn2}{Character. Allows the user to report on a column of their choosing in the output.}

\item{TrueAlwaysTop}{If TRUE, the quality (TRUE) points will always be displayed on top of other points.
If FALSE, then whichever layer was turned on most-recently will be displayed on top.}

\item{excludeSpecies}{Character. A character vector of species names to exclude, especially
if they are so large as to become a problem when making maps. For bee data, for example,
"Apis mellifera Linnaeus, 1758" has too many data points and is default excluded.}

\item{pointColours}{A character vector of colours. In order provide colour for TRUE, FALSE, countryOutlier, and customOutlier.
Default = c("blue", "darkred","#ff7f00", "black").}

\item{returnPlot}{Logical. If TRUE, return the plot to the environment. Default = FALSE.}
}
\value{
Exports .html interactive maps of bee occurrences to the specified directory.
}
\description{
Uses the occurrence data (preferably uncleaned) and outputs interactive .html maps that can be opened
in your browser to a specific directory. The maps can highlight if an occurrence has passed all filtering
(.summary == TRUE) or failed at least one filter (.summary == FALSE). This can be modified by first running
\code{\link[=summaryFun]{summaryFun()}} to set the columns that you want to be highlighted. It can also highlight occurrences
flagged as expert-identified or country outliers.
}
\examples{
if(require("htmlwidgets") && require("leaflet")){

OutPath_Figures <- tempdir()

interactiveMapR(
# occurrence data - start with entire dataset, filter down to these species
data = BeeBDC::bees3sp, # \%>\%
  # Select only those species in the 100 randomly chosen
  # dplyr::filter(scientificName \%in\% beeData_interactive$scientificName),
  # Select only one species to map
  # dplyr::filter(scientificName \%in\% "Agapostemon sericeus (Forster, 1771)"),
# Directory where to save files
outPath = paste0(OutPath_Figures, "/interactiveMaps_TEST"),
# lat long columns
lon = "decimalLongitude",
lat = "decimalLatitude",
# Occurrence dataset column with species names
speciesColumn = "scientificName",
# Which species to map - a character vector of names or "ALL"
# Note: "ALL" is defined AFTER filtering for country
speciesList = "ALL",
# studyArea
countryList = NULL, 
# Point jitter to see stacked points - jitters an amount in decimal degrees
jitterValue = 0.01,
# If TRUE, it will only map the .summary column. Otherwise, it will map .summary
# which will be over-written by countryOutliers and manualOutliers
onlySummary = TRUE,
excludeSpecies = c("Apis mellifera Linnaeus, 1758"),
overWrite = TRUE,
  # Colours for points which are flagged as TRUE, FALSE, countryOutlier, and customOutlier
pointColours = c("blue", "darkred","#ff7f00", "black")
)

} # END if require
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/beesTaxonomy.R
\name{beesTaxonomy}
\alias{beesTaxonomy}
\title{Download a nearly complete taxonomy of bees globally}
\usage{
beesTaxonomy(
  URL = "https://api.figshare.com/v2/file/download/64193731",
  mode = NULL,
  headers = NULL,
  token = NULL,
  alternateURL = FALSE,
  ...
)
}
\arguments{
\item{URL}{A character vector to the FigShare location of the dataset. The default will be to
the most-recent version.}

\item{mode}{A character passed on to \code{utils::download.file()}. Default = "wb" for Windows or "w" for Mac/Linux.}

\item{headers}{Character. Passed on to  \code{utils::download.file()}. Default = NULL.}

\item{token}{Character. Optional personal access token from FigShare account for authorisation.
Default = NULL.}

\item{alternateURL}{Logical. If TRUE then the function will use the an alernate version of the
download URL. Might be worth trying if it is failing. Default = FALSE.}

\item{...}{Extra variables that can be passed to \code{downloader::download()}.}
}
\value{
A downloaded beesTaxonomy.Rda file in the \code{\link[=tempdir]{tempdir()}} and the same tibble returned to
the environment.
}
\description{
Downloads the taxonomic information for the bees of the world.
Source of taxonomy is listed under "source" but are mostly derived from the Discover Life
website. The data will be sourced from the BeeBDC article's Figshare.

Note that sometimes the download might not work without restarting R. In this case, you could
alternatively download the dataset from the URL below and then read it in using
\code{base::readRDS("filePath.Rda")}. Note that as of version 1.3.2, this function internally uses
a modified version of the "download" function from the
\code{downloader} package on CRAN. Additionally, for Windows machines, BeeBDC will try a different
download method with each failed attempt (for Windows: "auto" > "internal" > "libcurl" > "wget" > "curl";
or "auto" if any particular method is not available).
}
\details{
\strong{Column details}

\strong{flags} Flags or comments about the taxon name.

\strong{taxonomic_status} Taxonomic status. Values are "accepted" or "synonym"

\strong{source} Source of the name.

\strong{accid} The id of the accepted taxon name or "0" if taxonomic_status == accepted.

\strong{id} The id number for the taxon name.

\strong{kingdom} The biological kingdom the taxon belongs to. For bees, kingdom == Animalia.

\strong{phylum} The biological phylum the taxon belongs to. For bees, phylum == Arthropoda.

\strong{class} The biological class the taxon belongs to. For bees, class == Insecta.

\strong{order} The biological order the taxon belongs to. For bees, order == Hymenoptera.

\strong{family} The family of bee which the species belongs to.

\strong{subfamily} The subfamily of bee which the species belongs to.

\strong{tribe} The tribe of bee which the species belongs to.

\strong{subtribe} The subtribe of bee which the species belongs to.

\strong{validName} The valid scientific name as it should occur in the “scientificName” column in a Darwin Core file.

\strong{canonical} The scientificName without the scientificNameAuthority.

\strong{canonical_withFlags} The scientificName without the scientificNameAuthority and with Discover Life taxonomy flags.

\strong{genus} The genus the bee species belongs to.

\strong{subgenus} The subgenus the bee species belongs to.

\strong{species} The specific epithet for the bee species.

\strong{infraspecies} The infraspecific epithet for the bee addressed.

\strong{authorship} The author who described the bee species.

\strong{taxon_rank} Rank for the bee taxon addressed in the entry.

\strong{notes} Additional notes about the name/taxon.

\strong{Previous taxonomies:}
\itemize{
\item 2026-05-04 \strong{current}: https://api.figshare.com/v2/file/download/64193731
\item 2026-01-12: https://open.flinders.edu.au/ndownloader/files/60945820
\item 2024-06-17: https://open.flinders.edu.au/ndownloader/files/47089969
\item 2023-11-29: https://open.flinders.edu.au/ndownloader/files/43331472
\item 2023-10-10: https://open.flinders.edu.au/ndownloader/files/42613126
\item 2023-09-20: https://open.flinders.edu.au/ndownloader/files/42402264
\item Original: https://open.flinders.edu.au/ndownloader/files/42320595
}
}
\examples{
\dontrun{
beesTaxonomy <- BeeBDC::beesTaxonomy()
}


}
\references{
This dataset was created using the Discover Life taxonomy.
Dataset is from the publication:
DOREY, J. B., CHESSHIRE, P. R., BOLAÑOS, A. N., O’REILLY, R. L., BOSSERT, S., COLLINS, S. M., LICHTENBERG, E. M., TUCKER, E., SMITH-PARDO, A., FALCON-BRINDIS, A., GUEVARA, D. A., RIBEIRO, B. R., DE PEDRO, D., FISCHER, E., HUNG, J. K.-L., PARYS, K. A., ROGAN, M. S., MINCKLEY, R. L., VELZCO, S. J. E., GRISWOLD, T., ZARRILLO, T. A., SICA, Y., ORR, M. C., GUZMAN, L. M., ASCHER, J., HUGHES, A. C. & COBB, N. S. In review. A globally synthesised and flagged bee occurrence dataset and cleaning workflow. Scientific Data.
The taxonomy data are mostly compiled from Discover Life data, www.discoverlife.org:
ASCHER, J. S. & PICKERING, J. 2020. Discover Life bee species guide and world checklist (Hymenoptera: Apoidea: Anthophila). http://www.discoverlife.org/mp/20q?guide=Apoidea_species.
}
\seealso{
\code{\link[=taxadbToBeeBDC]{taxadbToBeeBDC()}} to download any other taxonomy (of any taxa or of bees)
and \code{\link[=harmoniseR]{harmoniseR()}} for the
taxon-cleaning function where these taxonomies are implemented.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/repoFinder.R
\name{repoFinder}
\alias{repoFinder}
\title{Find GBIF, ALA, iDigBio, and SCAN files in a directory}
\usage{
repoFinder(path)
}
\arguments{
\item{path}{A directory as character. The path within which to recursively look for GBIF, ALA,
iDigBio, and SCAN files.}
}
\value{
Returns a list of directories to each of the above data downloads
}
\description{
Find GBIF, ALA, iDigBio, and SCAN files in a directory
}
\examples{
\dontrun{
# Where DataPath is made by [BeeBDC::dirMaker()]
BeeBDC::repoFinder(path = DataPath)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/formattedCombiner.R
\name{formattedCombiner}
\alias{formattedCombiner}
\title{Combine the formatted USGS data with the main dataset}
\usage{
formattedCombiner(path, strings, existingOccurrences, existingEMLs)
}
\arguments{
\item{path}{A directory as character. The directory to look in for the formatted USGS data.}

\item{strings}{A regex string. The string to find the most-recent formatted USGS dataset.}

\item{existingOccurrences}{A data frame. The existing occurrence dataset.}

\item{existingEMLs}{An EML file. The existing EML data file to be appended.}
}
\value{
A list with the combined occurrence dataset and the updated EML file.
}
\description{
Merges the Darwin Core version of the USGS dataset that was created using \code{\link[=USGS_formatter]{USGS_formatter()}}
with the main dataset.
}
\examples{
\dontrun{
DataPath <- tempdir()
strings = c("USGS_DRO_flat_27-Apr-2022")
    # Combine the USGS data and the existing big dataset
Complete_data <- formattedCombiner(path = DataPath, 
                                    strings = strings, 
                                    # This should be the list-format with eml attached
                                    existingOccurrences = DataImp$Data_WebDL,
                                    existingEMLs = DataImp$eml_files) 
                                    }
}
